// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 * chanctx scheduler
 */
#include <linux/platform_device.h>
#include <linux/bitfield.h>
#include <net/mac80211.h>

#define func_fmt(str)		"[CHAN] "str
#define func_trace              CHAN

#include "core.h"
#include "phy.h"
#include "util.h"
#include "trace.h"
#include "debug.h"
#ifdef SCM80211_DFS_CERTIFIED
#include "dfs.h"
#endif

/**
 * DOC: Channel context scheduler
 *
 * struct ieee80211_chanctx:
 * - Represent a channel parameters (band, center frequency, bandwidth, and etc).
 * - All virtual interfaces whose "operating channels" are compatible are "assigned"
 *   to one channel context.
 * - If HW does not support multi-band/channel operation , only one channel context
 *   can be active at any moment and all the other channel contexts are inactive.
 * - In turn, only virtual interfaces associated with the active channel context can
 *   access WM and can communicate with their peer stations.
 * - Created/assigned/changed/unassigned/changed/removed by mac80211.
 * - Mainly used to implement CSA. mac80211 does not implement channel scheduler.
 *
 * struct scm80211_chanctx:
 * - Extends struct ieee80211_chanctx to include other PHY parameters (struct phy_vector).
 * - These parameters include for example, VHT/HE PHY frame filter and MU setting.
 * - Each virtual interface maintains it own PHY parameters (struct ieee80211_vif_priv::pv)
 * - The PHY parameters of a channel context is a collection of all virtual interfaces
 *   associated with it, and preferably meet individual virtual interface's PHY setting
 * - These PHY parameters take effect at once when the bearing channel context becomes
 *   active.
 * - The driver updates virtual interface's pv. Actual PHY reconfiguration is done
 *   when the associated channel context becomes active.
 *
 * Channel context scheduler:
 * - Channel context scheduler autonomously switches channel context and configures
 *   PHY/RF accordingly.
 * - Any function wishing to reserve channel time presents its channel context(s) to
 *   the scheduler. This includes scanning, ROC, and TDSL channel switching and the
 *   above mentioned mac80211-managed channel contexts.
 * - All common tasks involved in switching channel, i.e., putting station in ps mode
 *   if it is leaving its operating channel, and vice vera, are done by the channel
 *   scheduer. Clients only need to specify its wish and add to the scheduler.
 * - When it swiches channel context from c(i) to c(j), it calls c(i)->end() and
 *   c(j)->begin() callback functions.
 *
 * TDMA channel context:
 * - Represents base (operating) channel of virtual interface
 * - Created/changed/removed/assigned/unaissnged by mac80211
 * - Mainly used for CSA (swiching from one channel to another)
 * - We allow more than one TDMA channel context. The channel context scheduler
 *   serves each TDMA channel one at a time in a round-robin fashion.
 *
 * SCAN channel context:
 * - Mananged by the driver to perform scanning.
 * - @restart:
 *
 * ROC channel context:
 *
 * Channel swiching is done by the channel context scheduler
 *
 * FIXME:
 * - Channel forcing for AP as before
 * - Block tx() if the target channel is not active
 * - tdls_xxx()
 * - Add debug logs or trace
 * - filter after making sure AP puts correct GID63 (rx.c)
 */

#define SCM80211_MAX_CHANCTX		(2 + 2)
#define SCM80211_CHANCTX_QUANTUM	25	/* ms */

#define PASSIVE_SCAN_DWELL		120
#define ACTIVE_SCAN_DWELL		60
#define SCM80211_HWSIM_DWELL_EXTEND	1

/**
 * set as (beacon_loss_count -1) defined in mlme.c
 * avoid beacon loss because of turning to tdls channel
 */
#define TDLS_TIMEOUT_COUNT	6

static bool experimental;

/**
 * Interface combination
 *
 * Combination 0: no concurrency
 * - #{STA, P2P-Client, P2P-GO, AP} = 1, on 1 channels, 1 total
 *
 * Combination 1: 1 channel
 * - #{STA, P2P-Client, P2P-GO, AP} = 1, on 1 channels, 2 total
 * - If AP starts first, and STA later wants a different channel, STA fails
 * - If STA starts first, and AP later wants a different channel, AP fails
 *
 * Combination 2: 2 channel
 * - #{STA} = 1, #{P2P-Client, P2P-GO, AP} = 1 on 2 channels, 2 total
 *
 * Combination 3: simultaneous multi-band (1 wiphy/2 hw mac)
 * - #{STA} = 2, #{P2P-Client, P2P-GO, AP} = 1 on 2 channels, 3 total
 *
 * WDS, MESH-POINT: TBD
 *
 *
 * See https://docs.microsfot.com/en-us/windows-hardware/drivers/network/wdi-objects
 */
static const struct ieee80211_iface_limit scm80211_iface_limits[] = {
	{
	 .max = 2,
	 .types = BIT(NL80211_IFTYPE_STATION) | BIT(NL80211_IFTYPE_P2P_CLIENT),
	  },
	{
	 .max = 2,
	 .types = BIT(NL80211_IFTYPE_AP) | BIT(NL80211_IFTYPE_P2P_GO),
	  },
#ifdef SCM80211_P2P_DEVICE
	{
	 .max = 1,
	 .types = BIT(NL80211_IFTYPE_P2P_DEVICE),
	  },
#endif
#ifdef CONFIG_MAC80211_MESH
	{
	 .max = 1,
	 .types = BIT(NL80211_IFTYPE_MESH_POINT),
	  },
#endif
};

static const struct ieee80211_iface_limit scm80211_iface5g_limits[] = {
	{
	 .max = 1,
	 .types = BIT(NL80211_IFTYPE_STATION),
	  },
	{
	 .max = 2,
	 .types = BIT(NL80211_IFTYPE_AP),
	}
};

static const struct ieee80211_iface_combination scm80211_iface_combination[] = {
	[NL80211_BAND_2GHZ] = {
			       .limits = scm80211_iface_limits,
			       .n_limits = ARRAY_SIZE(scm80211_iface_limits),
#ifdef SCM80211_P2P_DEVICE
			       .max_interfaces = 3,
#else
			       .max_interfaces = 2,
#endif
			       .num_different_channels = 1,
#ifdef SCM80211_DFS_CERTIFIED
			       .radar_detect_widths = BIT(NL80211_CHAN_WIDTH_20_NOHT) |
			       BIT(NL80211_CHAN_WIDTH_20) |
			       BIT(NL80211_CHAN_WIDTH_40) | BIT(NL80211_CHAN_WIDTH_80),
#endif
				},
	/* for 5G only */
	[NL80211_BAND_5GHZ] = {
			       .limits = scm80211_iface5g_limits,
			       .n_limits = ARRAY_SIZE(scm80211_iface5g_limits),
			       .max_interfaces = 2,
			       .num_different_channels = 1,
#ifdef SCM80211_DFS_CERTIFIED
			       .radar_detect_widths = BIT(NL80211_CHAN_WIDTH_20_NOHT) |
			       BIT(NL80211_CHAN_WIDTH_20) |
			       BIT(NL80211_CHAN_WIDTH_40) | BIT(NL80211_CHAN_WIDTH_80),
#endif
				}

};

struct scm80211_chanctx_sched *hw_sched(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return &priv->sched;
}

static struct scm80211_scanner *hw_scanner(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return &priv->scanner;
}

struct scm80211_chanctx *scm80211_current_chanctx(struct ieee80211_hw *hw)
{
	return READ_ONCE(hw_sched(hw)->cur_chan);
}

static inline bool scm80211_chanctx_is_active(struct ieee80211_hw *hw,
					      struct scm80211_chanctx *chanctx)
{
	return (scm80211_current_chanctx(hw) == chanctx);
}

static inline bool scm80211_chanctx_compat_current(struct ieee80211_hw *hw,
						   struct scm80211_chanctx *chanctx)
{
	return !!cfg80211_chandef_compatible(&chanctx->ctx.def,
					     &scm80211_current_chanctx(hw)->ctx.def);
}

bool scm80211_is_scanning(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return hw_scanner(hw)->scanning || priv->roc.running;
}

/**
 * There is a condition that the beacon lossed during scan
 * but bcn loss check work start at the end of scan
 * which cause driver take it as beacon loss at normal condition
 * and do disconnection.
 * Add scanned member to make correct judgment
 **/
bool scm80211_scan_finished(struct ieee80211_hw *hw)
{
	bool ret = false;

	if (hw_scanner(hw)->scanned)
		ret = true;
	hw_scanner(hw)->scanned = false;

	return ret;
}

bool scm80211_is_offchannel(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct scm80211_chanctx *chanctx = NULL;

	if (vif)
		chanctx = (void *)rcu_dereference_chanctx_conf(vif);

	if (!chanctx)
		return false;

	return !scm80211_chanctx_compat_current(hw, chanctx);
}

static void scm80211_enqueue_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx);
static void scm80211_dequeue_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx);
static void scm80211_tdls_work(struct work_struct *work);
static enum hrtimer_restart scm80211_tdls(struct hrtimer *timer);

int scm80211_send_nullfunc(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			   struct cfg80211_chan_def *def, bool ps)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct sk_buff *skb;
	struct ieee80211_hdr *nullfunc;
	struct ieee80211_tx_info *info;

	scm80211_trace(hw, "%s: (type: %s , ps %s)\n", __func__,
		       ieee80211_iftype_str[vif->type], ps ? "true" : "false");

	skb = ieee80211_nullfunc_get(hw, vif, vif->bss_conf.qos);
	if (!skb)
		return -ENOMEM;

	nullfunc = (struct ieee80211_hdr *)skb->data;
	nullfunc->seq_ctrl = cpu_to_le16(vifp->null_seqnum++ << 4) & IEEE80211_SCTL_SEQ;
	if (ps)
		nullfunc->frame_control |= cpu_to_le16(IEEE80211_FCTL_PM);
	else
		nullfunc->frame_control &= ~cpu_to_le16(IEEE80211_FCTL_PM);

	info = IEEE80211_SKB_CB(skb);
	info->flags |= IEEE80211_TX_CTL_USE_MINRATE;
	info->band = def->chan->band;
	info->control.vif = vif;
	info->control.rates[0].idx = 0;
	info->control.rates[0].flags = 0;
	info->control.rates[0].count = 5;

	scm80211_tx_channel(hw, skb, def->chan);

	return 0;
}

struct chanctx_switch_vote {
	bool dryrun;
	struct scm80211_chanctx *from;
	struct scm80211_chanctx *to;
	int veto;
	bool sleep_needed;
};

/**
 * scm80211_stop_vif() - stop virtual interfaces going off-channel
 * @data: current channel wiphy is leaving from
 * @mac: mac address of @pubvif
 * @pubvif: virtual interface
 *
 * This function is called over all active interface before channel context switch.
 * All housekeeping, such as STA sending a QoS-Null to its AP, should be done here.
 *
 */
static void scm80211_stop_vif(void *data, u8 *mac, struct ieee80211_vif *pubvif)
{
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(pubvif);
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(pubvif);
	enum nl80211_iftype iftype_p2p = ieee80211_vif_type_p2p(pubvif);
	struct chanctx_switch_vote *vote = data;
	u32 queues = scm80211_vif_queues(pubvif);

	if (get_ieee80211_chanctx_conf(pubvif) != &vote->from->ctx)
		return;		/* not in the same chanctx */

	scm80211_trace(hw, "+%s(%s) %s\n", __func__,
		       ieee80211_vif_name(pubvif), vote->dryrun ? "dryrun" : "");

	switch (iftype_p2p) {
	case NL80211_IFTYPE_STATION:
	case NL80211_IFTYPE_P2P_CLIENT:	/* FIXME: Check again */
		{
			if (!pubvif->bss_conf.assoc)
				break;

			if (!vote->dryrun) {
				/*
				 * flush here to make sure the null packet
				 * is the last tx packet before off channel
				 */
				if (hw->conf.flags & IEEE80211_CONF_PS)
					hw->conf.flags &= ~IEEE80211_CONF_PS;
				scm80211_disable_ps(hw);

				queues &= ~BIT(vifp->mgmt_hw_queue);
				 /* do not stop offchannel_txq */
				queues &= ~BIT(hw->offchannel_tx_hw_queue);
				scm80211_stop_queues(hw, queues);

				scm80211_flush(hw, pubvif, queues, false);

				scm80211_send_nullfunc(hw, pubvif, &vote->from->ctx.def, true);
				scm80211_flush(hw, pubvif, BIT(vifp->mgmt_hw_queue), false);
			}
			goto out;
		}
	case NL80211_IFTYPE_AP:
	case NL80211_IFTYPE_AP_VLAN:
		{
			if (!pubvif->bss_conf.enable_beacon)
				break;

			/* NB: Can AP switch to vote->from to vote->to? */
			if (!hw_scanner(hw)->scanning) {
				scm80211_warn(hw, "AP %pM declines channel switch\n", mac);
				vote->veto++;
				return;
			}
			break;
		}
	case NL80211_IFTYPE_P2P_GO:
		{
			if (!pubvif->bss_conf.enable_beacon)
				break;
			/* FIXME: NOA */
			if (vote->dryrun)
				scm80211_warn(hw, "Offchannel operation on GO, NOA not supported!\n");
			break;
		}
	default:
		break;
	}

	if (!vote->dryrun) {
		/* STOP TX */
		queues &= ~BIT(vifp->mgmt_hw_queue);
		queues &= ~BIT(hw->offchannel_tx_hw_queue);	/* do not stop offchannel_txq */
		scm80211_stop_queues(hw, queues);

		scm80211_flush(hw, pubvif, queues, false);
	}

out:
	scm80211_trace(hw, "-%s(%s)\n", __func__, ieee80211_vif_name(pubvif));
}

/**
 * scm80211_resume_vif() - resume virtual interface on current channel
 */
static void scm80211_resume_vif(void *data, u8 *mac, struct ieee80211_vif *pubvif)
{
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(pubvif);
	struct scm80211_chanctx *chan = data;
	enum nl80211_iftype iftype_p2p = ieee80211_vif_type_p2p(pubvif);
	u32 queues = scm80211_vif_queues(pubvif);

	if (get_ieee80211_chanctx_conf(pubvif) &&
	    get_ieee80211_chanctx_conf(pubvif) != &chan->ctx)
		return;

	scm80211_trace(hw, "+%s(%s)\n", __func__, ieee80211_vif_name(pubvif));

	switch (iftype_p2p) {
	case NL80211_IFTYPE_STATION:
	case NL80211_IFTYPE_P2P_CLIENT:
		if (pubvif->bss_conf.assoc)
			scm80211_send_nullfunc(hw, pubvif, &chan->ctx.def, false);
		break;
	default:
		break;
	}
	scm80211_wake_queues(hw, queues);

	scm80211_trace(hw, "-%s(%s)\n", __func__, ieee80211_vif_name(pubvif));
}

/**
 * add chanctx to chanctx scheduler
 */
static void scm80211_enqueue_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	enum scm80211_chanctx_type type = chanctx->type;

	lockdep_assert_held(&sched->mutex);

	chanctx->air_time = 0;

	if (list_empty(&chanctx->sched_list)) {
		list_add_tail(&chanctx->sched_list, &sched->queue[type].head);
		atomic_inc(&sched->queue[type].nr_chanctx);
		atomic_inc(&sched->nr_chanctx);
		chanctx->offchan = true;
	} else {
		scm80211_info(hw, "enqueue: chanctx %d already started?\n", chanctx->id);
	}
}

/**
 * remove chanctx from chanctx scheduler
 * sched->lock held.
 */
static void scm80211_dequeue_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	enum scm80211_chanctx_type type = chanctx->type;

	lockdep_assert_held(&sched->mutex);

	if (!list_empty(&chanctx->sched_list)) {
		list_del_init(&chanctx->sched_list);
		atomic_dec(&sched->queue[type].nr_chanctx);
		atomic_dec(&sched->nr_chanctx);
	}
}

/**
 * scm80211_next_chanctx() - select next channel to servce
 *
 * Do a simple search from the higher priority channel queue. The scheduler
 * moves the selected channel to the last of the queue it belongs to
 * avoid the same chan from keeping being selected.
 *
 * TODO:
 * Implement more elaborated channel scheduling algorithm
 */
static struct scm80211_chanctx *scm80211_next_chanctx(struct ieee80211_hw *hw)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *chanctx = NULL;
	struct list_head *head = NULL;
	int i;

	lockdep_assert_held(&sched->mutex);

	chanctx = scm80211_current_chanctx(hw);
	head = &sched->queue[chanctx->type].head;
	if (!list_empty(&chanctx->sched_list) && !list_is_last(&chanctx->sched_list, head))
		list_move_tail(&chanctx->sched_list, head);

	for (i = ARRAY_SIZE(sched->queue) - 1; i >= 0; i--) {
		head = &sched->queue[i].head;
		chanctx = list_first_entry_or_null(head, struct scm80211_chanctx, sched_list);
		if (!chanctx)
			continue;
		return chanctx;
	}
	return &sched->idle;
}

/**
 * scm80211_schedule_chanctx() - peform channel context switch
 * @hw: &struct ieee80211_hw
 * @preempt: chanctx list changed
 *
 */
static void scm80211_schedule_chanctx(struct ieee80211_hw *hw, bool preempt)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *old, *new;
	struct ieee80211_channel *chan;
	long dwell;
	struct chanctx_switch_vote vote = {
		.dryrun = true,
		.veto = 0,
	};

	lockdep_assert_held(&sched->mutex);

	old = scm80211_current_chanctx(hw);
	old->air_time += jiffies_to_msecs(jiffies - old->last_start_jiffies);
	if (old->air_time >= old->dwell_time)
		old->air_time = 0;

	chan = old->ctx.def.chan;

	if (old->end)
		old->end(hw, old);

	if (!old->restart)
		scm80211_dequeue_chanctx(hw, old);

	new = scm80211_next_chanctx(hw);

	vote.from = old;
	vote.to = new;

	if (old == new && chan == new->ctx.def.chan) {
		scm80211_trace(hw, "dont need to go to offchannel\n");
		if (new->begin)
			new->begin(hw, new);
		goto out;
	}

	/* First vote about going off-channel */
	ieee80211_iterate_active_interfaces(hw, IEEE80211_IFACE_ITER_NORMAL,
					    scm80211_stop_vif, &vote);
	if (vote.veto > 0) {
		/*
		 * What shell we do? - just keep using this channel
		 * hoping this is temporary.
		 */
		scm80211_warn(hw, "chanctx switching %d->%d rejected\n", old->id, new->id);

		if (experimental) {
			new = old;
			goto out;
		}
	}
	/* Good bye to off-channel-to-be interfaces */
	old->offchan = true;
	vote.dryrun = false;
	ieee80211_iterate_active_interfaces(hw, IEEE80211_IFACE_ITER_NORMAL,
					    scm80211_stop_vif, &vote);

	/* Actual switch - no way back */
	scm80211_trace(hw, "switching chanctx: %d --> %d\n", old->id, new->id);
	trace_scm80211_switch_chanctx(hw, new);

	if (new != &sched->idle)
		scm80211_set_chanctx(hw, new);

	new->offchan = false;
	sched->cur_chan = new;
	new->last_start_jiffies = jiffies;
	if (new->begin)
		new->begin(hw, new);

	/* Hello to now on-channel interfaces on @new */
	ieee80211_iterate_active_interfaces(hw, IEEE80211_IFACE_ITER_NORMAL,
					    scm80211_resume_vif, new);
out:

	/* Schedule next chanctx switch */
	if (new->type == CHANCTX_TYPE_BEACON) {
		list_del_init(&new->sched_list);
		new->type = CHANCTX_TYPE_TDMA;
		list_add_tail(&new->sched_list, &sched->queue[new->type].head);
		queue_delayed_work(sched->workqueue, &sched->work, msecs_to_jiffies(50));
	} else if (new->dwell_time != -1ULL || new->type != CHANCTX_TYPE_IDLE) {
		dwell = new->dwell_time - new->air_time;
		queue_delayed_work(sched->workqueue, &sched->work, msecs_to_jiffies(dwell));
	}
}

static void scm80211_chan_work(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv = work_to_hw_priv(work, sched.work.work);
	struct ieee80211_hw *hw = priv->hw;
	struct scm80211_chanctx_sched *sched = &priv->sched;

	mutex_lock(&sched->mutex);

	scm80211_schedule_chanctx(hw, false);

	mutex_unlock(&sched->mutex);
}

/**
 * scm80211_start_chanctx() - start a chanctx for scheduling
 * @hw:
 * @chan: channel context to switch to
 *
 * Returns true if the request will be fullfiled immediately,
 * false if it is queued.
 *
 * This function should be called within a context that can sleep.
 *
 * Callers:
 * - assign_vif_chanctx(): can sleep
 * - hw_scan(): can sleep
 * - remain_on_channel(): can sleep
 * - tdls_channel_switch(): can sleep (?)
 *
 * Because all these caller functions are called from a user-space process and
 * not from the single threaded mac80211 workqueue, this function can
 * call cancel_delayed_work_sync() to cancel and flush channel scheduler work
 * which is running on mac80211 workqueue.
 *
 * In addition, because scheduler requeues chanctx itself, there will be no deadlock
 * as long as we release @sched->mutex before doing so.
 *
 */
bool scm80211_start_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *curchanctx;

	mutex_lock(&sched->mutex);

	trace_scm80211_start_chanctx(hw, chanctx);

	scm80211_enqueue_chanctx(hw, chanctx);

	curchanctx = scm80211_current_chanctx(hw);

	if (!delayed_work_pending(&sched->work)) {
		/* Channel scheduler is idle, start right away */

		scm80211_trace(hw, "chanctx %d kicking scheduler\n", chanctx->id);

		/*queue_delayed_work(sched->workqueue, &sched->work, 0); */
		scm80211_schedule_chanctx(hw, true);

		mutex_unlock(&sched->mutex);
		return true;
	}

	if (curchanctx->type < chanctx->type) {
		/*
		 * @chanctx is of higher priority than the current channel.
		 * Wanted to use mod_delayed_work(), but that requires
		 * &struct workqueue_struct.
		 *
		 * Getting it from sched->work->wq does not look good.
		 *
		 * Should I use our own workqueue? Really don't want to create
		 * a thread of our own.
		 *
		 * So I went a dirty way, i.e., stop and start the work
		 * scm28011_stop_chanctx_sched() sets @sched->stop to %true
		 * before cancel/flush work.
		 *
		 * Note that mutex is unlocked. This function can be also called
		 * from mac80211's workqueue. If it calls scm80211_stop_chanctx_sched()
		 * with mutex acquired:
		 *
		 * Calling process A                  mac80211 worker thread
		 * scm80211_start_chanctx()
		 * - mutex acquired
		 * - scm80211_stop_chanctx_sched()
		 *   - cancel_delayed_work_sync()  --> cancled or flushed()
		 *              ....
		 * - scm80211_start_chanctx_sched()
		 *
		 * Finally, to make my life easier, I decided to have a dedicated
		 * workqueue. scheduler runs in the worker thread of that workqueue.
		 *
		 */
		scm80211_trace(hw, "chanctx %d preempting scheduler\n", chanctx->id);
		mod_delayed_work(sched->workqueue, &sched->work, 0);
		mutex_unlock(&sched->mutex);
		return true;
	}

	if (curchanctx->type == CHANCTX_TYPE_TDMA &&
	    curchanctx->type == chanctx->type && chanctx->dwell_time != -1ULL) {
		scm80211_trace(hw, "chanctx %d TDMA scheduler\n", chanctx->id);
		mod_delayed_work(sched->workqueue, &sched->work, 0);
		mutex_unlock(&sched->mutex);
		return true;
	}

	/* Higher priority chanctx is being served. Just wait for my turn */
	scm80211_trace(hw, "chanctx %d queued\n", chanctx->id);
	mutex_unlock(&sched->mutex);
	return false;
}

/**
 * scm80211_stop_chanctx() - deactivate chanctx
 * @chanctx: chanctx to stop(synchronous)
 *
 */
void scm80211_stop_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);

	mutex_lock(&sched->mutex);

	trace_scm80211_stop_chanctx(hw, chanctx);

	chanctx->restart = false;

	scm80211_dequeue_chanctx(hw, chanctx);

	if (chanctx == scm80211_current_chanctx(hw)) {
		/* TODO: race condition? */
		bool res = cancel_delayed_work(&sched->work);

		scm80211_dbg(hw, "stopping current chanctx %d (sync, res=%d)\n", chanctx->id, res);
		scm80211_schedule_chanctx(hw, true);
	}

	mutex_unlock(&sched->mutex);
}

static void scm80211_chanctx_idle_begin(struct ieee80211_hw *hw, struct scm80211_chanctx *idle)
{
	/* PHY/RF power off? */
#ifdef SCM80211_HOSTAP_AUTO_TEST
	struct ieee80211_hw_priv *priv = hw->priv;

	if (dev_is_platform(wiphy_dev(hw->wiphy))) {
		/* Set chan to 0, which disables rx in effect */
		scm80211_rf_writel(hw, 0, 0);
		phy_writel(&priv->phy[priv->devid], 0, PHY_MODEM(FW_OPBW_PRICH_VINTF(0)));
	}
#endif /*SCM80211_HOSTAP_AUTO_TEST */
}

static void scm80211_chanctx_idle_end(struct ieee80211_hw *hw, struct scm80211_chanctx *idle)
{
	/* PHY/RF power on? */
#ifdef SCM80211_HOSTAP_AUTO_TEST
	if (dev_is_platform(wiphy_dev(hw->wiphy))) {
		idle->ctx.def = hw->conf.chandef;
		phy_config(hw, idle, BIT(PHY_CHANNEL));
	}
#endif /*SCM80211_HOSTAP_AUTO_TEST */
}

/**
 * scm80211_init_chanctx_sched(@hw)
 * 1. setup interface combiantion rules
 * 2. init chanctx work
 * 3. init idle, scan, roc, tdls chanctx
 */
int scm80211_init_chanctx_sched(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_chanctx_sched *sched = &priv->sched;
	int i;
	enum nl80211_band band = (priv->band & BIT(NL80211_BAND_2GHZ)) ?
				 NL80211_BAND_2GHZ : NL80211_BAND_5GHZ;

	hw->wiphy->iface_combinations = &scm80211_iface_combination[band];
	hw->wiphy->n_iface_combinations = 1;

	idr_init(&sched->idr);

	/* FIXME: tdls */

	mutex_init(&sched->mutex);
	sched->workqueue = create_singlethread_workqueue("scm80211");
	if (!sched->workqueue) {
		scm80211_err(hw, "failed to create a workqueue\n");
		return -ENOMEM;
	}

	INIT_DELAYED_WORK(&sched->work, scm80211_chan_work);
	for (i = 0; i < ARRAY_SIZE(sched->queue); i++) {
		INIT_LIST_HEAD(&sched->queue[i].head);
		atomic_set(&sched->queue[i].nr_chanctx, 0);
	}

	memset(&sched->idle, 0, sizeof(sched->idle));
	/*
	 * Arbitrarily set idle channel as the first available band
	 * channel
	 */
	cfg80211_chandef_create(&sched->idle.ctx.def,
				&hw->wiphy->bands[band]->channels[0], NL80211_CHAN_NO_HT);
	sched->idle.ctx.rx_chains_static = 1;
	sched->idle.ctx.rx_chains_dynamic = 1;
	sched->idle.mac80211 = false;
	sched->idle.id = 0;
	sched->idle.type = CHANCTX_TYPE_IDLE;
	sched->idle.restart = true;
	sched->idle.begin = scm80211_chanctx_idle_begin;
	sched->idle.end = scm80211_chanctx_idle_end;
	sched->idle.dwell_time = -1UL;
	sched->idle.vif = NULL;
	sched->idle.pv.present = BIT(PHY_CHANNEL) | BIT(PHY_ACTIVE_RXCHAIN_SET);
	spin_lock_init(&sched->idle.lock);
	INIT_LIST_HEAD(&sched->idle.sched_list);
	sched->cur_chan = &sched->idle;

	priv->scanner.chanctx.id = 0x1001;
	spin_lock_init(&priv->scanner.chanctx.lock);
	INIT_LIST_HEAD(&priv->scanner.chanctx.sched_list);

	priv->roc.chanctx.id = 0x1002;
	spin_lock_init(&priv->roc.chanctx.lock);
	INIT_LIST_HEAD(&priv->roc.chanctx.sched_list);

	priv->tdls.chanctx.id = 0x1003;
	spin_lock_init(&priv->tdls.chanctx.lock);
	INIT_LIST_HEAD(&priv->tdls.chanctx.sched_list);
	hrtimer_init(&priv->tdls.tdls_timer, CLOCK_MONOTONIC, HRTIMER_MODE_ABS_SOFT);
	priv->tdls.tdls_timer.function = scm80211_tdls;
	INIT_WORK(&priv->tdls.work, scm80211_tdls_work);
	return 0;
}

void scm80211_deinit_chanctx_sched(struct ieee80211_hw *hw)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct ieee80211_hw_priv *priv = hw->priv;

	hrtimer_cancel(&priv->tdls.tdls_timer);
	cancel_delayed_work_sync(&sched->work);
	destroy_workqueue(sched->workqueue);
	sched->workqueue = NULL;
	idr_destroy(&sched->idr);
}

/**
 * TDMA chanctx - ieee80211_chanctx bound to vifs
 */

struct phy_vector_merger {
	struct scm80211_chanctx *chanctx;
	unsigned long conflicts;
};

/**
 * scm80211_combine_chanctx()
 * Iterate over the vifs on the same channel context to
 * find out the conflict combination or fixup some conflicts
 */
void scm80211_combine_chanctx(void *data, u8 *addr, struct ieee80211_vif *pubvif)
{
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(pubvif);
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(pubvif);
	struct phy_vector_merger *merger = data;
	struct scm80211_chanctx *chanctx = merger->chanctx;
	unsigned long conflicts;

	if (rcu_access_pointer_chanctx_conf(pubvif) != &merger->chanctx->ctx)
		return;

	conflicts = combine_phy_vector(&chanctx->pv, &vifp->pv, true);

	if (!conflicts) {
		combine_phy_vector(&chanctx->pv, &vifp->pv, false);
		return;
	}

	/*
	 * We are in a trouble: @pubvif is likely to undergo a probelm in receiving
	 * either VHT MU or HE MU PPDUs.
	 */
	scm80211_warn(hw, "vif %d has conflicting phy config %lx\n",
		      scm80211_vdev_id(pubvif), conflicts);

	merger->conflicts |= conflicts;
}

/**
 * scm80211_vif_update_chanctx() - update chanctx
 * @hw: hardware
 * @vif: requesting virtual interface
 * @changes: bitmap of changed PHY config vector parameters.
 *
 * Returns:
 * - < 0 if configuration fails,
 * - 1 if phy configuration changes immediately,
 * - 0 otherwise (it will be reflected next time @vif goes on channel
 *
 * This function is called by the following ieee8021_ops methods to combine
 * phy vectors of chanctx @vif is bound to.
 * - bss_info_changed()
 * - assign_vif_chanctx()
 * - unassign_vif_chanctx()
 *
 * FIXME: For the case we modify the AP's channel, mac80211/driver will maintain
 * two chanctxs but the same channel , how to deal with it now?
 *
 */
int scm80211_vif_update_chanctx(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				unsigned long changes)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *chanctx;
	struct phy_vector_merger merger;
	int ret = 0;

	rcu_read_lock();

	chanctx = (void *)rcu_dereference_chanctx_conf(vif);

	rcu_read_unlock();

	if (!ieee80211_vif_to_wdev(vif) || !chanctx)
		return 0;

	scm80211_trace(hw, "%s: vif=%s, chanctx=%d, changes=%08lx\n",
		       __func__, ieee80211_vif_name(vif), chanctx->id, changes);

	trace_scm80211_vif_update_chanctx(hw, vif, chanctx, changes);

	merger.chanctx = chanctx;
	merger.conflicts = 0;

	/*
	 * Iterate over vifs on the same chanctx and merge their phy vector
	 * into chanctx->pv
	 */
	ieee80211_iterate_active_interfaces(hw, IEEE80211_IFACE_ITER_NORMAL,
					    scm80211_combine_chanctx, &merger);

	if (merger.conflicts) {
		scm80211_warn(hw, "chanctx phy vector failed: conflicts=%lx\n", merger.conflicts);

		/* What can we do? */
	}

	mutex_lock(&sched->mutex);

	if (scm80211_chanctx_is_active(hw, chanctx)) {
		phy_config(hw, chanctx, changes);
		ret = 1;
	}

	mutex_unlock(&sched->mutex);

	return ret;
}

/**
 * scm80211_reassign_chanctx() - adjust and reassign vif's channel context

  */
static inline int scm80211_reassign_vif_chanctx(struct ieee80211_hw *hw,
						struct ieee80211_vif *vif,
						struct ieee80211_vif *vif2,
						struct scm80211_chanctx *ctx, bool is_switch)
{
	struct scm80211_chanctx *ctx2 = NULL;
	struct wireless_dev *wdev = NULL;
	enum nl80211_iftype type = ieee80211_vif_type_p2p(vif);
	enum nl80211_iftype type2 = ieee80211_vif_type_p2p(vif2);
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct list_head *head = NULL;
	struct list_head *pos;
	struct scm80211_chanctx *entry = NULL;
	unsigned long flags;

	rcu_read_lock();
	ctx2 = (void *)rcu_dereference_chanctx_conf(vif2);
	rcu_read_unlock();

	if (!ctx2 || ctx2 == ctx)
		return SCM80211_CHAN_UNCHANGE;

	scm80211_dbg(hw, "%s (type [%s] cenfreq [%d]; type2 [%s] cenfreq2 [%d])\n",
		     __func__, ieee80211_iftype_str[type],
		     (ctx->ctx.def.chan)->center_freq,
		     ieee80211_iftype_str[type2], (ctx2->ctx.def.chan)->center_freq);

	if (type == NL80211_IFTYPE_AP && vif2->type == NL80211_IFTYPE_STATION) {
		/*
		 * AP want to do channel switch , return unexcept
		 * for STA in other channel
		 */
		if (is_switch)
			return SCM80211_CHAN_UNEXCEPT;
		/*
		 * Case 1: vif to be an AP , but other vif is
		 * connected STA(include P2P GC).
		 * Smoothly notify the uplayer to change the ap's channel
		 * same as STA's op channel
		 */
		wdev = ieee80211_vif_to_wdev(vif);
		if (WARN_ON(!wdev || !wdev->netdev))
			return SCM80211_CHAN_UNEXCEPT;

		scm80211_trace(hw, "AP channel switch to freq %dMHz\n",
			       (ctx2->ctx.def.chan)->center_freq);

#if KERNEL_VERSION(5, 19, 0) > MAC80211_BASE
		cfg80211_ch_switch_notify(wdev->netdev, &ctx2->ctx.def);
#else
		cfg80211_ch_switch_notify(wdev->netdev, &ctx2->ctx.def, 0);
#endif

		spin_lock_irqsave(&ctx->lock, flags);
		ctx->ctx.def = ctx2->ctx.def;
		vif->bss_conf.chandef = ctx2->ctx.def;
		spin_unlock_irqrestore(&ctx->lock, flags);

		return SCM80211_CHAN_CHANGED;
	}

	if (type2 == NL80211_IFTYPE_AP && vif->type == NL80211_IFTYPE_STATION) {
		struct ieee80211_vif_priv *priv2;
		u32 queues = scm80211_vif_queues(vif2);
		/*
		 * Case 2: vif to be a STA and connect to an AP, but other vif
		 * is AP mode , force transform the AP's channel to STA want op channel
		 */
		priv2 = ieee80211_to_priv(vif2);
		wdev = ieee80211_vif_to_wdev(vif2);
		if (WARN_ON(!wdev || !wdev->netdev))
			return SCM80211_CHAN_UNEXCEPT;

		scm80211_trace(hw, "Force AP channel switch to freq %dMhz\n",
			       (ctx->ctx.def.chan)->center_freq);

		if (vif2->bss_conf.enable_beacon) {
			/**
			 * Set CSA counter 2 and sleep for 2 * TBTT + 2ms
			 * TODO: is there a better way?
			 */
			u32 freq = (ctx->ctx.def.chan)->center_freq;

			atomic_set(&priv2->csa_cnt, 2);
			priv2->sw_ch = ieee80211_frequency_to_channel(freq);
			msleep((vif2->bss_conf.beacon_int + 1) << 1);
		}

		/* flush queues on vif2 */
		scm80211_flush(hw, vif2, queues, false);

#if KERNEL_VERSION(5, 19, 0) > MAC80211_BASE
		cfg80211_ch_switch_notify(wdev->netdev, &ctx->ctx.def);
#else
		cfg80211_ch_switch_notify(wdev->netdev, &ctx->ctx.def, 0);
#endif

		ctx2 = (void *)rcu_access_pointer_chanctx_conf(vif2);
		spin_lock_irqsave(&ctx2->lock, flags);
		ctx2->ctx.def = ctx->ctx.def;
		vif2->bss_conf.chandef = ctx->ctx.def;
		spin_unlock_irqrestore(&ctx2->lock, flags);

		return SCM80211_CHAN_CHANGED;
	}

	/*
	 * For other cases for example
	 * STA+STA(P2P_GC)
	 * STA+P2P GO
	 * multi channel(real TDMA) enabled
	 *
	 * NOTE: Currently chanel switch cost a lot of time,
	 * so we do not support multi channels which is announced in
	 * interface combiation rules
	 */
	head = &sched->queue[CHANCTX_TYPE_TDMA].head;
	mutex_lock(&sched->mutex);
	if (unlikely(list_empty(head))) {
		mutex_unlock(&sched->mutex);
		return SCM80211_CHAN_UNCHANGE;
	}

	list_for_each(pos, head) {
		entry = list_entry(pos, struct scm80211_chanctx, sched_list);
		entry->dwell_time = SCM80211_CHANCTX_QUANTUM;
	}
	mutex_unlock(&sched->mutex);
	return SCM80211_CHAN_MULTI;
}

extern void scm80211_beacon_tx(void *arg, u8 *mac, struct ieee80211_vif *vif);
static void scm80211_tdma_chan_begin(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_chanctx_conf *ctx = &chanctx->ctx;
	struct ieee80211_hw_priv *priv = hw->priv;
	u64 bcn_int = priv->beacon_int;

	scm80211_trace(hw, "%s chanctx %d: fc=%d/width=%s/fp20=%d\n",
		       __func__, chanctx->id, ctx->def.center_freq1, chanwidth_str[ctx->def.width],
		       ctx->def.chan->center_freq);

	if (chanctx->type == CHANCTX_TYPE_BEACON) {
		ieee80211_iterate_active_interfaces_atomic(hw,
							   IEEE80211_IFACE_ITER_NORMAL,
							   scm80211_beacon_tx, NULL);

		priv->beacon_offchan = false;
		/* beacon at new TBTT + beacon interval */
		if (priv->bcn_delta) {
			bcn_int -= priv->bcn_delta;
			priv->bcn_delta = 0;
		}

		mod_timer(&priv->beacon_timer, jiffies + usecs_to_jiffies(bcn_int));
	}
}

static void scm80211_tdma_chan_end(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_chanctx_conf *ctx = &chanctx->ctx;

	/*
	 * Put an intelligence here about load-based channel time
	 * allocation and etc.
	 */
	scm80211_trace(hw, "%s chanctx %d: fc=%d/width=%s/fp20=%d\n",
		       __func__, chanctx->id, ctx->def.center_freq1, chanwidth_str[ctx->def.width],
		       ctx->def.chan->center_freq);
}

struct ieee80211_vif *scm80211_get_other_vif(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	int vdev;

	for_each_set_bit(vdev, &hwp->vif_map, SCM80211_NUM_VIF) {
		if (vdev != scm80211_vdev_id(vif) && hwp->vifs[vdev])
			return hwp->vifs[vdev];
	}
	return NULL;
}

/**
 *
 * Call chain leading to add_chanctx()
 *
 * ieee80211_assoc()
 * - ieee80211_mgd_assoc()
 *   - ieee80211_pre_connection()
 *     - ieee80211_prep_channel()

 * ieee80211_prep_channel(sdata, cbss)
 * - ieee80211_determine_chantype(sdata, sband, cbss->channel, ...)
 * - ieee80211_vif_use_channel() ---------------> (a)
 * - ieee80211_chandef_downgrade()
 * - ieee80211_vif_use_channel() ---------------> (a)
 *
 * (a) ieee80211_vif_use_channel()
 * - ieee80211_check_combinations()
 * - __ieee80211_vif_release_channel()
 * - ieee80211_find_chanctx() -----------------> (b)
 * - ieee80211_new_chanctx() ------------------> (c)
 * - ieee80211_vif_update_chandef(sdata, chandef)
 *   - sdata->vif.bss_conf.chandef = *chandef
 * - ieee80211_assign_vif_chanctx() -----------> (d)
 * - ieee80211_recalc_smps_chanctx()
 * - ieee80211_recalc_radar_chanctx()
 *
 * (b) ieee80211_find_chanctx()
 * - ieee80211_chanctx_reserved_chandef()
 * - ieee80211_change_chanctx()
 *   - ieee80211_change_chanctx()
 *     - drv_change_chanctx(local, ctx, IEEE80211_CHANCTX_CHANGE_WIDTH) (*)
 *     - ieee80211_recalc_chanctx_min_def()
 *     - ieee80211_hw_config()
 *
 * (c) ieee80211_new_chanctx()
 * - ieee80211_add_chanctx(local, ctx)
 * - ieee80211_idle_off(local)
 * - ieee80211_hw_config()
 * - drv_add_chanctx(local, ctx) (*)
 * - ieee80211_recalc_idle(local)
 *
 * (d) ieee80211_assign_vif_chanctx()
 * - drv_unassign_vif_chanctx() (*)
 * - drv_assign_vif_chanctx() (*)
 * - ieee80211_recalc_chanctx_num_assigned()
 * - ieee80211_recalc_chanctx_chantype()
 * - ieee80211_recalc_smps_chanctx()
 * - ieee80211_recalc_radar_chanctx()
 * - ieee80211_recalc_chanctx_min_def()
 * - ieee80211_recalc_txpower()
 * ...
 * - ieee80211_bss_info_change_notify(, BSS_CHANGED_IDLE)
 * - ieee80211_check_fast_xmit_iface()
 *
 */

/**
 * scm80211_add_chanctx() - struct ieee80211_ops::add_chanctx() callback
 *
 * @ctx: channel context being created
 */
int scm80211_add_chanctx(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *chanctx = (void *)ctx;
	int n;

	n = atomic_read(&sched->queue[CHANCTX_TYPE_TDMA].nr_chanctx);
	if (n <= 0)
		n = 1;

	chanctx->mac80211 = true;
	spin_lock_init(&chanctx->lock);
	INIT_LIST_HEAD(&chanctx->assigned_vif);
	INIT_LIST_HEAD(&chanctx->sched_list);
	chanctx->type = CHANCTX_TYPE_TDMA;
	chanctx->begin = scm80211_tdma_chan_begin;
	chanctx->end = scm80211_tdma_chan_end;
	chanctx->dwell_time = -1UL;
	chanctx->private = NULL;

#ifdef SCM80211_DFS_CERTIFIED
	scm80211_dfs_init(hw, chanctx);
#endif
	chanctx->id = idr_alloc(&hw_sched(hw)->idr, chanctx, 1, SCM80211_MAX_CHANCTX, GFP_KERNEL);

	scm80211_dbg(hw, "adding chanctx %d: fc=%d/width=%s/fp20=%d\n",
		     chanctx->id, ctx->def.center_freq1, chanwidth_str[ctx->def.width],
		     ctx->def.chan->center_freq);

	if (chanctx->id < 0) {
		scm80211_warn(hw, "failed to add chanctx: fc=%d/width=%d/fp20=%d\n",
			      ctx->def.center_freq1, ctx->def.width, ctx->def.chan->center_freq);
		return chanctx->id;
	}

	return 0;
}

/**
 * scm80211_remove_chanctx() - struct ieee80211_ops::remove_chanctx() callback
 *
 * @ctx: channel context to remove
 */
void scm80211_remove_chanctx(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *chanctx = (void *)ctx;
	unsigned long flags;

	scm80211_dbg(hw, "%s() chanctx %d: fc=%d/width=%s/fp20=%d\n",
		     __func__, chanctx->id, ctx->def.center_freq1,
		     chanwidth_str[ctx->def.width], ctx->def.chan->center_freq);

	if (!chanctx->mac80211)
		scm80211_warn(hw, "trying to remove internal chanctx %d\n", chanctx->id);

#ifdef SCM80211_DFS_CERTIFIED
	scm80211_dfs_deinit(hw, chanctx);
#endif
	/*
	 * mac80211 guarntees there is no vifs assigned to @ctx
	 */
	spin_lock_irqsave(&chanctx->lock, flags);

	if (!list_empty(&chanctx->assigned_vif))
		scm80211_warn(hw, "trying to remove active chanctx %d\n", chanctx->id);
	list_del_init(&chanctx->assigned_vif);	/* redundant */

	spin_unlock_irqrestore(&chanctx->lock, flags);

	chanctx->restart = false;
	/*
	 * No vif remains on this channel - remove @chanctx from
	 * scheduling list. The @remove_chanctx() will
	 * remove this one shortly.
	 */
	scm80211_stop_chanctx(hw, chanctx);

	idr_remove(&sched->idr, chanctx->id);
}

/**
 * scm80211_change_chanctx() - struct ieee80211_ops::change_chanctx() callback
 *
 * @ctx: channel context to change
 * @changed: change flags
 *
 * This function is called by mac80211 to notify changes in chanctx. Currently
 * the changes is one of enum ieee80211_chanctx_change, which only affect
 * @PHY_CHANNEL and @PHY_ACTIVE_RXCHAIN_SET, and not other parameters in
 * struct phy_vecotr. Therefore, this function does not try to iterate
 * over the interfaces sitting on @ctx and merge the phy vector together.
 *
 * What is the better whay of "applying" changes to PHY/RF instead of if statement
 * below?
 */
void scm80211_change_chanctx(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx,
			     u32 changed)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *chanctx = (void *)ctx;
	unsigned long changes = BIT(PHY_CHANNEL) | BIT(PHY_ACTIVE_RXCHAIN_SET);

	scm80211_dbg(hw, "%s: chanctx=%d(fc=%d/width=%s/fp20=%d, changed=0x%08x)\n",
		     __func__, chanctx->id, ctx->def.center_freq1, chanwidth_str[ctx->def.width],
		     ctx->def.chan->center_freq, changed);
	/*
	 * mac80211 only changes @PHY_CHANNEL and @PHY_ACTIVE_RXCHAIN_SET.
	 * No need to merge all interface's PHY vectors,.
	 */
	scm80211_dynamic_ps_disable(hw, true);
	mutex_lock(&sched->mutex);

#ifdef SCM80211_DFS_CERTIFIED
	if (changed & IEEE80211_CHANCTX_CHANGE_RADAR) {
		scm80211_trace(hw, "%s: chanctx=%d\n", __func__, chanctx->id);
		scm80211_dfs_detect_start(chanctx);
	}
#endif //SCM80211_DFS_CERTIFIED

	if (changed & IEEE80211_CHANCTX_CHANGE_RX_CHAINS)
		scm80211_smps_mode_set(hw, ctx);

	if (scm80211_chanctx_is_active(hw, chanctx)) {
		if (changed == IEEE80211_CHANCTX_CHANGE_MIN_WIDTH) {
			struct phy_vector *pv = &chanctx->pv;
			int width = CBW20;

			switch (ctx->def.width) {
			case NL80211_CHAN_WIDTH_80:
				width = CBW80;
				break;
			case NL80211_CHAN_WIDTH_40:
				width = CBW40;
				break;
			default:
				break;
			}
			if (pv->BANDWIDTH == width) {
				mutex_unlock(&sched->mutex);
				return;
			}
		}
		if (changed & IEEE80211_CHANCTX_CHANGE_RX_CHAINS)
			msleep(1);

		/*
		 * Apply the change right away if @ctx is currently active
		 * chanctx. Otherwise, the change will be applied next time
		 * @ctx is scheduled.
		 */
		phy_config(hw, chanctx, changes);
	}

	mutex_unlock(&sched->mutex);
}

/**
 * scm80211_assign_vif_chanctx() - assign_vif_chanctx() callback
 *
 * @hw:
 * @pubvif: &struct ieee80211_vif that is being assigned
 * @ctx: chanctx is being assigned to
 *
 * @ctx is added to scheduler
 */
int scm80211_assign_vif_chanctx(struct ieee80211_hw *hw,
				struct ieee80211_vif *pubvif, struct ieee80211_chanctx_conf *ctx)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(pubvif);
	struct scm80211_chanctx *chanctx = (void *)ctx;
	struct ieee80211_vif *other = NULL;
	unsigned long flags;
	bool first = true;
	int ret = SCM80211_CHAN_UNCHANGE;

	scm80211_dbg(hw,
		     "%s: vif=%s, chanctx=%d (fc=%d/width=%s/fp20=%d)\n",
		     __func__, ieee80211_vif_name(pubvif), chanctx->id,
		     ctx->def.center_freq1, chanwidth_str[ctx->def.width],
		     ctx->def.chan->center_freq);

	scm80211_dynamic_ps_disable(hw, true);
#ifdef SCM80211_DFS_CERTIFIED
	if (ieee80211_iface_is_ap(pubvif))
		scm80211_dfs_enable(hw, chanctx);
#endif //SCM80211_DFS_CERTIFIED

	/*
	 * AP + STA on different channels
	 *  - AP is forced to move to the same channel with STA
	 *  - If STA starts after AP, AP has to do CSA
	 *
	 * Plan:
	 *  - Check other active interface
	 *  - Call cfg80211_ch_switch_notify(dev, chandef) like this:
	 *
	 *      struct net_device *dev;
	 *
	 *      dev = ieee80211_vif_to_wdev(pubvif)->netdev;
	 *      cfg80211_ch_switch_notify(dev, &ctx->def);
	 */

	other = scm80211_get_other_vif(hw, pubvif);
	if (other)
		ret = scm80211_reassign_vif_chanctx(hw, pubvif, other, chanctx, false);

	switch (ret) {
	case SCM80211_CHAN_UNCHANGE:
		break;
	case SCM80211_CHAN_CHANGED:
		first = false;
		break;
	case SCM80211_CHAN_MULTI:
		chanctx->dwell_time = SCM80211_CHANCTX_QUANTUM;
		/**
		 * TODO:
		 * 1) probably we should enable IEEE80211_HW_CONNECTION_MONITOR
		 * here to prevent to many beacon loss detection by mac80211
		 * 2) We need to configure the PHY@VINTF1 for multi channel
		 **/
		break;
	default:
		return -EBUSY;
	}

	INIT_LIST_HEAD(&vifp->chanctx);

	spin_lock_irqsave(&chanctx->lock, flags);
	if (first)
		first = list_empty(&chanctx->assigned_vif);
	list_add_tail(&vifp->chanctx, &chanctx->assigned_vif);

	spin_unlock_irqrestore(&chanctx->lock, flags);

	if (first) {
		/* Start chanctx */
		chanctx->restart = true;
		scm80211_start_chanctx(hw, chanctx);
	} else {
		/*
		 * There are other vifs sitting on this @chanctx.
		 * @chanctx must have been added to the scheduler already.
		 * Just update chanctx and wait for turn, or start immediately
		 * if @chanctx is currently being served.
		 * FIXME: how to deal with two chanctx share one channel
		 */
		scm80211_vif_update_chanctx(hw, pubvif, -1UL);
	}

	return 0;
}

void scm80211_unassign_vif_chanctx(struct ieee80211_hw *hw,
				   struct ieee80211_vif *pubvif, struct ieee80211_chanctx_conf *ctx)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(pubvif);
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	struct scm80211_chanctx *chanctx = (void *)ctx;
	struct scm80211_chanctx *chanctx2 = NULL;
	unsigned long flags;
	bool last = true;
#ifdef SCM80211_DFS_CERTIFIED
	struct scm80211_dfs *dfs = chanctx->dfs;
#endif // SCM80211_DFS_CERTIFIED
	scm80211_dbg(hw,
		     "%s: vif=%s, chanctx=%d (fc=%d/width=%s/fp20=%d)\n",
		     __func__, ieee80211_vif_name(pubvif), chanctx->id,
		     ctx->def.center_freq1, chanwidth_str[ctx->def.width],
		     ctx->def.chan->center_freq);

#ifdef SCM80211_DFS_CERTIFIED
	if (dfs && dfs->enable)
		scm80211_dfs_disable(hw, chanctx);
#endif // SCM80211_DFS_CERTIFIED

	if (chanctx->dwell_time != -1ULL) {
		/* Exit from real TDMA mode now */
		struct list_head *head = &sched->queue[CHANCTX_TYPE_TDMA].head;
		struct scm80211_chanctx *entry = NULL;
		int cnt = 0;

		mutex_lock(&sched->mutex);
		if (likely(!list_empty(head))) {
			struct list_head *pos;

			list_for_each(pos, head) {
				entry = list_entry(pos, struct scm80211_chanctx, sched_list);
				if (entry != chanctx) {
					cnt++;
					chanctx2 = entry;
				}
			}
		}
		mutex_unlock(&sched->mutex);

		if (cnt == 1) {
			spin_lock_irqsave(&chanctx2->lock, flags);
			chanctx2->dwell_time = -1UL;
			spin_unlock_irqrestore(&chanctx2->lock, flags);
		}
	} else {
		struct ieee80211_vif *other = scm80211_get_other_vif(hw, pubvif);
		const struct cfg80211_chan_def *chandef;

		if (!other)
			goto out;

		rcu_read_lock();
		chanctx2 = (void *)rcu_dereference_chanctx_conf(other);
		rcu_read_unlock();

		if (!chanctx2 || chanctx2 == chanctx)
			goto out;

		chandef = cfg80211_chandef_compatible(&chanctx2->ctx.def, &chanctx->ctx.def);

		if (likely(chandef)) {
			mutex_lock(&sched->mutex);
			/**
			 * enqueue the chanctx2, maybe it has added in
			 */
			chanctx2->restart = true;
			scm80211_enqueue_chanctx(hw, chanctx2);
			mutex_unlock(&sched->mutex);
			last = false;
		}
	}

out:
	spin_lock_irqsave(&chanctx->lock, flags);

	list_del_init(&vifp->chanctx);

	if (last)
		last = list_empty(&chanctx->assigned_vif);

	memset(&vifp->pv, 0, sizeof(vifp->pv));

	spin_unlock_irqrestore(&chanctx->lock, flags);

	if (last) {
		/*
		 * No vif remains on this channel - remove @chanctx from
		 * scheduling list later.
		 */
		chanctx->restart = false;
	} else {
		/*
		 * There are still other vifs on this channel - update
		 * the chanctx hoping remaining ones can enable more
		 * features. Again, scm80211_vif_update_chanctx() will
		 * directly call phy_config() if chanctx is being
		 * served.
		 */
		scm80211_vif_update_chanctx(hw, pubvif, -1UL);
	}
	scm80211_try_to_doze(hw);
}

/**
 * scm80211_switch_vif_chanctx() -  switch_vif_chanctx() callback
 * mode type:
 * @CHANCTX_SWMODE_SWAP_CONTEXTS
 *	- 1) unassign old ctx
 *	- 2) add new ctx
 *	- 3) assign new ctx
 *	- 4) remove old ctx (should it move to the step-2 ?)
 * @CHANCTX_SWMODE_REASSIGN_VIF
 *	- 1) unassign old ctx
 *	- 2) assign new ctx
 *
 * case1: AP want to do channel switch and another VIF is act as connected STA
 *	  Return busy now, because AP + STA work in the same channel now do not
 *	  change it
 *
 * case2: STA want to do channel switch and another VIF is act as AP, move the
 *	  AP's channel to the same channel
 *
 * case3: STA+STA concurrent mode, we'd better compare the new ctx with
 *	  the other vif's assigned chanctx to deside whether need to enable or
 *	  disable real TDMA
 */

int scm80211_switch_vif_chanctx(struct ieee80211_hw *hw,
				struct ieee80211_vif_chanctx_switch *vifs,
				int n_vifs, enum ieee80211_chanctx_switch_mode mode)
{
	struct scm80211_chanctx_sched *sched = hw_sched(hw);
	unsigned long flags;
	int i;

	scm80211_dbg(hw, "switch chanctx n_vifs %d mode %s\n",
		     n_vifs, mode ? "SWAP_CONTEXTS" : "REASSIGN_VIF");

	for (i = 0; i < n_vifs; ++i) {
		struct ieee80211_vif_priv *vifp;
		struct scm80211_chanctx *old, *new;
		struct ieee80211_vif *vif2 = NULL;
		int ret = SCM80211_CHAN_UNCHANGE;
		bool first;

		vifp = ieee80211_to_priv(vifs[i].vif);
		old = (void *)vifs[i].old_ctx;
		new = (void *)vifs[i].new_ctx;

		vif2 = scm80211_get_other_vif(hw, vifs[i].vif);
		if (vif2)
			ret = scm80211_reassign_vif_chanctx(hw, vifs[i].vif, vif2, new, true);
		switch (ret) {
		case SCM80211_CHAN_UNCHANGE:
		case SCM80211_CHAN_CHANGED:
			break;
		case SCM80211_CHAN_MULTI:
			new->dwell_time = SCM80211_CHANCTX_QUANTUM;
			/**
			 * TODO:
			 * 1) probably we should enable IEEE80211_HW_CONNECTION_MONITOR
			 * here to prevent to many beacon loss detection by mac80211
			 * 2) We need to configure the PHY@VINTF1 for multi channel
			 **/
			break;
		default:
			return -EBUSY;
		}

		spin_lock_irqsave(&old->lock, flags);
		list_del_init(&vifp->chanctx);
		old->restart = false;
		spin_unlock_irqrestore(&old->lock, flags);

		if (mode == CHANCTX_SWMODE_SWAP_CONTEXTS && scm80211_add_chanctx(hw, &new->ctx))
			return -EBUSY;

		spin_lock_irqsave(&new->lock, flags);
		first = list_empty(&new->assigned_vif);
		list_add_tail(&vifp->chanctx, &new->assigned_vif);
		new->restart = true;
		spin_unlock_irqrestore(&new->lock, flags);

		mutex_lock(&sched->mutex);
		scm80211_enqueue_chanctx(hw, new);
		mutex_unlock(&sched->mutex);

		if (!first || ret == SCM80211_CHAN_CHANGED)
			scm80211_vif_update_chanctx(hw, vifs[i].vif, -1UL);

		mutex_lock(&sched->mutex);
		if (scm80211_current_chanctx(hw) == old || first) {
			if (delayed_work_pending(&sched->work))
				mod_delayed_work(sched->workqueue, &sched->work, 0);
			else
				scm80211_schedule_chanctx(hw, true);
		}
		mutex_unlock(&sched->mutex);

		if (mode == CHANCTX_SWMODE_SWAP_CONTEXTS) {
			spin_lock_irqsave(&old->lock, flags);
			old->mac80211 = false;
			spin_unlock_irqrestore(&old->lock, flags);
			scm80211_remove_chanctx(hw, &old->ctx);
		}
	}

	return 0;
}

/**
 * scm80211_restore_bss_config()
 *
 * restore the original configuration for STA/AP which is changed
 * while scanning
 */
static void scm80211_restore_bss_config(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_bss_conf *bss_conf = &vif->bss_conf;

	switch (vif->type) {
	case NL80211_IFTYPE_STATION:
		if (bss_conf->assoc) {
			scm80211_set_basic_rates(hw, vif, bss_conf);
			scm80211_config_bssid_filter(hw, vif, true);
		}
		break;
	case NL80211_IFTYPE_AP:
		if (bss_conf->enable_beacon)
			scm80211_set_basic_rates(hw, vif, bss_conf);
		break;
	default:
		break;
	}
}

static void scm80211_scan_start_one_chan(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_scanner *scanner = &priv->scanner;
	struct cfg80211_scan_request *req = &scanner->req->req;
	struct ieee80211_channel *chan = chanctx->ctx.def.chan;
	struct sk_buff *probereq;
	struct ieee80211_mgmt *mgmt;
	struct ieee80211_tx_info *info = NULL;
	bool passive = false;
	int i, dwell;

	priv->survey_data[scanner->index].start = jiffies;
	priv->survey_data[scanner->index].channel = chan;

	passive |= (chan->flags & (IEEE80211_CHAN_NO_IR | IEEE80211_CHAN_RADAR));
	passive |= (req->n_ssids == 0);

	dwell = passive ? PASSIVE_SCAN_DWELL : ACTIVE_SCAN_DWELL;
	if (req->duration_mandatory || (req->duration && dwell > req->duration))
		dwell = req->duration;

	if (dev_is_platform(wiphy_dev(hw->wiphy)))
		dwell *= SCM80211_HWSIM_DWELL_EXTEND;

	chanctx->dwell_time = dwell;
	chanctx->air_time = 0;

	scm80211_set_mandatory_hwrates(hw, scanner->vif, chan->band, false);
	scm80211_config_bssid_filter(hw, scanner->vif, false);

	scm80211_trace(priv->hw, "scanning index=%d: %d MHz(%c], dwell=%ld\n",
		       scanner->index, chanctx->ctx.def.chan->center_freq, "AP"[passive],
		       chanctx->dwell_time);

	if (passive)
		return;

	/* Send a probe request */
	for (i = 0; i < req->n_ssids; i++) {
		probereq = ieee80211_probereq_get(hw, scanner->addr,
						  req->ssids[i].ssid,
						  req->ssids[i].ssid_len, req->ie_len);
		if (!probereq)
			continue;

		info = IEEE80211_SKB_CB(probereq);
		info->band = req->channels[scanner->index]->band;
		info->control.vif = scanner->vif;

		mgmt = (struct ieee80211_mgmt *)probereq->data;
		memcpy(mgmt->da, req->bssid, ETH_ALEN);
		memcpy(mgmt->bssid, req->bssid, ETH_ALEN);

		if (req->ie_len)
			skb_put_data(probereq, req->ie, req->ie_len);
		local_bh_disable();
		scm80211_tx_channel(hw, probereq, chanctx->ctx.def.chan);
		local_bh_enable();
	}
}

static void scm80211_scan_end_one_chan(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_scanner *scanner = &priv->scanner;
	struct cfg80211_scan_request *req = &scanner->req->req;

	priv->survey_data[scanner->index].end = jiffies;
	scanner->index++;

	if (!scanner->scanning)
		return;

	if (scanner->index >= req->n_channels) {	/* final channel */
		struct ieee80211_vif *vif = scanner->vif;
		struct cfg80211_scan_info result = {
			.aborted = false,
		};

		scm80211_dbg(hw, "hw scan complete\n");
		if (scanner->scanning)
			ieee80211_scan_completed(hw, &result);
		/* resetore the vif address */
		if (req->flags & NL80211_SCAN_FLAG_RANDOM_ADDR)
			scm80211_set_addr(hw, vif, vif->addr);
		scanner->scanning = false;
		scanner->scanned = true;

#ifdef SCM80211_COEX
		priv->coex.reduce_streams_bypass = true;
#endif

		scm80211_restore_bss_config(hw, vif);
		/* Do not request @chanctx any more */
		chanctx->restart = false;
		scm80211_try_to_doze(hw);
	} else {
		/* Request for next channel */
		cfg80211_chandef_create(&chanctx->ctx.def,
					req->channels[scanner->index], NL80211_CHAN_NO_HT);

		chanctx->restart = true;	/* just to be sure */
	}
}

int scm80211_hw_scan(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		     struct ieee80211_scan_request *scan_req)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_scanner *scanner = hw_scanner(hw);
	struct scm80211_chanctx *chanctx = &scanner->chanctx;
	struct cfg80211_scan_request *req = &scan_req->req;
	int i, vidx = scm80211_vdev_id(vif);

	scm80211_dbg(hw, "%s: nssid=%d, nchannels=%d, vdev=%d,devid=%d\n",
		     __func__, req->n_ssids, req->n_channels, vidx, priv->devid);

	scm80211_wakeup(hw);

	mutex_lock(&priv->mutex);
	if (WARN_ON(scanner->scanning || priv->roc.running || priv->tdls.started)) {
		mutex_unlock(&priv->mutex);
		return -EBUSY;
	}
	scm80211_trace(hw, "hw scan started\n");
	scanner->scanning = true;
	scanner->scanned = false;
	mutex_unlock(&priv->mutex);

	for (i = 0; i < req->n_ssids; ++i) {
		if (!memcmp(req->ssids[i].ssid, "DIRECT-", 7)) {
			scm80211_set_p2p_basic_rate(hw, vidx);
			break;
		}
	}

	scanner->req = scan_req;
	scanner->vif = vif;
	scanner->index = 0;

	if (req->flags & NL80211_SCAN_FLAG_RANDOM_ADDR) {
		get_random_mask_addr(scanner->addr, req->mac_addr, req->mac_addr_mask);
		/* change multicast address to unicast address */
		if (is_multicast_ether_addr(scanner->addr))
			scanner->addr[0] &= ~0xFE;
		/*
		 * temporarily set mac address to random address
		 * otherwise AP's response can not be filtered in
		 */
		if (!is_zero_ether_addr(scanner->addr))
			scm80211_set_addr(hw, vif, scanner->addr);
		else
			memcpy(scanner->addr, vif->addr, ETH_ALEN);
	} else {
		memcpy(scanner->addr, vif->addr, ETH_ALEN);
	}

	memset(priv->survey_data, 0, sizeof(priv->survey_data));

	/* Build up scan chanctx */
	if (req->duration_mandatory) {
		chanctx->dwell_time = req->duration * req->n_channels;
	} else if (!req->n_ssids) {
		chanctx->dwell_time = PASSIVE_SCAN_DWELL * req->n_channels;
	} else {
		u32 dur = 0;

		if (req->duration)
			chanctx->dwell_time = req->duration * req->n_channels;

		for (i = 0; i < req->n_channels; i++) {
			if (req->channels[i]->flags & (IEEE80211_CHAN_RADAR | IEEE80211_CHAN_NO_IR))
				dur += PASSIVE_SCAN_DWELL;
			else
				dur += ACTIVE_SCAN_DWELL;
		}
		if (!chanctx->dwell_time || chanctx->dwell_time > dur)
			chanctx->dwell_time = dur;
	}

	cfg80211_chandef_create(&chanctx->ctx.def, req->channels[scanner->index],
				NL80211_CHAN_NO_HT);

	chanctx->ctx.rx_chains_static = 1;
	chanctx->ctx.rx_chains_dynamic = 1;
	chanctx->mac80211 = false;
	chanctx->restart = true;
	chanctx->begin = scm80211_scan_start_one_chan;
	chanctx->end = scm80211_scan_end_one_chan;
	chanctx->type = CHANCTX_TYPE_SCAN;
	chanctx->vif = vif;

	scm80211_start_chanctx(hw, chanctx);

	return 0;
}

void scm80211_cancel_hw_scan(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct scm80211_scanner *scanner = hw_scanner(hw);
	struct scm80211_chanctx *chanctx = &scanner->chanctx;
	struct ieee80211_scan_request *req = scanner->req;
	struct cfg80211_scan_info result = {
		.aborted = true,
	};

	if (scanner->scanning)
		ieee80211_scan_completed(hw, &result);

	if (req && (req->req.flags & NL80211_SCAN_FLAG_RANDOM_ADDR))
		scm80211_set_addr(hw, vif, vif->addr);

	scm80211_stop_chanctx(hw, chanctx);

	scm80211_restore_bss_config(hw, vif);
	scanner->scanning = false;
	scanner->scanned = true;
	scm80211_try_to_doze(hw);
}

/**
 * ROC chanctx
 */
static void scm80211_roc_start(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	/* FIXME: do it here? */
	scm80211_set_p2p_basic_rate(hw, scm80211_vdev_id(chanctx->vif));
	priv->roc.running = true;

	ieee80211_ready_on_channel(hw);
}

static void scm80211_roc_done(struct ieee80211_hw *hw, struct scm80211_chanctx *chan)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (priv->roc.running)
		ieee80211_remain_on_channel_expired(hw);
	priv->roc.running = false;
	scm80211_try_to_doze(hw);
}

int scm80211_roc(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		 struct ieee80211_channel *chan, int duration, enum ieee80211_roc_type type)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_chanctx *chanctx = &priv->roc.chanctx;

	scm80211_dbg(hw, "%s: (%d MHz, %d ms)\n", __func__, chan->center_freq, duration);

	scm80211_wakeup(hw);
	mutex_lock(&priv->mutex);
	if (WARN_ON(priv->scanner.scanning || priv->roc.running || priv->tdls.started)) {
		mutex_unlock(&priv->mutex);
		return -EBUSY;
	}
	priv->roc.vif = vif;
	mutex_unlock(&priv->mutex);

	cfg80211_chandef_create(&chanctx->ctx.def, chan, NL80211_CHAN_NO_HT);
	chanctx->ctx.rx_chains_static = 1;
	chanctx->ctx.rx_chains_dynamic = 1;
	chanctx->mac80211 = false;
	chanctx->restart = false;	/* one-shot */
	chanctx->begin = scm80211_roc_start;
	chanctx->end = scm80211_roc_done;
	chanctx->type = CHANCTX_TYPE_ROC;
	chanctx->dwell_time = duration;
	chanctx->vif = vif;

	scm80211_start_chanctx(hw, chanctx);
	return 0;
}

int scm80211_croc(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_chanctx *chanctx = &priv->roc.chanctx;

	priv->roc.running = false;
	scm80211_stop_chanctx(hw, chanctx);
	scm80211_try_to_doze(hw);
	return 0;
}

void scm80211_clr_tdls(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (priv->tdls.vif == vif && priv->tdls.started == true) {
		priv->tdls.started = false;
		ieee80211_queue_work(hw, &priv->tdls.work);
	}
}

static void scm80211_tdls_work(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv = container_of(work, struct ieee80211_hw_priv, tdls.work);
	struct ieee80211_hw *hw = priv->hw;
	struct scm80211_chanctx *chanctx = &priv->tdls.chanctx;

	if (priv->tdls.started)
		scm80211_start_chanctx(hw, chanctx);
	else
		scm80211_stop_chanctx(hw, chanctx);
}

static enum hrtimer_restart scm80211_tdls(struct hrtimer *timer)
{
	struct ieee80211_hw_priv *priv =
	    container_of(timer, struct ieee80211_hw_priv, tdls.tdls_timer);
	struct ieee80211_hw *hw = priv->hw;
	struct ieee80211_sta *sta = priv->tdls.sta;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	ktime_t next_tdls;

	if (priv->tdls.packets == sta_priv->tx_packets + sta_priv->rx_packets) {
		priv->tdls.started = false;
		ieee80211_queue_work(hw, &priv->tdls.work);
	} else {
		priv->tdls.packets = sta_priv->tx_packets + sta_priv->rx_packets;
		if (priv->tdls.switch_timeout > 0) {
			next_tdls = ns_to_ktime(priv->tdls.switch_timeout * 1000);
			hrtimer_start(&priv->tdls.tdls_timer, next_tdls, HRTIMER_MODE_REL_SOFT);
		}
	}

	return HRTIMER_NORESTART;
}

static void scm80211_tdls_start(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long queues = scm80211_vif_queues(chanctx->vif);

	priv->tdls.state = TDLS_CHANCTX_SWITCH_FINISH;
	scm80211_wake_queues(hw, queues);
}

static void scm80211_tdls_done(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->tdls.state = TDLS_CHANCTX_SWITCH_INIT;
	priv->tdls.started = false;
}

void scm80211_mgd_tdls_discover(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	/*
	 * tdls discover will send in AP's channel,
	 * when off-channel, frame will be queued, and after recover
	 * and it will be sent after going back to AP's operation channel.
	 */
}

int scm80211_tdls_channel_switch(struct ieee80211_hw *hw,
				 struct ieee80211_vif *vif,
				 struct ieee80211_sta *sta, u8 oper_class,
				 struct cfg80211_chan_def *chandef,
				 struct sk_buff *tmpl_skb, u32 ch_sw_tm_ie)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_chanctx *chanctx = &priv->tdls.chanctx;
	struct ieee80211_tx_control control = { };
	struct ieee80211_tx_info *info;
	struct sk_buff *skb;
	struct ieee80211_sta_priv *sta_priv = (struct ieee80211_sta_priv *)sta->drv_priv;
	struct ieee80211_hdr *hdr;

	scm80211_dbg(hw, "%s: (%d MHz)\n", __func__, chandef->chan->center_freq);

	mutex_lock(&priv->mutex);
	if (WARN_ON(priv->scanner.scanning || priv->roc.running || priv->tdls.started)) {
		mutex_unlock(&priv->mutex);
		return -EBUSY;
	}

	priv->tdls.vif = vif;
	priv->tdls.switch_time = ch_sw_tm_ie & 0xffff;
	priv->tdls.switch_timeout = (ch_sw_tm_ie & 0xffff0000) > 16;
	mutex_unlock(&priv->mutex);

	chanctx->ctx.def = *chandef;
	chanctx->ctx.rx_chains_static = 1;
	chanctx->ctx.rx_chains_dynamic = (sta_priv->n_ss > 1) ? 2 : 1;
	chanctx->mac80211 = false;
	chanctx->restart = false;	/* one-shot */
	chanctx->begin = scm80211_tdls_start;
	chanctx->end = scm80211_tdls_done;
	chanctx->type = CHANCTX_TYPE_TDLS;
	chanctx->dwell_time =
	    ieee80211_tu_to_usec(TDLS_TIMEOUT_COUNT * vif->bss_conf.beacon_int) / 1000;
	chanctx->vif = vif;

	skb = skb_copy(tmpl_skb, GFP_ATOMIC);
	info = IEEE80211_SKB_CB(skb);
	info->control.vif = vif;
	control.sta = sta;

	hdr = (void *)skb->data;
	if (ieee80211_is_data_qos(hdr->frame_control)) {
		/*
		 * use AC_VO to send this frame, because QOS-DATA 's seq is set by
		 * mac80211, if use AC_BE to send, we can't get correct seq which leads
		 * Receiver drop the frame if AMPDU has enabled.
		 */
		u8 *p = ieee80211_get_qos_ctl(hdr);
		*p = *p | 0x7;
	}

	scm80211_tx(hw, &control, skb);
	priv->tdls.state = TDLS_CHANCTX_SWITCH_REQUEST;
	return 0;
}

void scm80211_tdls_cancel_channel_switch(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif, struct ieee80211_sta *sta)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->tdls.started = false;
	ieee80211_queue_work(hw, &priv->tdls.work);
}

void scm80211_tdls_recv_channel_switch(struct ieee80211_hw *hw,
				       struct ieee80211_vif *vif,
				       struct ieee80211_tdls_ch_sw_params *params)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_chanctx *chanctx = &priv->tdls.chanctx;
	struct ieee80211_tx_control control = { };
	struct ieee80211_tx_info *info;
	struct ieee80211_sta_priv *sta_priv;
	struct sk_buff *skb;
	struct ieee80211_hdr *hdr;
	ktime_t next_tdls;

	scm80211_dbg(hw, "%s tdls code = %d\n", __func__, params->action_code);

	scm80211_wakeup(hw);
	if (params->action_code == WLAN_TDLS_CHANNEL_SWITCH_REQUEST) {
		mutex_lock(&priv->mutex);
		if (WARN_ON(priv->scanner.scanning || priv->roc.running) || priv->tdls.started) {
			mutex_unlock(&priv->mutex);
			scm80211_info(hw, "chanctx is busy now!\n");
			return;
		}

		priv->tdls.vif = vif;
		priv->tdls.switch_time = params->switch_time;
		priv->tdls.switch_timeout = params->switch_timeout;
		mutex_unlock(&priv->mutex);

		sta_priv = (struct ieee80211_sta_priv *)params->sta->drv_priv;
		chanctx->ctx.def = *params->chandef;
		chanctx->ctx.rx_chains_static = 1;
		chanctx->ctx.rx_chains_dynamic = (sta_priv->n_ss > 1) ? 2 : 1;
		chanctx->mac80211 = false;
		chanctx->restart = false;	/* one-shot */
		chanctx->begin = scm80211_tdls_start;
		chanctx->end = scm80211_tdls_done;
		chanctx->type = CHANCTX_TYPE_TDLS;
		chanctx->dwell_time =
		    ieee80211_tu_to_usec(TDLS_TIMEOUT_COUNT * vif->bss_conf.beacon_int) / 1000;
		chanctx->vif = vif;
		priv->tdls.sta = params->sta;
		priv->tdls.state = TDLS_CHANCTX_SWITCH_RESPONSE;

		skb = skb_copy(params->tmpl_skb, GFP_ATOMIC);
		info = IEEE80211_SKB_CB(skb);
		info->control.vif = vif;
		control.sta = params->sta;

		hdr = (void *)skb->data;
		if (ieee80211_is_data_qos(hdr->frame_control)) {
			/**
			 * use AC_VO to send this frame, because QOS-DATA 's seq is set by
			 * mac80211, if use AC_BE to send, we can't get correct seq which leads
			 * Receiver drop the frame if AMPDU has enabled.
			 */
			u8 *p = ieee80211_get_qos_ctl(hdr);
			*p = *p | 0x7;
		}

		scm80211_tx(hw, &control, skb);
	}

	if (params->action_code == WLAN_TDLS_CHANNEL_SWITCH_RESPONSE &&
	    priv->tdls.state == TDLS_CHANCTX_SWITCH_REQUEST) {
		priv->tdls.sta = params->sta;
		sta_priv = (void *)params->sta->drv_priv;
		priv->tdls.packets = sta_priv->tx_packets + sta_priv->rx_packets;
		priv->tdls.started = true;
		ieee80211_queue_work(hw, &priv->tdls.work);

		if (priv->tdls.switch_timeout > 0) {
			next_tdls = ns_to_ktime(priv->tdls.switch_timeout * 1000);
			hrtimer_start(&priv->tdls.tdls_timer, next_tdls, HRTIMER_MODE_REL_SOFT);
		}
	}
}
