#include <linux/compiler.h>
#include <linux/bitfield.h>
#include <linux/list_sort.h>
#include <linux/log2.h>

#define func_fmt(str)		"[COEX] "str
#define func_trace			COEX

#ifdef SCM80211_COEX

#include "hw.h"
#include "core.h"
#include "util.h"
#include "trace.h"
#include "debug.h"
#include "coex.h"


void scm80211_enable_coex_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	scm80211_trace(hw, "%s\n", __func__);

	priv->int_en |= INTR_ENABLE_COEX;
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

void scm80211_disable_coex_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	scm80211_trace(hw, "%s\n", __func__);

	priv->int_en &= ~INTR_ENABLE_COEX;
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

void scm80211_coex_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long queues = 0;
	int i;

	scm80211_trace(hw, "%s\n", __func__);

	for (i = 0; i < SCM80211_NUM_VIF; i++) {
		if ((priv->vif_map & BIT(i)))
			queues +=
			    ((1 << SCM80211_NUM_HWQ_PER_VIF) - 1) << (i * SCM80211_NUM_HWQ_PER_VIF);
	}
	/*
	 * When INTR_STATUS_COEX raise, BT stop, wake hwq to continue scheduling txop
	 * WARNNING: wakeup all hwq when intr raised.
	 */
	scm80211_wake_queues(hw, queues);
	hrtimer_try_to_cancel(&priv->coex.stop_queue_watchdog);
	priv->coex.coex_en = false;
	scm80211_disable_coex_irq(hw);
}

void scm80211_set_coex(struct ieee80211_hw *hw, struct scm80211_chanctx *ctx)
{
	enum nl80211_band band;
	u32 v = 0;
	struct phy_vector *pv = &ctx->pv;
	struct ieee80211_chanctx_conf *chanctx = &ctx->ctx;
	struct cfg80211_chan_def *chandef = &chanctx->def;

	if (!chandef->chan) {
		scm80211_warn(hw, "%s: no chan available\n", __func__);
		return;
	}

	v = scm80211_mac_readl(hw, MAC_REG_DEV_OPT2);
	v &= ~DEV_OPT2_COEX_WLAN_ABORT_EN;

	band = chandef->chan->band;
	if (band != NL80211_BAND_2GHZ) {
		v |= (FIELD_PREP(DEV_OPT2_COEX_WLAN_ABORT_EN, 0));
		scm80211_mac_writel(hw, v, MAC_REG_DEV_OPT2);
		scm80211_disable_coex_irq(hw);
		return;
	}

	v |= (FIELD_PREP(DEV_OPT2_COEX_WLAN_ABORT_EN, 1));
	scm80211_mac_writel(hw, v, MAC_REG_DEV_OPT2);

	v = 0;
	v |= FIELD_PREP(DEV_CH_COEX_WLAN_CHANNEL_FREQ, (chandef->center_freq1 - 2400));
	scm80211_trace(hw, "%s: center_freq=%d, center_freq1=%d\n",
		       __func__, chandef->chan->center_freq, chandef->center_freq1);
	scm80211_trace(hw, "%s: BANDWIDTH=%d PRIMARY_20MHZ_CHANNEL_LOC=%d\n",
		       __func__, pv->BANDWIDTH, pv->PRIMARY_20MHZ_CHANNEL_LOC);
	if (pv->BANDWIDTH && chandef->center_freq1 > chandef->chan->center_freq)
		v |= FIELD_PREP(DEV_CH_COEX_WLAN_CHANNEL_OFFSET, 1);

	v |= FIELD_PREP(DEV_CH_COEX_RESP_PRI0, 0xF);
	v |= FIELD_PREP(DEV_CH_COEX_RESP_PRI1, 0xF);

	scm80211_mac_writel(hw, v, MAC_REG_DEV_CH_COEX);
}

#if defined(SCM2020_Z1) || defined(SCM2020_A0) || defined(SCM2020_A1)
static void scm80211_omi_change(struct ieee80211_hw *hw, bool reduce_streams_flag)
{
	struct ieee80211_hw_priv *hw_priv = hw->priv;
	struct ieee80211_sta *sta = hw_priv->coex.sta;
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	unsigned long flag;

	hw_priv->coex.htc_omi.id = VIF_HTC_OMI_HETB_QOSNULL_OMI_EN;
	if (reduce_streams_flag) {
		/* Tx/Rx stream number set to 1 */
		hw_priv->coex.htc_omi.rx_nss = 0;
		hw_priv->coex.htc_omi.tx_nss = 0;
	} else {
		/* Tx/Rx stream number set to 2 */
		hw_priv->coex.htc_omi.rx_nss = 1;
		hw_priv->coex.htc_omi.tx_nss = 1;
	}

	scm80211_trace(hw, "OMI content %04x\n", hw_priv->coex.htc_omi.content);

	spin_lock_irqsave(&sta_priv->lock, flag);
	sta_priv->omi_req = HE_VARIANT | (hw_priv->coex.htc_omi.content << 2);
	spin_unlock_irqrestore(&sta_priv->lock, flag);
}
#endif

/*
 * Using 1000ms monitor period to polling BT event as default
 * Using 200ms monitor period to polling BT event when BT active
 */
enum hrtimer_restart scm80211_coex_2ant_mon_timer(struct hrtimer *timer)
{
#if defined(SCM2020_Z1) || defined(SCM2020_A0) || defined(SCM2020_A1)

	struct scm80211_coex *coex = container_of(timer, struct scm80211_coex,
						      coex_2ant_timer);
	struct ieee80211_hw *hw = coex->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	int vidx = 0;
	int mimo;

	scm80211_trace(hw, "check BT event mimo=%d\n",
					phy_readl(priv->phy, PHY_MODEM(FW_MIMO_EN_VINTF(vidx))));

	coex->bt_action_counter = scm80211_pta_readl(hw, PTA_STAT_BT_TX)
							+ scm80211_pta_readl(hw, PTA_STAT_BT_RX);
	/*
	 * bt_action_counter:
	 * 0 ~ 4 for bt on, no scanning, no connection
	 * 200 ~ 300 for bt on, connection, no scanning, no music
	 * for bt on, connection, no scanning, music
	 */
	scm80211_trace(hw, "coex_mon:bt_act_cnt=%d\n", coex->bt_action_counter);
	if (coex->bt_action_counter > 4) {
		if (coex->reduce_streams_bypass) {
			/* reduce tx rate stream in "get_txop_rate" */
			/* disable pv->MIMO_EN here? (Rx) */
			coex->reduce_streams_flag = true;
			mimo = 0;
			phy_writel(priv->phy, mimo, PHY_MODEM(FW_MIMO_EN_VINTF(vidx)));
			scm2020_set_gpio_config(hw, 40, 1, SCM_GPIO_OUT, true);
			scm2020_set_gpio_config(hw, 41, 1, SCM_GPIO_OUT, true);
			scm80211_omi_change(hw, coex->reduce_streams_flag);
			coex->reduce_streams_bypass = false;
		}
		hrtimer_start(&coex->coex_2ant_timer,
				      ms_to_ktime(SCM80211_COEX_2ANT_MONITOR_PERIOD_200),
				      HRTIMER_MODE_REL_SOFT);
	} else {
		if (!coex->reduce_streams_bypass) {
			coex->reduce_streams_flag = false;
			mimo = 1;
			phy_writel(priv->phy, mimo, PHY_MODEM(FW_MIMO_EN_VINTF(vidx)));
			scm2020_set_gpio_config(hw, 40, 5, SCM_GPIO_IN, false);
			scm2020_set_gpio_config(hw, 41, 5, SCM_GPIO_IN, false);
			scm80211_omi_change(hw, coex->reduce_streams_flag);
			coex->reduce_streams_bypass = true;
		}
		hrtimer_start(&coex->coex_2ant_timer,
				      ms_to_ktime(SCM80211_COEX_2ANT_MONITOR_PERIOD_1000),
				      HRTIMER_MODE_REL_SOFT);
	}

	coex->bt_action_counter = 0;
	scm80211_pta_writel(hw, 0x0, PTA_STAT_BT_TX);
	scm80211_pta_writel(hw, 0x0, PTA_STAT_BT_RX);

#endif

	return HRTIMER_NORESTART;

}

enum hrtimer_restart scm80211_hwq_stop_queue_watchdog_expire(struct hrtimer *timer)
{
	struct scm80211_coex *coex = container_of(timer, struct scm80211_coex,
						      stop_queue_watchdog);
	struct ieee80211_hw *hw = coex->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long queues = 0;
	int i;

	scm80211_info(hw, "stop_queue_wdg expired (miss_coex_int)\n");
	for (i = 0; i < SCM80211_NUM_VIF; i++) {
		if ((priv->vif_map & BIT(i)))
			queues +=
			    ((1 << SCM80211_NUM_HWQ_PER_VIF) - 1) << (i * SCM80211_NUM_HWQ_PER_VIF);
	}
	/* WARNNING: wakeup all hwq when wdg expire. */
	scm80211_wake_queues(hw, queues);
	coex->coex_en = false;
	scm80211_disable_coex_irq(hw);
	return HRTIMER_NORESTART;
}

void scm80211_pta_pri_set(struct ieee80211_hw *hw, u8 wlan)
{
#if defined(SCM2020_Z1) || defined(SCM2020_A0) || defined(SCM2020_A1)
	u32 v;

	if (wlan == 1) {
		/* set wlan in higher priority */
		v = PTA_ENABLE | PTA_PRIORITY | PTA_NO_SIM_TX |
		    PTA_NO_SIM_RX | CHAN_OVERLAPDECT_ENABLE | INT_ABORT_RX_IN_TX;
	} else {
		/* set BT in higher priority */
		v = PTA_ENABLE | CHAN_OVERLAPDECT_ENABLE | WLAN_PTI_MODE | INT_ABORT_RX_IN_TX;
	}
	
	scm80211_pta_writel(hw, v, PTA_CONFIG);
#endif
}

#endif
