/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifndef __SCM80211_CORE_H__
#define __SCM80211_CORE_H__

#include <linux/types.h>
#include <linux/workqueue.h>
#include <linux/slab.h>
#include <linux/regmap.h>
#include <linux/proc_fs.h>
#include <linux/platform_device.h>
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/average.h>
#include <linux/hrtimer.h>
#include <linux/version.h>

#include <net/mac80211.h>
#include <linux/version.h>

#include "cfg80211.h"
#include "hrtimer.h"
#include "skbuff.h"
#include "ieee80211ax.h"
#include "hw.h"
#include "phy.h"
#include "random.h"
#ifdef SCM80211_DFS_CERTIFIED
#include "dfs.h"
#endif
#ifdef SCM80211_WOW
#include "wow.h"
#endif
#ifdef SCM80211_FTM
#include "ftm.h"
#endif
#ifdef SCM80211_COEX
#include "coex.h"
#endif

extern bool hwsim_mode;

#define SCM2020_INTF_MAX			2
#define SCM80211_SKB_MAX_FRAGS			8
#define SCM80211_MAX_TXOP			2
#define SCM80211_MAX_STA			8
#define SCM80211_NUM_TXQ			(SCM80211_MAX_STA * 8)

/*
 *  since kernel 5.17, "NOMIMAL" -> "NOMINAL"
 */
#if KERNEL_VERSION(5, 17, 0) > MAC80211_BASE
#define IEEE80211_HE_PHY_CAP9_NOMINAL_PKT_PADDING_0US \
	IEEE80211_HE_PHY_CAP9_NOMIMAL_PKT_PADDING_0US
#define IEEE80211_HE_PHY_CAP9_NOMINAL_PKT_PADDING_8US \
	IEEE80211_HE_PHY_CAP9_NOMIMAL_PKT_PADDING_8US
#define IEEE80211_HE_PHY_CAP9_NOMINAL_PKT_PADDING_16US \
	IEEE80211_HE_PHY_CAP9_NOMIMAL_PKT_PADDING_16US
#define IEEE80211_HE_PHY_CAP9_NOMINAL_PKT_PADDING_RESERVED \
	IEEE80211_HE_PHY_CAP9_NOMIMAL_PKT_PADDING_RESERVED
#define IEEE80211_HE_PHY_CAP9_NOMINAL_PKT_PADDING_MASK \
	IEEE80211_HE_PHY_CAP9_NOMIMAL_PKT_PADDING_MASK
#endif

enum scm80211_tx_info_flags {
	SCM80211_XTI_HAS_TXTIME = BIT(0),
	SCM80211_XTI_HAS_TXVECTOR = BIT(1),
	SCM80211_XTI_DMA_MAPPED = BIT(2),
};

enum scm2020_gpio_dir {
	SCM_GPIO_IN = 0,
	SCM_GPIO_OUT,
	SCM_GPIO_TMAX,
};

/**
 * struct scm80211_tx_info - skb tx metadata
 *
 * @ti: copy of ieee80211_tx_info
 * @tv: &struct txvector which will be prepended to @skb->data
 * @mlen: mpdu length (including FCS)
 * @txtime: ppdu txtime
 * @per_time: air time for per MPDU (approximate not exact)
 * @duration: response ppdu txtime + sifs
 * @mmss_pad: the number of MPDU delimiters to be added before
 *     the subframe represented by @skb
 * @eof: true if skb is EOF MPDU
 * @ack: true if skb solicits an Ack or a BA
 * @sgt: struct sg_table
 * @sg: scatterlist
 * @bdesc: buffer descriptor array
 * @bd_dma_addr: dma address of buffer descriptor
 *
 * FIXME: tidy up the fields
 * FIXME:
 * Instead of maintinaing information in xtxi, put it into mdesc and
 * do dma_sync_xxx(mdesc) before starting tx.
 *
 */
struct scm80211_tx_info {
	struct txvector tv;

	u32 flags;
	size_t mlen;
	u32 txtime;
	u32 per_time;
	u16 duration;
	u16 sn;
	int mmss_pad;
	bool eof;
	bool ack;
	u8 tid;
	struct sg_table sgt;
	struct scatterlist sg[SCM80211_SKB_MAX_FRAGS + 1];
	u32 subtxtime;
	u32 subper_time;
};

struct ieee80211_subframe {
	struct sk_buff *skb;
	size_t len;		/* MPDU length */
	int mmss_pad;		/* number of MPDU delimiter before the subframe */
};

#define AMPDU_DATA_ENABLED		BIT(0)
#define AMPDU_IMMEDIATE_RESPONSE	BIT(1)
#define AMPDU_CONTROL			BIT(2)
#define AMPDU_PSMP			BIT(3)
#define AMPDU_SMPDU			BIT(4)
#define AMPDU_HE_ACK_ENABLED		BIT(5)
#define AMPDU_HE_MULTI_TID		BIT(6)

#define TID_TO_WME_AC(_tid)             \
	((((_tid) == 0) || ((_tid) == 3)) ? IEEE80211_AC_BE :   \
	 ((_tid) < 3) ? IEEE80211_AC_BK :   \
	 ((_tid) < 6) ? IEEE80211_AC_VI :   \
	 IEEE80211_AC_VO)

#define WME_AC_TO_TID(_ac)		\
	(((_ac) == IEEE80211_AC_VO) ? 6 :	\
	 ((_ac) == IEEE80211_AC_VI) ? 5 :	\
	 ((_ac) == IEEE80211_AC_BK) ? 1 :	\
	 0)

#define SCM80211_AC_TO_IEEE80211_AC(_ac) \
	(((_ac) == SCM80211_HWQ_VO) ? IEEE80211_AC_VO :	\
	 ((_ac) == SCM80211_HWQ_VI) ? IEEE80211_AC_VI :	\
	 ((_ac) == SCM80211_HWQ_BK) ? IEEE80211_AC_BK :	\
	 IEEE80211_AC_BE)

/**
 * struct scm80211_ampdu_info - A-MPDU information
 *
 * @enabled: HT immediate block ack agreement has been established
 * @maxlen: maximum A-MPDU length of the recipient
 * @bufsize: BA window size of the recipient
 * @mmss: minimum MPDU start spacing (in byte)
 * @len: total A-MPDU length (APEP_LENGTH)
 * @ssn: starting sequence number of this A-MPDU (obsolete)
 * @qosctrl: QoS Control field of the first QoS Data frame
 * @tids: tx scoreboard
 */
struct scm80211_ampdu_info {
	bool enabled;
	bool update_needed;
	size_t maxlen;
	size_t len;		/* pre-eof length */
	int bitrate;
	u32 mmss;

	struct {
		size_t bufsize;
		__le16 qosctl;
		u16 ssn;
		unsigned long bitmap[DIV_ROUND_UP((SCM80211_MAX_MPDU << 1), BITS_PER_LONG)];
		unsigned long ba[DIV_ROUND_UP((SCM80211_MAX_MPDU << 1), BITS_PER_LONG)];
	} tids[IEEE80211_NUM_TIDS + 1];

	unsigned long tidset_ba;
	unsigned long tidset_eof;

	bool has_bar;
	bool has_eof;
	int has_qos_data;
	bool has_immediate_response;
};

struct scm80211_txq;
struct scm80211_hwq;

enum txop_state {
	TXOP_STATE_IDLE,
	TXOP_STATE_NEEDS_RESCHED,
	TXOP_STATE_SCHEDULED,
	TXOP_STATE_ISSUING,
	TXOP_STATE_ISSUED,
	TXOP_STATE_TX_DONE,
	TXOP_STATE_TX_RETRY,
	TXOP_STATE_END,
};

extern const char *const txop_state_str[];
/**
 * struct txop - txop schedule information
 *
 * @limit: txoplimit (in microsecond)
 * @txtime: the expected total air time
 * @rts_time: RTS/CTS exchange time (including 2 SIFS)
 * @ack_time: Ack time (non A-MPDU) or BA time (A-MPDU)
 * @ti: tx info of this txop. All the frames in this txop are transmitted using
 *      this parameter
 * @xti: array of struct scm80211_tx_info structure.
 * @valid: 1 if txop is ready
 * @dma[0]: mpdu and buffer descriptor in the host memory
 * @dma[1]: mdpu and buffer descriptor in the target memory
 * @raddr: remote target address of the mpdu table base address
 * @bounce_buffer: remote target address of the bounce buffer
 */
struct txop {
	/* schedule/transmission context */
	struct sk_buff			*skb; /* first frame */
	struct scm80211_ampdu_info	ampdu;
	struct txvector		tv;
	struct sk_buff_head		sched;
	u32				cmd;
	struct txdesc			txd;
	__le64				bitmap;
	ktime_t				tstamp; /* start time */

	int		n_mpdu;
	int		sent;
	int		acked;
	int		nacked;
	int		retries;
	u8		reduce;
	/* Retransmition times on current TXOP, reset to 0 when init TXOP */
	int		retransmit;
	u32		status;
	u32		limit;
	u32		txtime;
	u32		rts_time;
	u32		ack_time;
	u32		actual_time;
	u8		protection;
	bool		he;
	bool		he_dcm;
	u16		rts_time_th;
	u32		bitrate;
	u32		bcnt;
	int		cts_timeout_cnt;
	int		winlost_cnt;
	enum gi_type	he_gi_type;
	enum ltf_type	he_ltf_type;
	bool		probe;
	u8		n_primary; /* To be removed */
	bool		tx_window_valid;
	bool		scheduled;
	u8		retry_cnt;

	void		*marker;

	/* ppdu context */

	struct ieee80211_tx_info	rc; /* to keep the rc rates */
	struct ieee80211_tx_info	rs; /* to report back to rc */
	struct ieee80211_tx_info	ti; /* modified working copy */

	struct ieee80211_tx_rate	*rate; /* current rate in ti->control.rates */

	struct scm80211_tx_info	xti[SCM80211_MAX_MPDU];
	bool				retry;

	/* shortcut */
	struct ieee80211_vif		*vif;
	struct ieee80211_sta		*sta;

	struct scm80211_txq		*owner;
	struct scm80211_txq		*multi_owner;
	struct scm80211_txq		*txq[2];
	unsigned int total_len;	/*total len in bytes*/

	/* subrate */
	struct ieee80211_tx_rate	subrate; /* rate for usb retainsmission */
	bool				subused;
	struct txvector                 subtv;

	/*
	 * permanent context
	 * @dma should be the first
	 */
	struct {
		struct tx_dma_desc	*desc; /* host virtual address */
		dma_addr_t		paddr;  /* host bus address */
		phys_addr_t		raddr;	/* target physical address */
	} dma[2];

	/* USB only */
	struct urb			*urb;
	struct delayed_work		urb_timeout;
	struct txop_info		txop_info;

	phys_addr_t		raddr;
	phys_addr_t		bounce_buffer;
	enum space_type			space;

	struct ieee80211_hw		*hw;
	struct scm80211_hwq		*hwq;

	int				id;
	enum txop_state			state;
	enum txop_state			old_state;
	struct list_head		list;

	char				buf[256];
};

enum scm80211_hwq_state {
	HWQ_STATE_IDLE = 0,
	HWQ_STATE_DELAY = 1,
	HWQ_STATE_BACKOFF = 2,
	HWQ_STATE_IN_TRANSMISSION = 3,
};

enum scm80211_hwq_flush {
	HWQ_FLUSH_NONE = 0,
	HWQ_FLUSH_DRAIN = 1,
	HWQ_FLUSH_DROP = 2,
};

struct scm80211_tx_debug;

#define HE_TX_FINISHED		3
#define HE_TX_SELECT_MAX	HE_TX_FINISHED
#define TX_RATE_BACKUP		HE_TX_FINISHED

#define MU_EDCA_CONFIGURED	BIT(0)
#define MU_EDCA_RUNNING		BIT(1)

struct tx_dma_desc {
	struct tx_mpdu_desc mdesc[SCM80211_MAX_MPDU];

	struct {
		struct tx_buffer_desc bdesc[SCM80211_SKB_MAX_FRAGS + 1] __aligned(64);
	} mpdu[SCM80211_MAX_MPDU];

	unsigned int buffer[128 / sizeof(unsigned int)];
};

/**
 * struct scm80211_hwq - hardware AC queue
 *
 * @id: hardware queue ID [0 .. hw->queue-1]
 * @shm: shared memory base address
 * @regbase: ?
 * @txdesc: tx descriptor (in shared memory)
 * @state: hardware queue state
 * @hw: mac80211 hardware the hardware queue belongs to
 * @he_sta: indicate hwq transmit pkt to an he STA
 * @vif:  virtual interface that the hardware queue belongs to
 * @mu_edca_timer: MU EDCA  timer for using MU EDCA parameter
 * @edca: EDCA parameter set for this hardware queue
 * @b_edca: Backup EDCA parameter set for this hardware queue when using MU EDCA parameter
 * @owner:  &struct scm80211_txq that is selected for transmission
 * @multi_owner:  &txq that is selected for multi-tid transmission
 * @n_primary: save the number of pkts for the primary txq(owner)
 * @list: list of ieee80211_txq's mapped to this hardware queue.
 * @mtab: mpdu descriptor base
 * @txop: txop schedule information
 * @mu_edca_status: current MU EDCA status to check if hwq is using MU EDCA parameter
 */
struct scm80211_hwq {
	int					id;

	struct txdesc __iomem			*txdesc;
	struct hrtimer				mu_edca_timer;
	struct ieee80211_tx_queue_params	edca;
	struct ieee80211_tx_queue_params	b_edca;
	enum scm80211_hwq_state			qstate;
	struct ieee80211_hw			*hw;
	struct ieee80211_vif			*vif;
	struct ieee80211_sta			*he_sta;
	struct list_head			txq_head;
	struct scm80211_txq			*txq;
	struct scm80211_txq			*p2p_txq;
	struct work_struct			work;
#ifdef SCM80211_FTM
	struct work_struct			tx_ac_wk;
	enum gi_type				he_gi_type;
	enum ltf_type				he_ltf_type;
#endif
	struct tasklet_struct			tasklet;
	struct completion			completion;
	spinlock_t				lock;
	bool					mac80211;
	bool					p2p_dev_enable;
	int					expire_cnt;

	u16					cw;
	u8					qsrc;
	u8					qlrc;

	struct {
		u32				order;
		struct txop			pool[SCM80211_MAX_TXOP];
		u32				rptr, wptr;
		struct list_head		head[TXOP_STATE_END];
		spinlock_t			lock;
	} ring;
	bool                                    need_resched;
	bool					enabled;
	enum scm80211_hwq_flush			flush;
	bool					idle;
	int					ac;
	u32					mu_edca_status;
	__le16					mu_timer;
	bool					use_recmd_rate;

	struct hrtimer				watchdog;
	int					bcnt;	/* backoff counter */
	ktime_t					tstamp; /* start time */
	s64					timeout;

	unsigned int limit;
	struct scm80211_tx_debug		*debug;
	bool					on_bar;
	u16					bar_ssn;

};

#define txop_to_hwq(x) container_of((x), struct scm80211_hwq, txop)

#define ieee80211_to_priv(ptr) ((ptr) ? (void *)((ptr))->drv_priv : (void *)(ptr))
#define ieee80211_hw_to_priv(hw) ((struct ieee80211_hw_priv *)((hw)->priv))
#define priv_to_ieee80211(type, priv) container_of(((void *)priv), type, drv_priv)

struct scm80211_txq {
	struct ieee80211_txq	txq;	/* Must be first */

	struct ieee80211_hw	*hw;
	struct list_head	list;
	struct list_head	txq_list;

	struct sk_buff_head	queue;
	struct sk_buff_head	retry;

	struct scm80211_hwq	*hwq;
	struct ieee80211_txq	*source;
	struct ieee80211_ampdu_params ba;
	spinlock_t		lock;

	char			name[32];
	u8			hw_queue;
	s8			idx;
	bool			internal;
	bool			offchannel;
	bool			stop;

	/* Transmit window */
	struct {
		u16 ssn;
		unsigned long bitmap[64 / BITS_PER_LONG];
	} sb;
};

static inline size_t scm80211_txq_data_size(void)
{
	return sizeof(struct scm80211_txq) - sizeof(struct ieee80211_txq);
}

/**
 * ieee80211_dbg_rate_fix
 * @rate: indicates the rate info of HT/VHT/HE
 * @he_gi_type: indicates the he GI type
 * @he_ltf_type: indicates the HE-LTF type
 * @flags: record the mac80211_tx_info_flags such as
 *	   IEEE80211_TX_CTL_LDPC
 * @rate_idx: indicates the NON-HT rate index
 */
struct ieee80211_dbg_rate_fix {
	struct ieee80211_tx_rate rate;
	enum nl80211_band band;
	enum gi_type he_gi_type;
	enum ltf_type he_ltf_type;
	u32 flags;

	/* Add more */
};

/**
 * Driver will keep statistics for rate higher than MCS6
 * from MCS7 ~ MCS11 5 rates
 * There are 6 groups (3-types bw * 2-types stream)
 **/
#define GROUP_COUNT		6
#define GROUP_RATES		5

#define PROB_SCALE		7
#define PROB_FRAC(val, div)	(((val) << PROB_SCALE) / div)
#define PROB_TRUNC(val)		((val) >> PROB_SCALE)

/**
 * @attempts: attempts tx counter on this rate
 * @success: success tx counter on this rate
 * @tot_atmp: total attempts counter
 * @tot_succ: total success counter
 * @prob: avg proportion on this rate
 * @jiffies the last update timestamp
 **/
struct he_higher_rates_stats {
	u16 attempts, success;
	u32 tot_atmp, tot_succ;
	s16 prob;
	unsigned long jiffies;
};

/**
 * @tot_ampdus: total tx ampdu number
 * @tot_mpdus: success tx mpdu number
 * @tot_succ: total tx success mpdu number
 * @mpdus_per_ampdu: avg mpdu number per ampdu
 **/
struct mpdu_stats {
	u32 tot_ampdus;
	u32 tot_mpdus;
	u32 tot_succ;
	u32 mpdus_per_ampdu;
};

/**
 * @asleep: indicates whether it is in sleep mode
 * @band: indicates sta operation band
 * @width: indicates sta operation width
 * @vht_la: VHT link Adaptation cap
 * @vht_msi: msi for vht
 * @he_la: HE link Adaptation cap
 * @he_msi: msi for he
 * @htc: record the LA htc received from this sta
 *	 but how to deal with it - FIXME
 * @n_ss: record then nss of this sta
 * @he_mcs: record he rx mcs cap
 * @vht_mcs: record vht rx mcs cap
 * @ht_mcs: record ht rx mcs cap
 * @he_multid: indicates Multi-TID Aggregation rx ability
 * @rate_available: indicates whether recomd_rate available
 * @recomd_rate: record the recommended rate in HT-control linked adapatation
 * @stbc_or_dcm: whether use stbc for vht or dcm modulation for he
 * @rate_fix_on: indicates whether rate fix on or off
 * @rate_fix: indicates rate information
 * @rate_stats: to record the rate_stats for this sta
 */
struct ieee80211_sta_priv {
	u16 max_mpdu_len;
	u32 max_ampdu_len;
	u8 mmss;
	spinlock_t lock;
	bool asleep;
	enum nl80211_band band;
	enum nl80211_chan_width width;
	u8 vht_la;
	u8 vht_msi;
	u8 he_la;
	u8 he_msi;
	u8 n_ss;
	u8 vht_mcs;
	u8 he_mcs[8];		/* he_mcs[i]: i-th MCS mask */
	u16 he_npp[2][CBW80 + 1];
	u8 ht_mcs;
	u8 he_multid;
	u16 txop_rts_th;
	bool rate_available;
	struct ieee80211_tx_rate recomd_rate;
	bool stbc_or_dcm;

	bool rate_fix_on;
	struct ieee80211_dbg_rate_fix rate_fix;
	u64 tx_packets;
	u64 rx_packets;
	struct he_higher_rates_stats rate_stats[GROUP_COUNT * GROUP_RATES + 1];
	struct mpdu_stats mpdu_stat[IEEE80211_NUM_ACS];
	struct rate_info txinfo;

	u32 omi_req;
	u32 he_aom;
	u32 he_nom;
};

struct ieee80211_ie_sr_reuse_param_set {
	u8 sr_control;
	u8 non_srg_obss_pd_max_offset;	/* optional */
	u8 srg_obss_pd_min_offset;	/* optional */
	u8 srg_obss_pd_max_offset;	/* optional */
	u64 srg_bss_colar_bitmap;	/* optional */
	u64 srg_partial_bssid_bitmap;	/* optional */
} __packed;

/**
 * struct ieee80211_vif_priv - ieee80211_vif driver data
 *
 * @mq: txq for tx()
 */
#define SCM80211_NUM_TIDS 8

struct scm80211_sr {
	u64 sr_non_srg_color_en;	/* REG_OBSS_SR_NON_SRG_COLOR */
	u64 sr_non_srg_bssid_en;	/* REG_OBSS_SR_NON_SRG_BSSID_EN */
	u64 sr_srg_color_en;	/* REG_SRG_SR_COLOR_EN */
	u64 sr_srg_bssid_en;	/* REG_SRG_SR_BSSID_EN */
	u8 sr_obss_pd_thr;	/* OBSS_PD_THR */
	u8 sr_obss_pd_er_thr;	/* OBSS_PD_ER_THR */
	u8 sr_srg_thr;		/* SRG_THR */
	u8 sr_srg_er_thr;	/* SRG_ER_THR */

};

enum scm80211_twt_state {
	TWT_DISABLED = 0,
	TWT_REQ,
	TWT_NEGO,
	TWT_DOZE,
	TWT_SP,
};

#ifdef SCM80211_TWT
#define SCM80211_TWT_AHEAD_SP		3000
#define SCM80211_TWT_AHEAD_SP_END	500
extern const char *twt_state[];
#endif

#define pm_regs(name, num)			\
	struct {				\
		struct {			\
			u32 addr;		\
			u32 val;		\
		} regs[num];			\
		int type;			\
		int n;				\
	} name;

#define pm_save(hw, sr, reg_type, regarray)					\
	do {									\
		int i;								\
		for (i = 0; i < ARRAY_SIZE(regarray); i++) {			\
			sr.regs[i].addr = regarray[i];				\
			switch (reg_type) {					\
			case MAC_REG:						\
				sr.regs[i].val = reg_readl(mac, hw, regarray[i]);	\
				break;						\
			case PHY_REG:						\
				sr.regs[i].val = reg_readl(phy, hw, regarray[i]);	\
				break;						\
			case RFI_REG:						\
				sr.regs[i].val = reg_readl(rf, hw, regarray[i]);	\
				break;						\
			case RAW_REG:						\
			default:						\
				sr.regs[i].val = reg_readl(raw, hw, regarray[i]);	\
				break;						\
			}							\
		}								\
		sr.type = reg_type;						\
		sr.n	= i;							\
	} while (0);

#define pm_restore(hw, sr)							\
	do {									\
		int i;								\
		if (sr.n <= 0)							\
			break;							\
		for (i = 0; i < sr.n; i++) {					\
			switch (sr.type) {					\
			case MAC_REG:						\
				reg_writel(mac, hw, sr.regs[i].val, sr.regs[i].addr);	\
				break;						\
			case PHY_REG:						\
				reg_writel(phy, hw, sr.regs[i].val, sr.regs[i].addr);	\
				break;						\
			case RFI_REG:						\
				reg_writel(rf, hw, sr.regs[i].val, sr.regs[i].addr);	\
				break;						\
			case RAW_REG:						\
			default:						\
				reg_writel(raw, hw, sr.regs[i].val, sr.regs[i].addr);	\
				break;						\
			}							\
		}								\
	} while (0);


struct ieee80211_vif_priv {
	struct ieee80211_hw *hw;
	u8 index;
	struct scm80211_txq mq;
	u8 mgmt_hw_queue;
	s8 def_key_idx;
	u8 bssid_filter;
	spinlock_t lock;
	/* P2P Client */
	struct ieee80211_noa_data noa;
	struct hrtimer p2p_ps_timer;
	struct delayed_work suspend_tx_work;

	struct {
		struct {
			u32 last;
			u32 next;
			u32 tsf0;
			u32 latency;
			u32 dur;
		} beacon;
		bool traffic_indicator_used;
		bool more_data_used;
	} bss;

	ktime_t	last_beacon;	/* last beacon time */

	bool wait_beacon;
#ifdef SCM80211_FTM
	struct scm80211_ftm ftm;
#endif //SCM80211_FTM

	struct {
		bool doze;
		struct timer_list doze_timer;
		struct timer_list wake_timer;

		pm_regs(assoc_awake,  16);
		pm_regs(assoc_active, 16);
		pm_regs(ps_active,    16);
	} pm;

	struct {
		enum tbtt_state {
			TBTT_DISABLE = 0,
			TBTT_START,
			TBTT_END,
		} state;
		u32 next_time;
	} tbtt;

	struct {
		enum dtim_state {
			DTIM_DISABLE = 0,
			DTIM_START,
			DTIM_END,
		} state;
		u32 next_time;
	} dtim;

	int tid_queue_size[SCM80211_NUM_TIDS];
	bool has_sr;
	struct ieee80211_ie_sr_reuse_param_set sr_param;
	struct scm80211_sr sr;
	atomic_t num_sta;
	u32 bss_changed;	/* bss_info_changed bitmap */
	u8 bcn_seqnum;
	u8 null_seqnum;
	u8 slot;
#ifdef SCM80211_WFA_TEST
	u32 hetb_psdu_len;
#endif
	atomic_t csa_cnt;
	u8 sw_ch;
	bool noa_enabled;
#ifdef SCM80211_TWT
	enum scm80211_twt_state twt_state;
	enum scm80211_twt_state old_state;
	struct work_struct twt_state_work;
	struct hrtimer twt_tsf_timer;
	struct sk_buff *twt_request;
	size_t twt_off;
	u64    twt;
	u32    twt_min_sp_dur;
	u32    twt_wake_int;
	bool   twt_trigger;
	bool   twt_implicit;
	bool   twt_flowtype;
	bool   twt_prot;
	bool   twt_teardown;
	int    twt_flow_id;
	u64    twt_next_irq;
	ktime_t   twt_ktime;
#endif /* SCM80211_TWT */
	u64    sw_tsf;

	__le32 reg_mmss;

	u64 rx_beacon;
	ktime_t tstamp;		/* last tx success time */
	u64 beacon_timeout;
	u32 beacon_loss_cnt;
	struct timer_list bcn_mon_timer;
	struct work_struct bcn_mon_work;
	struct work_struct omi_work;
	bool disable_qnull_irq;

	struct list_head chanctx;
	struct phy_vector pv;
};

enum scm80211_tdls_state {
	TDLS_CHANCTX_SWITCH_INIT = 0,
	TDLS_CHANCTX_SWITCH_REQUEST,
	TDLS_CHANCTX_SWITCH_RESPONSE,
	TDLS_CHANCTX_SWITCH_FINISH,
};

/**
 * enum scm80211_chanctx_type
 *
 * @SCM80211_CHANCTX_TDMA: virtual interface operating channel.
 *	These channels are created and removed by mac80211 via
 *	add_chanctx() and remove_chanctx() callbacks.
 * @SCM80211_CHANCTX_ROC: remain on channel
 * @SCM80211_CHANCTX_SCAN: channel is for scanning
 */
enum scm80211_chanctx_type {
	CHANCTX_TYPE_IDLE = 0,
	CHANCTX_TYPE_TDMA,
	CHANCTX_TYPE_ROC,
	CHANCTX_TYPE_SCAN,
	CHANCTX_TYPE_TDLS,
	CHANCTX_TYPE_BEACON,
	CHANCTX_TYPE_FTM,
	CHANCTX_TYPE_MAX,
};

/**
 * enum scm80211_chan_reassign_type
 *
 * @SCM80211_CHAN_UNEXCEPT: unexcept chan.
 * @SCM80211_CHAN_UNCHANGE: chan unchanged
 * @SCM80211_CHAN_CHANGED: channel changed
 * @SCM80211_CHAN_MULTI: enable multi-channel real TDMA enabled
 */
enum scm80211_chan_reassign_type {
	SCM80211_CHAN_UNEXCEPT = -1,
	SCM80211_CHAN_UNCHANGE,
	SCM80211_CHAN_CHANGED,
	SCM80211_CHAN_MULTI,
};

/**
 * struct scm80211_chanctx - channel context driver data
 *
 * @vif: viftual interface requesting channel (may be NULL)
 * @ctx: &struct ieee80211_chanctx_conf
 * @start_time: preferred start time
 * @dwell_time: duration to remain on the channel specified by @ctx
 * @timeout: timeout
 * @type: request type
 */
struct scm80211_chanctx {
	struct ieee80211_chanctx_conf ctx;

	struct ieee80211_vif *vif;
	int id;
	bool offchan;
	bool mac80211;
	bool restart;		/* automatic restart to simplify implementation */
	struct list_head assigned_vif;
	spinlock_t lock;

	unsigned long start_time;
	unsigned long dwell_time;

	unsigned long last_start_jiffies;
	unsigned long air_time;

	unsigned long timeout;
	enum scm80211_chanctx_type type;

	void (*begin)(struct ieee80211_hw *hw, struct scm80211_chanctx *ctx);
	void (*end)(struct ieee80211_hw *hw, struct scm80211_chanctx *ctx);
#ifdef SCM80211_DFS_CERTIFIED
	struct scm80211_dfs *dfs;
#endif				// SCM80211_DFS_CERTIFIED

	struct list_head sched_list;
	unsigned int state;
	int status;
	int retval;

	struct phy_vector pv;
	void *private;
};

static inline size_t scm80211_chanctx_data_size(void)
{
	return sizeof(struct scm80211_chanctx) - sizeof(struct ieee80211_chanctx_conf);
}

/**
 * enum scm80211_chanctx_sched_state
 *
 * @SCM80211_CHANCTX_SCHED_IDLE: scheduler is idle
 * @SCM80211_CHANCTX_SCHED_WAITING: scheduler is waiting for current chanctx to end
 * @SCM80211_CHANCTX_SCAN: scheduler is switching channel
 */

enum scm80211_chanctx_sched_state {
	SCM80211_CHANCTX_SCHED_IDLE = 0,
	SCM80211_CHANCTX_SCHED_WAITING,
	SCM80211_CHANCTX_SCHED_SWITCHING,
};

/**
 * struct scm80211_chanctx_sched - chanctx scheduler structure
 */
struct scm80211_chanctx_sched {
	struct scm80211_chanctx *cur_chan;
	struct workqueue_struct *workqueue;
	struct delayed_work work;
	struct {
		struct list_head head;
		atomic_t nr_chanctx;
	} queue[CHANCTX_TYPE_MAX];
	struct scm80211_chanctx idle;
	struct mutex mutex;
	bool stopped;
	atomic_t nr_chanctx;
	atomic_t nr_pending;
	struct rb_root root;
	enum scm80211_chanctx_sched_state state;
	struct idr idr;
};

struct scm80211_scanner {
	struct scm80211_chanctx chanctx;
	struct ieee80211_scan_request *req;
	struct ieee80211_vif *vif;
	u8 addr[ETH_ALEN];
	int index;
	bool scanning;
	bool scanned;
};

/**
 * struct soft_rxdesc - software rx descriptor
 * @index: ring index
 * @skb: skb
 * @page: do we need this?
 * @dma_addr: will be put to skb->cb
 * @size: not necessary
 * @hwdesc: struct rxdesc
 */
struct scm80211_rxdesc {
	int id;
	bool end;
#ifdef SCM80211_RX_PAGE_ALLOC
	struct page *page;
	dma_addr_t dma_addr;
#endif
	struct sk_buff *skb;
	__iomem struct rxdesc *config;
};

/**
 * struct scm80211_bus_ops - host-target interface
 *
 * @writel: write to a mac register
 * @readl: read from a mac register
 * @phy_writel: write to a phy register
 * @phy_readl: read from a phy register
 * @rf_writel: write to a rf register
 * @rf_readl: read from a rf register
 * @raw_writel: write to target register
 * @raw_readl: read from target register
 * @remap: map the host address to the target address
 * @unmap: unmap the host address from the target address
 * @inverse_map: get host address from target address
 *
 * @init_tx: bus dependent tx init
 * @deint_tx: bus dependent tx deint
 * @vif_init_tx: notify target init vif tx
 * @vif_deinit_tx: notify target deinit vif tx
 * @register_txq: register txq to target
 * @deregister_txq: degisetrt txq from target
 * @cancel_tx: notify target to abort tx
 * @hwq_alloc_dma: alloc dma for mdesc,bdesc for each hwq
 * @hwq_free_dma: free mdesc bdesc for each hwq
 * @fixup_tx: bus dependent tx fixup
 * @lookup_tx: lookup tx info
 * @start_tx: issue txop to target
 * @txop_len_limit: max_txop len limit by target
 * @prepare_tx: setup buff desc and mdesc desc
 * @cleanup_tx: unmap skb dma used for bus which support dma
 *
 * @init_rx: info target rx start and notice target to do rx init
 * @deinit_rx: notice target to do rx deinit
 * @rx_start: init rx_data struct for each skb
 * @refill_rxring: attach skb to rxring
 *
 * @register_bss_info: register bss info to target
 * @deregister_bss_info: deregister bss info from target
 *
 * @phy_start: info target to do phy start progress
 * @phy_stop: info target to do phy stop progress
 * @set_channel: set channel
 *
 * @set_fw_debug: set debug switch to taget fw
 *
 * @power_save
 * @vif_power_save
 *
 * @priv: bus-dependent data
 */
struct scm80211_bus_ops {
	/* common callback */
	void	(*mac_writel)(struct ieee80211_hw *hw, unsigned int val, unsigned int offset);
	u32	(*mac_readl)(struct ieee80211_hw *hw, unsigned int offset);
	void	(*phy_writel)(struct ieee80211_hw *hw, unsigned int val, unsigned int offset);
	u32	(*phy_readl)(struct ieee80211_hw *hw, unsigned int offset);
	void	(*rf_writel)(struct ieee80211_hw *hw, unsigned int val, unsigned int offset);
	u32	(*rf_readl)(struct ieee80211_hw *hw, unsigned int offset);
	void	(*pta_writel)(struct ieee80211_hw *hw, unsigned int val, unsigned int offset);
	u32	(*pta_readl)(struct ieee80211_hw *hw, unsigned int offset);
	void	(*raw_writel)(struct ieee80211_hw *hw, unsigned int val, unsigned int addr);
	u32	(*raw_readl)(struct ieee80211_hw *hw, unsigned int addr);
	void	(*efuse_writel)(void *dev_ep, void *val, unsigned int addr, unsigned int len);
	int	(*efuse_readl)(void *dev_ep, void *val, unsigned int addr, unsigned int len);
	void	(*param_write)(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			       int event, void *data, size_t len);

	phys_addr_t (*remap)(struct ieee80211_hw *hw, dma_addr_t start, size_t len);
	void	(*unmap)(struct ieee80211_hw *hw, dma_addr_t start, size_t len);
	dma_addr_t (*inverse_map)(struct ieee80211_hw *hw, phys_addr_t addr);

	/* tx */
	int	(*init_tx)(struct ieee80211_hw *hw);
	void	(*deinit_tx)(struct ieee80211_hw *hw);
	int	(*vif_init_tx)(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
	void	(*vif_deinit_tx)(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
	int	(*register_txq)(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				struct ieee80211_txq *txq);
	void	(*deregister_txq)(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				  struct ieee80211_txq *txq);
	int	(*cancel_tx)(struct ieee80211_hw *hw, struct txop *txop, bool discard);
	int	(*hwq_alloc_dma)(struct ieee80211_hw *hw, struct scm80211_hwq *hwq);
	void	(*hwq_free_dma)(struct ieee80211_hw *hw, struct scm80211_hwq *hwq);
	void	(*fixup_tx)(struct ieee80211_hw *hw, struct txop *txop);
	int	(*lookup_tx)(struct ieee80211_hw *hw, struct txop *txop);
	bool	(*start_tx)(struct ieee80211_hw *hw, struct txop *txop);
	int	(*prepare_tx)(struct ieee80211_hw *hw, struct txop *txop, int idx,
			      struct sk_buff *skb);
	int	(*cleanup_tx)(struct ieee80211_hw *hw, struct txop *txop, int idx);
	int	(*txop_len_limit)(struct ieee80211_hw *hw, struct txop *txop, int n_frag,
				  u32 total_len);

	/* rx */
	int	(*init_rx)(struct ieee80211_hw *hw);
	void	(*deinit_rx)(struct ieee80211_hw *hw);
	int	(*rx_start)(struct ieee80211_hw *hw, struct sk_buff *skb,
			    struct scm80211_rx_data *rx);
	int	(*refill_rxring)(struct ieee80211_hw *hw);

	/* cfg */
	int	(*associated)(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			      struct ieee80211_bss_conf *bss_info);
	void	(*disconnected)(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				struct ieee80211_bss_conf *bss_info);

	/* rf/phy */
	int	(*phy_start)(struct ieee80211_hw *hw);
	void	(*phy_stop)(struct ieee80211_hw *hw);
	int	(*set_channel)(struct ieee80211_hw *hw, struct scm80211_phy *phy,
			       struct phy_vector *pv);
	int	(*set_ant_sel)(struct ieee80211_hw *hw, int ant);
	int	(*get_ant_sel)(struct ieee80211_hw *hw);
	int	(*set_param_b)(struct ieee80211_hw *hw, int *buff, int length);
	int	(*get_param_b)(struct ieee80211_hw *hw, int *buff, int length);
	int (*get_tssi)(struct ieee80211_hw *hw);

	/* power save */
	int	(*power_save)(struct ieee80211_hw *hw, bool ps, bool link);
	int	(*vif_power_save)(struct ieee80211_hw *hw, struct ieee80211_vif *vif, bool enable);
	int	(*next_beacon)(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			       struct sk_buff *beacon);

#ifdef SCM80211_WOW
	/* wow */
	int	(*wow_set_wakeup)(struct ieee80211_hw *hw, struct scm80211_wow_context *wow);
#endif

	/* others */
	int	(*set_fw_debug)(struct ieee80211_hw *hw, u32 debug_info);

	// int	(*alloc_tx_urb)(struct ieee80211_hw *hw, struct scm80211_hwq *hwq);
	// void	(*free_tx_urb)(struct ieee80211_hw *hw, struct scm80211_hwq *hwq);
	// void	(*prepare_rx)(struct ieee80211_hw *hw);
	// void	(*tx_status)(struct ieee80211_hw *hw);
};

struct scm80211_bus {
	const struct scm80211_bus_ops *ops;
	u8 data[0] __aligned(sizeof(void *));
};
extern const struct scm80211_bus_ops *scm_bus_ops;

#if KERNEL_VERSION(4, 11, 0) > LINUX_VERSION_CODE
DECLARE_EWMA(tx_sched_time, 1024, 8);
#else
DECLARE_EWMA(tx_sched_time, 10, 8);
#endif

enum ps_mode {
	PS_DISABLED,
	PS_ENABLED,
	PS_AUTO_POLL,
	PS_MANUAL_POLL
};

/**
 * struct ieee80211_hw_priv - ieee80211_hw driver data
 *
 * @cur_chan: current channel that @hw is tuned to.
 * @devid - wiphy instance id (for multiple wiphy)
 *
 */
struct ieee80211_hw_priv {
	struct ieee80211_hw *hw;
	struct device *dev;
	int irq;
	int devid;
	bool beacon_offchan;

#ifdef SCM80211_RFKILL_POLL
	/* hw radio and rf state */
	bool rfkill_off;
#endif				/* SCM80211_RFKILL_POLL */

	struct mac_address addresses[SCM80211_NUM_VIF + 1];

	int channels, idx;
	bool use_chanctx;
	bool destroy_on_close;
	u32 portid;
	char alpha2[2];
	const struct ieee80211_regdomain *regd;

	struct scm80211_scanner scanner;
	struct scm80211_chanctx_sched sched;
	struct {
		struct scm80211_chanctx chanctx;
		struct ieee80211_vif *vif;
		bool running;
	} roc;

	struct {
		struct scm80211_chanctx chanctx;
		struct ieee80211_vif *vif;
		struct ieee80211_sta *sta;	/* TDLS Peer sta */
		u64 packets;
		u16 switch_time;
		u16 switch_timeout;
		bool started;	/* chan switch started */
		enum scm80211_tdls_state state;
		struct hrtimer tdls_timer;
		struct work_struct work;
	} tdls;

	struct {
		struct ieee80211_channel *channel;
		unsigned long next_start, start, end;
	} survey_data[60];

	struct {
		pm_regs(rxring,	2);
		pm_regs(doze_awake, 16);
		pm_regs(assoc_active, 20);
		pm_regs(phy_assoc_active, 18);
		enum ps_mode ps;
		bool doze;
		bool enable;
		bool acked;
		bool link;
		bool force_awake;
		spinlock_t lock;
		int  dynamic_ps_timeout;
		struct ieee80211_vif *vif;
		struct timer_list timer;
		struct work_struct enable_work;
		struct work_struct disable_work;
	} pm;

	u64 beacon_int	/* beacon interval in us */;
	unsigned int rx_filter;
	bool started, idle, scanning, hw_idle;
	struct mutex mutex;
	struct hrtimer phy_timer;
	u8 rx_statistic;

	struct timer_list beacon_timer;

#ifdef SCM80211_SMPS
	enum smps_mode {
		SMPS_OFF = 0,
		SMPS_DYNAMIC,
		SMPS_STATIC
	} smps;
#endif

#ifdef CONFIG_MAC80211_DEBUGFS
	struct dentry *debugfs_hwsim;
	struct dentry *debugfs_priv;
	u32 reg;
	int cnt;
	int reg_type;
#endif
	/*
	 * Only radios in the same group can communicate together (the
	 * channel has to match too). Each bit represents a group. A
	 * radio can be in more than one group.
	 */
	u64 group;

	/* difference between this hw's clock and the real clock, in usecs */
	s64 tsf_offset;
	s64 bcn_delta;
	/* absolute beacon transmission time. Used to cover up "tx" delay. */
	u64 abs_bcn_ts;

	/* Stats */
	u64 tx_pkts;
	u64 rx_pkts;
	u64 tx_bytes;
	u64 rx_bytes;
	u64 tx_dropped;
	u64 tx_failed;
	u64 tx_winlost;
	u64 total_winlost;
#ifdef SCM80211_RX_STATISTICS_DEBUG
	int rx_overrun;
#endif				// SCM80211_RX_STATISTICS_DEBUG
	struct ieee80211_low_level_stats stats;

	enum space_type space;

	/* scm80211 i/o regions */
	// void __iomem *mac_base[2];
	// void __iomem *phy_base;
	// void __iomem *rf_base;
	// void __iomem *macbase;	//mac based address used by SW

	bool dbdc;
	int band;

	bool hwsim;
	unsigned long vif_map;
	unsigned long vif_map_ref;
	struct ieee80211_vif *vifs[SCM80211_NUM_VIF];
	char version_str[128];
	u16 debug_trace;

	/* tx */
	struct scm80211_hwq hwq[SCM80211_NUM_HWQ];
	atomic_t n_he_sta;
	struct hrtimer ratestats_timer;
	u8 coverage_class;
	int rdesc;		/* 1 to use remote descriptor, 0 to use host descriptor */
	struct ewma_tx_sched_time sched_time;
	bool local_rc_used;
	spinlock_t dequeue_lock;
	struct hrtimer cca_timer;
#ifdef SCM80211_COEX
	struct scm80211_coex coex;
#endif

	/* rx */
	struct tasklet_struct tasklet;
	struct {
		struct scm80211_rxdesc rxd[SCM80211_NUM_RX_RING];
		u32 order;	/* the number of ring element is 2^order */
		u32 bptr, rptr, wptr;
		struct work_struct work;
#ifdef SCM80211_RX_PAGE_ALLOC
		struct page *pages;
#endif
		struct sk_buff_head pool;
		struct sk_buff_head rxskbs;
		struct sk_buff_head frags;
		spinlock_t lock;
	} rxring;

	struct {
		struct net_device	netdev;
		struct napi_struct	context;
		bool			enabled;
		spinlock_t		lock;
	} napi;

#ifdef SCM80211_DFS_CERTIFIED
	struct scm80211_dfs_radars_info radar_info;
#endif
	/* monitor - STA interface only */
	struct {
		u32			timestamp;
		u32			ampdu_reference;
		u8			ra[ETH_ALEN];
		u8			ta[ETH_ALEN];
		struct scm80211_ampdu_info ampdu;
	} monitor;

	/* record MAC_REG_INTR_ENABLE */
	u32 int_en;
	u32 mac_ver;
	u32 phy_ver;
	/* phy */
	struct scm80211_pme *pme;	/* phy management entity (TBD) */
	struct scm80211_phy *phy;

	unsigned int n_chain;	/* number of allocated phys */
	unsigned long chains;	/* bitmap of allocated phys */
	unsigned long active_chains;	/* bitmap of active phy paths */
	spinlock_t lock;

	u32 phy_mandatory_rates[NUM_NL80211_BANDS];

	/* rf */
	struct scm80211_rf_device *rf;
	struct delayed_work temp_work;
	int temperature[2];
	int dynamic_temp_timer;		/* in second */
	struct work_struct power_detect_work;
	bool pdet_en;

	struct work_struct debug_work;

#ifdef SCM80211_WOW
	struct scm80211_wow_context wow;
#endif

#ifdef AMPDU_SZ_SLIDE
	u32 tx_ampdu_limit;
	u32 tx_fail_ampdu_sz;
	u16 tx_suc_cnt;
	u16 tx_fail_cnt;
#endif

	enum scm80211_bus_type {
		PLATFORM_BUS = 0,
		PCIE_BUS = 1,
		USB_BUS = 2,
		SDIO_BUS = 3,
	} bus_type;
	u16 bcdusb;

	struct scm80211_bus bus;	/* should be the last */
};

#define is_pcie_device(hw) (((struct ieee80211_hw_priv *)(hw)->priv)->bus_type == PCIE_BUS)
#define is_usb_device(hw) (((struct ieee80211_hw_priv *)(hw)->priv)->bus_type == USB_BUS)
#define is_highspeed_usb_device(hw) \
	((((struct ieee80211_hw_priv *)(hw)->priv)->bus_type == USB_BUS) && \
	 ((struct ieee80211_hw_priv *)(hw)->priv)->bcdusb == 0x200)
#define is_platform_device(hw) (((struct ieee80211_hw_priv *)(hw)->priv)->bus_type == PLATFORM_BUS)
#define is_sdio_device(hw) (((struct ieee80211_hw_priv *)(hw)->priv)->bus_type == SDIO_BUS)
#define hw_to_priv(hw) ((struct ieee80211_hw_priv *)(hw)->priv)

#define work_to_hw_priv(wk, field)				\
	container_of(wk, struct ieee80211_hw_priv, field)

static inline struct scm80211_bus *bus(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return &priv->bus;
}

static inline void *scm80211_get_bus_data(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return (void *)priv->bus.data;
}

static inline struct ieee80211_hw *scm80211_bus_to_hw(void *data)
{
	struct ieee80211_hw_priv *priv;

	priv = container_of(data, struct ieee80211_hw_priv, bus.data);
	return priv->hw;
}

/* FIXME: scm80211_{writel,readl} -> scm80211_mac_{writel,readl} */
static inline void scm80211_mac_writel(struct ieee80211_hw *hw, unsigned int val,
				       unsigned int offset)
{
	bus(hw)->ops->mac_writel(hw, val, offset);
}

static inline u32 scm80211_mac_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return bus(hw)->ops->mac_readl(hw, offset);
}

static inline void scm80211_phy_writel(struct ieee80211_hw *hw, unsigned int val,
				       unsigned int offset)
{
	bus(hw)->ops->phy_writel(hw, val, offset);
}

static inline u32 scm80211_phy_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return bus(hw)->ops->phy_readl(hw, offset);
}

static inline void scm80211_rf_writel(struct ieee80211_hw *hw, unsigned int val,
				      unsigned int offset)
{
	bus(hw)->ops->rf_writel(hw, val, offset);
}

static inline u32 scm80211_rf_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return bus(hw)->ops->rf_readl(hw, offset);
}

static inline void scm80211_pta_writel(struct ieee80211_hw *hw, unsigned int val,
				       unsigned int offset)
{
	bus(hw)->ops->pta_writel(hw, val, offset);
}

static inline u32 scm80211_pta_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return bus(hw)->ops->pta_readl(hw, offset);
}

static inline void scm80211_raw_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int addr)
{
	bus(hw)->ops->raw_writel(hw, val, addr);
}

static inline u32 scm80211_raw_readl(struct ieee80211_hw *hw, unsigned int addr)
{
	u32 v = bus(hw)->ops->raw_readl(hw, addr);
	return v;
}

static inline void scm80211_efuse_writel(void *dev_ep, void *val, unsigned int addr,
				    unsigned int len)
{
	scm_bus_ops->efuse_writel(dev_ep, val, addr, len);
}

static inline int scm80211_efuse_readl(void *dev_ep, void *val, unsigned int addr,
				    unsigned int len)
{
	return scm_bus_ops->efuse_readl(dev_ep, val, addr, len);
}

static inline void scm80211_param_write(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					int event, void *data, size_t len)
{
	if (bus(hw)->ops->param_write)
		bus(hw)->ops->param_write(hw, vif, event, data, len);
}

static inline phys_addr_t scm80211_remap(struct ieee80211_hw *hw, dma_addr_t start, size_t len)
{
	if (bus(hw)->ops->remap)
		return bus(hw)->ops->remap(hw, start, len);
	else
		return 0;
}

static inline void scm80211_unmap(struct ieee80211_hw *hw, dma_addr_t start, size_t len)
{
	if (bus(hw)->ops->unmap)
		bus(hw)->ops->unmap(hw, start, len);
}

static inline dma_addr_t scm80211_inverse_map(struct ieee80211_hw *hw, phys_addr_t addr)
{
	if (bus(hw)->ops->inverse_map)
		return bus(hw)->ops->inverse_map(hw, addr);

	return (dma_addr_t)(~0);
}

#define PM_LOCAL_TEST	0
static inline int scm80211_power_save(struct ieee80211_hw *hw, bool powersave, bool link)
{
	if (bus(hw)->ops->power_save)
		return bus(hw)->ops->power_save(hw, powersave, link);
	return 0;
}

static inline int scm80211_vif_power_save(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					  bool enable)
{
	if (bus(hw)->ops->vif_power_save)
		return bus(hw)->ops->vif_power_save(hw, vif, enable);
	return 0;
}

static inline int scm80211_calc_next_beacon(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					    struct sk_buff *beacon)
{
	if (bus(hw)->ops->next_beacon)
		return bus(hw)->ops->next_beacon(hw, vif, beacon);
	return 0;
}

static inline bool scm80211_start_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	return bus(hw)->ops->start_tx(hw, txop);
}

static inline int scm80211_bus_phy_start(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->phy_start)
		return bus(hw)->ops->phy_start(hw);
	return -EINVAL;
}

static inline int scm80211_bus_phy_stop(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->phy_stop) {
		bus(hw)->ops->phy_stop(hw);
		return 0;
	}
	return -EINVAL;
}

static inline int scm80211_set_channel(struct ieee80211_hw *hw, struct scm80211_phy *phy,
				       struct phy_vector *pv)
{
	if (bus(hw)->ops->set_channel)
		return bus(hw)->ops->set_channel(hw, phy, pv);
	return 0;
}

static inline int scm80211_set_ant_sel(struct ieee80211_hw *hw, int ant)
{
	if (bus(hw)->ops->set_ant_sel)
		return bus(hw)->ops->set_ant_sel(hw, ant);
	return 0;
}

static inline int scm80211_get_ant_sel(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->get_ant_sel)
		return bus(hw)->ops->get_ant_sel(hw);
	return -1;
}

static inline int scm80211_set_param_b(struct ieee80211_hw *hw, int *buff, int length)
{
	if (bus(hw)->ops->set_param_b)
		return bus(hw)->ops->set_param_b(hw, buff, length);
	return 0;
}

static inline int scm80211_get_param_b(struct ieee80211_hw *hw, int *buff, int length)
{
	if (bus(hw)->ops->get_param_b)
		return bus(hw)->ops->get_param_b(hw, buff, length);
	return 0;
}


static inline int scm80211_get_tssi(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->get_tssi)
		return bus(hw)->ops->get_tssi(hw);
	return 0;
}

static inline int scm80211_associated(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				      struct ieee80211_bss_conf *bss_info)
{
	if (bus(hw)->ops->associated)
		return bus(hw)->ops->associated(hw, vif, bss_info);

	return 0;
}

static inline void scm80211_disconnected(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					 struct ieee80211_bss_conf *bss_info)
{
	if (bus(hw)->ops->disconnected)
		bus(hw)->ops->disconnected(hw, vif, bss_info);
}

static inline int scm80211_bus_init_tx(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->init_tx)
		return bus(hw)->ops->init_tx(hw);
	return 0;
}

static inline void scm80211_bus_deinit_tx(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->deinit_tx)
		bus(hw)->ops->deinit_tx(hw);
}

static inline int scm80211_bus_vif_init_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	if (bus(hw)->ops->vif_init_tx)
		return bus(hw)->ops->vif_init_tx(hw, vif);
	return 0;
}

static inline void scm80211_bus_vif_deinit_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	if (bus(hw)->ops->vif_deinit_tx)
		bus(hw)->ops->vif_deinit_tx(hw, vif);
}

static inline int scm80211_register_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					struct ieee80211_txq *txq)
{
	if (bus(hw)->ops->register_txq)
		return bus(hw)->ops->register_txq(hw, vif, txq);
	return 0;
}

static inline void scm80211_deregister_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					   struct ieee80211_txq *txq)
{
	if (bus(hw)->ops->deregister_txq)
		bus(hw)->ops->deregister_txq(hw, vif, txq);
}

/**
 * scm80211_abort_backoff() - cancel previous issued txop
 *
 * @hwq: hardware queue to cancel
 *
 * This function cancels the previously issued txop. This function must be
 * called after disabling @hwq.
 */
static inline int scm80211_cancel_tx(struct ieee80211_hw *hw, struct txop *txop, bool discard)
{
	if (bus(hw)->ops->cancel_tx)
		return bus(hw)->ops->cancel_tx(hw, txop, discard);
	return 0;
}

static inline int scm80211_hwq_alloc_dma_desc(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	if (bus(hw)->ops->hwq_alloc_dma)
		return bus(hw)->ops->hwq_alloc_dma(hw, hwq);
	return 0;
}

static inline void scm80211_hwq_free_dma_desc(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	if (bus(hw)->ops->hwq_free_dma)
		bus(hw)->ops->hwq_free_dma(hw, hwq);
}

static inline int scm80211_lookup_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	if (bus(hw)->ops->lookup_tx)
		return bus(hw)->ops->lookup_tx(hw, txop);
	else
		return -EINVAL;
}

static inline int scm80211_prepare_tx(struct ieee80211_hw *hw, struct txop *txop,
				      int idx, struct sk_buff *skb)
{
	if (bus(hw)->ops->prepare_tx)
		return bus(hw)->ops->prepare_tx(hw, txop, idx, skb);
	return 0;
}

static inline int scm80211_cleanup_tx(struct ieee80211_hw *hw, struct txop *txop, int idx)
{
	if (bus(hw)->ops->cleanup_tx)
		return bus(hw)->ops->cleanup_tx(hw, txop, idx);
	return 0;
}

#ifdef SCM80211_WOW
static inline int scm80211_wow_set_wakeup(struct ieee80211_hw *hw, struct scm80211_wow_context *wow)
{
	if (bus(hw)->ops->wow_set_wakeup)
		return bus(hw)->ops->wow_set_wakeup(hw, wow);
	return 0;
}
#endif

#ifdef TODO_TO_REMOVE
static inline int scm80211_setup_bdesc_dma(struct ieee80211_hw *hw, struct txop *txop,
					   struct tx_buffer_desc *bdesc,
					   struct scm80211_tx_info *xti, struct sk_buff *skb)
{
	if (bus(hw)->ops->setup_bdesc_dma)
		return bus(hw)->ops->setup_bdesc_dma(hw, txop, bdesc, xti, skb);
	return 0;
}

static inline void scm80211_setup_mdesc_priv(struct ieee80211_hw *hw, struct txop *txop,
					     struct tx_mpdu_desc *mdesc,
					     struct scm80211_tx_info *xti, struct sk_buff *skb)
{
	if (bus(hw)->ops->setup_mdesc_priv)
		bus(hw)->ops->setup_mdesc_priv(hw, txop, mdesc, xti, skb);
}

static inline int scm80211_unmap_skb_sg(struct ieee80211_hw *hw, struct scm80211_tx_info *xti)
{
	if (bus(hw)->ops->unmap_skb_sg)
		return bus(hw)->ops->unmap_skb_sg(hw, xti);
	return 0;
}
#endif

static inline int scm80211_bus_init_rx(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->init_rx)
		return bus(hw)->ops->init_rx(hw);
	return 0;
}

static inline void scm80211_bus_deinit_rx(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->deinit_rx)
		bus(hw)->ops->deinit_rx(hw);
}

static inline int scm80211_bus_rx_start(struct ieee80211_hw *hw, struct sk_buff *skb,
					struct scm80211_rx_data *rx)
{
	if (bus(hw)->ops->rx_start)
		return bus(hw)->ops->rx_start(hw, skb, rx);
	return 0;
}

static inline int scm80211_refill_rxring(struct ieee80211_hw *hw)
{
	if (bus(hw)->ops->refill_rxring)
		return bus(hw)->ops->refill_rxring(hw);
	return 0;
}

static inline int scm80211_set_fw_debug(struct ieee80211_hw *hw, u32 debug_info)
{
	if (bus(hw)->ops->set_fw_debug)
		return bus(hw)->ops->set_fw_debug(hw, debug_info);
	return 0;
}

#ifdef TODO_TO_REMOVE
static inline bool scm80211_alloc_tx_urb(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	return bus(hw)->ops->alloc_tx_urb(hw, hwq);
}

static inline void scm80211_free_tx_urb(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	bus(hw)->ops->free_tx_urb(hw, hwq);
}

static inline void scm80211_prepare_rx(struct ieee80211_hw *hw)
{
	bus(hw)->ops->prepare_rx(hw);
}

static inline void scm80211_prepare_tx_status(struct ieee80211_hw *hw)
{
	bus(hw)->ops->tx_status(hw);
}
#endif

static inline int scm80211_tx_bus_limit(struct ieee80211_hw *hw, struct txop *txop,
					int n_frags, u32 len)
{
	if (bus(hw)->ops->txop_len_limit)
		return bus(hw)->ops->txop_len_limit(hw, txop, n_frags, len);
	return 0;
}

#define SCM80211_TSF_TIME0_IRQ	BIT(0)
#define SCM80211_TSF_TIME1_IRQ	BIT(1)
#define TSF_OFFSET	25000	// 25ms before tbtt
#define TSF_PERIOD	40000	//wake 40ms
#define TSF_MASK 0xFFFFFFFF
#define TSF_MODULO  0x100000000
static inline bool scm80211_tsf_less(u32 tsf0, u32 tsf1)
{
	return ((tsf0 - tsf1) & TSF_MASK) > (TSF_MODULO >> 1);
}

static inline u32 scm80211_mac_version(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (priv->mac_ver == 0)
		priv->mac_ver = scm80211_mac_readl(hw, MAC_REG_VER_INFO);
	return priv->mac_ver;
}

static inline u32 scm80211_phy_version(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (priv->phy_ver == 0)
		priv->phy_ver = phy_readl(&priv->phy[priv->devid], PHY_VER_INFO);
	return priv->phy_ver;
}

static inline enum scm80211_twt_state scm80211_get_twt_state(struct ieee80211_vif *vif)
{
#ifdef SCM80211_TWT
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	return vifp->twt_state;
#else
	return TWT_DISABLED;
#endif
}

#ifdef SCM80211_TWT
bool scm80211_set_twt_state(struct ieee80211_vif *vif, enum scm80211_twt_state state);
void scm80211_twt_state_work(struct work_struct *work);
enum hrtimer_restart scm80211_twt_tsf_timer(struct hrtimer *timer);
int  ieee80211_parse_twt(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			 const u8 *twt, u8 len, struct ieee80211_sta *ta);
#endif /* SCM80211_TWT */

#define word(ptr, n)		(((__le32 *)(ptr)) + n)

#define ring_size(_priv)         (1 << (_priv)->order)
#define ring_ptr(_priv, _ptr)     ((_ptr) & GENMASK((_priv)->order, 0))
#define ring_index(_priv, _ptr)   ((_ptr) & GENMASK((_priv)->order - 1, 0))

static inline bool ring_ptr_equal(struct ieee80211_hw_priv *priv, u32 ptr1, u32 ptr2)
{
	return ring_ptr(&priv->rxring, ptr1) == ring_ptr(&priv->rxring, ptr2);
}

#ifdef SCM80211_RX_PAGE_ALLOC
#define SCM80211_RX_PAGE_ORDER		get_order(IEEE80211_RX_HEADROOM + rx_buflen)
#define SCM80211_RX_PAGE_SIZE		(PAGE_SIZE << SCM80211_RX_PAGE_ORDER)
#endif

#define SCM80211_SKB_RXCB(_skb)	((struct scm80211_rx_info *)(_skb)->cb)

#define product(ver)    (((ver) & 0xFFFF0000) >> 16)
#define version(ver)    (((ver) & 0x0000FFF0) >> 4)
#define is_chip(ver)    (((ver) & 0x0000000F) == 0xC ? 1 : 0)

/* main */
int scm80211_vdev_id(struct ieee80211_vif *vif);
int scm80211_vif_chan(struct ieee80211_vif *vif, struct cfg80211_chan_def *def);
extern struct ieee80211_ops scm80211_ops;
#ifdef SCM2020_A0
void scm2020_set_gpio_config(struct ieee80211_hw *hw, u32 pin, u32 mode,
			     enum scm2020_gpio_dir dir, bool val);
#endif
int scm80211_init_hw(struct ieee80211_hw *hw, u8 *addr);
int scm80211_deinit_hw(struct ieee80211_hw *hw);
void scm80211_set_addr(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u8 addr[ETH_ALEN]);
void scm80211_set_mandatory_hwrates(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				    enum nl80211_band band, bool erp);

irqreturn_t scm80211_handle_irq(int irq, void *data);

void scm80211_tsf_irq(struct ieee80211_hw *hw, u32 v);
void scm80211_handle_muedca_irq(struct ieee80211_hw *hw, u32 val);
void scm80211_set_basic_rates(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			      struct ieee80211_bss_conf *info);

void scm80211_config_bssid_filter(struct ieee80211_hw *hw, struct ieee80211_vif *vif, bool enable);
void scm80211_set_p2p_basic_rate(struct ieee80211_hw *hw, int vdev);
void scm80211_set_mac_channel(struct ieee80211_hw *hw, struct cfg80211_chan_def *chandef);
void scm80211_set_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx);
void scm80211_set_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u64 tsf);
u64 scm80211_get_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
u32 scm80211_get_tsf32(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_enable_tsf_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_disable_tsf_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

void scm80211_rfkill_init(struct ieee80211_hw *hw);
void scm80211_rfkill_deinit(struct ieee80211_hw *hw);
int _scm80211_set_key(struct ieee80211_hw *hw, enum set_key_cmd cmd,
		      struct ieee80211_vif *vif,
		      struct ieee80211_sta *sta,
		      struct ieee80211_key_conf *key,
		      bool by_idx);
void scm80211_register_oops_handler(void);
void scm80211_unregister_oops_handler(void);

#define SCM80211_UNLINK_TIMEOUT	10
#define SCM80211_UNLINK_SCAN	5000
/* power save */
static inline bool scm80211_ps_enabled(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return (priv->pm.vif && priv->pm.ps == PS_ENABLED);
}

void scm80211_disable_ps(struct ieee80211_hw *hw);
void scm80211_enable_ps(struct ieee80211_hw *hw, bool link);
void scm80211_vif_disable_ps(struct ieee80211_vif *vif);
void scm80211_vif_enable_ps(struct ieee80211_vif *vif);
void scm80211_set_debug_trace(struct ieee80211_hw *hw);
int scm80211_hw_band_set(struct ieee80211_hw *hw);
struct ieee80211_vif *scm80211_get_vif_from_vedv(struct ieee80211_hw *hw, int vdev);
#ifdef SCM80211_WOW
int scm80211_wow_op_suspend(struct ieee80211_hw *hw, struct cfg80211_wowlan *wowlan);
void scm80211_wow_op_set_wakeup(struct ieee80211_hw *hw, bool enabled);
int scm80211_wow_op_resume(struct ieee80211_hw *hw);
#endif
void scm80211_smps_mode_set(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx);
void scm80211_setup_link_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_cleanup_link_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_setup_unlink_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u16 msecs);
void scm80211_cleanup_unlink_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_init_pm(struct ieee80211_hw *hw);
void scm80211_deinit_pm(struct ieee80211_hw *hw);
void scm80211_dynamic_ps_disable(struct ieee80211_hw *hw, bool force);
int scm80211_wakeup(struct ieee80211_hw *hw);
int scm80211_try_to_doze(struct ieee80211_hw *hw);

/* tx */
void scm80211_enable_tx_irq(struct ieee80211_hw *hw, int queues);
void scm80211_disable_tx_irq(struct ieee80211_hw *hw, int queues);
void scm80211_init_tx(struct ieee80211_hw *hw);
void scm80211_deinit_tx(struct ieee80211_hw *hw);
int scm80211_vif_init_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_vif_deinit_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_vif_map_hwqueue(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_init_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		       struct ieee80211_txq *txq, bool internal);
void scm80211_deinit_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			 struct ieee80211_txq *txq);
int scm80211_conf_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		     u16 queue, const struct ieee80211_tx_queue_params *params);
void scm80211_wake_tx_queue(struct ieee80211_hw *hw, struct ieee80211_txq *txq);
void scm80211_tx(struct ieee80211_hw *hw, struct ieee80211_tx_control *control,
		 struct sk_buff *skb);
void scm80211_tx_channel(struct ieee80211_hw *hw, struct sk_buff *skb,
			 struct ieee80211_channel *chan);
int scm80211_sta_set_txpwr(struct ieee80211_hw *hw,
			   struct ieee80211_vif *vif, struct ieee80211_sta *sta);
void scm80211_flush(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u32 queues, bool drop);
void scm80211_handle_txop_complete(struct ieee80211_hw *hw, struct txop *txop);
void scm80211_tx_irq(struct ieee80211_hw *hw, unsigned long queues);
struct txop *scm80211_get_txop(struct scm80211_hwq *hwq, int state);

int scm80211_ampdu_tx_action(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			     struct ieee80211_ampdu_params *params);
void scm80211_sta_rate_tbl_update(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				  struct ieee80211_sta *sta);

unsigned long scm80211_vif_queues(struct ieee80211_vif *vif);
bool scm80211_vif_tx_frames_pending(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
bool scm80211_tx_frames_pending(struct ieee80211_hw *hw);
void scm80211_wake_queues(struct ieee80211_hw *hw, unsigned long queues);
void scm80211_stop_queues(struct ieee80211_hw *hw, unsigned long queues);
enum ppdu_format txvector_format(struct ieee80211_tx_rate *rate);
enum cbw txvector_ch_bandwidth(struct ieee80211_tx_rate *rate);
struct scm80211_hwq *scm80211_hwq(struct ieee80211_hw *hw, int queue);
void scm80211_flush_hwq(struct scm80211_hwq *hwq, bool drop);
void scm80211_enable_mu_edca(struct scm80211_hwq *hwq);
enum hrtimer_restart scm80211_ratestats_timer(struct hrtimer *t);
void scm80211_enable_trig_qnull_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_disable_trig_qnull_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_vif_queues_flush(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_tx_start_irq(struct ieee80211_hw *hw);

void __scm80211_txop_change_state(struct txop *txop, int state);
struct txop *__scm80211_get_txop(struct scm80211_hwq *hwq, int state);
void __scm80211_txop_add(struct txop *txop, int state);
void __scm80211_txop_add_back(struct txop *txop);
void scm80211_txop_add_back(struct txop *txop);
void scm80211_complete_txop(struct txop *txop, bool discard);
bool scm80211_txop_is_within_transmit_window(struct txop *txop);
void scm80211_hwq_start_watchdog(struct scm80211_hwq *hwq);
int scm80211_ba_buf_size(struct scm80211_txq *txq);
void scm80211_init_ampdu(struct txop *txop);
int scm80211_ampdu_mmss_bytes(struct ieee80211_hw *hw, struct ieee80211_sta *sta,
				     u32 bitrate);

int scm80211_net_dev_attach(struct ieee80211_hw *hw, struct ieee80211_vif *vif);


static inline struct scm80211_tx_info *txop_cur_xti(struct txop *txop)
{
	int n = skb_queue_len(&txop->sched);

	WARN_ON(n > ARRAY_SIZE(txop->xti));
	if (n == ARRAY_SIZE(txop->xti))
		return NULL;
	return &txop->xti[n];
}

static inline struct tx_mpdu_desc *txop_get_mdesc(struct txop *txop, int idx)
{
	BUG_ON(!txop);
	return &txop->dma[0].desc->mdesc[idx];
}

/**
 * txop_get_bdesc() - returns the bdesc of an mpdu @idx
 */
static inline struct tx_buffer_desc *txop_get_bdesc(struct txop *txop, int idx)
{
	BUG_ON(!txop);
	return &txop->dma[0].desc->mpdu[idx].bdesc[0];
}

/* rx */
extern const size_t rx_buflen;
extern int n_rx;

int scm80211_init_rx(struct ieee80211_hw *hw);
void scm80211_deinit_rx(struct ieee80211_hw *hw);
int scm80211_ampdu_rx_action(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			     struct ieee80211_ampdu_params *params);
void scm80211_wait_for_rx_idle(struct ieee80211_hw *hw);
struct sk_buff *scm80211_alloc_rx_skb(struct ieee80211_hw *hw);
void scm80211_refill_rxpool(struct ieee80211_hw *hw);
void scm80211_rx_statistics_update(struct ieee80211_hw *hw);
void scm80211_enable_rx_irq(struct ieee80211_hw *hw);
void scm80211_disable_rx_irq(struct ieee80211_hw *hw);
static inline int rxd_index(struct ieee80211_hw *hw, struct scm80211_rxdesc *rxd)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	return rxd - &priv->rxring.rxd[0];
}

bool scm80211_start_backoff(struct txop *txop);
struct txop *scm80211_txop_alloc(struct scm80211_hwq *hwq);
void scm80211_txop_free(struct txop *txop);
void scm80211_txop_attach(struct txop *txop, struct scm80211_txq *txq);
int scm80211_schedule_txop(struct txop *txop);
int scm80211_num_pending_frames(struct scm80211_txq *txq, bool buffered);
bool scm80211_txop_ring_empty(struct scm80211_hwq *hwq);
bool scm80211_txop_is_running(struct scm80211_hwq *hwq);
struct scm80211_txq *scm80211_get_next_txq(struct scm80211_hwq *hwq);
void scm80211_tx_power_set(struct ieee80211_hw *hw, int channel, int power);
int scm80211_dequeue_pending_frames(struct scm80211_txq *txq);
void scm80211_cleanup_txq(struct scm80211_txq *txq);


void scm80211_txop_add(struct txop *txop, int state);
void scm80211_txop_change_state(struct txop *txop, int state);
bool scm80211_has_txop(struct scm80211_hwq *hwq, int state);

/* chan */
int scm80211_init_chanctx_sched(struct ieee80211_hw *hw);

void scm80211_deinit_chanctx_sched(struct ieee80211_hw *hw);

int scm80211_add_chanctx(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx);

void scm80211_remove_chanctx(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx);

void scm80211_change_chanctx(struct ieee80211_hw *hw,
			     struct ieee80211_chanctx_conf *ctx, u32 changed);

int scm80211_assign_vif_chanctx(struct ieee80211_hw *hw,
				struct ieee80211_vif *pubvif, struct ieee80211_chanctx_conf *ctx);

void scm80211_unassign_vif_chanctx(struct ieee80211_hw *hw,
				   struct ieee80211_vif *vif, struct ieee80211_chanctx_conf *ctx);

int scm80211_switch_vif_chanctx(struct ieee80211_hw *hw,
				struct ieee80211_vif_chanctx_switch *vifs,
				int n_vifs, enum ieee80211_chanctx_switch_mode mode);

int scm80211_hw_scan(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		     struct ieee80211_scan_request *scan_req);

void scm80211_cancel_hw_scan(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

int scm80211_roc(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		 struct ieee80211_channel *chan, int duration, enum ieee80211_roc_type type);

int scm80211_croc(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

int scm80211_vif_update_chanctx(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				unsigned long changes);

void scm80211_mgd_tdls_discover(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

int scm80211_tdls_channel_switch(struct ieee80211_hw *hw,
				 struct ieee80211_vif *vif,
				 struct ieee80211_sta *sta, u8 oper_class,
				 struct cfg80211_chan_def *chandef,
				 struct sk_buff *tmpl_skb, u32 ch_sw_tm_ie);

void scm80211_tdls_cancel_channel_switch(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif, struct ieee80211_sta *sta);

void scm80211_tdls_recv_channel_switch(struct ieee80211_hw *hw,
				       struct ieee80211_vif *vif,
				       struct ieee80211_tdls_ch_sw_params *params);

void scm80211_clr_tdls(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

bool scm80211_start_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx);

struct ieee80211_vif *scm80211_get_other_vif(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

bool scm80211_is_scanning(struct ieee80211_hw *hw);
bool scm80211_scan_finished(struct ieee80211_hw *hw);
bool scm80211_is_offchannel(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
int scm80211_send_nullfunc(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			   struct cfg80211_chan_def *def, bool ps);
struct scm80211_chanctx *scm80211_current_chanctx(struct ieee80211_hw *hw);
struct scm80211_chanctx_sched *hw_sched(struct ieee80211_hw *hw);

/* phy */
u32 scm80211_ppdu_txtime(struct txvector *tv, enum nl80211_band);
int scm80211_bitrate(struct txvector *tv);
int he_constellation_index(int mcs);
int he_n_sym(struct txvector *tv, int *a);

/* rf */
int scm80211_init_rfi(struct ieee80211_hw *hw);
void scm80211_deinit_rfi(struct ieee80211_hw *hw);
struct spi_master *scm80211_get_rfi_spi_master(struct ieee80211_hw *hw, int busnum);

struct rf *scm80211_find_rf(struct ieee80211_hw *hw, char *name);
int scm80211_attach_rf(struct ieee80211_hw *hw, char *name);
void scm80211_detach_rf(struct ieee80211_hw *hw);

void scm2020_powersave(struct ieee80211_hw *hw, bool ps);

int scm80211_rf_config(struct ieee80211_hw *hw, int frequency, int bandwidth);
int scm80211_rf_power_up(struct ieee80211_hw *hw);
int scm80211_rf_power_down(struct ieee80211_hw *hw);
int scm80211_rx_poll(struct ieee80211_hw *hw, int budget);
void scm2020_rfreg_init(struct ieee80211_hw *hw);

void scm2020_set_comp_en(bool en);
void scm2020_pdet_set_para_b(int rf_path, int rf_2g, int rf_5g_low,
			     int rf_5g_mid, int rf_5g_high);
void scm2020_pdet_get_para_b(int rf_path, int *pdet);
void scm2020_set_xtal_cal_b(int xcal);
void scm2020_phy_temp_comp(struct ieee80211_hw *hw);
void scm2020_set_phy_dig_gain(struct ieee80211_hw *hw, int channel, int power);

/* debug */
void scm80211_snprint_frame(char *buf, size_t len, struct ieee80211_hdr *hdr);
void scm80211_print_frame(struct ieee80211_hw *hw, struct ieee80211_hdr *hdr, size_t len);

/* rc */
#ifdef CONFIG_SCM80211_RC
int scm80211_minstrel_init(void);
void scm80211_minstrel_exit(void);
#else
static inline int scm80211_minstrel_init(void)
{
	return 0;
}

static inline void scm80211_minstrel_exit(void)
{
}
#endif

#ifdef SCM80211_VENDOR_CMD
void scm80211_register_vendor_cmds(struct ieee80211_hw *hw);
#else
#define scm80211_register_vendor_cmds(hw)	((void)(hw))
#endif

void scm80211_omi_work(struct work_struct *work);
#ifdef CONFIG_MAC80211_DEBUGFS
/* degbugfs*/
void scm80211_vif_register_dbgfs(struct ieee80211_vif *vif);
void scm80211_vif_unregister_dbgfs(struct ieee80211_vif *vif);
void scm80211_register_dbgfs(struct ieee80211_hw *hw);
void scm80211_unregister_dbgfs(struct ieee80211_hw *hw);
void scm80211_sta_add_debugfs(struct ieee80211_hw *hw,
			      struct ieee80211_vif *vif,
			      struct ieee80211_sta *sta, struct dentry *dir);
#endif /*#CONFIG_MAC80211_DEBUGFS */

#endif /* __SCM80211_CORE_H__ */
