// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
#include <linux/types.h>
#include <linux/workqueue.h>
#include <linux/slab.h>
#include <linux/regmap.h>
#include <net/mac80211.h>

const char *mgt_subtype_name[] = {
	"Assoc Req",
	"Assoc Resp",
	"Reassoc req",
	"Reassoc resp",
	"Probe req",
	"Probe resp",
	"reserved#6",
	"reserved#7",
	"Beacon",
	"ATIM",
	"Disassoc",
	"Auth",
	"Deauth",
	"Action",
	"Action No Ack",
	"reserved#15"
};

const char *ctl_subtype_name[] = {
	"reserved#0",
	"reserved#1",
	"Trigger",
	"reserved#3",
	"BRP",
	"VHT/HE NDPA",
	"reserved#6",
	"Control Wrapper",
	"BAR",
	"BA",
	"PS-Poll",
	"RTS",
	"CTS",
	"Ack",
	"CF-End",
	"CF-End-Ack"
};

const char *data_subtype_name[] = {
	"Data",
	"reserved#1",
	"reserved#2",
	"reserved#3",
	"Null",
	"reserved#5",
	"reserved#6",
	"reserved#7",
	"QoS Data",
	"QoS Data+CF-Ack",
	"QoS Data+CF-Poll",
	"QoS Data+CF-Ack+CF-Poll",
	"QoS Null",
	"reserved#13",
	"QoS CF-Poll",
	"QoS CF-Ack+CF-Poll",
};

void scm80211_snprint_frame(char *buf, size_t len, struct ieee80211_hdr *hdr)
{
	char *ptr = buf, *end = buf + len;
	const char *name;
	u16 fc, stype, seq;

	fc = le16_to_cpu(hdr->frame_control);
	stype = (fc & IEEE80211_FCTL_STYPE) >> 4;
	seq = le16_to_cpu(hdr->seq_ctrl);

	switch (fc & IEEE80211_FCTL_FTYPE) {
	case IEEE80211_FTYPE_DATA:
		name = data_subtype_name[stype];
		break;
	case IEEE80211_FTYPE_MGMT:
		name = mgt_subtype_name[stype];
		break;
	case IEEE80211_FTYPE_CTL:
		name = ctl_subtype_name[stype];
		break;
	default:
		name = "Unknown Type";
		break;
	}

	ptr += snprintf(ptr, end - ptr, "%s FC=%c%c%c%c%c%c%c SN=%d",
			name,
			(fc & IEEE80211_FCTL_PROTECTED) ?  'p' : '.',
			(fc & IEEE80211_FCTL_MOREFRAGS) ?  'm' : '.',
			(fc & IEEE80211_FCTL_PM) ?         'P' : '.',
			(fc & IEEE80211_FCTL_RETRY) ?      'R' : '.',
			(fc & IEEE80211_FCTL_MOREDATA) ?   'M' : '.',
			(fc & IEEE80211_FCTL_FROMDS) ?     'F' : '.',
			(fc & IEEE80211_FCTL_TODS) ?       'T' : '.',
			(seq & IEEE80211_SCTL_SEQ) >> 4);

	if (ieee80211_is_data_qos(hdr->frame_control) ||
	    ieee80211_is_qos_nullfunc(hdr->frame_control)) {
		ptr += snprintf(ptr, end - ptr, " QoS [TID %u]", ieee80211_get_tid(hdr));
	}
}

void scm80211_print_frame(struct ieee80211_hw *hw, struct ieee80211_hdr *hdr, size_t len)
{
	char buf[128];

	scm80211_snprint_frame(buf, sizeof(buf), hdr);
	wiphy_dbg(hw->wiphy, "%s\n", buf);
	print_hex_dump_bytes("skb: ", DUMP_PREFIX_NONE, hdr, len);
}
