// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
#include <linux/types.h>
#include <linux/bitfield.h>
#include <net/mac80211.h>

#define func_fmt(str)		"[DEBUGFS] "str
#define func_trace              FS

#include "core.h"
#include "util.h"
#include "debug.h"

#define MAX_DEBUGFS_STRING_LEN 128

static void scm80211_vif_info(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	char *str = data;
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	enum nl80211_iftype type = ieee80211_vif_type_p2p(vif);

	if (ieee80211_iface_is_ap(type))
		sprintf(str, "vif index :%d\n\tname:%s\n\ttype:%s\n"
			"\tbssid_filter:0x%02x\n\tbss_changed:0x%08x\n\tsta_num:%d\n\tbeacon seqNum:%d\n",
			vp->index, ieee80211_vif_name(vif),
			ieee80211_iftype_str[type],
			vp->bssid_filter, vp->bss_changed,
			atomic_read(&vp->num_sta), vp->bcn_seqnum);
	else
		sprintf(str, "vif index :%d\n\tname:%s\n\ttype:%s\n"
			"\tbssid_filter:0x%02x\n\tbss_changed:0x%08x\n",
			vp->index, ieee80211_vif_name(vif),
			ieee80211_iftype_str[type], vp->bssid_filter, vp->bss_changed);
}

static ssize_t scm80211_dbgfs_state_read(struct file *file,
					 char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	char *buf;
	char vif_buf[MAX_DEBUGFS_STRING_LEN];
	int pos = 0;
	const unsigned int size = 1024;
	ssize_t retval = 0;

	buf = kzalloc(size, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;
	pos += scnprintf(buf + pos, size - pos, "%s:\n\tversion:%s\n"
			 "\tirq:%d\n\tdevid:%d\n\tchannels:%d\n"
			 "\tidx:%d\n\tuse_chanctx:%d\n\tportid:%d\n\tscan_addr:%pM\n"
			 "\tscanning:%d\n\troc:%d\n\n",
			 wiphy_name(priv->hw->wiphy),
			 priv->version_str,
			 priv->irq,
			 priv->devid,
			 priv->channels,
			 priv->idx,
			 priv->use_chanctx,
			 priv->portid,
			 priv->scanner.addr, priv->scanner.scanning, priv->roc.running);

	memset(vif_buf, 0, sizeof(vif_buf));
	ieee80211_iterate_active_interfaces(priv->hw,
					    IEEE80211_IFACE_ITER_NORMAL,
					    scm80211_vif_info, vif_buf);
	pos += scnprintf(buf + pos, size - pos, "%s", vif_buf);

	if (pos > size)
		pos = size;
	retval = simple_read_from_buffer(user_buf, count, ppos, buf, pos);
	kfree(buf);
	return retval;
}

static const struct file_operations scm80211_dbgfs_state_ops = {
	.read = scm80211_dbgfs_state_read,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

static ssize_t scm80211_debugfs_tx_read(struct file *file,
					char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	char buf[128];
	int pos = 0;
	const unsigned int size = sizeof(buf);
	ssize_t retval = 0;

	pos += scnprintf(buf + pos, size - pos, "sched time: %luns\n",
			 ewma_tx_sched_time_read(&priv->sched_time));
	if (pos > size)
		pos = size;
	retval = simple_read_from_buffer(user_buf, count, ppos, buf, pos);
	return retval;
}

static const struct file_operations scm80211_debugfs_tx_ops = {
	.read = scm80211_debugfs_tx_read,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

#define reg_resp(x)	\
	{		\
		int reg  = priv->reg;				\
		int cnt;					\
		for (cnt = priv->cnt; cnt > 0; cnt--) {		\
			pos += scnprintf(buf + pos, count - pos, "[0x%08x] 0x%08x\n",	\
					 reg, reg_readl(x, hw, reg));	\
			reg += 4;				\
		}						\
	}

static ssize_t scm80211_dbgfs_reg_response(struct file *file, char __user *user_buf,
					   size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	struct ieee80211_hw *hw = priv->hw;
	char *buf = NULL;
	int pos = 0;
	int ret;

	buf = kzalloc(count, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	switch (priv->reg_type) {
	case MAC_REG:
		reg_resp(mac);
		break;
	case PHY_REG:
		reg_resp(phy);
		break;
	case RFI_REG:
		reg_resp(rf);
		break;
	case RAW_REG:
		reg_resp(raw);
		break;
	default:
		break;
	}
	ret = simple_read_from_buffer(user_buf, count, ppos, buf, pos);
	kfree(buf);
	return ret;
}

#define reg_cmd(x)				\
	{					\
		if (!read) {			\
			reg_writel(x, priv->hw, val, reg);			\
		}			\
		priv->reg = reg;	\
		priv->reg_type = ad;	\
		priv->cnt = val;	\
		if (!read)		\
			priv->cnt = 1;	\
	}

/**
 * scm80211_dbgfs_reg_cmd() to read/write the register
 * echo reg_type,operater,register,value > senscomm/reg
 * e.g
 *	echo mac,write,0x803C,0x0001 > senscomm/reg
 *	indicates	write 0x0001 to mac_reg@0x803C
 *	echo mac,read,0x803C > senscomm/reg
 *	indicates	read reg from mac_reg@0x803C
 *	echo mac,read,0x803C,10 > senscomm/reg
 *	indicates	read reg from mac_reg@0x803C and continuous 10 registers
 * cat senscomm/reg to see current val in these registers for read operation
 */
static ssize_t scm80211_dbgfs_reg_cmd(struct file *file, const char __user *user_buf,
				      size_t count, loff_t *off)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	char buf[MAX_DEBUGFS_STRING_LEN];
	char param[17];
	char *str = NULL;
	u32 reg = 0, val = 0;
	ssize_t ret = 0;
	ssize_t idx = 0;
	int i = 0, ad = 0, shift;
	bool read = false;

	if (count >= sizeof(buf))
		return -ENOMEM;

	memset(buf, 0, sizeof(buf));

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, off, user_buf, count);
	if (ret < 0)
		return ret;

	scm80211_trace(priv->hw, "%s command:%s\n", __func__, buf);

	do {
		str = strnchr(buf + idx, ret - idx, ',');
		memset(param, '\0', sizeof(param));
		if (str) {
			strncpy(param, buf + idx, str - (buf + idx));
			str++;
			idx = str - buf;
		} else {
			strcpy(param, buf + idx);
			param[strlen(param) - 1] = '\0';
		}
		if (idx >= ret)
			break;

		i++;

		if (!strncmp(param, "mac", 3)) {
			ad = MAC_REG;
		} else if (!strncmp(param, "phy", 3)) {
			ad = PHY_REG;
		} else if (!strncmp(param, "rf", 2)) {
			ad = RFI_REG;
		} else if (!strncmp(param, "raw", 3)) {
			ad = RAW_REG;
		} else if (!strncmp(param, "read", 4)) {
			read = true;
		} else if (!strncmp(param, "write", 5)) {
			read = false;
		} else if (i == 3) {
			shift = 0;
			if (!strncasecmp(param, "0x", 2))
				shift = 2;
			if (kstrtou32(param + shift, 16, &reg))
				return -EINVAL;
		} else if (i == 4) {
			shift = 0;
			if (!strncasecmp(param, "0x", 2))
				shift = 2;
			if (kstrtou32(param + shift, shift ? 16 : 10, &val))
				return -EINVAL;

			break;
		}
	} while (str);

	if (i == 3 && read)
		val = 1;

	switch (ad) {
	case MAC_REG:
		reg_cmd(mac);
		break;
	case PHY_REG:
		reg_cmd(phy);
		break;
	case RFI_REG:
		reg_cmd(rf);
		break;
	case RAW_REG:
		reg_cmd(raw);
		break;
	default:
		break;
	}
	return ret;
}

static const struct file_operations scm80211_dbgfs_reg_ops = {
	.read = scm80211_dbgfs_reg_response,
	.write = scm80211_dbgfs_reg_cmd,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

static ssize_t scm80211_dbgfs_winlost_read(struct file *file, char __user *user_buf,
					   size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	char buf[64] = { '\0', };
	ssize_t size;

	size = scnprintf(buf, sizeof(buf), "win lost happened %llu times\n", priv->total_winlost);
	return simple_read_from_buffer(user_buf, count, ppos, buf, 64);
}

static ssize_t scm80211_dbgfs_winlost_write(struct file *file, const char __user *user_buf,
					    size_t count, loff_t *off)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	char buf[20] = { '\0', };
	int ret;
	u64 winlost;

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, off, user_buf, count);
	if (ret < 0)
		return ret;

	ret = kstrtou64(buf, 0, &winlost);
	if (!ret)
		priv->total_winlost = winlost;

	return ret;
}

static const struct file_operations scm80211_dbgfs_winlost_ops = {
	.read = scm80211_dbgfs_winlost_read,
	.write = scm80211_dbgfs_winlost_write,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

static ssize_t scm80211_dbgfs_amsdu_get(struct file *file, char __user *user_buf,
					size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	struct ieee80211_hw *hw = priv->hw;
	char *buf = NULL;
	int pos = 0;
	int ret;

	buf = kzalloc(count, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;
	pos += scnprintf(buf + pos, count - pos, "TX AMSDU %sabled\n",
			 _ieee80211_hw_check(hw, IEEE80211_HW_TX_AMSDU) ? "en" : "dis");
	pos += scnprintf(buf + pos, count - pos, "RX AMSDU %sabled\n",
			 _ieee80211_hw_check(hw, IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU) ?
			 "en" : "dis");
	ret = simple_read_from_buffer(user_buf, count, ppos, buf, pos);
	kfree(buf);
	return ret;
}

static ssize_t scm80211_dbgfs_amsdu_set(struct file *file, const char __user *user_buf,
					size_t count, loff_t *off)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	struct ieee80211_hw *hw = priv->hw;
	char buf[MAX_DEBUGFS_STRING_LEN];
	int ret;

	if (count >= sizeof(buf))
		return -ENOMEM;

	memset(buf, 0, sizeof(buf));

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, off, user_buf, count);
	if (ret < 0)
		return ret;

	if (buf[0] == '0') {
		scm80211_trace(priv->hw, "- disable amsdu cap\n");
		__clear_bit(IEEE80211_HW_TX_AMSDU, hw->flags);
		__clear_bit(IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU, hw->flags);
	} else {
		scm80211_trace(priv->hw, "- enable amsdu cap\n");
		_ieee80211_hw_set(hw, IEEE80211_HW_TX_AMSDU);
		_ieee80211_hw_set(hw, IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU);
	}

	return ret;
}

static const struct file_operations scm80211_dbgfs_amsdu_ops = {
	.read = scm80211_dbgfs_amsdu_get,
	.write = scm80211_dbgfs_amsdu_set,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

static const char *debug_section[16] = {
	"main",
	"tx",
	"rx",
	"calibration",
	"bus",
	"rate control",
	"chan",
	"phy",
	"rf",
	"powersave",
	"dfs",
	"filesys",
	"", "", "",
	"priv"
};

static ssize_t scm80211_dbgfs_trace_get(struct file *file, char __user *user_buf,
					size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	unsigned long bitmap = priv->debug_trace;
	char buf[MAX_DEBUGFS_STRING_LEN];
	size_t pos = 0;
	int i, ret;

	pos += scnprintf(buf + pos, count - pos, "trace on:\n");

	for_each_set_bit(i, &bitmap, 16) {
		pos += scnprintf(buf + pos, count - pos, " - [%s]\n", debug_section[i]);
	}

	ret = simple_read_from_buffer(user_buf, count, ppos, buf, pos);
	return ret;
}

static ssize_t scm80211_dbgfs_trace_set(struct file *file, const char __user *user_buf,
					size_t count, loff_t *off)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	struct ieee80211_hw *hw = priv->hw;
	char buf[MAX_DEBUGFS_STRING_LEN];
	char param[10];
	char *str;
	int ret, i, idx = 0;
	u8 debug_trace = 0;

	if (count >= sizeof(buf))
		return -ENOMEM;

	memset(buf, 0, sizeof(buf));

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, off, user_buf, count);
	if (ret < 0)
		return ret;

	scm80211_trace(hw, "%s command:%s\n", __func__, buf);

	do {
		str = strnchr(buf + idx, ret - idx, ',');
		memset(param, '\0', sizeof(param));
		if (str) {
			strncpy(param, buf + idx, str - (buf + idx));
			str++;
			idx = str - buf;
		} else {
			strcpy(param, buf + idx);
			param[strlen(param) - 1] = '\0';
		}

		if (idx >= ret)
			break;

		for (i = 0; i < 16; i++) {
			if (!strcmp(param, debug_section[i])) {
				debug_trace |= BIT(i);
				break;
			}
		}
	} while (str);

	priv->debug_trace = debug_trace;

	return ret;
}

static const struct file_operations scm80211_dbgfs_trace_ops = {
	.read = scm80211_dbgfs_trace_get,
	.write = scm80211_dbgfs_trace_set,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

/**
 * scm80211_dbgfs_fw_dbg_cmd() to open/close the some part of fw log
 * echo [main],[tx],[rx],[cal],[bus] > senscomm/fw_dbg
 * to turn on corresponding debug on fw, others will be turned off
 * e.g
 * echo main,tx > senscomm/fw_dbg
 * turn on main, tx log on fw and turn off others
 *
 * Notice: rx log can not opened even if set rx to fw_dbg
 */
static ssize_t scm80211_dbgfs_fw_dbg_set(struct file *file, const char __user *user_buf,
					 size_t count, loff_t *off)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	struct ieee80211_hw *hw = priv->hw;
	char buf[MAX_DEBUGFS_STRING_LEN];
	char param[10];
	int ret, idx = 0;
	char *str;
	u8 dbg_info = 0;

	if (count >= sizeof(buf))
		return -ENOMEM;

	memset(buf, 0, sizeof(buf));

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, off, user_buf, count);
	if (ret < 0)
		return ret;

	scm80211_trace(hw, "%s command:%s\n", __func__, buf);

	do {
		str = strnchr(buf + idx, ret - idx, ',');
		memset(param, '\0', sizeof(param));
		if (str) {
			strncpy(param, buf + idx, str - (buf + idx));
			str++;
			idx = str - buf;
		} else {
			strcpy(param, buf + idx);
			param[strlen(param) - 1] = '\0';
		}

		if (idx >= ret)
			break;

		if (!strcmp(param, "main"))
			dbg_info |= MAIN_DBG;
		else if (!strcmp(param, "tx"))
			dbg_info |= TX_DBG;
		else if (!strcmp(param, "rx"))
			dbg_info |= RX_DBG;
		else if (!strcmp(param, "cal"))
			dbg_info |= CAL_DBG;
		else if (!strcmp(param, "bus"))
			dbg_info |= BUS_DBG;
		else
			break;
	} while (str);

	scm80211_set_fw_debug(hw, (u32)dbg_info);
	mdelay(5);

	return ret;
}

static const struct file_operations scm80211_dbgfs_fw_dbg_ops = {
	.write = scm80211_dbgfs_fw_dbg_set,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

static int scm80211_fops_ps_read(void *dat, u64 *val)
{
	struct ieee80211_hw_priv *data = dat;
	*val = data->pm.ps;
	return 0;
}

static void scm80211_send_ps_poll(void *dat, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *data = dat;
	struct sk_buff *skb;
	struct ieee80211_tx_info *info = NULL;
	struct cfg80211_chan_def def;
	struct ieee80211_pspoll *pspoll;

	if (!vif->bss_conf.assoc)
		return;

	if (scm80211_vif_chan(vif, &def))
		return;

	scm80211_trace(data->hw, "%s: send PS-Poll to %pM for aid %d\n",
		       __func__, vif->bss_conf.bssid, vif->bss_conf.aid);

	skb = dev_alloc_skb(sizeof(*pspoll));
	if (!skb)
		return;
	pspoll = (struct ieee80211_pspoll *)skb_put(skb, sizeof(*pspoll));
	pspoll->frame_control = cpu_to_le16(IEEE80211_FTYPE_CTL |
					    IEEE80211_STYPE_PSPOLL | IEEE80211_FCTL_PM);
	pspoll->aid = cpu_to_le16(0xc000 | vif->bss_conf.aid);
	memcpy(pspoll->bssid, vif->bss_conf.bssid, ETH_ALEN);
	memcpy(pspoll->ta, mac, ETH_ALEN);

	info = IEEE80211_SKB_CB(skb);
	info->flags |= IEEE80211_TX_CTL_USE_MINRATE;
	info->band = def.chan->band;
	info->control.vif = vif;

	local_bh_disable();
	scm80211_tx_channel(data->hw, skb, def.chan);
	local_bh_enable();
}

static void scm80211_send_nullfunc_ps(void *dat, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(vif);
	struct cfg80211_chan_def def;
	u32 queues = scm80211_vif_queues(vif);
	bool *ps = dat;

	if (!vif->bss_conf.assoc)
		return;

	if (scm80211_vif_chan(vif, &def))
		return;

	if (!(*ps))
		scm80211_wake_queues(hw, queues);
	else
		scm80211_flush(hw, vif, queues, true);
	scm80211_send_nullfunc(hw, vif, &def, *ps);

	if (*ps) {
		scm80211_flush(hw, vif, queues, false);
		scm80211_stop_queues(hw, queues);
	}
}

static int scm80211_fops_ps_write(void *dat, u64 val)
{
	struct ieee80211_hw_priv *data = dat;
	enum ps_mode old_ps;
	bool ps;

	if (val != PS_DISABLED && val != PS_ENABLED && val != PS_AUTO_POLL && val != PS_MANUAL_POLL)
		return -EINVAL;

	if (val == PS_MANUAL_POLL) {
		if (data->pm.ps != PS_ENABLED)
			return -EINVAL;
		local_bh_disable();
		ieee80211_iterate_active_interfaces_atomic(data->hw, IEEE80211_IFACE_ITER_NORMAL,
							   scm80211_send_ps_poll, data);
		local_bh_enable();
		return 0;
	}
	old_ps = data->pm.ps;
	data->pm.ps = val;

	if (val == PS_ENABLED || val == PS_DISABLED) {
		ps = val == PS_ENABLED ? true : false;

		if (val == PS_ENABLED) {
			data->hw->conf.flags |= IEEE80211_CONF_PS;
			ieee80211_iterate_active_interfaces(data->hw, IEEE80211_IFACE_ITER_NORMAL,
							    scm80211_send_nullfunc_ps, &ps);
			scm80211_enable_ps(data->hw, true);
		} else {
			data->hw->conf.flags &= ~IEEE80211_CONF_PS;
			scm80211_disable_ps(data->hw);
			ieee80211_iterate_active_interfaces(data->hw, IEEE80211_IFACE_ITER_NORMAL,
							    scm80211_send_nullfunc_ps, &ps);
		}
	}

	return 0;
}

DEFINE_SIMPLE_ATTRIBUTE(scm80211_fops_ps, scm80211_fops_ps_read, scm80211_fops_ps_write, "%llu\n");

static int scm80211_fops_rx_statistic_read(void *dat, u64 *val)
{
	struct ieee80211_hw_priv *data = dat;
	*val = data->rx_statistic;
	return 0;
}

static int scm80211_fops_rx_statistic_write(void *dat, u64 val)
{
	struct ieee80211_hw_priv *data = dat;

	data->rx_statistic = val;
	return 0;
}

DEFINE_SIMPLE_ATTRIBUTE(scm80211_fops_rx_statistic,
			scm80211_fops_rx_statistic_read,
			scm80211_fops_rx_statistic_write, "%llu\n");

static int hwsim_fops_group_read(void *dat, u64 *val)
{
	struct ieee80211_hw_priv *data = dat;
	*val = data->group;
	return 0;
}

static int hwsim_fops_group_write(void *dat, u64 val)
{
	struct ieee80211_hw_priv *data = dat;

	data->group = val;
	return 0;
}

DEFINE_SIMPLE_ATTRIBUTE(hwsim_fops_group, hwsim_fops_group_read, hwsim_fops_group_write, "%llx\n");

static ssize_t scm80211_fops_space_read(struct file *file, char __user *user_buf,
					size_t count, loff_t *ppos)
{
	struct ieee80211_hw_priv *priv = file->private_data;
	char buf[64];
	int pos = 0;
	ssize_t retval = 0;

	switch (priv->space) {
	case MT_HST_BD_HST:
		pos += scnprintf(buf, sizeof(buf), "MPDU Table [Host],Buffer Desc [Host]\n");
		break;
	case MT_HST_BD_TGT:
		pos += scnprintf(buf, sizeof(buf), "MPDU Table [Host],Buffer Desc [Target]\n");
		break;
	case MT_TGT_BD_TGT:
		pos += scnprintf(buf, sizeof(buf), "MPDU Table [Target],Buffer Desc [Target]\n");
		break;
	default:
		pos += scnprintf(buf, sizeof(buf), "INVALID Postion\n");
		break;
	}

	retval = simple_read_from_buffer(user_buf, count, ppos, buf, pos);
	return retval;
}

static const struct file_operations scm80211_dbgfs_space_ops = {
	.read = scm80211_fops_space_read,
	.open = simple_open,
	.llseek = generic_file_llseek,
};

void scm80211_register_dbgfs(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (is_platform_device(hw)) {
		priv->debugfs_hwsim = debugfs_create_dir("hwsim", hw->wiphy->debugfsdir);
		if (IS_ERR_OR_NULL(priv->debugfs_hwsim)) {
			scm80211_err(priv->hw,
				     "%s Failed to create debugfs directory under %pd\n",
				     __func__, hw->wiphy->debugfsdir);
			return;
		}
		debugfs_create_file("group", 0400, priv->debugfs_hwsim, priv, &hwsim_fops_group);
	}

	priv->debugfs_priv = debugfs_create_dir("senscomm", hw->wiphy->debugfsdir);
	if (IS_ERR_OR_NULL(priv->debugfs_priv)) {
		scm80211_err(priv->hw,
			     "%s Failed to create debugfs directory under %pd\n",
			     __func__, hw->wiphy->debugfsdir);
		return;
	}

	debugfs_create_file("state", 0400, priv->debugfs_priv, priv, &scm80211_dbgfs_state_ops);
	debugfs_create_file("reg", 0600, priv->debugfs_priv, priv, &scm80211_dbgfs_reg_ops);
	debugfs_create_file("ps", 0664, priv->debugfs_priv, priv, &scm80211_fops_ps);
	debugfs_create_file("tx", 0664, priv->debugfs_priv, priv, &scm80211_debugfs_tx_ops);
	debugfs_create_file("amsdu", 0664, priv->debugfs_priv, priv, &scm80211_dbgfs_amsdu_ops);
	debugfs_create_file("trace", 0664, priv->debugfs_priv, priv, &scm80211_dbgfs_trace_ops);
	if (is_pcie_device(hw)) {
		debugfs_create_file("space", 0664, priv->debugfs_priv,
				    priv, &scm80211_dbgfs_space_ops);
		debugfs_create_file("winlost", 0664, priv->debugfs_priv,
				    priv, &scm80211_dbgfs_winlost_ops);
		debugfs_create_file("rx_statistic", 0664, priv->debugfs_priv,
				    priv, &scm80211_fops_rx_statistic);
	}
	if (is_usb_device(hw)) {
		debugfs_create_file("fw_dbg", 0664, priv->debugfs_priv,
				    priv, &scm80211_dbgfs_fw_dbg_ops);
	}
}

void scm80211_unregister_dbgfs(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct device *dev = wiphy_dev(hw->wiphy);

	if (dev_is_platform(dev))
		debugfs_remove_recursive(priv->debugfs_hwsim);

	debugfs_remove_recursive(priv->debugfs_priv);
}

static const char *const format_str[] = {
	[0] = "NON-HT",
	[1] = "HT",
	[2] = "VHT",
	[3] = "unused",
	[4] = "HE-SU",
};

static const char *const non_ht_rate_to_str[] = {
	"1M", "2M", "5.5M", "11M",
	"OFDM-6M", "OFDM-9M", "OFDM-12M", "OFDM-18M",
	"OFDM-24M", "OFDM-36M", "OFDM-48M", "OFDM-54M",
};

static ssize_t scm80211_dbg_sta_rate_ctl_read(struct file *file,
					      char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_dbg_rate_fix *dbg_rate = &sta_priv->rate_fix;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	unsigned long flag;
	ssize_t len = 0;

	spin_lock_irqsave(&sta_priv->lock, flag);
	if (sta_priv->rate_fix_on) {
		int rateidx = ieee80211_rate_get_vht_mcs(&dbg_rate->rate);
		enum ppdu_format fmt = txvector_format(&dbg_rate->rate);
		enum cbw bw = txvector_ch_bandwidth(&dbg_rate->rate);

		len = scnprintf(buf + len, sizeof(buf) - len, "Fix rate:\n");
		len += scnprintf(buf + len, sizeof(buf) - len, "Fmt = %s\n", format_str[fmt]);
		if (fmt == NON_HT) {
			if (sta_priv->band == NL80211_BAND_5GHZ)
				rateidx += 4;
			len += scnprintf(buf + len, sizeof(buf) - len,
					 "Rate = %s", non_ht_rate_to_str[rateidx]);
			if (dbg_rate->rate.flags & IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
				len += scnprintf(buf + len, sizeof(buf) - len,
						 "(short preamble)\n");
			else
				len += scnprintf(buf + len, sizeof(buf) - len, "\n");
			goto exit;
		} else {
			len += scnprintf(buf + len, sizeof(buf) - len, "Rate = MCS%d\n", rateidx);
		}

		len += scnprintf(buf + len, sizeof(buf) - len,
				 "BW = %s\n",
				 bw == CBW20 ? "20Mhz" :
				 bw == CBW40 ? "40Mhz" : bw == CBW80 ? "80Mhz" : "invalid value");
		len += scnprintf(buf + len, sizeof(buf) - len,
				 "Code type = %s\n",
				 dbg_rate->flags & IEEE80211_TX_CTL_LDPC ? "LDPC" : "BCC");
		if (fmt != HE_SU) {
			len += scnprintf(buf + len, sizeof(buf) - len,
					 "GI type = %s\n",
					 dbg_rate->rate.flags & IEEE80211_TX_RC_SHORT_GI ?
					 "Short GI" : "Long GI");
			if (fmt == HT_MF)
				goto exit;
		}

		len += scnprintf(buf + len, sizeof(buf) - len,
				 "NSS = %d\n", ieee80211_rate_get_vht_nss(&dbg_rate->rate));

		if (fmt == VHT)
			goto exit;

		len += scnprintf(buf + len, sizeof(buf) - len,
				 "GI type = %s\n",
				 dbg_rate->he_gi_type == GI_0u8s ? "GI_0u8s" :
				 dbg_rate->he_gi_type == GI_1u6s ? "GI_1u6s" : "GI_3u2s");

		len += scnprintf(buf + len, sizeof(buf) - len,
				 "HE-LTF mode: %s HE-LTF\n",
				 dbg_rate->he_ltf_type == HE_LTF_1x ? "1x" :
				 dbg_rate->he_ltf_type == HE_LTF_2x ? "2x" : "4x");
	} else {
		len = scnprintf(buf + len, sizeof(buf) - len, "RA: Controlled by mac80211\n");
	}
exit:
	spin_unlock_irqrestore(&sta_priv->lock, flag);

	return simple_read_from_buffer(user_buf, count, ppos, buf, len);
}

/*
 * fmt=NON-HT/HT/VHT/HE
 * rate=xx
 * bw=20M/40M/80M
 * gi=SGI/LGI/0u4s/0u8s/1u6s/3u2s
 * code=BCC/LDPC
 * nss=x -> for vht and he
 * ltf=1x/2x/4x -> only for he
 * fmt, rate must first and second
 * echo fmt=xx,rate=xx,[bw=xx],[gi=xx],[code=xx],[nss=x],[ltf=1x]
 * e,g echo fmt=HE,rate=8,bw=20M,gi=0u8s,code=BCC,nss=2,ltf=1x > dbg_rate_ctl
 *
 * echo disable > dbg_rate_ctl to turn off the dbg_rate_ctl
 */
static ssize_t scm80211_dbg_sta_rate_ctl_write(struct file *file,
					       const char __user *user_buf,
					       size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	int ret;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	char *str;
	char param[17];
	u8 idx = 0;
	enum ppdu_format fmt = NON_HT;
	enum cbw bw = CBW20;
	enum gi_type gtype = LONG_GI;
	enum ltf_type ltf_type = HE_LTF_4x;
	bool rate_valid = false, disable = false;
	u8 rate_idx = 0xFF, nss = 1;
	bool prot = false;
	bool is_ldpc = false;
	unsigned long flag;

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, ppos, user_buf, count);

	if (ret <= 0)
		return ret;

	pr_debug("%s: %s\n", __func__, buf);

	str = buf;
	ret = strlen(buf);

	do {
		str = strnchr(buf + idx, ret - idx, ',');
		memset(param, '\0', sizeof(param));
		if (str) {
			strncpy(param, buf + idx, str - (buf + idx));
			str++;
			idx = str - buf;
		} else {
			strcpy(param, buf + idx);
			param[strlen(param) - 1] = '\0';
		}

		if (idx >= ret)
			break;

		if (!strncmp(param, "fmt=", 4)) {
			if (!strcmp(param + 4, "NON-HT")) {
				fmt = NON_HT;
			} else if (!strcmp(param + 4, "HT")) {
				fmt = HT_MF;
			} else if (!strcmp(param + 4, "VHT")) {
				fmt = VHT;
			} else if (!strcmp(param + 4, "HE")) {
				fmt = HE_SU;
			} else {
				pr_err("PPDU format error\n");
				goto exit;
			}
		} else if (!strncmp(param, "rate=", 5)) {
			if (sscanf(param + 5, "%hhd", &rate_idx) != 1)
				goto exit;
		} else if (!strncmp(param, "prot=", 5)) {
			if (!strcmp(param + 5, "1"))
				prot = true;
		} else if (!strncmp(param, "bw=", 3)) {
			if (!strcmp(param + 3, "20M")) {
				bw = CBW20;
			} else if (!strcmp(param + 3, "40M")) {
				bw = CBW40;
			} else if (!strcmp(param + 3, "80M")) {
				bw = CBW80;
			} else {
				pr_err("unsupported bw\n");
				goto exit;
			}
		} else if (!strncmp(param, "gi=", 3)) {
			if (fmt == HE_SU) {
				if (!strcmp(param + 3, "0u8s")) {
					gtype = GI_0u8s;
				} else if (!strcmp(param + 3, "1u6s")) {
					gtype = GI_1u6s;
				} else if (!strcmp(param + 3, "3u2s")) {
					gtype = GI_3u2s;
				} else {
					pr_err("Invalid GI type\n");
					goto exit;
				}
			} else {
				if (!strcmp(param + 3, "SGI")) {
					gtype = SHORT_GI;
				} else if (!strcmp(param + 3, "LGI")) {
					gtype = LONG_GI;
				} else {
					pr_err("Invalid GI type\n");
					goto exit;
				}
			}
		} else if (!strncmp(param, "code=", 5)) {
			if (!strcmp(param + 5, "LDPC"))
				is_ldpc = true;
		} else if (!strncmp(param, "nss=", 4)) {
			if (sscanf(param + 4, "%hhd", &nss) != 1)
				goto exit;
			if (nss > 2) {
				pr_err("Invalid nss setting\n");
				goto exit;
			}
		} else if (!strncmp(param, "ltf=", 4)) {
			if (!strncmp(param + 4, "1x", 2))
				ltf_type = HE_LTF_1x;
			else if (!strncmp(param + 4, "2x", 2))
				ltf_type = HE_LTF_2x;
		} else if (!strcmp(param, "disable")) {
			disable = true;
			break;
		}
	} while (str);

	spin_lock_irqsave(&sta_priv->lock, flag);

	if (!disable)
		rate_valid = scm80211_dbg_construct_rate(sta, fmt, rate_idx,
							 bw, nss, gtype, ltf_type, is_ldpc);

	if (rate_valid) {
		pr_debug("rate setting success turn on rate fix flag\n");
		sta_priv->rate_fix_on = true;
	} else {
		if (disable) {
			pr_debug("turn off rate fix flag\n");
		} else {
			pr_debug("rate setting fail please check the param\n");
			ret = -EINVAL;
		}
		sta_priv->rate_fix_on = false;
	}

	spin_unlock_irqrestore(&sta_priv->lock, flag);

	return ret;
exit:
	return -EINVAL;
}

static const struct file_operations fops_rate_ctl = {
	.read = scm80211_dbg_sta_rate_ctl_read,
	.write = scm80211_dbg_sta_rate_ctl_write,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};

static ssize_t scm80211_he_rate_map_read(struct file *file,
					 char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct he_higher_rates_stats *stats = sta_priv->rate_stats;
	ssize_t size;
	int i, j;
	unsigned long flag;
	char *buf = kmalloc(10240, GFP_KERNEL);
	char *p;

	if (!buf)
		return -ENOMEM;

	p = buf;
	p += sprintf(p, "%-21s %8s %-8s %5s %16s %-16s\n",
		     "rate", "success", "attempt", "prob.", "#succes", "#attempts");

	spin_lock_irqsave(&sta_priv->lock, flag);

	for (i = 0; i < GROUP_COUNT; i++) {
		int bw, nss;

		bw = 20 << (i / 2);
		nss = i % 2 + 1;

		for (j = 0; j < GROUP_RATES; j++) {
			int index = i * GROUP_RATES + j;
			int mcs = j + 7;
			int prob = PROB_TRUNC(stats[index].prob * 1000);

			p += sprintf(p, "HE-SU-%dMHz-%dSS-MCS%-2d %8d %-8d %3u.%1u %16d %-16d\n",
				     bw, nss, mcs,
				     stats[index].success, stats[index].attempts,
				     prob / 10, prob % 10,
				     stats[index].tot_succ, stats[index].tot_atmp);
		}
	}

	spin_unlock_irqrestore(&sta_priv->lock, flag);

	size = simple_read_from_buffer(user_buf, count, ppos, buf, p - buf);
	kfree(buf);

	return size;
}

static const struct file_operations fops_he_rate_map = {
	.read = scm80211_he_rate_map_read,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};

static ssize_t scm80211_ampdu_stats_read(struct file *file,
					 char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct mpdu_stats *stats = sta_priv->mpdu_stat;
	ssize_t size;
	int i;
	unsigned long flag;
	char *buf = kmalloc(10240, GFP_KERNEL);
	char *p;

	if (!buf)
		return -ENOMEM;

	p = buf;

	spin_lock_irqsave(&sta_priv->lock, flag);
	p += sprintf(p, "\n");
	p += sprintf(p, "[AC] [AMPDUS] [N_MPDUS] [Success] [MPDU-PER_AMPDU]\n");
	for (i = 0; i < IEEE80211_NUM_ACS; ++i) {
		p += sprintf(p, "%2d %9d %10d %10d %10d\n",
			     i, stats[i].tot_ampdus, stats[i].tot_mpdus,
			     stats[i].tot_succ, stats[i].mpdus_per_ampdu);
	}
	spin_unlock_irqrestore(&sta_priv->lock, flag);

	size = simple_read_from_buffer(user_buf, count, ppos, buf, p - buf);
	kfree(buf);

	return size;
}

static const struct file_operations fops_ampdu_stats = {
	.read = scm80211_ampdu_stats_read,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};

union htc_om {
	__le16 content;
	struct {
		__le16	id		:bits(3,  0);
		__le16	rx_nss		:bits(6,  4);
		__le16	bw		:bits(8,  7);
		__le16	ul_mudis	:bits(9,  9);
		__le16	tx_nss		:bits(12, 10);
		__le16	ersu_dis	:bits(13, 13);
		__le16	dl_mu_rec	:bits(14, 14);
		__le16	ul_mudatadis	:bits(15, 15);
	};
};

static ssize_t scm80211_dbg_omi_read(struct file *file,
				     char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_vif *vif = sta->txq[0]->vif;
	struct ieee80211_hw *hw = NULL;
	union htc_om htc_om;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	u32 v;
	ssize_t len = 0;

	if (vif)
		hw = ieee80211_vif_to_hw(vif);
	if (!hw)
		return -EINVAL;

	v = scm80211_mac_readl(hw, MAC_REG_VIF_HTC_OMI(scm80211_vdev_id(vif)));

	htc_om.content = 0x01 | (FIELD_GET(VIF_HTC_OMI_HE_OMI, v) << 4);

	len += scnprintf(buf + len, sizeof(buf) - len, "OMI: Append in Qos null [%s] Qos [%s]\n",
			 FIELD_GET(VIF_HTC_OMI_HETB_QOSNULL_OMI_EN, v) ? "enable" : "disable",
			 FIELD_GET(VIF_HTC_OMI_HETB_DATA_OMI_EN, v) ? "enable" : "disable");
	len += scnprintf(buf + len, sizeof(buf) - len, " Control ID : %d\n", htc_om.id);
	len += scnprintf(buf + len, sizeof(buf) - len, " RX-NSS     : %d\n", htc_om.rx_nss);
	len += scnprintf(buf + len, sizeof(buf) - len, " Channel BW : %d\n", htc_om.bw);
	len += scnprintf(buf + len, sizeof(buf) - len, " UL-MU DIS  : %d\n", htc_om.ul_mudis);
	len += scnprintf(buf + len, sizeof(buf) - len, " TX-NSS     : %d\n", htc_om.tx_nss);
	len += scnprintf(buf + len, sizeof(buf) - len, " ER-SU DIS  : %d\n", htc_om.ersu_dis);
	len += scnprintf(buf + len, sizeof(buf) - len, " DL-MU REC  : %d\n", htc_om.dl_mu_rec);
	len += scnprintf(buf + len, sizeof(buf) - len, " UL DATA DIS: %d\n", htc_om.ul_mudatadis);

	return simple_read_from_buffer(user_buf, count, ppos, buf, len);
}

/*
 * tx/rx_nss=1/2
 * bw=20M/40M/80M
 * ul_midis=0/1
 * ul_mudatadis=0/1
 *
 * e,g echo rx_nss=1,tx_nss=1,bw=20M,ul_mudis=0,ul_mudatadis=0 > dbg_om_set
 *
 * echo disable > dbg_om_set to turn off the dbg_om_set
 */
static ssize_t scm80211_dbg_omi_write(struct file *file,
				      const char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	struct ieee80211_vif *vif = sta->txq[0]->vif;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = NULL;
	union htc_om htc_om;
	u8 idx = 0;
	int vidx, val;
	ssize_t ret = 0;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	char param[17];
	char *str;
	bool disable = false;
	unsigned long flag, flag2;

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, ppos, user_buf, count);

	if (vif)
		hw = ieee80211_vif_to_hw(vif);
	if (!hw)
		goto exit;

	if (ret <= 0)
		return ret;

	vidx = scm80211_vdev_id(vif);

	/* set defalut */
	memset(&htc_om, 0, sizeof(htc_om));
	htc_om.id = 1;
	htc_om.rx_nss = get_ieee80211_rx_nss(sta) - 1;
	htc_om.tx_nss = get_ieee80211_rx_nss(sta) - 1;
	htc_om.bw = get_ieee80211_rx_bandwidth(sta);

	str = buf;
	ret = strlen(buf);

	do {
		str = strnchr(buf + idx, ret - idx, ',');
		memset(param, '\0', sizeof(param));
		if (str) {
			strncpy(param, buf + idx, str - (buf + idx));
			str++;
			idx = str - buf;
		} else {
			strcpy(param, buf + idx);
			param[strlen(param) - 1] = '\0';
		}

		if (idx >= ret)
			break;

		if (!strncmp(param, "rx_nss=", 7)) {
			if (sscanf(param + 7, "%d", &val) != 1)
				goto exit;
			htc_om.rx_nss = val - 1;
		} else if (!strncmp(param, "tx_nss=", 7)) {
			if (sscanf(param + 7, "%d", &val) != 1)
				goto exit;
			htc_om.tx_nss = val - 1;
		} else if (!strncmp(param, "bw=", 3)) {
			if (!strcmp(param + 3, "20M")) {
				htc_om.bw = CBW20;
			} else if (!strcmp(param + 3, "40M")) {
				htc_om.bw = CBW40;
			} else if (!strcmp(param + 3, "80M")) {
				htc_om.bw = CBW80;
			} else {
				pr_err("unsupported bw\n");
				goto exit;
			}
		} else if (!strncmp(param, "ul_mudis=", 9)) {
			if (sscanf(param + 9, "%d", &val) != 1)
				goto exit;
			htc_om.ul_mudis = val;
		} else if (!strncmp(param, "ul_mudatadis=", 13)) {
			if (sscanf(param + 13, "%d", &val) != 1)
				goto exit;
			htc_om.ul_mudatadis = val;
		} else if (!strncmp(param, "ersu_dis=", 9)) {
			if (sscanf(param + 9, "%d", &val) != 1)
				goto exit;
			htc_om.ersu_dis = val;
		} else if (!strncmp(param, "dl_mu_rec=", 10)) {
			if (sscanf(param + 10, "%d", &val) != 1)
				goto exit;
			htc_om.dl_mu_rec = val;
		} else {
			disable = true;
		}
	} while (str);

	if (disable) {
		scm80211_dbg(hw, "disable omi");
		scm80211_mac_writel(hw, 0, MAC_REG_VIF_HTC_OMI(vidx));
		spin_lock_irqsave(&vifp->lock, flag);
		vifp->reg_mmss |= VIF_MIN_MPDU_SPACING_TRIG_EN;
		vifp->reg_mmss &= ~VIF_MIN_MPDU_SPACING_HETB_DATA_DIS;
		spin_unlock_irqrestore(&vifp->lock, flag);
		scm80211_mac_writel(hw, vifp->reg_mmss, MAC_REG_VIF_MIN_MPDU_SPACING(vidx));
		spin_lock_irqsave(&sta_priv->lock, flag2);
		sta_priv->omi_req = 0;
		spin_unlock_irqrestore(&sta_priv->lock, flag2);
	} else {
		u32 v = 0;

		scm80211_dbg(hw, "OMI content %04x\n", htc_om.content);

		v |= FIELD_PREP(VIF_HTC_OMI_HE_OMI, FIELD_GET(0xFFF0, htc_om.content));

		spin_lock_irqsave(&sta_priv->lock, flag2);
		sta_priv->omi_req = HE_VARIANT | (htc_om.content << 2);
		spin_unlock_irqrestore(&sta_priv->lock, flag2);
		scm80211_mac_writel(hw, v, MAC_REG_VIF_HTC_OMI(vidx));
	}

	return ret;
exit:
	return -EINVAL;
}

static const struct file_operations fops_om_set = {
	.read = scm80211_dbg_omi_read,
	.write = scm80211_dbg_omi_write,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};

static ssize_t scm80211_max_amsdu_len_read(struct file *file,
					   char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	char buf[32];
	ssize_t len = 0;

	len = scnprintf(buf, sizeof(buf), "max_amsdu_len %d\n", sta->max_amsdu_len);
	return simple_read_from_buffer(user_buf, count, ppos, buf, len);
}

static ssize_t scm80211_max_amsdu_len_write(struct file *file,
					    const char __user *user_buf,
					    size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	ssize_t ret = 0;
	char buf[16] = { '\0', };
	u32 max = 0, len;

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, ppos, user_buf, count);
	if (ret <= 0)
		return ret;

	ret = kstrtou32(buf, 0, &len);
	if (ret)
		return ret;

	if (ieee80211_is_vht_sta(sta)) {
		switch (get_ieee80211_vht_cap(sta).cap & IEEE80211_VHT_CAP_MAX_MPDU_MASK) {
		case IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_11454:
			max = IEEE80211_MAX_MPDU_LEN_VHT_11454;
			break;
		case IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_7991:
			max = IEEE80211_MAX_MPDU_LEN_VHT_7991;
			break;
		default:
			max = IEEE80211_MAX_MPDU_LEN_VHT_3895;
			break;
		}
	} else if (ieee80211_is_ht_sta(sta)) {
		if (get_ieee80211_ht_cap(sta).cap & IEEE80211_HT_CAP_MAX_AMSDU)
			max = IEEE80211_MAX_MPDU_LEN_HT_7935;
		else
			max = IEEE80211_MAX_MPDU_LEN_HT_3839;
	}

	if (len > max)
		len = max;
	if (len < IEEE80211_MAX_MPDU_LEN_HT_3839)
		len = IEEE80211_MAX_MPDU_LEN_HT_3839;

	sta->max_amsdu_len = len;

	return ret;
}

static const struct file_operations fops_max_amsdu_len = {
	.read = scm80211_max_amsdu_len_read,
	.write = scm80211_max_amsdu_len_write,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};

#ifdef SCM80211_TWT
static ssize_t scm80211_dbg_twt_read(struct file *file,
				     char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_vif *vif = sta->txq[0]->vif;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = NULL;
	unsigned long flag;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	ssize_t len = 0;

	if (vif)
		hw = ieee80211_vif_to_hw(vif);
	if (!hw)
		return -EINVAL;

	spin_lock_irqsave(&vifp->lock, flag);
	switch (scm80211_get_twt_state(vif)) {
	case TWT_DISABLED:
		len += scnprintf(buf + len, sizeof(buf) - len, "TWT disabled\n");
		break;
	case TWT_REQ:
		len += scnprintf(buf + len, sizeof(buf) - len, "TWT Request\n");
		break;
	case TWT_NEGO:
		len += scnprintf(buf + len, sizeof(buf) - len, "TWT under Negotiation\n");
		break;
	case TWT_DOZE:
	case TWT_SP:
		len += scnprintf(buf + len, sizeof(buf) - len, "TWT Nego Success: State[%s]\n",
				 twt_state[scm80211_get_twt_state(vif)]);
		if (vifp->twt_teardown)
			len += scnprintf(buf + len, sizeof(buf) - len, "Going to Teardown!\n");
		len += scnprintf(buf + len, sizeof(buf) - len, " - Trigger   %d\n",
				 vifp->twt_trigger);
		len += scnprintf(buf + len, sizeof(buf) - len, " - Implicit  %d\n",
				 vifp->twt_implicit);
		len += scnprintf(buf + len, sizeof(buf) - len, " - Flow type %d\n",
				 vifp->twt_flowtype);
		len += scnprintf(buf + len, sizeof(buf) - len, " - Flow ID   %d\n",
				 vifp->twt_flow_id);
		len += scnprintf(buf + len, sizeof(buf) - len, " - Protect   %d\n",
				 vifp->twt_prot);
		len += scnprintf(buf + len, sizeof(buf) - len, " - TWT       %llu\n",
				 vifp->twt);
		len += scnprintf(buf + len, sizeof(buf) - len, " - Wake Int  %d\n",
				 vifp->twt_wake_int);
		len += scnprintf(buf + len, sizeof(buf) - len, " - MinSPdur  %d\n",
				 vifp->twt_min_sp_dur);
		break;
	default:
		len += scnprintf(buf + len, sizeof(buf) - len, "Invalid STATE\n");
		break;
	}
	spin_unlock_irqrestore(&vifp->lock, flag);

	return simple_read_from_buffer(user_buf, count, ppos, buf, len);
}

/*
 * TWT Element:(Individual)
 * -----------------
 * |Element ID(216)|
 * -----------------
 * |    Length     |
 * --------------------------------------------------------------
 * |NDP PAD indi|Resp PM|Nego type|TWT info dis|Wake Dur UT|Resv|
 * ----------------------------------------------------------------------------------------
 * |TWT req|TWT Setup cmd|Trigger|Implicit|Flow type|TWT Flow ID|TWT Wake Int Exp|TWT Prot|
 * ----------------------------------------------------------------------------------------
 * |Target wakeup time|
 * -------------------------------
 * |Nominal Minimum TWT Wakeup Dur|
 * --------------------------------
 * |TWT Wake Interval Mantissa|
 * ----------------------------
 * |TWT Channel|NDP Padding(O)|
 * ----------------------------
 * Setup
 * e,g echo TWT_Ctl=0,Req_Type=0x10B3,MinWakeDur=16,WakeIntMant=512,TWT_Channel=0 > dbg_twt_set
 * for this cmd  TWT Wake Int Exp in Req_Type [14-10] = 4,
 * so the TWT wake interval = WakeIntMant * 2^4 = 8192us
 * TWT Minimum Wake Duration = MinWakeDur * 256 = 4096us
 * (wake duration unit in Ctl filed = 0, so use 256 default ,if set use TU 1024)
 *
 * Teardown
 * e.g echo TWT_teardown=1 > dbg_twt_set
 */
static ssize_t scm80211_dbg_twt_write(struct file *file,
				      const char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_vif *vif = sta->txq[0]->vif;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = NULL;
	struct sk_buff *skb = NULL;
	struct ieee80211_mgmt *mgmt;
	struct unprot_sig_action *action;
	struct element *elem;
	int val[5];
	ssize_t ret = 0, size;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	unsigned long flag;
	bool is_teardown = false;

	if (!ieee80211_is_he_sta(sta) || ieee80211_iface_is_ap(vif->type))
		return -EINVAL;

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, ppos, user_buf, count);
	if (ret <= 0)
		return ret;

	skb = alloc_skb(sizeof(*mgmt) + sizeof(struct txvector), GFP_KERNEL);
	if (!skb)
		return -ENOMEM;

	if (vif)
		hw = ieee80211_vif_to_hw(vif);
	if (!hw)
		goto exit;

	if (sscanf(buf, "TWT_teardown=%d", val) == 1)
		is_teardown = true;

	if (!is_teardown) {
		if (sscanf
		    (buf, "TWT_Ctl=%d,Req_Type=0x%x,MinWakeDur=%d,WakeIntMant=%d,TWT_Channel=%d",
		     val, val + 1, val + 2, val + 3, val + 4) != 5) {
			goto exit;
		}
	}

	skb_reserve(skb, hw->extra_tx_headroom);
	/* Set default */
	mgmt = (void *)skb->data;
	mgmt->frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION;
	memcpy(mgmt->da, sta->addr, ETH_ALEN);
	memcpy(mgmt->sa, vif->addr, ETH_ALEN);
	memcpy(mgmt->bssid, vif->bss_conf.bssid, ETH_ALEN);
	mgmt->seq_ctrl = cpu_to_le16((get_random_u32() % (4096)));
	size = offsetof(struct ieee80211_mgmt, u.action.category);
	action = (void *)(skb->data + size);
	action->category = WLAN_CATEGORY_UNPROT_SIG;
	if (is_teardown) {
		action->action_code = WLAN_SIG_TWT_TEARDOWN;
		action->dialog_token = val[0];
	} else {
		action->action_code = WLAN_SIG_TWT_SETUP;
		action->dialog_token = 0;
	}
	size += 3;
	if (is_teardown) {
		skb_put(skb, size);
		goto print;
	}
	elem = (void *)(skb->data + size);
	elem->id = WLAN_EID_TWT;
	elem->datalen = 15;
	if (sizeof(*mgmt) - size - 2 < 15) {
		scm80211_err(hw, "No enough memory total: %ld used %ld\n", sizeof(*mgmt),
			     size + 2);
		goto exit;
	}
	skb_put(skb, size + 2 + 15);

	elem->data[0] = (u8)val[0];
	put_unaligned_le16((u16)val[1], &elem->data[1]);
	elem->data[11] = (u8)val[2];
	put_unaligned_le16((u16)val[3], &elem->data[12]);
	elem->data[14] = (u8)val[4];
print:
	if (is_teardown)
		print_hex_dump_bytes("twt teardown", DUMP_PREFIX_OFFSET, skb->data, skb->len);
	else
		print_hex_dump_bytes("twt setup", DUMP_PREFIX_OFFSET, skb->data, skb->len);

	spin_lock_irqsave(&vifp->lock, flag);
	/* share using twt_request sk_buff */
	vifp->twt_request = skb;
	vifp->twt_teardown = is_teardown;
	if (!is_teardown && vifp->twt_state == TWT_DISABLED) {
		scm80211_set_twt_state(vif, TWT_REQ);
		vifp->twt_off = size + 2 + 3;
	}
	spin_unlock_irqrestore(&vifp->lock, flag);

	return ret;
exit:
	kfree_skb(skb);
	return -EINVAL;
}

static const struct file_operations fops_twt_set = {
	.read = scm80211_dbg_twt_read,
	.write = scm80211_dbg_twt_write,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};
#endif /* SCM80211_TWT */

#ifdef SCM80211_WFA_TEST
static ssize_t scm80211_dbg_action_read(struct file *file,
					char __user *user_buf, size_t count, loff_t *ppos)
{
	return 0;
}

/**
 * Send action frame used for wfa test 20/40 coex test case
 * echo coex2040,oper_class,channel_num > dbg_action_set
 * e.g echo coex2040,12,3 > dbg_action_set
 **/
static ssize_t scm80211_dbg_action_write(struct file *file,
					 const char __user *user_buf, size_t count, loff_t *ppos)
{
	struct ieee80211_sta *sta = file->private_data;
	struct ieee80211_vif *vif = sta->txq[0]->vif;
	struct ieee80211_hw *hw = NULL;
	struct sk_buff *skb = NULL;
	struct cfg80211_chan_def def;
	struct ieee80211_mgmt *mgmt;
	struct ieee80211_tx_info *info = NULL;
	ssize_t ret = 0, size;
	char buf[MAX_DEBUGFS_STRING_LEN] = { '\0', };
	u8 *pos;

	ret = simple_write_to_buffer(buf, sizeof(buf) - 1, ppos, user_buf, count);
	if (ret <= 0)
		return ret;

	skb = alloc_skb(sizeof(*mgmt) + sizeof(struct txvector), GFP_KERNEL);
	if (!skb)
		return -ENOMEM;

	if (vif)
		hw = ieee80211_vif_to_hw(vif);
	if (!hw)
		goto exit;

	skb_reserve(skb, hw->extra_tx_headroom);
	/* Set default */
	mgmt = (void *)skb->data;
	mgmt->frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION;
	memcpy(mgmt->da, sta->addr, ETH_ALEN);
	memcpy(mgmt->sa, vif->addr, ETH_ALEN);
	memcpy(mgmt->bssid, vif->bss_conf.bssid, ETH_ALEN);
	mgmt->seq_ctrl = cpu_to_le16((get_random_u32() % (4096)));
	if (!strncmp(buf, "coex2040", strlen("coex2040"))) {
		int op_class, chan_list;

		if (sscanf(buf, "coex2040,%d,%d", &op_class, &chan_list) != 2) {
			op_class = 12;
			chan_list = 3;
		}
		mgmt->u.action.category = WLAN_CATEGORY_PUBLIC;
		size = offsetof(struct ieee80211_mgmt, u.action.category) + 1;
		skb_put(skb, size + 8);
		pos = (u8 *)(skb->data + size);
		*pos++ = 0;
		*pos++ = WLAN_EID_BSS_COEX_2040;
		*pos++ = 1;
		*pos++ = 0x07;
		*pos++ = WLAN_EID_BSS_INTOLERANT_CHL_REPORT;
		*pos++ = 2;
		*pos++ = (u8)op_class;
		*pos++ = (u8)chan_list;
	}

	print_hex_dump_bytes("action", DUMP_PREFIX_OFFSET, skb->data, skb->len);

	if (scm80211_vif_chan(vif, &def))
		goto exit;
	info = IEEE80211_SKB_CB(skb);
	info->flags |= IEEE80211_TX_CTL_USE_MINRATE;
	info->band = def.chan->band;
	info->control.vif = vif;
	scm80211_tx_channel(hw, skb, def.chan);

	return ret;
exit:
	kfree_skb(skb);
	return -EINVAL;
}

static const struct file_operations fops_action_set = {
	.read = scm80211_dbg_action_read,
	.write = scm80211_dbg_action_write,
	.open = simple_open,
	.owner = THIS_MODULE,
	.llseek = default_llseek,
};
#endif

void scm80211_sta_add_debugfs(struct ieee80211_hw *hw,
			      struct ieee80211_vif *vif,
			      struct ieee80211_sta *sta, struct dentry *dir)
{
	debugfs_create_file("dbg_rate_ctl", 0644, dir, sta, &fops_rate_ctl);
	debugfs_create_file("dbg_he_rate_map", 0644, dir, sta, &fops_he_rate_map);
	debugfs_create_file("dbg_om_set", 0644, dir, sta, &fops_om_set);
	debugfs_create_file("dbg_ampdu_stat", 0644, dir, sta, &fops_ampdu_stats);
	debugfs_create_file("max_amsdu_len", 0644, dir, sta, &fops_max_amsdu_len);
#ifdef SCM80211_TWT
	debugfs_create_file("dbg_twt_set", 0644, dir, sta, &fops_twt_set);
#endif
#ifdef SCM80211_WFA_TEST
	debugfs_create_file("dbg_action_set", 0644, dir, sta, &fops_action_set);
#endif
}
