// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
#ifdef SCM80211_DFS_CERTIFIED
#include <linux/compiler.h>
#include <linux/bitfield.h>
#include <linux/netlink.h>
#include <linux/nl80211.h>
#include <linux/ieee80211.h>
#include <linux/net.h>
#include <net/regulatory.h>

#define func_fmt(str)           "[DFS] "str
#define func_trace              CHAN

#include "dfs.h"
#include "core.h"
#include "debug.h"

struct scm80211_dfs_pulse dfs_fcc_radars[] = {
	/* 0x0 mean NA */
	/* FCC TYPE 0 */
	{ 0, 0, 18, 0x0, 0x0, 700, 700, 1, 1 },

	/* FCC TYPE 1 */
	{ 1, 1, 18, 0x0, 0x0, 325, 1930, 1, 1 },

	/* FCC TYPE 2 */
	{ 2, 1, 23, 0x0, 0x0, 4347, 6666, 1, 5 },

	/* FCC TYPE 3 */
	{ 2, 1, 16, 0x0, 0x0, 2000, 5000, 6, 10 },

	/* FCC TYPE 4 */
	{ 2, 1, 12, 0x0, 0x0, 4347, 6666, 11, 20 },

};

struct scm80211_dfs_pulse dfs_etsi_radars[] = {
	/* ETSI TYPE 0 */
	{ 0, 0, 18, 1, 0x0, 0x0, 700, 700 }
};

#define DFS_DEBUGFS 1
#ifdef DFS_DEBUGFS
#define TSF_2_MS(_tsf) ktime_to_ms((_tsf))
static int short_pulse_cnt;
module_param(short_pulse_cnt, int, 0644);
MODULE_PARM_DESC(short_pulse_cnt, "short pulse cnt per second");

static int long_pulse_cnt;
module_param(long_pulse_cnt, int, 0644);
MODULE_PARM_DESC(long_pulse_cnt, "long pulse cnt per second");
#else
#define TSF_2_MS(_tsf) ((_tsf) / 1000)
#endif //DFS_DEBUGFS

static bool dfs_elem_ptr_equal(struct scm80211_dfs_phy_rpt *rpt, u32 ptr1, u32 ptr2)
{
	return ring_ptr(rpt, ptr1) == ring_ptr(rpt, ptr2);
}

static int dfs_elem_ring_full(struct scm80211_dfs_phy_rpt *rpt)
{
	return (ring_index(rpt, rpt->first) == ring_index(rpt, rpt->last)) &&
	    (ring_ptr(rpt, rpt->first) != ring_ptr(rpt, rpt->last));
}

static int dfs_elem_ring_empty(struct scm80211_dfs_phy_rpt *rpt)
{
	return ring_ptr(rpt, rpt->first) == ring_ptr(rpt, rpt->last);
}

/**
 * scm80211_radar_evt_get() - Get radar pulse statistics
 * Get short and long radar pulse count, and get the related timestamps as well.
 *
 */
u32 scm80211_radar_evt_get(struct scm80211_dfs *dfs, struct scm80211_dfs_phy_rpt_elem *elem)
{
#ifdef DFS_DEBUGFS
	struct scm80211_dfs_phy_rpt *rpt = &dfs->phy_rpt;
	int idx = ring_index(rpt, rpt->last ? rpt->last - 1 : 0);
	struct scm80211_dfs_phy_rpt_elem *latest = &rpt->elems[idx];

	elem->short_pulse_cnt =
		short_pulse_cnt * (rpt->last ? (TSF_2_MS(ktime_get() - latest->tsf) / 1000) : 1) +
		latest->short_pulse_cnt;
	elem->long_pulse_cnt =
		long_pulse_cnt * (rpt->last ? (TSF_2_MS(ktime_get() - latest->tsf) / 1000) : 1) +
		latest->long_pulse_cnt;
	elem->tsf = ktime_get();
#else
	// TODO:
	struct ieee80211_hw *hw = dfs->hw;
	int vdev = 0;		//scm80211_vdev_id(vif);
	u32 tsf[2];
	unsigned long flags;

	local_irq_save(flags);
	tsf[1] = scm80211_mac_readl(hw, MAC_REG_VIF_TSF_H(vdev));
	tsf[0] = scm80211_mac_readl(hw, MAC_REG_VIF_TSF_L(vdev));
	elem->short_pulse_cnt = scm80211_phy_readl(hw, PHY_AGC_DFSDET_RPT_SHT);
	elem->long_pulse_cnt = scm80211_phy_readl(hw, PHY_AGC_DFSDET_RPT_LNG);
	local_irq_restore(flags);
	elem->tsf = get_unaligned_le64(tsf);
#endif

	return 0;
}

int scm80211_radar_info_dump(struct scm80211_dfs *dfs)
{
	struct scm80211_dfs_phy_rpt *rpt = &dfs->phy_rpt;
	int first = rpt->first;

	scm80211_trace(dfs->hw, "%s: [first %d, last %d, order %d]\n", __func__,
		       rpt->first, rpt->last, rpt->order);
	while (!dfs_elem_ptr_equal(rpt, first, rpt->last)) {
		scm80211_trace(dfs->hw, "- [%d]: tsf %lld short_pulse %d long_pulse %d,\n",
			       first, rpt->elems[ring_index(rpt, first)].tsf,
			       rpt->elems[ring_index(rpt, first)].short_pulse_cnt,
			       rpt->elems[ring_index(rpt, first)].long_pulse_cnt);
		first++;
	}

	return 0;
}

/**
 * scm80211_radar_pulse_detect()
 * Parsing radar pulse based on statistics.
 *
 */
static int scm80211_radar_pulse_detect(struct scm80211_dfs *dfs)
{
	int found = 0;
	int first;
	int long_pulse_even;
	int short_pulse_even;
	struct scm80211_dfs_phy_rpt_elem *latest;
	struct scm80211_dfs_phy_rpt_elem *elem;
	struct scm80211_dfs_phy_rpt *rpt = &dfs->phy_rpt;
	u64 delta;

	scm80211_trace(dfs->hw, "%s: chanctx#%d(%dhz) target freq[%d~%d]\n", __func__,
		       dfs->chanctx->id, dfs->chanctx->ctx.def.center_freq1,
		       dfs->radar_info->pulse_freq_min, dfs->radar_info->pulse_freq_max);

	latest = &rpt->elems[ring_index(rpt, rpt->last)];
	first = rpt->first;
	while (!dfs_elem_ptr_equal(rpt, first, rpt->last)) {
		elem = &rpt->elems[ring_index(rpt, first)];
		delta = TSF_2_MS(latest->tsf - elem->tsf);
		if (!delta) {
			scm80211_trace(dfs->hw, "- current %d last %d tsf %lld %lld\n",
				       first, rpt->last, latest->tsf, elem->tsf);
			WARN_ON(1);
			goto next;
		}
		if (ktime_ms_delta(latest->tsf, elem->tsf) > DFS_RADAR_PULSE_DRATION_MONITOR) {
			WARN_ON(first != rpt->first);
			rpt->first++;
			/* the statistics is stale */
			scm80211_trace(dfs->hw, "- skip index %lu tsf %lld latest tsf %lld\n",
				       ring_index(rpt, first), elem->tsf, latest->tsf);
			goto next;
		}

		long_pulse_even =
		    PULSE_PER_SEC(latest->long_pulse_cnt, elem->long_pulse_cnt, delta);
		short_pulse_even =
		    PULSE_PER_SEC(latest->short_pulse_cnt, elem->short_pulse_cnt, delta);
		scm80211_trace(dfs->hw, "- calculated freq short=%d long=%d delta=%lldms\n",
			       short_pulse_even, long_pulse_even, delta);
		if (short_pulse_even >= dfs->radar_info->pulse_freq_min &&
		    short_pulse_even <= dfs->radar_info->pulse_freq_max)
			found++;

		if (long_pulse_even >= dfs->radar_info->pulse_freq_min &&
		    long_pulse_even <= dfs->radar_info->pulse_freq_max)
			found++;

		if (found >= DFS_RADAR_SIGNAL_DETECTION)
			break;
next:
		first++;
	}
	scm80211_trace(dfs->hw, "- chanctx#%d(%dhz) detected %s\n", dfs->chanctx->id,
		       dfs->chanctx->ctx.def.center_freq1, found ? "true" : "false");
	return found;
}

/**
 * scm80211_radar_pulse_detect() -
 * Get and parse radar statistics.
 *
 */
static int scm80211_radar_evt_proc(struct scm80211_dfs *dfs)
{
	int detect = 0;
	struct scm80211_dfs_phy_rpt *rpt = &dfs->phy_rpt;
	struct scm80211_dfs_phy_rpt_elem elem;
	struct scm80211_dfs_phy_rpt_elem *latest;

	scm80211_radar_info_dump(dfs);
	scm80211_radar_evt_get(dfs, &elem);
	scm80211_trace(dfs->hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       dfs->chanctx->id, dfs->chanctx->ctx.def.center_freq1);
	if (elem.long_pulse_cnt == 0 && elem.short_pulse_cnt == 0) {
		scm80211_trace(dfs->hw, "- no radar pulse detected\n");
		return 0;
	}

	if (dfs_elem_ring_empty(rpt)) {
		memcpy(&rpt->elems[ring_index(rpt, rpt->last)], &elem, sizeof(elem));
		rpt->last++;
		scm80211_trace(dfs->hw, "- first radar detected short %d long %d tsf %llu\n",
			       elem.short_pulse_cnt, elem.long_pulse_cnt, elem.tsf);
		return 0;
	}

	latest = &rpt->elems[ring_index(rpt, rpt->last - 1)];
	if (elem.long_pulse_cnt == latest->long_pulse_cnt &&
	    elem.short_pulse_cnt == latest->short_pulse_cnt) {
		dfs->interval = min_t(u32, DFS_REPORT_POLLING_INTERVAL_MAX, dfs->interval * 2);
		scm80211_trace(dfs->hw, "- no more radar pulse detected\n");
	} else {
		dfs->interval = max_t(u32, DFS_REPORT_POLLING_INTERVAL_MIN, dfs->interval / 2);
	}

	memcpy(&rpt->elems[ring_index(rpt, rpt->last)], &elem, sizeof(elem));
	detect = scm80211_radar_pulse_detect(dfs);
	/* Just for convenient, move last++ here */
	rpt->last++;

	if (dfs_elem_ring_full(rpt)) {
		/* It does not make sense when the statistics is full before radar pulse detected.
		 * In ideal condition, either will happen before ring full:
		 * 1. radar pulse detected.
		 * 2. the stale statistics is remove before ring full.
		 */
		scm80211_err(dfs->hw, "- full, no radar pulse detected\n");
		rpt->first += ring_size(rpt) / 2;
	}

	return detect;
}

/**
 * scm80211_dfs_timer_handle() - Radar timer handle
 *
 */
static enum hrtimer_restart scm80211_dfs_timer_handle(struct hrtimer *timer)
{
	struct scm80211_dfs *dfs = container_of(timer, struct scm80211_dfs, timer);

	scm80211_trace(dfs->hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       dfs->chanctx->id, dfs->chanctx->ctx.def.center_freq1);
	if (scm80211_radar_evt_proc(dfs)) {
		scm80211_trace(dfs->hw, "- Radar pulse detected\n");
		ieee80211_radar_detected(dfs->hw);
	} else {
		scm80211_trace(dfs->hw, "- restart timer interval %dms\n", dfs->interval);
		hrtimer_start(timer, ms_to_ktime(dfs->interval), HRTIMER_MODE_REL_SOFT);
	}

	return HRTIMER_NORESTART;
}

/**
 * scm80211_dfs_enable() -
 * Enable radar detection.
 */
void scm80211_dfs_enable(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_dfs *dfs = chanctx->dfs;
	struct scm80211_phy *phy;
	u32 v;
	u32 pulse_duration_max;

	dfs->chanctx = chanctx;
	dfs->interval = DFS_REPORT_POLLING_INTERVAL_MAX;
	dfs->timer.function = scm80211_dfs_timer_handle;
	scm80211_trace(dfs->hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       dfs->chanctx->id, dfs->chanctx->ctx.def.center_freq1);
	for_each_phy(phy, hw) {
		dfs->phy = phy;
		break;
	}
	/* enable radar pulse detection */
	v = scm80211_mac_readl(hw, PHY_AGC_DFSDET_CFG0);
	// short pulse duration
	pulse_duration_max = DURATION_TO_TICK(dfs->phy_param.short_pulse_dur_max, dfs->clock);
	v |= (FIELD_PREP(PHY_AGC_SHT_PULSE_TMAX_MASK, pulse_duration_max));
	// long pulse duration
	pulse_duration_max = DURATION_TO_TICK(dfs->phy_param.long_pulse_dur_max, dfs->clock);
	v |= (FIELD_PREP(PHY_AGC_LNG_PULSE_TMAX_MASK, pulse_duration_max));
	// enable
	v |= (FIELD_PREP(PHY_AGC_RADAR_DET_EN_MASK, 1));
	scm80211_phy_writel(hw, v, PHY_AGC_DFSDET_CFG0);
	/* set radar pulse attributes here */

	dfs->enable = 1;
	scm80211_trace(hw, "- domain=%d\n", dfs->radar_info->domain);
	scm80211_trace(hw, "- pps(short)   min=%d max=%d\n", dfs->radar_info->pulse_freq_min,
		       dfs->radar_info->pulse_freq_max);
	scm80211_trace(hw, "- pps(long)    min=%d max=%d\n", dfs->radar_info->long_pulse_freq_min,
		       dfs->radar_info->long_pulse_freq_max);
	scm80211_trace(hw, "- width(short) min=%dus max=%dus\n", dfs->radar_info->pulse_width_min,
		       dfs->radar_info->pulse_width_max);
	scm80211_trace(hw, "- width(long)  min=%dus max=%dus\n",
		       dfs->radar_info->long_pulse_width_min,
		       dfs->radar_info->long_pulse_width_max);
	scm80211_trace(hw, "- reg value 0x%x\n", v);
}

void scm80211_dfs_disable(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_dfs *dfs = chanctx->dfs;
	u32 v;

	scm80211_trace(dfs->hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       dfs->chanctx->id, dfs->chanctx->ctx.def.center_freq1);
	/* enable radar pulse detection */
	v = scm80211_mac_readl(hw, PHY_AGC_DFSDET_CFG0);
	v |= (FIELD_PREP(PHY_AGC_RADAR_DET_EN_MASK, 0));
	scm80211_phy_writel(hw, v, PHY_AGC_DFSDET_CFG0);
	hrtimer_try_to_cancel(&chanctx->dfs->timer);

	dfs->enable = 0;
}

void scm80211_dfs_detect_start(struct scm80211_chanctx *chanctx)
{
	struct scm80211_dfs *dfs = chanctx->dfs;

	scm80211_trace(dfs->hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       dfs->chanctx->id, dfs->chanctx->ctx.def.center_freq1);
	hrtimer_try_to_cancel(&dfs->timer);
	memset(&dfs->phy_rpt, 0, sizeof(dfs->phy_rpt));
	/* Get the origial radar statistics */
	dfs->phy_rpt.order = DFS_REPORT_ELEMENT_ORDER;

	hrtimer_start(&dfs->timer, ms_to_ktime(dfs->interval), HRTIMER_MODE_REL_SOFT);
	if (!dfs->enable) {
		// FIXME: enable should be set before here
		WARN_ON(1);
		dfs->enable = 1;
	}
}

static void scm80211_dfs_param_calc(struct scm80211_dfs_radars_info *radar_info)
{
	u32 i;
	u32 pps_min = UINT_MAX, pps_max = 0;
	u32 width_min = UINT_MAX, width_max = 0;
	struct scm80211_dfs_pulse *radars;

	radars = radar_info->dfs_radars;
	if (!radars)
		return;

	for (i = 0; i < radar_info->radar_num; i++) {
		// pulse PPS
		if (pps_min > radars->pulse_per_sec_min)
			pps_min = radars->pulse_per_sec_min;
		if (pps_max < radars->pulse_per_sec_max)
			pps_max = radars->pulse_per_sec_max;
		// pulse width
		if (width_min > radars->pulse_width_min)
			width_min = radars->pulse_width_min;
		if (width_max < radars->pulse_width_max)
			width_max = radars->pulse_width_max;
		radars++;
	}

	radar_info->pulse_freq_min = pps_min;
	radar_info->pulse_freq_max = pps_max;
	radar_info->pulse_width_min = width_min;
	radar_info->pulse_width_max = width_max;
}

static void scm80211_reg_notifier(struct wiphy *wiphy, struct regulatory_request *request)
{
	int support = false;
	struct ieee80211_hw *hw = wiphy_to_ieee80211_hw(wiphy);
	struct ieee80211_hw_priv *priv = hw->priv;

	if (!request)
		return;

	if (priv->radar_info.domain == request->dfs_region) {
		scm80211_trace(hw, "- dfs region %d not changed\n", request->dfs_region);
		return;
	}
	switch (request->dfs_region) {
	case NL80211_DFS_UNSET:
		scm80211_trace(hw, "- dfs region unset\n");
		break;
	case NL80211_DFS_FCC:
		scm80211_trace(hw, "- dfs region FCC\n");
		support = true;
		priv->radar_info.dfs_radars = dfs_fcc_radars;
		priv->radar_info.radar_num = ARRAY_SIZE(dfs_fcc_radars);
		break;
	case NL80211_DFS_ETSI:
		scm80211_trace(hw, "- dfs region ETSI\n");
		support = true;
		priv->radar_info.dfs_radars = dfs_etsi_radars;
		priv->radar_info.radar_num = ARRAY_SIZE(dfs_etsi_radars);
		break;
	case NL80211_DFS_JP:
		scm80211_trace(hw, "- dfs region JP\n");
		break;
	default:
		scm80211_trace(hw, "- unknown dfs region %d\n", request->dfs_region);
		break;
	}

	if (!support)
		return;

	priv->radar_info.domain = request->dfs_region;
	scm80211_dfs_param_calc(&priv->radar_info);
	scm80211_trace(hw, "- pps(short)   min=%d max=%d\n", priv->radar_info.pulse_freq_min,
		       priv->radar_info.pulse_freq_max);
	scm80211_trace(hw, "- pps(long)    min=%d max=%d\n", priv->radar_info.long_pulse_freq_min,
		       priv->radar_info.long_pulse_freq_max);
	scm80211_trace(hw, "- width(short) min=%dus max=%dus\n", priv->radar_info.pulse_width_min,
		       priv->radar_info.pulse_width_max);
	scm80211_trace(hw, "- width(long)  min=%dus max=%dus\n",
		       priv->radar_info.long_pulse_width_min,
		       priv->radar_info.long_pulse_width_max);
}

int scm80211_dfs_init(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct scm80211_dfs *dfs;
	struct ieee80211_hw_priv *priv = hw->priv;

	scm80211_trace(hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       chanctx->id, chanctx->ctx.def.center_freq1);

	dfs = kzalloc(sizeof(*dfs), GFP_ATOMIC);
	if (!dfs) {
		scm80211_err(hw, "- memory alloc fail\n");
		return -ENOMEM;
	}

	/*  */
	hrtimer_init(&dfs->timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
	dfs->hw = hw;
	dfs->chanctx = chanctx;
	dfs->radar_info = &priv->radar_info;
	dfs->clock = SCM2020_CLOCK_FEQ;

	dfs->phy_param.short_pulse_dur_max = dfs->radar_info->pulse_width_max;
	dfs->phy_param.long_pulse_dur_max = dfs->radar_info->long_pulse_width_max;
	chanctx->dfs = dfs;

	return 0;
}

int scm80211_dfs_deinit(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	scm80211_trace(hw, "%s: chanctx#%d(%dhz)\n", __func__,
		       chanctx->id, chanctx->ctx.def.center_freq1);
	if (!chanctx->dfs)
		return 0;

	hrtimer_try_to_cancel(&chanctx->dfs->timer);

	kfree(chanctx->dfs);
	chanctx->dfs = NULL;
	return 0;
}

void scm80211_dfs_reg_init(struct ieee80211_hw *hw)
{
	hw->wiphy->regulatory_flags = REGULATORY_CUSTOM_REG;
	hw->wiphy->reg_notifier = scm80211_reg_notifier;
}

void scm80211_dfs_reg_deinit(struct ieee80211_hw *hw)
{
	hw->wiphy->reg_notifier = NULL;
}

#endif //SCM80211_DFS_CERTIFIED
