/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifndef __SCM80211_DFS_H__
#define __SCM80211_DFS_H__
#include <linux/types.h>
#include <linux/workqueue.h>
#include <net/mac80211.h>
#define DFS_REPORT_POLLING_INTERVAL_MAX 5000L	//ms
#define DFS_REPORT_POLLING_INTERVAL_MIN 500L	//ms
#define DFS_REPORT_ELEMENT_ORDER 5
#define DFS_REPORT_ELEMENT_MAX 32
#define DFS_RADAR_PULSE_DRATION_MONITOR (30 * 1000)	//the last 30 Seconds
#define DFS_RADAR_SIGNAL_DETECTION 1	// more than 1 times will report radar detection
#define IS_MASTER_DEVICE(_vif) (((_vif)->type == NL80211_IFTYPE_AP) || \
				((_vif)->type == NL80211_IFTYPE_P2P_GO))
#define PULSE_PER_SEC(_end, _from, _delta_ms) (((_end) - (_from)) * 1000 / (_delta_ms))
#define TICK_PER_USEC(_clock) ((_clock) / (1000000L))
#define DURATION_TO_TICK(_dur_us, _clock) ((_dur_us) * TICK_PER_USEC(_clock))

#define SCM2020_CLOCK_FEQ (80 * 1000 * 1000)	//TODO: 80Mhz
/**
 * struct scm80211_dfs_pulse - Radar pulse attributes
 *
 * @type: Radar type, e.g. FCC type1.
 * @pattern_type: 0: Fixed or 1: variable pattern type
 * @pulse_num_per_burst: Number of pulses per burst
 * @pulse_width: Width/Duration of each pulse(usecs)
 * @burst_freq_min: Minimum frequency of burst(1/PRI)
 * @burst_freq_max: Maximum frequency of burst(1/PRI)
 *
 */
struct scm80211_dfs_pulse {
	u32 type;
	u32 pattern_type;
	u32 pulse_num_per_burst;
	u32 burst_freq_min;
	u32 burst_freq_max;
	//new added
	u32 pulse_per_sec_min;
	u32 pulse_per_sec_max;
	//new added
	u32 pulse_width_min;
	u32 pulse_width_max;
};

struct scm80211_dfs_phy_param {
	/*
	 * Max duration of (long) radar sign in 20ns units
	 */
	u32 long_pulse_dur_max;
	u32 long_pulse_count_threshold;
	u32 long_pulse_detection_threshold;
	u32 long_pulse_detect_pwr_min;
	/*
	 * Max duration of (short) radar sign in 20ns units
	 */
	u32 short_pulse_dur_max;
	/* Relative power threshold in 0.5dB steps */
	u32 short_pulse_detect_pwr_min;
	u32 ACR_internal_scale;
};

struct scm80211_dfs_phy_rpt_elem {
	u32 long_pulse_cnt;	// pulse number reported by phy
	u32 short_pulse_cnt;	// pulse number reported by phy
	u64 tsf;		// TODO
};

struct scm80211_dfs_phy_rpt {
	u32 order;		/* the number of ring element is 2^order */
	u32 first;
	u32 last;
	struct scm80211_dfs_phy_rpt_elem elems[DFS_REPORT_ELEMENT_MAX];
};

struct scm80211_dfs_radars_info {
	u32 pulse_freq_min;	// short pulse
	u32 pulse_freq_max;	// short pulse
	u32 pulse_width_min;	// short pulse
	u32 pulse_width_max;	// short pulse

	u32 long_pulse_freq_min;	// long pulse
	u32 long_pulse_freq_max;	// long pulse
	u32 long_pulse_width_min;	// long pulse
	u32 long_pulse_width_max;	// long pulse

	u32 domain;
	u32 radar_num;		// total element of dfs_radars
	struct scm80211_dfs_pulse *dfs_radars;
};

/**
 * struct scm80211_dfs - DFS(for radar pulse detection) metadata
 *
 * @work: Work for checking Radar pulse periodically(@interval).
 * @interval:
 * @default: default interval for work
 *
 */
struct scm80211_dfs {
	int interval;
	int enable;
	u32 clock;
	struct scm80211_chanctx *chanctx;
	struct ieee80211_hw *hw;
	struct scm80211_phy *phy;
	struct hrtimer timer;
	struct scm80211_dfs_phy_param phy_param;
	struct scm80211_dfs_phy_rpt phy_rpt;
	struct scm80211_dfs_radars_info *radar_info;
};

void scm80211_dfs_enable(struct ieee80211_hw *hw, struct scm80211_chanctx *ctx);
void scm80211_dfs_detect_start(struct scm80211_chanctx *chanctx);
void scm80211_dfs_disable(struct ieee80211_hw *hw, struct scm80211_chanctx *ctx);
void scm80211_dfs_reg_init(struct ieee80211_hw *hw);
void scm80211_dfs_reg_deinit(struct ieee80211_hw *hw);
int scm80211_dfs_deinit(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx);
int scm80211_dfs_init(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx);

#endif /* __SCM80211_DFS_H__ */
