// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/delay.h>
#include "efuse.h"


#ifdef EFUSE_ENABLE
#define	EFUSE_TYPE_VER				3
#define	RF_CAL_VER_ID				2

#define	NAME2STR(name)				(#name)

#define	EFUSE_BK_BASE				EFUSE_BANK_1
#define EFUSE_MAX_BITS				(400)

#define	MAP_UPDATA_BIT				0xA7000000
#define	EFUSE_BD_EFLAG				(0xA5FE7F5A)

#define	MAP_UPG_BIT_CHECK(v)		(((v) & MAP_UPDATA_BIT) == MAP_UPDATA_BIT ? 1 : 0)
#define	MAP_UPG_BIT_CLEAR(v)		((v) &= ~MAP_UPDATA_BIT)
#define	MAP_UPG_BIT_MASK(v)		((v) & ~MAP_UPDATA_BIT)
#define XTAL_BIT_SYMBOL				0x10

#define EFUSE_DBG(fmt, args...)		pr_debug(fmt, ##args)
//#define EFUSE_DBG(fmt, args...)

#define	FUSE_REG_BASE				0x4011F000
#define	FUSE_REG_FLAG				0x80000000
#define	FUSE_RD_OFFSET				0x80
#define	FUSE_WD_OFFSET				0x10
#define	FUSE_MAX_ADDR				0x1C

#define REG32_READ(ep, reg)			((u32)scm80211_efuse_readl(ep, NULL, reg, 4))
#define REG32_WRITE(ep, val, reg)		scm80211_efuse_writel(ep, &val, reg, 4)

#define	SYMBOL_FLAG				0x80
#define	SYMBOL(d)				(((d) & SYMBOL_FLAG) ? "-" : " ")
#define	CALTEMP2STR(d)			((d) & 0x7F)
#define CALDAT_DEC_STR(d)		(((d >> 4) & 0x7) * 10)
#define CALDAT_UNT_STR(d)		(((d) & 0xF) ? 5 : 0)
#define CALDAT2STR(d)			\
			(((d) & SYMBOL_FLAG) ? \
			(0 - (CALDAT_DEC_STR(d) + CALDAT_UNT_STR(d))) : \
			(CALDAT_DEC_STR(d) + CALDAT_UNT_STR(d)))


typedef int (*efuse_cmd)(void *bd_map, void *t_map, char *str, char *rsp);
struct efuse_cmd_proc {
	char *cmd_name;
	efuse_cmd cmd_proc;
};

static void *dev_ep;
static bool dev_flag;	// 0: PCIE_DEV, 1: USB_DEV
static u8 bit_map_t[EFUSE_MAX_BANK][EFUSE_BANK_SIZE] = { 0 };
static struct efuse_bdata_map board_data_map = { 0 };

static const u8 efuse_type_size_list[MAX_EFUSE_FLD_ID] = {
//id    0, 1, 2, 3, 4, 5, 6, 7, 8, 9, A, B, C, D, E, F
	0, 1, 1, 6, 6, 2, 2, 0, 0, 2, 4, 5, 0, 2, 4, 5
};

static void update_efuse_to_bdata_map(struct efuse_bdata_map *bd_map, struct efuse_bit_map *map_t);

/**
 * scm2020_efuse_read_bank()
 * start_addr:	0x0 - 0x1C
 * len:		min bytes 4, max bytes 32. must be multiple of 4
 */
static int scm2020_efuse_read_bank(enum sens_efuse_banks r_bank,
				   u8 start_addr,
				   u32 *rbuff, int len, char *rsp)
{
	char *ptr = rsp;
	u32 fuse_rdcfg_addr = FUSE_REG_BASE + 0x0;
	u32 fuse_bank_addr = FUSE_REG_BASE + FUSE_RD_OFFSET + (r_bank * 0x20);
	u32 rbyte = 0, val = 0;
	int idx = 0;
	int i = 0;

	ptr += sprintf(ptr, "\n");
	if (!dev_ep) {
		sprintf(ptr, "[EFUSE] dev_ep is NULL!\n");
		return -1;
	}
	if (start_addr > FUSE_MAX_ADDR) {
		sprintf(ptr, "[EFUSE] bad read addr 0x%08X\n", start_addr);
		return -2;
	}

	rbyte = REG32_READ(dev_ep, fuse_rdcfg_addr);
	val = rbyte & (~(0x1 << 2));
	REG32_WRITE(dev_ep, val, fuse_rdcfg_addr);

	rbyte = REG32_READ(dev_ep, fuse_rdcfg_addr);
	val = rbyte | (0x1 << 2);
	REG32_WRITE(dev_ep, val, fuse_rdcfg_addr);

	rbyte = REG32_READ(dev_ep, fuse_rdcfg_addr);
	val = rbyte | 0x1;
	REG32_WRITE(dev_ep, val, fuse_rdcfg_addr);

	do {
		mdelay(5);
		rbyte = REG32_READ(dev_ep, fuse_rdcfg_addr);
		i++;
		if (i == 10)
			break;
	} while ((rbyte & FUSE_REG_FLAG) != FUSE_REG_FLAG);

	if (i == 10) {
		sprintf(ptr, "[EFUSE] read timeout!\n");
		return -3;
	}

	rbyte = 0;
	for (i = start_addr; i <= FUSE_MAX_ADDR; i += 4) {
		rbuff[idx] = REG32_READ(dev_ep, fuse_bank_addr + i);
		idx++;
		len -= 4;
		rbyte += 4;
		if (len <= 0)
			break;
	}

	return rbyte;
}

/**
 * scm2020_efuse_write_bank()
 * start_addr:	0x0 - 0x1C
 * len:		min bytes 4, max bytes 32. must be multiple of 4
 */
#ifdef SCM80211_FTM
static int scm2020_efuse_write_bank(enum sens_efuse_banks w_bank,
				    u8 start_addr,
				    u32 *wbuff, int len, char *rsp)
{
	char *ptr = rsp;
	u32 fuse_wrcfg_addr = FUSE_REG_BASE + 0x004;
	u32 fuse_prog_addr = FUSE_REG_BASE + FUSE_WD_OFFSET + start_addr;
	u32 w_addr, rbyte = 0, val = 0;
	int i = 0;

	ptr += sprintf(ptr, "\n");
	if (!dev_ep) {
		sprintf(ptr, "[EFUSE] dev_ep is NULL!\n");
		return -1;
	}
	if (len < 4) {
		sprintf(ptr, "[EFUSE] write size(%d) must be >= 4 bytes!\n", len);
		return -2;
	}

	if ((start_addr > FUSE_MAX_ADDR) || ((start_addr + len / 4) > FUSE_MAX_ADDR)) {
		sprintf(ptr, "[EFUSE] write bad write addr 0x%08X\n", fuse_prog_addr);
		return -3;
	}

	for (i = 0; i < len / 4; i++) {
		w_addr = fuse_prog_addr + i * 4;
		val = wbuff[i];
		REG32_WRITE(dev_ep, val, w_addr);
	}

	val = (0x1 << 1) | (0x1 << 2) | (0x1 << 3) | (0x1 << 4) | (w_bank << 8);
	REG32_WRITE(dev_ep, val, fuse_wrcfg_addr);

	val = (0x1 << 1) | (0x1 << 2) | (0x1 << 3) | (0x1 << 4) | (w_bank << 8) | 0x1;
	REG32_WRITE(dev_ep, val, fuse_wrcfg_addr);

	i = 0;
	do {
		mdelay(20);
		rbyte = REG32_READ(dev_ep, fuse_wrcfg_addr);
		i++;
		if (i == 50)
			break;
	} while ((rbyte & FUSE_REG_FLAG) != FUSE_REG_FLAG);
	if (i >= 50) {
		ptr += sprintf(ptr, "[EFUSE] write timeout(%d)!\n", i);
		return -4;
	}

	val = (0x1 << 1) | (0x1 << 2) | (0x1 << 3) | (0x1 << 4) | (w_bank << 8);
	REG32_WRITE(dev_ep, val, fuse_wrcfg_addr);
	sprintf(ptr, "[EFUSE] write complete!\n");

	return 0;
}
#endif

/********************************************************************************/
__maybe_unused static u8 calc_bytes_sbit(u8 *buf, u8 sz)
{
	u8 ret = 0;
	u8 i, n, val;

	for (i = 0; i < sz; i++) {
		val = buf[i];
		for (n = 0; n < 8; n++) {
			val >>= i;
			if (val & 0x1)
				ret++;
		}
	}
	return ret;
}

static int efuse_free_map_calc(void *bd_map, void *t_map, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)t_map;
	u32 free_byte = 0;
	u8 free_bits = 0, i, *ptr = NULL;

	if (!MAP_UPG_BIT_MASK(bd_map_t->efuse_ver)) {
		if (rsp)
			sprintf(rsp, "ver(%d) is empty!\n", MAP_UPG_BIT_MASK(bd_map_t->efuse_ver));
		return -1;
	}

	ptr = &map_t->resv_next[EFUSE_WR_SIZE - 1];
	for (i = 0; i < sizeof(map_t->resv_next); i++) {
		if (*ptr)
			break;
		free_byte++;
		ptr--;
	}

	ptr = (u8 *)&map_t->efuse_rpwr_t;
	for (i = 0; i < 8; i++) {
		if (!calc_bytes_sbit(ptr, 1))
			free_byte++;
		ptr++;
	}

	/* BT MAC */
	if (!calc_bytes_sbit(map_t->efuse_tmac_t.efuse_bt_mac_H, 3) &&
		!calc_bytes_sbit(map_t->efuse_tmac_t.efuse_bt_mac_L, 3))
		free_byte += 6;

	/* WIFI MAC */
	if (!calc_bytes_sbit(map_t->efuse_tmac_t.efuse_wifi_mac_H, 3) &&
		!calc_bytes_sbit(map_t->efuse_tmac_t.efuse_wifi_mac_L, 3))
		free_byte += 6;

	free_bits = calc_bytes_sbit((u8 *)&map_t->flag_t, 2);
	free_bits = 16 - free_bits;

	if (rsp)
		sprintf(rsp, "%-8s%3d bits\n", "free", free_byte * 8 + free_bits);
	return 0;
}

static int efuse_map_read(void *bd_map, void *t_map, char *str, char *rsp);
static int get_efuse_free_bits(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map bd_map_t = { 0 };
	u8 bit_buff[EFUSE_MAX_BANK][EFUSE_BANK_SIZE] = { 0 };
	char *ptr = rsp;

	efuse_map_read(&bd_map_t, bit_buff, str, ptr);

	ptr += strlen(ptr);
	ptr += sprintf(ptr, "\nEFUSE:\n%-8s%3d bits\n", "total", EFUSE_MAX_BITS);
	efuse_free_map_calc(&bd_map_t, bit_buff, ptr);
	return strlen(rsp);
}

__maybe_unused static int efuse_bk2_size_calc(void *t_map)
{
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)t_map;
	int i, n = 8;
	union efuse_tag_type *ptag_t[3] = {
		&map_t->efuse_tmac_t.efuse_tag_t0,
		&map_t->efuse_tmac_t.efuse_tag_t1,
		&map_t->efuse_tmac_t.efuse_tag_t2,
	};

	for (i = 0; i < 3; i++) {
		n += efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bl];
		n += efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bh];
	}
	return n;
}

static int efuse_map_show(void *bd_map, void *t_map, char *str, char *rsp)
{
	u32 *map_t = (u32 *)t_map;
	char *ptr = rsp;
	u8 i, s;

	if (!rsp)
		return 0;

	ptr += sprintf(ptr, "\n");

	for (s = 0; s < EFUSE_MAX_BANK; s++) {
		for (i = 0; i < 8; i++)
			ptr += sprintf(ptr, "[%02X] %08X\n", i * 4, *map_t++);
		ptr += sprintf(ptr, "\n");
	}

	return strlen(rsp);
}

static int bdata_map_hex(void *bd_map, void *t_map, char *str, char *rsp)
{
	u8 i, n = 0;
	u32 *p = (u32 *)bd_map;
	char *ptr = rsp;

	ptr += sprintf(ptr, "\n");

	for (i = 0; i < (sizeof(struct efuse_bdata_map) / sizeof(u32)); i++) {
		ptr += sprintf(ptr, "%08X ", *p);
		p++;
		n++;

		if (n == 4) {
			n = 0;
			ptr += sprintf(ptr, "\n");
		}
	}
	return strlen(rsp);
}

static int bdata_map_show(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 *wmac = bd_map_t->bd_mac_t.wb_mac_t.wifi_mac;
	u8 *btmac = bd_map_t->bd_mac_t.wb_mac_t.bt_mac;
	char *ptr = rsp;
	u8 xcal = MAP_UPG_BIT_MASK(bd_map_t->xtal_cal);
	int xtal_val = xcal & 0xF;

	if (xcal & XTAL_BIT_SYMBOL)
		xtal_val = 0 - xtal_val;

	ptr += sprintf(ptr, "\n");

	// 630 Bytes
	ptr += sprintf(ptr,
		"ver [%d-%d]\n"
		"\n%-12s= %d\n%-12s= %d\n"
		"%-12s= %d\n%-12s= %d\n"
		"\n"
		"%-12s= %d\n%-12s= %d\n"
		"%-12s= %d\n%-12s= %d\n"
		"%-12s= %d\n%-12s= %d\n"
		"%-12s= %d\n"
		"\n"
		"%-16s= %s%-4d\n%-16s= %s%-4d\n"
		"%-16s= %s%-4d\n%-16s= %s%-4d\n"
		"\n"
		"%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n"
		"%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n"
		"\n"
		"%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n"
		"%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n%-16s= %3d\n"
		"\n"
		"%-9s= %02X:%02X:%02X:%02X:%02X:%02X\n"
		"%-9s= %02X:%02X:%02X:%02X:%02X:%02X\n"
		"\n",
		EFUSE_TYPE_VER, RF_CAL_VER_ID,
		NAME2STR(ANT), MAP_UPG_BIT_MASK(bd_map_t->ant_sel),
		NAME2STR(dn_flag), MAP_UPG_BIT_MASK(bd_map_t->bt_dn_flag),
		NAME2STR(w_disable), MAP_UPG_BIT_MASK(bd_map_t->bt_wdis_uart),
		NAME2STR(oui), MAP_UPG_BIT_MASK(bd_map_t->sens_oui),

		NAME2STR(efuse_ver), MAP_UPG_BIT_MASK(bd_map_t->efuse_ver),
		NAME2STR(rf_cal_ver), MAP_UPG_BIT_MASK(bd_map_t->rf_cal_ver),
		NAME2STR(t_t_d), MAP_UPG_BIT_MASK(bd_map_t->t_t_d),
		NAME2STR(t_r_d), MAP_UPG_BIT_MASK(bd_map_t->t_r_d),
		NAME2STR(f_t_d), MAP_UPG_BIT_MASK(bd_map_t->f_t_d),
		NAME2STR(f_r_d), MAP_UPG_BIT_MASK(bd_map_t->f_r_d),
		NAME2STR(xtal_cal),  xtal_val,

		NAME2STR(tx_temp_2g_cal), SYMBOL(bd_map_t->tx_temp_2g_cal),
		CALTEMP2STR(bd_map_t->tx_temp_2g_cal),
		NAME2STR(rx_temp_2g_cal), SYMBOL(bd_map_t->rx_temp_2g_cal),
		CALTEMP2STR(bd_map_t->rx_temp_2g_cal),
		NAME2STR(tx_temp_5g_cal), SYMBOL(bd_map_t->tx_temp_5g_cal),
		CALTEMP2STR(bd_map_t->tx_temp_5g_cal),
		NAME2STR(rx_temp_5g_cal), SYMBOL(bd_map_t->rx_temp_5g_cal),
		CALTEMP2STR(bd_map_t->rx_temp_5g_cal),

		NAME2STR(ch0_pwr_2g_cal), CALDAT2STR(bd_map_t->ch0_pwr_2g_cal),
		NAME2STR(ch1_pwr_2g_cal), CALDAT2STR(bd_map_t->ch1_pwr_2g_cal),
		NAME2STR(ch0_pwr_5gl_cal), CALDAT2STR(bd_map_t->ch0_pwr_5gl_cal),
		NAME2STR(ch1_pwr_5gl_cal), CALDAT2STR(bd_map_t->ch1_pwr_5gl_cal),
		NAME2STR(ch0_pwr_5gm_cal), CALDAT2STR(bd_map_t->ch0_pwr_5gm_cal),
		NAME2STR(ch1_pwr_5gm_cal), CALDAT2STR(bd_map_t->ch1_pwr_5gm_cal),
		NAME2STR(ch0_pwr_5gh_cal), CALDAT2STR(bd_map_t->ch0_pwr_5gh_cal),
		NAME2STR(ch1_pwr_5gh_cal), CALDAT2STR(bd_map_t->ch1_pwr_5gh_cal),

		NAME2STR(ch0_rssi_2g_cal), CALDAT2STR(bd_map_t->ch0_rssi_2g_cal),
		NAME2STR(ch1_rssi_2g_cal), CALDAT2STR(bd_map_t->ch1_rssi_2g_cal),
		NAME2STR(ch0_rssi_5gl_cal), CALDAT2STR(bd_map_t->ch0_rssi_5gl_cal),
		NAME2STR(ch1_rssi_5gl_cal), CALDAT2STR(bd_map_t->ch1_rssi_5gl_cal),
		NAME2STR(ch0_rssi_5gm_cal), CALDAT2STR(bd_map_t->ch0_rssi_5gm_cal),
		NAME2STR(ch1_rssi_5gm_cal), CALDAT2STR(bd_map_t->ch1_rssi_5gm_cal),
		NAME2STR(ch0_rssi_5gh_cal), CALDAT2STR(bd_map_t->ch0_rssi_5gh_cal),
		NAME2STR(ch1_rssi_5gh_cal), CALDAT2STR(bd_map_t->ch1_rssi_5gh_cal),
		NAME2STR(wifi_mac),
		wmac[0], wmac[1], wmac[2], wmac[3], wmac[4], wmac[5],
		NAME2STR(bt_mac),
		btmac[0], btmac[1], btmac[2], btmac[3], btmac[4], btmac[5]);

	ptr += sprintf(ptr, "\n");
	return strlen(rsp);
}

static int set_dev_2ant(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;

	if (!bd_map_t->ant_sel) {
		bd_map_t->ant_sel = 1;
		bd_map_t->ant_sel |= MAP_UPDATA_BIT;
	}
	return 0;
}

static int set_dev_oui(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;

	if (!bd_map_t->sens_oui) {
		bd_map_t->sens_oui = 1;
		bd_map_t->sens_oui |= MAP_UPDATA_BIT;
	}
	return 0;
}

static int set_dev_bt_dn(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;

	if (!bd_map_t->bt_dn_flag) {
		bd_map_t->bt_dn_flag = 1;
		bd_map_t->bt_dn_flag |= MAP_UPDATA_BIT;
	}
	return 0;
}

static int set_dev_bt_uart(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;

	if (!bd_map_t->bt_wdis_uart) {
		bd_map_t->bt_wdis_uart = 1;
		bd_map_t->bt_wdis_uart |= MAP_UPDATA_BIT;
	}
	return 0;
}

// str: "2g/5g <tx temperature> <rx temperature>"
static int set_tx_temp(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 n, bw, dec[2] = { 0 }, sbl[2] = { 0 };
	char *spl[3] = { 0 };
	int val;

	n = string_split(spl, str, ",", 3);
	if (n != 3) {
		sprintf(rsp, "set tx temper format error!\n");
		return -1;
	}

	bw = spl[0][0] - 0x30;

	val = decstr2int(spl[1], POS_10);
	if (val < 0)
		sbl[0] = SYMBOL_FLAG;
	dec[0] = abs(val);

	val = decstr2int(spl[2], POS_10);
	if (val < 0)
		sbl[1] = SYMBOL_FLAG;
	dec[1] = abs(val);

	if (bw == 2) {
		bd_map_t->tx_temp_2g_cal = (dec[0] & 0xFF) | MAP_UPDATA_BIT | sbl[0];
		bd_map_t->rx_temp_2g_cal = (dec[1] & 0xFF) | MAP_UPDATA_BIT | sbl[1];
		bd_map_t->t_t_d = 1 | MAP_UPDATA_BIT;
		bd_map_t->t_r_d = 1 | MAP_UPDATA_BIT;
	} else if (bw == 5) {
		bd_map_t->tx_temp_5g_cal = (dec[0] & 0xFF) | MAP_UPDATA_BIT | sbl[0];
		bd_map_t->rx_temp_5g_cal = (dec[1] & 0xFF) | MAP_UPDATA_BIT | sbl[1];
		bd_map_t->t_t_d = 1 | MAP_UPDATA_BIT;
		bd_map_t->t_r_d = 1 | MAP_UPDATA_BIT;
	} else {
		sprintf(rsp, "set tx temper parameters error!\n");
		return -2;
	}
	return 0;
}

// str: "2g/5g 0/1(ch) <2g/5gl*10> <5gm*10> <5gh*10>"
static int set_tx_power(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 bw, ch, dec[3] = { 0 }, dflag[3] = { 0 };
	u8 i, n, unt, sbl[3] = { 0 };
	char *spl[5] = { 0 };
	int val;

	n = string_split(spl, str, ",", 5);
	bw = spl[0][0] - 0x30;
	ch = spl[1][0] - 0x30;
	if ((bw == 2 && n != 3) || (bw == 5 && n != 5)) {
		sprintf(rsp, "set tx power format error!\n");
		return -1;
	}

	for (i = 0; i < 3; i++) {
		val = decstr2int(spl[i + 2], POS_10);
		if (val < 0)
			sbl[i] = SYMBOL_FLAG;
		dec[i] = abs(val) / 10;
		unt = abs(val) % 10;
		if (dec[i] > 7) {
			sprintf(rsp,
				"set tx power out of range ch(%d): spl: %s {sbl(%d) %X,%X}!\n",
				ch, spl[i + 2], sbl[i], dec[i], unt);
			return -2;
		}
		if (unt > 2 && unt < 8)
			dflag[i] = 1;
		else if (unt > 7)
			dec[i] += 1;
		if (bw == 2)
			break;
	}

	if (bw == 2) {		//2g
		if (!ch) {
			bd_map_t->ch0_pwr_2g_cal = ((dec[0] << 4) & 0xF0) | dflag[0] | sbl[0];
			bd_map_t->ch0_pwr_2g_cal |= MAP_UPDATA_BIT;
		} else {
			bd_map_t->ch1_pwr_2g_cal = ((dec[0] << 4) & 0xF0) | dflag[0] | sbl[0];
			bd_map_t->ch1_pwr_2g_cal |= MAP_UPDATA_BIT;
		}
		bd_map_t->f_t_d = 1 | MAP_UPDATA_BIT;
	} else if (bw == 5) {	//5g
		if (!ch) {
			bd_map_t->ch0_pwr_5gl_cal = ((dec[0] << 4) & 0xF0) | dflag[0] | sbl[0];
			bd_map_t->ch0_pwr_5gl_cal |= MAP_UPDATA_BIT;
			bd_map_t->ch0_pwr_5gm_cal = ((dec[1] << 4) & 0xF0) | dflag[1] | sbl[1];
			bd_map_t->ch0_pwr_5gm_cal |= MAP_UPDATA_BIT;
			bd_map_t->ch0_pwr_5gh_cal = ((dec[2] << 4) & 0xF0) | dflag[2] | sbl[2];
			bd_map_t->ch0_pwr_5gh_cal |= MAP_UPDATA_BIT;
		} else {
			bd_map_t->ch1_pwr_5gl_cal = ((dec[0] << 4) & 0xF0) | dflag[0] | sbl[0];
			bd_map_t->ch1_pwr_5gl_cal |= MAP_UPDATA_BIT;
			bd_map_t->ch1_pwr_5gm_cal = ((dec[1] << 4) & 0xF0) | dflag[1] | sbl[1];
			bd_map_t->ch1_pwr_5gm_cal |= MAP_UPDATA_BIT;
			bd_map_t->ch1_pwr_5gh_cal = ((dec[2] << 4) & 0xF0) | dflag[2] | sbl[2];
			bd_map_t->ch1_pwr_5gh_cal |= MAP_UPDATA_BIT;
		}
		bd_map_t->f_t_d = 1 | MAP_UPDATA_BIT;
	} else {
		sprintf(rsp, "set tx power parameters error!\n");
		return -3;
	}
	return 0;
}

static int set_efuse_ver(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u32 ver, ef_ver;

	ver = decstr2int(str, POS_10) & 0xF;
	ef_ver = MAP_UPG_BIT_MASK(bd_map_t->efuse_ver);
	if (ef_ver || (ver <= ef_ver)) {
		sprintf(rsp, "check efuse ver(%d) fail\n", ef_ver);
		return 0;
	}

	bd_map_t->efuse_ver = ver | MAP_UPDATA_BIT;
	return 0;
}

static int set_rf_cal_ver(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u32 ver, rf_ver;

	ver = decstr2int(str, POS_10) & 0xF;
	rf_ver = MAP_UPG_BIT_MASK(bd_map_t->rf_cal_ver);
	if (rf_ver || (ver <= rf_ver))
		return 0;

	bd_map_t->rf_cal_ver = ver | MAP_UPDATA_BIT;
	return 0;
}

static int str_mac2hex_mac(u8 *mac, char *str_mac)
{
	u8 i, n;
	char *spl[6] = { 0 };

	n = string_split(spl, str_mac, ":", 6);
	if (n != 6)
		return -1;

	for (i = 0; i < n; i++)
		*(mac + i) = decstr2int(spl[i], 16);

	return 0;
}

static int efuse_wifi_mac_set(void *bd_map,
			      void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 *w_mac = bd_map_t->bd_mac_t.wb_mac_t.wifi_mac;
	u8 mac[6] = { 0 };

	if (str_mac2hex_mac(mac, str) < 0) {
		sprintf(rsp, "MAC address format error!\n");
		return -1;
	}

	memcpy(w_mac, mac, 6);
	bd_map_t->bd_mac_t.wb_mac_t.wmac_cflag = MAP_UPDATA_BIT >> 24;

	return 0;
}

static int efuse_wifi_mac_get(void *bd_map,
			      void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 *mac = bd_map_t->bd_mac_t.wb_mac_t.wifi_mac;

	sprintf(rsp, "\n%02X:%02X:%02X:%02X:%02X:%02X\n",
		mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

	return strlen(rsp);
}

static int efuse_bt_mac_set(void *bd_map, void *t_map,
			    char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 *b_mac = bd_map_t->bd_mac_t.wb_mac_t.bt_mac;
	u8 mac[6] = { 0 };

	if (str_mac2hex_mac(mac, str) < 0) {
		sprintf(rsp, "MAC address format error!\n");
		return -1;
	}

	memcpy(b_mac, mac, 6);
	bd_map_t->bd_mac_t.wb_mac_t.bmac_cflag = MAP_UPDATA_BIT >> 24;

	return 0;
}

static int efuse_bt_mac_get(void *bd_map, void *t_map,
			    char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 *mac = bd_map_t->bd_mac_t.wb_mac_t.bt_mac;

	sprintf(rsp, "\n%02X:%02X:%02X:%02X:%02X:%02X\n",
		mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

	return strlen(rsp);
}

/***************************************************************************************/
void update_efuse_to_bdata_tags(struct efuse_bdata_map *bd_map, u8 *r_data, u8 tag_id)
{
	struct efuse_rssi_pwr_2g *cal_2g = NULL;
	struct efuse_rssi_pwr_5g *cal_5g = NULL;
	struct efuse_xx_flag_type *flag_t = NULL;
	struct efuse_resv2_map *resv2_t = NULL;

	switch (tag_id) {
	case SENS_EFUSE_XCAL_FLD_ID:
		if (bd_map->efuse_ver > 2) {
			resv2_t = (struct efuse_resv2_map *)r_data;
			bd_map->xtal_cal = resv2_t->xtal_cal & 0x1F;
		}
		break;
	case SENS_EFUSE_XXD_FLD_ID:
		flag_t = (struct efuse_xx_flag_type *)r_data;
		bd_map->t_t_d = flag_t->t_t_d;
		bd_map->t_r_d = flag_t->t_r_d;
		bd_map->f_t_d = flag_t->f_t_d;
		bd_map->f_r_d = flag_t->f_r_d;
		bd_map->ant_sel = flag_t->A_n;
		bd_map->bt_dn_flag = flag_t->b_d;
		bd_map->bt_wdis_uart = flag_t->b_u;
		bd_map->sens_oui = flag_t->S_0;
		break;
	case SENS_EFUSE_WIFI_MAC_FLD_ID:
		memcpy(&bd_map->bd_mac_t.wb_mac_t.wifi_mac, r_data, 6);
		break;
	case SENS_EFUSE_BT_MAC_FLD_ID:
		memcpy(&bd_map->bd_mac_t.wb_mac_t.bt_mac, r_data, 6);
		break;
	case SENS_EFUSE_TEMP_FLD_ID_CAL_2G:
		bd_map->tx_temp_2g_cal = r_data[0];
		bd_map->tx_temp_2g_cal = r_data[1];
		break;
	case SENS_EFUSE_TEMP_FLD_ID_CAL_5G:
		bd_map->tx_temp_2g_cal = r_data[0];
		bd_map->tx_temp_2g_cal = r_data[1];
		break;
	case SENS_EFUSE_TXPWR_FLD_ID_CAL_2G:
		cal_2g = (struct efuse_rssi_pwr_2g *)r_data;
		bd_map->ch0_pwr_2g_cal = (cal_2g->ch0_rpwr_2g << 4) | cal_2g->ch0_rpwr_2g_unt * 5;
		bd_map->ch1_pwr_2g_cal = (cal_2g->ch1_rpwr_2g << 4) | cal_2g->ch1_rpwr_2g_unt * 5;
		break;
	case SENS_EFUSE_RSSI_FLD_ID_CAL_2G:
		cal_2g = (struct efuse_rssi_pwr_2g *)r_data;
		bd_map->ch0_rssi_2g_cal = (cal_2g->ch0_rpwr_2g << 4) | cal_2g->ch0_rpwr_2g_unt * 5;
		bd_map->ch1_rssi_2g_cal = (cal_2g->ch1_rpwr_2g << 4) | cal_2g->ch1_rpwr_2g_unt * 5;
		break;
	case SENS_EFUSE_TXPWR_FLD_ID_CAL_5G:
		cal_5g = (struct efuse_rssi_pwr_5g *)r_data;
		bd_map->ch0_pwr_5gl_cal =
		    (cal_5g->ch0_rpwr_5g_l << 4) | cal_5g->ch0_rpwr_5g_l_unt * 5;
		bd_map->ch1_pwr_5gl_cal =
		    (cal_5g->ch1_rpwr_5g_l << 4) | cal_5g->ch1_rpwr_5g_l_unt * 5;
		bd_map->ch0_pwr_5gm_cal =
		    (cal_5g->ch0_rpwr_5g_m << 4) | cal_5g->ch0_rpwr_5g_m_unt * 5;
		bd_map->ch1_pwr_5gm_cal =
		    (cal_5g->ch1_rpwr_5g_m << 4) | cal_5g->ch1_rpwr_5g_m_unt * 5;
		bd_map->ch0_pwr_5gh_cal =
		    (cal_5g->ch0_rpwr_5g_h << 4) | cal_5g->ch0_rpwr_5g_h_unt * 5;
		bd_map->ch1_pwr_5gh_cal =
		    (cal_5g->ch1_rpwr_5g_h << 4) | cal_5g->ch1_rpwr_5g_h_unt * 5;
		break;
	case SENS_EFUSE_RSSI_FLD_ID_CAL_5G:
		cal_5g = (struct efuse_rssi_pwr_5g *)r_data;
		bd_map->ch0_rssi_5gl_cal =
		    (cal_5g->ch0_rpwr_5g_l << 4) | cal_5g->ch0_rpwr_5g_l_unt * 5;
		bd_map->ch1_rssi_5gl_cal =
		    (cal_5g->ch1_rpwr_5g_l << 4) | cal_5g->ch1_rpwr_5g_l_unt * 5;
		bd_map->ch0_rssi_5gm_cal =
		    (cal_5g->ch0_rpwr_5g_m << 4) | cal_5g->ch0_rpwr_5g_m_unt * 5;
		bd_map->ch1_rssi_5gm_cal =
		    (cal_5g->ch1_rpwr_5g_m << 4) | cal_5g->ch1_rpwr_5g_m_unt * 5;
		bd_map->ch0_rssi_5gh_cal =
		    (cal_5g->ch0_rpwr_5g_h << 4) | cal_5g->ch0_rpwr_5g_h_unt * 5;
		bd_map->ch1_rssi_5gh_cal =
		    (cal_5g->ch1_rpwr_5g_h << 4) | cal_5g->ch1_rpwr_5g_h_unt * 5;
		break;
	}
}

static void update_efuse_to_bdata_map(struct efuse_bdata_map *bd_map, struct efuse_bit_map *map_t)
{
	u8 i = 0, *p_map = NULL;
	struct efuse_resv_map *resv_t = NULL;
	union efuse_tag_type *ptag_t[3] = {
		&map_t->efuse_tmac_t.efuse_tag_t0,
		&map_t->efuse_tmac_t.efuse_tag_t1,
		&map_t->efuse_tmac_t.efuse_tag_t2,
	};

	if (!map_t->efuse_id_t.efuse_ver)
		goto END;
	else if (map_t->efuse_id_t.efuse_ver > EFUSE_TYPE_VER)
		return;

	bd_map->efuse_ver = map_t->efuse_id_t.efuse_ver;
	bd_map->rf_cal_ver = map_t->efuse_id_t.rf_cal_ver;

	bd_map->t_t_d = map_t->flag_t.t_t_d;
	bd_map->t_r_d = map_t->flag_t.t_r_d;
	bd_map->f_t_d = map_t->flag_t.f_t_d;
	bd_map->f_r_d = map_t->flag_t.f_r_d;
	bd_map->ant_sel = map_t->flag_t.A_n;
	bd_map->bt_dn_flag = map_t->flag_t.b_d;
	bd_map->bt_wdis_uart = map_t->flag_t.b_u;
	bd_map->sens_oui = map_t->flag_t.S_0;

	bd_map->tx_temp_2g_cal = map_t->efuse_temp_t.tx_temp_sloop_2g;
	bd_map->rx_temp_2g_cal = map_t->efuse_temp_t.rx_temp_sloop_2g;
	bd_map->tx_temp_5g_cal = map_t->efuse_temp_t.tx_temp_sloop_5g;
	bd_map->rx_temp_5g_cal = map_t->efuse_temp_t.rx_temp_sloop_5g;

	bd_map->ch0_pwr_2g_cal =
	    (map_t->efuse_rpwr_t.ch0_rpwr_2g << 4) | map_t->efuse_pwr_step.ch0_rpwr_2g_unt * 5;
	bd_map->ch1_pwr_2g_cal =
	    (map_t->efuse_rpwr_t.ch1_rpwr_2g << 4) | map_t->efuse_pwr_step.ch1_rpwr_2g_unt * 5;
	bd_map->ch0_pwr_5gl_cal =
	    (map_t->efuse_rpwr_t.ch0_rpwr_5g_l << 4) | map_t->efuse_pwr_step.ch0_rpwr_5g_l_unt * 5;
	bd_map->ch1_pwr_5gl_cal =
	    (map_t->efuse_rpwr_t.ch1_rpwr_5g_l << 4) | map_t->efuse_pwr_step.ch1_rpwr_5g_l_unt * 5;
	bd_map->ch0_pwr_5gm_cal =
	    (map_t->efuse_rpwr_t.ch0_rpwr_5g_m << 4) | map_t->efuse_pwr_step.ch0_rpwr_5g_m_unt * 5;
	bd_map->ch1_pwr_5gm_cal =
	    (map_t->efuse_rpwr_t.ch1_rpwr_5g_m << 4) | map_t->efuse_pwr_step.ch1_rpwr_5g_m_unt * 5;
	bd_map->ch0_pwr_5gh_cal =
	    (map_t->efuse_rpwr_t.ch0_rpwr_5g_h << 4) | map_t->efuse_pwr_step.ch0_rpwr_5g_h_unt * 5;
	bd_map->ch1_pwr_5gh_cal =
	    (map_t->efuse_rpwr_t.ch1_rpwr_5g_h << 4) | map_t->efuse_pwr_step.ch1_rpwr_5g_h_unt * 5;

	bd_map->bd_mac_t.wb_mac_t.wifi_mac[5] = map_t->efuse_tmac_t.efuse_wifi_mac_H[2];
	bd_map->bd_mac_t.wb_mac_t.wifi_mac[4] = map_t->efuse_tmac_t.efuse_wifi_mac_H[1];
	bd_map->bd_mac_t.wb_mac_t.wifi_mac[3] = map_t->efuse_tmac_t.efuse_wifi_mac_H[0];
	bd_map->bd_mac_t.wb_mac_t.wifi_mac[2] = map_t->efuse_tmac_t.efuse_wifi_mac_L[2];
	bd_map->bd_mac_t.wb_mac_t.wifi_mac[1] = map_t->efuse_tmac_t.efuse_wifi_mac_L[1];
	bd_map->bd_mac_t.wb_mac_t.wifi_mac[0] = map_t->efuse_tmac_t.efuse_wifi_mac_L[0];

	bd_map->bd_mac_t.wb_mac_t.bt_mac[5] = map_t->efuse_tmac_t.efuse_bt_mac_H[2];
	bd_map->bd_mac_t.wb_mac_t.bt_mac[4] = map_t->efuse_tmac_t.efuse_bt_mac_H[1];
	bd_map->bd_mac_t.wb_mac_t.bt_mac[3] = map_t->efuse_tmac_t.efuse_bt_mac_H[0];
	bd_map->bd_mac_t.wb_mac_t.bt_mac[2] = map_t->efuse_tmac_t.efuse_bt_mac_L[2];
	bd_map->bd_mac_t.wb_mac_t.bt_mac[1] = map_t->efuse_tmac_t.efuse_bt_mac_L[1];
	bd_map->bd_mac_t.wb_mac_t.bt_mac[0] = map_t->efuse_tmac_t.efuse_bt_mac_L[0];

	if (bd_map->efuse_ver > 2) {
		resv_t = (struct efuse_resv_map *)map_t->resv_next;
		bd_map->xtal_cal = resv_t->resv_2.xtal_cal & 0x1F;
		p_map = resv_t->resv_tags;
	} else
		p_map = map_t->resv_next;

	for (i = 0; i < 3; i++) {
		if (!ptag_t[i]->efuse_tag_x)
			break;

		update_efuse_to_bdata_tags(bd_map, p_map, ptag_t[i]->tag_b.efuse_tag_bl);
		p_map += efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bl];

		update_efuse_to_bdata_tags(bd_map, p_map, ptag_t[i]->tag_b.efuse_tag_bh);
		p_map += efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bh];
	}

END:
	bd_map->end_flag = EFUSE_BD_EFLAG;
}

static int efuse_map_read(void *bd_map, void *t_map, char *str, char *rsp)
{
	int i, ret = 0, r_bank = EFUSE_BK_BASE;
	u32 *ptr = (u32 *)t_map;

	memset(bd_map, 0, BD_MAP_SIZE);
	memset(t_map, 0, sizeof(struct efuse_bit_map));

	if (!dev_flag) {
		for (i = 0; i < 2; i++) {
			ret =
				scm2020_efuse_read_bank(r_bank, 0, ptr, EFUSE_BANK_SIZE, rsp);
			if (ret < 0)
				return ret;

			r_bank++;
			ptr += (EFUSE_BANK_SIZE / 4);
			rsp += strlen(rsp);
			mdelay(5);
		}
	} else {
		if (!dev_ep) {
			sprintf(rsp, "efuse read fail, hw_ep is NULL!\n");
			return -3;
		}
		scm80211_efuse_readl(dev_ep, t_map, 0, MAX_EFUSE_SIZE);
	}
	memset(t_map, 0, EFUSE_USED_SIZE);

	update_efuse_to_bdata_map(bd_map, t_map);

	return 0;
}

static int efuse_map_write_bk2(void *t_map, void *bd_map, u8 *w_data,
			       enum sens_efuse_field_ids tag_id, char *rsp)
{
	u8 i, len;
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)t_map;
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 *pw = NULL, *p_bk2 = NULL;
	struct efuse_resv_map *resv = NULL;
	union efuse_tag_type *ptag_t[3] = {
		&map_t->efuse_tmac_t.efuse_tag_t0,
		&map_t->efuse_tmac_t.efuse_tag_t1,
		&map_t->efuse_tmac_t.efuse_tag_t2,
	};

	tag_id &= 0xF;
	len = efuse_type_size_list[tag_id];
	if (!len)
		return -1;

	if (map_t->efuse_id_t.efuse_ver > 2) {
		resv = (struct efuse_resv_map *)map_t->resv_next;
		p_bk2 = resv->resv_tags;
	} else
		p_bk2 = map_t->resv_next;
	pw = p_bk2;

	for (i = 0; i < 3; i++) {
		if (!ptag_t[i]->tag_b.efuse_tag_bl) {
			break;
		} else if (!ptag_t[i]->tag_b.efuse_tag_bh) {
			pw += efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bl];
			break;
		}
		pw += efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bl] +
		      efuse_type_size_list[ptag_t[i]->tag_b.efuse_tag_bh];
	}
	if ((pw - p_bk2 + len) > EFUSE_WR_SIZE) {
		sprintf(rsp, "Error: %s out of range!\n", __func__);
		return -2;
	}
	if (i >= 3) {
		sprintf(rsp, "Error: tag id %d out of range!\n", i);
		return -3;
	}

	if (!ptag_t[i]->tag_b.efuse_tag_bl)
		ptag_t[i]->tag_b.efuse_tag_bl = tag_id;
	else if (!ptag_t[i]->tag_b.efuse_tag_bh)
		ptag_t[i]->tag_b.efuse_tag_bh = tag_id;

	bd_map_t->bk2_wsize += len;
	memcpy(pw, w_data, len);

	return len;
}

__maybe_unused static void update_bdata_to_efuse_map(struct efuse_bit_map *map_t,
				      struct efuse_bdata_map *bd_map, char *rsp)
{
	struct efuse_rssi_pwr_2g *rssi_2g = NULL;
	struct efuse_rssi_pwr_5g *rssi_5g = NULL;
	u8 val[4] = { 0 };

	memset(map_t, 0, sizeof(struct efuse_bit_map));

	// x_flag
	map_t->flag_t.t_t_d = bd_map->t_t_d;
	map_t->flag_t.t_r_d = bd_map->t_r_d;
	map_t->flag_t.f_t_d = bd_map->f_t_d;
	map_t->flag_t.f_r_d = bd_map->f_r_d;
	map_t->flag_t.A_n = bd_map->ant_sel;
	map_t->flag_t.b_d = bd_map->bt_dn_flag;
	map_t->flag_t.b_u = bd_map->bt_wdis_uart;
	map_t->flag_t.S_0 = bd_map->sens_oui;
	// ver
	map_t->efuse_id_t.efuse_ver = bd_map->efuse_ver & 0xF;
	map_t->efuse_id_t.rf_cal_ver = bd_map->rf_cal_ver & 0xF;
	// temp
	map_t->efuse_temp_t.tx_temp_sloop_2g = bd_map->tx_temp_2g_cal & 0xFF;
	map_t->efuse_temp_t.rx_temp_sloop_2g = bd_map->rx_temp_2g_cal & 0xFF;
	map_t->efuse_temp_t.tx_temp_sloop_5g = bd_map->tx_temp_5g_cal & 0xFF;
	map_t->efuse_temp_t.rx_temp_sloop_5g = bd_map->rx_temp_5g_cal & 0xFF;
	// w_mac
	map_t->efuse_tmac_t.efuse_wifi_mac_H[2] = bd_map->bd_mac_t.wb_mac_t.wifi_mac[5];
	map_t->efuse_tmac_t.efuse_wifi_mac_H[1] = bd_map->bd_mac_t.wb_mac_t.wifi_mac[4];
	map_t->efuse_tmac_t.efuse_wifi_mac_H[0] = bd_map->bd_mac_t.wb_mac_t.wifi_mac[3];
	map_t->efuse_tmac_t.efuse_wifi_mac_L[2] = bd_map->bd_mac_t.wb_mac_t.wifi_mac[2];
	map_t->efuse_tmac_t.efuse_wifi_mac_L[1] = bd_map->bd_mac_t.wb_mac_t.wifi_mac[1];
	map_t->efuse_tmac_t.efuse_wifi_mac_L[0] = bd_map->bd_mac_t.wb_mac_t.wifi_mac[0];
	// b_mac
	map_t->efuse_tmac_t.efuse_bt_mac_H[2] = bd_map->bd_mac_t.wb_mac_t.bt_mac[5];
	map_t->efuse_tmac_t.efuse_bt_mac_H[1] = bd_map->bd_mac_t.wb_mac_t.bt_mac[4];
	map_t->efuse_tmac_t.efuse_bt_mac_H[0] = bd_map->bd_mac_t.wb_mac_t.bt_mac[3];
	map_t->efuse_tmac_t.efuse_bt_mac_L[2] = bd_map->bd_mac_t.wb_mac_t.bt_mac[2];
	map_t->efuse_tmac_t.efuse_bt_mac_L[1] = bd_map->bd_mac_t.wb_mac_t.bt_mac[1];
	map_t->efuse_tmac_t.efuse_bt_mac_L[0] = bd_map->bd_mac_t.wb_mac_t.bt_mac[0];
	// pwr
	map_t->efuse_pwr_step.ch0_rpwr_2g_unt = (bd_map->ch0_pwr_2g_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch0_rpwr_2g = (bd_map->ch0_pwr_2g_cal >> 4) & 0xF;
	map_t->efuse_pwr_step.ch1_rpwr_2g_unt = (bd_map->ch1_pwr_2g_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch1_rpwr_2g = (bd_map->ch1_pwr_2g_cal >> 4) & 0xF;

	map_t->efuse_pwr_step.ch0_rpwr_5g_l_unt = (bd_map->ch0_pwr_5gl_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch0_rpwr_5g_l = (bd_map->ch0_pwr_5gl_cal >> 4) & 0xF;

	map_t->efuse_pwr_step.ch1_rpwr_5g_l_unt = (bd_map->ch1_pwr_5gl_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch1_rpwr_5g_l = (bd_map->ch1_pwr_5gl_cal >> 4) & 0xF;

	map_t->efuse_pwr_step.ch0_rpwr_5g_m_unt = (bd_map->ch0_pwr_5gm_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch0_rpwr_5g_m = (bd_map->ch0_pwr_5gm_cal >> 4) & 0xF;

	map_t->efuse_pwr_step.ch1_rpwr_5g_m_unt = (bd_map->ch1_pwr_5gm_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch1_rpwr_5g_m = (bd_map->ch1_pwr_5gm_cal >> 4) & 0xF;

	map_t->efuse_pwr_step.ch0_rpwr_5g_h_unt = (bd_map->ch0_pwr_5gh_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch0_rpwr_5g_h = (bd_map->ch0_pwr_5gh_cal >> 4) & 0xF;

	map_t->efuse_pwr_step.ch1_rpwr_5g_h_unt = (bd_map->ch1_pwr_5gh_cal & 0xF) ? 1 : 0;
	map_t->efuse_rpwr_t.ch1_rpwr_5g_h = (bd_map->ch1_pwr_5gh_cal >> 4) & 0xF;

	if (bd_map->ch0_rssi_2g_cal || bd_map->ch1_rssi_2g_cal) {
		rssi_2g = (struct efuse_rssi_pwr_2g *)val;
		rssi_2g->ch0_rpwr_2g_unt = (bd_map->ch0_pwr_2g_cal & 0xF) ? 1 : 0;
		rssi_2g->ch0_rpwr_2g = (bd_map->ch0_pwr_2g_cal >> 4) & 0xF;
		rssi_2g->ch1_rpwr_2g_unt = (bd_map->ch1_pwr_2g_cal & 0xF) ? 1 : 0;
		rssi_2g->ch1_rpwr_2g = (bd_map->ch1_pwr_2g_cal >> 4) & 0xF;
		efuse_map_write_bk2(map_t, bd_map, val, SENS_EFUSE_RSSI_FLD_ID_CAL_2G, rsp);
	}
	if (bd_map->ch0_rssi_5gl_cal || bd_map->ch1_rssi_5gl_cal ||
	    bd_map->ch0_rssi_5gm_cal || bd_map->ch1_rssi_5gm_cal ||
	    bd_map->ch0_rssi_5gh_cal || bd_map->ch1_rssi_5gh_cal) {
		rssi_5g = (struct efuse_rssi_pwr_5g *)val;

		rssi_5g->ch0_rpwr_5g_l_unt = (bd_map->ch0_pwr_5gl_cal & 0xF) ? 1 : 0;
		rssi_5g->ch0_rpwr_5g_l = (bd_map->ch0_pwr_5gl_cal >> 4) & 0xF;

		rssi_5g->ch1_rpwr_5g_l_unt = (bd_map->ch1_pwr_5gl_cal & 0xF) ? 1 : 0;
		rssi_5g->ch1_rpwr_5g_l = (bd_map->ch1_pwr_5gl_cal >> 4) & 0xF;

		rssi_5g->ch0_rpwr_5g_m_unt = (bd_map->ch0_pwr_5gm_cal & 0xF) ? 1 : 0;
		rssi_5g->ch0_rpwr_5g_m = (bd_map->ch0_pwr_5gm_cal >> 4) & 0xF;

		rssi_5g->ch1_rpwr_5g_m_unt = (bd_map->ch1_pwr_5gm_cal & 0xF) ? 1 : 0;
		rssi_5g->ch1_rpwr_5g_m = (bd_map->ch1_pwr_5gm_cal >> 4) & 0xF;

		rssi_5g->ch0_rpwr_5g_h_unt = (bd_map->ch0_pwr_5gh_cal & 0xF) ? 1 : 0;
		rssi_5g->ch0_rpwr_5g_h = (bd_map->ch0_pwr_5gh_cal >> 4) & 0xF;

		rssi_5g->ch1_rpwr_5g_h_unt = (bd_map->ch1_pwr_5gh_cal & 0xF) ? 1 : 0;
		rssi_5g->ch1_rpwr_5g_h = (bd_map->ch1_pwr_5gh_cal >> 4) & 0xF;
		efuse_map_write_bk2(map_t, bd_map, val, SENS_EFUSE_RSSI_FLD_ID_CAL_5G, rsp);
	}
}

static int efuse_check_update_idver(struct efuse_bit_map *map_t,
				    struct efuse_bdata_map *bd_map, char *rsp)
{
	if (MAP_UPG_BIT_CHECK(bd_map->efuse_ver) && !map_t->efuse_id_t.efuse_ver)
		map_t->efuse_id_t.efuse_ver = bd_map->efuse_ver & 0xF;

	if (MAP_UPG_BIT_CHECK(bd_map->rf_cal_ver) && !map_t->efuse_id_t.rf_cal_ver)
		map_t->efuse_id_t.rf_cal_ver = bd_map->rf_cal_ver & 0xF;

	return 0;
}

static int efuse_check_update_xflag(struct efuse_bit_map *map_t, struct efuse_bdata_map *bd_map)
{
	if (MAP_UPG_BIT_CHECK(bd_map->t_t_d))
		map_t->flag_t.t_t_d = bd_map->t_t_d & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->t_r_d))
		map_t->flag_t.t_r_d = bd_map->t_r_d & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->f_t_d))
		map_t->flag_t.f_t_d = bd_map->f_t_d & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->f_r_d))
		map_t->flag_t.f_r_d = bd_map->f_r_d & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->ant_sel))
		map_t->flag_t.A_n = bd_map->ant_sel & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->bt_dn_flag))
		map_t->flag_t.b_d = bd_map->bt_dn_flag & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->bt_wdis_uart))
		map_t->flag_t.b_u = bd_map->bt_wdis_uart & 0x1;
	if (MAP_UPG_BIT_CHECK(bd_map->sens_oui))
		map_t->flag_t.S_0 = bd_map->sens_oui & 0x1;

	return 0;
}

static int efuse_check_update_temp(struct efuse_bit_map *map_t,
				   struct efuse_bdata_map *bd_map, char *rsp)
{
	int ret = 0;
	u8 val[4] = { 0 };
	bool upd_flag = false;
	struct efuse_temper_type *pt = NULL;

	pt = (struct efuse_temper_type *)&val;
	if (MAP_UPG_BIT_CHECK(bd_map->tx_temp_2g_cal)) {
		if (map_t->efuse_temp_t.tx_temp_sloop_2g)
			upd_flag = true;
		else
			map_t->efuse_temp_t.tx_temp_sloop_2g = bd_map->tx_temp_2g_cal & 0xFF;
		pt->tx_temp_sloop_2g = bd_map->tx_temp_2g_cal & 0xFF;
	}
	if (MAP_UPG_BIT_CHECK(bd_map->rx_temp_2g_cal)) {
		if (map_t->efuse_temp_t.rx_temp_sloop_2g)
			upd_flag = true;
		else
			map_t->efuse_temp_t.rx_temp_sloop_2g = bd_map->rx_temp_2g_cal & 0xFF;
		pt->rx_temp_sloop_2g = bd_map->rx_temp_2g_cal & 0xFF;
	}
	if (upd_flag) {
		ret = efuse_map_write_bk2(map_t, bd_map, val, SENS_EFUSE_TEMP_FLD_ID_CAL_2G, rsp);
		MAP_UPG_BIT_CLEAR(bd_map->tx_temp_2g_cal);
	}

	if (MAP_UPG_BIT_CHECK(bd_map->tx_temp_5g_cal)) {
		if (map_t->efuse_temp_t.tx_temp_sloop_5g)
			upd_flag = true;
		else
			map_t->efuse_temp_t.tx_temp_sloop_5g = bd_map->tx_temp_5g_cal & 0xFF;
		pt->tx_temp_sloop_5g = bd_map->tx_temp_5g_cal & 0xFF;
	}
	if (MAP_UPG_BIT_CHECK(bd_map->rx_temp_5g_cal)) {
		if (map_t->efuse_temp_t.rx_temp_sloop_5g)
			upd_flag = true;
		else
			map_t->efuse_temp_t.rx_temp_sloop_5g = bd_map->rx_temp_5g_cal & 0xFF;
		pt->rx_temp_sloop_5g = bd_map->rx_temp_5g_cal & 0xFF;
	}
	if (upd_flag) {
		ret = efuse_map_write_bk2(map_t, bd_map, val, SENS_EFUSE_TEMP_FLD_ID_CAL_5G, rsp);
		MAP_UPG_BIT_CLEAR(bd_map->tx_temp_5g_cal);
	}
	return ret;
}

static int efuse_check_update_mac_addr(struct efuse_bit_map *map_t,
				       struct efuse_bdata_map *bd_map, char *rsp)
{
	int ret = 0;
	u8 val[6] = { 0 };
	u32 cflag = 0;
	bool upd_flag = false;

	cflag = bd_map->bd_mac_t.wb_mac_t.wmac_cflag;
	cflag <<= 24;

	if (MAP_UPG_BIT_CHECK(cflag)) {
		if (calc_bytes_sbit(map_t->efuse_tmac_t.efuse_wifi_mac_L, 3)) {
			upd_flag = true;
		} else {
			map_t->efuse_tmac_t.efuse_wifi_mac_H[2] =
			    bd_map->bd_mac_t.wb_mac_t.wifi_mac[5];
			map_t->efuse_tmac_t.efuse_wifi_mac_H[1] =
			    bd_map->bd_mac_t.wb_mac_t.wifi_mac[4];
			map_t->efuse_tmac_t.efuse_wifi_mac_H[0] =
			    bd_map->bd_mac_t.wb_mac_t.wifi_mac[3];
			map_t->efuse_tmac_t.efuse_wifi_mac_L[2] =
			    bd_map->bd_mac_t.wb_mac_t.wifi_mac[2];
			map_t->efuse_tmac_t.efuse_wifi_mac_L[1] =
			    bd_map->bd_mac_t.wb_mac_t.wifi_mac[1];
			map_t->efuse_tmac_t.efuse_wifi_mac_L[0] =
			    bd_map->bd_mac_t.wb_mac_t.wifi_mac[0];
		}
		memcpy(val, bd_map->bd_mac_t.wb_mac_t.wifi_mac, 6);
		bd_map->bd_mac_t.wb_mac_t.wmac_cflag = 0;
	}

	if (upd_flag)
		ret = efuse_map_write_bk2(map_t, bd_map, val, SENS_EFUSE_WIFI_MAC_FLD_ID, rsp);

	upd_flag = false;
	memset(val, 0, 6);
	cflag = bd_map->bd_mac_t.wb_mac_t.bmac_cflag;
	cflag <<= 24;

	if (MAP_UPG_BIT_CHECK(cflag)) {
		if (calc_bytes_sbit(map_t->efuse_tmac_t.efuse_bt_mac_L, 3)) {
			upd_flag = true;
		} else {
			map_t->efuse_tmac_t.efuse_bt_mac_H[2] = bd_map->bd_mac_t.wb_mac_t.bt_mac[5];
			map_t->efuse_tmac_t.efuse_bt_mac_H[1] = bd_map->bd_mac_t.wb_mac_t.bt_mac[4];
			map_t->efuse_tmac_t.efuse_bt_mac_H[0] = bd_map->bd_mac_t.wb_mac_t.bt_mac[3];
			map_t->efuse_tmac_t.efuse_bt_mac_L[2] = bd_map->bd_mac_t.wb_mac_t.bt_mac[2];
			map_t->efuse_tmac_t.efuse_bt_mac_L[1] = bd_map->bd_mac_t.wb_mac_t.bt_mac[1];
			map_t->efuse_tmac_t.efuse_bt_mac_L[0] = bd_map->bd_mac_t.wb_mac_t.bt_mac[0];
		}
		memcpy(val, bd_map->bd_mac_t.wb_mac_t.bt_mac, 6);
		bd_map->bd_mac_t.wb_mac_t.bmac_cflag = 0;
	}

	if (upd_flag)
		ret = efuse_map_write_bk2(map_t, bd_map, val, SENS_EFUSE_BT_MAC_FLD_ID, rsp);

	return ret;
}

static int efuse_check_update_rssi_power_2g(struct efuse_bit_map *map_t,
					    struct efuse_bdata_map *bd_map,
					    enum sens_efuse_field_ids id, char *rsp)
{
	int ret = 0;
	u8 val[2] = { 0 };
	u32 *cal_0v = NULL, *cal_1v = NULL;
	bool upd_flag = false;
	struct efuse_rssi_pwr_2g *pt = NULL;

	pt = (struct efuse_rssi_pwr_2g *)&val;
	if (id == SENS_EFUSE_TXPWR_FLD_ID_CAL_2G) {
		cal_0v = &bd_map->ch0_pwr_2g_cal;
		cal_1v = &bd_map->ch1_pwr_2g_cal;
	} else if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_2G) {
		cal_0v = &bd_map->ch0_rssi_2g_cal;
		cal_1v = &bd_map->ch1_rssi_2g_cal;
	} else
		return -1;

	if (MAP_UPG_BIT_CHECK(*cal_0v)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_2G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch0_rpwr_2g) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch0_rpwr_2g_unt = (*cal_0v & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch0_rpwr_2g = (*cal_0v >> 4) & 0xF;
			}
		}
		pt->ch0_rpwr_2g_unt = *cal_0v & 0xF;
		pt->ch0_rpwr_2g = (*cal_0v >> 4) & 0xF;
	}

	if (MAP_UPG_BIT_CHECK(*cal_1v)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_2G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch1_rpwr_2g) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch1_rpwr_2g_unt = (*cal_1v & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch1_rpwr_2g = (*cal_1v >> 4) & 0xF;
			}
		}
		pt->ch1_rpwr_2g_unt = *cal_1v & 0xF;
		pt->ch1_rpwr_2g = (*cal_1v >> 4) & 0xF;
	}

	if (upd_flag) {
		ret = efuse_map_write_bk2(map_t, bd_map, val, id, rsp);
		MAP_UPG_BIT_CLEAR(*cal_0v);
		MAP_UPG_BIT_CLEAR(*cal_1v);
	}
	return ret;
}

static int efuse_check_update_rssi_power_5g(struct efuse_bit_map *map_t,
					    struct efuse_bdata_map *bd_map,
					    enum sens_efuse_field_ids id, char *rsp)
{
	int ret = 0;
	u8 val[4] = { 0 };
	u32 *cal_0lv = NULL, *cal_1lv = NULL, *cal_0mv = NULL;
	u32 *cal_1mv = NULL, *cal_0hv = NULL, *cal_1hv = NULL;
	bool upd_flag = false;
	struct efuse_rssi_pwr_5g *pt = NULL;

	pt = (struct efuse_rssi_pwr_5g *)&val;
	if (id == SENS_EFUSE_TXPWR_FLD_ID_CAL_5G) {
		cal_0lv = &bd_map->ch0_pwr_5gl_cal;
		cal_1lv = &bd_map->ch1_pwr_5gl_cal;
		cal_0mv = &bd_map->ch0_pwr_5gm_cal;
		cal_1mv = &bd_map->ch1_pwr_5gm_cal;
		cal_0hv = &bd_map->ch0_pwr_5gh_cal;
		cal_1hv = &bd_map->ch1_pwr_5gh_cal;
	} else if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
		cal_0lv = &bd_map->ch0_rssi_5gl_cal;
		cal_1lv = &bd_map->ch1_rssi_5gl_cal;
		cal_0mv = &bd_map->ch0_rssi_5gm_cal;
		cal_1mv = &bd_map->ch1_rssi_5gm_cal;
		cal_0hv = &bd_map->ch0_rssi_5gh_cal;
		cal_1hv = &bd_map->ch1_rssi_5gh_cal;
	} else
		return -1;

	if (MAP_UPG_BIT_CHECK(*cal_0lv)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch0_rpwr_5g_l) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch0_rpwr_5g_l_unt = (*cal_0lv & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch0_rpwr_5g_l = (*cal_0lv >> 4) & 0xF;
			}
		}
		pt->ch0_rpwr_5g_l_unt = *cal_0lv & 0xF;
		pt->ch0_rpwr_5g_l = (*cal_0lv >> 4) & 0xF;
	}

	if (MAP_UPG_BIT_CHECK(*cal_1lv)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch1_rpwr_5g_l) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch1_rpwr_5g_l_unt = (*cal_1lv & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch1_rpwr_5g_l = (*cal_1lv >> 4) & 0xF;
			}
		}
		pt->ch1_rpwr_5g_l_unt = *cal_1lv & 0xF;
		pt->ch1_rpwr_5g_l = (*cal_1lv >> 4) & 0xF;
	}

	if (MAP_UPG_BIT_CHECK(*cal_0mv)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch0_rpwr_5g_m) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch0_rpwr_5g_m_unt = (*cal_0mv & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch0_rpwr_5g_m = (*cal_0mv >> 4) & 0xF;
			}
		}
		pt->ch0_rpwr_5g_m_unt = *cal_0mv & 0xF;
		pt->ch0_rpwr_5g_m = (*cal_0mv >> 4) & 0xF;
	}

	if (MAP_UPG_BIT_CHECK(*cal_1mv)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch1_rpwr_5g_m) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch1_rpwr_5g_m_unt = (*cal_1mv & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch1_rpwr_5g_m = (*cal_1mv >> 4) & 0xF;
			}
		}
		pt->ch1_rpwr_5g_m_unt = *cal_1mv & 0xF;
		pt->ch1_rpwr_5g_m = (*cal_1mv >> 4) & 0xF;
	}

	if (MAP_UPG_BIT_CHECK(*cal_0hv)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch0_rpwr_5g_h) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch0_rpwr_5g_h_unt = (*cal_0hv & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch0_rpwr_5g_h = (*cal_0hv >> 4) & 0xF;
			}
		}
		pt->ch0_rpwr_5g_h_unt = *cal_0hv & 0xF;
		pt->ch0_rpwr_5g_h = (*cal_0hv >> 4) & 0xF;
	}

	if (MAP_UPG_BIT_CHECK(*cal_1hv)) {
		if (id == SENS_EFUSE_RSSI_FLD_ID_CAL_5G) {
			upd_flag = true;
		} else {
			if (map_t->efuse_rpwr_t.ch1_rpwr_5g_h) {
				upd_flag = true;
			} else {
				map_t->efuse_pwr_step.ch1_rpwr_5g_h_unt = (*cal_1hv & 0xF) ? 1 : 0;
				map_t->efuse_rpwr_t.ch1_rpwr_5g_h = (*cal_1hv >> 4) & 0xF;
			}
		}
		pt->ch1_rpwr_5g_h_unt = *cal_1hv & 0xF;
		pt->ch1_rpwr_5g_h = (*cal_1hv >> 4) & 0xF;
	}

	if (upd_flag) {
		ret = efuse_map_write_bk2(map_t, bd_map, val, id, rsp);
		MAP_UPG_BIT_CLEAR(*cal_0lv);
		MAP_UPG_BIT_CLEAR(*cal_1lv);
		MAP_UPG_BIT_CLEAR(*cal_0mv);
		MAP_UPG_BIT_CLEAR(*cal_1mv);
		MAP_UPG_BIT_CLEAR(*cal_0hv);
		MAP_UPG_BIT_CLEAR(*cal_1hv);
	}
	return ret;
}

static int efuse_check_update_xtal_cal(struct efuse_bit_map *map_t,
				   struct efuse_bdata_map *bd_map, char *rsp)
{
	struct efuse_resv_map *resv = NULL;
	u8 val;

	if (MAP_UPG_BIT_CHECK(bd_map->xtal_cal)) {
		resv = (struct efuse_resv_map *)map_t->resv_next;
		val = bd_map->xtal_cal & 0x1F;
		if (!resv->resv_2.xtal_cal)
			resv->resv_2.xtal_cal = val;
		else
			efuse_map_write_bk2(map_t, bd_map, &val, SENS_EFUSE_XCAL_FLD_ID, rsp);
	}
	return 0;
}

static int efuse_map_write(void *bd_map, void *t_map, char *str, char *rsp)
{
	int ret = 0;
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)t_map;
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;

	if (bd_map_t->end_flag != EFUSE_BD_EFLAG) {
		sprintf(rsp, "efuse map cannot updated(0x%x)!\n", bd_map_t->end_flag);
		return -1;
	}

	if (!(bd_map_t->efuse_ver & 0xF) || !(bd_map_t->rf_cal_ver & 0xF)
		|| ((bd_map_t->efuse_ver & 0xF) > EFUSE_TYPE_VER)) {
		sprintf(rsp, "can not set efuse/rf ver(%d/%d)!\n",
			bd_map_t->efuse_ver, bd_map_t->rf_cal_ver);
		return -2;
	}

	efuse_check_update_xflag(map_t, bd_map_t);
	efuse_check_update_idver(map_t, bd_map_t, rsp);
	efuse_check_update_temp(map_t, bd_map_t, rsp);
	efuse_check_update_mac_addr(map_t, bd_map_t, rsp);
	efuse_check_update_rssi_power_2g(map_t, bd_map_t, SENS_EFUSE_TXPWR_FLD_ID_CAL_2G, rsp);
	efuse_check_update_rssi_power_2g(map_t, bd_map_t, SENS_EFUSE_RSSI_FLD_ID_CAL_2G, rsp);
	efuse_check_update_rssi_power_5g(map_t, bd_map_t, SENS_EFUSE_TXPWR_FLD_ID_CAL_5G, rsp);
	efuse_check_update_rssi_power_5g(map_t, bd_map_t, SENS_EFUSE_RSSI_FLD_ID_CAL_5G, rsp);
	efuse_check_update_xtal_cal(map_t, bd_map_t, rsp);

	memset(t_map, 0, EFUSE_USED_SIZE);
#ifdef SCM80211_FTM
	if (!dev_flag) {
		int i, wsize = EFUSE_BANK_SIZE, w_bank = EFUSE_BK_BASE;
		u32 *pdat = (u32 *)t_map;

		for (i = 0; i < 2; i++) {
			ret = scm2020_efuse_write_bank(w_bank, 0, pdat, wsize, rsp);
			if (ret < 0)
				return ret;

			w_bank++;
			pdat += (EFUSE_BANK_SIZE / 4);
			rsp += strlen(rsp);
			mdelay(5);
		}
	} else {
		if (!dev_ep) {
			sprintf(rsp, "efuse write fail, hw_ep is NULL!\n");
			return -3;
		}
		scm80211_efuse_writel(dev_ep, map_t, 0, MAX_EFUSE_SIZE);
	}
#else
	ret = -4;
	sprintf(rsp, "efuse map cannot updated(%d)!\n", ret);
#endif	/* #ifdef SCM80211_FTM */

	return ret;
}

static int set_efuse_reload(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)&bit_map_t[0];
	struct efuse_bdata_map *bd_map_t = &board_data_map;
	char temp_str[600] = { 0 };

	if (efuse_map_read(bd_map_t, map_t, NULL, temp_str) < 0)
		return -3;

	memset(temp_str, 0, sizeof(temp_str));
	EFUSE_DBG("str size: %d\n%s\n", efuse_map_show(NULL, map_t, NULL, temp_str), temp_str);

	return 0;
}

static int set_efuse_xtal_cal(void *bd_map, void *t_map, char *str, char *rsp)
{
	struct efuse_bdata_map *bd_map_t = (struct efuse_bdata_map *)bd_map;
	u8 cal;
	int val = decstr2int(str, POS_10);

	cal = abs(val);
	if (val < 0)
		cal |= XTAL_BIT_SYMBOL;
	bd_map_t->xtal_cal = cal | MAP_UPDATA_BIT;
	return 0;
}

enum EFUSE_SET_CMD_LIST {
	CMD_EMAP_WD = 0,
	CMD_DEV_2ANT,
	CMD_DEV_OUI,
	CMD_DEV_BT_DN,
	CMD_DEV_BT_SER,
	CMD_EFUSE_VER,
	CMD_RF_CAL_VER,
	CMD_WMAC_SET,
	CMD_BMAC_SET,
	CMD_TEMP_SET,
	CMD_PWR_SET,
	CMD_RELOAD,
	CMD_XTAL_SET,
	CMD_EFUSE_SET_MAX,
};

enum EFUSE_GET_CMD_LIST {
	CMD_EFUSE_HELP = 0,
	CMD_EMAP_RD,
	CMD_EMAP_SHOW,
	CMD_BDMAP_SHOW,
	CMD_WMAC_GET,
	CMD_BMAC_GET,
	CMD_EFUSE_FREE = 7,
	CMD_EFUSE_GET_MAX,
};

static int get_efuse_help(void *bd_map, void *t_map, char *str, char *rsp);

struct efuse_cmd_proc ef_set_cmd_proc[] = {
	{ "save ", efuse_map_write },

	{ "2ant ", set_dev_2ant },
	{ "oui ", set_dev_oui },
	{ "btdn ", set_dev_bt_dn },
	{ "btser ", set_dev_bt_uart },
	{ "ef_ver ", set_efuse_ver },
	{ "rf_ver ", set_rf_cal_ver },

	{ "wmac ", efuse_wifi_mac_set },
	{ "bmac ", efuse_bt_mac_set },
	{ "tx_temp ", set_tx_temp },
	{ "tx_pwr ", set_tx_power },
	{ "reload ", set_efuse_reload},
	{ "xtal_cal ", set_efuse_xtal_cal},
};

struct efuse_cmd_proc ef_get_cmd_proc[] = {
	{ "help ", get_efuse_help },
	{ "read ", efuse_map_read },
	{ "show ", efuse_map_show },
	{ "bd_show ", bdata_map_show },

	{ "wmac ", efuse_wifi_mac_get },
	{ "bmac ", efuse_bt_mac_get },
	{ "bd_hex ", bdata_map_hex },
	{ "free_sz ", get_efuse_free_bits },
};

#define	EF_MAX_SET_CMD_ID		(sizeof(ef_set_cmd_proc) / sizeof(struct efuse_cmd_proc))
#define	EF_MAX_GET_CMD_ID		(sizeof(ef_get_cmd_proc) / sizeof(struct efuse_cmd_proc))

static int get_efuse_help(void *bd_map, void *t_map, char *str, char *rsp)
{
	sprintf(rsp,
		"\nset commands:\n"
		"%-12s\twrite values into efuse\n"
		"%-12s\tset device 2-ANT module\n"
		"%-12s\tenable use internal uart download bt fw\n"
		"%-12s\tdisable internal uart pins after download bt fw\n"
		"%-12s\tef_ver <ver>\tset efuse ver\n"
		"%-12s\trf_ver <ver>\tset rf calibration ver\n"
		"%-12s\twmac <MAC address>\tset wifi mac address\n"
		"%-12s\tbmac <MAC address>\tset bt mac address\n"
		"%-12s\ttx_temp <2g tx>,<2g rx>,<5g tx>,<5g rx>\n"
		"%-12s\ttx_pwr <\"2g\"/\"5g\">,<chain \"0/1\">,"
			"<rf_2g/rf_5g_low>,<rf_5g_mid>,<rf_5g_high>\n"
		"%-12s\tefuse reload\n"
		"%-12s\tefuse xtal <xtal frequency calibration(-30 ~ 30)>\n"
		"\nget commands:\n"
		"%-12s\tread/load values from efuse\n"
		"%-12s\tbd_show configuration file\n"
		"%-12s\twmac\tget wifi mac address\n"
		"%-12s\tbmac\tget bt mac address\n"
		"%-12s\tfree_sz\tget efuse free space bits\n",
		ef_set_cmd_proc[CMD_EMAP_WD].cmd_name,
		ef_set_cmd_proc[CMD_DEV_2ANT].cmd_name,
		ef_set_cmd_proc[CMD_DEV_BT_DN].cmd_name,
		ef_set_cmd_proc[CMD_DEV_BT_DN].cmd_name,
		ef_set_cmd_proc[CMD_EFUSE_VER].cmd_name,
		ef_set_cmd_proc[CMD_RF_CAL_VER].cmd_name,
		ef_set_cmd_proc[CMD_WMAC_SET].cmd_name,
		ef_set_cmd_proc[CMD_BMAC_SET].cmd_name,
		ef_set_cmd_proc[CMD_TEMP_SET].cmd_name,
		ef_set_cmd_proc[CMD_PWR_SET].cmd_name,
		ef_set_cmd_proc[CMD_RELOAD].cmd_name,
		ef_set_cmd_proc[CMD_XTAL_SET].cmd_name,

		ef_get_cmd_proc[CMD_EMAP_RD].cmd_name,
		ef_get_cmd_proc[CMD_BDMAP_SHOW].cmd_name,
		ef_get_cmd_proc[CMD_WMAC_GET].cmd_name,
		ef_get_cmd_proc[CMD_BMAC_GET].cmd_name,
		ef_get_cmd_proc[CMD_EFUSE_FREE].cmd_name);

	return strlen(rsp);
}

/***************************************************************************************/
int scm80211_efuse_set_cmd_proc(char *cmd_str, char *rsp)
{
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)&bit_map_t[0];
	struct efuse_bdata_map *bd_map_t = &board_data_map;
	int i, ret = -1;

	if (!cmd_str || (strlen(cmd_str) < 3)) {
		sprintf(rsp, "Error: set params error!\n");
		return strlen(rsp);
	}

	for (i = 0; i < strlen(cmd_str); i++) {
		if ((cmd_str[i] == '\r') || (cmd_str[i] == '\n'))
			cmd_str[i] = '\0';
	}

	for (i = 0; i < EF_MAX_SET_CMD_ID; i++) {
		if (!strncmp(cmd_str, ef_set_cmd_proc[i].cmd_name,
		     strlen(ef_set_cmd_proc[i].cmd_name) - 1)) {
			ret = ef_set_cmd_proc[i].cmd_proc(bd_map_t, map_t,
				(cmd_str + strlen(ef_set_cmd_proc[i].cmd_name)), rsp);
			break;
		}
	}
	return ret;
}

int scm80211_efuse_get_cmd_proc(char *cmd_str, char *rsp)
{
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)&bit_map_t[0];
	struct efuse_bdata_map *bd_map_t = &board_data_map;
	int i, ret = -1;

	if (!cmd_str || (strlen(cmd_str)) < 3) {
		sprintf(rsp, "Error: get params error!\n");
		return strlen(rsp);
	}

	for (i = 0; i < strlen(cmd_str); i++) {
		if ((cmd_str[i] == '\r') || (cmd_str[i] == '\n'))
			cmd_str[i] = '\0';
	}

	for (i = 0; i < EF_MAX_GET_CMD_ID; i++) {
		if (!strncmp(cmd_str, ef_get_cmd_proc[i].cmd_name,
		     strlen(ef_get_cmd_proc[i].cmd_name) - 1)) {
			ret = ef_get_cmd_proc[i].cmd_proc(bd_map_t, map_t,
				(cmd_str + strlen(ef_get_cmd_proc[i].cmd_name)), rsp);
			break;
		}
	}
	return ret;
}

int scm80211_efuse_pcie_init(void *ep)
{
	if (!ep)
		return -2;
	dev_ep = ep;
	dev_flag = false;

	set_efuse_reload(NULL, NULL, NULL, NULL);

	return 0;
}

bool scm80211_efuse_usb_init(void *udev, void *buff, int length)
{
	struct efuse_bit_map *map_t = (struct efuse_bit_map *)&bit_map_t[0];
	struct efuse_bdata_map *bd_map_t = &board_data_map;

	if (length != MAX_EFUSE_SIZE)
		return false;

	if (!udev)
		return false;
	dev_ep = udev;
	dev_flag = true;

	memset(bd_map_t, 0, BD_MAP_SIZE);
	memcpy(map_t, buff, length);

	update_efuse_to_bdata_map(bd_map_t, map_t);

	return true;
}

bool scm80211_efuse_mac_update(u8 *wmac, u8 *btmac)
{
	struct efuse_bdata_map *bd_map_t = &board_data_map;
	u8 *mac;
	bool ret = true;

	if (bd_map_t->end_flag != EFUSE_BD_EFLAG || !bd_map_t->efuse_ver)
		return false;

	if (wmac) {
		mac = bd_map_t->bd_mac_t.wb_mac_t.wifi_mac;
		if (is_valid_ether_addr(mac))
			memcpy(wmac, mac, ETH_ALEN);
		else
			ret = false;
	}

	if (btmac) {
		mac = bd_map_t->bd_mac_t.wb_mac_t.bt_mac;
		if (is_valid_ether_addr(mac))
			memcpy(btmac, mac, ETH_ALEN);
	}
	return ret;
}

int scm80211_efuse_cal_update(u8 path, int *cal)
{
	struct efuse_bdata_map *bd_map_t = &board_data_map;

	if (bd_map_t->end_flag != EFUSE_BD_EFLAG || !bd_map_t->efuse_ver)
		return -1;

	if (path) {
		cal[0] = CALDAT2STR(bd_map_t->ch1_pwr_2g_cal);
		cal[1] = CALDAT2STR(bd_map_t->ch1_pwr_5gl_cal);
		cal[2] = CALDAT2STR(bd_map_t->ch1_pwr_5gm_cal);
		cal[3] = CALDAT2STR(bd_map_t->ch1_pwr_5gh_cal);
	} else {
		cal[0] = CALDAT2STR(bd_map_t->ch0_pwr_2g_cal);
		cal[1] = CALDAT2STR(bd_map_t->ch0_pwr_5gl_cal);
		cal[2] = CALDAT2STR(bd_map_t->ch0_pwr_5gm_cal);
		cal[3] = CALDAT2STR(bd_map_t->ch0_pwr_5gh_cal);
	}

	return 0;
}

int scm80211_efuse_get_ant_sel(void)
{
	struct efuse_bdata_map *bd_map_t = &board_data_map;

	if (bd_map_t->end_flag != EFUSE_BD_EFLAG || !bd_map_t->efuse_ver)
		return -1;
	return (bd_map_t->ant_sel & 0x1);
}

int scm80211_efuse_get_xtal_cal(void)
{
	struct efuse_bdata_map *bd_map_t = &board_data_map;
	int val = -255;

	if (bd_map_t->end_flag != EFUSE_BD_EFLAG || !bd_map_t->efuse_ver)
		return val;

	val = (bd_map_t->xtal_cal & 0xF);
	if (bd_map_t->xtal_cal & XTAL_BIT_SYMBOL)
		val = 0 - val;

	return val;
}

#endif /* EFUSE_ENABLE */
