// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */

/**
 * Limitations:
 * - Injected frames are not considered, and may not work
 * - At most A-MPDU in a TXOP
 *
 * TODO:
 * - Add a robust rate at the end of the rate
 * - Take into account frames currently in txop->sched
 * - Multi-TID A-MPDU
 */

#define func_fmt(str)			"[FTM] "str
#define func_trace				FTM

#ifdef SCM80211_FTM
#include <linux/compiler.h>
#include <linux/bitfield.h>
#include <linux/list_sort.h>
#include <linux/log2.h>
#include "core.h"
#include "util.h"
#include "hw.h"
#include "util.h"
#include "ftm.h"
#include "rf.h"
#ifdef EFUSE_ENABLE
#include "efuse.h"
#endif

static bool ftm_en;		//Factory Test Mode EN
static struct ieee80211_sta ftm_sta = { 0 };

module_param(ftm_en, bool, 0644);
MODULE_PARM_DESC(ftm_en, "Factory Test Mode Enable Flag");

#define FTM_DBG(hw, fmt, args...)   scm80211_trace(hw, fmt, ##args)


#define CTRL_QOS_NO_ACK					(0x0020)
#define CTRL_QOS_NORMAL_ACK				(0x0000)
#define PHY_REG_BANKSZ					(0x00004000)


static int scm80211_ftm_ctx_init(struct ieee80211_hw *hw, struct scm80211_ftm *ftm);
static void scm80211_ftm_hwq_tasklet(unsigned long data);
static void scm80211_ftm_work(struct work_struct *wk);
static void scm80211_ftm_tx_alloc_work(struct work_struct *wk);


static inline struct work_struct *scm80211_ftm_work_get(struct ieee80211_hw *hw,
							struct ieee80211_vif *vif)
{
	struct scm80211_hwq *hwq;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	hwq = scm80211_hwq(hw, ftm->ac);
	// TODO
	hwq->vif = vif;
	return &hwq->work;
}

static inline struct work_struct *scm80211_ftm_tx_ac_work_get(struct ieee80211_hw *hw,
							struct ieee80211_vif *vif)
{
	struct scm80211_hwq *hwq;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	hwq = scm80211_hwq(hw, ftm->ac);

	return &hwq->tx_ac_wk;
}

static inline void scm80211_ftm_tasklet_init(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct scm80211_hwq *hwq;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	hwq = scm80211_hwq(hw, ftm->ac);
	// TODO
	hwq->vif = vif;

	tasklet_init(&hwq->tasklet, scm80211_ftm_hwq_tasklet, (unsigned long)hwq);
	FTM_DBG(hw, "[- %s] get_txop done\n", __func__);
}

struct sk_buff *scm80211_ftm_skb_alloc(struct scm80211_ftm *ftm, struct ieee80211_vif *vif,
								int nsize)
{
	struct sk_buff *skb;
	struct ieee80211_hdr_3addr *nullfunc;
	struct ieee80211_qos_hdr *qoshdr;

	if (!nsize)
		return NULL;

	skb = dev_alloc_skb(nsize + sizeof(*qoshdr) + 2 + 64);
	if (!skb)
		return NULL;

	skb_reserve(skb, 64);

	if (ftm->qos) {
		qoshdr = skb_put_zero(skb, sizeof(*qoshdr) + nsize);
		qoshdr->frame_control = cpu_to_le16(IEEE80211_FTYPE_DATA |
						    IEEE80211_STYPE_QOS_DATA | IEEE80211_FCTL_TODS);
		if (!ftm->ack_policy)
			qoshdr->qos_ctrl |= cpu_to_le16(CTRL_QOS_NO_ACK);
		else
			qoshdr->qos_ctrl |= cpu_to_le16(CTRL_QOS_NORMAL_ACK);

		qoshdr->qos_ctrl |= cpu_to_le16(ftm->tid);
		memcpy(qoshdr->addr1, ftm->macaddr, ETH_ALEN);
		memcpy(qoshdr->addr2, vif->addr, ETH_ALEN);
		memcpy(qoshdr->addr3, ftm->macaddr, ETH_ALEN);
	} else {
		nullfunc = skb_put_zero(skb, sizeof(*nullfunc) + nsize);
		nullfunc->frame_control = cpu_to_le16(IEEE80211_FTYPE_DATA |
						      IEEE80211_STYPE_NULLFUNC |
						      IEEE80211_FCTL_TODS);
		memcpy(nullfunc->addr1, ftm->macaddr, ETH_ALEN);
		memcpy(nullfunc->addr2, vif->addr, ETH_ALEN);
		memcpy(nullfunc->addr3, ftm->macaddr, ETH_ALEN);
	}

	return skb;
}

static void scm80211_ftm_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				struct ieee80211_tx_control *control, struct sk_buff *skb)
{
	struct scm80211_txq *txq;
	struct scm80211_hwq *hwq;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	__le16 fc = hdr->frame_control;
	char buf[128];
	bool skip = false;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	skip |= ieee80211_is_beacon(fc);
	skip |= ieee80211_is_probe_req(fc);

	if (!skip) {
		scm80211_snprint_frame(buf, sizeof(buf), hdr);
		FTM_DBG(hw, "%s(): %s (vdev#%d)\n", __func__, buf, scm80211_vdev_id(vif));
	}

	txq = &vifp->mq;
	hwq = scm80211_hwq(hw, txq->hw_queue);

	skb_queue_tail(&txq->queue, skb);
}

static void
scm80211_ftm_tx_channel(struct ieee80211_hw *hw, struct sk_buff *skb,
			struct ieee80211_channel *chan)
{
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct ieee80211_vif *vif = info->control.vif;
	struct ieee80211_tx_control control = {
		.sta = NULL,
	};

	info->band = chan->band;

	if (ieee80211_hw_check(hw, SUPPORTS_RC_TABLE))
		ieee80211_get_tx_rates(info->control.vif, NULL, skb,
				       info->control.rates, ARRAY_SIZE(info->control.rates));

	scm80211_ftm_tx(hw, vif, &control, skb);
}

static int _scm80211_ftm_tx_frame(struct ieee80211_hw *hw,
				struct ieee80211_vif *vif, struct scm80211_ftm *ftm)
{
	struct sk_buff *skb;
	struct ieee80211_hdr *nullfunc;
	struct ieee80211_tx_info *info;
	struct scm80211_ftm_txskb *txs = &ftm->ftm_tx_skb;

	skb = txs->tx_skb[txs->skb_id][txs->txd_num];
	if (!skb) {
		FTM_DBG(hw, "[%s] get skb(%d-%d) is NULL!\n", __func__, txs->skb_id, txs->txd_num);
		return -ENOMEM;
	}

	ftm->txd_pktcnt++;
	nullfunc = (struct ieee80211_hdr *)skb->data;
	nullfunc->seq_ctrl = cpu_to_le16(ftm->seqnum++ << 4) & IEEE80211_SCTL_SEQ;
	nullfunc->frame_control &= ~cpu_to_le16(IEEE80211_FCTL_PM);

	info = IEEE80211_SKB_CB(skb);
	info->flags |= (IEEE80211_TX_CTL_USE_MINRATE | ftm->fixrate.flags);
	info->band = ftm->chan.band;
	info->control.vif = vif;
	info->control.rates[0].idx = ftm->fixrate.rate.idx;
	info->control.rates[0].count = ftm->fixrate.rate.count;
	info->control.rates[0].flags = ftm->fixrate.rate.flags;
	info->control.skip_table = 1;

	scm80211_ftm_tx_channel(hw, skb, &ftm->chan);

	return 0;
}

static void scm80211_ftm_set_he_hwq(struct scm80211_hwq *hwq, struct scm80211_ftm *ftm)
{
	if (ftm->ht_mod != HE_SU)
		return;
	hwq->he_gi_type = ftm->fixrate.he_gi_type;
	hwq->he_ltf_type = ftm->fixrate.he_ltf_type;
}

#ifdef FTM_DBG_PARMS
#define CHAN2G(_freq)  {	   \
	.band = NL80211_BAND_2GHZ, \
	.center_freq = (_freq),	   \
	.hw_value = (_freq),	   \
	.max_power = 20,	   \
}

#define SCM80211_FTM_DFT_CHAN_2GHZ 2437
static struct ieee80211_channel scm80211_ftm_default_chan_2ghz =
				CHAN2G(SCM80211_FTM_DFT_CHAN_2GHZ);	/* Channel 6 */
#endif

void scm80211_ftm_rx_filter(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	int vdev = scm80211_vdev_id(vif);
	u32 v = 0x0;		//mask all management frames

	scm80211_mac_writel(hw, v, MAC_REG_VIF_RX_FILTER1(vdev));
	FTM_DBG(hw, "%s: filter=%x dev=%d\n", __func__, v, vdev);
}

void scm80211_ftm_wake_ftm_queues(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_hwq *hwq;
	u32 queues;

	FTM_DBG(hw, "%s: ac=%x\n", __func__, ftm->ac);

	hwq = scm80211_hwq(hw, ftm->ac);
	hwq->enabled = true;

	/* start mgmt hwq */
	queues = scm80211_vif_queues(vif);
	queues |= BIT(vifp->mgmt_hw_queue);
	scm80211_wake_queues(hw, queues);
	FTM_DBG(hw,
		"%s: [FTM_DEBUG] vifp->mgmt_hw_queue=%d vif queues=0x%lx, wakeup queues=0x%x\n",
		__func__, vifp->mgmt_hw_queue, scm80211_vif_queues(vif), queues);
}

void scm80211_ftm_stop_ftm_queues(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_txq *txq = &vifp->mq;
	struct scm80211_hwq *hwq;
	int pending = 0;
	u32 queues;

	FTM_DBG(hw, "%s: ac=%x\n", __func__, ftm->ac);

	hwq = scm80211_hwq(hw, ftm->ac);
	do {
		pending = skb_queue_len(&txq->queue);
		FTM_DBG(hw, "wait for pending=%d\n\n", pending);
		if (pending)
			scm80211_cleanup_txq(txq);
	} while (pending);
	hwq->enabled = false;

	/* stop all hwq except mgmt */
	queues = scm80211_vif_queues(vif);
	queues &= ~BIT(vifp->mgmt_hw_queue);
	scm80211_stop_queues(hw, queues);

	FTM_DBG(hw,
		"%s: [FTM_DEBUG] vifp->mgmt_hw_queue=%d vif queues=0x%lx, stop queues=0x%x\n",
		__func__, vifp->mgmt_hw_queue, scm80211_vif_queues(vif), queues);
}

void scm80211_ftm_tx_frame(struct ieee80211_vif *vif, struct scm80211_ftm *ftm)
{
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(vif);
	struct scm80211_ftm_txskb *txs = &ftm->ftm_tx_skb;

	if (ftm->tx_start == FTM_TX_STOP)
		return;

	txs->txd_num = 0;
	while (txs->txd_num < txs->total_num) {
		_scm80211_ftm_tx_frame(hw, vif, ftm);
		if (ftm->seqnum > 4095)
			ftm->seqnum = 0;
		txs->tx_skb[txs->skb_id][txs->txd_num] = NULL;
		txs->txd_num++;
	}

	if (txs->txd_num) {
		txs->txd_cnt++;
		FTM_DBG(hw, "[%s] skb_id=%d, txd=%d\n", __func__, txs->skb_id, txs->txd_num);

		schedule_work(scm80211_ftm_tx_ac_work_get(hw, vif));
		flush_work(scm80211_ftm_tx_ac_work_get(hw, vif));
	}
}

static void
scm80211_sta_supp_rates_init(struct ieee80211_hw *hw,
			     struct ieee80211_sta *sta, enum nl80211_band band)
{
	int r;
	u32 mandatory = 0;
	struct ieee80211_rate *rate;

	if (band == NL80211_BAND_2GHZ)
		/*
		 * We use both here, even if we cannot really know for
		 * sure the station will support both, but the only use
		 * for this is when we don't know anything yet and send
		 * management frames, and then we'll pick the lowest
		 * possible rate anyway.
		 * If we don't include _G here, we cannot find a rate
		 * in P2P, and thus trigger the WARN_ONCE() in rate.c
		 */
		mandatory = IEEE80211_RATE_MANDATORY_B | IEEE80211_RATE_MANDATORY_G;
	else if (band == NL80211_BAND_5GHZ)
		mandatory = IEEE80211_RATE_MANDATORY_A;

	for (r = 0; r < hw->wiphy->bands[band]->n_bitrates; r++) {
		rate = &hw->wiphy->bands[band]->bitrates[r];

		if (!(rate->flags & mandatory))
			continue;
#if (KERNEL_VERSION(5, 19, 0) > MAC80211_BASE)
		sta->supp_rates[band] |= BIT(r);
#else
		sta->deflink.supp_rates[band] |= BIT(r);
#endif
	}
}

void
scm80211_sta_init(struct ieee80211_vif *vif, struct ieee80211_sta *sta, struct scm80211_ftm *ftm)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = vifp->hw;
	struct ieee80211_sta_priv *sta_priv;
	struct ieee80211_dbg_rate_fix *sta_rate;
	enum nl80211_band band = ftm->chan.band;
	struct ieee80211_sta_ht_cap *ht_cap;
	struct ieee80211_sta_vht_cap *vht_cap;
	struct ieee80211_sta_he_cap *he_cap;

	vifp->mq.txq.tid = ftm->tid;
	if (!ftm->ampdu) {
		vifp->mq.txq.sta = NULL;
		return;
	}

	vifp->mq.txq.sta = sta;
	sta_priv = (void *)sta->drv_priv;
	sta_rate = &sta_priv->rate_fix;

#if (KERNEL_VERSION(5, 19, 0) > MAC80211_BASE)
	sta->bandwidth = ftm->cbw;
	ht_cap = &sta->ht_cap;
	vht_cap = &sta->vht_cap;
	he_cap = &sta->he_cap;
#else
	sta->deflink.bandwidth = ftm->cbw;
	ht_cap = &sta->deflink.ht_cap;
	vht_cap = &sta->deflink.vht_cap;
	he_cap = &sta->deflink.he_cap;
#endif

	scm80211_sta_supp_rates_init(hw, sta, band);
	memcpy(ht_cap, &hw->wiphy->bands[band]->ht_cap, sizeof(struct ieee80211_sta_ht_cap));
	memcpy(vht_cap, &hw->wiphy->bands[band]->vht_cap, sizeof(struct ieee80211_sta_vht_cap));
	memcpy(he_cap, &hw->wiphy->bands[band]->iftype_data[band].he_cap,
	       sizeof(struct ieee80211_sta_he_cap));
	sta_priv->txop_rts_th = 0xFFFF;
	sta_priv->n_ss = ftm->rate_nss;
	memcpy(sta->addr, ftm->macaddr, ETH_ALEN);
	ht_cap->ampdu_density = 5;

	memcpy(sta_rate, &ftm->fixrate, sizeof(struct scm80211_ftm_rate));
	sta_priv->rate_fix_on = false;
}

int scm80211_ftm_rx_ampdu_init(struct ieee80211_hw *hw, struct ieee80211_vif_priv *vifp)
{
	u32 v;

	/* No matter enable or not, reset the scoreboard */
	v = scm80211_mac_readl(hw, MAC_REG_RX_SCB_CTRL);
	v &= ~RX_SCB_CTRL_SCB_RESET;
	v |= FIELD_PREP(RX_SCB_CTRL_SCB_RESET, BIT(vifp->ftm.tid));
	scm80211_mac_writel(hw, v, MAC_REG_RX_SCB_CTRL);

	return 0;
}

void scm80211_ftm_init_ampdu(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	struct ieee80211_tx_rate *rate = txop->rate;	/*&txop->ti.control.rates[0]; */
	int tid = txop->owner->txq.tid;

	ampdu->len = 0;
	ampdu->maxlen = ieee80211_max_ampdu_len(sta, rate);
	ampdu->bitrate = txop->bitrate;	/* To be removed */
	ampdu->mmss = scm80211_ampdu_mmss_bytes(hwq->hw, sta, txop->bitrate);
	ampdu->tids[tid].bufsize = scm80211_ba_buf_size(txop->owner);

	bitmap_zero(ampdu->tids[tid].bitmap, ampdu->tids[tid].bufsize);
	bitmap_zero(ampdu->tids[tid].ba, ampdu->tids[tid].bufsize);
	ampdu->tids[tid].ssn = 0;

	ampdu->has_bar = false;
	ampdu->has_eof = false;
	ampdu->has_qos_data = false;
	ampdu->has_immediate_response = false;
}

void scm80211_ftm_tx_ampdu_init(struct ieee80211_hw *hw, struct ieee80211_vif_priv *vifp)
{
	struct ieee80211_ampdu_params *ba = &vifp->mq.ba;
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_txq *txq = (struct scm80211_txq *)&vifp->mq.txq;
	struct scm80211_hwq *hwq = NULL;
	struct txop *txop = NULL;

	if (!ftm->ampdu) {
		memset(ba, 0, sizeof(struct ieee80211_ampdu_params));
		return;
	}

	hwq = scm80211_hwq(hw, ftm->ac);
	txop = scm80211_get_txop(hwq, TXOP_STATE_SCHEDULED);
	if (txop)
		scm80211_ftm_init_ampdu(txop);

	ba->action = IEEE80211_AMPDU_TX_OPERATIONAL;
	ba->sta = vifp->mq.txq.sta;
	ba->tid = ftm->tid;
	ba->ssn = 0;
	ba->buf_size = 128;
	ba->amsdu = false;
	ba->timeout = 0;
	memset(&txq->sb, 0, sizeof(txq->sb));

	FTM_DBG(hw, "AMPDU blockack session parameters:\n");
	FTM_DBG(hw, "\t txq=%p\n", &vifp->mq.txq);
	FTM_DBG(hw, "\t action=%d\n", ba->action);
	FTM_DBG(hw, "\t sta=%p\n", ba->sta);
	FTM_DBG(hw, "\t tid=%d\n", ba->tid);
	FTM_DBG(hw, "\t ssn=%d\n", ba->ssn);
	FTM_DBG(hw, "\t buf_size=%d\n", ba->buf_size);
	FTM_DBG(hw, "\t amsdu=%s\n", ba->amsdu ? "enable" : "disable");
	FTM_DBG(hw, "\t timeout=%d\n", ba->timeout);
}

static void scm80211_ftm_tx_skb_deinit(struct ieee80211_hw *hw,
				struct scm80211_ftm *ftm)
{
	struct scm80211_ftm_txskb *txs = &ftm->ftm_tx_skb;
	u8 i;

	FTM_DBG(hw, "[%s] skb_id=%d, total_num=%d\n", __func__, txs->skb_id, txs->total_num);
	for (i = 0; i < txs->total_num; i++) {
		if (txs->tx_skb[txs->skb_id][i]) {
			dev_kfree_skb_any(txs->tx_skb[txs->skb_id][i]);
			txs->tx_skb[txs->skb_id][i] = NULL;
		}
	}
}

int scm80211_ftm_tx_start(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct ieee80211_sta *sta = &ftm_sta;

	FTM_DBG(hw, "%s: [FTM_DEBUG] schedule_delayed_work\n", __func__);

	ftm->seqnum = 0;
	ftm->tx_start = FTM_TX_START;
	memset(sta, 0, sizeof(struct ieee80211_sta));
	INIT_WORK((scm80211_ftm_work_get(hw, vif)), scm80211_ftm_work);

	INIT_WORK((scm80211_ftm_tx_ac_work_get(hw, vif)), scm80211_ftm_tx_alloc_work);
	schedule_work(scm80211_ftm_tx_ac_work_get(hw, vif));
	flush_work(scm80211_ftm_tx_ac_work_get(hw, vif));

	scm80211_ftm_stop_ftm_queues(hw, vif);
	scm80211_ftm_wake_ftm_queues(hw, vif);
	scm80211_ftm_rx_filter(hw, vif);

	scm80211_tx_power_set(hw, ftm->channel, ftm->txpower);
	if (scm80211_ftm_ctx_init(hw, ftm) < 0) {
		FTM_DBG(hw, "[%s] ctx_init fail!\n", __func__);
		return -1;
	}

	scm80211_sta_init(vif, sta, ftm);
	scm80211_ftm_tx_ampdu_init(hw, vifp);

	//start txop schedule
	schedule_work(scm80211_ftm_work_get(hw, vif));

	return 0;
}

int scm80211_ftm_tx_stop(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	FTM_DBG(hw, "%s: [FTM_DEBUG]\n", __func__);

	vifp->ftm.tx_start = FTM_TX_STOP;

	cancel_work_sync(scm80211_ftm_tx_ac_work_get(hw, vif));
	cancel_work_sync(scm80211_ftm_work_get(hw, vif));
	scm80211_ftm_tx_skb_deinit(hw, ftm);
	scm80211_ftm_stop_ftm_queues(hw, vif);

	return 0;
}

static void scm80211_ftm_tx_alloc_work(struct work_struct *wk)
{
	struct scm80211_hwq *hwq = container_of(wk, struct scm80211_hwq, tx_ac_wk);
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_vif *vif = hwq->vif;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_ftm_txskb *txs = &ftm->ftm_tx_skb;
	int lst_size = FTM_MAX_PKT_SIZE, pkt_size = FTM_MAX_PKT_SIZE;
	u8 n, i;

	if (ftm->tx_start == FTM_TX_STOP)
		return;

	if (ftm->ampdu) {
		pkt_size = FTM_PER_AMPDU_SIZE;
		lst_size = FTM_PER_AMPDU_SIZE;
	} else {
		if (ftm->txpktsize > FTM_MAX_PKT_SIZE)
			ftm->txpktsize = FTM_MAX_PKT_SIZE;
		lst_size = ftm->txpktsize % pkt_size;
		txs->total_num = ftm->txpktsize / pkt_size;
	}

	if (ftm->txpktcnt == FTM_MAX_TX_CNT) {
		txs->total_cnt = FTM_MAX_TX_CNT;
		if (ftm->ampdu) {
			if (is_usb_device(hw))
				txs->total_num = FTM_MAX_PKT_NUM;
			else if (ftm->ht_mod < HE_SU)
				txs->total_num = 27;
			else
				txs->total_num = FTM_MAX_PKT_NUM - 1;
		}
	} else {
		txs->total_cnt = ftm->txpktcnt;
		if (ftm->txpktsize > FTM_MAX_PKT_SIZE)
			ftm->txpktsize = FTM_MAX_PKT_SIZE;
		lst_size = ftm->txpktsize % pkt_size;
		txs->total_num = ftm->txpktsize / pkt_size;
	}

	txs->skb_id = ((txs->skb_id & 0x1) ^ 0x1);
	n = txs->skb_id;
	for (i = 0; i < txs->total_num; i++)
		txs->tx_skb[n][i] = scm80211_ftm_skb_alloc(ftm, vif, pkt_size);
	if (lst_size) {
		txs->tx_skb[n][i] = scm80211_ftm_skb_alloc(ftm, vif, lst_size);
		txs->total_num += 1;
	}
}

static void scm80211_ftm_work(struct work_struct *wk)
{
	struct scm80211_hwq *hwq = container_of(wk, struct scm80211_hwq, work);
	struct ieee80211_vif *vif = hwq->vif;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_ftm_txskb *txs = &ftm->ftm_tx_skb;
	struct ieee80211_hw *hw = hwq->hw;
	struct txop *txop;
	struct scm80211_txq *txq;
	/* no_need: no need reschedule this hwq */
	bool no_need = true;
	int i, ret;

	// TODO MGMT
	WARN_ON_ONCE(hwq != scm80211_hwq(hw, 4));

	FTM_DBG(hw, "[+ %s] tx_start = %d\n", __func__, ftm->tx_start);
	if (ftm->tx_start != FTM_TX_START)
		goto wk_stop;

	if ((ftm->txpktcnt != FTM_MAX_TX_CNT) && (txs->txd_cnt >= txs->total_cnt))
		goto wk_stop;

	for (i = 0; i < ARRAY_SIZE(hwq->ring.pool); i++) {
		txop = scm80211_get_txop(hwq, TXOP_STATE_NEEDS_RESCHED);
		if (txop) {
			if (!scm80211_schedule_txop(txop))
				no_need = scm80211_start_backoff(txop);
			else
				scm80211_txop_free(txop);
		} else {
			break;
		}
	}

	txop = scm80211_get_txop(hwq, TXOP_STATE_SCHEDULED);
	if (txop)
		no_need = scm80211_start_backoff(txop);

	while ((txop = scm80211_txop_alloc(hwq)) != NULL) {
		int pending = 0;

		scm80211_ftm_tx_frame(vif, ftm);
		scm80211_ftm_set_he_hwq(hwq, ftm);
		txq = scm80211_get_next_txq(hwq);
		if (!txq) {
			scm80211_txop_free(txop);
			break;
		}

		scm80211_txop_attach(txop, txq);
		ret = scm80211_schedule_txop(txop);
		if (ret < 0) {
			FTM_DBG(hw, "- txq %s: schedule fails\n", txq->name);
			scm80211_txop_free(txop);
			/* internal error occur, need resched current hwq later */
			if (ret == -EAGAIN)
				no_need = false;
			break;
		}
		no_need = scm80211_start_backoff(txop);

		pending = scm80211_num_pending_frames(txq, true);
		if (hwq->mac80211)
			ieee80211_return_txq(hw, &txq->txq, pending > 0);
		FTM_DBG(hw, "ftm_work pending %d\n", pending);
	}

	if (scm80211_txop_ring_empty(hwq)) {
		/*
		 * No pending txop and no txq to schedule from. The hardware queue
		 * is parking. Next txq should schedule the work to get things
		 * started.
		 */
		hwq->idle = true;
		complete(&hwq->completion);
	} else {
		hwq->idle = false;
		reinit_completion(&hwq->completion);
	}

	hwq->need_resched = !(no_need || scm80211_txop_is_running(hwq));
	FTM_DBG(hw, "-%s(hwq#%d): idle=%d, %sneed reschedule, seqnum=%d\n",
		__func__, hwq->id, hwq->idle, hwq->need_resched ? "" : "no ", ftm->seqnum);
	return;
wk_stop:
	if ((ftm->txpktcnt != FTM_MAX_TX_CNT) && (txs->txd_cnt >= txs->total_cnt))
		ftm->tx_start = FTM_TX_COMPLETE;
}

bool scm80211_ftm_rx_enable_filter(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int vdev = scm80211_vdev_id(vif);
	u32 v = 0x0;		//mask all management frames

	if (ftm->enabled) {
		scm80211_mac_writel(hw, v, MAC_REG_VIF_RX_FILTER1(vdev));
		FTM_DBG(hw, "%s: filter1=%x\n", __func__, v);
	}
	return ftm->enabled;
}

void scm80211_ftm_tx_tssi_show(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, int ch)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy = priv->phy;

	memset(ftm->ftm_tx_rpt, 0, sizeof(ftm->ftm_tx_rpt));
	if (is_usb_device(hw)) {
		ftm->ftm_tssi = scm80211_get_tssi(hw);
	} else {
		if (!phy || !phy->radio) {
			FTM_DBG(hw, "[%s] get phy radio error!\n", __func__);
			return;
		}
		ftm->ftm_tssi = phy->radio->ops->get_cur_pwr(phy->radio);
	}
	sprintf(ftm->ftm_tx_rpt, "\n\tTSSI: %d\n", ftm->ftm_tssi);
}

void scm80211_ftm_rx_rssi_show(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	struct scm80211_ftm_rx_statistics *rxs = &ftm->rx_statistics;
	int i = 0, j = 0, len = 1;

	memset(ftm->ftm_rssi_rpt, 0, sizeof(ftm->ftm_rssi_rpt));
	ftm->ftm_rssi_rpt[0] = '\n';
	j = rxs->cur + SCM80211_FTM_RX_RSSI_MAX;
	for (i = 0; i < SCM80211_FTM_RX_RSSI_MAX; i++) {
		--j;
		len += sprintf(ftm->ftm_rssi_rpt + len, "index_%02d: %d\n",
			       i, rxs->rssi[j % SCM80211_FTM_RX_RSSI_MAX]);
	}
//      FTM_DBG(hw, "%s", ftm->ftm_rssi_rpt);
}

void scm80211_ftm_rx_statistics_show(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	u32 v, ppdu_cnt = 0, pmpdu_cnt = 0;
	struct scm80211_ftm_rx_statistics *rxs = &ftm->rx_statistics;

	FTM_DBG(hw, "%s: rx expectation\n", __func__);

	memset(ftm->ftm_rx_rpt, 0, sizeof(ftm->ftm_rx_rpt));
	if (ftm->rx_start == FTM_RX_START) {
		v = scm80211_mac_readl(hw, MAC_REG_DBG_RX_PPDU_CNT0);
		ppdu_cnt = (v & 0xFFFF);
		pmpdu_cnt = v >> 16;

		if (ppdu_cnt > rxs->rx_ppkt)
			rxs->rx_pkt = ppdu_cnt - rxs->rx_ppkt;
		else if (ppdu_cnt < rxs->rx_ppkt)
			rxs->rx_pkt = 0xFFFF - (rxs->rx_ppkt - ppdu_cnt);

		if (pmpdu_cnt > rxs->rx_pmpdu)
			rxs->rx_mpdu = pmpdu_cnt - rxs->rx_pmpdu;
		else if (pmpdu_cnt < rxs->rx_pmpdu)
			rxs->rx_mpdu = 0xFFFF - (rxs->rx_pmpdu - pmpdu_cnt);

		rxs->rx_err = rxs->rx_pkt - rxs->rx_mpdu;
	}
	sprintf(ftm->ftm_rx_rpt,
		"\nExpected rx set frame number = %d\n"
		"Received rx frame ppdu = %d\n"
		"Received rx frame mpdu = %d\n"
		"Received rx err frame number = %d\n",
		ftm->rxpktcnt, rxs->rx_pkt, rxs->rx_mpdu, rxs->rx_err);

	FTM_DBG(hw, "%s", ftm->ftm_rx_rpt);
	ftm->rx_start = FTM_RX_COMPLETE;
}

void scm80211_ftm_rx_statis_update(struct scm80211_rx_data *rx)
{
	static bool ftm_rx_logs = true;
	struct ieee80211_hw *hw = rx->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif *vif = priv->vifs[0];
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_ftm_rx_statistics *rxs = &ftm->rx_statistics;
	struct ieee80211_hdr *hdr = rx->hdr;

	if (ftm->rx_start != FTM_RX_START) {
		if (ftm_rx_logs) {
			ftm_rx_logs = false;
			FTM_DBG(hw, "%s: [FTM_DEBUG] rx_start=%d, return!\n", __func__,
				ftm->rx_start);
		}
		return;
	}

	if (hdr) {
		FTM_DBG(hw,
			"%s: [FTM_DEBUG] frame_control=0x%x duration_id=%x seq_ctrl=0x%x\n",
			__func__, hdr->frame_control, hdr->duration_id, hdr->seq_ctrl);
		FTM_DBG(hw, "%s: [FTM_DEBUG] addr1=%02x:%02x:%02x:%02x:%02x:%02x\n",
			__func__, hdr->addr1[0], hdr->addr1[1], hdr->addr1[2],
			hdr->addr1[3], hdr->addr1[4], hdr->addr1[5]);
		FTM_DBG(hw, "%s: [FTM_DEBUG] addr1=%02x:%02x:%02x:%02x:%02x:%02x\n",
			__func__, hdr->addr2[0], hdr->addr2[1], hdr->addr2[2],
			hdr->addr2[3], hdr->addr2[4], hdr->addr2[5]);
		FTM_DBG(hw, "%s: [FTM_DEBUG] addr1=%02x:%02x:%02x:%02x:%02x:%02x\n",
			__func__, hdr->addr3[0], hdr->addr3[1], hdr->addr3[2],
			hdr->addr3[3], hdr->addr3[4], hdr->addr3[5]);
	}
	//record ieee80211_rx_status on the first frames
	if (!rxs->rx_pkt)
		memcpy(&rxs->rxs, rx->rxs, sizeof(struct ieee80211_rx_status));

	ftm_rx_logs = true;

	//update rssi
	rxs->rssi[ftm->rx_statistics.cur] = rx->rv->RSSI;
	rxs->cur = (rxs->cur + 1) % SCM80211_FTM_RX_RSSI_MAX;

	FTM_DBG(hw, "%s: [FTM_DEBUG] rxpktcnt=%d\n", __func__, ftm->rxpktcnt);
}

int scm80211_ftm_rx_start(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	struct scm80211_ftm_rx_statistics *rxs = &ftm->rx_statistics;

	FTM_DBG(hw, "%s: [FTM_DEBUG] duration=%dms\n", __func__, ftm->rx_duration);

	rxs->rx_pkt = 0;
	rxs->rx_err = 0;
	rxs->cur = 0;
	memset(&rxs->rxs, 0, sizeof(struct ieee80211_rx_status));
	memset(rxs->rssi, 0, sizeof(rxs->rssi));

	ftm->rx_start = FTM_RX_START;
	//set channel
	scm80211_ftm_ctx_init(hw, &vifp->ftm);

	if (is_zero_ether_addr(vif->addr))
		memcpy(vif->addr, ftm->macaddr, ETH_ALEN);

	FTM_DBG(hw,
		"%s: [FTM_DEBUG] vif->type=%d vif=%02x:%02x:%02x:%02x:%02x:%02x\n",
		__func__, vif->type, vif->addr[0], vif->addr[1], vif->addr[2],
		vif->addr[3], vif->addr[4], vif->addr[5]);

	scm80211_set_addr(hw, vif, vif->addr);
	scm80211_ftm_rx_filter(hw, vif);

	scm80211_ftm_rx_ampdu_init(hw, vifp);

	return 0;
}

int scm80211_ftm_rx_stop(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	if (ftm->rx_start == FTM_RX_STOP)
		return 0;

	FTM_DBG(hw, "%s: [FTM_DEBUG] %d\n", __func__, __LINE__);
	ftm->rx_start = FTM_RX_STOP;

	return 0;
}

int ftm_clear_parms(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	if (ftm->tx_start == FTM_TX_START) {
		FTM_DBG(hw, "[%s] Error! FTM not stopped! tx=%d rx=%d\n", __func__,
			ftm->tx_start, ftm->rx_start);
		return -1;
	}

	ftm->txpktcnt = 0;
	ftm->rxpktcnt = 0;
	ftm->txpktsize = 0;
	ftm->channel = 0;
	ftm->center_freq1 = 0;
	ftm->center_freq2 = 0;
	ftm->width = 0;
	ftm->chain = 0;
	ftm->ack_policy = 0;
	ftm->ht_mod = 0;
	memset(&ftm->fixrate, 0, sizeof(struct scm80211_ftm_rate));
	memset(ftm->macaddr, 0xFF, ETH_ALEN);
	memset(&ftm->ftm_tx_skb, 0, sizeof(struct scm80211_ftm_txskb));
	ftm->ifs = 0;
	ftm->ampdu = 0;
	ftm->amsdu = 0;
	ftm->ftm_cfo = 0;
	ftm->freq_cal = 0;

	ftm->tx_start = 0;
	ftm->rx_start = 0;
	ftm->txpower = 0;
	ftm->gain_index[0] = FTM_GAINIDX_NOUSE;
	ftm->gain_index[1] = FTM_GAINIDX_NOUSE;
	memset(&ftm->chan, 0, sizeof(struct ieee80211_channel));
	memset(&ftm->rx_statistics, 0, sizeof(struct scm80211_ftm_rx_statistics));
	ftm->tid = 5;
	ftm->qos = 1;
	ftm->txd_pktcnt = 0;
	ftm->ac = 4;		//mgmt
	ftm->rate_nss = 0;
	ftm->rx_duration = SCM80211_FTM_RX_DURATION_MS;
	ftm->err_flag = false;
	memset(&ftm->ftm_tx_skb, 0, sizeof(ftm->ftm_tx_skb));
	ftm->tone_ofs = 0;
	ftm->pri_ch = 0;

	return 0;
}

int scm80211_ftm_rx_clear(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp;
	struct scm80211_ftm *ftm;
	struct scm80211_ftm_rx_statistics *rxs;
	u32 v;

	vifp = ieee80211_to_priv(vif);
	ftm = &vifp->ftm;
	rxs = &ftm->rx_statistics;

	v = scm80211_mac_readl(hw, MAC_REG_DBG_RX_PPDU_CNT0);
	rxs->rx_ppkt = (v & 0xFFFF);
	rxs->rx_pmpdu = v >> 16;
	rxs->rx_mpdu = 0;
	rxs->rx_pkt = 0;
	rxs->rx_err = 0;
	ftm->rx_start = FTM_RX_START;

	return 0;
}

static void scm80211_chanctx_ftm_begin(struct ieee80211_hw *hw, struct scm80211_chanctx *ftm)
{
	FTM_DBG(hw, "%s: [FTM_DEBUG] TODO\n", __func__);
}

static void scm80211_chanctx_ftm_end(struct ieee80211_hw *hw, struct scm80211_chanctx *ftm)
{
	FTM_DBG(hw, "%s: [FTM_DEBUG] TODO\n", __func__);
}

void scm80211_chandef_create(struct cfg80211_chan_def *chandef,
			struct ieee80211_channel *chan,
			enum nl80211_channel_type chan_type, struct scm80211_ftm *ftm)
{
	if (WARN_ON(!chan))
		return;

	chandef->chan = chan;
#if (KERNEL_VERSION(5, 8, 0) <= MAC80211_BASE)
	chandef->freq1_offset = chan->freq_offset;
#endif
	chandef->center_freq2 = 0;
	chandef->edmg.bw_config = 0;
	chandef->edmg.channels = 0;

	if (chan_type == SCM80211_CHAN_HT80MINUS) {
		chandef->width = NL80211_CHAN_WIDTH_80;
		chandef->center_freq1 = ftm->center_freq2;
	} else if (chan_type == SCM80211_CHAN_HT80PLUS) {
		chandef->width = NL80211_CHAN_WIDTH_80;
		chandef->center_freq1 = ftm->center_freq2;
	} else {
		switch (chan_type) {
		case NL80211_CHAN_NO_HT:
			chandef->width = NL80211_CHAN_WIDTH_20_NOHT;
			chandef->center_freq1 = chan->center_freq;
			break;
		case NL80211_CHAN_HT20:
			chandef->width = NL80211_CHAN_WIDTH_20;
			chandef->center_freq1 = chan->center_freq;
			break;
		case NL80211_CHAN_HT40PLUS:
			chandef->width = NL80211_CHAN_WIDTH_40;
			chandef->center_freq1 = chan->center_freq + 10;
			break;
		case NL80211_CHAN_HT40MINUS:
			chandef->width = NL80211_CHAN_WIDTH_40;
			chandef->center_freq1 = chan->center_freq - 10;
			break;
		default:
			WARN_ON(1);
		}
	}
}

int scm80211_get_gainidx(struct ieee80211_hw *hw, char *rsp)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy = priv->phy;
	u32 val = 0;

	if (!phy || !rsp)
		return -1;

	val = phy_readl(phy, PHY_RFIF(FW_TXLUT_CFG0));
	val = (val >> 8) & 0xFF;
	sprintf(rsp, "\ngain_idx = %d\n", val);
	return 0;
}

static int scm80211_set_gainidx(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy = priv->phy;
	int gain, flag, i, chain;
	u32 val = 0;

	chain = (ftm->chain < ftm_chain_mimo) ? ftm->chain : 0;
	gain = ftm->gain_index[chain];
	for (i = 0; i < 2; i++) {
		if (gain == FTM_GAINIDX_NOUSE) {
			sprintf(ftm->ftm_cmd_rsp, "gainidx is no used!\n");
			return 0;
		}
		if (ftm->tone_ofs)
			gain = ((ftm->gain_index[chain] + 1)/2);

		flag = i + 1;
		val = (gain << 8) | 0x10;
		if (flag & 0x2)
			phy = phy_next(hw, phy, false);
		if (flag & 0x4)
			phy = phy_next(hw, NULL, false);

		if (!phy) {
			sprintf(ftm->ftm_cmd_rsp, "set gainidx_%d fail, phy is NULL!\n", chain);
			return -1;
		}
		phy_writel(phy, val, PHY_RFIF(FW_TXLUT_CFG0));

		if (ftm->chain < ftm_chain_mimo)
			break;
		chain++;
	}

	return 0;
}


void scm80211_set_chain(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	FTM_DBG(hw, "[+ %s]chain=%d\n", __func__, ftm->chain);

	if (ftm->chain < ftm_chain_mimo)
		scm80211_set_ant_sel(hw, ftm->chain & 0x1);
	else
		scm80211_set_ant_sel(hw, ftm_chain_mimo);
}

void scm80211_set_hw_band(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->band &= (~BIT(NL80211_BAND_2GHZ) & ~BIT(NL80211_BAND_5GHZ));
	if (ftm->chan.band == NL80211_BAND_2GHZ)
		priv->band |= BIT(NL80211_BAND_2GHZ);
	else
		priv->band |= BIT(NL80211_BAND_5GHZ);
}

void scm80211_ftm_set_param_b(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	int buff[5] = { 0 };
	int i, chain = (ftm->chain < ftm_chain_mimo) ? ftm->chain : 0;

	for (i = 0; i < 2; i++) {
#ifdef EFUSE_ENABLE
		int ret, t = 0, *pdat = (int *)&ftm->cal_dat[chain];

		for (i = 0; i < 4; i++) {
			if (*(pdat + i))
				t++;
		}
		if (!t)
			ret = scm80211_efuse_cal_update(chain, (int *)&ftm->cal_dat[chain]);
#endif

		buff[0] = chain;
		memcpy(&buff[1], &ftm->cal_dat[chain], sizeof(struct scm80211_scm_cal));
		if (is_usb_device(hw))
			scm80211_set_param_b(hw, buff, sizeof(buff));
		else
			scm2020_pdet_set_para_b(chain, buff[1], buff[2], buff[3], buff[4]);

		FTM_DBG(hw, "[%s] chain%d path=%d {%d,%d,%d,%d}\n", __func__,
			chain, buff[0], buff[1], buff[2], buff[3], buff[4]);
		if (ftm->chain < ftm_chain_mimo)
			break;
		chain++;
	}
}

int scm80211_ftm_get_para_b(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, char *str)
{
	int pdet[5] = { 0 };
	char *ptr = str;

	ptr += sprintf(ptr, "\n");
	if (ftm->chain < ftm_chain_mimo) {
		if (is_usb_device(hw)) {
			pdet[0] = ftm->chain;
			scm80211_get_param_b(hw, pdet, sizeof(pdet));
		} else
			scm2020_pdet_get_para_b(ftm->chain, &pdet[1]);
		sprintf(ptr, "caldat_%d: %d,%d,%d,%d\n", ftm->chain, pdet[1], pdet[2], pdet[3],
				pdet[4]);
	} else
		sprintf(ptr, "cannot get caldat in MIMO mode\n");

	return strlen(str);
}

void scm80211_ftm_set_xtal_cal(struct ieee80211_hw *hw, int cal)
{
	u32 v;

	v = scm80211_raw_readl(hw, RF_BASE_START + 0x10);
	v &= ~0x3F;
	cal += 0x20;
	v += cal;
	scm80211_raw_writel(hw, v, RF_BASE_START + 0x10);
}

int scm80211_ftm_get_xtal_cal(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
	u32 v;

	v = scm80211_raw_readl(hw, RF_BASE_START + 0x10) & 0x3F;
	return (int)(v - 0x20);
}


static int scm80211_ftm_ctx_init(struct ieee80211_hw *hw, struct scm80211_ftm *ftm)
{
#ifdef SCM2020_CFO
	struct scm80211_phy *phy = NULL;
#endif
	struct scm80211_chanctx ftm_ctx;
	struct ieee80211_channel *chan = NULL;
	enum nl80211_channel_type chan_type = NL80211_CHAN_NO_HT;
	u8 i;

	BUG_ON(ftm->channel <= 0);

	scm80211_set_hw_band(hw, ftm);

	if (ftm->channel < 14) {
		chan = &hw->wiphy->bands[ftm->fixrate.band]->channels[ftm->channel - 1];
	} else {
		for (i = 0; i < 25; i++) {
			chan = &hw->wiphy->bands[ftm->fixrate.band]->channels[i];
			if (ftm->center_freq1 == chan->center_freq)
				break;
		}
	}
	if (!chan)
		return -1;

	if (ftm->ht_mod != NON_HT) {
		if (ftm->width == NL80211_CHAN_WIDTH_20) {
			chan_type = NL80211_CHAN_HT20;
		} else if (ftm->width == NL80211_CHAN_WIDTH_40) {
			if (ftm->center_freq1 > ftm->center_freq2)
				chan_type = NL80211_CHAN_HT40MINUS;
			else
				chan_type = NL80211_CHAN_HT40PLUS;
		} else if (ftm->width == NL80211_CHAN_WIDTH_80) {
			if (ftm->center_freq1 > ftm->center_freq2)
				chan_type = SCM80211_CHAN_HT80MINUS;
			else
				chan_type = SCM80211_CHAN_HT80PLUS;
		}
	}
	FTM_DBG(hw, "%s: ht_mod=%d, chan_type=%d\n", __func__, ftm->ht_mod, chan_type);

#ifdef SCM2020_CFO
	if (ftm->ftm_cfo) {
		for_each_active_phy(phy, hw) {
			if (phy && phy->radio)
				scm80211_phy_set_cfo_freq(phy, ftm->ftm_cfo);
		}
	}
#endif
	scm80211_chandef_create(&ftm_ctx.ctx.def, chan, chan_type, ftm);

	ftm_ctx.ctx.rx_chains_static = 1;
	ftm_ctx.ctx.rx_chains_dynamic = 1 + ((ftm->chain < ftm_chain_mimo) ? 0 : 1);
	ftm_ctx.mac80211 = false;
	ftm_ctx.id = 0x100;
	ftm_ctx.type = CHANCTX_TYPE_FTM;
	ftm_ctx.restart = true;
	ftm_ctx.begin = scm80211_chanctx_ftm_begin;
	ftm_ctx.end = scm80211_chanctx_ftm_end;
	ftm_ctx.dwell_time = -1UL;
	ftm_ctx.vif = NULL;
	ftm_ctx.pv.present = BIT(PHY_CHANNEL) | BIT(PHY_ACTIVE_RXCHAIN_SET);
	spin_lock_init(&ftm_ctx.lock);
	INIT_LIST_HEAD(&ftm_ctx.sched_list);

	scm80211_ftm_set_param_b(hw, ftm);
	scm2020_set_comp_en((ftm->chain < ftm_chain_mimo) ? true : false);

	scm80211_set_chain(hw, ftm);
	scm80211_set_chanctx(hw, &ftm_ctx);

	/* Rx no need set gainidx */
	if (ftm->ftm_flag == 2)
		return 0;

	return scm80211_set_gainidx(hw, ftm);
}

int scm80211_ftm_init(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	memset(ftm, 0, sizeof(struct scm80211_ftm));
	FTM_DBG(hw, "[+ %s] ftm_en=%d\n", __func__, ftm_en);
	if (!ftm_en)
		return 0;

	ftm_clear_parms(hw, vif);
	ftm->enabled = ftm_en;

	scm80211_ftm_tasklet_init(hw, vif);

	FTM_DBG(hw, "[- %s]: FTM_DEBUG\n", __func__);

	return 0;
}

int scm80211_ftm_deinit(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	if (!ftm_en)
		return 0;

	ftm->enabled = 0;

	return 0;
}

static void scm80211_ftm_process_done_txop(struct txop *txop)
{
	struct ieee80211_hw *hw = txop->hw;
	struct sk_buff *skb;
	int i = 0;

	if (txop->state != TXOP_STATE_TX_DONE)
		return;

	while ((skb = skb_dequeue(&txop->sched))) {
		/* Unmap tx buffers */
		scm80211_cleanup_tx(hw, txop, i);
		dev_kfree_skb_any(skb);

		i++;
		FTM_DBG(hw, " %s: free skb directly(%d)\n", __func__, i);
	}
}

static void scm80211_ftm_complete_txop_epilogue(struct txop *txop)
{
	struct ieee80211_hw *hw = txop->hw;
	struct txop *next;

	if (txop->state != TXOP_STATE_TX_DONE)
		goto out;

	FTM_DBG(hw, "[+ %s]\n", __func__);
	/*
	 * Full or partial failure.
	 *
	 * Try to transmit the next available schedule. If there is none
	 * or it is not eligible for transmission, the driver let
	 * the hardware queue work deal with it with a higher priority.
	 */

	next = scm80211_get_txop(txop->hwq, TXOP_STATE_SCHEDULED);
	if (next && scm80211_start_backoff(next)) {
		scm80211_txop_free(txop);
	} else {
		scm80211_txop_change_state(txop, TXOP_STATE_NEEDS_RESCHED);
		scm80211_txop_add(txop, TXOP_STATE_NEEDS_RESCHED);
	}
out:
	FTM_DBG(hw, "[- %s] schedule_work\n", __func__);
	schedule_work(&txop->hwq->work);
}

static void scm80211_ftm_hwq_tasklet(unsigned long data)
{
	struct scm80211_hwq *hwq = (struct scm80211_hwq *)data;
	struct txop *txop;

	if (is_pcie_device(hwq->hw)) {
		unsigned long flags;

		spin_lock_irqsave(&hwq->ring.lock, flags);
		if (!scm80211_has_txop(hwq, TXOP_STATE_ISSUED))
			hrtimer_try_to_cancel(&hwq->watchdog);
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
	}
	while ((txop = scm80211_get_txop(hwq, TXOP_STATE_TX_DONE))) {
		if (txop->state == TXOP_STATE_TX_RETRY) {
			if (!scm80211_start_backoff(txop))
				schedule_work(&hwq->work);
			continue;
		}
		scm80211_ftm_process_done_txop(txop);
		scm80211_ftm_complete_txop_epilogue(txop);
	}
}

static void scm80211_ftm_complete_txop_prologue(struct txop *txop)
{
	struct ieee80211_hw *hw = txop->hw;
	struct scm80211_hwq *hwq = txop->hwq;

	txop->actual_time = ktime_us_delta(ktime_get(), txop->tstamp);
	txop->sent = txop->acked = txop->nacked = 0;
	txop->txd.ok = 1;
	txop->txd.tb = 0;
	txop->txd.nack = 0;
	txop->txd.fail = 0;

	FTM_DBG(hw,
		"%s(hwq#%d.%d): cmd=%08x result=%08x (ok=%d tb=%d nack=%d fail=%d)\n",
		__func__,
		hwq->id, txop->id, txop->cmd, txop->txd.result,
		txop->txd.ok, txop->txd.tb, txop->txd.nack, txop->txd.fail);

	txop->sent = txop->acked = txop->nacked = 0;
	scm80211_txop_change_state(txop, TXOP_STATE_TX_DONE);
}

static void scm80211_ftm_process_successful_txop(struct txop *txop)
{
	int tid;
	struct scm80211_hwq *hwq;
	int n = skb_queue_len(&txop->sched);
	struct tx_mpdu_desc *start_md = &txop->dma[0].desc->mdesc[0];
	struct tx_mpdu_desc *end_md = &txop->dma[0].desc->mdesc[n - 1];

	if (txop->space == MT_TGT_BD_TGT) {
		memcpy_fromio(word(start_md, 2),
			      word(&txop->dma[1].desc->mdesc[0], 2), sizeof(__le32) * 3);
		memcpy_fromio(word(end_md, 2),
			      word(&txop->dma[1].desc->mdesc[n - 1], 2), sizeof(__le32) * 3);
	}
	if (!(txop->txd.ok && !txop->txd.nack) || !start_md->acked ||
	    (txop->txd.tb && !end_md->sent))
		return;

	if (!txop->ampdu.enabled)
		return;

	hwq = txop->hwq;

	tid = txop->owner->txq.tid;
	txop->txd.ok = 1;
	txop->txd.nack = 0;

	FTM_DBG(txop->hw,
		"%s: tid[%d]: txop->owner->sb.ssn=%d txop->ampdu.tids[tid].ssn=%d",
		__func__, tid, txop->owner->sb.ssn, txop->ampdu.tids[tid].ssn);
	txop->owner->sb.ssn = txop->ampdu.tids[tid].ssn;
}

void scm80211_ftm_handle_txop_complete(struct ieee80211_hw *hw, struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;

	FTM_DBG(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

	scm80211_ftm_complete_txop_prologue(txop);
	scm80211_ftm_process_successful_txop(txop);
	scm80211_txop_add(txop, TXOP_STATE_TX_DONE);

	FTM_DBG(hw, "[- %s] schedule_work\n", __func__);
	tasklet_hi_schedule(&hwq->tasklet);
}

bool scm80211_ftm_tx_irq(struct ieee80211_hw *hw, unsigned long queues)
{
	struct scm80211_hwq *hwq;
	struct txop *txop;
	int i;

	if (!ftm_en)
		return false;

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		txop = scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
		if (!txop) {
			FTM_DBG(hw, "%s(hwq#%d) bogus interrupt\n", __func__, hwq->id);
			continue;
		}
		FTM_DBG(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

		/* Read and store hardware txop context */
		txop->cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		txop->txd.result = readl(&hwq->txdesc->result);
		txop->retry_cnt = 1;
		txop->scheduled = true;

		scm80211_ftm_handle_txop_complete(hw, txop);
	}

	return true;
}


void scm80211_ftm_set_single_tone(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, char *rsp)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy = priv->phy;
	const u8 idx_base[2] = {0xAF, 0xB3};
	int ret, chain = (ftm->chain < ftm_chain_mimo) ? 1 : 0;
	u32 v = 0;

	if (ftm->ftm_flag) {
		//chan_freq
		ret = scm80211_ftm_ctx_init(hw, ftm);
		if (ret < 0) {
			sprintf(rsp, "set ctx_init fail(%d)\n", ret);
			return;
		}
		//tone
		phy_writel(phy, ftm->tone_ofs, PHY_RFIF(PHY_RFI_TX_TONEGEN_0));
		phy_writel(phy, 0x01130001, PHY_RFIF(PHY_RFI_TX_SIGGEN_CFG));
		//tx_manual
		v = idx_base[(ftm->channel < 14) ? 0 : 1];
		v += ((ftm->gain_index[chain] % 2) ? 0 : 1);
		scm80211_phy_writel(hw, v, PHY_RFI_TX_DIG_GAIN_MAN);
	} else
		phy_writel(phy, 0, PHY_RFIF(PHY_RFI_TX_SIGGEN_CFG));
}

#endif // SCM80211_FTM
