/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
 *
 * Copyright (c) 2020-2021 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifndef __SCM80211_FTM_H__
#define __SCM80211_FTM_H__

#include <linux/types.h>
#include <linux/workqueue.h>
#include <linux/slab.h>
#include <linux/regmap.h>
#include <linux/proc_fs.h>
#include <linux/platform_device.h>
#include <linux/pci.h>
#include <linux/usb.h>
#include <linux/average.h>
#include <linux/hrtimer.h>
#include <linux/version.h>

#include <net/mac80211.h>
#include <linux/version.h>

#include "cfg80211.h"
#include "hrtimer.h"
#include "skbuff.h"
#include "ieee80211ax.h"
#include "hw.h"
#include "phy.h"
#include "random.h"
#include "debug.h"


#define FTM_SKB_NUM					2
#define FTM_MAX_PKT_NUM				64
#define FTM_MAX_TX_CNT				0xFFFFFFFFUL

#define MAX_CMD_STR					2000
#define FTM_RSP_SIZE				1000
#define FTM_MAX_PKT_SIZE			1500
#define FTM_PER_AMPDU_SIZE			2304
#define FTM_MAX_AMPDU_SIZE			(FTM_MAX_PKT_NUM * 1024)

#define SCM80211_FTM_RX_RSSI_MAX	10
#define SCM80211_CHAN_HT80MINUS		4
#define SCM80211_CHAN_HT80PLUS		5

#define	FTM_POWER_NOUSE				-128
#define	FTM_GAINIDX_NOUSE			-99
#define	FTM_GAINIDX_MAX				63

struct scm80211_scm_cal {
	int rf_cal_2g;
	int rf_cal_5g_low;
	int rf_cal_5g_mid;
	int rf_cal_5g_high;
};

#ifdef SCM80211_FTM

enum ftm_chain_idx {
	ftm_chain_0 = 0,
	ftm_chain_1,
	ftm_chain_mimo,
	ftm_chain_max,
};

enum FTM_ERR_LIST {
	FTM_CMD_OK = 0,
	FTM_ERR_COMM,
	FTM_ERR_TX_PKTSZ,
	FTM_ERR_CHANNEL,
	FTM_ERR_HTMODE,
	FTM_ERR_AMPDU,		//5
	FTM_ERR_FREQ2,
	FTM_ERR_HT80_MCS,
	FTM_ERR_HE_MCS,
	FTM_ERR_HE_GI,
	FTM_ERR_HE_LTF,		//10
	FTM_ERR_CFO,
	FTM_ERR_MIMO,
	FTM_ERR_NSS,
	FTM_ERR_MAX,
};

enum FTM_SET_CMDS {
	CMD_CLEAR = 0,
	CMD_PKTCNT,
	CMD_PKTSIZE,
	CMD_SET_CHAN,
	CMD_POWER,
	CMD_CHAIN,
	CMD_RATE,
	CMD_MACADDR,
	CMD_IFS,
	CMD_AMPDU,
	CMD_ACK,
	CMD_TX,
	CMD_RX,
	CMD_SET_GAINIDX,
	CMD_SET_GPIO,
	CMD_SET_EFUSE,
	CMD_SET_CALDAT,
	CMD_SET_CFO,
	CMD_SET_STONE,
	CMD_SET_XTAL,
	CMD_SET_MAX,
};

enum FTM_GET_CMDS {
	CMD_FTM_HELP = 0,
	CMD_GET_RSP,
	CMD_TXPARMS,
	CMD_RXPARMS,
	CMD_RXCNT,
	CMD_GET_GPIO,
	CMD_GET_EFUSE,
	CMD_GET_RSSI,
	CMD_GET_TSSI,
	CMD_GET_CALDAT,
	CMD_GET_GAINIDX,
	CMD_GET_CHAN,
	CMD_GET_XTAL,
	CMD_GET_MAX,
};

enum FTM_PRINT_SEL {
	LOG_FTM_TX = 0,
	LOG_FTM_RX,
};

struct scm80211_ftm_rx_statistics {
#define SCM80211_RX_PKT_DFT 100
	int rx_pkt;
	int rx_ppkt;		// last packet count
	int rx_err;
	int rx_mpdu;
	int rx_pmpdu;
	// All packets should share the same MCS/NSS/BCC/LDPC/etc
	struct ieee80211_rx_status rxs;
	int cur;		// current index of rssi
	int rssi[SCM80211_FTM_RX_RSSI_MAX];
};

/**
 * Need check sync with core.h
 * ieee80211_dbg_rate_fix
 * @rate: indicates the rate info of HT/VHT/HE
 * @he_gi_type: indicates the he GI type
 * @he_ltf_type: indicates the HE-LTF type
 * @flags: record the mac80211_tx_info_flags such as
 *	   IEEE80211_TX_CTL_LDPC
 * @rate_idx: indicates the NON-HT rate index
 */
struct scm80211_ftm_rate {
	struct ieee80211_tx_rate rate;
	enum nl80211_band band;
	enum gi_type he_gi_type;
	enum ltf_type he_ltf_type;
	u32 flags;
	/* Add more */
};

enum scm80211_ftm_tx_state {
	FTM_TX_STOP = 0,
	FTM_TX_START,
	FTM_TX_COMPLETE,
};

enum scm80211_ftm_rx_state {
	FTM_RX_STOP = 0,
	FTM_RX_START,
	FTM_RX_COMPLETE,
	FTM_RX_TIMEOUT,
};

struct scm80211_ftm_txskb {
	u8 skb_id;
	u8 total_num;
	u8 txd_num;
	u32 total_cnt;
	u32 txd_cnt;

	struct sk_buff *tx_skb[FTM_SKB_NUM][FTM_MAX_PKT_NUM];
};

struct scm80211_ftm {
	struct scm80211_ftm_txskb ftm_tx_skb;
	bool enabled;
	bool dbdc;
	int ftm_flag;		//running flag, 0:stop; 1:tx start;     2:rx start
	u32 txpktcnt;
	int txpktsize;
	int channel;		//channel 1 ,6 ,11, etc
	int center_ch;		//20/40/80M center channel
	u32 center_freq1;
	u32 center_freq2;
	int width;
	int chain;
	struct scm80211_ftm_rate fixrate;
	u8 macaddr[ETH_ALEN];
	u32 ifs;
	u8 cbw;
	u8 rate_nss;
	u8 ampdu;
	u8 amsdu;
	enum scm80211_ftm_tx_state tx_start;
	enum scm80211_ftm_rx_state rx_start;
	struct scm80211_scm_cal cal_dat[2];

	u8 tid;
	u8 qos;
	u16 seqnum;
	u32 txd_pktcnt;
	int ac;			//be/bk/vi/vo/mgmt
	int ack_policy;
	int txpower;
	int gain_index[2];
	int ftm_cfo;
	int freq_cal;

	enum ppdu_format ht_mod;

	int rx_duration;	//ms
	struct ieee80211_channel chan;
#define SCM80211_FTM_RX_DURATION_MS (5 * 60 * 1000)
	struct completion completion;
	u32 rxpktcnt;		//the expected number
	struct scm80211_ftm_rx_statistics rx_statistics;
	int ftm_tssi;

	//single tone
	u32 tone_ofs;
	u8 pri_ch;

	char ftm_cmd_rsp[MAX_CMD_STR];
	char ftm_rx_rpt[FTM_RSP_SIZE];
	char ftm_rssi_rpt[FTM_RSP_SIZE];
	char ftm_tx_rpt[FTM_RSP_SIZE];
	int err_flag;
};

int ftm_clear_parms(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
int scm80211_ftm_get_para_b(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, char *str);
int scm80211_get_gainidx(struct ieee80211_hw *hw, char *rsp);
int scm80211_ftm_init(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
int scm80211_ftm_deinit(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

bool scm80211_ftm_rx_enable_filter(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
void scm80211_ftm_set_xtal_cal(struct ieee80211_hw *hw, int cal);
int scm80211_ftm_get_xtal_cal(struct ieee80211_hw *hw, struct scm80211_ftm *ftm);

void scm80211_ftm_tx_tssi_show(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, int ch);
void scm80211_ftm_rx_rssi_show(struct ieee80211_hw *hw, struct scm80211_ftm *ftm);
void scm80211_ftm_rx_statistics_show(struct ieee80211_hw *hw, struct scm80211_ftm *ftm);
void scm80211_ftm_rx_statis_update(struct scm80211_rx_data *rx);
int scm80211_ftm_rx_start(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
int scm80211_ftm_rx_stop(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
int scm80211_ftm_rx_clear(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

int scm80211_ftm_tx_start(struct ieee80211_hw *hw, struct ieee80211_vif *vif);
int scm80211_ftm_tx_stop(struct ieee80211_hw *hw, struct ieee80211_vif *vif);

bool scm80211_ftm_tx_irq(struct ieee80211_hw *hw, unsigned long queues);

void scm80211_ftm_set_single_tone(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, char *rsp);

#endif
#endif /* __SCM80211_FTM_H__ */
