// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 * RM: resource manager
 * HTC: host target communication
 */
#include <linux/pci.h>
#include <linux/aer.h>
#include <linux/rbtree.h>
#include <linux/firmware.h>
#include <linux/bitfield.h>
#include <linux/version.h>
#include <linux/io.h>

#define func_fmt(str)           "[PCIE HTC] "str
#define func_trace              BUS

#include "core.h"
#include "util.h"
#include "debug.h"
#include "htc.h"
#define SCM_WAIT_FOR_EVENT_MAX_RETRY 2
#define SCM_HTC_EVENT_COOKIE 0x1a2b3c4d
#define SCM_HTC_EVENT_WAIT_MAX 200	//MS
#define SRAM_MEM_SIZE           0x100000	//1M

#define HTC_DBG_PRINT   0	// HTC print flag
#ifdef HTC_DBG_PRINT
#define HTC_DBG(hw, fmt, args...)   scm80211_trace(hw, fmt, ##args)
#else
#define HTC_DBG(hw, fmt, args...)
#endif
#define HTC_ERR(hw, fmt, args...)   scm80211_err(hw, fmt, ##args)

#ifdef FW_LOG_UPLOAD
#define LOG_INFO_BUF_SIZE 300
#define PRINTK_BUFIDX(i, buf_len) ((i) & ((buf_len) - 1))

static int fw_log_upload_by_t2h_proc(struct ieee80211_hw *hw, u8 *data, u32 data_len)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	struct ieee80211_hw_priv *priv;
	unsigned __iomem *printk_head;
	u8 __iomem *printk_buf;
	u8 __iomem *ram;
	unsigned int tail;
	int n = 0;
	char c;
	u32 *buf;
	u32 printk_head_ram_idx;
	u32 print_buf_ram_idx;
	u32 PRINTK_BUFSIZ;
	u8 info[LOG_INFO_BUF_SIZE];
	u32 len = 0;

	buf = data ? (u32 *)data : NULL;
	printk_head_ram_idx = buf ? buf[0] : 0;
	print_buf_ram_idx = buf ? buf[1] : 0;
	PRINTK_BUFSIZ = buf ? buf[2] : 0;
	priv = hw ? hw->priv : NULL;
	ram = htc_ctx ? (htc_ctx->fw_log_ram) : NULL;
	printk_head = ram ? (unsigned int *)(ram + printk_head_ram_idx) : NULL;
	printk_buf = ram ? (ram + print_buf_ram_idx) : NULL;

	if (!hw || !data || !priv || !ram || !printk_head || !printk_buf)
		return -1;

	HTC_DBG(hw, "FW log upload below:\n");

	tail = htc_ctx->fw_log_printk_tail;
	while (tail != *printk_head && n++ < PRINTK_BUFSIZ) {
		if (*printk_head >= tail + PRINTK_BUFSIZ) {
			HTC_DBG(hw, "FW LOG : ... <truncated> ...");
			tail = *printk_head - (PRINTK_BUFSIZ - 1);

			/* Eat up the remaining bytes in the truncated line */
			while (printk_buf[PRINTK_BUFIDX(tail, PRINTK_BUFSIZ)] != '\n')
				tail++;
			tail++;
		}
		c = printk_buf[PRINTK_BUFIDX(tail++, PRINTK_BUFSIZ)];
		if (len >= (LOG_INFO_BUF_SIZE - 1) || c == '\n') {
			info[len++] = 0;
			HTC_DBG(hw, "FW LOG : %s", info);
			len = 0;
		} else {
			info[len++] = c;
		}
	}

	htc_ctx->fw_log_printk_tail = tail;

	return 0;
}

void log_upload_msg_proc(void *priv, void *msg, u16 len)
{
	struct ieee80211_hw *hw = priv;

	if (!hw)
		return;

	fw_log_upload_by_t2h_proc(hw, msg, len);
}

int log_upload_init(struct ieee80211_hw *hw)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	int ret = 0;
	u32 type = HTC_MSG_INDICATE_FW_LOG_UPLOAD;

	htc_ctx->fw_log_printk_tail = 0;
	htc_ctx->fw_log_ram = pcim_iomap_table(htc_ctx->pdev)[3];
	htc_register_msg_handle(hw, type, log_upload_msg_proc, hw);

	return ret;
}

int log_upload_deinit(struct ieee80211_hw *hw)
{
	int ret = 0;
	u32 type = HTC_MSG_INDICATE_FW_LOG_UPLOAD;

	htc_unregister_msg_handle(hw, type);

	return ret;
}

#endif

static inline int scm80211_htc_enqueue(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	int ret = 0;

	HTC_DBG(hw, "%s: %d\n", __func__, __LINE__);
	if (htc_ctx->htc_ops->htc_enqueue)
		ret = htc_ctx->htc_ops->htc_enqueue(hw, skb);
	return ret;
}

static inline void scm80211_notify_remote(struct ieee80211_hw *hw)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);

	HTC_DBG(hw, "%s: %d\n", __func__, __LINE__);
	if (htc_ctx->htc_ops->notify_remote)
		htc_ctx->htc_ops->notify_remote(hw);
}

static inline int scm80211_htc_send(struct ieee80211_hw *hw, struct sk_buff *skb, int type)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	int ret = 0;
	struct ieee80211_hw_priv *priv = hw->priv;

	HTC_DBG(hw, "%s: type=0x%x\n", __func__, type);
	if (htc_ctx->htc_ops)
		ret = htc_ctx->htc_ops->htc_msg_send(priv, skb, type);
	return ret;
}

static inline int scm80211_htc_recv(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	int ret = 0;
	struct ieee80211_hw_priv *priv = hw->priv;

	HTC_DBG(hw, "%s: %d\n", __func__, __LINE__);
	if (htc_ctx->htc_ops)
		ret = htc_ctx->htc_ops->htc_msg_recv(priv, skb);
	return ret;
}

static inline struct sk_buff *scm80211_htc_alloc(struct ieee80211_hw *hw, const void *data, int len)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	struct sk_buff *skb = NULL;
	struct ieee80211_hw_priv *priv = hw->priv;

	HTC_DBG(hw, "%s: %d\n", __func__, __LINE__);
	if (htc_ctx->htc_ops)
		skb = htc_ctx->htc_ops->htc_msg_alloc(priv, data, len);
	return skb;
}

static inline void scm80211_htc_free(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);

	HTC_DBG(hw, "%s: %d\n", __func__, __LINE__);
	if (htc_ctx->htc_ops)
		htc_ctx->htc_ops->htc_msg_free(priv, skb);
}

#ifdef HEX_DBG_PRINT
char hex_to_char(u8 hex)
{
	char ch = 0;

	if (hex >= 0 && hex <= 0x09)
		ch = hex + '0';
	else if (hex >= 0x0a && hex <= 0x0f)
		ch = (hex & 0x0f) + ('a' - 0x0a);

	return ch;
}

void array_to_string(u8 *data, u8 data_len, u8 *string)
{
	u8 i, *p_string = string;

	*(p_string++) = '{';
	*(p_string++) = ' ';
	for (i = 0; i < data_len; i++) {
		*(p_string++) = hex_to_char(data[i] >> 4);
		*(p_string++) = hex_to_char(data[i] & 0x0f);
		*(p_string++) = ' ';
	}
	*(p_string++) = '}';
	*p_string = '\0';
}

void htc_print_hex_dbg(struct ieee80211_hw_priv *dev, u8 *dat, u16 len)
{
	char *arry;
	u16 dat_len = len * 4;

	HTC_DBG(dev->hw, "%s malloc (%d)%d Bytes\n", __func__, len, dat_len);
	arry = kzalloc(dat_len, GFP_KERNEL);
	if (!arry) {
		HTC_ERR(dev->hw, "malloc htc dbg fail!\n");
		return;
	}
	array_to_string(dat, len, arry);
	HTC_DBG(dev->hw->wiphy, "%s: %s\n", __func__, arry);
	kfree(arry);
}
#endif

static inline void htc_msg_hdr_build(struct ieee80211_hw_priv *dev, struct sk_buff *skb, int type)
{
	u16 len = skb->len;
	struct htc_hdr *hdr = skb_push(skb, sizeof(struct htc_hdr));

	hdr->chip = 0;
	hdr->band = dev->devid;
	hdr->msg_type = type;
	hdr->len = len;
	hdr->cookie = *(u32 *)&skb;
	HTC_DBG(dev->hw, "%s: type=0x%x msg len=%d skb len=%d\n", __func__, type, len, skb->len);
}

/**
 * htc_msg_send	-	transmit htc message and return(without waiting).
 * @dev:
 * @skb:
 */
static int htc_msg_send(struct ieee80211_hw_priv *dev, struct sk_buff *skb, int type)
{
	int ret = -1;

	htc_msg_hdr_build(dev, skb, type);

	HTC_DBG(dev->hw, "%s: type=%d cookie=%p len=%d\n", __func__, type, skb, skb->len);
	ret = scm80211_htc_enqueue(dev->hw, skb);
	if (ret)
		HTC_ERR(dev->hw, "%s: send htc msg=%d failed\n", __func__, type);

	return ret;
}

/**
 * htc_msg_send_recv	-	send request and waiting for response
 * @dev:
 * @skb:	HTC message
 * @type:	HTC message type
 */
static int htc_msg_send_recv(struct ieee80211_hw_priv *dev, struct sk_buff *skb, int type)
{
	// not support yet
	return 0;
}

/**
 * htc_msg_recv	-	receive and process HTC message(e.g. from rx ring)
 * @dev:
 * @skb:
 */
static int htc_msg_recv(struct ieee80211_hw_priv *dev, struct sk_buff *skb)
{
	int ret = -1;
	struct htc_msg_reg_ops *item;
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(dev->hw);
	struct htc_msg *htc = (struct htc_msg *)skb->data;
	struct htc_hdr *common = &htc->hdr;

	HTC_DBG(dev->hw, "%s: skb->len=%d msg_type=0x%x\n",
		__func__, common->len, common->msg_type);

	if (is_valid_msg(common->msg_type)) {
		item = &htc_ctx->msg_reg_ops[common->msg_type];
	} else {
		HTC_ERR(dev->hw, "+ %s msg_type: 0x%X is invalid\n", __func__, common->msg_type);
		return ret;
	}

	if (item->msg_proc_cb) {
		item->msg_proc_cb(item->msg_priv, htc->data, common->len);
		ret = 0;
	}

	return ret;
}

struct sk_buff *htc_msg_alloc(struct ieee80211_hw_priv *dev, const void *data, int data_len)
{
	int length = HTC_HEADROOM + data_len;
	struct sk_buff *skb;

	skb = alloc_skb(length, GFP_KERNEL);
	if (!skb)
		return NULL;

	memset(skb->head, 0, length);
	skb_reserve(skb, HTC_HEADROOM);

	if (data && data_len)
		skb_put_data(skb, data, data_len);

	return skb;
}

void htc_msg_free(struct ieee80211_hw_priv *dev, struct sk_buff *skb)
{
	if (skb)
		dev_kfree_skb_any(skb);
}

#define order_update(_queue, _order) \
	iowrite32((_order), &((struct htc_shared_memory *)((_queue)->base))->order)
#define rptr_update(_queue, _rptr) \
	iowrite32((_rptr), &((struct htc_shared_memory *)((_queue)->base))->rptr)
#define wptr_update(_queue, _wptr) \
	iowrite32((_wptr), &((struct htc_shared_memory *)((_queue)->base))->wptr)

#define rptr_sync(_queue)	(((struct htc_shared_memory *)((_queue)->base))->rptr)
#define wptr_sync(_queue)	(((struct htc_shared_memory *)((_queue)->base))->wptr)

#define entry_get(_queue, _wptr) \
	(&(((struct htc_shared_memory *)((_queue)->base))->entries[(_wptr)]))

void scm80211_h2t_init(void *mem_base, struct htc_queue *h2t)
{
	h2t->base = mem_base;
	h2t->order = HTC_ENTRY_ORDER;
	h2t->rptr = 0;
	h2t->wptr = 0;
	order_update(h2t, h2t->order);
	rptr_update(h2t, 0);
	wptr_update(h2t, 0);
}

void scm80211_t2h_init(void *mem_base, struct htc_queue *t2h)
{
	t2h->base = (u8 *)mem_base + sizeof(struct htc_shared_memory);
	t2h->order = HTC_ENTRY_ORDER;
	t2h->rptr = 0;
	t2h->wptr = 0;

	order_update(t2h, t2h->order);
	rptr_update(t2h, 0);
	wptr_update(t2h, 0);
}

//Host side
#define ring_is_full(_ring) \
	((ring_index((_ring), (_ring)->rptr) == ring_index((_ring), (_ring)->wptr)) && \
		    (ring_ptr((_ring), (_ring)->rptr) != ring_ptr((_ring), (_ring)->wptr)))

#define ring_is_empty(_ring) \
	(ring_ptr((_ring), (_ring)->rptr) == ring_ptr((_ring), (_ring)->wptr))

#define ring_ptr_equal(_ring, _ptr1, _ptr2) \
	(ring_ptr((_ring), (_ptr1)) == ring_ptr((_ring), (_ptr2)))

static int _htc_enqueue(struct ieee80211_hw *hw, struct htc_queue *q, struct sk_buff *skb)
{
	int idx;
	struct ieee80211_hw_priv *dev = hw->priv;
	struct htc_msg *msg;

	q->rptr = rptr_sync(q);
	if (unlikely(ring_is_full(q))) {
		HTC_ERR(hw, "%s: ring is full devid=%d h2t->rptr=0x%x h2t->wptr=0x%x\n",
			__func__, dev->devid, q->rptr, q->wptr);
		return -1;
	}

	idx = ring_index(q, q->wptr);
	msg = entry_get(q, idx);

	__iowrite32_copy(msg, skb->data,
			 WARN_ON(skb->len > sizeof(*msg)) ? sizeof(*msg) : skb->len);

	/* TODO: is it necessary? add a proper reason. IO ops have implicit barrier */
	wmb();

	q->wptr++;
	wptr_update(q, q->wptr);

	HTC_DBG(hw, "%s: devid=%d w_addr=%#llx h2t->rptr=%d h2t->wptr=%d, idx=%d\n",
		__func__, dev->devid, (u64)msg, q->rptr, q->wptr, idx);

#ifdef HEX_DBG_PRINT
	htc_print_hex_dbg(dev, (u8 *) msg, (skb->len + sizeof(struct htc_hdr)));
#endif
	return 0;
}

int htc_enqueue(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	int ret;
	struct ieee80211_hw_priv *dev = hw->priv;
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	struct queue_context *ctx = &htc_ctx->htc_queues[dev->devid];
	unsigned long flags;

	spin_lock_irqsave(&ctx->lock, flags);
	ret = _htc_enqueue(hw, &ctx->h2t, skb);
	spin_unlock_irqrestore(&ctx->lock, flags);
	if (ret) {
		HTC_ERR(hw, "%s: devid=%d _htc_enqueue failed %d\n", __func__, dev->devid, ret);
		return ret;
	}

	scm80211_notify_remote(hw);

	return 0;
}

static int _htc_dequeue(struct ieee80211_hw_priv *dev, struct queue_context *ctx,
			struct htc_queue *q)
{
	int done = 0;
	struct htc_msg *msg;
	struct sk_buff *skb;

	q->wptr = wptr_sync(q);
	HTC_DBG(dev->hw, "%s: devid=%d q=%p, t2h->rptr=0x%x t2h->wptr=0x%x\n",
		__func__, dev->devid, q, q->rptr, q->wptr);

	if (unlikely(ring_is_empty(q))) {
		HTC_ERR(dev->hw, "%s: ring is empty devid=%d t2h->rptr=0x%x\n",
			__func__, dev->devid, q->rptr);
		return -1;
	}

	while (!ring_ptr_equal(q, q->rptr, q->wptr)) {
		msg = entry_get(q, ring_index(q, q->rptr));
		skb = scm80211_htc_alloc(dev->hw, msg, sizeof(struct htc_hdr) + msg->hdr.len);
		if (!skb) {
			// TODO
			goto next_entry;
		}
		done++;
		skb_queue_tail(&ctx->list, skb);
next_entry:
		q->rptr++;
		if (ring_ptr_equal(q, q->rptr, q->wptr))
			q->wptr = wptr_sync(q);
	}
	rptr_update(q, q->rptr);
	HTC_DBG(dev->hw, "%s: devid=%d t2h->rptr=0x%x t2h->wptr=0x%x done=%d\n",
		__func__, dev->devid, q->rptr, q->wptr, done);
	return done;
}

int htc_dequeue(struct ieee80211_hw *hw)
{
	struct scm80211_htc_ctx *htc_ctx = NULL;
	int done = 0;
	struct ieee80211_hw_priv *dev;
	struct queue_context *ctx;

	if (!hw)
		return -1;

	dev = hw->priv;
	if (!dev)
		return -1;

	htc_ctx = scm80211_get_htc_ctx(hw);
	ctx = &htc_ctx->htc_queues[dev->devid];
	if (!ctx)
		return -1;

	done = _htc_dequeue(dev, ctx, &ctx->t2h);
	if (done)
		ieee80211_queue_work(hw, &ctx->work);

	return 0;
}

static void scm80211_rx_htc_work(struct work_struct *work)
{
	unsigned long flags;
	struct sk_buff *skb;
	struct sk_buff_head head;
	struct queue_context *ctx = container_of(work, struct queue_context, work);
	struct ieee80211_hw_priv *dev = ctx->dev;

	skb_queue_head_init(&head);

	HTC_DBG(dev->hw, "%s: entering\n", __func__);
	spin_lock_irqsave(&ctx->list.lock, flags);
	if (!ctx->list.qlen) {
		spin_unlock_irqrestore(&ctx->list.lock, flags);
		HTC_DBG(dev->hw, "%s:ctx->list.qlen=0 %d\n", __func__, __LINE__);
		return;
	}
	skb_queue_splice_init(&ctx->list, &head);
	spin_unlock_irqrestore(&ctx->list.lock, flags);

	while (!skb_queue_empty(&head)) {
		skb = __skb_dequeue(&head);
		scm80211_htc_recv(dev->hw, skb);
		scm80211_htc_free(dev->hw, skb);
	}
	HTC_DBG(dev->hw, "%s: exiting\n", __func__);
}

int scm80211_htc_queue_init(void *mem_base, struct queue_context *ctx)
{
	scm80211_h2t_init(mem_base, &ctx->h2t);
	scm80211_t2h_init(mem_base, &ctx->t2h);
	skb_queue_head_init(&ctx->list);

	INIT_WORK(&ctx->work, scm80211_rx_htc_work);
	spin_lock_init(&ctx->lock);
	ctx->init = 1;

	return 0;
}

static int scm80211_hw_init_done(struct scm80211_htc_ctx *htc_ctx)
{
	struct ieee80211_hw *hw = htc_ctx->htc_queues[0].dev->hw;
	int ret = -1;
	struct htc_req_add_dev add_dev = {
		.devid = 0,
		.band = 0,
		.dbdc = 0
	};
	struct htc_req_add_vif add_vif = {
		.mode = 0,
		.vif = 0,
	};
#ifdef FW_LOG_UPLOAD
	u8 log_upload_enable = 1;
#endif

	//notify FW
	ret =
	    scm80211_send_to_target_by_h2t(hw, HTC_MSG_WIFI_ADD_DEV, &add_dev,
					   sizeof(struct htc_req_add_dev));

	ret =
	    scm80211_send_to_target_by_h2t(hw, HTC_MSG_WIFI_ADD_VIF, &add_vif,
					   sizeof(struct htc_req_add_vif));

#ifdef FW_LOG_UPLOAD
	log_upload_init(hw);
	/* since htc queue len is 2 now, and target is slower than host, so wait a moment */
	msleep(500);
	ret =
	    scm80211_send_to_target_by_h2t(hw, HTC_MSG_ENABLE_FW_LOG_UPLOAD, &log_upload_enable, 1);
#endif

	return ret;
}

void scm80211_hw_deinit_action(struct scm80211_htc_ctx *htc_ctx)
{
	struct ieee80211_hw *hw = htc_ctx->htc_queues[0].dev->hw;
	struct htc_req_add_vif del_vif = {
		.mode = 0,
		.vif = 0,
	};
#ifdef FW_LOG_UPLOAD
	u8 log_upload_enable = 0;

	scm80211_send_to_target_by_h2t(hw, HTC_MSG_ENABLE_FW_LOG_UPLOAD, &log_upload_enable, 1);
	log_upload_deinit(hw);
#endif
	scm80211_send_to_target_by_h2t(hw, HTC_MSG_WIFI_DEL_VIF, &del_vif, sizeof(del_vif));
}

// HTC APIs should be common for all chipsets
static const struct scm80211_htc_ops scm2020_htc_ops = {
	.htc_msg_alloc = htc_msg_alloc,
	.htc_msg_free = htc_msg_free,
	.htc_msg_recv = htc_msg_recv,
	.htc_msg_send = htc_msg_send,
	.htc_msg_send_recv = htc_msg_send_recv,
	.htc_enqueue = htc_enqueue,
	.htc_dequeue = htc_dequeue,
	.notify_remote = scm80211_pcie_rc_notify_ep,
};

int scm80211_htc_init(struct pci_dev *pdev)
{
	int i = 0;
	void *mem_base;
	void *htc_base;
	struct scm80211_htc_ctx *htc_ctx = pcie_get_htc_ctx(pdev);
	struct queue_context *ctx;

	mem_base = pcim_iomap_table(pdev)[3];
	htc_base =
	    mem_base + SRAM_MEM_SIZE -
	    (sizeof(struct tx_dma_desc) * SCM80211_NUM_HWQ * SCM80211_MAX_TXOP * SCM2020_INTF_MAX) -
	    HTC_SHM_SIZE_MAX;
	HTC_DBG(htc_ctx->htc_queues[0].dev->hw, "mem_base=%p, htc_base=%p\n", mem_base, htc_base);
	for (i = 0; i < SCM80211_NUM_DEV; i++) {
		htc_base += HTC_SHM_SIZE_MAX_PER_BAND * i;
		HTC_DBG(htc_ctx->htc_queues[0].dev->hw, "htc_queue_init: htc_base=%p\n", htc_base);
		ctx = &htc_ctx->htc_queues[i];
		scm80211_htc_queue_init(htc_base, ctx);
		HTC_DBG(htc_ctx->htc_queues[0].dev->hw, "%s: h2t->base=%p h2t->order=%d\n",
			__func__, ctx->h2t.base, ctx->h2t.order);
		HTC_DBG(htc_ctx->htc_queues[0].dev->hw, "%s: t2h->base=%p t2h->order=%d\n",
			__func__, ctx->t2h.base, ctx->t2h.order);
	}
	htc_ctx->htc_ops = (struct scm80211_htc_ops *)&scm2020_htc_ops;
	htc_ctx->pdev = pdev;

	scm80211_hw_init_done(htc_ctx);

	return 0;
}

int scm80211_htc_deinit(struct pci_dev *pdev)
{
	int i;
	struct scm80211_htc_ctx *htc_ctx = pcie_get_htc_ctx(pdev);
	struct queue_context *ctx;

	HTC_DBG(htc_ctx->htc_queues[0].dev->hw, "%s: succ\n", __func__);

	scm80211_hw_deinit_action(htc_ctx);

	for (i = 0; i < SCM80211_NUM_DEV; i++) {
		ctx = &htc_ctx->htc_queues[i];
		flush_work(&ctx->work);
	}

	return 0;
}

int scm80211_send_to_target_by_h2t(struct ieee80211_hw *hw, u32 msg_type, void *data, u16 len)
{
	int ret = -1;
	struct sk_buff *htc;

	do {
		if (!is_valid_msg(msg_type))
			break;

		htc = scm80211_htc_alloc(hw, data, len);
		if (!htc)
			break;

		ret = scm80211_htc_send(hw, htc, msg_type);

		scm80211_htc_free(hw, htc);
	} while (0);

	HTC_DBG(hw, "HTC send msg type: 0x%x %s!\n", msg_type, ret ? "fail" : "success");

	return ret;
}

int htc_register_msg_handle(struct ieee80211_hw *hw, u32 msg_type,
			    void (*handle)(void *, void *, u16), void *priv)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	int ret = -EAGAIN;
	struct htc_msg_reg_ops *item;

	if (!htc_ctx || !is_valid_msg(msg_type))
		return ret;

	item = &htc_ctx->msg_reg_ops[msg_type];
	if (item->msg_proc_cb)
		return ret;

	item->msg_priv = priv;
	item->msg_proc_cb = handle;

	return 0;
}

int htc_unregister_msg_handle(struct ieee80211_hw *hw, u32 msg_type)
{
	struct scm80211_htc_ctx *htc_ctx = scm80211_get_htc_ctx(hw);
	int ret = -EAGAIN;
	struct htc_msg_reg_ops *item;

	if (!htc_ctx || !is_valid_msg(msg_type))
		return ret;

	item = &htc_ctx->msg_reg_ops[msg_type];
	if (!(item->msg_proc_cb))
		return ret;

	item->msg_priv = NULL;
	item->msg_proc_cb = NULL;

	return 0;
}
