/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifndef __SCM80211_HOST_TARGET_COMMUNICATION_H__
#define __SCM80211_HOST_TARGET_COMMUNICATION_H__

#include <linux/ieee80211.h>
#include <net/cfg80211.h>
#include <net/mac80211.h>
#include <linux/version.h>
#include "hw.h"

#define SCM80211_NUM_DEV        SCM2020_INTF_MAX	//dbdc
#define HTC_HEADROOM		48

#ifdef __staging__
enum htc_msg_type {
	HTC_MSG_INVALID = 0,
	HTC_MSG_FW = 1,		//FW downloading
	HTC_MSG_KEEPALIVE = 2,
	HTC_MSG_ENABLE_FW_LOG_UPLOAD,
	HTC_MSG_INDICATE_FW_LOG_UPLOAD,
	//wifi start
	HTC_MSG_WIFI_ADD_DEV,
	HTC_MSG_WIFI_DEL_DEV,
	HTC_MSG_WIFI_ADD_VIF,
	HTC_MSG_WIFI_DEL_VIF,
	HTC_MSG_WIFI_RX_START,
	HTC_MSG_WIFI_RX_STOP,
	HTC_MSG_WIFI_TX_START,
	HTC_MSG_WIFI_TX_STOP,
	HTC_MSG_WIFI_WOW_ENABLE,
	HTC_MSG_WIFI_WOW_DISABLE,
	HTC_MSG_WIFI_VIF_PS_ENABLE,
	HTC_MSG_WIFI_VIF_PS_DISABLE,
	HTC_MSG_WIFI_HW_PS_ENABLE,
	HTC_MSG_WIFI_HW_PS_DISABLE,
	HTC_MSG_MAX
};
#endif /* __staging__ */

#define is_valid_msg(x) (((x) >= 0) && ((x) < HTC_MSG_MAX))

struct htc_msg_reg_ops {
	void *msg_priv;
	void (*msg_proc_cb)(void *priv, void *msg, u16 len);
};

#define HTC_SHM_SIZE_MAX	4096
#define HTC_SHM_SIZE_MAX_PER_BAND (HTC_SHM_SIZE_MAX / 2)

#define HTC_ENTRY_MAX		2
#define HTC_ENTRY_ORDER		1
#define HTC_PAYLOAD_MAX		384

struct htc_hdr {
	// common header
	u8 chip;
	u8 band;
	//for coding's convenience, len including all fields
	u16 len;
	u32 msg_type;		//htc_msg_type
	// cookie: opaque to the remote
	u32 cookie;
} __packed;

struct htc_msg {
	// common header
	struct htc_hdr hdr;
	u8 data[HTC_PAYLOAD_MAX];
} __packed;

struct htc_shared_memory {
	u32 order;
	u32 rptr;
	u32 wptr;
	struct htc_msg entries[HTC_ENTRY_MAX];
} __packed;

struct htc_queue {
	void __iomem *base;
	u32 order;
	u32 rptr, wptr;
};

struct queue_context {
	int init;
	struct htc_queue h2t;
	struct htc_queue t2h;
	//rx htc message list
	struct sk_buff_head list;
	//work for processing rx htc message
	struct work_struct work;
	struct ieee80211_hw_priv	*dev;
	spinlock_t	lock;
} queue_context;

/***************** Detail request data structure defination ******************/

// add and del device request information
struct htc_req_add_dev {
	int devid;
	int band;
	int dbdc;
} __packed;

// add and del virtual interface request information
struct htc_req_add_vif {
	u8 addr[ETH_ALEN];
	u16 mode;		//ap or sta
	u32 vif;		//vif id
} __packed;

#ifdef SCM80211_WOW
// enable/disable wow
// typedef struct scm80211_wow_context htc_req_enable_wow;
#endif

struct scm80211_htc_ops {
	struct sk_buff *(*htc_msg_alloc)(struct ieee80211_hw_priv *dev,
					 const void *data, int data_len);
	void (*htc_msg_free)(struct ieee80211_hw_priv *dev, struct sk_buff *skb);
	int (*htc_msg_recv)(struct ieee80211_hw_priv *dev, struct sk_buff *skb);
	int (*htc_msg_send)(struct ieee80211_hw_priv *dev, struct sk_buff *skb, int type);
	int (*htc_msg_send_recv)(struct ieee80211_hw_priv *dev, struct sk_buff *skb, int type);
	int (*htc_enqueue)(struct ieee80211_hw *hw, struct sk_buff *skb);
	int (*htc_dequeue)(struct ieee80211_hw *hw);
	void (*notify_remote)(struct ieee80211_hw *hw);
};

struct scm80211_htc_ctx {
	struct queue_context htc_queues[SCM80211_NUM_DEV];
	struct scm80211_htc_ops *htc_ops;
	struct pci_dev *pdev;
	struct htc_msg_reg_ops msg_reg_ops[HTC_MSG_MAX];

#ifdef FW_LOG_UPLOAD
	unsigned int fw_log_printk_tail;
	u8 __iomem *fw_log_ram;
#endif
};

int scm80211_htc_init(struct pci_dev *pdev);
int scm80211_htc_deinit(struct pci_dev *pdev);
int scm80211_send_to_target_by_h2t(struct ieee80211_hw *hw, u32 msg_type, void *data, u16 len);
int htc_register_msg_handle(struct ieee80211_hw *hw, u32 msg_type,
			    void (*handle)(void *, void *, u16), void *priv);
int htc_unregister_msg_handle(struct ieee80211_hw *hw, u32 msg_type);

void scm80211_pcie_rc_notify_ep(struct ieee80211_hw *hw);
struct scm80211_htc_ctx *pcie_get_htc_ctx(struct pci_dev *pdev);
struct scm80211_htc_ctx *scm80211_get_htc_ctx(struct ieee80211_hw *hw);

#endif /* __SCM80211_HOST_TARGET_COMMUNICATION_H__ */
