/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifndef __SCM80211_HW_H__
#define __SCM80211_HW_H__

#include <linux/types.h>
#include <net/mac80211.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>

#define SCM80211_NUM_VIF	2
#define SCM80211_NUM_HWQ	10
#define SCM80211_NUM_HWQ_PER_VIF ((SCM80211_NUM_HWQ) / (SCM80211_NUM_VIF))
#define SCM80211_NUM_RX_RING	256
#define SCM80211_NUM_RF_IFACE	2
#define SCM80211_NUM_PHY	2
#define SCM80211_VHT_CAP_BFM_STS_CAP 7

#ifdef ERRATA_TX_MMSS_4US
/* 12000(HE-MCS11-2NSS-BW80) * 4 / (8 * 10) */
#define SCM80211_MAX_MMSS_BYTES	600
#endif

#include "regs.h"

/* Per VIF HWQ index corresponding to table 9-154 (P80211REVmd_D3.0) */
enum {
	SCM80211_HWQ_BE,
	SCM80211_HWQ_BK,
	SCM80211_HWQ_VI,
	SCM80211_HWQ_VO,
	SCM80211_HWQ_HI,	/* Not in standard. HW specific */
};

/* Shared memory */
#define SHM_TX_MPDU_DESC_BASE_LO	0x0000
#define SHM_TX_MPDU_DESC_BASE_HI	0x0004
#define SHM_TX_NUM_MPDU			0x0008
#define SHM_TX_RESULT			0x0008
#define SHM_TX_RESULT_OK		BIT(8)
#define SHM_TX_RESULT_NACK		BIT(9)
#define SHM_TX_RESULT_FAIL_MASK		0x00003c00
#define SHM_TX_RESULT_FAIL_SHIFT	10
#define SHM_TX_RESULT_TB		BIT(14)
#define SHM_TX_RESULT_UORA		BIT(15)
#define SHM_TX_RESULT_CW		BIT(16)

#if defined(SCM2020_Z1) || defined(SCM2020_A0)

#ifdef RF_BASE_START
#undef RF_BASE_START
#define RF_BASE_START			0x40118000
#endif

#define PTA_BASE_START			0x40120000
#define PTA_CONFIG				0x4
#define PTA_STAT_BT_TX			0x8
#define PTA_STAT_BT_RX			0x10
#define PTA_ENABLE				BIT(0)
#define PTA_PRIORITY			BIT(1)	/* 0: BT pri, 1: WLAN pri */
#define PTA_NO_SIM_TX			BIT(2)
#define PTA_NO_SIM_RX			BIT(3)
#define PTI_ENABLE				BIT(4)
#define CHAN_OVERLAPDECT_ENABLE	BIT(5)
#define COEX_BT_EVENT			BIT(6)
#define WLAN_PTI_MODE			BIT(7)
#define INT_ABORT_RX_IN_TX		BIT(20)

#define WLAN_CFG_BASE			0x40000000
#define WLAN_TOP_CFG			(WLAN_CFG_BASE)
#define WLAN_MCU_CFG			(WLAN_CFG_BASE + 0x100)
#define WLAN_QSPI_CFG			(WLAN_CFG_BASE + 0x180)
#define WLAN_WMP_CFG			(WLAN_CFG_BASE + 0x400)
#define WLAN_PMU_CFG			(WLAN_CFG_BASE + 0x200)

#define MCU_CLK_DIV			0x00000003
#define MCU_CLK_SET			0x00000010
#define AHB_CLK_SET			0x00003000
#define AXI_CLK_SET			0x0001C000

#define MAC_CLK_M1X_EN(n)		(0x00000001 << (n))
#define PHY_CLK_EN			0x00000004
#define PHY_PCLK_EN			0x00000008
#define RX_EN(n)			(0x00000010 << (n))
#define TX_EN(n)			(0x00000040 << (n))
#define ADC_DIV_EN(n)			(0x00000100 << (n))
#define DAC_DIV_EN(n)			(0x00000400 << (n))
#define DPD_960M_CLK_EN(n)		(0x00001000 << (n))
#define DPD_320M_CLK_EN(n)              (0x00004000 << (n))

#define QSPI_CLK_DIV			0x00000070

#define PLL_CLK_EN			0x00000001
#define PLL_EN				0x00000004

#define WCN_CFG_BASE			0x40110000
#define WCN_RFDIG_CFG			(WCN_CFG_BASE + 0x90)

#define RFDIG_CLK_EN			0x00000001

#define WCN_GPIO_BASE                   0x40111000
#define WCM_GPIO_IN_BASE		WCN_GPIO_BASE
#define WCM_GPIO_MODE_BASE		(WCN_GPIO_BASE + 0x80)
#define WCM_GPIO_DIR_BASE		(WCN_GPIO_BASE + 0x20)
#define WCM_GPIO_VAL_BASE		(WCN_GPIO_BASE + 0x70)

#endif

#ifndef SCM80211_TXOP_AMPDU_LEN
#define SCM80211_TXOP_AMPDU_LEN			0x24000	//144K
#endif
#define SCM80211_TXOP_MPDU_LEN			0x1000	//4K

enum tx_fail_reason {
	ack_timeout		 = 0,
	cts_timeout		 = 1,
	ack_invalid		 = 2,
	cts_invalid		 = 3,
	cca_busy_secondary_20	 = 5,
	cca_busy_secondary_40    = 6,
	cca_busy_secondary_20_40 = 7,
	internal_collision	 = 8,
	sifs_lost	         = 10,
	internal_win_lost	 = 11,
	internal_tx_timeout	 = 12,
	internal_underrun	 = 13,
	internal_tx_abort	 = 14,
	internal_tx_rejected	 = 15,
};

#define bits(h, l) (((h) % 32) - ((l) % 32) + 1)

enum ppdu_format {
	NON_HT = 0,
	HT_MF = 1,
	HT_GF = 2,
	VHT = 3,
	HE_SU = 4,
	HE_ER_SU = 5,
	HE_MU = 6,
	HE_TB = 7,
};

/* SCM does not support ERP_XXX */
enum modulation {
	ERP_DSSS = -3,
	ERP_CCK = -2,
	ERP_OFDM = -1,
	OFDM = 0,
	NON_HT_DUP_OFDM = 1,
};

/* Table 17-8 */
enum cbw {
	ER_RU_242 = 0,
	ER_RU_H_106 = 1,

	CBW20 = 0,
	CBW40 = 1,
	CBW80 = 2,
	CBW160 = 3,
	CBW80PLUS80 = 3,
	CBW80S20P = 4,		/* HE-CBW-PUNC80-PRI */
	CBW80S40P = 5,		/* HE-CBW-PUNC80-SEC */
	CBW160S20P = 6,		/* HE-CBW-PUNC160-PRI120 */
	CBW160S40P = 7,		/* HE-CBW-PUNC160-SEC40 */
	CBW_NOT_PRESENT = 8
};

enum dyn_bandwidth {
	DYN_BW_STATIC = 0,
	DYN_BW_DYNAMIC = 1,
};

enum ltf_type {
	HE_LTF_1x = 0,
	HE_LTF_2x = 1,
	HE_LTF_4x = 2,
	HE_LTF_TYPE,		/* HE */
};

enum ltf_mode {
	HE_single_stream_pilot = 0,
	HE_masked_HE_LTF_sequence = 1,
};

enum trigger_mode {
	TRIGGER_FRAME = 0,
	TRS = 1,
};

/**
 * In order to make the implementation easier,
 * the actual enumeration values are chosen to
 * match the related cbw.
 */
enum cca_channel {
	primary = 0,
	secondary = 1,
	secondary40 = 2,
	secondary80 = 3,
};

enum coding_type {
	BCC_CODING = 0,
	LDPC_CODING = 1,
};

enum gi_type {
	SHORT_GI = 0,		/* 0.4 us */
	LONG_GI = 1,		/* 0.8 us */

	GI_0u4s = 0,
	GI_0u8s = 1,
	GI_1u6s = 2,
	GI_3u2s = 3,
};

enum preamble_type {
	LONG_PREAMBLE = 0,
	SHORT_PREAMBLE = 1,
};

/**
 * struct txvector (size: 224/8 = 28 bytes)
 *
 * @MCS: rate index (non-HT) or mcs
 * @PREAMBLE_TYPE: DSSS/CCK preamble type
 *
 * trxvector_r1_7.xls
 */
struct txvector {
	/* word 0 */
	__le32 MCS			:bits(6,  0);
	__le32 PREAMBLE_TYPE		:bits(7,  7);
	__le32 L_LENGTH			:bits(19,  8);
	__le32 FORMAT			:bits(22, 20);
	__le32 FEC_CODING		:bits(23, 23);
	__le32 TXPWR_LEVEL_INDEX	:bits(31, 24);

	/* word 1 */
	__le32 LENGTH			:bits(47, 32);
	__le32 SPATIAL_REUSE		:bits(63, 48);

	/* word 2 */
	__le32 APEP_LENGTH		:bits(83, 64);
	__le32 CH_BANDWIDTH		:bits(86, 84);
	__le32 STBC			:bits(87, 87);
	__le32 RU_ALLOCATION		:bits(95, 88);

	/* word 3 */
	__le32 HE_SIG_A2_RESERVED	:bits(104,  96);
	__le32 TXOP_DURATION		:bits(111, 105);
	__le32 PARTIAL_AID		:bits(120, 112);
	__le32 RESERVED3_0		:bits(123, 121);
	__le32 DYN_BANDWIDTH_IN_NON_HT	:bits(124, 124);
	__le32 CH_BANDWIDTH_IN_NON_HT	:bits(126, 125);
	__le32 MIDAMBLE_PERIODICITY	:bits(127, 127);

	/* word 4 */
	__le32 RESERVED_4_0		:bits(143, 128);
	__le32 GROUP_ID			:bits(149, 144);
	__le32 GI_TYPE			:bits(151, 150);
	__le32 RESERVED_4_1		:bits(159, 152);

	/* word 5 */
	__le32 SCRAMBLER_INITIAL_VALUE	:bits(166, 160);
	__le32 LDPC_EXTRA_SYMBOL	:bits(167, 167);
	__le32 BSS_COLOR		:bits(173, 168);
	__le32 HE_LTF_TYPE		:bits(175, 174);
	__le32 TOTAL_NUM_STS		:bits(178, 176);
	__le32 NUM_HE_LTF		:bits(181, 179);
	__le32 NOMINAL_PACKET_PADDING	:bits(183, 182);
	__le32 DEFAULT_PE_DURATION	:bits(186, 184);
	__le32 STARTING_STS_NUM		:bits(189, 187);
	__le32 SMOOTHING		:bits(190, 190);
	__le32 AGGREGATION		:bits(191, 191);

	/* word 6 */
	__le32 DCM			:bits(192, 192);
	__le32 NUM_STS			:bits(193, 193);
	__le32 DOPPLER			:bits(194, 194);
	__le32 HE_LTF_MODE		:bits(195, 195);
	__le32 TX_OP_PS_NOT_ALLOWED	:bits(196, 196);
	__le32 TRIGGER_METHOD		:bits(197, 197);
	__le32 UPLINK_FLAG		:bits(198, 198);
	__le32 TRIGGER_RESPONDING	:bits(199, 199);
	__le32 HE_TB_DATA_SYMBOLS	:bits(204, 200);
	__le32 HE_TB_PRE_FEC_FACTOR	:bits(206, 205);
	__le32 FEEDBACK_STATUS		:bits(207, 207);
	__le32 HE_TB_PE_DISAMBIGUITY	:bits(208, 208);
	__le32 RESERVED6		:bits(223, 209);
} __packed;

struct rxvector {
	/* word 0 */
	__le32 MCS			:bits(6,  0);
	__le32 PREAMBLE_TYPE		:bits(7,  7);
	__le32 L_LENGTH			:bits(19,  8);
	__le32 FORMAT			:bits(22, 20);
	__le32 FEC_CODING		:bits(23, 23);
	__le32 RSSI_LEGACY		:bits(31, 24);

	/* word 1 */
	__le32 LENGTH			:bits(47, 32);
	__le32 SPATIAL_REUSE		:bits(51, 48);
	__le32 MCS_SIG_B		:bits(54, 52);
	__le32 DCM_SIG_B		:bits(55, 55);
	__le32 RSSI			:bits(63, 56);

	/* word 2 */
	__le32 PSDU_LENGTH		:bits(83, 64);
	__le32 CH_BANDWIDTH		:bits(86, 84);
	__le32 STBC			:bits(87, 87);
	__le32 TXOP_DURATION		:bits(94, 88);
	__le32 DCM			:bits(95, 95);

	/* word 3 */
	__le32 RU_ALLOCATION0		:bits(103,  96);
	__le32 RU_ALLOCATION1		:bits(111, 104);
	__le32 RU_ALLOCATION2		:bits(119, 112);
	__le32 RU_ALLOCATION3		:bits(127, 120);

	/* word 4 */
	__le32 RU_ALLOCATION4		:bits(135, 128);
	__le32 RU_ALLOCATION5		:bits(143, 136);
	__le32 RU_ALLOCATION6		:bits(151, 144);
	__le32 RU_ALLOCATION7		:bits(159, 152);

	/* word 5 */
	__le32 STA_ID_LIST		:bits(170, 160);
	__le32 PE_DURATION		:bits(173, 171);
	__le32 GI_TYPE			:bits(175, 174);
	__le32 PARTIAL_AID		:bits(184, 176);
	__le32 NUM_STS			:bits(187, 185);
	__le32 DYN_BANDWIDTH_IN_NON_HT	:bits(188, 188);
	__le32 CH_BANDWIDTH_IN_NON_HT	:bits(190, 189);
	__le32 DOPPLER			:bits(191, 191);

	/* word 6 */
	__le32 SCRAMBLER_INITIAL_VALUE	:bits(198, 192);
	__le32 UPLINK_FLAG		:bits(199, 199);
	__le32 BSS_COLOR		:bits(205, 200);
	__le32 HE_LTF_TYPE		:bits(207, 206);
	__le32 GROUP_ID			:bits(213, 208);
	__le32 USER_POSITION		:bits(215, 214);
	__le32 BEAM_CHANGE		:bits(216, 216);
	__le32 BEAMFORMED		:bits(217, 217);
	__le32 TXOP_PS_NOT_ALLOWED	:bits(218, 218);
	__le32 NON_HT_MODULATION	:bits(219, 219);
	__le32 AGGREGATION		:bits(220, 220);
	__le32 SOUNDING			:bits(221, 221);
	__le32 SMOOTHING		:bits(222, 222);
	__le32 LSIGVALID		:bits(223, 223);

	/* word 7 */
	__le32 SPATIAL_REUSE2		:bits(227, 224);
	__le32 SPATIAL_REUSE3		:bits(231, 228);
	__le32 SPATIAL_REUSE4		:bits(235, 232);
	__le32 CFO_EST			:bits(247, 236);
	__le32 RCPI			:bits(255, 248);

} __packed;

enum scm80211_rx_frag_info {
	MIDDLE_FRAME	= 0,
	LAST_FRAME	= 1,
	START_FRAME	= 2,
	SIGNAL_FRAME	= 3,
};

enum scm80211_protection {
	NO_PROTECTION = 0,
	RTS_CTS_PROTECTION,
	RTS_CTS_DYN_BANDWIDTH_PROTECTION,
	CTS_TO_SELF_PROTECTION,
};

#define SCM80211_MCU_IRQ_NUM	6
enum scm80211_mcu_irq_desc {
	POWER_DETECT	= 0,
	HTC_IND0	= 2,
	HTC_IND1	= 3,
	HTC_IND2	= 4,
	HTC_IND3	= 5,
};

enum htc_cmd {
	RX_START = 0,
	RX_STOP,
	TX_START,
	TX_STOP,
	VIF0_START,
	VIF0_STOP,
	VIF1_START,
	VIF1_STOP,
	AUTO_CAL,
	CHAN_SET,
	DEBUG_CTL,
	FW_VER,
	ABORT_TX,
	TXQ_OPER,
	IND_TX_FAIL,
	QNULL_IRQ,
	HW_PS_ENABLE,
	HW_PS_DISABLE,
	VIF_PS_ENABLE,
	VIF_PS_DISABLE,
	CONNECTED,
	DISCONNECTED,
	WOW_ENABLE,
	WOW_DISABLE,
	ENABLE_FW_LOG_UPLOAD,
	ANT_SEL,
	SET_PARAM_B,
	GET_PARAM_B,
	GET_EFUSE,
	SET_EFUSE,
	GET_TSSI,
	DOZE_ENABLE,
	ACTIVE_ENABLE,
	TWT_DOZE_ENABLE,
	HTC_MSG_MAX,
};

/**
 * struct tx_mpdu_desc - MPDU descriptor
 *
 * References:
 * - WiFi_MAC_Architecture_V0.1.17_20200312.xlsx
 *
 * @aid: AID of transmitting STA
 * @tid: TID to which the mpdu belongs to
 * @sn: the sequence number (12-bit) of the mpdu
 * @noack: indicates if the mpdu solicits no immediate response
 * @eof: indicates the value of EOF bit of the MPDU descriptor
 * @sent: indicates whether the MPDU is sent(set by hw)
 * @acked: indicates whether ACK received(set by hw)
 *
 * @mpdu_len: the length of the mpdu (including FCS)
 * @aggregate: indicates more subframes follow
 * @bypass_encryption: indicates whether HW should bypass encyrption
 * @bdnum: Number of buffer descriptors occupied by this MPDU-1
 * @protection: indicates the protection mechanism to use (0: no protection,
 *              1; RTS, 2: Bandwidth-signaling RTS, 3: CTS-to-self)
 * @timestamp_update: indicates whether to put timestamp (beacon)
 * @timestamp_offset: indicates the byte offset of timestamp
 * @no_htc: indicates that there is no HT Control field in MPDU
 *
 * @prot_duration: indicates the value of Duration/ID field of the
 *                 protection frame (RTS or CTS-to-self)
 * @mpdu_spacing: the number of MPDU delimiter to be prepended before
 *                mpdu (valid only in A-MPDU context)
 * @retry_limit: indicates number of limit for retransmission (S/W use only)
 */
struct tx_mpdu_desc {
	/* MT0,1 */
	__le64 bdaddr;

	/* MT2 */
	__le32 aid		:bits(11,  0);
	__le32 tid		:bits(15, 12);
	__le32 sn		:bits(27, 16);
	__le32 noack		:bits(28, 28);
	__le32 eof		:bits(29, 29);
	__le32 sent		:bits(30, 30);
	__le32 acked		:bits(31, 31);

	/* MT3 */
	__le32 mpdu_len		:bits(13,  0);
	__le32 aggregate	:bits(14, 14);
	__le32 bypass_encrypt	:bits(15, 15);
	__le32 bdnum		:bits(20, 16);
	__le32 protection	:bits(22, 21);
	__le32 timestamp_update	:bits(23, 23);
	__le32 timestamp_offset :bits(30, 24);
	__le32 no_htc		:bits(31, 31);

	/* MT4 */
	__le32 prot_duration	:bits(15,  0);
	__le32 mpdu_spacing	:bits(25, 16);
	__le32 retry_limit	:bits(31, 26);
	union {
		/* MT5-7 */
		__le32 private[3];
		struct {
			__le32 txtime0:bits(15, 0);
			__le32 txtime1:bits(31, 16);
			__le32 buf_len;
			__le32 unused;
		};
	};
};

/**
 * struct tx_buffer_desc - data buffer descriptor
 */
struct tx_buffer_desc {
	__le32 addr_l;
	__le32 addr_h;
	__le32 len;
} __packed;

struct rx_mpdu_hdr {
	/* Header 0 */
	__le32 mpdu_len;

	/* Header 1 */
	union {
		struct {
			__le32 phy	 :bits(0,  0);
			__le32 ok	 :bits(1,  1);
			__le32 err_fcs	 :bits(2,  2);
			__le32 err_len	 :bits(3,  3);
			__le32 err_key	 :bits(4,  4);
			__le32 err_mic	 :bits(5,  5);
			__le32 err_icv	 :bits(6,  6);
			__le32 decrypted :bits(7,  7);
			__le32 cipher	 :bits(10,  8);
			__le32 ampdu	 :bits(11, 11);
			__le32 eof	 :bits(12, 12);
			__le32 rsvd	 :bits(13, 13);
			__le32 ack	 :bits(15, 14);
			__le32 tid	 :bits(19, 16);
			__le32 tail	 :bits(20, 20);
			__le32 unused	 :bits(31, 21);
		};
		__le32 hd1;
	};

	/* Header 2 */
	__le32 timestamp;
};

#define RXH_ERR		GENMASK(6, 2)
#define RXH_ERR_FCS	BIT(2)
#define RXH_ERR_LEN	BIT(3)
#define RXH_ERR_KEY	BIT(4)
#define RXH_ERR_MIC	BIT(5)
#define RXH_ERR_ICV	BIT(6)

/*
 * Shared memory layout
 */
struct txdesc {
	/* 0x0 - 0x7 */
	__le64 addr;

	/* 0x8 */
	union {
		__le32 result;
		struct {
#if SCM80211_MAX_MPDU == 32
			__le32 num	:bits(5, 0);
			__le32 unused	:bits(7, 6);
#else
			__le32 num	:bits(6, 0);
			__le32 unused	:bits(7, 7);
#endif
			__le32 ok	:bits(8, 8);
			__le32 nack	:bits(9, 9);
			__le32 fail	:bits(13, 10);
			__le32 tb	:bits(14, 14);
			__le32 uora	:bits(15, 15);
			__le32 cw	:bits(16, 16);
			__le32 unused1	:bits(23, 17);
			__le32 err	:bits(24, 24);
			__le32 to	:bits(25, 25);
			__le32 unused2	:bits(31, 26);
		};
	};

	/* 0xc */
	__le32 htc;
};

struct rxdesc {
	/*
	 * 0x0: Buffer start address low
	 * 0x7: Buffer start address high
	 */
	__le64 addr;

	/* 0x8: Buffer size */
	__le32 blen	:bits(15,  0);
	__le32 unused0	:bits(31, 16);

	/* 0xc: Buffer info */
	union {
		struct {
			__le32 dlen	:bits(15,  0);
			__le32 frag	:bits(17, 16);
			__le32 unused1  :bits(31, 18);
		};
		__le32 word2;
	};
};

struct shared_mem {
	union {
		struct txdesc tx[SCM80211_NUM_HWQ];
		u32 txspace[0xa0 / sizeof(u32)];
	};
	struct rxdesc rx[SCM80211_NUM_RX_RING];
};

/**
 * struct reg_vif_mmss mapping to REG_VIF_MMSS
 */

struct reg_vif_mmss {
	union {
		struct {
			__le32 mmss		:bits(2,  0);
			__le32 def_pe_dur	:bits(5,  3);
			__le32 bss_color	:bits(13,  8);
			__le32 color_en		:bits(14, 14);
			__le32 color_dis	:bits(15, 15);
			__le32 trig_en		:bits(16, 16);
			__le32 trig_assoc_uora	:bits(17, 17);
			__le32 trig_uassoc_uora	:bits(18, 18);
			__le32 trs_en		:bits(19, 19);
			__le32 partial_bss_color:bits(20, 20);
			__le32 basic_nav_en	:bits(21, 21);
			__le32 hetb_data_dis	:bits(22, 22);
			__le32 bsrp_bsr_en	:bits(23, 23);
			__le32 bsrp_more_qosnull:bits(24, 24);
			__le32 resvd		:bits(31, 25);
		};
		__le32 v;
	};
};

/**
 * struct reg_sec_key_cmd - REG_SEC_KEY_CMD
 * @kcmd Indicates the key command. After command operation is done
 *	 H/W automatically sets this value to zero
 *	 0 : No command
 *	 1 : Add key by tag
 *	 2 : Delete key by tag
 *	 3 : Delete all keys
 *	 4 : Read key by tag index
 * @kid Indicates the key id
 * @ktype Indicates whether it is a GTK or PTK
 * @cipher Indicates cipher type
 * @spp_en Indicate that both the STA and its peer have their SPP A-MSDU
 *	   Capable fields equal to 1
 * @idx Used for Read key command (handle by HW)
 */

struct reg_sec_key_cmd {
	union {
		struct {
			__le32 kcmd	:bits(2,  0);
			__le32 unused0	:bits(7,  3);
			__le32 keyid	:bits(9,  8);
			__le32 gtk	:bits(10, 10);
			__le32 unused1	:bits(15, 11);
			__le32 cipher	:bits(18, 16);
			__le32 spp	:bits(19, 19);
			__le32 unused2	:bits(23, 20);
			__le32 idx	:bits(27, 24);
			__le32 unused3	:bits(31, 28);
		};

		__le32 val;
	};
};

enum key_cmd {
	KEY_CMD_DONE		= 0,
	KEY_CMD_ADD		= 1,
	KEY_CMD_DEL		= 2,
	KEY_CMD_DELALL		= 3,
	KEY_CMD_READ		= 4,
	KEY_CMD_DEL_BY_IDX	= 5,
	KEY_CMD_ADD_BY_IDX	= 6,
};

enum cipher_suite {
	CIPHER_SUITE_WEP40	= 0,
	CIPHER_SUITE_WEP104	= 1,
	CIPHER_SUITE_TKIP	= 2,
	CIPHER_SUITE_CCMP128	= 4,
	CIPHER_SUITE_CCMP256	= 5,
	CIPHER_SUITE_GCMP128	= 6,
	CIPHER_SUITE_GCMP256	= 7,
	CIPHER_SUITE_UNKNOWN	= 3,
};

struct reg_aid_cmd {
	union {
		struct {
			__le32 aidcmd		:bits(2,  0);
			__le32 rx_aid_monitor	:bits(3,  3);
			__le32 unused0		:bits(23,  4);
			__le32 tag_idx		:bits(25, 24);
			__le32 unused1		:bits(31, 26);
		};

		__le32 val;
	};
};

enum aid_cmd {
	AID_CMD_DONE		= 0,
	AID_CMD_ADD		= 1,
	AID_CMD_DEL		= 2,
	AID_CMD_DELALL		= 3,
	AID_CMD_READ		= 4,
};

enum space_type {
	MT_HST_BD_HST = 0,
	MT_HST_BD_TGT,
	MT_TGT_BD_TGT,
};

struct txop_info {
	__le32 num		:bits(6,  0);
	__le32 ampdu		:bits(7,  7);
	__le32 he		:bits(8,  8);
	__le32 prot		:bits(10,  9);
	__le32 band		:bits(11, 11);
	__le32 mmss		:bits(22, 12);
	__le32 hwq_id		:bits(26, 23);
	__le32 txop_id		:bits(27, 27);
	__le32 mu_enable	:bits(28, 28);
	__le32 vidx		:bits(29, 29);
	__le32 probe		:bits(30, 30);
	__le32 rsvd0		:bits(31, 31);

	__le32 limit		:bits(15,  0);
	__le32 ecwmin		:bits(19, 16);
	__le32 ecwmax		:bits(23, 20);
	__le32 aifsn		:bits(27, 24);
	__le32 rate_idx		:bits(31, 28);

	__le32 rts_time		:bits(15,  0);
	__le32 ack_time		:bits(31, 16);

	__le32 muedca_timer	:bits(15,  0);
	__le32 ssn		:bits(28, 16);
	__le32 rsvd1		:bits(31, 29);  /* window size == 2^(wz_order + 3)*/

	__le32 rts_time_th	:bits(15,  0);
	__le32 txq_id		:bits(22, 16);
	__le32 twt_state	:bits(25, 23);
	__le32 twt_trigger	:bits(26, 26);
	__le32 in_window	:bits(27, 27);
	__le32 rsvd2		:bits(31, 28);

	__le32 wz	        :bits(7,  0);
	__le32 rsvd3		:bits(31,  8);

	struct ieee80211_tx_rate rates[IEEE80211_TX_MAX_RATES];
} __packed;

struct scm80211_bss_conf {
	u16	aid;
	u8	associated;

	u8	he_support;
	u8	he_bss_color;
	u8	bssid[ETH_ALEN];
	u8	transmitter_bssid[ETH_ALEN];
	u8	band;

	u8	ps;
	u32	sync_device_ts;
	u8	sync_dtim_count;
	u16	beacon_int;
	u8	dtim_period;
	u16	listen_interval;

} __packed;

struct scm80211_vif {
	u8				type;
	u8				idx;
	struct scm80211_bss_conf	bss_conf;
} __packed;

enum sent_info {
	SENT	= 0,
	ABORT,
	OUT_OF_WINDOW,
	UNDER_PS,
};

struct scm80211_tx_status_info {
	__le32 hwq_id		:bits(3,  0);
	__le32 txop_id		:bits(4,  4);
	__le32 info		:bits(7,  5);
	__le32 tb		:bits(8,  8);
	__le32 resv1		:bits(10, 9);
	__le32 reduce		:bits(11, 11);
	__le32 rsvd2		:bits(31, 12);

	__le32 tb_psdu_len;
	__le64 report;
	struct ieee80211_tx_rate rates[IEEE80211_TX_MAX_RATES];
} __packed;


struct htc_content {
	__le32 hw_id		:bits(0,  0);  /* hw for dbdc */
	__le32 vif_id		:bits(2,  1);	/* vif idx */
	__le32 resv		:bits(7,  3);  /* reserved */
	__le32 event		:bits(15,  8);	/* htc len */
	__le32 len		:bits(31, 16);  /* content len */
	u8 content[60];
} __packed;


struct scm80211_rx_info {
	dma_addr_t dma_addr;
	int index;
	struct rxdesc rxd;
};

enum scm80211_rx_result {
	RX_CONTINUE = 0,
	RX_DROP = 1,
	RX_MONITOR_ONLY = 2,
	RX_HANDLED = 3,
	RX_AGAIN = 4,
};

struct scm80211_rx_data {
	struct ieee80211_hw *hw;
	unsigned long interfaces;
	struct sk_buff *skb;
	struct ieee80211_rx_status *rxs;
	struct rxvector *rv;
	struct rx_mpdu_hdr *rh;
	struct ieee80211_hdr *hdr;
	struct ieee80211_sta *sta;
	struct ieee80211_vif *vif;
	struct scm80211_rx_info info;
	enum scm80211_rx_result result;
	struct sk_buff_head queue;
};

enum reg_type {
	MAC_REG,
	PHY_REG,
	RFI_REG,
	RAW_REG,
};

#define reg_readl(x, hw, reg) scm80211_##x##_readl(hw, reg)
#define reg_writel(x, hw, val, reg) scm80211_##x##_writel(hw, (val), (reg))



extern struct txvector default_txvector;

#define ppdu_is_he(v)		((v)->FORMAT >= HE_SU && (v)->FORMAT <= HE_TB)
#define ppdu_is_he_su(v)	((v)->FORMAT = HE_SU)
#define ppdu_is_he_er_su(v)	((v)->FORMAT == HE_ER_SU)
#define ppdu_is_he_mu(v)	((v)->FORMAT == HE_MU)

#define ppdu_is_vht(v)		((v)->FORMAT == VHT)
#define ppdu_is_ht(v)		((v)->FORMAT == HT_MF || (v)->FORMAT == HT_GF)
#define ppdu_is_non_ht(v)	((v)->FORMAT == NON_HT)

int scm2020_chip_init(struct ieee80211_hw *hw);

int scm80211_instantiate_hwsim(struct platform_device *pdev);
int scm80211_hwsim_request_threaded_irq(struct platform_device *pdev,
					irq_handler_t handler, void *data);

#endif /* __SCM80211_HW_H__ */
