// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#include <linux/platform_device.h>
#include "core.h"
#include "rf.h"
#include <linux/bitfield.h>

#define func_fmt(str)           "[HWSIM] "str
#define func_trace              BUS

#include "trace.h"
#include "debug.h"

struct hwsim_bus_data {
	void		*hws[SCM2020_INTF_MAX];
	atomic_t	ref;
	unsigned long	bitmap[256 / BITS_PER_LONG];
	size_t		page_size;
	unsigned int	page_shift;
	struct shared_mem __iomem *mem;
	void __iomem	*bp_base;
	void __iomem	*mac_base;
	void __iomem	*phy_base;
	phys_addr_t	last_bp_base;
	u32		intf_num;
	struct rb_root	root;
	spinlock_t	lock;
	struct regmap   *regmap;
	int		num_atu;

};

#define HWSIM_REG_PHY_BASE	0x01000000
#define HWSIM_REG_RF_BASE	(HWSIM_REG_PHY_BASE + 0x10000)
void scm80211_hwsim_mac_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	int err;

	if (offset >= MAC_SHM_OFFSET && offset < MAC_REG_OFFSET) {
		void *ptr = (void *)hwsim->mem;

		offset -= MAC_SHM_OFFSET;
		memcpy(ptr + offset, &val, sizeof(val));
	} else if (offset >= MAC_REG_OFFSET && offset < MAC_SYS_OFFSET) {
		offset -= MAC_REG_OFFSET;
		err = regmap_write(hwsim->regmap, offset, val);
		if (err) {
			scm80211_err(hw, "write (reg=%08x, val=%08x) failed (%d)\n",
				     offset, val, err);
		}
	}
}

unsigned int scm80211_hwsim_mac_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	unsigned int val = 0;
	int err;

	if (offset >= MAC_SHM_OFFSET && offset < MAC_REG_OFFSET) {
		void *ptr = (void *)hwsim->mem;

		offset -= MAC_SHM_OFFSET;
		memcpy(&val, ptr + offset, sizeof(val));
	} else if (offset >= MAC_REG_OFFSET && offset < MAC_SYS_OFFSET) {
		offset -= MAC_REG_OFFSET;
		err = regmap_read(hwsim->regmap, offset, &val);
		if (err)
			scm80211_err(hw, "read (reg=%08x) failed (%d)\n", offset, err);
	}
	return val;
}

/**
 * hwsim PHY and RF
 *
 * PHY register base is HWSIM_REG_PHY_BASE above MAC base address.
 * RF register base is 0x10000 above PHY registers.
 *
 */
void scm80211_hwsim_phy_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int reg)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	int err;

	//scm80211_trace(hw, "phy write: reg=%08x, val=%08x\n", reg, val);

	err = regmap_write(hwsim->regmap, HWSIM_REG_PHY_BASE + reg, val);
	if (err)
		scm80211_err(hw, "phy write (reg=%08x, val=%08x) failed (%d)\n", reg, val, err);
}

unsigned int scm80211_hwsim_phy_readl(struct ieee80211_hw *hw, unsigned int reg)
{
	/* FIXME */
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	unsigned int val = 0;
	int err;

	err = regmap_read(hwsim->regmap, HWSIM_REG_PHY_BASE + reg, &val);
	if (err)
		scm80211_err(hw, "read (reg=%08x) failed (%d)\n", reg, err);
	return val;
}

void scm80211_hwsim_rf_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int reg)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	int err;

	err = regmap_write(hwsim->regmap, HWSIM_REG_RF_BASE + reg, val);
	if (err)
		scm80211_err(hw, "rf write (reg=%08x, val=%08x) failed (%d)\n", reg, val, err);
}

unsigned int scm80211_hwsim_rf_readl(struct ieee80211_hw *hw, unsigned int reg)
{
	/* FIXME */
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	unsigned int val = 0;
	int err;

	err = regmap_read(hwsim->regmap, HWSIM_REG_RF_BASE + reg, &val);
	if (err)
		scm80211_err(hw, "read (reg=%08x) failed (%d)\n", reg, err);
	return val;
}

void scm80211_hwsim_raw_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int addr)
{
	/* */
}

u32 scm80211_hwsim_raw_readl(struct ieee80211_hw *hw, unsigned int addr)
{
	return 0;
}

/**
 * hwsim_addr_mapping - DMA to hwsim address mapping
 *
 * This is to test and evaluate PCI-to-local address mapping
 */
struct hwsim_mapping {
	struct rb_node rb;
	dma_addr_t addr;
	int atu;
	atomic_t refcount;
};

/**
 * pci_find_mapping() - find a mapping for @addr and @size
 * @addr: pci address
 * @size: pci address range
 *
 * [@addr, @addr+@size-1] should not span over multiple mappings
 */
struct hwsim_mapping *hwsim_find_mapping(struct hwsim_bus_data *hwsim, dma_addr_t addr, size_t size)
{
	struct rb_node *node = hwsim->root.rb_node;

	while (node) {
		struct hwsim_mapping *mapping;

		mapping = rb_entry(node, struct hwsim_mapping, rb);
		if (addr < mapping->addr)
			node = node->rb_left;
		else if (addr > mapping->addr)
			node = node->rb_right;
		else
			return mapping;
	}
	return NULL;
}

void hwsim_dump_mapping(struct ieee80211_hw *hw)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	struct hwsim_mapping *m;
	struct rb_node *node = rb_first(&hwsim->root);

	for (node = rb_first(&hwsim->root); node; node = rb_next(node)) {
		m = rb_entry(node, struct hwsim_mapping, rb);

		// scm80211_trace(hw, "- [%llx-%llx]: atu=%d, refs=%d\n",
		//	       m->addr, m->addr + hwsim->page_size - 1,
		//	       m->atu, atomic_read(&m->refcount));
	}
}

/**
 * hwsim_alloc_mapping - allocates a pci-to-local address mapping entry
 */
static struct hwsim_mapping *hwsim_alloc_mapping(dma_addr_t addr, size_t size)
{
	struct hwsim_mapping *mapping;

	mapping = kzalloc(sizeof(*mapping), GFP_ATOMIC);
	if (!mapping)
		return NULL;

	mapping->addr = addr;
	mapping->atu = -1;
	atomic_set(&mapping->refcount, 0);

	return mapping;
}

static unsigned long hwsim_pfn(struct hwsim_bus_data *hwsim, dma_addr_t addr)
{
	return addr >> (hwsim->page_shift);
}

/**
 * hwsim_map_addr() - map a pci address to a epc local address
 *
 * @ep: remote ep controller
 * @addr: pci address to map
 * @len: address range to map
 *
 * Returns: epc local address or 0 if mapping cannot be established.
 *
 * What if the range [addr, addr + len - 1] does not fit into a single pci page?
 */
phys_addr_t scm80211_hwsim_remap(struct ieee80211_hw *hw, dma_addr_t addr, size_t len)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	struct rb_node **node, *parent = NULL;
	struct hwsim_mapping *mapping;
	dma_addr_t start, end, offset;
	int atu = -1;
	unsigned long flags;

	end = addr + len - 1;
	start = addr & ~(hwsim->page_size - 1);
	offset = addr & (hwsim->page_size - 1);

	/*
	 * NB: If a scatterlist element cannot be mapped into a single
	 * target address range, it needs to be split.
	 *
	 * To the best of our knowledge, the maximum number of contiguosly
	 * allocatable pages is 1024 (4MB), which is smaller than the
	 * address mapping size (32MB, or 64MB). As long as we are dealing
	 * with dynamic memory areas allocated via kmalloc() or get_free_pages(),
	 * they fit into a single mapping.
	 */
	if (unlikely(hwsim_pfn(hwsim, addr) != hwsim_pfn(hwsim, end))) {
		scm80211_warn(hw, "dma range [%llx, %llx] crosses page boundary\n", addr, end);
		return 0;
	}

	spin_lock_irqsave(&hwsim->lock, flags);
	node = &hwsim->root.rb_node;
	while (*node) {
		mapping = rb_entry(*node, struct hwsim_mapping, rb);
		parent = *node;

		if (start < mapping->addr)
			node = &((*node)->rb_left);
		else if (start > mapping->addr)
			node = &((*node)->rb_right);
		else
			goto done;
	}

	/* No mapping exists */

	/* Allocate an ATU */
	atu = find_first_zero_bit(hwsim->bitmap, hwsim->num_atu);
	if (atu == hwsim->num_atu) {
		scm80211_warn(hw, "no outbound ATU available\n");
		hwsim_dump_mapping(hw);
		spin_unlock_irqrestore(&hwsim->lock, flags);
		return 0;
	}

	/* Create a new mapping */
	mapping = hwsim_alloc_mapping(start, hwsim->page_size);
	if (!mapping) {
		scm80211_warn(hw, "failed to alloc mapping\n");
		spin_unlock_irqrestore(&hwsim->lock, flags);
		return 0;
	}
	mapping->atu = atu;
	rb_link_node(&mapping->rb, parent, node);
	rb_insert_color(&mapping->rb, &hwsim->root);
	set_bit(atu, hwsim->bitmap);

done:
	atomic_inc(&mapping->refcount);
	// scm80211_trace(hw, "remap: dma=%llx+%llx (%ld), atu=%d (refs=%d)\n",
	//	       start, offset, len,
	//	       mapping->atu, atomic_read(&mapping->refcount));
	hwsim_dump_mapping(hw);
	spin_unlock_irqrestore(&hwsim->lock, flags);
	return cpu_to_le64(addr);
}

void scm80211_hwsim_unmap(struct ieee80211_hw *hw, dma_addr_t addr, size_t len)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	struct hwsim_mapping *mapping = NULL;
	dma_addr_t start, end, offset;
	unsigned long flags;

	end = addr + len - 1;
	start = addr & ~(hwsim->page_size - 1);
	offset = addr & (hwsim->page_size - 1);

	spin_lock_irqsave(&hwsim->lock, flags);
	mapping = hwsim_find_mapping(hwsim, start, len);
	if (!mapping) {
		scm80211_warn(hw, "failed to find the mapping for [%llx-%llx]\n", addr, end);
		spin_unlock_irqrestore(&hwsim->lock, flags);
		return;
	}

	scm80211_trace(hw, "unmap: dma=%llx+%llx (%ld), atu=%d (refs=%d)\n",
		       start, offset, len, mapping->atu, atomic_read(&mapping->refcount));

	if (atomic_dec_return(&mapping->refcount) > 0) {
		spin_unlock_irqrestore(&hwsim->lock, flags);
		return;
	}

	clear_bit(mapping->atu, hwsim->bitmap);
	rb_erase(&mapping->rb, &hwsim->root);
	kfree(mapping);

	hwsim_dump_mapping(hw);
	spin_unlock_irqrestore(&hwsim->lock, flags);
}

static dma_addr_t scm80211_hwsim_inverse_map(struct ieee80211_hw *hw, phys_addr_t addr)
{
	return (dma_addr_t)le64_to_cpu(addr);
}

static int scm80211_hwsim_hwq_alloc_dma(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	size_t size = sizeof(struct tx_dma_desc);
	int m, n;

	for (n = 0; n < SCM80211_MAX_TXOP; n++) {
		struct txop *txop = &hwq->ring.pool[n];

		txop->dma[0].desc = dma_alloc_coherent(dev, size, &txop->dma[0].paddr, GFP_KERNEL);
		if (!txop->dma[0].desc)
			return -ENOMEM;
		txop->dma[0].raddr = scm80211_remap(hw, txop->dma[0].paddr, size);

		txop->dma[1].desc = dma_alloc_coherent(dev, size, &txop->dma[1].paddr, GFP_KERNEL);
		if (!txop->dma[1].desc)
			return -ENOMEM;
		txop->dma[1].raddr = scm80211_remap(hw, txop->dma[1].paddr, size);

		/* Program >mdesc[m].baddr with the remote target buffer address */
		for (m = 0; m < SCM80211_MAX_MPDU; m++) {
			off_t offset = offsetof(struct tx_dma_desc, mpdu[m].bdesc);

			txop->dma[0].desc->mdesc[m].bdaddr = txop->dma[0].raddr + offset;
			txop->dma[1].desc->mdesc[m].bdaddr = txop->dma[1].raddr + offset;
		}

		txop->raddr = txop->dma[0].raddr + offsetof(struct tx_dma_desc, mdesc) /* = 0 */;
	}

	return 0;
}

static void scm80211_hwsim_hwq_free_dma(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	size_t size = sizeof(struct tx_dma_desc);
	int m, n;

	for (m = 0; m < SCM80211_MAX_TXOP; m++) {
		struct txop *txop = &hwq->ring.pool[m];

		for (n = 0; n < ARRAY_SIZE(txop->dma); n++) {
			scm80211_unmap(hw, txop->dma[n].paddr, size);
			dma_free_coherent(dev, size, txop->dma[n].desc, txop->dma[n].paddr);

			txop->dma[n].desc = NULL;
			txop->dma[n].paddr = DMA_MAPPING_ERROR;
			txop->dma[n].raddr = 0;
		}
	}
}

static void scm80211_update_rx_ring_ptrs(struct ieee80211_hw *hw, bool init)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 v, rptr, bptr, wptr;

	/* Take least significant (priv->rx.order + 1) bits */
	bptr = ring_ptr(&priv->rxring, priv->rxring.bptr);
	rptr = ring_ptr(&priv->rxring, priv->rxring.rptr);
	wptr = ring_ptr(&priv->rxring, priv->rxring.wptr);

	scm80211_trace(hw, "rx ptrs: bptr=0x%04x, rptr=0x%04x, wtpr=0x%04x\n", bptr, rptr, wptr);

	v = FIELD_PREP(RX_BUF_PTR0_READ_PTR, rptr);
	v |= FIELD_PREP(RX_BUF_PTR0_READY_PTR, bptr);
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_PTR0);
}

static int scm80211_hwsim_set_channel(struct ieee80211_hw *hw, struct scm80211_phy *phy,
				      struct phy_vector *pv)
{
	phy->radio->ops->set_channel(phy->radio, pv->CENTER_FREQUENCY, pv->BANDWIDTH);
	scm2020_phy_set_channel(phy, 0, pv);
	return 0;
}

static int scm80211_hwsim_rx_poll(struct napi_struct *napi, int budget)
{
	struct ieee80211_hw_priv *priv =
		container_of(napi, struct ieee80211_hw_priv, napi.context);
	struct ieee80211_hw *hw = priv->hw;

	return scm80211_rx_poll(hw, budget);
}

static int scm80211_hwsim_start_rx(struct ieee80211_hw *hw)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_rxdesc *rxd;
	u32 v;
	int i;

	netif_napi_add(&priv->napi.netdev, &priv->napi.context, scm80211_hwsim_rx_poll, 64);
	napi_enable(&priv->napi.context);
	priv->napi.enabled = true;

	scm80211_mac_writel(hw, 0, MAC_REG_RX_BUF_CFG);

	scm80211_update_rx_ring_ptrs(hw, true);
#ifdef SCM80211_RX_PAGE_ALLOC
	/* alloc 256 * buffer page size */
	priv->rxring.pages = dev_alloc_pages(SCM80211_RX_PAGE_ORDER + priv->rxring.order);
	if (!priv->rxring.pages) {
		scm80211_err(hw, "Can not alloc this much of pages\n");
		return -ENOMEM;
	}
#endif
	for (i = 0; i < ring_size(&priv->rxring); i++) {
		rxd = &priv->rxring.rxd[i];
		memset(rxd, 0, sizeof(*rxd));
		rxd->config = &hwsim->mem->rx[i];
		rxd->id = i;
#ifdef SCM80211_RX_PAGE_ALLOC
		rxd->page = priv->rxring.pages + (1 << SCM80211_RX_PAGE_ORDER) * i;
		page_ref_add(rxd->page, USHRT_MAX - 1);
		rxd->dma_addr = 0;
#endif
	}
	priv->rxring.bptr = 0;
	priv->rxring.rptr = 0;

	scm80211_refill_rxring(hw);
	scm80211_refill_rxpool(hw);
	v = FIELD_PREP(RX_BUF_CFG_BUF_NUM, priv->rxring.order);
	if (version(scm80211_mac_version(hw)) >= 0x16)
		v |= FIELD_PREP(RX_BUF_CFG_DMA_EN, 1);
	v |= FIELD_PREP(RX_BUF_CFG_EN, 1);
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_CFG);

#ifdef SCM80211_RX_STATISTICS_DEBUG
	/* For rx overrun debuging */
	priv->rx_overrun = 0;
#endif
	scm80211_enable_rx_irq(hw);
	return 0;
}

static void scm80211_hwsim_stop_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct scm80211_rxdesc *rxd;
#ifndef SCM80211_RX_PAGE_ALLOC
	struct scm80211_rx_info *info;
#endif
	int i, retries = 10;
	u32 v;

	v = scm80211_mac_readl(hw, MAC_REG_RX_BUF_CFG);
	v &= ~RX_BUF_CFG_EN;
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_CFG);

	/* Wait for RXE to park */
	v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
	while (retries-- > 0) {
		mdelay(1);
		v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
		if (FIELD_GET(MAC_STATUS_DMA_BUSY, v) == 0)
			break;
	}

	scm80211_disable_rx_irq(hw);
	/* Cleanup rx ring */
	for (i = 0; i < ring_size(&priv->rxring); i++) {
		rxd = &priv->rxring.rxd[i];
#ifdef SCM80211_RX_PAGE_ALLOC
		if (rxd->dma_addr) {
			scm80211_unmap(hw, rxd->dma_addr, rx_buflen);
			dma_unmap_page(dev, rxd->dma_addr, rx_buflen, DMA_FROM_DEVICE);
		}
		page_ref_sub(rxd->page, USHRT_MAX - 1);
#else
		if (!rxd->skb)
			continue;

		info = SCM80211_SKB_RXCB(rxd->skb);
#ifdef RDMA
		scm80211_unmap(hw, info->dma_addr, rx_buflen);
#endif
		dma_unmap_single(dev, info->dma_addr, rx_buflen, DMA_FROM_DEVICE);

		kfree_skb(rxd->skb);
		rxd->skb = NULL;
#endif /* SCM80211_RX_PAGE_ALLOC */
		memset_io(rxd->config, 0, sizeof(*rxd->config));
	}

#ifdef SCM80211_RX_PAGE_ALLOC
	__free_pages(priv->rxring.pages, SCM80211_RX_PAGE_ORDER + priv->rxring.order);
#endif
}

#ifdef SCM80211_RX_PAGE_ALLOC
static int scm80211_rxd_attach_page(struct ieee80211_hw *hw, struct scm80211_rxdesc *rxd)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	dma_addr_t dma_addr;

	dma_addr = dma_map_page(dev, rxd->page, IEEE80211_RX_HEADROOM, rx_buflen, DMA_FROM_DEVICE);

	if (dma_mapping_error(dev, dma_addr)) {
		scm80211_warn(hw, "rxd %d: failed to map page\n", rxd_index(hw, rxd));
		return -ENOMEM;
	}

	do {
		struct rxdesc rd;
		phys_addr_t addr = scm80211_remap(hw, dma_addr, rx_buflen);

		if (!addr) {
			dma_unmap_page(dev, dma_addr, rx_buflen, DMA_FROM_DEVICE);
			scm80211_warn(hw, "rxd %d: rx addr mapping fail\n", rxd_index(hw, rxd));
			return -ENOMEM;
		}
		memset(&rd, 0, sizeof(rd));
		rd.addr = addr;
		rd.blen = cpu_to_le16(rx_buflen);
		scm80211_trace(hw, " - %s rxd[%02x] page addr %p rxd addr %llx, blen %u\n",
			       __func__, rxd->id, rxd->page, rd.addr, rd.blen);
		memcpy_toio(rxd->config, &rd, sizeof(rd));

	} while (0);

	rxd->dma_addr = dma_addr;
	return 0;
}

#else

static int scm80211_rxd_attach_skb(struct ieee80211_hw *hw, struct scm80211_rxdesc *rxd)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	struct sk_buff *skb;
	struct scm80211_rx_info *info;
	dma_addr_t dma_addr;
	int err = 0;

	skb = scm80211_alloc_rx_skb(hw);
	if (!skb) {
		scm80211_warn(hw, "rxd %d: failed to alloc skb\n", rxd_index(hw, rxd));
		return -ENOMEM;
	}

	skb_reserve(skb, IEEE80211_RX_HEADROOM);

	dma_addr = dma_map_single(dev, skb->data, rx_buflen, DMA_FROM_DEVICE);
	if (dma_mapping_error(dev, dma_addr)) {
		scm80211_warn(hw, "rxd %d: failed to map rx skb\n", rxd_index(hw, rxd));
		err = -ENOMEM;
		goto free_skb;
	}

	info = SCM80211_SKB_RXCB(skb);
	info->dma_addr = dma_addr;
	info->index = rxd->id;

#ifdef RDMA
	do {
		struct rxdesc rd;
		phys_addr_t addr = scm80211_remap(hw, dma_addr, rx_buflen);

		if (!addr) {
			scm80211_warn(hw, "failed to translate rx buffer\n");
			err = -ENOMEM;
			goto dma_unmap;
		}

		memset(&rd, 0, sizeof(rd));
		info->rxd.addr = addr;
		rd.addr = addr;
		info->rxd.blen = cpu_to_le16(rx_buflen);
		rd.blen = cpu_to_le16(rx_buflen);
		memcpy_toio(rxd->config, &rd, sizeof(rd));
	} while (0);
#else
	rxd->config->addr = cpu_to_le64(dma_addr);
	rxd->config->blen = cpu_to_le16(rx_buflen);
	info->rxd.addr = cpu_to_le64(dma_addr);
	info->rxd.blen = cpu_to_le16(rx_buflen);
#endif
	rxd->skb = skb;

	return 0;

dma_unmap:
	dma_unmap_single(dev, info->dma_addr, rx_buflen, DMA_FROM_DEVICE);
free_skb:
	kfree_skb(skb);
	return err;
}
#endif /* SCM80211_RX_PAGE_ALLOC */

static int scm80211_hwsim_refill_rxring(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_rxdesc *rxd;
	u16 wptr, rptr;
	unsigned long flags;

	spin_lock_irqsave(&priv->rxring.lock, flags);
	wptr = priv->rxring.wptr;
	rptr = priv->rxring.rptr;

	scm80211_trace(hw,
		       "rx fill: bptr=0x%04lx, rptr=0x%04lx, wptr=0x%04lx\n",
		       ring_ptr(&priv->rxring, priv->rxring.bptr),
		       ring_ptr(&priv->rxring, rptr), ring_ptr(&priv->rxring, wptr));

	/* Advance one revolution */
	wptr = wptr + ring_size(&priv->rxring);
	rptr = rptr + ring_size(&priv->rxring);

	while (!ring_ptr_equal(priv, priv->rxring.bptr, rptr)) {
		rxd = &priv->rxring.rxd[ring_index(&priv->rxring, priv->rxring.bptr)];
#ifdef SCM80211_RX_PAGE_ALLOC
		if (scm80211_rxd_attach_page(hw, rxd) < 0)
			return -1;
#else
		if (scm80211_rxd_attach_skb(hw, rxd) < 0)
			return -1;
#endif
		priv->rxring.bptr++;
	}
	scm80211_update_rx_ring_ptrs(hw, false);
	spin_unlock_irqrestore(&priv->rxring.lock, flags);
	return 0;
}

static int scm80211_hwsim_rx_start(struct ieee80211_hw *hw, struct sk_buff *skb,
				   struct scm80211_rx_data *rx)
{
	struct scm80211_rx_info *info = SCM80211_SKB_RXCB(skb);

	info->rxd.dlen = __le16_to_cpu(info->rxd.dlen);
	info->rxd.blen = __le16_to_cpu(info->rxd.blen);

	/*
	 * This is to detect and discard the case where rxd.dlen > rxd.blen and
	 * therefore causes skb over panic
	 */
	if (unlikely(info->rxd.dlen > info->rxd.blen)) {
		scm80211_warn(hw, "rxd[%02x]: invalid dlen %d", info->index, info->rxd.dlen);
		info->rxd.dlen = sizeof(struct rxvector) + sizeof(struct rx_mpdu_hdr);
		return -EINVAL;
	}

	skb_put(skb, info->rxd.dlen);
	return 0;
}

static int scm80211_hwsim_start_tx(struct ieee80211_hw *hw)
{
	struct hwsim_bus_data *hwsim = scm80211_get_bus_data(hw);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		hwq = &priv->hwq[i];
		hwq->txdesc = hwsim->mem->tx + i;
	}
	return 0;
}

static void scm80211_hwsim_stop_tx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		struct txop *txop;
		enum scm80211_hwq_state qstate;
		int j, retries;
		u32 cmd;

		hwq = &hwp->hwq[i];
		for (j = 0; j < ARRAY_SIZE(hwq->ring.pool); j++) {
			txop = &hwq->ring.pool[j];
			if (txop->state == TXOP_STATE_ISSUED)
				break;
			txop = NULL;
		}
		if (!txop)
			continue;

		retries = txop->txtime / 10 + 1;
		scm80211_mac_writel(hw, 0, MAC_REG_TX_QUEN_CMD(i));
		while (retries-- > 0) {
			cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
			if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd) == 0)
				break;
			udelay(10);
		}
		if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
			scm80211_info(hw, "failed to clean hwq#%d: cmd=%08x\n", hwq->id, cmd);

		qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
		scm80211_trace(hw, "%s(hwq#%d) finished: cmd=%08x state=%d\n",
			       __func__, hwq->id, cmd, qstate);
		if (qstate > HWQ_STATE_DELAY) {
			retries = 20;
			while (retries-- > 0) {
				udelay(10);
				qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
				if (qstate <= HWQ_STATE_DELAY)
					break;
			}
		}
	}
}

static int scm80211_hwsim_abort_backoff(struct ieee80211_hw *hw, struct txop *txop, bool discard)
{
	struct scm80211_hwq *hwq = txop->hwq;
	enum scm80211_hwq_state qstate;
	u32 cmd;
	int retries = 10;

	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));

	scm80211_trace(hw, "%s(hwq#%d.%d) cmd=%08x state=%d\n",
		       __func__, hwq->id, txop->id, cmd, qstate);

	list_del_init(&txop->list);
	if (unlikely(txop->state != TXOP_STATE_ISSUED)) {
		scm80211_warn(hw, "- hwq#%d.%d not issued (state=%s)\n",
			      hwq->id, txop->id, txop_state_str[txop->state]);
		return 0;
	}

	if (!FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
		scm80211_warn(hw, "- hwq#%d.%d issue and sent out success?\n", hwq->id, txop->id);

	/* (1) Disable TX_DONE interrupt for this queue */
	scm80211_disable_tx_irq(hw, BIT(hwq->id));

	/* (2) Issue CMD_CANCEL blindly */
	scm80211_mac_writel(hw, 0, MAC_REG_TX_QUEN_CMD(hwq->id));

	/* (3) Loop until the MAC_REG_TX_QUEN_CMD[3:0] is read as zero. */
	while (retries-- > 0) {
		cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		if (!FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
			break;
		udelay(10);
	}

	/* (4) Check again. If hardware queue is in transmission, wait until done */
	retries = (txop->txtime + txop->ack_time) / 10 + 1;
	qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
	while (qstate == HWQ_STATE_IN_TRANSMISSION && retries-- > 0) {
		udelay(10);
		qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
	}
#ifdef ERRATA_HWQ_STUCK
	if (qstate == HWQ_STATE_IN_TRANSMISSION) {
		/* TX hung happened, force clear the queue */
		scm80211_mac_writel(hwq->hw, BIT(hwq->id), MAC_REG_DEV_FORCE_QUE_CLR);
	}
#endif

	/* (5) By this time, the hardware queue should have been parked. double check it */
	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	if (unlikely(FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd)))
		scm80211_warn(hw, "- queue still not parked: cmd=%08x state=%d\n", cmd, qstate);

	/*
	 * Because we masked TX_DONE interrupt before cancelling backoff,
	 * interrupt request is not raised even if the transmission was
	 * not cancelled and completed.
	 *
	 * We complete txop here on behalf of scm80211_tx_irq(). if TX_DONE
	 * interrupt is raised later after the interrupt is unmasked below,
	 * scm80211_tx_irq() will simply does nothing.
	 *
	 * If the txop is cancelled, mpdu descriptors will have their sent
	 * field set to 0. Such frames are discarded anyhow.
	 */
	scm80211_complete_txop(txop, discard);

	/* Re-enable TX_DONE interrupt */
	scm80211_enable_tx_irq(hw, BIT(hwq->id));

	return -qstate;
}

static int scm80211_hwsim_cleanup_tx(struct ieee80211_hw *hw, struct txop *txop, int idx)
{
	struct scm80211_tx_info *xti = &txop->xti[idx];
	struct device *dev = wiphy_dev(hw->wiphy);
	struct scatterlist *sg;
	int n;

	for_each_sg(xti->sgt.sgl, sg, xti->sgt.nents, n)
		scm80211_unmap(hw, sg_dma_address(sg), sg_dma_len(sg));

	dma_unmap_sg(dev, xti->sgt.sgl, xti->sgt.nents, DMA_TO_DEVICE);

	return 0;
}

static int scm80211_hwsim_prepare_tx(struct ieee80211_hw *hw, struct txop *txop,
				     int idx, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = &txop->xti[idx];
	struct tx_mpdu_desc *mdesc = txop_get_mdesc(txop, idx);
	struct tx_buffer_desc *bdesc = txop_get_bdesc(txop, idx);
	struct sg_table *sgt = &xti->sgt;
	struct scatterlist *sg;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)(skb->data + sizeof(xti->tv));
	struct scm80211_txq *txq = txop->multi_owner ? txop->multi_owner : txop->owner;
	int n = scm80211_get_skb_fragments(skb);
	__le64 addr;

#if KERNEL_VERSION(5, 3, 0) <= LINUX_VERSION_CODE
	__sg_alloc_table(sgt, n, ARRAY_SIZE(xti->sg), xti->sg,
			 ARRAY_SIZE(xti->sg), GFP_KERNEL, NULL);
#else
	__sg_alloc_table(sgt, n, ARRAY_SIZE(xti->sg), xti->sg, GFP_KERNEL, NULL);
#endif
	sgt->nents = skb_to_sgvec(skb, sgt->sgl, 0, skb->len);
	if (unlikely(((int)sgt->nents) <= 0)) {
		scm80211_err(hw, "- failed to create skb scatterlist\n");
		return -EAGAIN;
	}

	sgt->nents = dma_map_sg(dev, sgt->sgl, sgt->nents, DMA_TO_DEVICE);
	if (unlikely(sgt->nents == 0)) {
		scm80211_err(hw, "- failed to map scatterlist\n");
		return -EAGAIN;
	}

	BUG_ON(sgt->nents > SCM80211_SKB_MAX_FRAGS);

	/* Program TX buffer descriptors */
	memset(bdesc, 0, sizeof(*bdesc) * SCM80211_SKB_MAX_FRAGS);
	for_each_sg(sgt->sgl, sg, sgt->nents, n) {
		/* Map host address into the target local address */
		addr = scm80211_remap(hw, sg_dma_address(sg), sg_dma_len(sg));
		if (!addr) {
			scm80211_err(hw, "- fail to remap bdesc %d\n", n);
			sgt->nents = n;
			goto fail;
		}

		put_unaligned_le64(addr, (void *)&bdesc[n].addr_l);
		bdesc[n].len = cpu_to_le32(sg_dma_len(sg));
		// TODO: is it necessary? Add a proper comment.
		/* Every sg has no order dependency with each other */
		wmb();
	}
	xti->flags |= SCM80211_XTI_DMA_MAPPED;

	memset(word(mdesc, 2), 0, sizeof(__le32) * 3);
	mdesc->aid = txq->txq.vif->bss_conf.aid;
	mdesc->tid = xti->tid;
	xti->sn = IEEE80211_SEQ_TO_SN(le16_to_cpu(hdr->seq_ctrl));
	mdesc->sn = IEEE80211_SEQ_TO_SN(le16_to_cpu(hdr->seq_ctrl));

	/*
	 * NB: IEEE80211_TX_CTL_NO_ACK does not indicate the frame does not
	 * solicits a response frame. It only indicates the upper layer is
	 * not interested whether the frame is acknowledged or not. Therefore,
	 * @mdesc->noack should be set without referrering to it.
	 */
	mdesc->noack = !xti->ack;
	mdesc->mpdu_len = xti->mlen;

	/* PMF is done by software. H/W needs to bypass encyrption */
	/* PMF + unsupported cipher : key is NULL */
	if (ieee80211_has_protected(hdr->frame_control) &&
	    !IEEE80211_SKB_CB(skb)->control.hw_key) {
		mdesc->bypass_encrypt = true;
	}

	/* RTS protection is set in scm80211_finalize_txop() */
	mdesc->protection = NO_PROTECTION;
	/* EOF */
	mdesc->eof = xti->eof;

	/* FIXME: TSF */
	mdesc->timestamp_update = false;
	mdesc->timestamp_offset = false;

	if (txop->ampdu.enabled) {
		mdesc->aggregate = true;
		mdesc->mpdu_spacing = xti->mmss_pad;
	}

	mdesc->bdnum = cpu_to_le32(xti->sgt.nents - 1);
	if (ieee80211_is_mgmt(hdr->frame_control) || ieee80211_is_data_qos(hdr->frame_control))
		mdesc->no_htc = !ieee80211_has_order(hdr->frame_control);

	return 0;

fail:
	for_each_sg(sgt->sgl, sg, sgt->nents, n) {
		if (sg_dma_len(sg))
			scm80211_unmap(hw, sg_dma_address(sg), sg_dma_len(sg));
	}
	dma_unmap_sg(dev, sgt->sgl, sgt->nents, DMA_TO_DEVICE);
	return -EAGAIN;
}

static bool scm80211_hwsim_issue_txop(struct ieee80211_hw *hw, struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct txdesc txdesc;
	struct txop *issued = NULL;
	struct ieee80211_vif *vif;
	struct ieee80211_vif_priv *vifp;
	unsigned long flags;
	u32 bcnt, cmd;

	/* Check if txop is eligible for transmission */
	if (txop->ampdu.update_needed && !scm80211_txop_is_within_transmit_window(txop))
		goto exit;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	issued = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
	/* only one issued txop on hwq under pcie soluction */
	if (issued) {
		list_move(&issued->list, &hwq->ring.head[TXOP_STATE_ISSUED]);
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		goto exit;
	}

	if (txop->state != TXOP_STATE_SCHEDULED && txop->state != TXOP_STATE_TX_RETRY) {
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		scm80211_warn(hw, "- txop not ready for transmission status=%s\n",
			      txop_state_str[txop->state]);
		goto exit;
	}

	__scm80211_txop_change_state(txop, TXOP_STATE_ISSUED);
	__scm80211_txop_add(txop, TXOP_STATE_ISSUED);

	spin_unlock_irqrestore(&hwq->ring.lock, flags);

	scm80211_trace(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

	bcnt = get_random_u32() & hwq->cw;

	memset(&txdesc, 0, sizeof(txdesc));
	txdesc.addr = txop->raddr;
	txdesc.num = txop->n_mpdu;
	/* write to target memory */
	wmb();
	memcpy_toio(hwq->txdesc, &txdesc, offsetof(struct txdesc, htc));
	/* write to target memory */
	wmb();

#ifndef SCM80211_WFA_TEST
	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
		scm80211_warn(hw, "hwq#%d busy? cmd=%08x\n", hwq->id, cmd);
#endif
	cmd = 0;
	/*
	 * 9.4.2.28 EDCA Parameter Set element
	 *      The minimum value of the AIFSN subfield is 2.
	 *
	 * 26.2.7 EDCA operation using MU EDCA parameters.
	 * The backoff counter maintenance corresponding to the updated
	 *   state variables shall follow the rules in
	 *   10.23.2.2 (EDCA backoff procedure) except that if AIFSN[AC] is 0
	 *   then the EDCAF corresponding to that AC shall be suspended until
	 *    the MUEDCATimer[AC] reaches 0 or is reset to 0.
	 */
	/* When AIFSN[AC] != 0 or during scan, enable EDCA/MU EDCA */
	if (hwq->edca.aifs || priv->scanner.scanning)
		cmd |= FIELD_PREP(TX_QUEN_CMD_EDCA, 1);

	/* If mu_edca running, during scan, using non-mu_edca param */
	if ((hwq->mu_edca_status & MU_EDCA_RUNNING) && priv->scanner.scanning) {
		bcnt = get_random_u32() & hwq->b_edca.cw_min;
		hwq->edca.aifs = hwq->b_edca.aifs;
	}
	txop->bcnt = bcnt;

	cmd |= FIELD_PREP(TX_QUEN_CMD_BOFF_COUNT, bcnt);
	cmd |= FIELD_PREP(TX_QUEN_CMD_AIFSN, hwq->edca.aifs);

	vif = txop->owner->txq.vif;
	vifp = ieee80211_to_priv(vif);
	/* Enable HE TB */
	if (vif && vif->type == NL80211_IFTYPE_STATION) {
		struct ieee80211_bss_conf *bss_conf = &vif->bss_conf;
		u32 mmss = vifp->reg_mmss;

		if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_EN, mmss)) {
			if (bss_conf->aid && bss_conf->he_support)
				cmd |= FIELD_PREP(TX_QUEN_CMD_AID_MATCHED_QUEUE_EN, 1);

			if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_ASSO_UORA_EN, mmss) &&
			    bss_conf->he_support)
				cmd |= FIELD_PREP(TX_QUEN_CMD_ASSO_UORA_QUEUE_EN, 1);
			else if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_UNASSO_UORA_EN, mmss))
				cmd |= FIELD_PREP(TX_QUEN_CMD_UNASSO_UORA_QUEUE_EN, 1);
		}
	}

	/*
	 * Report bug when AIFS == 0 && no HE TB response bits set
	 * v == 0 means cancel TX
	 */
	if (WARN_ON((!hwq->edca.aifs && !(cmd & TX_QUEN_CMD_HE_TB_RESP_MASK)) || !cmd)) {
		/* set default value */
		cmd |= FIELD_PREP(TX_QUEN_CMD_AIFSN, 2);
		cmd |= FIELD_PREP(TX_QUEN_CMD_EDCA, 1);
	}

	trace_scm80211_start_txop(hw, txop, txop->owner, hwq->edca.aifs,
				  bcnt, txop->n_mpdu, txop->ampdu.enabled);

	scm80211_mac_writel(hw, txop->limit, MAC_REG_TX_QUEN_TXOP(hwq->id));
	scm80211_mac_writel(hw, cmd, MAC_REG_TX_QUEN_CMD(hwq->id));

	return true;

exit:
	spin_lock_irqsave(&hwq->ring.lock, flags);
	__scm80211_txop_change_state(txop, TXOP_STATE_SCHEDULED);
	__scm80211_txop_add(txop, TXOP_STATE_SCHEDULED);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
	return false;
}

const struct scm80211_bus_ops scm80211_hwsim_ops = {
	.mac_writel	= scm80211_hwsim_mac_writel,
	.mac_readl	= scm80211_hwsim_mac_readl,
	.phy_writel	= scm80211_hwsim_phy_writel,
	.phy_readl	= scm80211_hwsim_phy_readl,
	.rf_writel	= scm80211_hwsim_rf_writel,
	.rf_readl	= scm80211_hwsim_rf_readl,
	.raw_writel	= scm80211_hwsim_raw_writel,
	.raw_readl	= scm80211_hwsim_raw_readl,
	.remap		= scm80211_hwsim_remap,
	.unmap		= scm80211_hwsim_unmap,
	.inverse_map	= scm80211_hwsim_inverse_map,
	.hwq_alloc_dma	= scm80211_hwsim_hwq_alloc_dma,
	.hwq_free_dma	= scm80211_hwsim_hwq_free_dma,
	.set_channel	= scm80211_hwsim_set_channel,
	.init_rx	= scm80211_hwsim_start_rx,
	.deinit_rx      = scm80211_hwsim_stop_rx,
	.refill_rxring  = scm80211_hwsim_refill_rxring,
	.rx_start	= scm80211_hwsim_rx_start,
	.init_tx	= scm80211_hwsim_start_tx,
	.deinit_tx      = scm80211_hwsim_stop_tx,
	.cancel_tx	= scm80211_hwsim_abort_backoff,
	.cleanup_tx	= scm80211_hwsim_cleanup_tx,
	.prepare_tx	= scm80211_hwsim_prepare_tx,
	.start_tx	= scm80211_hwsim_issue_txop,

};
const struct scm80211_bus_ops *scm_bus_ops = &scm80211_hwsim_ops;

static u32 scm80211_sync_rx_wptr(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 wptr;

	wptr = scm80211_mac_readl(hw, MAC_REG_RX_BUF_PTR1);
	wptr = FIELD_GET(RX_BUF_PTR1_WRITE_PTR, wptr);
	priv->rxring.wptr = wptr;

	scm80211_trace(hw, "wtpr=0x%04x\n", wptr);
	return wptr;
}

/**
 * scm80211_rx_napi_irq() - handle rx done interrupt request
 *
 * This function check the read buffer invalid and schedule the napi callback function
 */
static void scm80211_rx_napi_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct scm80211_rxdesc *rxd;
	struct scm80211_rx_info *info;

	scm80211_sync_rx_wptr(hw);

	if (ring_ptr_equal(priv, priv->rxring.rptr, priv->rxring.wptr))
		return;

	scm80211_trace(hw,
		       "rx irq: bptr=0x%04lx, rptr=0x%04lx, wptr=0x%04lx\n",
		       ring_ptr(&priv->rxring, priv->rxring.bptr),
		       ring_ptr(&priv->rxring, priv->rxring.rptr),
		       ring_ptr(&priv->rxring, priv->rxring.wptr));

	trace_scm80211_rx_irq(hw,
			      ring_ptr(&priv->rxring, priv->rxring.bptr),
			      ring_ptr(&priv->rxring, priv->rxring.rptr),
			      ring_ptr(&priv->rxring, priv->rxring.wptr));

	//scm80211_disable_rx_irq(hw);
	//scm80211_disable_rx_irq(hw);
	scm80211_rx_statistics_update(hw);
again:
	/* Step 1: Detach buffer and update rptr */
	while (!ring_ptr_equal(priv, priv->rxring.rptr, priv->rxring.wptr)) {
		struct sk_buff *skb;

		rxd = &priv->rxring.rxd[ring_index(&priv->rxring, priv->rxring.rptr)];
#ifdef SCM80211_RX_PAGE_ALLOC
		scm80211_unmap(hw, rxd->dma_addr, rx_buflen);
		dma_unmap_page(dev, rxd->dma_addr, rx_buflen, DMA_FROM_DEVICE);
		rxd->dma_addr = 0;
		skb = scm80211_alloc_rx_skb(hw);
		if (!skb) {
			scm80211_info(hw, "scm alloc skb null\n");
			break;
		}
		info = SCM80211_SKB_RXCB(skb);
		memset(info, 0, sizeof(*info));
		memcpy_fromio(&info->rxd, rxd->config, sizeof(struct rxdesc));
		info->index = rxd->id;
		memcpy(skb->data, page_address(rxd->page), info->rxd.dlen + IEEE80211_RX_HEADROOM);
		skb_reserve(skb, IEEE80211_RX_HEADROOM);
		skb_queue_tail(&priv->rxring.rxskbs, skb);
#else
		info = SCM80211_SKB_RXCB(rxd->skb);
		memcpy_fromio((__le32 *)&info->rxd + 3, (__le32 *)rxd->config + 3,
			      sizeof(__le32));
#ifdef RDMA
		scm80211_unmap(hw, info->dma_addr, rx_buflen);
#endif
		dma_unmap_single(dev, info->dma_addr, rx_buflen, DMA_FROM_DEVICE);

		skb_queue_tail(&priv->rxring.rxskbs, rxd->skb);
		skb_queue_tail(&priv->rxring.rxskbs, rxd->skb);

		rxd->skb = NULL;
#endif
		priv->rxring.rptr++;

		/* Sync new skb */
		if (ring_ptr_equal(priv, priv->rxring.rptr, priv->rxring.wptr))
			scm80211_sync_rx_wptr(hw);
	}

	/* check last info */
	if (info->rxd.frag == START_FRAME || info->rxd.frag == MIDDLE_FRAME) {
		scm80211_trace(hw, "rx irq end in the middle polling data again!\n");
		scm80211_sync_rx_wptr(hw);
		goto again;
	}

	/* Step 2: Attach new buffer and update bptr */
	scm80211_refill_rxring(priv->hw);

	ieee80211_queue_work(hw, &priv->rxring.work);

	napi_schedule(&priv->napi.context);
}

void scm80211_tx_irq(struct ieee80211_hw *hw, unsigned long queues)
{
	struct scm80211_hwq *hwq;
	struct txop *txop;
	int i;

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		txop = scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
		if (!txop) {
			scm80211_warn(hw, "%s(hwq#%d) bogus interrupt\n", __func__, hwq->id);
			continue;
		}
		scm80211_trace(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

		/* Read and store hardware txop context */
		txop->cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		txop->txd.result = readl(&hwq->txdesc->result);
		txop->retry_cnt = 1;
		txop->scheduled = true;
		if (txop->sta && txop->txd.cw)
			txop->txd.htc = readl(&hwq->txdesc->htc);

		/* Debugging */
		if (hwq->txdesc->addr != txop->raddr) {
			scm80211_warn(hw, "- txdesc addr=%llx, txop addr=%lx\n",
				      hwq->txdesc->addr, (unsigned long)txop->raddr);
		}
		scm80211_handle_txop_complete(hw, txop);
	}
}

irqreturn_t scm80211_handle_irq(int irq, void *data)
{
	struct ieee80211_hw *hw = data;
	u32 v, en, queues;

	v = scm80211_mac_readl(hw, MAC_REG_INTR_STATUS);
	if (v == 0)
		return IRQ_NONE;

retries:
	en = ((struct ieee80211_hw_priv *)hw->priv)->int_en;
	scm80211_trace(hw, "irq=%08x enable=%08x\n", v, en);

	if (v & (INTR_STATUS_TSF0 | INTR_STATUS_TSF1))
		scm80211_tsf_irq(hw, v);

	if (v & (INTR_STATUS_RX_DONE | INTR_STATUS_RX_BUF))
		scm80211_rx_napi_irq(hw);

	queues = FIELD_GET(INTR_STATUS_TX_DONE, v);
	scm80211_tx_irq(hw, (unsigned long)queues);

	if (v & INTR_STATUS_BASIC_TRIG_QNULL_VIF)
		scm80211_handle_muedca_irq(hw, FIELD_GET(INTR_STATUS_BASIC_TRIG_QNULL_VIF, v));

	/* Acknowledge the interrupts */
	scm80211_mac_writel(hw, v, MAC_REG_INTR_CLEAR);

	/*
	 * See if new interrupts arrived while we are handling the older ones.
	 *
	 * NB: If new interrupts are not handled, the driver will not receive
	 *     any further interrupts.
	 *
	 *
	 *              +-------------------------+
	 *              |                         |
	 * TX_DONE(i)---+                         +-------------
	 *
	 *                                 +--------------------
	 *                                 |
	 * RX_DONE   ----------------------+
	 *
	 *               +--------------------------------------+
	 *               |
	 * WLAN_IRQ -----+
	 *
	 *               T0               T1     T2
	 *
	 * If interrupt handler hanldes only TX_DONE(i) and clears that
	 * interrupts at T2, WLAN_IRQ line keeps staying asserted, and
	 * the PCIe controller will not generate further interrupts.
	 *
	 * Therefore, the driver needs to read again the MAC_REG_INTR_STATUS
	 * register and if there is new interrupts sneaking in between T0 and T2,
	 * it needs to process it.
	 *
	 */
	v = scm80211_mac_readl(hw, MAC_REG_INTR_STATUS);
	if (v)
		goto retries;

	return IRQ_HANDLED;
}

static int scm80211_platform_probe(struct platform_device *pdev)
{
	struct ieee80211_hw *hw;
	struct ieee80211_hw_priv *priv;
	struct hwsim_bus_data *hwsim;
	struct resource *r;
	u8 addr[ETH_ALEN] = { 0x02, 0x00, 0x00, pdev->id >> 8, pdev->id, 0x00 };
	int err = 0;

	dev_notice(&pdev->dev, "Senscomm wireless device: %s\n", dev_name(&pdev->dev));

	scm80211_register_oops_handler();

	hw = ieee80211_alloc_hw(sizeof(*priv) + sizeof(*hwsim), &scm80211_ops);
	if (!hw) {
		dev_err(&pdev->dev, "failed to alloc hw\n");
		return -ENOMEM;
	}

	SET_IEEE80211_DEV(hw, &pdev->dev);
	platform_set_drvdata(pdev, hw);

	priv = hw->priv;
	priv->hw = hw;
	priv->dev = &pdev->dev;
	priv->bus.ops = &scm80211_hwsim_ops;
	priv->bus_type = PLATFORM_BUS;
	priv->idx = pdev->id;
	scm80211_set_debug_trace(hw);
	scm80211_hw_band_set(hw);

	hwsim = (void *)priv->bus.data;
	hwsim->num_atu = 64;
	hwsim->page_size = SZ_2G / hwsim->num_atu;
	hwsim->page_shift = ilog2(hwsim->page_size);
	bitmap_zero(hwsim->bitmap, hwsim->num_atu);
	spin_lock_init(&hwsim->lock);

	err = scm80211_instantiate_hwsim(pdev);
	if (err) {
		err = -EINVAL;
		goto fail;
	}

	/* Shared memory */
	r = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!r) {
		dev_err(&pdev->dev, "found no memory resource\n");
		err = -EINVAL;
		goto fail;
	}
	hwsim->mem = __va(r->start);

	/* Register interface */
	hwsim->regmap = dev_get_regmap(&pdev->dev, NULL);
	if (!hwsim->regmap) {
		dev_err(&pdev->dev, "failed to get regmap\n");
		err = -EINVAL;
		goto fail;
	}

	/* Irq */
	err = scm80211_hwsim_request_threaded_irq(pdev, scm80211_handle_irq, hw);
	if (err < 0) {
		dev_err(&pdev->dev, "failed to request irq\n");
		goto fail;
	}

	if (1) {
		struct scm80211_rf_device *rf;

		err = scm2020_chip_init(hw);
		if (err) {
			dev_err(&pdev->dev, "failed to initialize device %04x\n", 0x2020);
			goto fail;
		}

		rf = scm80211_alloc_rf_device(hw, NULL, 0);
		if (!rf) {
			dev_err(&pdev->dev, "failed to find RF\n");
			goto fail;
		}
		phy_attach_rf(hw, rf);
	}

	err = scm80211_init_hw(hw, addr);
	if (err) {
		dev_err(&pdev->dev, "scm80211_init_hw failed (%d)\n", err);
		goto fail;
	}

	err = ieee80211_register_hw(hw);
	if (err < 0) {
		dev_err(&pdev->dev, "ieee80211_register_hw failed (%d)\n", err);
		goto fail;
	}

#ifdef CONFIG_MAC80211_DEBUGFS
	scm80211_register_dbgfs(hw);
#endif
	dev_dbg(&pdev->dev, "hwaddr=%pM iomem=%pR\n", hw->wiphy->perm_addr, r);

	scm80211_rfkill_init(hw);
	return 0;

fail:
	scm80211_unregister_oops_handler();
	scm80211_deinit_hw(hw);
	platform_set_drvdata(pdev, NULL);
	return err;
}

static int scm80211_platform_remove(struct platform_device *pdev)
{
	struct ieee80211_hw *hw = platform_get_drvdata(pdev);

	if (!hw) {
		dev_err(&pdev->dev, "driver data staled\n");
		return -EINVAL;
	}

	scm80211_rfkill_deinit(hw);

	scm80211_deinit_chanctx_sched(hw);

#ifdef CONFIG_MAC80211_DEBUGFS
	scm80211_unregister_dbgfs(hw);
#endif

	/* Mind the order */
	ieee80211_unregister_hw(hw);
	scm80211_detach_rf(hw);
	scm80211_deinit_hw(hw);
	platform_set_drvdata(pdev, NULL);

	scm80211_unregister_oops_handler();
	return 0;
}

static struct platform_driver scm80211_platform_driver = {
	.driver = {
		   .name = "scm80211-hwsim",
		   .suppress_bind_attrs = true,
		    },
	.probe = scm80211_platform_probe,
	.remove = scm80211_platform_remove
};

/*module_platform_driver(scm80211_platform_driver);*/

static int scm80211_hwsim_init(void)
{
	int ret;

	scm80211_minstrel_init();
	ret = scm80211_init_rf_modules();
	if (ret)
		pr_err("failed to register rf drivers\n");

	ret = platform_driver_register(&scm80211_platform_driver);
	if (ret)
		pr_err("failed to register scm80211 hwsim driver: %d\n", ret);
	return ret;
}

module_init(scm80211_hwsim_init);

static void scm80211_hwsim_exit(void)
{
	platform_driver_unregister(&scm80211_platform_driver);
	scm80211_minstrel_exit();
	scm80211_exit_rf_modules();
}

module_exit(scm80211_hwsim_exit);

MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Sensecomm Semiconductor, Inc");
