// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifdef SCM80211_FTM
#define	func_fmt(str)		"[FTM_CMD] "str
#define	func_trace			FTM
#endif

#include <linux/wireless.h>
#include <net/cfg80211.h>
#include <net/iw_handler.h>
#include "core.h"
#include "util.h"
#ifdef SCM80211_FTM
#include "ftm.h"
#include "efuse.h"
#endif

#define MAX_CH_ID			169
#define	MAX_FR_CMD			8
#define	MAX_HT				4

#define	NAME2STR(name)		(#name)
#define	RATE_BCC_CODE		"BCC"
#define	RATE_LDPC_CODE		"LDPC"

#ifdef SCM80211_FTM
#define FTM_DBG(hw, fmt, args...)   scm80211_trace(hw, fmt, ##args)
#endif

#define	STRL_STR(cs, s)		((strlen(cs) > strlen(s)) ? strlen(cs) : strlen(s))
#define	COMP_STR(cs, s)		((strncmp(cs, s, STRL_STR(cs, s)) == 0) ? true : false)
#define	DIFF_CH(ch1, ch2)	(((ch1) > (ch2)) ? ((ch1) - (ch2)) : ((ch2) - (ch1)))
#define	GET_2G_CH2FREQ(ch)	(2407 + (ch) * 5)
#define	GET_5G_CH2FREQ(ch)	(5000 + (ch) * 5)
#define	GET_2G_FREQ2CH(freq)	(((freq) - 2407) / 5)
#define	GET_5G_FREQ2CH(freq)	(((freq) - 5000) / 5)

#ifdef SCM80211_FTM
#define SCM80211_IOCTL_SET		SIOCIWFIRSTPRIV
#define SCM80211_IOCTL_GET		(SIOCIWFIRSTPRIV + 0x01)
#else
#define SCM80211_IOCTL_HELP		(SIOCIWFIRSTPRIV + 0x03)
#ifdef SCM80211_IWPRIV_DEBUG
#define SCM80211_IOCTL_DEBUG_REG	(SIOCIWFIRSTPRIV + 0x05)
#define SCM80211_IOCTL_DEBUG_TRACE	(SIOCIWFIRSTPRIV + 0x07)
#endif
#define SCM80211_IOCTL_STATE	(SIOCIWFIRSTPRIV + 0x09)
#define SCM80211_IOCTL_RATE		(SIOCIWFIRSTPRIV + 0x0B)
#endif

static const char *const rate_format_str[] = {
	[0] = "NON-HT",
	[1] = "HT",
	[2] = "VHT",
	[3] = "unused",
	[4] = "HE-SU",
};

static const char *const non_ht_rate_str[] = {
	"1M", "2M", "5.5M", "11M",
	"OFDM-6M", "OFDM-9M", "OFDM-12M", "OFDM-18M",
	"OFDM-24M", "OFDM-36M", "OFDM-48M", "OFDM-54M",
};

#ifdef SCM80211_FTM
#define	SCM_GPIO_NUM			17
const unsigned char gpio_list[SCM_GPIO_NUM] = {
	10, 11, 12, 13, 14, 20, 31,
	46, 47, 48, 49, 50, 61, 62, 63
};

#define	SCM_MAX_CFO			300000
#define SCM_MAX_XCAL		15

enum FTM_PARMS_TYPE {
	FTM_PARMS_SET = 0,
	FTM_PARMS_GET,
	FTM_PARMS_END,
};

const char ftm_fixrate_cmds[MAX_FR_CMD][8] = {
	"fmt=", "rate=", "bw=", "gi=", "code=", "nss=", "ltf=", "cmt="
};

const char ftm_ht_str[MAX_HT][8] = {
	"NON-HT", "HT", "HE", "VHT",
};

#else
#define MAX_CMD_STR			2000
#ifdef SCM80211_IWPRIV_DEBUG

const char *reg_cmd_str[] = {
	[MAC_REG] = "mac",
	[PHY_REG] = "phy",
	[RFI_REG] = "rf",
	[RAW_REG] = "raw"
};

static const char *debug_section[16] = {
	"main",
	"tx",
	"rx",
	"cal",
	"bus",
	"rc",
	"chan",
	"phy",
	"rf",
	"ps",
	"dfs",
	"fs",
	"ftm",
	"coex",
	"",
	"priv"
};
#endif
#endif

const char *cmd_not_support = "This command not support in current version!\n";

struct iwpriv_private_ops {
	char *name;
	int (*handler)(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp);
};

#ifdef SCM80211_FTM
static int ftm_set_parms_clear(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			       char *rsp)
{
	ftm_clear_parms(hw, vif);
	return 0;
}

static int ftm_set_pktcnt(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->rxpktcnt = (u32)decstr2int(str, POS_10);
	ftm->txpktcnt = ftm->rxpktcnt;
	FTM_DBG(hw, "%s = %d\n", NAME2STR(txpktcnt), (int)ftm->txpktcnt);
	return 0;
}

static int ftm_set_pktsize(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->txpktsize = decstr2int(str, POS_10);
	FTM_DBG(hw, "%s = %d\n", NAME2STR(txpktsize), ftm->txpktsize);
	return 0;
}

static void ftm_set_center_freq(struct scm80211_ftm *ftm, int bw, bool plus)
{
	const int ch40_5g_center[12] = { 38, 46, 54, 62, 102, 110, 118, 126, 134, 142, 151, 159 };
	const int ch80_center[6] = { 42, 58, 106, 122, 138, 155 };
	struct ieee80211_channel chan = { 0 };
	int i, ch = ftm->center_ch;

	if (bw == 20) {
		ftm->channel = ch;
		ftm->center_freq2 = 0;
	} else if (bw == 40) {
		if (ch <= 14 && (ch < 3 || ch > 11)) {
			return;
		} else if (ch > 36) {
			for (i = 0; i < 12; i++) {
				if (ch == ch40_5g_center[i])
					break;
			}
			if (i == 12)
				return;
		}
		/* scm80211_chan_5ghz, do not support CH144 */
		if (ch == 142)
			plus = false;
		if (plus)
			ftm->channel = ch + 2;
		else
			ftm->channel = ch - 2;
	} else if (bw == 80) {
		for (i = 0; i < 6; i++) {
			if (ch == ch80_center[i])
				break;
		}
		if (i == 6)
			return;
		ftm->channel = ch - 6;
	} else {
		ftm->err_flag = -FTM_ERR_CHANNEL;
		return;
	}

	if (ftm->err_flag == -FTM_ERR_CHANNEL)
		ftm->err_flag = 0;

	if (bw != 20)
		ftm->center_freq2 = (ch < 15) ? GET_2G_CH2FREQ(ch) : GET_5G_CH2FREQ(ch);
	if (ftm->channel < 15) {
		ftm->center_freq1 = GET_2G_CH2FREQ(ftm->channel);
		chan.band = NL80211_BAND_2GHZ;
	} else if (ftm->channel > 30) {
		ftm->center_freq1 = GET_5G_CH2FREQ(ftm->channel);
		chan.band = NL80211_BAND_5GHZ;
	}

	chan.center_freq = ftm->center_freq1;
	chan.hw_value = ftm->center_freq1;
	chan.max_power = 20;

	ftm->fixrate.band = chan.band;
	ftm->fixrate.rate.count = 1;
	memcpy(&ftm->chan, &chan, sizeof(struct ieee80211_channel));
}

static int ftm_set_channel(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->center_ch = decstr2int(str, POS_10);
	if (ftm->center_ch > MAX_CH_ID)
		ftm->center_ch = 0;

	return 0;
}

static int ftm_set_txpwr(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->txpower = decstr2int(str, POS_10);

	if (ftm->txpower != FTM_POWER_NOUSE) {
		ftm->gain_index[0] = FTM_GAINIDX_NOUSE;
		ftm->gain_index[1] = FTM_GAINIDX_NOUSE;
	}
	/* ftm->txpower * 10 for 0.5db steps */
	ftm->txpower *= 10;

	return 0;
}

static int ftm_set_gain_index(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			      char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int idx = ftm->chain;
	char *spl[2];
	u8 i, n;

	FTM_DBG(hw, "[%s] str = %s\n", __func__, str);
	ftm->gain_index[ftm_chain_0] = FTM_GAINIDX_NOUSE;
	ftm->gain_index[ftm_chain_1] = FTM_GAINIDX_NOUSE;

	n = (u8)string_split(spl, str, ",", 2);
	if (!n && strlen(str)) {
		ftm->gain_index[idx] = decstr2int(str, POS_10);
	} else {
		for (i = 0; i < n; i++) {
			if (n == 2)
				idx = i;
			ftm->gain_index[idx] = decstr2int(spl[i], POS_10);

			if (ftm->gain_index[idx] > FTM_GAINIDX_MAX || ftm->gain_index[idx] < 0) {
				ftm->gain_index[idx] = FTM_GAINIDX_NOUSE;
				sprintf(rsp, "set gain_index_%d out of range(0~%d)\n", i,
						FTM_GAINIDX_MAX);
				return -FTM_ERR_COMM;
			}
		}
	}

	if (ftm->gain_index[idx] != FTM_GAINIDX_NOUSE)
		ftm->txpower = FTM_POWER_NOUSE;

	return 0;
}

static int ftm_set_chain(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	FTM_DBG(hw, "[%s] str = %s\n", __func__, str);

	ftm->chain = decstr2int(str, POS_10);

	return 0;
}

static int ftm_gi;
static int ftm_set_fixrate(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	const u8 cck_bitrates[4] = { 1, 2, 5, 11 };
	const u8 ofdm_bitrates[8] = { 6, 9, 12, 18, 24, 36, 48, 54 };
	bool rate_flag = false, plus_40 = false;
	u8 cbw, rate, r_id, cmt;
	char *p, *spl[MAX_FR_CMD], *ptr = rsp;
	int i, n;

	n = string_split(spl, str, " ", MAX_FR_CMD);
	if (n < 1) {
		n = string_split(spl, str, ",", MAX_FR_CMD);
		if (n < 1) {
			if (strlen(str) > 1) {
				spl[0] = str;
				n = 1;
			} else {
				ptr +=
				    sprintf(ptr, "\nError: set fixrate parameters error(%d)!\n", n);
				return -FTM_ERR_COMM;
			}
		}
	}

	ftm->fixrate.flags &= ~IEEE80211_TX_CTL_LDPC;	//Default BCC
	ftm->ht_mod = VHT;
	ftm->fixrate.rate.flags |= IEEE80211_TX_RC_VHT_MCS;	//if not set fmt, Default VHT
	ftm->fixrate.flags |= IEEE80211_TX_CTL_NO_ACK | IEEE80211_TX_INTFL_OFFCHAN_TX_OK;
	ftm_set_center_freq(ftm, 20, 0);

	for (i = 0; i < n; i++) {
		/* 1. set fmt */
		p = strstr(spl[i], ftm_fixrate_cmds[0]);
		if (p) {
			ftm->fixrate.rate.flags &= ~IEEE80211_TX_RC_VHT_MCS;	//if set fmt, clear
			ftm->fixrate.he_gi_type = LONG_GI;
			ftm->fixrate.he_ltf_type = 0;
			ftm_gi = 1;

			p += strlen(ftm_fixrate_cmds[0]);
			if (COMP_STR(p, ftm_ht_str[0])) {	//NON-HT
				ftm->ht_mod = NON_HT;
				ftm->width = NL80211_CHAN_WIDTH_20_NOHT;
				ftm_gi = 0;
			} else if (COMP_STR(p, ftm_ht_str[1])) {	//HT
				ftm->ht_mod = HT_MF;
				ftm->fixrate.rate.flags |= IEEE80211_TX_RC_MCS;
			} else if (COMP_STR(p, ftm_ht_str[2])) {	//HE
				ftm->ht_mod = HE_SU;
				ftm->fixrate.rate.flags |= IEEE80211_TX_RC_HE_MCS;
				ftm->fixrate.flags |= IEEE80211_TX_CTL_LDPC;	//LDPC
				ftm->fixrate.he_gi_type = GI_3u2s;
				ftm->fixrate.he_ltf_type = 2;
				ftm->fixrate.rate.flags |= IEEE80211_HE_GI_3u2;
				ftm_gi = 6;
			} else if (COMP_STR(p, ftm_ht_str[3])) {	//VHT
				ftm->ht_mod = VHT;
				ftm->fixrate.rate.flags |= IEEE80211_TX_RC_VHT_MCS;
			}
			continue;
		}

		/* 2. set rate */
		p = strstr(spl[i], ftm_fixrate_cmds[1]);
		if (p) {
			p += strlen(ftm_fixrate_cmds[1]);
			ftm->fixrate.rate.idx &= 0xF0;
			if (ftm->ht_mod == NON_HT) {
				rate = decstr2int(p, POS_10) & 0xFF;
				for (r_id = 0; r_id < 4; r_id++) {
					if (rate == cck_bitrates[r_id]) {
						ftm->fixrate.rate.idx |= r_id;
						rate_flag = true;
						break;
					}
				}
				if (rate_flag)
					continue;

				for (r_id = 0; r_id < 8; r_id++) {
					if (rate == ofdm_bitrates[r_id]) {
						ftm->fixrate.rate.idx |= r_id;
						break;
					}
				}
				if (ftm->fixrate.band == NL80211_BAND_2GHZ)
					ftm->fixrate.rate.idx += 4;

			} else {
				if ((*p == 'm') || (*p == 'M'))
					p += 3;

				rate = decstr2int(p, POS_10) & 0x0F;
				ftm->fixrate.rate.idx |= rate;
			}
			ptr +=
			    sprintf(ptr, "r_id=%d, rate.idx=0x%X\n", r_id, ftm->fixrate.rate.idx);
			continue;
		}

		/* 3. set bandwidth */
		p = strstr(spl[i], ftm_fixrate_cmds[2]);
		if (p) {
			p += strlen(ftm_fixrate_cmds[2]);
			if (strchr(p, '+'))
				plus_40 = true;

			cbw = decstr2int(p, POS_10);
			if (cbw == 20) {
				ftm->cbw = CBW20;
				ftm->width = NL80211_CHAN_WIDTH_20;
			} else if (cbw == 40) {
				ftm->cbw = CBW40;
				ftm->width = NL80211_CHAN_WIDTH_40;
				ftm->fixrate.rate.flags |= IEEE80211_TX_RC_40_MHZ_WIDTH;
			} else if (cbw == 80) {
				ftm->cbw = CBW80;
				ftm->width = NL80211_CHAN_WIDTH_80;
				ftm->fixrate.rate.flags |= IEEE80211_TX_RC_80_MHZ_WIDTH;
			}
			ftm_set_center_freq(ftm, cbw, plus_40);
			ptr += sprintf(ptr, "cbw=%d, cent: %d, ch: %d, cent_1: %d, cent_2: %d\n",
				       cbw, ftm->center_ch, ftm->channel,
				       ftm->center_freq1, ftm->center_freq2);
			continue;
		}

		/* 4. set GI type */
		p = strstr(spl[i], ftm_fixrate_cmds[3]);
		if (p) {
			ftm->fixrate.rate.flags &= ~IEEE80211_HE_GI_3u2;
			p += strlen(ftm_fixrate_cmds[3]);
			if (*p == 'L') {
				ftm->fixrate.he_gi_type = LONG_GI;
				ftm_gi = 1;
			} else if (*p == 'S') {
				ftm->fixrate.he_gi_type = SHORT_GI;
				ftm->fixrate.rate.flags |= IEEE80211_TX_RC_SHORT_GI;
				ftm_gi = 2;
			} else if (*p == 'G') {
				p += 2;
				ftm->fixrate.he_gi_type = (decstr2int(p, POS_10) >> 3);

				if (ftm->fixrate.he_gi_type == GI_0u8s) {
					ftm->fixrate.he_gi_type = GI_0u8s;
					ftm->fixrate.rate.flags |= IEEE80211_TX_RC_SHORT_GI;
					ftm_gi = 4;
				} else if (ftm->fixrate.he_gi_type == GI_1u6s) {
					ftm->fixrate.he_gi_type = GI_1u6s;
					ftm->fixrate.rate.flags |= IEEE80211_HE_GI_1u6;
					ftm_gi = 5;
				} else if (ftm->fixrate.he_gi_type > GI_3u2s) {
					ftm->fixrate.he_gi_type = GI_3u2s;
					ftm->fixrate.rate.flags |= IEEE80211_HE_GI_3u2;
					ftm_gi = 6;
				} else
					ftm_gi = 0;
			}
			continue;
		}

		/* 5. set code mode  */
		p = strstr(spl[i], ftm_fixrate_cmds[4]);
		if (p) {
			p += strlen(ftm_fixrate_cmds[4]);
			if (COMP_STR(p, RATE_BCC_CODE))
				ftm->fixrate.flags &= ~IEEE80211_TX_CTL_LDPC;	//BCC
			else if (COMP_STR(p, RATE_LDPC_CODE))
				ftm->fixrate.flags |= IEEE80211_TX_CTL_LDPC;	//LDPC
			continue;
		}

		/* 6. set NSS */
		p = strstr(spl[i], ftm_fixrate_cmds[5]);
		if (!ftm->dbdc && p) {
			p += strlen(ftm_fixrate_cmds[5]);
			ftm->rate_nss = decstr2int(p, POS_10);

			if (ftm->fixrate.rate.flags & IEEE80211_TX_RC_MCS) {
				ftm->fixrate.rate.idx &= 0x7;
				ftm->fixrate.rate.idx |= ((ftm->rate_nss - 1) << 3);
			} else if (ftm->fixrate.rate.flags &
				   (IEEE80211_TX_RC_VHT_MCS | IEEE80211_TX_RC_HE_MCS)) {
				ftm->fixrate.rate.idx &= 0xF;
				ftm->fixrate.rate.idx |= (ftm->rate_nss << 4);
			}
			continue;
		}

		/* 7. set ltf for he */
		p = strstr(spl[i], ftm_fixrate_cmds[6]);
		if (p) {
			p += strlen(ftm_fixrate_cmds[6]);
			ftm->fixrate.he_ltf_type = (decstr2int(p, POS_10) >> 1);
		}

		/* 8. cmt */
		p = strstr(spl[i], ftm_fixrate_cmds[7]);
		if (p) {
			p += strlen(ftm_fixrate_cmds[7]);
			cmt = decstr2int(p, POS_10);
			ftm->fixrate.rate.flags &= ~IEEE80211_HE_GI_3u2;

			switch (cmt) {
			case 0:
				ftm->fixrate.he_gi_type = GI_0u8s;
				ftm->fixrate.he_ltf_type = 0;
				ftm_gi = 4;
				break;
			case 1:
				ftm->fixrate.he_gi_type = GI_0u8s;
				ftm->fixrate.he_ltf_type = 1;
				ftm_gi = 4;
				break;
			case 2:
				ftm->fixrate.he_gi_type = GI_1u6s;
				ftm->fixrate.he_ltf_type = 1;
				ftm->fixrate.rate.flags |= IEEE80211_HE_GI_1u6;
				ftm_gi = 5;
				break;
			case 3:
				ftm->fixrate.he_gi_type = GI_3u2s;
				ftm->fixrate.he_ltf_type = 2;
				ftm->fixrate.rate.flags |= IEEE80211_HE_GI_3u2;
				ftm_gi = 6;
				break;
			default:
				ftm_gi = 0;
			}
		}
	}

	ptr += sprintf(ptr,
		       "fixrate:\n"
		       "flags=0x%X\ttx_rate.flags=0x%X\n",
		       ftm->fixrate.flags, ftm->fixrate.rate.flags);

	return 0;
}

static int ftm_set_macaddr(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	char *spl[6];
	int i, n;

	n = string_split(spl, str, ":", 6);
	if (n != 6) {
		sprintf(rsp, "Error: set macaddr parameters format error!\n");
		return -FTM_ERR_COMM;
	}

	for (i = 0; i < 6; i++)
		ftm->macaddr[i] = decstr2int(spl[i], POS_16);

	return 0;
}

static int ftm_set_ifs(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->ifs = decstr2int(str, POS_10);

	memcpy(rsp, cmd_not_support, strlen(cmd_not_support) + 1);
	return strlen(cmd_not_support);
}

static int ftm_set_ampdu(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->ampdu = decstr2int(str, POS_10);
	if (ftm->ampdu)
		ftm->tid = 0;

	return 0;
}

static int ftm_set_policy(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	ftm->ack_policy = decstr2int(str, POS_10);

	return 0;
}

#define FTM_GPIO_BASE	WCN_GPIO_BASE
#define FTM_GPIO_MODE	1

static int scm2020_gpio_get_value(struct ieee80211_hw *hw, u32 pin)
{
	u32 v = 0, addr;

	addr = FTM_GPIO_BASE + (pin / 32) * 4;
	v = scm80211_raw_readl(hw, addr) >> (pin & 31);
	return v & 0x1;
}

static int ftm_gpio_ctrl(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	const char *p_err = "gpio control parameters error!\n";
	unsigned char gpio_pin = 0, gpio_type = 0, gpio_val = 0;
	char *spl[3], *ptr = rsp;
	int i, n;

	FTM_DBG(hw, "[%s] str = %s\n", __func__, str);	// id in/out [dir: 0/1]
	n = string_split(spl, str, " ", 3);
	if (n < 2)
		goto gpio_err;

	gpio_pin = decstr2int(spl[0], POS_10);
	gpio_val = decstr2int(spl[2], POS_10);
	if (COMP_STR("in", spl[1]))
		gpio_type = SCM_GPIO_IN;
	else if (COMP_STR("out", spl[1]))
		gpio_type = SCM_GPIO_OUT;
	else
		goto gpio_err;

	for (i = 0; i < SCM_GPIO_NUM; i++) {
		if (gpio_list[i] == gpio_pin)
			break;
	}
	if (i == SCM_GPIO_NUM)
		goto gpio_err;

	scm2020_set_gpio_config(hw, gpio_pin, FTM_GPIO_MODE, gpio_type, gpio_val);

	return 0;
gpio_err:
	memcpy(ptr, p_err, strlen(p_err));
	return strlen(p_err);
}

static int ftm_gpio_status(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	const char *p_err = "gpio control parameters error!\n";
	unsigned char gpio_pin, gpio_val = 0;
	int i;

	FTM_DBG(hw, "[%s] str = %s\n", __func__, str);	// id
	gpio_pin = decstr2int(str, POS_10);
	for (i = 0; i < SCM_GPIO_NUM; i++) {
		if (gpio_list[i] == gpio_pin)
			break;
	}
	if (i == SCM_GPIO_NUM) {
		sprintf(rsp, "%s", p_err);
		goto gpio_end;
	}

	gpio_val = scm2020_gpio_get_value(hw, gpio_pin);
	sprintf(rsp, "\ngpio %d val: %d\n", gpio_pin, gpio_val);

gpio_end:
	return strlen(p_err);
}

int ftm_show_parms(struct ieee80211_hw *hw, struct scm80211_ftm *ftm, char *rsp,
		   enum FTM_PRINT_SEL p_sel)
{
	const int ht_mode[MAX_HT] = { NON_HT, HT_MF, HE_SU, VHT };
	const int bw_mode[3] = {
		NL80211_CHAN_WIDTH_20, NL80211_CHAN_WIDTH_40, NL80211_CHAN_WIDTH_80
	};
	const char bw_str[3][3] = { "20", "40", "80" };
	const char _band[2][5] = { "2GHz", "5GHz" };
	const char sel_chain[3][5] = { "0", "1", "MIMO" };
	const char noht_cck_rate[4][4] = { "1", "2", "5.5", "11" };
	const char noht_ofdm_rate[8][3] = {
		"6", "9", "12", "18", "24", "36", "48", "54"
	};
	const char gi_type[7][5] = { "NULL", "LGI", "SGI", "GI04", "GI08", "GI16", "GI32" };
	const char ltf_list[3][4] = { "1x", "2x", "4x" };
	char rate_b[10] = { 0 }, str_ltf[50] = { 0 };
	char str_log[500] = { 0 }, dbg_str[100] = { 0 };
	int center_freq = 0, len = 0, r_idx = 0, nss = ftm->rate_nss;
	char p_mode[10] = { 0 };
	int ht_md_idx = 0, bw_idx = 0;
	int tx_pwr = (ftm->txpower != FTM_POWER_NOUSE) ? ftm->txpower / 10 : FTM_POWER_NOUSE;

	if ((ftm->fixrate.rate.flags & IEEE80211_TX_RC_MCS) &&
	    ftm->ht_mod != HE_SU && ftm->ht_mod != NON_HT) {
		r_idx = ftm->fixrate.rate.idx & 0x7;
	} else {
		r_idx = ftm->fixrate.rate.idx & 0xF;
	}

	if (ftm->ht_mod == NON_HT) {
		if (ftm->fixrate.band == NL80211_BAND_2GHZ) {	//2G
			if (r_idx < 4) {
				snprintf(rate_b, 10, "%sM", noht_cck_rate[r_idx]);
			} else {
				r_idx -= 4;
				snprintf(rate_b, 10, "%sM", noht_ofdm_rate[r_idx]);
			}
		} else {
			snprintf(rate_b, 10, "%sM", noht_ofdm_rate[r_idx]);
		}
	} else {
		snprintf(rate_b, 10, "MCS%d", r_idx);
	}

	for (ht_md_idx = 0; ht_md_idx < MAX_HT; ht_md_idx++) {
		if (ftm->ht_mod == ht_mode[ht_md_idx])
			break;
	}
	if (ht_md_idx >= MAX_HT)
		ht_md_idx = 0;

	if (!ht_md_idx) {
		snprintf(p_mode, sizeof(p_mode), "%s", ftm_ht_str[ht_md_idx]);
	} else {
		for (bw_idx = 0; bw_idx < 3; bw_idx++) {
			if (ftm->width == bw_mode[bw_idx])
				break;
		}
		if (bw_idx >= 3)
			bw_idx = 0;

		snprintf(p_mode, sizeof(p_mode), "%s%s", ftm_ht_str[ht_md_idx], bw_str[bw_idx]);
	}

	if (ftm->ht_mod == HE_SU) {
		snprintf(str_ltf, sizeof(str_ltf), "%-9s = %-6s", NAME2STR(ltf),
			 ltf_list[ftm->fixrate.he_ltf_type]);
	}

	center_freq =
	    (ftm->center_ch < 15) ? GET_2G_CH2FREQ(ftm->center_ch) : GET_5G_CH2FREQ(ftm->center_ch);

	if (p_sel == LOG_FTM_TX) {
		sprintf(str_log,
			"%-9s = %-6d\t%-9s = %-6s\t%-9s = %d(%d)\n"
			"%-9s = %-6s\t%-9s = %-6s\t%-9s = %-6s\n"
			"%-9s = %-6d\t%-9s = %-6d\t%-9s = %-6d\n"
			"%-9s = %-6d\t%-9s = %-6s\t%-9s = %d/%d\n"
			"%-9s = %-6d\t%-9s = %-6d\t%s\n"
			"%-9s = %02X:%02X:%02X:%02X:%02X:%02X\n\n"
			"%-9s = %-6d\n"
			"%-9s = %-6d\t%-9s = %-6d\n",
			NAME2STR(channel), ftm->channel, NAME2STR(mode), p_mode,
			NAME2STR(cent_freq), ftm->center_ch, center_freq, NAME2STR(chain),
			sel_chain[ftm->chain], NAME2STR(band), _band[ftm->fixrate.band],
			NAME2STR(rate), rate_b, NAME2STR(txpower), tx_pwr, NAME2STR(ampdu),
			ftm->ampdu, NAME2STR(amsdu), ftm->amsdu, NAME2STR(nss), nss, NAME2STR(gi),
			gi_type[ftm_gi], NAME2STR(gain_idx), ftm->gain_index[0], ftm->gain_index[1],
			NAME2STR(pktcnt), (int)ftm->txpktcnt, NAME2STR(pktsize), ftm->txpktsize,
			str_ltf, NAME2STR(mac), ftm->macaddr[0], ftm->macaddr[1],
			ftm->macaddr[2], ftm->macaddr[3], ftm->macaddr[4], ftm->macaddr[5],
			NAME2STR(seqnum), ftm->txd_pktcnt, NAME2STR(ftm_stats),
			ftm->ftm_flag, NAME2STR(err_flag), ftm->err_flag);
	} else if (p_sel == LOG_FTM_RX) {
		sprintf(str_log,
			"%-9s = %-6d\t%-9s = %-6s\t%-9s = %d(%d)\n"
			"%-9s = %-6s\t%-9s = %-6s\t%-9s = %-6d\n"
			"%-9s = %02X:%02X:%02X:%02X:%02X:%02X\n\n"
			"%-9s = %-6d\t%-9s = %-6d\n",
			NAME2STR(channel), ftm->channel, NAME2STR(mode), p_mode,
			NAME2STR(cent_freq), ftm->center_ch, center_freq,
			NAME2STR(chain), sel_chain[ftm->chain],
			NAME2STR(band), _band[ftm->fixrate.band], NAME2STR(pktcnt),
			(int)ftm->rxpktcnt, NAME2STR(mac), ftm->macaddr[0],
			ftm->macaddr[1], ftm->macaddr[2], ftm->macaddr[3],
			ftm->macaddr[4], ftm->macaddr[5], NAME2STR(ftm_stats),
			ftm->ftm_flag, NAME2STR(err_flag), ftm->err_flag);
	}

	if (rsp) {
		sprintf(rsp, "\nFTM Configured Parameters:\n%s%s", str_log, dbg_str);
		len = strlen(rsp);
	} else {
		FTM_DBG(hw, "\nFTM Configured Parameters:\n%s\n%s\n", str_log, dbg_str);
	}

	return len;
}

static int ftm_check_parms_err(struct scm80211_ftm *ftm, bool flag)
{
	if (ftm->err_flag)
		return ftm->err_flag;

	if (flag && ftm->txpktsize <= 0)
		return -FTM_ERR_TX_PKTSZ;

	if (ftm->channel <= 0 ||
	    (ftm->width > NL80211_CHAN_WIDTH_20 && ftm->center_freq1 <= 0))
		return -FTM_ERR_CHANNEL;

	if (ftm->ht_mod == NON_HT && ftm->ampdu)
		return -FTM_ERR_HTMODE;

	if (ftm->ampdu && (ftm->txpktsize > FTM_MAX_AMPDU_SIZE))
		return -FTM_ERR_AMPDU;

	if (ftm->width > NL80211_CHAN_WIDTH_20 &&
	    (ftm->center_freq2 == 5000 || ftm->center_freq2 == 2407))
		return -FTM_ERR_FREQ2;

	if (flag && ftm->width == NL80211_CHAN_WIDTH_80) {
		if (!(ftm->fixrate.rate.flags & IEEE80211_TX_RC_VHT_MCS) &&
		    !(ftm->fixrate.rate.flags & IEEE80211_TX_RC_HE_MCS))
			return -FTM_ERR_HT80_MCS;
	}

	switch (ftm->ht_mod) {
	case HE_SU:
		if ((!(ftm->fixrate.rate.flags & IEEE80211_TX_RC_MCS)) &&
		    !(ftm->fixrate.rate.flags | IEEE80211_TX_CTL_LDPC))
			return -FTM_ERR_HE_MCS;
		if (flag && ftm->fixrate.he_gi_type == GI_0u4s)
			return -FTM_ERR_HE_GI;
		if ((ftm->fixrate.he_gi_type == GI_1u6s && ftm->fixrate.he_ltf_type == HE_LTF_4x) ||
		    (ftm->fixrate.he_gi_type == GI_3u2s && ftm->fixrate.he_ltf_type == HE_LTF_2x))
			return -FTM_ERR_HE_LTF;
		break;

	default:
		break;
	}
	return 0;
}

static int ftm_tx_ctrl(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int ret = 0;

	if (ftm->rx_start == FTM_RX_START) {
		ftm->err_flag = -FTM_ERR_COMM;
		sprintf(rsp, "[%s] Error! FTM Rx not stopped(%d)!\n", __func__, ftm->rx_start);
		goto FTM_ERR;
	}

	FTM_DBG(hw, "[+%s]ctrl=%s\n", __func__, str);

	if (COMP_STR(str, "start")) {
		if (ftm->ftm_flag) {
			if (ftm->ftm_flag == 2)
				return 0;
			FTM_DBG(hw, "[%s]scm80211_ftm_tx_stop()\n", __func__);
			scm80211_ftm_tx_stop(hw, vif);
		}

		ftm->err_flag = ftm_check_parms_err(ftm, true);
		if (ftm->err_flag < 0) {
			sprintf(rsp, "[%s]FTM Tx Parameters checked fail(%d)!\n", __func__,
				ftm->err_flag);
			goto FTM_ERR;
		}
		ftm_show_parms(hw, ftm, NULL, LOG_FTM_TX);
		if (scm80211_ftm_tx_start(hw, vif) >= 0)
			ftm->ftm_flag = 1;
	} else if (COMP_STR(str, "stop")) {
		if (ftm->ftm_flag == 1) {
			scm80211_ftm_tx_stop(hw, vif);
			ftm->ftm_flag = 0;
		}
	}

	return ret;
FTM_ERR:
	FTM_DBG(hw, "%s", rsp);

	return ftm->err_flag;
}

static int ftm_set_efuse_proc(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			      char *rsp)
{
	int ret = 0;

	FTM_DBG(hw, "[%s] str=%s\n", __func__, str);
#ifdef EFUSE_ENABLE
	ret = scm80211_efuse_set_cmd_proc(str, rsp);
#else
	memcpy(rsp, cmd_not_support, strlen(cmd_not_support) + 1);
#endif
	FTM_DBG(hw, "%s", rsp);

	return ret;
}

static int ftm_set_cal_data(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			    char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	char *spl[4] = { 0 };
	int ret = 0, i, n, val[4] = { 0 };
	int dec, unt, temp;
	bool sbl;

	if (ftm->chain == ftm_chain_mimo) {
		sprintf(rsp, "can not set caldat in MIMO mode!\n");
		ret = strlen(rsp);
		goto FTM_END;
	}

	FTM_DBG(hw, "[%s] str = %s\n", __func__, str);
	n = string_split(spl, str, ",", 4);
	if (n < 4) {
		sprintf(rsp, "set params error!\n");
		ret = strlen(rsp);
		ftm->err_flag = -FTM_ERR_COMM;
		goto FTM_END;
	}

	for (i = 0; i < 4; i++) {
		sbl = false;
		temp = decstr2int(spl[i], POS_10);
		if (temp < 0)
			sbl = true;
		temp = abs(temp);
		dec = temp / 10;
		unt = temp % 10;
		if (dec > 7) {
			sprintf(rsp, "Error: [%s] set caldat{%d,%d} out of range(-75 ~ 75)\n",
				__func__, dec, unt);
			ret = strlen(rsp);
			goto FTM_END;
		}
		if (unt > 2 && unt < 8)
			val[i] = 5;
		else if (unt > 7)
			dec += 1;
		val[i] += dec * 10;
		if (sbl)
			val[i] = 0 - val[i];
	}

	memcpy(&ftm->cal_dat[ftm->chain], val, sizeof(struct scm80211_scm_cal));

FTM_END:
	if (ret)
		FTM_DBG(hw, "%s", rsp);

	return ret;
}

static int ftm_set_cfo(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

#ifdef SCM2020_CFO
	ftm->ftm_cfo = decstr2int(str, POS_10);
	if (ftm->ftm_cfo > SCM_MAX_CFO || ftm->ftm_cfo < -SCM_MAX_CFO) {
		ftm->ftm_cfo = 0;
		sprintf(rsp, "Error: [%s]out of range(%d ~ %d)!\n", __func__, -SCM_MAX_CFO,
			SCM_MAX_CFO);
		ftm->err_flag = -FTM_ERR_CFO;
	}
#else
	memcpy(rsp, cmd_not_support, strlen(cmd_not_support) + 1);
	ftm->err_flag = -FTM_ERR_CFO;
#endif
	return ftm->err_flag;
}

/*
*	before: set chain, channel, fixrate(bw), gain_idx
*	<start/stop>1/0, <tone_ofs>0-127, <pri>0-3
*/
static int ftm_set_stone(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	char *spl[3] = { 0 }, *ptr = rsp;
	int n, center_freq;

	n = string_split(spl, str, ",", 3);
	if (n < 3) {
		ftm->ftm_flag = 0;
	} else {
		ftm->ftm_flag = decstr2int(spl[0], POS_10);
		ftm->tone_ofs = decstr2int(spl[1], POS_10);
		ftm->pri_ch = decstr2int(spl[2], POS_10);
		if ((ftm->tone_ofs > 127) || (ftm->pri_ch > 3)) {
			sprintf(ptr, "set parameter out of range!\n");
			return strlen(ptr);
		}
		center_freq = (ftm->center_ch < 15) ?
			GET_2G_CH2FREQ(ftm->center_ch) : GET_5G_CH2FREQ(ftm->center_ch);
		ptr += sprintf(ptr, "\nFTM Configured Parameters:\n"
			 "\t%-9s = %d(%d)\n",
			 NAME2STR(cent_freq), ftm->center_ch, center_freq);
	}
	scm80211_ftm_set_single_tone(hw, ftm, ptr);

	return strlen(rsp);
}

static int ftm_set_xtal_cal(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int freq_cal = decstr2int(str, POS_10);

	if ((freq_cal > SCM_MAX_XCAL) || (freq_cal < -SCM_MAX_XCAL)) {
		sprintf(rsp, "set freq cal out of range(%d ~ %d)!\n", SCM_MAX_XCAL, -SCM_MAX_XCAL);
		return strlen(rsp);
	}
	ftm->freq_cal = freq_cal;
	scm80211_ftm_set_xtal_cal(hw, ftm->freq_cal);

	return 0;
}

static int ftm_rx_ctrl(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int ret = 0;

	if (ftm->tx_start == FTM_TX_START) {
		ftm->err_flag = -FTM_ERR_COMM;
		sprintf(rsp, "[%s] Error! FTM Tx not stopped(%d)!\n", __func__, ftm->tx_start);
		goto FTM_ERR;
	}

	FTM_DBG(hw, "[+%s]ctrl=%s\n", __func__, str);

	if (COMP_STR(str, "start")) {
		if (ftm->ftm_flag) {
			if (ftm->ftm_flag == 1)
				return 0;
			scm80211_ftm_rx_stop(hw, vif);
		}

		ftm->err_flag = ftm_check_parms_err(ftm, false);
		if (ftm->err_flag < 0) {
			sprintf(rsp, "[%s]FTM Rx Parameters checked fail(%d)!\n", __func__,
				ftm->err_flag);
			goto FTM_ERR;
		}
		ftm_show_parms(hw, ftm, NULL, LOG_FTM_RX);
		scm80211_ftm_rx_start(hw, vif);
		ftm->ftm_flag = 2;
	} else if (COMP_STR(str, "stop")) {
		if (ftm->ftm_flag == 2) {
			scm80211_ftm_rx_stop(hw, vif);
			ftm->ftm_flag = 0;
		}
	} else if (COMP_STR(str, "cnt_clr")) {
		scm80211_ftm_rx_clear(hw, vif);
	}

	return ret;
FTM_ERR:
	FTM_DBG(hw, "%s", rsp);

	return ftm->err_flag;
}

static int ftm_tx_parms_print(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			      char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int ret;

	ftm->err_flag = ftm_check_parms_err(ftm, true);
	ret = ftm_show_parms(hw, ftm, rsp, LOG_FTM_TX);

	FTM_DBG(hw, "[- %s] ret=%d\n", __func__, ret);
	return ret;
}

static int ftm_rx_parms_print(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			      char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	int ret;

	ftm->err_flag = ftm_check_parms_err(ftm, false);
	ret = ftm_show_parms(hw, ftm, rsp, LOG_FTM_RX);

	FTM_DBG(hw, "[- %s] ret=%d\n", __func__, ret);
	return ret;
}

static int ftm_rx_result(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	char *pt = "rx result is empty!\n";
	int len = 0;

	scm80211_ftm_rx_statistics_show(hw, ftm);
	if (strlen(ftm->ftm_rx_rpt) > 1) {
		len = strlen(ftm->ftm_rx_rpt);
		memcpy(rsp, ftm->ftm_rx_rpt, len);
	} else {
		len = strlen(pt);
		memcpy(rsp, pt, len);
	}

	FTM_DBG(hw, "[- %s] ret=%d\n", __func__, len);
	return len;
}

static int ftm_help(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp);

static int ftm_get_cmd_rsp(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	memcpy(rsp, ftm->ftm_cmd_rsp, strlen(ftm->ftm_cmd_rsp));
	memset(ftm->ftm_cmd_rsp, 0, sizeof(ftm->ftm_cmd_rsp));
	return strlen(rsp);
}

static int ftm_get_efuse_proc(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			      char *rsp)
{
	int ret = 0;

	FTM_DBG(hw, "[%s] str=%s\n", __func__, str);
#ifdef EFUSE_ENABLE
	ret = scm80211_efuse_get_cmd_proc(str, rsp);
#else
	memcpy(rsp, cmd_not_support, strlen(cmd_not_support) + 1);
#endif
	FTM_DBG(hw, "%s", rsp);

	return ret;
}

static int ftm_get_rssi(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;
	char *pt = "cannot get RSSI!\n";
	int len = 0;

	scm80211_ftm_rx_rssi_show(hw, ftm);
	if (strlen(ftm->ftm_rssi_rpt) > 1) {
		len = strlen(ftm->ftm_rssi_rpt);
		memcpy(rsp, ftm->ftm_rssi_rpt, len);
	} else {
		len = strlen(pt);
		memcpy(rsp, pt, len);
	}

	FTM_DBG(hw, "[- %s] ret=%d\n", __func__, len);
	return len;
}

static int ftm_get_tssi(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	memcpy(rsp, cmd_not_support, strlen(cmd_not_support) + 1);
	return 0;

}

static int ftm_get_cal_data(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			    char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	return scm80211_ftm_get_para_b(hw, ftm, rsp);
}

static int ftm_get_gain_index(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str,
			      char *rsp)
{
	if (scm80211_get_gainidx(hw, rsp) < 0)
		sprintf(rsp, "\ngain_idx fail!\n");
	return strlen(rsp);
}

static int ftm_get_chain(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	sprintf(rsp, "\nchain %d\n", scm80211_get_ant_sel(hw));
	return strlen(rsp);
}

static int ftm_get_xtal_cal(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				char *str, char *rsp)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_ftm *ftm = &vifp->ftm;

	sprintf(rsp, "\nget xtal_cal = %d\n", scm80211_ftm_get_xtal_cal(hw, ftm));
	return strlen(rsp);
}

struct iwpriv_private_ops iw_private_set_ops[] = {
	{ "clear", ftm_set_parms_clear },
	{ "pktcnt", ftm_set_pktcnt },
	{ "txpktsize", ftm_set_pktsize },
	{ "channel", ftm_set_channel },
	{ "txpower", ftm_set_txpwr },
	{ "chain", ftm_set_chain },
	{ "fixrate", ftm_set_fixrate },
	{ "macaddr", ftm_set_macaddr },
	{ "ifs", ftm_set_ifs },
	{ "ampdu", ftm_set_ampdu },
	{ "ack_policy", ftm_set_policy },
	{ "tx", ftm_tx_ctrl },
	{ "rx", ftm_rx_ctrl },
	{ "gainidx", ftm_set_gain_index },
	{ "gpio", ftm_gpio_ctrl },
	{ "efuse", ftm_set_efuse_proc },
	{ "caldat", ftm_set_cal_data },
	{ "cfo", ftm_set_cfo },
	{ "stone", ftm_set_stone },
	{ "xtal_cal", ftm_set_xtal_cal },
};

struct iwpriv_private_ops iw_private_get_ops[] = {
	{ "help", ftm_help },
	{ "rsp", ftm_get_cmd_rsp },
	{ "tx_parms", ftm_tx_parms_print },
	{ "rx_parms", ftm_rx_parms_print },
	{ "rx_cnt", ftm_rx_result },
	{ "gpio_stu", ftm_gpio_status },
	{ "efuse", ftm_get_efuse_proc },
	{ "rssi", ftm_get_rssi },
	{ "tssi", ftm_get_tssi },
	{ "caldat", ftm_get_cal_data },
	{ "gainidx", ftm_get_gain_index },
	{ "chain", ftm_get_chain },
	{ "xtal_cal", ftm_get_xtal_cal },
};

static int ftm_help(struct ieee80211_hw *hw, struct ieee80211_vif *vif, char *str, char *rsp)
{
	sprintf(rsp,
		"\nSCM FTM Test Tool 2024-04-18 V3.7.2\n\n"
		"set commands:\n"
		"%s\t\t\tclear and reset parameters to default\n"
		"%s\t\t\tset pktcnt <Tx/Rx packents number>\n"
		"\t\t\t\tnumber: -1 for long term\n"
		"%s\t\tset txpktsize <Tx payload data size>\n"
		"%s\t\t\tset channel <Tx/Rx channel> eg: 1,6,7,36,140...\n"
		"%s\t\t\tset txpower <Tx power dBm>\n"
		"%s\t\t\tset chain <0:chain 0, 1:chain 1, 2:MIMO>\n"
		"%s\t\t\tset fixrate <fmt>,<rate>,[bw],[gi],[code]\n"
		"\t\t\t\tfmt: HT(20MHz)/HT40/VHT80\n"
		"\t\t\t\trate: 6-54Mbps(NON-HT) 0-15(HT/VHT MCS)\n"
		"\t\t\t\tbw: 20/40/80M\n"
		"\t\t\t\tgi: LGI/SGI/GI04/GI08/GI16/GI32\n"
		"\t\t\t\tcode: BBC\n"
		"%s\t\t\tset macaddr <MAC address>\n"
		"%s\t\t\tset ifs <ifs idx>\n"
		"%s\t\t\tset ampdu <0:disable, 1:enable>\n"
		"%s\t\tset ack_policy <0:NoAck, 1:Ack or Blockack>\n"
		"%s\t\t\tset tx <clear/start/stop>\n"
		"\t\t\t\tclear: clear all set tx parameters\n"
		"\t\t\t\tstart: start tx\n"
		"\t\t\t\tstop:  stop tx\n"
		"%s\t\t\tset rx <cnt_clr/start/stop>\n"
		"\t\t\t\tcnt_clr: clear received packets/mpdu/ppdu count\n"
		"\t\t\t\tstart: start rx\n"
		"\t\t\t\tstop:  stop rx\n\n"
		"%s\t\t\tset gainidx <0-63>\n"
		"%s\t\t\tset gpio <pin> <in/out> [out: 0/1]\n"
		"%s\t\t\tset \"efuse <commands>\"\n"
		"%s\t\t\tset caldat <rf_2g>,<rf_5g_low>,<rf_5g_mid>,<rf_5g_high>\n"
#ifdef SCM2020_CFO
		"%s\t\t\tset cfo <-300000 ~ 300000>Hz\n"
#endif
		"%s\t\t\tset \"stone <stop/start(0/1)>,[ofs(0~127)],[pri(0~3)]\"\n"
		"%s\t\tset xtal_cal <-15 ~ 15>\n"
		"\nget commands:\n"
		"%s\t\t\tget help\n"
		"%s\t\tget tx_parms (show set tx parameters)\n"
		"%s\t\tget rx_parms (show set rx parameters)\n"
		"%s\t\t\tget rx_cnt (show received packets/mpdu/ppdu/RSSI)\n"
		"%s\t\tget gpio_stu <pin>\n"
		"%s\t\t\tsee command \"efuse help\"\n"
		"%s\t\t\tget rssi (show rssi)\n"
		"%s\t\t\tget tssi (show tssi)\n"
		"%s\t\t\tget rsp (get latest cmd response)\n"
		"%s\t\t\tget caldat\n"
		"%s\t\t\tget current gain index\n"
		"%s\t\t\tget current rf chain\n"
		"%s\t\tget xtal calibration value\n"
		"\n",
		iw_private_set_ops[CMD_CLEAR].name,
		iw_private_set_ops[CMD_PKTCNT].name, iw_private_set_ops[CMD_PKTSIZE].name,
		iw_private_set_ops[CMD_SET_CHAN].name, iw_private_set_ops[CMD_POWER].name,
		iw_private_set_ops[CMD_CHAIN].name, iw_private_set_ops[CMD_RATE].name,
		iw_private_set_ops[CMD_MACADDR].name, iw_private_set_ops[CMD_IFS].name,
		iw_private_set_ops[CMD_AMPDU].name, iw_private_set_ops[CMD_ACK].name,
		iw_private_set_ops[CMD_TX].name, iw_private_set_ops[CMD_RX].name,
		iw_private_set_ops[CMD_SET_GAINIDX].name, iw_private_set_ops[CMD_SET_GPIO].name,
		iw_private_set_ops[CMD_SET_EFUSE].name, iw_private_set_ops[CMD_SET_CALDAT].name,
#ifdef SCM2020_CFO
		iw_private_set_ops[CMD_SET_CFO].name,
#endif
		iw_private_set_ops[CMD_SET_STONE].name, iw_private_set_ops[CMD_SET_XTAL].name,
		iw_private_get_ops[CMD_FTM_HELP].name,
		iw_private_get_ops[CMD_TXPARMS].name, iw_private_get_ops[CMD_RXPARMS].name,
		iw_private_get_ops[CMD_RXCNT].name, iw_private_get_ops[CMD_GET_GPIO].name,
		iw_private_get_ops[CMD_GET_EFUSE].name, iw_private_get_ops[CMD_GET_RSSI].name,
		iw_private_get_ops[CMD_GET_TSSI].name, iw_private_get_ops[CMD_GET_RSP].name,
		iw_private_get_ops[CMD_GET_CALDAT].name, iw_private_get_ops[CMD_GET_GAINIDX].name,
		iw_private_get_ops[CMD_GET_CHAN].name, iw_private_get_ops[CMD_GET_XTAL].name);
	return strlen(rsp);
}

static int ftm_parms_process(struct net_device *dev, char *cmd, char *rsp, enum FTM_PARMS_TYPE pt)
{
	struct ieee80211_vif *vif = NULL;
	struct ieee80211_hw *hw = NULL;
	struct ieee80211_hw_priv *priv = NULL;
	struct iwpriv_private_ops *ops = NULL;
	char *sp1 = NULL, *sp2 = NULL;
	int i = 0, n = 0, ret = 0;
	const char *ve = "is NULL!\n";
	struct ieee80211_vif_priv *vifp = NULL;
	struct scm80211_ftm *ftm = NULL;
	char *cmd_rsp;

	hw = wiphy_to_ieee80211_hw(dev->ieee80211_ptr->wiphy);
	if (!hw) {
		sprintf(rsp, "Error: [%s] hw %s", __func__, ve);
		pr_debug("%s", rsp);
		return strlen(rsp);
	}

	priv = hw->priv;
	vif = priv->vifs[0];
	if (!vif) {
		sprintf(rsp, "[%s] vif(%d) %s\n", __func__, i, ve);
		return strlen(rsp);
	}
	FTM_DBG(hw, "[+ %s]cmd=%s\n", __func__, cmd);

	vifp = ieee80211_to_priv(vif);
	ftm = &vifp->ftm;

	sp1 = cmd;
	while (*cmd != '\0') {
		if ((*cmd == ' ') || (*cmd == '='))
			break;
		cmd++;
	}
	if (*cmd != '\0') {
		sp2 = cmd + 1;
		*cmd = '\0';
	}

	if (pt == FTM_PARMS_SET) {
		memset(ftm->ftm_cmd_rsp, 0, sizeof(ftm->ftm_cmd_rsp));
		cmd_rsp = ftm->ftm_cmd_rsp;
		ops = iw_private_set_ops;
		n = (int)(sizeof(iw_private_set_ops) / sizeof(struct iwpriv_private_ops));
	} else if (pt == FTM_PARMS_GET) {
		cmd_rsp = rsp;
		ops = iw_private_get_ops;
		n = (int)(sizeof(iw_private_get_ops) / sizeof(struct iwpriv_private_ops));
	} else {
		memcpy(rsp, cmd_not_support, strlen(cmd_not_support) + 1);
		return strlen(cmd_not_support);
	}

	for (i = 0; i < n; i++) {
		if (COMP_STR(sp1, ops[i].name)) {
			ret = ops[i].handler(hw, vif, sp2, cmd_rsp);
			break;
		}
	}

	return ret;
}
#endif

static int scm_wx_set_encode_ext(struct net_device *dev,
				 struct iw_request_info *info, union iwreq_data *wrqu, char *extra)
{
	return 0;
}

#define IW_IOCTL(x) ((x) - SIOCSIWCOMMIT)
static iw_handler scm_wx_handlers[] = {
	[IW_IOCTL(SIOCSIWENCODEEXT)] = scm_wx_set_encode_ext,
};

#ifdef SCM80211_FTM
static int scm80211_iwpriv_set_process(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	char cmd[100] = { 0 };
	int ret;

	if (copy_from_user(cmd, wrqu->data.pointer, wrqu->data.length))
		return 0;

	pr_debug("[+ %s]str: %s\n", __func__, cmd);

	wrqu->data.length = 0;
	ret = ftm_parms_process(dev, cmd, extra, FTM_PARMS_SET);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
//      printk("[- %s]data.length=%d\n%s\n", __func__, wrqu->data.length, extra);

	return 0;
}

static int scm80211_iwpriv_get_process(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	char cmd[100] = { 0 };
	int ret;

	if (copy_from_user(cmd, wrqu->data.pointer, wrqu->data.length))
		return 0;

	pr_debug("[+ %s]str(%d): %s\n", __func__, wrqu->data.length, cmd);

	wrqu->data.length = 0;
	ret = ftm_parms_process(dev, cmd, extra, FTM_PARMS_GET);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
//      printk("[- %s]data.length=%d\n\n", __func__, wrqu->data.length);

	return 0;
}

#else
#ifdef SCM80211_IWPRIV_DEBUG

/**
 * e.g
 * get: iwpriv wlan0 reg "raw 0x40111098"
 * set: iwpriv wlan0 reg "raw 0x40111098=0x01101100"
 */
static int scm80211_iwpriv_debug_reg(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	struct ieee80211_hw *hw = NULL;
	char cmd[100] = { 0 };
	char *sp1 = NULL, *sp2 = NULL, *sp3 = NULL;
	char str_reg[100] = { 0 };
	char str_val[100] = { 0 };
	u32 reg = 0, val = 0;
	int shift = 0;
	int reg_type;
	int i = 0;
	int ret;
	bool cmd_invalid = true;

	if (copy_from_user(cmd, wrqu->data.pointer, wrqu->data.length))
		return 0;
	pr_debug("[+ %s]str(%d): %s\n", __func__, wrqu->data.length, cmd);
	wrqu->data.length = 0;
	hw = wiphy_to_ieee80211_hw(dev->ieee80211_ptr->wiphy);
	pr_debug("cmd:%s", cmd);//debug
	sp1 = cmd;
	for (i = 0; i < ARRAY_SIZE(reg_cmd_str); i++) {
		if (strnstr(sp1, reg_cmd_str[i], strlen(reg_cmd_str[i]))) {
			reg_type = i;
			cmd_invalid = false;
			break;
		}
	}
	if (cmd_invalid)
		return -EINVAL;
	sp2 = strnchr(sp1, strlen(sp1), ' ');
	while (*sp2 != '\0') {
		if (*sp2 == ' ')
			sp2++;
		else
			break;
	}
	sp3 = strnchr(sp2, strlen(sp2), '=');
	if (sp3 != NULL) {
		strscpy(str_reg, sp2, sp3 - sp2 + 1);
		sp3++;
		strscpy(str_val, sp3, strlen(sp3) + 1);
	} else
		strscpy(str_reg, sp2, strlen(sp2) + 1);
	if (!strncasecmp(str_reg, "0x", 2))
		shift = 2;
	if (kstrtou32(str_reg + shift, 16, &reg))
		return -EINVAL;
	switch (reg_type) {
	case MAC_REG:
		if (sp3 == NULL) {
			val = reg_readl(mac, hw, reg);
			sprintf(extra, "\n\nmac reg read:\n[0x%08x]\t0x%08x\n", reg, val);
		} else {
			shift = 0;
			if (!strncasecmp(str_val, "0x", 2))
				shift = 2;
			if (kstrtou32(str_val + shift, 16, &val))
				return -EINVAL;
			reg_writel(mac, hw, val, reg);
			sprintf(extra, "\n\nmac reg write:\n[0x%08x]\t0x%08x\n", reg, val);
		}
		break;
	case PHY_REG:
		if (sp3 == NULL) {
			val = reg_readl(phy, hw, reg);
			sprintf(extra, "\n\nphy reg read:\n[0x%08x]\t0x%08x\n", reg, val);
		} else {
			shift = 0;
			if (!strncasecmp(str_val, "0x", 2))
				shift = 2;
			if (kstrtou32(str_val + shift, 16, &val))
				return -EINVAL;
			reg_writel(phy, hw, val, reg);
			sprintf(extra, "\n\nphy reg write:\n[0x%08x]\t0x%08x\n", reg, val);
		}
		break;
	case RFI_REG:
		if (sp3 == NULL) {
			val = reg_readl(rf, hw, reg);
			sprintf(extra, "\n\nrf reg read:\n[0x%08x]\t0x%08x\n", reg, val);
		} else {
			shift = 0;
			if (!strncasecmp(str_val, "0x", 2))
				shift = 2;
			if (kstrtou32(str_val + shift, 16, &val))
				return -EINVAL;
			reg_writel(rf, hw, val, reg);
			sprintf(extra, "\n\nrf reg write:\n[0x%08x]\t0x%08x\n", reg, val);
		}
		break;
	case RAW_REG:
		if (sp3 == NULL) {
			val = reg_readl(raw, hw, reg);
			sprintf(extra, "\n\nraw reg read:\n[0x%08x]\t0x%08x\n", reg, val);
		} else {
			shift = 0;
			if (!strncasecmp(str_val, "0x", 2))
				shift = 2;
			if (kstrtou32(str_val + shift, 16, &val))
				return -EINVAL;
			reg_writel(raw, hw, val, reg);
			sprintf(extra, "\n\nraw reg write:\n[0x%08x]\t0x%08x\n", reg, val);
		}
		break;
	default:
		break;
	}
	ret = strlen(extra);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
	return 0;
}

/**
 * e.g
 * get: iwpriv wlan0 trace
 * set: iwpriv wlan0 trace main,tx,bus
 */
static int scm80211_iwpriv_debug_trace(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	struct ieee80211_hw *hw = NULL;
	struct ieee80211_hw_priv *priv = NULL;
	char cmd[100] = { 0 };
	char param[10];
	char *sp1 = NULL, *sp2 = NULL;
	unsigned long bitmap;
	size_t pos = 0;
	u16 debug_trace = 0;
	int i = 0;
	int idx = 0;
	int ret;
	bool cmd_invalid = true;

	if (copy_from_user(cmd, wrqu->data.pointer, wrqu->data.length))
		return 0;
	pr_debug("[+ %s]str(%d): %s\n", __func__, wrqu->data.length, cmd);
	wrqu->data.length = 0;
	hw = wiphy_to_ieee80211_hw(dev->ieee80211_ptr->wiphy);
	priv = hw->priv;
	sp1 = cmd;
	if (*sp1 != '\0') {
		do {
			sp2 = strnchr(cmd + idx, strlen(cmd) - idx, ',');
			memset(param, '\0', sizeof(param));
			if (sp2) {
				strscpy(param, cmd + idx, sp2 - (cmd + idx) + 1);
				sp2++;
				idx = sp2 - cmd;
			} else
				strscpy(param, cmd + idx, strlen(cmd + idx) + 1);
			if (idx >= strlen(cmd))
				break;
			for (i = 0; i < 16; i++) {
				if (!strncmp(param, debug_section[i], strlen(param))) {
					debug_trace |= BIT(i);
					cmd_invalid = false;
					break;
				}
			}
		} while (sp2);
		if (cmd_invalid)
			return -EINVAL;
		priv->debug_trace = debug_trace;
		pos += sprintf(extra + pos, "\n\ntrace set:\n");
	} else
		pos += sprintf(extra + pos, "\n\ntrace on:\n");
	bitmap = priv->debug_trace;
	for_each_set_bit(i, &bitmap, 16)
		pos += sprintf(extra + pos, " [%s]", debug_section[i]);
	pos += sprintf(extra + pos, "\n");
	ret = strlen(extra);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
	return 0;
}
#endif

static void scm80211_vif_info_show(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	char *str = data;
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	enum nl80211_iftype type = ieee80211_vif_type_p2p(vif);
	static size_t pos;

	if (ieee80211_vif_name(vif) == NULL)
		return;
	if (*str == '\0')
		pos = 0;
	if (ieee80211_iface_is_ap(type))
		pos += sprintf(str + pos, "\n\tvif index :%d\n\tname:%s\n\ttype:%s\n"
					"\tvif_addr:%pM\n"
					"\tbssid_filter:0x%02x\n\tbss_changed:0x%08x\n\tsta_num:%d\n\tbeacon seqNum:%d\n",
					vp->index, ieee80211_vif_name(vif),
					ieee80211_iftype_str[type],
					vif->addr,
					vp->bssid_filter, vp->bss_changed,
					atomic_read(&vp->num_sta), vp->bcn_seqnum);
	else
		pos += sprintf(str + pos, "\n\tvif index :%d\n\tname:%s\n\ttype:%s\n"
					"\tvif_addr:%pM\n"
					"\tbssid_filter:0x%02x\n\tbss_changed:0x%08x\n",
					vp->index, ieee80211_vif_name(vif),
					ieee80211_iftype_str[type],
					vif->addr,
					vp->bssid_filter, vp->bss_changed);
}

/**
 * e.g
 * get: iwpriv wlan0 state
 * set: iwpriv wlan0 state amsdu=1
 */
static int scm80211_iwpriv_state(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	struct ieee80211_hw *hw = NULL;
	struct ieee80211_hw_priv *hw_priv = NULL;
	char cmd[100] = { 0 };
	char vif_buf[512] = { 0 };
	char param[16];
	char *sp1 = NULL, *sp2 = NULL;
	size_t pos = 0;
	int idx = 0;
	int ret;
	bool cmd_invalid = true;

	if (copy_from_user(cmd, wrqu->data.pointer, wrqu->data.length))
		return 0;
	pr_debug("[+ %s]str(%d): %s\n", __func__, wrqu->data.length, cmd);
	wrqu->data.length = 0;
	hw = wiphy_to_ieee80211_hw(dev->ieee80211_ptr->wiphy);
	hw_priv = hw->priv;
	sp1 = cmd;

	if (*sp1 != '\0') { //set
		pos += sprintf(extra + pos, "\n\n");
		do {
			sp2 = strnchr(cmd + idx, strlen(cmd) - idx, ',');
			memset(param, '\0', sizeof(param));
			if (sp2) {
				strscpy(param, cmd + idx, sp2 - (cmd + idx) + 1);
				sp2++;
				idx = sp2 - cmd;
			} else
				strscpy(param, cmd + idx, strlen(cmd + idx) + 1);
			if (idx >= strlen(cmd))
				break;
			if (!strncmp(param, "amsdu=", 6)) {
				if (!strncmp(param + 6, "1", 1)) {
					_ieee80211_hw_set(hw, IEEE80211_HW_TX_AMSDU);
					_ieee80211_hw_set(hw, IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU);
					pos += sprintf(extra + pos, "\tenable tx/rx amsdu\n");
					cmd_invalid = false;
				} else if (!strncmp(param + 6, "0", 1)) {
					__clear_bit(IEEE80211_HW_TX_AMSDU, hw->flags);
					__clear_bit(IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU,
								hw->flags);
					pos += sprintf(extra + pos, "\tdisable tx/rx amsdu\n");
					cmd_invalid = false;
				}
			}
		} while (sp2);
		if (cmd_invalid)
			return -EINVAL;
	} else { //get
		pos += sprintf(extra + pos, "\n\n%s:\n\tversion:%s\n"
			 "\tirq:%d\n\tdevid:%d\n\tchannels:%d\n"
			 "\tidx:%d\n\tuse_chanctx:%d\n\tportid:%d\n\tscan_addr:%pM\n"
			 "\tscanning:%d\n\troc:%d\n"
			 "\n\ttx sched time: %luns\n"
			 "\ttx amsdu %sabled\n"
			 "\trx amsdu %sabled\n",
			 wiphy_name(hw_priv->hw->wiphy),
			 hw_priv->version_str,
			 hw_priv->irq,
			 hw_priv->devid,
			 hw_priv->channels,
			 hw_priv->idx,
			 hw_priv->use_chanctx,
			 hw_priv->portid,
			 hw_priv->scanner.addr,
			 hw_priv->scanner.scanning,
			 hw_priv->roc.running,
			 ewma_tx_sched_time_read(&hw_priv->sched_time),
			 _ieee80211_hw_check(hw, IEEE80211_HW_TX_AMSDU) ? "en" : "dis",
			 _ieee80211_hw_check(hw, IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU) ?
			 "en" : "dis");
		if (is_pcie_device(hw)) {
			pos += sprintf(extra + pos, "\n");
			switch (hw_priv->space) {
			case MT_HST_BD_HST:
			pos += sprintf(extra + pos,
				"\tmpdu table [Host], buffer desc [Host]\n");
			break;
			case MT_HST_BD_TGT:
			pos += sprintf(extra + pos,
				"\tmpdu table [Host], buffer desc [Target]\n");
			break;
			case MT_TGT_BD_TGT:
			pos += sprintf(extra + pos,
				"\tmpdu table [Target], buffer desc [Target]\n");
			break;
			default:
			pos += sprintf(extra + pos, "\tinvalid postion\n");
			break;
			}
			pos += sprintf(extra + pos, "\twin-lost happened %llu times\n",
				hw_priv->total_winlost);
		}
		memset(vif_buf, 0, sizeof(vif_buf));
		ieee80211_iterate_active_interfaces(hw,
							IEEE80211_IFACE_ITER_ACTIVE,
							scm80211_vif_info_show, vif_buf);
		pos += sprintf(extra + pos, "%s", vif_buf);
		memset(vif_buf, 0, sizeof(vif_buf));
	}
	ret = strlen(extra);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
	return 0;
}

static void scm80211_sta_rate_fix(void *data, struct ieee80211_sta *sta)
{
	char *str = data;
	char cmd[100] = { 0 };
	char param[32] = { 0 };
	char u_addr[32] = { 0 };
	char addr_str[32] = { 0 };
	char *sp1 = NULL;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_dbg_rate_fix *dbg_rate = &sta_priv->rate_fix;
	enum ppdu_format fmt = NON_HT;
	enum cbw bw = CBW20;
	enum gi_type gtype = LONG_GI;
	enum ltf_type ltf_type = HE_LTF_4x;
	u8 rate_idx = 0xFF;
	u8 nss = 1;
	bool prot = false;
	bool is_ldpc = false;
	unsigned long flag;
	int param_num = 0;
	size_t pos = 0;
	int idx = 0;
	bool rate_valid = false, disable = false, param_invalid = false;
	bool match = false;

	if (*str == '\t')
		return;
	strscpy(cmd, data, strlen(data) + 1);
	sprintf(addr_str, "%02X:%02X:%02X:%02X:%02X:%02X",
		sta->addr[0], sta->addr[1], sta->addr[2],
		sta->addr[3], sta->addr[4], sta->addr[5]);
	if (sta_priv->rate_fix_on) {
		//load initial rate parameters
		fmt = txvector_format(&dbg_rate->rate);
		rate_idx = ieee80211_rate_get_vht_mcs(&dbg_rate->rate);
		bw = txvector_ch_bandwidth(&dbg_rate->rate);
		is_ldpc = dbg_rate->flags & IEEE80211_TX_CTL_LDPC ? true : false;
		if (fmt < HE_SU) {
			if (fmt != NON_HT) {
				gtype = dbg_rate->rate.flags & IEEE80211_TX_RC_SHORT_GI ?
						SHORT_GI : LONG_GI;
			}
		} else { //HE
			gtype = dbg_rate->he_gi_type;
			ltf_type = dbg_rate->he_ltf_type;
		}
		if (fmt >= VHT)
			nss = ieee80211_rate_get_vht_nss(&dbg_rate->rate);
	}
	do {
		sp1 = strnchr(cmd + idx, strlen(cmd) - idx, ',');
		memset(param, '\0', sizeof(param));
		if (sp1) {
			strscpy(param, cmd + idx, sp1 - (cmd + idx) + 1);
			sp1++;
			idx = sp1 - cmd;
		} else
			strscpy(param, cmd + idx, strlen(cmd + idx) + 1);
		param_num++;
		if (idx >= strlen(cmd))
			break;
		if (param_num == 1) { //mac addr
			uppercase(u_addr, param);
			if (!strncmp(addr_str, u_addr, strlen(addr_str)))
				match = true;
		} else if (param_num == 2) { //fmt
			if (!strncmp(param, "fmt=", 4)) {
				if (!strcmp(param + 4, "NON-HT")) {
					fmt = NON_HT;
				} else if (!strcmp(param + 4, "HT")) {
					fmt = HT_MF;
				} else if (!strcmp(param + 4, "VHT")) {
					fmt = VHT;
				} else if (!strcmp(param + 4, "HE")) {
					fmt = HE_SU;
				} else {
					param_invalid = true;
					pos += sprintf(str + pos, "\tPPDU format error\n");
					break;
				}
			} else if (!strcmp(param, "disable")) {
				disable = true;
			} else {
				param_invalid = true;
				pos += sprintf(str + pos, "\tParam invalid\n");
				break;
			}
		} else if (param_num == 3) { //rate idx
			if (!strncmp(param, "rate=", 5)) {
				if (sscanf(param + 5, "%hhd", &rate_idx) != 1) {
					param_invalid = true;
					pos += sprintf(str + pos, "\tRate index setting error\n");
					break;
				}
			} else {
				param_invalid = true;
				pos += sprintf(str + pos, "\tParam invalid\n");
				break;
			}
		} else {
			if (!strncmp(param, "prot=", 5)) {
				if (!strcmp(param + 5, "1"))
					prot = true;
			} else if (!strncmp(param, "bw=", 3)) {
				if (!strcmp(param + 3, "20M")) {
					bw = CBW20;
				} else if (!strcmp(param + 3, "40M")) {
					bw = CBW40;
				} else if (!strcmp(param + 3, "80M")) {
					bw = CBW80;
				} else {
					param_invalid = true;
					pos += sprintf(str + pos, "\tUnsupported bw\n");
					break;
				}
			} else if (!strncmp(param, "gi=", 3)) {
				if (fmt == HE_SU) {
					if (!strcmp(param + 3, "0u8s")) {
						gtype = GI_0u8s;
					} else if (!strcmp(param + 3, "1u6s")) {
						gtype = GI_1u6s;
					} else if (!strcmp(param + 3, "3u2s")) {
						gtype = GI_3u2s;
					} else {
						param_invalid = true;
						pos += sprintf(str + pos, "\tInvalid GI type\n");
						break;
					}
				} else {
					if (!strcmp(param + 3, "SGI")) {
						gtype = SHORT_GI;
					} else if (!strcmp(param + 3, "LGI")) {
						gtype = LONG_GI;
					} else {
						param_invalid = true;
						pos += sprintf(str + pos, "\tInvalid GI type\n");
						break;
					}
				}
			} else if (!strncmp(param, "code=", 5)) {
				if (!strcmp(param + 5, "LDPC"))
					is_ldpc = true;
			} else if (!strncmp(param, "nss=", 4)) {
				if (sscanf(param + 4, "%hhd", &nss) != 1) {
					param_invalid = true;
					pos += sprintf(str + pos, "\tNss setting error\n");
					break;
				}
				if (nss > 2) {
					param_invalid = true;
					pos += sprintf(str + pos, "\tInvalid nss setting\n");
					break;
				}
			} else if (!strncmp(param, "ltf=", 4)) {
				if (!strncmp(param + 4, "1x", 2))
					ltf_type = HE_LTF_1x;
				else if (!strncmp(param + 4, "2x", 2))
					ltf_type = HE_LTF_2x;
			}
		}
		if (match)
			memset(str, 0, strlen(str));
		else
			break;
	} while (sp1);
	if (!match)
		return;
	spin_lock_irqsave(&sta_priv->lock, flag);
	if ((!disable) && (!param_invalid))
		rate_valid = scm80211_dbg_construct_rate(sta, fmt, rate_idx,
							 bw, nss, gtype, ltf_type, is_ldpc);
	if (rate_valid && (!param_invalid)) {
		sta_priv->rate_fix_on = true;
		pos += sprintf(str + pos, "\n\tRate setting success turn on rate fix flag\n");
	} else {
		if (disable) {
			sta_priv->rate_fix_on = false;
			pos += sprintf(str + pos, "\n\tTurn off rate fix flag\n");
		} else {
			pos += sprintf(str + pos, "\n\tRate setting fail please check the param\n");
		}
	}
	spin_unlock_irqrestore(&sta_priv->lock, flag);
}

static void scm80211_sta_rate_show(void *data, struct ieee80211_sta *sta)
{
	char *str = data;
	char u_str[32] = { 0 };
	char addr_str[32] = { 0 };
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_dbg_rate_fix *dbg_rate = &sta_priv->rate_fix;
	int rateidx;
	enum ppdu_format fmt;
	enum cbw bw;
	unsigned long flag;
	size_t pos = 0;

	if (*str == '\t')
		return;
	sprintf(addr_str, "%02X:%02X:%02X:%02X:%02X:%02X",
		sta->addr[0], sta->addr[1], sta->addr[2],
		sta->addr[3], sta->addr[4], sta->addr[5]);
	uppercase(u_str, str);
	if (!strncmp(addr_str, u_str, strlen(addr_str))) {
		memset(str, 0, strlen(str));
		spin_lock_irqsave(&sta_priv->lock, flag);
		if (sta_priv->rate_fix_on) {
			rateidx = ieee80211_rate_get_vht_mcs(&dbg_rate->rate);
			fmt = txvector_format(&dbg_rate->rate);
			bw = txvector_ch_bandwidth(&dbg_rate->rate);
			pos += sprintf(str + pos, "\tFix rate:\n");
			pos += sprintf(str + pos, "\tFmt = %s\n", rate_format_str[fmt]);
			if (fmt == NON_HT) {
				if (sta_priv->band == NL80211_BAND_5GHZ)
					rateidx += 4;
				pos += sprintf(str + pos,
					"\tRate = %s", non_ht_rate_str[rateidx]);
				if (dbg_rate->rate.flags & IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
					pos += sprintf(str + pos, "(short preamble)\n");
				else
					pos += sprintf(str + pos, "\n");
				goto exit;
			} else
				pos += sprintf(str + pos, "\tRate = MCS%d\n", rateidx);
			pos += sprintf(str + pos, "\tBW = %s\n",
				bw == CBW20 ? "20Mhz" :
				bw == CBW40 ? "40Mhz" : bw == CBW80 ? "80Mhz" : "invalid value");
			pos += sprintf(str + pos, "\tCode type = %s\n",
				dbg_rate->flags & IEEE80211_TX_CTL_LDPC ? "LDPC" : "BCC");
			if (fmt != HE_SU) {
				pos += sprintf(str + pos, "\tGI type = %s\n",
					dbg_rate->rate.flags & IEEE80211_TX_RC_SHORT_GI ?
					"Short GI" : "Long GI");
				if (fmt == HT_MF)
					goto exit;
			}
			pos += sprintf(str + pos, "\tNSS = %d\n",
				ieee80211_rate_get_vht_nss(&dbg_rate->rate));
			if (fmt == VHT)
				goto exit;
			pos += sprintf(str + pos, "\tGI type = %s\n",
				dbg_rate->he_gi_type == GI_0u8s ? "GI_0u8s" :
				dbg_rate->he_gi_type == GI_1u6s ? "GI_1u6s" : "GI_3u2s");
			pos += sprintf(str + pos, "\tHE-LTF mode: %s HE-LTF\n",
				dbg_rate->he_ltf_type == HE_LTF_1x ? "1x" :
				dbg_rate->he_ltf_type == HE_LTF_2x ? "2x" : "4x");
		} else
			pos += sprintf(str + pos, "\tRA: Controlled by mac80211\n");
exit:
		spin_unlock_irqrestore(&sta_priv->lock, flag);
	}
}

/**
 * [mac]: the MAC address of associated station
 * fmt=NON-HT/HT/VHT/HE
 * rate=xx
 * bw=20M/40M/80M
 * gi=SGI/LGI/0u4s/0u8s/1u6s/3u2s
 * code=BCC/LDPC
 * nss=x -> for vht and he
 * ltf=1x/2x/4x -> only for he
 * [mac], fmt, rate must the first, second and third
 * e.g
 * get station rate information: iwpriv wlan0 rate "[mac]"
 * set station rate: iwpriv wlan0 rate
 * "[mac],fmt=HE,rate=8,bw=20M,gi=0u8s,code=BCC,nss=2,ltf=1x"
 */
static int scm80211_iwpriv_rate(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	struct ieee80211_hw *hw = NULL;
	struct ieee80211_hw_priv *hw_priv = NULL;
	char cmd[100] = { 0 };
	char sta_buf[512] = { 0 };
	char param[256];
	char *sp1 = NULL, *sp2 = NULL;
	size_t pos = 0;
	int ret;

	if (copy_from_user(cmd, wrqu->data.pointer, wrqu->data.length))
		return 0;
	pr_debug("[+ %s]str(%d): %s\n", __func__, wrqu->data.length, cmd);
	wrqu->data.length = 0;
	hw = wiphy_to_ieee80211_hw(dev->ieee80211_ptr->wiphy);
	hw_priv = hw->priv;
	sp1 = cmd;
	if (*sp1 != '\0') {
		do {
			sp2 = strnchr(cmd, strlen(cmd), ',');
			memset(param, '\0', sizeof(param));
			if (!sp2) { //get station rate information
				strscpy(param, cmd, strlen(cmd) + 1);
				pos += sprintf(extra + pos, "\n\nGet rate information about %s:\n",
					param);
				memcpy(sta_buf, param, sizeof(param));
				ieee80211_iterate_stations_atomic(hw,
					scm80211_sta_rate_show, sta_buf);
				if (strncmp(sta_buf, param, strlen(param)))
					pos += sprintf(extra + pos, "%s", sta_buf);
				else
					pos += sprintf(extra + pos,
						"\tThis station is not associated!\n");
			} else { //set station rate
				strscpy(param, cmd, strlen(cmd) + 1);
				memcpy(sta_buf, param, sizeof(param));
				ieee80211_iterate_stations_atomic(hw,
					scm80211_sta_rate_fix, sta_buf);
				if (strncmp(sta_buf, param, strlen(param)))
					pos += sprintf(extra + pos, "%s", sta_buf);
				else
					pos += sprintf(extra + pos,
						"\tThis station is not associated!\n");
				break;
			}
		} while (sp2);
	} else
		pos += sprintf(extra + pos, "\n\nCommand format error.\n");
	ret = strlen(extra);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
	return 0;
}

/**
 * e.g
 * get: iwpriv wlan0 help
 */
static int scm80211_iwpriv_help(struct net_device *dev,
				       struct iw_request_info *info, union iwreq_data *wrqu,
				       char *extra)
{
	size_t pos = 0;
	int ret;

	wrqu->data.length = 0;
	pos += sprintf(extra + pos,
					"\n\nCommands:\n"
					"\tstate:\tget interface status\n"
					"\trate:\tfix rate operation\n");
#ifdef SCM80211_IWPRIV_DEBUG
	pos += sprintf(extra + pos,
					"\treg:\tget/set register value\n"
					"\ttrace:\tget/set debug trace\n");
#endif
	ret = strlen(extra);
	if (ret > 0)
		wrqu->data.length = ret;
	else
		memset(extra, 0, strlen(extra));
	return 0;
}
#endif

static const iw_handler scm80211_private_handler[] = {
#ifdef SCM80211_FTM
	/* cmd template: iwpriv wlan0 set ssid */
	(iw_handler)scm80211_iwpriv_set_process,
	/* cmd template: iwpriv wlan0 get ssid */
	(iw_handler)scm80211_iwpriv_get_process,
#else
	NULL,
	NULL,
	NULL,
	(iw_handler)scm80211_iwpriv_help,
#ifdef SCM80211_IWPRIV_DEBUG
	NULL,
	(iw_handler)scm80211_iwpriv_debug_reg,
	NULL,
	(iw_handler)scm80211_iwpriv_debug_trace,
#else
	NULL,
	NULL,
	NULL,
	NULL,
#endif
	NULL,
	(iw_handler)scm80211_iwpriv_state,
	NULL,
	(iw_handler)scm80211_iwpriv_rate,
#endif
};

static const struct iw_priv_args scm80211_private_args[] = {
#ifdef SCM80211_FTM
	{
	 .cmd = SCM80211_IOCTL_SET,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "set" },
	{
	 .cmd = SCM80211_IOCTL_GET,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .get_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "get" },
#else
	{
	 .cmd = SCM80211_IOCTL_HELP,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .get_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "help" },
#ifdef SCM80211_IWPRIV_DEBUG
	{
	 .cmd = SCM80211_IOCTL_DEBUG_REG,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .get_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "reg" },
	{
	 .cmd = SCM80211_IOCTL_DEBUG_TRACE,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .get_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "trace" },
#endif
	{
	 .cmd = SCM80211_IOCTL_STATE,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .get_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "state" },
	{
	 .cmd = SCM80211_IOCTL_RATE,
	 .set_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .get_args = IW_PRIV_TYPE_CHAR | MAX_CMD_STR,
	 .name = "rate" },
#endif
};

static struct iw_statistics *scm80211_get_wireless_stats(struct net_device *dev)
{
	static struct iw_statistics statistics = { 0 };

	return &statistics;
}

static const struct iw_handler_def scm80211_iw_handlers = {
	.num_standard = ARRAY_SIZE(scm_wx_handlers),
	.num_private = ARRAY_SIZE(scm80211_private_handler),
	.num_private_args = ARRAY_SIZE(scm80211_private_args),
	.standard = (iw_handler *)scm_wx_handlers,
	.private = (iw_handler *)scm80211_private_handler,
	.private_args = (struct iw_priv_args *)scm80211_private_args,
	.get_wireless_stats = scm80211_get_wireless_stats,
};


int scm80211_net_dev_attach(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct wireless_dev *wdev;

	if (!vif)
		return 0;

	list_for_each_entry(wdev, &hw->wiphy->wdev_list, list) {
		if (!wdev->netdev)
			continue;
		if (wdev->netdev->wireless_handlers)
			continue;
		wdev->netdev->wireless_handlers = &scm80211_iw_handlers;
	}

#ifdef SCM80211_FTM
	FTM_DBG(hw, "[%s] completed!\n", __func__);
#endif

	return 0;
}
