// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#define DEBUG

#include <linux/pci.h>
#include <linux/aer.h>
#include <linux/rbtree.h>
#include <linux/firmware.h>
#include <linux/bitfield.h>
#include <linux/version.h>
#include <linux/crc32.h>
#include <linux/version.h>
#if KERNEL_VERSION(5, 4, 0) > LINUX_VERSION_CODE
#include <linux/pci-aspm.h>
#endif
#include <linux/io.h>
#include <linux/iommu.h>

#define func_fmt(str)           "[PCIE] "str
#define func_trace              BUS

#include "trace.h"
#include "core.h"
#include "rf.h"
#include "util.h"
#include "auto_version.h"
#include "debug.h"

#ifdef CONFIG_HTC_IF
#include "htc.h"
#endif

#ifdef SCM80211_FTM
#include "ftm.h"
#endif

#ifdef EFUSE_ENABLE
#include "efuse.h"
#endif

static int dbdc;		//Dual Band Dual Concurrent
module_param(dbdc, int, 0444);
MODULE_PARM_DESC(dbdc, "Dual Band Dual Concurrent");

static bool pdet_en;
module_param(pdet_en, bool, 0644);
MODULE_PARM_DESC(pdet_en, "enable power detect feature");

static bool iommu_en;
module_param(iommu_en, bool, 0644);
MODULE_PARM_DESC(iommu_en, "check OS IOMMU feature and set DMA mask according to this feature");

/*
 * FIXME:
 * - Host to target interrupt
 */

#define PCIE_REG_ALIGN		BIT(12)	/* 4KB */

#define MCU_START_RETRIES	10
#define MCU_START_DELAY		1

/*
 * Designware specifics copied from drivers/pci/controller/dwc
 */

/* Parameters for the waiting for iATU enabled routine */
#define LINK_WAIT_MAX_IATU_RETRIES	5
#define LINK_WAIT_IATU			9

/*
 * iATU Unroll-specific register definitions
 * From 4.80 core version the address translation will be made by unroll
 */
#define PCIE_ATU_UNR_REGION_CTRL1	0x00
#define PCIE_ATU_UNR_REGION_CTRL2	0x04
#define PCIE_ATU_UNR_LOWER_BASE		0x08
#define PCIE_ATU_UNR_UPPER_BASE		0x0C
#define PCIE_ATU_UNR_LIMIT		0x10
#define PCIE_ATU_UNR_LOWER_TARGET	0x14
#define PCIE_ATU_UNR_UPPER_TARGET	0x18

#define PCIE_ATU_WIRE_OFFSET		0x2000

/* Register address builder */
#define PCIE_GET_ATU_OUTB_WIRE_REG_OFFSET(region) \
	(PCIE_ATU_WIRE_OFFSET + ((region) << 9))

#define PCIE_GET_ATU_INB_WIRE_REG_OFFSET(region) \
	(PCIE_ATU_WIRE_OFFSET + (((region) << 9) | BIT(8)))

/* Maximum number of inbound/outbound iATUs */
#define MAX_IATU_IN			256
#define MAX_IATU_OUT			256

#define PCIE_ATU_TYPE_MEM               0x0
#define PCIE_ATU_TYPE_IO                0x2
#define PCIE_ATU_TYPE_CFG0              0x4
#define PCIE_ATU_TYPE_CFG1              0x5
#define PCIE_ATU_ENABLE			BIT(31)
#define PCIE_ATU_BAR_MODE_ENABLE        BIT(30)

//fix the issue about PCIe BAR4 during downloading firmware
#define DLER_CTRL_MSG_OFFSET		0x000F4000
#define WAIT_TIME			2
#define WAIT_CNT			500
#define RETRY_CNT			3

/**
 * struct pci_addr_mapping - pci-to-local address mapping
 * @rb: rbtree node
 * @pci_addr: pci address aligned to address translation size.
 * @ep_addr: endpoint local address
 * @atu: index to outbound atu holding the translation
 * @refcount: reference count
 */
struct pcie_addr_mapping {
	struct rb_node	rb;
	dma_addr_t	pci_addr;
	phys_addr_t	ep_addr;
	int		atu;
	atomic_t	refcount;
};

/**
 * struct pcie_endpoint_config - the configuration for pcie ep
 * @firmare: HWMAC fw
 * @wise: soc fw
 * @axi_base: base address for axi
 * @axi_size: total size for axi
 * @num_atu: max num of atus
 */
struct pcie_endpoint_config {
	const char	*firmware;
	const char	*wise;
	phys_addr_t	axi_base;
	size_t		axi_size;
	size_t		num_atu;
};

/**
 * struct pcie_endpoint - the private data for pcie device
 * @config: pointer to the driver data configure
 * @hws: pointer to each hwmac
 * @bitmap: outbound ATU allocation bitmap
 * @page_size; size of a page (should be a power of 2)
 * @page_shift: log2(@page_shift)
 * @bp_base: bp_address
 * @last_bp_base: target address where bp bar point to
 * @intf_num: num of hwmac
 * @root: rb tree root
 *
 * A page is a unit of PCI to local AXI address translation. Its size is equal
 * to @size/@nr_atus in case of uniform page.
 *
 */
#define HTC_H2T_NR	8
#define HTC_T2H_NR	4
struct pcie_endpoint {
	const struct pcie_endpoint_config *config;
	void		*hws[SCM2020_INTF_MAX];
	atomic_t	ref;
	unsigned long	bitmap[256 / BITS_PER_LONG];
	size_t		page_size;
	unsigned int	page_shift;
	void __iomem	*pcie_base;
	void __iomem	*mac_base;
	void __iomem	*phy_base;
	void __iomem	*mem_base;
	phys_addr_t	ep_mem_base;
	void __iomem	*bp_base;
	phys_addr_t	last_bp_base;
	u32		elbi_offset;
	u32		intf_num;
	struct pci_dev	*pdev;
	struct rb_root	root;
	spinlock_t	lock;
	int		num_atu;
	u32		barmask;
	atomic_t	irq_chan;
	struct htc_content __iomem *h2t[HTC_H2T_NR];
	struct htc_content __iomem *t2h[HTC_T2H_NR];

#ifdef CONFIG_HTC_IF
	struct scm80211_htc_ctx	htc_ctx;
#endif
	bool		doze;
	u32		link_cotrol[2];
};

/**
 * struct hw_reg_section - iomem for mac,phy
 *
 * @mac_base: iomapped for mac
 * @phy_base: iomapped for phy
 * @pdev: pci dev address
 */
struct hw_reg_section {
	void __iomem    *mac_base;
	void __iomem	*phy_base;
	u32		rf_base;
	u32		pta_base;
	struct pci_dev  *pdev;
	struct shared_mem __iomem *mem;
};

/* Registers to be saved and restored */

static u32 mac_doze_awake_regs[] = {
	MAC_REG_DEV_CFG,
	MAC_REG_DEV_OPT1,
	MAC_REG_DEV_CS,
	MAC_REG_DEV_CH_COEX,
	MAC_REG_DEV_OPT2,

	MAC_REG_RX_CFG,
	MAC_REG_RX_SCB_CTRL,
	MAC_REG_RX_TO_CFG,
	MAC_REG_RX_TO_11B_CFG, REG_PHYRX_DELAY,

	MAC_REG_INTR_ENABLE,    /* enable interrupts */
	MAC_REG_RX_BUF_CFG,     /* enable RX */
	MAC_REG_MAC_CFG,        /* enable MAC */
};

static u32 mac_rxring_regs[] = {
	MAC_REG_RX_BUF_PTR0,
	MAC_REG_RX_BUF_PTR1,
};

static u32 mac_assoc_active_regs[] = {
	MAC_REG_DEV_PMU,

	MAC_REG_TX_TO_CFG,
	MAC_REG_TX_TO_11B_CFG,
	MAC_REG_TX_CFG,

	MAC_REG_OBSS_SR_NON_SRG_CFG,
	MAC_REG_OBSS_SR_NON_SRG_COLOR_EN0,
	MAC_REG_OBSS_SR_NON_SRG_COLOR_EN1,
	MAC_REG_OBSS_SR_NON_SRG_BSSID_EN0,
	MAC_REG_OBSS_SR_NON_SRG_BSSID_EN1,
	MAC_REG_SRG_CFG,
	MAC_REG_SRG_MAP_COLOR0,
	MAC_REG_SRG_MAP_COLOR1,
	MAC_REG_SRG_SR_COLOR_EN0,
	MAC_REG_SRG_SR_COLOR_EN1,
	MAC_REG_SRG_SR_BSSID_EN0,
	MAC_REG_SRG_SR_BSSID_EN1,
	MAC_REG_SR_CFG,
	MAC_REG_CTS_TX_PWR,
};

static u32 mac_vif_assoc_awake_regs[2][14] = {
	{
		MAC_REG_VIF_MAC0_L(0),
		MAC_REG_VIF_MAC0_H(0),
		MAC_REG_VIF_MAC1_L(0),
		MAC_REG_VIF_MAC1_H(0),
		MAC_REG_VIF_BSS0_L(0),
		MAC_REG_VIF_BSS0_H(0),
		MAC_REG_VIF_BSS1_L(0),
		MAC_REG_VIF_BSS1_H(0),
		MAC_REG_VIF_AID(0),
		MAC_REG_VIF_RX_FILTER0(0),
		MAC_REG_VIF_RX_FILTER1(0),
		MAC_REG_VIF_MBSSID_MASK_L32(0),
		MAC_REG_VIF_MBSSID_MASK_H16(0),
		MAC_REG_VIF_CFG0(0),
	},
	{
		MAC_REG_VIF_MAC0_L(1),
		MAC_REG_VIF_MAC0_H(1),
		MAC_REG_VIF_MAC1_L(1),
		MAC_REG_VIF_MAC1_H(1),
		MAC_REG_VIF_BSS0_L(1),
		MAC_REG_VIF_BSS0_H(1),
		MAC_REG_VIF_BSS1_L(1),
		MAC_REG_VIF_BSS1_H(1),
		MAC_REG_VIF_AID(1),
		MAC_REG_VIF_RX_FILTER0(1),
		MAC_REG_VIF_RX_FILTER1(1),
		MAC_REG_VIF_MBSSID_MASK_L32(1),
		MAC_REG_VIF_MBSSID_MASK_H16(1),
		MAC_REG_VIF_CFG0(1),
	},
};

static u32 mac_vif_assoc_active_regs[2][13] = {
	{
		MAC_REG_VIF_CFG1(0),
		MAC_REG_VIF_BFM0(0),
		MAC_REG_VIF_BFM1(0),
		MAC_REG_VIF_BSR(0),
		MAC_REG_VIF_MIN_MPDU_SPACING(0),
		MAC_REG_VIF_QNULL_CTRL_AC0(0),
		MAC_REG_VIF_QNULL_CTRL_AC1(0),
		MAC_REG_VIF_QNULL_CTRL_AC2(0),
		MAC_REG_VIF_QNULL_CTRL_AC3(0),
		MAC_REG_VIF_UORA_CFG(0),
		MAC_REG_VIF_UORA_CFG1(0),
		MAC_REG_VIF_NON_SRG_OBSS_PD(0),
		MAC_REG_VIF_SRG_OBSS_PD(0),
	},
	{
		MAC_REG_VIF_CFG1(1),
		MAC_REG_VIF_BFM0(1),
		MAC_REG_VIF_BFM1(1),
		MAC_REG_VIF_BSR(1),
		MAC_REG_VIF_MIN_MPDU_SPACING(1),
		MAC_REG_VIF_QNULL_CTRL_AC0(1),
		MAC_REG_VIF_QNULL_CTRL_AC1(1),
		MAC_REG_VIF_QNULL_CTRL_AC2(1),
		MAC_REG_VIF_QNULL_CTRL_AC3(1),
		MAC_REG_VIF_UORA_CFG(1),
		MAC_REG_VIF_UORA_CFG1(1),
		MAC_REG_VIF_NON_SRG_OBSS_PD(1),
		MAC_REG_VIF_SRG_OBSS_PD(1),
	},
};

static u32 mac_vif_ps_active_regs[2][12] = {
	{
		MAC_REG_VIF_CFG0(0),
		MAC_REG_VIF_TSF_OFFSET_L(0),
		MAC_REG_VIF_TSF_OFFSET_H(0),
		MAC_REG_VIF_TRIG_RESP_QUEUE_EN(0),
		MAC_REG_VIF_BSR(0),
		MAC_REG_VIF_TX_PWR(0),
		MAC_REG_VIF_CFG_INFO(0),
		MAC_REG_VIF_TRIG(0),
		MAC_REG_VIF_ESP(0),
		MAC_REG_VIF_HTC_OMI(0),
	},
	{
		MAC_REG_VIF_CFG0(1),
		MAC_REG_VIF_TSF_OFFSET_L(1),
		MAC_REG_VIF_TSF_OFFSET_H(1),
		MAC_REG_VIF_TRIG_RESP_QUEUE_EN(1),
		MAC_REG_VIF_BSR(1),
		MAC_REG_VIF_TX_PWR(1),
		MAC_REG_VIF_CFG_INFO(1),
		MAC_REG_VIF_TRIG(1),
		MAC_REG_VIF_ESP(1),
		MAC_REG_VIF_HTC_OMI(1),
	},
};

static u32 phy_assoc_active_regs[] = {
	PHY_MODEM(FW_VHT_CONFIG0_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG1_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG2_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG3_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG4_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG5_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG6_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG7_VINTF(0)),
	PHY_MODEM(FW_VHT_CONFIG8_VINTF(0)),
	PHY_MODEM(FW_HE_CONFIG1_VINTF(0)),
	PHY_MODEM(FW_HE_CONFIG2_VINTF(0)),
	PHY_MODEM(FW_11B_CSFO_CONV_VINTF(0)),
	PHY_MODEM(FW_11B_PLLADJ_VINTF(0)),
	PHY_MODEM(FW_11B_TX_PLLADJ_VINTF(0)),
	PHY_MODEM(PHY_11B_CS_PARAM),
	PHY_MODEM(PHY_CS_CONFIG_1),
	PHY_MODEM(PHY_CS_CONFIG_2),
};

#define FW_BASE				0x20000000
#define FW_MAX_SIZE			0xC0000	//768K
#define SRAM_MEM_SIZE                   0x100000	//1M

#if defined(SCM2020_Z1) || defined(SCM2020_A0)

#define CTRL_MEM_SIZE			0x1000	//4K
#define SC2_MSG_SIZE			64
#define SC2_DATA_SIZE			(CTRL_MEM_SIZE - 64)
#define VERSION_MAX_LEN			64
#define SC2_CRC32_INIT			0xFFFFFFFFU

#define CMD_MAP				"MAP"
#define CMD_BOOT			"BOOT"
#define CMD_CRC				"CRC"
#define CMD_VER				"VER"
#define CMD_END				"END"
#define CMD_PREAMBLE			"UUUU"

#define RESP_OK				"OKAY"
#define RESP_FAIL			"FAIL"
#define RESP_UPCI			"UPCI"

struct dler_ctrl {
	char msg[SC2_MSG_SIZE];
	char data[SC2_DATA_SIZE];
};
#endif

static phys_addr_t pcie_ep_mem_addr(struct pcie_endpoint *ep, int bar, void *addr);
/**
 * pcie_endpoint_init() - alloc and init driver private data
 * configure ATU infoman
 */
static int pcie_endpoint_init(struct pci_dev *pdev, const struct pci_device_id *id, u32 mask)
{
	struct pcie_endpoint *ep = devm_kzalloc(&pdev->dev, sizeof(*ep), GFP_KERNEL);
	int i;
	u32 v, shift;

	if (!ep) {
		dev_err(&pdev->dev, "%s: alloc mem fail size=%zu\n",
			__func__, sizeof(*ep));
		return -ENOMEM;
	}
	ep->config = (const struct pcie_endpoint_config *) id->driver_data;
	atomic_set(&ep->ref, 0);
	atomic_set(&ep->irq_chan, -1);
	bitmap_zero(ep->bitmap, ep->config->num_atu);
	ep->page_size = ep->config->axi_size / ep->config->num_atu;
	ep->page_shift = ilog2(ep->page_size);
	ep->intf_num = dbdc ? 2 : 1;
	ep->pdev = pdev;
	ep->root = RB_ROOT;
	/*
	 * - BAR0: Designware PCIe CDM register (not configurable)
	 * - BAR1: MAC0+MAC1 register
	 * - BAR2: PHY register
	 * - BAR3: target memory
	 * - BAR4: Generic backplane register/memory access
	 */
	ep->pcie_base   = pcim_iomap_table(pdev)[0];
	ep->mac_base	= pcim_iomap_table(pdev)[1];
	ep->phy_base    = pcim_iomap_table(pdev)[2];
	ep->mem_base    = pcim_iomap_table(pdev)[3];
	ep->ep_mem_base = pcie_ep_mem_addr(ep, 3, ep->mem_base);
	if (mask & BIT(4)) {
		ep->bp_base = pcim_iomap_table(pdev)[4];
		ep->last_bp_base = pcie_ep_mem_addr(ep, 4, ep->bp_base);
	} else {
		ep->last_bp_base = 0;
	}
	ep->barmask	= mask;

#ifdef SCM2020_A0
	for (i = 0; i < HTC_H2T_NR; i++) {
		shift = SRAM_MEM_SIZE - (i + 1) * sizeof(struct htc_content);
		ep->h2t[i] = ep->mem_base + shift;
		dev_dbg(&pdev->dev, "h2t[%d] ioaddr %lx, target addr %lx\n", i,
			(unsigned long) ep->h2t[i],
			(unsigned long) pcie_ep_mem_addr(ep, 3, ep->h2t[i]));
	}

	for (i = 0; i < HTC_T2H_NR; i++) {
		shift = SRAM_MEM_SIZE - (i + HTC_H2T_NR + 1) * sizeof(struct htc_content);
		ep->t2h[i] = ep->mem_base + shift;
		dev_dbg(&pdev->dev, "t2h[%d] ioaddr %lx, target addr %lx\n", i,
			(unsigned long) ep->t2h[i],
			(unsigned long) pcie_ep_mem_addr(ep, 3, ep->t2h[i]));
	}
#endif

	pci_read_config_dword(pdev, 0x700 + 0x1bc, &v);
	ep->elbi_offset = FIELD_GET(GENMASK(17, 8), v) * 4;
	ep->doze = false;
	spin_lock_init(&ep->lock);
	pci_set_drvdata(pdev, ep);

	dev_dbg(&pdev->dev, "%s success\n", __func__);
	return 0;
}


static u32 pcie_readl_outbound_atu(struct pcie_endpoint *ep, u32 index, u32 reg)
{
	u32 offset = PCIE_GET_ATU_OUTB_WIRE_REG_OFFSET(index);

	return readl(ep->pcie_base + offset + reg);
}

static void pcie_writel_outbound_atu(struct pcie_endpoint *ep, u32 index, u32 reg, u32 val)
{
	u32 offset = PCIE_GET_ATU_OUTB_WIRE_REG_OFFSET(index);

	writel(val, ep->pcie_base + offset + reg);
}

static void pcie_program_outbound_atu(struct pcie_endpoint *ep, int index,
				      int type, __le64 ep_addr, dma_addr_t pci_addr, size_t size)
{
	u32 retries, val;

	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_LOWER_BASE, lower_32_bits(ep_addr));
	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_UPPER_BASE, upper_32_bits(ep_addr));
	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_LIMIT, lower_32_bits(ep_addr + size - 1));
	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_LOWER_TARGET, lower_32_bits(pci_addr));
	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_UPPER_TARGET, upper_32_bits(pci_addr));
	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL1, type);
	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2, PCIE_ATU_ENABLE);

	/*
	 * Make sure ATU enable takes effect before any subsequent config
	 * and I/O accesses.
	 */
	for (retries = 0; retries < LINK_WAIT_MAX_IATU_RETRIES; retries++) {
		val = pcie_readl_outbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2);
		if (val & PCIE_ATU_ENABLE)
			return;

		mdelay(LINK_WAIT_IATU);
	}
	dev_err(&ep->pdev->dev, "Outbound iATU is not being enabled\n");
}

static void pcie_disable_outbound_atu(struct pcie_endpoint *ep, int index)
{
	u32 retries, val;

	pcie_writel_outbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2, (u32)~PCIE_ATU_ENABLE);
	for (retries = 0; retries < LINK_WAIT_MAX_IATU_RETRIES; retries++) {
		val = pcie_readl_outbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2);
		if (!(val & PCIE_ATU_ENABLE))
			return;

		mdelay(LINK_WAIT_IATU);
	}
	dev_err(&ep->pdev->dev, "Outbound iATU is not being enabled\n");
}

static u32 pcie_readl_inbound_atu(struct pcie_endpoint *ep, u32 index, u32 reg)
{
	u32 offset = PCIE_GET_ATU_INB_WIRE_REG_OFFSET(index);

	return readl(ep->pcie_base + offset + reg);
}

static void pcie_writel_inbound_atu(struct pcie_endpoint *ep, u32 index, u32 reg, u32 val)
{
	u32 offset = PCIE_GET_ATU_INB_WIRE_REG_OFFSET(index);

	writel(val, ep->pcie_base + offset + reg);
}

static void pcie_disable_inbound_atu(struct pcie_endpoint *ep, int index)
{
	u32 retries, val;

	pcie_writel_inbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2, (u32)~PCIE_ATU_ENABLE);
	/*
	 * Make sure ATU enable takes effect before any subsequent config
	 * and I/O accesses.
	 */
	for (retries = 0; retries < LINK_WAIT_MAX_IATU_RETRIES; retries++) {
		val = pcie_readl_inbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2);
		if (!(val & PCIE_ATU_ENABLE))
			return;

		mdelay(LINK_WAIT_IATU);
	}
	dev_err(&ep->pdev->dev, "Inbound iATU is not being disabled\n");
}

static int pcie_program_inbound_atu(struct pcie_endpoint *ep, int index,
				    int bar, phys_addr_t ep_addr)
{
	u32 retries, val;

	pcie_writel_inbound_atu(ep, index, PCIE_ATU_UNR_LOWER_TARGET, lower_32_bits(ep_addr));
	pcie_writel_inbound_atu(ep, index, PCIE_ATU_UNR_UPPER_TARGET, upper_32_bits(ep_addr));
	pcie_writel_inbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL1, PCIE_ATU_TYPE_MEM);
	pcie_writel_inbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2,
				PCIE_ATU_ENABLE | PCIE_ATU_BAR_MODE_ENABLE | (bar << 8));

	/* BAR-matched mode does not need to set PCIE_ATU_UNR_LIMIT */

	/*
	 * Make sure ATU enable takes effect before any subsequent config
	 * and I/O accesses.
	 */
	for (retries = 0; retries < LINK_WAIT_MAX_IATU_RETRIES; retries++) {
		val = pcie_readl_inbound_atu(ep, index, PCIE_ATU_UNR_REGION_CTRL2);
		if (val & PCIE_ATU_ENABLE)
			return 0;

		mdelay(LINK_WAIT_IATU);
	}
	dev_err(&ep->pdev->dev, "Inbound iATU is not being enabled\n");

	return -EBUSY;
}

static inline struct pcie_endpoint *pcie_endpoint_from_hw(struct ieee80211_hw *hw)
{
	return (struct pcie_endpoint *)(dev_get_drvdata(wiphy_dev(hw->wiphy)));
}

static inline struct pci_dev *pci_dev_from_hw(struct ieee80211_hw *hw)
{
	return to_pci_dev(wiphy_dev(hw->wiphy));
}

/**
 * pcie_mmio_writel() - generic mmio write
 * @ep: endpoint
 * @val: value
 * @addr: endpoint local address (32 bits)
 *
 * This function remaps the BAR4 to addr
 * 4KB
 */
static void pcie_mmio_writel(struct pcie_endpoint *ep, u32 val, unsigned int addr)
{
	phys_addr_t base, offset;
	unsigned long flags;

	base = ALIGN_DOWN(addr, PCIE_REG_ALIGN);
	offset = addr - base;

	/* Program inbound ATU 4 to translate BAR 4 */
	spin_lock_irqsave(&ep->lock, flags);

	if (base != ep->last_bp_base) {
		pcie_disable_inbound_atu(ep, 4);
		if (!pcie_program_inbound_atu(ep, 4, 4, base))
			ep->last_bp_base = base;
		else
			goto out;
	}
	writel(val, ep->bp_base + offset);
out:
	spin_unlock_irqrestore(&ep->lock, flags);
}

/**
 * pcie_mmio_readl() - generic mmio read
 * @ep: endpoint
 * @addr: endpoint local address (32 bits)
 *
 * This function remaps the BAR4 to addr
 *
 */
static u32 pcie_mmio_readl(struct pcie_endpoint *ep, unsigned int addr)
{
	phys_addr_t base, offset;
	unsigned long flags;
	u32 val = -1;

	base = ALIGN_DOWN(addr, PCIE_REG_ALIGN);
	offset = addr - base;

	/* Program inbound ATU 4 to translate BAR 4 */
	spin_lock_irqsave(&ep->lock, flags);

	if (base != ep->last_bp_base) {
		pcie_disable_inbound_atu(ep, 4);
		if (!pcie_program_inbound_atu(ep, 4, 4, base))
			ep->last_bp_base = base;
		else
			goto out;
	}
	val = readl(ep->bp_base + offset);
out:
	spin_unlock_irqrestore(&ep->lock, flags);

	return val;
}

static phys_addr_t pcie_ep_mem_addr(struct pcie_endpoint *ep, int bar, void *addr)
{
	off_t offset = addr - (void *)pcim_iomap_table(ep->pdev)[bar];
	u32 v;
	phys_addr_t ep_bar_addr;

	v = pcie_readl_inbound_atu(ep, bar, PCIE_ATU_UNR_REGION_CTRL2);
	if (!(v & PCIE_ATU_ENABLE)) {
		dev_err(&ep->pdev->dev, "atu %d not enabled\n", bar);
		return 0;
	}

	ep_bar_addr = pcie_readl_inbound_atu(ep, bar, PCIE_ATU_UNR_LOWER_TARGET);
	return ep_bar_addr + offset;
}

static unsigned long pcie_pfn(struct pcie_endpoint *ep, dma_addr_t addr)
{
	return addr >> (ep->page_shift);
}

static inline phys_addr_t pcie_atu_to_ep_addr(struct pcie_endpoint *ep, int atu)
{
	return ep->config->axi_base + atu * ep->page_size;
}

static inline int pcie_ep_addr_to_atu(struct pcie_endpoint *ep, phys_addr_t ep_addr)
{
	int atu = (ep_addr - ep->config->axi_base) >> ep->page_shift;

	WARN_ON(atu < 0 || atu > ep->config->num_atu);
	return atu;
}

/**
 * pci_endpoint_addr() - calculate pice slave interface address
 * @ep: pcie endpoint
 * @p: mapping
 */
phys_addr_t pci_endpoint_addr(struct pcie_endpoint *ep, struct pcie_addr_mapping *m, off_t offset)
{
	u64 addr;

	addr = pcie_atu_to_ep_addr(ep, m->atu) + offset;
	return addr;
}

void pcie_dump_mapping(struct ieee80211_hw *hw)
{
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct pcie_addr_mapping *m;
	struct rb_node *node;
	dma_addr_t end;

	for (node = rb_first(&ep->root); node; node = rb_next(node)) {
		m = rb_entry(node, struct pcie_addr_mapping, rb);
		end = m->pci_addr + ep->page_size - 1;
		scm80211_trace(hw, "- [%pad-%pad]: atu=%d, refs=%d\n",
			       &m->pci_addr, &end, m->atu, atomic_read(&m->refcount));
	}
}

/**
 * pcie_find_mapping() - find a mapping for @addr and @size
 * @addr: pci address
 * @size: pci address range
 *
 * This function should be called with ep->lock held.
 * [@addr, @addr+@size-1] should not span over multiple mappings
 */

struct pcie_addr_mapping *pcie_find_mapping(struct pcie_endpoint *ep, dma_addr_t addr, size_t size)
{
	struct rb_node *node = ep->root.rb_node;
	struct pcie_addr_mapping *mapping = NULL;

	while (node) {
		mapping = rb_entry(node, struct pcie_addr_mapping, rb);

		if (addr < mapping->pci_addr)
			node = node->rb_left;
		else if (addr > mapping->pci_addr)
			node = node->rb_right;
		else
			return mapping;
	}
	return NULL;
}

#define PC_TO_EP_INTERRUPT	GENMASK(31, 10)

static void pcie_raise_endpoint_irq(struct pci_dev *pdev, int irq)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	u32 where, v;
	int retries = 10;
	unsigned long flags;

	/*
	 * NB: currently target has problem in enabling interrupts.
	 * Do nothing here.
	 *
	 */

	/* Raise an EP interrupt */
	where = ep->elbi_offset + 0x8;

	spin_lock_irqsave(&ep->lock, flags);

	pci_read_config_dword(pdev, where, &v);
	v &= PC_TO_EP_INTERRUPT;
	v |= FIELD_PREP(PC_TO_EP_INTERRUPT, (1 << irq));
	pci_write_config_dword(pdev, where, v);

	/* Wait until EP clears the interrupt */
	do {
		udelay(10);
		pci_read_config_dword(pdev, where, &v);
		v = FIELD_GET(PC_TO_EP_INTERRUPT, v);
		if ((v & BIT(irq)) == 0) {
			spin_unlock_irqrestore(&ep->lock, flags);
			return;
		}
	} while (retries-- >= 0);

	spin_unlock_irqrestore(&ep->lock, flags);

	dev_warn(&pdev->dev, "endpoint does not handle interrupt %d\n", irq);
}

/**
 * scm80211_pcie_param_write()
 * program message and sent it to target
 * TODO
 */
static void scm80211_pcie_param_write(struct ieee80211_hw *hw, int vif_id, u16 event,
				      void *data, u16 size)
{
	struct pcie_endpoint *ep	= pcie_endpoint_from_hw(hw);
	struct ieee80211_hw_priv *priv	= hw->priv;
	struct pci_dev *pdev		= pci_dev_from_hw(hw);
	int idx = atomic_inc_return(&ep->irq_chan) % HTC_H2T_NR;
	struct htc_content __iomem *htc = ep->h2t[idx];

	scm80211_dbg(hw, "program idx %d, event %d\n", idx,  event);
	htc->hw_id	= priv->devid;
	htc->vif_id	= vif_id;
	htc->event	= event;
	htc->len	= size;
	/* operate target memory */
	wmb();

	if (data)
		memcpy_toio(htc->content, data, size);

	pcie_raise_endpoint_irq(pdev, idx);
}


/**
 * pcie_alloc_mapping() - allocates a pci-to-local address mapping entry
 */
static struct pcie_addr_mapping *pcie_alloc_mapping(struct ieee80211_hw *hw,
						    dma_addr_t pci_addr, size_t size)
{
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct pcie_addr_mapping *mapping;
	struct device *dev = wiphy_dev(hw->wiphy);
	int atu = -1;

	/* Find the free outbound atu */
	atu = find_first_zero_bit(ep->bitmap, ep->config->num_atu);
	if (atu == ep->config->num_atu) {
		dev_warn(dev, "failed to find free outbound ATU\n");
		pcie_dump_mapping(hw);
		return NULL;
	}

	/* Allocate a mapping object */
	mapping = kzalloc(sizeof(*mapping), GFP_ATOMIC);
	if (!mapping)
		return NULL;

	mapping->pci_addr = pci_addr;
	mapping->atu = atu;
	mapping->ep_addr = pcie_atu_to_ep_addr(ep, atu);
	atomic_set(&mapping->refcount, 0);
	set_bit(atu, ep->bitmap);

	/* Program the outbound ATU */
	pcie_program_outbound_atu(ep, atu, PCIE_ATU_TYPE_MEM, mapping->ep_addr,
				  pci_addr, ep->page_size);

	return mapping;
}

/**
 * scm80211_pcie_remap() - map host physical address into endpoint local address
 * @hw: hardware
 * @start: host phyiscal address
 * @len: sizeo of the address region to be translated
 *
 * Return: translated ep address
 *
 * This function establishes dma address to ep address mapping and returns the
 * translated address.
 *
 */
static phys_addr_t scm80211_pcie_remap(struct ieee80211_hw *hw, dma_addr_t start, size_t len)
{
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct rb_node **node, *parent = NULL;
	struct pcie_addr_mapping *mapping;
	dma_addr_t addr, end, offset;
	unsigned long flags;

	addr = start & ~(ep->page_size - 1);
	offset = start & (ep->page_size - 1);
	end = start + len - 1;

	scm80211_trace(hw, "pci remap: [%pad, %pad] (%zu bytes)\n", &start, &end, len);

	/*
	 * NB: If a scatterlist element cannot be mapped into a single
	 * target address range, it needs to be split.
	 *
	 * To the best of our knowledge, the maximum number of contiguosly
	 * allocatable pages is 1024 (4MB), which is smaller than the
	 * our address mapping size (32MB, or 64MB). As long as we are dealing
	 * with dynamic memory areas allocated via kmalloc() or get_free_pages(),
	 * they fit into a single mapping.
	 */
	if (WARN_ON(pcie_pfn(ep, start) != pcie_pfn(ep, end))) {
		scm80211_warn(hw, "pci address range [%pad, %pad] crosses page boundary\n",
			      &start, &end);
		return 0;
	}

	spin_lock_irqsave(&ep->lock, flags);

	/*
	 * Scan the rb tree to find the mapping or locate the (parent, node) to
	 * insert new mapping
	 */
	node = &ep->root.rb_node;
	while (*node) {
		mapping = rb_entry(*node, struct pcie_addr_mapping, rb);
		parent = *node;

		if (addr < mapping->pci_addr)
			node = &((*node)->rb_left);
		else if (addr > mapping->pci_addr)
			node = &((*node)->rb_right);
		else
			goto found;
	}

	/* No mapping found - create a new mapping and add it to rb tree */
	mapping = pcie_alloc_mapping(hw, addr, ep->page_size);
	if (!mapping) {
		scm80211_err(hw, "failed to remap: pci_addr=%pad\n", &start);

		spin_unlock_irqrestore(&ep->lock, flags);
		return 0;
	}
	rb_link_node(&mapping->rb, parent, node);
	rb_insert_color(&mapping->rb, &ep->root);

found:
	atomic_inc(&mapping->refcount);

	scm80211_trace(hw,
		       "remap: pci=%pad+%lx ep=%08x size=%zu atu=%d ref=%d\n",
		       &addr, (unsigned long)offset, lower_32_bits(mapping->ep_addr), len,
		       mapping->atu, atomic_read(&mapping->refcount));

	spin_unlock_irqrestore(&ep->lock, flags);

	return mapping->ep_addr + offset;
}

static void scm80211_pcie_unmap(struct ieee80211_hw *hw, dma_addr_t start, size_t len)
{
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct pcie_addr_mapping *mapping = NULL;
	dma_addr_t addr, end, offset;
	unsigned long flags;

	end = start + len - 1;
	offset = start & (ep->page_size - 1);
	addr = start & ~(ep->page_size - 1);

	spin_lock_irqsave(&ep->lock, flags);

	mapping = pcie_find_mapping(ep, addr, len);
	if (!mapping) {
		scm80211_warn(hw, "failed to find the mapping for %pad\n", &start);
		goto out;
	}

	scm80211_trace(hw, "unmap: pci=%pad (%zu) atu=%d refs=%d\n",
		       &start, len, mapping->atu, atomic_read(&mapping->refcount));

	if (atomic_dec_return(&mapping->refcount) > 0)
		goto out;

	/* Free the mapping */
	clear_bit(mapping->atu, ep->bitmap);
	rb_erase(&mapping->rb, &ep->root);
	pcie_disable_outbound_atu(ep, mapping->atu);
	kfree(mapping);

out:
	spin_unlock_irqrestore(&ep->lock, flags);
}

/*
 * Slow version
 */
static dma_addr_t scm80211_pcie_inverse_map(struct ieee80211_hw *hw, phys_addr_t addr)
{
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct rb_node *node;
	struct pcie_addr_mapping *mapping;
	dma_addr_t pci_addr = DMA_MAPPING_ERROR;
	off_t offset;
	int atu;
	unsigned long flags;

	offset = addr & (ep->page_size - 1);
	atu = pcie_ep_addr_to_atu(ep, addr);

	spin_lock_irqsave(&ep->lock, flags);

	for (node = rb_first(&ep->root); node; node = rb_next(node)) {
		mapping = rb_entry(node, struct pcie_addr_mapping, rb);
		if (mapping->atu == atu) {
			pci_addr = mapping->pci_addr + offset;
			goto out;
		}
	}

out:
	spin_unlock_irqrestore(&ep->lock, flags);
	return pci_addr;
}

static void scm80211_pcie_mac_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze && !(offset == 0x8004 || offset == 0x8000)) {
		scm80211_err(hw, "[ERROR] +%s reg %08x\n", __func__, offset);
		return;
	}
	writel(val, section->mac_base + offset);
}

static u32 scm80211_pcie_mac_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze && offset != 0x8004) {
		scm80211_err(hw, "[ERROR] +%s reg %08x\n", __func__, offset);
		return (u32) -1;
	}

	return readl(section->mac_base + offset);
}

static void scm80211_pcie_phy_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze) {
		scm80211_err(hw, "[ERROR] +%s reg %08x\n", __func__, offset);
		return;
	}

	writel(val, section->phy_base + offset);
}

static u32 scm80211_pcie_phy_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze) {
		scm80211_err(hw, "[ERROR] +%s reg %08x\n", __func__, offset);
		return (u32) -1;
	}

	return readl(section->phy_base + offset);
}

static void scm80211_pcie_raw_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	pcie_mmio_writel(pcie_endpoint_from_hw(hw), val, offset);
}

static u32 scm80211_pcie_raw_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return pcie_mmio_readl(pcie_endpoint_from_hw(hw), offset);
}

static void scm80211_pcie_efuse_writel(void *dev_ep, void *val, unsigned int addr, unsigned int len)
{
#ifdef SCM80211_FTM
	pcie_mmio_writel((struct pcie_endpoint *)dev_ep, *(u32 *)val, addr);
#endif
}

static int scm80211_pcie_efuse_readl(void *dev_ep, void *val, unsigned int addr, unsigned int len)
{
	return (int)pcie_mmio_readl((struct pcie_endpoint *)dev_ep, addr);
}

/* In case RF registers are memory mapped */
static void scm80211_pcie_rf_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze)
		scm80211_err(hw, "[ERROR] +%s reg %08x\n", __func__, offset);

	pcie_mmio_writel(pcie_endpoint_from_hw(hw), val, section->rf_base + offset);
}

static u32 scm80211_pcie_rf_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze)
		scm80211_err(hw, "[ERROR] +%s reg %08x\n", __func__, offset);
	return pcie_mmio_readl(pcie_endpoint_from_hw(hw), section->rf_base + offset);
}

static void scm80211_pcie_pta_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);

	pcie_mmio_writel(pcie_endpoint_from_hw(hw), val, section->pta_base + offset);
}

static u32 scm80211_pcie_pta_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);

	return pcie_mmio_readl(pcie_endpoint_from_hw(hw), section->pta_base + offset);
}

static struct tx_mpdu_desc *scm80211_txop_mdesc(struct txop *txop, int idx, int target)
{
	return &txop->dma[target].desc->mdesc[idx];
}

static struct tx_buffer_desc *scm80211_txop_bdesc(struct txop *txop, int idx, int target)
{
	return &txop->dma[target].desc->mpdu[idx].bdesc[0];
}

static phys_addr_t scm80211_mdesc_raddr(struct txop *txop, int index, int target)
{
	return txop->dma[target].raddr + offsetof(struct tx_dma_desc, mdesc[index]);
}

static phys_addr_t scm80211_bdesc_raddr(struct txop *txop, int index, int target)
{
	return txop->dma[target].raddr + offsetof(struct tx_dma_desc, mpdu[index].bdesc);
}

/**
 * scm80211_pcie_hwq_alloc_dma()
 *
 * In order to avoid TX WIN lost failure, use mpdu descriptor,  buffer descriptors,
 * and the bounce buffer in target remote memory.
 */
static int scm80211_pcie_hwq_alloc_dma(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct tx_dma_desc *ddesc;
	size_t size = sizeof(struct tx_dma_desc), shift;
	u8 m, n;

	if (ep->ep_mem_base == FW_BASE) {
		shift = SRAM_MEM_SIZE - (sizeof(struct htc_content) * (HTC_H2T_NR + HTC_T2H_NR)) -
			(size * SCM80211_NUM_HWQ * SCM80211_MAX_TXOP) * (priv->devid + 1);
	} else {
		shift = (size * SCM80211_NUM_HWQ * SCM80211_MAX_TXOP) * priv->devid;
	}
	ddesc = (struct tx_dma_desc *)(ep->mem_base + shift / sizeof(void));

	for (n = 0; n < SCM80211_MAX_TXOP; n++) {
		struct txop *txop = &hwq->ring.pool[n];

		/* Allocate and remap the mpdu and buffer descriptors on the host memory */
		txop->dma[0].desc = dma_alloc_coherent(dev, size, &txop->dma[0].paddr, GFP_KERNEL);
		if (!txop->dma[0].desc)
			return -ENOMEM;
		txop->dma[0].raddr = scm80211_remap(hw, txop->dma[0].paddr, size);

		/* Setup up the mpdu and buffer descriptor on the target memory */
		txop->dma[1].desc = ddesc + hwq->id + n * SCM80211_NUM_HWQ;
		txop->dma[1].raddr = pcie_ep_mem_addr(ep, 3, txop->dma[1].desc);

		txop->bounce_buffer = txop->dma[1].raddr + offsetof(struct tx_dma_desc, buffer);

		/* Program >mdesc[m].baddr with the remote target buffer address */
		for (m = 0; m < SCM80211_MAX_MPDU; m++) {
			txop->dma[0].desc->mdesc[m].bdaddr = scm80211_bdesc_raddr(txop, m, 0);
			txop->dma[1].desc->mdesc[m].bdaddr = scm80211_bdesc_raddr(txop, m, 1);
		}

		/* Use the mpdu descriptor in the host instead of the one in the target */
		txop->raddr = scm80211_mdesc_raddr(txop, 0, 0);
		txop->space = MT_HST_BD_HST;

		if (hwq->ring.order == 0 && n == 0) {
			hwq->txdesc->addr = txop->raddr;
			/* add write barrier to make sure it had write to memory */
			wmb();
		}
	}
	return 0;
}

static void scm80211_pcie_hwq_free_dma(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	size_t size = sizeof(struct tx_dma_desc);
	int i;

	for (i = 0; i < SCM80211_MAX_TXOP; i++) {
		struct txop *txop = &hwq->ring.pool[i];

		scm80211_unmap(hw, txop->dma[0].paddr, size);
		dma_free_coherent(dev, size, txop->dma[0].desc, txop->dma[0].paddr);
		txop->dma[0].desc = NULL;
		txop->dma[0].paddr = DMA_MAPPING_ERROR;
		txop->dma[0].raddr = 0;
	}
}

#ifdef ERRATA_RSN_PSIZE_ERROR
static u32 scm80211_get_header_len(struct sk_buff *skb)
{
	u32 hlen = 0;
	struct txvector *tv;
	struct ieee80211_hdr *hdr;
	struct ieee80211_key_conf *hw_key;
	bool isdata;

	if (!skb)
		return 0;

	tv = (void *)skb->data;
	hdr = (void *)(tv + 1);
	hw_key = IEEE80211_SKB_CB(skb)->control.hw_key;
	isdata = ieee80211_is_data(hdr->frame_control);

	if (hw_key && isdata && ieee80211_has_protected(hdr->frame_control)) {
		hlen = (sizeof(struct txvector) +
			ieee80211_hdrlen(hdr->frame_control) + hw_key->iv_len);
	}

	return hlen;
}
#endif

#define SKB_COPY_BYTES	80

/**
 * scm80211_pcie_fixup_tx() - fix up tx information
 *
 * The main tx routines set up the mpdu and buffer descriptor on the host memory, i.e.,
 * @txop->dma[0]. Depending on the configuration, we copy them to target memory
 *
 */
static void scm80211_pcie_fixup_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	struct tx_mpdu_desc *mdesc;
	struct tx_buffer_desc *bdesc;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct sk_buff *skb;
	size_t len = SKB_COPY_BYTES;
	int n_mpdus = skb_queue_len(&txop->sched);
	int m, ver;

	/*
	 * Copy the first SKB_COPY_BYTES of psdu into target memory.
	 * LDPC information block maximum size: 1620 bits = 202.5 bytes
	 */
	mdesc = scm80211_txop_mdesc(txop, 0, 0);
	bdesc = scm80211_txop_bdesc(txop, 0, 0);

	skb = skb_peek(&txop->sched);
#ifdef ERRATA_RSN_PSIZE_ERROR
	len = SKB_COPY_BYTES - ((SKB_COPY_BYTES - scm80211_get_header_len(skb)) % 16);
#endif
	len = min_t(size_t, skb->len, len);
	memcpy_toio(txop->dma[1].desc->buffer, skb->data, len);

	if (bdesc[0].len > len) {
		memmove(bdesc + 1, bdesc, sizeof(*bdesc) * (mdesc->bdnum + 1));
		bdesc[0].len = len;
		bdesc[1].addr_l += len;
		bdesc[1].len -= len;
		mdesc->bdnum++;
	}

	/* Make the bdesc[0] points to the bounce buffer */
	bdesc[0].addr_l = txop->bounce_buffer;
	/* make sure it had written to host memory */
	wmb();

	/*
	 * Copy tx descriptors from the host memory to the target memory
	 * if necessary.
	 */
	BUG_ON(priv->space < txop->space);
	ver = version(scm80211_mac_version(hw));

	/*
	 * Fixup the bdaddr which may modified by HWMAC under multi-hwq/vif test
	 * reload the bdaddr in for each condition
	 */
	for (m = 0; m < n_mpdus; m++) {
		if (priv->space == MT_HST_BD_HST) {
			txop->dma[0].desc->mdesc[m].bdaddr = scm80211_bdesc_raddr(txop, m, 0);
		} else if (ver < 0x14 || priv->space == MT_TGT_BD_TGT) {
			txop->dma[0].desc->mdesc[m].bdaddr = scm80211_bdesc_raddr(txop, m, 1);
		} else if (m == 0) {
			/*
			 * only set the bdescs in first mpdu in target if mac ver > 0x14
			 * which enabled muilti-fetch data from host
			 */
			txop->dma[0].desc->mdesc[m].bdaddr = scm80211_bdesc_raddr(txop, m, 1);
		} else {
			txop->dma[0].desc->mdesc[m].bdaddr = scm80211_bdesc_raddr(txop, m, 0);
		}
	}

	if (priv->space != txop->space)
		txop->space = priv->space;

	if (priv->space < MT_TGT_BD_TGT)
		txop->raddr = scm80211_mdesc_raddr(txop, 0, 0);
	else
		txop->raddr = scm80211_mdesc_raddr(txop, 0, 1);

	for (m = 0; m < n_mpdus; m++) {
		struct tx_mpdu_desc *md[2];
		struct tx_buffer_desc *bd[2];

		md[0] = scm80211_txop_mdesc(txop, m, 0);
		md[1] = scm80211_txop_mdesc(txop, m, 1);

		bd[0] = scm80211_txop_bdesc(txop, m, 0);
		bd[1] = scm80211_txop_bdesc(txop, m, 1);

		if (priv->space == MT_TGT_BD_TGT) {
			/* Copy mdesc except for pointer to the bdesc */
			memcpy_toio(md[1], md[0], sizeof(__le32) * 5);
		}

		if (priv->space > MT_HST_BD_HST) {
			if (ver < 0x14 || priv->space == MT_TGT_BD_TGT) {
				/* Copy buffer descriptors for all mpdus */
				memcpy_toio(bd[1], bd[0], sizeof(*bd[0]) * (md[0]->bdnum + 1));
			} else if (m == 0) {
				/* Copy buffer descriptors of the first mpdu */
				memcpy_toio(bd[1], bd[0], sizeof(*bd[0]) * (md[0]->bdnum + 1));
			}
		}
	}

	/* make sure it had written to target memory */
	wmb();
}

/**
 * scm80211_pcie_lookup_tx()
 * check bdaddr in md to identify hwmac rewrite it or not
 */
static int scm80211_pcie_lookup_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	switch (txop->space) {
	case MT_HST_BD_HST:
		return txop->dma[0].desc->mdesc[0].bdaddr != scm80211_bdesc_raddr(txop, 0, 0) ?
		    -EIO : 0;
	case MT_HST_BD_TGT:
	case MT_TGT_BD_TGT:
		return txop->dma[0].desc->mdesc[0].bdaddr != scm80211_bdesc_raddr(txop, 0, 1) ?
		    -EIO : 0;
	default:
		return -EINVAL;
	}
}

static int scm80211_pcie_set_channel(struct ieee80211_hw *hw, struct scm80211_phy *phy,
				     struct phy_vector *pv)
{
	phy->radio->ops->set_channel(phy->radio, pv->CENTER_FREQUENCY, pv->BANDWIDTH);
	scm2020_phy_set_channel(phy, 0, pv);
	return 0;
}

static int scm80211_pcie_set_ant_sel(struct ieee80211_hw *hw, int ant)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy = priv->phy;
	struct scm80211_radio radio, *radio_sw1, *radio_sw2;
	struct scm2020 *r_dev;
	int ant_sel = (ant == 1) ? 1 : 0;

	r_dev = (struct scm2020 *)phy->radio->priv;
#ifdef SCM2020_A0
	if (!ant_sel && (r_dev->path == 'B'))
		return 0;
	else if (ant_sel && (r_dev->path == 'A'))
		return 0;
#else
	if (!ant_sel && (r_dev->path == 'A'))
		return 0;
	else if (ant_sel && (r_dev->path == 'B'))
		return 0;
#endif

	phy->rf_ant_sel = ant_sel;
	memcpy(&radio, phy->radio, sizeof(radio));
	radio_sw1 = phy->radio;

	phy = priv->phy + 1;
	phy->rf_ant_sel = ant_sel;
	radio_sw2 = phy->radio;

	radio_sw1->phy = radio_sw2->phy;
	radio_sw1->priv = radio_sw2->priv;
	radio_sw2->phy = radio.phy;
	radio_sw2->priv = radio.priv;

	return 0;
}

static int scm80211_pcie_get_ant(struct ieee80211_hw *hw)
{
	u32 v;

	v = scm80211_raw_readl(hw, PHY_MODEM(FW_RFI_EN));
	v >>= 8;
	v &= 0x1;
#ifdef SCM2020_A0
	v ^= 0x1;
#endif

	return v;
}
#define PM_LOCAL_TEST	0

static void scm80211_enable_aspm(struct pci_dev *pdev);
static void scm80211_disable_aspm(struct pci_dev *pdev);
static void scm80211_pcie_vif_doze(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				   bool twt, bool link);
static int scm80211_pcie_power_save(struct ieee80211_hw *hw, bool powersave, bool link)
{
	struct pci_dev *pdev = pci_dev_from_hw(hw);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	struct ieee80211_vif *vif = NULL;
	struct ieee80211_vif_priv *vp = NULL;
	int i;

	if (powersave) {
		WARN_ON(ep->doze);
		for_each_set_bit(i, &priv->vif_map, SCM80211_NUM_VIF) {
			vif = priv->vifs[i];
			if (vif == NULL)
				continue;
			vp = ieee80211_to_priv(vif);
			pm_save(hw, vp->pm.ps_active, MAC_REG, mac_vif_ps_active_regs[vp->index]);
		}
		scm80211_pcie_param_write(hw, -1, (link << 8) | HW_PS_ENABLE, NULL, 0);
		msleep(2);
		scm80211_enable_aspm(pdev);
		if (!link && vp) {
			pm_save(hw, priv->pm.assoc_active, MAC_REG, mac_assoc_active_regs);
			pm_save(hw, vp->pm.assoc_awake, MAC_REG,
				mac_vif_assoc_awake_regs[vp->index]);
			scm80211_pcie_vif_doze(hw, priv->pm.vif, false, link);
		}
	} else {
		struct scm80211_chanctx *chanctx;
		struct scm80211_phy *phy;
		int timeout = 10;

		WARN_ON(hweight_long(priv->vif_map) != 1);
		vif = priv->vifs[find_first_bit(&priv->vif_map, SCM80211_NUM_VIF)];
		if (WARN_ON(!vif))
			return -EINVAL;

		vp = ieee80211_to_priv(vif);
		scm80211_disable_aspm(pdev);
		if (vif->bss_conf.assoc)
			chanctx = (struct scm80211_chanctx *)get_ieee80211_chanctx_conf(vif);
		else
			chanctx = &hw_sched(hw)->idle;

		priv->pm.ps = PS_DISABLED;
		del_timer_sync(&vp->pm.doze_timer);
		del_timer_sync(&vp->pm.wake_timer);

		scm80211_pcie_param_write(hw, -1, HW_PS_DISABLE, NULL, 0);

#if PM_LOCAL_TEST
		if (ep->doze) {
			/* test call wake up here, remove it when fw wakeup used */
			mod_timer(&vp->pm.wake_timer, jiffies + usecs_to_jiffies(2));
			msleep(10);
		}
#endif
		while (ep->doze) {
			/* cmd maybe lost try again */
			scm80211_pcie_param_write(hw, -1, HW_PS_DISABLE, NULL, 0);
			msleep(2);
			if (timeout-- < 0) {
				scm80211_err(hw, "awake fail\n");
				WARN_ON(1);
				return -ENODEV;
			}
		}
#ifdef SCM2020_A0
		for_each_phy(phy, hw) {
			phy->radio->ops->power_save(phy->radio, 0, false);
		}
#endif
		pm_restore(hw, priv->pm.assoc_active);
		pm_restore(hw, vp->pm.assoc_active);
		pm_restore(hw, vp->pm.ps_active);
		pm_restore(hw, priv->pm.phy_assoc_active);
		scm80211_set_chanctx(hw, chanctx);
	}
	return 0;
}

static int scm80211_pcie_vif_power_save(struct ieee80211_hw *hw,
					struct ieee80211_vif *vif, bool enable)
{
	return 0;
}

static u32 scm80211_sync_rx_wptr(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 wptr;

	wptr = scm80211_mac_readl(hw, MAC_REG_RX_BUF_PTR1);
	wptr = FIELD_GET(RX_BUF_PTR1_WRITE_PTR, wptr);
	priv->rxring.wptr = wptr;

	scm80211_trace(hw, "wtpr=0x%04x\n", wptr);

	return wptr;
}

static void scm80211_update_rx_ring_ptrs(struct ieee80211_hw *hw, bool init)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 v, rptr, bptr, wptr;

	/* Take least significant (priv->rx.order + 1) bits */
	bptr = ring_ptr(&priv->rxring, priv->rxring.bptr);
	rptr = ring_ptr(&priv->rxring, priv->rxring.rptr);
	wptr = ring_ptr(&priv->rxring, priv->rxring.wptr);

	scm80211_trace(hw, "rx ptrs: bptr=0x%04x, rptr=0x%04x, wtpr=0x%04x\n", bptr, rptr, wptr);

	v = FIELD_PREP(RX_BUF_PTR0_READ_PTR, rptr);
	v |= FIELD_PREP(RX_BUF_PTR0_READY_PTR, bptr);
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_PTR0);

	/* mac version > 0.9.4 wptr can be r/w  */
	if (version(scm80211_mac_version(hw)) >= 0x16 && init) {
		v = FIELD_PREP(RX_BUF_PTR1_WRITE_PTR, wptr);
		scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_PTR1);
	}
}

#ifdef SCM80211_RX_PAGE_ALLOC

/**
 * scm80211_rxd_attach_page() - attach page to rxd
 *
 * background: buff in host ram will be mapped to PCIe 2G address,
 * if skbs is separated too much (memory is fragmented after long time running) in host ram,
 * it may cause no ATU left and lead tx/rx fail (memory fragmented after)
 *
 * positive:
 * - will reduce the consumption of PCIe ATUs
 * negative:
 * - have some consumption to copy data from page to skb after get rx irq
 *
 * suggest to use it on the host which has more than 4G ram size
 */
static int scm80211_rxd_attach_page(struct ieee80211_hw *hw, struct scm80211_rxdesc *rxd)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	struct sk_buff *skb;
	dma_addr_t dma_addr;

	skb = scm80211_alloc_rx_skb(hw);
	if (!skb) {
		scm80211_warn(hw, "rxd %d: failed to alloc skb\n", rxd_index(hw, rxd));
		return -ENOMEM;
	}

	dma_addr = dma_map_page(dev, rxd->page, IEEE80211_RX_HEADROOM, rx_buflen, DMA_FROM_DEVICE);
	if (dma_mapping_error(dev, dma_addr)) {
		scm80211_warn(hw, "rxd %d: failed to map page\n", rxd_index(hw, rxd));
		goto free_skb;
	}

	do {
		struct rxdesc rd;
		phys_addr_t addr = scm80211_remap(hw, dma_addr, rx_buflen);

		if (!addr) {
			scm80211_warn(hw, "rxd %d: rx addr mapping fail\n", rxd_index(hw, rxd));
			goto dma_unmap;
		}
		memset(&rd, 0, sizeof(rd));
		rd.addr = addr;
		rd.blen = cpu_to_le16(rx_buflen);
		scm80211_trace(hw, " - rxd[%02x] page addr %p rxd addr %llx, blen %u\n",
			       rxd->id, rxd->page, rd.addr, rd.blen);
		memcpy_toio(rxd->config, &rd, sizeof(rd));

	} while (0);

	rxd->dma_addr = dma_addr;
	rxd->skb = skb;
	return 0;

dma_unmap:
	dma_unmap_page(dev, dma_addr, rx_buflen, DMA_FROM_DEVICE);
free_skb:
	kfree_skb(skb);
	return -ENOMEM;
}

#define scm80211_rxd_attach_buff scm80211_rxd_attach_page

#else
/**
 * scm80211_rxd_attach_skb() -- attach skb to txd
 *
 * positive:
 * - no copy, can use this skb directly
 * negative:
 * - may have no ATU for skbs
 */
static int scm80211_rxd_attach_skb(struct ieee80211_hw *hw, struct scm80211_rxdesc *rxd)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	struct sk_buff *skb;
	struct scm80211_rx_info *info;
	struct rxdesc rd;
	phys_addr_t addr;
	dma_addr_t dma_addr;

	skb = scm80211_alloc_rx_skb(hw);
	if (!skb) {
		scm80211_warn(hw, "rxd %d: failed to alloc skb\n", rxd_index(hw, rxd));
		goto err_exit;
	}

	skb_reserve(skb, IEEE80211_RX_HEADROOM);

	dma_addr = dma_map_single(dev, skb->data, rx_buflen, DMA_FROM_DEVICE);
	if (dma_mapping_error(dev, dma_addr)) {
		scm80211_warn(hw, "rxd %d: failed to map rx skb\n", rxd_index(hw, rxd));
		goto free_skb;
	}

	info = SCM80211_SKB_RXCB(skb);
	info->dma_addr = dma_addr;
	info->index = rxd->id;

	addr = scm80211_remap(hw, dma_addr, rx_buflen);
	if (!addr) {
		scm80211_warn(hw, "failed to translate rx buffer\n");
		goto dma_unmap;
	}

	memset(&rd, 0, sizeof(rd));
	info->rxd.addr = addr;
	rd.addr = addr;
	info->rxd.blen = cpu_to_le16(rx_buflen);
	rd.blen = cpu_to_le16(rx_buflen);
	memcpy_toio(rxd->config, &rd, sizeof(rd));
	rxd->skb = skb;

	return 0;

dma_unmap:
	dma_unmap_single(dev, info->dma_addr, rx_buflen, DMA_FROM_DEVICE);
free_skb:
	kfree_skb(skb);
err_exit:
	return -ENOMEM;
}

#define scm80211_rxd_attach_buff scm80211_rxd_attach_skb
#endif /* SCM80211_RX_PAGE_ALLOC */

/**
 * scm80211_refill_rxring() - refill rxring
 * @hw: hardware
 *
 * Returns:
 * - 0 if all the ring entries are refilled with new buffer
 * - < 0 otherwise
 */
static int scm80211_pcie_refill_rxring(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_rxdesc *rxd;
	u16 wptr, rptr;
	unsigned long flags;

	spin_lock_irqsave(&priv->rxring.lock, flags);
	wptr = priv->rxring.wptr;
	rptr = priv->rxring.rptr;

	if (priv->pm.doze) {
		scm80211_dbg(hw, "hw goes to doze state, rxring refill suspend\n");
		scm80211_update_rx_ring_ptrs(hw, false);
		spin_unlock_irqrestore(&priv->rxring.lock, flags);
		return 0;
	}

	scm80211_trace(hw,
		       "rx fill: bptr=0x%04lx, rptr=0x%04lx, wptr=0x%04lx\n",
		       ring_ptr(&priv->rxring, priv->rxring.bptr),
		       ring_ptr(&priv->rxring, rptr), ring_ptr(&priv->rxring, wptr));

	/* Advance one revolution */
	wptr = wptr + ring_size(&priv->rxring);
	rptr = rptr + ring_size(&priv->rxring);

	while (!ring_ptr_equal(priv, priv->rxring.bptr, rptr)) {
		rxd = &priv->rxring.rxd[ring_index(&priv->rxring, priv->rxring.bptr)];
		if (scm80211_rxd_attach_buff(hw, rxd) < 0)
			break;
		priv->rxring.bptr++;
	}
	scm80211_update_rx_ring_ptrs(hw, false);
	spin_unlock_irqrestore(&priv->rxring.lock, flags);
	return 0;
}

static void scm80211_pcie_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct sk_buff *skb;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct scm80211_rxdesc *rxd = NULL;
	struct scm80211_rx_info *info = NULL;
	unsigned long flags;

	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	if (ep->doze) {
		scm80211_info(hw, "pcie device goes to doze mode skip\n");
		return;
	}
	scm80211_sync_rx_wptr(hw);
	scm80211_trace(hw,
		       "rx napi: bptr=0x%04lx, rptr=0x%04lx, wptr=0x%04lx\n",
		       ring_ptr(&priv->rxring, priv->rxring.bptr),
		       ring_ptr(&priv->rxring, priv->rxring.rptr),
		       ring_ptr(&priv->rxring, priv->rxring.wptr));

	scm80211_rx_statistics_update(hw);
	spin_lock_irqsave(&priv->rxring.lock, flags);
again:
	/* Step 1: Detach buffer and update rptr */
	while (!ring_ptr_equal(priv, priv->rxring.rptr, priv->rxring.wptr)) {
		rxd = &priv->rxring.rxd[ring_index(&priv->rxring, priv->rxring.rptr)];
#ifdef SCM80211_RX_PAGE_ALLOC
		if (unlikely(!rxd->dma_addr)) {
			scm80211_err(hw,
				     "rx err: bptr=0x%04lx, rptr=0x%04lx, wptr=0x%04lx\n",
				     ring_ptr(&priv->rxring, priv->rxring.bptr),
				     ring_ptr(&priv->rxring, priv->rxring.rptr),
				     ring_ptr(&priv->rxring, priv->rxring.wptr));
			BUG_ON(1);
		}

		scm80211_unmap(hw, rxd->dma_addr, rx_buflen);
		dma_unmap_page(dev, rxd->dma_addr, rx_buflen, DMA_FROM_DEVICE);
		rxd->dma_addr = 0;
		skb = rxd->skb;
		rxd->skb = NULL;
		info = SCM80211_SKB_RXCB(skb);
		memset(info, 0, sizeof(*info));
		memcpy_fromio(&info->rxd, rxd->config, sizeof(struct rxdesc));
		info->index = rxd->id;
		memcpy(skb->data, page_address(rxd->page), info->rxd.dlen + IEEE80211_RX_HEADROOM);
		skb_reserve(skb, IEEE80211_RX_HEADROOM);
		skb_queue_tail(&priv->rxring.rxskbs, skb);
#else
		info = SCM80211_SKB_RXCB(rxd->skb);
		memcpy_fromio((__le32 *)&info->rxd + 3, (__le32 *)rxd->config + 3,
			      sizeof(__le32));
		scm80211_unmap(hw, info->dma_addr, rx_buflen);
		dma_unmap_single(dev, info->dma_addr, rx_buflen, DMA_FROM_DEVICE);

		skb_queue_tail(&priv->rxring.rxskbs, rxd->skb);
		rxd->skb = NULL;
#endif
		priv->rxring.rptr++;

		/* Sync new skb */
		if (ring_ptr_equal(priv, priv->rxring.rptr, priv->rxring.wptr))
			scm80211_sync_rx_wptr(hw);
	}
	skb = NULL;

	/* check last info */
	if (info && (info->rxd.frag == START_FRAME || info->rxd.frag == MIDDLE_FRAME)) {
		u32 wptr = priv->rxring.wptr;

		scm80211_trace(hw, "rx irq end in the middle polling data again!\n");
		if (scm80211_sync_rx_wptr(hw) != wptr)
			goto again;
	}
	spin_unlock_irqrestore(&priv->rxring.lock, flags);

	/* Step 2: Attach new buffer and update bptr */
	scm80211_refill_rxring(priv->hw);
}


static int scm80211_pcie_rx_bh(struct napi_struct *napi, int budget)
{
	struct ieee80211_hw_priv *priv =
		container_of(napi, struct ieee80211_hw_priv, napi.context);
	struct ieee80211_hw *hw = priv->hw;
	int done;

	scm80211_pcie_rx(hw);

	ieee80211_queue_work(hw, &priv->rxring.work);

	done = scm80211_rx_poll(hw, budget);
	if (done < budget)
		scm80211_enable_rx_irq(hw);
	else
		scm80211_trace(hw,
			       "rx bh: done=%d, budget=%d keep rx irq disabled\n", done, budget);

	return done;
}

static int scm80211_pcie_init_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_rxdesc *rxd;
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	u32 v;
	int i;

	netif_napi_add(&priv->napi.netdev, &priv->napi.context, scm80211_pcie_rx_bh, 64);
	napi_enable(&priv->napi.context);
	priv->napi.enabled = true;
	scm80211_mac_writel(hw, 0, MAC_REG_RX_BUF_CFG);

	scm80211_update_rx_ring_ptrs(hw, true);
#ifdef SCM80211_RX_PAGE_ALLOC
	/* alloc 256 * buffer page size */
	priv->rxring.pages = dev_alloc_pages(SCM80211_RX_PAGE_ORDER + priv->rxring.order);
	if (!priv->rxring.pages) {
		scm80211_err(hw, "Can not alloc this much of pages\n");
		return -ENOMEM;
	}
#endif
	for (i = 0; i < ring_size(&priv->rxring); i++) {
		rxd = &priv->rxring.rxd[i];
		memset(rxd, 0, sizeof(*rxd));
		rxd->config = &section->mem->rx[i];
		rxd->id = i;
#ifdef SCM80211_RX_PAGE_ALLOC
		rxd->page = priv->rxring.pages + (1 << SCM80211_RX_PAGE_ORDER) * i;
		page_ref_add(rxd->page, USHRT_MAX - 1);
		rxd->dma_addr = 0;
#endif
	}
	priv->rxring.bptr = 0;
	priv->rxring.rptr = 0;

	scm80211_refill_rxring(hw);
	scm80211_refill_rxpool(hw);
	v = FIELD_PREP(RX_BUF_CFG_BUF_NUM, priv->rxring.order);
	if (version(scm80211_mac_version(hw)) >= 0x16)
		v |= FIELD_PREP(RX_BUF_CFG_DMA_EN, 1);
	v |= FIELD_PREP(RX_BUF_CFG_EN, 1);
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_CFG);

#ifdef SCM80211_RX_STATISTICS_DEBUG
	/* For rx overrun debuging */
	priv->rx_overrun = 0;
#endif
	scm80211_enable_rx_irq(hw);
	return 0;
}

static void scm80211_pcie_deinit_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct scm80211_rxdesc *rxd;
#ifndef SCM80211_RX_PAGE_ALLOC
	struct scm80211_rx_info *info;
#endif
	int i, retries = 10;
	u32 v;

	v = scm80211_mac_readl(hw, MAC_REG_RX_BUF_CFG);
	v &= ~RX_BUF_CFG_EN;
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_CFG);

	/* Wait for RXE to park */
	v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
	while (retries-- > 0) {
		mdelay(1);
		v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
		if (FIELD_GET(MAC_STATUS_DMA_BUSY, v) == 0)
			break;
	}

	scm80211_disable_rx_irq(hw);

	/* Cleanup rx ring */
	for (i = 0; i < ring_size(&priv->rxring); i++) {
		rxd = &priv->rxring.rxd[i];
#ifdef SCM80211_RX_PAGE_ALLOC
		if (rxd->dma_addr) {
			scm80211_unmap(hw, rxd->dma_addr, rx_buflen);
			dma_unmap_page(dev, rxd->dma_addr, rx_buflen, DMA_FROM_DEVICE);
		}
		if (rxd->skb) {
			kfree_skb(rxd->skb);
			rxd->skb = NULL;
		}
		page_ref_sub(rxd->page, USHRT_MAX - 1);
#else
		if (!rxd->skb)
			continue;

		info = SCM80211_SKB_RXCB(rxd->skb);
		scm80211_unmap(hw, info->dma_addr, rx_buflen);
		dma_unmap_single(dev, info->dma_addr, rx_buflen, DMA_FROM_DEVICE);

		kfree_skb(rxd->skb);
		rxd->skb = NULL;
#endif /* SCM80211_RX_PAGE_ALLOC */
		memset_io(rxd->config, 0, sizeof(*rxd->config));
	}

#ifdef SCM80211_RX_PAGE_ALLOC
	__free_pages(priv->rxring.pages, SCM80211_RX_PAGE_ORDER + priv->rxring.order);
#endif
}

static int scm80211_pcie_rx_start(struct ieee80211_hw *hw, struct sk_buff *skb,
				  struct scm80211_rx_data *rx)
{
	struct scm80211_rx_info *info = SCM80211_SKB_RXCB(skb);

	info->rxd.dlen = __le16_to_cpu(info->rxd.dlen);
	info->rxd.blen = __le16_to_cpu(info->rxd.blen);

	/*
	 * This is to detect and discard the case where rxd.dlen > rxd.blen and
	 * therefore causes skb over panic
	 */
	if (unlikely(info->rxd.dlen > info->rxd.blen)) {
		scm80211_warn(hw, "rxd[%02x]: invalid dlen %d", info->index, info->rxd.dlen);
		info->rxd.dlen = sizeof(struct rxvector) + sizeof(struct rx_mpdu_hdr);
		return -EINVAL;
	}

	skb_put(skb, info->rxd.dlen);
	return 0;
}

static void scm80211_enable_mcu_irq(struct ieee80211_hw *hw, int idx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);

	priv->int_en |= FIELD_PREP(INTR_ENABLE_MCU, BIT(idx));
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);

	spin_unlock_irqrestore(&priv->lock, flags);
}

static void scm80211_disable_mcu_irq(struct ieee80211_hw *hw, int idx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);

	priv->int_en &= ~FIELD_PREP(INTR_ENABLE_MCU, BIT(idx));
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);

	spin_unlock_irqrestore(&priv->lock, flags);
}

static int scm80211_pcie_init_tx(struct ieee80211_hw *hw)
{
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		hwq = &priv->hwq[i];
		hwq->txdesc = section->mem->tx + i;
	}

	if (priv->pdet_en)
		scm80211_enable_mcu_irq(hw, POWER_DETECT);

#ifdef CONFIG_HTC_IF
	for (i = HTC_IND0; i <= HTC_IND3; i++)
		scm80211_enable_mcu_irq(hw, i);
#endif

	return 0;
}

static void scm80211_pcie_deinit_tx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	scm80211_disable_mcu_irq(hw, POWER_DETECT);	/* disable MCU[0] */

#ifdef CONFIG_HTC_IF
	for (i = HTC_IND0; i <= HTC_IND3; i++)
		scm80211_disable_mcu_irq(hw, i);
#endif

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		struct txop *txop;
		enum scm80211_hwq_state qstate;
		int j, retries;
		u32 cmd;

		hwq = &hwp->hwq[i];
		for (j = 0; j < ARRAY_SIZE(hwq->ring.pool); j++) {
			txop = &hwq->ring.pool[j];
			if (txop->state == TXOP_STATE_ISSUED)
				break;
			txop = NULL;
		}
		if (!txop)
			continue;

		retries = txop->txtime / 10 + 1;
		scm80211_mac_writel(hw, 0, MAC_REG_TX_QUEN_CMD(i));
		while (retries-- > 0) {
			cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
			if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd) == 0)
				break;
			udelay(10);
		}

		if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
			scm80211_warn(hw, "failed to clean hwq#%d: cmd=%08x\n", hwq->id, cmd);

		qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
		scm80211_trace(hw, "%s(hwq#%d) finished: cmd=%08x state=%d\n",
			       __func__, hwq->id, cmd, qstate);
		if (qstate > HWQ_STATE_DELAY) {
			retries = 20;
			while (retries-- > 0) {
				udelay(10);
				qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
				if (qstate <= HWQ_STATE_DELAY)
					break;
			}
		}
	}
}

static int scm80211_pcie_cancel_tx(struct ieee80211_hw *hw, struct txop *txop, bool discard)
{
	struct scm80211_hwq *hwq = txop->hwq;
	enum scm80211_hwq_state qstate;
	u32 cmd;
	int retries = 10;

	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));

	scm80211_trace(hw, "%s(hwq#%d.%d) cmd=%08x state=%d\n",
		       __func__, hwq->id, txop->id, cmd, qstate);

	if (unlikely(txop->state != TXOP_STATE_ISSUED)) {
		scm80211_warn(hw, "- hwq#%d.%d not issued (state=%s)\n",
			      hwq->id, txop->id, txop_state_str[txop->state]);
		return 0;
	}

	if (!FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
		scm80211_warn(hw, "- hwq#%d.%d issue and sent out success?\n", hwq->id, txop->id);

	/* (1) Disable TX_DONE interrupt for this queue */
	scm80211_disable_tx_irq(hw, BIT(hwq->id));

	/* (2) If hardware queue is in transmission, wait until done */
	retries = (txop->txtime + txop->ack_time) / 10 + 1;
	qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
	while (qstate == HWQ_STATE_IN_TRANSMISSION && retries-- > 0) {
		udelay(10);
		hwq->qstate = qstate;
		qstate = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
	}

	/* tx complete */
	if (hwq->qstate == HWQ_STATE_IN_TRANSMISSION && qstate != HWQ_STATE_IN_TRANSMISSION)
		goto complete_txop;

#ifdef ERRATA_HWQ_STUCK
	if (qstate == HWQ_STATE_IN_TRANSMISSION) {
		/* TX hung happened, force clear the queue */
		scm80211_mac_writel(hwq->hw, BIT(hwq->id), MAC_REG_DEV_FORCE_QUE_CLR);
	}
#endif

	/* (4) Issue CMD_CANCEL blindly */
	scm80211_mac_writel(hw, 0, MAC_REG_TX_QUEN_CMD(hwq->id));

	/* (5) Loop until the MAC_REG_TX_QUEN_CMD[3:0] is read as zero. */
	while (retries-- > 0) {
		cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		if (!FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
			break;
		udelay(10);
	}

	/* (6) By this time, the hardware queue should have been parked. double check it */
	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	if (unlikely(FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd)))
		scm80211_warn(hw, "- queue still not parked: cmd=%08x state=%d\n", cmd, qstate);

	/*
	 * Because we masked TX_DONE interrupt before cancelling backoff,
	 * interrupt request is not raised even if the transmission was
	 * not cancelled and completed.
	 *
	 * We complete txop here on behalf of scm80211_tx_irq(). if TX_DONE
	 * interrupt is raised later after the interrupt is unmasked below,
	 * scm80211_tx_irq() will simply does nothing.
	 *
	 * If the txop is cancelled, mpdu descriptors will have their sent
	 * field set to 0. Such frames are discarded anyhow.
	 */
complete_txop:
	scm80211_complete_txop(txop, discard);

	/* Re-enable TX_DONE interrupt */
	scm80211_enable_tx_irq(hw, BIT(hwq->id));

	return -qstate;
}

static int scm80211_pcie_cleanup_tx(struct ieee80211_hw *hw, struct txop *txop, int idx)
{
	struct scm80211_tx_info *xti = &txop->xti[idx];
	struct device *dev = wiphy_dev(hw->wiphy);
	struct scatterlist *sg;
	int n;

	for_each_sg(xti->sgt.sgl, sg, xti->sgt.nents, n)
		scm80211_unmap(hw, sg_dma_address(sg), sg_dma_len(sg));

	dma_unmap_sg(dev, xti->sgt.sgl, xti->sgt.nents, DMA_TO_DEVICE);

	return 0;
}

static int scm80211_pcie_prepare_tx(struct ieee80211_hw *hw, struct txop *txop,
				    int idx, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = &txop->xti[idx];
	struct tx_mpdu_desc *mdesc = txop_get_mdesc(txop, idx);
	struct tx_buffer_desc *bdesc = txop_get_bdesc(txop, idx);
	struct sg_table *sgt = &xti->sgt;
	struct scatterlist *sg;
	struct device *dev = wiphy_dev(hw->wiphy);
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)(skb->data + sizeof(xti->tv));
	struct scm80211_txq *txq = txop->multi_owner ? txop->multi_owner : txop->owner;
	int n = scm80211_get_skb_fragments(skb);
	__le64 addr;

#if KERNEL_VERSION(5, 3, 0) <= LINUX_VERSION_CODE
	__sg_alloc_table(sgt, n, ARRAY_SIZE(xti->sg), xti->sg,
			 ARRAY_SIZE(xti->sg), GFP_KERNEL, NULL);
#else
	__sg_alloc_table(sgt, n, ARRAY_SIZE(xti->sg), xti->sg, GFP_KERNEL, NULL);
#endif
	sgt->nents = skb_to_sgvec(skb, sgt->sgl, 0, skb->len);
	if (unlikely(((int)sgt->nents) <= 0)) {
		scm80211_err(hw, "- failed to create skb scatterlist\n");
		return -EAGAIN;
	}

	sgt->nents = dma_map_sg(dev, sgt->sgl, sgt->nents, DMA_TO_DEVICE);
	if (unlikely(sgt->nents == 0)) {
		scm80211_err(hw, "- failed to map scatterlist\n");
		return -EAGAIN;
	}

	BUG_ON(sgt->nents > SCM80211_SKB_MAX_FRAGS);

	/* Program TX buffer descriptors */
	memset(bdesc, 0, sizeof(*bdesc) * SCM80211_SKB_MAX_FRAGS);
	for_each_sg(sgt->sgl, sg, sgt->nents, n) {
		/* Map host address into the target local address */
		addr = scm80211_remap(hw, sg_dma_address(sg), sg_dma_len(sg));
		if (!addr) {
			scm80211_err(hw, "- fail to remap bdesc %d\n", n);
			sgt->nents = n;
			goto fail;
		}

		put_unaligned_le64(addr, (void *)&bdesc[n].addr_l);
		bdesc[n].len = cpu_to_le32(sg_dma_len(sg));
		/* make sure it had write to target memory */
		wmb();
	}
	xti->flags |= SCM80211_XTI_DMA_MAPPED;

	memset(word(mdesc, 2), 0, sizeof(__le32) * 3);
	mdesc->aid = txq->txq.vif->bss_conf.aid;
	mdesc->tid = xti->tid;
	xti->sn = IEEE80211_SEQ_TO_SN(le16_to_cpu(hdr->seq_ctrl));
	mdesc->sn = xti->sn;

	/*
	 * NB: IEEE80211_TX_CTL_NO_ACK does not indicate the frame does not
	 * solicits a response frame. It only indicates the upper layer is
	 * not interested whether the frame is acknowledged or not. Therefore,
	 * @mdesc->noack should be set without referrering to it.
	 */
	mdesc->noack = !xti->ack;
	mdesc->mpdu_len = xti->mlen;

	/* PMF is done by software. H/W needs to bypass encyrption */
	/* PMF + unsupported cipher : key is NULL */
	if (ieee80211_has_protected(hdr->frame_control) &&
	    !IEEE80211_SKB_CB(skb)->control.hw_key) {
		mdesc->bypass_encrypt = true;
	}

	/* RTS protection is set in scm80211_finalize_txop() */
	mdesc->protection = NO_PROTECTION;
	/* EOF */
	mdesc->eof = xti->eof;

	/* FIXME: TSF */
	mdesc->timestamp_update = false;
	mdesc->timestamp_offset = false;

	if (txop->ampdu.enabled) {
		mdesc->aggregate = true;
		mdesc->mpdu_spacing = xti->mmss_pad;
	}

	mdesc->bdnum = cpu_to_le32(xti->sgt.nents - 1);
	if (ieee80211_is_mgmt(hdr->frame_control) || ieee80211_is_data_qos(hdr->frame_control))
		mdesc->no_htc = !ieee80211_has_order(hdr->frame_control);

	return 0;
fail:
	for_each_sg(sgt->sgl, sg, sgt->nents, n) {
		if (sg_dma_len(sg))
			scm80211_unmap(hw, sg_dma_address(sg), sg_dma_len(sg));
	}
	dma_unmap_sg(dev, sgt->sgl, sgt->nents, DMA_TO_DEVICE);
	return -EAGAIN;
}

static bool scm80211_pcie_start_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct txdesc txdesc;
	struct txop *issued = NULL;
	struct ieee80211_vif *vif;
	struct ieee80211_vif_priv *vifp;
	unsigned long flags;
	u32 bcnt, cmd;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	/* Check if txop is eligible for transmission */
	if (txop->ampdu.update_needed && !scm80211_txop_is_within_transmit_window(txop))
		goto exit;

	issued = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
	/* only one issued txop on hwq under pcie soluction */
	if (issued) {
		list_move(&issued->list, &hwq->ring.head[TXOP_STATE_ISSUED]);
		goto exit;
	}

	if (txop->state != TXOP_STATE_SCHEDULED && txop->state != TXOP_STATE_TX_RETRY) {
		scm80211_warn(hw, "- txop not ready for transmission status=%s\n",
			      txop_state_str[txop->state]);
		goto exit;
	}

	scm80211_trace(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

	bcnt = get_random_u32() & hwq->cw;

	memset(&txdesc, 0, sizeof(txdesc));
	txdesc.addr = txop->raddr;
	txdesc.num = txop->n_mpdu;
	/* operate the target memory */
	wmb();
	memcpy_toio(hwq->txdesc, &txdesc, offsetof(struct txdesc, htc));
	/* operate the target memory */
	wmb();

#ifndef SCM80211_WFA_TEST
	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd))
		scm80211_warn(hw, "hwq#%d busy? cmd=%08x\n", hwq->id, cmd);
#endif
	cmd = 0;
	/*
	 * 9.4.2.28 EDCA Parameter Set element
	 *      The minimum value of the AIFSN subfield is 2.
	 *
	 * 26.2.7 EDCA operation using MU EDCA parameters.
	 * The backoff counter maintenance corresponding to the updated
	 *   state variables shall follow the rules in
	 *   10.23.2.2 (EDCA backoff procedure) except that if AIFSN[AC] is 0
	 *   then the EDCAF corresponding to that AC shall be suspended until
	 *    the MUEDCATimer[AC] reaches 0 or is reset to 0.
	 */
	/* When AIFSN[AC] != 0 or during scan, enable EDCA/MU EDCA */
	if (hwq->edca.aifs || priv->scanner.scanning)
		cmd |= FIELD_PREP(TX_QUEN_CMD_EDCA, 1);

	/* If mu_edca running, during scan, using non-mu_edca param */
	if ((hwq->mu_edca_status & MU_EDCA_RUNNING) && priv->scanner.scanning) {
		bcnt = get_random_u32() & hwq->b_edca.cw_min;
		hwq->edca.aifs = hwq->b_edca.aifs;
	}
	txop->bcnt = bcnt;

	cmd |= FIELD_PREP(TX_QUEN_CMD_BOFF_COUNT, bcnt);
	cmd |= FIELD_PREP(TX_QUEN_CMD_AIFSN, hwq->edca.aifs);

	vif = txop->owner->txq.vif;
	vifp = ieee80211_to_priv(vif);
	/* Enable HE TB */
	if (vif && vif->type == NL80211_IFTYPE_STATION) {
		struct ieee80211_bss_conf *bss_conf = &vif->bss_conf;
		u32 mmss = vifp->reg_mmss;

		if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_EN, mmss)) {
			if (bss_conf->aid && bss_conf->he_support)
				cmd |= FIELD_PREP(TX_QUEN_CMD_AID_MATCHED_QUEUE_EN, 1);

			if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_ASSO_UORA_EN, mmss) &&
			    bss_conf->he_support)
				cmd |= FIELD_PREP(TX_QUEN_CMD_ASSO_UORA_QUEUE_EN, 1);
			else if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_UNASSO_UORA_EN, mmss))
				cmd |= FIELD_PREP(TX_QUEN_CMD_UNASSO_UORA_QUEUE_EN, 1);
		}
	}

#ifdef SCM80211_TWT
	{
		struct sk_buff *skb = skb_peek(&txop->sched);
		struct ieee80211_hdr *hdr = NULL;
		enum scm80211_twt_state twt_state;

		twt_state = scm80211_get_twt_state(vif);
		if (twt_state == TWT_DOZE) {
			scm80211_err(hw, "- TWT DOZE state should not transmit frame!!\n");
			goto exit;
		}

		if (skb) {
			hdr = (void *)(skb->data + sizeof(struct txvector));
		} else {
			WARN(1, "-%s no skb?\n", __func__);
			goto exit;
		}

		/* Workaround: HE_TB Action frame sent but no ack response */
		if (twt_state == TWT_SP && vifp->twt_trigger) {
			scm80211_trace(hw, "Clear CMD_EACA\n");
			cmd &= ~TX_QUEN_CMD_EDCA;
		}

		if (hdr && ieee80211_is_action(hdr->frame_control) && twt_state != TWT_SP) {
			cmd &= ~TX_QUEN_CMD_HE_TB_RESP_MASK;
			cmd |= FIELD_PREP(TX_QUEN_CMD_EDCA, 1);
		}
	}
#endif
	/*
	 * Report bug when AIFS == 0 && no HE TB response bits set
	 * v == 0 means cancel TX
	 */
	if (WARN_ON((!hwq->edca.aifs && !(cmd & TX_QUEN_CMD_HE_TB_RESP_MASK)) || !cmd)) {
		/* set default value */
		cmd |= FIELD_PREP(TX_QUEN_CMD_AIFSN, 2);
		cmd |= FIELD_PREP(TX_QUEN_CMD_EDCA, 1);
	}

	trace_scm80211_start_txop(hw, txop, txop->owner, hwq->edca.aifs,
				  bcnt, txop->n_mpdu, txop->ampdu.enabled);

	scm80211_mac_writel(hw, txop->limit, MAC_REG_TX_QUEN_TXOP(hwq->id));
	scm80211_mac_writel(hw, cmd, MAC_REG_TX_QUEN_CMD(hwq->id));

#ifdef SCM80211_COEX
	if (vif->type == NL80211_IFTYPE_P2P_CLIENT ||
			 vif->type == NL80211_IFTYPE_STATION) {
		if (txop->hwq->id % SCM80211_NUM_HWQ_PER_VIF == SCM80211_HWQ_HI)
			scm80211_pta_pri_set(hw, 1);
		else
			scm80211_pta_pri_set(hw, 0);
	}
#endif

	hwq->tstamp = ktime_get();
	if (FIELD_GET(TX_QUEN_CMD_AIFSN, cmd))
		scm80211_hwq_start_watchdog(hwq);

	/* Make sure code below placed in the final */
	__scm80211_txop_change_state(txop, TXOP_STATE_ISSUED);
	__scm80211_txop_add(txop, TXOP_STATE_ISSUED);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
	return true;

exit:
	__scm80211_txop_change_state(txop, TXOP_STATE_SCHEDULED);
	__scm80211_txop_add(txop, TXOP_STATE_SCHEDULED);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
	return false;
}

static void scm2020_mac_check_sanity(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 v;
	u16 wptr, rptr, bptr;

	v = scm80211_mac_readl(hw, MAC_REG_RX_BUF_PTR0);
	rptr = FIELD_GET(RX_BUF_PTR0_READ_PTR, v);
	bptr = FIELD_GET(RX_BUF_PTR0_READY_PTR, v);

	v = scm80211_mac_readl(hw, MAC_REG_RX_BUF_PTR1);
	wptr = FIELD_GET(RX_BUF_PTR1_WRITE_PTR, v);

#define PTR_CHECK(xptr)			\
	{				\
		if (ring_ptr(&priv->rxring, priv->rxring.xptr) != xptr)	\
			scm80211_warn(hw, "rx ring " __stringify(xptr) " mismatch: %04lx, %04x\n",\
				      ring_ptr(&priv->rxring, priv->rxring.xptr), xptr); \
	}

	PTR_CHECK(rptr);
	PTR_CHECK(bptr);
	PTR_CHECK(wptr);

	scm80211_trace(hw, "rptr %04lx, bptr %04lx, wptr %04lx\n",
		       ring_ptr(&priv->rxring, priv->rxring.rptr),
		       ring_ptr(&priv->rxring, priv->rxring.bptr),
		       ring_ptr(&priv->rxring, priv->rxring.wptr));
}
/**
 * TODO:
 * scm80211_pcie_vif_doze info target to enter doze mode
 */
static void scm80211_pcie_vif_doze(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				   bool twt, bool link)
{

	struct ieee80211_hw_priv  *priv = hw->priv;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	s32 delta;
	u32 v, awake_tsf;
	unsigned long start = jiffies;
	int timeout = 0;

	if (link && !scm80211_ps_enabled(hw)) {
		scm80211_warn(hw, "enter into powersave mode is not allowed\n");
		return;
	}

	if (scm80211_tx_frames_pending(hw))
		return;

#ifdef SCM80211_TWT
	if (!link) {
		delta = S32_MAX;
	} else if (twt) {
		vifp->sw_tsf += ktime_us_delta(ktime_get(), vifp->twt_ktime);
		vifp->twt_ktime = ktime_get();
		vifp->twt_next_irq = vifp->twt - SCM80211_TWT_AHEAD_SP;
		scm80211_trace(hw, "twt %llu (wakeup at %llu) current tsf %llu\n",
			       vifp->twt, vifp->twt_next_irq, vifp->sw_tsf);
		if (vifp->twt_next_irq > vifp->sw_tsf)
			delta = (vifp->twt_next_irq - vifp->sw_tsf) & S32_MAX;
	} else {
		delta = (vifp->bss.beacon.next - scm80211_get_tsf32(hw, vif)) & S32_MAX;
	}
#else
	delta = (vifp->bss.beacon.next - scm80211_get_tsf32(hw, vif)) & S32_MAX;
#endif
	if (delta < ieee80211_tu_to_usec(30)) {
		scm80211_info(hw, "only %dus left, no need to enter doze\n", delta);
		return;
	}

	priv->pm.doze	= true;

	scm2020_powersave(hw, true);
	/* WLAN power domain power off sequence */
	pm_save(hw, priv->pm.doze_awake, MAC_REG, mac_doze_awake_regs);
	scm80211_mac_writel(hw, 0, MAC_REG_INTR_ENABLE);

	/* Park RX */
	v = scm80211_mac_readl(hw, MAC_REG_RX_BUF_CFG);
	v &= ~(RX_BUF_CFG_EN | RX_BUF_CFG_DMA_EN);
	scm80211_mac_writel(hw, v, MAC_REG_RX_BUF_CFG);

	/* Wait for RXE DMA to finish */
	v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
	while (FIELD_GET(MAC_STATUS_RXE_BUSY, v)) {
		udelay(2);
		v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
		if (timeout-- < 0)
			break;
	}

	v = scm80211_mac_readl(hw, MAC_REG_MAC_CFG) & ~MAC_CFG_EN;
	scm80211_mac_writel(hw, v, MAC_REG_MAC_CFG);

	/* Handle pending rx */
	scm80211_pcie_rx(hw);
	scm2020_mac_check_sanity(hw);

	pm_save(hw, priv->pm.rxring, MAC_REG, mac_rxring_regs);

	scm80211_mac_writel(hw, scm80211_mac_readl(hw, MAC_IMEM_OFFSET), MAC_IMEM_OFFSET);
	scm80211_mac_writel(hw, scm80211_mac_readl(hw, MAC_DMEM_OFFSET), MAC_DMEM_OFFSET);

	if (link) {
		awake_tsf = vifp->bss.beacon.next - vifp->bss.beacon.latency -
			    jiffies_to_usecs(jiffies - start) - vifp->bss.beacon.dur;
		if (!twt) {
			/* reserved some time for configuration */
			delta = ((awake_tsf - scm80211_get_tsf32(hw, vif)) & S32_MAX);
			scm80211_dbg(hw, "info target awake at %u, delta=%dus\n",
				     awake_tsf, delta);
		}
	} else {
		twt = true;
		delta = S32_MAX;
	}

#if PM_LOCAL_TEST
	/* to info target when to doze mode tsf... */
	mod_timer(&vifp->pm.wake_timer, jiffies + usecs_to_jiffies(delta));
#endif
	ep->doze = true;
	if (twt)
		scm80211_pcie_param_write(hw, vifp->index, TWT_DOZE_ENABLE,
					  &delta, sizeof(delta));
	else
		scm80211_pcie_param_write(hw, vifp->index, DOZE_ENABLE,
					  &awake_tsf, sizeof(awake_tsf));

}

static int scm80211_pcie_next_beacon(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				     struct sk_buff *skb)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_bss_conf *conf = &vif->bss_conf;
	struct ieee80211_mgmt	  *beacon = (struct ieee80211_mgmt *) skb->data;
	const struct element	  *elem;
	const struct ieee80211_tim_ie *tim;
	int n_tbtt = 1;


	vifp->bss.beacon.next = vifp->bss.beacon.last + ieee80211_tu_to_tsf(conf->beacon_int);
	if (!scm80211_ps_enabled(hw))
		return 0;

	elem = cfg80211_find_elem(WLAN_EID_TIM, beacon->u.beacon.variable,
				  skb->len - offsetof(struct ieee80211_mgmt, u.beacon.variable[0]));

	if (!elem || elem->datalen < sizeof(*tim))
		return -EINVAL;

	tim = (const struct ieee80211_tim_ie *) elem->data;
	n_tbtt = tim->dtim_count ? tim->dtim_count : tim->dtim_period;
	n_tbtt = min_t(int, n_tbtt, hw->conf.listen_interval);
	vifp->bss.beacon.next = vifp->bss.beacon.last +
				n_tbtt * ieee80211_tu_to_tsf(conf->beacon_int);

	if (ieee80211_check_tim(tim, elem->datalen, conf->aid)) {
		scm80211_trace(hw, "have more unicast for me,  stay awake state...\n");
		ieee80211_queue_work(priv->hw, &priv->pm.disable_work);
	} else if (tim->dtim_count == 0) {
		int ti = tim->bitmap_ctrl & BIT(0);

		if (!ti & vifp->bss.traffic_indicator_used) {
			scm80211_trace(hw, "goes to doze mode immediately\n");
			//mod_timer(&vifp->pm.doze_timer, jiffies + usecs_to_jiffies(1));
			scm80211_pcie_vif_doze(hw, vif, false, true);
		} else {
			scm80211_trace(hw, "goes to doze mode later (have broadcast)\n");
			if (ti)
				vifp->bss.traffic_indicator_used = true;
			mod_timer(&vifp->pm.doze_timer, jiffies + msecs_to_jiffies(2));
		}
	} else {
		scm80211_trace(hw, "goes to doze mode right away\n");
		scm80211_pcie_vif_doze(hw, vif, false, true);
		//mod_timer(&vifp->pm.doze_timer, jiffies + usecs_to_jiffies(1));
	}

	return 0;
}

static void scm80211_pcie_start_vif_doze(struct timer_list *doze)
{
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_vif *vif;
	struct ieee80211_hw *hw;

	vifp = from_timer(vifp, doze, pm.doze_timer);
	if (WARN_ON(!vifp))
		return;

	vif  = priv_to_ieee80211(struct ieee80211_vif, vifp);
	hw   = vifp->hw;

	scm80211_pcie_vif_doze(hw, vif, scm80211_get_twt_state(vif) != TWT_DISABLED, true);
}

static void scm80211_reinstall_keys(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif,
				    struct ieee80211_sta *sta,
				    struct ieee80211_key_conf *key,
				    void *data)
{
	_scm80211_set_key(hw, SET_KEY, vif, sta, key, true);
}

static void scm80211_pcie_vif_wake(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct ieee80211_chanctx_conf *chanctx = get_ieee80211_chanctx_conf(vif);
	struct phy_vector pv;
	struct scm80211_phy *phy = phy_next(hw, NULL, false);
	u32 v;
	int timeout = 10;

	del_timer(&vifp->pm.doze_timer);
	del_timer(&vifp->pm.wake_timer);

	WARN_ON(!ep->doze);
#ifdef SCM2020_A0
	/* check WIFI power domain first */
	v = scm80211_raw_readl(hw, WLAN_PMU_CFG);
	while (!(v & BIT(8)) && timeout--) {
		/* no power domain */
		udelay(10);
		v = scm80211_raw_readl(hw, WLAN_PMU_CFG);
	}
#endif

	ep->doze = false;
	timeout = 10;

	/* Run MCU */
	/* scm80211_mac_writel(hw, 0x1, MAC_REG_MCU_CFG); */
	do {
		v = scm80211_mac_readl(hw, MAC_REG_MCU_STATUS);
		if (v & MCU_STATUS_DONE)
			break;
		udelay(10);
	} while (timeout-- > 10);

	if (!(v & MCU_STATUS_DONE)) {
		scm80211_err(hw, "MAC MCU did not awaked up\n");
		return;
	}

	/* restore keys */
	rcu_read_lock();
	ieee80211_iter_keys_rcu(hw, vif, scm80211_reinstall_keys, NULL);
	rcu_read_unlock();

	pm_restore(hw, vifp->pm.assoc_awake);
	pm_restore(hw, priv->pm.rxring);
	scm2020_mac_check_sanity(hw);
	pm_restore(hw, priv->pm.doze_awake);
	/* enable rx irq manually to fixup no rx irq after wakeup */
	scm80211_enable_rx_irq(hw);
	/* restore channel */
	pv.BANDWIDTH = CBW20;
	if (vif->bss_conf.assoc) {
		pv.BAND = chanctx->def.chan->band;
		pv.CENTER_FREQUENCY = chanctx->def.chan->center_freq;
	} else {
		pv.BAND = priv->sched.idle.ctx.def.chan->band;
		pv.CENTER_FREQUENCY = priv->sched.idle.ctx.def.chan->center_freq;
	}
#ifdef SCM2020_A0
	phy->radio->ops->power_save(phy->radio, 0, false);
#endif
	scm80211_set_channel(hw, phy, &pv);

	priv->pm.doze = false;
	scm80211_refill_rxring(hw);
#ifdef SCM80211_TWT
	if (scm80211_get_twt_state(vif) == TWT_DOZE)
		scm80211_set_twt_state(vif, TWT_SP);
#endif
}

static void scm80211_pcie_start_vif_wake(struct timer_list *wake)
{
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_vif *vif;
	struct ieee80211_hw *hw;

	vifp = from_timer(vifp, wake, pm.wake_timer);
	if (WARN_ON(!vifp))
		return;

	vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	hw  = vifp->hw;

	scm80211_pcie_vif_wake(hw, vif);
}

static int scm80211_pcie_add_vif(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	int vid = scm80211_vdev_id(vif);


	timer_setup(&vifp->pm.doze_timer, scm80211_pcie_start_vif_doze, 0);
	timer_setup(&vifp->pm.wake_timer, scm80211_pcie_start_vif_wake, 0);


	scm80211_pcie_param_write(hw, vid, VIF0_START + (vid * 2), NULL, 0);
	msleep(10);
	return 0;
}

static void scm80211_pcie_del_vif(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	int vid = scm80211_vdev_id(vif);

	del_timer_sync(&vifp->pm.doze_timer);
	del_timer_sync(&vifp->pm.wake_timer);
	scm80211_pcie_param_write(hw, vid, VIF0_STOP + (vid * 2), NULL, 0);
	msleep(5);
}

static int scm80211_pcie_assoc_bss(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				   struct ieee80211_bss_conf *bss_info)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_vif   vif_info;

	pm_save(hw, priv->pm.assoc_active, MAC_REG, mac_assoc_active_regs);
	pm_save(hw, priv->pm.phy_assoc_active, PHY_REG, phy_assoc_active_regs);

	pm_save(hw, vifp->pm.assoc_awake, MAC_REG, mac_vif_assoc_awake_regs[vifp->index]);
	pm_save(hw, vifp->pm.assoc_active, MAC_REG, mac_vif_assoc_active_regs[vifp->index]);

	memset(&vif_info, 0, sizeof(vif_info));
	vif_info.type = (u8) vif->type;
	vif_info.idx  = vifp->index;
	vif_info.bss_conf.associated = true;
	memcpy(vif_info.bss_conf.bssid, vif->bss_conf.bssid, ETH_ALEN);
	memcpy(vif_info.bss_conf.transmitter_bssid,
	       vif->bss_conf.transmitter_bssid, ETH_ALEN);
	scm80211_pcie_param_write(hw, vifp->index, CONNECTED, &vif_info,
				  sizeof(struct scm80211_vif));
	msleep(5);
	return 0;
}

static void scm80211_pcie_leave_bss(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				    struct ieee80211_bss_conf *bss_info)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct scm80211_vif   vif_info;

	memset(&vif_info, 0, sizeof(vif_info));
	vif_info.type = (u8) vif->type;
	vif_info.idx  = vifp->index;

	scm80211_pcie_param_write(hw, vifp->index, DISCONNECTED, &vif_info,
				  sizeof(struct scm80211_vif));
	msleep(5);
}

const struct scm80211_bus_ops scm80211_pcie_ops = {
	.mac_writel	= scm80211_pcie_mac_writel,
	.mac_readl	= scm80211_pcie_mac_readl,
	.phy_writel	= scm80211_pcie_phy_writel,
	.phy_readl	= scm80211_pcie_phy_readl,
	.rf_writel	= scm80211_pcie_rf_writel,
	.rf_readl	= scm80211_pcie_rf_readl,
	.pta_writel	= scm80211_pcie_pta_writel,
	.pta_readl	= scm80211_pcie_pta_readl,
	.raw_writel	= scm80211_pcie_raw_writel,
	.raw_readl	= scm80211_pcie_raw_readl,
	.efuse_writel	= scm80211_pcie_efuse_writel,
	.efuse_readl	= scm80211_pcie_efuse_readl,
	.remap		= scm80211_pcie_remap,
	.unmap		= scm80211_pcie_unmap,
	.inverse_map	= scm80211_pcie_inverse_map,
	.hwq_alloc_dma	= scm80211_pcie_hwq_alloc_dma,
	.hwq_free_dma	= scm80211_pcie_hwq_free_dma,
	.fixup_tx	= scm80211_pcie_fixup_tx,
	.lookup_tx	= scm80211_pcie_lookup_tx,
	.set_channel	= scm80211_pcie_set_channel,
	.set_ant_sel	= scm80211_pcie_set_ant_sel,
	.get_ant_sel	= scm80211_pcie_get_ant,
	.power_save	= scm80211_pcie_power_save,
	.init_rx	= scm80211_pcie_init_rx,
	.deinit_rx	= scm80211_pcie_deinit_rx,
	.refill_rxring	= scm80211_pcie_refill_rxring,
	.rx_start	= scm80211_pcie_rx_start,
	.init_tx	= scm80211_pcie_init_tx,
	.deinit_tx	= scm80211_pcie_deinit_tx,
	.vif_init_tx	= scm80211_pcie_add_vif,
	.vif_deinit_tx	= scm80211_pcie_del_vif,
	.prepare_tx	= scm80211_pcie_prepare_tx,
	.start_tx	= scm80211_pcie_start_tx,
	.cleanup_tx	= scm80211_pcie_cleanup_tx,
	.cancel_tx	= scm80211_pcie_cancel_tx,
	.vif_power_save	= scm80211_pcie_vif_power_save,
	.next_beacon	= scm80211_pcie_next_beacon,
	.associated	= scm80211_pcie_assoc_bss,
	.disconnected	= scm80211_pcie_leave_bss,
};
const struct scm80211_bus_ops *scm_bus_ops = &scm80211_pcie_ops;

int scm80211_request_firmware(const struct firmware **fw, const char *name,
		     struct device *device)
{
	int ret = 0;
	u8 retry = 120;

	/*
	 * workaround for Sugon platform
	 * request FW failed occasionally as driver loads during system boot
	 * retry to request FW bin, and sleep 100ms after each failure
	 * until 12s or successful to exit retry
	 */
	do {
		if (ret)
			msleep(100);
		ret = request_firmware(fw, name, device);
	} while (ret && retry--);

	return ret;
}

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
static void __iomem *pcie_download_addr(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	struct dler_ctrl __iomem *ctl = ep->mem_base + DLER_CTRL_MSG_OFFSET;
	return ctl;
}

static int scm80211_response_check(struct pci_dev *pdev)
{
	struct dler_ctrl __iomem *ctl = pcie_download_addr(pdev);
	char *msg = ctl->msg;
	int i;

	for (i = 0; i < 3000; i++) {
		if (!strcmp(msg, RESP_OK))
			return 0;
		if (!strcmp(msg, RESP_FAIL))
			return -EINVAL;
		mdelay(1);
	}
	return -EAGAIN;
}

#define CA7_ALREADY_BOOTUP	1

#define scm80211_bp_in_sram(bp_base)  \
	((bp_base) >= FW_BASE && (bp_base) < FW_BASE + SRAM_MEM_SIZE)

static int wise_check(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	struct device *dev = &pdev->dev;
	const struct firmware *fw;
	int ret, i;

	ret = scm80211_request_firmware(&fw, ep->config->wise, dev);
	if (ret) {
		dev_err(dev, "%s request fw error (errno %d)!\n", __func__, ret);
		return ret;
	}

	ret = CA7_ALREADY_BOOTUP;
	/* Check the first 4 words */
	for (i = 0; i < 4; i++) {
		u32 v[2];

		v[0] = get_unaligned_le32(fw->data + i * 4);
		v[1] = readl(ep->mem_base + i * 4);

		if (v[0] != v[1]) {
			ret = -EINVAL;
			break;
		}
	}

	release_firmware(fw);
	return ret;
}

static int pcie_wait_for_dev_init(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	int ret = -EINVAL, i = WAIT_CNT;

	do {
		ep->last_bp_base = pcie_ep_mem_addr(ep, 4, ep->bp_base);
		if (scm80211_bp_in_sram(ep->last_bp_base)) {
			ret = 0;
			break;
		}
		mdelay(WAIT_TIME);
	} while (i--);

	return ret;
}

static int scm80211_pcie_ctrl_preamble(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	struct device *dev = &pdev->dev;
	struct dler_ctrl __iomem *ctl = pcie_download_addr(pdev);
	char __iomem *msg = ctl->msg;
	int ret = 0, i;

	dev_info(dev,
		 "bp_base=%p, last_bp_base=%llx, mem_base=%p, ep_memp_base=%llx\n",
		 ep->bp_base, ep->last_bp_base, ep->mem_base, ep->ep_mem_base);

	if (!scm80211_bp_in_sram(ep->last_bp_base)) {
		ret = wise_check(pdev);
		if (ret == CA7_ALREADY_BOOTUP)
			return ret;
		ret = pcie_wait_for_dev_init(pdev);
	}

	if (ret) {
		dev_err(dev,
			"wait for pcie download mode fail, bp_base=%p, last_bp_base=%llx\n",
			ep->bp_base, ep->last_bp_base);
		return ret;
	}

	/* do preamble */
	for (i = 0; i < RETRY_CNT; ++i) {
		int j;

		memset_io(msg, '\0', sizeof(ctl->msg));
		memcpy_toio(msg, CMD_PREAMBLE, strlen(CMD_PREAMBLE) + 1);
		for (j = 0; j < WAIT_CNT; j++) {
			mdelay(WAIT_TIME);
			if (!strncmp(msg, RESP_UPCI, strlen(RESP_UPCI)))
				return 0;
		}
	}

	ret = -EINVAL;
	return ret;
}

/**
 * scm80211_pcie_start_wise() - download wise to CA7 and boot it
 */
static int scm80211_pcie_start_wise(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	struct device *dev = &pdev->dev;
	const struct firmware *fw;
	void __iomem *ram = ep->mem_base;
	struct dler_ctrl __iomem *ctl = pcie_download_addr(pdev);
	char *msg = ctl->msg;
	int ret;
	bool boot = false;

	/* bp bar is not available */
	if (ep->last_bp_base == 0)
		return -EIO;

	ret = scm80211_pcie_ctrl_preamble(pdev);
	if (ret == CA7_ALREADY_BOOTUP) {
		dev_dbg(dev, "wise have been loaded\n");
		return 0;
	} else if (ret) {
		dev_err(dev, "preamble fail!\n");
		return -EAGAIN;
	}

	/* dump version */
	memset_io(msg, '\0', sizeof(ctl->msg));
	memcpy_toio(msg, CMD_VER, strlen(CMD_VER) + 1);
	ret = scm80211_response_check(pdev);
	if (ret == 0) {
		dev_dbg(dev, "Version info: %s\n", ctl->data);
	} else {
		dev_err(dev, "Get version fail\n");
		return ret;
	}

	/* download fw */
	ret = scm80211_request_firmware(&fw, ep->config->wise, dev);
	if (ret) {
		dev_err(dev, "%s request fw error (errno %d)!\n", __func__, ret);
		goto unmap;
	}

	dev_dbg(dev, "downloading wise %s (%zu bytes)\n", ep->config->wise, fw->size);

	__iowrite32_copy(ram, fw->data, fw->size / sizeof(u32));
	/* wait target memory operation done */
	wmb();

	/* check CRC */
	memset_io(msg, '\0', sizeof(ctl->msg));
	sprintf(ctl->data, "%x:%x:%lx", 0, FW_BASE, (unsigned long)fw->size);
	memcpy_toio(msg, CMD_CRC, strlen(CMD_CRC) + 1);
	ret = scm80211_response_check(pdev);
	if (ret == 0) {
		u32 crc, crc2;

		ret = kstrtou32(ctl->data, 16, &crc);
		if (ret)
			goto fw_release;
		crc2 = crc32(SC2_CRC32_INIT, fw->data, fw->size);
		dev_info(dev, "%s crc cal by ca7 0x%08x cal by driver 0x%08x\n",
			 __func__, crc, crc2);
		if (crc != crc2) {
			dev_err(dev, "pcie download fail for crc mismatch");
			ret = -EAGAIN;
			goto fw_release;
		}
	} else {
		dev_err(dev, "crc fail: %s\n", ret == -EINVAL ? "no response" : "crc mismatch");
		goto fw_release;
	}

	boot = true;

fw_release:
	release_firmware(fw);
unmap:
	if (boot) {
		/* boot */
		dev_dbg(dev, "%s boot flow\n", __func__);
		memset_io(msg, '\0', sizeof(ctl->msg));
		memcpy_toio(ctl->data, "0:20000000", strlen("0:20000000") + 1);
		memcpy_toio(msg, CMD_BOOT, strlen(CMD_BOOT) + 1);
		ret = scm80211_response_check(pdev);
		if (ret == 0)
			dev_dbg(dev, "wise boot success!\n");
		else
			dev_err(dev, "wise boot fail!\n");
	}

	/* end */
	memset_io(msg, '\0', sizeof(ctl->msg));
	memcpy_toio(msg, CMD_END, strlen(CMD_END) + 1);

	return ret;
}
#endif

/**
 * scm80211_pcie_start_mcu() - download MCU microcode and kick it
 */
static int scm80211_pcie_start_mcu(struct ieee80211_hw *hw)
{
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	struct device *dev = wiphy_dev(hw->wiphy);
	const struct firmware *fw;
	struct hw_reg_section *section = scm80211_get_bus_data(hw);
	void __iomem *imem = section->mac_base + MAC_IMEM_OFFSET;
	int err, retries;
	u32 v;

	err = scm80211_request_firmware(&fw, ep->config->firmware, dev);
	if (err) {
		scm80211_err(hw, "failed to request firmware (err=%d)\n", err);
		return err;
	}

	scm80211_trace(hw, "downloading ucode %s (%zu bytes)\n", ep->config->firmware, fw->size);

	__iowrite32_copy(imem, fw->data, fw->size / sizeof(u32));
	/* wait target memory operation done */
	wmb();

	/* Kick MCU */
	scm80211_trace(hw, "starting mcu\n");

	scm80211_mac_writel(hw, 0, MAC_REG_MCU_CFG);
	udelay(10);

	v = MCU_CFG_RSTB;
	scm80211_mac_writel(hw, v, MAC_REG_MCU_CFG);

	err = -EBUSY;
	for (retries = 0; retries < MCU_START_RETRIES; retries++) {
		v = scm80211_mac_readl(hw, MAC_REG_MCU_STATUS);
		if (FIELD_GET(MCU_STATUS_DONE, v) == 1) {
			err = 0;
			break;
		}
		scm80211_trace(hw, "retry=%d MCU STATUS=0x%x\n", retries, v);
		mdelay(MCU_START_DELAY);
	}
	if (err)
		scm80211_err(hw, "failed to start mcu\n");

	print_hex_dump_bytes("mcode", DUMP_PREFIX_OFFSET, imem + 0x200, 64);

	release_firmware(fw);
	return err;
}

static int scm2020_pcie_print_info(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	int bar;

	for (bar = 0; bar <= PCI_ROM_RESOURCE; bar++) {
		if (!pci_resource_len(pdev, bar))
			continue;
		dev_dbg(&pdev->dev, "bar%d: %pR, base=%#llx\n", bar,
			&pdev->resource[bar], (u64)pcim_iomap_table(pdev)[bar]);
	}

	dev_dbg(&pdev->dev, "ep bar3 mem base addr=%lx\n", (unsigned long)ep->ep_mem_base);
	dev_dbg(&pdev->dev, "ep bar4 backplane addr=%lx\n", (unsigned long)ep->last_bp_base);
	dev_dbg(&pdev->dev, "ep: num_atu=%zu, page_size=%zx, page_shift=%d\n",
		ep->config->num_atu, ep->page_size, ep->page_shift);

	dev_dbg(&pdev->dev, "ep: elbi offset=%08x\n", ep->elbi_offset);

	return 0;
}

static u16 aspm, aspm_parent;
static void scm80211_disable_aspm(struct pci_dev *pdev)
{
	struct pci_dev *parent = pdev->bus->self;

	/* disable the aspm manually */
	if (!aspm) {
		pcie_capability_read_word(pdev, PCI_EXP_LNKCTL, &aspm);
		if (parent)
			pcie_capability_read_word(parent, PCI_EXP_LNKCTL, &aspm_parent);

		aspm &= PCI_EXP_LNKCTL_ASPMC;
		aspm_parent &= PCI_EXP_LNKCTL_ASPMC;
	}

	if (!aspm && (!parent || !aspm_parent)) {
		dev_dbg(&pdev->dev, "aspm have disabled as default\n");
		return;
	}

	dev_info(&pdev->dev, "disabling ASPM %s %s\n",
		 (aspm & PCI_EXP_LNKCTL_ASPM_L0S) ? "L0s" : "",
		 (aspm & PCI_EXP_LNKCTL_ASPM_L1) ? "L1" : "");

	pcie_capability_clear_word(pdev, PCI_EXP_LNKCTL, aspm);
	if (parent)
		pcie_capability_clear_word(parent, PCI_EXP_LNKCTL, aspm_parent);
}

static void scm80211_enable_aspm(struct pci_dev *pdev)
{
	struct pci_dev *parent = pdev->bus->self;

	if (!aspm)
		return;
	if (parent)
		pcie_capability_clear_and_set_word(parent, PCI_EXP_LNKCTL,
						   PCI_EXP_LNKCTL_ASPMC,
						   aspm_parent);
	pcie_capability_clear_and_set_word(parent, PCI_EXP_LNKCTL,
					   PCI_EXP_LNKCTL_ASPMC,
					   aspm);

}

/**
 * scm80211_rx_napi_irq() - handle rx done interrupt request
 *
 * This function check the read buffer invalid and schedule the napi callback function
 */
static void scm80211_rx_napi_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	scm80211_sync_rx_wptr(hw);
	if (ring_ptr_equal(priv, priv->rxring.rptr, priv->rxring.wptr))
		return;

	scm80211_trace(hw,
		       "rx irq: bptr=0x%04lx, rptr=0x%04lx, wptr=0x%04lx\n",
		       ring_ptr(&priv->rxring, priv->rxring.bptr),
		       ring_ptr(&priv->rxring, priv->rxring.rptr),
		       ring_ptr(&priv->rxring, priv->rxring.wptr));

	trace_scm80211_rx_irq(hw,
			      ring_ptr(&priv->rxring, priv->rxring.bptr),
			      ring_ptr(&priv->rxring, priv->rxring.rptr),
			      ring_ptr(&priv->rxring, priv->rxring.wptr));

	scm80211_disable_rx_irq(hw);
	napi_schedule(&priv->napi.context);
}

void scm80211_tx_irq(struct ieee80211_hw *hw, unsigned long queues)
{
	struct scm80211_hwq *hwq;
	struct txop *txop;
	int i;

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		txop = scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
		if (!txop) {
			scm80211_warn(hw, "%s(hwq#%d) bogus interrupt\n", __func__, hwq->id);
			continue;
		}

		scm80211_trace(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

		/* Read and store hardware txop context */
		txop->cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		txop->txd.result = readl(&hwq->txdesc->result);
		txop->retry_cnt = 1;
		txop->scheduled = true;
		if (txop->sta && txop->txd.cw)
			txop->txd.htc = readl(&hwq->txdesc->htc);

		/* Debugging */
		if (hwq->txdesc->addr != txop->raddr) {
			scm80211_warn(hw, "- txdesc addr=%llx, txop addr=%lx\n",
				      hwq->txdesc->addr, (unsigned long)txop->raddr);
		}
		scm80211_handle_txop_complete(hw, txop);
	}
}


#ifdef SCM2020_A0
static void scm80211_pcie_htc_irq(struct ieee80211_hw *hw, int idx)
{
	struct pcie_endpoint *ep	= pcie_endpoint_from_hw(hw);
	struct ieee80211_hw_priv *priv	= hw->priv;
	struct htc_content __iomem *t2h = ep->t2h[idx];
	struct htc_content htc;

	memcpy_fromio(&htc, t2h, sizeof(htc));
	scm80211_trace(hw, "%s [%02x]\n", __func__, idx);
	switch (htc.event) {
	case ACTIVE_ENABLE:
		WARN_ON(htc.vif_id > 1 || !priv->vifs[htc.vif_id]);
#if !PM_LOCAL_TEST
		scm80211_pcie_vif_wake(hw, priv->vifs[htc.vif_id]);
#endif
		break;
	default:
		break;
	}
}
#endif

static void scm80211_mcu_irq(struct ieee80211_hw *hw, unsigned long bitmap)
{
#ifdef SCM2020_A0
	int i;

	for_each_set_bit(i, &bitmap, SCM80211_MCU_IRQ_NUM) {
		switch (i) {
		case POWER_DETECT:
			scm80211_tx_start_irq(hw);
			break;
		case HTC_IND0 ... HTC_IND3:
			scm80211_pcie_htc_irq(hw, i - HTC_IND0);
			break;
		default:
			break;
		}
	}
#endif
}

irqreturn_t scm80211_handle_irq(int irq, void *data)
{
	struct ieee80211_hw *hw = data;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);
	u32 v, en, queues, mcu;
	unsigned long flags;

retries:
	spin_lock_irqsave(&ep->lock, flags);
	v = scm80211_mac_readl(hw, MAC_REG_INTR_STATUS);
	if (v == 0) {
		spin_unlock_irqrestore(&ep->lock, flags);
		return IRQ_NONE;
	}

	/* Acknowledge the interrupts */
	scm80211_mac_writel(hw, v, MAC_REG_INTR_CLEAR);
	spin_unlock_irqrestore(&ep->lock, flags);

	en = ((struct ieee80211_hw_priv *)hw->priv)->int_en;
	scm80211_trace(hw, "irq=%08x enable=%08x\n", v, en);

	mcu = FIELD_GET(INTR_STATUS_MCU, v);
	scm80211_mcu_irq(hw, (unsigned long)mcu);

	if (v & (INTR_STATUS_TSF0 | INTR_STATUS_TSF1))
		scm80211_tsf_irq(hw, v);

	if (v & (INTR_STATUS_RX_DONE | INTR_STATUS_RX_BUF))
		scm80211_rx_napi_irq(hw);

	queues = FIELD_GET(INTR_STATUS_TX_DONE, v);
#ifdef SCM80211_FTM
	if (!scm80211_ftm_tx_irq(hw, (unsigned long)queues))
#endif
		scm80211_tx_irq(hw, (unsigned long)queues);

	if (v & INTR_STATUS_BASIC_TRIG_QNULL_VIF)
		scm80211_handle_muedca_irq(hw, FIELD_GET(INTR_STATUS_BASIC_TRIG_QNULL_VIF, v));

#ifdef SCM80211_COEX
	if (v & INTR_STATUS_COEX)
		scm80211_coex_irq(hw);
#endif

	/*
	 * See if new interrupts arrived while we are handling the older ones.
	 *
	 * NB: If new interrupts are not handled, the driver will not receive
	 *     any further interrupts.
	 *
	 *
	 *              +-------------------------+
	 *              |                         |
	 * TX_DONE(i)---+                         +-------------
	 *
	 *                                 +--------------------
	 *                                 |
	 * RX_DONE   ----------------------+
	 *
	 *               +--------------------------------------+
	 *               |
	 * WLAN_IRQ -----+
	 *
	 *               T0               T1     T2
	 *
	 * If interrupt handler hanldes only TX_DONE(i) and clears that
	 * interrupts at T2, WLAN_IRQ line keeps staying asserted, and
	 * the PCIe controller will not generate further interrupts.
	 *
	 * Therefore, the driver needs to read again the MAC_REG_INTR_STATUS
	 * register and if there is new interrupts sneaking in between T0 and T2,
	 * it needs to process it.
	 *
	 */
	v = scm80211_mac_readl(hw, MAC_REG_INTR_STATUS);
	if (v)
		goto retries;

	/* Is it necessary */
	if (priv->dbdc) {
		hw = ep->hws[(priv->devid + 1) % ep->intf_num];
		priv = hw->priv;

		v = scm80211_mac_readl(hw, MAC_REG_INTR_STATUS);
		if (v)
			goto retries;
	}

	return IRQ_HANDLED;
}

static int scm80211_soc_init(struct pci_dev *pdev)
{
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	int path;
	int ret;
	u32 v;
#endif

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	ret = scm80211_pcie_start_wise(pdev);
	if (ret) {
		if (ret == -EIO) {
			dev_info(&pdev->dev, "have no BAR4\n");
		} else {
			dev_err(&pdev->dev, "failed to start wise\n");
			return -ENODEV;
		}
	}

	mdelay(3000);

#ifndef SCM80211_PS
	/* quickly enable pll clk */
	v = pcie_mmio_readl(ep, WLAN_TOP_CFG);
	v |= FIELD_PREP(PLL_CLK_EN, 1);
	v |= FIELD_PREP(PLL_EN, 1);
	pcie_mmio_writel(ep, v, WLAN_TOP_CFG);

	v = pcie_mmio_readl(ep, WLAN_MCU_CFG);
#ifndef __TEMP_DEBUG__
	/* Use pll 960MHz clock source */
	v &= ~(AHB_CLK_SET | AXI_CLK_SET | MCU_CLK_SET | MCU_CLK_DIV);
	v |= FIELD_PREP(MCU_CLK_SET, 1);
	v |= FIELD_PREP(MCU_CLK_DIV, 3);
#ifdef SCM2020_A0
	v |= FIELD_PREP(AXI_CLK_SET, 4);	/* Setup bit:2 to enable pll clk source for AXI */
#endif
	v |= FIELD_PREP(AXI_CLK_SET, 2);	/* 1: div 4(240MHz) 2: div 8(120MHz) */
	v |= FIELD_PREP(AHB_CLK_SET, 2);
	dev_notice(&pdev->dev, "%s: WLAN_MCU_CFG=0x%x\n", __func__, v);
#else
	/* Use external 80MHz clock source */
	/* AXI clk to 40MHz / AHB clk to 20MHz */
	v &= ~(AHB_CLK_SET | AXI_CLK_SET | MCU_CLK_SET | MCU_CLK_DIV);
#endif
	pcie_mmio_writel(ep, v, WLAN_MCU_CFG);
#endif
	/* quickly enable mac/phy clk */
	v = pcie_mmio_readl(ep, WLAN_WMP_CFG);
	v |= PHY_CLK_EN;
	v |= PHY_PCLK_EN;
	v |= MAC_CLK_M1X_EN(0);
	if (ep->intf_num > 1)
		v |= MAC_CLK_M1X_EN(1);	//for path 1
	else
		v &= (~(MAC_CLK_M1X_EN(1)));	//disable path-1 explicitly
	for (path = 0; path < 2; path++) {
		// v |= DPD_960M_CLK_EN(path);
		// v |= DPD_320M_CLK_EN(path);
		v |= TX_EN(path);
		v |= RX_EN(path);
	}
	pcie_mmio_writel(ep, v, WLAN_WMP_CFG);

	/* reset and enable rfdig */
	v = pcie_mmio_readl(ep, WCN_RFDIG_CFG);
	if (!FIELD_GET(RFDIG_CLK_EN, v)) {
		pcie_mmio_writel(ep, 0, WCN_RFDIG_CFG);
		udelay(10);
		v = FIELD_PREP(RFDIG_CLK_EN, 1) | 0x30;
		pcie_mmio_writel(ep, v, WCN_RFDIG_CFG);
		mdelay(2);
	}
#endif
	return 0;
}

static struct ieee80211_hw *scm80211_alloc_hw(struct pci_dev *pdev, int idx)
{
	struct ieee80211_hw *hw = NULL;
	struct ieee80211_hw_priv *priv;
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	struct hw_reg_section *section;

	dev_dbg(&pdev->dev, "%s: WLAN interface#%d\n", __func__, idx);

	hw = ieee80211_alloc_hw(sizeof(*priv) + sizeof(*section), &scm80211_ops);
	if (!hw) {
		dev_err(&pdev->dev, "failed to alloc hw\n");
		return NULL;
	}

	SET_IEEE80211_DEV(hw, &pdev->dev);

	priv = hw->priv;
	priv->hw = hw;
	priv->devid = idx;
	priv->dbdc = (ep->intf_num > 1) ? true : false;
	priv->bus.ops = &scm80211_pcie_ops;
	priv->bus_type = PCIE_BUS;
	if (pdet_en)
		priv->pdet_en = true;
	else
		priv->pdet_en = false;
	scm80211_set_debug_trace(hw);
	scm80211_hw_band_set(hw);
	ewma_tx_sched_time_init(&priv->sched_time);

	section = (void *)priv->bus.data;
	section->pdev = pdev;
	BUG_ON(ep->hws[priv->devid]);
	ep->hws[priv->devid] = hw;
	atomic_inc(&ep->ref);

	section->mac_base	= ep->mac_base + 0x10000 * idx;
	section->phy_base	= ep->phy_base; /* FIXME should it be ep->phy_base + 0x4000 ? */
	section->rf_base	= RF_BASE_START; /* FIXME should handle it separately ? */
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	section->pta_base	= PTA_BASE_START;
#endif
	section->mem		= ep->mac_base + 0x10000 * idx + MAC_SHM_OFFSET;

	return hw;
}

int scm80211_phy_rf_attach(struct ieee80211_hw *hw, int idx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int ret;
	struct scm80211_rf_device *rf;

	scm80211_trace(hw, "init phy/rf for hwmac#%d\n", priv->devid);

	ret = scm2020_chip_init(hw);
	if (ret) {
		scm80211_err(hw, "failed to do int phy\n");
		return ret;
	}

	rf = scm80211_alloc_rf_device(hw, NULL, priv->devid);
	if (!rf) {
		scm80211_err(hw, "failed to find RF\n");
		return -ENOMEM;
	}

	phy_attach_rf(hw, rf);
	return 0;
}

static int scm80211_pcie_request_irq(struct ieee80211_hw *hw, struct pci_dev *pdev, int idx)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->irq = pci_irq_vector(pdev, idx);
	return devm_request_irq(&pdev->dev, priv->irq, scm80211_handle_irq,
				IRQF_SHARED, "scm80211_pcie", hw);
}

static int _scm80211_pcie_set_dma_mask(struct pci_dev *pdev, u64 dma_bit_mask)
{
	int ret;

#if KERNEL_VERSION(5, 17, 0) >= LINUX_VERSION_CODE
	ret = pci_set_dma_mask(pdev, dma_bit_mask);
#else
	ret = dma_set_mask(&pdev->dev, dma_bit_mask);
#endif
	if (ret) {
		dev_err(&pdev->dev, "failed to set dma mask ret=%d\n", ret);
		return ret;
	}

#if KERNEL_VERSION(5, 17, 0) >= LINUX_VERSION_CODE
	ret = pci_set_consistent_dma_mask(pdev, dma_bit_mask);
#else
	ret = dma_set_coherent_mask(&pdev->dev, dma_bit_mask);
#endif
	if (ret)
		dev_err(&pdev->dev, "failed to set coherent dma mask ret=%d\n", ret);

	return ret;
}

static int scm80211_pcie_set_dma_mask(struct pci_dev *pdev)
{
	int ret;
	/*
	 * our ATU can only map 2G range of memory, we will suffer from no ATU issue
	 * when IOMMU/SMMU is disabled and physical memory is larger than 2G.
	 */
	bool iommu_enabled = iommu_en ? iommu_get_domain_for_dev(&pdev->dev) != NULL : false;
	u64 dma_bit_mask = iommu_enabled ? DMA_BIT_MASK(31) : DMA_BIT_MASK(64);

	if (iommu_enabled)
		dev_dbg(&pdev->dev, "enable IOMMU backed DMA mapping\n");

	ret = _scm80211_pcie_set_dma_mask(pdev, dma_bit_mask);

	if (ret && !iommu_enabled) {
		dev_err(&pdev->dev, "DMA mask set 64 failed, try to set 32\n");
		ret = _scm80211_pcie_set_dma_mask(pdev, DMA_BIT_MASK(32));
	}

	return ret;
}

static int scm80211_pcie_dev_init(struct pci_dev *pdev, const struct pci_device_id *id)
{
	int ret, bar, pos;
	u32 mask = 0;

	aspm = 0;
	aspm_parent = 0;
	scm80211_disable_aspm(pdev);

	ret = pcim_enable_device(pdev);
	if (ret) {
		dev_err(&pdev->dev, "unable to enable pci device %s ret=%d\n", pci_name(pdev), ret);
		return ret;
	}

	ret = scm80211_pcie_set_dma_mask(pdev);
	if (ret)
		goto fail1;

	pci_set_master(pdev);

	for (bar = 0; bar <= PCI_ROM_RESOURCE; bar++) {
		if (pci_resource_len(pdev, bar) && pci_resource_flags(pdev, bar) & IORESOURCE_MEM)
			mask |= BIT(bar);
	}

	ret = pcim_iomap_regions(pdev, mask, pci_name(pdev));
	if (ret) {
		dev_err(&pdev->dev, "failed to request iomap regions %08x for %s ret=%d\n",
			mask, pci_name(pdev), ret);
		goto fail1;
	}

	ret = pcie_endpoint_init(pdev, id, mask);
	if (ret) {
		dev_err(&pdev->dev, "failed to init drvdata ret=%d\n", ret);
		goto fail2;
	}

	ret = pci_alloc_irq_vectors(pdev, 1, dbdc ? 2 : 1, PCI_IRQ_MSI);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to allocate msi vectors ret=%d\n", ret);
		goto fail3;
	}

	ret = scm80211_soc_init(pdev);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to init soc ret = %d\n", ret);
		goto fail4;
	}

	pci_enable_pcie_error_reporting(pdev);
	/*
	 * clear RxErr of AER(PCI_ERR_COR_RCVR)
	 * PCI_EXT_CAP_ID_ERR(0x100)->PCI_ERR_COR_STATUS(0x10)
	 */
	pos = pci_find_ext_capability(pdev, PCI_EXT_CAP_ID_ERR);
	if (pos)
		pci_write_config_dword(pdev, pos + PCI_ERR_COR_STATUS, PCI_ERR_COR_RCVR);

	return ret;
fail4:
	pci_free_irq_vectors(pdev);
fail3:
	devm_kfree(&pdev->dev, pci_get_drvdata(pdev));
fail2:
	pcim_iounmap_regions(pdev, mask);
fail1:
	pci_disable_device(pdev);

	return ret;
}

static void scm80211_pcie_dev_deinit(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);
	u32 mask = ep->barmask;

	pci_disable_pcie_error_reporting(pdev);
	pci_free_irq_vectors(pdev);
	pcim_iounmap_regions(pdev, mask);
	pci_set_drvdata(pdev, NULL);
	devm_kfree(&pdev->dev, ep);
	pci_disable_device(pdev);
}

static int scm80211_pcie_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	struct pcie_endpoint *ep;
	struct ieee80211_hw *hw;
	int ret, i, j, succ = 0;
	bool rand_mac = true;
	u8 addr[ETH_ALEN] = { 0x64, 0xf9, 0x47, 0x00, 0x00, 0x00 };
	enum {
		HW_ALLOC, START_MCU, PHY_RF_ATTACH, HW_INIT, HW_REGISTER,
	} fail_step;

#ifdef EFUSE_ENABLE
	int cal[4] = { 0 };
	int xcal = 0;
#endif

	dev_notice(&pdev->dev, "Senscomm wireless device: %s, build time=%s\n",
		   dev_name(&pdev->dev), DRIVER_DATE);

	scm80211_register_oops_handler();

	ret = scm80211_pcie_dev_init(pdev, id);
	if (ret) {
		dev_err(&pdev->dev, "failed to pcie dev pre init ret=%d\n", ret);
		goto fail1;
	}

	ep = pci_get_drvdata(pdev);
#ifdef EFUSE_ENABLE
	if (!scm80211_efuse_pcie_init(ep))
		rand_mac = !scm80211_efuse_mac_update(addr, NULL);
#endif
	for (i = 0; i < ep->intf_num; i++) {
		hw = scm80211_alloc_hw(pdev, i);
		if (!hw) {
			fail_step = HW_ALLOC;
			ret = -ENOMEM;
			goto fail2;
		}

		ret = scm80211_pcie_start_mcu(hw);
		if (ret < 0) {
			fail_step = START_MCU;
			scm80211_err(hw, "failed to start mac#%d mcu", i);
			goto fail2;
		};

		ret = scm80211_phy_rf_attach(hw, i);
		if (ret) {
			fail_step = PHY_RF_ATTACH;
			scm80211_err(hw, "failed to attach phy rf ret = %d", ret);
			goto fail2;
		}

		if (rand_mac) {
			addr[3] = PCI_FUNC(pdev->devfn) >> 8;
			addr[4] = PCI_FUNC(pdev->devfn) + i;
			addr[5] = prandom_u32() & 0xff;
		} else {
			addr[4] += i;
		}

		scm80211_pcie_request_irq(hw, pdev, i);

		ret = scm80211_init_hw(hw, addr);
		if (ret) {
			fail_step = HW_INIT;
			scm80211_err(hw, "failed to init hw = %d", ret);
			goto fail2;
		}

		ret = ieee80211_register_hw(hw);
		if (ret) {
			fail_step = HW_REGISTER;
			scm80211_err(hw, "failed to register hw = %d", ret);
			goto fail2;
		}
#ifdef CONFIG_MAC80211_DEBUGFS
		scm80211_register_dbgfs(hw);
#endif
		scm80211_rfkill_init(hw);
		succ++;
	}

#ifdef EFUSE_ENABLE
	if (!scm80211_efuse_cal_update(0, cal))
		scm2020_pdet_set_para_b(0, cal[0], cal[1], cal[2], cal[3]);
	if (!scm80211_efuse_cal_update(1, cal))
		scm2020_pdet_set_para_b(1, cal[0], cal[1], cal[2], cal[3]);
	xcal = scm80211_efuse_get_xtal_cal();
	if (xcal != -255)
		scm2020_set_xtal_cal_b(xcal);
#endif

#ifdef CONFIG_HTC_IF
	for (i = 0; i < SCM80211_NUM_DEV; i++) {
		ep->htc_ctx.htc_queues[i].dev =
			(ep->hws[i]) ?
				ieee80211_hw_to_priv(((struct ieee80211_hw *)(ep->hws[i]))) : NULL;
	}
	scm80211_htc_init(pdev);
#endif
	scm2020_pcie_print_info(pdev);
	return 0;

fail2:
	WARN_ON(ret == 0);
	if (hw) {
		switch (fail_step) {
		case HW_REGISTER:
		case HW_INIT:
			devm_free_irq(&pdev->dev, pci_irq_vector(pdev, i), hw);
			scm80211_detach_rf(hw);
			scm80211_deinit_hw(hw);
			break;
		case PHY_RF_ATTACH:
		case START_MCU:
			ieee80211_free_hw(hw);
			break;
		case HW_ALLOC:
		default:
			break;
		}
	}

	for (j = succ - 1; j >= 0; j--) {
		hw = ep->hws[j];
		BUG_ON(!hw);
		devm_free_irq(&pdev->dev, pci_irq_vector(pdev, j), hw);
		scm80211_rfkill_deinit(hw);
#ifdef CONFIG_MAC80211_DEBUGFS
		scm80211_unregister_dbgfs(hw);
#endif
		ieee80211_unregister_hw(hw);
		scm80211_detach_rf(hw);
		scm80211_deinit_hw(hw);
	}

	scm80211_pcie_dev_deinit(pdev);
fail1:
	scm80211_unregister_oops_handler();
	return ret;
}

static void scm80211_pcie_remove(struct pci_dev *pdev)
{
	int i = 0;
	struct ieee80211_hw *hw;
	struct ieee80211_hw_priv *priv;
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);

	if (!ep) {
		dev_err(&pdev->dev, "driver data staled\n");
		return;
	}

	dev_notice(&pdev->dev, "%s\n", __func__);

#ifdef CONFIG_HTC_IF
	scm80211_htc_deinit(pdev);
#endif
	for (i = 0; i < ep->intf_num; i++) {
		hw = ep->hws[i];
		if (!hw) {
			dev_err(&pdev->dev, "driver data staled\n");
			continue;
		}
		if (ep->doze) {
			scm80211_pcie_param_write(hw, -1, HW_PS_DISABLE, NULL, 0);
			msleep(2);
		}
		priv = hw->priv;
		scm80211_trace(hw, "removing %pM\n", hw->wiphy->perm_addr);

		/* Mind the order */
		devm_free_irq(&pdev->dev, pci_irq_vector(pdev, priv->devid), hw);
		scm80211_rfkill_deinit(hw);

		scm80211_deinit_chanctx_sched(hw);
#ifdef CONFIG_MAC80211_DEBUGFS
		scm80211_unregister_dbgfs(hw);
#endif
		ieee80211_unregister_hw(hw);
		scm80211_detach_rf(hw);
		scm80211_deinit_hw(hw);
	}

	scm80211_pcie_dev_deinit(pdev);
	scm80211_unregister_oops_handler();
}

static int scm80211_pcie_suspend(struct pci_dev *pdev, pm_message_t state)
{
#if defined(SCM2020_A0) || defined(SCM2020_Z1)
	dev_dbg(&pdev->dev, "pcie_suspend\n");
#endif
	return 0;
}

static int scm80211_pcie_resume(struct pci_dev *pdev)
{
#if defined(SCM2020_A0) || defined(SCM2020_Z1)
	struct ieee80211_hw *hw;
	struct pcie_endpoint *drvdata = pci_get_drvdata(pdev);

	hw = drvdata->hws[0];
	dev_dbg(&pdev->dev, "pcie_resume\n");
	scm80211_soc_init(pdev);
	scm80211_pcie_start_mcu(hw);
	scm2020_rfreg_init(hw);
#endif
	return 0;
}

static const struct pcie_endpoint_config scm2020_config = {
	.firmware = "senscomm/scm2020.bin",
	.wise = "senscomm/wise-pcie.bin",
	.axi_base = 0x80000000,
	.axi_size = SZ_2G,
	.num_atu = 64,
};

static const struct pci_device_id scm80211_pcie_ids[] = {
	{ PCI_DEVICE(0x1eb9, 0x2020),
	 .driver_data = (kernel_ulong_t)&scm2020_config },
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	{ PCI_DEVICE(0x1e5d, 0x5000),
	 .driver_data = (kernel_ulong_t)&scm2020_config },
#endif
	{ },
};

MODULE_DEVICE_TABLE(pci, scm80211_pcie_ids);

/**
 * scm80211_pcie_error_detected() - called when PCI error is detected
 * The error_detected() method should shut down any ongoing I/O operations,
 * but should not attempt to communicate with the adapter itself.
 * This method can take locks and sleep; it is called from process context.
 * The return value tells the error recovery subsystem how to proceed;
 * it can be PCIERR_RESULT_CAN_RECOVER (the driver thinks it will be able
 * to recover just by talking to the adapter), PCIERR_RESULT_NEED_RESET
 * (a hard reset of the adapter will be required), or PCIERR_RESULT_DISCONNECT
 * (the situation is hopeless, and the adapter should be considered permanently dead).
 * @pdev: pointer to PCI device
 * @state: current state of the PCI slot
 */
#if KERNEL_VERSION(5, 9, 0) > LINUX_VERSION_CODE
static pci_ers_result_t scm80211_pcie_error_detected(struct pci_dev *pdev,
						     enum pci_channel_state state)
#else
static pci_ers_result_t scm80211_pcie_error_detected(struct pci_dev *pdev,
						     pci_channel_state_t state)
#endif
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);

	if (!ep) {
		dev_err(&pdev->dev, "driver data staled\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}

	dev_err(&pdev->dev, "%s: PCIe error detect, state %d\n", __func__, state);
	if (state == pci_channel_io_perm_failure)
		return PCI_ERS_RESULT_DISCONNECT;

	/* Request that MMIO be enabled, so register dump can be taken. */
	return PCI_ERS_RESULT_CAN_RECOVER;
}

/**
 * scm80211_pcie_mmio_enabled - Enable MMIO and dump debug registers
 * If all drivers on an affected PCI segment think they can recover from the problem,
 * the next step is to turn memory-mapped I/O back on and let the drivers try.
 * To this end, each driver's mmio_enabled() callback will be invoked.
 * This callback should do whatever port banging is required to get the adapter
 * back into a reasonable state, then return one of PCIERR_RESULT_RECOVERED (it worked),
 * PCIERR_RESULT_NEED_RESET (it failed, try resetting), or PCIERR_RESULT_DISCONNECT
 * (it failed, abandon all hope). Regardless of the outcome,
 * the driver should not restart I/O from this callback.
 * @pdev: pointer to PCI device
 */
static pci_ers_result_t scm80211_pcie_mmio_enabled(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);

	if (!ep) {
		dev_err(&pdev->dev, "driver data staled\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}
	dev_err(&pdev->dev, "%s: PCIe mmio enabled\n", __func__);

	return PCI_ERS_RESULT_RECOVERED;
}

/**
 * scm80211_pcie_slot_reset() - called when the pci bus has been reset.
 * @pdev: pointer to PCI device
 */
static pci_ers_result_t scm80211_pcie_slot_reset(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);

	if (!ep) {
		dev_err(&pdev->dev, "driver data staled\n");
		return PCI_ERS_RESULT_DISCONNECT;
	}
	dev_err(&pdev->dev, "%s: PCIe slot reset\n", __func__);

	return PCI_ERS_RESULT_RECOVERED;
}

/**
 * scm80211_pcie_resume() - resume normal ops after PCI reset
 * Called when the error recovery driver tells us that its
 * OK to resume normal operation.
 * @pdev: pointer to PCI device
 */
static void scm80211_pcie_err_resume(struct pci_dev *pdev)
{
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);

	if (!ep) {
		dev_err(&pdev->dev, "driver data staled\n");
		return;
	}
	dev_err(&pdev->dev, "%s: PCIe resume\n", __func__);
}

static const struct pci_error_handlers scm80211_pcie_err_handler = {
	.error_detected	= scm80211_pcie_error_detected,
	.mmio_enabled	= scm80211_pcie_mmio_enabled,
	.slot_reset	= scm80211_pcie_slot_reset,
	.resume		= scm80211_pcie_err_resume,
};

static struct pci_driver scm80211_pcie_driver = {
	.name		= "scm80211-pcie",
	.id_table	= scm80211_pcie_ids,
	.probe		= scm80211_pcie_probe,
	.remove		= scm80211_pcie_remove,
	.suspend	= scm80211_pcie_suspend,
	.resume		= scm80211_pcie_resume,
	.err_handler	= &scm80211_pcie_err_handler,
#ifdef __not_yet__
	.driver		= NULL,
#endif
};

static int scm80211_pcie_init(void)
{
	int ret;

	ret = scm80211_minstrel_init();
	if (ret)
		pr_info("rate control module has been registered\n");

	ret = scm80211_init_rf_modules();
	if (ret) {
		pr_err("failed to register rf drivers\n");
		scm80211_minstrel_exit();
		goto exit;
	}

	ret = pci_register_driver(&scm80211_pcie_driver);
	if (ret) {
		pr_err("failed to register scm80211 pcie driver: %d\n", ret);
		scm80211_exit_rf_modules();
		scm80211_minstrel_exit();
	}

exit:
	return ret;
}

module_init(scm80211_pcie_init);

static void scm80211_pcie_exit(void)
{
	pci_unregister_driver(&scm80211_pcie_driver);
	scm80211_exit_rf_modules();
	scm80211_minstrel_exit();
}

module_exit(scm80211_pcie_exit);

MODULE_VERSION(DRIVER_VERSION);
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Senscomm Semiconductor, Inc.");

/*module_pci_driver(scm80211_pcie_driver);*/

#ifdef CONFIG_HTC_IF
struct scm80211_htc_ctx *pcie_get_htc_ctx(struct pci_dev *pdev)
{
	struct scm80211_htc_ctx *htc_ctx;
	struct pcie_endpoint *ep = pci_get_drvdata(pdev);

	htc_ctx = ep ? &ep->htc_ctx : NULL;

	return htc_ctx;
}

struct scm80211_htc_ctx *scm80211_get_htc_ctx(struct ieee80211_hw *hw)
{
	struct scm80211_htc_ctx *htc_ctx;
	struct pcie_endpoint *ep = pcie_endpoint_from_hw(hw);

	htc_ctx = ep ? &ep->htc_ctx : NULL;

	return htc_ctx;
}

void scm80211_pcie_rc_notify_ep(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	pcie_raise_endpoint_irq(pci_dev_from_hw(hw), priv->devid);
}
#endif
