// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#include <linux/bitfield.h>
#include <net/cfg80211.h>

#define func_fmt(str)           "[PHY] "str
#define func_trace              PHY

#include "core.h"
#include "phy.h"
#include "util.h"
#include "trace.h"
#include "debug.h"

static int temp_interval;
module_param(temp_interval, int, 0644);
MODULE_PARM_DESC(temp_interval, "interval for temp comp");

/**
 * PHY hardware abstraction layer
 */

void phy_writel(struct scm80211_phy *phy, u32 val, u32 offset)
{
	struct ieee80211_hw *hw = phy->hw;

	scm80211_phy_writel(hw, val, phy->start + offset);
}

u32 phy_readl(struct scm80211_phy *phy, u32 offset)
{
	struct ieee80211_hw *hw = phy->hw;

	return scm80211_phy_readl(hw, phy->start + offset);
}

struct scm80211_phy *phy_next(struct ieee80211_hw *hw, struct scm80211_phy *phy, bool mimo)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (priv->dbdc) {
		if (!phy)
			return &priv->phy[priv->devid];

		return NULL;
	}

	if (!phy)
		return &priv->phy[0];

	phy++;
	while (phy < priv->phy + priv->n_chain) {
		if (phy->hw == hw) {
			if (!mimo)
				return phy;
			else if (phy->active)
				return phy;
		}
		phy++;
	}
	return NULL;
}

#define is_present(v, which)	((v)->present & BIT(PHY_##which))

#define both_have(v1, v2, which)				\
	(is_present(v1, which) && is_present(v2, which))

#define has_conflict(v1, v2, which)				\
	(is_present(v1, which) && is_present(v2, which)	&&	\
	 ((v1)->which != (v2)->which))

#define is_different(v1, v2, which)		\
	((v1)->which != (v2)->which)

#define merge(v1, v2, which, op)	(v1)->which = (v1)->which op (v2)->which

#define bit(which)		BIT(PHY_##which)

#define copy(v1, v2, which)				\
	do {						\
		if (!is_present(v1, which)) {		\
			(v1)->which = (v2)->which;	\
			if ((v2)->present & bit(which))	\
				(v1)->present |= bit(which);	\
		}					\
	} while (0)					\

unsigned long combine_phy_vector(struct phy_vector *v1, struct phy_vector *v2, bool dryrun)
{
	struct phy_vector local, *v3;
	unsigned long ret = 0;

	if (dryrun) {
		v3 = &local;
		memcpy(v3, v1, sizeof(*v3));
	} else {
		v3 = v1;
	}

	v3->ACTIVE_RXCHAIN_SET = max(v2->ACTIVE_RXCHAIN_SET, v3->ACTIVE_RXCHAIN_SET);
	/* VHT */
	merge(v3, v2, LISTEN_TO_GID00, |);
	merge(v3, v2, LISTEN_TO_GID63, |);

	if (has_conflict(v3, v2, PARTIAL_AID_LIST_GID00))
		v3->VHT_SU_RX_FILTER_EN = false;
	else
		copy(v3, v2, PARTIAL_AID_LIST_GID00);

	if (has_conflict(v3, v2, PARTIAL_AID_LIST_GID63))
		v3->VHT_SU_RX_FILTER_EN = false;
	else
		copy(v3, v2, PARTIAL_AID_LIST_GID63);

	if (both_have(v3, v2, GROUP_ID_MANAGEMENT)) {
		/*
		 * We are in a trouble. There are more than one
		 * VHT STA vif in the same chanctx that has to receive VHT
		 * MU PPDUs.
		 */
		ret |= BIT(PHY_GROUP_ID_MANAGEMENT);
	} else if (!is_present(v1, GROUP_ID_MANAGEMENT)) {
		memcpy(&v1->GROUP_ID_MANAGEMENT, &v2->GROUP_ID_MANAGEMENT,
		       sizeof(v1->GROUP_ID_MANAGEMENT));
	}

	/* HE */
	if (has_conflict(v3, v2, BSS_COLOR)) {
		/*
		 * If more than two VHT interfaces want to use different
		 * BSS_COLOR, disable BSS_COLOR filter. This will at least
		 * allow two interfaces to receive HE SU PPDUs at the cost
		 * of increased power consumption.
		 */
		pr_warn("BSS COLOR conflicts: %d, %d\n", v3->BSS_COLOR, v2->BSS_COLOR);
		v3->BSS_COLOR_FILTER_EN = false;
	} else {
		copy(v3, v2, BSS_COLOR);
		copy(v3, v2, BSS_COLOR_FILTER_EN);
	}

	if (has_conflict(v3, v2, STA_ID_INDIVIDUAL)) {
		/*
		 * Cannot merge - v2 will not be able to receive
		 * individually addressed HE MU PPDUs.
		 */
		ret |= BIT(PHY_STA_ID_INDIVIDUAL);
	} else {
		copy(v3, v2, STA_ID_INDIVIDUAL);
	}

	if (has_conflict(v3, v2, STA_ID_GROUP)) {
		/*
		 * Cannot merge - v2 will not be able to receive
		 * group addressed HE MU PPDUs
		 */
		ret |= BIT(PHY_STA_ID_GROUP);
	} else {
		copy(v3, v2, STA_ID_GROUP);
	}

	merge(v3, v2, HE_DIR_EN, |);

	return ret;
}

static void phy_vector_set_channel(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct cfg80211_chan_def *chandef = &chanctx->ctx.def;
	struct phy_vector *pv = &chanctx->pv;
	u32 f0, bandwidth;

	pv->CENTER_FREQUENCY = chandef->center_freq1;
	pv->BAND = (chandef->chan->band == NL80211_BAND_2GHZ) ? 0 : 1;

	switch (chandef->width) {
	case NL80211_CHAN_WIDTH_80:
		pv->BANDWIDTH = 2;
		break;
	case NL80211_CHAN_WIDTH_40:
		pv->BANDWIDTH = 1;
		break;
	case NL80211_CHAN_WIDTH_20_NOHT:
	case NL80211_CHAN_WIDTH_20:
		pv->BANDWIDTH = 0;
	default:
		break;
	}

	bandwidth = 20 * (1 << pv->BANDWIDTH);
	f0 = chandef->center_freq1 - bandwidth / 2 + 10;
	pv->PRIMARY_20MHZ_CHANNEL_LOC = (chandef->chan->center_freq - f0) / 20;

	set_bit(PHY_CHANNEL, &pv->present);
}

static int phy_vector_set_rx_chain(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct phy_vector *pv = &chanctx->pv;
	struct scm80211_phy *phy;

	/*
	 * @priv->n_chains (A): # of PHY chains allocated to @hw
	 * @conf->rx_chains_static (B): # of RX chains that must be always active
	 * @conf->rx_chains_dynamic (C): # of RX chains for SMPS
	 *
	 * C >= B >= A
	 */
	if (chanctx->ctx.rx_chains_static > priv->n_chain) {
		scm80211_warn(hw,
			      "chains (%d) > # of hardware chains (%d)\n",
			      chanctx->ctx.rx_chains_static, priv->n_chain);
		return -EINVAL;
	}

	pv->ACTIVE_RXCHAIN_SET = (1 << priv->n_chain) - 1;
	pv->MIMO_EN = (chanctx->ctx.rx_chains_dynamic > 1) ? 1 : 0;
	priv->active_chains = pv->ACTIVE_RXCHAIN_SET;
	pv->present |= BIT(PHY_ACTIVE_RXCHAIN_SET);

	for_each_phy(phy, hw) {
		if (priv->active_chains & BIT(phy->chain))
			phy->active = true;
		else
			phy->active = false;
	}

	return 0;
}

void scm80211_dump_phy_vector(struct ieee80211_hw *hw, struct phy_vector *pv, unsigned long changed)
{
	int p = 0;

	for_each_set_bit(p, &changed, PHY_PARAM_MAX) {
		switch (p) {
		case PHY_CHANNEL:
			{
				u32 primary;

				primary = (pv->CENTER_FREQUENCY - (1 << pv->BANDWIDTH) * 10 + 10);
				primary += 20 * pv->PRIMARY_20MHZ_CHANNEL_LOC;
				scm80211_trace(hw,
					       "PHY:CHANNEL: primary=%d band=%dGHz bandwidth=%d center=%d\n",
					       primary,
					       pv->BAND ? 5 : 2,
					       20 * (1 << pv->BANDWIDTH), pv->CENTER_FREQUENCY);
				break;
			}
		case PHY_ACTIVE_RXCHAIN_SET:
			{
				scm80211_trace(hw, "PHY:ACTIVE_RXCHAIN_SET=%x\n",
					       pv->ACTIVE_RXCHAIN_SET);
				break;
			}
		case PHY_ANT_CONFIG:
			{
				/* FIXME: when we have PHY/RF matrix */
				break;
			}
		case PHY_RSSI_THRESHOLD:
			{
				scm80211_trace(hw, "PHY: RSSI_THRESHOLD=%d\n",
					       pv->ACTIVE_RXCHAIN_SET);
				break;
			}
		case PHY_LISTEN_TO_GID00:
			{
				scm80211_trace(hw, "PHY:PHY_LISTEN_TO_GID00=%d\n",
					       pv->LISTEN_TO_GID00);
				break;
			}
		case PHY_LISTEN_TO_GID63:
			{
				scm80211_trace(hw, "PHY: PHY_LISTEN_TO_GID63=%d\n",
					       pv->LISTEN_TO_GID63);
				break;
			}
		case PHY_PARTIAL_AID_LIST_GID00:
			{
				scm80211_trace(hw, "PHY:PHY_PARTIAL_AID_LIST_GID00=%d\n",
					       pv->PARTIAL_AID_LIST_GID00);
				break;
			}
		case PHY_PARTIAL_AID_LIST_GID63:
			{
				scm80211_trace(hw, "PHY:PHY_PARTIAL_AID_LIST_GID63=%d\n",
					       pv->PARTIAL_AID_LIST_GID63);
				break;
			}
		case PHY_VHT_SU_RX_FILTER_EN:
			{
				scm80211_trace(hw, "PHY:PHY_VHT_SU_RX_FILTER_EN=%d\n",
					       pv->VHT_SU_RX_FILTER_EN);
				break;
			}
		case PHY_GROUP_ID_MANAGEMENT:
			{
				break;
			}
		case PHY_BSS_COLOR:
			{
				scm80211_trace(hw, "PHY:BSS_COLOR=%d\n", pv->BSS_COLOR);
				break;
			}
		case PHY_STA_ID_INDIVIDUAL:
			{
				scm80211_trace(hw, "PHY:STA_ID_INDIVIDUAL=%08x\n",
					       pv->STA_ID_INDIVIDUAL);
				break;
			}
		case PHY_STA_ID_GROUP:
			{
				scm80211_trace(hw, "PHY:STA_ID_GROUP=%08x\n", pv->STA_ID_GROUP);
				break;
			}
		case PHY_BSS_COLOR_FILTER_EN:
			{
				scm80211_trace(hw, "PHY:BSS_COLOR_FILTER_EN=%d\n",
					       pv->BSS_COLOR_FILTER_EN);
				break;
			}
		case PHY_HE_DIR:
			{
				scm80211_trace(hw, "PHY:HE_DIR_EN=%x\n", pv->HE_DIR_EN);
				break;
			}
		default:
			scm80211_trace(hw, "unsupported parameter %d\n", p);
			break;
		}
	}
}

/**
 * phy_config() - configure phy
 * @hw: hardware
 * @vif: virtual interface
 * @changed: bitmap of changed PHY parameters (not used currently)
 *
 * Returns:
 * - < 0 if configuration fails,
 * - 1 if phy configuration changes immediately,
 * - 0 otherwise (it will be reflected next time @vif goes on channel
 *
 * This function calculates consolidated phy configuration (min of max)
 * that satisfies all virtual interfaces sharing a chanctx and puts
 * the result into chanctx->phyconfig.
 *
 * The callers of this function are:
 * 1. chanctx scheduler (scanning, ROC, TDMA, and etc)
 * 2. bss_conf_change()
 * 3. Other beacon/management/notification reception
 *
 * TODO:
 * 1. serialization (need a mutex)
 * 2. This function is likely to be called a lot. Nicer to have a cache.
 * 3. It would be nicer if we can update only changes, but seems difficult.
 * 4. Can there be recursion? I don't think so.
 * 5. debug log
 */
int phy_config(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx, unsigned long changes)
{
	struct phy_vector *pv = &chanctx->pv;
	struct scm80211_phy *phy;

	trace_phy_config(hw, chanctx, changes);

	if (test_bit(PHY_CHANNEL, &changes))
		phy_vector_set_channel(hw, chanctx);

	if (test_bit(PHY_ACTIVE_RXCHAIN_SET, &changes))
		phy_vector_set_rx_chain(hw, chanctx);

	scm80211_trace(hw, "phy config: chanctx=%d, changes=0x%lx\n", chanctx->id, changes);
	scm80211_dump_phy_vector(hw, pv, changes);

	for_each_active_phy(phy, hw) {
		phy->ops->config(phy, pv, changes);
	}
	return 0;
}

static void scm80211_temp_comp(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv = container_of(work, struct ieee80211_hw_priv,
						      temp_work.work);
	scm80211_info(priv->hw, "do temp comp\n");
	scm2020_phy_temp_comp(priv->hw);
}

/*
 * This function is intended to be called in start() and resume()
 *
 * Since PHY registers may not be retained, we may have to initialize everything.
 */
int scm80211_phy_start(struct ieee80211_hw *hw)
{
	struct scm80211_phy *phy;
	struct ieee80211_hw_priv *priv = hw->priv;
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	struct phy_vector pv;
	int band_sel = 0;
	int freq[] = { 2442, 5530, 5180, 5885 };
#endif

	scm80211_phy_writel(hw, 0, PHY_CONFIG0);

	if (!scm80211_bus_phy_start(hw))
		return 0;

	for (phy = priv->phy; phy < priv->phy + 2; phy++) {
		scm80211_warn(hw, "%s: enable phy#%d", __func__, phy->chain);
		phy->ops->enable(hw, phy, true);
		phy->ops->calibrate(phy);
	}

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	mdelay(500);
	for (band_sel = 0; band_sel < 4; band_sel++) {
		pv.BANDWIDTH = CBW80;
		pv.CENTER_FREQUENCY = freq[band_sel];

		if (band_sel == 0) {
			pv.BANDWIDTH = CBW40;
			pv.BAND = NL80211_BAND_2GHZ;
		} else {
			pv.BAND = NL80211_BAND_5GHZ;
		}

		if (!(priv->band & BIT(pv.BAND))) {
			scm80211_warn(hw, "%s: devid=%d dbdc=%d band=0x%x not support band=%d\n",
				      __func__, priv->devid, priv->dbdc, priv->band, pv.BAND);
			continue;
		}
		for_each_phy(phy, hw) {
			scm80211_set_channel(hw, phy, &pv);
		}
		mdelay(10);
		for_each_phy(phy, hw) {
			phy->radio->ops->power_on_cal(phy->radio, phy, pv.BAND,
						      pv.CENTER_FREQUENCY);
		}
		mdelay(10);
	}
#endif

	INIT_DELAYED_WORK(&priv->temp_work, scm80211_temp_comp);
	priv->dynamic_temp_timer = temp_interval;
	if (priv->dynamic_temp_timer)
		ieee80211_queue_delayed_work(hw, &priv->temp_work, temp_interval * HZ);

	return 0;
}

/*
 * This function is intended to be called in stop() and suspend()
 */
void scm80211_phy_stop(struct ieee80211_hw *hw)
{
	struct scm80211_phy *phy;
	struct ieee80211_hw_priv *priv = hw->priv;

	if (!scm80211_bus_phy_stop(hw))
		return;

	cancel_delayed_work_sync(&priv->temp_work);

	for (phy = priv->phy; phy < priv->phy + 2; phy++)
		phy->ops->enable(hw, phy, false);
}

void scm80211_phy_calibrate(struct ieee80211_hw *hw)
{
	struct scm80211_phy *phy;

	for_each_phy(phy, hw)
		phy->ops->calibrate(phy);
}

/**
 * PHY utilities
 */

#define T(x)	T_##x
#define N(x)	N_##x
#define L(x)	L_##x
#define ceil(x, y) DIV_ROUND_UP(x, y)

/**
 * struct mcs_info
 * @bpsc: bit per subcarrier
 * @r: coding rate multiplied by 12
 * @non_ht_rate: non-ht reference rate
 */
struct mcs_info {
	enum {
		BPSK = 1,
		QPSK = 2,
		QAM16 = 4,
		QAM64 = 6,
		QAM256 = 8,
		QAM1024 = 10,
	} bpsc;
	int r;
	u32 non_ht_rate;
};

static u32 aSignalExtension(enum nl80211_band band)
{
	return (band == NL80211_BAND_2GHZ) ? 6 : 0;
}

/* bitrates in 100kbps */
static u32 non_ht_rate[] = {
	10, 20, 55, 110,
	60, 90, 120, 180, 240, 360, 480, 540,
};

/**
 * DSSS/CCK PHY
 */
#define PreambleLength 144
#define PHYHeaderTime  48

u32 dsss_cck_txtime(struct txvector *tv, enum nl80211_band band)
{
	u32 txtime;
	u32 bitrate;

	txtime = PreambleLength + PHYHeaderTime;
	bitrate = non_ht_rate[tv->MCS];

	if (tv->PREAMBLE_TYPE == SHORT_PREAMBLE)
		txtime >>= 1;

	txtime += ceil(tv->L_LENGTH * 8 * 10, bitrate);
	return txtime;
}

/*
 * OFDM and ERP PHY
 */
#define OFDM_PLCP_BITS  22	/* 16 SERVICE + 6 Tail */
#define T_L_STF	8
#define T_L_LTF	8
#define T_L_SIG	4
#define T_HT_SIG	8
#define T_HT_STF	4
#define T_HT_LTF	4
#define T_SYM		4
#define T_SYSMS		(18 / 5)

#define R(x)  (12 * x)

static struct mcs_info non_ht_ofdm_mcs[] = {
	[4] =  { BPSK,  R(1 / 2), 60 },
	[5] =  { BPSK,  R(3 / 4), 90 },
	[6] =  { QPSK,  R(1 / 2), 120 },
	[7] =  { QPSK,  R(3 / 4), 180 },
	[8] =  { QAM16, R(1 / 2), 240 },
	[9] =  { QAM16, R(2 / 3), 360 },
	[10] = { QAM64, R(2 / 3), 480 },
	[11] = { QAM64, R(3 / 4), 540 },
};

u32 ofdm_txtime(struct txvector *tv, enum nl80211_band band)
{
	u32 txtime = 0;

	int N(DBPS), N(SYM);

	N(DBPS) = (non_ht_ofdm_mcs[tv->MCS].bpsc * 48 * non_ht_ofdm_mcs[tv->MCS].r / 12);
	N(SYM) = ceil(16 + 8 * tv->L_LENGTH + 6, N(DBPS));

	txtime = T(L_STF) + T(L_LTF) + T(L_SIG) + T(SYM) * N(SYM);
	txtime += aSignalExtension(band);	/* T(SE) */
	return txtime;
}

int erp_txtime(struct txvector *tv, enum nl80211_band band)
{
	if (tv->MCS < 4)
		return dsss_cck_txtime(tv, band);
	return ofdm_txtime(tv, band);
}

/**
 * HT PHY
 */
struct mcs_info ht_mcs_info[] = {
	[0] = { BPSK,  R(1/2),  6*2 },
	[1] = { QPSK,  R(1/2),  12*2 },
	[2] = { QPSK,  R(3/4),  18*2 },
	[3] = { QAM16, R(1/2),  24*2 },
	[4] = { QAM16, R(3/4),  36*2 },
	[5] = { QAM64, R(2/3),  48*2 },
	[6] = { QAM64, R(3/4),  54*2 },
	[7] = { QAM64, R(5/6),  54*2 },
};

/* HT N(SD) */
static int ht_n_sd[] = {
	[CBW20] = 52,
	[CBW40] = 108,
};

/**
 * ht_cbps() - HT coded bits per symbol
 */
static int ht_cbps(struct txvector *tv)
{
	int mcs, cbps, streams;

	mcs = tv->MCS % 8;
	streams = tv->MCS / 8 + 1;
	cbps = ht_mcs_info[mcs].bpsc * ht_n_sd[tv->CH_BANDWIDTH];
	cbps *= streams;
	return cbps;
}

/**
 * ht_dbps() - HT data bits per symbol
 */
static int ht_dbps(struct txvector *tv)
{
	int mcs = tv->MCS % 8;

	return ht_cbps(tv) * ht_mcs_info[mcs].r / 12;
}

static int ht_NSYM_ldpc(struct txvector *tv)
{
	int mcs = tv->MCS % 8;
	int mSTBC = 1, R;

	int N(SYM), N(pld), N(avbits), N(CBPS), N(CW) = 1;
	int N(shrt), N(punc);
	int L(LDPC);

	R = ht_mcs_info[mcs].r;
	mSTBC = (tv->STBC) ? 2 : 1;
	N(CBPS) = ht_cbps(tv);
	N(pld) = tv->LENGTH * 8 + 16;
	N(avbits) = N(CBPS) * mSTBC;
	N(avbits) *= ceil(N(pld), N(CBPS) * mSTBC * R / 12);

	if (N(avbits <= 648)) {
		L(LDPC) = (N(avbits) >= N(pld) + 912 - 912 * R / 12) ? 1296 : 648;
	} else if (N(avbits) <= 1296) {
		L(LDPC) = (N(avbits) >= N(pld) + 1464 - 1464 * R / 12) ? 1944 : 1296;
	} else if (N(avbits) <= 1944) {
		L(LDPC) = 1944;
	} else if (N(avbits) <= 2592) {
		N(CW) = 2;
		L(LDPC) = (N(avbits) >= N(pld) + 2916 - 2916 * R / 12) ? 1944 : 1296;
	} else {
		N(CW) = ceil(N(pld), 1994 * R / 12);
		L(LDPC) = 1944;
	}

	N(shrt) = max(0, N(CW) * L(LDPC) * R / 12 - N(pld));
	N(punc) = max(0, N(CW) * L(LDPC) - N(avbits) - N(shrt));

	if ((N(punc) > (N(CW) * (L(LDPC) - L(LDPC) * R / 12) / 10)) &&
	    N(shrt) < N(punc) * R / (12 - R) * 6 / 5) {
		N(avbits) += N(CBPS) * mSTBC;
		N(punc) = max(0, N(CW) * L(LDPC) - N(avbits) - N(shrt));
	}
	N(SYM) = N(avbits) / N(CBPS);
	return N(SYM);
}

int ht_txtime(struct txvector *tv, enum nl80211_band band)
{
	int txtime = 0;

	int N(SYM), N(SS), N(HT_LTF), N(DBPS), N(ES) = 1;
	int mSTBC = 1;

	N(SS) = (tv->MCS >> 3) + 1;
	N(HT_LTF) = N(SS) + tv->STBC;
	N(HT_LTF) += (N(HT_LTF) == 3) ? 1 : 0;
	N(DBPS) = ht_dbps(tv);

	if (tv->STBC > 0)
		mSTBC = 2;

	if (tv->FEC_CODING == BCC_CODING) {
		N(SYM) = mSTBC * ceil(8 * tv->LENGTH + 16 + 6 * N(ES), mSTBC * N(DBPS));
	} else {
		/* LDPC */
		N(SYM) = ht_NSYM_ldpc(tv);
	}

	if (tv->FORMAT == HT_MF) {
		txtime += T(L_STF) + T(L_LTF) + T(L_SIG);
		txtime += T(HT_SIG) + T(HT_STF) + N(HT_LTF) * T(HT_LTF);
		if (tv->GI_TYPE == SHORT_GI)
			txtime += T(SYM) * ceil(N(SYM) * 18, 5 * T(SYM));
		else
			txtime += T(SYM) * N(SYM);
	} else {
		txtime += T(HT_SIG) + T(HT_STF) + (N(HT_LTF) + 1) * T(HT_LTF);
	}

	txtime += aSignalExtension(band);
	return txtime;
}

/**
 * VHT PHY
 *
 * MU is not considered yet
 */

#define T_VHT_SIG_A	8
#define T_VHT_STF	4
#define T_VHT_LTF	4
#define T_VHT_SIG_B	4

struct mcs_info vht_mcs_info[] = {
	[0] = { BPSK,   R(1 / 2),   6*2 },
	[1] = { QPSK,   R(1 / 2),  12*2 },
	[2] = { QPSK,   R(3 / 4),  18*2 },
	[3] = { QAM16,  R(1 / 2),  24*2 },
	[4] = { QAM16,  R(3 / 4),  36*2 },
	[5] = { QAM64,  R(2 / 3),  48*2 },
	[6] = { QAM64,  R(3 / 4),  54*2 },
	[7] = { QAM64,  R(5 / 6),  54*2 },
	[8] = { QAM256, R(3 / 4),  54*2 },
	[9] = { QAM256, R(5 / 6),  54*2 },
};

/* VHT N(SD) */
static int vht_n_sd[] = {
	[CBW20] = 52,
	[CBW40] = 108,
	[CBW80] = 234,
	[CBW160] = 468,
};

/**
 * vht_cbps() - VHT coded bits per symbol
 */
static int vht_cbps(struct txvector *tv)
{
	int mcs = tv->MCS, cbps, streams;

	streams = (tv->NUM_STS + 1) >> ((tv->STBC) ? 1 : 0);
	cbps = vht_mcs_info[mcs].bpsc * vht_n_sd[tv->CH_BANDWIDTH];
	cbps *= streams;
	return cbps;
}

/**
 * vht_dbps() - VHT data bits per symbol
 */
static int vht_dbps(struct txvector *tv)
{
	return vht_cbps(tv) * vht_mcs_info[tv->MCS].r / 12;
}

static int vht_NSYM_ldpc(struct txvector *tv)
{
	int N(SYM), N(SYM0), N(pld), N(avbits), N(CBPS), N(DBPS);
	int N(CW) = 1, N(shrt), N(punc);
	int L(LDPC);
	int mSTBC = 1, R;
	bool cond;

	mSTBC = (tv->STBC) ? 2 : 1;
	N(CBPS) = vht_cbps(tv);
	N(DBPS) = vht_dbps(tv);
	N(SYM0) = mSTBC * roundup(8 * tv->APEP_LENGTH + 16, mSTBC * N(DBPS)) / (mSTBC * N(DBPS));
	N(pld) = N(SYM0) * N(DBPS);
	N(avbits) = N(SYM0) * N(CBPS);
	R = vht_mcs_info[tv->MCS].r;

	if (N(avbits) <= 648) {
		L(LDPC) = (N(avbits) >= N(pld) + 912 - 912 * R / 12) ? 1296 : 648;
	} else if (N(avbits) <= 1296) {
		L(LDPC) = (N(avbits) >= N(pld) + 1464 - 1464 * R / 12) ? 1944 : 1296;
	} else if (N(avbits) <= 1944) {
		L(LDPC) = 1944;
	} else if (N(avbits) <= 2592) {
		N(CW) = 2;
		L(LDPC) = (N(avbits) >= N(pld) + 2916 - 2916 * R / 12) ? 1944 : 1296;
	} else {
		N(CW) = ceil(N(pld), 1994 * R / 12);
		L(LDPC) = 1944;
	}

	N(shrt) = max(0, N(CW) * L(LDPC) * R / 12 - N(pld));
	N(punc) = max(0, N(CW) * L(LDPC) - N(avbits) - N(shrt));

	cond = (N(punc) > (N(CW) * (L(LDPC) - L(LDPC) * R / 12) / 10));
	cond &= N(shrt) < N(punc) * R / (12 - R) * 6 / 5;
	cond |= N(punc) > (N(CW) * L(LDPC) * (12 - R)) / 40;
	if (cond) {
		N(avbits) += N(CBPS) * mSTBC;
		N(punc) = max(0, N(CW) * L(LDPC) - N(avbits) - N(shrt));
	}
	N(SYM) = N(avbits) / N(CBPS);
	return N(SYM);
}

u32 vht_txtime(struct txvector *tv, enum nl80211_band band)
{
	u32 txtime = 0;

	int N(VHT_LTF), N(SYM), N(DBPS), N(ES) = 1;
	int mSTBC = 1;

	/* Table 21-13 */
	N(VHT_LTF) = tv->NUM_STS + 1;
	if (N(VHT_LTF) > 2 && (N(VHT_LTF) & 0x1))
		N(VHT_LTF)++;

	N(DBPS) = vht_dbps(tv);

	if (tv->STBC > 0)
		mSTBC = 2;
	else
		mSTBC = 1;

	if (tv->FEC_CODING == BCC_CODING)
		N(SYM) = mSTBC * ceil(8 * tv->APEP_LENGTH + 16 + 6 * N(ES), mSTBC * N(DBPS));
	else
		N(SYM) = vht_NSYM_ldpc(tv);

	txtime = T(L_STF) + T(L_LTF) + T(L_SIG)
	    + T(VHT_SIG_A) + T(VHT_STF) + N(VHT_LTF) * T(VHT_LTF);

	if (tv->GI_TYPE == SHORT_GI)
		txtime += T(SYM) * ceil(N(SYM) * 18, 5 * T(SYM));
	else
		txtime += T(SYM) * N(SYM);

	return txtime;
}

/**
 * Table 27-12—Timing-related constants (ten times magnification)
 **/

#define T_GI_LTF        16
#define T_GI1           8
#define T_GI2           16
#define T_GI3           32

#define T_RL_SIG        40
#define T_HE_SIG_A      80
#define T_HE_SIG_A_R    160
#define T_HE_SIG_B      40

#define T_HE_STF_T      80
#define T_HE_STF_NT     40

#define T_HE_LTF_1X	32
#define T_HE_LTF_2X     64
#define T_HE_LTF_4X     128

#define T_GI1_DATA	8
#define T_GI2_DATA	16
#define T_GI4_DATA	32

#define T_HE_SYM1	136
#define T_HE_SYM2	144
#define T_HE_SYM4	160

struct mcs_info he_mcs_info[] = {
	[0] =  {BPSK,    R(1 / 2),   6*2 },
	[1] =  {QPSK,    R(1 / 2),  12*2 },
	[2] =  {QPSK,    R(3 / 4),  18*2 },
	[3] =  {QAM16,   R(1 / 2),  24*2 },
	[4] =  {QAM16,   R(3 / 4),  36*2 },
	[5] =  {QAM64,   R(2 / 3),  48*2 },
	[6] =  {QAM64,   R(3 / 4),  54*2 },
	[7] =  {QAM64,   R(5 / 6),  54*2 },
	[8] =  {QAM256,  R(3 / 4),  54*2 },
	[9] =  {QAM256,  R(5 / 6),  54*2 },
	[10] = {QAM1024, R(3 / 4),  54*2 },
	[11] = {QAM1024, R(5 / 6),  54*2 },
};

enum he_ru {
	RU26	= 0,
	RU52	= 1,
	RU106	= 2,
	RU242	= 3,
	RU484	= 4,
	RU996	= 5,
	RU1992	= 6,
	RUMAX	= RU1992,
};

/* Table 27-14 11ax 6.0 */
static int he_n_sd_ru[] = {
	[RU26]	= 24,
	[RU52]	= 48,
	[RU106] = 102,
	[RU242] = 234,	/* full 20MHz */
	[RU484] = 468,	/* full 40MHz */
	[RU996] = 980,	/* full 80MHz */
	[RU1992] = 1960, /* full 160MHz */
};

/* Table 27-33 11ax 6.0 */
static int he_n_sd_short[][2] = {
	[RU26]	= {  6,   2},
	[RU52]	= { 12,   6},
	[RU106] = { 24,  12},
	[RU242] = { 60,  30},	/* full 20MHz */
	[RU484] = {120,  60},	/* full 40MHz */
	[RU996] = {240, 120},	/* full 80MHz */
	[RU1992] = {492, 246},	/* full 160MHz */

};

/* Table 27-46—Nominal TPE values */
static int T_PE_NOMINAL[][3] = {
	[0] = { 0, 0, 4 },
	[1] = { 0, 0, 8 },
	[2] = { 0, 4, 12 },
	[3] = { 0, 8, 16 },
};

int he_cbps(struct txvector *tv)
{
	int cbps, N(SS);
	enum he_ru ru_index;

	N(SS) = (tv->NUM_STS + 1) >> ((tv->STBC) ? 1 : 0);

	ru_index = RU242 + tv->CH_BANDWIDTH;
	if (WARN_ON(ru_index > RU996))
		ru_index = RU996;

	cbps = he_mcs_info[tv->MCS].bpsc * he_n_sd_ru[ru_index];

	return cbps * N(SS);
}

/**
 * he_dbps() - HE data bits per symbol
 */
static int he_dbps(struct txvector *tv)
{
	return he_cbps(tv) * he_mcs_info[tv->MCS].r / 12;
}

static int he_cbps_short(struct txvector *tv)
{
	int v;
	enum he_ru ru_index = RU242 + tv->CH_BANDWIDTH;

	int N(SS) = (tv->NUM_STS + 1) >> ((tv->STBC) ? 1 : 0);

	if (WARN_ON(ru_index > RU996))
		ru_index = RU996;

	v = he_n_sd_short[ru_index][tv->DCM] * N(SS) * he_mcs_info[tv->MCS].bpsc;

	return v;

}

static int he_dbps_short(struct txvector *tv)
{
	return he_cbps_short(tv) * he_mcs_info[tv->MCS].r / 12;
}

static bool is_ldpc_condition_met(struct txvector *tv, int N(SYM0), int mSTBC,
				  int N(DBPS), int N(DBPS_last0), int N(CBPS), int N(CBPS_last0))
{
	int N(pld), N(avbits), N(CW) = 1, L(LDPC);
	int N(shrt), N(punc);
	int R = he_mcs_info[tv->MCS].r;
	bool met = false;

	if (tv->FEC_CODING != LDPC_CODING)
		return met;

	N(pld) = (N(SYM0) - mSTBC) * N(DBPS) + mSTBC * N(DBPS_last0);

	N(avbits) = (N(SYM0) - mSTBC) * N(CBPS) + mSTBC * N(CBPS_last0);

	/**
	 * Table 19-16—PPDU encoding parameters
	 * because R = rate * 12 so made the constant div 12
	 * 912 -> 76,
	 * 1464 -> 122,
	 * 1944 -> 162,
	 * 2916 -> 243,
	 * ...
	 **/
	if (N(avbits) <= 648) {
		L(LDPC) = (N(avbits) >= N(pld) + 912 - 76 * R) ? 1296 : 648;
	} else if (N(avbits) <= 1296) {
		L(LDPC) = (N(avbits) >= N(pld) + 1464 - 122 * R) ? 1944 : 1296;
	} else if (N(avbits) <= 1944) {
		L(LDPC) = 1944;
	} else if (N(avbits) <= 2592) {
		N(CW) = 2;
		L(LDPC) = (N(avbits) >= N(pld) + 2916 - 243 * R) ? 1944 : 1296;
	} else {
		N(CW) = ceil(N(pld), 162 * R);
		L(LDPC) = 1944;
	}

	N(shrt) = max(0, N(CW) * L(LDPC) * R / 12 - N(pld));
	N(punc) = max(0, N(CW) * L(LDPC) - N(avbits) - N(shrt));

	met = (N(punc) > (N(CW) * (L(LDPC) - L(LDPC) * R / 12) / 10));
	met = met && (N(shrt) < N(punc) * R / (12 - R) * 6 / 5);
	met = met || (N(punc) > (N(CW) * L(LDPC) * (12 - R)) / 40);

	return met;
}

int he_n_sym(struct txvector *tv, int *a)
{
	int N(SYM), N(SYM0);
	int N(DBPS) = he_dbps(tv);
	int N(CBPS) = he_cbps(tv);
	int N(CBPS_last0), N(DBPS_last0);
	int N(DBPS_SHORT), N(CBPS_SHORT);
	int N(SERVICE) = 16;
	int N(TAIL) = tv->FEC_CODING == BCC_CODING ? 6 : 0;
	int mSTBC = tv->STBC ? 2 : 1;

	int N(EXCESS) = (8 * tv->APEP_LENGTH + N(TAIL) + N(SERVICE)) % (mSTBC * N(DBPS));
	int a0;

	N(CBPS_SHORT) = he_cbps_short(tv);
	N(DBPS_SHORT) = he_dbps_short(tv);

	a0 = N(EXCESS) == 0 ? 4 : DIV_ROUND_UP(N(EXCESS), mSTBC * N(DBPS_SHORT));

	a0 = min_t(int, a0, 4);

	/* Equation (27-64) */
	N(SYM0) = 8 * tv->APEP_LENGTH + N(TAIL) + N(SERVICE);
	N(SYM0) = DIV_ROUND_UP(N(SYM0), mSTBC * N(DBPS));
	N(SYM0) *= mSTBC;

	N(SYM) = N(SYM0);
	*a = a0;

	if (a0 == 4) {
		N(DBPS_last0) = N(DBPS);
		N(CBPS_last0) = N(CBPS);
	} else {
		N(DBPS_last0) = a0 * N(DBPS_SHORT);
		N(CBPS_last0) = a0 * N(CBPS_SHORT);
	}

	if (is_ldpc_condition_met(tv, N(SYM0), mSTBC,
				  N(DBPS), N(DBPS_last0), N(CBPS), N(CBPS_last0))) {
		if (a0 == 4) {
			*a = 1;
			N(SYM) += mSTBC;
		} else {
			*a += 1;
		}
	}

	return N(SYM);
}

/**
 * Reference: P802.11ax_D6.0 27.4.3
 * Notice: T(x) are ten times magnification
 */
u32 he_txtime(struct txvector *tv, enum nl80211_band band)
{
	u32 txtime = 200;

	int T(HE_PREAMBLE), T(HE_LTF_SYM), T(PE), T(HE_SYM);
	int N(SYM), N(HE_LTF), N(MA) = 0;
	int a;

	if (WARN_ON(tv->HE_LTF_TYPE > 2 || tv->GI_TYPE < 1))
		return 0;

	T(HE_LTF_SYM) = T(HE_LTF_1X) << tv->HE_LTF_TYPE;
	T(HE_LTF_SYM) += T(GI1_DATA) << (tv->GI_TYPE - 1);

	T(HE_PREAMBLE) = T(RL_SIG) + T(HE_STF_NT);

	if (tv->NUM_STS > 1)
		N(HE_LTF) = roundup(tv->NUM_STS + 1, 2);
	else
		N(HE_LTF) = tv->NUM_STS + 1;

	T(HE_PREAMBLE) += N(HE_LTF) * T(HE_LTF_SYM);

	if (tv->FORMAT == HE_SU)
		T(HE_PREAMBLE) += T(HE_SIG_A);
	else
		T(HE_PREAMBLE) += T(HE_SIG_A_R);

	N(SYM) = he_n_sym(tv, &a);

	T(HE_SYM) = tv->GI_TYPE == 1 ? T(HE_SYM1) : tv->GI_TYPE == 2 ? T(HE_SYM2) : T(HE_SYM4);

	if (tv->DOPPLER) {
		int MMA = tv->MIDAMBLE_PERIODICITY ? 20 : 10;

		N(MA) = DIV_ROUND_UP(N(SYM) - 1, MMA);
		N(MA) = N(MA) > 1 ? N(MA) - 1 : 0;
	}

	T(PE) = T_PE_NOMINAL[a - 1][tv->NOMINAL_PACKET_PADDING] * 10;

	txtime += T(HE_PREAMBLE) + N(SYM) * T(HE_SYM);
	txtime += N(MA) * N(HE_LTF) * T(HE_LTF_SYM) + T(PE);
	txtime = DIV_ROUND_UP(txtime, 10);
	txtime += aSignalExtension(band);
	return txtime;
}

int he_constellation_index(int mcs)
{
	if (mcs > ARRAY_SIZE(he_mcs_info))
		return 7;
	return he_mcs_info[mcs].bpsc / 2;
}

u32 scm80211_ppdu_txtime(struct txvector *tv, enum nl80211_band band)
{
	switch (tv->FORMAT) {
	case NON_HT:
		if (band == NL80211_BAND_2GHZ)
			return erp_txtime(tv, band);
		else
			return ofdm_txtime(tv, band);
	case HT_MF:
	case HT_GF:
		return ht_txtime(tv, band);
	case VHT:
		return vht_txtime(tv, band);
	case HE_SU:
	case HE_ER_SU:
		/* driver can only send HE_SU or HE_ER_SU */
		return he_txtime(tv, band);
	case HE_MU:
	case HE_TB:
		BUG_ON(1);
	}
	return 0;
}

/**
 * TXVECTOR to struct rate_info translation for bitrate computation
 */
static int legacy_rate_table[] = {
	[0] = 10,
	[1] = 20,
	[2] = 55,
	[3] = 110,
	[4] = 60,
	[5] = 90,
	[6] = 120,
	[7] = 180,
	[8] = 240,
	[9] = 360,
	[10] = 480,
	[11] = 540,
};

static enum rate_info_bw bw_tbl[] = {
	[CBW20] = RATE_INFO_BW_20,
	[CBW40] = RATE_INFO_BW_40,
	[CBW80] = RATE_INFO_BW_80,
	[CBW160] = RATE_INFO_BW_160,
};

static enum nl80211_he_gi gi_tbl[] = {
	[GI_0u8s] = NL80211_RATE_INFO_HE_GI_0_8,
	[GI_1u6s] = NL80211_RATE_INFO_HE_GI_1_6,
	[GI_3u2s] = NL80211_RATE_INFO_HE_GI_3_2,
};

/**
 * ieee80211_bitrate() - data bitrate in 100kbits/s
 * @tv: txvector
 *
 * Returns: bitrate in 100kbps
 *
 * This function calculates the transmission bitrate of the PPDU.
 * Internally, it populate a struct rate_info structure and calls
 * to cfg80211_calculate_bitrate().
 */
int scm80211_bitrate(struct txvector *tv)
{
	struct rate_info ri;

	memset(&ri, 0, sizeof(ri));

	if (ppdu_is_non_ht(tv))
		return legacy_rate_table[tv->MCS];

	ri.mcs = tv->MCS;
	ri.bw = bw_tbl[tv->CH_BANDWIDTH];
	ri.flags |= (tv->GI_TYPE == SHORT_GI) ? RATE_INFO_FLAGS_SHORT_GI : 0;

	if (ppdu_is_he(tv)) {
		ri.flags |= RATE_INFO_FLAGS_HE_MCS;
		ri.he_gi = gi_tbl[tv->GI_TYPE];
		ri.he_dcm = tv->DCM;
		ri.he_ru_alloc = 0;
		ri.nss = (tv->NUM_STS + 1) >> tv->STBC;
	} else if (ppdu_is_vht(tv)) {
		ri.flags |= RATE_INFO_FLAGS_VHT_MCS;
		ri.nss = (tv->NUM_STS + 1) >> tv->STBC;
	} else if (ppdu_is_ht(tv)) {
		ri.flags |= RATE_INFO_FLAGS_MCS;
	}

	return cfg80211_calculate_bitrate(&ri);
}

#ifdef __staging__
static u32 ieee80211_txtime_dsss_cck(struct ieee80211_hw *hw,
				     struct ieee80211_tx_info *ti, size_t len)
{
	struct ieee80211_tx_rate *txrate = &ti->control.rates[0];
	struct ieee80211_rate *rate;
	u32 txtime = PreambleLength + PHYHeaderTime;

	if (rate->flags & IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
		txtime >= 1;

	rate = hw->wiphy->bands[ti->band]->bitrates[txrate->idx];
	txtime += DIV_ROUND_UP(8 * len * 10, rate->bitrate);
	return txtime;
}

static u32 ieee80211_txtime_ofdm(struct ieee80211_hw *hw, struct ieee80211_tx_info *ti, size_t len)
{
	struct ieee80211_tx_rate *txrate = &ti->control.rates[0];
	struct ieee80211_rate *rate;
	u32 txtime = 0;

	rate = hw->wiphy->bands[ti->band]->bitrates[txrate->idx];

	/*
	 * N(DBPS) = rate->bitrate * 4 us (bits)
	 * N(SYM) = ceil((16 + 8*L + 6) / N(DBPS)
	 * TXTIME = T(L_STF) = T(L_LTF) + T(L_SIG) + N(SYM) * T(SYM) + T(SE)
	 */
	txtime = T(L_STF) + T(L_LTF) + T(L_SIG);
	txtime += 4 * DIV_ROUND_UP((16 + 8 * len + 6) * 10, rate->bitrate * 4);
	txtime += (rate->flags & IEEE80211_RATE_ERP_G) ? 6 : 0;	/* T(SE) */
	// txtime += (ti->band == NL80211_BAND_2GHZ) ? 6 : 0;	/* T(SE) */

	return txtime;
}

static u32 ieee80211_txtime_ht(struct ieee80211_hw *hw, struct ieee80211_tx_info *ti, size_t len)
{
	struct ieee80211_tx_rate *r = &ti->control.rates[0];
	u32 txtime = 0;

	u32 N(SYM), N(SS), N(HT_LTF), N(CBPS), N(DBPS), N(ES) = 1;
	u32 mSTBC = 1;

	N(SS) = txrate->idx % 7;
	N(HT_LTF) = N(SS);

	/*
	 * NB: we only support the case where N(SS) = 1 and N(STS) = 2
	 */
	if (r->flags & IEEE80211_TX_RC_USE_STBC) {
		N(HT_LTF) += 1;
		mSTBC = 2;
	}
}

#endif
