/* SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause */
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#ifndef __SCM80211_PHY_H__
#define __SCM80211_PHY_H__

#include <linux/spi/spi.h>

/**
 * enum scm80211_phy_param - PHYCONFIG VECTOR element
 *
 * @PHY_CHANNEL: This includes channel center frequency, channel bandwidth, band, and
 *	the location of the primary 20MHz channel.
 *
 * @PHY_ACTIVE_RXCHAIN_SET: always present
 *
 * @PHY_LISTEN_TO_GID00: AP and Mesh STA has to set this to be able to receive VHT UL SU PPDU
 * @PHY_PARTIAL_AID_LIST_GID00: AID to match for GID 0 frame
 * @PHY_LISTEN_TO_GID63: STA, TDLS STA has to set this to be able to receive VHT DL SU PPDU
 * @PHY_PARTIAL_AID_LIST_GID63: AID to match for GID 63 frame
 * @PHY_VHT_SU_RX_FILTER_EN: If set, GROUP_ID/PARTIAL_AID RX filtering is enabled. If clear,
 *	all VHT SU frames are not filtered out regardless of the above setting. Even if clear,
 *	VHT MU frames (GROUP_ID=1..62) will not be properly received unless
 *	@PHY_GROUP_ID_MANAGEMENT information is valid. This parameter is always present and
 *	should be valid.
 * @PHY_GROUP_ID_MANAGEMENT: should be present to be able to receive VHT MU-MIMO.
 *
 * @PHY_BSS_COLOR: BSS color filter value match with BSS_COLOR filed in HE-SIG-A
 *	This is always presented and always updated (even if it is not changed).
 * @PHY_STA_ID: STA_ID list to match with STA_ID in HE-SIG-B for HE DL MU PPDU.
 *      This includes both individually addressed frame, i.e., LSB11(AID), and
 *	group addressed frame, i.e., BSSID_INDEX or 0. Broadcast frames (2047) are always
 *	received. This should be always present and valid.
 * @PHY_HE_BSS_COLOR_FILTER: if set, BSS_COLOR filtering is enabled and HE PPDUs
 *	whose BSS_COLOR (in HE-SIG-A) does match @PHY_BSS_COLOR are discarded. If clear,
 *	all HE PPDUs are received. Even if a VHT MU PPDU passes BSS_COLOR filter, it will
 *	not be correctly decoded if @PHY_HE_STA_ID is not valid. This is always present
 *	and kept valid. Clear if received BSS Color Disabled or operating in monitor mode.
 *	This is always present and updated.
 * @PHY_HE_DIR_EN: always present (always updated)
 */
enum scm80211_phy_param {
	PHY_CHANNEL = 0,
	PHY_ACTIVE_RXCHAIN_SET = 1,
	PHY_ANT_CONFIG = 2,
	PHY_RSSI_THRESHOLD = 3,
	PHY_LISTEN_TO_GID00 = 4,
	PHY_LISTEN_TO_GID63 = 5,
	PHY_PARTIAL_AID_LIST_GID00 = 6,
	PHY_PARTIAL_AID_LIST_GID63 = 7,
	PHY_VHT_SU_RX_FILTER_EN = 8,
	PHY_GROUP_ID_MANAGEMENT = 9,
	PHY_BSS_COLOR = 10,
	PHY_STA_ID_INDIVIDUAL = 11,
	PHY_STA_ID_GROUP = 12,
	PHY_BSS_COLOR_FILTER_EN = 13,
	PHY_HE_DIR = 14,
	PHY_PARAM_MAX,
};

#define PHY_HE_PARAMS			  \
	(BIT(PHY_BSS_COLOR)		| \
	 BIT(PHY_STA_ID_INDIVIDUAL)	| \
	 BIT(PHY_STA_ID_GROUP)		| \
	 BIT(PHY_BSS_COLOR_FILTER_EN)	| \
	 BIT(PHY_HE_DIR))

#define PHY_HE_DIR_DL		0x1
#define PHY_HE_DIR_UL		0x2
#define PHY_HE_DIR_DL_UL	(PHY_HE_DIR_DL | PHY_HE_DIR_UL)

struct phy_vector {
	unsigned long			present;
	u32				CENTER_FREQUENCY; /* MHz */
	u32				BANDWIDTH; /* 0: 20MHz, 1: 40MHz, 2: 80MHz */
	u32				BAND; /* 0: 2.4GHZ, 1: 5GHZ */
	u32				PRIMARY_20MHZ_CHANNEL_LOC; /* 0,..3, */
	u32				ACTIVE_RXCHAIN_SET; /* Bitmap */
	int				MIMO_EN;
	bool				LISTEN_TO_GID00;
	bool				LISTEN_TO_GID63;
	int				PARTIAL_AID_LIST_GID00;
	int				PARTIAL_AID_LIST_GID63;
	bool				VHT_SU_RX_FILTER_EN;
	struct ieee80211_mu_group_data  GROUP_ID_MANAGEMENT;
	u32				BSS_COLOR;
	u32				STA_ID_INDIVIDUAL;
	u32				STA_ID_GROUP;
	bool				BSS_COLOR_FILTER_EN;
	u8				HE_DIR_EN;
	int				RSSI_THRESHOLD;
};

struct scm80211_phy;
struct scm80211_radio;
struct rf_cal_data;
struct rf_gspi_cfg;

/**
 * struct scm80211_phy_ops - PHY methods
 *
 * @enable: start/stop phy
 * @config: configure phy
 * @connect_spi: called by rf system to connect an spi interface
 * @attach: ?
 * @detach: ?
 * @calibrate: ?
 * @control_gpio: called by rf system to control GPIO related to AFE/RFIC
 */
struct scm80211_phy_ops {
	int (*init)(struct ieee80211_hw *hw, struct scm80211_phy *phy);
	void (*deinit)(struct ieee80211_hw *hw, struct scm80211_phy *phy);

	void (*enable)(struct ieee80211_hw *hw, struct scm80211_phy *phy, bool enable);
	void (*config)(struct scm80211_phy *phy, struct phy_vector *pv, unsigned long changes);

	struct spi_device *(*connect_spi)(struct scm80211_phy *phy,
					  struct spi_board_info *info, struct rf_gspi_cfg *gspi);
	int (*attach)(struct scm80211_phy *phy, struct scm80211_radio *radio);
	void (*detach)(struct scm80211_phy *phy, struct scm80211_radio *radio);
	void (*calibrate)(struct scm80211_phy *phy);
	void (*control_gpio)(struct scm80211_phy *phy, int id, u32 lmask, u32 hmask);
};

/**
 * struct scm80211_rf_iface - PHY/RF interface
 *
 * @base: ioremapped address of the RFI region (not used)
 * @start: RFI register offset from the PHY(x) base address
 * @spi: SPI controllers
 * @i2c: I2C controllers (not used)
 */
struct scm80211_rf_iface {
	void __iomem *base;
	u32 start;

	struct {
		struct spi_master *master[4];
		int n;
	} spi;

	struct {
		struct i2c_master *master[4];
		int n;
	} i2c;
};

/**
 * struct rf_board_info - RF board information
 */
struct rf_board_info {
	struct spi_board_info *spi_info;
	int n_spi;
	/* more to come */
};

/*
 *  struct scm80211_gspi - GSPI information
 *  @wr_nbits: write bit length ( <= 32)
 *  @rd_nbits: read bit length ( <= 32)
 *  @lf_nbits: LSB field bit length ( <= 16)
 *  @flags: bit#0: R/W polarity
 *          bit#1: LSB field is address
 *  @reg_cfg: control value for GSP#_CONFIG register
 */
struct rf_gspi_cfg {
	u8 wr_nbits;
	u8 rd_nbits;
	u8 lf_nbits;
	u8 flags;
	u32 reg_val;
};

/**
 * struct scm80211_phy
 *
 * @version: phy version
 * @mask: mask used in version matching
 * @chain: phy chain (0-2)
 *	If @chain > 1 can be only bound to mac instance 0 or @chain.
 *	This phy is connected to @chain-th radio chain in RF
 * @supported_bands: BIT(NL80211_BAND_2GHZ) | BIT(NL80211_BAND_5GHZ)
 * @hw: wiphy to which this phy is bound (NULL if not bound)
 * @active: indicates if phy is active (even mimo is enabed)
 * @rfi: RF interface
 * @ops: PHY driver methods
 * @radio: attached RF
 */
struct scm80211_phy {
	const unsigned int version;
	const unsigned int mask;
	int chain;
	int rf_ant_sel;
	const unsigned int supported_bands;

	struct ieee80211_hw *hw;
	bool active;

	const struct scm80211_phy_ops *ops;
	struct scm80211_rf_iface rfi;
	struct scm80211_radio *radio;
#ifdef SCM2020_A0
	u32 rfi_hpf[3];
#endif

	unsigned int start;
};

/**
 * struct scm80211_radio_ops - radio methods
 *
 * @init: initialize radio
 * @deinit: deinitialize radio
 * @attach: attach @radio to @phy
 * @detach: detach @radio from @phy
 * @set_channel: set RF frequency and bandwidth
 */
struct scm80211_radio_ops {
	int (*init)(struct scm80211_radio *radio);
	int (*deinit)(struct scm80211_radio *radio);

	int (*attach)(struct scm80211_radio *radio, struct scm80211_phy *phy);
	int (*detach)(struct scm80211_radio *radio);

	int (*power_up)(struct scm80211_radio *radio);
	int (*power_down)(struct scm80211_radio *radio);

	int (*set_channel)(struct scm80211_radio *radio, unsigned int freq, unsigned int bandwdith);
	int (*get_cal_data)(struct scm80211_radio *radio, struct rf_cal_data *data, int freq);

	int (*set_rtx_lut)(struct scm80211_radio *radio, int freq);
	int (*do_temp_comp)(struct scm80211_radio *radio);
	void (*set_dig_gain)(struct scm80211_radio *radio, int band, bool dig);
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	void (*power_on_cal)(struct scm80211_radio *radio, struct scm80211_phy *phy, int band,
			     int freq);
	int (*set_cal_path)(struct scm80211_radio *radio, u8 band, bool en, u8 cal_mode,
			    u8 sub_mode);
	int (*power_save)(struct scm80211_radio *radio, int freq, bool ps);

#ifdef SCM2020_CFO
	int	(*set_cfo_adj)(struct scm80211_radio *radio);
#endif
#ifdef SCM2020_A0
	int (*pdet_read)(struct scm80211_radio *radio, int power_set);
	int (*get_cur_pwr)(struct scm80211_radio *radio);
#endif
#endif
};

/**
 * struct scm80211_radio - one radio chain
 *
 */
struct scm80211_radio {
	const struct scm80211_radio_ops *ops;
	struct scm80211_phy		*phy;
	void				*priv;
	const char			*devname;
	struct rf_lut			*rf_lut;
};

/**
 * struct scm80211_rf_board - collection of radio chains
 *
 * @id: board number
 */

struct scm80211_rf_system {
	char *name;
	int id;
	struct scm80211_radio *radios;
	int n_chain;
	int (*probe)(struct scm80211_rf_system *rfs);
	int (*remove)(struct scm80211_rf_system *rfs);
};

enum rf_cal_flags {
	/* Skip */
	RFCAL_SKIP_TX_LO_COMP = (1 << 0),
	/* Add */
	RFCAL_RX_GDELAY = (1 << 16)
};

struct rf_lut {
	u32 tx_lut[2][64];
	int tx_lut_cnt;
	u32 rx_lut[2][128];
	int rx_lut_cnt;
};

struct rf_cal_data {
	int tx_i_offset[2];
	int tx_q_offset[2];
	int tx_i_dc_offset[2];
	int tx_q_dc_offset[2];
	int rx_i_offset[2];
	int rx_q_offset[2];
	int rx_i_dc_offset[2];
	int rx_q_dc_offset[2];
	u16 rx_i_gd_offset[2];
	u16 rx_q_gd_offset[2];
	u32 flags;
};

struct scm80211_chanctx;

unsigned long combine_phy_vector(struct phy_vector *v1, struct phy_vector *v2, bool dryrun);
void scm80211_dump_phy_vector(struct ieee80211_hw *hw,
			      struct phy_vector *pv, unsigned long changed);

int scm80211_phy_start(struct ieee80211_hw *hw);
void scm80211_phy_stop(struct ieee80211_hw *hw);
/*int  phy_config(struct ieee80211_hw *hw, struct phy_vector *pv, unsigned long changes);*/
int phy_config(struct ieee80211_hw *hw, struct scm80211_chanctx *chanctx, unsigned long changes);

void phy_writel(struct scm80211_phy *phy, u32 val, u32 offset);
u32 phy_readl(struct scm80211_phy *phy, u32 offset);

struct scm80211_phy *phy_next(struct ieee80211_hw *hw, struct scm80211_phy *phy, bool mimo);
void scm2020_phy_tx_mode(struct scm80211_phy *phy, int band, u8 mode, u8 lut_idx);
void scm2020_phy_rx_mode(struct scm80211_phy *phy, int band, u8 mode, u8 lut_idx);

void scm2020_phy_set_channel(struct scm80211_phy *phy, int vidx, struct phy_vector *pv);

#define for_each_phy(phy, hw) \
	for (phy = phy_next(hw, NULL, false); phy; phy = phy_next(hw, phy, false))

#define for_each_active_phy(phy, hw) \
	for (phy = phy_next(hw, NULL, true); phy; phy = phy_next(hw, phy, true))

static inline void scm80211_radio_power_up(struct scm80211_radio *radio)
{
	if (radio && radio->ops && radio->ops->power_up)
		radio->ops->power_up(radio);
}

static inline void scm80211_radio_power_down(struct scm80211_radio *radio)
{
	if (radio && radio->ops && radio->ops->power_down)
		radio->ops->power_down(radio);
}

static inline void scm80211_radio_set_chanel(struct scm80211_radio *radio,
					     unsigned int freq, unsigned int bandwidth)
{
	if (radio && radio->ops && radio->ops->set_channel)
		radio->ops->set_channel(radio, freq, bandwidth);
}

#ifdef SCM2020_CFO
extern int _scfo_freq_hz;
static inline int scm80211_phy_set_cfo_freq(struct scm80211_phy *phy, int cfo_freq_hz)
{
	_scfo_freq_hz = cfo_freq_hz;
	if (phy->radio->ops && phy->radio->ops->set_cfo_adj)
		return phy->radio->ops->set_cfo_adj(phy->radio);
	return 0;
}
#endif

#ifdef SCM2020_A0
u32 scm2020_iq_get(struct scm80211_phy *phy);
void scm2020_power_detect_work(struct work_struct *work);
#endif

#endif /* __SCM80211_PHY_H__ */
