// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#include <linux/bitfield.h>
#include <linux/kdebug.h>

#define func_fmt(str)           "[PS] "str
#define func_trace		PS

#include "core.h"
#include "util.h"
#include "phy.h"
#include "trace.h"
#include "cfg80211.h"
#include "debug.h"

static int dynamic_ps_timeout = 100;
module_param(dynamic_ps_timeout, int, 0644);
MODULE_PARM_DESC(dynamic_ps_timeout, "dynamic powersave timeout (in ms).");

#ifdef SCM80211_TWT
const char *twt_state[] = {
	[TWT_DISABLED]	= "TWT_DIS",
	[TWT_REQ]	= "TWT_REQ",
	[TWT_NEGO]	= "TWT_NEGO",
	[TWT_DOZE]	= "TWT_DOZE",
	[TWT_SP]	= "TWT_SP",
};


/**
 * scm80211_twt_doze() - info fw goes to doze mode as well as doze duration
 * TODO
 */
static void scm80211_twt_doze(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	if (is_pcie_device(hw)) {
		mod_timer(&vifp->pm.doze_timer, jiffies + usecs_to_jiffies(1));
	} else {
		s32 dur;

		vifp->sw_tsf += ktime_us_delta(ktime_get(), vifp->twt_ktime);
		vifp->twt_ktime = ktime_get();
		vifp->twt_next_irq = vifp->twt - SCM80211_TWT_AHEAD_SP;
		scm80211_trace(hw, "twt %llu (wakeup at %llu) current tsf %llu\n",
			       vifp->twt, vifp->twt_next_irq, vifp->sw_tsf);
		if (vifp->twt_next_irq <= vifp->sw_tsf + ieee80211_tu_to_usec(30)) {
			scm80211_trace(hw, "too short doze duration");
			return;
		}
		dur = (vifp->twt_next_irq - vifp->sw_tsf) & S32_MAX;
		scm80211_param_write(hw, vif, TWT_DOZE_ENABLE, &dur, sizeof(dur));
		hrtimer_start(&vifp->twt_tsf_timer, ns_to_ktime(dur),
			      HRTIMER_MODE_REL_SOFT);
	}

}

void scm80211_twt_state_work(struct work_struct *work)
{
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_vif *vif;
	struct ieee80211_hw *hw;
	enum scm80211_twt_state state;

	vifp = container_of(work, struct ieee80211_vif_priv, twt_state_work);
	vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	hw = ieee80211_vif_to_hw(vif);
	state = vifp->twt_state;

	if (state == TWT_DOZE || state == TWT_SP) {
		scm80211_mac_writel(hw, vifp->reg_mmss,
				    MAC_REG_VIF_MIN_MPDU_SPACING(scm80211_vdev_id(vif)));
	}

	if (state == TWT_DOZE) {
		if (vifp->old_state == TWT_NEGO) {
			/* Enable PM bit in Qos null */
			scm80211_mac_writel(hw, 0x91c8, MAC_MISC_NULL_HEADER);
		}
		scm80211_vif_queues_flush(hw, vif);
		scm80211_stop_queues(hw, scm80211_vif_queues(vif));
		scm80211_enable_ps(hw, true);
		scm80211_twt_doze(hw, vif);
	} else if (vifp->old_state == TWT_DISABLED) {
		/* Clear PM bit in Qos null */
		scm80211_mac_writel(hw, 0x81c8, MAC_MISC_NULL_HEADER);
	} else if (state == TWT_SP) {
		scm80211_disable_ps(hw);
	}

	if (FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_EN, vifp->reg_mmss))
		scm80211_wake_queues(hw, scm80211_vif_queues(vif));
	else
		scm80211_stop_queues(hw, scm80211_vif_queues(vif));
}

int ieee80211_parse_twt(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			 const u8 *twt_ie, u8 ielen, struct ieee80211_sta *sta)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u8 negotype;
	u16 reqtype;
	u32 du;
	int cmd;
	int ret = -1;
	unsigned long flag;

	if (!twt_ie)
		return ret;

	print_hex_dump_bytes("TWT IE", DUMP_PREFIX_OFFSET, twt_ie, ielen);

	negotype = FIELD_GET(TWT_NEGOTIATION_TYPE, twt_ie[0]);

	switch (negotype) {
	case INDIVIDUAL_TWT:
		if (FIELD_GET(TWT_NDP_PAGING_INDICATOR, twt_ie[0]) ||
		    FIELD_GET(TWT_RESPONDER_PM_MODE, twt_ie[0])) {
		/* FIXME Not clear about NDP paging and PM mode */
			break;
		}
		if (ielen < 15) {
			scm80211_err(hw, "len of twt ie is less than expect\n");
			break;
		}

		du = FIELD_GET(TWT_WAKE_DURATION_UNIT, twt_ie[0]) ?
		     ieee80211_tu_to_usec(1) : 256;

		reqtype = get_unaligned_le16(twt_ie + 1);
		cmd = FIELD_GET(TWT_SETUP_CMD, reqtype);
		/* Only Parse TWT response */
		if (FIELD_GET(TWT_REQUEST, reqtype))
			break;
		/* FIXME If twt channel is non-zero */
		if (twt_ie[14])
			break;
		/**
		 * Only Parse ACCECP_TWT
		 * TODO for others
		 * DICTATE_TWT
		 * REJECT_TWT
		 * ALTERNATE_TWT
		 **/
		switch (cmd) {
		case ACCEPT_TWT:

			spin_lock_irqsave(&vifp->lock, flag);
			vifp->twt_trigger  = FIELD_GET(TWT_TRIGGER, reqtype);
			vifp->twt_implicit = FIELD_GET(TWT_IMPLICIT, reqtype);
			vifp->twt_flowtype = FIELD_GET(TWT_FLOW_TYPE, reqtype);
			vifp->twt_flow_id  = FIELD_GET(TWT_FLOW_ID, reqtype);
			vifp->twt_prot     = FIELD_GET(TWT_PROTECTION, reqtype);
			vifp->twt	   = get_unaligned_le64(twt_ie + 3);
			vifp->twt_wake_int = 1 << FIELD_GET(TWT_WAKE_INT_EXP, reqtype);
			vifp->twt_wake_int *= get_unaligned_le16(twt_ie + 12);
			vifp->twt_min_sp_dur = du * twt_ie[11];

			scm80211_trace(hw, "twt negotiation success info:\n");
			scm80211_trace(hw, "twt wake interval %d minmal sp duration %d trigger %s implicit %s",
				       vifp->twt_wake_int,
				       vifp->twt_min_sp_dur,
				       vifp->twt_trigger ? "enabled" : "disabled",
				       vifp->twt_implicit ? "enabled" : "disabled");

#if PM_LOCAL_TEST
			vifp->sw_tsf += ktime_us_delta(ktime_get(), vifp->twt_ktime);
			vifp->twt_ktime = ktime_get();
			scm80211_dbg(hw, "twt SP %llu current tsf %llu\n",
				     vifp->twt, vifp->sw_tsf);
			vifp->twt_next_irq = vifp->twt - SCM80211_TWT_AHEAD_SP;

			if (WARN_ON(vifp->twt_next_irq < vifp->sw_tsf)) {
				hrtimer_start(&vifp->twt_tsf_timer, ns_to_ktime(5),
					      HRTIMER_MODE_REL_SOFT);
			} else {
				u64 ns = (vifp->twt_next_irq - vifp->sw_tsf) *
					  NSEC_PER_USEC;

				hrtimer_start(&vifp->twt_tsf_timer, ns_to_ktime(ns),
					      HRTIMER_MODE_REL_SOFT);
			}
#endif
			spin_unlock_irqrestore(&vifp->lock, flag);

			if (!scm80211_set_twt_state(vif, TWT_DOZE))
				break;

			ret = 0;
			break;
		case DICTATE_TWT:
		case ALTERNATE_TWT:
		case REJECT_TWT:
		default:
			break;
		}
		break;
	case INDIVIDUAL_TBTT:
		/* TODO */
	case BROADCAST_TWT_GROUP:
		/* TODO */
	case BROADCAST_TWT_SINGLE:
		/* TODO */
	default:
		break;
	}

	return ret;
}

bool scm80211_set_twt_state(struct ieee80211_vif *vif, enum scm80211_twt_state state)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(vif);
	struct cfg80211_chan_def def;
	enum scm80211_twt_state oldstate = vifp->twt_state;
	int changed = false;
	bool schedule_work = false, locked;
	unsigned long flags;

	if (ieee80211_iface_is_ap(vif->type) || scm80211_vif_chan(vif, &def))
		goto exit;

	if (vifp->twt_state == state) {
		scm80211_trace(hw, "Already in TWT STATE %s\n", twt_state[vifp->twt_state]);
		return true;
	}

	switch (state) {
	case TWT_DISABLED:
		if (oldstate == TWT_DOZE) {
			scm80211_disable_ps(hw);
			scm80211_send_nullfunc(hw, vif, &def, 0);
		}
		scm80211_wake_queues(hw, scm80211_vif_queues(vif));
		schedule_work = true;
		changed = true;
		break;
	case TWT_REQ:
		scm80211_disable_tsf_irq(hw, vif);
		changed = true;
		break;
	case TWT_NEGO:
		if (oldstate == TWT_REQ) {
			IEEE80211_SKB_CB(vifp->twt_request)->band = def.chan->band;
			IEEE80211_SKB_CB(vifp->twt_request)->control.vif = vif;
			scm80211_tx_channel(hw, vifp->twt_request, def.chan);
			vifp->twt_request = NULL;
			changed = true;
		}
		break;
	case TWT_DOZE:
		vifp->twt_state = state;
		locked = spin_trylock_irqsave(&vifp->lock, flags);

		if (vifp->twt_trigger)
			vifp->reg_mmss &= ~VIF_MIN_MPDU_SPACING_TRIG_EN;
		if (locked)
			spin_unlock_irqrestore(&vifp->lock, flags);

		schedule_work = true;
		changed = true;
		/* TODO: TWT PS */
		/* TODO: real power save */
		break;
	case TWT_SP:
		if (oldstate == TWT_DOZE) {
			struct sk_buff *skb;
			u32 us;

			locked = spin_trylock_irqsave(&vifp->lock, flags);
			vifp->reg_mmss |= VIF_MIN_MPDU_SPACING_TRIG_EN;
			changed = true;
			schedule_work = true;

			if (vifp->twt_teardown && vifp->twt_request) {
				skb = vifp->twt_request;
				vifp->twt_request = NULL;
				IEEE80211_SKB_CB(skb)->band = def.chan->band;
				IEEE80211_SKB_CB(skb)->control.vif = vif;
				scm80211_tx_channel(hw, skb, def.chan);
			} else if (!vifp->twt_flowtype) {
				struct ieee80211_pspoll *pspoll;

				skb = ieee80211_pspoll_get(hw, vif);
				if (!skb) {
					scm80211_err(hw, "%s alloc pspoll fail\n", __func__);
					break;
				}
				pspoll = (struct ieee80211_pspoll *)skb->data;
				pspoll->frame_control |= cpu_to_le16(IEEE80211_FCTL_PM);
				IEEE80211_SKB_CB(skb)->flags |= IEEE80211_TX_INTFL_DONT_ENCRYPT;
				IEEE80211_SKB_CB(skb)->band = def.chan->band;
				IEEE80211_SKB_CB(skb)->control.vif = vif;
				scm80211_tx_channel(hw, skb, def.chan);
			}

			vifp->twt_next_irq = vifp->twt + vifp->twt_min_sp_dur -
					     SCM80211_TWT_AHEAD_SP_END;
			vifp->twt += vifp->twt_wake_int;
			us = ktime_us_delta(ktime_get(), vifp->twt_ktime);
			vifp->sw_tsf += us;
			vifp->twt_ktime = ktime_get();

			/* set the next doze timer */
			if (WARN_ON(vifp->twt_next_irq < vifp->sw_tsf)) {
				hrtimer_start(&vifp->twt_tsf_timer, ns_to_ktime(5),
					      HRTIMER_MODE_REL_SOFT);
			} else {
				u64 ns = (vifp->twt_next_irq - vifp->sw_tsf) * NSEC_PER_USEC;

				hrtimer_start(&vifp->twt_tsf_timer, ns_to_ktime(ns),
					      HRTIMER_MODE_REL_SOFT);
			}

			if (locked)
				spin_unlock_irqrestore(&vifp->lock, flags);
		}
		schedule_work = true;
	}

exit:
	scm80211_trace(hw, "TWT STATE TRANS from [%s] to [%s] %s\n",
		       twt_state[oldstate], twt_state[state], changed ? "success" : "fail");

	if (changed) {
		vifp->old_state = oldstate;
		vifp->twt_state = state;
	}

	if (vifp->twt_state == TWT_DISABLED && oldstate != TWT_DISABLED) {
		hrtimer_cancel(&vifp->twt_tsf_timer);
		schedule_work = true;
	}

	if (schedule_work)
		ieee80211_queue_work(hw, &vifp->twt_state_work);

	return changed;
}

enum hrtimer_restart scm80211_twt_tsf_timer(struct hrtimer *timer)
{
	struct ieee80211_vif_priv *vifp =
	    container_of(timer, struct ieee80211_vif_priv, twt_tsf_timer);
	struct ieee80211_vif *vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	unsigned long flags;
	enum scm80211_twt_state state;

	spin_lock_irqsave(&vifp->lock, flags);
	state = scm80211_get_twt_state(vif);

	if (state == TWT_SP)
		scm80211_set_twt_state(vif, TWT_DOZE);
	else if (state == TWT_DOZE)
		scm80211_set_twt_state(vif, TWT_SP);

	spin_unlock_irqrestore(&vifp->lock, flags);

	return HRTIMER_NORESTART;
}

static void scm80211_twt_check(void *arg, u8 *mac, struct ieee80211_vif *vif)
{
	int *val = arg;
	enum scm80211_twt_state state =  scm80211_get_twt_state(vif);

	if (state != TWT_DOZE && state != TWT_DISABLED)
		*val = 0;
}
#endif /* SCM80211_TWT */

void __used scm80211_tsf_irq(struct ieee80211_hw *hw, u32 v)
{
}

void scm80211_smps_mode_set(struct ieee80211_hw *hw, struct ieee80211_chanctx_conf *ctx)
{
#ifdef SCM80211_SMPS
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_chanctx *chanctx = (void *)ctx;
	u32 v = scm80211_mac_readl(hw, MAC_REG_DEV_PMU);

	v &= ~(DEV_PMU_SMPS_EN | DEV_PMU_SMPS_MODE);
	/*
	 * according to ieee80211_recalc_smps_chanctx:
	 * ctx->rx_chains_static > 1 present that smps is disable;
	 * if 1 < ctx->rx_chains_dynamic <= priv->n_chain present dynamic smps,
	 * else present static smps
	 */
	if (ctx->rx_chains_static > 1 || priv->n_chain == 1) {
		v |= FIELD_PREP(DEV_PMU_SMPS_EN, 0);
		priv->smps = SMPS_OFF;
	} else if (ctx->rx_chains_dynamic > 1) {
		v |= FIELD_PREP(DEV_PMU_SMPS_EN, 1);
		v |= FIELD_PREP(DEV_PMU_SMPS_MODE, 1);
		priv->smps = SMPS_DYNAMIC;
	} else {
		v |= FIELD_PREP(DEV_PMU_SMPS_EN, 1);
		v |= FIELD_PREP(DEV_PMU_SMPS_MODE, 0);
		priv->smps = SMPS_STATIC;
	}

	scm80211_mac_writel(hw, v, MAC_REG_DEV_PMU);
	scm80211_dbg(hw, "%s: chanctx=%d(n_chain %d, rx_chains_static %d, rx_chains_dynamic %d)\n",
		     __func__, chanctx->id, priv->n_chain, ctx->rx_chains_static,
		     ctx->rx_chains_dynamic);
#endif //SCM80211_SMPS
}

void scm80211_vif_disable_ps(struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = vifp->hw;

	scm80211_vif_power_save(hw, vif, false);
}

void scm80211_vif_enable_ps(struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = vifp->hw;

#ifdef SCM80211_TWT
	if (scm80211_get_twt_state(vif) > TWT_DISABLED)
		return;
#endif

	scm80211_vif_power_save(hw, vif, true);
}

void scm80211_disable_ps(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	if (priv->pm.ps == PS_DISABLED)
		return;

	scm80211_power_save(hw, false, true);
	spin_lock_irqsave(&priv->pm.lock, flags);
	priv->pm.ps = PS_DISABLED;
	spin_unlock_irqrestore(&priv->pm.lock, flags);
	scm80211_trace(hw, "PS disabled\n");
}

static void scm80211_force_disable_ps(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	scm80211_power_save(hw, false, true);
	spin_lock_irqsave(&priv->pm.lock, flags);
	priv->pm.ps = PS_DISABLED;
	spin_unlock_irqrestore(&priv->pm.lock, flags);
	scm80211_trace(hw, "force disable PS\n");
}

void scm80211_enable_ps(struct ieee80211_hw *hw, bool link)
{
	int ps_allow = 1;
	struct ieee80211_conf *conf = &hw->conf;
	struct ieee80211_hw_priv *priv = hw->priv;

#ifdef SCM80211_TWT
	ieee80211_iterate_active_interfaces_atomic(hw,
						   IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_twt_check, &ps_allow);
#endif

	if (!ieee80211_hw_check(hw, SUPPORTS_DYNAMIC_PS) && !(conf->flags & IEEE80211_CONF_PS))
		ps_allow = 0;

	/* Not support PS mode under concurrent mode & dbdc mode */
	if (bitmap_weight(&priv->vif_map, SCM80211_NUM_VIF) > 1 ||
	    priv->dbdc)
		ps_allow = 0;

	/* Win lost will happen when using lower clk */
	if (ps_allow) {
		unsigned long flags;

		scm80211_power_save(hw, true, link);
		spin_lock_irqsave(&priv->pm.lock, flags);
		priv->pm.ps = PS_ENABLED;
		spin_unlock_irqrestore(&priv->pm.lock, flags);
		scm80211_trace(hw, "PS enabled\n");
	} else {
		scm80211_trace(hw, "PS disallow\n");
		scm80211_disable_ps(hw);
	}
}

void scm80211_dynamic_ps_timer(struct timer_list *t)
{
	struct ieee80211_hw_priv *priv =  from_timer(priv, t, pm.timer);

	ieee80211_queue_work(priv->hw, &priv->pm.enable_work);
}

void scm80211_dynamic_ps_enable_work(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv =
		container_of(work, struct ieee80211_hw_priv,
			     pm.enable_work);
	struct ieee80211_hw *hw = priv->hw;
	struct ieee80211_vif *vif = priv->pm.vif;
	unsigned long flags;

	if (!vif || READ_ONCE(priv->pm.ps) == PS_ENABLED)
		return;

	/* check pending frames */
	if (scm80211_tx_frames_pending(hw)) {
		spin_lock_irqsave(&priv->pm.lock, flags);
		priv->pm.acked = false;
		spin_unlock_irqrestore(&priv->pm.lock, flags);
		mod_timer(&priv->pm.timer, jiffies +
			  msecs_to_jiffies(priv->pm.dynamic_ps_timeout));
		return;
	}

	if (READ_ONCE(priv->pm.link) && !READ_ONCE(priv->pm.acked)) {
		struct cfg80211_chan_def def;
		u32 queues = scm80211_vif_queues(vif);

		if (!scm80211_vif_chan(vif, &def))
			scm80211_send_nullfunc(hw, vif, &def, true);

		scm80211_flush(hw, vif, queues, false);
	} else {
		/* goto ps mode */
		scm80211_trace(hw, "goes to powersave mode\n");
		scm80211_vif_enable_ps(vif);
		scm80211_enable_ps(hw, priv->pm.link);

		spin_lock_irqsave(&priv->pm.lock, flags);
		priv->pm.acked = false;
		spin_unlock_irqrestore(&priv->pm.lock, flags);
	}
}

void scm80211_dynamic_ps_disable(struct ieee80211_hw *hw, bool force)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif *vif = READ_ONCE(priv->pm.vif);
	struct cfg80211_chan_def def;

	if (!vif)
		return;

	if (!force && READ_ONCE(priv->pm.ps) != PS_ENABLED)
		return;

	if (force && !is_pcie_device(hw))
		scm80211_force_disable_ps(hw);
	else
		scm80211_disable_ps(hw);

	msleep(2);

	if (!force && !scm80211_vif_chan(vif, &def))
		scm80211_send_nullfunc(hw, vif, &def, false);
	scm80211_wake_queues(hw, scm80211_vif_queues(vif));
}

void scm80211_dynamic_ps_disable_work(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv =
		container_of(work, struct ieee80211_hw_priv,
			     pm.disable_work);
	bool force;
	unsigned long flags;

	spin_lock_irqsave(&priv->pm.lock, flags);
	force = priv->pm.force_awake;
	priv->pm.force_awake = false;
	spin_unlock_irqrestore(&priv->pm.lock, flags);

	if (priv)
		scm80211_dynamic_ps_disable(priv->hw, force);
}

/**
 * scm80211_setup_ps() - check conditions for PS 
 */
void scm80211_setup_link_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	if (!READ_ONCE(priv->pm.enable) || priv->dbdc)
		return;

	if (bitmap_weight((unsigned long *)&priv->vif_map, SCM80211_NUM_VIF) != 1)
		return;

	if (ieee80211_vif_type_p2p(vif) != NL80211_IFTYPE_STATION)
		return;

	scm80211_trace(hw, "setup link ps info for vif#%d\n", scm80211_vdev_id(vif));

	spin_lock_irqsave(&priv->pm.lock, flags);
	priv->pm.vif = vif;
	priv->pm.link = true;
	priv->pm.dynamic_ps_timeout = dynamic_ps_timeout;
	spin_unlock_irqrestore(&priv->pm.lock, flags);

	mod_timer(&priv->pm.timer, jiffies +
		  msecs_to_jiffies(priv->pm.dynamic_ps_timeout));
}

void scm80211_cleanup_link_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	if (!READ_ONCE(priv->pm.enable) || priv->dbdc)
		return;

	del_timer_sync(&priv->pm.timer);
	cancel_work_sync(&priv->pm.enable_work);

	if (READ_ONCE(priv->pm.ps) == PS_ENABLED) {
		scm80211_trace(hw, "info target to disable link powersave mode")
		ieee80211_queue_work(priv->hw, &priv->pm.disable_work);
		msleep(4);
	}

	spin_lock_irqsave(&priv->pm.lock, flags);
	priv->pm.vif = NULL;
	spin_unlock_irqrestore(&priv->pm.lock, flags);
}

void scm80211_setup_unlink_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u16 msecs)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	if (!READ_ONCE(priv->pm.enable) || priv->dbdc)
		return;

	if (bitmap_weight((unsigned long *)&priv->vif_map, SCM80211_NUM_VIF) != 1)
		return;

	if (ieee80211_vif_type_p2p(vif) != NL80211_IFTYPE_STATION)
		return;

	scm80211_trace(hw, "setup unlink ps info for vif#%d\n", scm80211_vdev_id(vif));

	spin_lock_irqsave(&priv->pm.lock, flags);
	priv->pm.vif = vif;
	priv->pm.link = false;
	spin_unlock_irqrestore(&priv->pm.lock, flags);

	mod_timer(&priv->pm.timer, jiffies + msecs_to_jiffies(msecs));
}

void scm80211_cleanup_unlink_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	unsigned long flags;

	if (READ_ONCE(priv->pm.vif) == NULL || READ_ONCE(priv->pm.link))
		return;

	del_timer_sync(&priv->pm.timer);
	cancel_work_sync(&priv->pm.enable_work);

	if (READ_ONCE(priv->pm.ps) == PS_ENABLED) {
		scm80211_trace(hw, "info target to disable unlink powersave mode")
		ieee80211_queue_work(priv->hw, &priv->pm.disable_work);
		msleep(4);
	}

	spin_lock_irqsave(&priv->pm.lock, flags);
	priv->pm.vif = NULL;
	spin_unlock_irqrestore(&priv->pm.lock, flags);
}

int scm80211_wakeup(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->pm.force_awake = false;

	if (READ_ONCE(priv->pm.vif) == NULL)
		return -ENODEV;

	if (READ_ONCE(priv->pm.link))
		scm80211_cleanup_link_ps(hw, priv->pm.vif);
	else
		scm80211_cleanup_unlink_ps(hw, priv->pm.vif);

	priv->pm.force_awake = true;

	return 0;
}

int scm80211_try_to_doze(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif *vif;
	unsigned long flags;

	spin_lock_irqsave(&priv->pm.lock, flags);
	if (!priv->pm.force_awake) {
		spin_unlock_irqrestore(&priv->pm.lock, flags);
		return -EINVAL;
	}

	priv->pm.force_awake = false;
	spin_unlock_irqrestore(&priv->pm.lock, flags);

	if (bitmap_weight((unsigned long *)&priv->vif_map, SCM80211_NUM_VIF) != 1)
		return -EINVAL;

	vif = priv->vifs[find_first_bit((unsigned long *)&priv->vif_map, SCM80211_NUM_VIF)];

	if (ieee80211_vif_type_p2p(vif) != NL80211_IFTYPE_STATION)
		return -EINVAL;

	if (priv->pm.link)
		scm80211_setup_link_ps(hw, vif);
	else
		scm80211_setup_unlink_ps(hw, vif, SCM80211_UNLINK_SCAN);

	return 0;
}

void scm80211_init_pm(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->pm.enable = false;
	timer_setup(&priv->pm.timer, scm80211_dynamic_ps_timer, 0);
	INIT_WORK(&priv->pm.enable_work, scm80211_dynamic_ps_enable_work);
	INIT_WORK(&priv->pm.disable_work, scm80211_dynamic_ps_disable_work);

	if (_ieee80211_hw_check(hw, IEEE80211_HW_SUPPORTS_PS))
		priv->pm.enable = true;

	spin_lock_init(&priv->pm.lock);
}

void scm80211_deinit_pm(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	del_timer_sync(&priv->pm.timer);
	cancel_work_sync(&priv->pm.enable_work);
	cancel_work_sync(&priv->pm.disable_work);
}
