// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 * Hwsim RF driver
 */
#define pr_fmt(fmt)  "hwsim-rf: " fmt
#define dev_fmt(fmt) pr_fmt(fmt)

#include <net/cfg80211.h>

#include "core.h"
#include "phy.h"
#include "rf.h"

static int hwsim_rf_power_up(struct scm80211_radio *radio)
{
	pr_debug("power up\n");
	return 0;
}

static int hwsim_rf_power_down(struct scm80211_radio *radio)
{
	pr_debug("power down\n");
	return 0;
}

static int hwsim_rf_set_channel(struct scm80211_radio *radio, u32 freq, u32 bandwidth)
{
	u32 channel;

	pr_debug("center frequency = %dMHz, bandwidth = %d\n", freq, 20 * (1 << bandwidth));

	channel = ieee80211_frequency_to_channel(freq);
	scm80211_rf_writel(radio->phy->hw, channel, 0x0);

	return 0;
}

static int hwsim_rf_get_cal_data(struct scm80211_radio *radio, struct rf_cal_data *data, int freq)
{
	memset(data, 0, sizeof(*data));
	return 0;
}

static int hwsim_rf_set_rtx_lut(struct scm80211_radio *radio, int band)
{
	pr_debug("set rtx lut\n");
	return 0;
}

static int hwsim_rf_attach(struct scm80211_radio *radio, struct scm80211_phy *phy)
{
	pr_debug("attached to phy\n");

	return 0;
}

static int hwsim_rf_dettach(struct scm80211_radio *radio)
{
	pr_debug("detached from phy\n");
	return 0;
}

static struct scm80211_radio_ops hwsim_radio_ops = {
	.attach = hwsim_rf_attach,
	.detach = hwsim_rf_dettach,
	.power_up = hwsim_rf_power_up,
	.power_down = hwsim_rf_power_down,
	.set_channel = hwsim_rf_set_channel,
	.get_cal_data = hwsim_rf_get_cal_data,
	.set_rtx_lut = hwsim_rf_set_rtx_lut,
};

void scm2020_pdet_set_para_b(int rf_path, int rf_2g, int rf_5g_low,
						   int rf_5g_mid, int rf_5g_high)
{
	pr_debug("hwsim set_para_b\n");
}

void scm2020_set_xtal_cal_b(int xcal)
{
	pr_debug("hwsim set xtal_cal offset\n");
}

static int hwsim_rf_probe(struct platform_device *pdev)
{
	struct scm80211_rf_device *rf = to_rf_device(pdev);
	struct ieee80211_hw_priv *priv = rf->hw->priv;
	int i;

	pr_debug("hwsim probed\n");
	rf->n_chain = priv->dbdc ? 1 : 2;
	rf->radios = devm_kzalloc(&pdev->dev, sizeof(struct scm80211_radio) * 2, GFP_KERNEL);

	if (!rf->radios)
		return -ENOMEM;
	for (i = 0; i < rf->n_chain; i++) {
		rf->radios[i].ops = &hwsim_radio_ops;
		rf->radios[i].devname = pdev->name;
	}
	priv->rf = rf;

	scm80211_notify_rf_ready(rf);

	return 0;
}

static int hwsim_rf_remove(struct platform_device *dev)
{
	pr_debug("removed\n");
	return 0;
}

struct scm80211_rf_driver hwsim_driver = {
	.platform_driver = {
			    .driver = {
				       .name = "hwsim",
				       .suppress_bind_attrs = true,
				       },
			    .probe = hwsim_rf_probe,
			    .remove = hwsim_rf_remove,
			     },
};

static int hwsim_init(void)
{
	return platform_driver_register(&hwsim_driver.platform_driver);
}

static void hwsim_exit(void)
{
	platform_driver_unregister(&hwsim_driver.platform_driver);
}

RF_MODULE(hwsim);
