// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 * Mango RF + AD/DA board driver
 */
#define pr_fmt(fmt)  "mango: " fmt

#include <linux/spi/spi.h>
#include <linux/regmap.h>

#include "core.h"
#include "phy.h"
#include "rf.h"

/**
 * DOC: Mango RF board
 *
 * Mango board has 2 RF chains, each chain consisting of an AD/DA (AD9963) and
 * a tranceiver (MAX2829). Both AD9963 and MAX2829 are accessible via spi
 * interface
 */

struct max2829_context {
	u16 cache[13];
};

struct ad9963 {
	struct spi_device *spi;
	struct regmap *regmap;
};

struct max2829 {
	struct spi_device *spi;
	struct regmap *regmap;
};

struct rf_chain {
	struct scm80211_phy *phy;
	struct ad9963 ad9963;
	struct max2829 max2829;
};

struct mango {
	struct scm80211_radio *radio;
	int id;			/* board id. I know it's ugly */
	struct ad9963 ad9963;
	struct max2829 max2829;
};

/**
 * AD9613: 16-bit instruction word + byte (N) (N=0,1,2)
 * MSB                                  LSB
 * R/W (1) | N1 N0 (2) | 0 (5) | A[7:0] (8)
 *
 */
static int ad9963_spi_reg_write(struct spi_device *spi, u8 addr, u8 data)
{
#ifdef MANGO_OLD_SPI
	u8 tx[3];
	ssize_t len;
	struct spi_transfer xfer = {
		.tx_buf = tx,
		.rx_buf = NULL,
		.len = 3,
	};

	tx[0] = 0;		/* write 1 byte */
	tx[1] = addr;
	tx[2] = data;

	dev_dbg(&spi->dev, "write: addr=%02x, data=%02x, tx=%02x %02x %02x\n",
		addr, data, tx[0], tx[1], tx[2]);

	len = spi_sync_transfer(spi, &xfer, 1);
	if (len < 0)
		return -EIO;
#else
	u32 txd;
	struct spi_transfer xfer = {
		.tx_buf = &txd,
		.rx_buf = NULL,
		.len = 4,
	};
	/* RW(1), N(2), Addr(13), Data(8) */
	txd = addr << 8 | data;

	dev_dbg(&spi->dev, "write: addr=%02x, data=%02x, tx=%08x", addr, data, txd);

	if (spi_sync_transfer(spi, &xfer, 1) < 0)
		return -EIO;
#endif /* MANGO_OLD_SPI */

	return 0;
}

static int ad9963_spi_reg_read(struct spi_device *spi, u8 addr)
{
#ifdef MANGO_OLD_SPI
	u8 tx[3], rx[3];
	int ret;
	struct spi_transfer xfer = {
		.tx_buf = tx,
		.rx_buf = rx,
		.len = 3,
	};

	tx[0] = (1 << 7);	/* read 1 byte */
	tx[1] = addr;

	if (spi_sync_transfer(spi, &xfer, 1) < 0) {
		dev_err(&spi->dev, "ad9963 read failed\n");
		return -EIO;
	}

	ret = rx[2];

	dev_dbg(&spi->dev, "read: addr=%02x, tx=%02x %02x, rx=%02x\n", addr, tx[0], tx[1], rx[2]);
#else
	u32 txd, rxd;
	int ret;
	struct spi_transfer xfer = {
		.tx_buf = &txd,
		.rx_buf = &rxd,
		.len = 4,
	};
	txd = (1 << 31) | addr << 8;

	if (spi_sync_transfer(spi, &xfer, 1) < 0)
		return -EIO;

	ret = rxd;

	dev_dbg(&spi->dev, "read: addr=%02x, tx=%08x, rx=%02x\n", addr, txd, rxd);
#endif /* MANGO_OLD_SPI */

	return ret & 0x00000000ff;
}

static int ad9963_readb(struct ad9963 *ad, u8 addr)
{
	return ad9963_spi_reg_read(ad->spi, addr);
}

static void ad9963_writeb(struct ad9963 *ad, u8 addr, u8 data)
{
	ad9963_spi_reg_write(ad->spi, addr, data);
}

static int ad9963_probe(struct spi_device *spi)
{
	dev_info(&spi->dev, "probed\n");
	return 0;
}

static int ad9963_remove(struct spi_device *spi)
{
	dev_info(&spi->dev, "removed\n");
	return 0;
}

struct spi_driver ad9963_driver = {
	.driver = {
		   .name = "ad9963",
		   .owner = THIS_MODULE,
		    },
	.probe = ad9963_probe,
	.remove = ad9963_remove,
};

/**
 * MAX2829
 */
static int max2829_spi_reg_write(struct spi_device *spi, u16 addr, u16 data)
{
	struct max2829_context *context = spi_get_drvdata(spi);
#ifdef MANGO_OLD_SPI
	u8 tx[3];
	ssize_t len;
	struct spi_transfer xfer = {
		.tx_buf = tx,
		.rx_buf = NULL,
		.len = 3,
	};

	/*
	 * tx[0]: '6b000000 | D[13:12]
	 * tx[1]: D[11:4]
	 * tx[2]: D[3:0] | A[3:0]
	 *
	 * According to MAX2829 datasheet, only the last 18 bits shifted into
	 * the device are retained in the shift register. No check is made on
	 * the number of clock pulses.
	 *
	 * We actually transmit 24 bits with the first 6 bits padded with zero.
	 */
	tx[0] = (data >> 12) & 3;
	tx[1] = (data >> 4) & 0xff;
	tx[2] = ((data & 0xf) << 4) | (addr & 0xf);

	dev_dbg(&spi->dev, "write: addr=%04x, data=%04x, tx=%02x %02x %02x\n",
		addr, data, tx[0], tx[1], tx[2]);

	len = spi_sync_transfer(spi, &xfer, 1);
	if (len < 0)
		return -EIO;

#else
	u32 txd;
	struct spi_transfer xfer = {
		.tx_buf = &txd,
		.rx_buf = NULL,
		.len = 4,
	};
	/* write only: Data[13:0], Addr[3:0] */
	txd = data << 4 | addr;

	dev_dbg(&spi->dev, "write: addr=%04x, data=%04x, tx=%08x", addr, data, txd);

	if (spi_sync_transfer(spi, &xfer, 1) < 0)
		return -EIO;
#endif /* MANGO_OLD_SPI */
	context->cache[addr] = data;
	return 0;
}

static int max2829_spi_reg_read(struct spi_device *spi, u16 addr)
{
	struct max2829_context *context = spi_get_drvdata(spi);

	if (addr > ARRAY_SIZE(context->cache))
		return -EINVAL;

	return context->cache[addr] & 0x0000ffff;
}

static int max2829_readw(struct max2829 *xceiver, u8 addr)
{
	return max2829_spi_reg_read(xceiver->spi, addr);
}

static int max2829_writew(struct max2829 *xceiver, u8 addr, u16 data)
{
	return max2829_spi_reg_write(xceiver->spi, addr, data);
}

static u16 max2829_default[] = {
	[0x0] = 0x1140,		/* Register 0 */
	[0x1] = 0x00CA,		/* Register 1 */
	[0x2] = 0x1007,		/* Standby */
	[0x3] = 0x30A2,		/* Integer-Divider Ratio */
	[0x4] = 0x1DDD,		/* Fractional-Divider Ratio */
	[0x5] = 0x1824,		/* Band Select and PLL */
	[0x6] = 0x1C00,		/* Calibration */
	[0x7] = 0x002A,		/* Lowpass Filter */
	[0x8] = 0x0025,		/* Rx Control/RSSI */
	[0x9] = 0x0200,		/* Tx Linearity/Baseband Gain */
	[0xa] = 0x03C0,		/* PA Bias DAC */
	[0xb] = 0x007F,		/* Rx Gain */
	[0xc] = 0x0000,		/* Tx VGA Gain */
};

/* temporary use for scm2020z2 fpga verification */
#define SCM2020Z2_FPGA

// mango RX/TX LUT
static struct rf_lut mango_rf_lut = {
	{
		{	// 2.4G TX LUT
			0x00000028, 0x00000128, 0x00000228, 0x00000328, 0x00000428, 0x00000528,
			0x00000628, 0x00000728, 0x00000828, 0x00000928, 0x00000A28, 0x00000B28,
			0x00000C28, 0x00000D28, 0x00000E28, 0x00000F28, 0x00001028, 0x00001128,
			0x00001228, 0x00001328, 0x00001428, 0x00001528, 0x00001628, 0x00001728,
			0x00001828, 0x00001928, 0x00001A28, 0x00001B28, 0x00001C28, 0x00001D28,
			0x00001E28, 0x00001F28, 0x00002028, 0x00002128, 0x00002228, 0x00002328,
			0x00002428, 0x00002528, 0x00002628, 0x00002728, 0x00002828, 0x00002928,
			0x00002A28, 0x00002B28, 0x00002C28, 0x00002D28, 0x00002E28, 0x00002F28,
			0x00003028, 0x00003128, 0x00003228, 0x00003328, 0x00003428, 0x00003528,
			0x00003628, 0x00003728, 0x00003828, 0x00003928, 0x00003A28, 0x00003B28,
			0x00003C28, 0x00003D28, 0x00003E28, 0x00003F28
		},
		{	// 5G TX LUT
			0x00000028, 0x00000128, 0x00000228, 0x00000328, 0x00000428, 0x00000528,
			0x00000628, 0x00000728, 0x00000828, 0x00000928, 0x00000A28, 0x00000B28,
			0x00000C28, 0x00000D28, 0x00000E28, 0x00000F28, 0x00001028, 0x00001128,
			0x00001228, 0x00001328, 0x00001428, 0x00001528, 0x00001628, 0x00001728,
			0x00001828, 0x00001928, 0x00001A28, 0x00001B28, 0x00001C28, 0x00001D28,
			0x00001E28, 0x00001F28, 0x00002028, 0x00002128, 0x00002228, 0x00002328,
			0x00002428, 0x00002528, 0x00002628, 0x00002728, 0x00002828, 0x00002928,
			0x00002A28, 0x00002B28, 0x00002C28, 0x00002D28, 0x00002E28, 0x00002F28,
			0x00003028, 0x00003128, 0x00003228, 0x00003328, 0x00003428, 0x00003528,
			0x00003628, 0x00003728, 0x00003828, 0x00003928, 0x00003A28, 0x00003B28,
			0x00003C28, 0x00003D28, 0x00003E28, 0x00003F28
		}
	},
	64,
#ifdef SCM2020Z2_FPGA
	{
		{
			0x00002026, 0x00002028, 0x00002126, 0x00002128,
			0x00002226, 0x00002228, 0x00002326, 0x00002328,
			0x00002426, 0x00002428, 0x00002526, 0x00002528,
			0x00002626, 0x00002628, 0x00002726, 0x00002728,
			0x00002826, 0x00002828, 0x00002926, 0x00002928,
			0x00002A26, 0x00002A28, 0x00002B26, 0x00002B28,
			0x00004428, 0x00004526, 0x00004528, 0x00004626,
			0x00004628, 0x00004726, 0x00004728, 0x00004826,
			0x00004828, 0x00004926, 0x00004928, 0x00004A26,
			0x00004A28, 0x00004B26, 0x00004B28, 0x00004C26,
			0x00004C28, 0x00006527, 0x00006529, 0x00006627,
			0x00006629, 0x00006727, 0x00006729, 0x00006827,
			0x00006829, 0x00006927, 0x00006929, 0x00006A27,
			0x00006A29, 0x00006B27, 0x00006B29, 0x00006C27,
			0x00006C29, 0x00006D27, 0x00006D29, 0x00006E27,
			0x00006E29, 0x00006F27, 0x00006F29, 0x00007027,
			0x00007029, 0x00007127, 0x00007129, 0x00007227,
			0x00007229, 0x00007327, 0x00007329, 0x00007427,
			0x00007429, 0x00007527, 0x00007529, 0x00007627,
			0x00007629, 0x00007727, 0x00007729, 0x00007827,
			0x00007829, 0x00007927, 0x00007929, 0x00007A27,
			0x00007A29, 0x00007B27, 0x00007B29, 0x00007C27,
			0x00007C29, 0x00007D27, 0x00007D29, 0x00007E27,
			0x00007E29, 0x00007F27, 0x00007F29
		},
		/* 5 GHz band */
		{
			0x00002026, 0x00002028, 0x00002126, 0x00002128,
			0x00002226, 0x00002228, 0x00002326, 0x00002328,
			0x00002426, 0x00002428, 0x00002526, 0x00002528,
			0x00002626, 0x00002628, 0x00002726, 0x00002728,
			0x00002826, 0x00002828, 0x00002926, 0x00002928,
			0x00002A26, 0x00002A28, 0x00002B26, 0x00002B28,
			0x00002C26, 0x00004426, 0x00004428, 0x00004526,
			0x00004528, 0x00004626, 0x00004628, 0x00004726,
			0x00004728, 0x00004826, 0x00004828, 0x00004926,
			0x00004928, 0x00004A26, 0x00004A28, 0x00004B26,
			0x00004B28, 0x00004C26, 0x00004C28, 0x00004D26,
			0x00006427, 0x00006429, 0x00006527, 0x00006529,
			0x00006627, 0x00006629, 0x00006727, 0x00006729,
			0x00006827, 0x00006829, 0x00006927, 0x00006929,
			0x00006A27, 0x00006A29, 0x00006B27, 0x00006B29,
			0x00006C27, 0x00006C29, 0x00006D27, 0x00006D29,
			0x00006E27, 0x00006E29, 0x00006F27, 0x00006F29,
			0x00007027, 0x00007029, 0x00007127, 0x00007129,
			0x00007227, 0x00007229, 0x00007327, 0x00007329,
			0x00007427, 0x00007429, 0x00007527, 0x00007529,
			0x00007627, 0x00007629, 0x00007727, 0x00007729,
			0x00007827, 0x00007829, 0x00007927, 0x00007929,
			0x00007A27, 0x00007A29, 0x00007B27, 0x00007B29,
			0x00007C27, 0x00007C29, 0x00007D27
		}

	},
#else
	{
		{	// 2.4G RX LUT
			0x00002026, 0x00002028, 0x00002126, 0x00002128, 0x00002226, 0x00002228,
			0x00002326, 0x00002328, 0x00002426, 0x00002428, 0x00002526, 0x00002528,
			0x00002626, 0x00002628, 0x00002726, 0x00002728, 0x00002826, 0x00002828,
			0x00002926, 0x00002928,	0x00002A26, 0x00002A28, 0x00002B26, 0x00002B28,
			0x00004428, 0x00004526, 0x00004528, 0x00004626,	0x00004628, 0x00004726,
			0x00004728, 0x00004826, 0x00004828, 0x00004926, 0x00004928, 0x00004A26,
			0x00004A28, 0x00004B26, 0x00004B28, 0x00004C26, 0x00004C28, 0x00006527,
			0x00006529, 0x00006627,	0x00006629, 0x00006727, 0x00006729, 0x00006827,
			0x00006829, 0x00006927, 0x00006929, 0x00006A27,	0x00006A29, 0x00006B27,
			0x00006B29, 0x00006C27, 0x00006C29, 0x00006D27, 0x00006D29, 0x00006E27,
			0x00006E29, 0x00006F27, 0x00006F29, 0x00007027, 0x00007029, 0x00007127,
			0x00007129, 0x00007227,	0x00007229, 0x00007327, 0x00007329, 0x00007427,
			0x00007429, 0x00007527, 0x00007529, 0x00007627,	0x00007629, 0x00007727,
			0x00007729, 0x00007827, 0x00007829, 0x00007927, 0x00007929, 0x00007A27,
			0x00007A29, 0x00007B27, 0x00007B29, 0x00007C27, 0x00007C29, 0x00007D27,
			0x00007D29, 0x00007E27,	0x00007E29, 0x00007F27, 0x00007F29
		},
		{	// 5G RX LUT
			0x00002026, 0x00002028, 0x00002126, 0x00002128,	0x00002226, 0x00002228,
			0x00002326, 0x00002328, 0x00002426, 0x00002428, 0x00002526, 0x00002528,
			0x00002626, 0x00002628, 0x00002726, 0x00002728, 0x00002826, 0x00002828,
			0x00002926, 0x00002928,	0x00002A26, 0x00002A28, 0x00002B26, 0x00002B28,
			0x00002C26, 0x00004426, 0x00004428, 0x00004526,	0x00004528, 0x00004626,
			0x00004628, 0x00004726, 0x00004728, 0x00004826, 0x00004828, 0x00004926,
			0x00004928, 0x00004A26, 0x00004A28, 0x00004B26, 0x00004B28, 0x00004C26,
			0x00004C28, 0x00004D26,	0x00006427, 0x00006429, 0x00006527, 0x00006529,
			0x00006627, 0x00006629, 0x00006727, 0x00006729,	0x00006827, 0x00006829,
			0x00006927, 0x00006929, 0x00006A27, 0x00006A29, 0x00006B27, 0x00006B29,
			0x00006C27, 0x00006C29, 0x00006D27, 0x00006D29, 0x00006E27, 0x00006E29,
			0x00006F27, 0x00006F29,	0x00007027, 0x00007029, 0x00007127, 0x00007129,
			0x00007227, 0x00007229, 0x00007327, 0x00007329,	0x00007427, 0x00007429,
			0x00007527, 0x00007529, 0x00007627, 0x00007629, 0x00007727, 0x00007729,
			0x00007827, 0x00007829, 0x00007927, 0x00007929, 0x00007A27, 0x00007A29,
			0x00007B27, 0x00007B29,	0x00007C27, 0x00007C29, 0x00007D27
		}
	},
#endif
	95,
};

static int max2829_probe(struct spi_device *spi)
{
	struct max2829_context *context;

	dev_info(&spi->dev, "%s probed\n", spi->modalias);

	context = devm_kzalloc(&spi->dev, sizeof(*context), GFP_KERNEL);
	if (!context)
		return -ENOMEM;
	spi_set_drvdata(spi, context);

	return 0;
}

static int max2829_remove(struct spi_device *spi)
{
	dev_info(&spi->dev, "%s removed\n", spi->modalias);
	return 0;
}

static int mango_tranceiver_init(struct mango *mango)
{
	struct max2829 *max2829 = &mango->max2829;
	int err, addr;

	/* Initialize max2829 registers with default values */
	for (addr = 0; addr < ARRAY_SIZE(max2829_default); addr++) {
		err = max2829_writew(max2829, addr, max2829_default[addr]);
		if (err) {
			dev_err(&max2829->spi->dev, "failed to write %04x to reg%02x\n",
				max2829_default[addr], addr);
			return -EIO;
		}
	}
	return 0;
}

struct spi_driver max2829_driver = {
	.driver = {
		   .name = "max2829",
		   .owner = THIS_MODULE,
		    },
	.probe = max2829_probe,
	.remove = max2829_remove,
};

static void mango_set_frequency(struct mango *mango, u32 freq)
{
	struct max2829 *max2829 = &mango->max2829;
	u16 r3, r4, r5;
	u32 div, R;

	r5 = max2829_readw(max2829, 0x5) & ~0x41;

	if (freq < 3000) {
		div = 3;
	} else {
		r5 |= (freq >= 5400) ? 0x41 : 0x01;
		div = 5;
	}

	/*
	 * R = freq x (4 / div) / 20 x 2^16
	 *   = freq / (div * 5) * 2^16
	 *   = (Rint + Rfrac) x 2^16 (Rint = R[31:16], Rfrac = R[15:0])
	 *
	 * r3 = (R[24:16] | R[1:0] << 12)
	 * r4 = R[15:3] >> 2
	 */
	R = (1 << 16) * freq / (div * 5);
	r3 = (R >> 16) | ((R & 0x3) << 12);
	r4 = (R >> 2) & GENMASK(13, 0);

	max2829_writew(max2829, 0x5, r5);
	max2829_writew(max2829, 0x3, r3);
	max2829_writew(max2829, 0x4, r4);
}

static void mango_set_bandwidth(struct mango *mango, int bandwidth)
{
	struct max2829 *max2829 = &mango->max2829;
	struct ad9963 *ad9963 = &mango->ad9963;
	u8 r7;

	/*
	 * R7: Low-pass filter register
	 * - D[6:5] Tx LPF corner frequency coarse adjustment
	 *   0: undefined, 1: 12MHz, 2: 18MHz, 3: 24MHz
	 * - D[4:3] Rx LPF corner frequency coarse adjustment
	 *   0: 7.5MHz, 1: 9.5MHz, 2: 14MHz, 3: 18MHz
	 */
	r7 = max2829_readw(max2829, 0x7);
	r7 &= ~0x78;
	r7 |= (bandwidth == 0) ? 0x1 << 5 : 0x3 << 5;	/* tx */
	r7 |= (bandwidth == 0) ? 0x1 << 3 : 0x3 << 3;	/* rx */
	max2829_writew(max2829, 0x7, r7);

	/* AD/DA */
	ad9963_writeb(ad9963, 0x71, (bandwidth == 0) ? 0x51 : 0xd1);
}

static void mango_tx_gain_manual_config(struct mango *mango, int gain_index)
{
	struct max2829 *max2829 = &mango->max2829;
	u16 r12;

	r12 = max2829_readw(max2829, 0xc);
	r12 &= ~0x3f;
	r12 |= gain_index & 0x3f;
	max2829_writew(max2829, 0xc, r12);
}

static void mango_rx_gain_mode_config(struct mango *mango, int mode)
{
	struct max2829 *max2829 = &mango->max2829;
	u8 r12;

	r12 = max2829_readw(max2829, 0xc);
	r12 &= ~(0x1 << 12);
	r12 |= (mode << 12);
	max2829_writew(max2829, 0xc, r12);
}

static void mango_rx_gain_manual_config(struct mango *mango, int lna_gain_index, int vga_gain_index)
{
	struct max2829 *max2829 = &mango->max2829;
	u8 r11;

	r11 = max2829_readw(max2829, 11);
	r11 &= ~0x7f;
	r11 |= (((lna_gain_index & 0x3) << 5) | (vga_gain_index & 0x1f));
	max2829_writew(max2829, 11, r11);
}

static int mango_afe_config_power(struct mango *mango, u8 pwrstate)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u16 v;

	if (pwrstate == 0)
		v = 0x3f;	/* everything powered down */
	else if (pwrstate == 1)
		v = 0x00;	/* DLL off, everything else on */
	else
		return -EINVAL;
	ad9963_writeb(ad9963, 0x60, v);
	return 0;
}

static int mango_afe_set_dll(struct mango *mango, u8 enable, u8 M, u8 N, u8 div)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u16 v, retries = 100;

	if ((N == 7) || (N == 0) || (N > 8))
		return -EINVAL;
	if (M > 31)
		return -EINVAL;
	if ((div == 0) || (div == 3) || (div > 4))
		return -EINVAL;

	if (enable == 0) {
		/* Assert DLL reset */
		ad9963_writeb(ad9963, 0x75, 0x08);

		/* Power down DLL block, leave all other blocks powered on */
		ad9963_writeb(ad9963, 0x60, 0x00);

		/* Disable DLL clock input, set ADC/DAC clock sources to ext ref clock */
		ad9963_writeb(ad9963, 0x71, 0xC0);

		return 0;
	}

	/* Assert DLL reset */
	ad9963_writeb(ad9963, 0x75, 0x08);

	/* Power up DLL block, leave all other blocks powered on */
	ad9963_writeb(ad9963, 0x60, 0x80);

	/* Assert DLL clock enable, set DLL_N */

	/* reg71 has bits to preserve, so handle it separately for each AD */
	v = ad9963_readb(ad9963, 0x71);
	v &= ~0x1f;
	v |= 0x10 | (N & 0xf);
	ad9963_writeb(ad9963, 0x71, v);

	/* Other registers are DLL-only, so we can write both ADs together */

	/* Set DLL_DIV and DLL_M */
	ad9963_writeb(ad9963, 0x72, ((div & 0x3) << 5) | (M & 0x1f));

	/* Release DLL reset (treating as active high) */
	udelay(100);
	ad9963_writeb(ad9963, 0x75, 0x00);

	/* Wait for both DLLs to lock */
	while (retries-- > 0) {
		v = ad9963_readb(ad9963, 0x72);
		if (v & 0x80)
			return 0;
		udelay(10);
	}
	pr_err("ad9963 failed to lock DLL\n");
	return -EIO;
}

static u8 mango_afe_set_clocks(struct mango *mango, u8 dac_clksrc,
			       u8 adc_clksrc, u8 adc_clkdiv, u8 adc_dcs)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u8 v;

	v = ad9963_readb(ad9963, 0x66);
	v &= ~0x7;
	v |= (adc_dcs << 2) | (adc_clkdiv & 0x3);
	ad9963_writeb(ad9963, 0x66, v);

	v = ad9963_readb(ad9963, 0x71);
	v &= ~0xc0;
	v |= (dac_clksrc << 6) | (adc_clksrc << 7);
	ad9963_writeb(ad9963, 0x71, v);

	return 0;
}

static int mango_afe_set_filters(struct mango *mango, u8 interpolation, u8 decimation)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u8 v = 0x3;

	switch (interpolation) {
	case 1:
		v |= 0x1c;
		break;
	case 2:
		v |= 0x14;
		break;
	case 4:
		v |= 0x04;
		break;
	case 8:
		break;
	default:
		return -EINVAL;
	}

	if (decimation == 1)
		v |= 0x20;
	else if (decimation != 2)
		return -EINVAL;

	ad9963_writeb(ad9963, 0x30, v);
	return 0;
}

static int mango_afe_set_tx_gain1(struct mango *mango, u8 iqsel, u8 gain)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u8 reg;

	if (iqsel != 0 && iqsel != 1)
		return -EINVAL;
	if (gain > 63)
		return -EINVAL;

	/* 6-bit Linear-in-dB gain +/- 6dB */
	reg = (iqsel == 0) ? 0x68 /* IGAIN 2 */ : 0x6b /* QGAIN 2 */;
	ad9963_writeb(ad9963, reg, (gain & 0x3F));
	return 0;
}

static int mango_afe_set_tx_gain2(struct mango *mango, u8 iqsel, u8 gain)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u8 reg;

	if (iqsel != 0 && iqsel != 1)
		return -EINVAL;

	/* 6-bit Linear gain +/- 2.5% */
	reg = (iqsel == 0) ? 0x69 /* IGAIN 2 */ : 0x6c /* QGAIN 2 */;
	ad9963_writeb(ad9963, reg, (gain & 0x3f));
	return 0;
}

static int mango_afe_set_tx_dco(struct mango *mango, u8 iqsel, u32 dco)
{
	struct ad9963 *ad9963 = &mango->ad9963;
	u8 reg;

	/* Sanity check inputs */
	if (iqsel != 0 && iqsel != 1)
		return -EINVAL;
	if (dco > 1023)
		return 0xff;

	reg = (iqsel == 0) ? 0x49 : 0x46;
	ad9963_writeb(ad9963, reg, ((dco & 0x3ff) >> 2));
	ad9963_writeb(ad9963, reg + 1, (dco & 0x3));
	return 0;
}

/* mango afe */
static int mango_afe_init(struct mango *mango)
{
	struct scm80211_radio *radio = mango->radio;
	struct ad9963 *ad9963 = &mango->ad9963;
	u32 phyver, macver;
	u8 v;

	v = ad9963_readb(ad9963, 0x5c);
	if (!(v & 0x8)) {
		pr_err("ad9963 reg 0x5c read value wrong: %02x\n", v);
		return -EIO;
	}
	v = ad9963_readb(ad9963, 0x72);
	if (!(v & 0x1)) {
		pr_err("ad9963 reg 0x72 read value wrong: %02x\n", v);
		return -EIO;
	}

	/* Power on/off blocks */
	ad9963_writeb(ad9963, 0x40, 0x00);	/* DAC12A, DAC12B off */
	ad9963_writeb(ad9963, 0x60, 0x00);	/* DLL off, everything else on */
	ad9963_writeb(ad9963, 0x61, 0x03);	/* LDOs on, AUXADC off */
	/* xil_printf("AD TEST: reg61=0x%08x\n", ad_spi_read(baseaddr,  (adSel), 0x61)); */

	/*
	 * Clocking setup
	 * [7]=0: ADCCLK=ext clk
	 * [6]=0: DACCLK=ext clk
	 * [4]=0: disable DLL input ref
	 * [3:0]: DLL divide ratio (only 1, 2, 3, 4, 5, 6, 8 valid)
	 */

	/* DLL ref input off, ADCCLK=extCLK, DACCLK=extCLK, DLL_DIV=1 */
	ad9963_writeb(ad9963, 0x71, 0x01);

	/* TXDBL_SEL */
	ad9963_writeb(ad9963, 0x39, 0x01);

	/*
	 * Reference resistor selection
	 * Datasheet says reg62[0]=0 for internal resistor, reg62[0]=1 for
	 * external resistor
	 * But experimentally DAC currents are much more stable with temperature
	 * when reg62[0]=0
	 * I'm guessing this bit is flipped in the datasheet
	 */
	ad9963_writeb(ad9963, 0x62, 0x00);

	/*
	 * Clock disables and DCS
	 * [7:3]=0: Enable internal clocks to ADCs/DACs
	 * [1:0]=0: Set ADCCLK=ext clock (no division)
	 * [2]=0: Enable ADC duty cycle stabilizer (recommended for ADC rates > 75MHz)
	 */
	ad9963_writeb(ad9963, 0x66, 0x00);	//Enable internal clocks, enable DCS

	/*
	 * Aux DACs (Tx DC offset adjustment)
	 *  DAC10B=Q DCO, DAC10A=I DCO
	 *  DAC outputs update after LSB write (configured by reg40[0])
	 */
	ad9963_writeb(ad9963, 0x45, 0x88);	/* DAC10B on, full scale = [0,+2]v */
	ad9963_writeb(ad9963, 0x46, 0x80);	/* DAC10B data[9:2] */
	ad9963_writeb(ad9963, 0x47, 0x00);	/* DAC10B {6'b0, data[1:0]} */

	ad9963_writeb(ad9963, 0x48, 0x88);	/* DAC10A on, full scale = [0,+2]v */
	ad9963_writeb(ad9963, 0x49, 0x80);	/* DAC10A data[9:2] */
	ad9963_writeb(ad9963, 0x50, 0x00);	/* DAC10A {6'b0, data[1:0]} */

	/* ADC common mode buffer: disabled for DC coupled inputs */
	ad9963_writeb(ad9963, 0x7E, 0x01);

	/*
	 * Spectral inversion
	 * Invert RxI (reg3D[2]=1) to match PCB
	 * TxI, TxQ also swapped on PCB, but ignored here since -1*(a+jb) is
	 * just one of many phase shifts the Tx signal sees
	 * ad9963_writeb(ad9963, 0x3D, 0x04);
	 */
	/* Invert RxI to match PCB (board swaps +/- for better routing) */
	ad9963_writeb(ad9963, 0x3d, 0x04);

	/*
	 * Rx clock and data order/format
	 * [7:1]=[0 x 1 0 1 0 1] to match design of ad_bridge input registers
	 *       (TRXD DDR relative to TRXCLK, I data first)
	 * [0]=1 for 2's compliment (Datasheet says reg32[0]=0 for 2's compliment,
	 *       but experiments show 0 is really straight-binary)
	 * ad9963_writeb(ad9963, 0x32, 0xA7); // ADC clock is "0" deg.
	 */
	ad9963_writeb(ad9963, 0x32, 0xAF);	/* ADC clock is "180" deg. */

	/* Full-duplex mode (DACs/TXD and ADCs/TRXD both active all the time) */
	ad9963_writeb(ad9963, 0x3F, 0x01);	/* FD mode */

	/*
	 * Tx data format
	 * (datasheet bug! reg[31][0] is flipped;
	 *  - 1=2's complement,
	 *  - 0=straight binary)
	 *
	 * 0x17 worked great with latest ad_bridge (where TXCLK is
	 * ODDR(D1=1,D0=0,C=ad_ref_clk_90) and TXD are ODDR
	 * (D1=I,D2=Q,C=ad_ref_clk_0))
	 * ad9963_writeb(ad9963, 0x31, 0x97); //Txdata=DDR two's complement,
	 */

	/*  DAC clock phase control */
	phyver = (scm80211_phy_version(radio->phy->hw) & 0xfff0) >> 4;
	macver = (scm80211_mac_version(radio->phy->hw) & 0xfff0) >> 4;

	if (phyver == 4 || phyver == 5 || (phyver == 6 && macver == 3)
	    || (phyver >= 9 && phyver < 17)
	    || (phyver >= 22 && macver != 29))
		ad9963_writeb(ad9963, 0x31, 0x97);	/* Txdata=DDR two's complement, */
	else
		ad9963_writeb(ad9963, 0x31, 0x9f);	/* Txdata=DDR two's complement, */

	/* Tx/Rx data paths */
	/* Bypass all rate change filters, enable Tx/Rx clocks */
	ad9963_writeb(ad9963, 0x30, 0x3F);

	/*
	 * ADC RXCML output buffer requires special register process
	 * (see AD9963 datasheet pg. 21 "sub serial interface communications")
	 */
	ad9963_writeb(ad9963, 0x05, 0x03);	/* Address both ADCs */
	ad9963_writeb(ad9963, 0x0F, 0x02);	/* Enable RXCML output */
	ad9963_writeb(ad9963, 0x10, 0x00);	/* Set I/Q offset to 0 */
	ad9963_writeb(ad9963, 0xFF, 0x01);	/* Trigger ADC param update */
	ad9963_writeb(ad9963, 0x05, 0x00);	/* De-Address both ADCs */

	//REFIO adjustment: set to default of 0.8v
	ad9963_writeb(ad9963, 0x6E, 0x40);

	/*
	 * Tx gains (it seems these registers default to non-zero, and maybe
	 * non-I/Q-matched values; safest to set them explicitly after reset)
	 */
	/*
	 * I/Q GAIN1[5:0]: Fix5_0 value, Linear-in-dB, 0.25dB per bit
	 *   0-25=>0dB-+6dB, 25-32:+6dB, 33-41:-6dB, 41-63=>-6dB-0dB
	 */
	ad9963_writeb(ad9963, 0x68, 0);	/* IGAIN1 */
	ad9963_writeb(ad9963, 0x6B, 0);	/* QGAIN1 */

	/*
	 * I/Q GAIN2[5:0]: Fix5_0 value, Linear +/-2.5%, 0.08% per bit
	 *  0:+0, 31:+max, 32:-max, 63:-0
	 */
	ad9963_writeb(ad9963, 0x69, 0);	/* IGAIN2 */
	ad9963_writeb(ad9963, 0x6C, 0);	/* QGAIN2 */

	/* I/Q RSET[5:0]: Fix5_0, Linear +/-20%, 0.625% per bit */
	/*  0:-0, 31:-max, 32:+max, 63:+0 */
	ad9963_writeb(ad9963, 0x6A, 0);	/* IRSET */
	ad9963_writeb(ad9963, 0x6D, 0);	/* QRSET */

	/* Digital output drive strengths: all 8mA */
	ad9963_writeb(ad9963, 0x63, 0xAA);	/* 2 bits each: TRXD TRXIQ TRXCLK TXCLK */

	/* Disable Tx and Rx BIST modes */
	ad9963_writeb(ad9963, 0x50, 0x00);	/* Tx BIST control */
	ad9963_writeb(ad9963, 0x51, 0x00);	/* Rx BIST control */

	return 0;
}

static int mango_power_up(struct scm80211_radio *radio)
{
	struct mango *mango = radio->priv;

	mango_afe_init(mango);
	mango_afe_set_dll(mango, 1, 1, 1, 1);
	mango_afe_set_clocks(mango, 1, 0, 1, 0);
	mango_afe_set_filters(mango, 1, 1);
	mango_afe_set_tx_dco(mango, 0, 0);

	mango_tranceiver_init(mango);

	return 0;
}

static int mango_power_down(struct scm80211_radio *radio)
{
	/* TBD */
	return 0;
}

static int mango_set_channel(struct scm80211_radio *radio, unsigned int frequency,
			     unsigned int bandwidth)
{
	struct mango *mango = radio->priv;

	mango_set_frequency(mango, frequency);
	mango_set_bandwidth(mango, bandwidth);
	return 0;
}

/* Mango board cal value apply*/
static struct rf_cal_data mango_board_cal[] = {
	[0] = { /* Board #1 */
		{4096, 4096},	//TX SSB param_1
		{   0,    0},	//TX SSB param_0
		{-256, -384},	//TX CL ofs_i
		{ 256,  384},	//TX CL ofs_q
		{4096, 4096},	//RX IMB param_1
		{   0,    0},	//RX IMB param_0
		{-704, -512},	//RX DC ofs_i
		{ 640,   32}	//RX DC ofs_q
	},
	[1] = { /* Board #2 */
		{4096, 4096},	//TX SSB param_1
		{   0,    0},	//TX SSB param_0
		{-256, -256},	//TX CL ofs_i
		{   0, -256},	//TX CL ofs_q
		{4096, 4096},	//RX IMB param_1
		{   0,    0},	//RX IMB param_0
		{   0,    0},	//RX DC ofs_i
		{ 128,  256},	//RX DC ofs_q
	},
	[2] = { /* Board #3 */
		{4096, 4096},	//TX SSB param_1
		{   0,    0},	//TX SSB param_0
		{-1024, -768},	//TX CL ofs_i
		{ 512,  128},	//TX CL ofs_q
		{4096, 4096},	//RX IMB param_1
		{   0,    0},	//RX IMB param_0
		{-576, -512},	//RX DC ofs_i
		{ 576,    0},	//RX DC ofs_q
	},
	[3] = { /* Board #4 */
		{4096, 4096},	//TX SSB param_1
		{   0,    0},	//TX SSB param_0
		{-128,    0},	//TX CL ofs_i
		{ 128,  192},	//TX CL ofs_q
		{4096, 4096},	//RX IMB param_1
		{   0,    0},	//RX IMB param_0
		{-384, -736},	//RX DC ofs_i
		{ 512,  192},	//RX DC ofs_q
	},
	[4] = { /* Board #5 */
		{4096, 4096},	//TX SSB param_1
		{   0,    0},	//TX SSB param_0
		{-512, -256},	//TX CL ofs_i
		{ 256,  256},	//TX CL ofs_q
		{4096, 4096},	//RX IMB param_1
		{   0,    0},	//RX IMB param_0
		{-288, -352},	//RX DC ofs_i
		{ 160,   64},	//RX DC ofs_q
	},
	[5] = { /* Board #6 */
		{4096, 4096},   //TX SSB param_1
		{   0,    0},   //TX SSB param_0
		{-528, -112},   //TX CL ofs_i
		{   0, -384},   //TX CL ofs_q
		{4096, 4096},   //RX IMB param_1
		{   0,    0},   //RX IMB param_0
		{-448, -448},   //RX DC ofs_i
		{ 448,  -32},   //RX DC ofs_q
	},
};

static int mango_get_cal_data(struct scm80211_radio *radio, struct rf_cal_data *data, int freq)
{
	struct mango *mango = radio->priv;

	pr_info("%s: board id = %d\n", __func__, mango->id);
	memcpy(data, &mango_board_cal[mango->id - 1], sizeof(*data));
	return 0;
}

static int mango_set_rtx_lut(struct scm80211_radio *radio, int freq)
{
	struct rf_lut *lut = radio->rf_lut;
	struct scm80211_phy *phy = radio->phy;
	int i;
	int band = freq < 3000 ? NL80211_BAND_2GHZ : NL80211_BAND_5GHZ;
	u32 phyver = (scm80211_phy_version(radio->phy->hw) & 0xfff0) >> 4;

	/* Rx LUT APB access en */
	phy_writel(phy, 0x00010000, PHY_RFIF(FW_RXLUT_CFG0));
	for (i = 0; i < lut->rx_lut_cnt; i++) {
		if (phyver < 22) {
			/* new rx gain tabl */
			phy_writel(phy, lut->rx_lut[band][i],
				   PHY_RFIF(FW_RX_GAIN_LUT_00 + (i * 4)));
		} else {
			phy_writel(phy, (2 << 8) | i, PHY_RFIF(FW_RFI_RXLUTV2_ADDR));
			phy_writel(phy, lut->rx_lut[band][i], PHY_RFIF(FW_RFI_RXLUTV2_WRDATA));
			phy_writel(phy, 0x1, PHY_RFIF(FW_RFI_RXLUTV2_WRSTART));
		}
	}
	/* Rx LUT APB access dis */
	phy_writel(phy, 0x00000000, PHY_RFIF(FW_RXLUT_CFG0));

	/* Tx LUT APB access en */
	phy_writel(phy, 0x00010000, PHY_RFIF(FW_TXLUT_CFG0));
	for (i = 0; i < lut->tx_lut_cnt; i++) {
		if (phyver < 22) {
			/* new rx gain tabl */
			phy_writel(phy, lut->tx_lut[band][i],
				   PHY_RFIF(FW_TX_GAIN_LUT_00 + (i * 4)));
		} else {
			phy_writel(phy, (2 << 8) | i, PHY_RFIF(FW_RFI_TXLUTV2_ADDR));
			phy_writel(phy, lut->tx_lut[band][i], PHY_RFIF(FW_RFI_TXLUTV2_WRDATA));
			phy_writel(phy, 0x1, PHY_RFIF(FW_RFI_TXLUTV2_WRSTART));
		}
	}
	/* Tx LUT APB access dis */
	phy_writel(phy, 0x00000000, PHY_RFIF(FW_TXLUT_CFG0));

	return 0;
}

/*
 * bus_num is encoded as follows:
 * bus_num[15:8]: chain number
 * bus_num[7:0]: spi master id
 */
static struct spi_board_info mango_spi_info[2][2] = {	/*[path][spi] */
	[0] = {
		{
			.modalias = "ad9963",
			.bus_num = 0, /* chain 0, spi 0 */
			.platform_data = "A"
		},
		{
			.modalias = "max2829",
			.bus_num = 1, /* chain 0, spi 1 */
			.platform_data = "A"
		},
	},
	[1] = {
		{
			.modalias = "ad9963",
			.bus_num = 1 << 8 | 0, /* chain 1, spi 0 */
			.platform_data = "B"
		},
		{
			.modalias = "max2829",
			.bus_num = 1 << 8 | 1, /* chain 1, spi 1 */
			.platform_data = "B"
		},
	}
};

/* GSPI config */
static struct rf_gspi_cfg mango_gspi_cfg[2] = {
#ifndef MANGO_OLD_SPI
	/* ADDA */
	{
		.wr_nbits = 0x18,
		.rd_nbits = 0x10,
		.lf_nbits = 0x08,
		.flags = 0x00,
		.reg_val = 0x00980806,
	},
	/* RFIC */
	{
		.wr_nbits = 0x12,
		.rd_nbits = 0x00,
		.lf_nbits = 0x04,
		.flags = 0x02,
		.reg_val = 0x00920806,
	}
#endif /* MANGO_OLD_SPI */
};

static int mango_attach(struct scm80211_radio *radio, struct scm80211_phy *phy)
{
	struct mango *mango = radio->priv;
	struct spi_device *spi;
	struct spi_board_info *info = mango_spi_info[phy->chain];
	int err, i;
	struct rf_gspi_cfg *gspi = mango_gspi_cfg;

	phy->rfi.spi.n = 2;
	/*
	 * Populate ad9963 and max2829 and bind them to
	 * repspective rfi spi channels.
	 */
	for (i = 0; i < 2; i++, info++, gspi++) {
		int path = (info->bus_num >> 8) & 0xff;

		spi = phy->ops->connect_spi(phy, info, gspi);
		if (!spi) {
			pr_err("failed to allocate ad9963-%c", 'A' + path);
			goto fail;
		}
		if (!strcmp(info->modalias, "ad9963"))
			mango->ad9963.spi = spi;
		else
			mango->max2829.spi = spi;
	}

	return 0;

fail:
	for (i = 0; i < 2; i++) {
		if (mango->ad9963.spi)
			spi_unregister_device(mango->ad9963.spi);
		if (mango->max2829.spi)
			spi_unregister_device(mango->max2829.spi);
	}
	return err;
}

static int mango_detach(struct scm80211_radio *radio)
{
	struct mango *mango = radio->priv;

	spi_unregister_device(mango->ad9963.spi);
	spi_unregister_device(mango->max2829.spi);
	return 0;
}

static struct scm80211_radio_ops mango_ops = {
	.attach = mango_attach,
	.detach = mango_detach,
	.power_up = mango_power_up,
	.power_down = mango_power_down,
	.set_channel = mango_set_channel,
	.get_cal_data = mango_get_cal_data,
	.set_rtx_lut = mango_set_rtx_lut,
};

static int mango_probe(struct platform_device *pdev)
{
	struct scm80211_rf_device *rf = to_rf_device(pdev);
	struct ieee80211_hw_priv *priv = rf->hw->priv;
	struct mango *mango;
	int i;

	dev_info(&pdev->dev, "probed (board id=%d)\n", rf->board_id);

	rf->n_chain = 2;
	rf->radios = devm_kzalloc(&pdev->dev, sizeof(struct scm80211_radio) * 2, GFP_KERNEL);

	for (i = 0; i < 2; i++) {
		mango = devm_kzalloc(&pdev->dev, sizeof(*mango), GFP_KERNEL);
		if (mango == NULL)
			return -ENOMEM;
		rf->radios[i].ops = &mango_ops;
		rf->radios[i].priv = mango;
		rf->radios[i].devname = pdev->name;
		rf->radios[i].rf_lut = &mango_rf_lut;
		mango->radio = &rf->radios[i];
		mango->id = rf->board_id;
	}
	priv->rf = rf;

#ifdef TEST_MANGO_RF
	scm80211_mac_writel(priv->hw, 0x0df60df6, MAC_REG_HE_TB_PWR01);
	scm80211_mac_writel(priv->hw, 0x0df60df6, MAC_REG_HE_TB_PWR23);
	scm80211_mac_writel(priv->hw, 0x0df60df6, MAC_REG_HE_TB_PWR45);
	scm80211_mac_writel(priv->hw, 0x0df60df6, MAC_REG_HE_TB_PWR67);
	scm80211_mac_writel(priv->hw, 0x0df60df6, MAC_REG_HE_TB_PWR89);
	scm80211_mac_writel(priv->hw, 0x0df60df6, MAC_REG_HE_TB_PWR1011);
#endif

	scm80211_notify_rf_ready(rf);
	return 0;
}

static int mango_remove(struct platform_device *pdev)
{
	dev_info(&pdev->dev, "removed\n");
	return 0;
}

struct scm80211_rf_driver mango_driver = {
	.platform_driver = {
			    .driver = {
				       .name = "mango",
				       .suppress_bind_attrs = true,
				       },
			    .probe = mango_probe,
			    .remove = mango_remove,
			     },
};

static int mango_init(void)
{
	int err;

	pr_debug("mango init\n");

	/* Register ad9963 and max2829 driver */
	err = spi_register_driver(&ad9963_driver);
	if (err)
		return err;

	err = spi_register_driver(&max2829_driver);
	if (err)
		return err;

	err = platform_driver_register(&mango_driver.platform_driver);
	if (err) {
		pr_err("failed to register mango driver\n");
		return err;
	}

	return 0;
}

static void mango_exit(void)
{
	pr_debug("mango exit\n");

	spi_unregister_driver(&ad9963_driver);
	spi_unregister_driver(&max2829_driver);

	platform_driver_unregister(&mango_driver.platform_driver);
}

RF_MODULE(mango);
