// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#include <linux/device.h>
#include <net/mac80211.h>
#include "core.h"
#include "phy.h"
#include "rf.h"

static char *rf_board = "scm2020-1";
module_param_named(rf, rf_board, charp, 0444);
MODULE_PARM_DESC(rf, "RF board <board>[-id]");

static char *rf_board_dbdc_2 = "scm2020_2-2";
module_param_named(rf2, rf_board_dbdc_2, charp, 0444);
MODULE_PARM_DESC(rf2, "DBDC RF board <board>[-id]");

static const char *seps = "-.#";

void scm80211_wait_for_rf_ready(struct scm80211_rf_device *rf)
{
	wait_for_completion(&rf->completion);
}

void scm80211_notify_rf_ready(struct scm80211_rf_device *rf)
{
	complete(&rf->completion);
}

static void rf_device_release(struct device *dev)
{
	/* Should I free all devm_xxx? */
}

struct scm80211_rf_device *scm80211_alloc_rf_device(struct ieee80211_hw *hw, const char *name,
						    int board_id)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_rf_device *rf;
	struct device *dev = wiphy_dev(hw->wiphy);
	int ret;

	rf = devm_kzalloc(dev, sizeof(*rf), GFP_KERNEL);
	if (!rf)
		return NULL;

	if (!name) {
		char *ptr;

		if (board_id == 0) {
			name = devm_kstrdup(dev, rf_board, GFP_KERNEL);
		} else {
			name = devm_kstrdup(dev, rf_board_dbdc_2, GFP_KERNEL);
		}

		/* Get the name and boardid from module parameter */
		ptr = strpbrk((const char *)name, seps);
		if (ptr) {
			ret = kstrtoint(ptr + 1, 10, &board_id);
			if (ret) {
				wiphy_warn(hw->wiphy, "invalid RF name: %s\n", name);
				board_id = 1;
			}
			*ptr = '\0';
		}
		wiphy_dbg(hw->wiphy, "%s: board_id %d, name=%s\n", __func__, board_id, name);

		rf->pdev.name = name;
	} else {
		rf->pdev.name = name;
	}

	wiphy_dbg(hw->wiphy, "allocating an rf device '%s' board id=%d\n", name, board_id);

	rf->pdev.id = priv->idx;
	rf->pdev.dev.release = rf_device_release;
	rf->board_id = board_id;
	rf->hw = hw;
	init_completion(&rf->completion);
	return rf;
}

#define radio_is_sane(x) true

/**
 * phy_attach_rf() - attach RF
 *
 */
int phy_attach_rf(struct ieee80211_hw *hw, struct scm80211_rf_device *rf)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy;
	struct scm80211_radio *radio;
	int ret;

	wiphy_dbg(hw->wiphy, "attaching RF '%s' (board number=%d\n", rf->pdev.name, rf->board_id);

#ifdef SCM2020_A0
	INIT_WORK(&priv->power_detect_work, scm2020_power_detect_work);
#endif

	ret = platform_device_register(&rf->pdev);
	if (ret) {
		wiphy_err(hw->wiphy, "failed to register rf device (%d)\n", ret);
		return ret;
	}

	scm80211_wait_for_rf_ready(rf);

	/* Wait until rf driver is probed */
	if (priv->rf != rf) {
		wiphy_err(hw->wiphy, "failed to find rf driver for '%s'\n",
			  dev_name(&rf->pdev.dev));
		return -ENODEV;
	}

	/* RF device is successfully registered and bound (probed) */

	for_each_phy(phy, hw) {
		radio = &rf->radios[phy->chain];

		if (!radio_is_sane(radio)) {
			wiphy_err(hw->wiphy, "%s has incomplete radion", dev_name(&rf->pdev.dev));
			goto fail;
		}

		ret = radio->ops->attach(radio, phy);
		if (ret) {
			wiphy_err(hw->wiphy, "failed to attach %s\n", dev_name(&rf->pdev.dev));
			priv->chains &= ~BIT(phy->chain);
			priv->n_chain--;
			continue;
		}

		ret = phy->ops->attach(phy, radio);
		if (ret) {
			wiphy_err(hw->wiphy, "failed to attach %s\n", dev_name(&rf->pdev.dev));
			priv->chains &= BIT(phy->chain);
			priv->n_chain--;
			continue;
		}
		phy->radio = radio;
		radio->phy = phy;
		wiphy_err(hw->wiphy, "%s: devid=%d phy->chain=%d phy->radio=%p devname=%s %d\n",
			  __func__, priv->devid, phy->chain, phy->radio, phy->radio->devname,
			  __LINE__);
	}

	/* FIXME - check if at least single phy has radio */
	if (priv->n_chain == 0)
		goto fail;

	wiphy_info(hw->wiphy, "successfully attached RF %s\n", dev_name(&rf->pdev.dev));

	priv->rf = rf;
	return 0;

fail:
	for_each_phy(phy, hw) {
		radio = &rf->radios[phy->chain];
		radio->ops->detach(radio);
		phy->ops->detach(phy, radio);
		radio->phy = NULL;
		phy->radio = NULL;
	}
	return ret;
}

void scm80211_detach_rf(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy;

	if (!priv->rf)
		return;

	wiphy_dbg(hw->wiphy, "detaching RF\n");

#ifdef SCM2020_A0
	cancel_work_sync(&priv->power_detect_work);
#endif

	for_each_phy(phy, hw) {
		struct scm80211_radio *radio = phy->radio;

		if (!radio || !radio->ops || !radio->ops->detach)
			continue;

		radio->ops->detach(radio);
		phy->ops->detach(phy, radio);
		radio->phy = NULL;
		phy->radio = NULL;
	}
	platform_device_unregister(&priv->rf->pdev);
	priv->rf = NULL;
}

int scm80211_init_rf_modules(void)
{
	struct scm80211_rf_module *rf;
	struct scm80211_rf_module *start = &rf_module_start;
	struct scm80211_rf_module *end = &rf_module_end;
	int num = 0;

	for (rf = start; rf < end; rf++) {
		pr_info("scm80211: loading %s\n", rf->name);
		if (!rf->init())
			num++;
	}
	return num > 0 ? 0 : -ENODEV;
}

void scm80211_exit_rf_modules(void)
{
	struct scm80211_rf_module *rf;
	struct scm80211_rf_module *start = &rf_module_start;
	struct scm80211_rf_module *end = &rf_module_end;

	for (rf = start; rf < end; rf++)
		rf->exit();
}

#ifdef TODO_TO_REMOVE
struct scm80211_rf_module *phy_find_rf_module(struct ieee80211_hw *hw, char *name)
{
	struct scm80211_rf_module *module;
	char *ptr;
	int id = -1;
	int ret;

	if (!name)
		name = rf_board;

	wiphy_dbg(hw->wiphy, "searching for RF '%s'\n", name);

	/* Break name into board model and board number */
	ptr = strpbrk((const char *)name, seps);
	if (ptr) {
		ret = kstrtoint(ptr + 1, 10, &id);
		if (ret) {
			wiphy_warn(hw->wiphy, "invalid RF name: %s\n", name);
			goto fail;
		}
	}

	for (module = &rf_module_start; module < &rf_module_end; module++) {
		if (strstr(name, module->name)) {
			if (id != -1)
				rf->id = id;
			wiphy_dbg(hw->wiphy, "RF: %s found\n", module->name);
			return rf;
		}
	}
fail:
	wiphy_err(hw->wiphy, "could not find rf %s\n", name);

	return NULL;
}
#endif
