// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#include <linux/compiler.h>
#include <linux/module.h>
#include <linux/bitfield.h>

#define func_fmt(str)		"[RX] "str
#define func_trace              RX

#include "hw.h"
#include "core.h"
#include "util.h"
#include "trace.h"
#include "debug.h"
//#include "usb.h"

/**
 * DOC: RX
 *
 * H/W RX ring consists of 256 rx descriptors (rxd's). Each rxd contains the
 * address and the size of the data buffer to store a received frame. The
 * buffer size is 3895 + overhead (28 + 3), which is long enough to store the
 * entire MPDU because STA will advertize the maximum mpdu size to be 3895 in
 * VHT capabilities.
 *
 * The rxd buffer is dma-mapped and the rx descriptor is programed with its dma
 * address and length.
 *
 * The rx interrupt service routine performs the following over rxds which H/W
 * completes RX DMA:
 * (1) Detaches skb from rxd
 * (2) DMA-unmaps the skb, and process it (peels off the rxvector and rx header,
 *    and forwards the skb to mac80211.
 * (3) Allocate a new skb and attach it to rxd for further rx operation.
 *
 * @order: 2^@order is the number of the ring elements
 * @bptr: buffer pointer
 * @rptr: read pointer
 * @wtpr: write pointer
 *
 * Pointer comparison is done using @order + 1 bits of linear pointer to
 * distinguish between full and empty states, whereas ring indexing is done
 * using @order bits.
 *
 * TODO:
 * - Extend the RX ring management such that one skb can span over multiple
 *   rxd's. If the requested size is too big (> 4KB), skb_alloc() may return
 *   a pskb. In this case, the driver needs to map each frags to different
 *   rxd's. In additions the interrupt service routine should be able to
 *   deal with such a skb. This is left for futher extension.
 */
#define MAC_RXE_OFFSET       0x9400
#define REG_RXE_MPDU_INFO    0x40
#define MAC_REG_RXE_MPDU_INFO   (MAC_RXE_OFFSET + REG_RXE_MPDU_INFO)

int ring_order = 8;		/* 2^8=256 ring entries */
module_param(ring_order, int, 0444);
MODULE_PARM_DESC(ring_order, "Number of rx ring (log2)");

/**
 * The number of the RX chain(NSS)
 */
int n_rx = 2;
module_param(n_rx, int, 0444);
MODULE_PARM_DESC(n_rx, "Number of rx chain");

/**
 * rx packet dump control, turned off as default
 **/
int rx_dump;
module_param(rx_dump, int, 0644);
MODULE_PARM_DESC(rx_dump, "RX packet dump switch");

/**
 * IEEE802.11 Header info dump
 **/
int header_info_dump;
module_param(header_info_dump, int, 0644);
MODULE_PARM_DESC(header_info_dump, "IEEE802.11 Header info dump");

static int n_nofull;
module_param(n_nofull, int, 0444);
MODULE_PARM_DESC(n_nofull, "counter for rxring no fully populated");

/* must 4-bytes alignment */
#ifdef SCM80211_USE_4K_MPDU

#ifdef SCM80211_SHRINK_MEM
	#undef IEEE80211_MAX_MPDU_LEN_VHT_3895
	#define IEEE80211_MAX_MPDU_LEN_VHT_3895 1600
#endif

const size_t rx_buflen = ALIGN((sizeof(struct rxvector) +
				sizeof(struct rx_mpdu_hdr) + IEEE80211_MAX_MPDU_LEN_VHT_3895), 16);
#else
const size_t rx_buflen = ALIGN((sizeof(struct rxvector) +
				sizeof(struct rx_mpdu_hdr) + IEEE80211_MAX_MPDU_LEN_VHT_7991), 16);
#endif

void scm80211_enable_rx_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

#ifdef TEST_INTR_RX_BUF
	priv->int_en |= INTR_ENABLE_RX_BUF;
#else
	priv->int_en |= INTR_ENABLE_RX_DONE;
#endif
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

void scm80211_disable_rx_irq(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

#ifdef TEST_INTR_RX_BUF
	priv->int_en &= ~INTR_ENABLE_RX_BUF;
#else
	priv->int_en &= ~INTR_ENABLE_RX_DONE;
#endif

	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

struct sk_buff *scm80211_alloc_rx_skb(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct sk_buff *skb;

	skb = skb_dequeue(&priv->rxring.pool);
#ifndef CONFIG_ALLWINNER_USB
	if (!skb) {
		gfp_t gfp = in_interrupt() ? GFP_ATOMIC : GFP_KERNEL;

		skb = alloc_skb(rx_buflen + IEEE80211_RX_HEADROOM, gfp);
	}
#endif
	return skb;
}

void scm80211_refill_rxpool(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct sk_buff *skb;
	int n = skb_queue_len(&priv->rxring.pool);
	gfp_t gfp = in_interrupt() ? GFP_ATOMIC : GFP_KERNEL;

again:
	/* Populate pool */
	n = (int)(ring_size(&priv->rxring)) - n;

	while (n-- > 0) {
		skb = alloc_skb(rx_buflen + IEEE80211_RX_HEADROOM, gfp);
		if (!skb)
			break;

		skb_queue_tail(&priv->rxring.pool, skb);
	}

	n = skb_queue_len(&priv->rxring.pool);
	if (n == 0) {
		scm80211_info(hw, "no skbs in pool\n");
		/* try again ? */
		goto again;
	} else if (n < ring_size(&priv->rxring)) {
		if (n_nofull++ % 1024 == 0)
			scm80211_info(hw, "rx pool is not fully populated\n");
		/* to prepare more */
		n = 0;
		goto again;
	}
}

/**
 * scm80211_rebuild_fragment_frame()
 * link skbs in the list to the head, and only report head skb to uplayer
 */
static struct sk_buff *scm80211_rebuild_fragment_frame(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct sk_buff *head = skb_peek(&priv->rxring.frags);
	struct sk_buff **frag_tail, *skb;

	if (WARN_ON(!head))
		return NULL;

	skb_unlink(head, &priv->rxring.frags);

	while ((skb = skb_dequeue(&priv->rxring.frags))) {
		frag_tail = &skb_shinfo(head)->frag_list;
		while (*frag_tail)
			frag_tail = &(*frag_tail)->next;
		head->len += skb->len;
		head->data_len += skb->len;
		*frag_tail = skb;
	}
	return head;
}

/**
 * scm80211_rx_start() populate @scm80211_rx_info and do sanity for rx frame
 */
static enum scm80211_rx_result scm80211_rx_start(struct scm80211_rx_data *rx,
						 struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct scm80211_rx_info *info = SCM80211_SKB_RXCB(skb);
	struct ieee80211_hw_priv *priv = hw->priv;

	memset(rx, 0, sizeof(*rx));

	/*
	 * populate rx_info param
	 */
	if (scm80211_bus_rx_start(hw, skb, rx))
		goto drop;

	/*
	 * Sanity check RXD,
	 * - To detect and discard the case where rxd.dlen > rxd.blen
	 * - For rx data which is fragmented we need to be re-construct the non-liner skb
	 *   after collecting all START,MIDDLE,END frames.
	 *   frags<->start<->middle<->...<->middle<->end
	 */
	switch (info->rxd.frag) {
	case SIGNAL_FRAME:
		if (info->rxd.dlen < (sizeof(struct rxvector) +
				      sizeof(struct rx_mpdu_hdr) + sizeof(struct ieee80211_cts))) {
			scm80211_warn(hw, "rxd[%02x]: rxd truncated? (%d)\n",
				      info->index, info->rxd.dlen);

			print_hex_dump_bytes("rx", DUMP_PREFIX_OFFSET, skb->data, info->rxd.dlen);
			goto drop;
		}
		break;
	case START_FRAME:
		scm80211_trace(hw, "rxd[%02x]: start fragmented\n", info->index);
		if (WARN_ON(!skb_queue_empty(&priv->rxring.frags)))
			skb_queue_purge(&priv->rxring.frags);
		skb_queue_tail(&priv->rxring.frags, skb);
		goto again;
	case MIDDLE_FRAME:
		scm80211_trace(hw, "rxd[%02x]: middle fragmented\n", info->index);
		if (WARN_ON(skb_queue_empty(&priv->rxring.frags)))
			goto drop;
		skb_queue_tail(&priv->rxring.frags, skb);
		goto again;
	case LAST_FRAME:
		{
			struct sk_buff *head;

			scm80211_trace(hw, "rxd[%02x]: end fragmented\n", info->index);
			if (WARN_ON(skb_queue_empty(&priv->rxring.frags)))
				goto drop;
			skb_queue_tail(&priv->rxring.frags, skb);
			head = scm80211_rebuild_fragment_frame(hw);
			if (!head)
				goto drop;
			else
				skb = head;
			break;
		}
	default:
		goto drop;
	}

	skb_queue_head_init(&rx->queue);
	skb_queue_tail(&rx->queue, skb);

	rx->hw = hw;
	rx->skb = skb;
	rx->rxs = IEEE80211_SKB_RXCB(skb);
	rx->rv = (void *)skb->data;
	rx->rh = (void *)skb_pull(skb, sizeof(*rx->rv));
	rx->hdr = (void *)skb_pull(skb, sizeof(*rx->rh));
	rx->result = RX_CONTINUE;

	/* Sanity check RXH */
	if (rx->rh->mpdu_len < sizeof(struct ieee80211_cts)) {
		scm80211_warn(rx->hw, "rxd[%02x]: len=0\n", info->index);
		goto drop;
	}

	memcpy(&rx->info, info, sizeof(*info));

	return RX_CONTINUE;

drop:
	return RX_DROP;
again:
	return RX_AGAIN;
}

static enum scm80211_rx_result scm80211_rxh_dump_rxb(struct scm80211_rx_data *rx)
{
	struct rxvector *rv = rx->rv;
	struct rx_mpdu_hdr *rh = rx->rh;
	struct ieee80211_hdr *hdr = rx->hdr;
	struct sk_buff *skb = rx->skb;
	char str[128];

	if (header_info_dump) {
		if (hdr)
			scm80211_snprint_frame(str, sizeof(str), hdr);
		else
			sprintf(str, "?");

		scm80211_info(rx->hw, "rx: %s\n", str);
	}

	if (rx_dump) {
		if (rv)
			print_hex_dump_bytes("rxv", DUMP_PREFIX_OFFSET, rv, sizeof(*rv));

		if (rh)
			print_hex_dump_bytes("rxh", DUMP_PREFIX_OFFSET, rh, sizeof(*rh));

		print_hex_dump_bytes("rxd", DUMP_PREFIX_OFFSET, hdr, min_t(int, 64, skb->len));
	}
	return RX_CONTINUE;
}

/**
 * scm80211_rxh_status() - populates &ieee80211_rx_info
 */
static enum scm80211_rx_result scm80211_rxh_status(struct scm80211_rx_data *rx)
{
	struct ieee80211_hw *hw = rx->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_rx_status *rxs = rx->rxs;
	struct rxvector *rv = rx->rv;
	struct rx_mpdu_hdr *rh = rx->rh;
	struct ieee80211_channel *chan;

	memset(rxs, 0, sizeof(*rxs));

	chan = hw->conf.chandef.chan;
	if (unlikely(!chan)) {
		scm80211_warn(hw, "%s channel invalid !!\n", __func__);
		return RX_MONITOR_ONLY;
	}

	rxs->freq = chan->center_freq;
	rxs->band = chan->band;
	rxs->signal = rv->RSSI;

	switch (rv->CH_BANDWIDTH) {
	case CBW80:
	case CBW80S20P:
	case CBW80S40P:
		rxs->bw = RATE_INFO_BW_80;
		break;
	case CBW40:
		rxs->bw = RATE_INFO_BW_40;
		break;
	case CBW20:
		rxs->bw = RATE_INFO_BW_20;
		break;
	case CBW160:
	case CBW160S20P:
	case CBW160S40P:
		rxs->bw = RATE_INFO_BW_160;
		break;
	default:
		break;
	}

	if (ppdu_is_he(rv)) {
		rxs->encoding = RX_ENC_HE;
		rxs->rate_idx = rv->MCS;
		rxs->nss = (rv->NUM_STS + 1) >> rv->STBC;
		rxs->he_gi = rv->GI_TYPE - 1;
		rxs->he_dcm = rv->DCM;

		if (ppdu_is_he_er_su(rv)) {
			rxs->bw = ((rv->CH_BANDWIDTH == ER_RU_242) ?
				   RATE_INFO_BW_20 : RATE_INFO_BW_10);
		}

	} else if (ppdu_is_vht(rv)) {
		rxs->encoding = RX_ENC_VHT;
		rxs->rate_idx = rv->MCS;
		rxs->nss = (rv->NUM_STS + 1) >> rv->STBC;
		if (rv->GI_TYPE == SHORT_GI)
			rxs->enc_flags |= RX_ENC_FLAG_SHORT_GI;

	} else if (ppdu_is_ht(rv)) {
		rxs->encoding = RX_ENC_HT;
		rxs->rate_idx = rv->MCS;
		rxs->nss = rv->NUM_STS + 1 - rv->STBC;
		if (rv->GI_TYPE == SHORT_GI)
			rxs->enc_flags |= RX_ENC_FLAG_SHORT_GI;
	} else {
		rxs->encoding = RX_ENC_LEGACY;

		rxs->rate_idx = rv->MCS;
		if (rxs->band == NL80211_BAND_5GHZ)
			rxs->rate_idx -= 4;

		if (rxs->rate_idx >= hw->wiphy->bands[rxs->band]->n_bitrates) {
			scm80211_warn(hw, "rxd[%02x] invalid rate %d\n",
				      rx->info.index, rxs->rate_idx);
			rxs->rate_idx = 0;
			return RX_MONITOR_ONLY;
		}

		if (rv->GI_TYPE == SHORT_GI)
			rxs->enc_flags |= RX_ENC_FLAG_SHORT_GI;
		rxs->signal = rv->RSSI_LEGACY;
	}

	if (rv->FEC_CODING == LDPC_CODING)
		rxs->enc_flags |= RX_ENC_FLAG_LDPC;

	if (rv->BEAMFORMED)
		rxs->enc_flags |= RX_ENC_FLAG_BF;

	if (rv->STBC)
		rxs->enc_flags |= rxs->nss << RX_ENC_FLAG_STBC_SHIFT;

	if (rh->ampdu) {
		rxs->flag |= RX_FLAG_AMPDU_DETAILS | RX_FLAG_AMPDU_EOF_BIT_KNOWN;

		if (rh->eof)
			rxs->flag |= RX_FLAG_AMPDU_EOF_BIT;

		/*
		 * NB: rh->timestamp seems to be PLCP start time for all
		 * subframes in an A-MPDU have the same value. We use
		 * this timestamp as an rxs->ampdu_reference.
		 */
		rxs->ampdu_reference = rh->timestamp;
	}
	/* Linux-5.9 is not ready for RX_ENC_HE */
	// rxs->flag |= RX_FLAG_MACTIME_START;

	rxs->mactime = rh->timestamp;
	rxs->device_timestamp = rh->timestamp;
#ifdef ERRATA_SW_WEP
	/* softap mode , wep cipher gtk will decrypt by sw */
	if (rh->err_key && !rh->decrypted)
		return RX_CONTINUE;
#endif
	if (rh->err_fcs || !rh->ok) {
		rxs->flag |= RX_FLAG_FAILED_FCS_CRC;
		priv->stats.dot11FCSErrorCount++;
		return RX_MONITOR_ONLY;
	}

	return RX_CONTINUE;
}

/**
 * scm80211_mark_rx_interface() -  mark which interface handle this packet
 */
static void scm80211_mark_rx_interface(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	struct scm80211_rx_data *rx = data;
	struct ieee80211_hdr *hdr = rx->hdr;

	if (ether_addr_equal(hdr->addr1, vif->addr)) {
		if (ieee80211_hdrlen(hdr->frame_control) > sizeof(struct ieee80211_cts)) {
			rx->sta = ieee80211_find_sta(vif, hdr->addr2);
			rx->vif = vif;
		}
		set_bit(scm80211_vdev_id(vif), &rx->interfaces);
		return;
	}

	if (is_multicast_ether_addr(hdr->addr1)) {
		struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
		u8 *bssid = ieee80211_get_bssid(hdr, rx->skb->len, vif->type);

		/*
		 * @vif accepts this frame if (1) BSSID is unknown,
		 * (2) BSSID is broadcast address (probe request), or
		 * (3) BSSID matches @vif->bss_conf.bssid
		 * (4) BSSID filtering is disabled (before assoc or for HW scan)
		 * (5) BSSID matches @vif->bss_conf.transmitter_bssid
		 */
		if (!bssid ||
		    !vifp->bssid_filter ||
		    is_broadcast_ether_addr(bssid) ||
		    ether_addr_equal(vif->bss_conf.bssid, bssid) ||
		    ether_addr_equal(vif->bss_conf.transmitter_bssid, bssid)) {
			set_bit(scm80211_vdev_id(vif), &rx->interfaces);
		}
	}
}

static enum scm80211_rx_result scm80211_rxh_address_filter(struct scm80211_rx_data *rx)
{
	rx->interfaces = 0UL;
	ieee80211_iterate_active_interfaces_atomic(rx->hw,
						   IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_mark_rx_interface, rx);
	if (rx->interfaces == 0UL && !scm80211_is_scanning(rx->hw)) {
		rx->rxs->flag |= RX_FLAG_ONLY_MONITOR;
		return RX_MONITOR_ONLY;
	}

	if (rx->sta) {
		struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(rx->sta);

		sta_priv->rx_packets++;
	}

	return RX_CONTINUE;
}

static enum scm80211_rx_result scm80211_rxh_decrypt(struct scm80211_rx_data *rx)
{
	struct ieee80211_rx_status *rxs = rx->rxs;
	struct rx_mpdu_hdr *rh = rx->rh;
	__le16 fc = rx->hdr->frame_control;

	if (!ieee80211_has_protected(fc))
		return RX_CONTINUE;

	/* H/W does not perform BIP - pass it to mac80211 */
	if (ieee80211_is_mgmt(fc))
		return RX_CONTINUE;

	/*
	 * H/W did not decrypt this frame, either because (1) it could not find
	 * the appropriate transient key or because (2) it is not destined to us.
	 *
	 * For (1) mac80211 will software decrypt this frame. For (2),
	 * scm80211_rxh_addr_filter() would already have set @RX_FLAG_ONLY_MONITOR
	 * and mac80211 will forward this frame only to monitor interfaces.
	 *
	 * We can simply return here.
	 */
	if (!rh->decrypted) {
		if (rx->interfaces)
			scm80211_trace(rx->hw, "software decryption");
		return RX_CONTINUE;
	}

	/*
	 * H/W did decrypt this frame either successfully or with errors.
	 * Frames with decryption error is demoted to monitor interface only
	 * except for ones with TKIP MMIC error.
	 */
	rxs->flag |= RX_FLAG_DECRYPTED;

	switch (rh->cipher) {
	case CIPHER_SUITE_WEP40:
	case CIPHER_SUITE_WEP104:
		rxs->flag |= RX_FLAG_ICV_STRIPPED;
		rxs->flag |= rh->err_icv ? RX_FLAG_ONLY_MONITOR : 0;
		break;
	case CIPHER_SUITE_TKIP:
		/*
		 * We pass MMIC failed frames so that mac80211 performs
		 * countermeasure procedure.
		 */
		rxs->flag |= RX_FLAG_MMIC_STRIPPED;
		rxs->flag |= RX_FLAG_ICV_STRIPPED;
		if (rh->err_mic) {
			rxs->flag |= RX_FLAG_MMIC_ERROR;
			scm80211_warn(rx->hw, "MMIC error detected\n");
		}
		break;
	case CIPHER_SUITE_CCMP128:
	case CIPHER_SUITE_CCMP256:
	case CIPHER_SUITE_GCMP128:
	case CIPHER_SUITE_GCMP256:
		rxs->flag |= RX_FLAG_MIC_STRIPPED;
		if (rh->err_key + rh->err_mic) {
			scm80211_warn(rx->hw,
				      "%s decapsulation error: %08x\n",
				      cipher_suite_str[rh->cipher], rh->hd1);
			rxs->flag |= RX_FLAG_ONLY_MONITOR;
		}
		break;
	default:
		/* Should not come to here */
		WARN_ONCE(1, "unsupported cipher %0x\n", rh->cipher);
		return RX_DROP;
	}

	/*
	 * NB:
	 * Do not clear IEEE80211_FCTL_PROTECTED for mac80211 to do remaining
	 * dirty works.
	 */
	return RX_CONTINUE;
}

extern struct txvector default_txvector;

/**
 * scm80211_sync_tsf()
 * Synchronization the tsf
 */
void scm80211_sync_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		       u64 bss_ts, u32 mac_ts)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	vifp->sw_tsf = bss_ts;
#ifdef SCM80211_TWT
	vifp->twt_ktime = ktime_get();
#endif
	/*
	 * for USB soluction it will be synchronized in fw
	 * use probable latency 2TU for usb
	 */
	if (is_pcie_device(hw)) {
		u64 tsf = scm80211_get_tsf(hw, vif);
		u32 lo, hi;

		lo = lower_32_bits(tsf);
		hi = upper_32_bits(tsf);
		/* calculate the rx beacon latency from hwmac to driver */
		vifp->bss.beacon.latency = (lo - mac_ts) & S32_MAX;
		if (mac_ts > lo)
			hi--;
		tsf = hi;
		scm80211_set_tsf(hw, vif, bss_ts);
		scm80211_trace(hw, "tsf=%llu, old=%llu, new=%llu latency = %u\n",
			       scm80211_get_tsf(hw, vif),
			       ((tsf << 32) | mac_ts),
			       bss_ts, vifp->bss.beacon.latency);
	} else {
		vifp->bss.beacon.latency = ieee80211_tu_to_usec(2);
	}
}

static void scm80211_rx_parse_rssi(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				   struct ieee80211_rx_status *rxs, bool force)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	unsigned long delay;
	u32 reserved = 4000;  /* reserved 4ms before TBTT to cleanup tx */

	/* add more latency for hwmac progress or fw upload */
	reserved += vifp->bss.beacon.latency;

	delay = usecs_to_jiffies(ieee80211_tu_to_usec(vif->bss_conf.beacon_int) * 2 - reserved);
	if (vifp->wait_beacon)
		scm80211_wake_queues(hw, scm80211_vif_queues(vif));

	if (delayed_work_pending(&vifp->suspend_tx_work))
		cancel_delayed_work(&vifp->suspend_tx_work);

	force |= vifp->wait_beacon;

	vifp->wait_beacon = false;

	/*
	 * for 2.4GHZ rssi signal under -82dbm beacon loss will happen while
	 * doing tx throughput test. try to stop tx during next next TBTT
	 */
	if (force || (rxs->signal) < (s8)(-78)) {
		scm80211_trace(hw, " rssi %d %sschedule suspend work %dms later\n",
			       rxs->signal, force ? "force " : "",
			       jiffies_to_msecs(delay));
		ieee80211_queue_delayed_work(hw, &vifp->suspend_tx_work, delay);
	}
}

static int scm80211_calculate_tbtt_delta(u32 tsf, u32 last, int bi)
{
	u32 tu, next;

	if (last == 0 || bi == 0)
		return -2;

	tu   = ieee80211_tsf_to_tu(tsf);
	next = ieee80211_tsf_to_tu(last);

	do {
		next += bi;
	} while (2 * abs(next - tu) > bi);

	return next - tu;
}

/**
 * scm80211_track_beacon_transmit_time() - track nominal beacon transmit time
 *
 * The nominal TBTTs form an arithmetic sequence represented by
 * TBTT(n) = BI * 1024 * n + tsf0. This function tries to deduce
 * tsf0 by finding the consecutive beacons whose TSFs are BI * 1024
 * apart and taking the modulo (BI * 1024) of the observed TBTT.
 *
 * We calculate tsf0 in the unit of TU (=1024us).
 */
static bool scm80211_track_beacon_transmit_time(struct ieee80211_hw *hw,
						struct ieee80211_vif *vif,
						u32 tsf)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_bss_conf *bss = &vif->bss_conf;
	u32 tsf0;
	int delta;
	bool sync = false;

	delta = scm80211_calculate_tbtt_delta(tsf, vifp->bss.beacon.last, bss->beacon_int);

	if (vifp->bss.beacon.tsf0 == -1U) {
		if (delta == 0) {
			tsf0 = tsf % ieee80211_tu_to_tsf(bss->beacon_int);
			tsf0 = round_down(tsf0, ieee80211_tu_to_tsf(1));
			vifp->bss.beacon.tsf0 = tsf0;
			vifp->bss.beacon.last = round_down(tsf, ieee80211_tu_to_tsf(1));
			vifp->bss.beacon.next = vifp->bss.beacon.last;
			scm80211_trace(hw, "beacon: tbtt(n) = %u*n + %u\n",
				       (u32) ieee80211_tu_to_tsf(bss->beacon_int),
				       vifp->bss.beacon.tsf0);
			sync = true;
		} else {
			scm80211_trace(hw, "beacon %08u out of sync\n", tsf);
			vifp->bss.beacon.last = tsf;
			vifp->bss.beacon.next = vifp->bss.beacon.last;
		}
	} else if (abs(delta) < 4) {
		tsf += ieee80211_tu_to_tsf(delta);
		tsf0 = tsf % ieee80211_tu_to_tsf(bss->beacon_int);
		tsf0 = round_down(tsf0, ieee80211_tu_to_tsf(1));
		if (tsf0 != vifp->bss.beacon.tsf0) {
			scm80211_trace(hw, "beacon: tbtt(n) = %u*n + %u (overflow)\n",
				       (u32) ieee80211_tu_to_tsf(bss->beacon_int),
				       vifp->bss.beacon.tsf0);
			vifp->bss.beacon.tsf0 = tsf0;
		}
		/* nominalize beacon tsf */
		vifp->bss.beacon.last = round_down(tsf, ieee80211_tu_to_tsf(1));
		scm80211_trace(hw, "beacon: tsf=%08u nominal=%08u\n",
			       tsf, vifp->bss.beacon.last);

		sync = true;
	} else {
		scm80211_warn(hw, "beacon: %08u out of sync last %08u delta=%d",
			      tsf, vifp->bss.beacon.last, delta);
		vifp->bss.beacon.tsf0 = -1U;
		vifp->bss.beacon.last = tsf;
	}

	return sync;
}

static void scm80211_handle_setup_twt(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				      u64 ts)
{
#ifdef SCM80211_TWT
	struct ieee80211_vif_priv	*vifp = ieee80211_to_priv(vif);
	enum scm80211_twt_state		state;
	unsigned long			flag;

	state = scm80211_get_twt_state(vif);
	/* TWT will be rewrite later */
	if (state != TWT_DISABLED) {
		spin_lock_irqsave(&vifp->lock, flag);

		if (state != TWT_REQ || !vifp->twt_request) {
			spin_unlock_irqrestore(&vifp->lock, flag);
			return;
		}

		/* Set TWT as ts + 100ms */
		vifp->twt	= ts + USEC_PER_MSEC * 100;
		put_unaligned_le64(vifp->twt, vifp->twt_request->data + vifp->twt_off);
		print_hex_dump_bytes("TWT", DUMP_PREFIX_OFFSET, vifp->twt_request->data,
				     vifp->twt_request->len);
		scm80211_set_twt_state(vif, TWT_NEGO);
		spin_unlock_irqrestore(&vifp->lock, flag);
	}
#endif
}
/**
 * scm80211_rx_beacon() - pre-handle beacon
 */
static void scm80211_rx_beacon(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			       void *data)
{
	struct scm80211_rx_data    *rx = data;
	struct rxvector		   *rv = rx->rv;
	struct ieee80211_vif_priv  *vifp = ieee80211_to_priv(vif);
	struct rx_mpdu_hdr	   *rh  = rx->rh;
	struct ieee80211_mgmt	   *beacon = (struct ieee80211_mgmt *)rx->skb->data;
	struct ieee80211_rx_status *rxs = rx->rxs;
	struct ieee80211_bss_conf  *bss = &vif->bss_conf;
	struct ieee80211_chanctx_conf *chan_conf;
	int miss = 0;
	u64 bss_ts = le64_to_cpu(beacon->u.beacon.timestamp);
	u32 tsf32  = lower_32_bits(bss_ts);

	scm80211_sync_tsf(hw, vif, bss_ts, le32_to_cpu(rh->timestamp));

	chan_conf = get_ieee80211_chanctx_conf(vif);
	if (vifp->bss.beacon.dur == 0 && chan_conf) {
		u32 dur;

		dur = scm80211_ppdu_txtime((void *)rv, chan_conf->def.chan->band);
		vifp->bss.beacon.dur = dur;
	}

	if (!scm80211_track_beacon_transmit_time(hw, vif, tsf32)) {
		vifp->bss.beacon.next = vifp->bss.beacon.last +
					ieee80211_tu_to_tsf(bss->beacon_int);
		return;
	}

	if (vifp->bss.beacon.last > vifp->bss.beacon.next) {
		miss = DIV_ROUND_UP(vifp->bss.beacon.last - vifp->bss.beacon.next,
				    ieee80211_tu_to_tsf(1));
		miss /= bss->beacon_int;
		if (miss > 0)
			scm80211_trace(hw, "missed %d beacons from %u\n",
				       miss, ieee80211_tsf_to_tu(vifp->bss.beacon.next));
	}

	scm80211_calc_next_beacon(hw, vif, rx->skb);

	if (rxs->band == NL80211_BAND_2GHZ)
		scm80211_rx_parse_rssi(hw, vif, rxs, miss > 3);

	scm80211_handle_setup_twt(hw, vif, bss_ts);
}

static enum scm80211_rx_result scm80211_rx_mgnt(void *data, struct ieee80211_vif *vif,
						struct ieee80211_sta *sta)
{
	struct scm80211_rx_data *rx = data;
	struct sk_buff *skb = rx->skb;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = vifp->hw;
	struct ieee80211_bss_conf *info = &vif->bss_conf;
	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *)skb->data;
	struct ieee802_11_elems elems;
	u8 *start = NULL;
	bool is_action = false;
	u16 fc;
	size_t baselen;

	memset(&elems, 0, sizeof(struct ieee802_11_elems));
	fc = le16_to_cpu(mgmt->frame_control);

	if (!scm80211_rx_parse_needed(mgmt->bssid, sta, info))
		goto exit;

	switch (fc & IEEE80211_FCTL_STYPE) {
	case IEEE80211_STYPE_BEACON:
		scm80211_rx_beacon(hw, vif, data);
		break;
	case IEEE80211_STYPE_PROBE_RESP:
		start = (u8 *)mgmt->u.probe_resp.variable;
		break;
	case IEEE80211_STYPE_ACTION:
		if (mgmt->u.action.category == WLAN_CATEGORY_UNPROT_SIG) {
			struct unprot_sig_action *action = (void *)&mgmt->u.action;

			if (action->action_code == WLAN_SIG_TWT_SETUP)
				start = (u8 *)action->variable;
		}
		is_action = true;
		break;
	default:
		goto exit;
	}

	baselen = start - (u8 *)mgmt;
	scm80211_parse_elems(start, skb->len - baselen, is_action, &elems);

	if (elems.he_operation)
		ieee80211_parse_he_operation(hw, vif, elems.he_operation, sta);
#ifdef SCM80211_TWT
	if (elems.twt && elems.twt_len)
		if (ieee80211_parse_twt(hw, vif, elems.twt, elems.twt_len, sta) == 0 &&
		    ieee80211_is_action(mgmt->frame_control)) {
			/**
			 * Set as handled and release this frame for mac80211 will treat this
			 * action frame as invalid frame
			 **/
			return RX_HANDLED;
		}
#endif
exit:
	return RX_CONTINUE;
}

static void scm80211_rx_mulitcast(struct ieee80211_hw *hw, struct ieee80211_vif *vif, bool more)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	if (!is_pcie_device(hw) || !scm80211_ps_enabled(hw) || vifp->pm.doze)
		return;

	if (timer_pending(&vifp->pm.doze_timer)) {
		if (vifp->bss.more_data_used && !more) {
			/* goto doze mode soon */
			mod_timer(&vifp->pm.doze_timer, jiffies + usecs_to_jiffies(2));
		} else {
			if (more)
				vifp->bss.more_data_used = true;
			mod_timer(&vifp->pm.doze_timer, jiffies + msecs_to_jiffies(2));
		}
	}
}

/**
 * scm80211_rx_handle_packet() - pre parse same ies in the frame
 *
 * for some advanced ies. mac80211 did not parse and handle it
 * driver will do pre parse these and handle them by driver self
 */
static void scm80211_rx_handle_packet(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	struct scm80211_rx_data *rx = data;
	struct sk_buff *skb = rx->skb;
	struct ieee80211_rx_status *status = IEEE80211_SKB_RXCB(skb);
	enum nl80211_iftype iftype = ieee80211_vif_type_p2p(vif);
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = vifp->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_bss_conf *info = &vif->bss_conf;
	struct ieee80211_sta *ap;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	u16 fc;

	if (iftype == NL80211_IFTYPE_STATION || iftype == NL80211_IFTYPE_P2P_CLIENT) {
		ap = ieee80211_find_sta(vif, info->bssid);
		if (!ap || !info->assoc)
			return;
		fc = le16_to_cpu(hdr->frame_control);
		switch (fc & IEEE80211_FCTL_FTYPE) {
		case IEEE80211_FTYPE_CTL:
			/* TODO */
			break;
		case IEEE80211_FTYPE_MGMT:
			rx->result = scm80211_rx_mgnt(data, vif, ap);
			break;
		case IEEE80211_FTYPE_DATA:
			if (is_multicast_ether_addr(hdr->addr1))
				scm80211_rx_mulitcast(hw, vif,
						      ieee80211_has_moredata(hdr->frame_control));
			else if (priv->pm.vif)
				mod_timer(&priv->pm.timer, jiffies +
					  msecs_to_jiffies(priv->pm.dynamic_ps_timeout));
			/* TODO */
			break;
		}
	}

	if (iftype == NL80211_IFTYPE_AP) {
		fc = le16_to_cpu(hdr->frame_control);
		if (ieee80211_is_beacon(fc))
			cfg80211_report_obss_beacon(hw->wiphy, skb->data, skb->len,
						    status->freq, status->signal);
	}
}

static enum scm80211_rx_result scm80211_rxh_prehandle(struct scm80211_rx_data *rx)
{
	ieee80211_iterate_active_interfaces_atomic(rx->hw, IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_rx_handle_packet, rx);
	return rx->result;
}

/**
 * scm80211_rxh_htc() - parse htc
 */
static enum scm80211_rx_result scm80211_rxh_htc(struct scm80211_rx_data *rx)
{
	struct ieee80211_hw *hw = rx->hw;
	__le32 htc;
	int offset;
	struct ieee80211_tx_rate rate;
	bool stbc_or_dcm;

	if (!rx->sta)
		return RX_CONTINUE;

	if (!ieee80211_has_order(rx->hdr->frame_control))
		return RX_CONTINUE;

	offset = ieee80211_hdrlen(rx->hdr->frame_control);
	offset -= IEEE80211_HT_CTL_LEN;

	/* FIXME: how to deal with this htc */
	htc = get_unaligned_le32(rx->skb->data + offset);

	scm80211_trace(hw, "%s htc: %08x\n", __func__, htc);

	memset(&rate, 0, sizeof(rate));

	if (!ieee80211_link_adap_to_rate(hw, htc, &rate, &stbc_or_dcm)) {
		struct ieee80211_sta_priv *priv = ieee80211_to_priv(rx->sta);
		unsigned long flag;

		spin_lock_irqsave(&priv->lock, flag);
		priv->rate_available = true;
		priv->stbc_or_dcm = stbc_or_dcm;
		priv->recomd_rate = rate;
		spin_unlock_irqrestore(&priv->lock, flag);
	} else {
		scm80211_process_omi(rx->hw, rx->vif, rx->sta, htc, true);
	}

	/* For mgmt frame remove htc field because ieee80211_mgmt has no htc filed */
	if (ieee80211_is_mgmt(rx->hdr->frame_control)) {
		u8 hdr[IEEE80211_MAX_HDR_LEN] = { 0, };

		rx->hdr->frame_control &= ~cpu_to_le16(IEEE80211_FCTL_ORDER);
		memcpy(hdr, rx->skb->data, offset);
		memcpy(skb_pull(rx->skb, IEEE80211_HT_CTL_LEN), hdr, offset);
	}

	return RX_CONTINUE;
}

/**
 * scm80211_rxh_he_radiotap_header() - adds radiotap headers for HE
 *
 * This function prepends HE radiotap information to @rx->skb. That is,
 * the @rx->skb->data no longer points to &struct ieee80211_hdr. Therefore,
 * this function is recommended to be called at the last stage of rx processing
 * chain.
 */
static enum scm80211_rx_result scm80211_rxh_he_radiotap_header(struct scm80211_rx_data *rx)
{
	struct sk_buff *skb = rx->skb;
	struct ieee80211_rx_status *rxs = rx->rxs;
	struct rxvector *rv = rx->rv;
	struct ieee80211_radiotap_he he = { };
	struct ieee80211_radiotap_he_mu he_mu = { };
	u32 v;

	if (!ppdu_is_he(rv))
		return RX_CONTINUE;

	rxs->flag |= RX_FLAG_RADIOTAP_HE;

#define HE_PREP(f, val)	le16_encode_bits(val, IEEE80211_RADIOTAP_HE_##f)

	he.data1 = (IEEE80211_RADIOTAP_HE_DATA1_BSS_COLOR_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_BEAM_CHANGE_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_UL_DL_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_DATA_MCS_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_DATA_DCM_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_CODING_KNOWN |
		    /* IEEE80211_RADIOTAP_HE_DATA1_LDPC_XSYMSEG_KNOWN   | */
		    IEEE80211_RADIOTAP_HE_DATA1_STBC_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_SPTL_REUSE_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_SPTL_REUSE2_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_SPTL_REUSE3_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_SPTL_REUSE4_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_BW_RU_ALLOC_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA1_DOPPLER_KNOWN);

	he.data2 = (IEEE80211_RADIOTAP_HE_DATA2_PRISEC_80_KNOWN |
		    IEEE80211_RADIOTAP_HE_DATA2_GI_KNOWN |
		    /* IEEE80211_RADIOTAP_HE_DATA2_NUM_LTF_SYMS_KNOWN   | */
		    /* IEEE80211_RADIOTAP_HE_DATA2_PRE_FEC_PAD_KNOWN    | */
		    /* IEEE80211_RADIOTAP_HE_DATA2_TXBF_KNOWN           | */
		    /* IEEE80211_RADIOTAP_HE_DATA2_PE_DISAMBIG_KNOWN    | */
		    IEEE80211_RADIOTAP_HE_DATA2_TXOP_KNOWN |
		    /*IEEE80211_RADIOTAP_HE_DATA2_MIDAMBLE_KNOWN        | */
		    IEEE80211_RADIOTAP_HE_DATA2_RU_OFFSET |
		    IEEE80211_RADIOTAP_HE_DATA2_RU_OFFSET_KNOWN);

	/* data1 */
	v = ((rv->FORMAT == HE_SU) ? IEEE80211_RADIOTAP_HE_DATA1_FORMAT_SU :
	     (rv->FORMAT == HE_ER_SU) ? IEEE80211_RADIOTAP_HE_DATA1_FORMAT_EXT_SU :
	     (rv->FORMAT == HE_TB) ? IEEE80211_RADIOTAP_HE_DATA1_FORMAT_TRIG :
	     IEEE80211_RADIOTAP_HE_DATA1_FORMAT_MU);
	he.data1 |= HE_PREP(DATA1_FORMAT_MASK, v);

	/*
	 * data2
	 * IEEE80211_RADIOTAP_HE_DATA2_RU_OFFSET: driver
	 * IEEE80211_RADIOTAP_HE_DATA2_PRISEC_80_SEC: driver
	 */
	v = rv->RU_ALLOCATION0;
	v = ((v < 37) ? v :	/* 26-tone RU */
	     (v < 53) ? v - 37 :	/* 52-tone RU */
	     (v < 61) ? v - 53 :	/* 106-tone RU */
	     (v < 67) ? v - 61 :	/* 242-tone RU */
	     (v < 68) ? v - 67 :	/* 484-tone RU */
	     0);		/* 996-tone and 996x2-tone RU */
	he.data2 |= HE_PREP(DATA2_RU_OFFSET, v);
	he.data2 |= HE_PREP(DATA2_PRISEC_80_SEC, 0);	/* primary */

	/*
	 * data3
	 * IEEE80211_RADIOTAP_HE_DATA3_BSS_COLOR: driver
	 * IEEE80211_RADIOTAP_HE_DATA3_BEAM_CHANGE: driver
	 * IEEE80211_RADIOTAP_HE_DATA3_UL_DL: driver
	 * IEEE80211_RADIOTAP_HE_DATA3_DATA_MCS: mac80211
	 * IEEE80211_RADIOTAP_HE_DATA3_DATA_DCM: mac80211
	 * IEEE80211_RADIOTAP_HE_DATA3_CODING: mac80211
	 * IEEE80211_RADIOTAP_HE_DATA3_LDPC_XSYMSEG: N/A
	 * IEEE80211_RADIOTAP_HE_DATA3_STBC: mac80211
	 */
	he.data3 |= HE_PREP(DATA3_BSS_COLOR, rv->BSS_COLOR);
	he.data3 |= HE_PREP(DATA3_BEAM_CHANGE, rv->BEAM_CHANGE);
	he.data3 |= HE_PREP(DATA3_UL_DL, rv->UPLINK_FLAG);

	/*
	 * data4:
	 IEEE80211_RADIOTAP_HE_DATA4_SU_MU_SPTL_REUSE: driver
	 IEEE80211_RADIOTAP_HE_DATA4_MU_STA_ID: driver
	 IEEE80211_RADIOTAP_HE_DATA4_TB_SPTL_REUSE1: N/A
	 IEEE80211_RADIOTAP_HE_DATA4_TB_SPTL_REUSE2: N/A
	 IEEE80211_RADIOTAP_HE_DATA4_TB_SPTL_REUSE3: N/A
	 IEEE80211_RADIOTAP_HE_DATA4_TB_SPTL_REUSE4: N/A
	 */
	he.data4 |= HE_PREP(DATA4_SU_MU_SPTL_REUSE, rv->SPATIAL_REUSE);
	he.data4 |= HE_PREP(DATA4_MU_STA_ID, rv->STA_ID_LIST);

	/*
	 * data 5
	 * IEEE80211_RADIOTAP_HE_DATA5_DATA_BW_RU_ALLOC: mac80211
	 * IEEE80211_RADIOTAP_HE_DATA5_GI: mac80211
	 * IEEE80211_RADIOTAP_HE_DATA5_LTF_SIZE: driver
	 * IEEE80211_RADIOTAP_HE_DATA5_NUM_LTF_SYMS: N/A
	 * IEEE80211_RADIOTAP_HE_DATA5_PRE_FEC_PAD: N/A
	 * IEEE80211_RADIOTAP_HE_DATA5_TXBF: N/A
	 * IEEE80211_RADIOTAP_HE_DATA5_PE_DISAMBIG: N/A
	 */
	v = rv->HE_LTF_TYPE - HE_LTF_1x + IEEE80211_RADIOTAP_HE_DATA5_LTF_SIZE_1X;
	he.data5 |= HE_PREP(DATA5_LTF_SIZE, v);

	/*
	 * data 6
	 * IEEE80211_RADIOTAP_HE_DATA6_NSTS: mac80211
	 * IEEE80211_RADIOTAP_HE_DATA6_DOPPLER: driver
	 * IEEE80211_RADIOTAP_HE_DATA6_TB_PPDU_BW_KNOWN: N/A
	 * IEEE80211_RADIOTAP_HE_DATA6_TB_PPDU_BW: N/A
	 * IEEE80211_RADIOTAP_HE_DATA6_TXOP: driver
	 * IEEE80211_RADIOTAP_HE_DATA6_MIDAMBLE_PDCTY: N/A
	 */
	he.data6 |= HE_PREP(DATA6_DOPPLER, rv->DOPPLER);
	he.data6 |= HE_PREP(DATA6_TXOP, rv->TXOP_DURATION);	/* raw value */

	if (rv->FORMAT == HE_MU) {
		rxs->flag |= RX_FLAG_RADIOTAP_HE_MU;
		rxs->bw = RATE_INFO_BW_HE_RU;

		v = rv->RU_ALLOCATION0;
		rxs->he_ru = ((v < 37) ? NL80211_RATE_INFO_HE_RU_ALLOC_26 :
			      (v < 53) ? NL80211_RATE_INFO_HE_RU_ALLOC_52 :
			      (v < 61) ? NL80211_RATE_INFO_HE_RU_ALLOC_106 :
			      (v < 65) ? NL80211_RATE_INFO_HE_RU_ALLOC_242 :
			      (v < 67) ? NL80211_RATE_INFO_HE_RU_ALLOC_484 :
			      (v < 68) ? NL80211_RATE_INFO_HE_RU_ALLOC_996 :
			      NL80211_RATE_INFO_HE_RU_ALLOC_2x996);

		he_mu.flags1 |= (IEEE80211_RADIOTAP_HE_MU_FLAGS1_SIG_B_MCS_KNOWN |
				 IEEE80211_RADIOTAP_HE_MU_FLAGS1_SIG_B_DCM_KNOWN |
				 /*IEEE80211_RADIOTAP_HE_MU_FLAGS1_CH1_RU_KNOWN          | */
				 /*IEEE80211_RADIOTAP_HE_MU_FLAGS1_CH2_RU_KNOWN          | */
				 /*IEEE80211_RADIOTAP_HE_MU_FLAGS1_CH1_CTR_26T_RU_KNOWN  | */
				 /*IEEE80211_RADIOTAP_HE_MU_FLAGS1_CH2_CTR_26T_RU_KNOWN  | */
				 /*IEEE80211_RADIOTAP_HE_MU_FLAGS1_SIG_B_COMP_KNOWN      | */
				 /*IEEE80211_RADIOTAP_HE_MU_FLAGS1_SIG_B_SYMS_USERS_KNOWN */
				 0);

		he_mu.flags2 |= IEEE80211_RADIOTAP_HE_MU_FLAGS2_BW_FROM_SIG_A_BW_KNOWN;

		he_mu.flags1 |= HE_PREP(MU_FLAGS1_SIG_B_MCS, rv->MCS_SIG_B);
		he_mu.flags1 |= HE_PREP(MU_FLAGS1_SIG_B_DCM, rv->DCM_SIG_B);

		v = ((rxs->bw == RATE_INFO_BW_20) ?
		     IEEE80211_RADIOTAP_HE_MU_FLAGS2_BW_FROM_SIG_A_BW_20MHZ :
		     (rxs->bw == RATE_INFO_BW_40) ?
		     IEEE80211_RADIOTAP_HE_MU_FLAGS2_BW_FROM_SIG_A_BW_40MHZ :
		     (rxs->bw == RATE_INFO_BW_80) ?
		     IEEE80211_RADIOTAP_HE_MU_FLAGS2_BW_FROM_SIG_A_BW_80MHZ :
		     IEEE80211_RADIOTAP_HE_MU_FLAGS2_BW_FROM_SIG_A_BW_160MHZ);
		he_mu.flags2 |= HE_PREP(MU_FLAGS2_BW_FROM_SIG_A_BW, v);

		/* We do not fill he_mu.ru_ch1[] and he_mu.ru_ch2[] */

		memcpy(skb_push(skb, sizeof(he_mu)), &he_mu, sizeof(he_mu));
	}

	memcpy(skb_push(skb, sizeof(he)), &he, sizeof(he));

	return RX_CONTINUE;
}

#define IEEE80211_SEQ_BA_RANGE		(IEEE80211_SN_MODULO >> 1)

/**
 * scm80211_rxh_ampdu() - updates BA scoreboard
 *
 * TODO:
 * - Multi-TID A-MPDU
 * - Ack-enabled A-MPDU
 */
static enum scm80211_rx_result scm80211_rxh_block_ack_scoreboard(struct scm80211_rx_data *rx)
{
	struct ieee80211_hw_priv *priv = rx->hw->priv;
	struct scm80211_ampdu_info *ampdu = &priv->monitor.ampdu;
	int tid;
	u16 sn;
	u8 *qc;

	if (!(rx->hw->conf.flags & IEEE80211_CONF_MONITOR))
		return RX_CONTINUE;

	if (!rx->rh->ampdu)
		return RX_CONTINUE;

	if (rx->rh->timestamp != priv->monitor.ampdu_reference) {
		/* A new PPDU hence a new A-MPDU */
		priv->monitor.ampdu_reference = rx->rh->timestamp;
		ampdu->enabled = true;
	}

	if (ieee80211_is_data_qos(rx->hdr->frame_control)) {
		unsigned long *bitmap;

		sn = ieee80211_get_sn(rx->skb);
		qc = ieee80211_get_qos_ctl(rx->hdr);
		tid = qc[0] & IEEE80211_QOS_CTL_TID_MASK;
		bitmap = ampdu->tids[tid].bitmap;

		if (!ether_addr_equal(rx->hdr->addr2, priv->monitor.ta) ||
		    !ether_addr_equal(rx->hdr->addr1, priv->monitor.ra)) {
			/*
			 * BA context is not for this <ta, ra, tid> tuple.
			 * Perform partial-state operation.
			 */
			memset(ampdu, 0, sizeof(*ampdu));

			memcpy(&ampdu->tids[tid].qosctl, qc, 2);
			ampdu->tids[tid].ssn = ieee80211_sn_sub(sn, 63);
			bitmap_zero(bitmap, 64);
			set_bit(63, bitmap);

			ether_addr_copy(priv->monitor.ra, rx->hdr->addr1);
			ether_addr_copy(priv->monitor.ta, rx->hdr->addr2);
		} else {
			u16 ssn, delta;

			/* Temporary BA record is found. Perform full-state operation */
			ssn = ampdu->tids[tid].ssn;
			delta = ieee80211_sn_sub(sn, ssn);

			if (delta <= 63) {
				set_bit(delta, bitmap);
			} else if (delta < IEEE80211_SEQ_BA_RANGE) {
				u16 shift = ieee80211_sn_sub(sn, ssn + 63);

				if (shift > 63) {
					bitmap_zero(bitmap, 64);
				} else {
					bitmap_shift_right(bitmap, bitmap, shift, 64);
					bitmap_clear(bitmap, 64 - shift, shift);
				}
				ampdu->tids[tid].ssn = ieee80211_sn_add(ssn, shift);
				set_bit(63, bitmap);
			}
		}

		ampdu->has_qos_data |= true;

		if (rx->rh->eof) {
			set_bit(tid, &ampdu->tidset_eof);
			ampdu->has_immediate_response |= true;
		} else {
			set_bit(tid, &ampdu->tidset_ba);

			ampdu->has_immediate_response |=
			    (qc[0] & IEEE80211_QOS_CTL_ACK_POLICY_MASK) ==
			    IEEE80211_QOS_CTL_ACK_POLICY_NORMAL;
		}
	}
	return RX_CONTINUE;
}

#define IEEE80211_MAX_BA_LEN	(sizeof(struct ieee80211_block_ack) + (16 * (4 + 32)))

/**
 * scm80211_get_BA() - creates a BlockAck frame
 *
 * TODO: M-BA
 */
static struct sk_buff *scm80211_get_BA(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_ampdu_info *ampdu = &priv->monitor.ampdu;
	struct sk_buff *skb = NULL;
	struct ieee80211_block_ack *ba;
	u8 *info;
	unsigned long tidset;
	int nr_ack_ctx = 0, nr_ba_ctx = 0;
	int ack_len = sizeof(struct ieee80211_block_ack);

	tidset = ampdu->tidset_ba | ampdu->tidset_eof;
	if (tidset == 0)
		return NULL;

	nr_ba_ctx = hweight32(ampdu->tidset_ba);
	nr_ack_ctx = hweight32(ampdu->tidset_eof);

	/* FIXME: M-BA is not considered for the time being */
	if (nr_ack_ctx > 0 || nr_ba_ctx + nr_ack_ctx > 1)
		return NULL;

	ack_len += nr_ba_ctx * (2 + 2 + 8);	/* only 64-bit bitmap */
	ack_len += nr_ack_ctx * 2;

	skb = alloc_skb(ack_len, GFP_ATOMIC);
	ba = (struct ieee80211_block_ack *)skb_put(skb, sizeof(*ba));

	ba->frame_control = cpu_to_le16(IEEE80211_FTYPE_CTL | IEEE80211_STYPE_BACK);
	ba->duration = 0;

	ether_addr_copy(ba->ra, priv->monitor.ta);
	ether_addr_copy(ba->ta, priv->monitor.ra);

	if (nr_ack_ctx == 0 && nr_ba_ctx == 1) {
		struct ieee80211_rx_status *rxs;
		u16 ssn;
		int tid;

		/* Compressed BlockAck */
		tid = find_first_bit(&tidset, sizeof(tidset) * BITS_PER_BYTE);

		ba->control[0] = IEEE80211_BA_CONTROL_COMPRESSED_BA;
		ba->control[1] = FIELD_PREP(IEEE80211_BA_CONTROL_TID_INFO, tid);

		ssn = ampdu->tids[tid].ssn;
		info = skb_put(skb, 2 + 8);
		put_unaligned_le16(IEEE80211_SN_TO_SEQ(ssn), info);
		info += 2;
		memcpy(info, ampdu->tids[tid].bitmap, 8);

		rxs = IEEE80211_SKB_RXCB(skb);
		memset(rxs, 0, sizeof(*rxs));

		rxs->flag |= RX_FLAG_ONLY_MONITOR;
		rxs->freq = hw->conf.chandef.chan->center_freq;
		rxs->band = hw->conf.chandef.chan->band;
		rxs->signal = -50;
		rxs->bw = RATE_INFO_BW_20;
		rxs->encoding = RX_ENC_LEGACY;
		rxs->rate_idx = 0;
	}

	return skb;
}

/**
 * scm80211_rxh_monitor_block_ack() - generate a fictitious BlockAck for monitor
 *
 * Hardware does not "loopback" response frames it generates to software. This
 * function generates a BlockAck if the A-MPDU carried in the previous ppdu
 * solicits an immediate response and forwards it to monitor interfaces. Since
 * hardware does not indicate whether or not a subframe is the last one of an A-MPDU,
 * the driver can do this only after receiving a new ppdu. Hence, the capture time
 * of these BlockAck frames are not accurate.
 */
static enum scm80211_rx_result scm80211_rxh_monitor_block_ack(struct scm80211_rx_data *rx)
{
	struct ieee80211_hw_priv *priv = rx->hw->priv;
	struct scm80211_ampdu_info *ampdu = &priv->monitor.ampdu;

	/* No monitor inerface. Don't bother */
	if (!(rx->hw->conf.flags & IEEE80211_CONF_MONITOR))
		return RX_CONTINUE;

	/* We are in the middle of an A-MPDU */
	if (rx->rh->timestamp == priv->monitor.ampdu_reference)
		return RX_CONTINUE;

	/* The last PPDU does not carry an A-MPDU */
	if (!priv->monitor.ampdu.enabled)
		return RX_CONTINUE;

	/* The last A-MPDU solicits an immediate response */
	if (ampdu->has_immediate_response) {
		struct sk_buff *skb;

		skb = scm80211_get_BA(rx->hw);
		if (skb)
			skb_queue_head(&rx->queue, skb);

		ampdu->enabled = false;
		ampdu->has_immediate_response = false;
		ampdu->tidset_ba = 0;
		ampdu->tidset_eof = 0;
	}

	/* Starts a new PPDU */
	priv->monitor.timestamp = rx->rh->timestamp;

	return RX_CONTINUE;
}

/**
 * scm80211_rxh_monitor_ack() - generates an Ack for monitor interface
 *
 * This function generates an Ack if the (S)-MPDU carried in the previous ppdu
 * solicits an immediate response. The generated Ack is forwarded only to
 * monitor interfaces.
 */
static enum scm80211_rx_result scm80211_rxh_monitor_ack(struct scm80211_rx_data *rx)
{
	struct ieee80211_hdr *hdr;
	struct ieee80211_rx_status *rxs;
	struct sk_buff *skb;

	if (!(rx->hw->conf.flags & IEEE80211_CONF_MONITOR))
		return RX_CONTINUE;	/* don't bother */

	if (!ieee80211_solicits_ack(rx->hdr))
		return RX_CONTINUE;

	if (!rx->rh->ampdu && !ieee80211_is_back_req(rx->hdr->frame_control)) {
		/* Ack */
		skb = dev_alloc_skb(100);
		if (!skb)
			return RX_CONTINUE;

		hdr = (struct ieee80211_hdr *)skb_put(skb, 10);
		hdr->frame_control = cpu_to_le16(IEEE80211_FTYPE_CTL | IEEE80211_STYPE_ACK);
		hdr->duration_id = cpu_to_le16(0);
		memcpy(hdr->addr1, rx->hdr->addr2, ETH_ALEN);

		rxs = IEEE80211_SKB_RXCB(skb);
		memset(rxs, 0, sizeof(*rxs));

		rxs->flag |= RX_FLAG_ONLY_MONITOR;
		rxs->freq = rx->hw->conf.chandef.chan->center_freq;
		rxs->band = rx->hw->conf.chandef.chan->band;
		rxs->signal = rx->rxs->signal;
		rxs->bw = RATE_INFO_BW_20;
		rxs->encoding = RX_ENC_LEGACY;
		rxs->rate_idx = 0;

		/* Add ack at the tail */
		skb_queue_tail(&rx->queue, skb);

		return RX_CONTINUE;
	}

	return RX_CONTINUE;
}

#define CALL_RXH(rxh)						\
	do {							\
		ret = rxh(&rx);					\
		if (ret == RX_DROP || ret == RX_HANDLED)	\
			goto drop;				\
		if (ret == RX_MONITOR_ONLY) {			\
			rx.rxs->flag |= RX_FLAG_ONLY_MONITOR;	\
			scm80211_trace((rx.hw), #rxh " MONITOR_ONLY\n");\
		}						\
		if (rx.rxs->flag & RX_FLAG_ONLY_MONITOR)	\
			goto out;				\
	} while (0)

/**
 * scm80211_rx()
 */
static int scm80211_rx(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_rx_data rx;
	enum scm80211_rx_result ret;
	struct sk_buff *skb1;
	bool in_task = in_task();
#ifdef SCM80211_FTM
	struct ieee80211_hdr *hdr;
#endif

	rcu_read_lock();

	ret = scm80211_rx_start(&rx, hw, skb);
	if (ret == RX_DROP)
		goto drop;
	else if (ret == RX_AGAIN)
		return -EAGAIN;

	CALL_RXH(scm80211_rxh_dump_rxb);
	/* CALL_RXH(scm80211_rxh_bad_frame); */
	CALL_RXH(scm80211_rxh_monitor_block_ack);
	CALL_RXH(scm80211_rxh_status);
	CALL_RXH(scm80211_rxh_address_filter);
	CALL_RXH(scm80211_rxh_decrypt);
	CALL_RXH(scm80211_rxh_htc);
	CALL_RXH(scm80211_rxh_prehandle);
	CALL_RXH(scm80211_rxh_block_ack_scoreboard);
	CALL_RXH(scm80211_rxh_monitor_ack);
#ifdef SCM80211_FTM
	hdr = rx.hdr;
	if (ieee80211_is_data_present(hdr->frame_control))
		scm80211_ftm_rx_statis_update(&rx);
#endif // SCM80211_FTM
out:
	scm80211_rxh_he_radiotap_header(&rx);

	while ((skb1 = skb_dequeue(&rx.queue))) {
		struct ieee80211_sta *sta = (skb == skb1) ? rx.sta : NULL;

		if (in_task)
			local_bh_disable();
		ieee80211_rx_napi(hw, sta, skb1, priv->napi.enabled ? &priv->napi.context : NULL);
		if (in_task)
			local_bh_enable();
	}

	rcu_read_unlock();
	return 0;

drop:
	rcu_read_unlock();
	kfree_skb(skb);
	return -EINVAL;
}

/**
 * scm80211_rx_statistics_update()
 * Calculate rx statistics (roughly)
 */
void scm80211_rx_statistics_update(struct ieee80211_hw *hw)
{
#ifdef SCM80211_RX_STATISTICS_DEBUG
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 val = scm80211_mac_readl(hw, MAC_REG_RXE_MPDU_INFO);

	if (val & BIT(7)) {
		priv->rx_overrun++;
		scm80211_info(hw, "rx_statistics: overrun for %d times\n", priv->rx_overrun);
	}
#endif // SCM80211_RX_STATISTICS_DEBUG
}

int scm80211_rx_poll(struct ieee80211_hw *hw, int budget)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct sk_buff *skb;
	int done = 0;
	int ret;

	/* Process received frames */
	while ((skb = skb_dequeue(&priv->rxring.rxskbs))) {
		ret = scm80211_rx(hw, skb);
		if (!ret)
			done++;

		if (done >= budget && ret != -EAGAIN)
			break;
	}

	if (WARN_ON(!skb_queue_empty(&priv->rxring.frags)))
		skb_queue_purge(&priv->rxring.frags);

	if (priv->napi.enabled && done < budget) {
		napi_complete_done(&priv->napi.context, done);
		if (!skb_queue_empty(&priv->rxring.rxskbs))
			napi_reschedule(&priv->napi.context);
	}

	/**
	 * There will be some warning in net stack
	 * if return value bigger than budget,
	 * so cut it down under this case
	 */
	return done > budget ? budget : done;
}

static void scm80211_rxpool_work(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv;

	priv = container_of(work, struct ieee80211_hw_priv, rxring.work);

	scm80211_refill_rxpool(priv->hw);
	scm80211_refill_rxring(priv->hw);
}

void scm80211_deinit_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	scm80211_bus_deinit_rx(hw);

	cancel_work_sync(&priv->rxring.work);

	if (priv->napi.enabled) {
		napi_synchronize(&priv->napi.context);
		napi_disable(&priv->napi.context);
		netif_napi_del(&priv->napi.context);
	}
	priv->napi.enabled = false;

	flush_work(&priv->rxring.work);
	skb_queue_purge(&priv->rxring.pool);
	skb_queue_purge(&priv->rxring.rxskbs);
	skb_queue_purge(&priv->rxring.frags);
}

int scm80211_init_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 v = 0;

	/* Disable RX interrupt */
	scm80211_disable_rx_irq(hw);

	memset(&priv->rxring, 0, sizeof(priv->rxring));

	spin_lock_init(&priv->rxring.lock);
	priv->rxring.order = min_t(int, ring_order, 8);
	skb_queue_head_init(&priv->rxring.pool);
	skb_queue_head_init(&priv->rxring.rxskbs);
	skb_queue_head_init(&priv->rxring.frags);

	/* Set RX chain */
	v = scm80211_mac_readl(hw, MAC_REG_DEV_CFG);
	v |= FIELD_PREP(DEV_CFG_N_RX, priv->n_chain - 1);
	scm80211_mac_writel(hw, v, MAC_REG_DEV_CFG);

	/* Initialize BA scoreboard */
	v = FIELD_PREP(RX_SCB_CTRL_SCB_RESET, 0xff);
	v |= FIELD_PREP(RX_SCB_CTRL_SCB_WINDOW, 64 - 1);
	scm80211_mac_writel(hw, v, MAC_REG_RX_SCB_CTRL);

	/* OFDM RX timeout */
	v = (FIELD_PREP(RX_TO_CFG_RX_TO_VALUE, 0x1c00));
	v |= (FIELD_PREP(RX_TO_CFG_RX_TO_PPDU_END_EN, 1));
	scm80211_mac_writel(hw, v, MAC_REG_RX_TO_CFG);

	/* CCK rx timeout register for 11b */
	v = (FIELD_PREP(RX_TO_11B_CFG_RX_TO_11B_VALUE, 0x8000));
	scm80211_mac_writel(hw, v, MAC_REG_RX_TO_11B_CFG);

	INIT_WORK(&priv->rxring.work, scm80211_rxpool_work);

	scm80211_refill_rxpool(hw);

	return scm80211_bus_init_rx(hw);
}

int scm80211_ampdu_rx_action(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			     struct ieee80211_ampdu_params *params)
{
	u32 v;

	switch (params->action) {
	case IEEE80211_AMPDU_RX_START:
	case IEEE80211_AMPDU_RX_STOP:
		scm80211_trace(hw,
			       "RX BA session %s: %pM: tid=%d, ssn=%u, bufsize=%u, amsdu=%d\n",
			       (params->action == IEEE80211_AMPDU_RX_START) ?
			       "established" : "destroyed",
			       params->sta->addr, params->tid, params->ssn, params->buf_size,
			       params->amsdu);

		/* Reset the scoreboard */
		v = scm80211_mac_readl(hw, MAC_REG_RX_SCB_CTRL);
		v &= ~RX_SCB_CTRL_SCB_RESET;
		v |= FIELD_PREP(RX_SCB_CTRL_SCB_RESET, BIT(params->tid));
		scm80211_mac_writel(hw, v, MAC_REG_RX_SCB_CTRL);
		break;
	default:
		return 0;
	}
	return 0;
}

void scm80211_wait_for_rx_idle(struct ieee80211_hw *hw)
{
	int retries = 10;
	u32 v;

	/* Wait for RXE to park, here is no guarantee */
	do {
		v = scm80211_mac_readl(hw, MAC_REG_MAC_STATUS);
		if (FIELD_GET(MAC_STATUS_DMA_BUSY, v) == 0)
			break;
		mdelay(1);
	} while (retries-- > 0);
}
