// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 * SENSCOMM scm2020
 */
#include <linux/bitfield.h>
#include <linux/spi/spi.h>
#include <linux/platform_device.h>

#define func_fmt(str)           "[PHY] "str
#define func_trace              PHY

#include "core.h"
#include "util.h"
#include "phy.h"
#include "regs.h"
#include "debug.h"

/* RFIF control signal assignment */
#define RFI_CTRLSIG_SHDN	RFI_CTRLSIG_CFG(0)
#define RFI_CTRLSIG_2G_PA_EN	RFI_CTRLSIG_CFG(1)
#define RFI_CTRLSIG_5G_PA_EN	RFI_CTRLSIG_CFG(2)
#define RFI_CTRLSIG_EXT_SW_RX	RFI_CTRLSIG_CFG(3)
#define RFI_CTRLSIG_EXT_SW_TX	RFI_CTRLSIG_CFG(4)
#define RFI_CTRLSIG_RF_RX_EN	RFI_CTRLSIG_CFG(5)
#define RFI_CTRLSIG_RF_TX_EN	RFI_CTRLSIG_CFG(6)
#define RFI_CTRLSIG_ADC		RFI_CTRLSIG_CFG(7)
#define RFI_CTRLSIG_DAC		RFI_CTRLSIG_CFG(8)
#define RFI_CTRLSIG_RX_EN	RFI_CTRLSIG_CFG(12)
#define RFI_CTRLSIG_TX_EN	RFI_CTRLSIG_CFG(13)
#define RFI_CTRLSIG_RX_CLK	RFI_CTRLSIG_CFG(14)
#define RFI_CTRLSIG_TX_CLK	RFI_CTRLSIG_CFG(15)

/* Bitfield */
#define bfset_p(v, f, x)	((v) |= FIELD_PREP(f##_MASK, x))
#define bfclr_p(v, f)		((v) &= ~(f##_MASK))
#define bfmod_p(v, f, x)	{bfclr_p(v, f); bfset_p(v, f, x); }
#define bfget_p(v, f)		(FIELD_GET(f##_MASK, v))

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
/* ugly code for temporary use */
static int test_band;
module_param(test_band, int, 0644);
MODULE_PARM_DESC(test_band, "set test band 0 for 2.4G 1 for 5G");
#endif

#ifdef SCM2020_A0
static u32 iq_read_ch[2] = { 0, 0 };

u32 scm2020_iq_get(struct scm80211_phy *phy)
{
	return ((iq_read_ch[phy->chain] & 0x3ff) <= 0x1c0) ? 0x00000200 : iq_read_ch[phy->chain];
}
#endif

struct reg_set {
	u32 reg;
	u32 value;
};

static struct {
	u8 phy_11b_cs_rssi_thr;
	u8 phy_cs_xcr_minrssi;
	u8 phy_cs_acr_minrssi;
} cs;

#define RSET(region, reg, value)	{region(reg), value}

#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
static struct reg_set phy_init[] = {
	/* PHY Control mode */
	{ PHY_RFIF(FW_RFI_EN), 0x0 },
	/* SHDN: - ctrl[0], always on */
	{ RFI_CTRLSIG_SHDN, RFCTRL(1, HIGH, ALWAYS, TXEN, 0, 0) },
	/* 2G PA En:1 - ctrl[1], during TXVALID */
	{ RFI_CTRLSIG_2G_PA_EN, RFCTRL(1, HIGH, TXVALID, TXVALID, 0, 0) },
	/* 5G PA En - ctrl[2], always off */
	{ RFI_CTRLSIG_5G_PA_EN, RFCTRL(1, HIGH, TXVALID, ALWAYS, 0, 0) },
	/* ExtSwitch Rx - ctrl[3], during RXEN */
	{ RFI_CTRLSIG_EXT_SW_RX, RFCTRL(1, HIGH, RXEN, RXEN, 0, 0) },
	/* ExtSwitch Tx - ctrl[4], during TXEN */
	{ RFI_CTRLSIG_EXT_SW_TX, RFCTRL(1, HIGH, TXEN, TXEN, 0, 0) },
	/* RF Rx Enable - ctrl[5], during RXEN */
	{ RFI_CTRLSIG_RF_RX_EN, RFCTRL(1, HIGH, RXEN, RXEN, 0, 0) },
	/* RF Tx Enable - ctrl[6], during TXEN */
	{ RFI_CTRLSIG_RF_TX_EN, RFCTRL(1, HIGH, TXEN, TXEN, 0, 0) },
	/* ADC control - ctrl[7], always on */
	{ RFI_CTRLSIG_ADC, RFCTRL(1, HIGH, ALWAYS, RXEN, 0, 0) },
	/* DAC control - ctrl[8], during TXEN */
	{ RFI_CTRLSIG_DAC, RFCTRL(1, HIGH, TXEN, TXEN, 1, 5) },
	/* RX_EN (digital I/Q) - ctrl[12], during RXEN */
	{ RFI_CTRLSIG_RX_EN, RFCTRL(1, HIGH, RXEN, RXEN, 0, 0) },
	/* TX_EN (digital I/Q) - ctrl[13], during TXEN */
	{ RFI_CTRLSIG_TX_EN, RFCTRL(1, HIGH, TXEN, TXEN, 0, 0) },
	/* RX_CLK  - ctrl[14], always (in FPGA) */
	{ RFI_CTRLSIG_RX_CLK, RFCTRL(1, HIGH, ALWAYS, TXEN, 0, 0) },
	/* TX_CLK  - ctrl[15], always on (in FGPA) */
	{ RFI_CTRLSIG_TX_CLK, RFCTRL(1, HIGH, ALWAYS, TXEN, 0, 0) },

	/* Tx DAC interface ON */
	{ PHY_RFIF(FW_RFI_TXDAC_INTF), 0x00000001 },
	/* Rx ADC interface ON */
	{ PHY_RFIF(FW_RFI_RXADC_INTF), 0x00000001 },
	/* TX digital manual gain */
	{ PHY_RFIF(PHY_RFI_TX_DIG_GAIN_MAN), 0x000000A8 },
	/* RX digital manual gain */
	/* {PHY_RFIF(PHY_RFI_RX_DIG_GAIN_MAN),0x000000A8), */
	/* RX digital manual gain disable */
	{ PHY_RFIF(PHY_RFI_RX_DIG_GAIN_MAN), 0x00000000 },
	/* Rx HPF duration */
	/* {PHY_RFIF(FW_RXHPF_CFG1), 0x00040004), */
	{ PHY_RFIF(FW_RXHPF_CFG1), 0x00000808 },
	/* Rx HPF BW selection */
	{ PHY_RFIF(FW_RXHPF_CFG2), 0x00000000 },

	/* PHY power en */
	{ PHY_MODEM(FW_PWR_EN), 0x00000001 },
	/* AGC LUT index range */
	{ PHY_MODEM(PHY_AGC_INDEX_CFG), 0x00005E5A },
	/* Power low TH change */
	{ PHY_MODEM(PHY_AGC_PWRMSR_CFG), 0x001100FE },
	/* Init power range */
	{ PHY_MODEM(PHY_AGC_INITPWR_RANGE), 0x01A2012C },
	/* Target power */
	{ PHY_MODEM(PHY_AGC_TARGET_PWR), 0x015E01A2 },
	/* Gain adj wait time */
	{ PHY_MODEM(PHY_AGC_WAIT_CNT2), 0x00280020 },
	/* Gain step control */
	{ PHY_MODEM(PHY_AGC_GDOWN_CFG), 0x18181818 },
	/* AGC debug config default setting */
	{ PHY_MODEM(PHY_AGC_TEST_OPTION), 0x00007F11 },

	/* Default value update for All version */
	/* Initial fine/coarse gain adj wait counter */
	{ PHY_MODEM(PHY_AGC_WAIT_CNT1), 0x00280020 },

	/* SW reset */
	{ PHY_MODEM(FW_PHY_SW_RESET), 0x00000001 },
	/* Temp Config */
	{ PHY_MODEM(PHY_DONT_BE_STUCK), 0xFFFFFFFF },

	/* Turn off (GID, PAID) filtering */
	{ PHY_MODEM(FW_VHT_CONFIG0_VINTF(0)), 0x00000000 },
	/* Turn off (GID, PAID) filtering */
	{ PHY_MODEM(FW_VHT_CONFIG0_VINTF(1)), 0x00000000 },

	/* RSSI offset */
	{ PHY_MODEM(PHY_RSSI_OFFSET), 0x000000F4 },
	/* SAT det config */
	{ PHY_RFIF(PHY_RFI_SATDET_CFG), 0x000C0708 },
	/* HPF config when Init gain change */
	{ PHY_RFIF(FW_RXHPF_CFG2), 0x000C0000 },
	/* HPF config when Init gain change */
	{ PHY_RFIF(FW_RXHPF_CFG1), 0x00280808 },
	/* 11B min sen config */
	{ PHY_MODEM(PHY_11B_DET), 0x0003E802 },
};
#else
static const struct reg_set phy_init[] = {
	/* Don't Stop reg */
	{ PHY_MODEM(PHY_DONT_BE_STUCK), 0xFFFFFFFF },
	/* 11B min sen config */
	{ PHY_MODEM(PHY_11B_DET), 0x0003E802 },
	/* RSSI offset */
	{ PHY_MODEM(PHY_RSSI_OFFSET), 0x000000E5 },
	/* AGC LUT index range */
	{ PHY_MODEM(PHY_AGC_INDEX_CFG), 0x00005650 },
	/* Gain step control */
	{ PHY_MODEM(PHY_AGC_GDOWN_CFG), 0x0f142018 },
	/* Initial fine/coarse gain adj wait counter */
	{ PHY_MODEM(PHY_AGC_WAIT_CNT1), 0x00280028 },
	/* Gain adj wait time */
	{ PHY_MODEM(PHY_AGC_WAIT_CNT2), 0x00280028 },
	/* Power low TH change */
	{ PHY_MODEM(PHY_AGC_PWRMSR_CFG), 0x001100FE },
	/* Init power range */
	{ PHY_MODEM(PHY_AGC_INITPWR_RANGE), 0x01A2012C },
	/* Target power */
	{ PHY_MODEM(PHY_AGC_TARGET_PWR), 0x015E01A2 },
	/* TXEN_to_RF */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_00), RFCTRL(1, HIGH, TXEN, TXEN, 0x50, 0x00) },
	/* TXEN_DIG */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_13), RFCTRL(1, HIGH, TXEN, TXEN, 0x88, 0x00) },
	/* CLKEN_TXDAC */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_11), RFCTRL(1, HIGH, TXEN, TXEN, 0x00, 0x7a) },
	/* CLKEN_TXRFI */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_15), RFCTRL(1, HIGH, TXEN, TXEN, 0x00, 0x7a) },
	/* RXEN_to_RF */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_01), RFCTRL(1, LOW, TXEN, TXEN, 0x00, 0x50) },
	/* RXEN_DIG */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_12), RFCTRL(1, LOW, TXEN, TXEN, 0x00, 0x80) },
	/* CLKEN_RXADC */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_10), RFCTRL(1, LOW, TXEN, TXEN, 0x50, 0x00) },
	/* CLKEN_RXRFI */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_14), RFCTRL(1, LOW, TXEN, TXEN, 0x50, 0x00) },
	/* TRX_SEL_CONFIG_1 */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_18), RFCTRL(1, LOW, TXEN, TXEN, 0x58, 0x00) },
	/* TRX_SEL_CONFIG_2 */
	{ PHY_RFIF(FW_RFI_CTRLSIG_CFG_19), RFCTRL(1, LOW, TXEN, TXEN, 0x58, 0x00) },
	/* Tx DAC interface ON */
	{ PHY_RFIF(FW_RFI_TXDAC_INTF), 0x00000041 },
	/* Rx ADC interface ON */
	{ PHY_RFIF(FW_RFI_RXADC_INTF), 0x000000A7 },
	/* Rx LPF CFG */
	{ PHY_RFIF(FW_RFI_RXLPF_CFG), 0x00001011 },
	/* RX digital manual gain disable */
	{ PHY_RFIF(PHY_RFI_RX_DIG_GAIN_MAN), 0x00000000 },
	/* SAT det config */
	{ PHY_RFIF(PHY_RFI_SATDET_CFG), 0x000C0708 },
	/* HPF config when Init gain change */
	{ PHY_RFIF(FW_RXHPF_CFG1), 0x00002828 },
	/* HPF config when Init gain change */
	{ PHY_RFIF(FW_RXHPF_CFG2), 0x00003210 },
	/* PHY_RFI_SATDET_CFG */
	{ PHY_RFIF(PHY_RFI_SATDET_CFG), 0x003C0708 },
	/* PHY_STR_PKT_ON */
	{ PHY_MODEM(PHY_STR_PKT_ON), 0x00000e21 },
	/* PHY_SMOOTHING_ON */
	{ PHY_MODEM(PHY_CHESMT_CFG0), 0x00000001 },
	/* PHY_TXWAIT1 40M SIFS tuning */
	{ PHY_MODEM(PHY_TXWAIT1), 0x00028002 },
	/* PHY_TXWAIT2 80M SIFS tuning */
	{ PHY_MODEM(PHY_TXWAIT2), 0x00038401 },
	/* PHY BCC clock boosting */
	{ PHY_MODEM(FW_VITERBI_CLK_CTRL), 0x00000001 },
#ifdef SCM2020_A0
	/* disable CG and enable path SWAP */
	{ PHY_RFIF(FW_RFI_EN), 0x00100100 },
	/* Need update */
	{ PHY_MODEM(PHY_CS_ACISAT_FLT), 0x000014AB },
	/* IRR HW cal tuning -70~-40dBm, CS_ready dis */
	{ PHY_MODEM(PHY_HWMMCAL_TRK_CFG), 0xBAF6001E },
	/* CS Th control tuning for min sen */
	{ PHY_MODEM(PHY_CS_CONFIG_THRD), 0x0012161a },

	/* phase amp tracker coefficient tuning for tuning min sen */
	{ PHY_MODEM(PHY_PHAMP_CFFS), 0x397C8288 },
	/* LLR_SCL value change for 256/1024QAM to improve 2SS performance */
	{ PHY_MODEM(PHY_SMD_LLR_SCL1), 0x00001214 },
#else
	/* PHY Control mode */
	{ PHY_RFIF(FW_RFI_EN), 0x00110000 },
#endif
	/* PHY power en */
	{ PHY_MODEM(FW_PWR_EN), 0x00001FF1 },
	/* SW reset */
	{ PHY_MODEM(FW_PHY_SW_RESET), 0x00000001 },
};
#endif

static inline bool scm80211_is_adveos(struct scm80211_phy *phy)
{
	return phy->radio->devname[0] == 'a';
}

void scm2020_phy_rtx_lut(struct scm80211_phy *phy, int freq)
{
	phy->radio->ops->set_rtx_lut(phy->radio, freq);
}

static void scm2020_phy_trx_set_boost(struct scm80211_phy *phy, u8 boost)
{
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	u32 v = phy_readl(phy, PHY_RFIF(FW_RFI_RXLPF_CFG));
	u32 v2 = phy_readl(phy, PHY_RFIF(FW_RFI_EN));

	v &= ~(0x00001F1F);	//bits 12-8, 4-0
	v2 &= ~(0x11000000);	//bits 28, 23

	if (boost) {
		v |= 0x00001110;
		v2 |= 0x11000000;
	} else {
		v |= 0x00001011;
	}

#ifdef SCM2020_Z1
	v2 &= ~(1 << 8);
	if (phy->rf_ant_sel)
		v2 |= (1 << 8);
#elif defined SCM2020_A0
	v2 |= (1 << 8);
	if (phy->rf_ant_sel)
		v2 &= ~(1 << 8);
#endif
	phy_writel(phy, v, PHY_RFIF(FW_RFI_RXLPF_CFG));
	phy_writel(phy, v2, PHY_RFIF(FW_RFI_EN));
#endif
}

static void scm80211_reconfig_hpf_setting(struct scm80211_phy *phy)
{
#ifdef SCM2020_A0
	struct ieee80211_hw_priv *priv = phy->hw->priv;
	struct scm80211_phy *phy0 = priv->dbdc ? phy : (phy - phy->chain);

	if (!(phy->rfi_hpf[0] | phy->rfi_hpf[1] | phy->rfi_hpf[2])) {
		phy->rfi_hpf[0] = phy_readl(phy, PHY_RFIF(FW_RXLUT_CFG0));
		phy->rfi_hpf[1] = phy_readl(phy, PHY_RFIF(FW_RXHPF_CFG1));
		phy->rfi_hpf[2] = phy_readl(phy, PHY_RFIF(FW_RFI_RXDC_HPF_CFG));
	}

	phy_writel(phy, 0x100000, PHY_RFIF(FW_RXLUT_CFG0));
	phy_writel(phy, 0xff2828, PHY_RFIF(FW_RXHPF_CFG1));
	phy_writel(phy, 0x018552, PHY_RFIF(FW_RFI_RXDC_HPF_CFG));

	if (phy != phy0) {
		phy_writel(phy0, 0x100000, PHY_RFIF(FW_RXLUT_CFG0));
		phy_writel(phy0, 0xff2828, PHY_RFIF(FW_RXHPF_CFG1));
		phy_writel(phy0, 0x018552, PHY_RFIF(FW_RFI_RXDC_HPF_CFG));
	}
#endif
}

static void scm80211_restore_hpf_setting(struct scm80211_phy *phy)
{
#ifdef SCM2020_A0
	if (phy->rfi_hpf[0] | phy->rfi_hpf[1] | phy->rfi_hpf[2]) {
		phy_writel(phy, phy->rfi_hpf[0], PHY_RFIF(FW_RXLUT_CFG0));
		phy_writel(phy, phy->rfi_hpf[1], PHY_RFIF(FW_RXHPF_CFG1));
		phy_writel(phy, phy->rfi_hpf[2], PHY_RFIF(FW_RFI_RXDC_HPF_CFG));
	}
#endif
}

void scm2020_phy_fdscl_adj(struct scm80211_phy *phy, u8 en)
{
	if (en == 1) {
		/* FDSCL target +2dB specially for some rate to improve anti-spur performance */
		phy_writel(phy, 0xd4e4e4d4, PHY_MODEM(PHY_FDSCL_CFG2));
		phy_writel(phy, 0xe4e4d4e4, PHY_MODEM(PHY_FDSCL_CFG3));
		phy_writel(phy, 0xd4d4e4e4, PHY_MODEM(PHY_FDSCL_CFG4));
		phy_writel(phy, 0xd4e4d4d4, PHY_MODEM(PHY_FDSCL_CFG5));
		phy_writel(phy, 0xe4e4d4e4, PHY_MODEM(PHY_FDSCL_CFG6));
		phy_writel(phy, 0xd4d4e4e4, PHY_MODEM(PHY_FDSCL_CFG7));
	} else {
		// Default value
		phy_writel(phy, 0xd4d4d4d4, PHY_MODEM(PHY_FDSCL_CFG2));
		phy_writel(phy, 0xd4d4d4d4, PHY_MODEM(PHY_FDSCL_CFG3));
		phy_writel(phy, 0xd4d4d4d4, PHY_MODEM(PHY_FDSCL_CFG4));
		phy_writel(phy, 0xd4d4d4d4, PHY_MODEM(PHY_FDSCL_CFG5));
		phy_writel(phy, 0xd4d4d4d4, PHY_MODEM(PHY_FDSCL_CFG6));
		phy_writel(phy, 0xd4d4d4d4, PHY_MODEM(PHY_FDSCL_CFG7));
	}
}

void scm2020_phy_set_bandwidth(struct scm80211_phy *phy, u8 bw)
{
	if (bw == CBW20) {
		phy_writel(phy, 0x015E0172, PHY_MODEM(PHY_AGC_TARGET_PWR));
		phy_writel(phy, 0x04C2B0AE, PHY_MODEM(PHY_CCA_THRD));
		phy_writel(phy, 0x2100070E, PHY_MODEM(PHY_SYNC_LLTF_DET));
		phy_writel(phy, 0x09800032, PHY_MODEM(PHY_CS_CONFIG_1));
		phy_writel(phy, 0x09800032, PHY_MODEM(PHY_CS_CONFIG_2));
	} else if (bw == CBW40) {
		phy_writel(phy, 0x015E0172, PHY_MODEM(PHY_AGC_TARGET_PWR));
		phy_writel(phy, 0x04C2B0AE, PHY_MODEM(PHY_CCA_THRD));
		phy_writel(phy, 0x2100070E, PHY_MODEM(PHY_SYNC_LLTF_DET));
		phy_writel(phy, 0x09800032, PHY_MODEM(PHY_CS_CONFIG_1));
		phy_writel(phy, 0x09800032, PHY_MODEM(PHY_CS_CONFIG_2));
	} else {
		phy_writel(phy, 0x015E0172, PHY_MODEM(PHY_AGC_TARGET_PWR));
		phy_writel(phy, 0x00F6B0AE, PHY_MODEM(PHY_CCA_THRD));
		phy_writel(phy, 0x2100020E, PHY_MODEM(PHY_SYNC_LLTF_DET));
		phy_writel(phy, 0x09800014, PHY_MODEM(PHY_CS_CONFIG_1));
		phy_writel(phy, 0x09800014, PHY_MODEM(PHY_CS_CONFIG_2));
	}

#ifdef SCM80211_FTM
	scm2020_phy_fdscl_adj(phy, 1);
#else
	scm2020_phy_fdscl_adj(phy, 0);
#endif
}

void scm2020_phy_set_channel(struct scm80211_phy *phy, int vidx, struct phy_vector *pv)
{
	u32 freq;
	u32 conv_ratio[2], conv_11b_ratio;
	u32 v, macver, phyver;
	u32 oldband;
	u8 td_comp = 0;		/* Time domain csfo compensation enable from v0.7.3 */

	macver = scm80211_mac_version(phy->hw);
	phyver = scm80211_phy_version(phy->hw);

	/* Time domain csfo compensation enable from v0.7.3 */
	if (is_chip(phyver) || version(phyver) >= 13)
		td_comp = 1;

	freq = pv->CENTER_FREQUENCY;

	/* Time domain csfo compensation enable from v0.7.3 */
	conv_11b_ratio = ((5 * 1048576) + freq / 2) / freq;
	conv_ratio[0] = (20 * 262144 + freq / 2) / freq | (td_comp << 16);
	conv_ratio[1] = (20 * 4194304 + freq / 2) / freq;

	oldband = (phy_readl(phy, FW_SIGEXT_VINTF(vidx)) == 0);

	if (pv->BAND == 0) {
		/* 2.4GHz */
		if (!is_chip(phyver) && version(phyver) == 5) {
			/* old sync value for v0.5.0 */
			phy_writel(phy, 0x21000a16, PHY_MODEM(PHY_SYNC_LLTF_DET));
		}
		if (!is_chip(phyver) && version(phyver) < 5)
			conv_11b_ratio = 0x7ff;

		/* Signal extension */
		phy_writel(phy, 1, FW_SIGEXT_VINTF(vidx));

		if (!is_chip(phyver)) {
			/* PA 2.4G mode enable */
			/* 2G PA En    - (ctrl[1]), If 5G mode */
			phy_writel(phy, RFCTRL(1, HIGH, TXVALID, TXVALID, 0, 0),
				   RFI_CTRLSIG_2G_PA_EN);
			/* 5G PA En    - (ctrl[2]), If 2G mode */
			phy_writel(phy, RFCTRL(1, HIGH, TXVALID, ALWAYS, 0, 0),
				   RFI_CTRLSIG_5G_PA_EN);
		}
		phy_writel(phy, conv_11b_ratio, PHY_MODEM(FW_11B_CSFO_CONV_VINTF0));

	} else {
		/* 5GHZ */
		if (!is_chip(phyver) && version(phyver) == 5) {
			/* new sync value for v0.5.0 */
			phy_writel(phy, 0x21000a0e, PHY_MODEM(PHY_SYNC_LLTF_DET));
		}

		/* Signal extension */
		phy_writel(phy, 0, FW_SIGEXT_VINTF(vidx));

		if (!is_chip(phyver)) {
			/* PA 5G mode enable */
			/* 2G PA En    - (ctrl[1]), If 5G mode */
			phy_writel(phy, RFCTRL(1, HIGH, TXVALID, ALWAYS, 0, 0),
				   RFI_CTRLSIG_2G_PA_EN);

			/* 5G PA En    - (ctrl[2]), If 2G mode */
			phy_writel(phy, RFCTRL(1, HIGH, TXVALID, TXVALID, 0, 0),
				   RFI_CTRLSIG_5G_PA_EN);
		}
	}

	phy_writel(phy, conv_ratio[0], PHY_MODEM(PHY_SYNC_CCFO_CSFO));
	phy_writel(phy, conv_ratio[1], PHY_MODEM(PHY_PHAMP_CONV));
	if (pv->BAND == 0) {
		/* set init_wait_t = 7.5us in 2G */
		phy_writel(phy, 0x002800c8, PHY_MODEM(PHY_AGC_WAIT_CNT0));
	} else {
		/* set init_wait_t = 1us in 5G */
		phy_writel(phy, 0x00280028, PHY_MODEM(PHY_AGC_WAIT_CNT0));
	}
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	if (oldband != pv->BAND)
		scm2020_phy_rtx_lut(phy, freq);
#endif

	if (dev_is_platform(wiphy_dev(phy->hw->wiphy))) {
		u32 channel;

		/* Fictitous hwsim RF */
		channel = ieee80211_frequency_to_channel(freq);
		scm80211_rf_writel(phy->hw, channel, 0x0);
	}
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	/* RFI/ADC interface disable */
	phy_writel(phy, 0x10, PHY_RFIF(FW_RFI_EN));
	phy_writel(phy, 0x002B0000, PHY_RFIF(FW_RFI_CTRLSIG_CFG_07));
#endif

#ifdef SCM2020_CFO
	/* SCFO value init */
	phy_writel(phy, 0x00000000, PHY_MODEM(PHY_PHAMP_PLLADJ));
	phy_writel(phy, 0x00000000, PHY_MODEM(FW_11B_PLLADJ_VINTF0));
	phy_writel(phy, 0x00000000, PHY_MODEM(FW_11B_PLLADJ_VINTF1));
	phy_writel(phy, 0x00000000, PHY_MODEM(FW_11B_TX_PLLADJ_VINTF0));
	phy_writel(phy, 0x00000000, PHY_MODEM(FW_11B_TX_PLLADJ_VINTF1));
	phy_writel(phy, 0x00000000, PHY_MODEM(PHY_LOADJ_CSFO_TX));
	phy_writel(phy, 0x00000000, PHY_MODEM(PHY_LOADJ_CSFO_RX));
	_scfo_freq_hz = 0;
#endif

	/* OPBW and PRICH */
	v = (FIELD_PREP(FW_OPBW_VINTF0_MASK, pv->BANDWIDTH) |
	     FIELD_PREP(FW_PRICH_VINTF0_MASK, pv->PRIMARY_20MHZ_CHANNEL_LOC));
	phy_writel(phy, v, PHY_MODEM(FW_OPBW_PRICH_VINTF(vidx)));
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	/* RFI Enable */
	phy_writel(phy, 0x00, PHY_RFIF(FW_RFI_EN));
#endif

	scm2020_phy_set_bandwidth(phy, pv->BANDWIDTH);
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	/* PHY SW reset after MIMO config */
	udelay(500);		// PLL lock time delay
#endif

	if (is_chip(phyver))
		scm2020_phy_trx_set_boost(phy, 0);

#ifdef SCM2020_A0
	iq_read_ch[phy->chain] = phy_readl(phy, PHY_RFIF(PHY_RFI_RXIQ_HWCAL_RES));
#endif
	/* If mimo mode on, only operate the primary phy */
	if (phy_readl(phy, PHY_MODEM(FW_MIMO_EN_VINTF(vidx))) == 0x01)
		phy_writel(phy - phy->chain, 0x1, PHY_MODEM(FW_PHY_SW_RESET));
	else
		phy_writel(phy, 0x1, PHY_MODEM(FW_PHY_SW_RESET));

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	//scm2020_phy_apply_calibration(phy, NULL, freq);
	if (phy->chain == 1)
		scm2020_phy_rtx_lut(phy, freq);
	else if (phy->chain == 0 && !(phy + 1)->active)
		scm2020_phy_rtx_lut(phy, freq);

#else
	/* ADC Enable */
	phy_writel(phy, 0x002E0000, PHY_RFIF(FW_RFI_CTRLSIG_CFG_07));
#endif
	if (pv->BANDWIDTH == CBW20)
		scm80211_reconfig_hpf_setting(phy);
	else
		scm80211_restore_hpf_setting(phy);
}

#define CENTER_FREQ_MASK	0xFFFF
#define PRIMARY_LOC_MASK	0xF0000
#define	BANDWIDTH_MASK		0x300000
#define MIMO_ENABLE_MASK	0xC00000
static void scm2020_phy_config(struct scm80211_phy *phy, struct phy_vector *pv,
			       unsigned long changes)
{
	int p = 0, rssi, vidx = 0;
	u32 v;

	for_each_set_bit(p, &changes, PHY_PARAM_MAX) {
		WARN_ON(!pv->present);
		switch (p) {
		case PHY_CHANNEL:
			{
				scm80211_set_channel(phy->hw, phy, pv);
				break;
			}
		case PHY_ACTIVE_RXCHAIN_SET:
			{
				bool mimo = pv->MIMO_EN;
				struct ieee80211_hw_priv *priv = phy->hw->priv;

				if (priv->dbdc)
					BUG_ON(mimo);

				v = phy_readl(phy, PHY_MODEM(FW_MIMO_EN_VINTF(vidx)));
				if (v != mimo) {
					struct scm80211_phy *phy0 = phy - phy->chain;

#ifdef SCM2020_A0
					iq_read_ch[0] =
					    phy_readl(phy0, PHY_RFIF(PHY_RFI_RXIQ_HWCAL_RES));
					if (phy != phy0) {
						iq_read_ch[1] = phy_readl(phy,
									  PHY_RFIF
									  (PHY_RFI_RXIQ_HWCAL_RES));
					}
#endif

					phy_writel(phy, mimo, PHY_MODEM(FW_MIMO_EN_VINTF(vidx)));
					if (mimo)
						phy_writel(phy0, 0x1, PHY_MODEM(FW_PHY_SW_RESET));
					else
						phy_writel(phy, 0x1, PHY_MODEM(FW_PHY_SW_RESET));
#ifdef SCM2020_A0
					udelay(500);
					phy_writel(phy0, 0x0, PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
					phy_writel(phy0, scm2020_iq_get(phy0),
						   PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG1));
					phy_writel(phy0, 0x11, PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
					if (phy0 != phy) {
						phy_writel(phy, 0x0,
							   PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
						phy_writel(phy, scm2020_iq_get(phy),
							   PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG1));
						phy_writel(phy, 0x11,
							   PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
					}
#else
					phy_writel(phy0, 0x0, PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
					phy_writel(phy0, 0x1, PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
					if (phy0 != phy) {
						phy_writel(phy, 0x0,
							   PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
						phy_writel(phy, 0x1,
							   PHY_RFIF(PHY_RFI_RXIQ_HWCAL_CFG0));
					}
#endif
				}
				break;
			}
		case PHY_ANT_CONFIG:
			{
				/* FIXME: when we have PHY/RF matrix */
				break;
			}
		case PHY_RSSI_THRESHOLD:
			{
				rssi = min(-82, pv->RSSI_THRESHOLD);

				v = phy_readl(phy, PHY_MODEM(PHY_11B_CS_PARAM));
				bfmod_p(v, PHY_11B_CS_RSSI_THR, (u8)rssi);
				phy_writel(phy, v, PHY_MODEM(PHY_11B_CS_PARAM));

				v = phy_readl(phy, PHY_MODEM(PHY_CS_CONFIG_1));
				bfmod_p(v, PHY_CS_XCR_MINRSSI, (u8)rssi);
				phy_writel(phy, v, PHY_MODEM(PHY_CS_CONFIG_1));

				v = phy_readl(phy, PHY_MODEM(PHY_CS_CONFIG_2));
				bfmod_p(v, PHY_CS_ACR_MINRSSI, (u8)rssi);
				phy_writel(phy, v, PHY_MODEM(PHY_CS_CONFIG_2));

				break;
			}
		case PHY_LISTEN_TO_GID00:
			{
				v = phy_readl(phy, PHY_MODEM(FW_VHT_CONFIG0_VINTF(vidx)));
				v &= ~FW_VHT_LISTEN_TO_GID00_VINTF0_MASK;
				v |= FIELD_PREP(FW_VHT_LISTEN_TO_GID00_VINTF0_MASK,
						pv->LISTEN_TO_GID00);
				phy_writel(phy, v, PHY_MODEM(FW_VHT_CONFIG0_VINTF(vidx)));
				break;
			}
		case PHY_LISTEN_TO_GID63:
			{
				v = phy_readl(phy, PHY_MODEM(FW_VHT_CONFIG0_VINTF(vidx)));
				v &= ~FW_VHT_LISTEN_TO_GID63_VINTF0_MASK;
				v |= FIELD_PREP(FW_VHT_LISTEN_TO_GID63_VINTF0_MASK,
						pv->LISTEN_TO_GID63);
				phy_writel(phy, v, PHY_MODEM(FW_VHT_CONFIG0_VINTF(vidx)));
				break;
			}
		case PHY_PARTIAL_AID_LIST_GID00:
			{
				v = pv->PARTIAL_AID_LIST_GID00;
				phy_writel(phy, v, PHY_MODEM(FW_VHT_CONFIG1_VINTF(vidx)));
				break;
			}
		case PHY_PARTIAL_AID_LIST_GID63:
			{
				v = pv->PARTIAL_AID_LIST_GID63;
				phy_writel(phy, v, PHY_MODEM(FW_VHT_CONFIG2_VINTF(vidx)));
				break;
			}
		case PHY_VHT_SU_RX_FILTER_EN:
			{
				v = phy_readl(phy, PHY_MODEM(FW_VHT_CONFIG0_VINTF(vidx)));
				v &= ~FW_VHT_FLTRDO_ON_VINTF0_MASK;
				v |= FIELD_PREP(FW_VHT_FLTRDO_ON_VINTF0_MASK, v);
				phy_writel(phy, v, PHY_MODEM(FW_VHT_CONFIG0_VINTF(vidx)));
				break;
			}
		case PHY_GROUP_ID_MANAGEMENT:
			{
				v = get_unaligned_le32(&pv->GROUP_ID_MANAGEMENT.membership[0]);
				phy_writel(phy, v, FW_VHT_CONFIG3_VINTF(vidx));
				v = get_unaligned_le32(&pv->GROUP_ID_MANAGEMENT.membership[4]);
				phy_writel(phy, v, FW_VHT_CONFIG4_VINTF(vidx));

				v = get_unaligned_le32(&pv->GROUP_ID_MANAGEMENT.position[0]);
				phy_writel(phy, v, FW_VHT_CONFIG5_VINTF(vidx));
				v = get_unaligned_le32(&pv->GROUP_ID_MANAGEMENT.position[4]);
				phy_writel(phy, v, FW_VHT_CONFIG6_VINTF(vidx));
				v = get_unaligned_le32(&pv->GROUP_ID_MANAGEMENT.position[8]);
				phy_writel(phy, v, FW_VHT_CONFIG7_VINTF(vidx));
				v = get_unaligned_le32(&pv->GROUP_ID_MANAGEMENT.position[12]);
				phy_writel(phy, v, FW_VHT_CONFIG8_VINTF(vidx));
				break;
			}
		case PHY_BSS_COLOR:
			{
				v = phy_readl(phy, FW_HE_CONFIG2_VINTF(vidx));
				v &= ~FW_HESIGA_BSSCOLOR_VINTF0_MASK;
				v |= FIELD_PREP(FW_HESIGA_BSSCOLOR_VINTF0_MASK, pv->BSS_COLOR);
				phy_writel(phy, v, FW_HE_CONFIG2_VINTF(vidx));
				break;
			}
		case PHY_STA_ID_INDIVIDUAL:
			{
				v = phy_readl(phy, FW_HE_CONFIG2_VINTF(vidx));
				v &= ~FW_HESIGB_MYSTAID_VINTF0_MASK;
				v |= FIELD_PREP(FW_HESIGB_MYSTAID_VINTF0_MASK,
						pv->STA_ID_INDIVIDUAL);
				phy_writel(phy, v, FW_HE_CONFIG2_VINTF(vidx));
				break;
			}
		case PHY_STA_ID_GROUP:
			{
				v = phy_readl(phy, FW_HE_CONFIG2_VINTF(vidx));
				v &= ~FW_HESIGB_BSSIDX_VINTF0_MASK;
				v |= FIELD_PREP(FW_HESIGB_BSSIDX_VINTF0_MASK, pv->STA_ID_GROUP);
				phy_writel(phy, v, FW_HE_CONFIG2_VINTF(vidx));
				break;
			}
		case PHY_BSS_COLOR_FILTER_EN:
			{
				v = phy_readl(phy, FW_HE_CONFIG1_VINTF(vidx));
				v &= ~FW_HESIGA_FLTRDO_ON_VINTF0_MASK;
				v |= FIELD_PREP(FW_HESIGA_FLTRDO_ON_VINTF0_MASK,
						pv->BSS_COLOR_FILTER_EN);
				phy_writel(phy, v, FW_HE_CONFIG1_VINTF(vidx));
				break;
			}
		case PHY_HE_DIR:
			{
				v = phy_readl(phy, FW_HE_CONFIG1_VINTF(vidx));
				v &= ~FW_HESIGA_UL_FLTRDO_ON_VINTF0_MASK;
				v &= ~FW_HESIGA_DL_FLTRDO_ON_VINTF0_MASK;
				v |= FIELD_PREP(FW_HESIGA_UL_FLTRDO_ON_VINTF0_MASK,
						(pv->HE_DIR_EN & PHY_HE_DIR_UL) ? 0 : 1);
				v |= FIELD_PREP(FW_HESIGA_DL_FLTRDO_ON_VINTF0_MASK,
						(pv->HE_DIR_EN & PHY_HE_DIR_DL) ? 0 : 1);
				phy_writel(phy, v, FW_HE_CONFIG1_VINTF(vidx));
				break;
			}
		default:
			pr_debug("unsupported parameter %d\n", p);
			break;
		}
	}
}

void scm2020_phy_tx_mode(struct scm80211_phy *phy, int band, u8 mode, u8 lut_idx)
{
	struct rf_lut *lut = phy->radio->rf_lut;

	switch (mode) {
	case 0:		/* manual input mode */
		phy_writel(phy, 0x10 | (lut_idx << 8), PHY_RFIF(FW_TXLUT_CFG0));
		break;
	case 1:		/* manual output mode */
		phy_writel(phy, 0x1, PHY_RFIF(FW_TXLUT_CFG0));
		phy_writel(phy, lut->tx_lut[band][lut_idx], PHY_RFIF(FW_TXLUT_MANSET_TXGAIN_OUT));
		break;
	default:		/* normal mode */
		{
			u32 v = phy_readl(phy, PHY_RFIF(FW_RFI_EN)) & ~(0xff);

			phy_writel(phy, v, PHY_RFIF(FW_RFI_EN));
			phy_writel(phy, 0x0, PHY_RFIF(FW_TXLUT_CFG0));
		}
		break;
	}
}

void scm2020_phy_rx_mode(struct scm80211_phy *phy, int band, u8 mode, u8 lut_idx)
{
	struct rf_lut *lut = phy->radio->rf_lut;

	switch (mode) {
	case 0:		/* manual input mode */
		phy_writel(phy, 0x10 | (lut_idx << 8), PHY_RFIF(FW_RXLUT_CFG0));
		break;
	case 1:		/* manual output mode */
		phy_writel(phy, 0x1, PHY_RFIF(FW_RXLUT_CFG0));
		phy_writel(phy, lut->rx_lut[band][lut_idx], PHY_RFIF(FW_RXLUT_MANSET_RXGAIN_OUT));
		break;
	default:		/* normal mode */
		{
			u32 v = phy_readl(phy, PHY_RFIF(FW_RFI_EN)) & ~(0xff);

			phy_writel(phy, v, PHY_RFIF(FW_RFI_EN));
			phy_writel(phy, 0x0, PHY_RFIF(FW_RXLUT_CFG0));
		}
		break;
	}
}

void scm2020_phy_auto_mode(struct scm80211_phy *phy)
{
	/* PHY control mode */
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	u32 v = phy_readl(phy, PHY_RFIF(FW_RFI_EN)) & ~(0xff);

	phy_writel(phy, v, PHY_RFIF(FW_RFI_EN));
#else
	phy_writel(phy, 0x00000000, PHY_RFIF(FW_RFI_EN));
#endif
	/* TX tone_gen path disable */
	phy_writel(phy, 0x00000000, PHY_RFIF(PHY_RFI_TX_SIGGEN_CFG));
}

#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
static void scm2020_phy_fixup(struct scm80211_phy *phy)
{
	u32 macver, phyver;

	macver = scm80211_mac_version(phy->hw);
	phyver = scm80211_phy_version(phy->hw);

	/*
	 * In FPGA version, the clock speed of LDPC decoder is slower than
	 * chip version. (320/640M vs 160M).
	 * The latency of LDPC decoder may have caused the "LDPC Rx ACK Tx fail"
	 * issues. So the number of iteration should be adjusted in FPGA version.
	 */
	if (!is_chip(phyver) && version(phyver) >= 12) {
		phy_writel(phy, 0x02020302, PHY_MODEM(FW_LDPCDEC_ITER));
		phy_writel(phy, 0x02020202, PHY_MODEM(FW_LDPCDEC_STBC_ITER));
	} else if (!is_chip(phyver)) {
		phy_writel(phy, 0x05030504, PHY_MODEM(FW_LDPCDEC_ITER));
		phy_writel(phy, 0x05030503, PHY_MODEM(FW_LDPCDEC_STBC_ITER));
	}

	/*
	 * Temporary patch to clear the issue related to 11B TX :
	 * Change RFI_CTRLSIG_CFG:End delay for ExtSwitch Tx, Tx Enable,
	 * DAC control, and TX_EN to +1.6us based on Tommy's test results
	 * (This is a temporary patch for MAC version earlier than v0.65)
	 */
	if (!is_chip(phyver) && version(macver) <= 1) {
		phy_writel(phy, 0x00200040, RFI_CTRLSIG_EXT_SW_TX);
		phy_writel(phy, 0x00200040, RFI_CTRLSIG_RF_TX_EN);
		phy_writel(phy, 0x00200140, RFI_CTRLSIG_DAC);
		phy_writel(phy, 0x00200040, RFI_CTRLSIG_TX_EN);
	}

	/* Improving PER performance of NonHT 2020.07.13 */
	if (!is_chip(phyver) && version(phyver) >= 7) {
		/* The weight value for low SNR detection */
		phy_writel(phy, 0x000000B5, PHY_MODEM(PHY_SNRE_LOWSNR_WGT));
		/* LLTF detection threshold */
		phy_writel(phy, 0x21000A0F, PHY_MODEM(PHY_SYNC_LLTF_DET));
	}
	/* V0.6.1 REG Default Config */
	if (!is_chip(phyver) && version(phyver) >= 10) {
		/* RIFS Config */
		phy_writel(phy, 0x00000002, PHY_MODEM(PHY_COSIM_MODE));
		/* STR_PKT config */
		phy_writel(phy, 0x00000E21, PHY_MODEM(PHY_STR_PKT_ON));
		/* Init gain sattling time tuning */
		phy_writel(phy, 0x00500050, PHY_MODEM(PHY_AGC_WAIT_CNT0));
		/* AGC debug config default setting */
		phy_writel(phy, 0x00007B11, PHY_MODEM(PHY_AGC_TEST_OPTION));
	}

	/* 11B stuck prevention */
	if (!is_chip(phyver) && version(phyver) == 6)
		phy_writel(phy, 0x00780028, PHY_MODEM(PHY_AGC_WAIT_CNT0));

	/* ADV RF update */
	if (scm80211_is_adveos(phy)) {
		scm80211_trace(phy->hw, "fixup for Adveos RF-%c(phy=%x,mac=%x)\n", 'A' + phy->chain,
			       phyver, macver);
		phy_writel(phy, 0x000000A8, PHY_RFIF(PHY_RFI_RX_DIG_GAIN_MAN));
		/* Init gain sattling time tuning */
		phy_writel(phy, 0x00500050, PHY_MODEM(PHY_AGC_WAIT_CNT0));
		/* Init gain sattling time tuning */
		phy_writel(phy, 0x00300030, PHY_MODEM(PHY_AGC_WAIT_CNT1));
		/* Init gain sattling time tuning */
		phy_writel(phy, 0x002c002c, PHY_MODEM(PHY_AGC_WAIT_CNT2));
		/* Init gain sattling time tuning */
		phy_writel(phy, 0x00300030, PHY_MODEM(PHY_AGC_WAIT_CNT3));
		/* AGC gain step tuning */
		phy_writel(phy, 0x20202020, PHY_MODEM(PHY_AGC_GDOWN_CFG));
		/* Init AGC gain tuning */
		phy_writel(phy, 0x0000504c, PHY_MODEM(PHY_AGC_INDEX_CFG));
		/* Fixed init config */
		phy_writel(phy, 0x00000001, PHY_MODEM(PHY_AGC_MODE_CFG));
		/* Digital HPF */
		phy_writel(phy, 0x0001AAAA, PHY_RFIF(FW_RFI_RXDC_HPF_CFG));
		/* ADC IQ swap */
		phy_writel(phy, 0x00000021, PHY_RFIF(FW_RFI_RXADC_INTF));
	}

	if (!is_chip(phyver) && version(phyver) >= 22)
		phy_writel(phy, 0x397c2288, PHY_MODEM(PHY_PHAMP_CFFS));
}
#endif

void scm2020_phy_apply_calibration(struct scm80211_phy *phy, struct rf_cal_data *cal, int freq)
{
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	/* update calibration val for PATH-A and PATH-B both if mimo on */
	if (phy_readl(phy, PHY_MODEM(FW_MIMO_EN_VINTF(0))) == 0x01 && phy->chain == 1) {
		struct scm80211_phy *phy0 = phy - phy->chain;

		phy->radio->ops->get_cal_data(phy0->radio, cal, freq);
	}
	phy->radio->ops->get_cal_data(phy->radio, cal, freq);
#else
	u32 rx_iq, rx_dc, tx_iq, tx_dc, rx_gd;
	u8 path = phy->chain, i = 0;

#define mk32(cal, i, q) ((cal->i[path] & 0xffff) << 16 | (cal->q[path] & 0xffff))

	tx_iq = mk32(cal, tx_i_offset, tx_q_offset);
	tx_dc = mk32(cal, tx_i_dc_offset, tx_q_dc_offset);
	rx_iq = mk32(cal, rx_i_offset, rx_q_offset);
	rx_dc = mk32(cal, rx_i_dc_offset, rx_q_dc_offset);
	rx_gd = ((cal->rx_i_gd_offset[path] & 0x1ff) << 12) |
	    ((cal->rx_q_gd_offset[path] & 0x1ff)) | (1 << 22);

	phy_writel(phy, 1, PHY_RFIF(FW_RFI_TX_LO_LEAK_MAN));
	phy_writel(phy, 1, PHY_RFIF(FW_RFI_RX_DC_MAN));

	/* FW_RFI_TX_IQ_MIS_COMP@0214 */
	phy_writel(phy, tx_iq, PHY_RFIF(FW_RFI_TX_IQ_MIS_COMP));

	/* FW_RFI_TX_LO_LEAK_PARA_COMP@021c */
	phy_writel(phy, tx_dc, PHY_RFIF(FW_RFI_TX_LO_LEAK_PARA));

	/* FW_RFI_RX_IQ_MIS_COMP@0514 */
	phy_writel(phy, rx_iq, PHY_RFIF(FW_RFI_RX_IQ_MIS_COMP));

	/* FW_RFI_RX_DC_PARA@051c */
	phy_writel(phy, rx_dc, PHY_RFIF(FW_RFI_RX_DC_PARA));

	/* FW_TXLUT_MANTSET_LO@0x1004 */
	if (!(cal->flags & RFCAL_SKIP_TX_LO_COMP))
		phy_writel(phy, tx_dc, PHY_RFIF(FW_TXLUT_MANSET_LO));

	/* @0804 ? */
	/*phy_writel(phy, v, PHY_RFIF(FW_RXLUT_MANSET_DC)); */
	phy_writel(phy, rx_dc, PHY_RFIF(0x0804));

	/* FW_LO_COMP_LUT_00 - FW_LO_COMP_LUT_63 */
	if (!(cal->flags & RFCAL_SKIP_TX_LO_COMP)) {
		phy_writel(phy, BIT(16), PHY_RFIF(FW_TXLUT_CFG0));
		for (i = 0; i < 64; i++)
			phy_writel(phy, tx_dc, PHY_RFIF(FW_LO_COMP_LUT_00 + i * 4));
		phy_writel(phy, 0, PHY_RFIF(FW_TXLUT_CFG0));
	}

	/* set TXPWR MINMAX reg */
	if (!scm80211_is_adveos(phy))
		phy_writel(phy, 0x000018D9, PHY_RFIF(PHY_RFI_TXPWR_MINMAX));

	if (cal->flags & RFCAL_RX_GDELAY)
		phy_writel(phy, rx_gd, PHY_RFIF(FW_RFI_RX_GDELAY_COMP));
#endif
}

void scm2020_phy_calibrate_auto(struct scm80211_phy *phy, struct rf_cal_data *data)
{
	/* TBD */
}

void scm2020_phy_calibrate_manual(struct scm80211_phy *phy, struct rf_cal_data *data)
{
	phy->radio->ops->get_cal_data(phy->radio, data, 2412);
}

void scm2020_phy_calibrate(struct scm80211_phy *phy)
{
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	struct rf_cal_data data;

	scm80211_info(phy->hw, "PHY-%c starts calibration\n", phy->chain + 'A');

	scm2020_phy_calibrate_manual(phy, &data);
	scm2020_phy_apply_calibration(phy, &data, 2412);
#endif
}

void scm2020_powersave(struct ieee80211_hw *hw, bool ps)
{
#if defined SCM2020_Z1 || defined SCM2020_A0
	struct scm80211_phy *phy;

	for_each_phy(phy, hw) {
		phy->radio->ops->power_save(phy->radio, 0, ps);
	}
#endif
}

void scm2020_set_phy_dig_gain(struct ieee80211_hw *hw, int channel, int power)
{
	struct scm80211_phy *phy;
	int band = channel <= 14 ? NL80211_BAND_2GHZ : NL80211_BAND_5GHZ;

	for_each_phy(phy, hw) {
		if (phy->radio->ops->set_dig_gain)
			phy->radio->ops->set_dig_gain(phy->radio, band, !!(power % 10));
	}
}

/**
 * scm2020_do_temp_comp()
 * hw:
 * do temperature compensation
 */
void scm2020_phy_temp_comp(struct ieee80211_hw *hw)
{
	struct scm80211_phy *phy;
	struct ieee80211_hw_priv *priv = hw->priv;
	int i = 0, diff;
	int temp[2] = {0, 0};
	int delay;

	for_each_phy(phy, hw) {
		if (phy->radio->ops->do_temp_comp)
			temp[i++] = phy->radio->ops->do_temp_comp(phy->radio);
	}

	diff = max(abs(temp[1] - priv->temperature[1]), abs(temp[0] - priv->temperature[0]));

	delay = priv->dynamic_temp_timer;
	if (!delay)
		delay = 2;

	if (diff > 10)
		ieee80211_queue_delayed_work(hw, &priv->temp_work, (HZ >> 1) * delay);
	else if (diff > 5)
		ieee80211_queue_delayed_work(hw, &priv->temp_work, HZ * delay);
	else
		ieee80211_queue_delayed_work(hw, &priv->temp_work, (HZ << 1) * delay);

	scm80211_trace(hw, "current temperature [%d], [%d]\n", temp[0], temp[1]);
	priv->temperature[0] = temp[0];
	priv->temperature[1] = temp[1];

}

#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
/**
 * DOC: scm2020 rfi spi
 *
 * scm2020 rfi spi is not a generic spi master controller, but is designed
 * to be compatible only with AD9963 and MAX2829.
 * We do very funny things here. Later when we have a generic SPI controller,
 * this routine will be revised.
 */
struct scm2020_rfi_spi {
	void __iomem *base;
	u32 start;
	struct scm80211_phy *phy;
	struct rf_gspi_cfg gspi;
};

static int scm2020_rfi_spi_setup(struct spi_device *spi)
{
	/* TBD */
	return 0;
}

static inline int scm2020_rfi_spi_wait_completion(struct scm2020_rfi_spi *ctx)
{
	struct scm80211_phy *phy = ctx->phy;
	u32 v, base = ctx->start;
	int retries = 10;

	if (ctx->gspi.wr_nbits)
		base += 0x08;
	else
		base += 0x18;

	udelay(10);

	while (--retries > 0) {
		v = phy_readl(phy, base);
		if (v == 0)
			return 0;
		udelay(10);
	}

	scm80211_warn(phy->hw, "spi transfer incomplete\n");
	return -EBUSY;
}

/*
 * read/write data via GSPI
 */
static int scm2020_gspi_write(struct scm2020_rfi_spi *ctx, unsigned int wdata)
{
	struct scm80211_phy *phy = ctx->phy;
	u32 base = ctx->start;
	int wr_nbits;

	wr_nbits = ctx->gspi.wr_nbits;

	if (ctx->gspi.flags & 1)
		wdata |= 1 << (wr_nbits - 1);

	phy_writel(phy, wr_nbits, base + 0x18);	/* WR_LEN */
	phy_writel(phy, wdata, base + 0x0c);	/* rw + n + addr + data */
	phy_writel(phy, 0x01, base + 0x04);	/* Run */

	return scm2020_rfi_spi_wait_completion(ctx);
}

static int scm2020_gspi_read(struct scm2020_rfi_spi *ctx, unsigned int wdata, unsigned int *rdata)
{
	struct scm80211_phy *phy = ctx->phy;
	u32 v, base = ctx->start;
	int ret, wr_nbits, rd_nbits;

	wr_nbits = ctx->gspi.wr_nbits;
	rd_nbits = ctx->gspi.rd_nbits;
	if (!rd_nbits)
		return -EINVAL;

	if (!(ctx->gspi.flags & 1))
		wdata |= 1 << (wr_nbits - 1);

	phy_writel(phy, rd_nbits, base + 0x18);	/* RD_LEN */
	phy_writel(phy, wdata, base + 0x0c);	/* rw + n + data + addr */
	phy_writel(phy, 0x01, base + 0x04);	/* Run */

	ret = scm2020_rfi_spi_wait_completion(ctx);

	v = phy_readl(phy, base + 0x10);
	if (rdata)
		*rdata = v;

	return ret;
}

#endif /* SCM2020_Z1 */

#ifdef MANGO_OLD_SPI
/**
 * scm2020_ad9963_read() - read ad9963 register
 * @base: ioremapped address of rf spi controller base register
 * @addr: ad9963 register
 * @val: read buffer
 *
 * Looks ugly. Will go away once we have generic spi master controller
 */
static int scm2020_ad9963_read(struct scm2020_rfi_spi *ctx, unsigned int addr, unsigned int *data)
{
	struct scm80211_phy *phy = ctx->phy;
	u32 v, base = ctx->start;
	int ret;

	/* CMD: byte count | addr | read */
	phy_writel(phy, (addr << 1) | 0x1, base + 0x4);
	/* RUN */
	phy_writel(phy, 0x1, base + 0x10);

	ret = scm2020_rfi_spi_wait_completion(ctx);

	/* RFE READ DATA */
	v = phy_readl(phy, base + 0x14) & 0xff;
	if (data)
		*data = v & 0xff;
	return ret;
}

static int scm2020_ad9963_write(struct scm2020_rfi_spi *ctx, unsigned int addr, unsigned int data)
{
	struct scm80211_phy *phy = ctx->phy;
	u32 base = ctx->start;

	/* CMD */
	phy_writel(phy, addr << 1, base + 0x4);
	/* RF WRITE DATA */
	phy_writel(phy, data & 0xff, base + 0x8);
	/* RUN */
	phy_writel(phy, 0x1, base + 0x10);

	return scm2020_rfi_spi_wait_completion(ctx);
}

static int scm2020_max2829_write(struct scm2020_rfi_spi *ctx, unsigned int addr, unsigned int data)
{
	struct scm80211_phy *phy = ctx->phy;
	u32 base = ctx->start;

	/* CMD */
	phy_writel(phy, addr << 1, base + 0x4);

	/* RF WRITE DATA */
	phy_writel(phy, data, base + 0xc);

	/* RUN */
	phy_writel(phy, 0x1, base + 0x10);

	return scm2020_rfi_spi_wait_completion(ctx);
}
#endif /*MANGO_OLD_SPI */

#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
static int scm2020_rfi_spi_transfer(struct spi_controller *ctlr, struct spi_message *msg)
{
	struct scm2020_rfi_spi *ctx = spi_controller_get_devdata(ctlr);
	struct spi_transfer *xfer;
#ifdef MANGO_OLD_SPI
	u32 addr, data;
#endif
	const u8 *tx;
	u8 *rx;
	int ret = 0;

	xfer = list_first_entry(&msg->transfers, struct spi_transfer, transfer_list);

	/* We only handle single transfer messages */
	WARN_ON(!spi_transfer_is_last(ctlr, xfer));

	reinit_completion(&ctlr->xfer_completion);
	tx = xfer->tx_buf;
	rx = xfer->rx_buf;

	if (ctx->gspi.wr_nbits) {
		u32 wdata;

		if (!tx || xfer->len != 4) {
			dev_warn(&msg->spi->dev, "invalid spi xfer(%p,%d)", tx, xfer->len);
			ret = -EINVAL;
			goto out;
		}
		wdata = *(u32 *)tx;
		if (wdata >> 31) {
			/* Read */
			wdata &= ~(1 << 31);
			ret = scm2020_gspi_read(ctx, wdata, (u32 *)rx);
		} else {
			/* Write */
			ret = scm2020_gspi_write(ctx, wdata);
		}
#ifdef MANGO_OLD_SPI
	} else if (!strcmp(msg->spi->modalias, "ad9963")) {	/* right? */
		/*
		 * Write: 1 transfer (tx: 3 byte, rx: null)
		 * Read: 1 transfers (tx: 3 byte: rx: 3 byte)
		 */
		if (!tx) {
			dev_warn(&msg->spi->dev, "invalid spi xfer (buffer null)\n");
			ret = -EINVAL;
			goto out;
		}
		if (xfer->len != 3) {
			dev_err(&msg->spi->dev, "wrong transfer size: %d\n", xfer->len);
			ret = -EINVAL;
			goto out;
		}
		addr = (tx[1]) & 0xff;
		if (tx[0] & 0x80) {
			/* Read */
			scm2020_ad9963_read(ctx, addr, &data);
			if (rx)
				rx[2] = data & 0xff;
		} else {
			/* Write */
			data = tx[2] & 0xff;
			scm2020_ad9963_write(ctx, addr, data);
		}
	} else if (!strcmp(msg->spi->modalias, "max2829")) {
		/*
		 * Write: 1 transfer (tx: 3 byte, rx: null)
		 * Read: not supported
		 */
		if (!tx) {
			dev_warn(&msg->spi->dev, "invalid spi xfer (tx buffer null)\n");
			ret = -EINVAL;
			goto out;
		}
		if (xfer->len != 3) {
			dev_err(&msg->spi->dev, "wrong transfer size: %d\n", xfer->len);
			ret = -EINVAL;
			goto out;
		}

		addr = tx[2] & 0xf;
		data = (((tx[2] >> 4) & 0xf) |	/* D[3:0] */
			((tx[1] << 4) & 0xff0) |	/* D[11:4] */
			((tx[0] << 12) & 0x3000));	/* D[13:12] */

		scm2020_max2829_write(ctx, addr, data);
#endif /*MANGO_OLD_SPI */
	} else {
		dev_err(&msg->spi->dev, "unknown driver %s", msg->spi->modalias);
		ret = -EINVAL;
	}

out:
	msg->actual_length += xfer->len;
	msg->status = ret;
	spi_res_release(ctlr, msg);
	spi_finalize_current_message(ctlr);
	return ret;
}

/**
 * scm2020_init_rfi_spi_master() - create a rfi spi master controller
 *
 * @dev: device parenting spi master.
 * @offset: rfi spi register offset from PHY base address
 *
 */
static struct spi_controller *scm2020_rfi_add_spi_master(struct scm80211_phy *phy,
							 struct device *dev, u32 offset,
							 struct rf_gspi_cfg *gspi)
{
	struct spi_controller *master;
	struct scm2020_rfi_spi *context;
	int err;

	master = spi_alloc_master(dev, sizeof(*context));
	if (!master) {
		dev_err(dev, "failed to alloc spi master\n");
		return NULL;
	}

	master->mode_bits = SPI_MODE_0;
	master->bus_num = -1;	/* Let the system decide it */
	master->num_chipselect = 1;
	master->setup = scm2020_rfi_spi_setup;
	master->transfer_one_message = scm2020_rfi_spi_transfer;

	context = spi_master_get_devdata(master);
	context->phy = phy;
	context->start = offset;
	if (gspi)
		context->gspi = *gspi;
	else
		context->gspi.wr_nbits = 0;
	err = devm_spi_register_controller(dev, master);
	if (err) {
		dev_err(dev, "failed to register spi master\n");
		goto fail;
	}
	dev_dbg(dev, "phy %d created spi@%08x (bus_num=%d)\n", phy->chain, offset, master->bus_num);

	return master;
fail:
	return NULL;
}

/**
 * scm2020_rfi_spi_connect() - create a spi device and connect it to spi master
 * @phy: phy
 * @info: spi board info
 *
 * Note:
 * We use the following convention:
 * @info->bus_num[15:8]: path, @info->bus_num[7:0]: spi master id
 *
 * This is for radio driver.
 */
static struct spi_device *scm2020_rfi_spi_connect(struct scm80211_phy *phy,
						  struct spi_board_info *info,
						  struct rf_gspi_cfg *gspi)
{
	struct spi_master *master;
	int id = info->bus_num & 0xff;
	u32 offset;

	if (id < 0 || id > phy->rfi.spi.n)
		return NULL;

	/* Populate SPI master controller */
	offset = phy->rfi.start + FW_RFSPI0_CONFIG + (id * 0x20);
	if (gspi && gspi->wr_nbits) {
		if (id < 2)
			phy_writel(phy, 0x8000000a, offset);
		offset = phy->rfi.start + FW_RF_GSPI0_CONFIG + (id * 0x30);
	}

	master = scm2020_rfi_add_spi_master(phy, wiphy_dev(phy->hw->wiphy), offset, gspi);
	if (!master) {
		scm80211_err(phy->hw, "failed to create phy %d spi %d\n", phy->chain, id);
		return NULL;
	}
	phy->rfi.spi.master[id] = master;

	if (gspi && gspi->wr_nbits) {
		phy_writel(phy, gspi->reg_val, offset);
#ifdef MANGO_OLD_SPI
	} else if (!strcmp(info->modalias, "ad9963")) {
		phy_writel(phy, 0x0a, offset + 0x0);
	} else if (!strcmp(info->modalias, "max2829")) {
		phy_writel(phy, 0x2a, offset + 0x0);
#endif
	} else {
		scm80211_info(phy->hw, "PHY-%c: %s.%d not supported\n",
			      phy->chain + 'A', info->modalias, id);
		return NULL;
	}

	/* Populate a SPI device on the created master and return it back */
	scm80211_info(phy->hw, "PHY-%c: %s.%d connected to spi%d\n",
		      phy->chain + 'A', info->modalias, id, master->bus_num);

	return spi_new_device(master, info);
}

#endif /* SCM2020_Z1 */

static int scm2020_phy_attach(struct scm80211_phy *phy, struct scm80211_radio *radio)
{
	/* cal? */
	return 0;
}

void scm2020_phy_detach(struct scm80211_phy *phy, struct scm80211_radio *radio)
{
}

int scm2020_phy_init(struct ieee80211_hw *hw, struct scm80211_phy *phy)
{
	int i;
	u32 v;
	bool is5g;

	/*
	 * Save power-on reset values for later use
	 */
	v = phy_readl(phy, PHY_MODEM(PHY_11B_CS_PARAM));
	cs.phy_11b_cs_rssi_thr = bfget_p(v, PHY_11B_CS_RSSI_THR);
	v = phy_readl(phy, PHY_MODEM(PHY_CS_CONFIG_1));
	cs.phy_cs_xcr_minrssi = bfget_p(v, PHY_CS_XCR_MINRSSI);
	v = phy_readl(phy, PHY_MODEM(PHY_CS_CONFIG_2));
	cs.phy_cs_acr_minrssi = bfget_p(v, PHY_CS_ACR_MINRSSI);

	/* Initialize Modem */
	scm80211_dbg(hw, "initializing PHY-%c", 'A' + phy->chain);
	for (i = 0; i < ARRAY_SIZE(phy_init); i++)
		phy_writel(phy, phy_init[i].value, phy_init[i].reg);

	scm80211_trace(hw, "initializing PHY-%c LUTs\n", 'A' + phy->chain);
	is5g = !phy_readl(phy, FW_SIGEXT_VINTF(0));
	scm2020_phy_auto_mode(phy);
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	scm2020_phy_rtx_lut(phy, is5g ? 5180 : 2412);
	scm80211_dbg(hw, "fixup PHY-%c\n", 'A' + phy->chain);
	scm2020_phy_fixup(phy);
#endif

	/* Enable 11b RSSI check */
	v = phy_readl(phy, PHY_MODEM(PHY_11B_CS_PARAM));
	bfmod_p(v, PHY_11B_CS_RSSICHK_OFF, 0);
	phy_writel(phy, v, PHY_MODEM(PHY_11B_CS_PARAM));
	return 0;
}

void scm2020_phy_deinit(struct ieee80211_hw *hw, struct scm80211_phy *phy)
{
}

static void scm2020_phy_enable(struct ieee80211_hw *hw, struct scm80211_phy *phy, bool enable)
{
	scm80211_trace(hw, "%s PHY %c\n",
		       enable ? "enabling" : "disabling", 'A' + phy->chain);

	if (enable) {
		scm2020_phy_init(hw, phy);
		scm80211_radio_power_up(phy->radio);
	} else {
		phy_writel(phy, 0x00000000, PHY_MODEM(FW_PWR_EN));
		scm80211_radio_power_down(phy->radio);
	}
}

/**
 * scm2020_gpio_control - Control GPIO value related to ADDA and RFIC
 * @lmask: reset bit mask
 * @hmask: set bit mask
 */
static void scm2020_gpio_control(struct scm80211_phy *phy, int id, u32 lmask, u32 hmask)
{
	u32 base, vmask, mask, data;

	if (id == 0) {
		/* JF80 GPIO: ADC_OEB[3], ADV_SPI_CLK_EN[2], DAC_RST[1], ADC_PWDN[0] */
		base = phy->rfi.start + 0x0770;
		mask = ~(phy->start ? 4 : 0);
		vmask = 0xf;
	} else {
		return;
	}

	phy_writel(phy, vmask & mask, base + 4);	/* OE */
	data = (phy_readl(phy, base + 8) & ~lmask) | hmask;
	phy_writel(phy, data & mask, base + 8);	/* DOUT */
	phy_writel(phy, vmask & mask, base + 0);	/* CTRL -> Update */
	phy_writel(phy, 0x00 & mask, base + 0);	/* CTRL -> Hold */
}

/**
 * PHY driver
 */
const struct scm80211_phy_ops scm2020_phy_ops = {
	/* RF interface */
#if !defined(SCM2020_Z1) && !defined(SCM2020_A0)
	.connect_spi = scm2020_rfi_spi_connect,
#endif
	.attach = scm2020_phy_attach,
	.detach = scm2020_phy_detach,
	.calibrate = scm2020_phy_calibrate,
	.enable = scm2020_phy_enable,
	.config = scm2020_phy_config,
	.control_gpio = scm2020_gpio_control,
};

struct scm80211_phy scm2020_phy[2] = {
	[0] = {
	       .chain = 0,
	       .ops = &scm2020_phy_ops,
	       .supported_bands = BIT(NL80211_BAND_2GHZ) | BIT(NL80211_BAND_5GHZ),
	       .start = PHY_BASE(0),
	       .rfi = {
		       .start = PHY_RFI_OFFSET,
		       .spi.n = 2,
		       },
		},
	[1] = {			/* This PHY supports only OFDM PHY */
	       .chain = 1,
	       .ops = &scm2020_phy_ops,
	       .supported_bands = BIT(NL80211_BAND_2GHZ) | BIT(NL80211_BAND_5GHZ),
	       .start = PHY_BASE(1),
	       .rfi = {
		       .start = PHY_RFI_OFFSET,
		       .spi.n = 2,
		       },
		},
};

int scm2020_chip_init(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct device *dev = wiphy_dev(hw->wiphy);
	int i;
	static struct scm80211_phy *phy;

	if (priv->dbdc) {
		if (!phy) {
			phy = devm_kmemdup(dev, scm2020_phy, sizeof(scm2020_phy), GFP_KERNEL);
			if (!phy)
				return -ENOMEM;
		}
		priv->phy = phy;
	} else {
		priv->phy = devm_kmemdup(dev, scm2020_phy, sizeof(scm2020_phy), GFP_KERNEL);
		if (!priv->phy)
			return -ENOMEM;
	}
	priv->n_chain = min_t(int, 2, (priv->dbdc ? 1 : n_rx));
	priv->chains = GENMASK(priv->n_chain, 0);
	if (priv->dbdc) {
		priv->phy[priv->devid].hw = hw;
		scm80211_trace(hw, "dbdc enabled, init phy for mac#%d\n", priv->devid);
	} else {
		/* One time initialization */
		for (i = 0; i < SCM80211_NUM_PHY; i++)
			priv->phy[i].hw = hw;
	}
	return 0;
}
