// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
#define DEBUG

#include <linux/rbtree.h>
#include <linux/firmware.h>
#include <linux/bitfield.h>
#include <linux/usb.h>

#define func_fmt(str)           "[USB] "str
#define func_trace              BUS

#include "core.h"
#include "rf.h"
#include "trace.h"
#include "auto_version.h"
#include "debug.h"
#include "hw.h"
#ifdef EFUSE_ENABLE
#include "efuse.h"
#endif

/* constants */
#define TS_URB_COUNT			16
#define RX_URB_COUNT			128
#define MAX_SG_NUM			(SCM80211_MAX_MPDU + 2)

#define SCM80211_USB_PIPE_INVALID SCM80211_USB_PIPE_MAX

/* USB endpoint definitions */
#define SCM80211_USB_EP_ADDR_CTRL_IN		0x80
#define SCM80211_USB_EP_ADDR_DATA_IN		0x81
#define SCM80211_USB_EP_ADDR_RSTATE_IN		0x82

#define SCM80211_USB_EP_ADDR_CTRL_OUT		0x00
#define SCM80211_USB_EP_ADDR_IS_DATA_OUT(addr)	(!((addr) & 0x80))

#define SCM80211_USB_IS_BULK_EP(attr)	(((attr) & 3) == 0x02)
#define SCM80211_USB_IS_INT_EP(attr)	(((attr) & 3) == 0x03)
#define SCM80211_USB_IS_ISOC_EP(attr)	(((attr) & 3) == 0x01)
#define SCM80211_USB_IS_DIR_IN(addr)	((addr) & 0x80)
#define SCM80211_USB_IS_BULK_OUT(attr, addr) \
					((((attr) & 3) == 0x02) && !((addr) & 0x80))
#define SCM80211_DATA_HWQ_MEM_LEN		0x10000	//64K

enum usb_vendor_request {
	REG_WRITE = 0,
	REG_READ = 1,
	PARAM_WRITE = 2,
	PARAM_READ = 3,
	FW_DOWNLOAD = 4,
	FW_STATUS = 5
};

enum scm80211_usb_pipe_id {
	SCM80211_USB_PIPE_TXQ0 = 0,
	SCM80211_USB_PIPE_TXQ1,
	/* SCM80211_USB_PIPE_TXQ2, */
	SCM80211_USB_PIPE_RX_DATA,
	SCM80211_USB_PIPE_RX_RSTATE,
	SCM80211_USB_PIPE_MAX,
};

#define SCM80211_USB_BUFF_NUM		0x04
struct scm80211_usb_tx_buff {
	u8 *data;
	struct list_head list;
	dma_addr_t dma;
};

#define NR_TXBUFF	4
#define THREAD_RUNNING	0
struct scm80211_usb_pipe {
	struct list_head urb_list_head;
	struct list_head urb_start_head;
	struct list_head urb_complete_head;
	struct usb_anchor urb_submitted;
	u32 urb_alloc;
	u32 urb_cnt;
	unsigned int usb_pipe_handle;
	u8 interval;
	u32 flags;
	u8 ep_address;
	u8 pipe_num;
	void *hw_priv;
	u16 max_packet_size;
	unsigned long	   state;
	struct work_struct rx_work;
	struct task_struct *rx_thread;
	/* protects pipe->urb_list_head and  pipe->urb_cnt */
	spinlock_t cs_lock;
	/* TX buff for tx ep which not supprt urb sg */
	struct list_head buff_head;
	struct scm80211_usb_tx_buff buff[NR_TXBUFF];
	struct usb_endpoint_descriptor *ep_desc;
};

#define SCM80211_USB_PIPE_FLAG_TX BIT(0)
#define SCM80211_USB_PIPE_FLAG_RX BIT(1)
#define SCM80211_USB_PIPE_FLAG_INIT BIT(2)

/* usb device object */
struct scm80211_usb_priv {
	struct usb_device		*udev;
	struct usb_interface		*interface;
	struct ieee80211_hw		*hw;
	struct scm80211_usb_pipe	pipes[SCM80211_USB_PIPE_MAX];
	u8				err_cnt[SCM80211_USB_PIPE_MAX];
	unsigned long			halt_bitmap;
	unsigned int			sg_size;
	spinlock_t			lock;
	struct list_head		tx_head;
	struct work_struct		usb_bus_monitor_work;
};

/* usb urb object */
struct scm80211_urb_context {
	struct list_head	 link;
	struct scm80211_usb_pipe *pipe;
	int			 idx;
	struct sk_buff		 *skb;
	struct sk_buff_head	 head;
	struct urb		 *urb;
	dma_addr_t		 dma;
	struct scm80211_tx_status_info *ts;
};

#define hw_priv_to_usb_priv(hw_priv)	((void *)((hw_priv)->bus.data))
#define hw_to_usb_priv(hw)	\
	((void *)(((struct ieee80211_hw_priv *)((hw)->priv))->bus.data))
#define MCU_START_RETRIES		10
#define MCU_START_DELAY			1

#define MAC_BASE_0			0x40200000
#define MAC_BASE_1			0x40210000
#define PHY_BASE_0			0x40220000

#define CTRL_MEM_SIZE			0x1000	//4K
#define SC2_MSG_SIZE			64
#define SC2_DATA_SIZE			(CTRL_MEM_SIZE - 64)
#define VERSION_MAX_LEN			64
#define SC2_CRC32_INIT			0xFFFFFFFFU

#define SCM80211_POOL_SIZE		1664
#define SCM80211_MAX_RX_SG		4

/**
 * scm80211_usb_alloc_urb_from_pipe() - get urb from urb pools
 */
struct scm80211_urb_context *scm80211_usb_alloc_urb_from_pipe(struct scm80211_usb_pipe *pipe)
{
	struct scm80211_urb_context *urb_context = NULL;
	unsigned long flags;

	spin_lock_irqsave(&pipe->cs_lock, flags);
	if (!list_empty(&pipe->urb_list_head)) {
		urb_context = list_first_entry(&pipe->urb_list_head,
					       struct scm80211_urb_context, link);
		list_del(&urb_context->link);
		pipe->urb_cnt--;
	}
	spin_unlock_irqrestore(&pipe->cs_lock, flags);

	return urb_context;
}

/**
 * scm80211_usb_free_urb_to_pipe() - put usb to pool
 */
static void scm80211_usb_free_urb_to_pipe(struct scm80211_usb_pipe *pipe,
					  struct scm80211_urb_context *urb_context)
{
	unsigned long flags;

	spin_lock_irqsave(&pipe->cs_lock, flags);

	pipe->urb_cnt++;
	list_add_tail(&urb_context->link, &pipe->urb_list_head);

	spin_unlock_irqrestore(&pipe->cs_lock, flags);
}

bool scm80211_usb_has_free_rx_urb(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	struct scm80211_usb_pipe *rxpipe = &usb_priv->pipes[SCM80211_USB_PIPE_RX_DATA];

	return !!READ_ONCE(rxpipe->urb_cnt);
}

void scm80211_usb_cleanup_recv_urb(struct ieee80211_hw *hw,
				   struct scm80211_urb_context *urb_context)
{
	struct sk_buff *skb = urb_context->skb;

	if (skb)
		dev_kfree_skb(skb);

	urb_context->skb = NULL;

	scm80211_usb_free_urb_to_pipe(urb_context->pipe, urb_context);
}

/**
 * scm80211_usb_free_pipe_resources() - free all the urb context resources
 */
static void scm80211_usb_free_pipe_resources(struct ieee80211_hw *hw,
					     struct scm80211_usb_pipe *pipe)
{
	struct scm80211_urb_context *urb_context;
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);

	if (pipe->urb_alloc != pipe->urb_cnt) {
		scm80211_trace(hw,
			       "usb urb leak pipe_n %d pipe_h 0x%x urbs %d avail %d\n",
			       pipe->pipe_num, pipe->usb_pipe_handle,
			       pipe->urb_alloc, pipe->urb_cnt);
	}

	if (usb_priv->sg_size == 0 && (pipe->flags & SCM80211_USB_PIPE_FLAG_TX)) {
		int i;

		for (i = 0; i < NR_TXBUFF; i++)
			usb_free_coherent(usb_priv->udev, SCM80211_TXOP_AMPDU_LEN,
					  pipe->buff[i].data,
					  pipe->buff[i].dma);
	}

	do {
		urb_context = scm80211_usb_alloc_urb_from_pipe(pipe);
		if (!urb_context)
			break;

		if (urb_context->ts) {
			usb_free_coherent(usb_priv->udev, sizeof(struct scm80211_tx_status_info),
					  urb_context->ts, urb_context->dma);
		}

		if (urb_context->skb)
			kfree_skb(urb_context->skb);

		if (urb_context->urb) {
			usb_kill_urb(urb_context->urb);
			usb_free_urb(urb_context->urb);
		}

		kfree(urb_context);
	} while (1);
}

static void scm80211_usb_cleanup_pipe_resources(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int i;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);

	for (i = 0; i < SCM80211_USB_PIPE_MAX; i++)
		scm80211_usb_free_pipe_resources(hw, &usb_priv->pipes[i]);
}

#ifdef FW_LOG_UPLOAD
static int fw_log_upload_proc(struct urb *urb)
{
	struct scm80211_urb_context *urb_context = urb->context;
	struct scm80211_usb_pipe *pipe = urb_context->pipe;
	struct ieee80211_hw_priv *priv = pipe->hw_priv;
	u8 upload_log_flag[8] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
	u32 len;
	u8 info[300];
	u8 *str_begin = ((u8 *)(urb_context->skb->data)) + sizeof(upload_log_flag);
	u8 *str_end = ((u8 *)(urb_context->skb->data)) + urb->actual_length;
	u8 *str_idx = str_begin;
	u8 *f_str = NULL;

	if (memcmp(upload_log_flag, urb_context->skb->data, sizeof(upload_log_flag)))
		return 0;

	wiphy_dbg(priv->hw->wiphy, "FW log upload below:\n");

	while (str_idx < str_end) {
		f_str = strchr(str_idx, '\n');
		if (f_str) {
			len = (f_str + 1) - str_idx;
		} else {
			wiphy_dbg(priv->hw->wiphy, "FW LOG : %s", str_idx);
			break;
		}

		if (len >= sizeof(info))
			len = sizeof(info) - 1;
		memcpy(info, str_idx, len);
		info[len] = 0;
		str_idx += len;
		wiphy_dbg(priv->hw->wiphy, "FW LOG : %s", info);
	}

	return 1;
}
#endif

static int scm80211_usb_bulk_rx(struct ieee80211_hw *hw);

/**
 * scm80211_usb_bulk_rx_complete() - complete handler used for RX packets EP
 *
 * (1) check urb status
 * (2) check rx actual_length
 * (3) set skb length and push it into rxskb list
 * (4) calling NAPI to handle rx skbs
 */
static void scm80211_usb_bulk_rx_complete(struct urb *urb)
{
	struct scm80211_urb_context *urb_context = urb->context;
	struct scm80211_usb_pipe *pipe = urb_context->pipe;
	struct ieee80211_hw_priv *priv = pipe->hw_priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	//struct scm80211_usb_pipe *rxpipe = &usb_priv->pipes[SCM80211_USB_PIPE_RX_DATA];
	//struct sk_buff *skb;
	int status = 0;
	unsigned long flags;

	urb_context->urb = usb_get_urb(urb);
#ifdef FW_LOG_UPLOAD
	if (fw_log_upload_proc(urb))
		urb->actual_length = 0;
#endif

	if (urb->status != 0) {
		status = -EIO;

		if (urb->status == -EPROTO || urb->status == -ETIMEDOUT) {
			unsigned long flags;

			spin_lock_irqsave(&usb_priv->lock, flags);
			if (usb_priv->err_cnt[pipe->pipe_num]++ > 1) {
				usb_priv->halt_bitmap |= BIT(pipe->pipe_num);
				ieee80211_queue_work(priv->hw, &usb_priv->usb_bus_monitor_work);
			}
			spin_unlock_irqrestore(&usb_priv->lock, flags);
		}

		switch (urb->status) {
		case -ECONNRESET:
		case -ENOENT:
		case -ESHUTDOWN:
			/* no need to spew these errors when device
			 * removed or urb killed due to driver shutdown
			 */
			status = -ECANCELED;
			break;
		default:
			scm80211_warn(priv->hw,
				      "usb recv [%02x] pipe %d ep 0x%2.2x failed: %d on transfer %p\n",
				      urb_context->idx,
				      pipe->pipe_num,
				      pipe->ep_address, urb->status,
				      urb_context->skb->data);
			break;
		}
		scm80211_usb_cleanup_recv_urb(priv->hw, urb_context);
	} else {
		spin_lock_irqsave(&pipe->cs_lock, flags);
		list_add_tail(&urb_context->link, &pipe->urb_complete_head);
		spin_unlock_irqrestore(&pipe->cs_lock, flags);
		if (!test_bit(THREAD_RUNNING, &pipe->state))
			wake_up_process(pipe->rx_thread);
	}
}

static int scm80211_fill_rx_urb(struct ieee80211_hw *hw, struct scm80211_urb_context *urb_context)
{
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *rxpipe = &usb_priv->pipes[SCM80211_USB_PIPE_RX_DATA];
	struct urb *urb = urb_context->urb;
	struct sk_buff *skb;
	int ret;

	urb_context->skb = scm80211_alloc_rx_skb(hw);
	if (!urb_context->skb)
		return -ENOMEM;

	skb = urb_context->skb;
	skb_reserve(skb, IEEE80211_RX_HEADROOM);
	usb_fill_bulk_urb(urb, usb_priv->udev, rxpipe->usb_pipe_handle,
			  skb->data, rx_buflen,
			  scm80211_usb_bulk_rx_complete, urb_context);
	urb->num_sgs    = 0;

	scm80211_trace(hw,
		       "rx bulk submit[%02x] pipe %d  ep 0x%2.2x len %u\n",
		       urb_context->idx, rxpipe->pipe_num,
		       rxpipe->ep_address, urb->transfer_buffer_length);

	usb_anchor_urb(urb, &rxpipe->urb_submitted);
	ret = usb_submit_urb(urb, GFP_ATOMIC);
	if (ret) {
		scm80211_warn(hw, "rx bulk submit failed: %d\n", ret);
		usb_unanchor_urb(urb);
	} else {
		usb_put_urb(urb);
	}
	return ret;
}

static void scm80211_usb_rx_process(struct scm80211_usb_pipe *pipe)
{
	struct scm80211_urb_context *urb_context;
	struct ieee80211_hw_priv *priv = pipe->hw_priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	struct urb *urb;
	struct sk_buff *skb;
	unsigned long flags;

	ieee80211_queue_work(priv->hw, &priv->rxring.work);
again:
	spin_lock_irqsave(&pipe->cs_lock, flags);
	urb_context = list_first_entry_or_null(&pipe->urb_complete_head,
					       struct scm80211_urb_context, link);
	if (!urb_context || WARN_ON(!urb_context->skb)) {
		spin_unlock_irqrestore(&pipe->cs_lock, flags);
		if (priv->napi.enabled)
			napi_schedule(&priv->napi.context);
		else
			scm80211_rx_poll(priv->hw, 128);
		return;
	}
	list_del_init(&urb_context->link);
	spin_unlock_irqrestore(&pipe->cs_lock, flags);

	urb = urb_context->urb;
	scm80211_trace(priv->hw,
		       "usb recv pipe %d stat %d len %d urb[%02x]\n",
		       pipe->pipe_num, urb->status,
		       urb->actual_length, urb_context->idx);

	if (urb->actual_length > 0) {
		usb_priv->err_cnt[pipe->pipe_num] = 0;
		skb = urb_context->skb;
		urb_context->skb = NULL;
		skb_put(skb, urb->actual_length);
		skb_queue_tail(&priv->rxring.rxskbs, skb);
	}

	if (scm80211_fill_rx_urb(priv->hw, urb_context))
		scm80211_usb_cleanup_recv_urb(priv->hw, urb_context);

	goto again;
}

static int scm80211_usb_rx_thread(void *data)
{
	struct scm80211_usb_pipe *pipe = data;

	while (!kthread_should_stop()) {
		set_current_state(TASK_INTERRUPTIBLE);

		if (kthread_should_park()) {
			kthread_parkme();
			continue;
		}

		if (list_empty(&pipe->urb_complete_head)) {
			schedule();
			continue;
		}

		set_bit(THREAD_RUNNING, &pipe->state);
		set_current_state(TASK_RUNNING);
		scm80211_usb_rx_process(pipe);
		cond_resched();
		clear_bit(THREAD_RUNNING, &pipe->state);
	}
	return 0;
}

/**
 * scm80211_usb_bulk_rx() - submit urb for RX packet EP
 */
static int scm80211_usb_bulk_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	struct scm80211_urb_context *urb_context;
	struct scm80211_usb_pipe *rxpipe = &usb_priv->pipes[SCM80211_USB_PIPE_RX_DATA];
	int usb_status;

	do {
		usb_status = -1;
		urb_context = scm80211_usb_alloc_urb_from_pipe(rxpipe);
		if (!urb_context)
			break;

		usb_status = scm80211_fill_rx_urb(hw, urb_context);
		if (usb_status)
			break;
	} while (1);

	if (usb_status && urb_context)
		scm80211_usb_cleanup_recv_urb(hw, urb_context);

	return usb_status;
}

static struct scm80211_tx_status_info fake_status;
static struct scm80211_tx_status_info __used *build_tx_status(int hwqid, int txopid)
{
	memset(&fake_status, 0, sizeof(fake_status));
	return &fake_status;
}

/**
 * scm80211_tx_complete() - tx complete handler for usb
 *
 * (1) sanity check the tx status info
 * (2) get issued txop and fill up tx status info for it
 * (3) calling common tx handler
 */
static void scm80211_tx_complete(struct ieee80211_hw *hw, struct scm80211_tx_status_info *status)
{
	struct scm80211_hwq *hwq = scm80211_hwq(hw, status->hwq_id);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_sta *sta;
	struct scm80211_tx_status_info *ts;
	struct txop *txop;
	int i, n, weight;
	unsigned long flags;

	scm80211_trace(hw, " - tx status for hwq#%d.%d (%s):\n",
		       status->hwq_id, status->txop_id,
		       status->info == SENT ? "normal" :
		       status->info == ABORT ? "tx abort" :
		       status->info == OUT_OF_WINDOW ? "out of window" :
		       status->info == UNDER_PS ? "under power save" :
		       "unsent for unknown reason");

	for (n = 0; n < IEEE80211_TX_MAX_RATES; n++) {
		scm80211_trace(hw, " - rate[%d]: idx=%d, flags=%08x, count=%d\n", n,
			       status->rates[n].idx,
			       status->rates[n].flags, status->rates[n].count);
	}
	scm80211_trace(hw, "  tb %d, psdu_len %u\n", status->tb, status->tb_psdu_len);
	scm80211_trace(hw, "  ack status 0x%llx\n", status->report);

	if (status->info == UNDER_PS && READ_ONCE(priv->pm.ps) != PS_ENABLED) {
		/* power state mismatch happened,force wake device up */
		scm80211_info(hw, "power state mismatch happened,force wake device up\n");
		priv->pm.force_awake = true;
		ieee80211_queue_work(priv->hw, &priv->pm.disable_work);
	}

	if (!hwq) {
		scm80211_err(hw, "hwq#%d in %s is NULL !!\n", status->hwq_id, __func__);
		return;
	}

again:
	spin_lock_irqsave(&hwq->ring.lock, flags);
	txop = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
	if (!txop) {
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		scm80211_info(hw, "%s(hwq#%d) bogus interrupt\n", __func__, hwq->id);
		return;
	}

	ts = status;
	if (txop->id != ts->txop_id) {
		scm80211_err(hw,
			     "txop from target hwq#%d.%d != hwq#%d.%d host want\n",
			     ts->hwq_id, ts->txop_id, hwq->id, txop->id);
		__scm80211_txop_add_back(txop);
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		return;
	}
	spin_unlock_irqrestore(&hwq->ring.lock, flags);

	BUG_ON(bitmap_weight((unsigned long *)&ts->report, 64) > txop->n_mpdu);
	weight = bitmap_weight((unsigned long *)&ts->report, txop->n_mpdu);

	txop->reduce = ts->reduce;
	txop->bitmap = ts->report;
	txop->scheduled = ts->info == SENT;
	txop->txd.tb = ts->tb;

	if (txop->bitmap)
		txop->txd.ok = 1;

	if (weight != txop->n_mpdu)
		txop->txd.nack = 1;

	for (i = 0; i < IEEE80211_TX_MAX_RATES; i++) {
		if (ts->rates[i].count > 0 && ts->rates[i].idx >= 0)
			txop->retry_cnt += ts->rates[i].count;
	}

	/* fillup mpdu desc */
	for (i = 0; i < txop->n_mpdu; i++) {
		struct tx_mpdu_desc *md = &txop->dma[0].desc->mdesc[i];

		if (test_bit(i, (void *)&txop->bitmap)) {
			md->sent = 1;
			md->acked = 1;
		} else if (txop->scheduled) {
			md->sent = 1;
		}
	}

	sta = txop->owner->txq.sta;
	if (sta && txop->scheduled) {
		txop->rs.status.ampdu_len = txop->n_mpdu;
		txop->rs.status.ampdu_ack_len = weight;
		scm80211_trace(hw, "- mpdu len %d acked len %d\n",
			       txop->rs.status.ampdu_len, weight);
		memcpy(txop->rs.status.rates, ts->rates,
		       sizeof(*ts->rates) * IEEE80211_TX_MAX_RATES);
	}

#ifdef SCM80211_WFA_TEST
	if (txop->vif && ts->tb && ts->tb_psdu_len > 0) {
		struct ieee80211_vif_priv *vifp = ieee80211_to_priv(txop->vif);

		vifp->hetb_psdu_len = ts->tb_psdu_len;
	}
#endif

	scm80211_handle_txop_complete(hw, txop);
	if (ts != status)
		goto again;
}

static void scm80211_usb_bulk_tx_status(struct ieee80211_hw *hw);
static void scm80211_usb_bulk_ts_complete(struct urb *urb)
{
	struct scm80211_urb_context *urb_context = urb->context;
	struct scm80211_usb_pipe *pipe = urb_context->pipe;
	struct ieee80211_hw_priv *priv = pipe->hw_priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	int status = 0;

	scm80211_trace(priv->hw,
		       "usb tx status pipe %d stat %d len %d urb 0x%p\n",
		       pipe->pipe_num, urb->status, urb->actual_length, urb);

	if (urb->status != 0) {
		status = -EIO;

		if (urb->status == -EPROTO || urb->status == -ETIMEDOUT) {
			unsigned long flags;

			spin_lock_irqsave(&usb_priv->lock, flags);
			if (usb_priv->err_cnt[pipe->pipe_num]++ > 1) {
				usb_priv->halt_bitmap |= BIT(pipe->pipe_num);
				ieee80211_queue_work(priv->hw, &usb_priv->usb_bus_monitor_work);
			}
			spin_unlock_irqrestore(&usb_priv->lock, flags);
		}

		switch (urb->status) {
		case -ECONNRESET:
		case -ENOENT:
		case -ESHUTDOWN:
			/* no need to spew these errors when device
			 * removed or urb killed due to driver shutdown
			 */
			status = -ECANCELED;
			break;
		default:
			scm80211_warn(priv->hw,
				      "usb tx status pipe %d ep 0x%2.2x failed: %d\n",
				      pipe->pipe_num, pipe->ep_address, urb->status);
			break;
		}
	} else if (urb->actual_length != 0) {
		/* reset usb bus error count */
		usb_priv->err_cnt[pipe->pipe_num] = 0;
		scm80211_tx_complete(priv->hw, urb_context->ts);
	}
	urb_context->urb = usb_get_urb(urb);
	scm80211_usb_cleanup_recv_urb(priv->hw, urb_context);

	if (status == 0)
		scm80211_usb_bulk_tx_status(priv->hw);
}

static void scm80211_usb_bulk_tx_status(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	struct scm80211_urb_context *urb_context;
	struct scm80211_usb_pipe *tspipe = &usb_priv->pipes[SCM80211_USB_PIPE_RX_RSTATE];
	struct urb *urb;
	int usb_status;

	do {
		usb_status = -1;
		urb_context = scm80211_usb_alloc_urb_from_pipe(tspipe);
		if (!urb_context)
			break;

		if (!urb_context->ts)
			break;

		memset(urb_context->ts, 0, sizeof(struct scm80211_tx_status_info));
		urb = urb_context->urb;
		urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
		urb->transfer_dma    = urb_context->dma;

		usb_fill_bulk_urb(urb, usb_priv->udev, tspipe->usb_pipe_handle,
				  urb_context->ts, sizeof(struct scm80211_tx_status_info),
				  scm80211_usb_bulk_ts_complete, urb_context);

		scm80211_trace(hw,
			       "tx status ep submit %d 0x%x ep 0x%2.2x len %zu buf 0x%p\n",
			       tspipe->pipe_num, tspipe->usb_pipe_handle,
			       tspipe->ep_address, sizeof(struct scm80211_tx_status_info),
			       urb_context->ts);

		usb_anchor_urb(urb, &tspipe->urb_submitted);
		usb_status = usb_submit_urb(urb, GFP_ATOMIC);
		if (usb_status) {
			usb_unanchor_urb(urb);
			scm80211_warn(hw, "tx status ep submit fail: %d\n", usb_status);
			break;
		}
		usb_put_urb(urb);
	} while (1);

	if (usb_status && urb_context)
		scm80211_usb_cleanup_recv_urb(hw, urb_context);
}

static struct scm80211_usb_tx_buff *scm80211_get_tx_buff(struct scm80211_usb_pipe *pipe, u8 *addr)
{
	int i;

	for (i = 0; i < NR_TXBUFF; i++) {
		if (addr == pipe->buff[i].data)
			return &pipe->buff[i];
	}
	BUG_ON(1);
	return NULL;
}

static struct scm80211_usb_pipe *scm80211_usb_get_pipe(struct urb *urb)
{
	struct txop *txop = urb->context;
	struct ieee80211_hw *hw = txop->hw;
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *pipe;
	int i;

	for (i = 0; i < SCM80211_USB_PIPE_RX_DATA; i++) {
		pipe = &usb_priv->pipes[i];
		if (urb->pipe == pipe->usb_pipe_handle)
			break;
	}

	if (i != SCM80211_USB_PIPE_RX_DATA)
		return pipe;
	else
		return NULL;
}

/**
 * scm80211_usb_bulk_tx_complete() - called after submit tx buffer to target
 */
static void scm80211_usb_bulk_tx_complete(struct urb *urb)
{
	struct txop *txop = urb->context;
	struct ieee80211_hw *hw = txop->hw;
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	struct scm80211_usb_pipe *pipe = scm80211_usb_get_pipe(urb);
	unsigned long flags;

	scm80211_trace(hw, "%s hwq#%d.%d (length %d status %d)\n", __func__,
		       txop->hwq->id,
		       txop->id,
		       urb->transfer_buffer_length,
		       urb->status);

	switch (urb->status) {
	case 0:
		break;
	case -ECONNRESET:
	case -ENOENT:
	case -ESHUTDOWN:
		return;
	default:
		usb_priv->halt_bitmap |= BIT(pipe->pipe_num);
		ieee80211_queue_work(priv->hw, &usb_priv->usb_bus_monitor_work);

		spin_lock_irqsave(&hwq->ring.lock, flags);
		list_del_init(&txop->list);
		__scm80211_txop_change_state(txop, TXOP_STATE_SCHEDULED);
		/* to put this txop in the head of the queue */
		txop->old_state = TXOP_STATE_NEEDS_RESCHED;
		__scm80211_txop_add(txop, TXOP_STATE_SCHEDULED);
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		return;
	}

	if (usb_priv->sg_size == 0) {
		spin_lock_irqsave(&pipe->cs_lock, flags);

		list_add_tail(&scm80211_get_tx_buff(pipe, urb->transfer_buffer)->list,
			      &pipe->buff_head);

		spin_unlock_irqrestore(&pipe->cs_lock, flags);
	}

	txop->urb = usb_get_urb(urb);
	/* Add this txop to issue list */
	spin_lock_irqsave(&hwq->ring.lock, flags);
	/* protect remove driver crash */
	list_del_init(&txop->list);
	__scm80211_txop_change_state(txop, TXOP_STATE_ISSUED);
	__scm80211_txop_add(txop, TXOP_STATE_ISSUED);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);

	if (hwq->need_resched) {
		schedule_work(&hwq->work);
		hwq->need_resched = false;
	}

	trace_scm80211_tx_return(txop->hwq, __func__);
}

static void scm80211_usb_cancel_all_request(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	int i;

	for (i = 0; i < SCM80211_USB_PIPE_MAX; i++) {
		if (usb_priv->pipes[i].flags & SCM80211_USB_PIPE_FLAG_INIT)
			usb_kill_anchored_urbs(&usb_priv->pipes[i].urb_submitted);

		if (i == SCM80211_USB_PIPE_RX_DATA) {
			kthread_stop(usb_priv->pipes[i].rx_thread);
			usb_priv->pipes[i].rx_thread = NULL;
		}
	}
}

static void scm80211_usb_destroy(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);

	scm80211_usb_cancel_all_request(hw);
	scm80211_usb_cleanup_pipe_resources(hw);
	usb_set_intfdata(usb_priv->interface, NULL);
}

/**
 * workaroud for EP in bulk hang
 */
void scm80211_usb_ep_status_check(struct work_struct *work)
{
	struct scm80211_usb_priv *usb_priv = container_of(work, struct scm80211_usb_priv,
							  usb_bus_monitor_work);
	struct ieee80211_hw *hw = usb_priv->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_pipe *pipe = NULL;
	int i;
	unsigned long flags;

	for_each_set_bit(i, &usb_priv->halt_bitmap, SCM80211_USB_PIPE_MAX) {
		pipe = &usb_priv->pipes[i];
		usb_kill_anchored_urbs(&pipe->urb_submitted);
		usb_priv->err_cnt[i] = 0;
		scm80211_dbg(usb_priv->hw, "try to reset ep#%d\n", i);
		usb_clear_halt(usb_priv->udev, pipe->usb_pipe_handle);

		if (i == SCM80211_USB_PIPE_RX_DATA) {
			/* Prepare rx urb again */
			scm80211_usb_bulk_rx(priv->hw);
		} else if (i == SCM80211_USB_PIPE_RX_RSTATE) {
			/* Prepare tx status urb again */
			scm80211_usb_bulk_tx_status(priv->hw);
		}
	}
	spin_lock_irqsave(&usb_priv->lock, flags);
	usb_priv->halt_bitmap = 0;
	spin_unlock_irqrestore(&usb_priv->lock, flags);
}

/**
 * scm80211_get_usb_pipe() - get corresponding bulk out ep
 */
static struct scm80211_usb_pipe *scm80211_get_usb_pipe(struct ieee80211_hw *hw, struct txop *txop)
{
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *pipe;
	struct scm80211_hwq *hwq = txop->hwq;

	if (hwq->ac != IEEE80211_AC_VO) {
		pipe = &usb_priv->pipes[SCM80211_USB_PIPE_TXQ0];
	} else {
		pipe = &usb_priv->pipes[SCM80211_USB_PIPE_TXQ1];
		BUG_ON(txop->total_len > hwq->limit);
	}

	return pipe;
}

/**
 * scm80211_usb_start_tx() - construct tx buffer and submit to target
 */
bool scm80211_usb_start_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *pipe = scm80211_get_usb_pipe(hw, txop);
	struct urb *urb = txop->urb;
	struct sk_buff *skb;
	struct scatterlist *sg = urb->sg;
	struct scm80211_usb_tx_buff *buff = NULL;
	struct scm80211_hwq *hwq = txop->hwq;
	struct txop_info *info = &txop->txop_info;
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	int i = 0, ret = -1;
	unsigned long flags;
#ifdef SCM80211_TWT
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct sk_buff *skb0 = skb_peek(&txop->sched);
	struct ieee80211_hdr *hdr = NULL;
	enum scm80211_twt_state twt_state;
#endif

	if (!pipe || !vif)
		return false;

	memset(info, 0, sizeof(*info));

	scm80211_trace(hw, "%s hwq#%d.%d pipe %p buffer len %u\n", __func__,
		       txop->hwq->id, txop->id, pipe, txop->total_len);

	if (txop->state != TXOP_STATE_SCHEDULED && txop->state != TXOP_STATE_TX_RETRY) {
		scm80211_warn(hw, "- txop not ready for transmission status=%s\n",
			      txop_state_str[txop->state]);
		goto exit;
	}
#ifdef SCM80211_TWT
	twt_state = scm80211_get_twt_state(vif);
	if (twt_state == TWT_DOZE) {
		scm80211_err(hw, "- TWT DOZE state should not transmit frame!!\n");
		goto exit;
	} else {
		if (skb0) {
			hdr = (void *)(skb0->data + sizeof(struct txvector));
		} else {
			WARN(1, "-%s no skb?\n", __func__);
			goto exit;
		}

		info->twt_state = twt_state;
		info->twt_trigger = vifp->twt_trigger;
		/* Workaround: HE_TB Action frame sent but no ack response */
		if (hdr && ieee80211_is_action(hdr->frame_control) && twt_state != TWT_SP)
			info->he = 0;
		else if (twt_state == TWT_SP && vifp->twt_trigger)
			info->he = 1;
	}
#endif
	info->num       = txop->n_mpdu;
	info->ampdu     = txop->ampdu.enabled;
	info->he        = ieee80211_rate_is_he(txop->rate);
	info->prot      = txop->protection;
	info->limit     = txop->limit;
	info->hwq_id    = hwq->id;
	info->txop_id   = txop->id;
	info->ecwmin    = ilog2(hwq->edca.cw_min + 1);
	info->ecwmax    = ilog2(hwq->edca.cw_max + 1);
	info->aifsn     = hwq->edca.aifs;
	info->rts_time  = txop->rts_time;
	info->ack_time  = txop->ack_time;
	info->vidx      = scm80211_vdev_id(vif) ? 1 : 0;
	info->probe     = !!txop->probe;

	if (!txop->multi_owner) {
		s16 wz = scm80211_ba_buf_size(txop->owner);

		info->txq_id = txop->owner->idx < 0 ? SCM80211_NUM_TXQ : txop->owner->idx;
		info->ssn = txop->owner->sb.ssn;
		/* If txop out of window, fw will check it */
		info->in_window = (!txop->ampdu.enabled && !txop->ampdu.update_needed) ||
				  scm80211_txop_is_within_transmit_window(txop);
		info->wz = wz;
	} else {
		info->txq_id = SCM80211_NUM_TXQ;
		if (txop->ampdu.enabled && !scm80211_txop_is_within_transmit_window(txop))
			goto exit;
		info->in_window = 1;
	}

	if ((hwq->mu_edca_status & MU_EDCA_RUNNING)) {
		info->mu_enable = 1;
		info->muedca_timer = hwq->mu_timer;
	}

	//info->band    where to get the band info
	if (txop->ampdu.enabled)
		info->mmss = txop->ampdu.mmss;
	info->rts_time_th = txop->rts_time_th;
	memcpy(info->rates, txop->rs.control.rates, sizeof(info->rates));
	if (txop->subused) {
		memcpy(info->rates + 1, &txop->subrate, sizeof(txop->subrate));
		info->rates[1].flags = info->rates[0].flags;
		info->rate_idx = txop->subtv.MCS;
	}

	if (usb_priv->sg_size > 0) {
		BUG_ON(txop->n_mpdu + 2 > usb_priv->sg_size);

		sg_init_table(urb->sg, txop->n_mpdu + 2);

		sg_set_buf(&sg[i++], &txop->txop_info, sizeof(txop->txop_info));
		sg_set_buf(&sg[i++], txop->dma[0].desc, sizeof(struct tx_mpdu_desc) * txop->n_mpdu);

		skb_queue_walk(&txop->sched, skb) {
			sg_set_buf(&sg[i++], skb->data, skb->len);
		}
		BUG_ON(i != txop->n_mpdu + 2);
		urb->num_sgs = i;
	} else {
		size_t pos = 0;

		spin_lock_irqsave(&pipe->cs_lock, flags);
		buff =
		    list_first_entry_or_null(&pipe->buff_head, struct scm80211_usb_tx_buff, list);
		if (buff)
			list_del_init(&buff->list);
		spin_unlock_irqrestore(&pipe->cs_lock, flags);

		if (!buff)
			goto exit;

		memcpy(buff->data + pos, &txop->txop_info, sizeof(txop->txop_info));
		pos += sizeof(txop->txop_info);

		memcpy(buff->data + pos, txop->dma[0].desc,
		       sizeof(struct tx_mpdu_desc) * txop->n_mpdu);
		pos += sizeof(struct tx_mpdu_desc) * txop->n_mpdu;

		skb_queue_walk(&txop->sched, skb) {
			memcpy(buff->data + pos, skb->data, skb->len);
			pos += skb->len;
		}
		urb->transfer_buffer = buff->data;
		urb->transfer_dma    = buff->dma;
		urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	}

	urb->transfer_buffer_length = txop->total_len;
	urb->pipe	= pipe->usb_pipe_handle;
	urb->dev	= usb_priv->udev;
	urb->context	= txop;
	urb->complete	= scm80211_usb_bulk_tx_complete;
	urb->transfer_flags |= URB_ZERO_PACKET;

	/*
	 * lock before submit urb to fixup urb complete before
	 * txop->state change to issuing
	 */
	spin_lock_irqsave(&hwq->ring.lock, flags);
	ret = usb_submit_urb(urb, GFP_ATOMIC);
	if (ret) {
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		scm80211_warn(hw, "submit tx urb fail %d\n", ret);
		if (usb_priv->sg_size == 0) {
			spin_lock_irqsave(&pipe->cs_lock, flags);
			list_add(&buff->list, &pipe->buff_head);
			spin_unlock_irqrestore(&pipe->cs_lock, flags);
		}
		goto exit;
	}

	__scm80211_txop_change_state(txop, TXOP_STATE_ISSUING);
	__scm80211_txop_add(txop, TXOP_STATE_ISSUING);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);

	//schedule_delayed_work(&txop->urb_timeout, msecs_to_jiffies(5000));

	trace_scm80211_tx_return(txop->hwq, __func__);
	return true;
exit:
	spin_lock_irqsave(&hwq->ring.lock, flags);
	__scm80211_txop_change_state(txop, TXOP_STATE_SCHEDULED);
	txop->old_state = TXOP_STATE_NEEDS_RESCHED;
	__scm80211_txop_add(txop, TXOP_STATE_SCHEDULED);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
	return false;
}

static void scm80211_usb_txop_transfer_timeout(struct work_struct *work)
{
	struct txop *txop = container_of(work, struct txop, urb_timeout.work);
	struct ieee80211_hw *hw = txop->hw;
	struct urb *urb = txop->urb;
	unsigned long flags;

	scm80211_warn(hw, "hwq%d.%d transfer timeout %d\n",
		      txop->hwq->id, txop->id, urb->status);

	/* 1. cancel this transfer request note: this function might sleep */
	usb_kill_urb(urb);

	/* 2. tried to sumbit stop again */
	usb_get_urb(urb);
	spin_lock_irqsave(&txop->hwq->ring.lock, flags);
	list_del_init(&txop->list);
	__scm80211_txop_change_state(txop, TXOP_STATE_SCHEDULED);
	spin_unlock_irqrestore(&txop->hwq->ring.lock, flags);
	scm80211_start_tx(hw, txop);
}

static int scm80211_usb_alloc_tx_urb(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	struct txop *txop;
	struct urb *urb;
	unsigned int size = sizeof(*urb);
	int n;

	size += usb_priv->sg_size * sizeof(struct scatterlist);
	for (n = 0; n < SCM80211_MAX_TXOP; n++) {
		txop = &hwq->ring.pool[n];
		urb = kzalloc(size, GFP_KERNEL);
		if (!urb)
			return -ENOMEM;
		txop->urb = urb;
		usb_init_urb(urb);
		if (usb_priv->sg_size != 0)
			urb->sg = (struct scatterlist *)(urb + 1);
		else
			urb->sg = NULL;

		INIT_DELAYED_WORK(&txop->urb_timeout, scm80211_usb_txop_transfer_timeout);
	}

	return 0;
}

static void scm80211_usb_free_tx_urb(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	int n;

	for (n = 0; n < SCM80211_MAX_TXOP; n++) {
		usb_kill_urb(hwq->ring.pool[n].urb);
		usb_free_urb(hwq->ring.pool[n].urb);
	}
}

static const char *const param_to_strings[] = {
	[RX_START]      = "rx_start",
	[RX_STOP]       = "rx_stop",
	[TX_START]      = "tx_start",
	[TX_STOP]       = "tx_stop",
	[VIF0_START]    = "start_vif0",
	[VIF0_STOP]     = "stop_vif0",
	[VIF1_START]    = "start_vif1",
	[VIF1_STOP]     = "stop_vif1",
	[AUTO_CAL]      = "auto_cal",
	[CHAN_SET]      = "set_chan",
	[DEBUG_CTL]     = "debug_ctl",
	[FW_VER]	= "fw_version",
	[ABORT_TX]	= "abort tx",
	[IND_TX_FAIL]   = "tx_fail",
	[QNULL_IRQ]     = "null irq",
	[HW_PS_ENABLE]	= "hw ps enable",
	[HW_PS_DISABLE] = "hw ps disable",
	[VIF_PS_ENABLE] = "vif ps enable",
	[VIF_PS_DISABLE] = "vif ps disable",
	[CONNECTED]	= "connect",
	[DISCONNECTED]	= "disconnect",
};

static void scm80211_usb_ctrl_rsp(struct ieee80211_hw *hw,
				      u8 req, u16 value, u16 index, void *data,
				      int ret, bool ctrl_flag)
{
	scm80211_warn(hw, "Failed, submit usb ctrl %s message ret: %d\n",
		      (ctrl_flag ? "out" : "in"), ret);
	switch (req) {
	case FW_DOWNLOAD:
		scm80211_info(hw,
			      " - HW MAC ucode download fail to program 0x%04x%04x\n",
			      index, value);
		break;
	case PARAM_WRITE:
		scm80211_info(hw,
			      " - Set/get %s fail\n", param_to_strings[index]);
		break;
	case REG_WRITE:
		scm80211_info(hw,
			      " - Write reg 0x%04x%04x to 0x%08x fail\n",
			      index, value, *(u32 *)data);
		break;
	case REG_READ:
		scm80211_info(hw,
			      " - Read reg 0x%04x%04x  fail\n", index, value);
		break;
	case PARAM_READ:
		scm80211_info(hw,
			      " - Set/get %s fail\n", param_to_strings[index]);
		break;
	default:
		scm80211_info(hw, " - Unknown request !! req(%02x)\n", req);
		break;
	}
}

static inline int scm80211_usb_dev_out_request(struct usb_device *udev,
				      u8 req, u16 value, u16 index, void *data,
				      u32 size)
{
	return usb_control_msg(udev,
			      usb_sndctrlpipe(udev, 0),
			      req,
			      USB_DIR_OUT | USB_TYPE_VENDOR |
			      USB_RECIP_DEVICE, value, index, data,
			      size, USB_CTRL_SET_TIMEOUT);
}

static int scm80211_usb_ctrl_out_request(struct ieee80211_hw *hw,
					 u8 req, u16 value, u16 index, void *data, u32 size)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	int ret;

	ret = scm80211_usb_dev_out_request(usb_priv->udev, req, value, index, data, size);
	if (ret < 0)
		scm80211_usb_ctrl_rsp(hw, req, value, index, data, ret, true);

	return ret;
}

static inline int scm80211_usb_dev_in_request(struct usb_device *udev,
					u8 req, u16 value, u16 index, void *data,
					u32 size)
{
	return usb_control_msg(udev,
			      usb_rcvctrlpipe(udev, 0),
			      req,
			      USB_DIR_IN | USB_TYPE_VENDOR |
			      USB_RECIP_DEVICE, value, index, data,
			      size, USB_CTRL_GET_TIMEOUT);
}

static int scm80211_usb_ctrl_in_request(struct ieee80211_hw *hw,
					u8 req, u16 value, u16 index, void *data, u32 size)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);
	int ret;

	if (object_is_on_stack(data))
		scm80211_warn(hw, "transfer buffer is on stack");

	ret = scm80211_usb_dev_in_request(usb_priv->udev, req, value, index, data, size);
	if (ret < 0)
		scm80211_usb_ctrl_rsp(hw, req, value, index, data, ret, false);

	return ret;
}

static int scm80211_usb_fw_download(struct ieee80211_hw *hw, u32 addr, void *data, u16 size)
{
	u16 value = (u16)(addr & 0xffff);
	u16 index = (u16)((addr & 0xffff0000) >> 16);

	return scm80211_usb_ctrl_out_request(hw, FW_DOWNLOAD, value, index, data, size);
}


static void scm80211_usb_param_write(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				     int param_indx, void *data, size_t size)
{
	void *buf = kzalloc(size, GFP_KERNEL);

	if (!buf)
		return;
	memcpy(buf, data, size);

	scm80211_usb_ctrl_out_request(hw, PARAM_WRITE, 0, param_indx, buf, size);
	kfree(buf);
	msleep(1);
}

static u32 scm80211_usb_param_read(struct ieee80211_hw *hw, u16 param_indx, void *data, u16 size)
{
	u8 *buf = kzalloc(size, GFP_KERNEL);

	if (!buf)
		return -ENOMEM;

	scm80211_usb_ctrl_in_request(hw, PARAM_READ, 0, param_indx, buf, size);

	memcpy(data, buf, size);
	kfree(buf);

	return 0;
}

static void scm80211_usb_raw_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int addr);

static u32 scm80211_usb_raw_readl(struct ieee80211_hw *hw, unsigned int addr);

static void scm80211_usb_mac_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	scm80211_usb_raw_writel(hw, val, MAC_BASE_0 + offset);
}

static u32 scm80211_usb_mac_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return scm80211_usb_raw_readl(hw, MAC_BASE_0 + offset);
}

static void scm80211_usb_phy_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	scm80211_usb_raw_writel(hw, val, PHY_BASE_0 + offset);
}

static u32 scm80211_usb_phy_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return scm80211_usb_raw_readl(hw, PHY_BASE_0 + offset);
}

/* In case RF registers are memory mapped */
static void scm80211_usb_rf_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	scm80211_usb_raw_writel(hw, val, RF_BASE_START + offset);
}

static u32 scm80211_usb_rf_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return scm80211_usb_raw_readl(hw, RF_BASE_START + offset);
}

#if defined(SCM2020_Z1) || defined(SCM2020_A0)

static void scm80211_usb_pta_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int offset)
{
	scm80211_usb_raw_writel(hw, val, PTA_BASE_START + offset);
}

static u32 scm80211_usb_pta_readl(struct ieee80211_hw *hw, unsigned int offset)
{
	return scm80211_usb_raw_readl(hw, PTA_BASE_START + offset);
}

#endif

static void scm80211_usb_raw_writel(struct ieee80211_hw *hw, unsigned int val, unsigned int addr)
{
	u16 value = (u16)(addr & 0xffff);
	u16 index = (u16)((addr & 0xffff0000) >> 16);
	u32 *buf = kzalloc(sizeof(u32), GFP_KERNEL);

	if (!buf)
		return;
	*buf = val;

	scm80211_usb_ctrl_out_request(hw, REG_WRITE, value, index, buf, 4);

	kfree(buf);
}

static u32 scm80211_usb_raw_readl(struct ieee80211_hw *hw, unsigned int addr)
{
	u16 value = (u16)(addr & 0xffff);
	u16 index = (u16)((addr & 0xffff0000) >> 16);
	u32 val;
	u32 *buf = kmalloc(sizeof(u32), GFP_KERNEL);

	if (!buf)
		return -ENOMEM;

	scm80211_usb_ctrl_in_request(hw, REG_READ, value, index, buf, sizeof(val));

	val = *buf;
	kfree(buf);
	return val;
}

static void scm80211_usb_efuse_writel(void *dev_ep, void *val, unsigned int addr, unsigned int len)
{
#ifdef SCM80211_FTM
	struct usb_device *udev = dev_ep;
	void *buf = kzalloc(len, GFP_KERNEL);

	if (!buf)
		return;

	memcpy(buf, val, len);

	scm80211_usb_dev_out_request(udev, PARAM_WRITE, 0, SET_EFUSE, buf, len);
	kfree(buf);
#endif
}

static int scm80211_usb_efuse_readl(void *dev_ep, void *val, unsigned int addr, unsigned int len)
{
	struct usb_device *udev = dev_ep;
	u8 *buf = kmalloc(len, GFP_KERNEL);
	int ret;

	if (!buf)
		return -ENOMEM;

	ret = scm80211_usb_dev_in_request(udev, PARAM_READ, 0, GET_EFUSE, buf, len);
	if (ret >= 0)
		memcpy(val, buf, len);
	kfree(buf);
	return ret;
}

static int scm80211_usb_alloc_dma_desc(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	int n;
	size_t size = sizeof(struct tx_dma_desc);

	for (n = 0; n < SCM80211_MAX_TXOP; n++) {
		struct txop *txop = &hwq->ring.pool[n];

		txop->dma[0].desc = kzalloc(size, GFP_KERNEL);
		if (!txop->dma[0].desc) {
			int i;

			scm80211_err(hw, "hwq#%d: failed to allocate descriptors\n", hwq->id);
			for (i = n - 1; i >= 0; i--) {
				kfree(hwq->ring.pool[i].dma[0].desc);
				hwq->ring.pool[i].dma[0].desc = NULL;
			}
			return -ENOMEM;
		}
	}
	return 0;
}

static void scm80211_usb_free_dma_desc(struct ieee80211_hw *hw, struct scm80211_hwq *hwq)
{
	int n;

	for (n = 0; n < SCM80211_MAX_TXOP; n++) {
		struct txop *txop = &hwq->ring.pool[n];

		if (txop->dma[0].desc) {
			kfree(txop->dma[0].desc);
			txop->dma[0].desc = NULL;
		}
	}
}

#define CENTER_FREQ_MASK        0xFFFF
#define PRIMARY_LOC_MASK        0xF0000
#define BANDWIDTH_MASK          0x300000
#define MIMO_ENABLE_MASK        0xC00000

static int scm80211_usb_set_channel(struct ieee80211_hw *hw, struct scm80211_phy *phy,
				    struct phy_vector *pv)
{
	u32 ch;
	u8 done;
	int cnt = 0;

	/*
	 * FW will set both phy-0/phy-1 rf-0/rf-1 once if mimo used
	 * No need to set channel for phy-1 again, it will reduce the
	 * timing cost for channel setting
	 */
	if (phy->chain != 0)
		return 0;

	/* bit[0~15] center freq, bit[16-19] pri_loc, bit[20-23] bw*/
	ch = FIELD_PREP(CENTER_FREQ_MASK, pv->CENTER_FREQUENCY);
	ch |= FIELD_PREP(PRIMARY_LOC_MASK, pv->PRIMARY_20MHZ_CHANNEL_LOC);
	ch |= FIELD_PREP(BANDWIDTH_MASK, pv->BANDWIDTH);
	scm80211_usb_param_write(phy->hw, NULL, CHAN_SET, &ch, sizeof(ch));

	mdelay(10);
	scm80211_usb_param_read(phy->hw, CHAN_SET, &done, sizeof(done));
	while (done == 0) {
		mdelay(1);
		scm80211_usb_param_read(phy->hw, CHAN_SET, &done, sizeof(done));
		if (cnt++ > 50)
			break;
	}

	scm80211_trace(hw, "channel set to %d bw %d (chinfo %08x) %s (%d)\n",
		       pv->CENTER_FREQUENCY, pv->BANDWIDTH, ch, done ? "success" : "fail", cnt);

	return done ? 0 : -EAGAIN;
}

static int scm80211_usb_ant_select(struct ieee80211_hw *hw, int ant)
{
	scm80211_usb_param_write(hw, NULL, ANT_SEL, &ant, sizeof(ant));
	return 0;
}

static int scm80211_usb_get_ant(struct ieee80211_hw *hw)
{
	int ant = 0;

	scm80211_usb_param_read(hw, ANT_SEL, &ant, sizeof(ant));
	return ant;
}

static int scm80211_usb_set_param_b(struct ieee80211_hw *hw, int *buff, int length)
{
	scm80211_usb_param_write(hw, NULL, SET_PARAM_B, buff, length);
	return 0;
}

static int scm80211_usb_get_param_b(struct ieee80211_hw *hw, int *buff, int length)
{
	return scm80211_usb_param_read(hw, GET_PARAM_B, buff, length);
}

static int scm80211_usb_get_tssi(struct ieee80211_hw *hw)
{
	int tssi = 0;

	scm80211_usb_param_read(hw, GET_TSSI, &tssi, sizeof(tssi));
	return tssi;
}

static int scm80211_usb_power_save(struct ieee80211_hw *hw, bool enable, bool link)
{
	struct scm80211_usb_priv *p = hw_to_usb_priv(hw);
	struct usb_device *udev = p->udev;
	enum htc_cmd cmd = enable ? HW_PS_ENABLE : HW_PS_DISABLE;
	u32 v;
	u8 set = link ? 1 : 0;

	scm80211_usb_param_write(hw, NULL, cmd, &set, 1);

	if (enable) {
		usb_enable_lpm(udev);
	} else {
		int timeout = 100;

		if (udev->lpm_disable_count == 0)
			usb_disable_lpm(udev);
#ifdef SCM2020_A0
		/*
		 * Wait until PHY clock is enabled
		 * It is observed that TX frames on OUT endpoints reach the target
		 * before this.
		 */
		do {
			v = scm80211_raw_readl(hw, WLAN_WMP_CFG);
			if (v & PHY_CLK_EN) {
				msleep(2);
				goto out;
			}
			msleep(1);
		} while (timeout-- > 0);
#endif
		scm80211_warn(hw, "disabling ps mode failed\n");
	}

out:
	scm80211_trace(hw, "power save %s: lpm_disable_count=%d\n",
		       enable ? "enabled" : "disabled",
		       p->udev->lpm_disable_count);
	return 0;
}

static int scm80211_usb_vif_power_save(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				       bool enable)
{
	int vdev = scm80211_vdev_id(vif);

	scm80211_usb_param_write(hw, NULL, enable ? VIF_PS_ENABLE : VIF_PS_DISABLE,
				 &vdev, sizeof(int));

	return 0;
}

static int scm80211_usb_phy_start(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u8 done = 0;
	int i = 0;

	scm80211_usb_param_write(hw, NULL, AUTO_CAL, &priv->n_chain, sizeof(unsigned int));
	mdelay(100);

	scm80211_usb_param_read(hw, AUTO_CAL, &done, sizeof(u8));
	while (done == 0 && i++ < 1000) {
		mdelay(10);
		scm80211_usb_param_read(hw, AUTO_CAL, &done, sizeof(u8));
	}

	scm80211_trace(hw, "rf calibration %s\n", done ? "done" : "fail");
	return 0;
}

/* nothing need to do */
static void scm80211_usb_phy_stop(struct ieee80211_hw *hw)
{
}

static int scm80211_usb_register_bss_info(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					  struct ieee80211_bss_conf *info)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct scm80211_vif usb_vif;

	memset(&usb_vif, 0, sizeof(struct scm80211_vif));
	usb_vif.type = (u8)vif->type;
	usb_vif.idx = vp->index;
	usb_vif.bss_conf.associated = true;
	usb_vif.bss_conf.sync_device_ts = info->sync_device_ts;
	usb_vif.bss_conf.dtim_period = info->dtim_period;
	usb_vif.bss_conf.sync_dtim_count = info->sync_dtim_count;
	usb_vif.bss_conf.listen_interval = hw->conf.listen_interval;
	usb_vif.bss_conf.beacon_int = info->beacon_int;
	memcpy(usb_vif.bss_conf.bssid, info->bssid, ETH_ALEN);
	memcpy(usb_vif.bss_conf.transmitter_bssid, info->transmitter_bssid, ETH_ALEN);
	scm80211_usb_param_write(hw, NULL, CONNECTED, &usb_vif, sizeof(struct scm80211_vif));
	return 0;
}

static void scm80211_usb_deregister_bss_info(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					     struct ieee80211_bss_conf *bss_info)
{
	struct scm80211_vif usb_vif;

	memset(&usb_vif, 0, sizeof(struct scm80211_vif));
	usb_vif.idx = scm80211_vdev_id(vif);
	usb_vif.type = vif->type;
	scm80211_usb_param_write(hw, NULL, DISCONNECTED, &usb_vif, sizeof(struct scm80211_vif));
}

static void scm80211_clear_halt(struct ieee80211_hw *hw)
{
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *pipe = NULL;
	int i;

	/* usb clear halt */
	for (i = 0; i < SCM80211_USB_PIPE_MAX; i++) {
		pipe = &usb_priv->pipes[i];
		scm80211_trace(hw, "halt clear ep#%d\n", i);
		if (i >= SCM80211_USB_PIPE_RX_DATA)
			usb_kill_anchored_urbs(&pipe->urb_submitted);
		usb_clear_halt(usb_priv->udev, pipe->usb_pipe_handle);
	}
}

static __used int scm80211_usb_rx_poll(struct napi_struct *napi, int budget)
{
	struct ieee80211_hw_priv *priv =
		container_of(napi, struct ieee80211_hw_priv, napi.context);
	struct ieee80211_hw *hw = priv->hw;

	return scm80211_rx_poll(hw, budget);
}

#define USB_USE_NAPI	0

static int scm80211_usb_init_rx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

#if USB_USE_NAPI
	netif_napi_add(&priv->napi.netdev, &priv->napi.context, scm80211_usb_rx_poll, 64);
	napi_enable(&priv->napi.context);
	priv->napi.enabled = true;
#else
	priv->napi.enabled = false;
#endif

	scm80211_refill_rxpool(hw);

	/* init rx urb */
	scm80211_usb_bulk_rx(hw);
	scm80211_refill_rxpool(hw);

	/* Info the fw to start rx */
	scm80211_usb_param_write(hw, NULL, RX_START, NULL, 0);
	msleep(10);

#ifdef FW_LOG_UPLOAD
	scm80211_usb_param_write(hw, NULL, ENABLE_FW_LOG_UPLOAD, NULL, 0);
#endif
	return 0;
}

static void scm80211_usb_deinit_rx(struct ieee80211_hw *hw)
{
	scm80211_usb_param_write(hw, NULL, RX_STOP, NULL, 0);
}

static int scm80211_usb_rx_start(struct ieee80211_hw *hw, struct sk_buff *skb,
				 struct scm80211_rx_data *rx)
{
	struct scm80211_rx_info *info = SCM80211_SKB_RXCB(skb);

	info->rxd.frag = SIGNAL_FRAME;
	info->rxd.dlen = skb->len;
	info->rxd.blen = rx_buflen;

	return 0;
}

static int scm80211_usb_len_limit(struct ieee80211_hw *hw, struct txop *txop, int n_frags, u32 len)
{
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_hwq *hwq = txop->hwq;
	u32 limit = hwq->limit;

#ifdef __unused__
	if (bitmap_weight(&priv->vif_map, SCM80211_NUM_VIF) == 1 &&
	    usb_priv->sg_size != 0 && hwq->ac != IEEE80211_AC_VO) {
		limit <<= 1;
	}
#endif
	if (limit < len || (usb_priv->sg_size != 0 && usb_priv->sg_size < n_frags))
		return -EIO;

	return 0;
}

static int scm80211_usb_init_tx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		hwq = &hwp->hwq[i];
		scm80211_usb_alloc_tx_urb(hw, hwq);
	}
	scm80211_usb_bulk_tx_status(hw);

	/* Info the fw to start rx */
	scm80211_usb_param_write(hw, NULL, TX_START, NULL, 0);
	return 0;
}

static void scm80211_usb_deinit_tx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		hwq = &hwp->hwq[i];
		scm80211_usb_free_tx_urb(hw, hwq);
	}

	/* Info the fw to stop tx */
	scm80211_usb_param_write(hw, NULL, TX_STOP, NULL, 0);
	scm80211_clear_halt(hw);
	msleep(10);
}

static int scm80211_usb_vif_init_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	int vid = scm80211_vdev_id(vif);
	u8 done = 0;
	int retry = 0;

	BUG_ON(vid < 0 || vid > 1);

	scm80211_usb_param_write(hw, NULL, VIF0_START + (vid * 2), NULL, 0);

	do {
		msleep(10);
		scm80211_usb_param_read(hw, VIF0_START + (vid * 2), &done, sizeof(done));
	} while (retry++ < 10 && done == 0);

	scm80211_trace(hw, "%s done = %u retries %d\n", __func__, done, retry);

	return 0;
}

static void scm80211_usb_vif_deinit_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int vid = scm80211_vdev_id(vif);
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *pipe;
	u8 done = 0;
	int retry = 0;

	BUG_ON(vid < 0 || vid > 1);

	switch (bitmap_weight(&priv->vif_map_ref, SCM80211_NUM_VIF)) {
	case 0:
		scm80211_trace(hw, "clear TX_PIPE\n");
		pipe = &usb_priv->pipes[SCM80211_USB_PIPE_TXQ0];
		usb_clear_halt(usb_priv->udev, pipe->usb_pipe_handle);
		pipe = &usb_priv->pipes[SCM80211_USB_PIPE_TXQ1];
		usb_clear_halt(usb_priv->udev, pipe->usb_pipe_handle);
		break;
	default:
		break;
	}

	msleep(10);
	scm80211_usb_param_write(hw, NULL, VIF0_STOP + (vid * 2), NULL, 0);

	do {
		msleep(10);
		scm80211_usb_param_read(hw, VIF0_STOP + (vid * 2), &done, sizeof(done));
	} while (retry++ < 10 && done == 0);

	scm80211_trace(hw, "%s done = %u retries %d\n", __func__, done, retry);
}

#define ADD_TXQ_IDX	BIT(7)
#define DEL_TXQ_IDX	0
static int scm80211_usb_register_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				     struct ieee80211_txq *itxq)
{
	struct scm80211_txq *txq = (struct scm80211_txq *)itxq;
	u8 txq_op = txq->idx | ADD_TXQ_IDX;

	scm80211_trace(hw, "register txq%s idx %d\n", txq->name, txq->idx);
	scm80211_usb_param_write(hw, NULL, TXQ_OPER, &txq_op, sizeof(txq_op));
	return 0;
}

static void scm80211_usb_deregister_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
					struct ieee80211_txq *itxq)
{
	struct scm80211_txq *txq = (struct scm80211_txq *)itxq;
	u8 txq_op = txq->idx | DEL_TXQ_IDX;

	scm80211_trace(hw, "deregister txq%s idx %d\n", txq->name, txq->idx);
	scm80211_usb_param_write(hw, NULL, TXQ_OPER, &txq_op, sizeof(txq_op));
}

static int scm80211_usb_cancel_tx(struct ieee80211_hw *hw, struct txop *txop, bool discard)
{
	struct scm80211_hwq *hwq = txop->hwq;

	scm80211_trace(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

	scm80211_txop_add_back(txop);
	scm80211_usb_param_write(hw, NULL, ABORT_TX, &hwq->id, sizeof(hwq->id));

	/*
	 * there will be 40ms delay from cmd to fw, and handle the abort cmd
	 * so need more delay to wait fw finsh aborting backoff
	 */
	msleep(60);
	return 0;
}

static int scm80211_usb_prepare_tx(struct ieee80211_hw *hw, struct txop *txop,
				   int idx, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = &txop->xti[idx];
	struct sg_table *sgt = &xti->sgt;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)(skb->data + sizeof(xti->tv));
	struct tx_mpdu_desc *mdesc = txop_get_mdesc(txop, idx);
	struct scm80211_txq *txq = txop->multi_owner ? txop->multi_owner : txop->owner;
	int n = scm80211_get_skb_fragments(skb);

	WARN_ON(n > 1);

	memset(sgt, 0, sizeof(*sgt));
	sgt->nents = n;

	memset(word(mdesc, 2), 0, sizeof(__le32) * 3);
	mdesc->aid = txq->txq.vif->bss_conf.aid;
	mdesc->tid = xti->tid;
	mdesc->sn = IEEE80211_SEQ_TO_SN(le16_to_cpu(hdr->seq_ctrl));
	xti->sn = mdesc->sn;

	/*
	 * NB: IEEE80211_TX_CTL_NO_ACK does not indicate the frame does not
	 * solicits a response frame. It only indicates the upper layer is
	 * not interested whether the frame is acknowledged or not. Therefore,
	 * @mdesc->noack should be set without referrering to it.
	 */
	mdesc->noack = !xti->ack;
	mdesc->mpdu_len = xti->mlen;

	/* PMF is done by software. H/W needs to bypass encyrption */
	/* PMF + unsupported cipher : key is NULL */
	if (ieee80211_has_protected(hdr->frame_control) &&
	    !IEEE80211_SKB_CB(skb)->control.hw_key) {
		mdesc->bypass_encrypt = true;
	}

	/* RTS protection is set in scm80211_finalize_txop() */
	mdesc->protection = NO_PROTECTION;

	/* EOF */
	mdesc->eof = xti->eof;

	/* FIXME: TSF */
	mdesc->timestamp_update = false;
	mdesc->timestamp_offset = false;

	if (txop->ampdu.enabled) {
		mdesc->aggregate = true;
		mdesc->mpdu_spacing = xti->mmss_pad;
	}

	mdesc->bdnum = cpu_to_le32(xti->sgt.nents - 1);
	if (ieee80211_is_mgmt(hdr->frame_control) || ieee80211_is_data_qos(hdr->frame_control))
		mdesc->no_htc = !ieee80211_has_order(hdr->frame_control);

	/* fillup mdesc->priv parameter */
	mdesc->buf_len = skb->len;
	if (txop->ampdu.enabled) {
		mdesc->txtime0 = xti->per_time;
		if (txop->subused)
			mdesc->txtime1 = xti->subper_time;
	} else {
		mdesc->txtime0 = xti->txtime;
	}

	txop->total_len += skb->len + sizeof(*mdesc);

	return 0;
}

#ifdef SCM80211_WOW
static int scm80211_usb_wow_set_wakeup(struct ieee80211_hw *hw, struct scm80211_wow_context *wow)
{
	scm80211_usb_param_write(hw, NULL, WOW_ENABLE, wow, sizeof(scm80211_wow_context));
	return 0;
}
#endif

static int scm80211_usb_set_fw_debug(struct ieee80211_hw *hw, u32 debug_info)
{
	u8 dbg_info = (u8)debug_info;

	scm80211_usb_param_write(hw, NULL,  DEBUG_CTL, &dbg_info, sizeof(dbg_info));
	msleep(2);
	return 0;
}

static int scm80211_usb_next_beacon(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				    struct sk_buff *skb)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_bss_conf *conf = &vif->bss_conf;
	struct ieee80211_mgmt     *beacon = (struct ieee80211_mgmt *) skb->data;
	const struct element      *elem;
	const struct ieee80211_tim_ie *tim;
	int n_tbtt = 1;

	if (scm80211_ps_enabled(hw)) {
		n_tbtt = conf->sync_dtim_count > 1 ?
			 conf->sync_dtim_count - 1 :
			 conf->sync_dtim_count == 1 ?
			 conf->dtim_period :
			 conf->dtim_period > 1 ?
			 conf->dtim_period - 1 : 1;
		n_tbtt = min_t(int, hw->conf.listen_interval, n_tbtt);
	}
	/* simply calulated */
	vifp->bss.beacon.next = vifp->bss.beacon.last +
				ieee80211_tu_to_tsf(conf->beacon_int) * n_tbtt;

	if (!scm80211_ps_enabled(hw))
		return 0;

	elem = cfg80211_find_elem(WLAN_EID_TIM, beacon->u.beacon.variable,
				  skb->len - offsetof(struct ieee80211_mgmt, u.beacon.variable[0]));

	if (!elem || elem->datalen < sizeof(*tim))
		return -EINVAL;

	tim = (const struct ieee80211_tim_ie *) elem->data;
	if (ieee80211_check_tim(tim, elem->datalen, conf->aid)) {
		scm80211_trace(hw, "have more unicast for me,  stay awake state...\n");
		ieee80211_queue_work(priv->hw, &priv->pm.disable_work);
	}

	return 0;
}

const struct scm80211_bus_ops scm80211_usb_ops = {
	.mac_writel		= scm80211_usb_mac_writel,
	.mac_readl		= scm80211_usb_mac_readl,
	.phy_writel		= scm80211_usb_phy_writel,
	.phy_readl		= scm80211_usb_phy_readl,
	.rf_writel		= scm80211_usb_rf_writel,
	.rf_readl		= scm80211_usb_rf_readl,
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	.pta_writel		= scm80211_usb_pta_writel,
	.pta_readl		= scm80211_usb_pta_readl,
#endif
	.raw_writel		= scm80211_usb_raw_writel,
	.raw_readl		= scm80211_usb_raw_readl,
	.efuse_writel		= scm80211_usb_efuse_writel,
	.efuse_readl		= scm80211_usb_efuse_readl,
	.param_write		= scm80211_usb_param_write,
	.start_tx		= scm80211_usb_start_tx,
	.hwq_alloc_dma		= scm80211_usb_alloc_dma_desc,
	.hwq_free_dma		= scm80211_usb_free_dma_desc,
	.set_channel		= scm80211_usb_set_channel,
	.set_ant_sel		= scm80211_usb_ant_select,
	.get_ant_sel		= scm80211_usb_get_ant,
	.set_param_b		= scm80211_usb_set_param_b,
	.get_param_b		= scm80211_usb_get_param_b,
	.get_tssi		= scm80211_usb_get_tssi,
	.power_save		= scm80211_usb_power_save,
	.vif_power_save		= scm80211_usb_vif_power_save,
	.phy_start		= scm80211_usb_phy_start,
	.phy_stop		= scm80211_usb_phy_stop,
	.associated		= scm80211_usb_register_bss_info,
	.disconnected		= scm80211_usb_deregister_bss_info,
	.init_rx		= scm80211_usb_init_rx,
	.deinit_rx		= scm80211_usb_deinit_rx,
	.rx_start		= scm80211_usb_rx_start,
	.txop_len_limit		= scm80211_usb_len_limit,
	.init_tx		= scm80211_usb_init_tx,
	.deinit_tx		= scm80211_usb_deinit_tx,
	.refill_rxring		= scm80211_usb_bulk_rx,
	.vif_init_tx		= scm80211_usb_vif_init_tx,
	.vif_deinit_tx		= scm80211_usb_vif_deinit_tx,
	.register_txq		= scm80211_usb_register_txq,
	.deregister_txq		= scm80211_usb_deregister_txq,
	.cancel_tx		= scm80211_usb_cancel_tx,
	.prepare_tx		= scm80211_usb_prepare_tx,
#ifdef SCM80211_WOW
	.wow_set_wakeup		= scm80211_usb_wow_set_wakeup,
#endif
	.set_fw_debug		= scm80211_usb_set_fw_debug,
	.next_beacon		= scm80211_usb_next_beacon,
};
const struct scm80211_bus_ops *scm_bus_ops = &scm80211_usb_ops;

static u8 scm80211_usb_get_pipe_urb_num(u8 ep_address, int *urb_count, u8 base_in, u8 base_out)
{
	u8 pipe_num = SCM80211_USB_PIPE_INVALID;

	if (SCM80211_USB_EP_ADDR_IS_DATA_OUT(ep_address)) {
		pipe_num = SCM80211_USB_PIPE_TXQ0 + (ep_address - base_out);
		//*urb_count = TX_URB_COUNT;
		*urb_count = 0;	//no need for tx, txop will be used;
	} else if (ep_address == base_in) {
		pipe_num = SCM80211_USB_PIPE_RX_DATA;
		*urb_count = RX_URB_COUNT;
	} else if (ep_address == (base_in + 0x1)) {
		pipe_num = SCM80211_USB_PIPE_RX_RSTATE;
		*urb_count = TS_URB_COUNT;
	}

	return pipe_num;
}

static int scm80211_usb_alloc_pipe_urb_context(struct ieee80211_hw *hw,
					       struct usb_interface *interface,
					       struct scm80211_usb_pipe *pipe,
					       int urb_cnt)
{
	struct scm80211_urb_context *urb_context;
	size_t size = sizeof(struct urb);
	int i;

	for (i = 0; i < urb_cnt; i++) {
		urb_context = kzalloc(sizeof(*urb_context), GFP_KERNEL);
		if (!urb_context)
			return -ENOMEM;

		urb_context->urb = kzalloc(size, GFP_KERNEL);

		if (!urb_context->urb) {
			kfree(urb_context);
			return -ENOMEM;
		}

		urb_context->pipe = pipe;
		urb_context->idx = i;

		pipe->urb_alloc++;
		if (pipe->pipe_num == SCM80211_USB_PIPE_RX_RSTATE) {
			urb_context->ts =  usb_alloc_coherent(interface_to_usbdev(interface),
					   sizeof(struct scm80211_tx_status_info), GFP_KERNEL,
					   &urb_context->dma);
			if (!urb_context->ts) {
				usb_free_urb(urb_context->urb);
				kfree(urb_context);
				return -ENOMEM;
			}
		}

		usb_init_urb(urb_context->urb);
		scm80211_usb_free_urb_to_pipe(pipe, urb_context);
	}

	return 0;
}

static int scm80211_usb_init_pipe(struct ieee80211_hw *hw, struct usb_interface *interface)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct usb_host_interface *iface_desc = interface->cur_altsetting;
	struct usb_endpoint_descriptor *endpoint;
	struct scm80211_usb_pipe *pipe;
	struct usb_device *udev = interface_to_usbdev(interface);
	int ret, i, urb_count;
	u8 pipe_num;
	u8 base_in = 0;
	u8 base_out = 0;

	spin_lock_init(&usb_priv->lock);
	INIT_LIST_HEAD(&usb_priv->tx_head);
	/* setup pipes */
	for (i = 0; i < iface_desc->desc.bNumEndpoints; ++i) {
		endpoint = &iface_desc->endpoint[i].desc;

		urb_count = 0;
		if (base_in == 0 && !SCM80211_USB_EP_ADDR_IS_DATA_OUT(endpoint->bEndpointAddress))
			base_in = endpoint->bEndpointAddress;

		if (base_out == 0 && SCM80211_USB_EP_ADDR_IS_DATA_OUT(endpoint->bEndpointAddress))
			base_out = endpoint->bEndpointAddress;

		pipe_num = scm80211_usb_get_pipe_urb_num(endpoint->bEndpointAddress,
							 &urb_count, base_in, base_out);

		if (pipe_num == SCM80211_USB_PIPE_INVALID)
			continue;

		pipe = &usb_priv->pipes[pipe_num];
		if (pipe->flags & SCM80211_USB_PIPE_FLAG_INIT)
			continue;

		scm80211_trace(hw, "ep idx %d, pipe num %d\n", i, pipe_num);

		pipe->hw_priv = priv;
		pipe->pipe_num = pipe_num;
		pipe->ep_address = endpoint->bEndpointAddress;
		pipe->max_packet_size = le16_to_cpu(endpoint->wMaxPacketSize);
		if (pipe_num == SCM80211_USB_PIPE_RX_DATA) {
			pipe->rx_thread = kthread_run(scm80211_usb_rx_thread, pipe, "usb-rx");
			BUG_ON(IS_ERR(pipe->rx_thread));
		}

		if (SCM80211_USB_IS_BULK_EP(endpoint->bmAttributes)) {
			if (SCM80211_USB_IS_DIR_IN(pipe->ep_address)) {
				pipe->usb_pipe_handle =
				    usb_rcvbulkpipe(udev, pipe->ep_address);
			} else {
				pipe->usb_pipe_handle =
				    usb_sndbulkpipe(udev, pipe->ep_address);
			}
		}

		if (SCM80211_USB_IS_BULK_OUT(endpoint->bmAttributes, pipe->ep_address)) {
			if (usb_priv->sg_size == 0) {
				int j;

				INIT_LIST_HEAD(&pipe->buff_head);
				for (j = 0; j < NR_TXBUFF; j++) {
					pipe->buff[j].data =
						usb_alloc_coherent(udev,
								   SCM80211_TXOP_AMPDU_LEN,
								   GFP_KERNEL,
								   &pipe->buff[j].dma);

					if (!pipe->buff[j].data) {
						scm80211_err(hw, "No buff for tx\n");
						return -ENOMEM;
					}
					list_add_tail(&pipe->buff[j].list, &pipe->buff_head);
				}
			}
		}

		pipe->ep_desc = endpoint;

		if (!SCM80211_USB_IS_DIR_IN(pipe->ep_address))
			pipe->flags |= SCM80211_USB_PIPE_FLAG_TX;
		else
			pipe->flags |= SCM80211_USB_PIPE_FLAG_RX;

		ret = scm80211_usb_alloc_pipe_urb_context(hw, interface, pipe, urb_count);
		if (ret)
			return ret;

		pipe->flags |= SCM80211_USB_PIPE_FLAG_INIT;
	}

	return 0;
}

#define	MAJOR_MASK	0xFF0000
#define SUB_MASK	0x00FF00
#define PATCH_MASK	0x0000FF
#define VERSION		KERNEL_VERSION

static u32 scm80211_get_wise_version(struct ieee80211_hw *hw)
{
	u32 ver = 0;

	scm80211_usb_param_read(hw, FW_VER, &ver, sizeof(ver));
	scm80211_info(hw, "WISE VERSION %d.%d.%d\n",
		      FIELD_GET(MAJOR_MASK, ver), FIELD_GET(SUB_MASK, ver),
		      FIELD_GET(PATCH_MASK, ver));
	return ver;
}

static int scm80211_usb_device_init(struct ieee80211_hw *hw, struct usb_interface *interface)
{
	struct usb_device *dev = interface_to_usbdev(interface);
	struct scm80211_usb_priv *usb_priv = hw_to_usb_priv(hw);
	struct scm80211_usb_pipe *pipe;
	int ret = -1, i;
	u32 want, real;

	usb_set_intfdata(interface, hw);
	usb_priv->udev = dev;
	usb_priv->interface = interface;
	usb_priv->hw = hw;
	want = VERSION(3, 0, 0);

	real = scm80211_get_wise_version(hw);
	if (real < want) {
		scm80211_err(hw, "wrong fw version %d.%d.%d needed but %d.%d.%d got\n",
			     FIELD_GET(MAJOR_MASK, want),
			     FIELD_GET(SUB_MASK, want),
			     FIELD_GET(PATCH_MASK, want),
			     FIELD_GET(MAJOR_MASK, real),
			     FIELD_GET(SUB_MASK, real),
			     FIELD_GET(PATCH_MASK, real));
		return -EINVAL;
	}

	if (dev->bus->sg_tablesize > 0 && dev->bus->no_sg_constraint)
		usb_priv->sg_size = min(dev->bus->sg_tablesize, (unsigned int)MAX_SG_NUM);
	else
		usb_priv->sg_size = 0;

#ifdef ERRATA_USB_HCD_SG
	usb_priv->sg_size = 0;
#endif

	scm80211_trace(hw, "+ %s urb sg %ssupported %s(size: %d)\n",
		       __func__, usb_priv->sg_size ? "" : "un",
		       dev->bus->no_sg_constraint ? "" : "have sg constraint ", usb_priv->sg_size);

	for (i = 0; i < SCM80211_USB_PIPE_MAX; i++) {
		pipe = &usb_priv->pipes[i];
		spin_lock_init(&pipe->cs_lock);
		INIT_LIST_HEAD(&pipe->urb_list_head);
		INIT_LIST_HEAD(&pipe->urb_start_head);
		INIT_LIST_HEAD(&pipe->urb_complete_head);
		init_usb_anchor(&pipe->urb_submitted);
	}

	ret = scm80211_usb_init_pipe(hw, interface);

	if (ret)
		return ret;

	for (i = 0; i < SCM80211_USB_PIPE_MAX; i++) {
		pipe = &usb_priv->pipes[i];
		scm80211_trace(hw, " %s pipe_num %d\n", __func__, pipe->pipe_num);
		if (!(pipe->flags & SCM80211_USB_PIPE_FLAG_INIT))
			return -1;
	}

	return 0;
}

static int scm80211_phy_rf_module_init(struct ieee80211_hw *hw, struct usb_device *dev)
{
	struct scm80211_rf_device *rf;
	int ret = -1;

	do {
		/*
		 * Chip-specific initialization
		 * At this moment just scm2020
		 */
		ret = scm2020_chip_init(hw);
		if (ret) {
			dev_err(&dev->dev, "failed to initialize device\n");
			break;
		}
		rf = NULL;
		ret = 0;
	} while (0);

	return ret;
}

/* defined not used on usb */
void scm80211_tx_irq(struct ieee80211_hw *hw, unsigned long queues)
{
}

irqreturn_t scm80211_handle_irq(int irq, void *data)
{
	return IRQ_NONE;
}

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
static int scm80211_soc_init(struct ieee80211_hw *hw)
{
	int path;
	u32 v;

	/* quickly enable mac/phy clk */
	v = scm80211_raw_readl(hw, WLAN_WMP_CFG);
	v |= PHY_CLK_EN;
	v |= PHY_PCLK_EN;
	v |= MAC_CLK_M1X_EN(0);
	for (path = 0; path < 2; path++) {
		//v |= DPD_960M_CLK_EN(path);
		//v |= DPD_320M_CLK_EN(path);
		v |= TX_EN(path);
		v |= RX_EN(path);
	}
	scm80211_raw_writel(hw, v, WLAN_WMP_CFG);

	udelay(10);

	return 0;
}
#endif

static struct ieee80211_hw *scm80211_usb_alloc_ieee80211_hw(struct usb_device *dev)
{
	struct ieee80211_hw *hw;
	struct ieee80211_hw_priv *priv;

	hw = ieee80211_alloc_hw(sizeof(*priv) + sizeof(struct scm80211_usb_priv), &scm80211_ops);
	if (!hw) {
		dev_err(&dev->dev, "failed to alloc hw\n");
		return NULL;
	}

	SET_IEEE80211_DEV(hw, &dev->dev);
	priv = hw->priv;
	priv->hw = hw;
	priv->devid = 0;
	priv->bus_type = USB_BUS;
	priv->bus.ops = &scm80211_usb_ops;
	priv->bcdusb = le16_to_cpu(dev->descriptor.bcdUSB) & 0xf00;
	dma_set_coherent_mask(&dev->dev, DMA_BIT_MASK(64));
	ewma_tx_sched_time_init(&priv->sched_time);

	scm80211_set_debug_trace(hw);
	scm80211_hw_band_set(hw);

	return hw;
}

#define FWDL_BLOCK_SIZE		64
static int scm80211_usb_mac_mcu_fw_download(struct ieee80211_hw *hw)
{
	struct device *dev = wiphy_dev(hw->wiphy);
	u32 addr = MAC_BASE_0 + MAC_IMEM_OFFSET;
	const struct firmware *fw;
	const char *firmware = "senscomm/scm2020.bin";
	int err;
	u32 size = 0;
	u8 *data = NULL;

	err = request_firmware(&fw, firmware, dev);
	if (err) {
		scm80211_err(hw, "failed to request firmware (err=%d)\n", err);
		return err;
	}

	scm80211_trace(hw, "downloading ucode %s (%zu bytes)\n", firmware, fw->size);

	data = kmalloc(fw->size, GFP_KERNEL);
	if (!data)
		return -ENOMEM;
	memcpy(data, fw->data, fw->size);

	while ((fw->size - size) > 0) {
		if ((fw->size - size) <= FWDL_BLOCK_SIZE) {
			err = scm80211_usb_fw_download(hw, addr + size, data + size,
						       fw->size - size);
			break;
		}
		err = scm80211_usb_fw_download(hw, addr + size, data + size, FWDL_BLOCK_SIZE);
		if (err == FWDL_BLOCK_SIZE)
			size += FWDL_BLOCK_SIZE;
		else
			break;
	}

	if (err <= 0)
		scm80211_err(hw, "ucode download fail\n");

	release_firmware(fw);
	kfree(data);

	return err;
}

/**
 * scm80211_usb_start_mcu() - download MCU microcode and kick it
 */
static int scm80211_usb_start_mcu(struct ieee80211_hw *hw)
{
	int err, retries;
	u32 v;

	err = scm80211_usb_mac_mcu_fw_download(hw);
	if (err < 0)
		return err;

	/* Kick MCU */
	scm80211_trace(hw, "starting mcu\n");

	scm80211_mac_writel(hw, 0, MAC_REG_MCU_CFG);
	udelay(10);

	v = MCU_CFG_RSTB;
	scm80211_mac_writel(hw, v, MAC_REG_MCU_CFG);

#ifdef SCM80211_FTM
	v = scm80211_raw_readl(hw, WLAN_WMP_CFG);
	v |= MAC_CLK_M1X_EN(1); //for path 1
	scm80211_raw_writel(hw, v, WLAN_WMP_CFG);
#endif

	err = -EBUSY;
	for (retries = 0; retries < MCU_START_RETRIES; retries++) {
		v = scm80211_mac_readl(hw, MAC_REG_MCU_STATUS);
		if (FIELD_GET(MCU_STATUS_DONE, v) == 1) {
			err = 0;
			break;
		}
		mdelay(MCU_START_DELAY);
	}
	if (err)
		scm80211_err(hw, "failed to start mcu\n");

	return err;
}

static int scm80211_usb_init_hw_info(struct ieee80211_hw *hw, struct usb_device *udev)
{
	int ret = -1;
	bool rand_mac = true;
	u8 addr[ETH_ALEN] = { 0x64, 0xf9, 0x47, 0x00, 0x00, 0x00 };

#ifdef EFUSE_ENABLE
	u8 buff[MAX_EFUSE_SIZE] = { 0 };

	ret = scm80211_usb_efuse_readl(udev, buff, 0, MAX_EFUSE_SIZE);
	scm80211_efuse_usb_init(udev, buff, MAX_EFUSE_SIZE);
	if (ret >= 0)
		rand_mac = !scm80211_efuse_mac_update(addr, NULL);
	else
		dev_info(&udev->dev, "usb_init cannot get efuse data(%d)\n", ret);
#endif

	do {
#if defined(SCM2020_Z1) || defined(SCM2020_A0)
		scm80211_soc_init(hw);
		mdelay(2);
#endif
		ret = scm80211_usb_start_mcu(hw);
		if (ret) {
			dev_err(&udev->dev, "failed to start MCU\n");
			break;
		}

		scm80211_phy_rf_module_init(hw, udev);

		if (rand_mac) {
			addr[3] = prandom_u32() & 0xff;
			addr[4] = prandom_u32() & 0xff;
			addr[5] = prandom_u32() & 0xff;
		}

		ret = scm80211_init_hw(hw, addr);
		if (ret) {
			dev_err(&udev->dev, "scm80211_init_hw failed (%d)\n", ret);
			break;
		}

		ret = ieee80211_register_hw(hw);
		if (ret) {
			dev_err(&udev->dev, "ieee80211_register_hw failed (%d)\n", ret);
			break;
		}

#ifdef CONFIG_MAC80211_DEBUGFS
		scm80211_register_dbgfs(hw);
#endif
		scm80211_rfkill_init(hw);

		ret = 0;
	} while (0);

	return ret;
}

extern int scm2020_usb_dl_probe(struct usb_interface *intf, const struct usb_device_id *id);

static int scm80211_usb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct ieee80211_hw *hw = NULL;
	struct usb_device *dev = interface_to_usbdev(interface);
	int ret = -1;

	usb_get_dev(dev);
	dev_notice(&interface->dev, "Senscomm wireless device: %s, build time=%s\n",
		   dev_name(&interface->dev), DRIVER_DATE);
	/*
	 * temporary disable auto suspend for some pc device can not wakeup
	 * Need take a look further
	 */
	pm_runtime_set_autosuspend_delay(&dev->dev, -1);
	if (interface->cur_altsetting->desc.bInterfaceClass == USB_CLASS_COMM)
		return scm2020_usb_dl_probe(interface, id);

	scm80211_register_oops_handler();
	dma_set_mask(&dev->dev, DMA_BIT_MASK(32));

	if (dev->lpm_disable_count == 0)
		usb_disable_lpm(dev);
	dev_info(&dev->dev, "lpm_disable_count=%d\n", dev->lpm_disable_count);

	do {
		hw = scm80211_usb_alloc_ieee80211_hw(dev);
		if (!hw) {
			ret = -ENOMEM;
			break;
		}

		ret = scm80211_usb_device_init(hw, interface);
		if (ret) {
			if (ret == -EINVAL) {
				ieee80211_free_hw(hw);
				hw = NULL;
			}
			dev_err(&dev->dev, "Senscomm usb device init fail\n");
			break;
		}
		ret = scm80211_usb_init_hw_info(hw, dev);
		if (ret) {
			dev_err(&dev->dev, "Senscomm init hw fail\n");
			break;
		}
		ret = 0;
	} while (0);

	if (ret) {
		if (hw) {
			scm80211_usb_destroy(hw);
			ieee80211_free_hw(hw);
		}
		scm80211_unregister_oops_handler();
		usb_put_dev(dev);
	} else {
		struct ieee80211_hw_priv *priv = hw->priv;
		struct scm80211_usb_priv *usb_priv = hw_priv_to_usb_priv(priv);

		INIT_WORK(&usb_priv->usb_bus_monitor_work, scm80211_usb_ep_status_check);
		usb_enable_lpm(dev);
		dev_info(&dev->dev, "lpm_disable_count=%d (expected=0)\n", dev->lpm_disable_count);
	}

	return ret;
}

extern void scm2020_usb_dl_disconnect(struct usb_interface *intf);
static void scm80211_usb_remove(struct usb_interface *interface)
{
	struct ieee80211_hw *hw = NULL;
	struct ieee80211_hw_priv *priv = NULL;
	struct scm80211_usb_priv *usb_priv = NULL;

	if (interface->cur_altsetting->desc.bInterfaceClass == USB_CLASS_COMM)
		return scm2020_usb_dl_disconnect(interface);

	hw = usb_get_intfdata(interface);
	priv = hw->priv;
	usb_priv = hw_priv_to_usb_priv(priv);

	scm80211_rfkill_deinit(hw);

#ifdef SCM80211_USES_CHANCTX
	scm80211_deinit_chanctx_sched(hw);
#endif
#ifdef CONFIG_MAC80211_DEBUGFS
	scm80211_unregister_dbgfs(hw);
#endif
	ieee80211_unregister_hw(hw);
	scm80211_detach_rf(hw);

	scm80211_usb_destroy(hw);
	usb_put_dev(interface_to_usbdev(interface));
	cancel_work_sync(&usb_priv->usb_bus_monitor_work);
	ieee80211_free_hw(hw);
	scm80211_unregister_oops_handler();
}

static int scm80211_usb_suspend(struct usb_interface *interface, pm_message_t message)
{
	struct ieee80211_hw *hw = usb_get_intfdata(interface);

	scm80211_trace(hw, "SCM2020 Going to susupend!\n");
	/* scm80211_usb_cancel_all_request(hw); */
	return 0;
}

static int scm80211_usb_resume(struct usb_interface *interface)
{
	struct ieee80211_hw *hw = usb_get_intfdata(interface);

	scm80211_trace(hw, "SCM2020 Going to resume!\n");

	// struct ieee80211_hw_priv *priv = hw->priv;
	// scm80211_usb_bulk_rx(hw, &(priv->usb_priv.pipes[SCM80211_USB_PIPE_RX_DATA]));

	return 0;
}

/* list usb device supported */
static struct usb_device_id scm80211_usb_id_table[] = {
	{ USB_DEVICE_INTERFACE_CLASS(0x3464, 0x2020, USB_CLASS_VENDOR_SPEC) },
	{ USB_DEVICE_INTERFACE_CLASS(0x3464, 0x2020, USB_CLASS_COMM) },
	{ USB_DEVICE_INTERFACE_CLASS(0x2ecc, 0x3004, USB_CLASS_COMM) },
	{ },
};

MODULE_DEVICE_TABLE(usb, scm80211_usb_id_table);

static struct usb_driver scm80211_usb_driver = {
	.name = "scm80211-usb",
	.probe = scm80211_usb_probe,
	.suspend = scm80211_usb_suspend,
	.resume = scm80211_usb_resume,
	.disconnect = scm80211_usb_remove,
	.id_table = scm80211_usb_id_table,
	.supports_autosuspend = true,
#ifndef SCM80211_PS
	.disable_hub_initiated_lpm = 1
#endif
};

static int __init scm80211_usb_init(void)
{
	int ret;

	ret = scm80211_minstrel_init();
	if (ret)
		pr_info("rate control module has been registered\n");

	ret = usb_register(&scm80211_usb_driver);
	if (ret) {
		pr_err("failed to register usb driver");
		scm80211_minstrel_exit();
	}

	return ret;
}

static void __exit scm80211_usb_exit(void)
{
	usb_deregister(&scm80211_usb_driver);
	scm80211_minstrel_exit();
}

module_init(scm80211_usb_init);
module_exit(scm80211_usb_exit);

MODULE_VERSION(DRIVER_VERSION);
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Senscomm Semiconductor, Inc.");
MODULE_DESCRIPTION("Senscomm 802.11ax USB wifi");
