// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 1999 Armin Fuerst	<fuerst@in.tum.de>
 * Copyright (c) 1999 Pavel Machek	<pavel@ucw.cz>
 * Copyright (c) 1999 Johannes Erdfelt	<johannes@erdfelt.com>
 * Copyright (c) 2000 Vojtech Pavlik	<vojtech@suse.cz>
 * Copyright (c) 2004 Oliver Neukum	<oliver@neukum.name>
 * Copyright (c) 2005 David Kubicek	<dave@awk.cz>
 * Copyright (c) 2011 Johan Hovold	<jhovold@gmail.com>
 * Copyright (c) 2022 Senscomm Semiconductor Co., Ltd
 *
 */

#undef DEBUG
#undef VERBOSE_DEBUG

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/log2.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/uaccess.h>
#include <linux/usb.h>
#include <linux/usb/cdc.h>
#include <asm/byteorder.h>
#include <asm/unaligned.h>
#include <linux/idr.h>
#include <linux/list.h>
#include <linux/version.h>
#include <linux/firmware.h>
#include <net/mac80211.h>

#include "usb_dl.h"
#include "rf.h"

static char cmd_line[ABOOT_COMMAND_SZ];
static char response[ABOOT_COMMAND_SZ];
static bool is_crane_bootrom;

static const char cmd_getvar[] = "getvar:";
static const char cmd_download[] = "download:";
static const char cmd_call[] = "call";
static const char cmd_nop[] = "nop";
static const char cmd_reboot[] = "reboot";
static const char cmd_complete[] = "complete";
static const char cmd_disconnect[] = "disconnect";
static const char cmd_flash[] = "flash";
static const char crane_version_bootrom[] = "2019.01.15";

/* Little helpers: write/read buffers free */
static void scm2020_write_buffers_free(struct scm2020_usb_dl *usb_dl)
{
	int i;
	struct scm2020_usb_buff *wb;

	for (wb = &usb_dl->wb[0], i = 0; i < ACM_NW; i++, wb++) {
		if (wb->urb) {
			usb_kill_urb(wb->urb);
			usb_free_urb(wb->urb);
		}

		if (wb->buf)
			usb_free_coherent(usb_dl->udev, usb_dl->writesize, wb->buf, wb->dmah);
	}
}

/* Little helper: write buffers allocate */
static int scm2020_write_buffers_alloc(struct scm2020_usb_dl *usb_dl)
{
	int i;
	struct scm2020_usb_buff *wb;

	for (wb = &usb_dl->wb[0], i = 0; i < ACM_NW; i++, wb++) {
		wb->buf = usb_alloc_coherent(usb_dl->udev, usb_dl->writesize, GFP_KERNEL,
					     &wb->dmah);
		wb->urb = usb_alloc_urb(0, GFP_KERNEL);
		if (!wb->buf || !wb->urb) {
			while (i != 0) {
				if (wb->buf) {
					usb_free_coherent(usb_dl->udev, usb_dl->writesize,
							  wb->buf, wb->dmah);
				}
				usb_free_urb(wb->urb);
				i--;
				wb--;
			}
			return -ENOMEM;
		}
		wb->idx = i;
		wb->instance = usb_dl;
		list_add_tail(&wb->list, &usb_dl->write_list);
	}
	return 0;
}

static int scm2020_write_buffer_init(struct scm2020_usb_dl *usb_dl)
{
	spin_lock_init(&usb_dl->wb_lock);
	INIT_LIST_HEAD(&usb_dl->write_list);
	return scm2020_write_buffers_alloc(usb_dl);
}

/* Little helpers: write/read buffers free */
static void scm2020_read_buffers_free(struct scm2020_usb_dl *usb_dl)
{
	int i;
	struct scm2020_usb_buff *rb;

	for (rb = &usb_dl->rb[0], i = 0; i < ACM_NR; i++, rb++) {
		if (rb->urb) {
			usb_kill_urb(rb->urb);
			usb_free_urb(rb->urb);
		}

		if (rb->buf)
			usb_free_coherent(usb_dl->udev, usb_dl->readsize, rb->buf, rb->dmah);
	}
}

/* Little helper: read buffers allocate */
static int scm2020_read_buffers_alloc(struct scm2020_usb_dl *usb_dl)
{
	int i;
	struct scm2020_usb_buff *rb;

	for (rb = &usb_dl->rb[0], i = 0; i < ACM_NR; i++, rb++) {
		rb->buf = usb_alloc_coherent(usb_dl->udev, usb_dl->readsize, GFP_KERNEL,
					     &rb->dmah);
		rb->urb = usb_alloc_urb(0, GFP_KERNEL);
		if (!rb->buf || !rb->urb) {
			while (i != 0) {
				if (rb->buf) {
					usb_free_coherent(usb_dl->udev, usb_dl->writesize,
							  rb->buf, rb->dmah);
				}
				usb_free_urb(rb->urb);
				i--;
				rb--;
			}
			return -ENOMEM;
		}
		rb->idx = i;
		rb->instance = usb_dl;
		list_add_tail(&rb->list, &usb_dl->read_list);
	}
	return 0;
}

static int scm2020_read_buffer_init(struct scm2020_usb_dl *usb_dl)
{
	spin_lock_init(&usb_dl->rb_lock);
	INIT_LIST_HEAD(&usb_dl->read_list);
	return scm2020_read_buffers_alloc(usb_dl);
}

static const char * const step_string[] = {
	"init", "start", "status", "fw_download", "progress", "reboot", "finsh"
};

static void scm2020_preamble_callback(struct scm2020_usb_dl *usb_dl, struct scm2020_usb_buff *rb)
{
	struct device *dev = &usb_dl->intf->dev;
	char preamble_buffer[ABOOT_PREAMBLE_SIZE + 1];
	int preamble_index = -1;
	u8 *data = rb->buf;
	int size = rb->len;

	while (size--) {
		if (preamble_index == -1) {
			if (*data == ABOOT_FRAME_DELIMITER)
				preamble_index = 0;
		} else if (preamble_index < ABOOT_PREAMBLE_SIZE) {
			if (*data == ABOOT_FRAME_DELIMITER)
				preamble_index = 0;
			else
				preamble_buffer[preamble_index++] = *data;
		} else if (preamble_index == ABOOT_PREAMBLE_SIZE) {
			if (*data == ABOOT_FRAME_DELIMITER) {
				/* notice start success */
				if (!strncmp
				    (preamble_buffer, ABOOT_PREAMBLE_UUUU, ABOOT_PREAMBLE_SIZE)) {
					break;
				}
				if (!strncmp
				    (preamble_buffer, ABOOT_PREAMBLE_UABT, ABOOT_PREAMBLE_SIZE)) {
					dev_dbg(dev, "notice aboot-tiny start success\n");
					usb_dl->step = DL_START;
					if (delayed_work_pending(&usb_dl->work))
						mod_delayed_work(usb_dl->workqueue, &usb_dl->work,
								 msecs_to_jiffies(5));
					else
						queue_delayed_work(usb_dl->workqueue, &usb_dl->work,
								   msecs_to_jiffies(5));
					break;
				}
				preamble_index = 0;
			}
		} else {
			preamble_index = -1;
		}
		data++;
	}

}

static int scm2020_parse_incoming_cmd(struct scm2020_usb_dl *usb_dl, const char *cmd)
{
	struct device *dev = &usb_dl->intf->dev;

	dev_dbg(dev, "cmd [%s]\n", cmd);
	if (!memcmp(cmd, "DATA", 4) || !memcmp(cmd, "OKAY", 4) || !memcmp(cmd, "FAIL", 4)) {
		usb_dl->step = DL_STATUS;

		WARN_ON(!list_empty(&usb_dl->cmd[usb_dl->cmd_idx].list));
		memcpy(usb_dl->cmd[usb_dl->cmd_idx].resp, cmd, usb_dl->resp_len);
		usb_dl->cmd[usb_dl->cmd_idx].len = usb_dl->resp_len;
		list_add_tail(&usb_dl->cmd[usb_dl->cmd_idx].list, &usb_dl->cmd_list);

		if (++usb_dl->cmd_idx == ACM_NW)
			usb_dl->cmd_idx = 0;

		if (delayed_work_pending(&usb_dl->work))
			mod_delayed_work(usb_dl->workqueue, &usb_dl->work, 0);
		else
			queue_delayed_work(usb_dl->workqueue, &usb_dl->work, msecs_to_jiffies(1));
		return 1;
	}
	return 0;
}

static void scm2020_incoming_data(struct scm2020_usb_dl *usb_dl, struct scm2020_usb_buff *rb)
{
	int i;
	u8 c;
	enum line_state state = WAIT_FRAMESTART;

	memset(usb_dl->response, 0, ABOOT_COMMAND_SZ + 1);
	usb_dl->resp_len = 0;

	for (i = 0; i < rb->len; i++) {
		c = rb->buf[i];
		switch (state) {
		case WAIT_FRAMESTART:
			if (c == SMUX_FRAME_DELIMITER)
				state = IN_FRAME;
			break;
		case IN_FRAME:
			if (c == SMUX_ESC_CHAR) {
				state = IN_ESCAPE;
			} else if (c == SMUX_FRAME_DELIMITER) {
				if (usb_dl->resp_len == 0 || usb_dl->resp_len < 4) {
					usb_dl->resp_len = 0;
					state = IN_FRAME;
					continue;
				}
				usb_dl->response[usb_dl->resp_len] = '\0';
				if (usb_dl->rtype == SMUX_FRAME_TYPE_ABOOT_CMD)
					scm2020_parse_incoming_cmd(usb_dl, usb_dl->response);
				usb_dl->resp_len = 0;
				state = IN_FRAME;

				continue;
			} else {
				usb_dl->response[usb_dl->resp_len++] = c;
			}
			break;
		case IN_ESCAPE:
			switch (c) {
			case (SMUX_FRAME_DELIMITER ^ 0x20):
			case (SMUX_ESC_CHAR ^ 0x20):
				usb_dl->response[usb_dl->resp_len++] = c;
				break;
			case (SMUX_FRAME_TYPE_STDIO ^ 0x20):
				usb_dl->rtype = SMUX_FRAME_TYPE_STDIO;
				break;
			case (SMUX_FRAME_TYPE_HELLO ^ 0x20):
				usb_dl->rtype = SMUX_FRAME_TYPE_HELLO;
				break;
			case (SMUX_FRAME_TYPE_HELLO_REPLY ^ 0x20):
				usb_dl->rtype = SMUX_FRAME_TYPE_HELLO_REPLY;
				break;
			case (SMUX_FRAME_TYPE_ABOOT_CMD ^ 0x20):
				usb_dl->rtype = SMUX_FRAME_TYPE_ABOOT_CMD;
				break;
			case (SMUX_FRAME_TYPE_ABOOT_DATA ^ 0x20):
				usb_dl->rtype = SMUX_FRAME_TYPE_ABOOT_DATA;
				break;
			case (SMUX_FRAME_TYPE_HEART_BEAT ^ 0x20):
				usb_dl->rtype = SMUX_FRAME_TYPE_HEART_BEAT;
				break;
			default:
				break;
			}
			state = IN_FRAME;
		default:
			break;
		}

		if (usb_dl->resp_len >= ABOOT_COMMAND_SZ)
			break;
	}
}

static void scm2020_process_received_packet(struct scm2020_usb_buff *rb)
{
	struct scm2020_usb_dl *usb_dl = rb->instance;
	unsigned long flags;

	spin_lock_irqsave(&usb_dl->lock, flags);
	switch (usb_dl->step) {
	case DL_INIT:
		scm2020_preamble_callback(usb_dl, rb);
		break;
	case DL_START:
	case DL_STATUS:
		scm2020_incoming_data(usb_dl, rb);
	default:
		break;
	}
	spin_unlock_irqrestore(&usb_dl->lock, flags);
}

static int scm2020_prepare_rx_urbs(struct scm2020_usb_dl *usb_dl);
static void scm2020_read_bulk_callback(struct urb *urb)
{
	struct scm2020_usb_buff *rb = urb->context;
	struct scm2020_usb_dl *usb_dl = rb->instance;
	unsigned long flags;

	if (urb->status == 0) {
		rb->len = urb->actual_length;
		rb->buf[rb->len] = (u8)'\0';
		scm2020_process_received_packet(rb);
	}

	spin_lock_irqsave(&usb_dl->rb_lock, flags);
	list_add_tail(&rb->list, &usb_dl->read_list);
	spin_unlock_irqrestore(&usb_dl->rb_lock, flags);
	if (!urb->status)
		scm2020_prepare_rx_urbs(usb_dl);
}

static int scm2020_prepare_rx_urbs(struct scm2020_usb_dl *usb_dl)
{
	struct device *dev = &usb_dl->intf->dev;
	struct scm2020_usb_buff *rb, *save;
	unsigned long flags;
	int ret = -ENOMEM;

	if (usb_dl->device_lost)
		return -ENODEV;

	if (list_empty(&usb_dl->read_list)) {
		dev_dbg(dev, "no rx buffer left\n");
		return ret;
	}

	spin_lock_irqsave(&usb_dl->rb_lock, flags);
	list_for_each_entry_safe(rb, save, &usb_dl->read_list, list) {
		if (!rb->urb) {
			dev_info(dev, "no prepared rb->urb\n");
			rb->urb = usb_alloc_urb(0, GFP_ATOMIC);
			if (!rb->urb)
				goto unlock;
		}
		rb->urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
		rb->urb->transfer_dma = rb->dmah;
		usb_fill_bulk_urb(rb->urb, usb_dl->udev,
				  usb_dl->in, rb->buf, usb_dl->readsize,
				  scm2020_read_bulk_callback, rb);
		ret = usb_submit_urb(rb->urb, GFP_ATOMIC);
		if (ret) {
			dev_err(dev, "usb_submit_urb rb error %d\n", ret);
			usb_dl->device_lost = true;
			goto unlock;
		}
		list_del_init(&rb->list);
	}
unlock:
	spin_unlock_irqrestore(&usb_dl->rb_lock, flags);
	return ret;
}

static struct scm2020_usb_buff *scm2020_get_usb_wb(struct scm2020_usb_dl *usb_dl)
{
	struct scm2020_usb_buff *buff;
	unsigned long flags;

	spin_lock_irqsave(&usb_dl->wb_lock, flags);
	buff = list_first_entry_or_null(&usb_dl->write_list, struct scm2020_usb_buff, list);
	if (buff)
		list_del_init(&buff->list);
	else
		dev_info(&usb_dl->intf->dev, "no wb get\n");
	spin_unlock_irqrestore(&usb_dl->wb_lock, flags);
	return buff;
}

static void scm2020_write_bulk_callback(struct urb *urb)
{
	struct scm2020_usb_buff *wb = urb->context;
	struct scm2020_usb_dl *usb_dl = wb->instance;
	unsigned long flags;

	spin_lock_irqsave(&usb_dl->wb_lock, flags);
	list_add_tail(&wb->list, &usb_dl->write_list);
	spin_unlock_irqrestore(&usb_dl->wb_lock, flags);
}

static void scm2020_submit_wb(struct scm2020_usb_dl *usb_dl, struct scm2020_usb_buff *wb)
{
	struct device *dev = &usb_dl->intf->dev;
	int ret;

	if (!wb->urb) {
		dev_info(dev, "no prepared wb->urb\n");
		wb->urb = usb_alloc_urb(0, GFP_ATOMIC);
		if (!wb->urb)
			goto error;
	}
	wb->urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
	wb->urb->transfer_dma = wb->dmah;
	usb_fill_bulk_urb(wb->urb, usb_dl->udev, usb_dl->out,
			  wb->buf, wb->len, scm2020_write_bulk_callback, wb);
	ret = usb_submit_urb(wb->urb, GFP_ATOMIC);
	if (ret) {
		dev_err(dev, "usb_submit_urb wb error %d usb state %d\n", ret, usb_dl->udev->state);
		goto error;
	}
	return;
error:
	list_add_tail(&wb->list, &usb_dl->write_list);
}

static int download_read_line(struct scm2020_usb_dl *usb_dl, u8 *buffer)
{
	const u8 *data, *p;
	size_t size;

	size = usb_dl->fw->size - usb_dl->off;

	while (size > 0) {
		size_t len;

		data = usb_dl->fw->data + usb_dl->off;
		size = usb_dl->fw->size - usb_dl->off;
		p = memchr(data, '\n', size > ABOOT_COMMAND_SZ ? ABOOT_COMMAND_SZ : size);
		if (!p)
			return -1;
		len = p - data;
		usb_dl->off += (len + 1);
		if (len == 0)
			continue;
		memcpy(buffer, data, len);
		buffer[len] = '\0';
		return len;
	}
	return 0;
}

static int scm2020_add_delimiter(struct scm2020_usb_dl *usb_dl, u8 *buff, int size)
{
	struct scm2020_usb_buff *wb;

	wb = scm2020_get_usb_wb(usb_dl);
	if (!wb)
		return -ENOMEM;

	memset(wb->buf, 0, usb_dl->writesize);
	wb->len = 0;

	while (size) {
		u8 *pos;
		int mtu = 1024;

		if (wb->len + 4 > usb_dl->writesize) {
			scm2020_submit_wb(usb_dl, wb);
			wb = scm2020_get_usb_wb(usb_dl);
			if (!wb)
				return -ENOMEM;
			memset(wb->buf, 0, usb_dl->writesize);
			wb->len = 0;
			continue;
		}
		pos = wb->buf + wb->len;
		*pos++ = SMUX_FRAME_DELIMITER;
		if (usb_dl->wtype) {
			*pos++ = SMUX_ESC_CHAR;
			*pos++ = usb_dl->wtype ^ 0x20;
		}

		wb->len = pos - wb->buf;
		while (size && (wb->len + 3 <= usb_dl->writesize) && mtu--) {
			u8 c = *buff++;

			switch (c) {
			case SMUX_FRAME_DELIMITER:
				*pos++ = SMUX_ESC_CHAR;
				*pos++ = SMUX_FRAME_DELIMITER ^ 0x20;
				break;
			case SMUX_ESC_CHAR:
				*pos++ = SMUX_ESC_CHAR;
				*pos++ = SMUX_ESC_CHAR ^ 0x20;
				break;
			default:
				*pos++ = c;
				break;
			}
			size--;
			wb->len = pos - wb->buf;
		}

		*pos++ = SMUX_FRAME_DELIMITER;
		wb->len = pos - wb->buf;
		if (!size)
			scm2020_submit_wb(usb_dl, wb);
	}

	return 0;
}

static void download_cmd_read_response(struct scm2020_usb_dl *usb_dl)
{
	if (download_read_line(usb_dl, cmd_line) == 0) {
		/* read to end reboot it */
		strcpy(cmd_line, cmd_reboot);
		strcpy(response, "OKAY");
		usb_dl->step = DL_REBOOT;
		queue_delayed_work(usb_dl->workqueue, &usb_dl->work, msecs_to_jiffies(10));
	} else {
		download_read_line(usb_dl, response);
	}
	usb_dl->wtype = SMUX_FRAME_TYPE_ABOOT_CMD;
	scm2020_add_delimiter(usb_dl, cmd_line, strlen(cmd_line));
}

static int download_handle_getvar(struct scm2020_usb_dl *usb_dl, const char *variable, char *expect)
{
	struct device *dev = &usb_dl->intf->dev;
	char *rx_resp;
	struct scm2020_msg *msg;

	msg = list_first_entry_or_null(&usb_dl->cmd_list, struct scm2020_msg, list);
	if (!msg)
		return -1;

	list_del_init(&msg->list);
	rx_resp = msg->resp;

	if (memcmp(expect, rx_resp, 4))
		return -1;

	expect += 4;
	rx_resp += 4;

	if (!strcmp(variable, "max-download-size")) {
		u32 expect_size = 0;
		u32 response_size = 0;

		while (!((*expect > '0' && *expect <= '9') || (*expect >= 'a' && *expect <= 'f')))
			expect++;

		while (!((*rx_resp > '0' && *rx_resp <= '9') ||
			 (*rx_resp >= 'a' && *rx_resp <= 'f'))) {
			rx_resp++;
		}

		if (kstrtou32(expect, 16, (void *)&expect_size) ||
		    kstrtou32(rx_resp, 16, (void *)&response_size)) {
			dev_err(dev, "get len fail %u, %u, %d\n", expect_size, response_size,
				kstrtou32(expect, 16, (void *)&expect_size));
			return -1;
		}
		if (expect_size > response_size) {
			dev_err(dev, "len compare fail %u, %u\n", expect_size, response_size);
			return -1;
		}
	} else if (!strcmp(variable, "version-bootrom")) {
		if (strcmp(expect, rx_resp))
			return -1;
		if (strcmp(crane_version_bootrom, rx_resp))
			is_crane_bootrom = false;
		else
			is_crane_bootrom = true;
	} else {
		char *token;

		while ((token = strsep(&expect, "|"))) {
			if (!strcmp(token, rx_resp))
				break;
		}
		if (!token)
			return -1;
	}

	return 0;
}

static int download_handle_download(struct scm2020_usb_dl *usb_dl, u32 size, char *expect)
{
	char *rx_resp;
	u32 expect_size, response_size;
	struct scm2020_msg *msg;

	msg = list_first_entry_or_null(&usb_dl->cmd_list, struct scm2020_msg, list);
	if (!msg)
		return -1;

	list_del_init(&msg->list);
	rx_resp = msg->resp;

	if (memcmp(expect, rx_resp, 4))
		return -1;

	expect += 4;
	rx_resp += 4;

	if (kstrtou32(expect, 16, (void *)&expect_size) ||
	    kstrtou32(rx_resp, 16, (void *)&response_size))
		return -1;

	if (size != response_size || size != expect_size)
		return -1;

	return 0;
}

#define MAX_DOWNLOAD_BUF_SIZE 1024
static int download_do_download_data(struct scm2020_usb_dl *usb_dl, u32 size)
{
	const u8 *data = usb_dl->fw->data + usb_dl->off;
	u8 *buff = kmalloc(size, GFP_ATOMIC);

	if (!buff)
		return -ENOMEM;

	if (usb_dl->off + size > usb_dl->fw->size)
		return -EINVAL;

	memcpy(buff, data, size);
	usb_dl->wtype = SMUX_FRAME_TYPE_ABOOT_DATA;
	scm2020_add_delimiter(usb_dl, buff, size);
	usb_dl->off += size;
	kfree(buff);
	return 0;
}

static void scm2020_fw_download_work(struct work_struct *work)
{
	struct scm2020_usb_dl *usb_dl = container_of(work, struct scm2020_usb_dl, work.work);
	struct device *dev = &usb_dl->intf->dev;
	struct scm2020_usb_buff *wb;
	unsigned long flags;

	spin_lock_irqsave(&usb_dl->lock, flags);

	dev_dbg(dev, "%s, in [%s] step\n", __func__, step_string[usb_dl->step]);
	switch (usb_dl->step) {
	case DL_INIT:
		scm2020_prepare_rx_urbs(usb_dl);

		wb = scm2020_get_usb_wb(usb_dl);
		if (!wb)
			break;

		memset(wb->buf, 0, ABOOT_PREAMBLE_SIZE + 1);
		memcpy(wb->buf, ABOOT_PREAMBLE_UUUU, ABOOT_PREAMBLE_SIZE);
		wb->buf[ABOOT_PREAMBLE_SIZE] = ABOOT_FRAME_DELIMITER;
		wb->len = ABOOT_PREAMBLE_SIZE + 1;
		scm2020_submit_wb(usb_dl, wb);
		queue_delayed_work(usb_dl->workqueue, &usb_dl->work, msecs_to_jiffies(10));
		break;
	case DL_START:
		{
			char magic[9];
			u32 size;

			if (download_read_line(usb_dl, cmd_line) <= 0)
				break;
			if (sscanf(cmd_line, "%08s%08x", magic, &size) != 2) {
				dev_info(dev, "sscanf error parameter\n");
				break;
			}

			if (strcmp(magic, "!CRANE!") && size == usb_dl->fw->size) {
				dev_err(dev,
					"magic %s mismatch with %s or len %08x mismath %08zx\n",
					magic, "!CRANE!", size, usb_dl->fw->size);
			}
			download_cmd_read_response(usb_dl);
			break;
		}
	case DL_STATUS:
		{
			struct scm2020_msg *msg;

			if (strlen(cmd_line) > strlen(cmd_getvar) &&
			    !memcmp(cmd_line, cmd_getvar, strlen(cmd_getvar))) {
				char variable[ABOOT_COMMAND_SZ];

				strcpy(variable, cmd_line + strlen(cmd_getvar));
				if (download_handle_getvar(usb_dl, variable, response) < 0) {
					dev_err(dev, "version check fail\n");
					break;
				}
			} else if (strlen(cmd_line) > strlen(cmd_download) &&
				   !memcmp(cmd_line, cmd_download, strlen(cmd_download))) {
				u32 size;

				if (kstrtou32(cmd_line + strlen(cmd_download), 16, (void *)&size))
					break;
				download_handle_download(usb_dl, size, response);
				download_do_download_data(usb_dl, size);
				memset(cmd_line, '\0', ABOOT_COMMAND_SZ);
				if (download_read_line(usb_dl, response) > 0)
					break;
			} else {
				msg = list_first_entry_or_null(&usb_dl->cmd_list,
							       struct scm2020_msg, list);
				if (!msg)
					break;

				list_del_init(&msg->list);
				if (!memcmp(msg->resp, "FAIL", 4))
					break;
			}
			download_cmd_read_response(usb_dl);
			break;
		}
	case DL_REBOOT:
		strcpy(cmd_line, cmd_complete);
		strcpy(response, "OKAY");
		usb_dl->step = DL_FINISH;
		usb_dl->wtype = SMUX_FRAME_TYPE_ABOOT_CMD;
		scm2020_add_delimiter(usb_dl, cmd_line, strlen(cmd_line));
		queue_delayed_work(usb_dl->workqueue, &usb_dl->work, msecs_to_jiffies(1000));
		break;
	case DL_FINISH:
		strcpy(cmd_line, cmd_disconnect);
		strcpy(response, "OKAY");
		usb_dl->wtype = SMUX_FRAME_TYPE_ABOOT_CMD;
		scm2020_add_delimiter(usb_dl, cmd_line, strlen(cmd_line));
		break;
	default:
		break;
	}
	spin_unlock_irqrestore(&usb_dl->lock, flags);
}

static void scm2020_msg_init(struct scm2020_usb_dl *usb_dl)
{
	int i;

	INIT_LIST_HEAD(&usb_dl->cmd_list);
	usb_dl->cmd_idx = 0;
	for (i = 0; i < ACM_NW; i++)
		INIT_LIST_HEAD(&usb_dl->cmd[i].list);
}

int scm2020_usb_dl_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_interface *control_interface;
	struct usb_interface *data_interface;
	struct usb_endpoint_descriptor *epctrl = NULL;
	struct usb_endpoint_descriptor *epread = NULL;
	struct usb_endpoint_descriptor *epwrite = NULL;
	struct usb_device *usb_dev = interface_to_usbdev(intf);
	struct device *dev = &intf->dev;
	struct scm2020_usb_dl *usb_dl;
	const char *firmware = "senscomm/wise-usb.bin";
	int rv = -ENOMEM;

	dev_dbg(dev, "usb interface info:\n");
	dev_dbg(dev, "- bNumInterfaces %d\n", intf->cur_altsetting->desc.bInterfaceNumber);
	dev_dbg(dev, "- bNumEndpoints %d\n", intf->cur_altsetting->desc.bNumEndpoints);

	/* interface have been init */
	if (intf->cur_altsetting->desc.bInterfaceNumber > 0)
		return 0;

	usb_dl = devm_kzalloc(dev, sizeof(*usb_dl), GFP_KERNEL);
	if (!usb_dl)
		goto exit;

	usb_dl->udev = usb_dev;

	if (intf->cur_altsetting->desc.bNumEndpoints == 3) {
		control_interface = intf;
		data_interface = intf;
	} else {
		control_interface = usb_ifnum_to_if(usb_dev, 0);
		data_interface = usb_ifnum_to_if(usb_dev, 1);
	}

	if (!control_interface || !data_interface)
		goto exit;

	usb_dl->intf = intf;
	usb_dl->control = control_interface;
	usb_dl->data = data_interface;

	if (data_interface == control_interface) {
		struct usb_endpoint_descriptor *epd;
		int i;

		for (i = 0; i < data_interface->cur_altsetting->desc.bNumEndpoints; ++i) {
			epd = &control_interface->cur_altsetting->endpoint[i].desc;
			if (usb_endpoint_type(epd) == USB_ENDPOINT_XFER_BULK) {
				if (usb_endpoint_dir_in(epd))
					epread = epd;
				else
					epwrite = epd;
			} else {
				epctrl = epd;
			}
		}
	} else {
		epctrl = &control_interface->cur_altsetting->endpoint[0].desc;
		epread = &data_interface->cur_altsetting->endpoint[1].desc;
		epwrite = &data_interface->cur_altsetting->endpoint[0].desc;
	}

	if (!epctrl || !epread || !epwrite) {
		dev_err(dev, "no ep got\n");
		rv = -EIO;
		goto exit;
	}

	usb_dl->in = usb_rcvbulkpipe(usb_dev, epread->bEndpointAddress);
	usb_dl->out = usb_sndbulkpipe(usb_dev, epwrite->bEndpointAddress);
	usb_dl->ctrlsize = usb_endpoint_maxp(epctrl);
	usb_dl->readsize = usb_endpoint_maxp(epread);
	usb_dl->writesize = usb_endpoint_maxp(epwrite) * 20;

	dev_dbg(dev, "in pipe %x, out pile %x, max r[%x] w[%x] c[%x]\n",
		usb_dl->in, usb_dl->out, usb_dl->readsize, usb_dl->writesize, usb_dl->ctrlsize);

	if (request_firmware(&usb_dl->fw, firmware, dev))
		goto exit;
	usb_dl->off = 0;

	/* init write buffer */
	scm2020_write_buffer_init(usb_dl);

	/* init read buff */
	scm2020_read_buffer_init(usb_dl);

	/* init msg list */
	scm2020_msg_init(usb_dl);
	mutex_init(&usb_dl->mutex);
	spin_lock_init(&usb_dl->lock);

	usb_dl->workqueue = create_singlethread_workqueue("usb_dl");
	if (!usb_dl->workqueue) {
		release_firmware(usb_dl->fw);
		scm2020_write_buffers_free(usb_dl);
		scm2020_read_buffers_free(usb_dl);
	}

	usb_dl->step = DL_INIT;
	INIT_DELAYED_WORK(&usb_dl->work, scm2020_fw_download_work);
	usb_set_intfdata(intf, usb_dl);

	queue_delayed_work(usb_dl->workqueue, &usb_dl->work, msecs_to_jiffies(10));
	return 0;
exit:
	return rv;
}

void scm2020_usb_dl_disconnect(struct usb_interface *intf)
{
	struct scm2020_usb_dl *usb_dl = usb_get_intfdata(intf);
	struct device *dev = &intf->dev;

	dev_dbg(dev, "device disconnected\n");

	cancel_delayed_work_sync(&usb_dl->work);
	destroy_workqueue(usb_dl->workqueue);
	release_firmware(usb_dl->fw);
	scm2020_write_buffers_free(usb_dl);
	scm2020_read_buffers_free(usb_dl);
}

#ifdef TODO_TO_REMOVE

/*
 * USB driver structure.
 */

static const struct usb_device_id scm2020_usb_dl_ids[] = {
	{ USB_DEVICE_AND_INTERFACE_INFO(0x2ecc, 0x3004,
					USB_CLASS_COMM,
					USB_CDC_SUBCLASS_ACM,
					USB_CDC_PROTO_NONE) },	/* SCM2020 Tiny usb device */

	{ }
};

MODULE_DEVICE_TABLE(usb, scm2020_usb_dl_ids);

static struct usb_driver scm2020_usb_dl_driver = {
	.name = "scm2020-usb-dl",
	.probe = scm2020_usb_dl_probe,
	.disconnect = scm2020_usb_dl_disconnect,
	.id_table = scm2020_usb_dl_ids,
};

/*
 * Init / exit.
 */

static int __init scm2020_usb_init(void)
{
	int retval;

	retval = usb_register(&scm2020_usb_dl_driver);
	if (retval)
		return retval;

	return 0;
}

static void __exit scm2020_usb_exit(void)
{
	usb_deregister(&scm2020_usb_dl_driver);
}

module_init(scm2020_usb_init);
module_exit(scm2020_usb_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Senscomm Semiconductor, Inc.");
#endif
