// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
#include <net/cfg80211.h>
#include <linux/crc32.h>
#include <linux/bitfield.h>

#include "core.h"
#include "util.h"
#include "regs.h"

/**
 *
 * Returns: non-HT reference rate in 100 kbps
 *
 *  Modulation R    HT    VHT   HE   Non-HT reference rate (Mb/s)
 *  -------------------------------------------------------------
 *  BPSK 1/2        0     0     0    6
 *  BPSK 3/4        -     -     -    9
 *  QPSK 1/2        1     1     1    12
 *  QPSK 3/4        2     2     2    18
 *  16-QAM 1/2      3     3     3    24
 *  16-QAM 3/4      4     4     4    36
 *  64-QAM 1/2      -     -     -    48
 *  64-QAM 2/3      5     5     5    48
 *  64-QAM 3/4      6     6     6    54
 *  64-QAM 5/6      7     7     7    54
 *  256-QAM 3/4     -     8     8    54
 *  256-QAM 5/6     -     -     9    54
 *  1024-QAM 3/4    -     -     10   54
 *  1024-QAM 5/6    -     -     11   54
 */
static int non_ht_ref_rate[] = {
	[0] = 60,
	[1] = 120,
	[2] = 180,
	[3] = 360,
	[4 ... 5] = 480,
	[5 ... 15] = 540,
};

static int ieee80211_non_ht_ref_rate(struct ieee80211_supported_band *sband,
				     struct ieee80211_tx_rate *r)
{
	int mcs = r->idx;

	if (ieee80211_rate_is_non_ht(r))
		return sband->bitrates[r->idx].bitrate;

	mcs = ieee80211_rate_get_mcs(r);

	return non_ht_ref_rate[mcs];
}

/**
 * __ieee80211_compute_ack_rate() - get control response frame rate
 * @hw: hardware
 * @vif: virtual interface
 * @ti: txop tx_info (&struct ieee80211_tx_info)
 * @rateset: set (bitmask) of rate indices from which to choose the rate
 */
static struct ieee80211_rate *__ieee80211_compute_ack_rate(struct ieee80211_hw *hw,
							   struct ieee80211_vif *vif,
							   struct ieee80211_tx_info *ti,
							   u32 rateset)
{
	struct ieee80211_supported_band *sband = hw->wiphy->bands[ti->band];
	struct ieee80211_tx_rate *rate = &ti->control.rates[0];
	u32 bitrate;
	int i, idx = -1;
	bool tx_dsss = false, resp_dsss;

	if (rateset == 0)
		return NULL;

	if (ieee80211_rate_is_non_ht(rate) &&
	    (sband->bitrates[rate->idx].flags & IEEE80211_RATE_MANDATORY_B))
		tx_dsss = true;

	bitrate = ieee80211_non_ht_ref_rate(sband, rate);

	for (i = 0; i < sband->n_bitrates; i++) {
		if (!(rateset & BIT(i)))
			continue;

		resp_dsss = !!(sband->bitrates[i].flags & IEEE80211_RATE_MANDATORY_B);
		if (tx_dsss ^ resp_dsss)
			continue;

		if (sband->bitrates[i].bitrate <= bitrate)
			idx = i;
	}

	if (idx == -1)
		return NULL;

	return &sband->bitrates[idx];
}

u32 scm80211_compute_phy_mandatory_rates(struct ieee80211_hw *hw, enum nl80211_band band)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_supported_band *sband = hw->wiphy->bands[band];
	u32 flags;
	int i;

	if (priv->phy_mandatory_rates[band])
		goto out;

	if (band == NL80211_BAND_2GHZ)
		flags = IEEE80211_RATE_MANDATORY_B | IEEE80211_RATE_MANDATORY_G;
	else
		flags = IEEE80211_RATE_MANDATORY_A;

	priv->phy_mandatory_rates[band] = 0;
	for (i = 0; i < sband->n_bitrates; i++) {
		if (sband->bitrates[i].flags & flags)
			priv->phy_mandatory_rates[band] |= BIT(i);
	}

out:
	return priv->phy_mandatory_rates[band];
}

/**
 * ieee80211_compute_ack_rate - compute the rate of the control response frame
 * @hw: hardware
 * @vif: virtual interface
 * @ti: tx info of the frame soliciting the response
 * @rate_idx: index to rates
 *
 * The function computes the rate of the control response frame.
 */
struct ieee80211_rate *ieee80211_compute_ack_rate(struct ieee80211_hw *hw,
						  struct ieee80211_vif *vif,
						  struct ieee80211_tx_info *ti)
{
	struct ieee80211_rate *ack_rate = NULL;
	u32 rateset = 0;

	if (vif)
		rateset = vif->bss_conf.basic_rates;

	/* Find the primary rate from BSSBasicRateset */
	ack_rate = __ieee80211_compute_ack_rate(hw, vif, ti, rateset);
	if (ack_rate)
		return ack_rate;

	rateset = scm80211_compute_phy_mandatory_rates(hw, ti->band);
	ack_rate = __ieee80211_compute_ack_rate(hw, vif, ti, rateset);
	WARN_ON(rateset == 0 || !ack_rate);
	return ack_rate;
}

/**
 * ieee80211_ack_duration() - compute Ack/BA duration
 * @hw: hardware
 * @vif: virtual interface
 * @ti: &struct ieee80211_tx_info of the frame soliciting Ack/BA response
 * @len: length of the response frame (including FCS)
 *
 * Returns: SIFS + Ack time
 */
u16 ieee80211_compute_ack_time(struct ieee80211_hw *hw,
			       struct ieee80211_vif *vif, struct ieee80211_tx_info *ti, size_t len)
{
	struct ieee80211_rate *rate;
	__le16 dur;

	if (len <= FCS_LEN)
		return 0;

	/*
	 * NB: If @ack_rate is OFDM rate, either ERP-OFDM or OFDM,
	 * the ieee80211_generic_frame_duration() returns correct
	 * duration in both cases. See the diagram below.
	 *
	 *        |       Data      |           |     Ack         |
	 *
	 * 11a:   |<-   TXTIME(a) ->|<- SIFS  ->|<-TXTIME(Ack,a)->|
	 *                             (16us)
	 *                          |<------ Duration(a) -------->|
	 *
	 * 11g:   |<-   TXTIME(g)    ->|<-SIFS->|<-TXTIME(Ack,g)   ->|
	 *                          |<>| (10us) |                 |<>| (<>: 6us)
	 *                             |<------ Duration(g) -------->|
	 *
	 * where,
	 * - TXTIME(g) = TXTIME(a) + aSignalExtension(=6us)
	 * - TXTIME(Ack,g) = TXTIME(Ack,a) + aSignalExtension
	 * - Duration(g) = Duration(a)
	 */
	rate = ieee80211_compute_ack_rate(hw, vif, ti);
	if (rate)
		dur = ieee80211_generic_frame_duration(hw, vif, ti->band, len - FCS_LEN, rate);
	else
		BUG_ON(1);

	return le16_to_cpu(dur);
}

/**
 * ieee80211_compute_rts_time() - calculate RTS/CTS exchange time
 * @hw: hardware
 * @vif: virtual interface
 * @ti: &struct ieee80211_tx_info of the frame to be protected
 */
u16 ieee80211_compute_rts_time(struct ieee80211_hw *hw,
			       struct ieee80211_vif *vif, struct ieee80211_tx_info *ti)
{
	u16 dur = 0;

	/* RTS + SIFS */
	dur += ieee80211_compute_ack_time(hw, vif, ti, IEEE80211_RTS_LEN);
	/* CTS + SIFS */
	dur += ieee80211_compute_ack_time(hw, vif, ti, IEEE80211_CTS_LEN);
	return dur;
}

static int ieee80211_max_ampdu_len_ht(struct ieee80211_sta *sta)
{
	int exp;

	/*
	 * HT Capabilities::A-MPDU Parameters
	 *   ::Maximum A-MPDU Length Exponent (0-3)
	 */
	if (get_ieee80211_ht_cap(sta).ht_supported) {
		exp = get_ieee80211_ht_cap(sta).ampdu_factor;
		return (1 << (13 + exp)) - 1;
	}
	return -1;
}

static int ieee80211_max_ampdu_len_vht(struct ieee80211_sta *sta)
{
	int exp;

	/*
	 * VHT Capabilities Info[23:25] (0-7)
	 */
	if (get_ieee80211_vht_cap(sta).vht_supported) {
		exp = get_ieee80211_vht_cap(sta).cap;
		exp &= IEEE80211_VHT_CAP_MAX_A_MPDU_LENGTH_EXPONENT_MASK;
		exp >>= IEEE80211_VHT_CAP_MAX_A_MPDU_LENGTH_EXPONENT_SHIFT;
		return (1 << (13 + exp)) - 1;
	}
	return -1;
}

#define BYTE_OFFSET(bit)  ((bit) / BITS_PER_BYTE)
#define BITS_IN_BYTE(bit) ((bit) % BITS_PER_BYTE)

static int ieee80211_max_ampdu_len_he(struct ieee80211_sta *sta)
{
	struct ieee80211_sta_he_cap he_cap_tmp = get_ieee80211_he_cap(sta);
	u8 *cap = he_cap_tmp.he_cap_elem.mac_cap_info;
	int ext = (cap[BYTE_OFFSET(27)] >> BITS_IN_BYTE(27)) & 0x3;
	int len = -1;

	/* P802.11ax_D4.0 26.6.1 */
	if (ext == 0) {
		len = ieee80211_max_ampdu_len_vht(sta);
		if (len < 0)
			len = ieee80211_max_ampdu_len_ht(sta);
	} else if (get_ieee80211_vht_cap(sta).vht_supported) {
		len = min((1 << (20 + ext)) - 1, 6500631);
	} else if (get_ieee80211_ht_cap(sta).ht_supported) {
		len = (1 << (16 + ext)) - 1;
	} else {
		pr_err("HE STA without both VHT and HT capabilities\n");
	}
	return len;
}

int ieee80211_max_ampdu_len(struct ieee80211_sta *sta, struct ieee80211_tx_rate *rate)
{
	BUG_ON(!sta);

	if (ieee80211_rate_is_he(rate))
		return ieee80211_max_ampdu_len_he(sta);
	else if (ieee80211_rate_is_vht(rate))
		return ieee80211_max_ampdu_len_vht(sta);
	else if (ieee80211_rate_is_ht(rate))
		return ieee80211_max_ampdu_len_ht(sta);
	return -1;
}

/*
 * Table 10-12
 */
u16 ieee80211_vht_partial_aid(u16 aid, const u8 *addr, int group)
{
	u16 v = 0;

	if (aid == 0 || !addr)
		return 0;

	if (group == 0) {
		v = ((addr[5] << 1) | ((addr[4] >> 7) & 0x1));
	} else if (group == 63) {
		v = (((addr[5] >> 4) ^ (addr[5])) & 0xf) << 5;
		v += aid;
		v &= 0x1ff;	/* Take the 9 least significant bits */
	}
	return v;
}

/* Copied from net/mac80211/rx.c */
u8 *ieee80211_get_bssid(struct ieee80211_hdr *hdr, size_t len, enum nl80211_iftype type)
{
	__le16 fc = hdr->frame_control;

	if (ieee80211_is_data(fc)) {
		if (len < 24)	/* drop incorrect hdr len (data) */
			return NULL;

		if (ieee80211_has_a4(fc))
			return NULL;
		if (ieee80211_has_tods(fc))
			return hdr->addr1;
		if (ieee80211_has_fromds(fc))
			return hdr->addr2;

		return hdr->addr3;
	}

	if (ieee80211_is_mgmt(fc)) {
		if (len < 24)	/* drop incorrect hdr len (mgmt) */
			return NULL;
		return hdr->addr3;
	}

	if (ieee80211_is_ctl(fc)) {
		if (ieee80211_is_pspoll(fc))
			return hdr->addr1;

		if (ieee80211_is_back_req(fc)) {
			switch (type) {
			case NL80211_IFTYPE_STATION:
				return hdr->addr2;
			case NL80211_IFTYPE_AP:
			case NL80211_IFTYPE_AP_VLAN:
				return hdr->addr1;
			default:
				break;	/* fall through to the return */
			}
		}
	}

	return NULL;
}

int ieee80211_chandef_create(struct cfg80211_chan_def *chandef,
			     const struct ieee80211_sta *ap,
			     const struct ieee80211_ht_operation *ht_oper,
			     const struct ieee80211_vht_operation *vht_oper)
{
	enum nl80211_channel_type channel_type;
	u32 cf0, cf1;
	u8 ch_offset;

	if (!ht_oper)
		return -EINVAL;

	if (!(get_ieee80211_ht_cap(ap).cap & IEEE80211_HT_CAP_SUP_WIDTH_20_40))
		goto valid_check;

	switch (ht_oper->ht_param & IEEE80211_HT_PARAM_CHA_SEC_OFFSET) {
	case IEEE80211_HT_PARAM_CHA_SEC_ABOVE:
		channel_type = NL80211_CHAN_HT40PLUS;
		break;
	case IEEE80211_HT_PARAM_CHA_SEC_BELOW:
		channel_type = NL80211_CHAN_HT40MINUS;
		break;
	case IEEE80211_HT_PARAM_CHA_SEC_NONE:
		channel_type = NL80211_CHAN_HT20;
		break;
	default:
		return -EOPNOTSUPP;
	}

	cfg80211_chandef_create(chandef, chandef->chan, channel_type);

	if (chandef->chan->band == NL80211_BAND_2GHZ)
		goto valid_check;
	if (chandef->chan->band == NL80211_BAND_5GHZ && !vht_oper)
		return -EINVAL;

	cf0 = vht_oper->center_freq_seg0_idx;
	cf1 = vht_oper->center_freq_seg1_idx;
	ch_offset = abs(cf0 - cf1);
	cf0 = ieee80211_channel_to_frequency(cf0, chandef->chan->band);
	cf1 = ieee80211_channel_to_frequency(cf1, chandef->chan->band);

	switch (vht_oper->chan_width) {
	case IEEE80211_VHT_CHANWIDTH_160MHZ:
		chandef->width = NL80211_CHAN_WIDTH_160;
		chandef->center_freq1 = cf0;
		break;
	case IEEE80211_VHT_CHANWIDTH_80P80MHZ:
		chandef->width = NL80211_CHAN_WIDTH_80P80;
		chandef->center_freq1 = cf0;
		chandef->center_freq2 = cf1;
		break;
	case IEEE80211_VHT_CHANWIDTH_80MHZ:
		chandef->width = NL80211_CHAN_WIDTH_80;
		chandef->center_freq1 = cf0;
		if (cf1 && ch_offset == 8) {
			chandef->width = NL80211_CHAN_WIDTH_160;
			chandef->center_freq1 = cf1;
		} else if (cf1 && ch_offset > 8) {
			chandef->width = NL80211_CHAN_WIDTH_80P80;
			chandef->center_freq2 = cf1;
		}
		break;
	default:
		break;
	}

valid_check:
	if (!cfg80211_chandef_valid(chandef))
		return -EINVAL;
	return 0;
}

int ieee80211_link_adap_to_rate(struct ieee80211_hw *hw, __le32 htc,
				struct ieee80211_tx_rate *rate, bool *stbc_dcm)
{
	u8 nsts, mcs, bw;
	bool is_la = false;

	if ((htc & HE_VARIANT) == HE_VARIANT) {
		u8 ctl_id = FIELD_GET(HE_CTLID_MASK, htc);

		if (ctl_id == HE_LA_CTL && !(htc & BIT(HE_MRQ_SHIFT))) {
			rate->flags = IEEE80211_TX_RC_HE_MCS;
			nsts = FIELD_GET(HE_NSS_MASK, htc);
			mcs = FIELD_GET(HE_MCS_MASK, htc);
			*stbc_dcm = !!(htc & BIT(HE_DCM_SHIFT));
			bw = FIELD_GET(HE_BW_MASK, htc);
			if (mcs == 0x0F && nsts == 7)
				return -1;
			if (htc & BIT(HE_UNSOL_SHIFT)) {
				if (!(htc & BIT(HE_TXB_SHIFT)) && !(htc & BIT(HE_UL_TB_SHIFT)))
					is_la = true;
			} else {
				if (FIELD_GET(HE_MSI_MASK, htc) < 0x07)
					is_la = true;
			}
		}
	} else if ((htc & VHT_VARIANT) && !(htc & BIT(VHT_MRQ_SHIFT))) {
		u16 vht_mfb = FIELD_GET(VHT_MFB_MASK, htc);

		rate->flags = IEEE80211_TX_RC_VHT_MCS;
		nsts = vht_mfb & 0x07;
		mcs = (vht_mfb >> 3) & 0x0F;
		bw = (vht_mfb >> 7) & 0x03;
		if (mcs == 0x0f && nsts == 0x07)
			return -1;

		if (htc & BIT(VHT_UNSOL_SHIFT)) {
			u8 gid_l = FIELD_GET(VHT_MFSI_MASK, htc);
			*stbc_dcm = !!(htc & BIT(5));
			if (gid_l == 0x07 && !(htc & BIT(VHT_FB_TYPE_SHIFT)))
				is_la = true;
		} else {
			if (FIELD_GET(VHT_MFSI_MASK, htc) < 0x07)
				is_la = true;
		}
	}

	if (!is_la)
		return -1;

	switch (bw) {
	case CBW20:
		break;
	case CBW40:
		rate->flags |= IEEE80211_TX_RC_40_MHZ_WIDTH;
		break;
	default:
		rate->flags |= IEEE80211_TX_RC_80_MHZ_WIDTH;
		break;
	}
	nsts = (nsts + 1) > 2 ? 2 : nsts + 1;

	ieee80211_rate_set_vht(rate, mcs, nsts);
	rate->count = 1;

	return 0;
}

const char *ieee80211_iftype_str[] = {
	"unspecified",
	"ibss",
	"managed",
	"ap",
	"ap vlan",
	"wds",
	"monitor",
	"mesh",
	"p2p client",
	"p2p go",
	"p2p dev",
	"ocb",
	"nan",
};

const char *ieee80211_sta_state_str[] = {
	"non-existent",
	"none",
	"authenticated",
	"associated",
	"authorized",
};

const char *cipher_suite_str[] = {
	[CIPHER_SUITE_WEP40]   = "WEP40",
	[CIPHER_SUITE_WEP104]  = "WEP104",
	[CIPHER_SUITE_TKIP]    = "TKIP",
	[CIPHER_SUITE_CCMP128] = "CCMP",
	[CIPHER_SUITE_CCMP256] = "CCMP256",
	[CIPHER_SUITE_GCMP128] = "GCMP",
	[CIPHER_SUITE_CCMP256] = "CCMP256",
};

bool scm80211_rx_parse_needed(const u8 *tx_bssid, struct ieee80211_sta *sta,
			      struct ieee80211_bss_conf *info)
{
	if (!info->assoc)
		return false;
	if (ether_addr_equal(tx_bssid, info->bssid))
		return true;
	return (info->nontransmitted && ether_addr_equal(tx_bssid, info->transmitter_bssid));
}

void scm80211_parse_elems(const u8 *start, size_t len, bool action, struct ieee802_11_elems *elems)
{
	const struct element *elem = NULL;

	memset(elems, 0, sizeof(*elems));

	if (!start || len < 2)
		return;

	elems->ie_start = start;
	elems->total_len = len;

	for_each_element(elem, start, len) {
		u8 id = elem->id;
		u8 elen = elem->datalen;
		const u8 *pos = elem->data;

		if (id == WLAN_EID_EXTENSION && pos[0] == WLAN_EID_EXT_HE_OPERATION &&
		    elen >= sizeof(*elems->he_operation) &&
		    elen >= ieee80211_he_oper_size(&pos[1])) {
			elems->he_operation = (void *)&pos[1];
		} else if (id == WLAN_EID_TWT && elen >= 10) {
			elems->twt = pos;
			elems->twt_len = elen;
		} else {
			continue;
		}
	}
}

void scm80211_update_sr(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u8 vidx = scm80211_vdev_id(vif);
	u32 v;

	/*
	 *      SR option
	 *  CFG#SR - SR_MPDU_EN/SR_PLCP_HEADER_EN
	 */
	v = scm80211_mac_readl(hw, MAC_REG_SR_CFG);
	v |= FIELD_PREP(SR_CFG_SR_PLCP_HEADER_EN, 1);
	v |= FIELD_PREP(SR_CFG_SR_MPDU_EN, 1);
	scm80211_mac_writel(hw, v, MAC_REG_SR_CFG);

	/* non_srg */
	if (vifp->has_sr &&
	    (vifp->sr_param.sr_control & IEEE80211_SR_NON_SRG_OBSS_PD_SR_DISALLOWED_PRESENT)) {
		/*
		 *  CFG#OBSS_SR_NON_SRG - EN
		 */
		v = scm80211_mac_readl(hw, MAC_REG_OBSS_SR_NON_SRG_CFG);
		v |= FIELD_PREP(OBSS_SR_NON_SRG_CFG_NON_SRG_AP_DIS, 1);
		v &= ~OBSS_SR_NON_SRG_CFG_NON_SRG_OBSS_PD_EN;
		scm80211_mac_writel(hw, v, MAC_REG_OBSS_SR_NON_SRG_CFG);
	} else {
		/*
		 *  NON_SRG_OBSS_PD#vif - MIN_VALUE/MAX_VALUE
		 */
		v = 0;
		v |= FIELD_PREP(VIF_NON_SRG_OBSS_PD_MIN_VALUE,
				IEEE80211_DEF_NON_SRG_OBSS_PD_MIN_VALUE);
		v |= FIELD_PREP(VIF_NON_SRG_OBSS_PD_MAX_VALUE,
				vifp->sr_param.non_srg_obss_pd_max_offset);
		scm80211_mac_writel(hw, v, MAC_REG_VIF_NON_SRG_OBSS_PD(vidx));

		/*
		 *  COLOR_EN0/1#OBSS_SR_NON_SRG - COLOR_EN0/1
		 */
		scm80211_mac_writel(hw, lower_32_bits(vifp->sr.sr_non_srg_color_en),
				    MAC_REG_OBSS_SR_NON_SRG_COLOR_EN0);
		scm80211_mac_writel(hw, upper_32_bits(vifp->sr.sr_non_srg_color_en),
				    MAC_REG_OBSS_SR_NON_SRG_COLOR_EN1);

		/*
		 *  BSSID_EN0/1#OBSS_SR_NON_SRG - BSSID_EN0/1
		 */
		scm80211_mac_writel(hw, lower_32_bits(vifp->sr.sr_non_srg_bssid_en),
				    MAC_REG_OBSS_SR_NON_SRG_BSSID_EN0);
		scm80211_mac_writel(hw, upper_32_bits(vifp->sr.sr_non_srg_bssid_en),
				    MAC_REG_OBSS_SR_NON_SRG_BSSID_EN1);

		/*
		 *  CFG#OBSS_SR_NON_SRG - EN
		 */
		v = scm80211_mac_readl(hw, MAC_REG_OBSS_SR_NON_SRG_CFG);
		if (vifp->sr_param.sr_control & IEEE80211_SR_NON_SRG_OBSS_PD_SR_DISALLOWED_PRESENT)
			v |= FIELD_PREP(OBSS_SR_NON_SRG_CFG_NON_SRG_AP_DIS, 1);
		else
			v &= ~OBSS_SR_NON_SRG_CFG_NON_SRG_AP_DIS;

		v |= FIELD_PREP(OBSS_SR_NON_SRG_CFG_NON_SRG_NOBSSID_EN, 1);
		v |= FIELD_PREP(OBSS_SR_NON_SRG_CFG_NON_SRG_ER_THR, vifp->sr.sr_obss_pd_er_thr);
		v |= FIELD_PREP(OBSS_SR_NON_SRG_CFG_NON_SRG_THR, vifp->sr.sr_obss_pd_thr);
		v |= FIELD_PREP(OBSS_SR_NON_SRG_CFG_NON_SRG_OBSS_PD_EN, 1);
		scm80211_mac_writel(hw, v, MAC_REG_OBSS_SR_NON_SRG_CFG);
	}

	/* srg */
	if (vifp->has_sr && (vifp->sr_param.sr_control & IEEE80211_SR_SRG_INFORMATION_PRESENT)) {
		/*
		 *  SRG_OBSS_PD#vif - MIN_VALUE/MAX_VALUE
		 */
		v = 0;
		v |= FIELD_PREP(VIF_NON_SRG_OBSS_PD_MIN_VALUE,
				vifp->sr_param.srg_obss_pd_min_offset);
		v |= FIELD_PREP(VIF_NON_SRG_OBSS_PD_MAX_VALUE,
				vifp->sr_param.srg_obss_pd_max_offset);
		scm80211_mac_writel(hw, v, MAC_REG_VIF_SRG_OBSS_PD(vidx));

		/*
		 *  COLOR_EN0/1#SRG_MAP - SRG_MAP_COLOR0/1
		 */
		scm80211_mac_writel(hw, lower_32_bits(vifp->sr_param.srg_bss_colar_bitmap),
				    MAC_REG_SRG_MAP_COLOR0);
		scm80211_mac_writel(hw, upper_32_bits(vifp->sr_param.srg_bss_colar_bitmap),
				    MAC_REG_SRG_MAP_COLOR1);

		/*
		 *  BSSID_EN0/1#SRG_MAP - SRG_MAP_BSSOD0/1
		 */
		scm80211_mac_writel(hw, lower_32_bits(vifp->sr_param.srg_partial_bssid_bitmap),
				    MAC_REG_SRG_MAP_BSSID0);
		scm80211_mac_writel(hw, upper_32_bits(vifp->sr_param.srg_partial_bssid_bitmap),
				    MAC_REG_SRG_MAP_BSSID1);

		/*
		 *  COLOR_EN0/1#SRG_SR - COLOR_EN0/1
		 */
		scm80211_mac_writel(hw, lower_32_bits(vifp->sr.sr_srg_color_en),
				    MAC_REG_SRG_SR_COLOR_EN0);
		scm80211_mac_writel(hw, upper_32_bits(vifp->sr.sr_srg_color_en),
				    MAC_REG_SRG_SR_COLOR_EN1);

		/*
		 *  BSSID_EN0/1#SRG_SR - BSSID_EN0/1
		 */
		scm80211_mac_writel(hw, lower_32_bits(vifp->sr.sr_srg_bssid_en),
				    MAC_REG_SRG_SR_BSSID_EN0);
		scm80211_mac_writel(hw, upper_32_bits(vifp->sr.sr_srg_bssid_en),
				    MAC_REG_SRG_SR_BSSID_EN1);

		/*
		 *  CFG#SRG - SRG_EN/SRG_ER_THR/SRG_THR
		 */
		v = scm80211_mac_readl(hw, MAC_REG_SRG_CFG);
		v |= FIELD_PREP(SRG_CFG_SRG_ER_THR, vifp->sr.sr_srg_er_thr);
		v |= FIELD_PREP(SRG_CFG_SRG_THR, vifp->sr.sr_srg_thr);
		v |= FIELD_PREP(SRG_CFG_SRG_EN, 1);
		scm80211_mac_writel(hw, v, MAC_REG_SRG_CFG);
	} else {
		/*
		 *  CFG#SRG - SRG_CFG_SRG_EN
		 */
		v = scm80211_mac_readl(hw, MAC_REG_SRG_CFG);
		v &= ~SRG_CFG_SRG_EN;
		scm80211_mac_writel(hw, v, MAC_REG_SRG_CFG);
	}

	scm80211_mac_writel(hw, 0, MAC_REG_OBSS_PD_TX_PWR_OPT);
}

void scm80211_init_sr(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u32 v;
	unsigned long flags;

	/*
	 *  CFG#SRG - SRG_CFG_SRG_EN
	 *  CFG#OBSS_SR_NON_SRG - OBSS_SR_NON_SRG_CFG_EN
	 */
	v = scm80211_mac_readl(hw, MAC_REG_SRG_CFG);
	v &= ~SRG_CFG_SRG_EN;
	scm80211_mac_writel(hw, v, MAC_REG_SRG_CFG);

	v = scm80211_mac_readl(hw, MAC_REG_OBSS_SR_NON_SRG_CFG);
	v &= ~OBSS_SR_NON_SRG_CFG_NON_SRG_OBSS_PD_EN;
	scm80211_mac_writel(hw, v, MAC_REG_OBSS_SR_NON_SRG_CFG);

	spin_lock_irqsave(&vifp->lock, flags);

	vifp->has_sr = false;
	memset(&vifp->sr_param, 0, sizeof(struct ieee80211_ie_sr_reuse_param_set));
	vifp->sr_param.non_srg_obss_pd_max_offset = IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
	vifp->sr_param.srg_obss_pd_min_offset = IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
	vifp->sr_param.srg_obss_pd_max_offset = IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;

	vifp->sr.sr_non_srg_color_en = -1ULL;
	vifp->sr.sr_non_srg_bssid_en = -1ULL;
	vifp->sr.sr_srg_color_en = -1ULL;
	vifp->sr.sr_srg_bssid_en = -1ULL;
	vifp->sr.sr_obss_pd_thr = 0;
	vifp->sr.sr_obss_pd_er_thr = 0;
	vifp->sr.sr_srg_thr = 0;
	vifp->sr.sr_srg_er_thr = 0;

	spin_unlock_irqrestore(&vifp->lock, flags);
}

void ieee80211_parse_he_sr(struct ieee80211_hw *hw, struct ieee80211_vif *vif, const u8 *ie)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u8 changed = 0;
	u64 val;
	unsigned long flags;

	spin_lock_irqsave(&vifp->lock, flags);

	if (ie) {
		if (!vifp->has_sr) {
			memset(&vifp->sr_param, 0, sizeof(vifp->sr_param));
			vifp->has_sr = true;
			changed = 1;
		}

		if (vifp->sr_param.sr_control != *ie)
			changed = 1;

		vifp->sr_param.sr_control = *ie;
		ie++;
		if (vifp->sr_param.sr_control & IEEE80211_SR_NON_SRG_OFFSET_PRESENT) {
			if (vifp->sr_param.non_srg_obss_pd_max_offset !=
			    (*ie + IEEE80211_DEF_NON_SRG_OBSS_PD_MAX_VALUE_M82))
				changed = 1;
			if ((vifp->sr_param.sr_control &
			     IEEE80211_SR_NON_SRG_OBSS_PD_SR_DISALLOWED_PRESENT))
				vifp->sr_param.non_srg_obss_pd_max_offset =
				    IEEE80211_DEF_NON_SRG_OBSS_PD_MAX_VALUE_M82;
			else
				vifp->sr_param.non_srg_obss_pd_max_offset =
				    *ie + IEEE80211_DEF_NON_SRG_OBSS_PD_MAX_VALUE_M82;
			ie++;
		} else {
			if ((vifp->sr_param.sr_control &
			     IEEE80211_SR_NON_SRG_OBSS_PD_SR_DISALLOWED_PRESENT))
				vifp->sr_param.non_srg_obss_pd_max_offset =
				    IEEE80211_DEF_NON_SRG_OBSS_PD_MAX_VALUE_M82;
			else
				vifp->sr_param.non_srg_obss_pd_max_offset =
				    IEEE80211_DEF_NON_SRG_OBSS_PD_MAX_VALUE_M62;
		}

		if (vifp->sr_param.sr_control & IEEE80211_SR_SRG_INFORMATION_PRESENT) {
			if (vifp->sr_param.srg_obss_pd_min_offset !=
			    (*ie + IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE))
				changed = 1;
			vifp->sr_param.srg_obss_pd_min_offset =
			    *ie + IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
			ie++;
			if (vifp->sr_param.srg_obss_pd_max_offset !=
			    (*ie + IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE))
				changed = 1;
			vifp->sr_param.srg_obss_pd_max_offset =
			    *ie + IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
			ie++;

			val = get_unaligned_le64(ie);
			if (vifp->sr_param.srg_bss_colar_bitmap != val)
				changed = 1;
			vifp->sr_param.srg_bss_colar_bitmap = val;
			ie += 8;
			val = get_unaligned_le64(ie);
			if (vifp->sr_param.srg_partial_bssid_bitmap != val)
				changed = 1;

			vifp->sr_param.srg_partial_bssid_bitmap = val;
			ie += 8;
		} else {
			vifp->sr_param.srg_obss_pd_min_offset =
			    IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
			vifp->sr_param.srg_obss_pd_max_offset =
			    IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
			vifp->sr_param.srg_bss_colar_bitmap = 0;
			vifp->sr_param.srg_partial_bssid_bitmap = 0;
		}
	} else {
		if (vifp->has_sr) {
			changed = 1;
			vifp->has_sr = false;
			memset(&vifp->sr_param, 0, sizeof(vifp->sr_param));
		}
		vifp->sr_param.non_srg_obss_pd_max_offset =
		    IEEE80211_DEF_NON_SRG_OBSS_PD_MAX_VALUE_M62;
		vifp->sr_param.srg_obss_pd_min_offset = IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
		vifp->sr_param.srg_obss_pd_max_offset = IEEE80211_DEF_SRG_OBSS_PD_MIN_MAX_VALUE;
		vifp->sr_param.srg_bss_colar_bitmap = 0;
		vifp->sr_param.srg_partial_bssid_bitmap = 0;
	}

	spin_unlock_irqrestore(&vifp->lock, flags);

	if (changed)
		scm80211_update_sr(hw, vif);
}

void ieee80211_parse_he_operation(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				  const struct ieee80211_he_operation *ie,
				  struct ieee80211_sta *sta)
{
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	unsigned long flags;
	u32 rts_th;

	if (!ieee80211_is_he_sta(sta))
		return;

	spin_lock_irqsave(&sta_priv->lock, flags);
	rts_th = le32_get_bits(ie->he_oper_params, IEEE80211_HE_OPERATION_RTS_THRESHOLD_MASK);

	rts_th = cpu_to_le16(rts_th);
	if (rts_th && rts_th != 1023)
		sta_priv->txop_rts_th = rts_th << 5;

	spin_unlock_irqrestore(&sta_priv->lock, flags);
}

bool ieee80211_is_highest_mcs(struct ieee80211_sta *sta, struct ieee80211_tx_rate *r)
{
	struct ieee80211_sta_priv *priv = ieee80211_to_priv(sta);
	int idx, nss;

	if (ieee80211_rate_is_non_ht(r))
		return false;

	idx = ieee80211_rate_get_mcs(r);
	nss = ieee80211_rate_get_nss(r);

	if (ieee80211_rate_is_ht(r) && idx == 0x07)
		return true;

	if (ieee80211_rate_is_vht(r)) {
		enum cbw cbw = ieee80211_ppdu_bandwidth(r);

		if (idx == priv->vht_mcs)
			return true;
		if (cbw == CBW20 && idx == 0x08)
			return true;
	}

	if (ieee80211_rate_is_he(r)) {
		if (idx == priv->he_mcs[0] * 2 + 7)
			return true;
	}

	return false;
}

bool scm80211_is_radar_freq(u16 center_freq)
{
	return (center_freq >= 5260 && center_freq <= 5700);
}

void scm80211_omi_work(struct work_struct *work)
{
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_vif *vif;
	struct ieee80211_hw *hw;

	vifp = container_of(work, struct ieee80211_vif_priv, omi_work);
	vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	hw = ieee80211_vif_to_hw(vif);

	if (vifp->disable_qnull_irq)
		scm80211_disable_trig_qnull_irq(hw, vif);
	else
		scm80211_enable_trig_qnull_irq(hw, vif);
	scm80211_mac_writel(hw, vifp->reg_mmss,
			    MAC_REG_VIF_MIN_MPDU_SPACING(scm80211_vdev_id(vif)));
}

/* Reference: P80211ax_D6.0, 26.9 Operating mode indication */
int scm80211_process_omi(struct ieee80211_hw *hw,
			 struct ieee80211_vif *vif, struct ieee80211_sta *sta, u32 htc, bool is_rx)
{
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u32 aom_old, nom_old;
	bool i_am_ap;
	bool changed = false;
	unsigned long flag;

	if (!IS_OMI(htc))
		return -1;

	i_am_ap = ieee80211_iface_is_ap(vif->type);

	/* transmitted OMI is saved without valid tag just for reference. */
	if (!is_rx)
		htc &= ~HE_VARIANT;

	aom_old = sta_priv->he_aom;
	nom_old = sta_priv->he_nom;
	if (i_am_ap != is_rx) {
		sta_priv->he_aom = htc;	/* AP ROM */
	} else {
		sta_priv->he_nom = htc;	/* non-AP TOM+ROM */
		changed = !i_am_ap;
	}

	wiphy_dbg(hw->wiphy, "OMI update(%c,%c,OM=%x,%x->%x,%x)\n", "SA"[i_am_ap],
		  "TR"[is_rx], aom_old, nom_old, sta_priv->he_aom, sta_priv->he_nom);

	if (changed) {
		changed = false;
		spin_lock_irqsave(&vifp->lock, flag);
		if (OMI_GET_UL_MU_DIS(htc) &&
		    FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_EN, vifp->reg_mmss)) {
			changed = true;
			vifp->disable_qnull_irq = true;
			vifp->reg_mmss &= ~VIF_MIN_MPDU_SPACING_TRIG_EN;
		} else if (!OMI_GET_UL_MU_DIS(htc) &&
			   !FIELD_GET(VIF_MIN_MPDU_SPACING_TRIG_EN, vifp->reg_mmss)) {
			changed = true;
			vifp->disable_qnull_irq = false;
			vifp->reg_mmss |= VIF_MIN_MPDU_SPACING_TRIG_EN;
		}

		if (OMI_GET_UL_MU_DATA_DIS(htc) &&
		    !FIELD_GET(VIF_MIN_MPDU_SPACING_HETB_DATA_DIS, vifp->reg_mmss)) {
			changed = true;
			vifp->disable_qnull_irq = true;
			vifp->reg_mmss |= VIF_MIN_MPDU_SPACING_HETB_DATA_DIS;
		} else if (!OMI_GET_UL_MU_DATA_DIS(htc) &&
			   FIELD_GET(VIF_MIN_MPDU_SPACING_HETB_DATA_DIS, vifp->reg_mmss)) {
			changed = true;
			vifp->disable_qnull_irq = false;
			vifp->reg_mmss &= ~VIF_MIN_MPDU_SPACING_HETB_DATA_DIS;
		}

		sta_priv->omi_req = 0;
		spin_unlock_irqrestore(&vifp->lock, flag);

		if (changed)
			ieee80211_queue_work(hw, &vifp->omi_work);
	}
	return 0;
}

bool scm80211_dbg_construct_rate(struct ieee80211_sta *sta,
					enum ppdu_format fmt, u8 rate_idx,
					enum cbw bw, u8 nss,
					enum gi_type gtype, enum ltf_type ltf_type, bool is_ldpc)
{
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_vif *vif = sta->txq[0]->vif;
	struct ieee80211_dbg_rate_fix *dbg_rate = &sta_priv->rate_fix;
	bool rate_valid = false;
	u8 tmp_idx;
	enum cbw maxbw = sta_priv->width == NL80211_CHAN_WIDTH_80 ? CBW80 :
	    sta_priv->width == NL80211_CHAN_WIDTH_40 ? CBW40 : CBW20;

	pr_debug("%s : fmt=%d, rateidx=%d, bw=%d, nss=%d, gi=%d, ltf=%d, %s\n",
		 __func__, fmt, rate_idx, bw, nss, gtype, ltf_type, is_ldpc ? "LDPC" : "BCC");
	memset(dbg_rate, 0, sizeof(struct ieee80211_dbg_rate_fix));
	/* check rate valid */
	switch (fmt) {
	case NON_HT:
		tmp_idx = sta_priv->band == NL80211_BAND_5GHZ ? rate_idx - 4 : rate_idx;

		if (!(BIT(tmp_idx) & get_ieee80211_supp_rates(sta, sta_priv->band)))
			break;

		dbg_rate->rate.idx = (s8)tmp_idx;
		/*
		 * for NON-HT translate sgi to short preamble
		 */
		if (gtype == SHORT_GI) {
			if (rate_idx > 0 && rate_idx <= 3) {
				dbg_rate->rate.flags |= IEEE80211_TX_RC_USE_SHORT_PREAMBLE;
				rate_valid = true;
			}
		} else {
			rate_valid = true;
		}
		break;
	case HT_MF:
		if (!get_ieee80211_ht_cap(sta).ht_supported)
			break;
		if (bw > CBW40 || bw > maxbw)
			break;
		if (rate_idx > sta_priv->ht_mcs)
			break;

		dbg_rate->rate.flags |= IEEE80211_TX_RC_MCS;
		dbg_rate->rate.idx = (s8)rate_idx;

		if (bw == CBW40)
			dbg_rate->rate.flags |= IEEE80211_TX_RC_40_MHZ_WIDTH;

		if (gtype == SHORT_GI &&
		    ((bw == CBW20 && (get_ieee80211_ht_cap(sta).cap & IEEE80211_HT_CAP_SGI_20)) ||
		     (bw == CBW40 && (get_ieee80211_ht_cap(sta).cap & IEEE80211_HT_CAP_SGI_40))))
			dbg_rate->rate.flags |= IEEE80211_TX_RC_SHORT_GI;

		if (is_ldpc && (get_ieee80211_ht_cap(sta).cap & IEEE80211_HT_CAP_LDPC_CODING))
			dbg_rate->flags |= IEEE80211_TX_CTL_LDPC;

		rate_valid = true;
		break;
	case VHT:
		if (!ieee80211_is_vht_sta(sta))
			break;
		if (rate_idx > sta_priv->vht_mcs)
			break;
		if (!nss || nss > min_t(u8, 2, sta_priv->n_ss))
			break;
		if (bw == CBW20 && rate_idx > 8)
			break;
		if (bw > maxbw)
			bw = maxbw;

		dbg_rate->rate.flags |= IEEE80211_TX_RC_VHT_MCS;
		ieee80211_rate_set_vht(&dbg_rate->rate, rate_idx, nss);

		if (bw == CBW40)
			dbg_rate->rate.flags |= IEEE80211_TX_RC_40_MHZ_WIDTH;
		else if (bw == CBW80)
			dbg_rate->rate.flags |= IEEE80211_TX_RC_80_MHZ_WIDTH;

		if (gtype == SHORT_GI &&
		    ((bw == CBW20 && (get_ieee80211_ht_cap(sta).cap & IEEE80211_HT_CAP_SGI_20)) ||
		     (bw == CBW40 && (get_ieee80211_ht_cap(sta).cap & IEEE80211_HT_CAP_SGI_40)) ||
		     (bw == CBW80 && (get_ieee80211_vht_cap(sta).cap &
				      IEEE80211_VHT_CAP_SHORT_GI_80))))
			dbg_rate->rate.flags |= IEEE80211_TX_RC_SHORT_GI;

		if (is_ldpc && (get_ieee80211_vht_cap(sta).cap & IEEE80211_VHT_CAP_RXLDPC))
			dbg_rate->flags |= IEEE80211_TX_CTL_LDPC;

		rate_valid = true;
		break;
	case HE_SU:
		if (!ieee80211_is_he_sta(sta))
			break;
		if (rate_idx > (7 + 2 * sta_priv->he_mcs[0]))
			break;
		if (!nss)
			break;
		if ((rate_idx > 9 || bw > CBW20) && !is_ldpc)
			break;
		if (gtype == GI_0u4s)
			break;
		if ((gtype == GI_1u6s && ltf_type == HE_LTF_4x) ||
		    (gtype == GI_3u2s && ltf_type == HE_LTF_2x))
			break;

		bw = min(bw, maxbw);
		nss = min_t(u8, 2, min(nss, sta_priv->n_ss));

		dbg_rate->rate.flags |= IEEE80211_TX_RC_HE_MCS;
		ieee80211_rate_set_vht(&dbg_rate->rate, rate_idx, nss);
		dbg_rate->he_gi_type = gtype;
		dbg_rate->he_ltf_type = ltf_type;

		if (bw == CBW40)
			dbg_rate->rate.flags |= IEEE80211_TX_RC_40_MHZ_WIDTH;
		else if (bw == CBW80)
			dbg_rate->rate.flags |= IEEE80211_TX_RC_80_MHZ_WIDTH;

		if (is_ldpc)
			dbg_rate->flags |= IEEE80211_TX_CTL_LDPC;

		rate_valid = true;
		break;
	default:
		break;
	}
	dbg_rate->band = vif->bss_conf.chandef.chan->band;

	return rate_valid;
}

int string_split(char *spl[], char *str, char *delim, int n)
{
	int i, c = 0;
	char *tmp;
	char *ps = str;

	if (!ps || !delim)
		return 0;

	tmp = ps;
	for (i = 0; i < n; i++) {
		while ((*ps != '\0') && (*ps != *delim))
			ps++;
		if (*ps == '\0')
			break;

		spl[c] = tmp;
		tmp = ps + 1;
		*ps = '\0';
		if (!tmp)
			break;

		ps = tmp;
		c++;
		spl[c] = ps;
	}

	return (c ? (c + 1) : c);
}

int decstr2int(char *str, enum HEX_STR_POS pos)
{
	int re = 0, i = 0;
	int s = (strlen(str) > 8) ? 8 : strlen(str);
	bool c = false;

	if (pos != POS_10 && pos != POS_16)
		return 0;

	if (str[0] == '-') {
		c = true;
		i = 1;
	}
	for (; i < s; i++) {
		if (str[i] == ' ')
			break;

		if (str[i] >= '0' && str[i] <= '9')
			re = re * pos + str[i] - '0';
		else if (pos == POS_10)
			break;	// only convert first byte is 0-9

		if (pos == POS_16) {
			if (str[i] >= 'A' && str[i] <= 'F')
				re = re * pos + 10 + str[i] - 'A';
			else if (str[i] >= 'a' && str[i] <= 'f')
				re = re * pos + 10 + str[i] - 'a';
		}
	}
	if (c)
		re = 0 - re;
	return re;
}

void uppercase(char *u_str, const char *str)
{
	int i;

	strscpy(u_str, str, strlen(str) + 1);
	for (i = 0; i < strlen(u_str); i++) {
		if (u_str[i] >= 'a' && u_str[i] <= 'z')
			u_str[i] -= 32;
	}
}
