// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
#ifdef SCM80211_WOW
#include <linux/compiler.h>
#include <linux/bitfield.h>

#define func_fmt(str)	"[WOW] "str
#define func_trace	PS
#include "core.h"
#include "wow.h"

//wow

void scm80211_wow_add_bitmap_pattern(struct scm80211_wow_context *wow,
				     struct cfg80211_wowlan *ntrig)
{
	struct scm80211_wow_bmap_filter *filter;
	int i, index = 0;

	for (i = 0; i < ntrig->n_patterns; i++) {
		if (ntrig->patterns[i].pattern_len > SCM80211_WOW_BMAP_PATTERN_SIZE_MAX) {
			//error
			continue;
		}
		filter = &wow->bmap_filter[i];
		filter->bmap_filter_pattern.pattern_len = ntrig->patterns[i].pattern_len;
		filter->bmap_filter_pattern.bitmask_len =
		    DIV_ROUND_UP(ntrig->patterns[i].pattern_len, 8);
		filter->bmap_filter_pattern.pattern_offset = ntrig->patterns[i].pkt_offset;
		memcpy(filter->bmap_filter_pattern.pattern,
		       ntrig->patterns[i].pattern, ntrig->patterns[i].pattern_len);
		memcpy(filter->bmap_filter_pattern.bitmask,
		       ntrig->patterns[i].mask, filter->bmap_filter_pattern.bitmask_len);
		filter->valid = 1;
		filter->index = index;
		index++;
		if (index >= wow->bmap_filter_max)
			break;
	}
}

int scm80211_wow_set_events(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct wiphy *wiphy = hw->wiphy;
	struct scm80211_wow_context *wow = &priv->wow;
	struct cfg80211_wowlan *ntrig = wiphy->wowlan_config;

	if (!ntrig) {
		scm80211_warn(hw, "%s: [wow] no parameters set yet %d\n", __func__, __LINE__);
		return -1;
	}

	scm80211_warn(hw, "[wow]: any=[%d] disconnect=[%d] magic_pkt=[%d] gtk_rekey_failure=[%d]\n",
		      ntrig->any, ntrig->disconnect, ntrig->magic_pkt, ntrig->gtk_rekey_failure);
	scm80211_warn(hw, "eap_identity_req=[%d] four_way_handshake=[%d] rfkill_release=[%d]\n",
		      ntrig->eap_identity_req, ntrig->four_way_handshake, ntrig->rfkill_release);

	if (ntrig->magic_pkt)
		wow->event_flag |= WOW_WAKEUP_E_MAGIC_PKT_RECVD_EVENT;
	else
		wow->event_flag &= (~WOW_WAKEUP_E_MAGIC_PKT_RECVD_EVENT);

	if (ntrig->n_patterns) {
		scm80211_wow_add_bitmap_pattern(wow, ntrig);
		wow->event_flag |= WOW_WAKEUP_E_PATTERN_MATCH_EVENT;
	} else {
		wow->event_flag &= (~WOW_WAKEUP_E_PATTERN_MATCH_EVENT);
	}

	scm80211_trace(hw, "%s: [wow] event_flag=0x%x\n", __func__, wow->event_flag);
	return 0;
}

int scm80211_wow_attach(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	scm80211_wow_context *wow = &priv->wow;

	scm80211_trace(hw, "%s: [wow] %d\n", __func__, __LINE__);
	memset(wow, 0, sizeof(struct scm80211_wow_context));
	wow->bmap_filter_max = SCM80211_WOW_BMAP_FILTER_MAX;
	return 0;
}

int scm80211_wow_detach(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_wow_context *wow = &priv->wow;

	scm80211_warn(hw, "%s: [wow] %d\n", __func__, __LINE__);
	memset(wow, 0, sizeof(struct scm80211_wow_context));
	return 0;
}

int scm80211_wow_op_suspend(struct ieee80211_hw *hw, struct cfg80211_wowlan *wowlan)
{
	scm80211_trace(hw, "rx:wow suspend\n");
	return 0;
}

void scm80211_wow_op_set_wakeup(struct ieee80211_hw *hw, bool enabled)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_wow_context *wow = &priv->wow;
	int i;

	if (enabled) {
		scm80211_wow_attach(hw);
		scm80211_wow_set_events(hw);	//write iw pattern into scm driver private struct
		for (i = 0; i < wow->bmap_filter_max; i++) {	//dump wow pattern
			scm80211_trace(hw, "wow pattern ofst:%d,len:%d\n",
				       wow->bmap_filter[i].bmap_filter_pattern.pattern_offset,
				       wow->bmap_filter[i].bmap_filter_pattern.pattern_len);
			print_hex_dump_bytes("rx_wow_pattern", DUMP_PREFIX_OFFSET,
					     wow->bmap_filter[i].bmap_filter_pattern.pattern,
					     wow->bmap_filter[i].bmap_filter_pattern.pattern_len *
					     SCM80211_WOW_BMAP_PATTERN_SIZE_MAX);
			scm80211_trace(hw, "wow pattern ofst:%d,len:%d\n",
				       wow->bmap_filter[i].bmap_filter_pattern.pattern_offset,
				       wow->bmap_filter[i].bmap_filter_pattern.pattern_len);
			print_hex_dump_bytes("rx_wow_pattern", DUMP_PREFIX_OFFSET,
					     wow->bmap_filter[i].bmap_filter_pattern.pattern,
					     wow->bmap_filter[i].bmap_filter_pattern.pattern_len *
					     SCM80211_WOW_BMAP_PATTERN_SIZE_MAX);
		}
		scm80211_wow_set_wakeup(hw, wow);	//only for USB platform
		scm80211_trace(hw, "%s: [wow] %d, sizeof wow:%ld\n", __func__, __LINE__,
			       sizeof(struct scm80211_wow_context));
	} else {
		scm80211_wow_detach(hw);
		scm80211_trace(hw, "%s: [wow] %d\n", __func__, __LINE__);
	}
	scm80211_trace(hw, "rx:wow set wakeup %d\n", enabled);	//wow enable:1;wow disable:0
}

int scm80211_wow_op_resume(struct ieee80211_hw *hw)
{
	scm80211_trace(hw, "rx:wow resume\n");
	return 0;
}
#endif //SCM80211_WOW
