// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 *
 */
#include <linux/platform_device.h>
#include <linux/bitfield.h>
#include <linux/reboot.h>
#include <linux/kdebug.h>
#include <linux/pci.h>

#define func_fmt(str)           "[MAIN] "str
#define func_trace		MAIN
#include "core.h"
#include "util.h"
#include "phy.h"
#include "trace.h"
#include "auto_version.h"
#include "cfg80211.h"
#include "debug.h"
#include "efuse.h"

/**
 * DOC: channel definition (struct cfg80211_chan_def)
 *
 * (1) @hw->conf.chandef: channel @hw is currently tuned to
 * (2) @vif->chanctx_conf.def: @vif operating channel
 * (3) @vif->bss_conf.chandef: @vif bss channel
 *     - @vif->bss_conf.chandef.chan can be NULL before association
 *
 * (1) can be different from (2): off-channel or scanning
 * (2) and (3) are identical?
 */

/**
 * DOC: survey
 * H/W needs to provide how much time is spent for each channel.
 * We will not support this for the time being.
 *
 */

/**
 * bcn_loss_discon_cnt in driver
 * refers to beacon_loss_count in net/mac80211/mlme.c
 * which is mac80211's beacon_loss_count * 3 - 1
 * if driver detect the beacon loss cnt grow more than 1
 * in this number of beacon intervals then do diconnection
 */
static int bcn_loss_discon_cnt = 20;
module_param(bcn_loss_discon_cnt, int, 0644);
MODULE_PARM_DESC(bcn_loss_discon_cnt, "Number of beacon intervals before we decide disconnection.");

static int rts_threshold = SCM80211_RTS_THRESHOLD;
module_param(rts_threshold, int, 0644);
MODULE_PARM_DESC(rts_threshold, "rts threshold base on ppdu txtime.");

static int debug_trace = MAIN_DBG;
module_param(debug_trace, int, 0644);
MODULE_PARM_DESC(debug_trace, "function debug control.");

static int dev_ant;
module_param(dev_ant, int, 0644);
MODULE_PARM_DESC(dev_ant, "select RF-ANT on board. 0: efuse cfg; 1: 2-ANT(default); 2: 3-ANT;");

#define CHAN2G(_freq)  {	   \
	.band = NL80211_BAND_2GHZ, \
	.center_freq = (_freq),	   \
	.hw_value = (_freq),	   \
	.max_power = 20,	   \
}

#define CHAN5G(_freq) {		   \
	.band = NL80211_BAND_5GHZ, \
	.center_freq = (_freq),	   \
	.hw_value = (_freq),	   \
	.max_power = 20,	   \
}

static const struct ieee80211_channel scm80211_chan_2ghz[] = {
	CHAN2G(2412),		/* Channel 1 */
	CHAN2G(2417),		/* Channel 2 */
	CHAN2G(2422),		/* Channel 3 */
	CHAN2G(2427),		/* Channel 4 */
	CHAN2G(2432),		/* Channel 5 */
	CHAN2G(2437),		/* Channel 6 */
	CHAN2G(2442),		/* Channel 7 */
	CHAN2G(2447),		/* Channel 8 */
	CHAN2G(2452),		/* Channel 9 */
	CHAN2G(2457),		/* Channel 10 */
	CHAN2G(2462),		/* Channel 11 */
	CHAN2G(2467),		/* Channel 12 */
	CHAN2G(2472),		/* Channel 13 */
	CHAN2G(2484),		/* Channel 14 */
};

static const struct ieee80211_channel scm80211_chan_5ghz[] = {
	CHAN5G(5180),		/* Channel 36 */
	CHAN5G(5200),		/* Channel 40 */
	CHAN5G(5220),		/* Channel 44 */
	CHAN5G(5240),		/* Channel 48 */

	CHAN5G(5260),		/* Channel 52 */
	CHAN5G(5280),		/* Channel 56 */
	CHAN5G(5300),		/* Channel 60 */
	CHAN5G(5320),		/* Channel 64 */

	CHAN5G(5500),		/* Channel 100 */
	CHAN5G(5520),		/* Channel 104 */
	CHAN5G(5540),		/* Channel 108 */
	CHAN5G(5560),		/* Channel 112 */
	CHAN5G(5580),		/* Channel 116 */
	CHAN5G(5600),		/* Channel 120 */
	CHAN5G(5620),		/* Channel 124 */
	CHAN5G(5640),		/* Channel 128 */
	CHAN5G(5660),		/* Channel 132 */
	CHAN5G(5680),		/* Channel 136 */
	CHAN5G(5700),		/* Channel 140 */
	CHAN5G(5745),		/* Channel 149 */
	CHAN5G(5765),		/* Channel 153 */
	CHAN5G(5785),		/* Channel 157 */
	CHAN5G(5805),		/* Channel 161 */
	CHAN5G(5825),		/* Channel 165 */
	CHAN5G(5845),		/* Channel 169 */

};

#define RATE(b, f, h) { .bitrate = (b), .flags = (f), .hw_value = (h) }

static const struct ieee80211_rate scm80211_rates[] = {
	RATE(10, 0, 0),
	RATE(20, IEEE80211_RATE_SHORT_PREAMBLE, 1),
	RATE(55, IEEE80211_RATE_SHORT_PREAMBLE, 2),
	RATE(110, IEEE80211_RATE_SHORT_PREAMBLE, 3),
	RATE(60, 0, 4),
	RATE(90, 0, 5),
	RATE(120, 0, 6),
	RATE(180, 0, 7),
	RATE(240, 0, 8),
	RATE(360, 0, 9),
	RATE(480, 0, 10),
	RATE(540, 0, 11)
};

#ifdef TODO_TO_REMOVE
// 9.4.2.55.5 HT Extended Capabilities field
#define SCM80211_HT_EXT_CAP_HTC_SUPPORT 0x400

// 9.4.2.55.6 Transmit Beamforming Capabilities
#define SCM80211_HT_CAP_TX_BF_RCV_NDP_CAP  0x8
#define SCM80211_HT_CAP_TX_BF_EXPLICIT_COMP_FEED_SHIFT 15
#define SCM80211_HT_CAP_TX_BF_MIN_GRP_SHIFT 17
#define SCM80211_HT_CAP_TX_BF_CSI_NUM_SHIFT 19
#define SCM80211_HT_CAP_TX_BF_NONCOMP_STEER_NUM_SHIFT 21
#define SCM80211_HT_CAP_TX_BF_COMP_STEER_NUM_SHIFT 23
#define SCM80211_HT_CAP_TX_BF_CSI_MAX_ROW_SHIFT 25
#define SCM80211_HT_CAP_TX_BF_CH_EST_CAP_SHIFT 27

static u32 extended_ht_cap_info = SCM80211_HT_EXT_CAP_HTC_SUPPORT;
static u32 tx_BF_cap_info = (SCM80211_HT_CAP_TX_BF_RCV_NDP_CAP |
			     (1 << SCM80211_HT_CAP_TX_BF_EXPLICIT_COMP_FEED_SHIFT) |
			     (3 << SCM80211_HT_CAP_TX_BF_MIN_GRP_SHIFT) |
			     (3 << SCM80211_HT_CAP_TX_BF_CSI_NUM_SHIFT) |
			     (3 << SCM80211_HT_CAP_TX_BF_NONCOMP_STEER_NUM_SHIFT) |
			     (3 << SCM80211_HT_CAP_TX_BF_COMP_STEER_NUM_SHIFT) |
			     (1 << SCM80211_HT_CAP_TX_BF_CSI_MAX_ROW_SHIFT) |
			     (3 << SCM80211_HT_CAP_TX_BF_CH_EST_CAP_SHIFT));
#endif

static struct ieee80211_sta_ht_cap scm80211_ht_cap = {
	.ht_supported = true,
	.cap = (IEEE80211_HT_CAP_LDPC_CODING |
		IEEE80211_HT_CAP_SUP_WIDTH_20_40 |
		IEEE80211_HT_CAP_SGI_20 |
		IEEE80211_HT_CAP_SGI_40 |
		IEEE80211_HT_CAP_TX_STBC |
		(1 << IEEE80211_HT_CAP_RX_STBC_SHIFT) |
#ifndef SCM80211_USE_4K_MPDU
		IEEE80211_HT_CAP_MAX_AMSDU |
#endif
		IEEE80211_HT_CAP_DSSSCCK40),
	.ampdu_factor = IEEE80211_HT_MAX_AMPDU_64K,
	.ampdu_density = IEEE80211_HT_MPDU_DENSITY_NONE,
	.mcs = {
		.rx_mask = { 0xff, 0xff, 0, 0, 0, 0, 0, 0, 0, 0,},
		.tx_params = IEEE80211_HT_MCS_TX_DEFINED,
		 },
};

static struct ieee80211_sta_vht_cap scm80211_vht_cap = {
	.vht_supported = true,
	.cap = (
#ifdef SCM80211_HOSTAP_AUTO_TEST
		       IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_11454 |
#elif defined SCM80211_USE_4K_MPDU
		       IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_3895 |
#else
		       IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_7991 |
#endif
		       IEEE80211_VHT_CAP_RXLDPC |
		       IEEE80211_VHT_CAP_SHORT_GI_80 |
		       IEEE80211_VHT_CAP_TXSTBC |
		       IEEE80211_VHT_CAP_RXSTBC_1 |
		       IEEE80211_VHT_CAP_SU_BEAMFORMEE_CAPABLE |
		       IEEE80211_VHT_CAP_MU_BEAMFORMEE_CAPABLE |
		       (SCM80211_VHT_CAP_BFM_STS_CAP <<
			IEEE80211_VHT_CAP_BEAMFORMEE_STS_SHIFT) |
		       IEEE80211_VHT_CAP_HTC_VHT |
		       (IEEE80211_VHT_MAX_AMPDU_1024K <<
			IEEE80211_VHT_CAP_MAX_A_MPDU_LENGTH_EXPONENT_SHIFT)),
	.vht_mcs = {
		    .rx_mcs_map = cpu_to_le16(0xfffa),
		    .tx_mcs_map = cpu_to_le16(0xfffa),
		     },
};

/* 2G HE STA mode capabilities */
#define scm80211_he_2G_cap_sta							\
	{									\
		.has_he = true,							\
		.he_cap_elem = {						\
			.mac_cap_info[0] =					\
				IEEE80211_HE_MAC_CAP0_HTC_HE |			\
				IEEE80211_HE_MAC_CAP0_TWT_REQ,			\
			.mac_cap_info[1] =					\
				IEEE80211_HE_MAC_CAP1_MULTI_TID_AGG_RX_QOS_8,	\
			.mac_cap_info[2] =					\
				IEEE80211_HE_MAC_CAP2_TRS |			\
				IEEE80211_HE_MAC_CAP2_BSR |			\
				IEEE80211_HE_MAC_CAP2_MU_CASCADING,		\
			.mac_cap_info[3] =					\
				IEEE80211_HE_MAC_CAP3_OMI_CONTROL |		\
				IEEE80211_HE_MAC_CAP3_OFDMA_RA |		\
				IEEE80211_HE_MAC_CAP3_MAX_AMPDU_LEN_EXP_EXT_2,	\
			.mac_cap_info[4] =					\
				IEEE80211_HE_MAC_CAP4_BSRP_BQRP_A_MPDU_AGG |	\
				IEEE80211_HE_MAC_CAP4_BQR |			\
				IEEE80211_HE_MAC_CAP4_AMSDU_IN_AMPDU |		\
				IEEE80211_HE_MAC_CAP4_MULTI_TID_AGG_TX_QOS_B39,	\
			.mac_cap_info[5] =					\
				IEEE80211_HE_MAC_CAP5_HT_VHT_TRIG_FRAME_RX,	\
			.phy_cap_info[0] =					\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_40MHZ_IN_2G |		\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_RU_MAPPING_IN_2G,	\
			.phy_cap_info[1] =							\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_20MHZ |\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_40MHZ |\
				IEEE80211_HE_PHY_CAP1_LDPC_CODING_IN_PAYLOAD |			\
				IEEE80211_HE_PHY_CAP1_MIDAMBLE_RX_TX_MAX_NSTS,			\
			.phy_cap_info[2] =					\
				IEEE80211_HE_PHY_CAP2_NDP_4x_LTF_AND_3_2US |	\
				IEEE80211_HE_PHY_CAP2_STBC_TX_UNDER_80MHZ |	\
				IEEE80211_HE_PHY_CAP2_STBC_RX_UNDER_80MHZ |	\
				IEEE80211_HE_PHY_CAP2_DOPPLER_TX |		\
				IEEE80211_HE_PHY_CAP2_DOPPLER_RX |		\
				IEEE80211_HE_PHY_CAP2_UL_MU_FULL_MU_MIMO |	\
				IEEE80211_HE_PHY_CAP2_UL_MU_PARTIAL_MU_MIMO,	\
			.phy_cap_info[3] =					\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_TX_16_QAM |	\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_RX_16_QAM,	\
			.phy_cap_info[4] =					\
				IEEE80211_HE_PHY_CAP4_SU_BEAMFORMEE |		\
				IEEE80211_HE_PHY_CAP4_BEAMFORMEE_MAX_STS_UNDER_80MHZ_8,	\
			.phy_cap_info[5] =							\
				IEEE80211_HE_PHY_CAP5_NG16_SU_FEEDBACK |			\
				IEEE80211_HE_PHY_CAP5_NG16_MU_FEEDBACK,			\
			.phy_cap_info[6] =					\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_42_SU |			\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_75_MU |			\
				IEEE80211_HE_PHY_CAP6_TRIG_SU_BEAMFORMING_FB |	\
				IEEE80211_HE_PHY_CAP6_TRIG_MU_BEAMFORMING_PARTIAL_BW_FB |	\
				IEEE80211_HE_PHY_CAP6_TRIG_CQI_FB |				\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BW_EXT_RANGE |			\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BANDWIDTH_DL_MUMIMO |		\
				IEEE80211_HE_PHY_CAP6_PPE_THRESHOLD_PRESENT,			\
			.phy_cap_info[7] =							\
				IEEE80211_HE_PHY_CAP7_HE_SU_MU_PPDU_4XLTF_AND_08_US_GI,		\
			.phy_cap_info[8] =							\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_PPDU_4XLTF_AND_08_US_GI |	\
				IEEE80211_HE_PHY_CAP8_20MHZ_IN_40MHZ_HE_PPDU_IN_2G |		\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_1XLTF_AND_08_US_GI |		\
				IEEE80211_HE_PHY_CAP8_MIDAMBLE_RX_TX_2X_AND_1XLTF |		\
				IEEE80211_HE_PHY_CAP8_DCM_MAX_RU_484,				\
			.phy_cap_info[9] =							\
				IEEE80211_HE_PHY_CAP9_LONGER_THAN_16_SIGB_OFDM_SYM |		\
				IEEE80211_HE_PHY_CAP9_TX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_COMP_SIGB |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_NON_COMP_SIGB,\
		},										\
		.he_mcs_nss_supp = {								\
			.rx_mcs_80 = cpu_to_le16(0xfffa),					\
			.tx_mcs_80 = cpu_to_le16(0xfffa),					\
			.rx_mcs_160 = cpu_to_le16(0xffff),					\
			.tx_mcs_160 = cpu_to_le16(0xffff),					\
			.rx_mcs_80p80 = cpu_to_le16(0xffff),					\
			.tx_mcs_80p80 = cpu_to_le16(0xffff),					\
		},										\
	}

/* 2G HE AP mode capabilities */
#define scm80211_he_2G_cap_ap							\
	{									\
		.has_he = true,							\
		.he_cap_elem = {						\
			.mac_cap_info[0] =					\
				IEEE80211_HE_MAC_CAP0_HTC_HE |			\
				IEEE80211_HE_MAC_CAP0_TWT_REQ,			\
			.mac_cap_info[1] =					\
				IEEE80211_HE_MAC_CAP1_MULTI_TID_AGG_RX_QOS_8,	\
			.mac_cap_info[2] =					\
				IEEE80211_HE_MAC_CAP2_TRS |			\
				IEEE80211_HE_MAC_CAP2_BSR |			\
				IEEE80211_HE_MAC_CAP2_MU_CASCADING,		\
			.mac_cap_info[3] =					\
				IEEE80211_HE_MAC_CAP3_OMI_CONTROL |		\
				IEEE80211_HE_MAC_CAP3_OFDMA_RA |		\
				IEEE80211_HE_MAC_CAP3_MAX_AMPDU_LEN_EXP_EXT_2,	\
			.mac_cap_info[4] =					\
				IEEE80211_HE_MAC_CAP4_BSRP_BQRP_A_MPDU_AGG |	\
				IEEE80211_HE_MAC_CAP4_BQR |			\
				IEEE80211_HE_MAC_CAP4_AMSDU_IN_AMPDU |		\
				IEEE80211_HE_MAC_CAP4_MULTI_TID_AGG_TX_QOS_B39,	\
			.mac_cap_info[5] =					\
				IEEE80211_HE_MAC_CAP5_HT_VHT_TRIG_FRAME_RX,	\
			.phy_cap_info[0] =					\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_40MHZ_IN_2G,	\
			.phy_cap_info[1] =							\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_20MHZ |\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_40MHZ |\
				IEEE80211_HE_PHY_CAP1_LDPC_CODING_IN_PAYLOAD |			\
				IEEE80211_HE_PHY_CAP1_MIDAMBLE_RX_TX_MAX_NSTS,			\
			.phy_cap_info[2] =					\
				IEEE80211_HE_PHY_CAP2_NDP_4x_LTF_AND_3_2US |	\
				IEEE80211_HE_PHY_CAP2_STBC_TX_UNDER_80MHZ |	\
				IEEE80211_HE_PHY_CAP2_STBC_RX_UNDER_80MHZ |	\
				IEEE80211_HE_PHY_CAP2_DOPPLER_TX |		\
				IEEE80211_HE_PHY_CAP2_DOPPLER_RX |		\
				IEEE80211_HE_PHY_CAP2_UL_MU_FULL_MU_MIMO |	\
				IEEE80211_HE_PHY_CAP2_UL_MU_PARTIAL_MU_MIMO,	\
			.phy_cap_info[3] =					\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_TX_16_QAM |	\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_RX_16_QAM,	\
			.phy_cap_info[4] =					\
				IEEE80211_HE_PHY_CAP4_SU_BEAMFORMEE |		\
				IEEE80211_HE_PHY_CAP4_BEAMFORMEE_MAX_STS_UNDER_80MHZ_8,	\
			.phy_cap_info[5] =							\
				IEEE80211_HE_PHY_CAP5_NG16_SU_FEEDBACK |			\
				IEEE80211_HE_PHY_CAP5_NG16_MU_FEEDBACK,			\
			.phy_cap_info[6] =					\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_42_SU |			\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_75_MU |			\
				IEEE80211_HE_PHY_CAP6_TRIG_SU_BEAMFORMING_FB |	\
				IEEE80211_HE_PHY_CAP6_TRIG_MU_BEAMFORMING_PARTIAL_BW_FB |	\
				IEEE80211_HE_PHY_CAP6_TRIG_CQI_FB |				\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BW_EXT_RANGE |			\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BANDWIDTH_DL_MUMIMO |		\
				IEEE80211_HE_PHY_CAP6_PPE_THRESHOLD_PRESENT,			\
			.phy_cap_info[7] =							\
				IEEE80211_HE_PHY_CAP7_HE_SU_MU_PPDU_4XLTF_AND_08_US_GI,		\
			.phy_cap_info[8] =							\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_PPDU_4XLTF_AND_08_US_GI |	\
				IEEE80211_HE_PHY_CAP8_20MHZ_IN_40MHZ_HE_PPDU_IN_2G |		\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_1XLTF_AND_08_US_GI |		\
				IEEE80211_HE_PHY_CAP8_MIDAMBLE_RX_TX_2X_AND_1XLTF |		\
				IEEE80211_HE_PHY_CAP8_DCM_MAX_RU_484,				\
			.phy_cap_info[9] =							\
				IEEE80211_HE_PHY_CAP9_LONGER_THAN_16_SIGB_OFDM_SYM |		\
				IEEE80211_HE_PHY_CAP9_TX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_COMP_SIGB |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_NON_COMP_SIGB,\
		},										\
		.he_mcs_nss_supp = {								\
			.rx_mcs_80 = cpu_to_le16(0xfffa),					\
			.tx_mcs_80 = cpu_to_le16(0xfffa),					\
			.rx_mcs_160 = cpu_to_le16(0xffff),					\
			.tx_mcs_160 = cpu_to_le16(0xffff),					\
			.rx_mcs_80p80 = cpu_to_le16(0xffff),					\
			.tx_mcs_80p80 = cpu_to_le16(0xffff),					\
		},										\
	}

/* 5G HE STA mode capabilities */
#define scm80211_he_5G_cap_sta									\
	{											\
		.has_he = true,									\
		.he_cap_elem = {								\
			.mac_cap_info[0] =							\
				IEEE80211_HE_MAC_CAP0_HTC_HE |					\
				IEEE80211_HE_MAC_CAP0_TWT_REQ,					\
			.mac_cap_info[1] =							\
				IEEE80211_HE_MAC_CAP1_MULTI_TID_AGG_RX_QOS_8,			\
			.mac_cap_info[2] =							\
				IEEE80211_HE_MAC_CAP2_TRS |					\
				IEEE80211_HE_MAC_CAP2_BSR |					\
				IEEE80211_HE_MAC_CAP2_MU_CASCADING,				\
			.mac_cap_info[3] =							\
				IEEE80211_HE_MAC_CAP3_OMI_CONTROL |				\
				IEEE80211_HE_MAC_CAP3_OFDMA_RA,					\
			.mac_cap_info[4] =							\
				IEEE80211_HE_MAC_CAP4_BSRP_BQRP_A_MPDU_AGG |			\
				IEEE80211_HE_MAC_CAP4_BQR |					\
				IEEE80211_HE_MAC_CAP4_AMSDU_IN_AMPDU |				\
				IEEE80211_HE_MAC_CAP4_MULTI_TID_AGG_TX_QOS_B39,			\
			.mac_cap_info[5] =							\
				IEEE80211_HE_MAC_CAP5_HT_VHT_TRIG_FRAME_RX,			\
			.phy_cap_info[0] =							\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_40MHZ_80MHZ_IN_5G |	\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_RU_MAPPING_IN_5G,	\
			.phy_cap_info[1] =							\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_20MHZ |\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_40MHZ |\
				IEEE80211_HE_PHY_CAP1_LDPC_CODING_IN_PAYLOAD |			\
				IEEE80211_HE_PHY_CAP1_MIDAMBLE_RX_TX_MAX_NSTS,			\
			.phy_cap_info[2] =							\
				IEEE80211_HE_PHY_CAP2_NDP_4x_LTF_AND_3_2US |			\
				IEEE80211_HE_PHY_CAP2_STBC_TX_UNDER_80MHZ |			\
				IEEE80211_HE_PHY_CAP2_STBC_RX_UNDER_80MHZ |			\
				IEEE80211_HE_PHY_CAP2_DOPPLER_TX |				\
				IEEE80211_HE_PHY_CAP2_DOPPLER_RX |				\
				IEEE80211_HE_PHY_CAP2_UL_MU_FULL_MU_MIMO |			\
				IEEE80211_HE_PHY_CAP2_UL_MU_PARTIAL_MU_MIMO,			\
			.phy_cap_info[3] =							\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_TX_16_QAM |			\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_RX_16_QAM,			\
			.phy_cap_info[4] =							\
				IEEE80211_HE_PHY_CAP4_SU_BEAMFORMEE |				\
				IEEE80211_HE_PHY_CAP4_BEAMFORMEE_MAX_STS_UNDER_80MHZ_8,		\
			.phy_cap_info[5] =							\
				IEEE80211_HE_PHY_CAP5_NG16_SU_FEEDBACK |			\
				IEEE80211_HE_PHY_CAP5_NG16_MU_FEEDBACK,				\
			.phy_cap_info[6] =							\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_42_SU |			\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_75_MU |			\
				IEEE80211_HE_PHY_CAP6_TRIG_SU_BEAMFORMING_FB |			\
				IEEE80211_HE_PHY_CAP6_TRIG_MU_BEAMFORMING_PARTIAL_BW_FB |	\
				IEEE80211_HE_PHY_CAP6_TRIG_CQI_FB |				\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BW_EXT_RANGE |			\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BANDWIDTH_DL_MUMIMO |		\
				IEEE80211_HE_PHY_CAP6_PPE_THRESHOLD_PRESENT,			\
			.phy_cap_info[7] =							\
				IEEE80211_HE_PHY_CAP7_HE_SU_MU_PPDU_4XLTF_AND_08_US_GI,		\
			.phy_cap_info[8] =							\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_PPDU_4XLTF_AND_08_US_GI |	\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_1XLTF_AND_08_US_GI |		\
				IEEE80211_HE_PHY_CAP8_MIDAMBLE_RX_TX_2X_AND_1XLTF |		\
				IEEE80211_HE_PHY_CAP8_DCM_MAX_RU_996,				\
			.phy_cap_info[9] =							\
				IEEE80211_HE_PHY_CAP9_LONGER_THAN_16_SIGB_OFDM_SYM |		\
				IEEE80211_HE_PHY_CAP9_TX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_COMP_SIGB |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_NON_COMP_SIGB,\
		},										\
		.he_mcs_nss_supp = {								\
			.rx_mcs_80 = cpu_to_le16(0xfffa),					\
			.tx_mcs_80 = cpu_to_le16(0xfffa),					\
			.rx_mcs_160 = cpu_to_le16(0xffff),					\
			.tx_mcs_160 = cpu_to_le16(0xffff),					\
			.rx_mcs_80p80 = cpu_to_le16(0xffff),					\
			.tx_mcs_80p80 = cpu_to_le16(0xffff),					\
		},										\
	}

/* 5G HE AP mode capabilities */
#define scm80211_he_5G_cap_ap									\
	{											\
		.has_he = true,									\
		.he_cap_elem = {								\
			.mac_cap_info[0] =							\
				IEEE80211_HE_MAC_CAP0_HTC_HE |					\
				IEEE80211_HE_MAC_CAP0_TWT_REQ,					\
			.mac_cap_info[1] =							\
				IEEE80211_HE_MAC_CAP1_MULTI_TID_AGG_RX_QOS_8,			\
			.mac_cap_info[2] =							\
				IEEE80211_HE_MAC_CAP2_TRS |					\
				IEEE80211_HE_MAC_CAP2_BSR |					\
				IEEE80211_HE_MAC_CAP2_MU_CASCADING,				\
			.mac_cap_info[3] =							\
				IEEE80211_HE_MAC_CAP3_OMI_CONTROL |				\
				IEEE80211_HE_MAC_CAP3_OFDMA_RA,					\
			.mac_cap_info[4] =							\
				IEEE80211_HE_MAC_CAP4_BSRP_BQRP_A_MPDU_AGG |			\
				IEEE80211_HE_MAC_CAP4_BQR |					\
				IEEE80211_HE_MAC_CAP4_AMSDU_IN_AMPDU |				\
				IEEE80211_HE_MAC_CAP4_MULTI_TID_AGG_TX_QOS_B39,			\
			.mac_cap_info[5] =							\
				IEEE80211_HE_MAC_CAP5_HT_VHT_TRIG_FRAME_RX,			\
			.phy_cap_info[0] =							\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_40MHZ_80MHZ_IN_5G |	\
				IEEE80211_HE_PHY_CAP0_CHANNEL_WIDTH_SET_RU_MAPPING_IN_5G,	\
			.phy_cap_info[1] =							\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_20MHZ |\
				IEEE80211_HE_PHY_CAP1_PREAMBLE_PUNC_RX_80MHZ_ONLY_SECOND_40MHZ |\
				IEEE80211_HE_PHY_CAP1_LDPC_CODING_IN_PAYLOAD |			\
				IEEE80211_HE_PHY_CAP1_MIDAMBLE_RX_TX_MAX_NSTS,			\
			.phy_cap_info[2] =							\
				IEEE80211_HE_PHY_CAP2_NDP_4x_LTF_AND_3_2US |			\
				IEEE80211_HE_PHY_CAP2_STBC_TX_UNDER_80MHZ |			\
				IEEE80211_HE_PHY_CAP2_STBC_RX_UNDER_80MHZ |			\
				IEEE80211_HE_PHY_CAP2_DOPPLER_TX |				\
				IEEE80211_HE_PHY_CAP2_DOPPLER_RX |				\
				IEEE80211_HE_PHY_CAP2_UL_MU_FULL_MU_MIMO |			\
				IEEE80211_HE_PHY_CAP2_UL_MU_PARTIAL_MU_MIMO,			\
			.phy_cap_info[3] =							\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_TX_16_QAM |			\
				IEEE80211_HE_PHY_CAP3_DCM_MAX_CONST_RX_16_QAM,			\
			.phy_cap_info[4] =							\
				IEEE80211_HE_PHY_CAP4_SU_BEAMFORMEE |				\
				IEEE80211_HE_PHY_CAP4_BEAMFORMEE_MAX_STS_UNDER_80MHZ_8,		\
			.phy_cap_info[5] =							\
				IEEE80211_HE_PHY_CAP5_NG16_SU_FEEDBACK |			\
				IEEE80211_HE_PHY_CAP5_NG16_MU_FEEDBACK,				\
			.phy_cap_info[6] =							\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_42_SU |			\
				IEEE80211_HE_PHY_CAP6_CODEBOOK_SIZE_75_MU |			\
				IEEE80211_HE_PHY_CAP6_TRIG_SU_BEAMFORMING_FB |			\
				IEEE80211_HE_PHY_CAP6_TRIG_MU_BEAMFORMING_PARTIAL_BW_FB |	\
				IEEE80211_HE_PHY_CAP6_TRIG_CQI_FB |				\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BW_EXT_RANGE |			\
				IEEE80211_HE_PHY_CAP6_PARTIAL_BANDWIDTH_DL_MUMIMO |		\
				IEEE80211_HE_PHY_CAP6_PPE_THRESHOLD_PRESENT,			\
			.phy_cap_info[7] =							\
				IEEE80211_HE_PHY_CAP7_HE_SU_MU_PPDU_4XLTF_AND_08_US_GI,		\
			.phy_cap_info[8] =							\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_PPDU_4XLTF_AND_08_US_GI |	\
				IEEE80211_HE_PHY_CAP8_HE_ER_SU_1XLTF_AND_08_US_GI |		\
				IEEE80211_HE_PHY_CAP8_MIDAMBLE_RX_TX_2X_AND_1XLTF |		\
				IEEE80211_HE_PHY_CAP8_DCM_MAX_RU_996,				\
			.phy_cap_info[9] =							\
				IEEE80211_HE_PHY_CAP9_LONGER_THAN_16_SIGB_OFDM_SYM |		\
				IEEE80211_HE_PHY_CAP9_TX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_1024_QAM_LESS_THAN_242_TONE_RU |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_COMP_SIGB |	\
				IEEE80211_HE_PHY_CAP9_RX_FULL_BW_SU_USING_MU_WITH_NON_COMP_SIGB,\
		},										\
		.he_mcs_nss_supp = {								\
			.rx_mcs_80 = cpu_to_le16(0xfffa),					\
			.tx_mcs_80 = cpu_to_le16(0xfffa),					\
			.rx_mcs_160 = cpu_to_le16(0xffff),					\
			.tx_mcs_160 = cpu_to_le16(0xffff),					\
			.rx_mcs_80p80 = cpu_to_le16(0xffff),					\
			.tx_mcs_80p80 = cpu_to_le16(0xffff),					\
		},										\
	}

static const struct ieee80211_sband_iftype_data scm80211_he_cap_2ghz[] = {
	{
	 .types_mask = BIT(NL80211_IFTYPE_STATION),
	 .he_cap = scm80211_he_2G_cap_sta,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_AP),
	 .he_cap = scm80211_he_2G_cap_ap,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_P2P_CLIENT),
	 .he_cap = scm80211_he_2G_cap_sta,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_P2P_GO),
	 .he_cap = scm80211_he_2G_cap_ap,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_MESH_POINT),
	 .he_cap = scm80211_he_2G_cap_ap,
	  },
};

static const struct ieee80211_sband_iftype_data scm80211_he_cap_5ghz[] = {
	{
	 .types_mask = BIT(NL80211_IFTYPE_STATION),
	 .he_cap = scm80211_he_5G_cap_sta,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_AP),
	 .he_cap = scm80211_he_5G_cap_ap,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_P2P_CLIENT),
	 .he_cap = scm80211_he_5G_cap_sta,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_P2P_GO),
	 .he_cap = scm80211_he_5G_cap_ap,
	  },
	{
	 .types_mask = BIT(NL80211_IFTYPE_MESH_POINT),
	 .he_cap = scm80211_he_5G_cap_ap,
	  },
};

/**
 * This is to override regulatory domains defined in cfg80211 module (reg.c)
 * By default world regulatory domain defined in reg.c puts the flags
 * NL80211_RRF_NO_IR for 5GHz channels (for * 36..48 and 149..165).
 * With respect to these flags. All the changes in world regulatory
 * domain are to be done here.
 */
static const struct ieee80211_regdomain scm80211_regdom = {
	.n_reg_rules = 4,
	.alpha2 = "99",
	.reg_rules = {
		      /* IEEE 802.11b/g, channels 1..11 */
		      REG_RULE(2412 - 10, 2472 + 10, 40, 6, 20, 0),
		      /* If any */
		      /* IEEE 802.11 channel 14 - Only JP enables
		       * this and for 802.11b only
		       */
		      REG_RULE(2484 - 10, 2484 + 10, 20, 6, 20, 0),
		      /* IEEE 802.11a, channel 36..64 */
		      REG_RULE(5150 - 10, 5350 + 10, 80, 6, 20, 0),
		      /* IEEE 802.11a, channel 100..165 */
		      REG_RULE(5470 - 10, 5850 + 10, 80, 6, 20, 0), }
};

static const u32 scm80211_ciphers[] = {
	WLAN_CIPHER_SUITE_WEP40,
	WLAN_CIPHER_SUITE_WEP104,
	WLAN_CIPHER_SUITE_TKIP,
	WLAN_CIPHER_SUITE_CCMP,
	WLAN_CIPHER_SUITE_CCMP_256,
	WLAN_CIPHER_SUITE_GCMP,
	WLAN_CIPHER_SUITE_GCMP_256,
	WLAN_CIPHER_SUITE_AES_CMAC,
	WLAN_CIPHER_SUITE_BIP_CMAC_256,
	WLAN_CIPHER_SUITE_BIP_GMAC_128,
	WLAN_CIPHER_SUITE_BIP_GMAC_256,
};

/* FIXME it doesn't work need FW support */
static const struct wiphy_wowlan_support scm80211_wowlan_support = {
	.flags = WIPHY_WOWLAN_ANY | WIPHY_WOWLAN_MAGIC_PKT,
	.n_patterns = 4,
	.pattern_max_len = 128,
	.pattern_min_len = 8,
	.max_pkt_offset = 96,
};

#ifdef SCM2020_A0
#define NR_PINS				61
#define WIFI_SYS_EN			22

static void scm2020_gpio_set_pmx(struct ieee80211_hw *hw, u32 pin, u8 func)
{
	u32 bit, mask, v, addr;

	addr = (pin / 8) * 4; /* 8 pins per register */
	bit = (pin & 7) * 4;
	mask = GENMASK(bit + 3, bit);

	v = scm80211_raw_readl(hw, WCM_GPIO_MODE_BASE + addr);
	v &= ~mask;
	v |= (func << bit) & mask;
	scm80211_raw_writel(hw, v, WCM_GPIO_MODE_BASE + addr);
}

static int scm2020_gpio_set_direction(struct ieee80211_hw *hw, u32 pin, enum scm2020_gpio_dir dir)
{
	u32 v, bit, addr;

	addr = (pin / 32) * 4;
	bit = pin & 31;

	v = scm80211_raw_readl(hw, WCM_GPIO_DIR_BASE + addr) & ~GENMASK(bit, bit);
	if (dir)
		v |= GENMASK(bit, bit);
	else
		v &= ~GENMASK(bit, bit);
	scm80211_raw_writel(hw, v, WCM_GPIO_DIR_BASE + addr);
	return 0;
}

static void scm2020_gpio_set_value(struct ieee80211_hw *hw, u32 pin, bool val)
{
	u32 bit, v, addr;

	addr = (pin / 32) * 4;
	bit = pin & 31;

	v = scm80211_raw_readl(hw, WCM_GPIO_VAL_BASE + addr);
	if (val)
		v |= GENMASK(bit, bit);
	else
		v &= ~GENMASK(bit, bit);
	scm80211_raw_writel(hw, v, WCM_GPIO_VAL_BASE + addr);
}

static int scm2020_gpio_get_value(struct ieee80211_hw *hw, u32 pin)
{
	u32 offset = (pin / 32) * 4;
	u32 bit = pin % 32;

	if (scm80211_raw_readl(hw, WCM_GPIO_IN_BASE + offset) & BIT(bit))
		return 1;

	return 0;
}

void scm2020_set_gpio_config(struct ieee80211_hw *hw, u32 pin, u32 mode,
				enum scm2020_gpio_dir dir, bool val)
{
	WARN_ON_ONCE(pin >= NR_PINS);
	scm2020_gpio_set_direction(hw, pin, dir);
	/* Set to GPIO mode */
	scm2020_gpio_set_pmx(hw, pin, mode);
	scm2020_gpio_set_value(hw, pin, val);
}

static void scm2020_pta_pinctrl_enable(struct ieee80211_hw *hw, u32 pin, int mode)
{
	WARN_ON_ONCE(pin >= NR_PINS);

	/* Safe to set the direction input before enabling GPIO function */
	scm2020_gpio_set_direction(hw, pin, SCM_GPIO_IN);

	scm2020_gpio_set_pmx(hw, pin, mode);
}

static void scm2020_ant_pinctrl(struct ieee80211_hw *hw)
{
	int pta_logic_pins[6] = {40, 41, 53, 54, 52, 55};
	u32 wifi_bt_ctrl_list[3] = {0x4011004C, 0x401100E4, 0x401100f4};
	u32 wifi_bt_ctrl_values[3] = {0x06060606, 0x07060e0d, 0x1e1f1514};
	int i, mode = 5;

	for (i = 0; i < 4; i++)
		scm2020_pta_pinctrl_enable(hw, pta_logic_pins[i], mode);

#ifndef SCM80211_FTM
	if (is_pcie_device(hw)) {
		/* GPIO40/41 used for controlling shared antenna */
		scm2020_set_gpio_config(hw, 40, 1, SCM_GPIO_OUT, true);
		scm2020_set_gpio_config(hw, 41, 1, SCM_GPIO_OUT, true);
	}
#endif

	/* GPIO52 default as WF_JRSTN, will let 16dBm reduction in RSSI */
	scm2020_pta_pinctrl_enable(hw, pta_logic_pins[4], 0);
	/* GPIO55 used for RF 5G SW */
	scm2020_pta_pinctrl_enable(hw, pta_logic_pins[5], 0);

	for (i = 0; i < 3; i++)
		scm80211_raw_writel(hw, wifi_bt_ctrl_values[i], wifi_bt_ctrl_list[i]);
}
#endif

static int scm80211_alloc_vdev(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	int i;

	/* Find the free vdev */
	i = find_first_zero_bit(&hwp->vif_map, SCM80211_NUM_VIF);
	hwp->vif_map |= BIT(i);
	return (i == SCM80211_NUM_VIF) ? -ENOENT : i;
}

static void scm80211_free_vdev(struct ieee80211_hw *hw, int index)
{
	struct ieee80211_hw_priv *hwp = hw->priv;

	hwp->vif_map &= ~BIT(index);
}

int scm80211_vdev_id(struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp;

	if (!vif)
		return -1;
	vp = ieee80211_to_priv(vif);
	return vp->index;
}

struct ieee80211_vif *scm80211_get_vif_from_vedv(struct ieee80211_hw *hw, int vdev)
{
	struct ieee80211_hw_priv *hwp = hw->priv;

	if ((hwp->vif_map & BIT(vdev)) && hwp->vifs[vdev])
		return hwp->vifs[vdev];
	else
		return NULL;
}

/**
 * DOC: Hardware interface
 */

#define sshift		__bf_shf(IEEE80211_FCTL_STYPE)
#define mshift		__bf_shf(VIF_RX_FILTER1_MGNT_FRAME_FILTER_BITMAP)
#define cshift		__bf_shf(VIF_RX_FILTER1_CTRL_FRAME_FILTER_BITMAP)

#define mbit(stype)	(BIT((stype) >> sshift) << mshift)
#define cbit(stype)	(BIT((stype) >> sshift) << cshift)

#define rx_filter_control	     (cbit(IEEE80211_STYPE_TRIGGER) |		\
				      cbit(IEEE80211_STYPE_BRPOLL) |		\
				      cbit(IEEE80211_STYPE_VHT_NDPA) |		\
				      cbit(IEEE80211_STYPE_CTL_EXT) |		\
				      cbit(IEEE80211_STYPE_BACK_REQ) |		\
				      cbit(IEEE80211_STYPE_BACK) |		\
				      cbit(IEEE80211_STYPE_RTS) |		\
				      cbit(IEEE80211_STYPE_CTS) |		\
				      cbit(IEEE80211_STYPE_ACK) |		\
				      cbit(IEEE80211_STYPE_CFEND) |		\
				      0)

#define  rx_filter_common	     (mbit(IEEE80211_STYPE_PROBE_RESP) |	\
				      mbit(IEEE80211_STYPE_BEACON) |		\
				      mbit(IEEE80211_STYPE_DISASSOC) |		\
				      mbit(IEEE80211_STYPE_AUTH) |		\
				      mbit(IEEE80211_STYPE_DEAUTH) |		\
				      mbit(IEEE80211_STYPE_ACTION) |		\
				      mbit(IEEE80211_STYPE_ACTION_NOACK) |	\
				      cbit(IEEE80211_STYPE_BACK_REQ) |		\
				      1)

#define rx_filter_station            (rx_filter_common |			\
				      mbit(IEEE80211_STYPE_ASSOC_RESP) |	\
				      mbit(IEEE80211_STYPE_REASSOC_RESP) |	\
				      0)

#define rx_filter_ap		     (rx_filter_common |			\
				      mbit(IEEE80211_STYPE_ASSOC_REQ) |		\
				      mbit(IEEE80211_STYPE_REASSOC_REQ) |	\
				      mbit(IEEE80211_STYPE_PROBE_REQ) |		\
				      cbit(IEEE80211_STYPE_PSPOLL) |		\
				      0)

#define rx_filter_ibss		     (rx_filter_common |			\
				      mbit(IEEE80211_STYPE_ASSOC_RESP) |	\
				      mbit(IEEE80211_STYPE_REASSOC_RESP) |	\
				      mbit(IEEE80211_STYPE_ASSOC_REQ) |		\
				      mbit(IEEE80211_STYPE_REASSOC_REQ) |	\
				      mbit(IEEE80211_STYPE_PROBE_REQ) |		\
				      cbit(IEEE80211_STYPE_PSPOLL) |		\
				      mbit(IEEE80211_STYPE_ATIM) |		\
				      0)

/**
 * scm80211_iftype_filter array
 * set default filters for each type of interfaces
 */
static const u32 scm80211_iftype_filter[NUM_NL80211_IFTYPES] = {
	[NL80211_IFTYPE_UNSPECIFIED]	= 0,
	[NL80211_IFTYPE_ADHOC]		= rx_filter_ibss,
	[NL80211_IFTYPE_STATION]	= rx_filter_station,
	[NL80211_IFTYPE_AP]		= rx_filter_ap,
	[NL80211_IFTYPE_AP_VLAN]	= rx_filter_ap,
	[NL80211_IFTYPE_WDS]		= rx_filter_station | rx_filter_ap,
	[NL80211_IFTYPE_MONITOR]	= -1U,
	[NL80211_IFTYPE_MESH_POINT]	= rx_filter_station | rx_filter_ap,
	[NL80211_IFTYPE_P2P_CLIENT]	= rx_filter_station, /* FIXME */
	[NL80211_IFTYPE_P2P_GO]		= rx_filter_station | rx_filter_ap, /* FIXME */
	[NL80211_IFTYPE_P2P_DEVICE]	= rx_filter_station | rx_filter_ap, /* FIXME */
	[NL80211_IFTYPE_OCB]		= rx_filter_station | rx_filter_ap, /* FIXME */
	[NL80211_IFTYPE_NAN]		= rx_filter_station | rx_filter_ap, /* FIXME */
};

#ifdef SCM80211_FTM
#include "ftm.h"
#endif

/**
 * scm80211_config_frame_filter() - configure interface frame filter
 * @hw: hardware
 * @vif: interface
 * @set: bitmap of frames to enable (pass)
 * @clear: bitmap of frames to disable (block)
 *
 * This function programs RX frame filter. If the intersection of
 * @set and @clear is non-empty, they are set.
 */
static void scm80211_config_frame_filter(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif, u32 set, u32 clear)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int vdev = scm80211_vdev_id(vif);
	bool ps = scm80211_ps_enabled(hw);
	u32 v;

	if (ps) {
		if (priv->pm.link)
			scm80211_cleanup_link_ps(hw, vif);
		else
			scm80211_cleanup_unlink_ps(hw, vif);
	}

#ifdef	SCM80211_FTM
	if (!scm80211_ftm_rx_enable_filter(hw, vif))
#endif
	{
		v = scm80211_mac_readl(hw, MAC_REG_VIF_RX_FILTER1(vdev));
		v &= ~(set | clear);
		v |= set;
		scm80211_mac_writel(hw, v, MAC_REG_VIF_RX_FILTER1(vdev));
	}

	if (ps) {
		if (priv->pm.link)
			scm80211_setup_link_ps(hw, vif);
		else
			scm80211_setup_unlink_ps(hw, vif, SCM80211_UNLINK_TIMEOUT);
	}
}

/**
 * scm80211_config_bssid_filter() - configure BSSID filter
 * @hw: hardware
 * @vif: interface
 * @change: Bitmap figure out which BSSID filter need to be turn on
 * @check_bssid: Indicates whether need to check the bssid in hw first
 *
 * This function configures the interface's BSSID filter. If disabled, BSSID
 * filter passes all the group addressed frames: otherwise, the filter
 * passes only those frames whose BSSID matches at least one of BSSIDs of the
 * @vif.
 *
 * Note that we turn on or off both MAC_REG_VIF_RX_FILTER0 BSS_EN and BFE.
 */
void scm80211_config_bssid_filter(struct ieee80211_hw *hw, struct ieee80211_vif *vif, bool enable)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_bss_conf *info = &vif->bss_conf;
	int vidx = scm80211_vdev_id(vif);
	u32 v;

	/* AP seems to have to listen to OBSS */
	if (vif->type == NL80211_IFTYPE_AP)
		enable = false;

	v = scm80211_mac_readl(hw, MAC_REG_VIF_RX_FILTER0(vidx));
	v &= ~VIF_RX_FILTER0_BSS_EN;
	v &= ~VIF_RX_FILTER0_BFE;

	if (enable) {
		if (info->bssid && is_valid_ether_addr(info->bssid))
			v |= FIELD_PREP(VIF_RX_FILTER0_BSS_EN, 0x3);
		v |= FIELD_PREP(VIF_RX_FILTER0_BFE, 1);
	}

	scm80211_mac_writel(hw, v, MAC_REG_VIF_RX_FILTER0(vidx));
	vifp->bssid_filter = enable;
}

/**
 * scm80211_set_addr() - set interface address
 * @hw: hardware
 * @vif: interface
 * @addr: MAC address to be bound to @vif
 *
 * This function sets the MAC address of @vif and additionally enable RA
 * filter. Once enabled, RA filter passes individually addressed frames
 * whose RA matches @vif's MAC address.
 *
 * when p2p_dev is enabled and created, the 2nd MAC address is used,
 * p2p_dev will share use the VIF0
 *
 * If @addr is equal to 00:00:00:00:00:00, this function turns off the RA
 * filter which discards all individually addressed frames.
 */
void scm80211_set_addr(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u8 addr[ETH_ALEN])
{
	int vidx = scm80211_vdev_id(vif);
	u32 mac[2], v;
	bool ap = ieee80211_iface_is_ap(vif->type);
	u32 mac_filter = BIT(0);

	mac[0] = get_unaligned_le32(&addr[0]);
	mac[1] = get_unaligned_le16(&addr[4]);
	if (vif->type != NL80211_IFTYPE_P2P_DEVICE) {
		scm80211_mac_writel(hw, mac[0], MAC_REG_VIF_MAC0_L(vidx));
		scm80211_mac_writel(hw, mac[1], MAC_REG_VIF_MAC0_H(vidx));
	} else {
		scm80211_mac_writel(hw, mac[0], MAC_REG_VIF_MAC1_L(vidx));
		scm80211_mac_writel(hw, mac[1], MAC_REG_VIF_MAC1_H(vidx));
		mac_filter |= BIT(1);
	}

	v = scm80211_mac_readl(hw, MAC_REG_VIF_RX_FILTER0(vidx));
	if (vif->type != NL80211_IFTYPE_P2P_DEVICE &&
	    (hw->wiphy->interface_modes & BIT(NL80211_IFTYPE_P2P_DEVICE))) {
		v &= ~FIELD_PREP(VIF_RX_FILTER0_MAC_EN, BIT(0));
	} else {
		v &= ~VIF_RX_FILTER0_MAC_EN;
	}
	v &= ~VIF_RX_FILTER0_AP_MODE;

	/* Enable RA address filter */
	if (!is_zero_ether_addr(addr)) {
		v |= FIELD_PREP(VIF_RX_FILTER0_MAC_EN, mac_filter);
		v |= FIELD_PREP(VIF_RX_FILTER0_AP_MODE, ap);
	}
	scm80211_mac_writel(hw, v, MAC_REG_VIF_RX_FILTER0(vidx));
}

static void scm80211_enable_vif(struct ieee80211_hw *hw, struct ieee80211_vif *vif, bool enable)
{
	int vidx = scm80211_vdev_id(vif);
	u32 v;

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vidx));
	v &= ~VIF_CFG0_EN;
	if (enable)
		v |= FIELD_PREP(VIF_CFG0_EN, 1);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG0(vidx));
}

/**
 * scm80211_set_bssid() - set bssid
 *
 * BSSID0: associated BSSID
 * BSSID1: transmitted BSSID
 */
static void scm80211_set_bssid(struct ieee80211_hw *hw,
			       struct ieee80211_vif *vif, const u8 *bssid, const u8 *tran_bssid)
{
	int vidx = scm80211_vdev_id(vif);
	u32 bssid_l, bssid_h;

	scm80211_trace(hw, " bssid: %pM, transmitted bssid: %pM\n", bssid, tran_bssid);

	if (is_valid_ether_addr(bssid)) {
		/* BSSID0: nontransmitted BSSID */
		bssid_l = get_unaligned_le32(bssid);
		bssid_h = 0x0 | get_unaligned_le16(bssid + 4);
		scm80211_mac_writel(hw, bssid_l, MAC_REG_VIF_BSS0_L(vidx));
		scm80211_mac_writel(hw, bssid_h, MAC_REG_VIF_BSS0_H(vidx));
		/* Set BSSID1 as the same value as bssid0 first */
		scm80211_mac_writel(hw, bssid_l, MAC_REG_VIF_BSS1_L(vidx));
		scm80211_mac_writel(hw, bssid_h, MAC_REG_VIF_BSS1_H(vidx));
	} else {
		/* Clear BSSID0 */
		scm80211_mac_writel(hw, 0, MAC_REG_VIF_BSS0_L(vidx));
		scm80211_mac_writel(hw, 0, MAC_REG_VIF_BSS0_H(vidx));
		/* Clear BSSID1 */
		scm80211_mac_writel(hw, 0, MAC_REG_VIF_BSS1_L(vidx));
		scm80211_mac_writel(hw, 0, MAC_REG_VIF_BSS1_H(vidx));

		scm80211_config_bssid_filter(hw, vif, false);
		return;
	}

	if (is_valid_ether_addr(tran_bssid)) {
		/* BSSID1: transmitted BSSID */
		bssid_l = get_unaligned_le32(tran_bssid);
		bssid_h = 0x0 | get_unaligned_le16(tran_bssid + 4);
		scm80211_mac_writel(hw, bssid_l, MAC_REG_VIF_BSS1_L(vidx));
		scm80211_mac_writel(hw, bssid_h, MAC_REG_VIF_BSS1_H(vidx));
	}

	if (vif->type == NL80211_IFTYPE_AP)
		scm80211_config_bssid_filter(hw, vif, false);
	else
		scm80211_config_bssid_filter(hw, vif, true);
}

/**
 * scm80211_set_bssid_mask()
 * @bssid_indicator 2 power bssid_indicator us the maximum of BSSIDs in
 * the multiple BSSID set.
 */
static void scm80211_set_bssid_mask(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif, u8 bssid_indicator)
{
	int vidx = scm80211_vdev_id(vif);
	u8 mask[ETH_ALEN + 2];
	u64 mask_u64;

	eth_broadcast_addr(mask);

	mask_u64 = ether_addr_to_u64(mask);

	if (bssid_indicator > 0)
		mask_u64 &= ~GENMASK_ULL(bssid_indicator - 1, 0);

	u64_to_ether_addr(mask_u64, mask);

	scm80211_trace(hw, "mbssid mask: %pM\n", mask);

	scm80211_mac_writel(hw, get_unaligned_le32(mask), MAC_REG_VIF_MBSSID_MASK_L32(vidx));
	scm80211_mac_writel(hw, get_unaligned_le16(mask + 4), MAC_REG_VIF_MBSSID_MASK_H16(vidx));
}

/*
 * Before BSS basic rate is known, set the basic rate registers
 * with PHY mandatory rates.
 */
void scm80211_set_mandatory_hwrates(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif, enum nl80211_band band, bool erp)
{
	struct ieee80211_supported_band *sband = hw->wiphy->bands[band];
	u32 rates = 0, flags, v;
	int i;
	int vidx = scm80211_vdev_id(vif);

	if (band == NL80211_BAND_2GHZ) {
		flags = IEEE80211_RATE_MANDATORY_B;
		flags |= erp ? IEEE80211_RATE_MANDATORY_G : 0;
	} else if (band == NL80211_BAND_5GHZ) {
		flags = IEEE80211_RATE_MANDATORY_A;
	} else {
		return;
	}

	for (i = 0; i < sband->n_bitrates; i++) {
		if (sband->bitrates[i].flags & flags)
			rates |= BIT(sband->bitrates[i].hw_value);
	}

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG1(vidx));
	/* TODO: To clear the value, we need to consider P2P */
	/*v = clrfield(v, MAC_REG_VIF_CFG1_BASIC_RATE_BITMAP); */
	v |= FIELD_PREP(VIF_CFG1_BASIC_RATE_BITMAP, rates);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG1(vidx));
}

/**
 * scm80211_set_basic_rates() - set BSS basic rate set
 *
 * This function sets basic rates from @info->basic_rates. It assumes
 * (1) @info->assoc is true and (2) @info->chandef.chan is valid.
 *
 */
void scm80211_set_basic_rates(struct ieee80211_hw *hw,
			      struct ieee80211_vif *vif, struct ieee80211_bss_conf *info)
{
	struct ieee80211_supported_band *sband;
	u32 bitmap = 0;
	unsigned long brates = info->basic_rates;
	int i = 0, vidx = scm80211_vdev_id(vif);
	u32 v;

	if (info->chandef.chan)
		sband = hw->wiphy->bands[info->chandef.chan->band];
	else
		sband = hw->wiphy->bands[NL80211_BAND_2GHZ];

	for_each_set_bit(i, &brates, sband->n_bitrates)
		bitmap |= BIT(sband->bitrates[i].hw_value);
	scm80211_trace(hw, " basic rates: 0x%llx (hw=%08x)\n",
		       (unsigned long long)info->basic_rates, bitmap);

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG1(vidx));
	v &= ~VIF_CFG1_BASIC_RATE_BITMAP;
	v |= FIELD_PREP(VIF_CFG1_BASIC_RATE_BITMAP, bitmap);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG1(vidx));
}

/**
 * scm80211_set_p2p_basic_rate()
 * for p2p 11b rate will not be used, 6M/12M/24M basic rate should be enabled
 */
void scm80211_set_p2p_basic_rate(struct ieee80211_hw *hw, int vdev)
{
	u32 v;

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG1(vdev));
	v |= FIELD_PREP(VIF_CFG1_BASIC_RATE_BITMAP, BIT(4) | BIT(5) | BIT(8));
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG1(vdev));
}

static void scm80211_set_erp(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			     struct ieee80211_bss_conf *info)
{
	int vidx = scm80211_vdev_id(vif);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif_priv *iface = ieee80211_to_priv(vif);
	u32 v;

	iface->slot = info->use_short_slot ? 9 : 20;

	scm80211_trace(hw, " erp: prot=%d, short_preamble=%d, short slot=%d\n",
		       info->use_cts_prot, info->use_short_preamble, info->use_short_slot);

	if (priv->coverage_class > 0 && priv->coverage_class <= 31)
		iface->slot += 3 * priv->coverage_class;

	/* for mac version higher than 0x16 same solt time will be used for vif0/1 */
	if (version(scm80211_mac_version(hw)) >= 0x16) {
		v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vidx));
		v &= ~(VIF_CFG0_BP | VIF_CFG0_PT);
		v |= (FIELD_PREP(VIF_CFG0_BP, info->use_cts_prot) |
		      FIELD_PREP(VIF_CFG0_PT, info->use_short_preamble));

		scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG0(vidx));

		v = scm80211_mac_readl(hw, MAC_REG_DEV_CFG) & ~DEV_CFG_SLOT_TIME;
		v |= FIELD_PREP(DEV_CFG_SLOT_TIME, iface->slot);
		scm80211_mac_writel(hw, v, MAC_REG_DEV_CFG);
	} else {
		v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vidx));
		v &= ~(VIF_CFG0_BP | VIF_CFG0_PT | VIF_CFG0_SLOT_TIME);
		v |= (FIELD_PREP(VIF_CFG0_BP, info->use_cts_prot) |
		      FIELD_PREP(VIF_CFG0_PT, info->use_short_preamble) |
		      FIELD_PREP(VIF_CFG0_SLOT_TIME, iface->slot));
		scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG0(vidx));
	}
}

/**
 * scm80211_vif_chan()
 * @def: copy vif->chanctx
 */
int scm80211_vif_chan(struct ieee80211_vif *vif, struct cfg80211_chan_def *def)
{
	struct ieee80211_chanctx_conf *conf;

	rcu_read_lock();
	conf = rcu_dereference_chanctx_conf(vif);
	if (!conf) {
		rcu_read_unlock();
		return -ENOENT;
	}
	*def = conf->def;
	rcu_read_unlock();
	return 0;
}

const char *const chanwidth_str[] = {
	[NL80211_CHAN_WIDTH_20_NOHT] = "non-ht",
	[NL80211_CHAN_WIDTH_20] = "ht20",
	[NL80211_CHAN_WIDTH_40] = "ht40",
	[NL80211_CHAN_WIDTH_80] = "vht80",
	[NL80211_CHAN_WIDTH_80P80] = "vht80p80",
	[NL80211_CHAN_WIDTH_160] = "vht160",
};

static void scm80211_set_basic_ht_mcs_set(struct ieee80211_hw *hw,
					  struct ieee80211_vif *vif,
					  struct ieee80211_bss_conf *info)
{
	struct cfg80211_bss *bss;
	struct cfg80211_chan_def def;
	int vidx = scm80211_vdev_id(vif);
	u32 v, ht_basic_mcs = 0xff;	/* mandatory 20MHz NSS=1 ESS=1 */

	if (scm80211_vif_chan(vif, &def)) {
		scm80211_err(hw, "failed to get vif channel\n");
		goto out;
	}

	bss = cfg80211_get_bss(hw->wiphy, def.chan, info->bssid, NULL, 0,
			       IEEE80211_BSS_TYPE_ANY, IEEE80211_PRIVACY_ANY);
	if (bss) {
		const struct ieee80211_ht_operation *ht_oper = NULL;
		const u8 *ie;

		/*
		 * SCM2020 supports up to MCS0-15 and MCS32, of which
		 * MCS32 is not used for control response frame.
		 */
		ie = ieee80211_bss_get_ie(bss, WLAN_EID_HT_OPERATION);
		if (ie && ie[1] >= sizeof(*ht_oper)) {
			ht_oper = (void *)(ie + 2);
			v = get_unaligned_le16(ht_oper->basic_set);
			if (v)
				ht_basic_mcs = v;
		}
	}

	scm80211_trace(hw, " HT basic MCS: %08x\n", ht_basic_mcs);
out:
	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG1(vidx));
	v &= ~VIF_CFG1_BASIC_HT_MCS_SET;
	v |= FIELD_PREP(VIF_CFG1_BASIC_HT_MCS_SET, ht_basic_mcs);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG1(vidx));
}

static void scm80211_set_ht_mmss(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				 struct ieee80211_sta *ap, struct ieee80211_bss_conf *info)
{
	int vdev = scm80211_vdev_id(vif);
	u32 v;

	/* MMSS */
	scm80211_trace(hw, " HT MMSS: %d\n", get_ieee80211_ht_cap(ap).ampdu_density);

	v = scm80211_mac_readl(hw, MAC_REG_VIF_MIN_MPDU_SPACING(vdev));
	v &= ~VIF_MIN_MPDU_SPACING_MMSS;
	v |= FIELD_PREP(VIF_MIN_MPDU_SPACING_MMSS, get_ieee80211_ht_cap(ap).ampdu_density);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_MIN_MPDU_SPACING(vdev));
}

static void scm80211_set_aid(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u16 aid)
{
	int vdev = scm80211_vdev_id(vif);

	scm80211_mac_writel(hw, aid, MAC_REG_VIF_AID(vdev));
}

/**
 * scm80211_conf_hw_power()
 * init tx power which will be used for mac HW to send ctrl frame
 * RTS/CTS/ACK/BA/CBF/CF-End
 */
static void scm80211_conf_hw_power(struct ieee80211_hw *hw, struct ieee80211_vif *vif, s8 power)
{
	u32 v;
	u8 vidx = scm80211_vdev_id(vif);

	v = scm80211_mac_readl(hw, MAC_REG_VIF_TX_PWR(vidx));
	v &= ~VIF_TX_PWR_HW_PWR;
	v = FIELD_PREP(VIF_TX_PWR_HW_PWR, power);

	scm80211_mac_writel(hw, v, MAC_REG_VIF_TX_PWR(vidx));
}

static void scm80211_init_bfm_reg(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	u32 v;
	u8 vidx = scm80211_vdev_id(vif);

	/*
	 * NB: use OFDM base rate (6 Mbps) for CBF for now
	 * BFM0#vif - CBF format, MCS to HT/VHT NDPA/NDP
	 */

	v = scm80211_mac_readl(hw, MAC_REG_VIF_BFM0(vidx));

	v &= ~(VIF_BFM0_HT_FORMAT | VIF_BFM0_HT_MCS | VIF_BFM0_VHT_FORMAT | VIF_BFM0_VHT_MCS);

	v |= FIELD_PREP(VIF_BFM0_HT_FORMAT, 0);
	v |= FIELD_PREP(VIF_BFM0_HT_MCS, 4);

	v |= FIELD_PREP(VIF_BFM0_VHT_FORMAT, 0);
	v |= FIELD_PREP(VIF_BFM0_VHT_MCS, 4);

	scm80211_mac_writel(hw, v, MAC_REG_VIF_BFM0(vidx));

	/*
	 * BFM1#vif - CBF format, MCS to HE NDPA/NDP
	 */
	v = scm80211_mac_readl(hw, MAC_REG_VIF_BFM1(vidx));

	v &= ~(VIF_BFM1_HE_FORMAT | VIF_BFM1_HE_MCS);

	v |= FIELD_PREP(VIF_BFM1_HE_FORMAT, 0);
	v |= FIELD_PREP(VIF_BFM1_HE_MCS, 4);

	scm80211_mac_writel(hw, v, MAC_REG_VIF_BFM1(vidx));
}

static void scm80211_set_vht_partial_aid(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif,
					 enum nl80211_iftype iftype, u16 aid)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct phy_vector *pv = &vp->pv;
	unsigned long changed = 0;
	u32 partial_aid;

	/*
	 * PHY_LISTEN_TO_GID00
	 * PHY_PARTIAL_AID_LIST_GID00
	 * PHY_LISTEN_TO_GID63
	 * PHY_PARTIAL_AID_LIST_GID63
	 * PHY_VHT_SU_RX_FILTER_ON
	 */
	switch (vif->type) {
	case NL80211_IFTYPE_STATION:
		partial_aid = ieee80211_vht_partial_aid(aid, vif->bss_conf.bssid, 63);
		pv->LISTEN_TO_GID63 = true;
		pv->PARTIAL_AID_LIST_GID63 = partial_aid;
		pv->VHT_SU_RX_FILTER_EN = false;	/* FIMXE: late enable */
		changed |= BIT(PHY_LISTEN_TO_GID63) | BIT(PHY_PARTIAL_AID_LIST_GID63);
		changed |= BIT(PHY_VHT_SU_RX_FILTER_EN);
		break;
	case NL80211_IFTYPE_AP:
		partial_aid = ieee80211_vht_partial_aid(aid, vif->addr, 0);
		pv->LISTEN_TO_GID00 = true;
		pv->PARTIAL_AID_LIST_GID00 = partial_aid;
		pv->VHT_SU_RX_FILTER_EN = false;	/* FIXME: enable after seeing it */
		changed |= BIT(PHY_LISTEN_TO_GID00) | BIT(PHY_PARTIAL_AID_LIST_GID00);
		changed |= BIT(PHY_VHT_SU_RX_FILTER_EN);
		break;
	case NL80211_IFTYPE_MONITOR:
		pv->VHT_SU_RX_FILTER_EN = false;
		changed |= BIT(PHY_VHT_SU_RX_FILTER_EN);
		break;
	default:
		break;
	}
	pv->present |= changed;
	scm80211_vif_update_chanctx(hw, vif, changed);
}

static void scm80211_clr_vht_partial_aid(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif, enum nl80211_iftype iftype)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct phy_vector *pv = &vp->pv;
	unsigned long changes;

	pv->VHT_SU_RX_FILTER_EN = false;
	changes = BIT(PHY_VHT_SU_RX_FILTER_EN);
	pv->present |= changes;
	scm80211_vif_update_chanctx(hw, vif, changes);
}

static void scm80211_set_vht_mu_group(struct ieee80211_hw *hw,
				      struct ieee80211_vif *vif,
				      struct ieee80211_mu_group_data *mu_group)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct phy_vector *pv = &vp->pv;
	unsigned long changes = BIT(PHY_GROUP_ID_MANAGEMENT);

	memcpy(&pv->GROUP_ID_MANAGEMENT, mu_group, sizeof(*mu_group));
	pv->present |= changes;
	scm80211_vif_update_chanctx(hw, vif, changes);
}

static void scm80211_clr_vht_mu_group(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct phy_vector *pv = &vp->pv;
	unsigned long changes = BIT(PHY_GROUP_ID_MANAGEMENT);

	memset(&pv->GROUP_ID_MANAGEMENT, 0, sizeof(pv->GROUP_ID_MANAGEMENT));
	pv->present |= changes;
	scm80211_vif_update_chanctx(hw, vif, changes);
}

static int scm80211_set_he_uora_cfg_ch_info(struct ieee80211_hw *hw,
					    struct ieee80211_vif *vif,
					    struct cfg80211_chan_def *chdef,
					    struct cfg80211_chan_def *apchdef)
{
	u8 vidx = scm80211_vdev_id(vif);
	u8 np20, width, chan_en = 0;
	u8 bwmode = 1;
	u8 opmode = 0;
	u32 f0;
	u32 v;

	switch (apchdef->width) {
	case NL80211_CHAN_WIDTH_160:
		bwmode = 0;
		width = 160;
		break;
	case NL80211_CHAN_WIDTH_80P80:
	case NL80211_CHAN_WIDTH_80:
		width = 80;
		break;
	case NL80211_CHAN_WIDTH_40:
		width = 40;
		break;
	case NL80211_CHAN_WIDTH_20:
		width = 20;
		break;
	default:
		scm80211_warn(hw, "AP channel info invalid!\n");
		goto exit;
	}

	f0 = apchdef->center_freq1 - width / 2 + 10;
	np20 = (apchdef->chan->center_freq - f0) / 20;
	if (apchdef->width == NL80211_CHAN_WIDTH_80P80 &&
	    apchdef->center_freq1 > apchdef->center_freq2)
		np20 += 4;

	switch (chdef->width) {
	case NL80211_CHAN_WIDTH_80:
		opmode = 2;
		chan_en = 0x0F;
		break;
	case NL80211_CHAN_WIDTH_40:
		opmode = 1;
		chan_en = BIT(0) | BIT(1);
		if ((np20 % 4) >= 2)
			chan_en <<= 2;
		break;
	case NL80211_CHAN_WIDTH_20:
		opmode = 0;
		chan_en = BIT(np20 % 4);
		break;
	default:
		scm80211_warn(hw, "STA channel info invalid!\n");
		goto exit;
	}

	scm80211_dbg(hw, "uora info set: channel_en %d , primary chanel index %d\n", chan_en, np20);

	v = scm80211_mac_readl(hw, MAC_REG_VIF_UORA_CFG1(vidx));
	v &= ~(VIF_UORA_CFG1_80P80_MODE | VIF_UORA_CFG1_CHANNEL_EN | VIF_UORA_CFG1_PRI_CH);
	v |= FIELD_PREP(VIF_UORA_CFG1_CHANNEL_EN, chan_en);
	v |= FIELD_PREP(VIF_UORA_CFG1_80P80_MODE, bwmode);
	v |= FIELD_PREP(VIF_UORA_CFG1_PRI_CH, np20);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_UORA_CFG1(vidx));

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG_INFO(vidx));
	v &= ~VIF_CFG_INFO_OP_MODE;
	v |= FIELD_PREP(VIF_CFG_INFO_OP_MODE, opmode);
	v |= FIELD_PREP(VIF_CFG_INFO_COLOR_INBSS, 1);
	v |= FIELD_PREP(VIF_CFG_INFO_QUEUE_HTC_APPEND_EN, 1);
	v |= FIELD_PREP(VIF_CFG_INFO_QUEUE_HTC_CONVERT_EN, 1);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG_INFO(vidx));

	return 0;
exit:
	return -1;
}

static void scm80211_set_he_uora_cfg(struct ieee80211_hw *hw,
				     struct ieee80211_vif *vif,
				     struct ieee80211_bss_conf *conf, struct ieee80211_sta *ap)
{
	struct cfg80211_bss *bss;
	struct ieee80211_ht_operation *ht_oper = NULL;
	struct ieee80211_vht_operation *vht_oper = NULL;
	struct cfg80211_chan_def def;
	struct cfg80211_chan_def bss_def = {
		.chan = conf->chandef.chan,
		.width = NL80211_CHAN_WIDTH_20,
		.center_freq1 = conf->chandef.chan->center_freq,
		.center_freq2 = 0,
	};
	u8 ocw_min, ocw_max;
	u8 vidx = scm80211_vdev_id(vif);
	u32 v;

	if (scm80211_vif_chan(vif, &def)) {
		scm80211_err(hw, "failed to get vif channel\n");
		return;
	}

	bss = cfg80211_get_bss(hw->wiphy, def.chan, conf->bssid, NULL, 0,
			       IEEE80211_BSS_TYPE_ANY, IEEE80211_PRIVACY_ANY);

	if (bss) {
		const struct cfg80211_bss_ies *ies = NULL;
		const u8 *elem;

		rcu_read_lock();

		ies = rcu_dereference(bss->ies);

		if (!ies) {
			scm80211_warn(hw, "Beacon ies invailded!\n");
			return;
		}

		elem = cfg80211_find_ie(WLAN_EID_HT_OPERATION, ies->data, ies->len);
		if (elem && elem[1] >= sizeof(*ht_oper))
			ht_oper = (void *)(elem + 2);

		elem = cfg80211_find_ie(WLAN_EID_VHT_OPERATION, ies->data, ies->len);
		if (elem && elem[1] >= sizeof(*vht_oper))
			vht_oper = (void *)(elem + 2);

		rcu_read_unlock();
	}

	bss_def.chan = conf->chandef.chan;
	bss_def.width = NL80211_CHAN_WIDTH_20;
	bss_def.center_freq1 = conf->chandef.chan->center_freq;
	bss_def.center_freq2 = 0;

	if (ieee80211_chandef_create(&bss_def, ap, ht_oper, vht_oper)) {
		scm80211_err(hw, "Get AP channel info fail!\n");
		return;
	}

	if (!cfg80211_chandef_compatible(&bss_def, &def)) {
		scm80211_err(hw, "Channel does not compatible!\n");
		return;
	}

	scm80211_trace(hw, "uora operate primary freq %d MHz(%s), center1 %d MHz\n",
		       def.chan->center_freq, chanwidth_str[def.width], def.center_freq1);

	scm80211_trace(hw,
		       "uora bss primary freq %d MHz(%s), center1 %d MHz, center2 %d MHz\n",
		       bss_def.chan->center_freq,
		       chanwidth_str[bss_def.width], bss_def.center_freq1, bss_def.center_freq2);

	if (def.width > bss_def.width) {
		scm80211_warn(hw, "Channel config does not match!\n");
		return;
	}

	if (scm80211_set_he_uora_cfg_ch_info(hw, vif, &def, &bss_def))
		return;

	if (conf->uora_exists) {
		scm80211_dbg(hw, "uora_ocw_range %d,\n", conf->uora_ocw_range);

		ocw_min = conf->uora_ocw_range & 0x07;
		ocw_max = (conf->uora_ocw_range >> 3) & 0x07;

		if (ocw_min && ocw_max) {
			/* config MAC_REG_VIF_UORA_CFG */
			v = scm80211_mac_readl(hw, MAC_REG_VIF_UORA_CFG(vidx));
			v &= ~(VIF_UORA_CFG_CW_MAX | VIF_UORA_CFG_CW_MIN);
			v |= FIELD_PREP(VIF_UORA_CFG_CW_MAX, ocw_max);
			v |= FIELD_PREP(VIF_UORA_CFG_CW_MIN, ocw_min);
			v |= FIELD_PREP(VIF_UORA_CFG_CW_INIT, 1);
			scm80211_mac_writel(hw, v, MAC_REG_VIF_UORA_CFG(vidx));
		}
	}
}

/**
 * scm80211_set_he_phycfg()
 * FIXME:
 * - For AP condition
 * - For concurrent/muti-channel mode how/when to operate FW_SEL_VIF reg
 *   and when to turn on the filter
 **/
static void scm80211_set_he_phycfg(struct ieee80211_hw *hw,
				   struct ieee80211_vif *vif,
				   struct ieee80211_bss_conf *conf, const u8 color)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct phy_vector *pv = &vp->pv;	/* Add this to ieee80211_vif_priv */
	unsigned long present;

	present = BIT(PHY_BSS_COLOR_FILTER_EN) | BIT(PHY_HE_DIR);

	if (!is_he_bss_color_enabled(vif)) {
		pv->BSS_COLOR_FILTER_EN = false;
		present &= ~BIT(PHY_BSS_COLOR);
	} else {
		pv->BSS_COLOR_FILTER_EN = true;
		pv->BSS_COLOR = color;

		present |= BIT(PHY_BSS_COLOR);
	}

	switch (vif->type) {	/* Assuming this is not iftype_p2p */
	case NL80211_IFTYPE_STATION:
		/*
		 * STA_ID_GROUP and STA_ID_INDIVIDUAL: set
		 * HE_DIR_EN: DL
		 */
		pv->STA_ID_GROUP = conf->nontransmitted ? conf->bssid_index : 0;
		pv->STA_ID_INDIVIDUAL = conf->aid;
		pv->HE_DIR_EN = PHY_HE_DIR_DL;
		present |= BIT(PHY_STA_ID_GROUP) | BIT(PHY_STA_ID_INDIVIDUAL);

		break;
	case NL80211_IFTYPE_AP:
		/*
		 * BSS color: enabled
		 * STA_ID: not set - PHY does not do STA_ID filtering if UL_FLAG=1
		 * HE_DIR_EN: UL
		 */
		pv->HE_DIR_EN |= PHY_HE_DIR_UL;
		break;
	case NL80211_IFTYPE_MONITOR:
		/*
		 * BSS_COLOR filter disabled
		 * STA_ID: not set (PHY cannot process HE DL MU)
		 * HE_DIR_EN: DL + UL
		 * At least monitor can receive all HE SU and HE UL M (single user)
		 */
		pv->BSS_COLOR_FILTER_EN = false;
		pv->HE_DIR_EN = PHY_HE_DIR_DL_UL;
		break;
	default:
		break;
	}

	pv->present |= present;
	scm80211_dump_phy_vector(hw, pv, pv->present);
	scm80211_vif_update_chanctx(hw, vif, present);
}

static void scm80211_clr_he_phycfg(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct phy_vector *pv = &vp->pv;
	unsigned long changes = BIT(PHY_BSS_COLOR_FILTER_EN);
	/*
	 * BSS color: disable
	 * STA_ID_GROUP and STA_ID_INDIVIDUAL: unset
	 * HE_DIR_EN: DL
	 */
	pv->BSS_COLOR_FILTER_EN = false;
	pv->present |= changes;
	scm80211_vif_update_chanctx(hw, vif, changes);
}

static void scm80211_set_he_sr(struct ieee80211_hw *hw,
			       struct ieee80211_vif *vif, struct ieee80211_bss_conf *conf)
{
	struct cfg80211_bss *bss;
	struct cfg80211_chan_def def;

	if (scm80211_vif_chan(vif, &def)) {
		scm80211_err(hw, "failed to get vif channel\n");
		return;
	}

	bss = cfg80211_get_bss(hw->wiphy, def.chan, conf->bssid, NULL, 0,
			       IEEE80211_BSS_TYPE_ANY, IEEE80211_PRIVACY_ANY);
	if (bss) {
		const struct cfg80211_bss_ies *ies;
		const struct element *ie = NULL;

		rcu_read_lock();
		ies = rcu_dereference(bss->ies);
		rcu_read_unlock();
		ie = cfg80211_find_ext_elem(WLAN_EID_EXT_HE_SPR, ies->data, ies->len);
		if (ie && ie->datalen >= sizeof(struct ieee80211_he_spr))
			ieee80211_parse_he_sr(hw, vif, ie->data + 1);
	}
}

static void scm80211_set_he_info(struct ieee80211_hw *hw,
				 struct ieee80211_vif *vif,
				 struct ieee80211_sta *ap, struct ieee80211_bss_conf *conf)
{
	struct ieee80211_sta_he_cap he_cap_tmp = get_ieee80211_he_cap(ap);
	struct ieee80211_sta_he_cap *he_cap = &he_cap_tmp;
	struct ieee80211_vif_priv *vifp = (void *)vif->drv_priv;
	u8 vidx = scm80211_vdev_id(vif);
	int i, param;
	__le32 v = 0;
	unsigned long flag;

	if (!conf->he_support)
		return;

	scm80211_trace(hw, "%s(%s:'%s')\n", __func__,
		       ieee80211_vif_name(vif), ieee80211_iftype_str[vif->type]);

	v = scm80211_mac_readl(hw, MAC_REG_VIF_MIN_MPDU_SPACING(vidx));

	/*
	 * Default PE Duration
	 *
	 * NB:
	 * The IEEE80211_HE_OPERATION_DFLT_PE_DURATION_MASK definition in ieee80211.h
	 * is wrong and @conf->htc_trig_based_pkt_ext calculated base on it is also
	 * wrong. We make the correction here.
	 */
	param = le32_get_bits(ieee80211_he_oper_params(conf), GENMASK(2, 0));
	conf->htc_trig_based_pkt_ext = param;

	scm80211_trace(hw, " HE default PE duration: %d\n", param);

	v &= ~VIF_MIN_MPDU_SPACING_DEFAULT_PE_DURATION;
	v |= FIELD_PREP(VIF_MIN_MPDU_SPACING_DEFAULT_PE_DURATION, param);

	scm80211_trace(hw, " HE BSS color element: %08x (color=%d partial=%d enabled=%d)\n",
		       le32_to_cpu(ieee80211_he_oper_params(conf)),
		       ieee80211_he_bss_color(conf),
		       ieee80211_he_partial_bss_color(conf), is_he_bss_color_enabled(vif));

	v &= ~(VIF_MIN_MPDU_SPACING_COLOR_DIS |
	       VIF_MIN_MPDU_SPACING_COLOR_EN |
	       VIF_MIN_MPDU_SPACING_PARTIAL_BSS_COLOR_EN |
	       VIF_MIN_MPDU_SPACING_BSS_COLOR | VIF_MIN_MPDU_SPACING_HETB_DATA_DIS);

	/* BSS color */
	v |= FIELD_PREP(VIF_MIN_MPDU_SPACING_BSS_COLOR, ieee80211_he_bss_color(conf));

	/* BSS color disabled */
	if (!is_he_bss_color_enabled(vif)) {
		v |= VIF_MIN_MPDU_SPACING_COLOR_DIS;
		v &= ~VIF_MIN_MPDU_SPACING_COLOR_EN;
	} else {
		v &= ~VIF_MIN_MPDU_SPACING_COLOR_DIS;
		v |= VIF_MIN_MPDU_SPACING_COLOR_EN;
	}

	/* BSS partial color enabled */
	if (ieee80211_he_partial_bss_color(conf))
		v |= VIF_MIN_MPDU_SPACING_PARTIAL_BSS_COLOR_EN;

	/* Trigger */
	v |= VIF_MIN_MPDU_SPACING_TRIG_EN;

	scm80211_trace(hw, " HE TRS %d BSR %d UORA %d\n",
		       !!(he_cap->he_cap_elem.mac_cap_info[2] & IEEE80211_HE_MAC_CAP2_TRS),
		       !!(he_cap->he_cap_elem.mac_cap_info[2] & IEEE80211_HE_MAC_CAP2_BSR),
		       conf->uora_exists);

	/* BSR */
	if (get_ieee80211_he_cap(ap).he_cap_elem.mac_cap_info[2] & IEEE80211_HE_MAC_CAP2_BSR) {
		v |= VIF_MIN_MPDU_SPACING_BSRP_BSR_EN;
		v &= ~VIF_MIN_MPDU_SPACING_BSRP_MORE_QOSNULL_EN;	/* ? */
	} else {
		v &= ~VIF_MIN_MPDU_SPACING_BSRP_BSR_EN;
		v |= VIF_MIN_MPDU_SPACING_BSRP_MORE_QOSNULL_EN;	/* ? */
	}

	/* UORA */
	v &= ~VIF_MIN_MPDU_SPACING_TRIG_UNASSO_UORA_EN;	/* already associated */
	if (conf->uora_exists)
		v |= VIF_MIN_MPDU_SPACING_TRIG_ASSO_UORA_EN;
	else
		v &= ~VIF_MIN_MPDU_SPACING_TRIG_ASSO_UORA_EN;

	/* OMI? */
	// ul_data_dis = !!(ap->he_cap.he_cap_elem.mac_cap_info[5] & BIT(4));

	v &= ~VIF_MIN_MPDU_SPACING_BASIC_NAV_EN;	/* set only for AP */

	scm80211_mac_writel(hw, v, MAC_REG_VIF_MIN_MPDU_SPACING(vidx));

	spin_lock_irqsave(&vifp->lock, flag);
	vifp->reg_mmss = v;
	spin_unlock_irqrestore(&vifp->lock, flag);

	for (i = 0; i < SCM80211_NUM_TIDS; ++i)
		vifp->tid_queue_size[i] = 0;

	scm80211_set_he_uora_cfg(hw, vif, conf, ap);

	scm80211_set_he_phycfg(hw, vif, conf, ieee80211_he_bss_color(conf));

	/* SR */
	scm80211_set_he_sr(hw, vif, conf);

#ifdef SCM80211_WFA_TEST
	/*
	 * this is for debug purpose which addes seq numfer for ACT-NO-ACk frame
	 * related case is HE_5.57.1_24G
	 */
	v = scm80211_mac_readl(hw, MAC_REG_DEV_OPT2);
	v |= (FIELD_PREP(DEV_OPT2_CBF_SEQ_EN, 1));
	scm80211_mac_writel(hw, v, MAC_REG_DEV_OPT2);
#endif
}

static void scm80211_clr_he_info(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u8 vidx = scm80211_vdev_id(vif);
	u32 v;
	unsigned long flag;

	scm80211_trace(hw, "%s (iftype %s)\n", __func__, ieee80211_iftype_str[vif->type]);

	v = VIF_MIN_MPDU_SPACING_TRIG_EN | VIF_MIN_MPDU_SPACING_TRIG_UNASSO_UORA_EN |
	    VIF_MIN_MPDU_SPACING_TRS_EN | VIF_MIN_MPDU_SPACING_BSRP_MORE_QOSNULL_EN;

	spin_lock_irqsave(&vifp->lock, flag);
	vifp->reg_mmss = v;
	spin_unlock_irqrestore(&vifp->lock, flag);

	scm80211_mac_writel(hw, v, MAC_REG_VIF_MIN_MPDU_SPACING(vidx));

	/*
	 * FIXME
	 * maybe we need consider more for somehow chip uora ongoing
	 * flush the hwq first?
	 */
	v = scm80211_mac_readl(hw, MAC_REG_VIF_UORA_CFG(vidx));
	v &= ~VIF_UORA_CFG_CW_INIT;
	scm80211_mac_writel(hw, v, MAC_REG_VIF_UORA_CFG(vidx));

	v = scm80211_mac_readl(hw, MAC_REG_VIF_UORA_CFG1(vidx));
	v &= ~(VIF_UORA_CFG1_CHANNEL_EN | VIF_UORA_CFG1_PRI_CH);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_UORA_CFG1(vidx));

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG_INFO(vidx));
	v &= ~(VIF_CFG_INFO_OP_MODE |
	       VIF_CFG_INFO_COLOR_INBSS |
	       VIF_CFG_INFO_QUEUE_HTC_CONVERT_EN | VIF_CFG_INFO_QUEUE_HTC_APPEND_EN);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG_INFO(vidx));

	scm80211_clr_he_phycfg(hw, vif);

#ifdef SCM80211_WFA_TEST
	v = scm80211_mac_readl(hw, MAC_REG_DEV_OPT2);
	v &= ~(FIELD_PREP(DEV_OPT2_CBF_SEQ_EN, 1));
	scm80211_mac_writel(hw, v, MAC_REG_DEV_OPT2);
#endif
}

void scm80211_set_mac_channel(struct ieee80211_hw *hw, struct cfg80211_chan_def *chandef)
{
	enum nl80211_band band;
	u32 v;

	if (!chandef || !chandef->chan)
		return;

	hw->conf.chandef = *chandef;

	/* Set frequency band */
	band = chandef->chan->band;
	if (band != NL80211_BAND_2GHZ && band != NL80211_BAND_5GHZ) {
		scm80211_warn(hw, "invalid band %d\n", band);
		return;
	}
	v = scm80211_mac_readl(hw, MAC_REG_DEV_CFG) & ~DEV_CFG_BAND;
	v |= FIELD_PREP(DEV_CFG_BAND, ((band == NL80211_BAND_2GHZ) ? 0 : 1));
	scm80211_mac_writel(hw, v, MAC_REG_DEV_CFG);

	v = scm80211_mac_readl(hw, MAC_REG_DEV_OPT1);
	v &= ~DEV_OPT1_PHYRX_TD_DELAY_MASK;
	v |= FIELD_PREP(DEV_OPT1_PHYRX_TD_DELAY, ((band == NL80211_BAND_2GHZ) ? 0 : 2));
	scm80211_mac_writel(hw, v, MAC_REG_DEV_OPT1);
}

void scm80211_set_chanctx(struct ieee80211_hw *hw, struct scm80211_chanctx *ctx)
{
	struct ieee80211_chanctx_conf *chanctx = &ctx->ctx;
	u32 v;
	bool mac_en;

	if (!ctx || !chanctx)
		return;

	mac_en = !!(scm80211_mac_readl(hw, MAC_REG_MAC_CFG) & MAC_CFG_EN);

	if (mac_en) {
		/*Disable MAC RX */
		scm80211_mac_writel(hw, 0, MAC_REG_MAC_CFG);
	}

	/*waiting for RX MAC IDLE to avoid RX SM corruption */
	scm80211_wait_for_rx_idle(hw);

	ctx->pv.present |= BIT(PHY_CHANNEL) | BIT(PHY_ACTIVE_RXCHAIN_SET);
	phy_config(hw, ctx, ctx->pv.present);
#ifdef SCM80211_COEX
	scm80211_set_coex(hw, ctx);
#endif

	scm80211_set_mac_channel(hw, &chanctx->def);

	v = scm80211_mac_readl(hw, MAC_REG_DEV_CFG);
	v |= FIELD_PREP(DEV_CFG_N_RX, chanctx->rx_chains_static - 1);
	scm80211_mac_writel(hw, v, MAC_REG_DEV_CFG);

	if (mac_en) {
		/*Enable MAC RX */
		scm80211_mac_writel(hw, MAC_CFG_EN, MAC_REG_MAC_CFG);
	}
}

static enum cipher_suite scm80211_cipher_suite(u32 cipher_suite)
{
	switch (cipher_suite) {
	case WLAN_CIPHER_SUITE_WEP40:
		return CIPHER_SUITE_WEP40;
	case WLAN_CIPHER_SUITE_WEP104:
		return CIPHER_SUITE_WEP104;
	case WLAN_CIPHER_SUITE_TKIP:
		return CIPHER_SUITE_TKIP;
	case WLAN_CIPHER_SUITE_CCMP:
		return CIPHER_SUITE_CCMP128;
	case WLAN_CIPHER_SUITE_CCMP_256:
		return CIPHER_SUITE_CCMP256;
	case WLAN_CIPHER_SUITE_GCMP:
		return CIPHER_SUITE_GCMP128;
	case WLAN_CIPHER_SUITE_GCMP_256:
		return CIPHER_SUITE_GCMP256;
	default:
		return CIPHER_SUITE_UNKNOWN;
	}
}

static int scm80211_wait_for_key_cmd_completion(struct ieee80211_hw *hw)
{
	long long start;

	struct reg_sec_key_cmd cmd = {.val = 0, };

	start = jiffies;
	do {
		cmd.val = scm80211_mac_readl(hw, MAC_REG_SEC_KEY_CMD);
		if (cmd.kcmd == KEY_CMD_DONE)
			return 0;
		if (is_pcie_device(hw))
			udelay(100);
	} while (jiffies_to_usecs(jiffies - start) < 10000);

	return -1;
}

static int scm80211_install_key(struct ieee80211_hw *hw,
				struct ieee80211_vif *vif,
				struct ieee80211_key_conf *key,
				const u8 *addr,
				bool by_idx)
{
	struct reg_sec_key_cmd cmd = {.val = 0, };
	bool spp = false;
	u32 reg, v;
	unsigned long bitmap;
	u8 i, remain, *tk;
	u8 count = version(scm80211_mac_version(hw)) >= 0x16 ? 32 : 16;

	if (by_idx && key->hw_key_idx == count)
		by_idx = false;

	/* FIXME how to get spp */
	cmd.cipher = scm80211_cipher_suite(key->cipher);
	cmd.gtk    = !(key->flags & IEEE80211_KEY_FLAG_PAIRWISE);
	cmd.keyid  = key->keyidx;
	cmd.kcmd   = by_idx ? KEY_CMD_ADD_BY_IDX : KEY_CMD_ADD;
	cmd.idx   = by_idx ? key->hw_key_idx : 0;
	cmd.spp    = spp;
	if (cmd.cipher == CIPHER_SUITE_UNKNOWN)
		return -EOPNOTSUPP;

	if (by_idx)
		goto set_addr;

	/* Check for room */
	bitmap = scm80211_mac_readl(hw, MAC_REG_SEC_KEY_STATUS);

	if (find_first_zero_bit(&bitmap, count) == count) {
		scm80211_err(hw, "failed to add a key (key memory full)\n");
		return -ENOSPC;
	}

set_addr:
	/* Key address */
	v = get_unaligned_le32(addr);
	scm80211_mac_writel(hw, v, MAC_REG_SEC_MAC_ADDR_L32);
	v = get_unaligned_le16(addr + 4);
	scm80211_mac_writel(hw, v, MAC_REG_SEC_MAC_ADDR_H16);

	/* Key data */
	tk = key->key;
	reg = MAC_REG_SEC_TEMP_KEY0;
	remain = key->keylen % sizeof(u32);
	for (i = 0; i < key->keylen / sizeof(u32); i++) {
		v = get_unaligned_le32(tk);
		scm80211_mac_writel(hw, v, reg);
		tk += sizeof(u32);
		reg += sizeof(u32);
	}

	if (remain) {
		v = 0;
		memcpy(&v, tk, remain);
		scm80211_mac_writel(hw, v, reg);
	}

	/* Key command */
	scm80211_mac_writel(hw, cmd.val, MAC_REG_SEC_KEY_CMD);
	if (scm80211_wait_for_key_cmd_completion(hw) < 0) {
		scm80211_warn(hw, "failed to intall key (timeout)\n");
		return -ENOSPC;
	}
	if (!by_idx) {
		/* Get the key index by XORing status */
		bitmap ^= scm80211_mac_readl(hw, MAC_REG_SEC_KEY_STATUS);
		key->hw_key_idx = find_first_bit(&bitmap, count);
		WARN_ON(key->hw_key_idx == count);
	}

	scm80211_dbg(hw,
		     "successfully installed key %d <%s, %d, %pM, %s>\n",
		     key->hw_key_idx,
		     cmd.gtk ? "GTK" : "PTK", key->keyidx, addr, cipher_suite_str[cmd.cipher]);
	return 0;
}

/**
 * scm80211_remove_key() - remove a key by index
 */
static int scm80211_remove_key(struct ieee80211_hw *hw, struct ieee80211_key_conf *key)
{
	struct reg_sec_key_cmd cmd = {.val = 0, };
	unsigned long bitmap;

	cmd.kcmd = KEY_CMD_DEL_BY_IDX;
	cmd.idx = key->hw_key_idx;
	cmd.cipher = scm80211_cipher_suite(key->cipher);

	bitmap = scm80211_mac_readl(hw, MAC_REG_SEC_KEY_STATUS);
	if (!test_bit(key->hw_key_idx, &bitmap)) {
		scm80211_warn(hw,
		       "failed to removed key %d bitmap 0x%08lx <%s, %d, %s>\n",
		       key->hw_key_idx, bitmap,
		       !(key->flags & IEEE80211_KEY_FLAG_PAIRWISE) ? "GTK" : "PTK",
		       key->keyidx, cipher_suite_str[cmd.cipher]);
		return 0;
	}

	if (cmd.cipher == CIPHER_SUITE_UNKNOWN)
		return -EOPNOTSUPP;

	/* Key command */
	scm80211_mac_writel(hw, cmd.val, MAC_REG_SEC_KEY_CMD);
	if (scm80211_wait_for_key_cmd_completion(hw) < 0) {
		scm80211_warn(hw, "failed to remove key (timeout)\n");
		return -ENOSPC;
	}

	/* Check if the key is really gone */
	bitmap = scm80211_mac_readl(hw, MAC_REG_SEC_KEY_STATUS);
	WARN_ON(test_bit(key->hw_key_idx, &bitmap));

	scm80211_trace(hw,
		       "successfully removed key %d <%s, %d, %s>\n",
		       key->hw_key_idx,
		       !(key->flags & IEEE80211_KEY_FLAG_PAIRWISE) ? "GTK" : "PTK",
		       key->keyidx, cipher_suite_str[cmd.cipher]);

	return 0;
}

void scm80211_enable_tsf_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
#if PM_LOCAL_TEST
	struct ieee80211_hw_priv *priv = hw->priv;
	int vdev = scm80211_vdev_id(vif);
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);

	if (vdev)
		priv->int_en |= INTR_ENABLE_TSF1;
	else
		priv->int_en |= INTR_ENABLE_TSF0;

	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);

	spin_unlock_irqrestore(&priv->lock, flags);
#endif
}

void scm80211_disable_tsf_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
#if PM_LOCAL_TEST
	struct ieee80211_hw_priv *priv = hw->priv;
	int vdev = scm80211_vdev_id(vif);
	unsigned long flags;

	spin_lock_irqsave(&priv->lock, flags);

	if (vdev)
		priv->int_en &= ~INTR_ENABLE_TSF1;
	else
		priv->int_en &= ~INTR_ENABLE_TSF0;

	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);

	spin_unlock_irqrestore(&priv->lock, flags);
#endif
}

static inline u64 scm80211_get_tsf_from_sys(void)
{
	return ktime_to_us(ktime_get_real());
}

static __used __le64 __scm80211_get_tsf(struct ieee80211_hw_priv *data)
{
	u64 now = scm80211_get_tsf_from_sys();

	return cpu_to_le64(now + data->tsf_offset);
}

u64 scm80211_get_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	int vdev = scm80211_vdev_id(vif);
	u32 tsf[2];
	u64 v;

	if (!is_pcie_device(hw))
		return scm80211_get_tsf_from_sys();

	do {
		tsf[1] = scm80211_mac_readl(hw, MAC_REG_VIF_TSF_H(vdev));
		tsf[0] = scm80211_mac_readl(hw, MAC_REG_VIF_TSF_L(vdev));
	} while (scm80211_mac_readl(hw, MAC_REG_VIF_TSF_H(vdev)) != tsf[1]);

	v = get_unaligned_le64(tsf);
	//scm80211_trace(hw, "%s (%llu)\n", __func__, v);

	return v;
}

u32 scm80211_get_tsf32(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	u32 tsf = 0;

	tsf = scm80211_mac_readl(hw, MAC_REG_VIF_TSF_L(scm80211_vdev_id(vif)));
	if (tsf == 0)
		tsf  = (u32)ktime_to_us(ktime_get_boottime());
	return le32_to_cpu(tsf);
}

static void scm80211_offset_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif, s64 offset)
{
	int vdev = scm80211_vdev_id(vif);
	u64 delta = abs(offset);
	u32 v[2];
	u8 op = offset < 0 ? 1 : 0;

	scm80211_trace(hw, "+%s (%lld) (%llu) op %d\n", __func__, offset, delta, op);

	put_unaligned_le64(delta, v);
	scm80211_mac_writel(hw, v[0], MAC_REG_VIF_TSF_OFFSET_L(vdev));
	scm80211_mac_writel(hw, v[1], MAC_REG_VIF_TSF_OFFSET_H(vdev));

	v[0] = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vdev));
	v[0] &= ~(VIF_CFG0_TU | VIF_CFG0_TOP);
	v[0] |= FIELD_PREP(VIF_CFG0_TU, 1);
	v[0] |= FIELD_PREP(VIF_CFG0_TOP, op);
	scm80211_mac_writel(hw, v[0], MAC_REG_VIF_CFG0(vdev));

	/* for debug hw/hwsim tsf */
	// udelay(2);
	// v[0] = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vdev));
	// WARN_ON(FIELD_GET(VIF_CFG0_TU, v[0]));
	// scm80211_trace(hw, "-%s (%llu)\n", __func__, scm80211_get_tsf(hw, vif));
}

static void scm80211_reset_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	int vdev = scm80211_vdev_id(vif);
	u32 v;

	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vdev));
	v |= FIELD_PREP(VIF_CFG0_TC, 1);
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG0(vdev));

	/* for debug hw/hwsim tsf */
	// udelay(10);
	// v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vdev));
	// WARN_ON(FIELD_GET(VIF_CFG0_TC, v));
}

void scm80211_set_tsf(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u64 tsf)
{
	u32 v[2];
	int vdev = scm80211_vdev_id(vif);

	//scm80211_trace(hw, "+%s (%llu)\n", __func__, tsf);

	if (!is_pcie_device(hw)) {
		return;
	}

	scm80211_reset_tsf(hw, vif);

	put_unaligned_le64(tsf, v);
	scm80211_mac_writel(hw, v[0], MAC_REG_VIF_TSF_OFFSET_L(vdev));
	scm80211_mac_writel(hw, v[1], MAC_REG_VIF_TSF_OFFSET_H(vdev));

	v[0] = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(vdev));
	v[0] &= ~(VIF_CFG0_TU | VIF_CFG0_TOP);
	v[0] |= FIELD_PREP(VIF_CFG0_TU, 1);
	scm80211_mac_writel(hw, v[0], MAC_REG_VIF_CFG0(vdev));

	// udelay(2);
	// scm80211_get_tsf(hw, vif);
}

/* Stop tx traffic 1ms before the GO goes away */
#define P2P_PS_STOP_TIME       1000	//1000

static inline void scm80211_update_p2p_ps_timer(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u32 tsf32, target_tsf;

	tsf32 = scm80211_get_tsf32(hw, vif);
	target_tsf = vifp->noa.next_tsf;

	if (!vifp->noa.absent)
		target_tsf -= P2P_PS_STOP_TIME;
	else
		target_tsf += P2P_PS_STOP_TIME;

	if (target_tsf - tsf32 < P2P_PS_STOP_TIME)
		target_tsf = tsf32 + P2P_PS_STOP_TIME;

	hrtimer_start(&vifp->p2p_ps_timer, ns_to_ktime((target_tsf - tsf32) * 1000),
		      HRTIMER_MODE_REL_SOFT);
}

static enum hrtimer_restart scm80211_p2p_ps_timer(struct hrtimer *timer)
{
	struct ieee80211_vif_priv *vifp =
	    container_of(timer, struct ieee80211_vif_priv, p2p_ps_timer);
	struct ieee80211_vif *vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	struct ieee80211_hw *hw = vifp->hw;
	u32 tsf32;
	bool absent;
	unsigned long flags;

	tsf32 = scm80211_get_tsf32(hw, vif);

	absent = vifp->noa.absent;

	if (!vifp->noa.absent)
		tsf32 += P2P_PS_STOP_TIME;
	else
		tsf32 -= P2P_PS_STOP_TIME;

	scm80211_trace(hw, "%s() current tsf32=%u absent =%d vifp->noa.next_tsf =%u\n",
		       __func__, tsf32, vifp->noa.absent, vifp->noa.next_tsf);

	spin_lock_irqsave(&vifp->lock, flags);

	if (!vifp->noa.has_next_tsf || vifp->noa.next_tsf - tsf32 > BIT(31))
		ieee80211_update_p2p_noa(&vifp->noa, tsf32);

	if (!(absent == vifp->noa.absent && vifp->noa.next_tsf == tsf32)) {
		/* not the final noa timer calback(noa.count >1) */
		scm80211_update_p2p_ps_timer(hw, vif);
	} else {
		/*
		 * In the final noa timer callback for entering noa, noa.
		 * desc[i].start will be not update
		 * In the final noa timer callback for leaving noa, noa.next_tsf
		 * will be the tsf value input,noa.absent will not change
		 */
		memset(&vifp->noa, 0, sizeof(struct ieee80211_noa_data));
		scm80211_trace(hw, "%s() absend =%d quit noa!\n", __func__, vifp->noa.absent);
	}

	spin_unlock_irqrestore(&vifp->lock, flags);

	if (vifp->noa.absent)
		/* TODO  stop tx */
		;
	else
		/* TODO can tx */
		;
	return HRTIMER_NORESTART;
}

static void scm80211_update_p2p_ps(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	u32 tsf32;
	unsigned long flags;

	/* lower 4 bytes */
	tsf32 = scm80211_get_tsf32(hw, vif);

	if (vif->bss_conf.p2p_noa_attr.desc[0].count < 1)
		return;

	/* FIXME  we use start_time  relative to now */
	vif->bss_conf.p2p_noa_attr.desc[0].start_time =
	    tsf32 + vif->bss_conf.p2p_noa_attr.desc[0].start_time;

	spin_lock_irqsave(&vifp->lock, flags);
	ieee80211_parse_p2p_noa(&vif->bss_conf.p2p_noa_attr, &vifp->noa, tsf32);
	spin_unlock_irqrestore(&vifp->lock, flags);

	scm80211_update_p2p_ps_timer(hw, vif);
}

/**
 * DOC: mac80211 interface
 */

/**
 * scm80211_start() - start hardware
 * @hw: hardware
 *
 * This function is &struct ieee80211_ops @start callback. This function is
 * called when the first interface is added.
 */
static int scm80211_start(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 v;

	priv->started = true;
	priv->coverage_class = 0;
	priv->int_en = 0;
	priv->mac_ver = 0;
	priv->phy_ver = 0;
	priv->pm.ps = PS_DISABLED;
	priv->tsf_offset = 0;
	memset(&priv->stats, 0, sizeof(priv->stats));

	scm80211_mac_writel(hw, 0, MAC_REG_INTR_ENABLE);
	scm80211_mac_writel(hw, 0, MAC_REG_MAC_CFG);

	scm80211_init_tx(hw);
	scm80211_init_rx(hw);
	/* to wait for usb firmware preparation rx/tx */
	mdelay(100);
	priv->int_en |= FIELD_PREP(INTR_ENABLE_RX_BUF, 1);

	v = MAC_CFG_EN;
	scm80211_mac_writel(hw, v, MAC_REG_MAC_CFG);

	scm80211_phy_start(hw);

	/* Put PHY/RF to a known state */
	scm80211_set_chanctx(hw, &priv->sched.idle);

	/* Begin idle chanctx */
	priv->sched.idle.begin(hw, &priv->sched.idle);

	return 0;
}

/**
 * scm80211_stop() - stop hardware
 * @hw: hardware
 *
 * This function is &struct ieee80211_ops @stop callback. The function is
 * called when all the interfaces are removed.
 */
static void scm80211_stop(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	u32 v;
	int retry = 0;

	priv->started = false;

	v = scm80211_mac_readl(hw, MAC_REG_MAC_CFG);
	v &= ~MAC_CFG_EN;
	scm80211_mac_writel(hw, v, MAC_REG_MAC_CFG);

	/* wait for all RX FSM finish */
	while (scm80211_mac_readl(hw, MAC_REG_DBG_RXE_STATE) & (~DBG_RXE_STATE_RX_BUSY_MASK)) {
		msleep(1);
		if (retry++ > 10)
			break;
	}

	scm80211_deinit_rx(hw);
	scm80211_deinit_tx(hw);

	scm80211_phy_stop(hw);

	v = scm80211_mac_readl(hw, MAC_REG_INTR_STATUS);
	scm80211_mac_writel(hw, v, MAC_REG_INTR_STATUS);
}

/*
 * This function controls hardware for idle state
 */
static void scm80211_change_idle(struct ieee80211_hw *hw)
{
	/* We will handle this differently */
	/* TODO: enable/disable HW rx-chain */
}

static int scm80211_config(struct ieee80211_hw *hw, u32 changed)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_conf *conf = &hw->conf;
	u32 v;

	scm80211_trace(hw, "%s(changed=%08x)\n", __func__, changed);

	priv->idle = !!(conf->flags & IEEE80211_CONF_IDLE);

	if (changed & IEEE80211_CONF_CHANGE_CHANNEL)
		scm80211_set_mac_channel(hw, &conf->chandef);

	// if (changed & IEEE80211_CONF_CHANGE_SMPS) {

	// }

	// if (changed & IEEE80211_CONF_CHANGE_LISTEN_INTERVAL) {

	// }

	if (changed & IEEE80211_CONF_CHANGE_MONITOR) {
		v = scm80211_mac_readl(hw, MAC_REG_RX_CFG);
		v &= ~RX_CFG_PRO;
		if (conf->flags & IEEE80211_CONF_MONITOR)
			v |= RX_CFG_PRO;
		scm80211_trace(hw, "monitor mode %s\n", (v & RX_CFG_PRO) ? "on" : "off");
		scm80211_mac_writel(hw, v, MAC_REG_RX_CFG);
	}
	if (changed & IEEE80211_CONF_CHANGE_PS) {
#ifdef SCM80211_HOSTAP_AUTO_TEST
		if (!(hw->conf.flags & IEEE80211_CONF_PS)) {
			unsigned long queues = 0;
			int i;

			for (i = 0; i < SCM80211_NUM_VIF; i++) {
				if ((priv->vif_map & BIT(i)))
					queues += ((1 << SCM80211_NUM_HWQ_PER_VIF) - 1) <<
					    (i * SCM80211_NUM_HWQ_PER_VIF);
			}
			scm80211_wake_queues(hw, queues);
		}
#endif
		if (hw->conf.flags & IEEE80211_CONF_PS)
			scm80211_enable_ps(hw, true);
		else
			scm80211_disable_ps(hw);
	}
	// if (changed & IEEE80211_CONF_CHANGE_POWER) {

	// }
	// if (changed & IEEE80211_CONF_CHANGE_RETRY_LIMITS) {

	// }
	if (changed & IEEE80211_CONF_CHANGE_IDLE)
		scm80211_change_idle(hw);

	return 0;
}

static void scm80211_bcn_monitor_work(struct work_struct *work)
{
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_vif *vif;
	struct ieee80211_hw *hw;
	struct ieee80211_hw_priv *priv;
	struct station_info sinfo;
	struct wireless_dev *wdev;

	bool scanning;

	vifp = container_of(work, struct ieee80211_vif_priv, bcn_mon_work);
	vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	hw = ieee80211_vif_to_hw(vif);
	priv = hw->priv;
	wdev = ieee80211_vif_to_wdev(vif);

	if (!hw || !wdev)
		return;

	scanning = scm80211_is_scanning(hw);

	if (!wdev->netdev || cfg80211_get_station(wdev->netdev, vif->bss_conf.bssid, &sinfo))
		return;

	scm80211_trace(hw, "calculate beacon loss (curr loss cnt = %d, prev loss cnt = %d)\n",
		       sinfo.beacon_loss_count, vifp->beacon_loss_cnt);

	if ((sinfo.beacon_loss_count - vifp->beacon_loss_cnt > 1) &&
	    scm80211_get_twt_state(vif) != TWT_DOZE) {
		/*
		 * calculate the duration of last tx, if it tx and acked in 2s
		 * do not disconnect the network
		 */
		if (ktime_ms_delta(ktime_get(), vifp->tstamp) > 2000)
			ieee80211_connection_loss(vif);
	}

	vifp->beacon_loss_cnt = sinfo.beacon_loss_count;
	vifp->rx_beacon = sinfo.rx_beacon;

	mod_timer(&vifp->bcn_mon_timer,
		  round_jiffies_up(jiffies + vifp->beacon_timeout * (scanning ? 5 : 1)));
}

static void scm80211_bcn_mon_timer(struct timer_list *t)
{
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_vif *vif;
	struct ieee80211_hw *hw;

	vifp = from_timer(vifp, t, bcn_mon_timer);
	vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	if (WARN_ON(!vif || vif->type != NL80211_IFTYPE_STATION))
		return;
	if (WARN_ON(!vif->bss_conf.assoc))
		return;

	hw = ieee80211_vif_to_hw(vif);

	scm80211_trace(hw, "%s (bssid=%pM)\n", __func__, vif->bss_conf.bssid);

	if (!is_valid_ether_addr(vif->bss_conf.bssid))
		return;

	ieee80211_queue_work(hw, &vifp->bcn_mon_work);
}

static int scm80211_add_p2p_device_interface(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw_priv *priv = hw->priv;
	int i;

	vifp->index = find_first_bit(&priv->vif_map, SCM80211_NUM_VIF);
	if (vifp->index == SCM80211_NUM_VIF) {
		scm80211_err(hw, "no primary vif used\n");
		scm80211_try_to_doze(hw);
		return -ENOENT;
	}
	scm80211_set_addr(hw, vif, vif->addr);
	spin_lock_init(&vifp->lock);

	for (i = 0; i < IEEE80211_NUM_ACS; i++)
		vif->hw_queue[i] = vifp->index + SCM80211_HWQ_VO;

	scm80211_vif_init_tx(hw, vif);
	scm80211_try_to_doze(hw);

	return 0;
}

static int scm80211_add_interface(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw_priv *priv = hw->priv;
	enum nl80211_iftype iftype = ieee80211_vif_type_p2p(vif);
	u32 ffset = scm80211_iftype_filter[iftype];
	int vdev, err;

	scm80211_trace(hw, "%s (ifname='%s', type='%s', addr=%pM)\n", __func__,
		       ieee80211_vif_name(vif), ieee80211_iftype_str[iftype], vif->addr);

	vifp->hw = hw;
	vifp->def_key_idx = -1;
	atomic_set(&vifp->num_sta, 0);
	atomic_set(&vifp->csa_cnt, 0);
#ifdef SCM80211_WFA_TEST
	vifp->hetb_psdu_len = 0;
#endif
	vifp->bss_changed = 0;

	scm80211_wakeup(hw);

	if (iftype == NL80211_IFTYPE_P2P_DEVICE)
		return scm80211_add_p2p_device_interface(hw, vif);

	vdev = scm80211_alloc_vdev(hw);
	vifp->index = vdev;
	if (vdev < 0) {
		scm80211_err(hw, "failed to alloc vdev resource\n");
		return -EBUSY;
	}

	if (priv->pm.enable && priv->pm.ps == PS_ENABLED &&
	    priv->pm.vif && priv->pm.vif != vif) {
		ieee80211_queue_work(hw, &priv->pm.disable_work);
		msleep(5);
	}

	priv->vif_map_ref |= BIT(vdev);

	scm80211_net_dev_attach(hw, vif);

	err = scm80211_vif_init_tx(hw, vif);
	if (err) {
		scm80211_err(hw, "failed to init vdev %d hw queues\n", vdev);
		scm80211_free_vdev(hw, vdev);
		return err;
	}
	priv->vifs[vdev] = vif;

	scm80211_set_addr(hw, vif, vif->addr);
	scm80211_config_frame_filter(hw, vif, ffset, 0);
	scm80211_enable_vif(hw, vif, true);

	/*
	 * Do this after beacon is enabled in scm80211_bss_info_changed()
	 * Or implment scm80211_start_ap() and do it there
	 */
	if (ieee80211_iface_is_ap(iftype))
		scm80211_set_vht_partial_aid(hw, vif, iftype, 0);
	spin_lock_init(&vifp->lock);

	if (iftype == NL80211_IFTYPE_P2P_CLIENT || iftype == NL80211_IFTYPE_P2P_GO)
		scm80211_set_p2p_basic_rate(hw, vdev);

	if (iftype == NL80211_IFTYPE_P2P_CLIENT)
		__clear_bit(IEEE80211_HW_PS_NULLFUNC_STACK, hw->flags);
	else
		ieee80211_hw_set(hw, PS_NULLFUNC_STACK);

	hrtimer_init(&vifp->p2p_ps_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
	vifp->p2p_ps_timer.function = scm80211_p2p_ps_timer;

#ifdef SCM80211_TWT
	INIT_WORK(&vifp->twt_state_work, scm80211_twt_state_work);
	vifp->twt_state = TWT_DISABLED;
	vifp->twt_request = NULL;
	hrtimer_init(&vifp->twt_tsf_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
	vifp->twt_tsf_timer.function = scm80211_twt_tsf_timer;
#endif

	/* initialized to mac default val */
	vifp->reg_mmss = VIF_MIN_MPDU_SPACING_TRIG_EN |
	    VIF_MIN_MPDU_SPACING_TRS_EN | VIF_MIN_MPDU_SPACING_BSRP_MORE_QOSNULL_EN;

	if (iftype == NL80211_IFTYPE_STATION || iftype == NL80211_IFTYPE_P2P_CLIENT)
		scm80211_init_sr(hw, vif);

	scm80211_init_bfm_reg(hw, vif);
	/* reference 20211025_SCM2020Z1_EVB3_TX EVM to init hw power to 19dbm */
	scm80211_conf_hw_power(hw, vif, 15);

	INIT_WORK(&vifp->bcn_mon_work, scm80211_bcn_monitor_work);
	timer_setup(&vifp->bcn_mon_timer, scm80211_bcn_mon_timer, 0);

	INIT_WORK(&vifp->omi_work, scm80211_omi_work);

	scm80211_setup_unlink_ps(hw, vif, SCM80211_UNLINK_TIMEOUT);

	/* And so on and so on */
	scm80211_trace(hw, "successfully added interface %pM (vdev=%d)\n", vif->addr, vdev);
	return 0;
}

static int scm80211_change_interface(struct ieee80211_hw *hw,
				     struct ieee80211_vif *vif,
				     enum nl80211_iftype newtype, bool newp2p)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	enum nl80211_iftype iftype = ieee80211_vif_type_p2p(vif);
	u32 ffset;
	int vdev;

	scm80211_cleanup_unlink_ps(hw, vif);

	vdev = vifp->index;
	newtype = ieee80211_iftype_p2p(newtype, newp2p);
	ffset = scm80211_iftype_filter[newtype];

	scm80211_trace(hw, "%s (ifname '%s', '%s'->'%s', addr=%pM)\n", __func__,
		       ieee80211_vif_name(vif),
		       ieee80211_iftype_str[iftype], ieee80211_iftype_str[newtype], vif->addr);

	vif->type = newtype;
	vif->p2p = newp2p;
	scm80211_vif_map_hwqueue(hw, vif);
	scm80211_set_addr(hw, vif, vif->addr);
	scm80211_config_frame_filter(hw, vif, ffset, 0);

	/*
	 * These should not be done here - either when beacon is disabled or
	 * stop_ap().
	 */
	if (ieee80211_vif_to_wdev(vif)) {
		scm80211_clr_vht_partial_aid(hw, vif, iftype);
		scm80211_clr_he_phycfg(hw, vif);
		if (ieee80211_iface_is_ap(newtype))
			scm80211_set_vht_partial_aid(hw, vif, newtype, 0);
	}

	atomic_set(&vifp->num_sta, 0);
	atomic_set(&vifp->csa_cnt, 0);
#ifdef SCM80211_WFA_TEST
	vifp->hetb_psdu_len = 0;
#endif
	vifp->bcn_seqnum = 0;
	vifp->null_seqnum = 0;
#ifdef SCM80211_TWT
	if ((scm80211_get_twt_state(vif) == TWT_REQ || vifp->twt_teardown) && vifp->twt_request)
		kfree_skb(vifp->twt_request);
	vifp->twt_request = NULL;
	scm80211_set_twt_state(vif, TWT_DISABLED);
#endif
	/* initialized to mac default val */
	vifp->reg_mmss = VIF_MIN_MPDU_SPACING_TRIG_EN |
	    VIF_MIN_MPDU_SPACING_TRS_EN | VIF_MIN_MPDU_SPACING_BSRP_MORE_QOSNULL_EN;

	if (newtype == NL80211_IFTYPE_P2P_CLIENT || newtype == NL80211_IFTYPE_P2P_GO)
		scm80211_set_p2p_basic_rate(hw, vdev);

	if (newtype == NL80211_IFTYPE_STATION || newtype == NL80211_IFTYPE_P2P_CLIENT)
		scm80211_init_sr(hw, vif);

	if (newtype == NL80211_IFTYPE_P2P_CLIENT)
		__clear_bit(IEEE80211_HW_PS_NULLFUNC_STACK, hw->flags);
	else
		ieee80211_hw_set(hw, PS_NULLFUNC_STACK);

	scm80211_setup_unlink_ps(hw, vif, SCM80211_UNLINK_TIMEOUT);

	return 0;
}

static void scm80211_remove_interface(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw_priv *priv = hw->priv;
	u8 zero_mac[ETH_ALEN];

	scm80211_trace(hw, "%s (ifname='%s' addr=%pM)\n", __func__,
		       ieee80211_vif_name(vif), vif->addr);

	scm80211_cleanup_unlink_ps(hw, vif);
	eth_zero_addr(zero_mac);
	scm80211_set_addr(hw, vif, zero_mac);
	if (vif->type == NL80211_IFTYPE_P2P_DEVICE) {
		scm80211_deinit_txq(hw, vif, (struct ieee80211_txq *)&vifp->mq);
		return;
	}

	/*
	 * Let's do not update these after unbound from netdev
	 */
	if (ieee80211_vif_to_wdev(vif)) {
		scm80211_clr_vht_partial_aid(hw, vif, ieee80211_vif_type_p2p(vif));
		scm80211_clr_he_phycfg(hw, vif);
	}
#ifdef SCM80211_TWT
	if ((scm80211_get_twt_state(vif) == TWT_REQ || vifp->twt_teardown) && vifp->twt_request)
		kfree_skb(vifp->twt_request);
	vifp->twt_request = NULL;
	scm80211_set_twt_state(vif, TWT_DISABLED);
	hrtimer_cancel(&vifp->twt_tsf_timer);
	cancel_work_sync(&vifp->twt_state_work);
#endif
	priv->vif_map_ref &= ~BIT(scm80211_vdev_id(vif));
#ifdef SCM80211_COEX
	scm80211_pta_pri_set(hw, 0);
	scm80211_disable_coex_irq(hw);
#endif
	scm80211_vif_deinit_tx(hw, vif);

	scm80211_config_frame_filter(hw, vif, 0, -1U);
	scm80211_set_bssid(hw, vif, zero_mac, zero_mac);
	scm80211_enable_vif(hw, vif, false);
	scm80211_set_bssid_mask(hw, vif, 0);
	hrtimer_cancel(&vifp->p2p_ps_timer);
	del_timer_sync(&vifp->bcn_mon_timer);
	cancel_work_sync(&vifp->bcn_mon_work);
	cancel_work_sync(&vifp->omi_work);
	scm80211_free_vdev(hw, scm80211_vdev_id(vif));

	priv->vifs[scm80211_vdev_id(vif)] = NULL;
}

void scm80211_beacon_tx(void *arg, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = vifp->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_channel *chan;
	struct ieee80211_tx_info *info;
	struct ieee80211_rate *txrate;
	struct ieee80211_mgmt *mgmt;
	struct sk_buff *skb;
	struct sk_buff *skb_cpy;
	int csa_cnt;
	u64 tsf;

	if (vif->type != NL80211_IFTYPE_AP &&
	    vif->type != NL80211_IFTYPE_MESH_POINT && vif->type != NL80211_IFTYPE_ADHOC)
		return;

	skb = ieee80211_beacon_get(hw, vif);
	if (!skb)
		return;

	info = IEEE80211_SKB_CB(skb);
	if (ieee80211_hw_check(hw, SUPPORTS_RC_TABLE))
		ieee80211_get_tx_rates(vif, NULL, skb, info->control.rates,
				       ARRAY_SIZE(info->control.rates));

	txrate = ieee80211_get_tx_rate(hw, info);

	mgmt = (struct ieee80211_mgmt *)skb->data;
	/* fake header transmission time */
	tsf = scm80211_get_tsf(hw, vif);

	if (tsf == 0 || tsf <= priv->abs_bcn_ts) {
		/* HW TSF does not work, get TSF from sys */
		priv->abs_bcn_ts = scm80211_get_tsf_from_sys();
	} else {
		priv->abs_bcn_ts = tsf;
	}
	if (info->flags & IEEE80211_TX_CTL_ASSIGN_SEQ) {
		mgmt->seq_ctrl &= cpu_to_le16(IEEE80211_SCTL_FRAG);
		mgmt->seq_ctrl = cpu_to_le16(vifp->bcn_seqnum++ << 4) & IEEE80211_SCTL_SEQ;
	}
	mgmt->u.beacon.timestamp = cpu_to_le64(priv->abs_bcn_ts +
					       priv->tsf_offset + 24 * 8 * 10 / txrate->bitrate);

	csa_cnt = atomic_read(&vifp->csa_cnt);
	if (csa_cnt > 0) {
		struct element *csa_elem;
		struct ieee80211_channel_sw_ie *sw_ie;

		skb_cpy = skb_copy_expand(skb, hw->extra_tx_headroom,
					  sizeof(*sw_ie) + 2, GFP_ATOMIC);
		if (!skb_cpy)
			goto out;

		csa_elem = (struct element *)skb_put(skb_cpy, sizeof(*sw_ie) + 2);
		csa_elem->id = WLAN_EID_CHANNEL_SWITCH;
		csa_elem->datalen = sizeof(*sw_ie);
		sw_ie = (void *)csa_elem->data;
		sw_ie->mode = 1;
		sw_ie->new_ch_num = vifp->sw_ch;
		sw_ie->count = csa_cnt;

		kfree_skb(skb);
		skb = skb_cpy;
		atomic_dec(&vifp->csa_cnt);
	}

	if (vifp->noa_enabled) {
		u8 *data;
		struct ieee80211_p2p_noa_attr *noa;
		u8 ext_len = sizeof(struct ieee80211_p2p_noa_desc) + 11;

		mgmt = (struct ieee80211_mgmt *)skb->data;
		skb_cpy = skb_copy_expand(skb, hw->extra_tx_headroom, ext_len, GFP_ATOMIC);
		if (!skb_cpy)
			goto out;

		data = skb_put(skb_cpy, ext_len);
		*data++ = WLAN_EID_VENDOR_SPECIFIC;
		*data++ = ext_len - 2;
		*data++ = (WLAN_OUI_WFA >> 16) & 0xff;
		*data++ = (WLAN_OUI_WFA >> 8) & 0xff;
		*data++ = (WLAN_OUI_WFA >> 0) & 0xff;
		*data++ = WLAN_OUI_TYPE_WFA_P2P;
		*data++ = IEEE80211_P2P_ATTR_ABSENCE_NOTICE;
		*data++ = 2 + sizeof(struct ieee80211_p2p_noa_desc);
		*data++ = 0;
		noa = (void *)data;
		noa->index = 0;
		noa->oppps_ctwindow = 0;
		noa->desc[0].count = 1;
		/* absent for 4s */
		noa->desc[0].duration = cpu_to_le32(4000000);
		noa->desc[0].interval = cpu_to_le32(4000000);
		noa->desc[0].start_time = cpu_to_le32(priv->abs_bcn_ts +
						      priv->tsf_offset +
						      24 * 8 * 10 / txrate->bitrate + 102400);
		kfree_skb(skb);
		skb = skb_cpy;
		vifp->noa_enabled = false;
	}

out:
	scm80211_wake_queues(hw, BIT(vifp->mgmt_hw_queue));
	chan = rcu_dereference_chanctx_conf(vif)->def.chan;
	scm80211_tx_channel(hw, skb, chan);

	while ((skb = ieee80211_get_buffered_bc(hw, vif))) {
		/* The mac80211 does not set control.vif for buffered_bc. */
		info = IEEE80211_SKB_CB(skb);
		info->control.vif = vif;

		scm80211_tx_channel(hw, skb, chan);
	}
}

static void scm80211_check_beacon(void *arg, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = arg;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw *hw = priv->hw;

	if (vif->type != NL80211_IFTYPE_AP &&
	    vif->type != NL80211_IFTYPE_MESH_POINT && vif->type != NL80211_IFTYPE_ADHOC)
		return;

	if (scm80211_is_scanning(hw) || scm80211_is_offchannel(hw, vif)) {
		struct scm80211_chanctx *chanctx = NULL;
		struct scm80211_chanctx_sched *sched = hw_sched(hw);

		scm80211_trace(hw, "tx beacon during scan or offchannel\n");
		chanctx = (void *)rcu_dereference_chanctx_conf(vif);

		if (!chanctx)
			return;

		/* temporary improve CHANCTX type priority */
		if (chanctx->type != CHANCTX_TYPE_BEACON) {
			/* stop mgmt queue */
			scm80211_stop_queues(hw, BIT(vifp->mgmt_hw_queue));
			list_del_init(&chanctx->sched_list);
			chanctx->type = CHANCTX_TYPE_BEACON;
			list_add(&chanctx->sched_list, &sched->queue[chanctx->type].head);
			priv->beacon_offchan = true;
		}
	}
}

/** Using timer_list timer function instead of hrtimer
 *  Background: For kernel older than 4.16.0 hrtimer will be called as
 *  hardirq which cause some bug when calling ieee80211_beacon_get
 *  _raw_spin_lock_bh can not be called in hardirq.
 *  So changed to timer_list
 *
 * static enum hrtimer_restart scm80211_beacon(struct hrtimer *timer)
 */
static void scm80211_beacon(struct timer_list *timer)
{
	struct ieee80211_hw_priv *data =
	    container_of(timer, struct ieee80211_hw_priv, beacon_timer);
	struct ieee80211_hw *hw = data->hw;
	u64 bcn_int = data->beacon_int;

	if (!data->started)
		goto out;

	ieee80211_iterate_active_interfaces_atomic(hw,
						   IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_check_beacon, data);

	if (data->beacon_offchan)
		return;

	ieee80211_iterate_active_interfaces_atomic(hw,
						   IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_beacon_tx, NULL);

	/* beacon at new TBTT + beacon interval */
	if (data->bcn_delta) {
		bcn_int -= data->bcn_delta;
		data->bcn_delta = 0;
	}
	mod_timer(&data->beacon_timer, jiffies + usecs_to_jiffies(bcn_int));

out:
	return;
}

static enum hrtimer_restart scm80211_phy(struct hrtimer *timer)
{
#define BEACON_INTERVAL	1
	struct ieee80211_hw_priv *data = container_of(timer, struct ieee80211_hw_priv, phy_timer);
	struct ieee80211_hw *hw = data->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	ktime_t next_bcn;
	static u32 m_770[2] = { 0 };
	static u32 m_774[2] = { 0 };
	static u32 m_780[2] = { 0 };
	static u32 m_784[2] = { 0 };
	static u32 p_17c[2] = { 0 };
	u32 a, b, c, d, e, aa, bb, cc, dd, ee, ff, gg, hh, jj;

	if (priv->rx_statistic) {
		a = scm80211_mac_readl(hw, 0x8770);
		aa = (a & 0xffff) - (m_770[priv->devid] & 0xffff);
		bb = (a >> 16) - (m_770[priv->devid] >> 16);

		b = scm80211_mac_readl(hw, 0x8774);
		cc = (b & 0xffff) - (m_774[priv->devid] & 0xffff);
		dd = (b >> 16) - (m_774[priv->devid] >> 16);

		c = scm80211_mac_readl(hw, 0x8780);
		ee = (c & 0xffff) - (m_780[priv->devid] & 0xffff);
		ff = (c >> 16) - (m_780[priv->devid] >> 16);

		d = scm80211_mac_readl(hw, 0x8784);
		gg = (d & 0xffff) - (m_784[priv->devid] & 0xffff);
		hh = (d >> 16) - (m_784[priv->devid] >> 16);

		e = phy_readl(&priv->phy[priv->devid], 0x17c);
		jj = e - p_17c[priv->devid];

		m_770[priv->devid] = a;
		m_774[priv->devid] = b;
		m_780[priv->devid] = c;
		m_784[priv->devid] = d;
		p_17c[priv->devid] = e;

		scm80211_trace(hw,
			       "rx: ppdu[%d],data[%d],mgmt[%d],ctrl[%d]\n",
			       aa, bb, cc, dd);
		scm80211_trace(hw,
			       "fcs_err[%d],fcs ok[%d],buf_overflow[%d],filter_in[%d],filter_out[%d]\n",
			       ee, jj, ff, gg, hh);
	}

	next_bcn = ktime_add(hrtimer_get_expires(timer),
			     ns_to_ktime(BEACON_INTERVAL * 100 * 1024 * 1000));
	hrtimer_start(&data->phy_timer, next_bcn, HRTIMER_MODE_ABS_SOFT);

	return HRTIMER_NORESTART;
}

#define fif(x)[__bf_shf(x)] = #x

static const char * const fif_string[] = {
	fif(FIF_ALLMULTI),
	fif(FIF_FCSFAIL),
	fif(FIF_PLCPFAIL),
	fif(FIF_BCN_PRBRESP_PROMISC),
	fif(FIF_CONTROL),
	fif(FIF_OTHER_BSS),
	fif(FIF_PSPOLL),
	fif(FIF_PROBE_REQ),
};

static void scm80211_config_iface_filter(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif,
					 unsigned int flags, unsigned int changed_flags)
{
	unsigned long changes = (unsigned long)changed_flags;
	u32 bitmap[2] = { 0, 0 };
	int i;

	scm80211_trace(hw, "%s(vdev=%d, flags=%08x, changed=%lx)\n",
		       __func__, scm80211_vdev_id(vif), flags, changes);

	for_each_set_bit(i, &changes, bitsizeof(changed_flags)) {
		int set = flags & BIT(i) ? 1 : 0;

		scm80211_dbg(hw, "%c%s\n", set ? '+' : '-', fif_string[i]);

		switch (BIT(i)) {
		case FIF_BCN_PRBRESP_PROMISC:
			bitmap[1] |= (mbit(IEEE80211_STYPE_BEACON) |
				      mbit(IEEE80211_STYPE_PROBE_RESP));
			scm80211_config_bssid_filter(hw, vif, !set);
			break;
		case FIF_PROBE_REQ:
			bitmap[set] |= mbit(IEEE80211_STYPE_PROBE_REQ);
			break;
		case FIF_CONTROL:
			bitmap[set] |= rx_filter_control;
			break;
		case FIF_PSPOLL:
			bitmap[set] |= cbit(IEEE80211_STYPE_PSPOLL);
			break;
		case FIF_OTHER_BSS:
			scm80211_config_bssid_filter(hw, vif, !set);
			break;
		case FIF_FCSFAIL:
		case FIF_PLCPFAIL:
		default:
			break;
		}
	}

	if (bitmap[0] || bitmap[1])
		scm80211_config_frame_filter(hw, vif, bitmap[1], bitmap[0]);
}

struct iface_filter_data {
	struct ieee80211_hw *hw;
	unsigned int changed;
	unsigned int flags;
};

static void scm80211_set_iface_filter(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	struct iface_filter_data *ifd = data;

	scm80211_config_iface_filter(ifd->hw, vif, ifd->flags, ifd->changed);
}

/**
 * scm80211_configure_filter() - configure RX filter
 *
 * We apply the same configuration for all vdevs
 */
static void scm80211_configure_filter(struct ieee80211_hw *hw, unsigned int changed,
				      unsigned int *flags, u64 multicast)
{
	struct iface_filter_data data = {
		.hw = hw,
		.changed = changed,
		.flags = *flags,
	};

	scm80211_trace(hw, "%s(change=%08x, total=%08x, mc=%lld)\n",
		       __func__, changed, *flags, multicast);

	ieee80211_iterate_interfaces(hw, 0, scm80211_set_iface_filter, &data);

	/* We will not support these flags for the time being */
	*flags &= ~(FIF_FCSFAIL | FIF_PLCPFAIL | BIT(31));
}

/*
 * struct cfg80211_chan_def
 * struct cfg80211_bss
 * struct ieee80211_bss_conf
 * struct ieee80211_sta
 */

static void scm80211_assoc_ht(struct ieee80211_hw *hw,
			      struct ieee80211_vif *vif,
			      struct ieee80211_sta *ap, struct ieee80211_bss_conf *info)
{
	if (!get_ieee80211_ht_cap(ap).ht_supported)
		return;

	scm80211_set_basic_ht_mcs_set(hw, vif, info);
	scm80211_set_ht_mmss(hw, vif, ap, info);
}

static void scm80211_assoc_vht(struct ieee80211_hw *hw,
			       struct ieee80211_vif *vif,
			       struct ieee80211_sta *ap, struct ieee80211_bss_conf *info)
{
	if (!get_ieee80211_vht_cap(ap).vht_supported)
		return;

	scm80211_set_vht_partial_aid(hw, vif, ieee80211_vif_type_p2p(vif), info->aid);

	/* FXIME */
}

static void scm80211_assoc_he(struct ieee80211_hw *hw,
			      struct ieee80211_vif *vif,
			      struct ieee80211_sta *ap, struct ieee80211_bss_conf *info)
{
	if (!get_ieee80211_he_cap(ap).has_he || !info->he_support)
		return;

#ifdef TEST_NO_HE
	return;
#endif

	if (info->nontransmitted) {
		u8 *trans_addr = info->transmitter_bssid;

		scm80211_trace(hw, "bssid %pM, trans bssid %pM\n", info->bssid, trans_addr);
		scm80211_set_bssid(hw, vif, info->bssid, trans_addr);
		scm80211_set_bssid_mask(hw, vif, info->bssid_indicator);
	}

	scm80211_set_he_info(hw, vif, ap, info);
	/* FIXME */
}

/**
 * scm80211_assoc_bss() - join a BSS as a station
 *
 * This function performs extra processing that is not handled by
 * scm80211_bss_info_changed().
 *
 * @hw: hardware
 * @vif: interface
 * @info: bss configuration
 *
 * Registers that need to be programmed:
 * MAC_REG_DEV_CFG:
 * - band (2G or 5G)
 * MAC_REG_VIF_CFG0:
 * - erp
 * MAC_REG_VIF_CFG1:
 * - basic mcs
 * - basic rates
 * MAC_REG_VIF_RX_FILTER0:
 * MAC_REG_VIF_RX_FILTER1:
 * MAC_REG_VIF_MAC0_L:
 * MAC_REG_VIF_MAC0_H:
 * MAC_REG_VIF_MAC1_L:
 * MAC_REG_VIF_MAC1_H:
 * MAC_REG_VIF_BSS0_L:
 * MAC_REG_VIF_BSS0_H:
 * MAC_REG_VIF_BSS1_L:
 * MAC_REG_VIF_BSS1_H:
 * MAC_REG_VIF_AID:
 * MAC_REG_VIF_
 *
 * must be called in an rcu_read_lock() critical section.
 */

static int scm80211_assoc_bss(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = (void *)vif->drv_priv;
	struct ieee80211_bss_conf *info = &vif->bss_conf;
	struct ieee80211_sta *ap;
	int vidx = scm80211_vdev_id(vif);

	scm80211_trace(hw, " vdev %d joining bss %pM, aid=%d\n", vidx, info->bssid, info->aid);

	scm80211_cleanup_unlink_ps(hw, vif);
	ap = ieee80211_find_sta(vif, info->bssid);
	if (!ap) {
		scm80211_warn(hw, "failed to find sta for bss %pM vdev %d\n", info->bssid, vidx);
		return -ENOENT;
	}

	vp->null_seqnum = 0;
	vp->last_beacon = ktime_get();
	vp->bss.beacon.last = 0;
	vp->bss.beacon.tsf0 = -1U;
	vp->bss.beacon.dur = 0;

	scm80211_set_aid(hw, vif, info->aid);
	scm80211_assoc_ht(hw, vif, ap, info);
	scm80211_assoc_vht(hw, vif, ap, info);
	scm80211_assoc_he(hw, vif, ap, info);
	scm80211_associated(hw, vif, info);
	scm80211_setup_link_ps(hw, vif);

	return 0;
}

static int scm80211_leave_bss(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_bss_conf *info = &vif->bss_conf;
	int vidx = scm80211_vdev_id(vif);
	u8 zero_mac[ETH_ALEN];

	scm80211_trace(hw, " vdev %i leaving bssid %pM\n", vidx, info->bssid);

	eth_zero_addr(zero_mac);
	scm80211_cleanup_link_ps(hw, vif);
	scm80211_set_bssid(hw, vif, zero_mac, zero_mac);
	scm80211_set_bssid_mask(hw, vif, 0);
	scm80211_clr_vht_partial_aid(hw, vif, ieee80211_vif_type_p2p(vif));
	scm80211_clr_he_info(hw, vif);

	scm80211_set_aid(hw, vif, 2345);
	scm80211_clr_tdls(hw, vif);
	scm80211_disconnected(hw, vif, info);

	scm80211_setup_unlink_ps(hw, vif, SCM80211_UNLINK_TIMEOUT);

	return 0;
}

static void scm80211_bcn_en_iter(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	unsigned int *count = data;

	if (vif->bss_conf.enable_beacon)
		(*count)++;
}

static const char * const power_type_string[] = {
	[NL80211_TX_POWER_AUTOMATIC] = "automatically",
	[NL80211_TX_POWER_LIMITED] = "limit",
	[NL80211_TX_POWER_FIXED] = "fixed"
};

static void scm80211_set_ap_bssid_and_color(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	u32 offset;
	struct sk_buff *skb;
	struct ieee80211_mgmt *mgmt;
	const struct element *elem;
	const struct ieee80211_he_operation *he_oper = NULL;
	u8 bssid[ETH_ALEN];

	scm80211_trace(hw, "+%s\n", __func__);
	skb = ieee80211_beacon_get(hw, vif);
	if (!skb)
		return;

	mgmt = (struct ieee80211_mgmt *)skb->data;
	offset = offsetof(struct ieee80211_mgmt, u.beacon.variable);
	elem = cfg80211_find_elem(WLAN_EID_MULTIPLE_BSSID,
				  mgmt->u.beacon.variable, skb->len - offset);
	if (elem && elem->datalen > 1 && elem->data[0] == 1) {
		u8 *non_bssid = bssid;

		scm80211_trace(hw, " have multi-bssid!\n");
		/**
		 *  9.4.2.45 Multiple BSSID element
		 * BSSID(i) = BSSID_A | BSSID_B
		 * BSSID_A = REF_BSSID & ZERO[(47-n+1):47]
		 * BSSID_B = ZERO[0:(47-n)] &
		 *	bin[ ((dec(REF_BSSID[(47-n+1):47]) + i) mod 2^n), n]
		 *
		 * simplify:
		 * non_bssid =
		 * REF_BSSID[0:46] | ((REF_BSSID[47:47]+1) mod 2)
		 * only change the MSB 1 bit
		 * non_bssid[ETH_ALEN-1] = tansbssid[ETH_ALEN-1] ^ 0x80
		 */
		memcpy(non_bssid, mgmt->bssid, ETH_ALEN);
		non_bssid[ETH_ALEN - 1] ^= 0x80;
		scm80211_set_bssid(hw, vif, non_bssid, mgmt->bssid);
		scm80211_set_bssid_mask(hw, vif, 1);
	} else {
		u8 *zero_bssid = bssid;

		eth_zero_addr(zero_bssid);
		scm80211_set_bssid(hw, vif, mgmt->bssid, zero_bssid);
	}

	elem = cfg80211_find_ext_elem(WLAN_EID_EXT_HE_OPERATION,
				      mgmt->u.beacon.variable, skb->len - offset);
	if (elem && elem->datalen >= sizeof(*he_oper) + 1) {
		u8 color = 0;

		scm80211_trace(hw, " configured as an he ap!\n");

		he_oper = (void *)(elem->data + 1);

		color = le32_get_bits(he_oper->he_oper_params,
				      IEEE80211_HE_OPERATION_BSS_COLOR_MASK);
		scm80211_set_he_phycfg(hw, vif, &vif->bss_conf, color);
	}

	kfree_skb(skb);
}

/**
 * DOC: call chain example
 *
 * ieee80211_iface_work()
 * - ieee80211_sta_rx_queued_mgmt()
 *   - ieee80211_rx_mgmt_assoc_resp()
 *     - ieee80211_assoc_success()
 *       - ieee80211_set_associated()
 *         - ieee80211_bss_info_change_notify()
 *           - drv_bss_info_changed()
 *     - drv_event_callback()
 */

/**
 * scm80211_bss_info_changed() - handle changes in BSS
 */
static void scm80211_bss_info_changed(struct ieee80211_hw *hw,
				      struct ieee80211_vif *vif,
				      struct ieee80211_bss_conf *info, u32 changed)
{
	struct ieee80211_vif_priv *vp = (void *)vif->drv_priv;
	struct ieee80211_hw_priv *data = hw->priv;

	scm80211_trace(hw, "%s(%s, changed=0x%x):\n", __func__, ieee80211_vif_name(vif), changed);

	vp->bss_changed |= changed;

	if (changed & BSS_CHANGED_BSSID) {
		scm80211_trace(hw, " BSSID %pM (nontrans %s), trans bssid %pM",
			       info->bssid, info->nontransmitted ? "true" : "false",
			       info->transmitter_bssid);
		/*
		 * FIXME
		 * need to check whether BSSID is transmited bssid
		 * at the beginning
		 */
		scm80211_set_bssid(hw, vif, info->bssid, info->transmitter_bssid);
	}

	if (changed & BSS_CHANGED_BASIC_RATES)
		scm80211_set_basic_rates(hw, vif, info);

	if (changed & (BSS_CHANGED_ERP_CTS_PROT | BSS_CHANGED_ERP_PREAMBLE | BSS_CHANGED_ERP_SLOT))
		scm80211_set_erp(hw, vif, info);

	if (changed & BSS_CHANGED_HT)
		scm80211_trace(hw, " ht: op_mode=0x%x\n", info->ht_operation_mode);
	if (changed & BSS_CHANGED_BEACON_INT)
		scm80211_trace(hw, " beacon interval=%d\n", info->beacon_int);
	if (changed & BSS_CHANGED_BEACON) {
		scm80211_trace(hw, " beacon changed\n");
		/* FIXME: set HT, VHT, and HE configuration again */
	}
	if (changed & BSS_CHANGED_BEACON_ENABLED) {
		scm80211_trace(hw, " beacon %s (BI=%u)\n",
			       info->enable_beacon ? "enabled" : "disabled", info->beacon_int);
		if (data->started && info->enable_beacon) {
			data->beacon_int = ieee80211_tu_to_usec(info->beacon_int);
			mod_timer(&data->beacon_timer,
				  jiffies + usecs_to_jiffies(data->beacon_int));
			if (ieee80211_iface_is_ap(vif->type)) {
				scm80211_set_vht_partial_aid(hw, vif, vif->type, 0);
				scm80211_set_ap_bssid_and_color(hw, vif);
			}
		} else if (!info->enable_beacon) {
			unsigned int count = 0;

			scm80211_clr_vht_partial_aid(hw, vif, ieee80211_vif_type_p2p(vif));
			scm80211_clr_he_phycfg(hw, vif);

			ieee80211_iterate_active_interfaces_atomic(data->hw,
								   IEEE80211_IFACE_ITER_NORMAL,
								   scm80211_bcn_en_iter, &count);
			scm80211_trace(hw, " beaconing vifs remaining: %u", count);
			if (count == 0) {
				del_timer_sync(&data->beacon_timer);
				data->beacon_int = 0;
			}
		}
		vp->bcn_seqnum = 0;
	}

	if (changed & BSS_CHANGED_IBSS)
		scm80211_trace(hw, " IBSS joined: %s\n", info->ibss_joined ? "true" : "false");
	if (changed & BSS_CHANGED_ARP_FILTER)
		scm80211_trace(hw, " arp filtering: %d addresses\n", info->arp_addr_cnt);
	if (changed & BSS_CHANGED_QOS)
		scm80211_trace(hw, " qos enabled: %s\n", info->qos ? "true" : "false");
	if (changed & BSS_CHANGED_IDLE)
		scm80211_trace(hw, " interface idle: %s\n", info->idle ? "true" : "false");
	if (changed & BSS_CHANGED_SSID) {
		char ssid[IEEE80211_MAX_SSID_LEN] = { 0, };

		memcpy(ssid, info->ssid, info->ssid_len);
		scm80211_trace(hw, " softap ssid: %s\n", ssid);
		if (ieee80211_vif_type_p2p(vif) == NL80211_IFTYPE_P2P_GO &&
		    !memcmp(ssid, "DIRECT-", 7)) {
			scm80211_set_p2p_basic_rate(hw, scm80211_vdev_id(vif));
		}
	}

	if (changed & BSS_CHANGED_AP_PROBE_RESP)
		scm80211_trace(hw, " probe response changed\n");

	if (changed & BSS_CHANGED_TXPOWER) {
		scm80211_trace(hw, " tx power: %d dBm, power type [%s]\n",
			       info->txpower, power_type_string[info->txpower_type]);
	}
	if (changed & BSS_CHANGED_P2P_PS) {
		enum nl80211_iftype iftype = ieee80211_vif_type_p2p(vif);

		scm80211_trace(hw, " %s P2P GO use P2P ps\n",
			       info->allow_p2p_go_ps ? "allow" : "disallow");
		if (iftype == NL80211_IFTYPE_P2P_CLIENT)
			scm80211_update_p2p_ps(hw, vif);
	}
	if (changed & BSS_CHANGED_BEACON_INFO)
		scm80211_trace(hw, " bss changed dtim period %d\n", info->dtim_period);
	if (changed & BSS_CHANGED_BANDWIDTH) {
		scm80211_trace(hw, " configured bandwidth to %s\n",
			       chanwidth_str[info->chandef.width]);
	}
	if (changed & BSS_CHANGED_MU_GROUPS) {
		scm80211_trace(hw, " mu groups info changed mu owner %s\n",
			       get_ieee80211_mu_mimo_owner(vif) ? "true" : "false");
		if (get_ieee80211_mu_mimo_owner(vif))
			scm80211_set_vht_mu_group(hw, vif, &info->mu_group);
		else
			scm80211_clr_vht_mu_group(hw, vif);
	}
	if (changed & BSS_CHANGED_KEEP_ALIVE) {
		scm80211_trace(hw, " max idle period %d prot keep alive %s\n",
			       info->max_idle_period,
			       info->protected_keep_alive ? "true" : "false");
	}
	if (changed & BSS_CHANGED_MCAST_RATE)
		scm80211_trace(hw, " multicast rate changed\n");
	if (changed & BSS_CHANGED_FTM_RESPONDER) {
		scm80211_trace(hw, " %s FTM responder\n",
			       info->ftm_responder ? "enable" : "disable");
	}
	if (changed & BSS_CHANGED_TWT) {
		/* FIXME need to do more config about TWT */
		scm80211_trace(hw, " TWT requester support %s\n",
			       info->twt_requester ? "true" : "false");
		scm80211_trace(hw, " TWT responder support %s\n",
			       info->twt_responder ? "true" : "false");
	}

	if (changed & BSS_CHANGED_ASSOC) {
		if (info->assoc)
			scm80211_assoc_bss(hw, vif);
		else
			scm80211_leave_bss(hw, vif);
	}

	(void)vp;
	(void)data;
}

static int scm80211_wait_for_aid_cmd_completion(struct ieee80211_hw *hw)
{
	long long start;

	struct reg_aid_cmd cmd = {.val = 0, };

	start = jiffies;
	do {
		cmd.val = scm80211_mac_readl(hw, MAC_REG_AID_CMD);
		if (cmd.aidcmd == AID_CMD_DONE)
			return 0;
		udelay(100);
	} while (jiffies_to_usecs(jiffies - start) < 10000);

	return -1;
}

static int scm80211_set_sta_aid(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				struct ieee80211_sta *sta, bool add)
{
	struct reg_aid_cmd cmd = {.val = 0, };
	unsigned long bitmap;
	u32 v;
	u16 aid = sta->aid;
	const u8 *addr;
#ifdef SCM80211_SAP_31_CLIENT
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
#endif

	/* For non-AP mode and aid(0) cases return directly */
	if (!ieee80211_iface_is_ap(vif->type) || sta->aid == 0)
		return 0;

	/* Check for room */
	if (add) {
		unsigned int max = version(scm80211_mac_version(hw)) >= 0x16 ? 8 : 4;

		bitmap = scm80211_mac_readl(hw, MAC_REG_AID_STATUS);

#ifdef SCM80211_SAP_31_CLIENT
		scm80211_info(hw, "add station num %d\n", atomic_read(&vifp->num_sta));
		if (atomic_read(&vifp->num_sta) >= 31) {
			scm80211_err(hw, "failed to add a station\n");
			return -ENOSPC;
		}
#endif
		if (find_first_zero_bit(&bitmap, max) == max) {
#ifdef SCM80211_SAP_31_CLIENT
			return 0;
#else
			scm80211_info(hw,
				      "failed to add a station aid (aid memory full, 0x%x)\n",
				      (u32)bitmap);

			return -ENOSPC;
#endif
		}
	}

	/* STA address */
	addr = (const u8 *)sta->addr;
	v = get_unaligned_le32(addr);
	scm80211_mac_writel(hw, v, MAC_REG_AID_MAC_ADDR0);
	v = get_unaligned_le16(addr + 4);
	scm80211_mac_writel(hw, v, MAC_REG_AID_MAC_ADDR1);

	scm80211_mac_writel(hw, aid, MAC_REG_AID_MAC_AID);

	/* AID command */
	if (add)
		cmd.aidcmd = AID_CMD_ADD;
	else
		cmd.aidcmd = AID_CMD_DEL;
	scm80211_mac_writel(hw, cmd.val, MAC_REG_AID_CMD);
	if (scm80211_wait_for_aid_cmd_completion(hw) < 0) {
		scm80211_warn(hw,
			      "%s failed to %s sta aid (timeout)\n",
			      __func__, add ? "add" : "remove");
		return add ? -ENOSPC : -EBUSY;
	}

	scm80211_dbg(hw,
		     "%s: successfully %s sta aid <%d, %pM>\n",
		     __func__, add ? "add" : "remove", aid, addr);
	return 0;
}

struct he_obss_narrow_bw_ru_data {
	bool obss_narrow_bw_ru;
};

static void scm80211_get_he_obss_narrow_bw_ru(struct wiphy *wiphy,
					      struct cfg80211_bss *bss, void *_data)
{
	struct he_obss_narrow_bw_ru_data *data = _data;
	const struct element *elem;

	elem = cfg80211_find_elem(WLAN_EID_EXT_CAPABILITY, bss->ies->data, bss->ies->len);

	if (!elem || elem->datalen < 10 ||
	    !(elem->data[9] & WLAN_EXT_CAPA10_OBSS_NARROW_BW_RU_TOLERANCE_SUPPORT)) {
		data->obss_narrow_bw_ru = false;
	}
}

static void scm80211_check_he_obss_narrow_bw_ru(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	u8 vidx = scm80211_vdev_id(vif);
	u32 v;
	bool dfs = false;
	struct he_obss_narrow_bw_ru_data data = {
		.obss_narrow_bw_ru = true,
	};

	cfg80211_bss_iter(hw->wiphy, &vif->bss_conf.chandef,
			  scm80211_get_he_obss_narrow_bw_ru, &data);

	/* We don't use chan->flags bit IEEE80211_CHAN_RADAR because the bit is
	 * always 0 sometimes
	 * dfs = !!(vif->bss_conf.chandef.chan->flags & IEEE80211_CHAN_RADAR);
	 */
	dfs = scm80211_is_radar_freq(vif->bss_conf.chandef.chan->center_freq);

	/*
	 * If there is at least one AP on radar channel that cannot
	 * tolerate 26-tone RU UL OFDMA transmissions using HE TB PPDU.
	 */
	v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG_INFO(vidx));
	if (dfs)
		v |= (FIELD_PREP(VIF_CFG_INFO_DFS_50_100_CON, 1));
	else
		v &= ~VIF0_CFG_INFO_DFS_50_100_CON_MASK;
	if (!data.obss_narrow_bw_ru)
		v |= (FIELD_PREP(VIF_CFG_INFO_OBSS_NARROW_BW_RU_IN_OFDMA_LOST, 1));
	else
		v &= ~VIF0_CFG_INFO_OBSS_NARROW_BW_RU_IN_OFDMA_LOST_MASK;
	scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG_INFO(vidx));
}

/**
 * According to Draft P802.11ax_D7.0 9.4.2.248.5 PPE Thresholds field
 *
 */
static void scm80211_calc_he_npp(struct ieee80211_hw *hw, struct ieee80211_sta *sta)
{
	struct ieee80211_hw_priv *hp = hw->priv;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_sta_he_cap he_cap_tmp = get_ieee80211_he_cap(sta);
	int n, b, nsts, n_ru = 0;
	u8 *ppe = he_cap_tmp.ppe_thres;

	nsts = ppe[0] & IEEE80211_PPE_THRES_NSS_MASK;
	nsts++;
	nsts = min_t(int, hp->n_chain, nsts);

	for (b = 0; b < 5; b++)
		if (ppe[0] & BIT(IEEE80211_PPE_THRES_RU_INDEX_BITMASK_POS + b))
			n_ru++;

	/*
	 * NB: PPE Threshold Info is a bit unclear.
	 * Is the hiearchy NSTS-RU or, RU-NSTS? The current implementation
	 * is based on NSTS-RU.
	 */
	for (n = 0; n < nsts; n++) {
		for (b = 0; b <= CBW80; b++) {
			u8 ppet, ppet16, ppet8;
			u16 npp = 0;
			int bits, idx;
			int pos = 7 + (n_ru * IEEE80211_PPE_THRES_INFO_PPET_SIZE * 2) * n;
			int x = he_constellation_index(7 + (2 * sta_priv->he_mcs[n]));

			if (!(ppe[0] & BIT(IEEE80211_PPE_THRES_RU_INDEX_BITMASK_POS + b)))
				break;

			pos += 2 * IEEE80211_PPE_THRES_INFO_PPET_SIZE * b;
			idx = pos / BITS_PER_BYTE;
			pos = pos % BITS_PER_BYTE;
			bits = min(6, 8 - pos);
			ppet = (ppe[idx] >> pos) & GENMASK(bits - 1, 0);
			if (bits < 6)
				ppet |= (ppe[idx + 1] & GENMASK(6 - bits - 1, 0)) << bits;

			ppet16 = FIELD_GET(GENMASK(2, 0), ppet);
			ppet8 = FIELD_GET(GENMASK(5, 3), ppet);

			/*
			 * The bits [2*x+1, 2*x] of sta_priv->he_npp[n][b] records the encoded NPP
			 * for constellation index x of NSTS n RU b.
			 * The encoding is 0: 0us, 1: 8us, 2: 16us
			 */
			for (; x >= 0; x--) {
				if (x >= ppet8 && (x < ppet16 || ppet16 == 7))
					npp |= 1 << (2 * x);
				else if ((x > ppet8 || ppet8 == 7) && x >= ppet16)
					npp |= 2 << (2 * x);
			}
			sta_priv->he_npp[n][b] = npp;
		}
	}
}

#define VHT_CAP(cap)	(IEEE80211_VHT_CAP_VHT_##cap)
#define HE_MAC_CAP(n, cap) (IEEE80211_HE_MAC_CAP##n##_##cap)

static int scm80211_sta_add(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			    struct ieee80211_sta *sta)
{
	int i, j = 0;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_hw_priv *hw_priv = hw->priv;
	struct ieee80211_sta_he_cap he_cap_tmp = get_ieee80211_he_cap(sta);
	const u8 *he_phy_cap_info = NULL;
	struct scm80211_chanctx *chanctx;
	unsigned long changes;

	scm80211_dbg(hw, "%s(vif=%pM, sta=%pM)\n", __func__, vif->addr, sta->addr);

	if (scm80211_set_sta_aid(hw, vif, sta, true)) {
		scm80211_trace(hw, "%s(vif=%pM, failed to add sta=%pM)\n",
			       __func__, vif->addr, sta->addr);
		return -ENOSPC;
	}

	atomic_inc(&vifp->num_sta);

	for (i = 0; i < ARRAY_SIZE(sta->txq); i++)
		scm80211_init_txq(hw, vif, sta->txq[i], false);

	memset(sta_priv, 0, sizeof(struct ieee80211_sta_priv));
	sta_priv->band = (vif->bss_conf.chandef.chan)->band;
	sta_priv->width = vif->bss_conf.chandef.width;
	sta_priv->tx_packets = 0;
	sta_priv->rx_packets = 0;
	sta_priv->txop_rts_th = rts_threshold;

	/*
	 * reduce the ampdu size from 64-subframe to 32-subframes when bw higher than
	 * BW40 under highspeed usb for ieee80211 phy rate is much higher(720Mbps@2NSS-MCS7)
	 * than bus rate (480Mbps) which could add more interval for uploading packets
	 * through USB bus and prepare more rx buffers which could make peer STA have stable
	 * tx rate for us
	 */
	if (is_highspeed_usb_device(hw) &&
	    sta_priv->width > NL80211_CHAN_WIDTH_40 &&
	    sta_priv->width < NL80211_CHAN_WIDTH_5) {
		hw->max_rx_aggregation_subframes = 0x20;
	} else {
		hw->max_rx_aggregation_subframes = 0x40;
	}

	if (get_ieee80211_ht_cap(sta).ht_supported) {
		unsigned long mcs = 0;

		for (i = 0; i < 2; ++i)
			mcs |= get_ieee80211_ht_cap(sta).mcs.rx_mask[i] << (8 * i);

		WARN_ON(!mcs);
		sta_priv->ht_mcs = bitmap_weight(&mcs, 16) - 1;
		sta_priv->n_ss = (sta_priv->ht_mcs >> 3) + 1;
	} else {
		goto print_info;
	}

	if (ieee80211_is_vht_sta(sta)) {
		u8 vht_mcs;

		for (i = 0; i < ARRAY_SIZE(sta_priv->he_mcs); ++i) {
			vht_mcs = 0x03 & (get_ieee80211_vht_cap(sta).vht_mcs.rx_mcs_map >> (2 * i));
			if (vht_mcs == IEEE80211_VHT_MCS_NOT_SUPPORTED)
				break;

			sta_priv->vht_mcs = 7 + vht_mcs;
		}
		sta_priv->n_ss = i;

		if (get_ieee80211_vht_cap(sta).cap & IEEE80211_VHT_CAP_HTC_VHT) {
			sta_priv->vht_la = (get_ieee80211_vht_cap(sta).cap &
					    VHT_CAP(LINK_ADAPTATION_VHT_MRQ_MFB)) >> 26;
			sta_priv->vht_msi = 0;
		}

	}

	if (!ieee80211_is_he_sta(sta))
		goto print_info;

	if (get_ieee80211_he_cap(sta).he_cap_elem.mac_cap_info[0] & HE_MAC_CAP(0, HTC_HE)) {
		if (get_ieee80211_he_cap(sta).he_cap_elem.mac_cap_info[2] &
		    HE_MAC_CAP(2, LINK_ADAPTATION))
			sta_priv->he_la |= BIT(1);
		if (get_ieee80211_he_cap(sta).he_cap_elem.mac_cap_info[1] &
		    HE_MAC_CAP(1, LINK_ADAPTATION))
			sta_priv->he_la |= BIT(0);

		sta_priv->he_msi = 0;
	}

	/* recored STA supported he rx rate */
	for (i = 0; i < ARRAY_SIZE(sta_priv->he_mcs); i++) {
		sta_priv->he_mcs[i] = 0x03 &
		    (get_ieee80211_he_cap(sta).he_mcs_nss_supp.rx_mcs_80 >> (2 * i));

		if (sta_priv->he_mcs[i] == IEEE80211_HE_MCS_NOT_SUPPORTED)
			break;
	}
	sta_priv->n_ss = i;

	/*
	 * calculate the npp
	 * 1 - reset all he_npp
	 * 2 - check ppe size if 0 Use Nominal Packet Padding field
	 * 3 - ppe size none 0 calculate the ppe for each nss/bw/constellation
	 */
	memset(sta_priv->he_npp, 0, sizeof(sta_priv->he_npp));

	he_phy_cap_info = he_cap_tmp.he_cap_elem.phy_cap_info;

	if (ieee80211_he_ppe_size(get_ieee80211_he_cap(sta).ppe_thres[0], he_phy_cap_info) == 0) {
		u16 npp;

		npp = (u16)he_phy_cap_info[78 / BITS_PER_BYTE];
		npp = FIELD_GET(IEEE80211_HE_PHY_CAP9_NOMINAL_PKT_PADDING_MASK, npp);
		npp = npp | (npp << 2) | (npp << 4) | (npp << 6) | (npp << 8) | (npp << 10);

		for (i = 0; i < min_t(int, 2, sta_priv->n_ss); i++) {
			for (j = 0; j <= CBW80; j++)
				sta_priv->he_npp[i][j] = npp;
		}
	} else {
		scm80211_calc_he_npp(hw, sta);
	}

	/*
	 * Reference P802.11ax_D6.0 26.2.1 TXOP duration-based RTS/CTS
	 * Get frame_time_rts_th from bss_conf which indicates
	 * TXOP Duration RTS Threshold in units of 32us
	 * For NON-AP HE STA it should be non-zero.
	 * the value 1023 indicates that TXOP duration-based RTS is disabled
	 */
	if (vif->type == NL80211_IFTYPE_STATION) {
		u32 rts_th = cpu_to_le16(vif->bss_conf.frame_time_rts_th);

		if (rts_th && rts_th != 1023)
			sta_priv->txop_rts_th = rts_th << 5;

		WARN_ON(!sta_priv->txop_rts_th);

		/*
		 * the mac8011 connection detection have on effect under
		 * HE STA, so start periodic timer to cherk beacon loss cnt
		 * and deside whether need to do disconnection by ourself
		 */
		vifp->beacon_timeout =
		    usecs_to_jiffies(ieee80211_tu_to_usec
				     (bcn_loss_discon_cnt * vif->bss_conf.beacon_int));
		mod_timer(&vifp->bcn_mon_timer, round_jiffies_up(jiffies + vifp->beacon_timeout));

#ifdef SCM80211_WFA_TEST
		vifp->hetb_psdu_len = 0;
#endif
	}

	/* Get Multi-TID Aggregation Rx Support info */
	sta_priv->he_multid = get_ieee80211_he_cap(sta).he_cap_elem.mac_cap_info[1] >> 4;
	sta_priv->he_multid &= 0x07;

	if ((vif->type == NL80211_IFTYPE_STATION || vif->type == NL80211_IFTYPE_P2P_CLIENT))
		scm80211_check_he_obss_narrow_bw_ru(hw, vif);

	if (atomic_inc_return(&hw_priv->n_he_sta) == 1 && !hw_priv->local_rc_used)
		hrtimer_start(&hw_priv->ratestats_timer, ms_to_ktime(MSEC_PER_SEC),
			      HRTIMER_MODE_REL_SOFT);

print_info:
	scm80211_trace(hw, "%s add a %s sta\n", __func__,
		       ieee80211_is_he_sta(sta) ? "HE" :
		       ieee80211_is_vht_sta(sta) ? "VHT" :
		       get_ieee80211_ht_cap(sta).ht_supported ? "HT" : "NON-HT");
	scm80211_trace(hw, " he rate support MCS0~%d\n", sta_priv->he_mcs[0] * 2 + 7);
	scm80211_trace(hw, " vht mcs rate support MCS0~%d\n", sta_priv->vht_mcs);
	scm80211_trace(hw, " ht mcs rate support MCS0~%d\n", sta_priv->ht_mcs);
	scm80211_trace(hw, " non-ht rate bitmap %08x\n",
		       get_ieee80211_supp_rates(sta, sta_priv->band));
	scm80211_trace(hw, " nss %d\n", sta_priv->n_ss);
	scm80211_trace(hw, " link adaptation (VHT: %d, HE: %d)\n",
		       sta_priv->vht_la, sta_priv->he_la);
	scm80211_trace(hw, " multi-tid support: %d\n", sta_priv->he_multid);
	scm80211_trace(hw, " txop duration rts threshold: %d\n", sta_priv->txop_rts_th);

#ifdef SCM80211_COEX
	if ((is_pcie_device(hw)) && (hw_priv->coex.dev_ant == 1)
		 && (sta_priv->band == NL80211_BAND_2GHZ)) {
		/* 2 ant configuration selected */
		hw_priv->coex.sta = sta;
		/* GPIO40/41 used for controlling shared antenna */
#ifdef SCM2020_A0
		scm2020_set_gpio_config(hw, 40, 5, SCM_GPIO_IN, false);
		scm2020_set_gpio_config(hw, 41, 5, SCM_GPIO_IN, false);
		if (hw_priv->n_chain == 2)
			hrtimer_start(&hw_priv->coex.coex_2ant_timer,
						ms_to_ktime(SCM80211_COEX_2ANT_MONITOR_PERIOD_1000),
						HRTIMER_MODE_REL_SOFT);
#endif
	}
#endif

	/* fixup the max amsdu setting under linux 5.10 */
	if (ieee80211_is_vht_sta(sta) && sta->max_amsdu_len < IEEE80211_MAX_MPDU_LEN_VHT_3895) {
		switch (get_ieee80211_vht_cap(sta).cap & IEEE80211_VHT_CAP_MAX_MPDU_MASK) {
		case IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_11454:
			sta->max_amsdu_len = IEEE80211_MAX_MPDU_LEN_VHT_11454;
			break;
		case IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_7991:
			sta->max_amsdu_len = IEEE80211_MAX_MPDU_LEN_VHT_7991;
			break;
		default:
			sta->max_amsdu_len = IEEE80211_MAX_MPDU_LEN_VHT_3895;
			break;
		}
	}
	for (i = 0; i < 2; ++i)
		scm80211_trace(hw, "he_npp 0x%04x %04x %04x\n",
			       sta_priv->he_npp[i][0], sta_priv->he_npp[i][1],
			       sta_priv->he_npp[i][2]);
	scm80211_trace(hw, "max amsdu len %d, max rc amsdu_len %d\n", sta->max_amsdu_len,
		       sta->max_rc_amsdu_len);

	spin_lock_init(&sta_priv->lock);

	if (vif->type == NL80211_IFTYPE_STATION) {
		u8 max_nss = min_t(u8, sta_priv->n_ss, hw_priv->n_chain);

		rcu_read_lock();
		chanctx = (void *)rcu_dereference_chanctx_conf(vif);
		rcu_read_unlock();

		if (chanctx->ctx.rx_chains_dynamic < max_nss) {
			chanctx->ctx.rx_chains_dynamic = max_nss;
			changes = BIT(PHY_ACTIVE_RXCHAIN_SET);
			scm80211_vif_update_chanctx(hw, vif, changes);
		}
	}
	return 0;
}

static int scm80211_sta_remove(struct ieee80211_hw *hw,
			       struct ieee80211_vif *vif, struct ieee80211_sta *sta)
{
	int i;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_hw_priv *hw_priv = hw->priv;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;

	scm80211_trace(hw, "+%s(vif=%pM, sta=%pM)\n", __func__, vif->addr, sta->addr);

	/* Calling scm80211_flush would be neater. However, it is postponed
	 * until this function is confirmed to be free from re-entrance issue.
	 */
#ifdef SCM80211_TWT
	if ((scm80211_get_twt_state(vif) == TWT_REQ || vifp->twt_teardown) && vifp->twt_request)
		kfree_skb(vifp->twt_request);
	vifp->twt_request = NULL;
	scm80211_set_twt_state(vif, TWT_DISABLED);
#endif /* SCM80211_TWT */

#ifndef IEEE80211_NEXT_TXQ
	/* Why is this necessary? */
	if (scm80211_vif_tx_frames_pending(hw, vif))
		msleep(1);
#endif

	scm80211_set_sta_aid(hw, vif, sta, false);

	if (vif->type == NL80211_IFTYPE_STATION && vifp->wait_beacon)
		scm80211_wake_queues(hw, scm80211_vif_queues(vif));
	for (i = 0; i < ARRAY_SIZE(sta->txq); i++)
		scm80211_deinit_txq(hw, vif, sta->txq[i]);

	atomic_dec(&vifp->num_sta);

#ifdef SCM80211_COEX
	if ((is_pcie_device(hw)) && (hw_priv->coex.dev_ant == 1)
		 && (sta_priv->band == NL80211_BAND_2GHZ)) {
#ifdef SCM2020_A0
		if (hw_priv->n_chain == 2)
			hrtimer_cancel(&hw_priv->coex.coex_2ant_timer);
		/* GPIO40/41 used for controlling shared antenna */
		scm2020_set_gpio_config(hw, 40, 1, SCM_GPIO_OUT, true);
		scm2020_set_gpio_config(hw, 41, 1, SCM_GPIO_OUT, true);
#endif
	}
#endif

	if (ieee80211_is_he_sta(sta)) {
		if (atomic_dec_return(&hw_priv->n_he_sta) == 0 && !hw_priv->local_rc_used)
			hrtimer_cancel(&hw_priv->ratestats_timer);

		if (vif->type == NL80211_IFTYPE_STATION)
			del_timer_sync(&vifp->bcn_mon_timer);
	}

	return 0;
}

static void scm80211_sta_notify(struct ieee80211_hw *hw,
				struct ieee80211_vif *vif,
				enum sta_notify_cmd cmd, struct ieee80211_sta *sta)
{
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;

	scm80211_trace(hw, "+%s(vif=%pM, sta=%pM STA_NOTIFY_%s)\n",
		       __func__, vif->addr, sta->addr, cmd ? "AWAKE" : "SLEEP");

	switch (cmd) {
	case STA_NOTIFY_SLEEP:
		sta_priv->asleep = true;
		break;
	case STA_NOTIFY_AWAKE:
		{
			u32 queues = scm80211_vif_queues(vif);

			sta_priv->asleep = false;
			scm80211_wake_queues(hw, queues);
			break;
		}
	default:
		WARN(1, "Invalid sta notify: %d\n", cmd);
		break;
	}
}

static int scm80211_set_tim(struct ieee80211_hw *hw, struct ieee80211_sta *sta, bool set)
{
	return 0;
}

int _scm80211_set_key(struct ieee80211_hw *hw, enum set_key_cmd cmd,
		      struct ieee80211_vif *vif,
		      struct ieee80211_sta *sta,
		      struct ieee80211_key_conf *key,
		      bool by_idx)
{
	enum cipher_suite cipher;
	enum nl80211_iftype iftype = ieee80211_vif_type_p2p(vif);
	int ret = 0;
	bool gtk;
	const u8 *peer;
	u32 queues;

	cipher = scm80211_cipher_suite(key->cipher);
	gtk = !(key->flags & IEEE80211_KEY_FLAG_PAIRWISE);

#ifdef ERRATA_SW_WEP
	if (ieee80211_iface_is_ap(iftype) &&
	    gtk && (cipher == CIPHER_SUITE_WEP104 || cipher == CIPHER_SUITE_WEP40)) {
		/*  softap mode default wep mapping key can not decrypt */
		ret = -EOPNOTSUPP;
		return ret;
	}
#endif
	if (!gtk && (cipher == CIPHER_SUITE_WEP104 || cipher == CIPHER_SUITE_WEP40) &&
	    !(key->keyidx == 0)) {
		/*  dot11WEPKeyMappings keyindx shall be 0 */
		ret = -EOPNOTSUPP;
		return ret;
	}

	if (!by_idx) {
		/* make sure all frames are sent out before install/remove key */
		queues = scm80211_vif_queues(vif);
		scm80211_flush(hw, vif, queues, false);
	}

	/*
	 * Key address
	 *
	 * PTK: sta != NULL, peer is set to sta->addr
	 * GTK: sta == NULL, peer is set to vif->bss_conf.bssid (both AP/STA)
	 *
	 * FIXME given correct peer when it's AP_VLAN, ADHOC or MESH ...
	 */
	if (!gtk && sta)
		peer = (const u8 *)sta->addr;
	else if (gtk && vif->type == NL80211_IFTYPE_MESH_POINT)
		peer = vif->addr;
	else
		peer = vif->bss_conf.bssid;

	if (cmd == SET_KEY) {
		ret = scm80211_install_key(hw, vif, key, peer, by_idx);
		if (ret == 0) {
			key->flags |= IEEE80211_KEY_FLAG_GENERATE_IV;
			key->flags |= IEEE80211_KEY_FLAG_SW_MGMT_TX;
		}
	} else {
		ret = scm80211_remove_key(hw, key);
	}

	return ret;
}

static int scm80211_set_key(struct ieee80211_hw *hw, enum set_key_cmd cmd,
			    struct ieee80211_vif *vif,
			    struct ieee80211_sta *sta,
			    struct ieee80211_key_conf *key)
{
	int ret;

	ret =  _scm80211_set_key(hw, cmd, vif, sta, key, false);

	return ret;
}

static void scm80211_set_default_unicast_key(struct ieee80211_hw *hw,
					     struct ieee80211_vif *vif, int idx)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	if (idx < 0 || idx > 3)
		return;
	vifp->def_key_idx = idx;
}

static void scm80211_get_sta_op_chan(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_channel **chan = data;
	struct cfg80211_chan_def def;
	int ret = -1;

	if (vif->type == NL80211_IFTYPE_STATION && vif->bss_conf.assoc)
		ret = scm80211_vif_chan(vif, &def);

	if (!ret)
		*chan = def.chan;
}

/**
 * scm2_get_survey()
 *
 * FIXME: HW does not provide the necessary information.
 * This callback is likely not to be supported.
 */
static int scm80211_get_survey(struct ieee80211_hw *hw, int idx, struct survey_info *survey)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_channel *chan = NULL;

	if (idx < 0 || idx >= ARRAY_SIZE(priv->survey_data))
		return -ENOENT;

	ieee80211_iterate_active_interfaces_atomic(hw,
						   IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_get_sta_op_chan, &chan);

	mutex_lock(&priv->mutex);
	survey->channel = priv->survey_data[idx].channel;
	if (!survey->channel) {
		mutex_unlock(&priv->mutex);
		return -ENOENT;
	}

	/* FIXME */
// #if 1
	survey->filled = (SURVEY_INFO_NOISE_DBM | SURVEY_INFO_TIME | SURVEY_INFO_TIME_BUSY);

	survey->time = jiffies_to_msecs(priv->survey_data[idx].end - priv->survey_data[idx].start);
	if (chan && survey->channel == chan) {
		/**
		 * report the fake noise as mini_noise - 1 to make acs choose
		 * the same channel as sta's and do not effect roaming cases
		 **/
		survey->noise = -93;
		survey->time_busy = 0;
	} else {
		/* FIXME how to get the real channel noise */
		survey->noise = -92;
		survey->time_busy = survey->time / 8;
	}

// #endif
	mutex_unlock(&priv->mutex);

	return 0;
}

int scm80211_ampdu_action(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			  struct ieee80211_ampdu_params *params)
{
	switch (params->action) {
	case IEEE80211_AMPDU_RX_START:
	case IEEE80211_AMPDU_RX_STOP:
		return scm80211_ampdu_rx_action(hw, vif, params);
	case IEEE80211_AMPDU_TX_START:
	case IEEE80211_AMPDU_TX_OPERATIONAL:
	case IEEE80211_AMPDU_TX_STOP_CONT:
	case IEEE80211_AMPDU_TX_STOP_FLUSH:
	case IEEE80211_AMPDU_TX_STOP_FLUSH_CONT:
		return scm80211_ampdu_tx_action(hw, vif, params);
	default:
		return -EOPNOTSUPP;
	}

	return 0;
}

static const char * const mlme_event[] = {
	[AUTH_EVENT] = "AUTH",
	[ASSOC_EVENT] = "ASSOC",
	[DEAUTH_RX_EVENT] = "RX_DEAUTH",
	[DEAUTH_TX_EVENT] = "TX_DEAUTH",
};

static const char * const mlme_status[] = {
	"successful",
	"denied",
	"timeout",
};

static void scm80211_event_mlme_callback(struct ieee80211_hw *hw,
					 struct ieee80211_vif *vif,
					 const struct ieee80211_event *event)
{
	scm80211_dbg(hw, "mlme event %s is %s reason %d\n",
		     mlme_event[event->u.mlme.data],
		     mlme_status[event->u.mlme.status], event->u.mlme.reason);
}

static void scm80211_event_bar_rx_callback(struct ieee80211_hw *hw,
					   struct ieee80211_vif *vif,
					   const struct ieee80211_event *event)
{
	scm80211_dbg(hw,
		     "BAR received from %pM, tid %d, ssn %d\n",
		     event->u.ba.sta->addr, event->u.ba.tid, event->u.ba.ssn);
}

static void scm80211_event_frame_timeout_callback(struct ieee80211_hw *hw,
						  struct ieee80211_vif *vif,
						  const struct ieee80211_sta *sta, u16 tid)
{
	scm80211_dbg(hw, "Frame from %pM timed out, tid %d\n", sta->addr, tid);
}

static void scm80211_event_callback(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif, const struct ieee80211_event *event)
{
	switch (event->type) {
	case MLME_EVENT:
		scm80211_event_mlme_callback(hw, vif, event);
		break;
	case BAR_RX_EVENT:
		scm80211_event_bar_rx_callback(hw, vif, event);
		break;
	case BA_FRAME_TIMEOUT:
		scm80211_event_frame_timeout_callback(hw, vif, event->u.ba.sta, event->u.ba.tid);
		break;
	default:
		break;
	}
}

static int scm80211_get_txpower(struct ieee80211_hw *hw, struct ieee80211_vif *vif, int *dbm)
{
	u32 v;

	v = scm80211_mac_readl(hw, MAC_REG_DBG_LAST_TXVEC0);
	*dbm = FIELD_GET(DBG_LAST_TXVEC0_TXPWR_LEVEL_INDEX_MASK, v);
	return 0;
}

#if KERNEL_VERSION(5, 14, 0) <= MAC80211_BASE
static void scm80211_mgd_prepare_tx(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif, struct ieee80211_prep_tx_info *info)
#else
static void scm80211_mgd_prepare_tx(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif, u16 duration)
#endif
{
	/**
	 *when in auth and assoc, we have stay in vif's assgined chanctx
	 * only for P2P Client, we need to check P2P GO's absent state
	 */
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	enum nl80211_iftype iftype = ieee80211_vif_type_p2p(vif);
	struct ieee80211_noa_data *noa;
	u32 tsf32;
	int delta;
	unsigned long flags;

	if (iftype != NL80211_IFTYPE_P2P_CLIENT)
		return;

	spin_lock_irqsave(&vifp->lock, flags);
	noa = &vifp->noa;

	if (noa->absent == 0) {
		spin_unlock_irqrestore(&vifp->lock, flags);
		return;
	}

	tsf32 = scm80211_get_tsf32(hw, vif);
	delta = noa->next_tsf - tsf32;
	spin_unlock_irqrestore(&vifp->lock, flags);
	if (delta < 0)
		return;
	msleep((round_up(delta, 1000) + P2P_PS_STOP_TIME) / 1000);
}

#ifdef SCM80211_RFKILL_POLL

static bool scm80211_radio_enabled(struct ieee80211_hw *hw)
{
#ifdef SCM2020_A0
	/* TODO if module updated need to be updated here */
	if (is_pcie_device(hw)) {
#if PCIE_CARD_VERSION == 5
		return scm2020_gpio_get_value(hw, WIFI_SYS_EN);
#endif
	} else if (is_usb_device(hw)) {
		/* FIXME */
#if USB_CARD_VERSION == 0xFF
		return scm2020_gpio_get_value(hw, WIFI_SYS_EN);
#endif
	}
#endif
	return true;
}

static void scm80211_rfkill_poll(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	bool rfkill_off = scm80211_radio_enabled(hw);

	wiphy_rfkill_set_hw_state(hw->wiphy, !priv->rfkill_off);
	if (priv->rfkill_off != rfkill_off) {
		scm80211_trace(hw, "rfkill set radio %s\n", rfkill_off ? "on" : "off");
		wiphy_rfkill_set_hw_state(hw->wiphy, !rfkill_off);
		priv->rfkill_off = rfkill_off;
	}
}
#endif /* SCM80211_RFKILL_POLL */

void scm80211_rfkill_init(struct ieee80211_hw *hw)
{
#ifdef SCM80211_RFKILL_POLL
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->rfkill_off = true;
	wiphy_rfkill_set_hw_state(hw->wiphy, !priv->rfkill_off);
	wiphy_rfkill_start_polling(hw->wiphy);
#endif /* SCM80211_RFKILL_POLL */
}

void scm80211_rfkill_deinit(struct ieee80211_hw *hw)
{
#ifdef SCM80211_RFKILL_POLL
	if (!hw)
		return;
	wiphy_rfkill_stop_polling(hw->wiphy);
#endif /* SCM80211_RFKILL_POLL */
}

static const char scm80211_gstrings_stats[][ETH_GSTRING_LEN] = {
	"tx_pkts_nic",
	"tx_bytes_nic",
	"rx_pkts_nic",
	"rx_bytes_nic",
	"d_tx_dropped",
	"d_tx_failed",
	"d_ps_mode",
	"d_group",
};

#define SCM80211_SSTATS_LEN ARRAY_SIZE(scm80211_gstrings_stats)

static void scm80211_get_et_strings(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif, u32 sset, u8 *data)
{
	if (sset != ETH_SS_STATS)
		return;

	memcpy(data, *scm80211_gstrings_stats, sizeof(scm80211_gstrings_stats));
}

static int scm80211_get_et_sset_count(struct ieee80211_hw *hw, struct ieee80211_vif *vif, int sset)
{
	if (sset == ETH_SS_STATS)
		return SCM80211_SSTATS_LEN;
	return 0;
}

static void scm80211_get_et_stats(struct ieee80211_hw *hw,
				  struct ieee80211_vif *vif,
				  struct ethtool_stats *stats, u64 *data)
{
	struct ieee80211_hw_priv *ar = hw->priv;
	int i = 0;

	data[i++] = ar->tx_pkts;
	data[i++] = ar->tx_bytes;
	data[i++] = ar->rx_pkts;
	data[i++] = ar->rx_bytes;
	data[i++] = ar->tx_dropped;
	data[i++] = ar->tx_failed;
	data[i++] = ar->pm.ps;
	data[i++] = ar->group;

	WARN_ON(i != SCM80211_SSTATS_LEN);
}

static void scm80211_get_ringparam(struct ieee80211_hw *hw,
				   u32 *tx, u32 *tx_max, u32 *rx, u32 *rx_max)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	*rx_max = SCM80211_NUM_RX_RING;
	*rx = 1 << priv->rxring.order;

	*tx = *tx_max = SCM80211_MAX_MPDU * SCM80211_NUM_HWQ;
}

static int scm80211_get_stats(struct ieee80211_hw *hw, struct ieee80211_low_level_stats *stats)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	memcpy(stats, &priv->stats, sizeof(*stats));
	return 0;
}

/**
 * Reference to Table 9-97—Coverage Class field parameters
 * It will be called in set_wiphy_params and ieee80211_reconfig
 * when uplayer set NL80211_ATTR_WIPHY_COVERAGE_CLASS or
 * NL80211_ATTR_WIPHY_DYN_ACK attribute. Currently wpa_supplicant
 * doesn't set this so this callback
 * Maybe unused callback
 **/
static void scm80211_vif_set_coverage_class(void *data, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_hw *hw = data;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif_priv *iface = ieee80211_to_priv(vif);
	u32 v;

	iface->slot += 3 * priv->coverage_class;

	if (version(scm80211_mac_version(hw)) >= 0x16) {
		v = scm80211_mac_readl(hw, MAC_REG_DEV_CFG) & ~DEV_CFG_SLOT_TIME;
		v |= FIELD_PREP(DEV_CFG_SLOT_TIME, iface->slot);
		scm80211_mac_writel(hw, v, MAC_REG_DEV_CFG);
	} else {
		v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG0(iface->index));
		v &= ~VIF_CFG0_SLOT_TIME;
		v |= FIELD_PREP(VIF_CFG0_SLOT_TIME, iface->slot);
		scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG0(iface->index));
	}
}

static void scm80211_set_coverage_class(struct ieee80211_hw *hw, s16 coverage_class)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	scm80211_trace(hw, "set coverage_class %d\n", coverage_class);

	if (coverage_class < 0)
		return;

	priv->coverage_class = clamp_t(u8, coverage_class, 0, 31);

	ieee80211_iterate_active_interfaces(hw, IEEE80211_IFACE_ITER_NORMAL,
					    scm80211_vif_set_coverage_class, hw);
}

static void scm80211_sta_statistics(struct ieee80211_hw *hw,
				    struct ieee80211_vif *vif,
				    struct ieee80211_sta *sta, struct station_info *sinfo)
{
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	unsigned long flags;
	struct rate_info *txinfo = &sta_priv->txinfo;

	spin_lock_irqsave(&sta_priv->lock, flags);
	if (txinfo->flags & RATE_INFO_FLAGS_HE_MCS) {
		memcpy(&sinfo->txrate, txinfo, sizeof(*txinfo));
		sinfo->filled |= BIT_ULL(NL80211_STA_INFO_TX_BITRATE);
	}

	/* TODO: Add more */

	spin_unlock_irqrestore(&sta_priv->lock, flags);
}

void scm80211_handle_muedca_irq(struct ieee80211_hw *hw, u32 val)
{
	struct ieee80211_vif *vif = NULL;
	struct scm80211_hwq *hwq = NULL;
	int vdev;
	unsigned long vifs = val;

	for_each_set_bit(vdev, &vifs, SCM80211_NUM_VIF) {
#ifdef SCM80211_WFA_TEST
		unsigned long queues;
		unsigned long flags;
		int i;
		struct ieee80211_vif_priv *vifp;
#else
		u32 reg, ac;
#endif
		vif = scm80211_get_vif_from_vedv(hw, vdev);

		if (!vif)
			continue;
#ifdef SCM80211_WFA_TEST
		vifp = ieee80211_to_priv(vif);
		queues = scm80211_vif_queues(vif) & (~BIT(vifp->mgmt_hw_queue));
		for_each_set_bit(i, &queues, hw->queues) {
			hwq = scm80211_hwq(hw, i);
#ifdef IEEE80211_NEXT_TXQ

			scm80211_enable_mu_edca(hwq);
#else
			if (list_first_entry_or_null(&hwq->txq_head, struct scm80211_txq, txq_list))
				scm80211_enable_mu_edca(hwq);
#endif
		}

		spin_lock_irqsave(&vifp->lock, flags);
		vifp->hetb_psdu_len = scm80211_mac_readl(hw, MAC_REG_VIF_BASIC_TRIG_PSDU_LEN(vdev));
		spin_unlock_irqrestore(&vifp->lock, flags);

#else
		reg = scm80211_mac_readl(hw, MAC_REG_TX_HETB_QNULL);

		if (vdev == 0)
			ac = FIELD_GET(TX_HETB_QNULL_RX_BASIC_TRIGGER_FIRST_AC_VIF0, reg);
		else
			ac = FIELD_GET(TX_HETB_QNULL_RX_BASIC_TRIGGER_FIRST_AC_VIF1, reg);
		hwq = scm80211_hwq(hw, vif->hw_queue[SCM80211_AC_TO_IEEE80211_AC(ac)]);

		scm80211_enable_mu_edca(hwq);
#endif
	}
}

static int scm80211_alloc_sband(struct wiphy *wiphy, enum nl80211_band band,
				const struct ieee80211_channel *chans,
				int n_chans, const struct ieee80211_rate *rates, int n_rates)
{
	struct ieee80211_supported_band *sband;
	struct device *dev = &wiphy->dev;
	size_t size;

	sband = devm_kzalloc(dev, sizeof(*sband), GFP_KERNEL);
	if (!sband)
		goto fail;
	sband->band = band;

	size = sizeof(struct ieee80211_channel) * n_chans;
	sband->channels = devm_kmemdup(dev, chans, size, GFP_KERNEL);
	if (!sband->channels)
		goto fail;
	sband->n_channels = n_chans;

	size = sizeof(struct ieee80211_rate) * n_rates;
	sband->bitrates = devm_kmemdup(dev, rates, size, GFP_KERNEL);
	if (!sband->bitrates)
		goto fail;
	sband->n_bitrates = n_rates;

	wiphy->bands[band] = sband;
	return 0;
fail:
	return -ENOMEM;
}

void scm80211_set_debug_trace(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->debug_trace = (u16)(((u32)debug_trace) & 0xFFFF);
}

int scm80211_hw_band_set(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

#ifndef SCM80211_FTM
	if (priv->dbdc)
		priv->band = (priv->devid == 0) ? BIT(NL80211_BAND_2GHZ) : BIT(NL80211_BAND_5GHZ);
	else
#endif
		priv->band = BIT(NL80211_BAND_2GHZ) | BIT(NL80211_BAND_5GHZ);

	scm80211_trace(hw, "%s: devid=%d band=0x%x\n", __func__, priv->devid, priv->band);
	return 0;
}

/**
 *  Content of PPE Thresholds field for different band and n_ss
 */
static u8 ppe_thres[][IEEE80211_HE_PPE_THRES_MAX_LEN] = {
	/*  ---------- 2.4G Band --------------- */
	/*  ----------   1-NSS   ---------------
	 * BIT(n)         <--------          BIT(0)
	 * [00000]   [[111000] [111000]] [0011] [000]
	 *   PAD          PPE Info       RU-MASK NSS
	 *          PPET8=None PPET16=0  242/484  1
	 **/
	[0] = { 0x18, 0x1c, 0x07 },

	/*  ----------   2-NSS   ---------------
	 * BIT(n)         <--------          BIT(0)
	 *  [0]   [[111000] .. [111000]] [0011] [001]
	 *  PAD         PPE Info         RU-MASK NSS
	 *            PPET8=None PPET16=0  242/484  2
	 */
	[1] = { 0x19, 0x1c, 0xc7, 0x71 },

	/*  ----------  5G Band  --------------- */
	/*  ----------   1-NSS   ---------------
	 * BIT(n)            <--------          BIT(0)
	 *  [0000000]   [[111000] . [111000]] [0111]   [000]
	 *    PAD           PPE Info          RU-MASK   NSS
	 *               PPET8=None PPET16=0  242/484/996 1
	 **/
	[2] = { 0x38, 0x1c, 0xc7, 0x01 },

	/*  ----------   2-NSS   ---------------
	 * BIT(n)               <--------           BIT(0)
	 *[00000]   [[111000] .... [111000]] [0111]   [001]
	 *   PAD           PPE Info          RU-MASK   NSS
	 *           PPET8=None PPET16=0      242/484/996 2
	 **/
	[3] = { 0x39, 0x1c, 0xc7, 0x71, 0x1c, 0x07 },
};

static int scm80211_init_sband_2ghz(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_supported_band *sband;
	struct ieee80211_sband_iftype_data *iftype_data;
	int i, err;
	u16 mask = 0xffff;

	if (!(priv->band & BIT(NL80211_BAND_2GHZ))) {
		scm80211_info(hw, "%snot support 2G band on hwmac#%d\n",
			      priv->dbdc ? "[DBDC used] " : "", priv->devid);
		return 0;
	}

	err = scm80211_alloc_sband(hw->wiphy, NL80211_BAND_2GHZ,
				   scm80211_chan_2ghz,
				   ARRAY_SIZE(scm80211_chan_2ghz),
				   scm80211_rates, ARRAY_SIZE(scm80211_rates));

	if (err)
		return err;

	sband = hw->wiphy->bands[NL80211_BAND_2GHZ];
	memcpy(&sband->ht_cap, &scm80211_ht_cap, sizeof(scm80211_ht_cap));
	for (i = priv->n_chain; i < sizeof(sband->ht_cap.mcs.rx_mask); i++)
		sband->ht_cap.mcs.rx_mask[i] = 0;

	iftype_data = devm_kmemdup(wiphy_dev(hw->wiphy),
				   scm80211_he_cap_2ghz, sizeof(scm80211_he_cap_2ghz), GFP_KERNEL);
	if (!iftype_data)
		return -ENOMEM;

	mask = 0xffff << (2 * priv->n_chain);
	mask = cpu_to_le16(mask);

	sband->n_iftype_data = ARRAY_SIZE(scm80211_he_cap_2ghz);
	for (i = 0; i < sband->n_iftype_data; i++) {
		struct ieee80211_sta_he_cap *cap = &iftype_data[i].he_cap;

		cap->he_mcs_nss_supp.rx_mcs_80 |= mask;
		cap->he_mcs_nss_supp.tx_mcs_80 |= mask;
		memcpy(cap->ppe_thres, ppe_thres[priv->n_chain - 1],
		       IEEE80211_HE_PPE_THRES_MAX_LEN);
	}

	sband->iftype_data = iftype_data;
	priv->phy_mandatory_rates[NL80211_BAND_2GHZ] = 0;
	scm80211_trace(hw, "hwmac#%d init 2Ghz capability\n", priv->devid);
	return 0;
}

static int scm80211_init_sband_5ghz(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_supported_band *sband;
	struct ieee80211_sband_iftype_data *iftype_data;
	int i, err;
	u32 mask;

	if (!(priv->band & BIT(NL80211_BAND_5GHZ))) {
		scm80211_info(hw, "%snot support 5G band on hwmac#%d\n",
			      priv->dbdc ? "[DBDC used] " : "", priv->devid);
		return 0;
	}
	err = scm80211_alloc_sband(hw->wiphy, NL80211_BAND_5GHZ,
				   scm80211_chan_5ghz,
				   ARRAY_SIZE(scm80211_chan_5ghz),
				   scm80211_rates + 4, ARRAY_SIZE(scm80211_rates) - 4);
	if (err)
		return err;

	sband = hw->wiphy->bands[NL80211_BAND_5GHZ];
	memcpy(&sband->ht_cap, &scm80211_ht_cap, sizeof(scm80211_ht_cap));

	for (i = priv->n_chain; i < sizeof(sband->ht_cap.mcs.rx_mask); i++)
		sband->ht_cap.mcs.rx_mask[i] = 0;

	mask = (0xffff) << (2 * priv->n_chain);
	mask = cpu_to_le16(mask);

	memcpy(&sband->vht_cap, &scm80211_vht_cap, sizeof(scm80211_vht_cap));
	/* downgrade the cap to only support recieving VHT-MCS7 */
	if (is_highspeed_usb_device(hw))
		sband->vht_cap.vht_mcs.rx_mcs_map = 0xFFF0;

	sband->vht_cap.vht_mcs.rx_mcs_map |= mask;
	sband->vht_cap.vht_mcs.tx_mcs_map |= mask;

	/*
	 * use MAX-MPDU 3895 bytes on usb aim to aggragate more mpdus and improve the
	 * mpdu sent success probability
	 */
	if (is_usb_device(hw)) {
		sband->vht_cap.cap &= ~IEEE80211_VHT_CAP_MAX_MPDU_MASK;
		sband->vht_cap.cap |= IEEE80211_VHT_CAP_MAX_MPDU_LENGTH_3895;
	}

	if (priv->n_chain <= 1)
		sband->vht_cap.cap &= ~(IEEE80211_VHT_CAP_RXSTBC_2);
	iftype_data = devm_kmemdup(wiphy_dev(hw->wiphy),
				   scm80211_he_cap_5ghz, sizeof(scm80211_he_cap_5ghz), GFP_KERNEL);
	if (!iftype_data)
		return -ENOMEM;

	sband->n_iftype_data = ARRAY_SIZE(scm80211_he_cap_5ghz);
	for (i = 0; i < sband->n_iftype_data; i++) {
		struct ieee80211_sta_he_cap *cap = &iftype_data[i].he_cap;

		/* downgrade the cap to only support recieving HE-MCS7 */
		if (is_highspeed_usb_device(hw))
			cap->he_mcs_nss_supp.rx_mcs_80 = 0xfff0;

		cap->he_mcs_nss_supp.rx_mcs_80 |= mask;
		cap->he_mcs_nss_supp.tx_mcs_80 |= mask;
		memcpy(cap->ppe_thres, ppe_thres[priv->n_chain + 1],
		       IEEE80211_HE_PPE_THRES_MAX_LEN);
	}

	sband->iftype_data = iftype_data;
	priv->phy_mandatory_rates[NL80211_BAND_5GHZ] = 0;

	scm80211_trace(hw, "hwmac#%d init 5Ghz capability succ\n", priv->devid);
	return 0;
}

/**
 * DOC: ieee80211_hw_flags
 *
 * set:
 * IEEE80211_HW_HOST_BROADCAST_PS_BUFFERING
 * IEEE80211_HW_SIGNAL_DBM
 * IEEE80211_HW_AMPDU_AGGREGATION
 * IEEE80211_HW_SUPPORTS_PS
 * IEEE80211_HW_PS_NULLFUNC_STACK
 * IEEE80211_HW_SUPPORTS_DYNAMIC_PS,
 * IEEE80211_HW_WANT_MONITOR_VIF,
 * IEEE80211_HW_SUPPORT_FAST_XMIT (FIXME)
 * IEEE80211_HW_SUPPORTS_RC_TABLE,
 * IEEE80211_HW_REPORTS_TX_ACK_STATUS,
 * IEEE80211_HW_QUEUE_CONTROL,
 * IEEE80211_HW_P2P_DEV_ADDR_FOR_INTF,
 * IEEE80211_HW_SUPPORTS_HT_CCK_RATES,
 * IEEE80211_HW_SUPPORTS_CLONED_SKBS,
 * IEEE80211_HW_TX_FRAG_LIST, (FIXME)
 * IEEE80211_HW_REPORTS_LOW_ACK, (FIXME)
 * IEEE80211_HW_BUFF_MMPDU_TXQ
 * IEEE80211_HW_STA_MMPDU_TXQ
 * IEEE80211_HW_SINGLE_SCAN_ON_ALL_BANDS,
 * IEEE80211_HW_MFP_CAPABLE,
 *
 * unset:
 * IEEE80211_HW_HAS_RATE_CONTROL
 * IEEE80211_HW_RX_INCLUDES_FCS
 * IEEE80211_HW_SIGNAL_UNSPEC,
 * IEEE80211_HW_NO_AUTO_VIF,
 * IEEE80211_HW_SW_CRYPTO_CONTROL,
 * IEEE80211_HW_CONNECTION_MONITOR,
 * IEEE80211_HW_SUPPORTS_PER_STA_GTK,
 * IEEE80211_HW_AP_LINK_PS,
 * IEEE80211_HW_TX_AMPDU_SETUP_IN_HW,
 * IEEE80211_HW_SUPPORTS_REORDERING_BUFFER,
 * IEEE80211_HW_SUPPORTS_TX_FRAG,
 *
 * TBD:
 * IEEE80211_HW_NEED_DTIM_BEFORE_ASSOC,
 * IEEE80211_HW_SPECTRUM_MGMT,
 * IEEE80211_HW_TIMING_BEACON_ONLY,
 * IEEE80211_HW_CHANCTX_STA_CSA,
 * IEEE80211_HW_TDLS_WIDER_BW,
 * IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU,
 * IEEE80211_HW_BEACON_TX_STATUS,
 * IEEE80211_HW_NEEDS_UNIQUE_STA_ADDR,
 * IEEE80211_HW_USES_RSS,
 * IEEE80211_HW_TX_AMSDU,
 * IEEE80211_HW_SUPPORTS_TDLS_BUFFER_STA,
 * IEEE80211_HW_DEAUTH_NEED_MGD_TX_PREP,
 * IEEE80211_HW_DOESNT_SUPPORT_QOS_NDP,
 * IEEE80211_HW_SUPPORTS_VHT_EXT_NSS_BW,
 */

static enum ieee80211_hw_flags scm80211_hw_flags[] = {
	IEEE80211_HW_HOST_BROADCAST_PS_BUFFERING,
	IEEE80211_HW_SIGNAL_DBM,
	IEEE80211_HW_AMPDU_AGGREGATION,
#ifdef SCM80211_PS
	IEEE80211_HW_SUPPORTS_PS,
	IEEE80211_HW_SUPPORTS_DYNAMIC_PS,
	IEEE80211_HW_PS_NULLFUNC_STACK,
#endif
	IEEE80211_HW_TIMING_BEACON_ONLY,
	IEEE80211_HW_WANT_MONITOR_VIF,
	IEEE80211_HW_SUPPORT_FAST_XMIT,	/* FIXME */
	IEEE80211_HW_SUPPORTS_RC_TABLE,
	IEEE80211_HW_REPORTS_TX_ACK_STATUS,
	IEEE80211_HW_QUEUE_CONTROL,
	IEEE80211_HW_P2P_DEV_ADDR_FOR_INTF,
	IEEE80211_HW_SUPPORTS_HT_CCK_RATES,
	IEEE80211_HW_SUPPORTS_CLONED_SKBS,
	IEEE80211_HW_REPORTS_LOW_ACK,	/* FIXME */
/*	IEEE80211_HW_BUFF_MMPDU_TXQ,	*/
/*	IEEE80211_HW_STA_MMPDU_TXQ,	*/
	IEEE80211_HW_CHANCTX_STA_CSA,
	IEEE80211_HW_SUPPORTS_MULTI_BSSID,
	IEEE80211_HW_SUPPORTS_ONLY_HE_MULTI_BSSID,
	IEEE80211_HW_DEAUTH_NEED_MGD_TX_PREP,
	IEEE80211_HW_SINGLE_SCAN_ON_ALL_BANDS,
	IEEE80211_HW_MFP_CAPABLE,
	IEEE80211_HW_TX_FRAG_LIST,
#ifndef SCM80211_SHRINK_MEM
	IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU,
	IEEE80211_HW_TX_AMSDU,
#endif
	IEEE80211_HW_TDLS_WIDER_BW,
};

/* FIXME is that right? */
static const u8 he_if_types_ext_capa_sta[] = {
	[0] = WLAN_EXT_CAPA1_EXT_CHANNEL_SWITCHING,
	[2] = WLAN_EXT_CAPA3_MULTI_BSSID_SUPPORT,
	[7] = WLAN_EXT_CAPA8_OPMODE_NOTIF,
	[9] = WLAN_EXT_CAPA10_TWT_REQUESTER_SUPPORT,
};

static const u8 he_if_types_ext_capa_ap[] = {
	[0] = WLAN_EXT_CAPA1_EXT_CHANNEL_SWITCHING,
	[2] = WLAN_EXT_CAPA3_MULTI_BSSID_SUPPORT,
	[7] = WLAN_EXT_CAPA8_OPMODE_NOTIF,
};

static const struct wiphy_iftype_ext_capab he_iftypes_ext_capa[] = {
	{
	 .iftype = NL80211_IFTYPE_STATION,
	 .extended_capabilities = he_if_types_ext_capa_sta,
	 .extended_capabilities_mask = he_if_types_ext_capa_sta,
	 .extended_capabilities_len = sizeof(he_if_types_ext_capa_sta),
	  },
	{
	 .iftype = NL80211_IFTYPE_AP,
	 .extended_capabilities = he_if_types_ext_capa_ap,
	 .extended_capabilities_mask = he_if_types_ext_capa_ap,
	 .extended_capabilities_len = sizeof(he_if_types_ext_capa_ap),
	  },
};

int scm80211_deinit_hw(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	del_timer_sync(&priv->beacon_timer);
	scm80211_deinit_pm(hw);

	if (is_pcie_device(hw))
		hrtimer_cancel(&priv->phy_timer);
#ifdef SCM80211_DFS_CERTIFIED
	scm80211_dfs_reg_init(hw);
#endif
	ieee80211_free_hw(hw);
	return 0;
}

static void scm80211_antenna_config(struct ieee80211_hw *hw, int antenna)
{
#ifdef SCM2020_A0
#ifdef EFUSE_ENABLE
	if (!antenna)
		dev_ant = (scm80211_efuse_get_ant_sel() == 0) ? 2 : 1;
#endif
	if (antenna == 2) {
		u32 v;

		v = scm80211_raw_readl(hw, WCM_GPIO_MODE_BASE + 0x18);
		v &= 0xFFFF;
		v |= 0x01100000;
		scm80211_raw_writel(hw, v, WCM_GPIO_MODE_BASE + 0x18);
	} else {
		scm2020_ant_pinctrl(hw);
	}
#endif
}

int scm80211_init_hw(struct ieee80211_hw *hw, u8 *addr)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int i, ret;

	for (i = 0; i < ARRAY_SIZE(scm80211_hw_flags); i++) {
		//Currently, disable PS when dbdc is enabled.
		if (priv->dbdc && (scm80211_hw_flags[i] == IEEE80211_HW_SUPPORTS_PS ||
				   scm80211_hw_flags[i] == IEEE80211_HW_PS_NULLFUNC_STACK ||
				   scm80211_hw_flags[i] == IEEE80211_HW_SUPPORTS_DYNAMIC_PS)) {
			scm80211_trace(hw, "hwmac#%d skip setting hw flags=0x%x\n",
				       priv->devid, scm80211_hw_flags[i]);
			continue;
		}

		/* disable frag skb and amsdu rx for usb device */
		if ((!is_usb_device(hw) && !is_sdio_device(hw)) ||
		    (scm80211_hw_flags[i] != IEEE80211_HW_SUPPORTS_AMSDU_IN_AMPDU &&
		     scm80211_hw_flags[i] != IEEE80211_HW_TX_FRAG_LIST))
			_ieee80211_hw_set(hw, scm80211_hw_flags[i]);
	}

	scm80211_init_pm(hw);

	hw->wiphy->flags |= WIPHY_FLAG_SUPPORTS_TDLS;
	hw->wiphy->flags |= WIPHY_FLAG_HAS_REMAIN_ON_CHANNEL;
	hw->wiphy->flags |= WIPHY_FLAG_AP_UAPSD;
	hw->wiphy->flags |= WIPHY_FLAG_HAS_CHANNEL_SWITCH;
	hw->wiphy->flags |= WIPHY_FLAG_IBSS_RSN;	/* for MESH */

	hw->wiphy->features |= NL80211_FEATURE_ACTIVE_MONITOR;
	hw->wiphy->features |= NL80211_FEATURE_AP_MODE_CHAN_WIDTH_CHANGE;
	hw->wiphy->features |= NL80211_FEATURE_SCAN_RANDOM_MAC_ADDR;
	hw->wiphy->features |= NL80211_FEATURE_TDLS_CHANNEL_SWITCH;
#ifdef SCM80211_SMPS
	hw->wiphy->features |= (NL80211_FEATURE_DYNAMIC_SMPS | NL80211_FEATURE_STATIC_SMPS);
#endif
	wiphy_ext_feature_set(hw->wiphy, NL80211_EXT_FEATURE_VHT_IBSS);
	wiphy_ext_feature_set(hw->wiphy, NL80211_EXT_FEATURE_STA_TX_PWR);
	wiphy_ext_feature_set(hw->wiphy, NL80211_EXT_FEATURE_BEACON_PROTECTION);

	hw->wiphy->interface_modes = (BIT(NL80211_IFTYPE_STATION) |
				      BIT(NL80211_IFTYPE_AP) | BIT(NL80211_IFTYPE_MESH_POINT) | 0);

	/* for hw which only support 5G band, do not support p2p */
	if (priv->band & BIT(NL80211_BAND_2GHZ))
		hw->wiphy->interface_modes |= (BIT(NL80211_IFTYPE_P2P_CLIENT) |
					       BIT(NL80211_IFTYPE_P2P_GO) |
#ifdef SCM80211_P2P_DEVICE
					       BIT(NL80211_IFTYPE_P2P_DEVICE) |
#endif
					       0);
	hw->queues = SCM80211_NUM_HWQ;
	hw->offchannel_tx_hw_queue = SCM80211_HWQ_VO;
	hw->max_rates = 4;
	hw->max_rate_tries = 3;	/* each rate try max 3 times */
	hw->max_tx_fragments = SCM80211_SKB_MAX_FRAGS;
	hw->tx_sk_pacing_shift = 5;
	hw->extra_tx_headroom = sizeof(struct txvector) + IEEE80211_HT_CTL_LEN;
	hw->vif_data_size = sizeof(struct ieee80211_vif_priv);
	hw->sta_data_size = sizeof(struct ieee80211_sta_priv);
	hw->chanctx_data_size = scm80211_chanctx_data_size();
	hw->txq_data_size = scm80211_txq_data_size();
#ifdef CONFIG_SCM80211_RC
	hw->rate_control_algorithm = "scm80211_rc";
	priv->local_rc_used = true;
#else
	priv->local_rc_used = false;
	hrtimer_init(&priv->ratestats_timer, CLOCK_MONOTONIC, HRTIMER_MODE_ABS_SOFT);
	priv->ratestats_timer.function = scm80211_ratestats_timer;
#endif
	ret = scm80211_init_sband_2ghz(hw);
	if (ret)
		goto fail;

	ret = scm80211_init_sband_5ghz(hw);
	if (ret) {
		pr_debug("could not init 5ghz band\n");
		goto fail;
	}
	hw->wiphy->n_cipher_suites = ARRAY_SIZE(scm80211_ciphers);
	hw->wiphy->cipher_suites = scm80211_ciphers;
#ifdef SCM80211_WOW
	hw->wiphy->wowlan = &scm80211_wowlan_support;
#endif
	hw->wiphy->iftype_ext_capab = he_iftypes_ext_capa;
	hw->wiphy->num_iftype_ext_capab = ARRAY_SIZE(he_iftypes_ext_capa);

#ifdef SCM80211_USES_CHANCTX_OLD
	hw->wiphy->iface_combinations = scm80211_if_comb;
	hw->wiphy->n_iface_combinations = ARRAY_SIZE(scm80211_if_comb);
#endif
	scm80211_register_vendor_cmds(hw);

	hw->wiphy->max_scan_ssids = 4;
	hw->wiphy->max_scan_ie_len = IEEE80211_MAX_DATA_LEN;
	hw->wiphy->max_remain_on_channel_duration = 5000;

#ifndef SCM80211_HOSTAP_AUTO_TEST
	hw->wiphy->regulatory_flags |= REGULATORY_CUSTOM_REG;
	wiphy_apply_custom_regulatory(hw->wiphy, &scm80211_regdom);
#endif /* !SCM80211_HOSTAP_AUTO_TEST */

	mutex_init(&priv->mutex);
	spin_lock_init(&priv->lock);

	scm80211_init_chanctx_sched(hw);

	timer_setup(&priv->beacon_timer, scm80211_beacon, 0);

	if (is_pcie_device(hw)) {
		priv->space = MT_HST_BD_HST;

		hrtimer_init(&priv->phy_timer, CLOCK_MONOTONIC, HRTIMER_MODE_ABS_SOFT);
		priv->phy_timer.function = scm80211_phy;
		hrtimer_start(&priv->phy_timer, ns_to_ktime(100 * 100 * 1000),
			      HRTIMER_MODE_REL_SOFT);
	}

	atomic_set(&priv->n_he_sta, 0);

	sprintf(priv->version_str, "%s Built Time: %s", DRIVER_VERSION, DRIVER_DATE);

	SET_IEEE80211_PERM_ADDR(hw, addr);
	for (i = 0; i <= SCM80211_NUM_VIF; i++)
		memcpy(&priv->addresses[i], addr, ETH_ALEN);
	priv->addresses[1].addr[3] ^= BIT(7);	//convert the bit(7) of mac->addr[3]
	priv->addresses[2].addr[3] ^= BIT(3);	//convert the bit(3) of mac->addr[3]
	hw->wiphy->n_addresses = SCM80211_NUM_VIF + 1;
	hw->wiphy->addresses = priv->addresses;

	spin_lock_init(&priv->napi.lock);
	init_dummy_netdev(&priv->napi.netdev);

#ifdef SCM80211_DFS_CERTIFIED
	scm80211_dfs_reg_init(hw);
#endif
	/* FIXME: scan parameters such as hw->wiphy->max_scan_ssids */
	/* FIXME: iface combination */
	scm80211_antenna_config(hw, dev_ant);

#ifdef SCM80211_COEX
	priv->coex.hw = hw;
	priv->coex.dev_ant = dev_ant;
#endif

	return 0;
fail:
	return ret;
}

struct ieee80211_ops scm80211_ops = {
	.tx				= scm80211_tx,
	.wake_tx_queue			= scm80211_wake_tx_queue,
	.start				= scm80211_start,
	.stop				= scm80211_stop,
	.add_interface			= scm80211_add_interface,
	.change_interface		= scm80211_change_interface,
	.remove_interface		= scm80211_remove_interface,
	.config				= scm80211_config,
	.configure_filter		= scm80211_configure_filter,
	.bss_info_changed		= scm80211_bss_info_changed,
	.sta_add			= scm80211_sta_add,
	.sta_remove			= scm80211_sta_remove,
	.sta_notify			= scm80211_sta_notify,
	.set_tim			= scm80211_set_tim,
	.set_key			= scm80211_set_key,
	.set_default_unicast_key	= scm80211_set_default_unicast_key,
	.conf_tx			= scm80211_conf_tx,
	.sta_set_txpwr			= scm80211_sta_set_txpwr,
	.get_survey			= scm80211_get_survey,
	.ampdu_action			= scm80211_ampdu_action,
	.flush				= scm80211_flush,
	.get_tsf			= scm80211_get_tsf,
	.set_tsf			= scm80211_set_tsf,
	.offset_tsf			= scm80211_offset_tsf,
	.reset_tsf			= scm80211_reset_tsf,
	.get_et_sset_count		= scm80211_get_et_sset_count,
	.get_et_stats			= scm80211_get_et_stats,
	.get_et_strings			= scm80211_get_et_strings,
	.get_ringparam			= scm80211_get_ringparam,
	.get_stats			= scm80211_get_stats,
	.hw_scan			= scm80211_hw_scan,
	.cancel_hw_scan			= scm80211_cancel_hw_scan,
	.add_chanctx			= scm80211_add_chanctx,
	.remove_chanctx			= scm80211_remove_chanctx,
	.change_chanctx			= scm80211_change_chanctx,
	.assign_vif_chanctx		= scm80211_assign_vif_chanctx,
	.unassign_vif_chanctx		= scm80211_unassign_vif_chanctx,
	.switch_vif_chanctx		= scm80211_switch_vif_chanctx,
	.remain_on_channel		= scm80211_roc,
	.cancel_remain_on_channel	= scm80211_croc,
	.event_callback			= scm80211_event_callback,
	.get_txpower			= scm80211_get_txpower,
	.mgd_prepare_tx			= scm80211_mgd_prepare_tx,
	.tdls_channel_switch		= scm80211_tdls_channel_switch,
	.tdls_cancel_channel_switch	= scm80211_tdls_cancel_channel_switch,
	.tdls_recv_channel_switch	= scm80211_tdls_recv_channel_switch,
	.mgd_protect_tdls_discover	= scm80211_mgd_tdls_discover,
	.tx_frames_pending		= scm80211_tx_frames_pending,
	.set_coverage_class		= scm80211_set_coverage_class,
	.sta_rate_tbl_update		= scm80211_sta_rate_tbl_update,
	.sta_statistics			= scm80211_sta_statistics,
#ifdef SCM80211_RFKILL_POLL
	.rfkill_poll			= scm80211_rfkill_poll,
#endif /* SCM80211_RFKILL_POLL */
#ifdef CONFIG_MAC80211_DEBUGFS
	.sta_add_debugfs		= scm80211_sta_add_debugfs,
#endif
#ifdef SCM80211_WOW
	.suspend			= scm80211_wow_op_suspend,
	.resume				= scm80211_wow_op_resume,
	.set_wakeup			= scm80211_wow_op_set_wakeup,
#endif
};

static int scm80211_oops_handler(struct notifier_block *self, unsigned long val, void *data)
{
	if (val == DIE_OOPS)
		pr_err("Senscomm WLAN Driver, Version (%s)\n", DRIVER_VERSION);
	return 0;
}

static struct notifier_block scm80211_oops_notifier = {
	.notifier_call = scm80211_oops_handler,
};

static bool is_oops_registered;

void scm80211_register_oops_handler(void)
{
	if (!is_oops_registered) {
		is_oops_registered = true;
		register_die_notifier(&scm80211_oops_notifier);
	}
}

void scm80211_unregister_oops_handler(void)
{
	if (is_oops_registered) {
		is_oops_registered = false;
		unregister_die_notifier(&scm80211_oops_notifier);
	}
}

#ifdef __staging__

#endif /* __staging__ */
