// SPDX-License-Identifier: GPL-2.0 OR BSD-2-Clause
/*
 * Copyright (c) 2020-2023 Senscomm Semiconductor Co., Ltd.
 */
/**
 * Limitations:
 * - Injected frames are not considered, and may not work
 * - At most A-MPDU in a TXOP
 *
 * TODO:
 * - Add a robust rate at the end of the rate
 * - Take into account frames currently in txop->sched
 * - Multi-TID A-MPDU
 */

#include <linux/compiler.h>
#include <linux/bitfield.h>
#include <linux/list_sort.h>
#include <linux/log2.h>

#define func_fmt(str)		"[TX] "str
#define func_trace              TX

#include "hw.h"
#include "core.h"
#include "util.h"
#include "trace.h"
#include "debug.h"
#ifdef SCM80211_FTM
#include "ftm.h"
#endif

/*
 * SCM80211_TXOP_RETRANSMIT_MAX:
 * The maximum times that one TXOP could be (re)transmit.
 * This machanism is only applied to no ack received condition(e.g underrun).
 */
#define SCM80211_TXOP_RETRANSMIT_MAX 8

#define SCM80211_TX_RETRY_RATE	(IEEE80211_TX_MAX_RATES - 2)	/* the last entry */
#define SCM80211_TX_TIMEOUT	100	/* ms */
#define SCM80211_CCA_TIMEOUT	1000	/* 1s */
//#define SCM80211_MGMT_HWQ_MEM_LEN       0x5000 //20K

static int retry_thr = 5;
module_param(retry_thr, int, 0644);
MODULE_PARM_DESC(retry_thr, "If retry cnt less than this threshold destroy the prepared txop");

static int txop_order = 1;
module_param(txop_order, int, 0644);
MODULE_PARM_DESC(txop_order, "TXOP order 0 means only one txop 1 means two txops");

static int rate_prob_level = 96;	/* 75% */
module_param(rate_prob_level, int, 0644);
MODULE_PARM_DESC(rate_prob_level,
		 "Try to select higher rate only when its proportion higer than rate_prob_level");

static int rate_prob_higher_level = 102;	/* 80% */
module_param(rate_prob_higher_level, int, 0644);
MODULE_PARM_DESC(rate_prob_higher_level,
		 "Try to probe higher rate when its proportion higer than rate_prob_higher_level");

static int tx_power = -128;
module_param(tx_power, int, 0644);
MODULE_PARM_DESC(tx_power, "force to using fixed power");

void scm80211_tx_power_set(struct ieee80211_hw *hw, int channel, int power)
{
	tx_power = DIV_ROUND_UP(power, 10);
	if (WARN_ON(abs(tx_power) > 20)) {
		scm80211_warn(hw, "power setting out of range must be [-200-200] but %d\n", power);
		tx_power = 10;
		power = 100;
	}

	if (is_pcie_device(hw))
		scm2020_set_phy_dig_gain(hw, channel, power);
}

static unsigned long power_bitmap;
static s8 power_index[SCM80211_NUM_HWQ];
static unsigned long txq_bitmap[SCM80211_NUM_TXQ / BITS_PER_LONG];

static const s16 higher_tp_level[GROUP_RATES] = {
	101,			/* MCS7 / MCS8  - 4% = 79% */
	110,			/* MCS8 / MCS9  - 4% = 86% */
	108,			/* MCS9 / MCS10 - 4% = 84% */
	110,			/* MCS10/ MCS11 - 4% = 86% */
	128
};

/* Reference to 20211025_SCM2020Z1_EVB3_TX EVM vs Power_Add_2G_DPD */
static const s8 tx_nonht_table[2][12] = {
	{
		[0 ...  3]	= 22,
		[4 ... 10]	= 20,
		[11]		= 18
	},
	{
		[0 ...  3]	= -128,
		[4 ...  8]	= 20,
		[9]		= 19,
		[10]		= 17,
		[11]		= 15,
	}
};

static const s8 tx_ht_table[2][CBW40 + 1][8] = {
	{
		[0 ... 1] = {
			[0 ... 4] = 20,
			[5]	  = 19,
			[6]	  = 18,
			[7]	  = 18,
		}
	},
	{
		[0 ... 1] = {
			[0 ... 3] = 18,
			[4]       = 17,
			[5]       = 16,
			[6]       = 15,
			[7]       = 15
		}
	}
};

static const s8 tx_vht_table[2][CBW80 + 1][10] = {
	{
		/* not used */
		{
			[0 ... 5] = 19,
			[6]	  = 18,
			[7]	  = 17,
			[8]	  = 13,
			[9]	  = -128
		},
		{
			[0 ... 5] = 19,
			[6]	  = 18,
			[7]	  = 17,
			[8]       = 14,
			[9]       = 11
		},
		{
			[0 ... 9] = -128
		}
	},
	{
		{
			[0 ... 3] = 18,
			[4]       = 17,
			[5]       = 16,
			[6]	  = 15,
			[7]	  = 15,
			[8]	  = 13,
			[9]	  = -128
		},
		[1 ... 2] = {
			[0 ... 3] = 18,
			[4]       = 17,
			[5]       = 16,
			[6]	  = 15,
			[7]	  = 15,
			[8]	  = 13,
			[9]	  = 12
		}
	}
};

static const s8 tx_he_table[2][CBW80 + 1][12] = {
	{
		[0 ... 1] = {
			[0 ...  4]	= 20,
			[5]		= 19,
			[6]		= 19,
			[7]		= 18,
			[8]		= 17,
			[9]		= 16,
			[10 ... 11]	= 15
		},
		[2] = {
			[0 ... 11]	= -128
		}
	},
	{
		[0 ... 2] = {
			[0 ...  3]	= 18,
			[4]		= 17,
			[5]		= 16,
			[6]		= 15,
			[7]		= 15,
			[8]		= 14,
			[9]		= 13,
			[10 ... 11]	= 11
		}
	}
};

/**
 * DOC: HW queue usage
 *
 * HW queues for 1st vif: 0x001f
 * - 0 (VO), 1 (VI), 2 (BE), 3 (BK), 4 (MGMT/off-chan/cab)
 * HW queues for 2nd vif: 0x03e0
 * - 5 (VO), 6 (VI), 7 (BE), 8 (BK), 9 (MGMT/off-chan/cab)
 *
 * The last hwqueues of each vif, i.e., 4 and 9, are used for
 * transmitting frames sent via tx() callback. These hardware queues
 * follow VO EDCA setting.
 */

static void scm80211_discard_schedule(struct txop *txop);
static void scm80211_update_tid_queue_size(struct txop *txop, struct sk_buff *skb, bool increase);
struct scm80211_txq *scm80211_get_next_txq(struct scm80211_hwq *hwq);

static int scm80211_schedule_hwq_work(struct scm80211_hwq *hwq)
{
#ifdef SCHEDULE_ON_SPECIFIC_CPU
	return schedule_work_on(SCHEDULE_ON_SPECIFIC_CPU, &hwq->work);
#else
	return schedule_work(&hwq->work);
#endif
}


/**
 * scm80211_hwq() - get hardware queue
 * @hw: hardware
 * @queue: queue index
 */
struct scm80211_hwq *scm80211_hwq(struct ieee80211_hw *hw, int queue)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	if (queue < 0 || queue >= hw->queues)
		return NULL;
	return &priv->hwq[queue];
}

/**
 * scm80211_vif_queues() - get bitmap of vif hardware queues
 */
unsigned long scm80211_vif_queues(struct ieee80211_vif *vif)
{
	unsigned long queues = 0;

	queues = ((1 << SCM80211_NUM_HWQ_PER_VIF) - 1);
	queues <<= scm80211_vdev_id(vif) * SCM80211_NUM_HWQ_PER_VIF;
	return queues;
}

static int scm80211_vif_mgmt_hw_queue(struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);

	return vp->mgmt_hw_queue;
}

/**
 * scm80211_wake_queues() - wakes up (activates) hardware queues
 *
 */
void scm80211_wake_queues(struct ieee80211_hw *hw, unsigned long queues)
{
	int i;
	struct scm80211_hwq *hwq;

	scm80211_trace(hw, "%s: queues=%lx\n", __func__, queues);

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		hwq->enabled = true;

		ieee80211_wake_queue(hw, i);
		scm80211_schedule_hwq_work(hwq);
	}
}

/**
 * scm80211_stop_queues() - stops hardware queues from transmitting
 */
void scm80211_stop_queues(struct ieee80211_hw *hw, unsigned long queues)
{
	int i;
	struct scm80211_hwq *hwq;

	scm80211_trace(hw, "%s: queues=%lx\n", __func__, queues);

	for_each_set_bit(i, &queues, hw->queues) {
		ieee80211_stop_queue(hw, i);

		hwq = scm80211_hwq(hw, i);
		hwq->enabled = false;
	}
}

/**
 * DOC: txq to hw queue mapping
 *
 * - sta->txq[IEEE80211_NUM_TIDS]
 *   : bufferable management frames
 *   : tid = IEEE80211_NUM_TIDS, ac = IEEE80211_AC_VO, hwqueue = 0
 * - sta->txq[i] (i < IEEE80211_NUM_TIDS)
 *   : individually addressed data frames
 *   : tid = i, ac = ieee80211_ac_from_tid(tid), hwqueue = vif->hw_queue[ac]
 * - vif->txq
 *   : group addressed data frames (only AP)
 *   : tid = 0, ac = IEEE80211_AC_BE, hwqueue = 3 or cab queue
 * - vif->mq
 *   : management frames sent via tx()
 *   : tid = 7, ac = IEEE80211_AC_VO, hwqueue = 4
 */

/**
 * DOC: software queue (struct scm80211_txq)
 *
 * Software queue type
 * ===================
 * There are two types of software queue:
 * - Ones that are associated with an intermediate queue mananged by mac80211.
 *   The @source field is set to that queue (of &struct ieee80211_txq). These
 *   queues are handed over to the driver by wake_tx_queue() callback. These
 *   belong to sta per-TID queue or vif queue for multicast tx.
 * - The other ones are not associated with an intermediate queue. They are
 *   internally created per vif and handle frames delivered to the driver
 *   via tx() callback. The @source field is @NULL for this type of queues.
 * Once they are added to hardware queue, they undergo the same procedure,
 * except that the former one pulls more frames from the upper queue before
 * starting scheduling.
 *
 * Software queue life cycle
 * =========================
 * 1. @wake_tx_queue callback schedules a corresponding hardware queue work if it
 *    is idle.
 * 2. The hardware queue work (the work in short) allocates a txop resource and
 *    calls @ieee80211_next_txq() to select the software queue to transmit.
 * 3. The work then dequeues enough frames from the software queue and forms
 *    an A-MPDU or a SIFS-burst txop.
 * 4. The work starts backoff to transmit an eligible txop.
 * 5. After a txop frame exchange sequence completes, hardware raises TX DONE
 *    interrupt. The interrupt handler collects unacknowledged frames for later
 *    retranmission and retires the acknowledged frames and frames that failed
 *    too many times.
 * 6. The interrupt handler updates transmit window if the software queue
 *    has a BlockAck agreement, and further starts backoff if there is a txop
 *    that finished schedduling and all the scheduled frames are within the
 *    transmit window.
 * 7. After completion of one txop from the software queue, if the queue has
 *    pending frames, it calls @ieee80211_schedule_txq() to return the
 *    software queue and schedules the hardware queue work.
 * Step 2-7 are repeated until there is no software queue that has frames to
 * transmit.
 */

static struct scm80211_txq *scm80211_vif_mgmt_txq(struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);

	return &vp->mq;
}

static bool is_mac80211_txq(struct scm80211_txq *txq)
{
	return (txq->source == &txq->txq);
}

void scm80211_enable_tx_irq(struct ieee80211_hw *hw, int queues)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->int_en |= FIELD_PREP(INTR_ENABLE_TX_DONE, queues);
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

void scm80211_disable_tx_irq(struct ieee80211_hw *hw, int queues)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	priv->int_en &= ~FIELD_PREP(INTR_ENABLE_TX_DONE, queues);
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

void scm80211_enable_trig_qnull_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int vdev = scm80211_vdev_id(vif);

	priv->int_en |= FIELD_PREP(INTR_ENABLE_BASIC_TRIG_QNULL_VIF, (1 << vdev));
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

void scm80211_disable_trig_qnull_irq(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int vdev = scm80211_vdev_id(vif);

	priv->int_en &= ~FIELD_PREP(INTR_ENABLE_BASIC_TRIG_QNULL_VIF, (1 << vdev));
	scm80211_mac_writel(hw, priv->int_en, MAC_REG_INTR_ENABLE);
}

/**
 * scm80211_txq_init() - initialize txq driver private members
 *
 * @internal: %true if @itxq is the driver internal queue
 */
void scm80211_init_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		       struct ieee80211_txq *itxq, bool internal)
{
	struct scm80211_txq *txq = (struct scm80211_txq *)itxq;
	struct ieee80211_sta *sta;
	struct scm80211_hwq *hwq;
	unsigned long flags;

	if (!txq)
		return;

	sta = itxq->sta;
	txq->hw = hw;
	txq->stop = false;
	INIT_LIST_HEAD(&txq->list);
	INIT_LIST_HEAD(&txq->txq_list);
	spin_lock_init(&txq->lock);
	skb_queue_head_init(&txq->queue);
	skb_queue_head_init(&txq->retry);

	memset(&txq->ba, 0, sizeof(txq->ba));

	if (internal) {
		txq->hw_queue = scm80211_vif_mgmt_hw_queue(vif);
		txq->source = NULL;
		txq->txq.ac = IEEE80211_AC_VO;
		txq->txq.tid = 15;
		txq->idx = -1;
		sprintf(txq->name, "<%pM:mgmt>", vif->addr);

		hwq = scm80211_hwq(hw, txq->hw_queue);
		if (vif->type == NL80211_IFTYPE_P2P_DEVICE) {
			hwq->p2p_dev_enable = true;
			hwq->p2p_txq = txq;
		} else {
			hwq->txq = txq;
		}
		spin_lock_irqsave(&hwq->lock, flags);
		list_add_tail(&txq->txq_list, &hwq->txq_head);
		spin_unlock_irqrestore(&hwq->lock, flags);
	} else {
		txq->hw_queue = vif->hw_queue[itxq->ac];
		txq->source = itxq;
		txq->idx = -1;
		sprintf(txq->name, "<%pM:%d>",
			sta ? sta->addr : (u8 *)"\xff\xff\xff\xff\xff\xff", itxq->tid);
		hwq = scm80211_hwq(hw, txq->hw_queue);
		spin_lock_irqsave(&hwq->lock, flags);
		list_add_tail(&txq->txq_list, &hwq->txq_head);
		spin_unlock_irqrestore(&hwq->lock, flags);
		if (itxq->tid > 7 || !sta)
			return;
		txq->idx = find_first_zero_bit(txq_bitmap, SCM80211_NUM_TXQ);
		if (txq->idx == SCM80211_NUM_TXQ) {
			scm80211_warn(hw, "txq%s alloc idx fail\n", txq->name);
		} else {
			set_bit(txq->idx, txq_bitmap);
			scm80211_register_txq(hw, vif, itxq);
		}
	}
}

static bool __used scm80211_txq_ba_session_started(struct scm80211_txq *txq)
{
	return (READ_ONCE(txq->ba.action) == IEEE80211_AMPDU_TX_START);
}

static bool scm80211_txq_is_ampdu_enabled(struct scm80211_txq *txq)
{
	return (txq->ba.action == IEEE80211_AMPDU_TX_OPERATIONAL);
}

/**
 * scm80211_ba_buf_size() - return the number of subframes
 *
 * This function returns the maximum number of subframes in an
 * AMPDU that the peer station @txq->sta can receive.
 */
int scm80211_ba_buf_size(struct scm80211_txq *txq)
{
	struct ieee80211_ampdu_params *param = &txq->ba;

	if (param->action != IEEE80211_AMPDU_TX_OPERATIONAL)
		return 0;
	return param->buf_size;
}

/**
 * scm80211_pending_frames() - returns the number of pending frames
 *
 * @buffered: if %true, account for frames kept in the driver as well.
 */
int scm80211_num_pending_frames(struct scm80211_txq *txq, bool buffered)
{
	unsigned long n = 0;

	if (is_mac80211_txq(txq))
		ieee80211_txq_get_depth(&txq->txq, &n, NULL);

	if (buffered)
		n += skb_queue_len(&txq->queue) + skb_queue_len(&txq->retry);
	return n;
}

/**
 * scm80211_dequeue_pending_frames() - dequeue frames from software queue
 *
 * This function pulls in a enough number of frames and put them into
 * the driver internal queue for txop scheduling.
 *
 * Returns the number of frames available for txop scheduling
 */
int scm80211_dequeue_pending_frames(struct scm80211_txq *txq)
{
	struct ieee80211_hw *hw = txq->hw;
	struct scm80211_hwq *hwq = scm80211_hwq(hw, txq->hw_queue);
	struct sk_buff *skb, *skb_cpy = NULL;
	int pending, max, m, n = 0;

	if (!is_mac80211_txq(txq))
		goto out;

	pending = scm80211_num_pending_frames(txq, false);
	m = skb_queue_len(&txq->retry) + skb_queue_len(&txq->queue);
	/* if ampdu not enabled only dequeue max 8 mpdus */
	max = (scm80211_txq_is_ampdu_enabled(txq)) ? SCM80211_MAX_MPDU :
	    hwq->edca.txop == 0 ? 1 : 8;
	m = (m >= max) ? 0 : max - m;
	m = min_t(int, m, pending);

	while (n < m && (skb = ieee80211_tx_dequeue_ni(txq->hw, &txq->txq))) {
		if (scm80211_txq_is_ampdu_enabled(txq) && txq->sb.ssn > IEEE80211_MAX_SN) {
			/*
			 * This is the very first frame after a BA session is
			 * established. Initialize the transmit window
			 */
			txq->sb.ssn = ieee80211_get_sn(skb);
			scm80211_trace(hw,
				       "- transmit window initialized to [%d-%d]\n",
				       txq->sb.ssn,
				       ieee80211_sn_add(txq->sb.ssn, txq->ba.buf_size - 1));
		}

		if (skb_has_frag_list(skb)) {
			/* Converts non-linear &sk_buff to linear one */
			if (is_usb_device(hw) ||
			    (version(scm80211_mac_version(hw)) < 0x14 &&
			     (skb->len - skb->data_len) <= SCM80211_MAX_MMSS_BYTES)) {
				skb_cpy = skb_copy(skb, GFP_ATOMIC);
			}

			/* for usb soluction non-linear skb can not be handled, free it */
			if (skb_cpy || is_usb_device(hw)) {
				kfree_skb(skb);
				skb = skb_cpy;
			}
			skb_cpy = NULL;

			if (!skb)
				continue;
		}

		skb_queue_tail(&txq->queue, skb);
		n++;
	}

	if (n < m) {
		scm80211_trace(hw, "- %s: dequeue failed %d/%d pending=%d/%d/%d",
			       txq->name, n, m, skb_queue_len(&txq->retry),
			       skb_queue_len(&txq->queue), scm80211_num_pending_frames(txq, false));
	}

out:
	return skb_queue_len(&txq->retry) + skb_queue_len(&txq->queue);
}

/**
 * scm80211_update_tid_queue_size() - update queue size for bsr
 *
 * TODO:
 * update per tid queue size
 * do not report error
 */
static void scm80211_update_tid_queue_size(struct txop *txop, struct sk_buff *skb, bool increase)
{
	struct scm80211_txq *txq = txop->owner;
	struct txvector *tv = (void *)skb->data;
	struct ieee80211_hdr *hdr = (void *)(tv + 1);
	struct ieee80211_vif *vif = txq->txq.vif;
	struct ieee80211_vif_priv *vifp;
	struct ieee80211_sta *sta = txq->txq.sta;
	int plen, tid;
	unsigned long flags;
	bool locked = false;

	if (!vif || vif->type != NL80211_IFTYPE_STATION)
		return;

	if (!vif->bss_conf.assoc)
		return;

	vifp = (void *)vif->drv_priv;

	if (!ieee80211_is_he_sta(sta))
		return;

	if (ppdu_is_non_ht(tv) || ppdu_is_ht(tv))
		plen = ieee80211_mpdu_len(skb);
	else
		plen = ieee80211_subframe_len(skb);

	if (ieee80211_is_data_qos(hdr->frame_control))
		tid = ieee80211_get_tid(hdr);
	else if (ieee80211_is_mgmt(hdr->frame_control))
		tid = WME_AC_TO_TID(IEEE80211_AC_VO);
	else
		return;

	locked = !!spin_trylock_irqsave(&vifp->lock, flags);
	if (increase)
		vifp->tid_queue_size[tid] += plen;
	else
		vifp->tid_queue_size[tid] -= plen;

	if (vifp->tid_queue_size[tid] < 0)
		vifp->tid_queue_size[tid] = 0;
	if (locked)
		spin_unlock_irqrestore(&vifp->lock, flags);
}

#if KERNEL_VERSION(5, 13, 0) <= LINUX_VERSION_CODE
static int scm80211_commpare_sn(void *priv, const struct list_head *a, const struct list_head *b)
#else
static int scm80211_commpare_sn(void *priv, struct list_head *a, struct list_head *b)
#endif
{
	struct sk_buff *skb[2];
	u16 sn[2];

	/*
	 * Note: list member is available after Kerner v4.19.0, container_of function
	 * will be failed on kernel before this it, the next/prev pointer is at the beginning
	 * of sk_buff struct so we can force translate it to sk_buff, ugly but works
	 */
	skb[0] = (struct sk_buff *)a;
	skb[1] = (struct sk_buff *)b;

	sn[0] = ieee80211_get_sn(skb[0]);
	sn[1] = ieee80211_get_sn(skb[1]);

	return ieee80211_sn_less(sn[1], sn[0]);
}

/**
 * scm80211_sort_skb_queue() - sorts frames in a queue in an ascending order of sn
 * @head: a socket buffer queue that contains frames to sort
 *
 * Note:
 * This function uses @list_sort() based on the fact that the first two members of
 * the struct sk_buff are the same with struct list_head. If the data
 * structure changes, this will not work.
 */
void scm80211_sort_skb_queue(struct sk_buff_head *head)
{
	unsigned long flags;

	spin_lock_irqsave(&head->lock, flags);
	list_sort(NULL, (struct list_head *)head, scm80211_commpare_sn);
	spin_unlock_irqrestore(&head->lock, flags);
}

/**
 * scm80211_txq_is_scheduled() - check whether the txq is scheduled by any txop
 */
static bool scm80211_txq_is_scheduled(struct scm80211_txq *txq)
{
	struct scm80211_hwq *hwq = scm80211_hwq(txq->hw, txq->hw_queue);
	unsigned long flags;
	int i;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	for (i = 0; i < ARRAY_SIZE(hwq->ring.pool); i++) {
		if (hwq->ring.pool[i].owner == txq) {
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			return true;
		}
	}
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
	return false;
}

/**
 * scm80211_complete_txq() - complete txq
 *
 * @txq: software queue whose transmission just finished
 */
static void scm80211_complete_txq(struct txop *txop, struct scm80211_txq *txq)
{
	bool force;

	if (!txq)
		return;

	scm80211_sort_skb_queue(&txq->retry);
	force = scm80211_num_pending_frames(txq, true) > 0;
	if (is_mac80211_txq(txq)) {
		ieee80211_return_txq(txq->hw, &txq->txq, force);
		if (force)
			ieee80211_schedule_txq(txq->hw, &txq->txq);
	}
}

/**
 * scm80211_flush_txq() - flush (discard) queued frames
 *
 * @txq: software queue
 *
 * Flushes internal frame queue. The upstream queue is not flushed
 */
void scm80211_cleanup_txq(struct scm80211_txq *txq)
{
	struct sk_buff *skb;

	skb_queue_splice_init(&txq->retry, &txq->queue);

	while ((skb = skb_dequeue(&txq->queue)) != NULL)
		ieee80211_free_txskb(txq->hw, skb);
}

/**
 * scm80211_deinit_txq() - deinitialize txq
 *
 * This function is called from sta_remove(). After this call, txq
 * should not be referred to any more. This is tricky.
 *
 * All the unsent frames are discarded.
 */
void scm80211_deinit_txq(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			 struct ieee80211_txq *itxq)
{
	struct scm80211_txq *txq = (struct scm80211_txq *)itxq;
	struct txop *txop;
	struct scm80211_hwq *hwq;
	unsigned long flags;
	enum txop_state state;
	int i, retry;

	if (!txq)
		return;

	txq->stop = true;
	hwq = scm80211_hwq(hw, txq->hw_queue);
	if (vif->type == NL80211_IFTYPE_P2P_DEVICE) {
		hwq->p2p_dev_enable = false;
		hwq->p2p_txq = NULL;
	} else {
		hwq->txq = NULL;
	}

	if (txq->idx >= 0 && txq->idx < SCM80211_NUM_TXQ) {
		scm80211_deregister_txq(hw, vif, itxq);
		clear_bit(txq->idx, txq_bitmap);
	}

	for (i = 0; i < ARRAY_SIZE(hwq->ring.pool); i++) {
		bool first = true;

		retry = 0;
again:

		spin_lock_irqsave(&hwq->ring.lock, flags);
		txop = &hwq->ring.pool[i];
		state = txop->state;

		if (txop->txq[0] != txq && txop->txq[1] != txq) {
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			continue;
		}

		if (list_empty(&txop->list) || state == TXOP_STATE_ISSUING) {
			scm80211_dbg(txop->hw, "hwq#%d.%d: state %s is in use, wait\n",
				     txop->hwq->id, txop->id,
				     txop_state_str[txop->state]);
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			if (retry++ < 10) {
				msleep(1);
				goto again;
			} else {
				WARN(1, "- hwq#%d.%d state [%s] unfinished\n",
				     txop->hwq->id, txop->id, txop_state_str[state]);
				if (!list_empty(&txop->list)) {
					list_del_init(&txop->list);
					scm80211_handle_txop_complete(hw, txop);
					msleep(2);
					continue;
				}
			}
		}

		first = list_is_first(&txop->list, &hwq->ring.head[txop->state]);
		if (state == TXOP_STATE_TX_DONE) {
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			/* wait for tx progress done txop */
			msleep(2);
		} else {
			list_del_init(&txop->list);
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			if (state == TXOP_STATE_SCHEDULED) {
				/* Discard all scheduled txop */
				scm80211_discard_schedule(txop);
				scm80211_txop_free(txop);
			} else if (state == TXOP_STATE_ISSUED) {
				/* abort backoff has free txop to IDLE list
				 * for pcie only one txop in hwq can be stay in issued mode.
				 * So it is awlays the first txop in the issued_list.
				 * But for USB both two txops can be stay in issued list,
				 * which may cause some mis-order pickup if picking out
				 * second txop and put back to header of the issue list
				 */
				if (is_pcie_device(hw))
					hrtimer_try_to_cancel(&hwq->watchdog);
				if (first) {
					scm80211_cancel_tx(hw, txop, true);
				} else {
					spin_lock_irqsave(&hwq->ring.lock, flags);
					list_move_tail(&txop->list, &hwq->ring.head[state]);
					spin_unlock_irqrestore(&hwq->ring.lock, flags);
				}
			} else {
				scm80211_txop_free(txop);
			}
		}
	}

	spin_lock_irqsave(&hwq->lock, flags);
	list_del_init(&txq->txq_list);
	spin_unlock_irqrestore(&hwq->lock, flags);

	scm80211_dequeue_pending_frames(txq);
	scm80211_cleanup_txq(txq);
}

/**
 * DOC: TXOP scheduling
 *
 * @txop: txop
 * @skb: 802.11 frame to be transmitted
 * @ti: pointer to struct ieee80211_tx_info obtained by IEEE80211_SKB_CB(@skb)
 * @xti: pointer to struct scm80211_tx_info
 *
 * The unit of transmission is a txop, not a frame. The driver attempts to
 * 'schedule' as many frames as possible as long as the total txop time
 * including possible response frames does not exceed the txop limit.
 *
 * The driver relies on the information in @ti->control. In particular,
 * it relies on @ti->control.key information to correctly calculate the MPDU
 * length during scheduling. Therefore, the driver must not corrupt this
 * information during its entire liefspan within the driver (i.e., from the
 * moment it is handed over to the driver via wake_tx_queue() or tx() to
 * the moment it is retired via ieee80211_tx_status() or its friends after
 *
 * The driver calls to ieee80211_get_tx_rate() once for a txop transmission.
 * - All the frames transmitted in this txop uses the same parameters (PPDU
 *   format, MCS, bandwidth, and etc).
 *
 * @txop->rc is struct ieee80211_tx_info that is filled by rc80211
 * @txop->ti is struct ieee80211_tx_info that is modified by the driver. The
 * driver uses @txop->ti.control.rates[] for rate selection.
 * @txop->rs is struct ieee80211_tx_info that the driver uses for @ieee80211_tx_status()
 *
 * @txop->txvector is the corresponding TXVECTOR of @txop.
 */

static void scm80211_sta_rate_update(void *data, struct ieee80211_sta *sta)
{
	struct ieee80211_sta_priv *priv;
	struct he_higher_rates_stats *stats;
	int *n = data;
	int i, j;

	if (!ieee80211_is_he_sta(sta))
		return;

	priv = ieee80211_to_priv(sta);
	stats = priv->rate_stats;
	*n += 1;

	for (i = 0; i < GROUP_COUNT; ++i) {
		bool downgrade = true;
		int bw = i / 2;
		int nss = i % 2;

		if (bw > get_ieee80211_rx_bandwidth(sta) || nss > get_ieee80211_rx_nss(sta) - 1)
			continue;

		for (j = GROUP_RATES - 1; j >= 0; j--) {
			int index = i * GROUP_RATES + j;
			unsigned long ts = stats[index].jiffies;

			if (ts == 0)
				continue;
			if (time_before(jiffies, ts + (2 * HZ))) {
				if (likely(stats[index].attempts > 0)) {
					s16 tmp_prob = PROB_FRAC(stats[index].success,
								 stats[index].attempts);
					if (stats[index].prob == 0 ||
					    stats[index].attempts > 1000) {
						stats[index].tot_succ += stats[index].success;
						stats[index].tot_atmp += stats[index].attempts;
						stats[index].success = 0;
						stats[index].attempts = 0;
					}

					if (stats[index].prob == 0) {
						stats[index].prob = tmp_prob;
					} else if (stats[index].attempts > 1000) {
						stats[index].prob +=
						    (tmp_prob - stats[index].prob) / 16;
					}

					downgrade = false;
				}
			} else if (downgrade && stats[index].prob < rate_prob_level) {
				/*
				 * This rate no updates for a long time
				 * for the success proportion is not good
				 * Shrink the attempts/success value when it
				 * reach to 0, clear the prob rate, So it will have
				 * chance to try again. This will take effect
				 * when the interference turns better
				 *
				 * If tried but all failed do not try it for 20s
				 */
				if (stats[index].prob == 0 && stats[index].tot_atmp > 0)
					stats[index].jiffies = jiffies + 20 * HZ;

				stats[index].prob /= 2;
				if (stats[index].prob == 0) {
					stats[index].tot_atmp = 0;
					stats[index].tot_succ = 0;
					stats[index].jiffies = 0;
				} else {
					stats[index].jiffies = jiffies + 5 * HZ;
				}

				stats[index].success = 0;
				stats[index].attempts = 0;
			}

			if (j < GROUP_RATES - 1 && stats[index].prob < stats[index + 1].prob) {
				/*
				 * Higer rate have better success proportion
				 * make the proportion equal to higher rate when
				 * This will take effect when the interference
				 * turns worse
				 */
				stats[index].prob = stats[index + 1].prob;
			}
		}
	}
}

enum hrtimer_restart scm80211_ratestats_timer(struct hrtimer *t)
{
	struct ieee80211_hw_priv *priv = container_of(t, struct ieee80211_hw_priv, ratestats_timer);
	struct ieee80211_hw *hw = priv->hw;
	int n = 0;

	ieee80211_iterate_stations_atomic(hw, scm80211_sta_rate_update, &n);

	if (n > 0)
		hrtimer_start(&priv->ratestats_timer, ms_to_ktime(MSEC_PER_SEC),
			      HRTIMER_MODE_REL_SOFT);

	return HRTIMER_NORESTART;
}

/**
 * scm80211_tx_rate_index() - returns the index of current rate in rates table
 */
static inline int scm80211_tx_rate_index(struct txop *txop)
{
	if (!txop->rate)
		return 0;
	return txop->rate - txop->ti.control.rates;
}

static void scm80211_fixup_tx_info_flags(struct ieee80211_hw *hw,
					 struct ieee80211_sta *sta,
					 struct ieee80211_tx_info *ti, struct ieee80211_tx_rate *r)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	int streams = 0;

	if (!ieee80211_rate_is_non_ht(r))
		streams = ieee80211_rate_get_nss(r);

	BUG_ON(streams > priv->n_chain);

	/*
	 * The RX NSS configuration is different between HT and HE capability
	 * Force downgrade the stream if streams is more than he rx nss
	 */
	if (sta) {
		struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);

		if (ieee80211_rate_is_he(r) && streams > sta_priv->n_ss) {
			streams = sta_priv->n_ss;
			ieee80211_rate_set(r, ieee80211_rate_get_mcs(r), streams);
		}

		if (ieee80211_rate_is_he(r) &&
		    !(get_ieee80211_he_cap(sta).he_cap_elem.phy_cap_info[2] &
		      IEEE80211_HE_PHY_CAP2_STBC_RX_UNDER_80MHZ))
			ti->flags &= ~IEEE80211_TX_CTL_STBC;
	}

	if (streams == 0 || streams == priv->n_chain)
		ti->flags &= ~IEEE80211_TX_CTL_STBC;

#ifdef SCM80211_SMPS
	if (priv->smps == SMPS_STATIC)
		ti->flags &= ~IEEE80211_TX_CTL_STBC;
#endif
	if (streams == 0)
		ti->flags &= ~IEEE80211_TX_CTL_LDPC;
}

static bool scm80211_fixup_he_omi(struct ieee80211_sta *sta, struct ieee80211_tx_rate *r)
{
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	int nss, bw;
	u16 flags;
	bool changed = false;

	if (!IS_OMI(sta_priv->he_aom))
		return false;

	nss = OMI_GET_RX_NSS(sta_priv->he_aom) + 1;
	bw = OMI_GET_CHAN_WIDTH(sta_priv->he_aom);

	flags = (bw == CBW80) ? IEEE80211_TX_RC_80_MHZ_WIDTH :
	    (bw == CBW40) ? IEEE80211_TX_RC_40_MHZ_WIDTH : 0;

#ifdef SCM80211_WFA_TEST
	if (nss != ieee80211_rate_get_nss(r)) {
		ieee80211_rate_set_vht(r, ieee80211_rate_get_mcs(r), nss);
		changed = true;
	}

	if (!(r->flags & flags)) {
		r->flags &= ~(IEEE80211_TX_RC_80_MHZ_WIDTH | IEEE80211_TX_RC_40_MHZ_WIDTH);
		r->flags |= flags;
		changed = true;
	}
#else
	if (nss < ieee80211_rate_get_nss(r)) {
		ieee80211_rate_set_vht(r, ieee80211_rate_get_mcs(r), nss);
		changed = true;
	}

	if ((r->flags & IEEE80211_TX_RC_80_MHZ_WIDTH) && bw < CBW80) {
		r->flags &= ~IEEE80211_TX_RC_80_MHZ_WIDTH;
		r->flags |= flags;
		changed = true;
	} else if ((r->flags & IEEE80211_TX_RC_40_MHZ_WIDTH) && bw < CBW40) {
		r->flags &= ~IEEE80211_TX_RC_40_MHZ_WIDTH;
		changed = true;
	}
#endif
	return changed;
}

#define MAP_HE_RATE_DIRECT		0
#define MAP_HE_RATE_HIGHER		1
#define PROBE_HE_RATE_HIGHER		2

static int scm80211_map_to_he_rate(struct ieee80211_tx_rate *r,
				   struct ieee80211_sta *sta, u8 n_ss, u8 max_mcs)
{
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	struct he_higher_rates_stats *stats = sta_priv->rate_stats;
	bool is_highest;
	enum cbw cbw = ieee80211_ppdu_bandwidth(r);
	int nss = ieee80211_rate_get_nss(r);
	int mcs = ieee80211_rate_get_mcs(r);
	int ret = MAP_HE_RATE_DIRECT;
	unsigned long flags;

	if (r->flags & IEEE80211_TX_RC_HE_MCS) {
		scm80211_fixup_he_omi(sta, r);
		return 0;
	}

	if (!(r->flags & (IEEE80211_TX_RC_VHT_MCS | IEEE80211_TX_RC_MCS)))
		return -1;

	is_highest = ieee80211_is_highest_mcs(sta, r);

	if (r->flags & IEEE80211_TX_RC_VHT_MCS) {
		r->flags &= ~IEEE80211_TX_RC_VHT_MCS;
		r->flags |= IEEE80211_TX_RC_HE_MCS;
	} else {
		r->flags &= ~IEEE80211_TX_RC_MCS;
		r->flags |= IEEE80211_TX_RC_HE_MCS;
		ieee80211_rate_set_vht(r, mcs, nss);
	}

	/* rate info changed */
	if (scm80211_fixup_he_omi(sta, r)) {
		nss = ieee80211_rate_get_nss(r);
		mcs = ieee80211_rate_get_mcs(r);
		cbw = ieee80211_ppdu_bandwidth(r);
	}

	if (mcs >= max_mcs) {
		ieee80211_rate_set_vht(r, max_mcs, nss);
		return ret;
	}

	/* check whether mapping ht/vht rate to higher he rate */
	if (is_highest) {
		int group = (cbw * 2 + (nss - 1));
		int i = mcs - 7;	/* start from MCS7 */

		spin_lock_irqsave(&sta_priv->lock, flags);
		for (; i < GROUP_RATES - 1; i++) {
			int j = group * GROUP_RATES + i;
			u8 prob2 = stats[j + 1].prob;

			if (i + 7 == max_mcs)
				break;

			if (stats[j].prob <= rate_prob_level || stats[j].tot_atmp < 128)
				break;

			if (stats[j].prob > rate_prob_higher_level && stats[j].tot_atmp >= 128 &&
			    ((stats[j + 1].prob == 0 &&
			      time_in_range(stats[j + 1].jiffies, 0UL, 0UL)) ||
			     (stats[j + 1].attempts + stats[j + 1].tot_atmp <= 8 &&
			      stats[j + 1].prob > 0))) {
				ret = PROBE_HE_RATE_HIGHER;	/* probe this higher rate */
				i++;
				break;
			}

			if (stats[j].prob >= rate_prob_level &&
			    PROB_FRAC(prob2, stats[j].prob) > higher_tp_level[i]) {
				if (prob2 > rate_prob_level && stats[j + 1].tot_atmp >= 128) {
					continue;
				} else {
					if (stats[j + 1].attempts + stats[j + 1].tot_atmp <= 8)
						ret = PROBE_HE_RATE_HIGHER;
					i++;
					break;
				}
			} else {
				break;
			}
		}

		spin_unlock_irqrestore(&sta_priv->lock, flags);

		if ((mcs != i + 7) && ret != PROBE_HE_RATE_HIGHER)
			ret = MAP_HE_RATE_HIGHER;

		ieee80211_rate_set_vht(r, i + 7, nss);
	}

	return ret;
}

/**
 * scm80211_fixup_tx_info_he() - fix up tx rates
 * @hw: hardware
 * @sta: peer sta (can be NULL)
 * @ti: &struct ieee80211_tx_info set by mac80211
 *
 * If both ends of the link is HE STAs and mac80211 chose to send @skb in
 * VHT PPDU, we promote it to send in HE SU PPDU.
 *
 * The flag IEEE80211_TX_RC_GREEN_FIELD is not used internally, and redefined
 * as IEEE80211_TX_RC_HE_MCS to denote HE SU PPDU.
 */
static int scm80211_fixup_tx_info_he(struct ieee80211_hw *hw,
				     struct ieee80211_vif *vif,
				     struct ieee80211_sta *sta,
				     struct ieee80211_tx_info *ti, struct ieee80211_tx_rate *rate)
{
	struct ieee80211_sta_priv *sta_priv;
	struct ieee80211_hw_priv *priv = hw->priv;
	int ret = -1;
	u8 max_mcs = 7;

	if (priv->local_rc_used) {
		if (ieee80211_rate_is_he(rate)) {
			if (ieee80211_is_he_sta(sta))
				scm80211_fixup_he_omi(sta, rate);
			return 0;
		}
		goto exit;
	}

	if (!vif)
		goto exit;

	if (!ieee80211_get_he_iftype_cap(hw->wiphy->bands[ti->band], vif->type))
		goto exit;

	if (!vif->bss_conf.he_support)
		goto exit;

	if (!ieee80211_is_he_sta(sta))
		goto exit;

	sta_priv = (void *)sta->drv_priv;
	if (!sta_priv->n_ss)
		goto exit;

	max_mcs += 2 * sta_priv->he_mcs[0];

	ret = scm80211_map_to_he_rate(rate, sta, sta_priv->n_ss, max_mcs);

exit:
	return ret;
}

static bool is_p2p_mgmt_frame(struct sk_buff *skb)
{
	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *)skb->data;
	struct element *elem = NULL;

	if (ieee80211_is_probe_resp(mgmt->frame_control) ||
	    ieee80211_is_beacon(mgmt->frame_control)) {
		elem = (void *)mgmt->u.beacon.variable;
	} else if (ieee80211_is_probe_req(mgmt->frame_control)) {
		elem = (void *)mgmt->u.probe_req.variable;
	} else if (ieee80211_is_assoc_req(mgmt->frame_control)) {
		elem = (void *)mgmt->u.assoc_req.variable;
	} else {
		return false;
	}

	if (elem && elem->id == WLAN_EID_SSID && elem->datalen >= 7) {
		if (!memcmp(elem->data, "DIRECT-", 7))
			return true;
	}

	return false;
}

static void scm80211_fixup_tx_info(struct txop *txop,
				   struct ieee80211_sta *sta, struct ieee80211_tx_info *ti)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_tx_rate *rate = txop->rate;
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	u32 basic_rates = vif->bss_conf.basic_rates;
	int ret;

	if (rate->idx == -1) {
		struct ieee80211_supported_band *sband;

		/*
		 * In case mac80211 rate controller could not provide a
		 * suitable rate, the driver tries the lowest rate sta or
		 * hw support.
		 */
		sband = hw->wiphy->bands[ti->band];
		rate->idx = rate_lowest_index(sband, sta);
		rate->count = ((ti->flags & IEEE80211_TX_CTL_NO_ACK) ? 1 : hw->max_rate_tries);

		scm80211_warn(hw, "- txop rate adjusted to %d\n", rate->idx);
	}

	/* To cope with malfunction of mac80211 stack after CSA */
	if (ieee80211_iface_is_ap(vif->type)) {
		if (vif->bss_conf.chandef.width < NL80211_CHAN_WIDTH_40)
			rate->flags &= ~(IEEE80211_TX_RC_40_MHZ_WIDTH |
					 IEEE80211_TX_RC_80_MHZ_WIDTH);
		else if (vif->bss_conf.chandef.width < NL80211_CHAN_WIDTH_80)
			rate->flags &= ~IEEE80211_TX_RC_80_MHZ_WIDTH;
	}

	/*
	 * FIXME:
	 * If BSSbasicrates do not set DSSS/CCK rates in 2GHz
	 * Assume that this bss do not support 11b mode
	 */
	if (ti->band == NL80211_BAND_2GHZ &&
	    ((basic_rates && (basic_rates & 0xF) == 0) ||
	     (is_p2p_mgmt_frame(txop->skb))) && ieee80211_rate_is_non_ht(rate) && rate->idx < 4) {
		scm80211_warn(hw, "- improve basic rates %d\n", rate->idx);
		rate->idx = 4;
	}

#ifdef ERRATA_TX_RFLIMIT
	if (ieee80211_rate_is_he(rate)) {
		int mcs = ieee80211_rate_get_vht_mcs(rate);
		int nss = ieee80211_rate_get_vht_nss(rate);

		if (mcs > 9) {
			ti->flags &= ~IEEE80211_TX_CTL_RATE_CTRL_PROBE;
			ieee80211_rate_set_vht(rate, 9, nss);
		}
	}
#endif

	ret = scm80211_fixup_tx_info_he(hw, vif, sta, ti, txop->rate);
	if (ret != -1) {
		struct ieee80211_tx_info *info = IEEE80211_SKB_CB(txop->skb);

		if (ret == PROBE_HE_RATE_HIGHER) {
			info->flags |= IEEE80211_TX_CTL_RATE_CTRL_PROBE;
			scm80211_trace(hwq->hw, "- new rate[%d]: idx=%d flags=%08x\n",
				       scm80211_tx_rate_index(txop),
				       txop->rate->idx, txop->rate->flags);
		}

		txop->he = true;
	}

	scm80211_fixup_tx_info_flags(hw, sta, &txop->ti, txop->rate);
	if (txop->subused) {
		if (txop->rate->flags != txop->subrate.flags)
			txop->subrate.flags = txop->rate->flags;

		if (ieee80211_rate_get_nss(txop->rate) != ieee80211_rate_get_nss(&txop->subrate))
			ieee80211_rate_set(&txop->subrate,
					   ieee80211_rate_get_mcs(&txop->subrate),
					   ieee80211_rate_get_nss(txop->rate));

		scm80211_trace(hw, "- subrate-1 idx=%d flags=0x%08x\n",
			       txop->subrate.idx, txop->subrate.flags);
	}
}

/**
 * scm80211_get_tx_retries() - returns the number of retries
 *
 * See scm80211_process_done_txop().
 */
static int scm80211_get_tx_retries(struct ieee80211_tx_info *ti)
{
	int retries;

	retries = ti->status.rates[SCM80211_TX_RETRY_RATE].count;
	return (retries > 0) ? retries - 1 : 0;
}

/**
 * scm80211_clear_tx_rates() - clear tx rates
 * @ti: &struct ieee80211_tx_info
 *
 * This function is called when the skb which is the owner of @ti
 * is handed over to the driver for the first time.
 */
static inline void scm80211_clear_tx_rates(struct ieee80211_tx_info *ti, int start)
{
	int i;

	for (i = start; i < IEEE80211_TX_MAX_RATES - 1; i++) {
		ti->status.rates[i].idx = -1;
		ti->status.rates[i].flags = 0;
		ti->status.rates[i].count = 0;
	}
}

struct txvector default_txvector = {
	.FORMAT			= NON_HT,
	.MCS			= 0,
	.PREAMBLE_TYPE		= LONG_PREAMBLE,
	.L_LENGTH		= 0,
	.FEC_CODING		= BCC_CODING,
	.TXPWR_LEVEL_INDEX	= 0,
	.LENGTH			= 0,
	.SPATIAL_REUSE		= 0,
	.APEP_LENGTH		= 0,
	.CH_BANDWIDTH		= CBW20,
	.STBC			= false,
	.RU_ALLOCATION		= 0,
	.HE_SIG_A2_RESERVED	= 0,
	.TXOP_DURATION		= 0,
	.DYN_BANDWIDTH_IN_NON_HT = DYN_BW_STATIC,
	.CH_BANDWIDTH_IN_NON_HT	= CBW20,
	.MIDAMBLE_PERIODICITY	= 0,
	.PARTIAL_AID		= 0,
	.GROUP_ID		= 0,
	.GI_TYPE		= LONG_GI,
	.SCRAMBLER_INITIAL_VALUE = 0,
	.LDPC_EXTRA_SYMBOL	= 0,
	.BSS_COLOR		= 0,
	.HE_LTF_TYPE		= 0,
	.TOTAL_NUM_STS		= 0,
	.NUM_HE_LTF		= 0,
	.NOMINAL_PACKET_PADDING	= 0,
	.DEFAULT_PE_DURATION	= 0,
	.STARTING_STS_NUM	= 0,
	.SMOOTHING		= 0,
	.AGGREGATION		= false,
	.DCM			= 0,
	.NUM_STS		= 0,
	.DOPPLER		= 0,
	.HE_LTF_MODE		= 0,
	.TX_OP_PS_NOT_ALLOWED	= 0,
	.TRIGGER_METHOD		= 0,
	.UPLINK_FLAG		= 0,
	.TRIGGER_RESPONDING	= 0,
	.HE_TB_DATA_SYMBOLS	= 0,
	.FEEDBACK_STATUS	= 0,
	.HE_TB_PE_DISAMBIGUITY	= 0,
};

enum ppdu_format txvector_format(struct ieee80211_tx_rate *rate)
{
	if (rate->flags & IEEE80211_TX_RC_HE_MCS)
		return HE_SU;
	else if (rate->flags & IEEE80211_TX_RC_VHT_MCS)
		return VHT;
	else if (rate->flags & IEEE80211_TX_RC_MCS)
		return HT_MF;
	else
		return NON_HT;
}

enum cbw txvector_ch_bandwidth(struct ieee80211_tx_rate *rate)
{
	if (rate->flags & IEEE80211_TX_RC_160_MHZ_WIDTH)
		return CBW160;
	if (rate->flags & IEEE80211_TX_RC_80_MHZ_WIDTH)
		return CBW80;
	if (rate->flags & IEEE80211_TX_RC_40_MHZ_WIDTH)
		return CBW40;
	return CBW20;
}

static void tv_set_he_ppe(struct txvector *tv, struct ieee80211_sta *sta)
{
	struct ieee80211_sta_priv *sta_priv = NULL;
	u16 npp;
	int x, b;

	/**
	 * Remove warning here
	 * Null Data will be send on he rate by using mgmt queue
	 * Setting padding to max 2 means 16us
	 **/
	if (!ieee80211_is_he_sta(sta)) {
		tv->NOMINAL_PACKET_PADDING = 2;
		return;
	}

	b = tv->CH_BANDWIDTH + tv->DCM;
	if (WARN_ON(b > CBW80))
		b = CBW80;

	sta_priv = ieee80211_to_priv(sta);
	npp = sta_priv->he_npp[tv->NUM_STS][b];
	x = he_constellation_index(tv->MCS);
	npp >>= x * 2;
	tv->NOMINAL_PACKET_PADDING = (npp & 0x3);
}

static void tv_set_he_txop_duration(struct txvector *tv, struct ieee80211_vif *vif, u16 duration)
{
	if (!ppdu_is_he(tv))
		return;
	if (!vif)
		return;

	if (!is_he_bss_color_enabled(vif)) {
		duration = 0x7f;	/* UNSPECIFIED */
	} else if (duration < 512) {
		duration = (duration / 8) << 1;
	} else {
		if (duration > 8448)
			duration = 8448;
		duration = ((duration - 512) / 128) << 1;
		duration |= 0x1;
	}

	tv->TXOP_DURATION = duration;
}

static void tv_set_he_ltf_gi(struct txop *txop, struct txvector *tv, struct ieee80211_tx_info *info)
{
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	struct ieee80211_hw_priv *priv = txop->hw->priv;
	/*
	 * PHY capability information
	 * - HE SU PPDU with 1x HE-LTF and 0.8us GI (B14)
	 * - HE SU PPDU and HE MU PPDU with 4x HE-LTF and 0.8us GI (B58)
	 * - HE ER SU PPDU with 1x HE-LTF and 0.8us (B68)
	 * - Midamble Tx/Rx 2x and 1x HE-LTF (B69)
	 *
	 * HE-LTF type and GI duration combination for HE SU formats (Table 27-31)
	 * HE-LTF   GI         HE SU             HE-SIG-A1 GI+LTF Size
	 * ------------------------------------------------------------
	 * 1x       0.8us      O                 0
	 * 1x       1.6us      N/A               X
	 * 2x       0.8us      M                 1
	 * 2x       1.6us      M                 2
	 * 4x       0.8us      CM1               3 (DCM=1, STBC=1)
	 * 4x       3.2us      M                 3 (otherwise)
	 */
	if (priv->local_rc_used) {
		struct ieee80211_tx_rate *rate = txop->rate;

		if (rate->flags & IEEE80211_HE_GI_3u2) {
			tv->HE_LTF_TYPE = HE_LTF_4x;
			tv->GI_TYPE = GI_3u2s;
		} else if (rate->flags & IEEE80211_HE_GI_1u6) {
			tv->GI_TYPE = GI_1u6s;
			tv->HE_LTF_TYPE = HE_LTF_2x;
		} else {
			if (ieee80211_is_he_sta(sta) &&
			    (get_ieee80211_he_cap(sta).he_cap_elem.phy_cap_info[1] &
			     IEEE80211_HE_PHY_CAP1_HE_LTF_AND_GI_FOR_HE_PPDUS_0_8US)) {
				tv->HE_LTF_TYPE = HE_LTF_1x;
			} else {
				tv->HE_LTF_TYPE = HE_LTF_2x;
			}
			tv->GI_TYPE = GI_0u8s;
		}
	} else {
		int retries;

		retries = scm80211_get_tx_retries(info);
		if (retries == 0) {
			if (ieee80211_is_he_sta(sta) &&
			    (get_ieee80211_he_cap(sta).he_cap_elem.phy_cap_info[1] &
			     IEEE80211_HE_PHY_CAP1_HE_LTF_AND_GI_FOR_HE_PPDUS_0_8US)) {
				tv->HE_LTF_TYPE = HE_LTF_1x;
				tv->GI_TYPE = GI_0u8s;
			} else {
				tv->HE_LTF_TYPE = HE_LTF_2x;
				tv->GI_TYPE = GI_0u8s;
			}
#ifdef TEST_NO_HE_LTF1x_GI0u8s
			/*
			 * - Intel AX sniffer does not display HE SU PPDU with 1x HE-LTF 0.8us GI.
			 */
			tv->HE_LTF_TYPE = HE_LTF_2x;
			tv->GI_TYPE = GI_0u8s;
#endif
		} else if (retries < 4) {
			tv->HE_LTF_TYPE = HE_LTF_2x;
			tv->GI_TYPE = GI_1u6s;
		} else {
			tv->HE_LTF_TYPE = HE_LTF_4x;
			tv->GI_TYPE = GI_3u2s;
		}
	}
}

static s8 scm80211_get_txpower(struct txop *txop)
{
	struct txvector *tv = &txop->tv;
	enum nl80211_band band = txop->ti.band;
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	s8 power = -128;

	if (tx_power != -128)
		return (s8)tx_power;

	switch (tv->FORMAT) {
	case NON_HT:
		power = tx_nonht_table[band][tv->MCS];
		break;
	case HT_MF:
		power = tx_ht_table[band][tv->CH_BANDWIDTH][tv->MCS & 0x07];
		break;
	case VHT:
		power = tx_vht_table[band][tv->CH_BANDWIDTH][tv->MCS];
		break;
	case HE_SU:
		power = tx_he_table[band][tv->CH_BANDWIDTH][tv->MCS];
		break;
	default:
		scm80211_warn(txop->hw, "- fmt invalid %d\n", tv->FORMAT);
		power = 8;
		break;
	}

	if (sta && get_ieee80211_txpwr(sta).type != NL80211_TX_POWER_AUTOMATIC &&
	    get_ieee80211_txpwr(sta).power < S8_MAX && get_ieee80211_txpwr(sta).power > S8_MIN) {
		switch (get_ieee80211_txpwr(sta).type) {
		case NL80211_TX_POWER_LIMITED:
			if (power > (s8)get_ieee80211_txpwr(sta).type)
				power = (s8)get_ieee80211_txpwr(sta).type;
			break;
		case NL80211_TX_POWER_FIXED:
			power = (s8)get_ieee80211_txpwr(sta).type;
			break;
		default:
			break;
		}
	}

	if (WARN_ON(power == -128))
		power = 8;
	return power;
}

/**
 * scm80211_build_txvector() - populate TXVECTOR
 *
 * @hw: the hardware
 * @vif: virtual interface
 * @sta: peer sta (can be NULL)
 * @skb: frame to be transmitted
 * @rate: selected ppdu parameters to be used for transmission
 *
 * This function populates tv fields of skb metadata
 * (&struct scm80211_tx_info).
 *
 * This function must be called once the transmission
 * rate @rate has been selected.
 */
static void scm80211_build_txvector(struct txop *txop, struct sk_buff *skb,
				    struct txvector *tv, struct ieee80211_tx_rate *rate)
{
	struct ieee80211_hw *hw = txop->hwq->hw;
	struct ieee80211_tx_info *ti = &txop->ti;
	struct ieee80211_supported_band *sband = hw->wiphy->bands[ti->band];
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	int nsts;

	*tv = default_txvector;

	tv->FORMAT = txvector_format(rate);
	tv->CH_BANDWIDTH = txvector_ch_bandwidth(rate);

	if ((ti->flags & IEEE80211_TX_CTL_STBC))
		tv->STBC = true;
	if (ti->flags & IEEE80211_TX_CTL_LDPC)
		tv->FEC_CODING = LDPC_CODING;
	if (rate->flags & IEEE80211_TX_RC_SHORT_GI)
		tv->GI_TYPE = SHORT_GI;

	if (ppdu_is_he(tv)) {
		/* D5.0 26.11 */
		tv->MCS = ieee80211_rate_get_vht_mcs(rate);
		nsts = ieee80211_rate_get_vht_nss(rate) * (tv->STBC ? 2 : 1);
		tv->NUM_STS = nsts - 1;
		tv->AGGREGATION = false;
		tv->BSS_COLOR = ieee80211_he_bss_color(&vif->bss_conf);
		tv->DEFAULT_PE_DURATION = vif->bss_conf.htc_trig_based_pkt_ext;

		if (vif && vif->type == NL80211_IFTYPE_STATION)
			tv->UPLINK_FLAG = 1;
		else
			tv->UPLINK_FLAG = 0;

		if (txop->he_gi_type) {
			/* now rate fixed */
			tv->DCM = txop->he_dcm;
			tv->GI_TYPE = txop->he_gi_type;
			tv->HE_LTF_TYPE = txop->he_ltf_type;
		} else {
			tv_set_he_ltf_gi(txop, tv, IEEE80211_SKB_CB(skb));
		}
		tv_set_he_ppe(tv, sta);

		/**
		 * BCC coding is not use in following cases
		 */
		if (tv->MCS > 9 || tv->CH_BANDWIDTH > CBW20)
			tv->FEC_CODING = LDPC_CODING;
		/*
		 * FIXME: fill other fields such as SPATIAL_REUSE
		 */
		tv->APEP_LENGTH = ieee80211_subframe_len(skb);

	} else if (ppdu_is_vht(tv)) {
		struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
		const u8 *addr = NULL;
		u16 aid = 0;

		tv->MCS = ieee80211_rate_get_vht_mcs(rate);
		tv->MCS = (tv->MCS > 9) ? 9 : tv->MCS;
		nsts = ieee80211_rate_get_vht_nss(rate) * (tv->STBC ? 2 : 1);
		tv->NUM_STS = nsts - 1;
		tv->AGGREGATION = false;
		tv->GROUP_ID = 63;

		switch (vif->type) {
		case NL80211_IFTYPE_STATION:
			if (ether_addr_equal(hdr->addr1, vif->bss_conf.bssid)) {
				tv->GROUP_ID = 0;
				aid = 0;
				addr = vif->bss_conf.bssid;
			} else if (ieee80211_is_vht_sta(sta) && sta->tdls &&
				   ether_addr_equal(hdr->addr1, sta->addr)) {
				tv->GROUP_ID = 63;
				aid = sta->aid;
				addr = vif->bss_conf.bssid;
			}
			break;
		case NL80211_IFTYPE_MESH_POINT:
			tv->GROUP_ID = 0;
			aid = 0;
			addr = hdr->addr1;	/* RA */
			break;
		case NL80211_IFTYPE_AP:
		case NL80211_IFTYPE_AP_VLAN:
		case NL80211_IFTYPE_P2P_GO:
			tv->GROUP_ID = 63;
			if (ieee80211_is_vht_sta(sta) && sta->aid > 0) {
				aid = sta->aid;
				addr = vif->addr;	/* BSSID */
			}
			break;
		default:
			break;
		}
		tv->PARTIAL_AID = ieee80211_vht_partial_aid(aid, addr, tv->GROUP_ID);
		tv->APEP_LENGTH = ieee80211_subframe_len(skb);
	} else if (ppdu_is_ht(tv)) {
		tv->MCS = rate->idx;
		nsts = ieee80211_rate_get_nss(rate) + tv->STBC;
		tv->NUM_STS = nsts - 1;
		if ((ti->flags & IEEE80211_TX_CTL_AMPDU) && txop->ampdu.enabled) {
			tv->AGGREGATION = true;
			tv->LENGTH = ieee80211_subframe_len(skb);
		} else {
			tv->AGGREGATION = false;
			tv->LENGTH = ieee80211_mpdu_len(skb);
		}
	} else {
		tv->MCS = sband->bitrates[rate->idx].hw_value;
		tv->CH_BANDWIDTH_IN_NON_HT = tv->CH_BANDWIDTH;
		if (rate->flags & IEEE80211_TX_RC_USE_SHORT_PREAMBLE)
			tv->PREAMBLE_TYPE = SHORT_PREAMBLE;
		tv->LENGTH = ieee80211_mpdu_len(skb);
	}

	tv->TXPWR_LEVEL_INDEX = scm80211_get_txpower(txop);

	if (txop->subused) {
		memcpy(&txop->subtv, tv, sizeof(txop->subtv));
		if (ppdu_is_ht(tv))
			txop->subtv.MCS = txop->subrate.idx;
		else
			txop->subtv.MCS = ieee80211_rate_get_mcs(&txop->subrate);
	}
}

static void txvector_set_psdu_len(struct txvector *tv, size_t len)
{
	switch (tv->FORMAT) {
	case NON_HT:
		tv->L_LENGTH = len;
		break;
	case HT_MF:
	case HT_GF:
		tv->LENGTH = len;
		break;
	case VHT:
	case HE_SU:
	case HE_ER_SU:
	case HE_MU:
	case HE_TB:
		tv->APEP_LENGTH = len;
	}
}

static char *ppdu_format_str[] = {
	"NON-HT",
	"HT-MF",
	"HT-GF",
	"VHT",
	"HE-SU",
	"HE-ER",
	"HE-MU",
	"HE-TB",
};

static const char * const fec_coding_str[] = {
	"BCC", "LDPC",
};

static const char * const gi_str[] = {
	"0.4u", "0.8u", "1.6u", "3.2u",
};

static int n_ss(struct txvector *tv)
{
	switch (tv->FORMAT) {
	case NON_HT:
		return 1;
	case HT_MF:
	case HT_GF:
		return (tv->NUM_STS + 1) - tv->STBC;
	case VHT:
	case HE_SU:
	case HE_ER_SU:
	case HE_MU:
	case HE_TB:
		return (tv->NUM_STS + 1) >> tv->STBC;
	}
	return 0;
}

static void scm80211_print_txvector(char *buf, size_t len, struct ieee80211_hw *hw,
				    struct ieee80211_sta *sta, struct txvector *tv)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	const char *prefix = "- txvector: ";
	char *ptr = buf, *end = buf + len;
	enum nl80211_band band = (priv->band & BIT(NL80211_BAND_2GHZ)) ?
	    NL80211_BAND_2GHZ : NL80211_BAND_5GHZ;
	struct ieee80211_supported_band *sband = hw->wiphy->bands[band];

	ptr += snprintf(ptr, end - ptr, "%s", prefix);

	if (tv->FORMAT == NON_HT) {
		int index;
		u16 rate;

		index = tv->MCS;
		rate = sband->bitrates[index].bitrate;	/* in 100Kbps */

		ptr += snprintf(ptr, end - ptr,
				"NON-HT %s %d%sMbps ",
				tv->MCS < 4 ? "11b" : "11a/g", rate / 10, (rate % 10) ? ".5" : "");

		if (tv->MCS >= 4 && tv->CH_BANDWIDTH_IN_NON_HT)
			ptr += snprintf(ptr, end - ptr, "%dMhz dup ",
					20 * (1 << tv->CH_BANDWIDTH_IN_NON_HT));

		if (tv->MCS < 4)
			ptr += snprintf(ptr, end - ptr, "%s ",
					tv->PREAMBLE_TYPE ? "short-preamble" : "long-preamble");

		ptr += snprintf(ptr, end - ptr, "len:%d ", tv->L_LENGTH);
		if (sta && ieee80211_is_vht_sta(sta))
			ptr += snprintf(ptr, end - ptr, " VHT BW oper:%s",
					tv->DYN_BANDWIDTH_IN_NON_HT ? "dynamic" : "static");
		ptr += snprintf(ptr, end - ptr, "\n");
		goto out;
	}

	/* HT, VHT, and HE */
	ptr += snprintf(ptr, end - ptr, "%s-%dMHz-%dSS%s-%s-MCS%d-%sGI len:%d ",
			ppdu_format_str[tv->FORMAT],
			20 * (1 << tv->CH_BANDWIDTH),
			n_ss(tv),
			tv->STBC ? "-STBC" : "",
			fec_coding_str[tv->FEC_CODING],
			tv->MCS,
			gi_str[tv->GI_TYPE], (tv->FORMAT < VHT) ? tv->LENGTH : tv->APEP_LENGTH);

	if (ppdu_is_ht(tv)) {
		ptr += snprintf(ptr, end - ptr, "%s", tv->AGGREGATION ? "ampdu" : "mpdu");
	} else if (ppdu_is_vht(tv)) {
		ptr +=
		    snprintf(ptr, end - ptr, "PARTIAL_AID:%d GROUP_ID:%d TXOP_PS_NOT_ALLOWED:%d\n",
			     tv->PARTIAL_AID, tv->GROUP_ID, tv->TX_OP_PS_NOT_ALLOWED);
	} else {
		ptr += snprintf(ptr, end - ptr, "HE-LTF-%dx NPP:%dus ",
				1 << tv->HE_LTF_TYPE, tv->NOMINAL_PACKET_PADDING * 8);

		ptr += snprintf(ptr, end - ptr, "SR:%04x TXOP:%d BSS_COLOR:%d\n",
				tv->SPATIAL_REUSE, tv->TXOP_DURATION, tv->BSS_COLOR);
	}
out:
	scm80211_trace(hw, "%s", buf);
}

static void scm80211_print_txvector_raw(struct ieee80211_hw *hw, struct txvector *tv)
{
#ifdef SCM80211_TXOP_DUMP
	struct wiphy *wiphy = hw->wiphy;

	wiphy_dbg(wiphy, "- txvector:");

	wiphy_dbg(wiphy,
		  "  [0] MCS=%d PREAMBLE=%d L_LENGTH=%d FORMAT=%d FEC=%d PWR=%d\n",
		  tv->MCS,
		  tv->PREAMBLE_TYPE,
		  tv->L_LENGTH, tv->FORMAT, tv->FEC_CODING, tv->TXPWR_LEVEL_INDEX);
	wiphy_dbg(wiphy, "  [1] LENGTH=%d SPATIAL_REUSE=%x\n", tv->LENGTH, tv->SPATIAL_REUSE);
	wiphy_dbg(wiphy,
		  "  [2] APEP_LENGTH=%d CH_BANDWIDTH=%d STBC=%d\n",
		  tv->APEP_LENGTH, tv->CH_BANDWIDTH, tv->STBC);
	wiphy_dbg(wiphy,
		  "  [3] TXOP=%d PARTIAL_AID=%d DYN_BW_NON_HT=%d CH_BW_NON_HT=%d\n",
		  tv->TXOP_DURATION,
		  tv->PARTIAL_AID, tv->DYN_BANDWIDTH_IN_NON_HT, tv->CH_BANDWIDTH_IN_NON_HT);
	wiphy_dbg(wiphy, "  [4] GROUP_ID=%d, GI_TYPE=%d\n", tv->GROUP_ID, tv->GI_TYPE);
	wiphy_dbg(wiphy,
		  "  [5] BSS_COLOR=%d HE_LTF_TYPE=%d NPP=%d DPE=%d SMOOTHING=%d AGGR=%d\n",
		  tv->BSS_COLOR,
		  tv->HE_LTF_TYPE,
		  tv->NOMINAL_PACKET_PADDING,
		  tv->DEFAULT_PE_DURATION, tv->SMOOTHING, tv->AGGREGATION);
	wiphy_dbg(wiphy,
		  "  [6] DCM=%d, NUM_STS=%d, HE_LTF_MODE=%d UPLINK_FLAG=%d\n",
		  tv->DCM, tv->NUM_STS, tv->HE_LTF_MODE, tv->UPLINK_FLAG);
#endif
}

const char *const txop_state_str[] = {
	[TXOP_STATE_IDLE] = "idle",
	[TXOP_STATE_NEEDS_RESCHED] = "resched",
	[TXOP_STATE_SCHEDULED] = "sched",
	[TXOP_STATE_ISSUING] = "issuing",
	[TXOP_STATE_ISSUED] = "issued",
	[TXOP_STATE_TX_DONE] = "done",
	[TXOP_STATE_TX_RETRY] = "retry",
	[TXOP_STATE_END] = "completed",
};

/**
 * DOC: hardware queue manipulation
 */
bool scm80211_txop_is_running(struct scm80211_hwq *hwq)
{
	unsigned long flags;
	bool empty;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	empty = list_empty(&hwq->ring.head[TXOP_STATE_ISSUED]);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);

	return !empty;
}

struct txop *__scm80211_get_txop(struct scm80211_hwq *hwq, int state)
{
	struct txop *txop = NULL;

	txop = list_first_entry_or_null(&hwq->ring.head[state], struct txop, list);

	if (txop) {
		list_del_init(&txop->list);
		scm80211_trace(txop->hw,
			       "hwq#%d.%d: get txop from state %s list, txop in state %s\n",
			       txop->hwq->id, txop->id, txop_state_str[state],
			       txop_state_str[txop->state]);
	}

	return txop;
}

bool scm80211_has_txop(struct scm80211_hwq *hwq, int state)
{
	return !list_empty(&hwq->ring.head[state]);
}

struct txop *scm80211_get_txop(struct scm80211_hwq *hwq, int state)
{
	struct txop *txop = NULL;
	unsigned long flags;

	spin_lock_irqsave(&hwq->ring.lock, flags);

	txop = __scm80211_get_txop(hwq, state);

	spin_unlock_irqrestore(&hwq->ring.lock, flags);

	if (txop)
		BUG_ON(txop->state != state);

	return txop;
}

void __scm80211_txop_add_back(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;

	scm80211_trace(txop->hw, "hwq#%d.%d: return to state %s list\n",
		       txop->hwq->id, txop->id, txop_state_str[txop->state]);

	list_move(&txop->list, &hwq->ring.head[txop->state]);
}

void scm80211_txop_add_back(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	unsigned long flags;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	__scm80211_txop_add_back(txop);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
}

void __scm80211_txop_add(struct txop *txop, int state)
{
	struct scm80211_hwq *hwq = txop->hwq;

	scm80211_trace(txop->hw, "hwq#%d.%d: add state %s -> %s list\n",
		       txop->hwq->id, txop->id,
		       txop_state_str[txop->old_state], txop_state_str[state]);

	BUG_ON(!list_empty(&txop->list));

	if (txop->old_state == TXOP_STATE_NEEDS_RESCHED && state == TXOP_STATE_SCHEDULED)
		list_move(&txop->list, &hwq->ring.head[state]);
	else
		list_move_tail(&txop->list, &hwq->ring.head[state]);
}

void scm80211_txop_add(struct txop *txop, int state)
{
	struct scm80211_hwq *hwq = txop->hwq;
	unsigned long flags;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	__scm80211_txop_add(txop, state);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
}

void __scm80211_txop_change_state(struct txop *txop, int state)
{
	txop->old_state = txop->state;
	txop->state = state;

	scm80211_trace(txop->hw, "hwq#%d.%d: state %s -> %s\n",
		       txop->hwq->id, txop->id,
		       txop_state_str[txop->old_state], txop_state_str[state]);
}

void scm80211_txop_change_state(struct txop *txop, int state)
{
	struct scm80211_hwq *hwq = txop->hwq;
	unsigned long flags;

	spin_lock_irqsave(&hwq->ring.lock, flags);
	__scm80211_txop_change_state(txop, state);
	spin_unlock_irqrestore(&hwq->ring.lock, flags);
}

void scm80211_txop_free(struct txop *txop)
{
	memset(txop, 0, offsetof(struct txop, dma[0]));
	scm80211_txop_change_state(txop, TXOP_STATE_IDLE);
	scm80211_txop_add(txop, TXOP_STATE_IDLE);
}

struct txop *scm80211_txop_alloc(struct scm80211_hwq *hwq)
{
	struct txop *txop;

	txop = scm80211_get_txop(hwq, TXOP_STATE_IDLE);
	return txop;
}

bool scm80211_txop_ring_empty(struct scm80211_hwq *hwq)
{
	int i;
	unsigned long flags;
	bool idle = true;

	spin_lock_irqsave(&hwq->ring.lock, flags);

	for (i = 0; i < ARRAY_SIZE(hwq->ring.pool); i++) {
		if (hwq->ring.pool[i].state >= TXOP_STATE_NEEDS_RESCHED) {
			idle = false;
			break;
		}
		if (list_empty(&hwq->ring.pool[i].list)) {
			idle = false;
			break;
		}
	}

	spin_unlock_irqrestore(&hwq->ring.lock, flags);
	return idle;
}

static struct scm80211_txq *scm80211_get_txq_from_tid(struct txop *txop, int tid)
{
	struct scm80211_txq *txq;

	if (!txop->txq[1])
		txq = txop->txq[0];
	else
		txq = (tid == txop->txq[0]->txq.tid) ? txop->txq[0] : txop->txq[1];

	/* debugging */
	WARN_ON(!txq);
	return txq;
}

/**
 * scm80211_discard_schedule() - discards the scheduled txop
 *
 * FIXME: multi-tid
 */
static void scm80211_discard_schedule(struct txop *txop)
{
	struct sk_buff_head queue[2];
	struct sk_buff *skb;
	struct scm80211_txq *txq;
	struct scm80211_hwq *hwq = txop->hwq;
	int i = 0;

	if (!txop || txop->state != TXOP_STATE_SCHEDULED)
		return;

	skb_queue_head_init(&queue[0]);
	skb_queue_head_init(&queue[1]);

	while ((skb = skb_dequeue(&txop->sched))) {
		struct scm80211_tx_info *xti = &txop->xti[i];

		scm80211_update_tid_queue_size(txop, skb, false);
		if (!txop->retry && (xti->flags & SCM80211_XTI_DMA_MAPPED))
			scm80211_cleanup_tx(hwq->hw, txop, i);

		if (!txop->retry && xti->flags & SCM80211_XTI_HAS_TXVECTOR)
			skb_pull(skb, sizeof(struct txvector));

		txq = scm80211_get_txq_from_tid(txop, xti->tid);
		if (txq == txop->txq[0])
			skb_queue_tail(&queue[0], skb);
		else
			skb_queue_tail(&queue[1], skb);
		i++;
	}
	skb_queue_splice(&queue[0], &txop->txq[0]->queue);
	if (txop->txq[1])
		skb_queue_splice(&queue[1], &txop->txq[1]->queue);
}

/**
 * scm80211_discard_frame() - tells whether or not to drop a frame
 * @skb: frame under inspection
 */
static bool scm80211_discard_frame(struct ieee80211_hw *hw, struct sk_buff *skb)
{
	struct ieee80211_tx_info *ti = IEEE80211_SKB_CB(skb);
	struct ieee80211_vif *vif = ti->control.vif;
	int limit = hw->conf.short_frame_max_tx_count - 1;
	int retries = scm80211_get_tx_retries(ti);

	/*
	 * NB: QoS STA maintains only short retry counter and discards
	 * MSDU, A-MSDU or MMPDU if it reaches dot11ShortRetryLimit
	 */
	if ((!vif || vif->bss_conf.qos == false) &&
	    ti->control.rates[0].flags & IEEE80211_TX_RC_USE_RTS_CTS)
		limit = hw->conf.long_frame_max_tx_count - 1;

	return retries >= limit;
}

static inline bool scm80211_block_ack_is_missed(struct txdesc *txd)
{
	if (txd->ok)
		return false;
	return (txd->fail == ack_timeout || txd->fail == ack_invalid);
}

static void scm80211_save_recmded_rate(struct ieee80211_hw *hw,
				       struct ieee80211_sta *sta, __le32 htc)
{
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_tx_rate rate;
	bool stbc_or_dcm;

	memset(&rate, 0, sizeof(rate));

	if (!ieee80211_link_adap_to_rate(hw, htc, &rate, &stbc_or_dcm)) {
		unsigned long flag;

		spin_lock_irqsave(&sta_priv->lock, flag);
		sta_priv->rate_available = true;
		sta_priv->stbc_or_dcm = stbc_or_dcm;
		sta_priv->recomd_rate = rate;
		spin_unlock_irqrestore(&sta_priv->lock, flag);
	}
}

void scm80211_disable_mu_edca(struct scm80211_hwq *hwq)
{
	int aifs;
	unsigned long flags;

	if (!hwq->edca.mu_edca)
		return;

	scm80211_trace(hwq->hw, "mu-edca disable hwq#%d\n", hwq->id);

	/**
	 * when the timer is currently executing the callback function
	 * no-MU EDCA will be reloaded in that function, just return here
	 * deadlock will be happen if using hrtimer_cancel function
	 */
	if (hrtimer_try_to_cancel(&hwq->mu_edca_timer) < 0)
		return;

	spin_lock_irqsave(&hwq->lock, flags);
	aifs = hwq->edca.aifs;
	/* Restore non-MU EDCA parameter */
	hwq->edca.aifs = hwq->b_edca.aifs;
	hwq->edca.cw_max = hwq->b_edca.cw_max;
	hwq->edca.cw_min = hwq->b_edca.cw_min;
	hwq->cw = hwq->edca.cw_min;

	if ((hwq->mu_edca_status & MU_EDCA_RUNNING) && aifs == 0) {
		struct txop *txop;
		unsigned long flags2;

		hwq->mu_edca_status &= ~MU_EDCA_RUNNING;
		spin_unlock_irqrestore(&hwq->lock, flags);

		spin_lock_irqsave(&hwq->ring.lock, flags2);
		txop = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
		if (txop && is_pcie_device(hwq->hw)) {
			u32 cmd = scm80211_mac_readl(hwq->hw, MAC_REG_TX_QUEN_CMD(hwq->id));

			if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd)) {
				/* scm80211_complete_txop(txop, false); */
				spin_unlock_irqrestore(&hwq->ring.lock, flags2);
				scm80211_cancel_tx(hwq->hw, txop, true);
			} else {
				__scm80211_txop_add_back(txop);
				spin_unlock_irqrestore(&hwq->ring.lock, flags2);
			}
		} else {
			spin_unlock_irqrestore(&hwq->ring.lock, flags2);
		}
		return;
	}
	spin_unlock_irqrestore(&hwq->lock, flags);
}

void scm80211_enable_mu_edca(struct scm80211_hwq *hwq)
{
	u8 ecw_min_max = hwq->edca.mu_edca_param_rec.ecw_min_max;
	u64 mu_edca_timer = hwq->edca.mu_edca_param_rec.mu_edca_timer;
	unsigned long flags;

	if (!ecw_min_max)
		ecw_min_max = 0xf;

	if (!mu_edca_timer)
		mu_edca_timer = 0xff;

	spin_lock_irqsave(&hwq->lock, flags);

	if (!(hwq->mu_edca_status & MU_EDCA_RUNNING)) {
		/* backup EDCA param */
		hwq->b_edca = hwq->edca;
		hwq->edca.cw_max = (1 << FIELD_GET(0xF0, ecw_min_max)) - 1;
		hwq->edca.cw_min = (1 << FIELD_GET(0x0F, ecw_min_max)) - 1;
		hwq->edca.aifs = FIELD_GET(0x0F, hwq->edca.mu_edca_param_rec.aifsn);
	}

	scm80211_trace(hwq->hw, "mu-edca enable hwq#%d cwmax %d cwmin %d aifs %d\n",
		       hwq->id, hwq->edca.cw_max, hwq->edca.cw_min, hwq->edca.aifs);
	hwq->mu_timer = cpu_to_le16(ieee80211_tu_to_usec(mu_edca_timer << 3));
	/* start mu edca timer */
	mu_edca_timer = ieee80211_tu_to_usec(mu_edca_timer << 3) * 1000;
	hrtimer_start(&hwq->mu_edca_timer, ns_to_ktime(mu_edca_timer), HRTIMER_MODE_REL_SOFT);
	hwq->mu_edca_status |= MU_EDCA_RUNNING | MU_EDCA_CONFIGURED;
	spin_unlock_irqrestore(&hwq->lock, flags);
}

static void scm80211_handle_tdls(struct scm80211_hwq *hwq, struct sk_buff *skb)
{
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_hdr *hdr = (void *)skb->data;
	struct ieee80211_sta *sta;
	struct ieee80211_sta_priv *sta_priv;
	ktime_t next_tdls;
	struct tdls_action {
		u8 hdr[ETH_ALEN];
		__be16 proto;
		u8 type;
		u8 category;
		u8 action_code;
	} __packed;

	if (ieee80211_is_data_qos(hdr->frame_control) &&
	    priv->tdls.state == TDLS_CHANCTX_SWITCH_RESPONSE) {
		struct tdls_action *tdls;

		tdls = (struct tdls_action *)(skb->data + sizeof(struct ieee80211_qos_hdr));

		if (ether_addr_equal(tdls->hdr, rfc1042_header) &&
		    tdls->proto == htons(ETH_P_TDLS) &&
		    tdls->type == WLAN_TDLS_SNAP_RFTYPE &&
		    tdls->category == WLAN_CATEGORY_TDLS &&
		    tdls->action_code == WLAN_TDLS_CHANNEL_SWITCH_RESPONSE) {
			sta = priv->tdls.sta;
			sta_priv = (void *)sta->drv_priv;
			priv->tdls.packets = sta_priv->tx_packets + sta_priv->rx_packets;
			priv->tdls.started = true;

			ieee80211_queue_work(hw, &priv->tdls.work);

			if (priv->tdls.switch_timeout > 0) {
				next_tdls = ns_to_ktime(priv->tdls.switch_timeout * 1000);
				hrtimer_start(&priv->tdls.tdls_timer, next_tdls,
					      HRTIMER_MODE_REL_SOFT);
			}
		}
	}
}

/**
 * Adjust CW[AC] according to 10.24.2.2
 *
 * NB: We do not double CW[AC] in event e) if the TXOP was protected
 * by RTS/CTS for this is more likely because of temporary channel
 * disturbance, not because of collision in WM.
 */
static
void scm80211_adjust_cw(struct scm80211_hwq *hwq, bool all_ok, int fail, int succ_cnt)
{
	/*this may causes by some internal failure, skip to change cw */
	if (fail > 0)
		return;

	if (all_ok) {
		/* b) */
		hwq->qsrc = 0;
		hwq->cw = hwq->edca.cw_min;
	} else if (succ_cnt == 0) {
		/* c), d) */
		if (hwq->qsrc < hwq->hw->wiphy->retry_short) {
			hwq->qsrc++;
			hwq->cw = ((hwq->edca.cw_min + 1) << hwq->qsrc) - 1;
			hwq->cw = min(hwq->edca.cw_max, hwq->cw);
		} else {
			hwq->qsrc = 0;
			hwq->cw = hwq->edca.cw_min;
		}
	} else {
		/* e) non-final MPDU failure - cw unchanged */
	}
}

static void scm80211_he_rate_record(struct ieee80211_hw *hw, struct ieee80211_sta *sta,
				    struct ieee80211_tx_rate *rate, int n_mpdu, int n_ok)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);
	int index;

	if (ieee80211_rate_get_vht_mcs(rate) < 0x07)
		return;

	index = (ieee80211_ppdu_bandwidth(rate) * 2) + (ieee80211_rate_get_vht_nss(rate) - 1);
	index = index * GROUP_RATES + ieee80211_rate_get_vht_mcs(rate) - 7;

	if (index < GROUP_RATES * GROUP_COUNT) {
		sta_priv->rate_stats[index].attempts += n_mpdu;
		sta_priv->rate_stats[index].success += n_ok;
		sta_priv->rate_stats[index].jiffies = jiffies;
	}

	/* Environment condition turns bad, speed up the timer (10us) */
	if (n_mpdu / 2 >= n_ok)
		hrtimer_start(&priv->ratestats_timer, ns_to_ktime(10 * NSEC_PER_USEC),
			      HRTIMER_MODE_REL_SOFT);
}

static void scm80211_tx_bar_handle(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				struct sk_buff *skb)
{
	struct ieee80211_bar *bar = (struct ieee80211_bar *)skb->data;
	u16 tid  = le16_to_cpu(bar->control) >> IEEE80211_BAR_CTRL_TID_INFO_SHIFT;
	u16 ssn = IEEE80211_SEQ_TO_SN(le16_to_cpu(bar->start_seq_num));
	struct scm80211_hwq *hwq = scm80211_hwq(hw, vif->hw_queue[TID_TO_WME_AC(tid)]);

	if (hwq->on_bar && ssn == hwq->bar_ssn) {
		hwq->on_bar = false;
		scm80211_wake_queues(hw, BIT(hwq->id));
	}
}

#ifndef SCM80211_WFA_TEST
static void scm80211_ampdu_stat_record(struct ieee80211_hw *hw, struct ieee80211_sta *sta,
				       int ac, int n_mpdu, int n_ok)
{
	struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);

	if (ac >= IEEE80211_NUM_ACS)
		return;

	sta_priv->mpdu_stat[ac].tot_ampdus++;
	sta_priv->mpdu_stat[ac].tot_mpdus += n_mpdu;
	sta_priv->mpdu_stat[ac].tot_succ += n_ok;
	if (sta_priv->mpdu_stat[ac].tot_ampdus)
		sta_priv->mpdu_stat[ac].mpdus_per_ampdu =
		    sta_priv->mpdu_stat[ac].tot_mpdus / sta_priv->mpdu_stat[ac].tot_ampdus;
}
#endif

static void scm80211_tx_he_rate_update(struct ieee80211_sta *sta, u16 flags, s8 rate)
{
	struct ieee80211_sta_priv *sta_priv;
	struct rate_info *txinfo;
	unsigned long lockflags;

	if (!sta || !(flags & IEEE80211_TX_RC_HE_MCS))
		return;

	sta_priv = ieee80211_to_priv(sta);
	txinfo = &sta_priv->txinfo;

	spin_lock_irqsave(&sta_priv->lock, lockflags);

	txinfo->flags = RATE_INFO_FLAGS_HE_MCS;
	txinfo->mcs = rate & 0xf;
	txinfo->nss = (rate >> 4) + 1;
	if (flags & IEEE80211_TX_RC_80_MHZ_WIDTH)
		txinfo->bw = RATE_INFO_BW_80;
	else if (flags & IEEE80211_TX_RC_40_MHZ_WIDTH)
		txinfo->bw = RATE_INFO_BW_40;
	else
		txinfo->bw = RATE_INFO_BW_20;

	txinfo->he_dcm = 0;
	if (flags & IEEE80211_HE_GI_3u2)
		txinfo->he_gi = NL80211_RATE_INFO_HE_GI_3_2;
	else if (flags & IEEE80211_HE_GI_1u6)
		txinfo->he_gi = NL80211_RATE_INFO_HE_GI_1_6;
	else
		txinfo->he_gi = NL80211_RATE_INFO_HE_GI_0_8;

	spin_unlock_irqrestore(&sta_priv->lock, lockflags);
}

static void __used scm80211_tx_rate_update(struct txop *txop)
{
	struct ieee80211_hw *hw = txop->hwq->hw;
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	struct ieee80211_hw_priv *priv = hw->priv;

	if (sta && !txop->txd.tb) {
		struct ieee80211_sta_priv *sta_priv = ieee80211_to_priv(sta);

		if (ieee80211_rate_is_non_ht(txop->rate) && txop->limit == 0 && txop->acked)
			txop->rs.flags |= IEEE80211_TX_STAT_ACK;

		if (!sta_priv->rate_fix_on)
			ieee80211_tx_rate_update(hw, sta, &txop->rs);
		/*
		 * register tx airtime for per STA per TID, mac80211 will do TXQ scheduler
		 * in airtime fairness
		 */
		if (is_mac80211_txq(txop->owner))
			ieee80211_sta_register_airtime(sta, txop->owner->txq.tid, txop->txtime, 0);
	}

	if (txop->sta && txop->txd.cw && txop->txd.htc)
		scm80211_save_recmded_rate(hw, txop->sta, txop->txd.htc);

	if (!priv->local_rc_used && txop->sta && ieee80211_rate_is_he(txop->rate) && !txop->txd.tb)
		scm80211_he_rate_record(hw, txop->sta, txop->rate,
					txop->rs.status.ampdu_len, txop->rs.status.ampdu_ack_len);

#ifndef SCM80211_WFA_TEST
	if (txop->ampdu.enabled && sta)
		scm80211_ampdu_stat_record(hw, sta, txop->owner->txq.ac,
					   txop->rs.status.ampdu_len,
					   txop->rs.status.ampdu_ack_len);
#endif

	/*
	 * if local rc is not used update he rate info every time when send and all acked
	 * otherwise update it in @sta_rate_tbl_update()
	 */
	if (txop->rs.status.ampdu_len == txop->rs.status.ampdu_ack_len)
		scm80211_tx_he_rate_update(sta, txop->rate->flags, txop->rate->idx);
}

/**
 * bitmap_find_msb() - find the most significant bit in a bitmap
 * @addr: bitmap
 * @nbits: number of bits
 *
 * Returns: @nbits if all zero. [0, @nbits-1] otherwise.
 *
 */
static int bitmap_find_msb(unsigned long *addr, int nbits)
{
	int pos, last;

	last = nbits;
	pos = find_first_bit(addr, nbits);

	while (pos < nbits) {
		last = pos;
		pos = find_next_bit(addr, nbits, pos + 1);
	}
	return last;
}

/**
 * scm80211_txop_is_within_transmit_window() - checks the transmission window
 *
 * This function works for ampdu schedule. What should it return
 * if @txop is not ampdu-enabled?
 */
bool scm80211_txop_is_within_transmit_window(struct txop *txop)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(txop->txq); i++) {
		struct scm80211_txq *txq = txop->txq[i];
		unsigned long *bitmap;
		int wsize, sn[2];

		if (!txq)
			break;

		if (!scm80211_txq_is_ampdu_enabled(txq))
			continue;

		bitmap = txop->ampdu.tids[txq->txq.tid].bitmap;
		wsize = scm80211_ba_buf_size(txq);

		sn[0] = txq->sb.ssn;
		sn[1] = txop->ampdu.tids[txq->txq.tid].ssn;
		sn[1] = ieee80211_sn_add(sn[1], bitmap_find_msb(bitmap, wsize));

		if (ieee80211_sn_less(sn[1], sn[0] + wsize))
			continue;

		scm80211_trace(txop->hw,
			       "- hwq#%d.%d: sched [%d-%d] beyond transmit window [%d-%d]\n",
			       txop->hwq->id, txop->id,
			       txop->ampdu.tids[txq->txq.tid].ssn,
			       sn[1], sn[0], ieee80211_sn_add(sn[0], wsize - 1));

		return false;
	}
	return true;
}

static void scm80211_txop_update_transmit_window(struct txop *txop)
{
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	struct scm80211_txq *txq;
	unsigned long bitmap[128 / BITS_PER_LONG];
	unsigned long tidset;
	int tid, nbits, delta, shift;
	u16 ssn;

	if (!txop->ampdu.update_needed)
		return;

	if (txop->tx_window_valid)
		return;

	tidset = ampdu->tidset_ba | ampdu->tidset_eof;
	for_each_set_bit(tid, &tidset, IEEE80211_NUM_TIDS) {
		unsigned long flags;

		txq = scm80211_get_txq_from_tid(txop, tid);

		spin_lock_irqsave(&txq->lock, flags);
		tid = txq->txq.tid;
		nbits = ampdu->tids[tid].bufsize;
		ssn = txq->sb.ssn;

		scm80211_trace(txop->hw, "- tid[%d]: BA ssn=%d %016llx/%016llx (%d/%d) ",
			       tid, ampdu->tids[tid].ssn,
			       get_unaligned_le64(ampdu->tids[tid].ba),
			       get_unaligned_le64(ampdu->tids[tid].bitmap),
			       bitmap_weight(ampdu->tids[tid].ba, nbits),
			       bitmap_weight(ampdu->tids[tid].bitmap, nbits));

		/* update no need */
		if (bitmap_weight(ampdu->tids[tid].ba, nbits) == 0) {
			spin_unlock_irqrestore(&txq->lock, flags);
			continue;
		}

		if (txq->sb.ssn > IEEE80211_MAX_SN && scm80211_txq_is_ampdu_enabled(txq)) {
			struct txop *next = scm80211_get_txop(txop->hwq, TXOP_STATE_SCHEDULED);
			/*
			 * sb ssn has been reset, for this case double check ampdu is enbaled
			 * on this txq, and whether it has SCHEDULEDD txop on this txq and the
			 * ssn is lower than current one, if that is the case don't reinit the
			 * sb ssn
			 */
			if (next) {
				if (next->owner == txq && next->ampdu.enabled &&
				    ieee80211_sn_less(next->ampdu.tids[tid].ssn,
						      ampdu->tids[tid].ssn)) {
					scm80211_txop_add_back(next);
					spin_unlock_irqrestore(&txq->lock, flags);
					continue;
				}
				scm80211_txop_add_back(next);
			}
			txq->sb.ssn = ampdu->tids[tid].ssn;	/* initial condition */
		} else if (!scm80211_txq_is_ampdu_enabled(txq)) {
			spin_unlock_irqrestore(&txq->lock, flags);
			continue;
		}

		if (unlikely(ieee80211_sn_less(ampdu->tids[tid].ssn, txq->sb.ssn))) {
			unsigned long ba[128 / BITS_PER_LONG];

			shift = find_last_bit(ampdu->tids[tid].ba, nbits);
			scm80211_warn(txop->hw,
				      "WARN: tid#%d, txq ssn %d, AMPDU ssn %d, shift %d\n", tid,
				      txq->sb.ssn, ampdu->tids[tid].ssn, shift);
			if (ieee80211_sn_less(ampdu->tids[tid].ssn + shift, txq->sb.ssn)) {
				// No overlap, do nothing on the AMPDU/txq
				spin_unlock_irqrestore(&txq->lock, flags);
				continue;
			}
			bitmap_zero(ba, nbits);
			bitmap_zero(bitmap, nbits);
			delta = ieee80211_sn_sub(txq->sb.ssn, ampdu->tids[tid].ssn);
			bitmap_shift_right(ba, ampdu->tids[tid].ba, delta, nbits);
			bitmap_or(bitmap, txq->sb.bitmap, ba, nbits);
		} else {
			delta = ieee80211_sn_sub(ampdu->tids[tid].ssn, txq->sb.ssn);
			BUG_ON(delta > nbits);
			shift = find_last_bit(ampdu->tids[tid].ba, nbits);
			/**
			 * There is a corner case here on usb driver,
			 * when txop0 sent but not all acked but txop1 send and all acked
			 * txop1 will have chance to update sb bitmap before txop0
			 * which cause some issue -- sb bitmap will not be updated correctly
			 * So check all packet in transmit window here,
			 * if out of sb window do not update sb bitmap it here.
			 * it will update in scm80211_complete_txop_epilogue as txop0.
			 **/
			if (shift == nbits ||
			    (shift != nbits && unlikely(delta + shift >= nbits))) {
				spin_unlock_irqrestore(&txq->lock, flags);
				return;
			}
			bitmap_zero(bitmap, nbits);
			bitmap_shift_left(bitmap, ampdu->tids[tid].ba, delta, nbits);
			bitmap_or(bitmap, txq->sb.bitmap, bitmap, nbits);
		}
		shift = find_first_zero_bit(bitmap, nbits);
		bitmap_shift_right(txq->sb.bitmap, bitmap, shift, nbits);
		txq->sb.ssn = ieee80211_sn_add(txq->sb.ssn, shift);

		scm80211_trace(txop->hw, "- tid[%d]: tx window: [%d-%d] %s [%d-%d] (%016llx)",
			       tid, ssn, ieee80211_sn_add(ssn, nbits - 1),
			       shift > 0 ? "slides to" : "stays at",
			       txq->sb.ssn, ieee80211_sn_add(txq->sb.ssn, nbits - 1),
			       get_unaligned_le64(txq->sb.bitmap));

		spin_unlock_irqrestore(&txq->lock, flags);
	}

	txop->tx_window_valid = true;
}

/**
 * __scm80211_complete_txop_prologue() - prepares txop completion
 *
 * This function is called from hardirq.
 */
static void scm80211_complete_txop_prologue(struct txop *txop, bool discard)
{
	struct ieee80211_hw *hw = txop->hw;
	struct scm80211_hwq *hwq = txop->hwq;

	txop->actual_time = ktime_us_delta(ktime_get(), txop->tstamp);
	txop->sent = 0;
	txop->acked = 0;
	txop->nacked = 0;

	scm80211_trace(hw, "scm80211_complete_txop(hwq#%d.%d): cmd=%08x result=%08x\n",
		       hwq->id, txop->id, txop->cmd, txop->txd.result);
	scm80211_trace(hw, "(ok=%d tb=%d nack=%d fail=%d)\n",
		       txop->txd.ok, txop->txd.tb, txop->txd.nack, txop->txd.fail);

	txop->sent = txop->acked = txop->nacked = 0;
	scm80211_txop_change_state(txop, TXOP_STATE_TX_DONE);
}

/**
 * scm80211_process_successful_txop() - processes txop that are fully acknowledged
 */
static void scm80211_process_successful_txop(struct txop *txop, bool discard)
{
	unsigned long tidset;
	int tid;
	struct scm80211_hwq *hwq;
	struct txop *next;
	int n = skb_queue_len(&txop->sched);
	struct tx_mpdu_desc *start_md = &txop->dma[0].desc->mdesc[0];
	struct tx_mpdu_desc *end_md = &txop->dma[0].desc->mdesc[n - 1];

	if (txop->space == MT_TGT_BD_TGT) {
		memcpy_fromio(word(start_md, 2), word(&txop->dma[1].desc->mdesc[0], 2),
			      sizeof(__le32) * 3);
		memcpy_fromio(word(end_md, 2), word(&txop->dma[1].desc->mdesc[n - 1], 2),
			      sizeof(__le32) * 3);
	}

	if (!(txop->txd.ok && !txop->txd.nack) || !start_md->acked ||
	    (txop->txd.tb && !end_md->sent))
		return;

	if (!txop->ampdu.enabled)
		return;

	hwq = txop->hwq;

	tidset = txop->ampdu.tidset_ba | txop->ampdu.tidset_eof;
	for_each_set_bit(tid, &tidset, IEEE80211_NUM_TIDS) {
		bitmap_copy(txop->ampdu.tids[tid].ba,
			    txop->ampdu.tids[tid].bitmap, txop->ampdu.tids[tid].bufsize);
	}
	scm80211_txop_update_transmit_window(txop);

	/**
	 * Some corner case:
	 * The AMPDU sent by HE_TB, but the mu edca param was not applied,
	 * the next txop will using the normal edca parameter and trainsmit using HE_SU
	 * which cause the mu edca turning on / off unexpectedly
	 * So if AMPDU sent by HE_TB, but the mu edca param was not applied, we do not start
	 * backoff for next txop, it will to start in scm80211_complete_txop_epilogue
	 **/
	if (!discard && (!txop->txd.tb ||
			 (txop->txd.tb && txop->hwq->mu_edca_status & MU_EDCA_RUNNING))) {
		next = scm80211_get_txop(txop->hwq, TXOP_STATE_SCHEDULED);
		if (next)
			scm80211_start_backoff(next);
	}
}

/**
 * scm80211_process_incomplete_txop() - processes txop whose FES failed.
 *
 * @txop: txop whose transmission is complete
 *
 * In most cases of transmission failures, we just need to retransmit @txop again
 * only with minor changes and can skip the lengthy processing.
 *
 * @txop->state changes to either of the following
 * - TXOP_STATE_DONE: takes the normal processing step
 * - TXOP_STATE_ISSUED: retransmit @txop
 * - TXOP_STATE_NEEDS_RESCHED: let the work reschedule @txop and transmit again
 *   later before the next scheduled txop.
 */
static void scm80211_process_incomplete_txop(struct txop *txop, bool discard)
{
	bool retransmit = false;
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw_priv *hw_priv = txop->hw->priv;

	if (discard)
		return;

	if (txop->txd.fail != internal_win_lost)
		hw_priv->tx_winlost = 0;

	if (txop->txd.ok)
		return;

	switch (txop->txd.fail) {
	case ack_timeout:
	case ack_invalid:
		/*
		 * The ppdu is transmitted but failed to receive Ack/BA.
		 * Let this be processed in the same way with BA with
		 * partially filled bitmap.
		 */
		break;

	case cts_timeout:
	case cts_invalid:
		/*
		 * RTS/CTS exchange failed. The mpdus in this txop were
		 * not transmitted. Just transmit again.
		 */
		if (txop->cts_timeout_cnt++ > 2) {
			struct tx_mpdu_desc *mdesc;

			scm80211_trace(txop->hw,
				      "-  rts/cts timeout %d times continuously!!",
				      txop->cts_timeout_cnt);

			mdesc = &txop->dma[0].desc->mdesc[0];
			mdesc->protection = NO_PROTECTION;
			mdesc = &txop->dma[1].desc->mdesc[0];
			mdesc->protection = NO_PROTECTION;
			/* make sure memory operation done */
			wmb();
		}
		/* FIXME: report? */
		retransmit = true;
		break;
	case internal_collision:
		hwq->qsrc++;
		hwq->cw = ((hwq->edca.cw_min + 1) << hwq->qsrc) - 1;
		hwq->cw = min(hwq->edca.cw_max, hwq->cw);
		retransmit = true;
		break;
	case internal_win_lost:
		/* Temporary failure. Transmit again as is */
		hw_priv->tx_winlost++;
		hw_priv->total_winlost++;
		scm80211_trace(txop->hw, "-  win-lost cnt: %lld\n", hw_priv->tx_winlost);
		/* FIXME: switch to using the tx descriptors in the target memory */

		/* check retry count field in mdesc is changed or not */
		if (hw_priv->tx_winlost < SCM80211_TXOP_RETRANSMIT_MAX &&
		    (!scm80211_lookup_tx(hwq->hw, txop))) {
			scm80211_txop_change_state(txop, TXOP_STATE_TX_RETRY);
		} else if (hw_priv->space < MT_TGT_BD_TGT) {
			scm80211_info(txop->hw,
				      "-  win-lost too many times change dma space to %d\n",
				      hw_priv->space + 1);
			hw_priv->space++;
			hw_priv->tx_winlost = 0;
		} else {
			scm80211_err(txop->hw, "-  win-lost can not recover !!!\n");
		}
		break;
	case internal_tx_abort:
		/* Temporary failure because of PHY reset. Trasnmit again as is */
		retransmit = true;
		break;
	case internal_tx_rejected:
		scm80211_trace(txop->hw, "-  Tx reject by coex\n");
		retransmit = true;
		break;
	case cca_busy_secondary_20:
	case cca_busy_secondary_40:
	case cca_busy_secondary_20_40:
		/* What should we do? Change the badwidth? */
		break;
	case internal_underrun:
		/* Hoping this is temporary */
		scm80211_trace(txop->hw, "-  Tx underrun\n");
		retransmit = true;
		break;
	default:
		break;
	}

	if (retransmit) {
		if (txop->txd.fail == internal_tx_rejected) {
#ifdef SCM80211_COEX
			hw_priv->coex.coex_en = true;
			/* make sure retry txop will not discard when meet internal_tx_rejected */
			scm80211_txop_change_state(txop, TXOP_STATE_TX_RETRY);
			/*
			 * stop hwq, do not schedule txop
			 * WARNNING: stop all hwq when tx abort by coex.
			 * It would meet problem when channel switch operation (multipul hwq work).
			 */
			scm80211_stop_queues(txop->hw, scm80211_vif_queues(txop->vif));
			scm80211_enable_coex_irq(txop->hw);
			/* prevent hwq always stop */
			hrtimer_start(&hw_priv->coex.stop_queue_watchdog,
				      ms_to_ktime(SCM80211_COEX_TIMEOUT), HRTIMER_MODE_REL_SOFT);
#else
			scm80211_txop_change_state(txop, TXOP_STATE_TX_RETRY);
#endif
			txop->retransmit = 0;
		} else if (++txop->retransmit > SCM80211_TXOP_RETRANSMIT_MAX) {
			/*
			 * retransmit maximum times on this TXOP, discard it silently.
			 * The MPDU retry number may violate standard.
			 * but for the sake of avoiding stability/endless loop issue,
			 * it's necessary.
			 */
			scm80211_err(txop->hw,
				     "%s(hwq#%d.%d): no more retransmit[retransmit=%d, fail=%d]\n",
				     __func__,
				     txop->hwq->id, txop->id, txop->retransmit, txop->txd.fail);
		} else {
			scm80211_txop_change_state(txop, TXOP_STATE_TX_RETRY);
		}
	}
}

static void scm80211_tx_status(struct txop *txop, struct sk_buff *skb, bool retire)
{
	struct ieee80211_hw *hw = txop->hw;
	struct txvector *tv = &txop->tv;
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct ieee80211_tx_rate *rate = txop->rate;
	struct ieee80211_tx_status status;
	struct rate_info ri;

	/*
	 * The driver sets IEEE80211_TX_CTL_AMPDU and clears
	 * IEEE80211_TX_STAT_AMPDU of IEEE80211_SKB(skb)->flags
	 * NOT to make ieee80211_tx_status_xxx() update the
	 * rates. Instead, the driver calls ieee80211_tx_rate_update()
	 * later once for an a-mpdu.
	 */
	info->flags |= IEEE80211_TX_CTL_AMPDU;
	info->flags &= ~IEEE80211_TX_STAT_AMPDU;

	if (!(hw->conf.flags & IEEE80211_CONF_MONITOR)) {
		if (retire)
			ieee80211_tx_status_irqsafe(hw, skb);
		return;
	}

	/* All the rest is to capture skb via monitor interface */

	/* skb is retry and needs further processing - copy it */
	if (!retire)
		skb = skb_copy(skb, GFP_ATOMIC);

	/* netif_rx() does not handle frag_list */
	if (skb_has_frag_list(skb))
		skb_linearize(skb);

	memset(&status, 0, sizeof(status));
	memset(&ri, 0, sizeof(ri));

	status.sta = txop->sta;
	status.info = info;
	status.skb = skb;
#if KERNEL_VERSION(5, 19, 0) > MAC80211_BASE
	status.rate = &ri;
#else
	status.rates->rate_idx = ri;
#endif

	/* RATE_INFO_BW_HE_RU TBD */
	if (rate->flags & IEEE80211_TX_RC_40_MHZ_WIDTH)
		ri.bw = RATE_INFO_BW_40;
	else if (rate->flags & IEEE80211_TX_RC_80_MHZ_WIDTH)
		ri.bw = RATE_INFO_BW_80;
	else if (rate->flags & IEEE80211_TX_RC_160_MHZ_WIDTH)
		ri.bw = RATE_INFO_BW_160;
	else
		ri.bw = RATE_INFO_BW_20;

	if (ieee80211_rate_is_non_ht(rate)) {
		ri.legacy = hw->wiphy->bands[txop->ti.band]->bitrates[rate->idx].bitrate;
		ri.flags |= ((rate->flags & IEEE80211_TX_RC_SHORT_GI) ?
			     RATE_INFO_FLAGS_SHORT_GI : 0);
		return;
	}

	ri.mcs = ieee80211_rate_get_mcs(rate);
	ri.nss = ieee80211_rate_get_nss(rate);

	if (ieee80211_rate_is_he(rate)) {
		ri.flags |= RATE_INFO_FLAGS_HE_MCS;
		ri.he_gi = ((tv->GI_TYPE == GI_0u8s) ? NL80211_RATE_INFO_HE_GI_0_8 :
			    (tv->GI_TYPE == GI_1u6s) ? NL80211_RATE_INFO_HE_GI_1_6 :
			    NL80211_RATE_INFO_HE_GI_3_2);
		ri.he_dcm = tv->DCM;
		ri.he_ru_alloc = 0;
	} else if (ieee80211_rate_is_vht(rate)) {
		ri.flags |= RATE_INFO_FLAGS_VHT_MCS;
	} else if (ieee80211_rate_is_ht(rate)) {
		ri.flags |= RATE_INFO_FLAGS_MCS;
	}

	ieee80211_tx_status_ext(txop->hw, &status);
}

#ifdef AMPDU_SZ_SLIDE

#define LARGE_AMPDU	10	// MPDU count of AMPDU
#define TX_FAIL_LMT	2	// continue TX fail count
#define TX_SUC_LMT	200	// continue TX success count

/* ampdu_sz_slide_proc() - check the condition about NOACK for
 * last MPDU of AMPDU and limit the AMPDU size
 * This function is fixed the issue about the last mpdu of AMPDU is no-ack
 */
inline void ampdu_sz_slide_proc(struct ieee80211_hw_priv *priv,
				struct txop *txop, int i, u64 bitmap)
{
	unsigned long flag;
	u32 idx = 0;

	idx = find_next_zero_bit((void *)&bitmap, 64, 0);
	// last mpdu of ampdu didn't recevie ACK and at least it's a large size of ampdu
	if (i > LARGE_AMPDU && (idx == (i - 1) || txop->reduce)) {
		struct scm80211_tx_info *xti = &txop->xti[0];

		spin_lock_irqsave(&priv->lock, flag);

		if (priv->tx_fail_ampdu_sz == txop->ampdu.len) {
			// same error len continue appear
			priv->tx_fail_cnt++;
		} else {
			// different error len appear or the first error len appear, record it.
			priv->tx_fail_cnt = 1;
			priv->tx_fail_ampdu_sz = txop->ampdu.len;
		}

		if (priv->tx_fail_cnt > TX_FAIL_LMT) {
			// meet the limit condition, enable limit
			priv->tx_ampdu_limit = priv->tx_fail_ampdu_sz - xti->mlen;
			priv->tx_suc_cnt = 0;
			priv->tx_fail_cnt = 0;
			priv->tx_fail_ampdu_sz = 0;
		}

		spin_unlock_irqrestore(&priv->lock, flag);
		scm80211_trace(txop->hw, "%s, AMPDU=%zu, MPDU=%zu, limit=%d, latest=%d cnt=%d\n",
			       txop->reduce ? "USB" : "PCIe", txop->ampdu.len, xti->mlen,
			       priv->tx_ampdu_limit, priv->tx_fail_ampdu_sz, priv->tx_fail_cnt);
	} else if (priv->tx_ampdu_limit) {
		spin_lock_irqsave(&priv->lock, flag);
		if (priv->tx_suc_cnt++ > TX_SUC_LMT) {
			// meet the recover condition, disable limit
			priv->tx_ampdu_limit = 0;
			priv->tx_suc_cnt = 0;
			priv->tx_fail_ampdu_sz = 0;
			priv->tx_fail_cnt = 0;
		}
		spin_unlock_irqrestore(&priv->lock, flag);
		scm80211_trace(txop->hw, "tx suc cnt=%d,  ampdu slide len limit=%d\n",
			       priv->tx_suc_cnt, priv->tx_ampdu_limit);
	}
}

#undef LARGE_AMPDU
#undef TX_FAIL_LMT
#undef TX_SUC_LMT

/* check_tx_ampdu_limit() - check current TX AMPDU size is over tx_ampdu_limit
 *
 * the current TX AMPDU size is limit with tx_ampdu_limit if tx_ampdu_limit is valid
 */
inline int check_tx_ampdu_limit(struct txop *txop, size_t totlen)
{
	struct ieee80211_hw_priv *hw_priv = txop->hw->priv;

	if (hw_priv->tx_ampdu_limit && totlen >= hw_priv->tx_ampdu_limit) {
		scm80211_dbg(txop->hw, "ampdu length=%zu, ampdu slide len limit=%u\n",
			     totlen, hw_priv->tx_ampdu_limit);
		return -EAGAIN;
	}

	return 0;
}

#else

inline void ampdu_sz_slide_proc(struct ieee80211_hw_priv *priv,
				struct txop *txop, int i, u64 bitmap)
{
}

inline int check_tx_ampdu_limit(struct txop *txop, size_t totlen)
{
	return 0;
}

#endif /* AMPDU_SZ_SLIDE */

/**
 * scm80211_process_done_txop() - process a txop transmission
 *
 * This function is called either to complete the issued txop on tx done
 * interrupt request from hardware, or to tidy up things done during txop
 * scheduling in case of scheduling failure.
 *
 */
static void scm80211_process_done_txop(struct txop *txop, bool discard)
{
	struct ieee80211_hw *hw = txop->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_vif *vif = txop->vif;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	struct ieee80211_sta *sta = txop->sta;
	struct sk_buff *skb;

	bool TB = false, mu_edca = false;
	int i = 0;
	int tid, bit;
	bool unsent_retry = false;
	u64 bitmap = 0;

	if (txop->state != TXOP_STATE_TX_DONE)
		return;

	trace_scm80211_complete_txop(hw, hwq, txop->txq[0], discard, txop->cmd, txop->txd.result);

	if ((hwq->mu_edca_status & MU_EDCA_CONFIGURED) && hwq->edca.mu_edca_param_rec.mu_edca_timer)
		TB = txop->txd.tb;

	scm80211_print_txvector(txop->buf, sizeof(txop->buf), hw, txop->sta, &txop->tv);

	while ((skb = skb_dequeue(&txop->sched))) {
		struct tx_mpdu_desc *mdesc = txop_get_mdesc(txop, i);
		struct scm80211_tx_info *xti = &txop->xti[i];
		struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
		struct scm80211_txq *txq;
		struct ieee80211_hdr *hdr;
		bool ok = !discard, drop = discard;

		/* Unmap tx buffers */
		scm80211_cleanup_tx(hw, txop, i);

		if (txop->space == MT_TGT_BD_TGT) {
			/* Copy back the second word (MT2) into the host descriptor */
			memcpy_fromio(word(mdesc, 2),
				      word(&txop->dma[1].desc->mdesc[i], 2), sizeof(__le32));
		}

		ok &= (mdesc->sent && mdesc->acked) && !discard;

		scm80211_update_tid_queue_size(txop, skb, false);
		txq = scm80211_get_txq_from_tid(txop, mdesc->tid);

		skb_pull(skb, sizeof(struct txvector));
		hdr = (void *)skb->data;

		drop = (!ok && scm80211_discard_frame(hw, skb)) ? true : drop;

		txop->sent += (mdesc->sent) ? 1 : 0;

		/**
		 * under concurrent mode, and multi queue test, the sn sometimes will be changed
		 * Workaround:
		 * Driver record the sn num, and do the double check
		 **/
		if (unlikely(mdesc->sn != xti->sn)) {
			scm80211_info(hw, " SN-mismatch with orginal one\n");
			mdesc->sn = xti->sn;
		}

		if (!txop->ampdu.enabled) {
			scm80211_trace(hw, "- tid[%d] ssn=%d sent=%d acked=%d\n",
				       mdesc->tid, mdesc->sn, mdesc->sent, mdesc->acked);

			trace_scm80211_tx_result_sifs_burst(hw, hwq, txq, mdesc->sn,
							    mdesc->sent, mdesc->acked);
		}

		if (txop->ampdu.update_needed && (ok || drop)) {
			tid = mdesc->tid;
			WARN_ON(tid == 15);

			bit = ieee80211_sn_sub(mdesc->sn, txop->ampdu.tids[tid].ssn);
			set_bit(bit, txop->ampdu.tids[tid].ba);
		}

		if (ok) {
			txop->acked++;
			if (mdesc->protection)
				priv->stats.dot11RTSSuccessCount++;

			if (ieee80211_is_beacon(hdr->frame_control)) {
				if (scm80211_csa_is_finished(vif))
					ieee80211_csa_finish(vif);
			} else if (unlikely(ieee80211_has_order(hdr->frame_control))) {
				int hdrlen = ieee80211_hdrlen(hdr->frame_control);
				u32 htc_tx = get_unaligned_le32((u8 *)hdr + hdrlen - 4);
				/* Apply new OM after tx complete */
				if (IS_OMI(htc_tx))
					scm80211_process_omi(hw, vif, sta, htc_tx, false);
			} else if (ieee80211_is_bar(hdr->frame_control))
				scm80211_tx_bar_handle(hw, vif, skb);

			mu_edca |= TB && ieee80211_is_data_qos(hdr->frame_control);

			scm80211_handle_tdls(hwq, skb);

#ifdef SCM80211_TWT
			if (ieee80211_is_action(hdr->frame_control)) {
				struct unprot_sig_action *action;

				action = (void *)(skb->data + sizeof(struct ieee80211_hdr_3addr));
				if (action->category == WLAN_CATEGORY_UNPROT_SIG &&
				    action->action_code == WLAN_SIG_TWT_TEARDOWN &&
				    vifp && vifp->twt_teardown) {
					vifp->twt_teardown = false;
					scm80211_set_twt_state(vif, TWT_DISABLED);
				}
			}
#endif
			if (ieee80211_has_pm(hdr->frame_control) &&
			    !scm80211_is_scanning(hw) &&
			    !scm80211_is_offchannel(hw, vif) &&
			    !mdesc->noack) {
				unsigned long flags;

				spin_lock_irqsave(&priv->pm.lock, flags);
				priv->pm.acked = true;
				spin_unlock_irqrestore(&priv->pm.lock, flags);
				mod_timer(&priv->pm.timer,
					  jiffies + msecs_to_jiffies(priv->pm.dynamic_ps_timeout));
			}

			info->flags |= mdesc->noack ? IEEE80211_TX_STAT_NOACK_TRANSMITTED :
			    IEEE80211_TX_STAT_ACK;

			scm80211_tx_status(txop, skb, true);
			set_bit(i, (void *)&bitmap);
		} else if (drop) {
			scm80211_warn(hw, "- sn %d dropped\n", mdesc->sn);

			if (ieee80211_is_bar(hdr->frame_control))
				scm80211_tx_bar_handle(hw, vif, skb);

			scm80211_tx_status(txop, skb, true);
		} else {
			if (mdesc->sent) {
				scm80211_tx_status(txop, skb, false);

				hdr->frame_control |= cpu_to_le16(IEEE80211_FCTL_RETRY);
				if (!txop->txd.tb) {
					struct ieee80211_tx_rate *rate;

					rate = &info->status.rates[SCM80211_TX_RETRY_RATE];
					rate->count += txop->retry_cnt;
				}
			} else {
				scm80211_trace(hw, "- sn %d not sent (discard=%d)\n",
					       mdesc->sn, discard);

				if (!unsent_retry && !txop->txd.tb && txop->scheduled) {
					info->control.rates[0].count++;
					unsent_retry = true;
				}
			}

			/* To skip mdesc construction for fast retransmission */
			txop->xti[txop->nacked] = *xti;
			memcpy(word(&txop->dma[0].desc->mdesc[txop->nacked++], 2), word(mdesc, 2),
			       sizeof(__le32) * 3);

			if (mdesc->protection)
				priv->stats.dot11RTSFailureCount++;
			if (!mdesc->noack)
				priv->stats.dot11ACKFailureCount++;

			/*
			 * Clear IEEE80211_TX_CTL_RATE_PROBE flag. This allows the
			 * next txop schedule to include other frames as well.
			 */
			if (i == 0 && info->flags & IEEE80211_TX_CTL_RATE_CTRL_PROBE) {
				/* invalid probe rate */
				info->control.rates[0].idx = -1;
			}
			info->flags &= ~IEEE80211_TX_CTL_RATE_CTRL_PROBE;
			skb_queue_tail(&txq->retry, skb);
		}
		i++;
	}

	// check the condition about NOACK for last MPDU of AMPDU
	ampdu_sz_slide_proc(priv, txop, i, bitmap);

	if (txop->ampdu.enabled && (txop->nacked == txop->n_mpdu) && txop->txd.ok) {
		u16 sn = ieee80211_sn_add(txop->owner->sb.ssn, 2048);

		hwq->bar_ssn = ieee80211_sn_sub(txop->owner->sb.ssn, scm80211_ba_buf_size(txop->owner));
		hwq->on_bar = true;

		ieee80211_send_bar(vif, txop->owner->txq.sta->addr, txop->owner->txq.tid, IEEE80211_SN_TO_SEQ(sn));
		ieee80211_send_bar(vif,  txop->owner->txq.sta->addr, txop->owner->txq.tid, IEEE80211_SN_TO_SEQ(hwq->bar_ssn));

		scm80211_stop_queues(hw, BIT(hwq->id));
	}

	/* For debugging */
	if (txop->txd.ok && !txop->txd.nack && txop->nacked != 0) {
		int i;

		scm80211_warn(hw, "- txd mismatches: %d fails\n", txop->nacked);

		for (i = 0; i < txop->n_mpdu; i++) {
			struct tx_mpdu_desc *mdesc = txop_get_mdesc(txop, i);

			scm80211_warn(hw, "mdesc[%d]: sent=%d acked=%d\n",
				      i, mdesc->sent, mdesc->acked);
		}

		/* Let's try to believe txdesc content */
		txop->nacked = 0;
	}

	if (txop->acked)
		vifp->tstamp = ktime_get();

	/* If HE TB is disabled, do not enable MU EDCA. */
	if (!(vifp->reg_mmss & VIF_MIN_MPDU_SPACING_TRIG_EN) ||
	    (vifp->reg_mmss & VIF_MIN_MPDU_SPACING_HETB_DATA_DIS)) {
		mu_edca = false;
	}

	if (mu_edca || ((hwq->mu_edca_status & MU_EDCA_RUNNING) && !txop->scheduled)) {
		scm80211_enable_mu_edca(hwq);
	} else if (!TB || !(hwq->mu_edca_status & MU_EDCA_RUNNING)) {
		/**
		 * Don't apply the normal edca parameter when
		 * all the mpdus in the ampdu sent by HE_TB but noack received
		 * when mu edca timer is not expired.
		 **/
		scm80211_disable_mu_edca(hwq);
		if (hwq->edca.aifs != 0)
			scm80211_adjust_cw(hwq, txop->txd.ok, txop->txd.fail, txop->acked);
	}

	scm80211_complete_txq(txop, txop->txq[0]);
	scm80211_complete_txq(txop, txop->txq[1]);

	trace_scm80211_tx_return(hwq, __func__);
}

static void scm80211_complete_txop_epilogue(struct txop *txop, bool discard)
{
	struct ieee80211_hw *hw = txop->hw;
	struct txop *next;
	int slot, aifs, backoff, txtime;

	if (txop->state != TXOP_STATE_TX_DONE)
		goto out;
	txop->rs.status.ampdu_len = txop->txd.tb ? txop->sent : txop->n_mpdu;
	txop->rs.status.ampdu_ack_len = txop->acked;

	scm80211_txop_update_transmit_window(txop);

	scm80211_trace(hw,
		       "- status: rate[%d]: idx=%d flags=%08x retries=%d ampdu ack/len=%d/%d\n",
		       scm80211_tx_rate_index(txop),
		       txop->rs.status.rates[0].idx,
		       txop->rs.status.rates[0].flags,
		       txop->retries, txop->rs.status.ampdu_ack_len, txop->rs.status.ampdu_len);

	scm80211_trace(hw, "- ampdu=%d prot=%d probe=%d\n",
		       txop->ampdu.enabled, txop->protection, txop->probe);

	/* Time report */
	slot = 9;
	if (txop->vif && !txop->vif->bss_conf.use_short_slot)
		slot = 20;

	aifs = FIELD_GET(TX_QUEN_CMD_AIFSN, txop->cmd) * slot + ieee80211_sifs(txop->ti.band);
	backoff = FIELD_GET(TX_QUEN_CMD_BOFF_COUNT, txop->cmd) * slot;
	txtime = aifs + backoff + txop->txtime + txop->ack_time;
	txtime += txop->protection ? txop->rts_time : 0;

	scm80211_trace(hw,
		       "- txtime=%dus expected=%dus (backoff:%d rts/cts:%d ppdu:%d ack:%d) defer=%dus\n",
		       txop->actual_time, txtime,
		       aifs + backoff,
		       txop->protection ? txop->rts_time : 0,
		       txop->txtime, txop->ack_time, txop->actual_time - txtime);
	WARN_ON(txop->actual_time < txtime / 100);
	scm80211_tx_rate_update(txop);

	if (!txop->nacked) {
		struct scm80211_txq *txq = txop->owner;
		/* Successful transmission. We are done */
		scm80211_txop_free(txop);
		/* check whether driver is ready to do aggregation */
		if (txq->ba.action == IEEE80211_AMPDU_TX_START &&
		    skb_queue_len(&txq->queue) + skb_queue_len(&txq->retry) == 0 &&
		    !scm80211_txq_is_scheduled(txq) && ieee80211_is_ht_sta(txq->txq.sta)) {
			ieee80211_start_tx_ba_cb_irqsafe(txq->txq.vif, txq->txq.sta->addr,
							 txq->ba.tid);
		}
		if (!discard) {
			next = scm80211_get_txop(txop->hwq, TXOP_STATE_SCHEDULED);
			if (next)
				scm80211_start_backoff(next);
		}
	} else {
		/*
		 * Full or partial failure.
		 *
		 * Try to transmit the next available schedule. If there is none
		 * or it is not eligible for transmission, the driver let
		 * the hardware queue work deal with it with a higher priority.
		 */

		next = scm80211_get_txop(txop->hwq, TXOP_STATE_SCHEDULED);
		if (next && scm80211_start_backoff(next)) {
			scm80211_txop_free(txop);
		} else {
			scm80211_txop_change_state(txop, TXOP_STATE_NEEDS_RESCHED);
			scm80211_txop_add(txop, TXOP_STATE_NEEDS_RESCHED);
		}
	}
out:
	scm80211_schedule_hwq_work(txop->hwq);
}

void scm80211_complete_txop(struct txop *txop, bool discard)
{
	struct ieee80211_hw *hw = txop->hw;
	struct ieee80211_hw_priv *hw_priv = hw->priv;

	if (!txop) {
		scm80211_warn(hw, "%s: no txop?\n", __func__);
		return;
	} else if (txop->state != TXOP_STATE_ISSUED) {
		scm80211_warn(hw, "%s(hwq#%d.%d): invalid state %d\n",
			      __func__, txop->hwq->id, txop->id, txop->state);
		return;
	}

	/* the same action as in scm80211_handle_tx_irq except add txop to tx done list */
	scm80211_complete_txop_prologue(txop, discard);
	scm80211_process_successful_txop(txop, discard);
	scm80211_process_incomplete_txop(txop, discard);

#ifdef SCM80211_COEX
	if (hw_priv->coex.coex_en) {
		if (txop->state == TXOP_STATE_TX_RETRY) {
			scm80211_txop_add(txop, TXOP_STATE_TX_RETRY);
			return;
		}
	}
#endif

	if (txop->state == TXOP_STATE_TX_RETRY) {
		if (!scm80211_start_backoff(txop))
			scm80211_schedule_hwq_work(txop->hwq);
		return;
	}

	scm80211_process_done_txop(txop, discard);
	scm80211_complete_txop_epilogue(txop, discard);
}

static enum hrtimer_restart scm80211_enable_cca_detect(struct hrtimer *timer)
{
	struct ieee80211_hw_priv *priv = container_of(timer, struct ieee80211_hw_priv, cca_timer);
	struct ieee80211_hw *hw = priv->hw;
	u32 val;

	val = scm80211_mac_readl(hw, MAC_REG_DEV_CS);
	val &= ~FIELD_PREP(DEV_CS_IGNORE_CCA, 1);
	scm80211_mac_writel(hw, val, MAC_REG_DEV_CS);
	scm80211_trace(hw, "%s REG_DEV_CS 0x%08x\n", __func__, val);

	return HRTIMER_NORESTART;
}

static enum hrtimer_restart scm80211_hwq_watchdog_expire(struct hrtimer *timer)
{
	struct scm80211_hwq *hwq = container_of(timer, struct scm80211_hwq, watchdog);
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct txop *txop;
	unsigned long flags;
	u32 cmd, result, state, en, intr;

	scm80211_warn(hw, "hwq#%d watchdog expired\n", hwq->id);

	cmd = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_CMD(hwq->id));
	state = scm80211_mac_readl(hw, MAC_REG_TX_QUEN_STATE(hwq->id));
	result = readl(&hwq->txdesc->result);
	intr = scm80211_mac_readl(hwq->hw, MAC_REG_INTR_STATUS);
	en = priv->int_en;

	scm80211_warn(hw,
		      "- cmd=%08x result=%08x state=%d enable=%08x intr=%08x\n",
		      cmd, result, state, en, intr);

	spin_lock_irqsave(&hwq->ring.lock, flags);
	txop = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
	if (!txop) {
		scm80211_warn(hw, "- no txop?\n");
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		return HRTIMER_NORESTART;
	}

	scm80211_info(hw, "- issued txop: hwq#%d.%d\n", hwq->id, txop->id);

	if ((FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd) == 0) && !(en & BIT(hwq->id))) {
		scm80211_warn(hw, "- tx irq turned off, completing txop manually\n");
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		scm80211_complete_txop(txop, false);

	} else if (intr & BIT(hwq->id)) {
		scm80211_warn(hw, "- start handling pending irq %d\n", priv->irq);
		__scm80211_txop_add_back(txop);
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		disable_irq(priv->irq);
		scm80211_handle_irq(priv->irq, hw);
		enable_irq(priv->irq);
	} else if (state == HWQ_STATE_IN_TRANSMISSION) {
		if (hwq->qstate == HWQ_STATE_IN_TRANSMISSION) {
			scm80211_err(hw, "- hardware queue stuck\n");
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			scm80211_cancel_tx(hwq->hw, txop, false);

#ifdef ERRATA_HWQ_STUCK
			scm80211_mac_writel(hwq->hw, BIT(hwq->id), MAC_REG_DEV_FORCE_QUE_CLR);
#endif
		} else {
			__scm80211_txop_add_back(txop);
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			hwq->qstate = state;
			hrtimer_start(timer, ms_to_ktime(5), HRTIMER_MODE_REL_SOFT);
		}
	} else if (FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd) != 0) {
		if (hwq->expire_cnt++ < 5) {
			__scm80211_txop_add_back(txop);
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			if (state == HWQ_STATE_DELAY && hwq->qstate == state &&
			    hwq->expire_cnt == 3) {
				u32 val;
				/* ignore CCA workaround */
				scm80211_warn(hw,
					      "- cca stay busy %d times, ignore cca for a while\n",
					      hwq->expire_cnt);
				val = scm80211_mac_readl(hw, MAC_REG_DEV_CS);
				val |= FIELD_PREP(DEV_CS_IGNORE_CCA, 1);
				scm80211_mac_writel(hw, val, MAC_REG_DEV_CS);
				hrtimer_start(&priv->cca_timer,
					      ms_to_ktime(SCM80211_CCA_TIMEOUT),
					      HRTIMER_MODE_REL_SOFT);
			}
			hwq->qstate = state;
			hrtimer_start(timer, ms_to_ktime(SCM80211_TX_TIMEOUT * hwq->expire_cnt),
				      HRTIMER_MODE_REL_SOFT);
		} else {
#ifdef SCM80211_SMPS
			if (state == HWQ_STATE_DELAY) {
				u32 v = scm80211_mac_readl(hw, MAC_REG_DEV_PMU);

				if ((v & (DEV_PMU_SMPS_EN | DEV_PMU_SMPS_MODE)) == 0x3) {
					v &= ~(DEV_PMU_SMPS_EN | DEV_PMU_SMPS_MODE);
					scm80211_mac_writel(hw, v, MAC_REG_DEV_PMU);
					udelay(10);
					v |= FIELD_PREP(DEV_PMU_SMPS_EN, 1);
					v |= FIELD_PREP(DEV_PMU_SMPS_MODE, 1);
					scm80211_mac_writel(hw, v, MAC_REG_DEV_PMU);
				}
			}
#endif
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
			scm80211_warn(hw,
				      "- trying to cancel txop in %d state expires for %d times\n",
				      state, hwq->expire_cnt);
			scm80211_cancel_tx(hwq->hw, txop, false);
		}

	} else {
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
		scm80211_warn(hw,
			      "- trying to cancel txop in %d state expires for %d times\n",
			      state, hwq->expire_cnt);
		scm80211_cancel_tx(hwq->hw, txop, false);
	}
	return HRTIMER_NORESTART;
}

void scm80211_hwq_start_watchdog(struct scm80211_hwq *hwq)
{
	if (!dev_is_pci(wiphy_dev(hwq->hw->wiphy)))
		return;

	hwq->qstate = HWQ_STATE_IDLE;
	hwq->expire_cnt = 0;
	hwq->watchdog.function = scm80211_hwq_watchdog_expire;
	hrtimer_start(&hwq->watchdog, ms_to_ktime(SCM80211_TX_TIMEOUT), HRTIMER_MODE_REL_SOFT);
}

/**
 * scm80211_fill_rate() fill the fixed or recommended rate
 **/
static void scm80211_fill_rate(struct ieee80211_hw *hw, struct txop *txop,
			       struct ieee80211_sta *sta)
{
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_dbg_rate_fix *dbg_rate = &sta_priv->rate_fix;
	struct ieee80211_tx_rate *rate = &txop->ti.control.rates[0];
	struct ieee80211_tx_rate *rc_rate = &txop->rc.control.rates[0];

	if (sta_priv->rate_fix_on) {
		if (dbg_rate->flags & IEEE80211_TX_CTL_LDPC)
			txop->ti.flags |= IEEE80211_TX_CTL_LDPC;
		else
			txop->ti.flags &= ~IEEE80211_TX_CTL_LDPC;

		if (rate->flags & IEEE80211_TX_RC_HE_MCS) {
			txop->he_gi_type = dbg_rate->he_gi_type;
			txop->he_ltf_type = dbg_rate->he_ltf_type;
		}
		txop->he_dcm = false;
	} else if ((rate->flags & IEEE80211_TX_RC_VHT_MCS) && sta_priv->stbc_or_dcm) {
		txop->ti.flags |= IEEE80211_TX_CTL_STBC;
	}

	/* Downgrade the bw setting sometimes AP changed the bw setting */
	if (!ieee80211_rate_is_non_ht(rate) &&
	    ((int)txvector_ch_bandwidth(rate) > (int)get_ieee80211_rx_bandwidth(sta))) {
		u16 flags;

		flags = IEEE80211_TX_RC_160_MHZ_WIDTH | IEEE80211_TX_RC_80_MHZ_WIDTH |
		    IEEE80211_TX_RC_40_MHZ_WIDTH;

		rate->flags &= ~flags;
		rc_rate->flags &= ~flags;
		switch (get_ieee80211_rx_bandwidth(sta)) {
		case CBW80:
			flags = IEEE80211_TX_RC_80_MHZ_WIDTH;
			break;
		case CBW40:
			flags = IEEE80211_TX_RC_40_MHZ_WIDTH;
			break;
		default:
			break;
		}
		rate->flags |= flags;
		rc_rate->flags |= flags;
	}
	if (ieee80211_rate_is_ht(rate))
		txop->ti.flags |= IEEE80211_TX_CTL_AMPDU;

	scm80211_fixup_tx_info_flags(hw, sta, &txop->ti, rate);
}

static bool sta_uses_dbg_fix_rate(struct txop *txop, struct ieee80211_sta *sta)
{
	struct ieee80211_tx_info *info = &txop->ti;
	struct ieee80211_tx_rate *rate = &info->control.rates[0];
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct ieee80211_dbg_rate_fix *dbg_rate = &sta_priv->rate_fix;
	unsigned long flag;

	spin_lock_irqsave(&sta_priv->lock, flag);

	if (!sta_priv->rate_fix_on) {
		spin_unlock_irqrestore(&sta_priv->lock, flag);
		return false;
	}

	*rate = dbg_rate->rate;
	scm80211_fill_rate(txop->hwq->hw, txop, sta);
	txop->rate = rate;
	info->band = dbg_rate->band;

	spin_unlock_irqrestore(&sta_priv->lock, flag);
	return true;
}

static bool sta_uses_recommended_rate(struct txop *txop, struct ieee80211_sta *sta)
{
	struct ieee80211_tx_info *info = &txop->ti;
	struct ieee80211_tx_rate *rate = &info->control.rates[0];
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	unsigned long flag;

	spin_lock_irqsave(&sta_priv->lock, flag);

	if (!sta_priv->vht_la && !sta_priv->he_la) {
		spin_unlock_irqrestore(&sta_priv->lock, flag);
		return false;
	}

	if (!sta_priv->rate_available) {
		spin_unlock_irqrestore(&sta_priv->lock, flag);
		return false;
	}

	*rate = sta_priv->recomd_rate;
	scm80211_fill_rate(txop->hwq->hw, txop, sta);
	txop->rate = rate;

	spin_unlock_irqrestore(&sta_priv->lock, flag);
	return true;
}

/**
 * Copy from rate_control_fill_sta_table in net/mac80211/rate.c
 * to Ignore rate check in mac80211/rate.c
 * HE rate will be modified by rate.c when bitrate mask set
 */
static void scm80211_get_tx_rates(struct ieee80211_vif *vif, struct ieee80211_sta *sta,
				  struct sk_buff *skb, struct ieee80211_tx_rate *dest,
				  int max_rates)
{
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(vif);
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_sta_rates *ratetbl = NULL;
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	int i, j = 0;

	if (!priv->local_rc_used) {
		ieee80211_get_tx_rates(vif, sta, skb, dest, max_rates);
		return;
	}

	/* low rate used */
	if (!sta ||
	    info->flags & (IEEE80211_TX_CTL_NO_ACK | IEEE80211_TX_CTL_USE_MINRATE) ||
	    !ieee80211_is_data(hdr->frame_control)) {
		memset(dest, 0, sizeof(*dest) * max_rates);
		if (info->band == NL80211_BAND_2GHZ && is_p2p_mgmt_frame(skb))
			dest->idx = 4;	/* 11g mode */
		dest->count = 1;
		for (i = 1; i < max_rates; i++)
			dest[i].idx = -1;
		return;
	}

	max_rates = min_t(int, max_rates, IEEE80211_TX_RATE_TABLE_SIZE);
	rcu_read_lock();
	ratetbl = rcu_dereference(sta->rates);
	rcu_read_unlock();
	for (i = 0; i < max_rates; i++) {
		if (i < ARRAY_SIZE(info->control.rates) &&
		    info->control.rates[i].idx >= 0 && info->control.rates[i].count) {
			if (dest != info->control.rates)
				dest[i] = info->control.rates[i];
		} else if (ratetbl) {
			dest[i].idx = ratetbl->rate[j].idx;
			dest[i].flags = ratetbl->rate[j].flags;
			if (info->control.use_rts)
				dest[i].count = ratetbl->rate[j].count_rts;
			if (info->control.use_cts_prot)
				dest[i].count = ratetbl->rate[j].count_cts;
			else
				dest[i].count = ratetbl->rate[j].count;
			j++;
		} else {
			dest[i].idx = -1;
			dest[i].count = 0;
		}

		if (info->control.use_rts) {
			dest[i].flags |= IEEE80211_TX_RC_USE_RTS_CTS;
			info->control.use_cts_prot = false;
		}

		if (dest[i].idx < 0 || !dest[i].count)
			break;
	}
}

/**
 * scm80211_txop_get_rate() - selects a tx rate
 * @txop: txop that is being scheduled
 *
 */
static struct ieee80211_tx_rate *scm80211_txop_get_rate(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_txq *txq = txop->owner;
	struct ieee80211_vif *vif = txq->txq.vif;
	struct ieee80211_sta *sta = txq->txq.sta;
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(txop->skb);
	int n, trial;
#ifdef SCM80211_COEX
	struct ieee80211_hw_priv *hw_priv = txop->hw->priv;
#endif

	txop->retries = scm80211_get_tx_retries(info);
	trial = txop->retries + 1;

	/*
	 * info->control.rates[]:
	 * - mac80211/rc80211 fills this table for management frames
	 *   such as Probe Req., Auth, and etc.
	 * - It also fills info->control.rates[0] for rate probing.
	 * - For other frames, the table entries are empty.
	 */

	/* FIXME: add more robus rates at the end */

	/* We need rc.flags and rc.control.skip_table */
	memcpy(&txop->rc, info, sizeof(txop->rc));

	if (!info->control.skip_table) {
		scm80211_get_tx_rates(vif, sta, txop->skb, txop->rc.control.rates,
				      IEEE80211_TX_MAX_RATES);
	}

#ifdef SCM80211_COEX
	if (hw_priv->coex.reduce_streams_flag) {
		scm80211_trace(txop->hw, "configure rc rate to 1ss\n");
		for (n = 0; n < IEEE80211_TX_MAX_RATES; n++) {
			int mcs = ieee80211_rate_get_mcs(&txop->rc.control.rates[n]);

			ieee80211_rate_set(&txop->rc.control.rates[n], mcs, 1);
		}
	}
#endif

	/* Copy to the working info */
	memcpy(&txop->ti, &txop->rc, sizeof(txop->ti));
	for (n = 0; n < IEEE80211_TX_MAX_RATES; n++)
		txop->ti.control.rates[n].count = 0;

	/* Status */
	memcpy(&txop->rs, &txop->rc, sizeof(txop->rc));
	txop->rs.flags |= IEEE80211_TX_CTL_AMPDU;
	txop->rs.flags |= IEEE80211_TX_STAT_AMPDU;

	/* FIXME: add more robus rates at the end */

	scm80211_trace(txop->hw, "- tx_info: skip_table=%d flags=%08x count=%d\n",
		       info->control.skip_table, info->flags, trial);

	for (n = 0; n < IEEE80211_TX_MAX_RATES; n++) {
		scm80211_trace(txop->hw,
			       " - rate[%d]: idx=%d, flags=%08x, count=%d\n",
			       n,
			       txop->rc.control.rates[n].idx,
			       txop->rc.control.rates[n].flags, txop->rc.control.rates[n].count);
	}

	for (n = 0; n < IEEE80211_TX_MAX_RATES; n++) {
		if (txop->rc.control.rates[n].idx == -1)
			break;
		if (trial <= txop->rc.control.rates[n].count)
			break;
		trial -= txop->rc.control.rates[n].count;
	}

	if (n >= IEEE80211_TX_MAX_RATES)
		n = 0;

	txop->rate = &txop->ti.control.rates[n];
	if (n >= IEEE80211_TX_MAX_RATES || txop->rate->idx == -1) {
		scm80211_trace(hwq->hw, "- no tx rate\n");
		txop->rate = &txop->ti.control.rates[0];
		n = 0;
	}

	if (txop->rate) {
		txop->rate->count++;

		txop->rs.status.rates[0] = txop->rs.control.rates[n];
		txop->rs.status.rates[0].count = 1;
		txop->rs.status.rates[1].count = 0;
		if (ieee80211_rate_is_non_ht(txop->rate) && txop->limit == 0) {
			txop->rs.flags &= ~IEEE80211_TX_CTL_AMPDU;
			txop->rs.flags &= ~IEEE80211_TX_STAT_AMPDU;
			txop->rs.status.rates[1].idx = -1;
		}

		/* setup second rate */
		memset(&txop->subrate, 0, sizeof(txop->subrate));
		txop->subrate.idx = -1;
		txop->subused = false;
		if (!is_pcie_device(hwq->hw) && (n < IEEE80211_TX_MAX_RATES - 1) &&
		    !(info->flags & IEEE80211_TX_CTL_RATE_CTRL_PROBE)) {
			struct ieee80211_tx_rate *rate = &txop->ti.control.rates[n + 1];
			/*
			 * check the second rate use the same flag(BW, GI) NSS as primary rate
			 * or choose the lower rate(rate with lower constellation)
			 */
			if (rate->flags == txop->rate->flags && rate->idx >= 0 &&
			    ieee80211_rate_get_nss(rate) == ieee80211_rate_get_nss(txop->rate)) {
				memcpy(&txop->subrate, rate, sizeof(txop->subrate));
				txop->subused = true;
				txop->subrate.count = 0;
			} else {
				int mcs = ieee80211_rate_get_mcs(txop->rate);
				int nss = ieee80211_rate_get_nss(txop->rate);
				int x = he_constellation_index(mcs);
				int iter = 0;

				memcpy(&txop->subrate, txop->rate, sizeof(txop->subrate));
				while (mcs - iter > 0) {
					iter++;
					if (x != he_constellation_index(mcs - iter))
						break;
				}

				if (iter > 0) {
					ieee80211_rate_set(&txop->subrate, mcs - iter, nss);
					txop->subused = true;
					txop->subrate.count = 0;
				}
			}
			scm80211_trace(hwq->hw,
				       "- subrate idx=%d flags=0x%08x\n",
				       txop->subrate.idx, txop->subrate.flags);
		}
		scm80211_trace(hwq->hw, "- rate[%d]: idx=%d flags=%08x count=%d/%d/%d\n",
			       n,
			       txop->rate->idx,
			       txop->rate->flags,
			       trial, txop->rc.control.rates[n].count, txop->retries + 1);
	} else {
		scm80211_info(hwq->hw, "- no tx rate\n");
	}
	return txop->rate;
}

/**
 * scm80211_setup_txvector() - determine txvector parameters
 *
 * This function picks up ppdu parameters from the first frame to be transmitted
 * in the txop, and build txvector accordingly. These ppdu parameters and
 * txvector are applied to all the psdu's scheduled for transmission within
 * this txop.
 */
static int scm80211_setup_txvector(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_txq *txq = txop->owner;
	struct ieee80211_vif *vif = txq->txq.vif;
	struct ieee80211_sta *sta = txq->txq.sta;
	struct ieee80211_hdr *hdr;
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(txop->skb);
	struct ieee80211_tx_rate *rate;
	char str[128];

	BUG_ON(!txop->skb);

	hdr = (struct ieee80211_hdr *)txop->skb->data;
	scm80211_snprint_frame(str, sizeof(str), hdr);
	scm80211_trace(hwq->hw, "- tx: %s\n", str);

	/* Sanity checker for control.vif */
	if (vif != info->control.vif) {
		WARN(1, "- %p != %p\n", vif, info->control.vif);
		info->control.vif = vif;
	}

	hwq->use_recmd_rate = false;

	if (sta && sta_uses_dbg_fix_rate(txop, sta)) {
		scm80211_trace(hwq->hw, "- fixed rate tx: %d flags %04x\n",
			       txop->rate->idx, txop->rate->flags);

		info->flags &= ~IEEE80211_TX_CTL_RATE_CTRL_PROBE;

	} else if (sta && sta_uses_recommended_rate(txop, sta)) {
		hwq->use_recmd_rate = true;
		scm80211_trace(hwq->hw, "- recommended rate idx: %d, flags %04x\n",
			       txop->rate->idx, txop->rate->flags);

		info->flags &= ~IEEE80211_TX_CTL_RATE_CTRL_PROBE;
	} else {
		if (info->flags & IEEE80211_TX_CTL_INJECTED)
			scm80211_trace(hwq->hw, "- injected\n");
		rate = scm80211_txop_get_rate(txop);
		if (!rate) {
			wiphy_warn(hwq->hw->wiphy, "- no more tx rates\n");
			return -ENOSPC;
		}
		scm80211_fixup_tx_info(txop, sta, &txop->ti);
	}

	if (!scm80211_txq_is_ampdu_enabled(txq)) {
		txop->ampdu.enabled = false;
		txop->ampdu.update_needed = false;
	} else if (ieee80211_rate_is_non_ht(txop->rate)) {
		txop->ampdu.enabled = false;
		txop->ampdu.update_needed = true;
	} else {
		txop->ampdu.enabled = true;
		txop->ampdu.update_needed = true;
	}

#ifdef SCM80211_FTM
	if (ieee80211_rate_is_he(txop->rate)) {
		txop->he_gi_type = hwq->he_gi_type;
		txop->he_ltf_type = hwq->he_ltf_type;
	}
#endif

	scm80211_build_txvector(txop, txop->skb, &txop->tv, txop->rate);
	txop->bitrate = scm80211_bitrate(&txop->tv);
	return 0;
}

static void scm80211_enable_protection(struct txop *txop, enum scm80211_protection prot)
{
	if (prot == NO_PROTECTION || txop->protection == RTS_CTS_PROTECTION)
		return;

	if (prot == RTS_CTS_PROTECTION) {
		bool bandwidth_signal = false;

		bandwidth_signal |= ieee80211_rate_is_vht(txop->rate);
		bandwidth_signal |= ieee80211_rate_is_he(txop->rate);
		bandwidth_signal &= (ieee80211_ppdu_bandwidth(txop->rate) > CBW20);

		/*
		 * NB:
		 * When we try RTS with bandwidth signaling TA, we never get CTS.
		 * This is disabled for the time being, and we use legacy RTS/CTS.
		 */
		bandwidth_signal = 0;

		if (bandwidth_signal) {
			prot = RTS_CTS_DYN_BANDWIDTH_PROTECTION;

			txop->tv.CH_BANDWIDTH_IN_NON_HT = txop->tv.CH_BANDWIDTH;
			txop->tv.DYN_BANDWIDTH_IN_NON_HT = DYN_BW_STATIC;
		}
	}

	txop->protection = prot;
	txop->txtime += txop->rts_time;
}

/**
 * scm80211_setup_protetion() - evaluate protection for HT/VHT frames
 *
 * NOTE:
 * The mac80211 rate controller sets IEEE80211_TX_RC_USE_RTS_CTS or
 * IEEE80211_TX_RC_USE_CTS_PROTECT flags based on the content
 * of @ti->control:
 *
 * @ti(&struct ieee80211_tx_info)->control.use_rts:
 * - Set by mac80211 tx if the frame length exceeds the interface's RTS
 *   threshold.
 * - The rate controller sets IEEE80211_TX_RC_USE_RTS_CTS
 * @ti->control.use_cts_prot:
 * - This is set by mac80211 tx based on @vif->bss_conf.use_cts_prot,
 *   which reflects the value of Use_Protection field of ERP element
 *   in Beacon or Probe Response frames sent from the 2.4GHz BSS
 * @ti->control.rts_cts_rate_idx:
 * - This is set by mac80211 rc, and is the index to
 *   @wiphy->bands[@ti->band].bitrates[]
 *   array. Currently, mac80211 rc does not set this field for HT and VHT
 *   frames (need double check)
 *
 * However, the rate controller neither sets these flags, nor records
 * correct rate index to @ti->control.rts_cts_rate_idx for HT and VHT frames.
 * This function tries to set protection for HT and VHT frames according to
 * P802.11REVmd_D2.2 10.28.3.
 *
 * The function now refers to rate->flags instead of &txop->ti
 */
static void scm80211_setup_protection(struct txop *txop)
{
	struct ieee80211_tx_rate *rate = txop->rate;
	struct ieee80211_bss_conf *bss = &txop->owner->txq.vif->bss_conf;
	struct sk_buff *skb = txop->skb;
	struct ieee80211_hdr *hdr = (void *)skb->data;
	u16 ht_oper_mode = bss->ht_operation_mode;
	enum scm80211_protection prot = NO_PROTECTION;

	txop->protection = NO_PROTECTION;

	if (is_multicast_ether_addr(hdr->addr1))
		return;

	/*
	 * rc80211 enables RTS/CTS protection for rates whose index is greater than 0.
	 * This sometimes lead to poorer performance.
	 */
	if (rate->flags & (IEEE80211_TX_RC_USE_CTS_PROTECT | IEEE80211_TX_RC_USE_RTS_CTS))
		prot = RTS_CTS_PROTECTION;

	/*
	 * Draft P802.11REVmd_D2.2 10.28.3
	 */
	if (ieee80211_rate_is_ht(rate)) {
		enum cbw bandwidth = ieee80211_ppdu_bandwidth(rate);

		switch (ht_oper_mode & IEEE80211_HT_OP_MODE_PROTECTION) {
		case IEEE80211_HT_OP_MODE_PROTECTION_NONE:
			/*
			 * - All STAs detected (by any means) in the primary and
			 *   the secondary channel are HT STAs, and
			 * - All STAs that are known by the transmitting STA
			 *   to be a member of this BSS are either
			 *   - 20/40MHZ HT STAs in a 20/40 MHz BSS, or
			 *   - 20 MHz HT STAs in a 20 MHz BSS
			 *
			 * Necessary protection (PPDU, condition)
			 *  - (HT_GF PPDU, Nongreenfield HT STAs Present is 1):
			 *    N/A
			 */
			return;
		case IEEE80211_HT_OP_MODE_PROTECTION_20MHZ:
			/*
			 * - All STAs detected (by any means) in the primary
			 *   channel and all STAs detected (by any means) in the
			 *   secondary channel are HT STAs and all STAs that are
			 *   member of this BSS are HT STAs, and
			 * - This BSS is a 20/40 MHz BSS, and
			 * - There is at least one 20 MHz HT STA associated with
			 *   this BSS.
			 * Necessary protections (PPDU, condition)
			 * - (HT_GF, Nongreenfield HT STAs Present is 1): N/A
			 * - (40MHz, always): check again if this is necessary
			 */
			if (bandwidth > CBW20)
				prot = RTS_CTS_PROTECTION;

			break;
		case IEEE80211_HT_OP_MODE_PROTECTION_NONMEMBER:
			/*
			 * - A non-HT STA is detected (by any means) in either
			 *   the primary or the secondary channel or in both the
			 *   primary and secondary channels, that is not known
			 *   by the transmitting STA to be a member of this BSS,
			 *   and
			 * - All STAs that are known by the transmitting STA to
			 *   be a member of this BSS are HT STAs
			 *
			 * - All HT_MF and HT_GF
			 */
			/* fall through */
		case IEEE80211_HT_OP_MODE_PROTECTION_NONHT_MIXED:
			prot = RTS_CTS_PROTECTION;
			break;
		}
	}

	scm80211_enable_protection(txop, prot);
}

/**
 * scm80211_compute_overhead() - calculate txtime overhead
 * @txop: txop be in scheduling
 *
 * This function pre-calculates the time for RTS/CTS exchange and
 * for receiving Ack/BA based on the selected ppdu parameters.
 *
 * NB: ieee80211_rts_duration() does not serve our purpose.
 */
static void scm80211_compute_overhead(struct txop *txop)
{
	struct ieee80211_hw *hw = txop->hwq->hw;
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	struct ieee80211_tx_info *ti = &txop->ti;
	int ack_len;

	ack_len = (txop->ampdu.enabled ? IEEE80211_BACK_LEN : IEEE80211_ACK_LEN);

	txop->rts_time = ieee80211_compute_rts_time(hw, vif, ti);
	txop->ack_time = ieee80211_compute_ack_time(hw, vif, ti, ack_len);
}

/**
 * txop schedule with A-MPDU
 */

static inline int scm80211_num_subframe(struct txop *txop)
{
	return skb_queue_len(&txop->sched);
}

static inline int ieee80211_qos_ack_policy(__le16 qos)
{
	return le16_to_cpu(qos) & IEEE80211_QOS_CTL_ACK_POLICY_MASK;
}

static bool ampdu_is_data_enabled(struct scm80211_ampdu_info *ampdu)
{
	return ampdu->has_qos_data || ampdu->has_bar;
}

static bool ampdu_has_immediate_response(struct scm80211_ampdu_info *ampdu)
{
	unsigned long tidset = ampdu->tidset_ba | ampdu->tidset_eof;

	return !bitmap_empty(&tidset, IEEE80211_NUM_TIDS);
}

static bool ampdu_data_enabled_immediate_response(struct scm80211_ampdu_info *ampdu)
{
	return (ampdu_is_data_enabled(ampdu) && ampdu_has_immediate_response(ampdu));
}

/**
 * scm80211_ampdu_mmss_bytes() - minimum mpdu start spacing in byte
 * @sta: peer station
 * @bitrate: data rate in unit of 100kbps
 *
 * This function calculates the number of bytes necessary to satisfy the minimum
 * MPDU start spacing mandated by @sta.
 */
int scm80211_ampdu_mmss_bytes(struct ieee80211_hw *hw, struct ieee80211_sta *sta,
				     u32 bitrate)
{
	int mmss_4x = 0;	/* 4X mmss (us) */
	int bytes;
	int ampdu_density = get_ieee80211_ht_cap(sta).ampdu_density;

	BUG_ON(!get_ieee80211_ht_cap(sta).ht_supported);

	if (version(scm80211_mac_version(hw)) < 0x14 && ampdu_density < 5)
		ampdu_density = 5;	/* Fix to 4us under current FPGA */

	if (ampdu_density == 0)
		return 0;

	mmss_4x = 1 << (ampdu_density - 1);	/* 4 x mmss (us) */
	bytes = DIV_ROUND_UP(bitrate * mmss_4x, 8 * 4 * 10);
	return bytes;
}

/**
 * scm80211_num_mmss_delimiter() - compute number of MPDU delimiters
 *
 * This function calculate the number of MPDU delimiters necessary to
 * honor the minimum mpdu start spacing of the receiver by inspecting
 * the length of the last subframe.
 */
static int scm80211_num_mmss_delimiter(struct ieee80211_hw *hw, struct txop *txop)
{
	struct sk_buff *last = skb_peek_tail(&txop->sched);
	ssize_t len;
	int n = 1;

	if (!last)
		return 0;

	if (version(scm80211_mac_version(hw)) < 0x14)
		n = scm80211_get_skb_fragments(last);

	len = ieee80211_subframe_len(last);

	if (len >= txop->ampdu.mmss * n)
		return 0;

	return DIV_ROUND_UP(txop->ampdu.mmss * n - len, MPDU_DELIMITER_SIZE);
}

/**
 * scm80211_is_eof() - returns MPDU descriptor EOF setting
 *
 * S-MPDU: no ba && VHT/HE PPDU && solicits ack
 * Ack-enabled A-MPDU: (Management|| Data || QoS Data (TID: no BA)) &&
 *   solicts_ack
 */
static bool scm80211_is_eof(struct txop *txop, struct sk_buff *skb)
{
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	struct ieee80211_tx_info *ti = IEEE80211_SKB_CB(skb);
	struct txvector *tv = &txop->tv;
	bool ack = ieee80211_solicits_ack(hdr);

	if (!ack)
		return false;

	if (!txop->ampdu.enabled) {
		/* Not in A-MPDU: VHT and HE S-MPDU */
		return !ppdu_is_ht(tv);
	} else if (!ieee80211_is_data_qos(hdr->frame_control)) {
		return true;
	} else if (!(ti->flags & IEEE80211_TX_CTL_AMPDU)) {
		scm80211_info(txop->hw, "- eof=1: IEEE80211_TX_CTL_AMPDU not set\n");
		return true;
	}
	return false;
}

static void scm80211_print_mdesc(struct ieee80211_hw *hw, struct tx_mpdu_desc *mdesc,
				 struct tx_buffer_desc *bdesc, int indent)
{
#ifdef SCM80211_TXOP_DUMP
	int i;

	scm80211_trace(hw, "- mdesc@%#llx:\n", (u64)mdesc);
	scm80211_trace(hw, "  MT0: baddr=0x%llx\n", mdesc->bdaddr);
	scm80211_trace(hw, "  MT2: aid=%d tid=%d sn=%d noack=%d eof=%d\n",
		       mdesc->aid, mdesc->tid, mdesc->sn, mdesc->noack, mdesc->eof);
	scm80211_trace(hw, "  MT3: len=%d aggr=%d skip_enc=%d prot=%d no_htc=%d\n",
		       mdesc->mpdu_len, mdesc->aggregate, mdesc->bypass_encrypt,
		       mdesc->protection, mdesc->no_htc);
	scm80211_trace(hw, "  MT4: mmss=%d prot_duration=%d\n",
		       mdesc->mpdu_spacing, mdesc->prot_duration);
	scm80211_trace(hw, "  MT5: txtime0=%d us txtime1=%d\n", mdesc->txtime0, mdesc->txtime1);
	scm80211_trace(hw, "  MT6: buf_len=%d\n", mdesc->buf_len);

	if (bdesc) {
		scm80211_trace(hw, "- bdesc@%#llx (num=%d):\n", (u64)bdesc, mdesc->bdnum);

		for (i = 0; i < mdesc->bdnum + 1; i++) {
			scm80211_trace(hw, "  [%d]: addr=0x%08x%08x, len=%d\n",
				       i, bdesc->addr_h, bdesc->addr_l, bdesc->len);
			bdesc++;
		}
	}
#endif
}

static __used bool do_dump(struct txop *txop, struct ieee80211_hdr *hdr,
			   struct ieee80211_tx_info *ti)
{
#ifdef SCM80211_TXOP_DUMP
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	struct ieee80211_tx_rate *r = txop->rate;	/*&ti->control.rates[0]; */
	__le16 fc = hdr->frame_control;
	u16 stype = le16_to_cpu(fc) & IEEE80211_FCTL_STYPE;
	bool pass = false;

	pass |= ieee80211_is_mgmt(fc);
	if (ieee80211_iface_is_ap(vif->type))
		pass &= !ieee80211_is_beacon(fc);
	pass |= ieee80211_is_data(fc);
	pass |= ieee80211_is_ctl(fc);

	pass |= ieee80211_rate_is_vht(r);
	pass |= ieee80211_rate_is_he(r);

	(void)stype;
	return pass;
#else
	return false;
#endif
}

/**
 * scm80211_init_ampdu() - prepare for A-MPDU aggregation
 *
 * FIXME: what if @txop->skb is not a QoS Data frame?
 */
void scm80211_init_ampdu(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	struct ieee80211_tx_rate *rate = txop->rate;	/*&txop->ti.control.rates[0]; */
	int tid = txop->owner->txq.tid;

	ampdu->len = 0;
	ampdu->maxlen = ieee80211_max_ampdu_len(sta, rate);
	ampdu->bitrate = txop->bitrate;	/* To be removed */
	ampdu->mmss = scm80211_ampdu_mmss_bytes(hwq->hw, sta, txop->bitrate);
	ampdu->tids[tid].bufsize = scm80211_ba_buf_size(txop->owner);

	bitmap_zero(ampdu->tids[tid].bitmap, ampdu->tids[tid].bufsize);
	bitmap_zero(ampdu->tids[tid].ba, ampdu->tids[tid].bufsize);

	ampdu->has_bar = false;
	ampdu->has_eof = false;
	ampdu->has_qos_data = false;
	ampdu->has_immediate_response = false;
}

/**
 * reference to
 * P802.11ax_D6.0 9.2.4.6a.3 HLA Control
 * and P802.11REVmd_D2.2 9.2.4.6.3 VHT variant
 */
static void scm80211_insert_htc(struct txop *txop, struct ieee80211_sta *sta, struct sk_buff *skb)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	struct ieee80211_sta_priv *priv = (void *)sta->drv_priv;
	struct txvector *tv = &txop->tv;
	int hdrlen = ieee80211_hdrlen(hdr->frame_control);
	__le32 htc = 0;
	unsigned long flag;
	u8 hdr_backup[IEEE80211_MAX_HDR_LEN] = { 0, };

	if (priv->he_la != HE_LA_BOTH && priv->vht_la != VHT_LA_BOTH && !IS_OMI(priv->omi_req))
		return;

	if (ieee80211_has_order(hdr->frame_control))
		return;

	if (!ieee80211_rate_is_vht(txop->rate) && !ieee80211_rate_is_he(txop->rate))
		return;

	if (!ieee80211_is_action(hdr->frame_control) && !ieee80211_is_data_qos(hdr->frame_control))
		return;

	if (!vif || vif->type != NL80211_IFTYPE_STATION || !vif->bss_conf.assoc)
		return;

	hdr->frame_control |= cpu_to_le16(IEEE80211_FCTL_ORDER);
	memcpy(hdr_backup, skb->data, hdrlen);
	memcpy(skb_push(skb, IEEE80211_HT_CTL_LEN), hdr_backup, hdrlen);

	if (IS_OMI(priv->omi_req)) {
		u32 v;
		int vidx = scm80211_vdev_id(vif);

		spin_lock_irqsave(&priv->lock, flag);

		v = priv->omi_req;
		v |= FIELD_PREP(VIF_HTC_OMI_HETB_QOSNULL_OMI_EN, 1);
		v |= FIELD_PREP(VIF_HTC_OMI_HETB_DATA_OMI_EN, 1);
		scm80211_mac_writel(hwq->hw, v, MAC_REG_VIF_HTC_OMI(vidx));
		htc = priv->omi_req;

		spin_unlock_irqrestore(&priv->lock, flag);

	} else if (priv->he_la) {
		int hla_id = 2;
		int ru = 67;

		htc = HE_VARIANT | (hla_id << HE_CTLID_SHIFT);
		htc |= BIT(HE_MRQ_SHIFT);
		htc |= tv->CH_BANDWIDTH << HE_BW_SHIFT;
		if (tv->CH_BANDWIDTH == CBW20)
			ru = 61;
		else if (tv->CH_BANDWIDTH == CBW40)
			ru = 65;
		htc |= ru << HE_RU_ALLOC_SHIFT;
		spin_lock_irqsave(&priv->lock, flag);
		htc |= (priv->he_msi++ % 7) << HE_MSI_SHIFT;
		spin_unlock_irqrestore(&priv->lock, flag);
	} else {
		htc = VHT_VARIANT | BIT(VHT_MRQ_SHIFT);
		spin_lock_irqsave(&priv->lock, flag);
		htc |= (priv->vht_msi++ % 7) << VHT_MSI_SHIFT;
		spin_unlock_irqrestore(&priv->lock, flag);
	}
	put_unaligned_le32(htc, skb->data + hdrlen);

	scm80211_trace(hwq->hw, "%s (%08x)\n", __func__, htc);
}

/**
 * DOC: TX handlers

 * @TXOP_CONTINUE: keep scheduling
 * @TXOP_FINISH:   include the lasst one and terminate
 * @TXOP_STOP:     stop scheduling
 */
#define TXOP_CONTINUE     BIT(0)
#define TXOP_FINISH       BIT(1)
#define TXOP_STOP         BIT(2)

static int scm80211_txh_dynamic_ps(struct txop *txop, struct sk_buff *skb)
{
	struct ieee80211_hw *hw = txop->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_vif *vif = priv->pm.vif;

	if (!priv->pm.enable || priv->dbdc)
		return TXOP_CONTINUE;

	if (!vif)
		return TXOP_CONTINUE;

	if (scm80211_is_offchannel(hw, vif) || scm80211_is_scanning(hw))
		return TXOP_CONTINUE;

	if (vif->type != NL80211_IFTYPE_STATION)
		return TXOP_CONTINUE;

	/* wake up it if device under ps mode */
	scm80211_dynamic_ps_disable(hw, false);

	if (vif && !vif->bss_conf.assoc)
		return TXOP_CONTINUE;

	mod_timer(&priv->pm.timer, jiffies + msecs_to_jiffies(priv->pm.dynamic_ps_timeout));

	return TXOP_CONTINUE;
}

static int scm80211_txh_prepare_sched(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct txvector *tv = &txop->tv;
	struct scm80211_tx_info *xti;
#if (defined SCM80211_WFA_TEST) || (defined SCM80211_TWT)
	struct ieee80211_vif *vif = txop->owner->txq.vif;
#endif
#ifdef SCM80211_WFA_TEST
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
#endif
	struct ieee80211_sta *sta = txop->owner->txq.sta;
	struct ieee80211_hdr *hdr = (void *)skb->data;
	__le16 fc = hdr->frame_control;
	u32 qlen = skb_queue_len(&txop->sched);
	size_t plen = 0;
	unsigned int total_len;

	xti = txop_cur_xti(txop);
	if (unlikely(!xti))
		return TXOP_STOP;

	memset(xti, 0, sizeof(*xti));
	memcpy(&xti->tv, tv, sizeof(struct txvector));

#ifdef SCM80211_TWT
	if (scm80211_get_twt_state(vif) == TWT_SP)
		hdr->frame_control |= cpu_to_le16(IEEE80211_FCTL_PM);
#endif

	if (ieee80211_is_vht_sta(sta) || ieee80211_is_he_sta(sta))
		scm80211_insert_htc(txop, sta, skb);

	if (ppdu_is_non_ht(tv) || ppdu_is_ht(tv))
		plen = ieee80211_mpdu_len(skb);
	else
		plen = ieee80211_subframe_len(skb);

	total_len = txop->total_len + plen + sizeof(struct txvector) + sizeof(struct tx_mpdu_desc);
	if (scm80211_tx_bus_limit(hwq->hw, txop, qlen + 3, total_len))
		return TXOP_STOP;

#ifdef SCM80211_WFA_TEST
	if (txop->ampdu.enabled && vifp->hetb_psdu_len &&
	    txop->ampdu.len + plen + qlen * IEEE80211_HTC_SIZE > vifp->hetb_psdu_len) {
		scm80211_trace(hwq->hw, "- HE-TB len exceeded\n");
		return TXOP_STOP;
	}
#endif

	xti->mlen = ieee80211_mpdu_len(skb);
	xti->eof = scm80211_is_eof(txop, skb);
	xti->ack = ieee80211_solicits_ack((struct ieee80211_hdr *)skb->data);
	if (txop->ampdu.enabled)
		xti->tid = txop->owner->ba.tid;
	else
		xti->tid = ieee80211_is_data_qos(fc) ? ieee80211_get_tid(hdr) : 15;

	txvector_set_psdu_len(&xti->tv, plen);

	if (!txop->ampdu.enabled && qlen == 0)
		txvector_set_psdu_len(&txop->tv, plen);

	trace_scm80211_tx_return(hwq, __func__);
	return TXOP_CONTINUE;
}

/**
 * scm80211_txh_txop_limit() - check if txtime exceeds the txop limit
 *
 * 10.24.2.9
 *
 */
static int scm80211_txh_txop_limit(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	size_t n_mpdu = skb_queue_len(&txop->sched);
	u32 txtime = txop->txtime;

	if (WARN_ON(txop->ampdu.enabled))
		return TXOP_CONTINUE;

	if (txop->limit == 0 && n_mpdu > 0) {
		/*
		 * TXOP limit is zero and one MPDU is already scheduled
		 * We are done - stop further scheduling
		 */
		return TXOP_STOP;
	}

	xti->txtime = scm80211_ppdu_txtime(&xti->tv, txop->ti.band);
	if (WARN_ON(!xti->txtime))
		return TXOP_STOP;

	if (xti->ack)
		xti->duration = txop->ack_time;
	else
		xti->duration = 0;

	txtime += (n_mpdu > 0) ? ieee80211_sifs(txop->ti.band) : 0;
	txtime += xti->txtime + xti->duration;

	if (txop->limit > 0 && n_mpdu > 0 && txtime > txop->limit)
		return TXOP_STOP;

	if (txtime > IEEE80211_MAX_DURATION)
		return TXOP_STOP;

	if (txtime >= txop->rts_time_th && !txop->protection)
		scm80211_enable_protection(txop, RTS_CTS_PROTECTION);

	return TXOP_FINISH;
}

static int scm80211_txh_add_txvector(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	ssize_t size = sizeof(xti->tv);

	memcpy(skb_push(skb, size), &xti->tv, size);
	xti->flags |= SCM80211_XTI_HAS_TXVECTOR;

	trace_scm80211_tx_return(txop->hwq, __func__);
	return TXOP_CONTINUE;
}

static int scm80211_txh_setup_desc(struct txop *txop, struct sk_buff *skb)
{
	int n = skb_queue_len(&txop->sched);
	unsigned int ret;

	ret = scm80211_prepare_tx(txop->hw, txop, n, skb);
	trace_scm80211_tx_return(txop->hwq, __func__);

	if (ret)
		return TXOP_STOP;
	else
		return TXOP_CONTINUE;
}

static int scm80211_txh_admit_skb(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	struct txvector *tv = (void *)skb->data;
	struct ieee80211_hdr *hdr = (void *)(tv + 1);
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	u16 sn = IEEE80211_SEQ_TO_SN(le16_to_cpu(hdr->seq_ctrl));
	int tid, bit;

	if (txop->ampdu.update_needed) {
		__le16 fc = hdr->frame_control;
		size_t subframe_len;
		bool is_data_qos = ieee80211_is_data_qos(fc);

		ampdu->has_eof |= xti->eof;
		ampdu->has_bar |= ieee80211_is_bar(fc);
		ampdu->has_qos_data |= is_data_qos;

		tid = xti->tid;

		if (is_data_qos) {
			unsigned long *bitmap = ampdu->tids[tid].bitmap;
			unsigned int n_bits = ampdu->tids[tid].bufsize;

			if (bitmap_empty(bitmap, n_bits)) {
				memcpy(&ampdu->tids[tid].qosctl,
				       ieee80211_get_qos_ctl(hdr), sizeof(__le16));
				ampdu->tids[tid].ssn = sn;
			}

			bit = ieee80211_sn_sub(sn, ampdu->tids[tid].ssn);
			if (bit < n_bits)
				set_bit(bit, bitmap);
			else
				BUG_ON(1);
		}
		set_bit(tid, xti->eof ? &ampdu->tidset_eof : &ampdu->tidset_ba);

		/* Length */
		subframe_len = MPDU_DELIMITER_SIZE + xti->mlen;
		if (!ppdu_is_ht(tv))
			subframe_len = roundup(subframe_len, 4);

		ampdu->len = roundup(ampdu->len, 4);
		ampdu->len += xti->mmss_pad * MPDU_DELIMITER_SIZE;
		ampdu->len += subframe_len;

		/* txtime is calculated in scm80211_txh_ampdu_txop_limit() */
		txop->txtime = xti->txtime;
		txop->ack_time = xti->duration;
	} else {
		if (skb_queue_len(&txop->sched))
			txop->txtime += ieee80211_sifs(txop->ti.band);
		txop->txtime += xti->duration + xti->txtime;
	}

	/* Add skb to txop->sched list */
	skb_queue_tail(&txop->sched, skb);

	trace_scm80211_tx_return(txop->hwq, __func__);

	return skb_queue_len(&txop->sched) == SCM80211_MAX_MPDU ? TXOP_FINISH : TXOP_CONTINUE;
}

static int scm80211_txh_increase_queue_size(struct txop *txop, struct sk_buff *skb)
{
	scm80211_update_tid_queue_size(txop, skb, true);
	trace_scm80211_tx_return(txop->hwq, __func__);

	return TXOP_CONTINUE;
}

static int scm80211_txh_rate_control(struct txop *txop, struct sk_buff *skb)
{
	struct ieee80211_sta_priv *sta;
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct txvector tv[2];
	int n = skb_queue_len(&txop->sched);
	u32 bitrate[2];

	if (!(info->flags & IEEE80211_TX_CTL_RATE_CTRL_PROBE))
		return TXOP_CONTINUE;

	sta = ieee80211_to_priv(txop->owner->txq.sta);
	if (sta && sta->rate_fix_on == true) {
		scm80211_trace(txop->hw, "- rc probe ignored (fixed rate in use)\n");
		return TXOP_CONTINUE;
	}

	if (n == 0)
		txop->probe = true;
	else if (txop->probe && n >= 7)
		return TXOP_FINISH;

	/*
	 * txop->rs.control.rates[0] should be built later for it will rebuild the subtv with
	 * the same fmt, otherwise it may not , which may cause some side effect
	 */
	scm80211_build_txvector(txop, skb, &tv[1], &info->control.rates[0]);
	scm80211_build_txvector(txop, skb, &tv[0], &txop->rs.control.rates[0]);

	bitrate[0] = scm80211_bitrate(&tv[0]);
	bitrate[1] = scm80211_bitrate(&tv[1]);

	scm80211_trace(txop->hw, "- probe rate(idx=%d, flags=%08x) %c tx rate",
		       info->control.rates[0].idx, info->control.rates[0].flags,
		       bitrate[0] > bitrate[1] ? '<' : bitrate[0] < bitrate[1] ? '>' : '=');

	if (n > 0 && bitrate[0] < bitrate[1])
		/* Probe bit rate is higher than that of current tx rate */
		return TXOP_STOP;

	/* Don't break aggregation at the cost of not reporting to rc80211 */
	return TXOP_CONTINUE;
}

static int scm80211_txh_undo_schedule(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_tx_info *xti = txop_cur_xti(txop);

	if (!xti)
		return 0;

	if (xti->flags & SCM80211_XTI_DMA_MAPPED)
		scm80211_cleanup_tx(hwq->hw, txop, skb_queue_len(&txop->sched));

	if (xti->flags & SCM80211_XTI_HAS_TXVECTOR) {
		scm80211_trace(hwq->hw, "- undo: skb=%p\n", skb);

		skb_pull(skb, sizeof(struct txvector));
	}
	return 0;
}

/**
 * scm80211_txh_ampdu_content() - check if a frame can be added to A-MPDU
 *
 * This function figures out which of A-MPDU contexts in Table 9-529 the
 * current A-MPDU is beling built on and checks whether a frame can be
 * aggregated to it or not.
 *
 * Draft P802.11ax_D5.0 9.7.3 with delayed block ack and DMG excluded.
 *
 * (a) Non-HE data enabled immediate response context
 *  - Ack or HT-immediate BA (RD)
 *  - Action No Ack
 *  - Data frames sent under an HT-immediate BA agreement or
 *    Immediate BlockAckReq
 *
 * (b) Non-HT data enabled no immediate response context
 *  - Data without a BA agreement (ack policy = No Ack, A-MSDU Present = 0)
 *  - Action No Ack
 *  - Trigger (HE AP STA)
 *  - QoS Null frames with No Ack ack policy (HE STA)
 *
 * (c) PSMP
 *  : not supported
 *
 * (d) Control response context
 *  : handled by MAC hardware
 *
 * (e) S-MPDU
 *  : handled by the mpdu scheduler
 *
 * (f) HE non-ack-enabled singe-TID immediate response
 *  - Ack or HT-immediate BA (non-EOF)
 *  - Action No Ack (non-EOF)
 *  - QoS Null frame with No Ack ack policy (non-EOF)
 *  - Data frames sent under an HT-immediate BA agreement (same TID, same ack
 *   policy) (non_EOF)
 *  - Compressed BAR
 *  - Trigger
 *
 * (g) HE ack-enabled single-TID immediate response
 *  : not supported
 *
 * (h) HE non-ack-enabled multi-TID immediate response
 *  - Ack or HT-immediate BA
 *  - Action No Ack
 *  - QoS Null frame with No Ack ack policy
 *  - Data frames sent under a HT-immediate BA agreement (different TIDs)
 *  - Immediate BAR
 *  - Trigger
 *
 * (i) HE ack-enabled multi-TID immediate response
 *  : not supported
 *
 * In summary, only following frames can be aggregated:
 * - Ack or BA (should be the first, not sent by the driver)
 * - Trigger (second to Ack or BA, HE)
 * - Action No Ack
 * - QoS Data belonging to a TID having BA agreement
 * - QoS Data sent without a BA agreement (ack policy = No Ack)
 * - QoS Null with No Ack ack policy (HE)
 * - BAR (should be the last, mutually exclusive with QoS Data)
 *
 * FIXME
 * - If hardware supports Ack-enabled A-MPDU TX, support it.
 */
static int scm80211_txh_ampdu_content(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw *hw = hwq->hw;
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	struct ieee80211_hdr *hdr = (void *)skb->data;
	__le16 fc = hdr->frame_control;
	bool allowed = false;

	if (!txop->owner->txq.sta)
		return -EINVAL;

	/* 9.7.3 */
	if (!ether_addr_equal(hdr->addr1, txop->owner->txq.sta->addr)) {
		scm80211_warn(hw, "- wrong RA=%pM (TA=%pM)\n", hdr->addr1, hdr->addr2);

		scm80211_trace(hw, " - skb=%p, xti.flags=%08x\n", skb, xti->flags);
		print_hex_dump_bytes("skb", DUMP_PREFIX_OFFSET, skb->data, 64);
		return -EINVAL;
	}

	allowed = (ieee80211_is_data_qos(fc) ||
		   ieee80211_is_action_no_ack(fc) || ieee80211_is_bar(fc) || 0);

	if (!allowed)
		return -EAGAIN;

	if (ampdu->has_bar)
		/* BAR is the last MPDU. No more MPDU can be added */
		return -EAGAIN;

	if (ieee80211_is_data_qos(fc)) {
		int ack[2], tid = xti->tid;
		__le16 qos[2];
		unsigned int n_bits = ampdu->tids[tid].bufsize;

		n_bits = n_bits > 64 ? 64 : n_bits;
		BUG_ON(!n_bits);

		qos[0] = ampdu->tids[tid].qosctl;
		memcpy(&qos[1], ieee80211_get_qos_ctl(hdr), sizeof(qos[0]));
		ack[0] = ieee80211_qos_ack_policy(qos[0]);
		ack[1] = ieee80211_qos_ack_policy(qos[1]);
		if (bitmap_empty(ampdu->tids[tid].bitmap, n_bits))
			return 0;
		if (ack[0] != ack[1]) {
			/* QoS Ack Policy mismatch with former QoS Data frames */
			return -EAGAIN;
		}
		if (ieee80211_is_he_sta(txop->owner->txq.sta)) {
			if (ampdu->tidset_eof & BIT(tid)) {
				/* TID already has eof frame */
				scm80211_trace(hw, "- tid %d: non-eof is not scheduled with eof\n",
					       tid);
				return -EAGAIN;
			} else if (xti->eof && (ampdu->tidset_ba & BIT(tid))) {
				/* This is EOF and the schedule has non-EOF subframes */
				scm80211_trace(hw, "- tid %d: eof is not scheduled with non-eof\n",
					       tid);
				return -EAGAIN;
			}
		}
	} else if (ieee80211_is_action_no_ack(fc)) {
		/* Action No Ack frame can be in an A-MPDU of any context */
		return 0;
	} else if (ieee80211_is_bar(fc)) {
		/* BAR can be added as long as there are no QoS Data frames */
		return ampdu->has_qos_data ? -EAGAIN : 0;
	} else if (ieee80211_is_qos_nullfunc(fc)) {
		__le16 qos;
		int ack;

		/* Non-HE non-DMG A-MPDU does not include QoS Null */
		if (!ppdu_is_he(&txop->tv))
			return -EAGAIN;

		memcpy(&qos, ieee80211_get_qos_ctl(hdr), sizeof(qos));
		ack = ieee80211_qos_ack_policy(qos);
		if (ack != IEEE80211_QOS_CTL_ACK_POLICY_NOACK)
			return -EAGAIN;
	}
	trace_scm80211_tx_return(txop->hwq, __func__);
	return 0;
}

/**
 *
 * BUG: This needs to take into account scheduled-but-not-yet-transmitted other txop
 */
static inline int scm80211_txh_ba_window(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	struct ieee80211_hdr *hdr = (void *)skb->data;
	u16 sn = IEEE80211_SEQ_TO_SN(le16_to_cpu(hdr->seq_ctrl));
	int tid, delta;
	unsigned int n_bits;

	if (!ieee80211_is_data_qos(hdr->frame_control))
		return 0;

	tid = xti->tid;
	n_bits = ampdu->tids[tid].bufsize > SCM80211_MAX_MPDU << 1 ?
	    SCM80211_MAX_MPDU << 1 : ampdu->tids[tid].bufsize;
	BUG_ON(!n_bits);

	if (!bitmap_empty(ampdu->tids[tid].bitmap, n_bits)) {
		delta = ieee80211_sn_sub(sn, ampdu->tids[tid].ssn);
		if (delta < n_bits)
			return 0;

		scm80211_dbg(txop->hw, "- sn=%u, ssn=%u, delta=%d, bufsize=%zu\n",
			     sn, ampdu->tids[tid].ssn, delta, ampdu->tids[tid].bufsize);
		return -EAGAIN;
	}

	/* FIXME: check txq->sb.ssn */

	trace_scm80211_tx_return(txop->hwq, __func__);
	return 0;
}

/**
 * scm80211_txh_ampdu_len() - check ampdu maximum length limitation
 * @hwq: hardware ac queue
 * @skb: frame to add to @hwq->txop
 *
 */
static int scm80211_txh_ampdu_len(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	size_t totlen, len = ieee80211_subframe_len(skb);
	int ret = 0;

	/* Add subframe padding to the last subframe */
	if (ppdu_is_ht(&txop->tv))
		len = MPDU_DELIMITER_SIZE + xti->mlen;

	xti->mmss_pad = scm80211_num_mmss_delimiter(txop->hwq->hw, txop);
	len += xti->mmss_pad * MPDU_DELIMITER_SIZE;
	totlen = roundup(txop->ampdu.len, 4) + len;
	if (totlen >= txop->ampdu.maxlen) {
		scm80211_dbg(txop->hw, "ampdu length=%zu, max length= %zu\n",
			     totlen, txop->ampdu.maxlen);
		return -EAGAIN;
	}

	ret = check_tx_ampdu_limit(txop, totlen);

	trace_scm80211_tx_return(txop->hwq, __func__);
	return ret;
}

static int scm80211_compute_ba_len(struct txvector *tv,
				   unsigned long tidset_ba, unsigned long tidset_eof)
{
	int nr_ba_ctx, nr_ack_ctx;
	int ack_len = 18;

	nr_ba_ctx = bitmap_weight(&tidset_ba, IEEE80211_NUM_TIDS);
	nr_ack_ctx = bitmap_weight(&tidset_eof, IEEE80211_NUM_TIDS);

	if (nr_ack_ctx + nr_ba_ctx == 0) {
		return 0;
	} else if (nr_ack_ctx == 1 && nr_ba_ctx == 0) {
		/* Ack in response to an Ack-enabled A-MPDU */
		/*WARN_ON(ppdu_is_he(tv)); */
		return IEEE80211_ACK_LEN;
	} else if (nr_ack_ctx == 0 && nr_ba_ctx == 1) {
		/* C-BA in response to a single-TID A-MPDU */
		return IEEE80211_BACK_LEN;
	}

	/* M-BA */
	ack_len += nr_ack_ctx * 2;
	ack_len += nr_ba_ctx * (2 + 2 + 8);

	return ack_len + FCS_LEN;
}

/**
 * scm80211_txh_ampdu_txop_limit() - check ampdu txtime
 *
 * 10.24.2.9
 */
static int scm80211_txh_ampdu_txop_limit(struct txop *txop, struct sk_buff *skb)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_vif *vif = txop->owner->txq.vif;
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	struct ieee80211_hdr *hdr = (void *)skb->data;
	unsigned long tidset_ba, tidset_eof;
	size_t plen;
	int n = skb_queue_len(&txop->sched);
	int tid = 15;
	int ack_len;

	/* Calculate the txtime if skb is added including BA */
	plen = roundup(txop->ampdu.len, 4);
	if (ppdu_is_ht(&txop->tv))
		plen += xti->mlen + MPDU_DELIMITER_SIZE;
	else
		plen += ieee80211_subframe_len(skb);
	plen += xti->mmss_pad * MPDU_DELIMITER_SIZE;
	txvector_set_psdu_len(&txop->tv, plen);

	xti->txtime = scm80211_ppdu_txtime(&txop->tv, txop->ti.band);
	if (WARN_ON(!xti->txtime))
		return TXOP_STOP;

	if (n == 0)
		xti->per_time = xti->txtime;
	else
		xti->per_time = xti->txtime - txop->xti[n - 1].txtime;

	if (!txop->subused)
		goto skip;

	txvector_set_psdu_len(&txop->subtv, plen);
	xti->subtxtime = scm80211_ppdu_txtime(&txop->subtv, txop->ti.band);
	if (WARN_ON(!xti->subtxtime))
		return TXOP_STOP;

	if (n == 0)
		xti->subper_time = xti->subtxtime;
	else
		xti->subper_time = xti->subtxtime - txop->xti[n - 1].subtxtime;
skip:
	/* Calculate the Ack/BA length */
	tidset_ba = txop->ampdu.tidset_ba;
	tidset_eof = txop->ampdu.tidset_eof;

	tid = xti->tid;

	set_bit(tid, xti->eof ? &tidset_eof : &tidset_ba);
	ack_len = scm80211_compute_ba_len(&txop->tv, tidset_ba, tidset_eof);

	if (ieee80211_is_back_req(hdr->frame_control) && ack_len < IEEE80211_BACK_LEN)
		ack_len = IEEE80211_BACK_LEN;

	if (ack_len)
		xti->duration = ieee80211_compute_ack_time(hw, vif, &txop->ti, ack_len);

	if (xti->txtime + xti->duration >= txop->rts_time_th ||
	    (hw->wiphy->rts_threshold != (u32)-1 && plen > hw->wiphy->rts_threshold) ||
	    txop->retries >= 3) {
		scm80211_enable_protection(txop, RTS_CTS_PROTECTION);
	}

	/* TODO: to make sure txtime do not exceed the txtime of phy */
	if ((xti->txtime >= IEEE80211_PPDU_MAX_TXTIME ||
	     (txop->limit > 0 && n > 0 && xti->txtime + xti->duration > txop->limit)) && n != 0) {
		scm80211_trace(hw,
			       "- bitrate=%d.%dMbps len=%zu txtime=%u+%u txop_limit=%u\n",
			       txop->bitrate / 10, txop->bitrate % 10, plen,
			       xti->txtime, xti->duration, txop->limit);
		return TXOP_STOP;
	}

	if (txop->subused &&
	    (xti->subtxtime >= IEEE80211_PPDU_MAX_TXTIME ||
	     (txop->limit > 0 && n > 0 && xti->subtxtime + xti->duration > txop->limit))) {
		scm80211_trace(hw,
			       "sub rate excceed txop: len=%zu subtxtime=%u+%u txop_limit=%u\n",
			       plen, xti->subtxtime, xti->duration, txop->limit);
		return TXOP_STOP;
	}

	trace_scm80211_tx_return(hwq, __func__);
	return 0;
}

#define CALL_TXH(txh)							\
	do {								\
		res |= txh(txop, skb);					\
		if (res & TXOP_STOP) {					\
			scm80211_trace(txop->hwq->hw,			\
				       "- %s stop at n=%d\n",		\
				       #txh,				\
				       skb_queue_len(&txop->sched));	\
			goto txh_done;					\
		}							\
	} while (0)

/**
 * scm80211_sched_txop_sifs_burst() - schedule txop with multiple frames
 * @hwq: hardware AC queue
 * Returns: the number of scheduled frames.
 *
 * This function tries to schedule a TXOP with as many frames as possible
 * subject to TXOP limit.
 */
static int scm80211_sched_txop_sifs_burst(struct txop *txop)
{
	struct scm80211_txq *txq = txop->owner;
	struct sk_buff_head *queue;
	struct sk_buff *skb;
	int res = 0;

	if (unlikely(!txop->rate))
		return 0;

	queue = &txq->retry;

again:
	while ((skb = skb_dequeue(queue))) {
		CALL_TXH(scm80211_txh_dynamic_ps);
		CALL_TXH(scm80211_txh_prepare_sched);
		CALL_TXH(scm80211_txh_rate_control);
		CALL_TXH(scm80211_txh_txop_limit);
		CALL_TXH(scm80211_txh_add_txvector);
		CALL_TXH(scm80211_txh_setup_desc);
		CALL_TXH(scm80211_txh_admit_skb);
		CALL_TXH(scm80211_txh_increase_queue_size);

txh_done:
		if (res & TXOP_STOP) {
			/*
			 * @skb is not eligible for transmission within this
			 * txop. Undo all the changes made to @skb and
			 */
			scm80211_txh_undo_schedule(txop, skb);
			skb_queue_head(queue, skb);
			goto out;
		} else if (res & TXOP_FINISH) {
			goto out;
		}
	}

	if (queue == &txq->retry) {
		queue = &txq->queue;
		goto again;
	}

out:
	scm80211_trace(txop->hw, "- sifs burst n=%d\n", skb_queue_len(&txop->sched));
	return skb_queue_len(&txop->sched);
}

/**
 * scm80211_sched_txop_ampdu() - schedule an A-MPDU
 * @hwq: hardware ac queue
 *
 * FIXME: multi-tid A-MPDU construction
 */
static int scm80211_sched_txop_ampdu(struct txop *txop)
{
	struct ieee80211_hw_priv *hwp = txop->hwq->hw->priv;
	struct scm80211_txq *txq;
	struct sk_buff *skb;
	struct ieee80211_sta *sta;
	struct ieee80211_sta_priv *sta_priv;
	struct sk_buff_head *queue;
	int tid;
	unsigned int res = 0;

	if (unlikely(!txop->rate)) {
		scm80211_trace(txop->hw, "- no tx rate\n");
		return 0;
	}

	if (ppdu_is_non_ht(&txop->tv)) {
		scm80211_trace(txop->hw, "- non-ht ppdu\n");
		return 0;
	}

	txq = txop->multi_owner ? txop->multi_owner : txop->owner;
	sta = txq->txq.sta;
	if (sta)
		sta_priv = (void *)sta->drv_priv;

	queue = &txq->retry;

again:
	while ((skb = skb_dequeue(queue))) {
		ktime_t start, end;

		start = ktime_get();
		CALL_TXH(scm80211_txh_dynamic_ps);
		CALL_TXH(scm80211_txh_prepare_sched);
		CALL_TXH(scm80211_txh_rate_control);
		CALL_TXH(scm80211_txh_ampdu_content);
		CALL_TXH(scm80211_txh_ba_window);
		CALL_TXH(scm80211_txh_ampdu_len);
		CALL_TXH(scm80211_txh_ampdu_txop_limit);
		CALL_TXH(scm80211_txh_add_txvector);
		CALL_TXH(scm80211_txh_setup_desc);
		CALL_TXH(scm80211_txh_admit_skb);
		CALL_TXH(scm80211_txh_increase_queue_size);
		end = ktime_get();

		ewma_tx_sched_time_add(&hwp->sched_time, (ktime_to_ns(end) - ktime_to_ns(start)));

txh_done:
		if (res & TXOP_STOP) {
			/*
			 * @skb is not eligible for transmission within this
			 * txop. Undo all the changes made to @skb and
			 */
			scm80211_txh_undo_schedule(txop, skb);
			skb_queue_head(queue, skb);
			goto out;
		} else if (res & TXOP_FINISH)
			goto out;
	}

	if (queue == &txq->retry) {
		queue = &txq->queue;
		goto again;
	}

out:
	if (1) {
		struct scm80211_hwq *hwq = txop->hwq;
		struct ieee80211_hw *hw = hwq->hw;
		unsigned long tidset;

		scm80211_print_txvector(txop->buf, sizeof(txop->buf), hw,
					txop->owner->txq.sta, &txop->tv);

		tidset = txop->ampdu.tidset_ba | txop->ampdu.tidset_eof;
		for_each_set_bit(tid, &tidset, IEEE80211_NUM_TIDS) {
			int nbits;
			u16 ssn, end;

			nbits = txop->ampdu.tids[tid].bufsize;
			ssn = txop->ampdu.tids[tid].ssn;
			end = bitmap_find_msb(txop->ampdu.tids[tid].bitmap, nbits);

			scm80211_trace(hw,
				       "- tid[%d]: sched=[%d-%d] (%016llx)\n", tid,
				       ssn, ieee80211_sn_add(ssn, end),
				       get_unaligned_le64(txop->ampdu.tids[tid].bitmap));
		}

		scm80211_trace(hw, "- ampdu=%d prot=%d probe=%d\n",
			       txop->ampdu.enabled, txop->protection, txop->probe);

		if (skb_queue_len(&txq->retry) != 0)
			scm80211_warn(hw, "- unsched retries=%d\n", skb_queue_len(&txq->retry));
	}

	return skb_queue_len(&txop->sched);
}

/**
 * scm80211_init_txop() - initializes a txop
 * @hwq: hardware queue
 *
 */
static int scm80211_init_txop(struct txop *txop)
{
	struct scm80211_txq *txq = txop->owner;
	struct ieee80211_sta_priv *sta = ieee80211_to_priv(txq->txq.sta);

	memset(txop, 0, offsetof(struct txop, marker));
	txop->retransmit = 0;
	txop->vif = txq->txq.vif;
	txop->sta = txq->txq.sta;
	txop->limit = txop->hwq->edca.txop << 5;	/* in microsecond */
	txop->rts_time_th = U16_MAX;	/* no limit while no STA connected */
	if (sta)
		txop->rts_time_th = sta->txop_rts_th;

	txop->total_len = sizeof(txop->txop_info);

	txop->skb = skb_peek(&txq->retry);
	if (!txop->skb)
		txop->skb = skb_peek(&txq->queue);
	if (WARN_ON(!txop->skb))
		return -EINVAL;

	skb_queue_head_init(&txop->sched);

	if (scm80211_setup_txvector(txop) < 0)
		return -EINVAL;

	scm80211_compute_overhead(txop);
	scm80211_setup_protection(txop);

	if (txop->ampdu.update_needed)
		scm80211_init_ampdu(txop);

	return 0;
}

static int scm80211_pre_fec_padding[2][12] = {
	{12, 28, 44,  60,  88, 120, 132, 148, 180, 200, 224, 248},
	{28, 60, 88, 120, 180, 240, 268, 300, 360, 400, 448, 500},
};

/**
 * scm80211_fixup_he() -- workaround for post fcs padding
 */
static void scm80211_fixup_he(struct ieee80211_hw *hw, struct txop *txop)
{
	struct sk_buff *skb = skb_peek(&txop->sched);
	struct txvector *tv = (void *) skb->data;
	struct tx_mpdu_desc *mdesc = txop_get_mdesc(txop, 0);
	int a, nss, pad;

	if (!ppdu_is_he(tv))
		return;

	if (tv->CH_BANDWIDTH < CBW80)
		return;

	he_n_sym(tv, &a);
	if (a != 1)
		return;

	nss = tv->NUM_STS >> tv->STBC;
	pad = scm80211_pre_fec_padding[nss][tv->MCS];

	tv->APEP_LENGTH += pad;
	mdesc->mpdu_spacing += pad / MPDU_DELIMITER_SIZE;
}

static inline void scm80211_fixup_tx(struct ieee80211_hw *hw, struct txop *txop)
{
	scm80211_fixup_he(hw, txop);

	if (bus(hw)->ops->fixup_tx)
		bus(hw)->ops->fixup_tx(hw, txop);
}

/**
 * scm80211_finalize_txop() - finalizes the txop schedule
 * @hwq: hardware queue
 *
 * This function adjusts Duration/ID field of scheduled frames
 */
static void scm80211_finalize_txop(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct device *dev = wiphy_dev(hwq->hw->wiphy);
	struct sk_buff *skb;
	struct ieee80211_hdr *hdr;
	struct scm80211_txq *txq = txop->owner;
	struct ieee80211_vif *vif = txq->txq.vif;
	struct scm80211_tx_info *xti;
	int n;
	u16 sifs = ieee80211_sifs(txop->ti.band);
	u16 duration = 0;
	struct txvector *tv;

	txop->n_mpdu = skb_queue_len(&txop->sched);
	n = txop->n_mpdu - 1;

	if (txop->ampdu.enabled) {
		struct tx_mpdu_desc *md = txop_get_mdesc(txop, n);

		md->aggregate = false;
		if (ampdu_data_enabled_immediate_response(&txop->ampdu))
			duration = txop->ack_time;
	}

	skb_queue_reverse_walk(&txop->sched, skb) {
		struct sg_table *sg;
		struct tx_mpdu_desc *md = txop_get_mdesc(txop, n);

		xti = &txop->xti[n];
		sg = &xti->sgt;

		if (is_pcie_device(hwq->hw) || is_platform_device(hwq->hw))
			dma_sync_sg_for_cpu(dev, sg->sgl, sg->nents, DMA_TO_DEVICE);

		tv = (void *)skb->data;
		hdr = (void *)(skb->data + sizeof(xti->tv));

		if (!ieee80211_is_pspoll(hdr->frame_control)) {
			duration += txop->ampdu.enabled ? 0 : xti->duration;
			hdr->duration_id = cpu_to_le16(duration);
		}

		tv_set_he_txop_duration(tv, vif, duration);

		if (txop->ampdu.enabled)
			txvector_set_psdu_len(tv, txop->ampdu.len);

		if (is_pcie_device(hwq->hw) || is_platform_device(hwq->hw))
			dma_sync_sg_for_device(dev, sg->sgl, sg->nents, DMA_TO_DEVICE);

		if (n == 0 && txop->protection) {
			duration += 2 * sifs + txop->ack_time + txop->txtime;
			md->protection = txop->protection;
			md->prot_duration = cpu_to_le16(duration);
		}

		n--;
	}

	scm80211_fixup_tx(hwq->hw, txop);

	trace_scm80211_tx_return(hwq, __func__);
}

/**
 * scm80211_dump_txop() - finalizes the txop schedule
 * @hwq: hardware queue
 *
 * This function adjusts Duration/ID field of scheduled frames
 */
static void scm80211_dump_txop(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_tx_info *ti = &txop->ti;
	struct ieee80211_tx_rate *r = txop->rate;	/*&ti->control.rates[0]; */
	struct sk_buff *skb = skb_peek(&txop->sched);
	struct ieee80211_hdr *hdr;
	char str[64];
	int n = 0;

#ifndef SCM80211_TXOP_DUMP
	return;
#endif

	hdr = (void *)skb->data + sizeof(struct txvector);

	if (!do_dump(txop, hdr, ti))
		return;

	scm80211_dbg(txop->hw, "hwq#%d: txop schedule\n", hwq->id);
	scm80211_dbg(txop->hw, "- txq=%s\n", txop->owner->name);
	scm80211_dbg(txop->hw,
		     "- ti: flags=%08x, rate={idx=%d, count=%d, flags=%04x}\n",
		     ti->flags, r->idx, r->count, r->flags);

	if (!txop->ampdu.enabled) {
		scm80211_dbg(txop->hw,
			     "- sifs burst: %d frames, airtime=%u, txop limit=%u\n",
			     skb_queue_len(&txop->sched), txop->txtime, txop->limit);
	} else {
		struct scm80211_ampdu_info *ampdu = &txop->ampdu;
		int tid;

		scm80211_dbg(txop->hw,
			     "- ampdu: %d subframes, airtime=%u+%u, txop limit=%u\n",
			     skb_queue_len(&txop->sched),
			     txop->txtime, txop->ack_time, txop->limit);

		scm80211_dbg(txop->hw, "- tidset_ba=%04lx, tidset_eof=%04lx\n",
			     ampdu->tidset_ba, ampdu->tidset_eof);

		for_each_set_bit(tid, &ampdu->tidset_ba, IEEE80211_NUM_TIDS) {
			scm80211_dbg(txop->hw, "- tid[%d]: bufsiz=%zu, ssn=%d, bitmap=%016llx\n",
				     tid, ampdu->tids[tid].bufsize, ampdu->tids[tid].ssn,
				     get_unaligned_le64(ampdu->tids[tid].bitmap));
		}
	}

	n = 0;
	skb_queue_walk(&txop->sched, skb) {
		struct scm80211_tx_info *xti = &txop->xti[n];

		hdr = (void *)skb->data + sizeof(struct txvector);

		scm80211_snprint_frame(str, sizeof(str), hdr);
		scm80211_dbg(txop->hw, "[%d] %s\n", n, str);
		scm80211_dbg(txop->hw, "- txtime=%u+%u, flags=%08x, rates=%s\n",
			     xti->txtime, xti->duration, IEEE80211_SKB_CB(skb)->flags, str);
		scm80211_dbg(txop->hw, "  txop raddr 0x%lx\n", (unsigned long)txop->raddr);
		if (txop->space == MT_HST_BD_HST)
			scm80211_print_mdesc(hwq->hw, &txop->dma[0].desc->mdesc[n],
					     txop->dma[0].desc->mpdu[n].bdesc, 0);
		else if (txop->space == MT_HST_BD_TGT)
			scm80211_print_mdesc(hwq->hw, &txop->dma[0].desc->mdesc[n],
					     txop->dma[1].desc->mpdu[n].bdesc, 0);
		else
			scm80211_print_mdesc(hwq->hw, &txop->dma[1].desc->mdesc[n],
					     txop->dma[1].desc->mpdu[n].bdesc, 0);
		scm80211_print_txvector_raw(hwq->hw, (void *)skb->data);

		n++;
	}
	trace_scm80211_tx_return(hwq, __func__);
}

static void scm80211_txop_priority_get(struct txop *txop)
{
#ifdef SCM80211_COEX
	int pri = 0;
	int id = txop->hwq->id % SCM80211_NUM_HWQ_PER_VIF;

	switch (id) {
	case SCM80211_HWQ_HI:
		pri = 15;
		break;
	case SCM80211_HWQ_VO:
		pri = 7;
		break;
	case SCM80211_HWQ_VI:
		pri = 3;
		break;
	default:
		break;
	}
	txop->limit |= FIELD_PREP(TX_QUEN_TXOP_TX_PRI, pri);
	scm80211_trace(txop->hw, "%s(hwq#%d.%d): ac=%d, pri=%d, limit=0x%x\n",
		       __func__, txop->hwq->id, txop->id, id, pri, txop->limit);

#endif //SCM80211_COEX
}

/**
 * scm80211_start_backoff() - start a backoff procedure
 * @hwq: hardware queue
 */
bool scm80211_start_backoff(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct ieee80211_hw *hw = hwq->hw;
	bool ret;

	scm80211_txop_priority_get(txop);
	ret = scm80211_start_tx(hw, txop);
	if (ret)
		txop->tstamp = ktime_get();

	power_bitmap |= BIT(hwq->id);
	power_index[hwq->id] = txop->tv.TXPWR_LEVEL_INDEX;

	return ret;
}

/**
 * According to Draft P802.11ax_D7.0 9.2.4.5.6 Queue Size subfield
 */
static __le16 scm80211_build_qc_for_qos_null(struct ieee80211_sta *sta, int ac, int queue_size)
{
	u16 qc;

	if (!get_ieee80211_he_cap(sta).has_he) {
		qc = queue_size > 64768 ? 254 : DIV_ROUND_UP(queue_size, 256);
	} else {
		u8 sf, uv;

		if (queue_size <= 1008) {
			sf = 0;
			uv = DIV_ROUND_UP(queue_size, 16);
		} else if (queue_size <= 17152) {
			sf = 1;
			uv = queue_size > 1024 ? DIV_ROUND_UP(queue_size - 1024, 256) : 0;
		} else if (queue_size <= 146432) {
			sf = 2;
			uv = queue_size > 17408 ? DIV_ROUND_UP(queue_size - 17408, 2048) : 0;
		} else {
			sf = 3;
			uv = queue_size <= 148480 ? 0 :
			    queue_size > 2147328 ? 62 : DIV_ROUND_UP(queue_size - 148480, 32768);
		}
		qc = (sf << 6) | uv;
	}

	qc <<= 8;
	qc |= (IEEE80211_QOS_CTL_EOSP | IEEE80211_QOS_CTL_ACK_POLICY_NOACK);
	qc |= WME_AC_TO_TID(ac);

	return qc;
}

/**
 * According to Draft P802.11ax_D7.0 9.2.4.6a.4 BSR Control
 */
static const u16 bsr_sf[] = { 16, 256, 2048, 32768 };

static __le32 scm80211_build_htc_bsr(struct ieee80211_hw *hw, int *ac_queue_size, int n_tids)
{
	union {
		struct {
			__le32 acibm		:bits(3,  0);
			__le32 dtid		:bits(5,  4);
			__le32 acih		:bits(7,  6);
			__le32 sf		:bits(9,  8);
			__le32 qszh		:bits(17, 10);
			__le32 qszall		:bits(25, 18);
			__le32 unused		:bits(31, 26);
		};
		__le32 v;
	} bsr;

	u32 all_size = 0;
	int n_aci = 0;
	int i;

	bsr.v = 0;

	for (i = 0; i < IEEE80211_NUM_ACS; ++i) {
		if (ac_queue_size[i]) {
			bsr.acibm |= BIT(i);
			n_aci++;
			/*
			 * Choose AC which has longest ac queue as
			 * ACI High
			 */
			if (ac_queue_size[i] > ac_queue_size[bsr.acih])
				bsr.acih = i;
		} else {
			bsr.acibm &= ~BIT(i);
		}
		all_size += ac_queue_size[i];
	}

	for (i = 0; i < ARRAY_SIZE(bsr_sf); ++i) {
		if (all_size <= bsr_sf[i] * 255)
			break;
	}

	bsr.sf = i;
	bsr.qszall = DIV_ROUND_UP(all_size, bsr_sf[i]);
	bsr.qszh = DIV_ROUND_UP(ac_queue_size[bsr.acih], bsr_sf[i]);

	if (n_tids == 8) {
		bsr.acibm = 0;
		bsr.dtid = 3;
	} else {
		bsr.dtid = n_tids - n_aci;
	}

	return bsr.v;
}

static void scm80211_update_bsr_content(struct txop *txop)
{
	struct scm80211_txq *txq = txop->owner;
	struct ieee80211_vif *vif = txq->txq.vif;
	struct ieee80211_sta *sta = txq->txq.sta;
	struct ieee80211_hw *hw = txop->hwq->hw;
	struct ieee80211_vif_priv *vifp = (void *)vif->drv_priv;
	int ac_queue[IEEE80211_NUM_ACS] = { 0, };
	int i, n_tids = 0;
	u32 v;

	if (!vif || vif->type != NL80211_IFTYPE_STATION)
		return;

	if (!vif->bss_conf.assoc)
		return;

	vifp = (void *)vif->drv_priv;

	if (!sta || !(get_ieee80211_he_cap(sta).has_he))
		return;

	for (i = 0; i < SCM80211_NUM_TIDS; ++i) {
		if (vifp->tid_queue_size[i])
			n_tids++;
		ac_queue[TID_TO_WME_AC(i)] += vifp->tid_queue_size[i];
	}

	if (FIELD_GET(VIF_MIN_MPDU_SPACING_BSRP_BSR_EN, vifp->reg_mmss)) {
		u32 v = scm80211_build_htc_bsr(hw, ac_queue, n_tids);

		scm80211_trace(hw, "%s bsr content: 0x%08x\n", __func__, v);
		scm80211_mac_writel(hw, v, MAC_REG_VIF_BSR(vifp->index));
	} else {
		u16 qc;

		for (i = 0; i < IEEE80211_NUM_ACS; ++i) {
			u32 hwqidx = vif->hw_queue[i] % (SCM80211_NUM_HWQ / SCM80211_NUM_VIF);
			u32 reg = MAC_REG_VIF_QNULL_CTRL_AC(vifp->index, hwqidx);

			if (ac_queue[i] == 0)
				continue;

			qc = scm80211_build_qc_for_qos_null(sta, i, ac_queue[i]);

			v = scm80211_mac_readl(hw, reg);
			v &= ~VIF_QNULL_CTRL_AC_QOS_CTRL;
			v |= FIELD_PREP(VIF_QNULL_CTRL_AC_QOS_CTRL, qc);
			scm80211_mac_writel(hw, v, reg);
		}
	}

	trace_scm80211_tx_return(txop->hwq, __func__);
}

/**
 * scm80211_sched_txop_multid_ampdu() for multid AMPDU
 * Note: Multi-TID did not verify
 */
static int scm80211_sched_txop_multid_ampdu(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_txq *txq = txop->owner;
	struct scm80211_txq *txq2 = NULL;
	struct ieee80211_sta *sta = txq->txq.sta;
	struct ieee80211_sta_priv *sta_priv = (void *)sta->drv_priv;
	struct scm80211_ampdu_info *ampdu = &txop->ampdu;
	struct sk_buff *skb;
	struct ieee80211_hdr *hdr;
	int tid;
	int n = 0, m;
	u16 seq;

	if (!sta_priv->he_multid)
		goto exit;

	txq2 = scm80211_get_next_txq(hwq);
	if (!txq2)
		goto exit;

	if (txq2->txq.tid == txq->txq.tid)
		goto exit;

	if (txq2->txq.sta != sta)
		goto exit;

	if (!skb_queue_len(&txq2->queue))
		goto exit;

	if (!ieee80211_is_he_sta(txq->txq.sta) || !ieee80211_rate_is_he(txop->rate))
		goto exit;

	if (skb_queue_len(&txop->sched) >= SCM80211_MAX_MPDU - 4)
		goto exit;

	if (txop->limit > 0 && txop->txtime + txop->ack_time > txop->limit / 2)
		goto exit;

	scm80211_trace(hwq->hw, "STA support multi tid, handle next txq\n");

	skb = skb_peek(&txq->retry);
	if (!skb)
		skb = skb_peek(&txq2->queue);

	hdr = (void *)skb->data;
	seq = le16_to_cpu(hdr->seq_ctrl);
	tid = txq2->txq.tid;
	txop->multi_owner = txq2;
	ampdu->tids[tid].bufsize = scm80211_ba_buf_size(txq2);
	ampdu->tids[tid].ssn = IEEE80211_SEQ_TO_SN(seq);

	n = scm80211_sched_txop_ampdu(txop);
exit:
	if (txq2) {
		m = scm80211_num_pending_frames(txq2, true);
		if (hwq->mac80211)
			ieee80211_return_txq(hwq->hw, &txq2->txq, m > 0);
	}

	return n;
}

static void scm80211_aggr_check(struct txop *txop, struct sk_buff *skb, struct scm80211_txq *txq)
{
	struct ieee80211_hw_priv *priv = txop->hw->priv;
	struct ieee80211_sta *sta = txq->txq.sta;

	if (!priv->local_rc_used)
		return;

	if (scm80211_txq_is_ampdu_enabled(txq) || scm80211_txq_ba_session_started(txq))
		return;
	if (txq->txq.ac == IEEE80211_AC_VO || !sta ||
	    !get_ieee80211_ht_cap(sta).ht_supported || !sta->wme)
		return;
	if (ieee80211_rate_is_non_ht(txop->rate))
		return;
	if (unlikely(skb->protocol == cpu_to_be16(ETH_P_PAE)))
		return;
	if (skb_queue_len(&txq->retry) + skb_queue_len(&txq->queue) > 10)
		return;

	ieee80211_start_tx_ba_session(txq->txq.sta, txq->txq.tid, 0);
}

/**
 * scm80211_schedule_txop() - schedule a txop
 * @hwq: hardware queue
 *
 * BUG:
 * Because of speculative scheduling, we need to take into account
 * the schedule that has not been yet transmitted and acknowledged in
 * the other txop. We need to keep track of transmission window of txq.
 *
 * txop[0] resched can go beyond txop[1].
 *
 */
int scm80211_schedule_txop(struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;
	struct scm80211_txq *txq = txop->owner;
	int n;

	scm80211_trace(hwq->hw,
		       "%s(hwq#%d.%d): txq=%s pending=%d/%d/%d\n", __func__,
		       hwq->id, txop->id, txq->name,
		       skb_queue_len(&txq->retry),
		       skb_queue_len(&txq->queue), scm80211_num_pending_frames(txq, false));

	n = scm80211_dequeue_pending_frames(txop->owner);

	if (n == 0) {
		scm80211_trace(hwq->hw, "- txq empty\n");
		return -EINVAL;
	} else if (!hwq->enabled) {
		scm80211_warn(hwq->hw, "- hwq is disabled\n");
		return -EINVAL;
	} else if (txq->stop) {
		scm80211_warn(hwq->hw, "- txq is stopped\n");
		return -EINVAL;
	}

	n = 0;
	scm80211_init_txop(txop);
	if (txop->ampdu.enabled) {
		n = scm80211_sched_txop_ampdu(txop);
		n += scm80211_sched_txop_multid_ampdu(txop);
		WARN_ON(n == 0);
		/*
		 * NB: We now do not downgrade to sifs burst. Doing so breaks
		 * the transmit window tracking.
		 */
	} else {
		scm80211_aggr_check(txop, txop->skb, txop->owner);
		n = scm80211_sched_txop_sifs_burst(txop);
	}

	if (n > 0) {
		scm80211_update_bsr_content(txop);
		scm80211_finalize_txop(txop);
		scm80211_dump_txop(txop);
		scm80211_txop_change_state(txop, TXOP_STATE_SCHEDULED);
	}

	return n == 0 ? -EAGAIN : 0;
}

struct scm80211_txq *scm80211_get_next_txq(struct scm80211_hwq *hwq)
{
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq2;
	struct scm80211_txq *txq, *txq2;
	struct list_head txq_unmatch;
	bool ps = false;
	bool offchan = false;

	scm80211_trace(hw, "%s(hwq#%d)\n", __func__, hwq->id);

	if (!hwq->mac80211) {
		if (hwq->p2p_dev_enable) {
			txq = hwq->p2p_txq;
			if (txq && skb_queue_len(&txq->retry) + skb_queue_len(&txq->queue) > 0)
				return txq;
		}

		txq = hwq->txq;
		if (txq && skb_queue_len(&txq->retry) + skb_queue_len(&txq->queue) > 0)
			return txq;

		return NULL;
	}

	INIT_LIST_HEAD(&txq_unmatch);
	spin_lock_bh(&hwp->dequeue_lock);
	ieee80211_txq_schedule_start(hw, hwq->ac);

again:
	txq = (struct scm80211_txq *)ieee80211_next_txq(hw, hwq->ac);

	if (!txq)
		goto out;

	if (txq->stop) {
		scm80211_trace(hw, "- hwq#%d txq %s stopped\n", hwq->id, txq->name);
		ieee80211_return_txq(hw, &txq->txq, false);
		goto again;
	}

	if (txq->txq.vif != hwq->vif) {
		scm80211_trace(hw, "- hwq#%d vif mismatch with txq %s, skip it\n",
			       hwq->id, txq->name);

		if (list_empty(&txq->list))
			list_add_tail(&txq->list, &txq_unmatch);
		goto again;
	}

	offchan = scm80211_is_offchannel(hw, txq->txq.vif);
	if (txq->txq.sta) {
		struct ieee80211_sta_priv *sta = ieee80211_to_priv(txq->txq.sta);

		ps = sta->asleep;
		if (ps && (skb_queue_len(&txq->queue) + skb_queue_len(&txq->retry) > 0)) {
			/*
			 * Notify that driver buffered some packet for powersave STA
			 * to mac80211, mac80211 will set TIM info for this STA
			 */
			ieee80211_sta_set_buffered(txq->txq.sta, txq->txq.tid, true);
		}
		if (hwp->tdls.state == TDLS_CHANCTX_SWITCH_FINISH && txq->txq.sta->tdls)
			offchan = false;
	}

	if (ps || (offchan && txq != scm80211_vif_mgmt_txq(txq->txq.vif))) {
		/*
		 * txq is offchan or txq.sta is in ps, therefore we should not
		 * transmit to this station.
		 *
		 * We do not return txq to mac80211
		 * When return to recoverd channel,  scm80211_resume_vif
		 * will do ieee80211_wake_queues to wake all txq in  scm80211_wake_tx_queue
		 */

		scm80211_trace(hw, "- skipping %s ps=%d scanning=%d offchan=%d\n",
			       txq->name, ps, scm80211_is_scanning(hw), offchan);

		if (list_empty(&txq->list))
			list_add_tail(&txq->list, &txq_unmatch);
		goto again;
	}

	if (scm80211_dequeue_pending_frames(txq) == 0)
		goto again;

out:
	while (!list_empty(&txq_unmatch)) {
		txq2 = list_first_entry(&txq_unmatch, struct scm80211_txq, list);
		hwq2 = scm80211_hwq(hw, txq2->hw_queue);
		list_del_init(&txq2->list);
		ieee80211_return_txq(hw, &txq2->txq, true);
		scm80211_schedule_hwq_work(hwq2);
	}
	spin_unlock_bh(&hwp->dequeue_lock);

	return txq;
}

/**
 * scm80211_flush_hwq() - flush all pending frames from hardware queue
 *
 * @hwq: hardware AC queue
 *
 * If drop is 1, this function tries to empty @hwq by dropping pending
 * frames. If drop is 0, this function waits until all pending frames
 * are transmitted and retired.
 *
 */
void scm80211_flush_hwq(struct scm80211_hwq *hwq, bool drop)
{
	struct ieee80211_hw *hw = hwq->hw;

	scm80211_trace(hw, "%s(hwq#%d) drop=%d\n", __func__, hwq->id, drop);

	/* This eventually makes the txop scheduling fail */
	hwq->flush = drop ? HWQ_FLUSH_DROP : HWQ_FLUSH_DRAIN;

	/* Wait for the hardware queue to become idle. There must be a better way */
	flush_work(&hwq->work);

	if (!scm80211_txop_ring_empty(hwq)) {
		if (wait_for_completion_timeout(&hwq->completion, HZ) == 0)
			scm80211_warn(hw, "- hwq%d flush timeout\n", hwq->id);
	}

	hwq->flush = HWQ_FLUSH_NONE;
	scm80211_trace(hw, "- hwq#%d flush done\n", hwq->id);
}

/**
 * scm80211_txop_attach() -  attach txq/sta/vif to txop
 */
void scm80211_txop_attach(struct txop *txop, struct scm80211_txq *txq)
{
	txop->owner = txq;
	txop->multi_owner = NULL;

	txop->txq[0] = txq;
	txop->txq[1] = NULL;

	txop->sta = txq->txq.sta;
	txop->vif = txq->txq.vif;
}

/*
 * scm80211_hwq_tasklet() - performs time-consuming cleanup
 *
 * Note that another tx done irq can be raised while the tasklet is running.
 */
static void scm80211_hwq_tasklet(unsigned long data)
{
	struct scm80211_hwq *hwq = (struct scm80211_hwq *)data;
	struct ieee80211_hw_priv *hw_priv = hwq->hw->priv;
	struct txop *txop;

	if (is_pcie_device(hwq->hw)) {
		unsigned long flags;

		spin_lock_irqsave(&hwq->ring.lock, flags);
		if (!scm80211_has_txop(hwq, TXOP_STATE_ISSUED))
			hrtimer_try_to_cancel(&hwq->watchdog);
		spin_unlock_irqrestore(&hwq->ring.lock, flags);
	}

	while ((txop = scm80211_get_txop(hwq, TXOP_STATE_TX_DONE))) {
		scm80211_process_incomplete_txop(txop, false);
#ifdef SCM80211_COEX
		if (hw_priv->coex.coex_en) {
			if (txop->state == TXOP_STATE_TX_RETRY) {
				scm80211_txop_add(txop, TXOP_STATE_TX_RETRY);
				break;
			}
		}
#endif
		if (txop->state == TXOP_STATE_TX_RETRY) {
			if (!scm80211_start_backoff(txop))
				scm80211_schedule_hwq_work(hwq);
			continue;
		}
		scm80211_process_done_txop(txop, false);
		scm80211_complete_txop_epilogue(txop, false);
	}
}

static void scm80211_hwq_work(struct work_struct *wk)
{
	struct scm80211_hwq *hwq = container_of(wk, struct scm80211_hwq, work);
	struct ieee80211_hw *hw = hwq->hw;
	struct txop *txop;
	struct scm80211_txq *txq;
	u8 i;
	int ret;
	/* no_need: no need reschedule this hwq */
	bool no_need = true;

	scm80211_trace(hw, "%s(hwq#%d)\n", __func__, hwq->id);

	if (!unlikely(hwq->enabled)) {
		scm80211_trace(hw, "- hwq#%d disabled\n", hwq->id);
		return;
	}

	for (i = 0; i < ARRAY_SIZE(hwq->ring.pool); i++) {
		txop = scm80211_get_txop(hwq, TXOP_STATE_NEEDS_RESCHED);
		if (txop) {
			if (!scm80211_schedule_txop(txop))
				no_need = scm80211_start_backoff(txop);
			else
				scm80211_txop_free(txop);
		} else {
			break;
		}
	}

	txop = scm80211_get_txop(hwq, TXOP_STATE_SCHEDULED);
	if (txop)
		no_need = scm80211_start_backoff(txop);

#ifdef SCM80211_COEX
	txop = scm80211_get_txop(hwq, TXOP_STATE_TX_RETRY);
	if (txop)
		no_need = scm80211_start_backoff(txop);
#endif

	while ((txop = scm80211_txop_alloc(hwq)) != NULL) {
		int pending = 0;

		txq = scm80211_get_next_txq(hwq);
		if (!txq) {
			scm80211_txop_free(txop);
			break;
		}

		scm80211_txop_attach(txop, txq);

		ret = scm80211_schedule_txop(txop);
		if (ret < 0) {
			scm80211_warn(hw, "- txq %s: schedule fails\n", txq->name);
			scm80211_txop_free(txop);
			/* internal error occur, need resched current hwq later */
			if (ret == -EAGAIN)
				no_need = false;
			break;
		}
		no_need = scm80211_start_backoff(txop);

		pending = scm80211_num_pending_frames(txq, true);
		if (hwq->mac80211)
			ieee80211_return_txq(hw, &txq->txq, pending > 0);
	}

	if (scm80211_txop_ring_empty(hwq)) {
		/*
		 * No pending txop and no txq to schedule from. The hardware queue
		 * is parking. Next txq should schedule the work to get things
		 * started.
		 */
		hwq->idle = true;
		complete(&hwq->completion);
	} else {
		hwq->idle = false;
		reinit_completion(&hwq->completion);
	}

	hwq->need_resched = !(no_need || scm80211_txop_is_running(hwq));
	scm80211_trace(hw, "-%s(hwq#%d): idle=%d, %sneed reschedule\n", __func__,
		       hwq->id, hwq->idle, hwq->need_resched ? "" : "no ");
}

/**
 * DOC: mac80211 methods
 *
 * @conf_tx
 * @wake_tx_queue
 * @tx
 * @flush
 * @sta_rate_tbl_update
 */

void scm80211_sta_rate_tbl_update(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
				  struct ieee80211_sta *pubsta)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct ieee80211_sta_rates *rates;
	int i;

	scm80211_trace(hw, "%s: sta=%pM\n", __func__, pubsta->addr);

	rcu_read_lock();
	rates = rcu_dereference(pubsta->rates);
	if (!rates) {
		rcu_read_unlock();
		return;
	}

	for (i = 0; i < ARRAY_SIZE(rates->rate); i++) {
		if (rates->rate[i].idx == -1)
			break;
		scm80211_trace(hw, "- sta rate[%d]: idx=%d flags=%08x count=%d\n",
			       i, rates->rate[i].idx, rates->rate[i].flags, rates->rate[i].count);

		if (i == 0 && priv->local_rc_used) {
			scm80211_tx_he_rate_update(pubsta, rates->rate[i].flags,
						   rates->rate[i].idx);
		}
	}

	trace_scm80211_sta_rate_tbl_update(hw, vif, rates);

	rcu_read_unlock();
}

/**
 * scm80211_conf_tx() - ieee80211_ops conf_tx() callback
 *
 * This function records the EDCA parameter for @ac in the vif driver private
 * structure for later reference.
 */
int scm80211_conf_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
		     u16 ac, const struct ieee80211_tx_queue_params *params)
{
	struct scm80211_hwq *hwq;
	unsigned long flags;

	scm80211_trace(hw,
		       "%s(ac=%d): txop=%d, cw_min=%d, cw_max=%d, aifs=%d, acm=%d,uapsd:%d, mu_edca:%d\n",
		       __func__, ac,
		       params->txop, params->cw_min, params->cw_max,
		       params->aifs, params->acm, params->uapsd, params->mu_edca);

	if (params->mu_edca) {
		scm80211_trace(hw, "mu edca aifsn %d, ecw_min_max 0x%02x mu_timer %d",
			       params->mu_edca_param_rec.aifsn,
			       params->mu_edca_param_rec.ecw_min_max,
			       params->mu_edca_param_rec.mu_edca_timer);
	}

	hwq = scm80211_hwq(hw, vif->hw_queue[ac]);

	spin_lock_irqsave(&hwq->lock, flags);
	memcpy(&hwq->edca, params, sizeof(*params));
	memcpy(&hwq->b_edca, params, sizeof(*params));
	hwq->cw = hwq->edca.cw_min;

	if (params->mu_edca)
		hwq->mu_edca_status |= MU_EDCA_CONFIGURED;
	else
		hwq->mu_edca_status = 0;

	if (hwq->mu_edca_status == 0 && hrtimer_is_queued(&hwq->mu_edca_timer))
		hrtimer_cancel(&hwq->mu_edca_timer);

	spin_unlock_irqrestore(&hwq->lock, flags);

	if (ac == IEEE80211_AC_VO) {
		/* management hardware queue */
		hwq = scm80211_hwq(hw, scm80211_vif_mgmt_hw_queue(vif));

		spin_lock_irqsave(&hwq->lock, flags);
		memcpy(&hwq->edca, params, sizeof(*params));
		memcpy(&hwq->b_edca, params, sizeof(*params));
		hwq->edca.txop = 0;
		hwq->cw = hwq->edca.cw_min;

		if (params->mu_edca)
			hwq->mu_edca_status |= MU_EDCA_CONFIGURED;
		else
			hwq->mu_edca_status = 0;

		if (hwq->mu_edca_status == 0 && hrtimer_is_queued(&hwq->mu_edca_timer))
			hrtimer_cancel(&hwq->mu_edca_timer);

		spin_unlock_irqrestore(&hwq->lock, flags);
	}
	return 0;
}

int scm80211_ampdu_tx_action(struct ieee80211_hw *hw, struct ieee80211_vif *vif,
			     struct ieee80211_ampdu_params *params)
{
	struct ieee80211_sta *sta = params->sta;
	enum ieee80211_ampdu_mlme_action action = params->action;
	u16 tid = params->tid;
	struct scm80211_txq *txq = (struct scm80211_txq *)sta->txq[tid];

	switch (action) {
	case IEEE80211_AMPDU_TX_START:
		scm80211_trace(hw,
			       "TX BA session start: %s ssn=%u bufsize=%u amsdu=%d\n",
			       txq->name, params->ssn, params->buf_size, params->amsdu);

		memcpy(&txq->ba, params, sizeof(*params));
		memset(&txq->sb, 0, sizeof(txq->sb));
		txq->sb.ssn = IEEE80211_MAX_SN + 1;
		/* check whether driver is ready to do aggregation */
		if (skb_queue_len(&txq->queue) + skb_queue_len(&txq->retry) == 0 &&
		    !scm80211_txq_is_scheduled(txq)) {
			ieee80211_start_tx_ba_cb_irqsafe(vif, sta->addr, tid);
		} else {
			return IEEE80211_AMPDU_TX_START_DELAY_ADDBA;
		}
		break;
	case IEEE80211_AMPDU_TX_STOP_CONT:
	case IEEE80211_AMPDU_TX_STOP_FLUSH:
	case IEEE80211_AMPDU_TX_STOP_FLUSH_CONT:{
			const char * const action_str[] = { "cont", "flush", "flush-cont" };
			struct scm80211_hwq *hwq = scm80211_hwq(hw, txq->hw_queue);

			scm80211_trace(hw,
				       "TX BA session stopped: %s action %s\n", txq->name,
				       action_str[action - IEEE80211_AMPDU_TX_STOP_CONT]);

			/* FIXME - to do something */
			if (scm80211_txq_is_scheduled(txq))
				scm80211_flush_hwq(hwq, false);
			ieee80211_stop_tx_ba_cb_irqsafe(vif, sta->addr, tid);
			memcpy(&txq->ba, params, sizeof(*params));
		}
		break;
	case IEEE80211_AMPDU_TX_OPERATIONAL:
		scm80211_trace(hw,
			       "TX BA session established: %s ssn=%u bufsize=%u amsdu=%d\n",
			       txq->name, params->ssn, params->buf_size, params->amsdu);
		memcpy(&txq->ba, params, sizeof(*params));
		break;
	default:
		return -EOPNOTSUPP;
	}

	return 0;
}

/**
 * scm80211_wake_tx_queue() - ieee80211_ops wake_tx_queue callback
 *
 * @hw: the hardware
 * @queue: software tx queue (&struct ieee80211_txq)
 *
 * This function is called in a tasklet context
 *
 */
void scm80211_wake_tx_queue(struct ieee80211_hw *hw, struct ieee80211_txq *queue)
{
	struct scm80211_txq *txq = (struct scm80211_txq *)queue;
	struct scm80211_hwq *hwq = scm80211_hwq(hw, txq->hw_queue);
	struct ieee80211_hw_priv *hw_priv = hw->priv;
	int queued, retries, pending;
	bool kick = false;

	ieee80211_schedule_txq(hw, queue);

	retries = skb_queue_len(&txq->retry);
	queued = skb_queue_len(&txq->queue);
	pending = scm80211_num_pending_frames(txq, false);
#ifdef SCM80211_COEX
	if (!(hw_priv->coex.coex_en))
		kick = scm80211_schedule_hwq_work(hwq);
#else
	kick = scm80211_schedule_hwq_work(hwq);
#endif

	scm80211_trace(hw, "- %s(%s): pending=%d/%d/%d kick=%d\n", __func__,
		       txq->name, retries, queued, pending, kick);
}

/**
 * scm80211_tx() - transmit a skb
 * @hw: hardware
 * @control: peer information
 * @skb: frame to be sent
 *
 * This function, as far as we figured out, transmits non-bufferable MMPDUs.
 * If we denote @info to be IEEE80211_SKB_CB(@skb), @info->hw_queue is set to
 * either @hw->offchanel_tx_hw_queue (if sent over off-channels) or
 * @info->control.vif->hw_queue[IEEE80211_AC_VO] otherwise.
 *
 * We need to deal with the cases where the corresponding hardware queue
 * is already serving other txq's. The idea is to have a dedicated txq
 * and insert it at the head of the target hwq's list.
 *
 */
void scm80211_tx(struct ieee80211_hw *hw, struct ieee80211_tx_control *control, struct sk_buff *skb)
{
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct ieee80211_vif *vif = info->control.vif;
#ifdef SCM80211_SAP_31_CLIENT
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
#endif
	struct scm80211_txq *txq;
	struct scm80211_hwq *hwq;
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	__le16 fc = hdr->frame_control;
	char buf[128];
	bool skip = false;

	skip |= ieee80211_is_beacon(fc);
	skip |= ieee80211_is_probe_req(fc);

	if (!skip) {
		scm80211_snprint_frame(buf, sizeof(buf), hdr);
		scm80211_trace(hw, "%s(): %s (vdev#%d)\n", __func__, buf, scm80211_vdev_id(vif));
	}

#ifdef SCM80211_SAP_31_CLIENT
	if ((atomic_read(&vifp->num_sta) > (version(scm80211_mac_version(hw)) >= 0x16 ? 8 : 4)) &&
		(ieee80211_is_assoc_resp(fc) || ieee80211_is_reassoc_resp(fc))) {
		struct ieee80211_mgmt *mgmt;
		u8 *pos, *end, *cap_ie;
		u8 tag_info_len;

		mgmt = (struct ieee80211_mgmt *) skb->data;
		if (ieee80211_is_assoc_resp(fc))
			pos = (u8 *)mgmt->u.assoc_resp.variable;
		else
			pos = (u8 *)mgmt->u.reassoc_resp.variable;
		end = (u8 *)mgmt + (skb->len - FCS_LEN);

		while (pos + 1 < end) {
			if (pos[0] == WLAN_EID_EXTENSION)
				tag_info_len = 3;
			else
				tag_info_len = 2;
			if (pos + tag_info_len + pos[1] > end)
				return;
			if (pos[0] == WLAN_EID_EXTENSION && pos[2] == WLAN_EID_EXT_HE_CAPABILITY) {
				cap_ie = pos + tag_info_len; //HE_MAC_CAP
				cap_ie[1] &= ~IEEE80211_HE_MAC_CAP1_MULTI_TID_AGG_RX_QOS_MASK;
				break;
			}
			pos += tag_info_len + pos[1];
		}
		scm80211_info(hw, "%s assoc_resp, num_sta=%d, {ie}=%x\n",
			      __func__, atomic_read(&vifp->num_sta), cap_ie[1]);
	}
#endif

	txq = scm80211_vif_mgmt_txq(vif);
	scm80211_trace(hw, "%s, txq->hw_queue %d\n", __func__, txq->hw_queue);
	hwq = scm80211_hwq(hw, txq->hw_queue);

	if (ieee80211_is_beacon(fc) || ieee80211_is_assoc_resp(fc) || ieee80211_is_reassoc_resp(fc))
		skb_queue_head(&txq->queue, skb);
	else
		skb_queue_tail(&txq->queue, skb);

	if (!scm80211_is_offchannel(hw, vif) || ieee80211_maybe_offchannel_frame(fc))
		scm80211_schedule_hwq_work(hwq);
}

bool scm80211_vif_tx_frames_pending(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	unsigned long queues = scm80211_vif_queues(vif);
	struct scm80211_hwq *hwq = NULL;
	int i;
	bool pending = false;

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);

		if (i == scm80211_vif_mgmt_hw_queue(vif))
			pending |= !!scm80211_get_next_txq(hwq);
		else
			pending |= !scm80211_txop_ring_empty(hwq);

		if (pending)
			break;
	}

	scm80211_trace(hw, "%s (have %s pending frames)\n", __func__, pending ? "" : "no");

	return pending;
}

bool scm80211_tx_frames_pending(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_hwq *hwq = NULL;
	int i;
	bool pending = false;

	for (i = 0; i < SCM80211_NUM_HWQ; ++i) {
		hwq = &priv->hwq[i];

		if (i % SCM80211_NUM_HWQ_PER_VIF == SCM80211_HWQ_HI)
			pending |= !!scm80211_get_next_txq(hwq);
		else
			pending |= !scm80211_txop_ring_empty(hwq);

		if (pending)
			break;
	}

	scm80211_trace(hw, "%s (have %spending frames)\n", __func__, pending ? "" : "no ");
	return pending;
}

/*
 * FIXME: handle off-channel tx
 */
void scm80211_tx_channel(struct ieee80211_hw *hw, struct sk_buff *skb,
			 struct ieee80211_channel *chan)
{
	struct ieee80211_tx_info *info = IEEE80211_SKB_CB(skb);
	struct ieee80211_tx_control control = {
		.sta = NULL,
	};

	scm80211_trace(hw, "%s(channel=%d MHz [%d])\n", __func__,
		       chan->center_freq, ieee80211_frequency_to_channel(chan->center_freq));

	info->band = chan->band;

	if (ieee80211_hw_check(hw, SUPPORTS_RC_TABLE))
		ieee80211_get_tx_rates(info->control.vif, NULL, skb,
				       info->control.rates, ARRAY_SIZE(info->control.rates));

	scm80211_tx(hw, &control, skb);
}

static void scm80211_active_vif_queues(void *arg, u8 *mac, struct ieee80211_vif *vif)
{
	unsigned long *queues = arg;

	*queues |= scm80211_vif_queues(vif);
}

/**
 * scm80211_vif_queues_flush() - stops vif hardware queues
 *
 * This function is used by TWT to start TWT doze period
 */
void scm80211_vif_queues_flush(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	unsigned long queues = scm80211_vif_queues(vif);
	int i;

	for_each_set_bit(i, &queues, hw->queues) {
		struct scm80211_hwq *hwq = scm80211_hwq(hw, i);
		struct txop *txop;
		unsigned long flags;
		u32 cmd;
		bool edca = false;

		cmd = scm80211_mac_readl(hwq->hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		edca = !!FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd);

		spin_lock_irqsave(&hwq->ring.lock, flags);
		txop = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
		if (txop) {
			if (edca || is_usb_device(hw)) {
				/* scm80211_complete_txop(txop, false); */
				spin_unlock_irqrestore(&hwq->ring.lock, flags);
				scm80211_cancel_tx(hwq->hw, txop, true);
			} else {
				__scm80211_txop_add_back(txop);
				spin_unlock_irqrestore(&hwq->ring.lock, flags);
			}
		} else {
			spin_unlock_irqrestore(&hwq->ring.lock, flags);
		}
	}
}

/**
 * scm80211_flush() - flush hardware queues
 *
 * @queues: bitmap of hardware queue to flush
 *
 * This function is called by mac80211. @queues are stopped and woken up before and
 * after this function, respectively.
 *
 * All hardware queues must be atomically flushed.
 */
void scm80211_flush(struct ieee80211_hw *hw, struct ieee80211_vif *vif, u32 queues, bool drop)
{
	struct scm80211_hwq *hwq;
	unsigned long queues_to_flush = 0;
	unsigned long all_queues = BIT(SCM80211_HWQ_BE) | BIT(SCM80211_HWQ_BK) |
	    BIT(SCM80211_HWQ_VI) | BIT(SCM80211_HWQ_VO);
	int i;

	scm80211_trace(hw, "+%s(vif=%d, queues=%08x, drop=%d)\n",
		       __func__, scm80211_vdev_id(vif), queues, drop);

	ieee80211_iterate_active_interfaces_atomic(hw,
						   IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_active_vif_queues, &queues_to_flush);

	if (vif) {
#if KERNEL_VERSION(5, 14, 0) > LINUX_VERSION_CODE
		/*
		 * for linux version < 5.14 it disabled qos null with PM on wifi6
		 * so it can not enter into power save mode when connecting with wifi6
		 * driver has to sent it manually.
		 *
		 * mac80211 will get tx status for NULL Qos data with PM bit,
		 * then the legacy PM will be enabled on mac80211
		 */
		struct ieee80211_bss_conf *bss = &vif->bss_conf;
		struct cfg80211_chan_def def;

		if (!bss->assoc || !bss->he_support || !ieee80211_hw_check(hw, SUPPORTS_PS))
			goto flush;

		if (scm80211_vif_chan(vif, &def))
			goto flush;

		scm80211_send_nullfunc(hw, vif, &def, true);
flush:
#endif

		/* flush HIQ if all_queues need flush */
		queues_to_flush &= scm80211_vif_queues(vif);
		if (queues != all_queues)
			queues_to_flush &= queues;
	}

	for_each_set_bit(i, &queues_to_flush, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		if (hwq->enabled)
			scm80211_flush_hwq(hwq, drop);
	}

	scm80211_trace(hw, "-%s(vif=%d, queues=%08x, drop=%d)\n",
		       __func__, scm80211_vdev_id(vif), queues, drop);
}

#ifdef SCM80211_WFA_TEST
static void scm80211_reset_he_psdu_len(void *arg, u8 *mac, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);

	vifp->hetb_psdu_len = 0;
}
#endif

static enum hrtimer_restart scm80211_mu_edca_timer(struct hrtimer *timer)
{
	struct scm80211_hwq *hwq = container_of(timer, struct scm80211_hwq, mu_edca_timer);
	unsigned long flags;

	int aifs = hwq->edca.aifs;

	spin_lock_irqsave(&hwq->lock, flags);

	aifs = hwq->edca.aifs;

	if (hwq->mu_edca_status & MU_EDCA_CONFIGURED) {
		/* restore EDCA param */
		hwq->edca.aifs = hwq->b_edca.aifs;
		hwq->edca.cw_max = hwq->b_edca.cw_max;
		hwq->edca.cw_min = hwq->b_edca.cw_min;
		hwq->cw = hwq->edca.cw_min;
		/* initialize timer */
		hwq->mu_edca_status &= ~MU_EDCA_RUNNING;
	}
	spin_unlock_irqrestore(&hwq->lock, flags);

	scm80211_trace(hwq->hw, "(HWQ#%d) mu-edca timer expried reload edca[%d:%d] aifs %d\n",
		       hwq->id, hwq->edca.cw_min, hwq->edca.cw_max, hwq->edca.aifs);

#ifdef SCM80211_WFA_TEST
	ieee80211_iterate_active_interfaces_atomic(hwq->hw, IEEE80211_IFACE_ITER_NORMAL,
						   scm80211_reset_he_psdu_len, NULL);
#endif

	if (aifs == 0 && is_pcie_device(hwq->hw)) {
		struct txop *txop;
		unsigned long flags2;
		u32 cmd = scm80211_mac_readl(hwq->hw, MAC_REG_TX_QUEN_CMD(hwq->id));
		bool edca = FIELD_GET(TX_QUEN_CMD_CMD_EDCA, cmd);

		spin_lock_irqsave(&hwq->ring.lock, flags2);
		txop = __scm80211_get_txop(hwq, TXOP_STATE_ISSUED);
		if (txop) {
			if (edca) {
				/* scm80211_complete_txop(txop, false); */
				spin_unlock_irqrestore(&hwq->ring.lock, flags2);
				scm80211_cancel_tx(hwq->hw, txop, false);
			} else {
				__scm80211_txop_add_back(txop);
				spin_unlock_irqrestore(&hwq->ring.lock, flags2);
			}
		} else {
			spin_unlock_irqrestore(&hwq->ring.lock, flags2);
		}
	}

	return HRTIMER_NORESTART;
}

int scm80211_sta_set_txpwr(struct ieee80211_hw *hw,
			   struct ieee80211_vif *vif, struct ieee80211_sta *sta)
{
	if (get_ieee80211_txpwr(sta).type != NL80211_TX_POWER_AUTOMATIC)
		scm80211_trace(hw, "set sta txpower to %d\n", get_ieee80211_txpwr(sta).power);
	return 0;
}

/**
 * scm80211_init_hwq() - initialize hw queues
 * @hw: hardware
 * This function is called during @hw initialization
 */
void scm80211_init_tx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq;
	u32 v;
	int i, m, n;

	/* TX timeout */
	v = scm80211_mac_readl(hw, MAC_REG_TX_TO_CFG);
	/* set timeout to 8192us */
	v |= (FIELD_PREP(TX_TO_CFG_TO_EN, 1) | FIELD_PREP(TX_TO_CFG_TO_VALUE, 0x1fff));
	scm80211_mac_writel(hw, v, MAC_REG_TX_TO_CFG);

	/*set tx timeout as 16ms for CCK rate */
	v = (FIELD_PREP(TX_TO_11B_CFG_TO_11B_VALUE, 0x3fff));
	scm80211_mac_writel(hw, v, MAC_REG_TX_TO_11B_CFG);

	/*config tx/rx busy */
	v = scm80211_mac_readl(hw, MAC_REG_DEV_OPT2);
	v |= (FIELD_PREP(DEV_OPT2_SIFS_LOST_STOP_TX_EN, 1));
	v |= (FIELD_PREP(DEV_OPT2_AID_MATCH_BASIC_TRIG_END_EN, 1));
	v |= (FIELD_PREP(DEV_OPT2_RXBUSY_CHK_EN, 1));
	v |= (FIELD_PREP(DEV_OPT2_CONT_TX_TERMI_FSMIDLE_EN, 1));
	v |= (FIELD_PREP(DEV_OPT2_AID_MATCH_BASIC_TRIG_END_EN, 1));
	v |= (FIELD_PREP(DEV_OPT2_EDCA_TXVEC_PROT_EN, 1));
	scm80211_mac_writel(hw, v, MAC_REG_DEV_OPT2);
	spin_lock_init(&hwp->dequeue_lock);

	power_bitmap = 0;
	memset(txq_bitmap, 0, sizeof(txq_bitmap));

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		hwq = &hwp->hwq[i];
		memset(hwq, 0, sizeof(*hwq));
		hwq->id = i;
		hwq->hw = hw;
		hwq->qstate = HWQ_STATE_IDLE;
		hwq->he_sta = NULL;
		hwq->edca.cw_min = 15;
		hwq->edca.cw_max = 1023;
		hwq->edca.aifs = 2;
		spin_lock_init(&hwq->lock);
		INIT_LIST_HEAD(&hwq->txq_head);

		for (m = 0; m < TXOP_STATE_END; m++)
			INIT_LIST_HEAD(&hwq->ring.head[m]);

		for (n = 0; n < SCM80211_MAX_TXOP; n++) {
			hwq->ring.pool[n].hwq = hwq;
			hwq->ring.pool[n].hw = hw;
			hwq->ring.pool[n].id = n;
			INIT_LIST_HEAD(&hwq->ring.pool[n].list);
			scm80211_txop_change_state(&hwq->ring.pool[n], TXOP_STATE_IDLE);
			scm80211_txop_add(&hwq->ring.pool[n], TXOP_STATE_IDLE);
		}
		spin_lock_init(&hwq->ring.lock);

		hrtimer_init(&hwq->mu_edca_timer, CLOCK_MONOTONIC, HRTIMER_MODE_ABS_SOFT);
		hwq->mu_edca_timer.function = scm80211_mu_edca_timer;

		scm80211_mac_writel(hw, 0, MAC_REG_TX_QUEN_CMD(i));
	}

#if defined(SCM2020_Z1) || defined(SCM2020_A0)
	//MAC timing
	scm80211_mac_writel(hw, 0x00a0, TXE_TIM_INFO0);
	//MAC SIFS timing
	scm80211_mac_writel(hw, 0x0238, TIM_IFS_CFG1);
#else
/* used for FPGA */
#ifdef SCM80211_WFA_TEST

#ifdef ERRATA_TX_TIMING_GAP
	/**
	 * timing gap should be within 15.6us ~ 16.4us in responding he_tb
	 **/
	scm80211_mac_writel(hw, 0x11F, TIM_IFS_CFG1);
#endif

#else

#ifdef ERRATA_TX_BA_SIFS_LOST
	scm80211_mac_writel(hw, 0, TIM_IFS_CFG1);
#endif

#endif /* SCM80211_WFA_TEST */
#endif

	phy_writel(&hwp->phy[hwp->devid], 1, PHY_TXBE_MON_CNT_RST);
	phy_writel(&hwp->phy[hwp->devid], 1, PHY_TXBE_MON_CNT_EN);
	phy_writel(&hwp->phy[hwp->devid], 1, PHY_MON_CNT_RST);

	for (i = 0; i < SCM80211_NUM_VIF; i++) {
		v = scm80211_mac_readl(hw, MAC_REG_VIF_CFG_INFO(i));
		v |= (FIELD_PREP(VIF_CFG_INFO_TRIG_CS_ASSO_CHK_DIS, 1));
		v |= (FIELD_PREP(VIF_CFG_INFO_TRIG_CS_UNASSO_CHK_DIS, 1));
		scm80211_mac_writel(hw, v, MAC_REG_VIF_CFG_INFO(i));
	}

	hrtimer_init(&hwp->cca_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
	hwp->cca_timer.function = scm80211_enable_cca_detect;
#ifdef SCM80211_COEX
	hwp->coex.coex_en = false;
	hrtimer_init(&hwp->coex.stop_queue_watchdog, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
	hwp->coex.stop_queue_watchdog.function = scm80211_hwq_stop_queue_watchdog_expire;
	if (hwp->coex.dev_ant == 1) {
		hwp->coex.reduce_streams_flag = false;
		hwp->coex.reduce_streams_bypass = false;
		hwp->coex.bt_action_counter = 0;
		hrtimer_init(&hwp->coex.coex_2ant_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
		hwp->coex.coex_2ant_timer.function = scm80211_coex_2ant_mon_timer;
	}
#endif
	scm80211_bus_init_tx(hw);
}

void scm80211_vif_map_hwqueue(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct scm80211_hwq *hwq;
	unsigned long queues = scm80211_vif_queues(vif);
	int i, ac;

	i = find_first_bit(&queues, hw->queues);
	vif->hw_queue[IEEE80211_AC_VO] = i + SCM80211_HWQ_VO;
	vif->hw_queue[IEEE80211_AC_VI] = i + SCM80211_HWQ_VI;
	vif->hw_queue[IEEE80211_AC_BE] = i + SCM80211_HWQ_BE;
	vif->hw_queue[IEEE80211_AC_BK] = i + SCM80211_HWQ_BK;
	/* we use the 5th hardware queue */
	vif->cab_queue = i + SCM80211_HWQ_HI;
	vp->mgmt_hw_queue = i + SCM80211_HWQ_HI;

	for (ac = 0; ac < IEEE80211_NUM_ACS; ac++) {
		hwq = scm80211_hwq(hw, vif->hw_queue[ac]);
		hwq->ac = ac;
		hwq->mac80211 = true;
		/*
		 * hwq->limit will be used in bus->ops->bus_limit function
		 * for pcie soluction there is no bus_limit api which means
		 * no len limitation on pcie bus
		 */
		if (ac == IEEE80211_AC_VO)
			hwq->limit = SCM80211_TXOP_MPDU_LEN;
		else
			hwq->limit = SCM80211_TXOP_AMPDU_LEN;
	}
	hwq = scm80211_hwq(hw, vp->mgmt_hw_queue);
	hwq->ac = IEEE80211_AC_VO;
	hwq->mac80211 = false;
	hwq->limit = SCM80211_TXOP_MPDU_LEN;
}

void scm80211_deinit_tx(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *hwp = hw->priv;
	struct scm80211_hwq *hwq;
	int i;

	for (i = 0; i < SCM80211_NUM_HWQ; i++) {
		hwq = &hwp->hwq[i];
		hrtimer_cancel(&hwq->mu_edca_timer);
	}

	hrtimer_cancel(&hwp->cca_timer);

#ifdef SCM80211_COEX
	hrtimer_cancel(&hwp->coex.stop_queue_watchdog);
#endif

	scm80211_bus_deinit_tx(hw);
}

/**
 * scm80211_suspend_tx_work() - temporarily suspend tx to wait for beacon
 */
static void scm80211_suspend_tx_work(struct work_struct *work)
{
	struct ieee80211_vif_priv *vifp = container_of(work, struct ieee80211_vif_priv,
						       suspend_tx_work.work);
	struct ieee80211_vif *vif = priv_to_ieee80211(struct ieee80211_vif, vifp);
	struct ieee80211_hw *hw = ieee80211_vif_to_hw(vif);

	scm80211_trace(hw, "+%s\n", __func__);

	if (is_pcie_device(hw))
		scm80211_vif_queues_flush(hw, vif);
	scm80211_stop_queues(hw, scm80211_vif_queues(vif));

	vifp->wait_beacon = true;
}

/**
 * scm80211_vif_init_tx() - initialize hwq's belonging to vif
 *
 */
int scm80211_vif_init_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct scm80211_txq *mq;
	struct scm80211_hwq *hwq;
	struct ieee80211_vif_priv *vifp = ieee80211_to_priv(vif);
	unsigned long queues;
	int i, ret = 0;

	scm80211_vif_map_hwqueue(hw, vif);
	if (vif->txq)
		scm80211_init_txq(hw, vif, vif->txq, false);

	/* Initialize internal mgmt txq */
	mq = scm80211_vif_mgmt_txq(vif);
	mq->txq.vif = vif;
	mq->txq.sta = NULL;
	mq->txq.tid = IEEE80211_NUM_TIDS;
	mq->txq.ac = IEEE80211_AC_VO;
	scm80211_init_txq(hw, vif, &mq->txq, true);
	if (vif->type == NL80211_IFTYPE_P2P_DEVICE)
		return 0;

	/* Activate hardware queues */
	queues = scm80211_vif_queues(vif);

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		ret = scm80211_hwq_alloc_dma_desc(hw, hwq);
		if (ret)
			goto fail;

		INIT_WORK(&hwq->work, scm80211_hwq_work);
		tasklet_init(&hwq->tasklet, scm80211_hwq_tasklet, (unsigned long)hwq);
		init_completion(&hwq->completion);
		/* Initialize completion as done */
		complete(&hwq->completion);
		hrtimer_init(&hwq->watchdog, CLOCK_MONOTONIC, HRTIMER_MODE_REL_SOFT);
		hwq->enabled = true;
		hwq->need_resched = false;
		hwq->idle = true;
		hwq->vif = vif;
	}
#ifdef SCM80211_FTM
	scm80211_ftm_init(hw, vif);
#endif //SCM80211_FTM

	INIT_DELAYED_WORK(&vifp->suspend_tx_work, scm80211_suspend_tx_work);
	scm80211_enable_tx_irq(hw, queues);
	scm80211_enable_trig_qnull_irq(hw, vif);
	ret = scm80211_bus_vif_init_tx(hw, vif);
	if (ret)
		goto fail;
	else
		return ret;

fail:
	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		scm80211_hwq_free_dma_desc(hw, hwq);
		hwq->enabled = false;
	}
	return ret;
}

/**
 * scm80211_vif_deinit_tx() - clean up vif tx queues
 */
void scm80211_vif_deinit_tx(struct ieee80211_hw *hw, struct ieee80211_vif *vif)
{
	struct ieee80211_vif_priv *vp = ieee80211_to_priv(vif);
	struct scm80211_hwq *hwq;
	unsigned long queues = scm80211_vif_queues(vif);
	int i;

#ifdef SCM80211_FTM
	scm80211_ftm_deinit(hw, vif);
#endif //SCM80211_FTM

	if (vif->txq)
		scm80211_deinit_txq(hw, vif, vif->txq);

	scm80211_deinit_txq(hw, vif, (struct ieee80211_txq *)&vp->mq);

	for_each_set_bit(i, &queues, hw->queues) {
		hwq = scm80211_hwq(hw, i);
		hwq->enabled = false;
		hwq->need_resched = false;
		cancel_work_sync(&hwq->work);
		tasklet_kill(&hwq->tasklet);
		hrtimer_cancel(&hwq->watchdog);
		scm80211_hwq_free_dma_desc(hw, hwq);
	}

	cancel_delayed_work_sync(&vp->suspend_tx_work);
	scm80211_disable_tx_irq(hw, queues);

	scm80211_disable_trig_qnull_irq(hw, vif);
	scm80211_bus_vif_deinit_tx(hw, vif);
}

void scm80211_handle_txop_complete(struct ieee80211_hw *hw, struct txop *txop)
{
	struct scm80211_hwq *hwq = txop->hwq;

	scm80211_trace(hw, "%s(hwq#%d.%d)\n", __func__, hwq->id, txop->id);

	scm80211_complete_txop_prologue(txop, false);
	scm80211_process_successful_txop(txop, false);
	scm80211_txop_add(txop, TXOP_STATE_TX_DONE);

	tasklet_hi_schedule(&hwq->tasklet);
}

void scm80211_tx_start_irq(struct ieee80211_hw *hw)
{
#ifdef SCM2020_A0
#ifndef __TEMP_DEBUG__
	struct ieee80211_hw_priv *priv = hw->priv;
	struct scm80211_phy *phy;
	int pdet[2] = { 0, 0 };
	int i;
	unsigned long map = power_bitmap;

	if (!priv->pdet_en)
		return;

	WRITE_ONCE(power_bitmap, 0);

	if (bitmap_weight(&map, SCM80211_NUM_HWQ) != 1)
		return;

	i = find_first_bit(&map, SCM80211_NUM_HWQ);
	if (i >= SCM80211_NUM_HWQ || power_index[i] < 6)
		return;

	for_each_active_phy(phy, hw) {
		pdet[phy->chain] = phy->radio->ops->pdet_read(phy->radio, power_index[i]);
	}
	scm80211_info(hw, "power detect [%04d, %04d]\n", pdet[1], pdet[0]);
#else
	struct ieee80211_hw_priv *priv = hw->priv;

	ieee80211_queue_work(hw, &priv->power_detect_work);

#endif
#endif
}

#ifdef __staging__

/*
 * TX debugging - this will be moved to chip-specific file after
 * TX is stabilized.
 */

/*
 * TX debugging
 */
struct mac_phy_counter {
	const char *name;
	u32 addr;
	u32 field;
};

static void scm80211_tx_done_debug(struct scm80211_hwq *hwq, struct tx_mpdu_desc *mdesc,
				   struct sk_buff *skb);

struct scm80211_tx_debug;
static int scm80211_alloc_tx_debug(struct scm80211_hwq *hwq);
static void scm80211_read_tx_debug_registers(struct scm80211_hwq *hwq, int index);
static void scm80211_tx_done_debug(struct scm80211_hwq *hwq, struct tx_mpdu_desc *mdesc,
				   struct sk_buff *skb);
static void scm80211_phy_start_dump(struct ieee80211_hw *hw);
static void scm80211_phy_stop_dump(struct ieee80211_hw *hw);
static void scm80211_phy_print_dump(struct ieee80211_hw *hw);

static const struct mac_phy_counter phy_txbe[] = {
	{"ON",		PHY_COUNTER_TXBE_ON, -1U},
	{"PPDU",	PHY_COUNTER_PPDU_DATA_GEN, -1U},
	{"NSYM",	PHY_COUNTER_FD_SYMBOL_GEN, -1U},
	{"MIMO A",	PHY_COUNTER_TXBE_MIMO_MODE, -1U},
	{"MIMO A+B",	PHY_COUNTER_TXBE_MIMO_2SS, -1U},
	{"BCC",		PHY_COUNTER_TXBE_BCC, -1U},
	{"LDPC",	PHY_COUNTER_TXBE_LDPC, -1U},
	{"NONHT",	PHY_COUNTER_TXBE_NONHT, -1U},
	{"HTMF",	PHY_COUNTER_TXBE_HTMF, -1U},
	{"VHT",		PHY_COUNTER_TXBE_VHT, -1U},
	{"HE-SU",	PHY_COUNTER_TXBE_HESU, -1U},
	{"HE-ER",	PHY_COUNTER_TXBE_HEER, -1U},
};

static const struct mac_phy_counter phy_mon_tx[] = {
	{"OFDM ERROR",   PHY_MON_00, GENMASK(15, 0)},
	{"11B ERROR",    PHY_MON_00, GENMASK(31, 16)},
	{"TX UNDERFLOW", PHY_MON_01, GENMASK(15, 0)},
	{"TX OVERFLOW",  PHY_MON_01, GENMASK(31, 16)},
	{"TX DONE",      PHY_MON_33, -1U},
};

static const struct mac_phy_counter mac_tx[] = {
	{ "QUEUE0 READ", MAC_REG_DBG_QUE_READ01, GENMASK(15, 0) },
	{ "QUEUE0 SENT", MAC_REG_DBG_TX_PPDU_CNT01, GENMASK(15, 0) },
	{ "QUEUE1 READ", MAC_REG_DBG_QUE_READ01, GENMASK(31, 16) },
	{ "QUEUE1 SENT", MAC_REG_DBG_TX_PPDU_CNT01, GENMASK(31, 16) },
	{ "QUEUE2 READ", MAC_REG_DBG_QUE_READ23, GENMASK(15, 0) },
	{ "QUEUE2 SENT", MAC_REG_DBG_TX_PPDU_CNT23, GENMASK(15, 0) },
	{ "QUEUE3 READ", MAC_REG_DBG_QUE_READ23, GENMASK(31, 16) },
	{ "QUEUE3 SENT", MAC_REG_DBG_TX_PPDU_CNT23, GENMASK(31, 16) },
	{ "QUEUE4 READ", MAC_REG_DBG_QUE_READ45, GENMASK(15, 0) },
	{ "QUEUE4 SENT", MAC_REG_DBG_TX_PPDU_CNT45, GENMASK(15, 0) },
};

static int sign_extend_32(u32 v, u32 mask)
{
	u32 msb, lsb;
	int res;

	lsb = __bf_shf(mask);
	mask >>= lsb;
	mask += 1;
	if (mask == 0)
		msb = 31;
	else
		msb = __bf_shf(mask) - 1;
	res = v << (31 - msb);
	res >>= (31 - msb);
	return res;
}

static void phy_read_counters(struct ieee80211_hw *hw, u32 *counters,
			      const struct mac_phy_counter *array, int size)
{
	int i;
	u32 v;
	struct ieee80211_hw_priv *priv = hw->priv;

	for (i = 0; i < size; i++) {
		v = phy_readl(&priv->phy[priv->devid], array[i].addr);
		v = (v & array[i].field) >> __bf_shf(array[i].field);
		counters[i] = sign_extend_32(v, array[i].field);
	}
}

static void mac_read_counters(struct ieee80211_hw *hw, u32 *counters,
			      const struct mac_phy_counter *array, int size)
{
	int i;
	u32 v;

	for (i = 0; i < size; i++) {
		v = scm80211_mac_readl(hw, array[i].addr);
		v = (v & array[i].field) >> __bf_shf(array[i].field);
		counters[i] = v;
	}
}

struct scm80211_tx_debug {
	u32 phy_txbe_counters[ARRAY_SIZE(phy_txbe)];
	u32 phy_mon_tx_counters[ARRAY_SIZE(phy_mon_tx)];
	u32 mac_tx_ppdu_counters[ARRAY_SIZE(mac_tx)];
};

static void scm80211_read_tx_debug_registers(struct scm80211_hwq *hwq, int index)
{
	struct ieee80211_hw *hw = hwq->hw;
	struct scm80211_tx_debug *d;

	if (!hwq->debug)
		return;

	d = &hwq->debug[index];

	phy_read_counters(hw, d->phy_txbe_counters, phy_txbe, ARRAY_SIZE(phy_txbe));
	phy_read_counters(hw, d->phy_mon_tx_counters, phy_mon_tx, ARRAY_SIZE(phy_mon_tx));
	mac_read_counters(hw, d->mac_tx_ppdu_counters, mac_tx, ARRAY_SIZE(mac_tx));
}

static void scm80211_dump_tx_debug_counters(struct scm80211_hwq *hwq)
{
	struct ieee80211_hw *hw = hwq->hw;
	struct scm80211_tx_debug *d = hwq->debug;
	int i;

	if (!d)
		return;

	for (i = 0; i < ARRAY_SIZE(d->phy_txbe_counters); i++) {
		scm80211_dbg(hw, "%-20s %10d\n",
			     phy_txbe[i].name,
			     d[1].phy_txbe_counters[i] - d[0].phy_txbe_counters[i]);
	}

	for (i = 0; i < ARRAY_SIZE(d->phy_mon_tx_counters); i++) {
		scm80211_dbg(hw, "%-20s %10d\n",
			     phy_mon_tx[i].name,
			     d[1].phy_mon_tx_counters[i] - d[0].phy_mon_tx_counters[i]);
	}

	for (i = 0; i < ARRAY_SIZE(d->mac_tx_ppdu_counters); i++) {
		scm80211_dbg(hw, "%-20s %10d\n",
			     mac_tx[i].name,
			     d[1].mac_tx_ppdu_counters[i] - d[0].mac_tx_ppdu_counters[i]);
	}

	for (i = 0; i < 7; i++) {
		scm80211_dbg(hw, "tv[%d]: %08x\n", i,
			     scm80211_mac_readl(hw, MAC_REG_DBG_LAST_TXVEC0 + i * 4));
	}
}

static void scm80211_phy_start_dump(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	phy_writel(&priv->phy[priv->devid], 0x0, PHY_RFIF(PHY_RFI_MEMDP_RUN));
	phy_writel(&priv->phy[priv->devid], 0xefff, PHY_MODEM(PHY_DUMP_SIG_SEL));
	phy_writel(&priv->phy[priv->devid], 0x21, PHY_RFIF(PHY_RFI_MEMDP_CFG0));
	phy_writel(&priv->phy[priv->devid], 0x10, PHY_RFIF(PHY_RFI_MEMDP_CFG1));
	phy_writel(&priv->phy[priv->devid], 0x1, PHY_RFIF(PHY_RFI_MEMDP_RUN));
}

static void scm80211_phy_stop_dump(struct ieee80211_hw *hw)
{
	struct ieee80211_hw_priv *priv = hw->priv;

	phy_writel(&priv->phy[priv->devid], 0x0, PHY_RFIF(PHY_RFI_MEMDP_RUN));
}

static void scm80211_phy_print_dump(struct ieee80211_hw *hw)
{
	int i;
	u32 addr, ts, state;
	struct ieee80211_hw_priv *priv = hw->priv;

	for (i = 0; i < 0x1000; i += 4) {
		addr = PHY_RFIF(FW_TESTDUMP_MEM_0000 + i);
		ts = phy_readl(&priv->phy[priv->devid], addr);
		state = phy_readl(&priv->phy[priv->devid], addr + 0x1000);
		pr_debug("0x%08X 0x%08X\n", ts, state);
	}
}

static void phy_debug_dump_work(struct work_struct *work)
{
	struct ieee80211_hw_priv *priv = container_of(work, struct ieee80211_hw_priv, debug_work);

	scm80211_phy_print_dump(priv->hw);
}

/*
 * Only for MPDU or S-MPDU
 */
static void scm80211_tx_done_debug(struct scm80211_hwq *hwq, struct tx_mpdu_desc *mdesc,
				   struct sk_buff *skb)
{
	struct ieee80211_hw *hw = hwq->hw;
	struct ieee80211_hw_priv *priv = hw->priv;
	int i = mdesc - &hwq->mtab[0];
	void __iomem *fifo;
	size_t bytes;
	bool fail = false;

	scm80211_trace(hw, "hwq#%d: sent=%d acked=%d result=%08x\n",
		       hwq->id, mdesc->sent, mdesc->acked, hwq->txdesc->result);

	scm80211_phy_stop_dump(hw);

	scm80211_read_tx_debug_registers(hwq, 1);

	/*
	 * (1) sent, !ok, fail reason > 8
	 * (2) !sent, fail reason > 8
	 */
	if (mdesc->sent && !hwq->txdesc->ok)
		fail |= hwq->txdesc->fail > 8;	/* (1) */
	else if (!mdesc->sent)
		fail |= hwq->txdesc->fail == 11;	/* (2) */

	if (fail && 0) {
		struct txvector *tv = (void *)skb->data;
		u8 *buf;

		buf = kmalloc(1024, GFP_ATOMIC);
		if (!buf)
			goto counters;

		/* mpdu descriptor */
		fifo = priv->mac_base[0] + MAC_FIFO_OFFSET;
		bytes = sizeof(struct tx_mpdu_desc);
		memcpy_fromio(buf, fifo + i * bytes, bytes);
		if (memcmp(buf, mdesc, sizeof(*mdesc))) {
			scm80211_warn(hw, "mtab mismatch\n");
			print_hex_dump_bytes(" mtb(mem)", DUMP_PREFIX_OFFSET, mdesc, bytes);
			print_hex_dump_bytes("mtb(fifo)", DUMP_PREFIX_OFFSET, buf, bytes);
		}

		/* txvector */
		fifo += 1024 + 4;	/* 4 byte in front of txvector */
		bytes = min_t(int, hwq->dma_desc->mpdu[i].bdesc[0].len, 1024);
		memcpy_fromio(buf, fifo, bytes);

		if (memcmp(buf, tv, sizeof(*tv))) {
			scm80211_warn(hw, "txvector mismatch\n");
			print_hex_dump_bytes(" txv(mem)", DUMP_PREFIX_OFFSET, tv, sizeof(*tv));
			print_hex_dump_bytes("txv(fifo)", DUMP_PREFIX_OFFSET, buf, sizeof(*tv));
		}

		/* 802.11 frame */
		bytes -= sizeof(*tv);

		if (memcmp(buf + sizeof(*tv), skb->data + sizeof(*tv), bytes)) {
			scm80211_warn(hw, "tx frame data mismatch\n");
			print_hex_dump_bytes(" txd(mem)", DUMP_PREFIX_OFFSET,
					     skb->data + sizeof(*tv), bytes);
			print_hex_dump_bytes("txd(fifo)", DUMP_PREFIX_OFFSET,
					     buf + sizeof(*tv), bytes);
		}

		kfree(buf);
	}

counters:
	if (0) {
		scm80211_dump_tx_debug_counters(hwq);
		ieee80211_queue_work(hw, &priv->debug_work);
	}
}

static int scm80211_alloc_tx_debug(struct scm80211_hwq *hwq)
{
	struct device *dev = wiphy_dev(hwq->hw->wiphy);

	hwq->debug = devm_kzalloc(dev, sizeof(*hwq->debug) * 2, GFP_KERNEL);

	return hwq->debug ? 0 : -ENODEV;
}

/**
 * DOC: Heuristics to protect a txop with RTS/CTS exchange
 *
 * RTS/CTS
 * - Protect ERP-OFDM PPDU against non-ERP STAs
 * - Protect HT-GF or HT RIF against non-HT STAs
 * - Set NAV of hidden nodes
 * - Detect collision
 *
 */
static int scm80211_txop_enable_protection(struct txop *txop)
{
	if (txop->protection)
		return 1;

	return 0;
}

static struct sk_buff *ampdu_first_subframe(struct scm80211_ampdu_info *ampdu)
{
	struct txop *txop = container_of(ampdu, struct txop, ampdu);

	return skb_peek(&txop->sched);
}

static struct sk_buff *ampdu_last_subframe(struct scm80211_ampdu_info *ampdu)
{
	struct txop *txop = container_of(ampdu, struct txop, ampdu);

	return skb_peek_tail(&txop->sched);
}

static int scm80211_txh_qos_data(struct scm80211_hwq *hwq, struct sk_buff *skb)
{
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
	struct txop *txop = &hwq->txop;
	struct scm80211_tx_info *xti = txop_cur_xti(txop);
	size_t n_mpdu = skb_queue_len(&txop->sched);
	u32 txtime = txop->txtime;

	if (!ieee80211_is_data_qos(hdr) && skb_queue_len(&txop->sched) > 0)
		return -EAGAIN;

	return 0;
}

#endif /* __staging__ */
